%This is a quasar implementation of the paper
            "gSLIC: a real-time implementation of SLIC superpixel segmentation"
                "-Carl Yuheng Ren"

import "colortransform.q"
import "inttypes.q"

function [] = main()
    img = imread("lena_big.tif")
    [M,N] = size(img, 0..1)
    img += 15*randn(size(img))
    
    compactness = 1.0
    superpixel_size = 256
    
    labimg = rgb2lab(img)
    iterations = 20
    
    tic()
    for i = 1..iterations
        [labels,seeds] = perform_gSLIC_clustering(labimg, compactness, superpixel_size)
    end
    t = toc("")
    printf("[SLIC] Average time for static images (no connectivity): %f ms\n",t/iterations)
    
    tic()
    for i = 1..iterations
        clabels = enforce_connectivity(labels,superpixel_size)
    end
    t = toc("")
    printf("[SLIC] Average time for connectivity enforcing: %f ms\n",t/iterations)
    
    %usage example for video: simply one iteration every frame. Should be plenty!
    tic()
    for i = 1..iterations
        [vlabels,vseeds] = perform_gSLIC_clustering(labimg,compactness,superpixel_size,1,labels,seeds)
    end
    t = toc("")
    %this needs a footnote: after all, it's constant input without any overhead as far as that is concerned
    %additional footnote: you might get away with slightly higher FPS by using persistent buffers for new_seeds (check function internals)
    printf("[SLIC] Average fps for video: %f FPS\n",1000*iterations/t)
    
%    glass_img = lab2rgb(create_glass_image(labels, seeds))
%    imshow(glass_img)
    dispimg1 = copy(img)
    apply_gSLIC_borders(dispimg1,labels)
    imshow(dispimg1)
    dispimg2 = copy(img)
    apply_gSLIC_connected_borders(dispimg2,clabels)
    imshow(dispimg2)
end

function [] = apply_gSLIC_borders(img:cube, labels:cube[int16])
    function [] = __kernel__ draw_outlines(klabels:cube[int16]'clamped,pos:ivec2)
        if pos[0] == 0 || pos[0] == size(klabels,0)-1 || pos[1] == 0 || pos[1] == size(klabels,1)-1
            return
        endif
        equal = 1
        if klabels[pos[0],pos[1],0] != klabels[pos[0],pos[1]-1,0]
            equal = 0
        elseif klabels[pos[0],pos[1],1] != klabels[pos[0],pos[1]-1,1]
            equal = 0
        elseif klabels[pos[0],pos[1],0] != klabels[pos[0]-1,pos[1],0]
            equal = 0
        elseif klabels[pos[0],pos[1],1] != klabels[pos[0]-1,pos[1],1]
            equal = 0
        endif
        if equal
            %interior pixel
        else
            %edge pixel
            img[pos[0],pos[1],0] = 0.5
            img[pos[0],pos[1],1] = 0.1
            img[pos[0],pos[1],2] = 0.1
        endif
    end
    parallel_do(size(img,0..1), labels, draw_outlines)
end

function [] = apply_gSLIC_connected_borders(img:cube, labels:mat)
    function [] = __kernel__ draw_outlines(labels:mat'clamped,pos:ivec2)
        if pos[0] == 0 || pos[0] == size(labels,0)-1 || pos[1] == 0 || pos[1] == size(labels,1)-1
            return
        endif
        equal = 1
        if labels[pos[0],pos[1]] != labels[pos[0],pos[1]-1]
            equal = 0
        elseif labels[pos[0],pos[1]] != labels[pos[0]-1,pos[1]]
            equal = 0
        endif
        if equal
            %interior pixel
        else
            %edge pixel
            img[pos[0],pos[1],0] = 0.5
            img[pos[0],pos[1],1] = 0.1
            img[pos[0],pos[1],2] = 0.1
        endif
    end
    parallel_do(size(img,0..1), labels, draw_outlines)
end

function [clabels:mat,labels:cube[int16]] = gSLIC(image:cube, compactness:scalar,superpixel_size:int, iterations:int=10)
    %do the clustering without connected components
    [labels,seeds] = perform_gSLIC_clustering(image, compactness, superpixel_size, iterations)
    
    %do connected components on the labels
    clabels = enforce_connectivity(labels,superpixel_size)
end

function [clabels:mat] = gSLICO(image:cube, superpixel_size:int, iterations:int=10)
    %do the clustering without connected components
    [labels,_] = perform_gSLICO_clustering(image, superpixel_size, iterations)
    
    %do connected components on the labels
    clabels = enforce_connectivity(labels,superpixel_size)
end

function [nlabels:mat] = enforce_connectivity(labels:cube[int16],superpixel_size:int)
    [M,N] = size(labels, 0..1)
    dx4 = [-1,0,1,0]
    dy4 = [0,-1,0,1]
    
    nlabels = -ones(M,N)
    xvec:vec[int] = vec[int](superpixel_size*4)
    yvec:vec[int] = vec[int](superpixel_size*4)
    
    function [] = __device__ connectivity_check()
    label = 0
    adjlabel = 0
    for j = 0..M-1
    for k = 0..N-1
        if nlabels[j,k] < 0
            nlabels[j,k] = label
            
            %start a new segment
            xvec[0] = k
            yvec[0] = j
            
            %quickly find an adjacent label for later use if required
            for n = 0..3
                x = k+dx4[n]
                y = j+dy4[n]
                if x>=0 && x < N && y>=0 && y<M
                    if nlabels[y,x] >= 0
                        adjlabel = nlabels[y,x]
                    endif
                endif
            end
            
            count=1
            c=0
            repeat
                for n = 0..3
                    x = xvec[c] + dx4[n]
                    y = yvec[c] + dy4[n]
                    if x >= 0 && x < N && y >= 0 && y < M
                        if 0 > nlabels[y,x] && labels[j,k,0] == labels[y,x,0] && labels[j,k,1] == labels[y,x,1]
                            xvec[count] = x
                            yvec[count] = y
                            nlabels[y,x] = label
                            count = count + 1
                        endif
                    endif
                end
                c = c + 1
            until c>=count
            
            %if segment size is less than a limit, assign an
            %adjacent label found before, and decrement label count
            if count <= superpixel_size/6
                for c = 0..count-1
                    nlabels[yvec[c],xvec[c]] = adjlabel
                end
                label -= 1
            endif
            label += 1
        endif
    end
    end
    end
    connectivity_check()
end

%this function can be called on its own to result in the clustering without enforced connectivity
function [labels:cube[int16], seeds:cube] = perform_gSLIC_clustering(image:cube,compactness:scalar, super_pixel_size:scalar, iterations:int=10,init_labels:cube[int16]=cube[int16](0,0,0),init_seeds:cube=zeros(0,0,0))
    [M,N] = size(image,0..1)
    
    %initialize cluster centers on a (rectangular) grid at regular grid steps S
    %pixels are assigned to the cluster center they are closest to
    if numel(init_labels) == 0
        [labels,seeds] = generate_seeds(image, super_pixel_size)
    else
        labels = copy(init_labels)
        seeds = copy(init_seeds)
    endif
    [Mk,Nk] = size(seeds,0..1)
    
    new_seeds = zeros(Mk,Nk,6)
    
    function [] = __kernel__ find_new_best_friend(seeds:cube,pos:ivec2)
        this_point = [image[pos[0],pos[1],0],image[pos[0],pos[1],1],image[pos[0],pos[1],2],pos[0]*compactness,pos[1]*compactness]
        
        yk = labels[pos[0],pos[1],0]
        xk = labels[pos[0],pos[1],1]
        best_distance = 1e9
        best_yko = yk
        best_xko = xk
        for yko = max(yk-1,0)..min(yk+1,Mk-1)
        for xko = max(xk-1,0)..min(xk+1,Nk-1)
            %what is the error to this cluster center?
            distance = sum((this_point-seeds[yko,xko,0..4]).^2)
            
            if distance < best_distance
                best_distance = distance
                best_yko = yko
                best_xko = xko
            endif
        end
        end
        
        labels[pos[0],pos[1],0] = best_yko
        labels[pos[0],pos[1],1] = best_xko
        
        %and already let our new friend know we support him
        for c = 0..4
            new_seeds[best_yko,best_xko,c] += this_point[c]
        end
        new_seeds[best_yko,best_xko,5] += 1
    end
    function [] = __kernel__ update_cluster_descriptions(pos:ivec2)
        if new_seeds[pos[0],pos[1],5] > 0
            seeds[pos[0],pos[1],0..4] = new_seeds[pos[0],pos[1],0..4] / new_seeds[pos[0],pos[1],5]
            seeds[pos[0],pos[1],5] = new_seeds[pos[0],pos[1],5]
        endif
        
        %zero the buffer for the next run
        new_seeds[pos[0],pos[1],0..5] = 0
    end
    
    for iteration = 1..iterations
        parallel_do([M,N],seeds,find_new_best_friend)
        parallel_do([Mk,Nk], update_cluster_descriptions)
    end
end

%this function can be called on its own to result in the clustering without enforced connectivity
function [labels:cube[int16], seeds:cube] = perform_gSLICO_clustering(image:cube,super_pixel_size:scalar, iterations:int=10,init_labels:cube[int16]=cube[int16](0,0,0),init_seeds:cube=zeros(0,0,0))
    [M,N] = size(image,0..1)
    
    %initialize cluster centers on a (rectangular) grid at regular grid steps
    if numel(init_labels) == 0
        [labels,seeds] = generate_seeds(image, super_pixel_size,1)%one extra fields for maxlab values
    else
        labels = copy(init_labels)
        seeds = copy(init_seeds)
    endif
    [Mk,Nk] = size(seeds,0..1)
    
    invxywt = 1/super_pixel_size
    inv_initial_compactness = 1/10
    new_seeds = uninit(Mk,Nk,7)
    function [] = __kernel__ initialise_new_seeds(pos:ivec2)
        for c = 0..5
            new_seeds[pos[0],pos[1],c] = 0.0
        end
        new_seeds[pos[0],pos[1],6] = inv_initial_compactness
            seeds[pos[0],pos[1],6] = inv_initial_compactness
    end
    parallel_do([Mk,Nk],initialise_new_seeds)
    
    
    function [] = __kernel__ find_new_best_friend(seeds:cube'hwtex_const,pos:ivec2)
        this_point = [image[pos[0],pos[1],0],image[pos[0],pos[1],1],image[pos[0],pos[1],2],pos[0],pos[1]]
    
        yk = labels[pos[0],pos[1],0]
        xk = labels[pos[0],pos[1],1]
        best_distance = 1e9
        best_colour_distance = 0.0
        best_yko = yk
        best_xko = xk
        for yko = max(yk-1,0)..min(yk+1,Mk-1)
        for xko = max(xk-1,0)..min(xk+1,Nk-1)
            %what is the error to this cluster center?
            colour_distance = sum((this_point[0..2] - seeds[yko,xko,0..2]).^2)
            this_distance = colour_distance/seeds[yko,xko,6] + sum((this_point[3..4]-seeds[yko,xko,3..4]).^2)*invxywt
            
            if this_distance < best_distance
                best_distance = this_distance
                best_colour_distance = colour_distance
                best_yko = yko
                best_xko = xko
            endif
        end
        end
        
        labels[pos[0],pos[1],0] = best_yko
        labels[pos[0],pos[1],1] = best_xko
        
        %and already let our new friend know we support him
        for c = 0..4
            new_seeds[best_yko,best_xko,c] += this_point[c]
        end
        new_seeds[best_yko,best_xko,5] += 1.0
        new_seeds[best_yko,best_xko,6] ^^= best_colour_distance
    end
    
    function [] = __kernel__ update_cluster_descriptions(pos:ivec2)
        if new_seeds[pos[0],pos[1],5] > 0
            seeds[pos[0],pos[1],0..4] = new_seeds[pos[0],pos[1],0..4] / new_seeds[pos[0],pos[1],5]
            seeds[pos[0],pos[1],5..6] = new_seeds[pos[0],pos[1],5..6]
        else
            seeds[pos[0],pos[1],5] = 0
        endif
        
        %zero the buffer for the next run
        for i = 0..5
            new_seeds[pos[0],pos[1],i] = 0
        end
        new_seeds[pos[0],pos[1],6] = inv_initial_compactness
    end
    
    for iteration = 1..iterations
        parallel_do([M,N], seeds, find_new_best_friend)
        parallel_do([Mk,Nk], update_cluster_descriptions)
    end
end

function [glass:cube] = create_glass_image(labels:cube, seeds:cube)
    [M,N] = size(labels, 0..1)
    [Mk,Nk,C] = size(seeds, 0..2)
    
    glass = uninit(M,N,3)
    
    function [] = __kernel__ glassblowing(pos:ivec2)
        [yk,xk] = labels[pos[0],pos[1],0..1]
        for c = 0..2
            glass[pos[0],pos[1],c] = seeds[yk,xk,c]
        end
    end
    parallel_do([M,N], glassblowing)
end

function [labels:cube[int16],seeds:cube] = generate_seeds(img:cube, super_pixel_size:scalar, extra_fields:int=0)
    step = sqrt(super_pixel_size)
    [M,N] = size(img,0..1)
    [Mk,Nk] = ceil([M,N]/step)
    
    
    labels = cube[int16](M,N,2)
    function [] = __kernel__ filllabels(pos:ivec2)
        labels[pos[0],pos[1],0] = int(pos[0]/step)
        labels[pos[0],pos[1],1] = int(pos[1]/step)
    end
    parallel_do([M,N],filllabels)
    
    seeds = uninit(Mk,Nk,6+extra_fields) %L,a,b,y,x,count
    function [] = __kernel__ fill_seed_colour(ratM:scalar,ratN:scalar,pos:ivec2)
        y = int((0.5+pos[0])*ratM)
        x = int((0.5+pos[1])*ratN)
        seeds[pos[0],pos[1],3] = float(y)
        seeds[pos[0],pos[1],4] = float(x)
        seeds[pos[0],pos[1],0] = img[y,x,0]
        seeds[pos[0],pos[1],1] = img[y,x,1]
        seeds[pos[0],pos[1],2] = img[y,x,2]
        seeds[pos[0],pos[1],5] = ratM*ratN
    end
    parallel_do([Mk,Nk],M/Mk,N/Nk,fill_seed_colour)
end
