%
% bayer_mosaic.q
% Quasar library for Bayer demosaicing
%
{!author name="Bart Goossens"}
{!doc category="Image Processing/Demosaicing"}

% Function: bayer_mosaicimage
%
% Create a Bayer mosaic image based on a specific Bayer pattern
% The pattern matrix is a 2x2 matrix, in which 0=red, 1=green1, 2=blue, 3=green2
%
% :function [im_mosaic : mat] = bayer_mosaicimage(im_bayer : cube(:,:,4), pattern : mat(2,2) = [[0,1],[3,2]])
%
% Parameters:
% im_bayer - an RGB color image
% pattern - the Bayer pattern to be used (default: [[0,1],[3,2]])
%
% Remarks:
% For the Bayer pattern, the following values are used:
% 0: Red channel (R)
% 1: Green channel #1 (G1)
% 2: Blue channel (B)
% 3: Green channel #2 (G2)
%
% See also:
% <bayer_upsampling>
function [im_mosaic : mat] = bayer_mosaicimage(im_bayer : cube(:,:,4), pattern : mat(2,2) = [[0,1],[3,2]])
    function [] = __kernel__ bayer_mosaicimage_kernel(x : cube'unchecked, y : mat'unchecked, _
        pattern : mat'unchecked'const(2,2), pos : ivec2)
        m = and(pos[0],1)
        n = and(pos[1],1)
        ch = int(pattern[m,n])
        y[pos] = x[int(pos[0]/2),int(pos[1]/2),ch]
    endfunction
    im_mosaic = zeros(size(im_bayer,0)*2,size(im_bayer,1)*2)
    parallel_do(size(im_mosaic,0..1),im_bayer,im_mosaic,pattern,bayer_mosaicimage_kernel)
endfunction

% Function: bayer_upsampling
%
% Create an RGB mosaic image, with unknown sensor values set to zero
% This is effectively the adjoint operation of bayer_subsampling
%
% :function [im_rgb : cube] = bayer_upsampling(im_bayer : cube(:,:,4), pattern : mat(2,2) = [[0,1],[3,2]])
%
% Parameters:
% im_bayer - a Bayer mosaic images (as obtained using the function <bayer_mosaicimage>
% pattern - the Bayer pattern to be used (default: [[0,1],[3,2]])
%
% See also:
% <bayer_mosaicimage>
function [im_rgb : cube] = bayer_upsampling(im_bayer : cube(:,:,4), pattern : mat(2,2) = [[0,1],[3,2]])
    function [] = __kernel__ kernel(x : cube'unchecked, y : cube'unchecked, _
        pattern : mat'unchecked'const(2,2), pos : ivec2)
        m = and(pos[0],1)
        n = and(pos[1],1)
        ch_in = int(pattern[m,n])
        ch_out = (ch_in == 3) ? 1 : ch_in % green channel
        y[pos[0], pos[1], ch_out] = x[int(pos[0]/2),int(pos[1]/2),ch_in]
    endfunction
    im_rgb = zeros(size(im_bayer,0)*2,size(im_bayer,1)*2,3)
    parallel_do(size(im_rgb,0..1),im_bayer,im_rgb,pattern,kernel)
endfunction

% Function: bayer_subsampling
%
% Bayer subsampling to 4-channel RG1BG2 image, using a specific Bayer pattern
%
% :function [im_bayer : cube] = bayer_subsampling(im_rgb : cube(:,:,3), pattern : mat(2,2) = [[0,1],[3,2]])
%
% In contrast to <bayer_mosaicimage>, an image with four color channels is formed.
% The color channels are R, G1, B, G2. During this operation, the image is subsampled by
% a factor two.
%
% Parameters:
% im_rgb - an RGB input image
% pattern - the Bayer pattern to be used (default: [[0,1],[3,2]])
%
% See also:
% <bayer_subsampling_rgb>
function [im_bayer] = bayer_subsampling(im_rgb : cube(:,:,3), pattern : mat(2,2) = [[0,1],[3,2]])
    function [] = __kernel__ bayer_subsampling_kernel(x : cube'unchecked, y : cube'unchecked, _
        pattern : mat'unchecked'const, pos : ivec2)

        m = and(pos[0],1)
        n = and(pos[1],1)

        ch_out = int(pattern[m,n])
        ch_in = (ch_out == 3) ? 1 : ch_out % green channel
        y[int(pos[0]/2),int(pos[1]/2),ch_out] = x[pos[0],pos[1],ch_in]
    endfunction

    im_bayer = zeros(size(im_rgb,0)/2,size(im_rgb,1)/2,4)
    parallel_do(size(im_rgb,0..1),im_rgb,im_bayer,pattern,bayer_subsampling_kernel)
endfunction

% Function: bayer_subsampling_rgb
%
% Bayer subsampling to 4-channel image, using a specific Bayer pattern
%
% :function [im_bayer : cube] = bayer_subsampling_rgb(im_rgb : cube(:,:,3), pattern : mat(2,2) = [[0,1],[3,2]])
%
% In contrast to <bayer_mosaicimage>, an image with four color channels is formed.
% The color channels are R, G1, B, G2. Compared to <bayer_subsampling>, no subsampling is
% being performed. Instead, the samples with missing data are set to zero.
%
% See also:
% <bayer_subsampling>
function [im_bayer : cube] = bayer_subsampling_rgb(im_rgb : cube(:,:,3), pattern : mat(2,2) = [[0,1],[3,2]])
    function [] = __kernel__ kernel(x : cube'unchecked, y : cube'unchecked, _
        pattern : mat'unchecked'const, pos : ivec2)

        m = and(pos[0],1)
        n = and(pos[1],1)

        ch_out = int(pattern[m,n])
        ch_in = (ch_out == 3) ? 1 : ch_out % green channel
        y[pos[0],pos[1],0..2] = 0.0
        y[pos[0],pos[1],ch_in] = x[pos[0],pos[1],ch_in]
    endfunction

    im_bayer = zeros(size(im_rgb))
    parallel_do(size(im_rgb,0..1),im_rgb,im_bayer,pattern,kernel)
endfunction

% Function: demosaic_bilinear
%
% Bilinear demosaicing for Bayer CFAs
%
% :function im_out : cube(:,:,3) = demosaic_bilinear(im : cube(:,:,4), bayer_pattern : mat(2,2))
%
% Parameters:
% im - the Bayer RG1BG2 image (e.g. obtained using <bayer_subsampling>) to
%      be interpolated
% pattern - the Bayer pattern to be used (default: [[0,1],[3,2]])
%
function im_out : cube(:,:,3) = demosaic_bilinear(im : cube(:,:,4), bayer_pattern : mat(2,2))
    function [] = __kernel__ mosaic(im_out : cube'unchecked, im_in : cube'unchecked, i_pattern : ivec4, o_pattern : ivec4, pos : ivec2)
        x = im_in[pos[0],pos[1],0..3]
        im_out[2*pos[0],2*pos[1],o_pattern[0]]     = x[i_pattern[0]]
        im_out[2*pos[0],2*pos[1]+1,o_pattern[1]]   = x[i_pattern[1]]
        im_out[2*pos[0]+1,2*pos[1],o_pattern[2]]   = x[i_pattern[2]]
        im_out[2*pos[0]+1,2*pos[1]+1,o_pattern[3]] = x[i_pattern[3]]
    endfunction
    function [] = __kernel__ demos_bilinear_step1(im_out : cube, o_pattern : ivec4, pos : ivec3)
        m = and(pos[0],1)
        n = and(pos[1],1)
        v = o_pattern[and(m+1,1)*2+n] == pos[2]
        h = o_pattern[m*2+and(n+1,1)] == pos[2]
        if o_pattern[m*2+n] != pos[2]
            nb_up = im_out[pos-[1,0,0]]
            nb_down = im_out[pos+[1,0,0]]
            nb_left = im_out[pos-[0,1,0]]
            nb_right = im_out[pos+[0,1,0]]
            if h && v
                im_out[pos] = (nb_up + nb_down + nb_left + nb_right) * 0.25
            elseif v
                im_out[pos] = (nb_up + nb_down) * 0.5
            elseif h
                im_out[pos] = (nb_left + nb_right) * 0.5
            endif
        endif
    endfunction
    function [] = __kernel__ demos_bilinear_step2(im_out : cube, o_pattern : ivec4, pos : ivec3)
        m = and(pos[0],1)
        n = and(pos[1],1)
        v = o_pattern[and(m+1,1)*2+n] == pos[2]
        h = o_pattern[m*2+and(n+1,1)] == pos[2]
        if o_pattern[m*2+n] != pos[2]
            if !h && !v
                nb_up = im_out[pos-[1,0,0]]
                nb_down = im_out[pos+[1,0,0]]
                nb_left = im_out[pos-[0,1,0]]
                nb_right = im_out[pos+[0,1,0]]
                im_out[pos] = (nb_up + nb_down + nb_left + nb_right) * 0.25
            endif
        endif
    endfunction
    [rows,cols,C] = size(im,0..2)
    im_out = uninit(2*rows,2*cols,3)
    %i_pattern = reshape(bayer_pattern,4) %TODO:
    i_pattern = [bayer_pattern[0,0],bayer_pattern[0,1],bayer_pattern[1,0],bayer_pattern[1,1]]
    o_pattern = copy(i_pattern)
    o_pattern = o_pattern - 2 * (o_pattern == 3)
    parallel_do([rows,cols],im_out,im,i_pattern,o_pattern,mosaic)
    parallel_do(size(im_out),im_out,o_pattern,demos_bilinear_step1)
    parallel_do(size(im_out),im_out,o_pattern,demos_bilinear_step2)
endfunction

function [] = main()
    pattern = [[0,1],[3,2]]
    im = imread("lighthouse.png")
    im_bayer = bayer_subsampling(im,pattern)
    im_mosaic = bayer_mosaicimage(im_bayer,pattern)
    print size(im_mosaic)
    %imshow(bayer_subsampling_rgb(im, pattern))
    imshow(demosaic_bilinear(im_bayer, pattern))
endfunction
