% Bregman utilities
import "system.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Restoration"}


% Section: bregman_utils.q
% Contains some general routines that are often used in Bregman/Primal Dual
% iterative algorithms. These algorithms solve non-linear inverse problems
% that are encountered within the field of image reconstruction and restoration.

% Function: kernelshift
%
% Hiep's kernelshift function: pads a given filter kernel (mask) with zeros
% so that it has the dimensions rows x cols. This function is useful for 
% implementing a convolution in the frequency domain
%
% : function result = kernelshift(kernel, rows, cols)
%
% Parameters:
% kernel - the filter kernel (mask) to extend
% rows - the new number of rows
% cols - the new number of cols
function result = kernelshift(kernel : mat, rows : int, cols : int)

    result = zeros(rows, cols)
    height = size(kernel,0)
    width = size(kernel,1)
    halfheight = floor(height/2)
    halfwidth = floor(width/2)

    print numel(0..halfheight),",",numel(halfheight..height-1)
    result[0..halfheight,0..halfwidth]=kernel[halfheight..height-1,halfwidth..width-1]
    result[rows-halfheight..rows-1,0..halfwidth]=kernel[0..halfheight-1,halfwidth..width-1]
    result[0..halfheight,cols-halfwidth..cols-1]=kernel[halfheight..height-1,0..halfwidth-1]
    result[rows-halfheight..rows-1,cols-halfwidth..cols-1]=kernel[0..halfheight-1,0..halfwidth-1]
end

% Hiep's kernel shift function (here for even kernel sizes)
%function y = kernelshift(x, dims)
%    function [] = __kernel__ kernelshift_kernel(x : mat, y : mat'unchecked, pos : ivec2)             
%        src_pos = pos +  size(x)/2
%        if src_pos[0]>=size(y,0)
%            src_pos[0] -= size(y,0)
%        endif
%        if src_pos[1]>=size(y,1)
%            src_pos[1] -= size(y,1)
%        endif
%        y[pos] = x[src_pos]
%    end
%    y = zeros(dims)
%    parallel_do(size(y),x,y,kernelshift_kernel)
%end

% Function: displacement_diff
% Computes a displacement difference. Useful for implementing discrete derivatives, such 
% as used in Total Variation (TV) approaches. The function uses circular boundary 
% extension.
%
% : function y = displacement_diff(x, d)
%
% Parameters:
% x - input matrix/cube to add displacement
% d - displacement vector
%
% Returns:
% y - The displaced difference, according to the relationship
% : y[p] = x[p] - x[p+d]
%
% See also:
%    <circshift>
%
function y = displacement_diff(x, d)
    % RGB version
    function [] = __kernel__ process_RGB(x : cube'circular, y : cube'unchecked, d : ivec2, pos : ivec2)
        y[pos[0],pos[1],0..2] = x[pos[0],pos[1],0..2] - x[pos[0]+d[0],pos[1]+d[1],0..2]
    end

    % 2D version
    function [] = __kernel__ process(x : mat'circular, y : mat'unchecked, d : ivec2, pos : ivec2)
        y[pos]= x[pos] - x[pos + d]
    end

    % Generic 3D version
    function [] = __kernel__ process_3D(x : cube'circular, y : cube'unchecked, d : ivec2, pos : ivec3)
        y[pos] = x[pos] - x[pos + [d[0],d[1],0]]
    end

    y = zeros(size(x))

    if size(x,2)==1
        parallel_do(size(x,0..1),x,y,d,process)
    elseif size(x,2)==3
        parallel_do(size(x,0..1),x,y,d,process_RGB)
    else
        parallel_do(size(x),x,y,d,process_3D)
    endif
end

%
% Function: shrink
% Soft-shrinkage of a subband (or cell array)
%
% : function y = shrink(x, T)
%
% The function performs soft-shrinkage, according to the formula
% y = sign(x)*max(abs(x)-T,0)
%
% Parameters:
% x - input data. Can be "cell", "scalar" or "cube"
% T - threshold
%
% Notes:
% Only real-valued data is accepted.
%
function y = shrink(x, T)
    function [] = __kernel__ process_RGB(x : cube, y : cube, T : scalar, pos : ivec2)
       xi = x[pos[0],pos[1],0..2]
       y[pos[0],pos[1],0..2] = sign(xi).*max(abs(xi)-T,[0.0,0.0,0.0])
    end

    function [] = __kernel__ process(x : mat'unchecked, y : mat'unchecked, T : scalar, pos : ivec2)
        xi = x[pos]
        y[pos] = sign(xi).*max(abs(xi)-T,0.0)
    end

    function [] = __kernel__ process_3D(x : cube'unchecked, y : cube'unchecked, T : scalar, pos : ivec3)
        xi = x[pos]
        y[pos] = sign(xi).*max(abs(xi)-T,0.0)
    end
    
    % x is a cell array
    if iscell(x)
        y = cell(size(x))
        % recursively apply the shrinkage to every element of the cell array
        for i=0..numel(y)-1
            y[...ind2pos(size(y), i)] = shrink(x[...ind2pos(size(y),i)], T)
        end
        a=1
    elseif isscalar(x)
        y = sign(x)*max(abs(x)-T,0.0)
    else
        y = zeros(size(x))
        if size(x, 2)==1
            parallel_do(size(x,0..1),x,y,T,process)
        elseif size(x, 2)==3
            parallel_do(size(x,0..1),x,y,T,process_RGB)
        else
            parallel_do(size(x),x,y,T,process_3D)
        endif
    endif
end

% Function: Dx
% Horizontal derivative operation
%
% : function y = Dx(x)
%
% See also:
%   <Dxt>
Dx = x -> displacement_diff(x,[0,1])

% Function: Dy
% Vertical derivative operation
%
% : function y = Dy(x)
%
% See also:
%   <Dyt>
Dy = x -> displacement_diff(x,[1,0])

% Function: Dxt
% Horizontal derivative operation (adjoint operation)
%
% : function y = Dxt(x)
%
% See also:
%   <Dx>
Dxt = x -> displacement_diff(x,[0,-1])

% Function: Dyt
% Vertical derivative operation (adjoint operation)
%
% : function y = Dyt(x)
%
% See also:
%   <Dy>
Dyt = x -> displacement_diff(x,[-1,0])


% Function: build_tvtransform
% 
% Constructs the total variation transform
% 
% Usage:
%   : function [S : [??->vec[??]], S_H : [vec[??]->??]] = build_tvtransform()
% 
% Returns:
% S - forward TV operator
% S_H - backward TV operator
function [S, S_H] = build_tvtransform()
    S = x -> {Dx(x), Dy(x)}
    S_H = x -> Dxt(x[0]) + Dyt(x[1])
endfunction

% Function: psnr
% Computation of the Peak-Signal-To-Noise (PSNR) ratio
%
% : psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))

