% colored_noise.q
% modeling and generation of colored (non-white) Gaussian noise
% Bart Goossens
% June 8, 2012
import "system.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Restoration"}

% Function: soft_step
%
% A soft step function at x=1 (based on the cosine function)
%
% Parameters:
% x - the matrix to apply the soft stepping to.
% t - transition width (t >= 0)
function y = soft_step(x, t)   
    function [] = __kernel__ process (x : cube'unchecked, y : cube'unchecked, t : scalar, pos : ivec3) 
        u = x[pos]
        if u < 1.0 - t
            y[pos] = 1.0
        elseif u < 1.0 + t
            y[pos] = cos((u - 1.0 + t) / (4.0 * t) * pi)
        else
            y[pos] = 0.0
        endif
    end

    y = zeros(size(x))
    parallel_do(size(x),x,y,t,process)
end

% Function: psd_noise1f
%
% Computes a 2D power spectral density for 1/f noise (vertical line noise)
psd_noise1f = (x, y) -> 1.0 ./ (1e-2 + abs(x))

% Function: psd_noise1f2
%
% Computes a 2D power spectral density for radial 1/f noise
psd_noise1f2 = (x, y) -> 1.0 ./ (1e-2 + abs(complex(x, y)))

% Function: psd_ct_ramp_spectrum_noise
%
% Computes a 2D power spectral density for noise in CT images caused by the
% filtered back projection algorithm (with ramp filtering)
psd_ct_ramp_spectrum_noise = (x, y) -> abs(complex(x, y))

% Function: psd_ct_hann_filter_noise
%
% Computes a 2D power spectral density for noise in CT images caused by the
% filtered back projection algorithm (with Hann smoothing filter)
psd_ct_hann_filter_noise = (x, y) -> (r = abs(complex(x, y)); r .* (r <= 1) .* (1 + cos(pi * r)))

% Function: psd_mri_elliptical_filter_noise
%
% Computes a 2D power spectral density for noise in MRI images with the
% elliptical filter
psd_mri_elliptical_filter_noise = (x, y) -> soft_step(abs(complex(x,y)), 0.02)

% Function: psd_partial_kspace_noise
%
% Computes a 2D power spectral density for noise in MRI images caused by
% partial kspace acquisition
psd_partial_kspace_noise = (x, y) -> soft_step(2*abs(x), 0.02)

% Function: psd_ef_partial_kspace_noise
%
% Computes a 2D power spectral density for noise in MRI images caused by
% combined effects of the elliptical filter and partial kspace acquisition
psd_ef_partial_kspace_noise = (x, y) -> psd_mri_elliptical_filter_noise(x, y) .* psd_partial_kspace_noise(x, y)

% Function: psd_paltv_noise
%
% Computes a simple (example) model for noise in PAL TV
psd_paltv_noise = (x, y) -> 0.0001 + exp(-100 * abs(complex(x - 0.2, y - 0.15))) + exp(-100 * abs(complex(x + 0.2, y + 0.15)))

% Function: psd_glb_noise
%
% Computes the power spectral density for Gaussian Lumpy backgrounds
% (mainly intended for medical image quality assessment, used in 
% channelized Hotelling observer models)
psd_glb_noise = (x, y) -> exp(-250 * (x.^2 + y.^2))
psd_sr_noise = (x, y) -> (x += 0.25 * y; gamma = 50; x1 = 0.4; y1 = 0.5; _
    0.05 + exp(-gamma * abs(complex(x - x1, y - y1))) + _
           exp(-gamma * abs(complex(x + x1, y - y1))) + _
           exp(-gamma * abs(complex(x - x1, y + y1))) + _
           exp(-gamma * abs(complex(x + x1, y + y1))) + _
           exp(-gamma * abs(complex(x - x1, y))) + _
           exp(-gamma * abs(complex(x + x1, y))) + _
           exp(-gamma * abs(complex(x, y + y1))) + _
           exp(-gamma * abs(complex(x, y - y1))))

% Function: generate_psd
%
% Generates the power spectral density matrix, corresponding to the
% given point-wise function)
%
% Parameters:
% sz - the dimensions of the output matrix
% func - the function to apply to every grid point
% scale - a scaling factor for the frequency spectrum
% visualization - the output is intended to be used for visualization
%    (In this case, the DC coefficient is in the center of the image)
function P = generate_psd(sz, func, scale = 1, visualisation = 0)
   
    [x, y] = meshgrid((0..sz[1]-1)/(sz[1]/2)*scale-scale, _
                      (0..sz[0]-1)/(sz[0]/2)*scale-scale)

    P = func(x,y)

    if visualisation
        P /= max(P)  % normalize by dividing by the maximum
    else
        P /= mean(P)
        P = fftshift2(P) % PSD is directly ready for use with the FFT
    endif
    P = sqrt(P)
end

% Function: cgn_rand
%
% Generate a colored gaussian random noise, with the given power spectral
% density.
%
% Parameters:
% sz - the size of the output matrix
% P - the power spectral density, as generated by <generate_psd> with
%      visualization parameter set to 0.
function y : cube = cgn_rand(sz : ivec3, P)
    if numel(sz) == 2
        sz = [sz[0],sz[1],1]
    endif
    extend_dim = x -> repmat(x,[1,1,sz[2]])

    spec = exp(1j*2*pi*rand(sz)) .* extend_dim(P) * sqrt(sz[0]*sz[1])
    y = irealfft2(spec)
end

function [] = main()
    P = generate_psd([256,384], psd_ct_hann_filter_noise, 2, 0)
    %imshow(P,[0,1])
    y = cgn_rand([256,384,3], P)
    imshow(y,[])
    print std(y)
    %imshow(fftshift2(abs(fft2(y))),[])
end