% MATLAB Like color maps
% Note - the function overrides the default implementation of imshow, for
% mat arguments, using a reduction.
import "system.q"
import "Quasar.UI.dll"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Colors"}

% Section: colormap.q
% Import colormap.q if you wish to use MATLAB style color maps for visualization
% using the imshow function.

% Initialization
current_color_map = object()

% Function: colormap
% Changes the current color map to the specified name. 
%
% : function y = colormap(name : string)
%
% Parameters:
% name - the name of the colormap. Accepted color maps are "jet", "hsv",
%          "hot", "hotter", "cool", "rgb", "cmy", "gray".
%        The default color map is "yet".
%
% Returns:
% y - a matrix of size Nx3 containing RGB-values of the colors of the
%     map. Note that the number of components N can vary.
%
function y : mat = colormap(name : string)
    function y = hsv2rgb (c : scalar)
        h = floor(c / 60.0)
        f = frac(c / 60.0)
        v = 255.0
        s = 1.0
        p = v * (1 - s)
        q = v * (1 - f * s)
        t = v * (1 - (1 - f) * s)
        match h with
        | 0 -> y = [v, t, p]
        | 1 -> y = [q, v, p]
        | 2 -> y = [p, v, t]
        | 3 -> y = [p, q, v]
        | 4 -> y = [t, p, v]
        | _ -> y = [v, p, q]
        end
    end

    match name with
    | "jet" ->  y = [[0, 0, 255], [0, 255, 0], [255, 255, 0], [255, 0, 0]]
    | "hsv" ->  y = zeros(360,3)
                for k=0..359
                    y[k,:] = hsv2rgb (k) 
                end
    | "hot" ->  y = [[0, 0, 0], [255, 0, 0], [255, 255, 0], [255, 255, 255]]
    | "hotter" ->  y = [[0, 0, 0], [255, 0, 0], [255, 255, 0], [255, 255, 255], [255, 255, 0], [255, 0, 0], [0, 0, 0]]
    | "cool" -> y = [[0, 255, 255], [127.5, 127.5, 255], [255, 0, 0]]
    | "rgb" ->  y = [[255, 0, 0], [0, 255, 0], [0, 0, 255]]
    | "cmy" ->  y = [[0, 255, 255], [255, 0, 255], [255, 255, 0]]
    | "gray" -> y = [[0, 0, 0], [255, 255, 255]]
    | _     ->  error("Colormap undefined!")
    end

    current_color_map.name = name
    current_color_map.data = y

end

colormap("gray") % Setup the default colormap

% Function: cmap
%
% Applies a colormap to a grayscale image, resulting in an RGB image
%
function y : cube = cmap(r : mat, map : mat, range)
    if isscalar(range) || numel(range)==0 % No range specified
        [r_min, r_max] = [min(r), max(r)]
    else
        [r_min, r_max] = range[0..1]
    endif
    function [] = __kernel__ cmap_kernel(x : mat'unchecked, y : cube'unchecked, map : mat'unchecked, r_min : scalar, r_max : scalar, pos : ivec2)
        val = (x[pos] - r_min) / (r_max - r_min) * size(map,0)
        % linearly interpolate
        if val < size(map,0)-1
            i = int(val)
            f = val - i
            y[pos[0],pos[1],0..2] = (1 - f) * map[i, 0..2] + f * map[i+1, 0..2]
        else
            y[pos[0],pos[1],0..2] = map[size(map,0)-1,0..2]
        endif
    end
    y = zeros(size(r,0),size(r,1),3)
    parallel_do(size(y,0..1),r,y,map,r_min,r_max,cmap_kernel)
end

% Function: imshow
% An overload of the imshow function; to allow color maps to be applied
%
% : imshow(x : mat)
% : imshow(x : mat, range : vec)
reduction (x : mat, range : vec) -> imshow(x, range) = imshow(cmap(x, current_color_map.data, range))
reduction (x : mat) -> imshow(x) = imshow(cmap(x, current_color_map.data, [0, 255]))
reduction (d : qdisplay, x : mat, range : vec) -> d.imshow(x, range) = d.imshow(cmap(x, current_color_map.data, [0, 255]), range)
reduction (d : qdisplay, x : mat) -> d.imshow(x) = d.imshow(cmap(x, current_color_map.data, [0, 255]))

% Example
function [] = main()

    [x, y] = meshgrid(-255..256,-255..256)
    assert(type(x,"mat")) % Now it is necessary to state that x has type 'mat'
    assert(type(y,"mat")) % but this will later be done automatically
    r = sqrt(x.^2+y.^2).*(x.^2+y.^2<256^2)    
    colormap("hsv")
    imshow(r)

end
