%==============================================================================
% Quasar compact core library
% (C) 2011-2015 Bart Goossens
%==============================================================================
% This file contains an implementation of the core matrix operations in Quasar.
% Originally, all functionality in this file was implemented in the Quasar
% runtime library, independently for each computation engine. Obviously, this
% generates a lot of functionality to be implemented at the back-end side and
% also makes possible future extensions more difficult.
%
% In this module, the core runtime functionality of Quasar is implemented in
% Quasar itself. The advantage is that the hybrid execution engine then
% handles all functions automatically. On the other hand the hybrid execution
% engine also involves some overhead and some of the functions _may_ be
% slower using this approach. To alleviate this overhead, the compact core
% is JIT'ed internally. Hence all functions in this module are device-
% independent, which is a big benefit when it comes to supporting more
% back-end devices.
%
% To keep the implementation practical, we heavily rely on Quasar generics.
%
% Please only edit this file if you know what you are doing - mistakes may
% affect *all* Quasar programs.

#pragma out_of_bounds_checks(off)

nextpow2    = x -> exp2(ceil(log2(x)))

% The following reduction modifies the pattern matching for this file.
% It ensures that fixed length vectors are not recognized as vectors (kind of
% overruling the type system).
{!reduction treat_vecX_as_vec=false}

%==============================================================================
% OpenGL/Hardware texture unit helper functions
% These functions are used by the Hyperion runtime engine
%==============================================================================

% Device -> Texture copy functions
function [] = __kernel__ __copy_vectotex(x : vec'unchecked, y : vec'hwtex_nearest, pos : int)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_mattotex(x : mat'unchecked, y : mat'hwtex_nearest, pos : ivec2)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_cubetotex(x : cube'unchecked, y : cube'hwtex_nearest, pos : ivec3)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_mat2totex(x : cube'unchecked, y : cube'hwtex_nearest(2), pos : ivec2)
    y[pos[0], pos[1], 0..1] = x[pos[0], pos[1], 0..1]
end

function [] = __kernel__ __copy_mat3totex(x : cube'unchecked, y : cube'hwtex_nearest(4), pos : ivec2)
    tmp = x[pos[0], pos[1], 0..2]
    y[pos[0], pos[1], 0..3] = [...tmp, 0.0]
end

function [] = __kernel__ __copy_mat4totex(x : cube'unchecked, y : cube'hwtex_nearest(4), pos : ivec2)
    y[pos[0], pos[1], 0..3] = x[pos[0], pos[1], 0..3]
end

% Texture -> Texture copy functions
function [] = __kernel__ __copy_textotex1(x : vec'hwtex_nearest, y : vec'hwtex_nearest, pos : int)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_textotex2(x : mat'hwtex_nearest, y : mat'hwtex_nearest, pos : ivec2)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_textotex3(x : cube'hwtex_nearest, y : cube'hwtex_nearest, pos : ivec3)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_tex2totex2(x : cube'hwtex_nearest(2), y : cube'hwtex_nearest(2), pos : ivec3)
    y[pos[0], pos[1], 0..1] = x[pos[0], pos[1], 0..1]
end

function [] = __kernel__ __copy_tex4totex4(x : cube'hwtex_nearest(4), y : cube'hwtex_nearest(4), pos : ivec3)
    y[pos[0], pos[1], 0..3] = x[pos[0], pos[1], 0..3]
end

% Texture -> Device copy functions
function [] = __kernel__ __copy_textovec(x : vec'hwtex_nearest, y : vec'unchecked, pos : int)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_textomat(x : mat'hwtex_nearest, y : mat'unchecked, pos : ivec2)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_textocube(x : cube'hwtex_nearest, y : cube'unchecked, pos : ivec3)
    y[pos] = x[pos]
end

function [] = __kernel__ __copy_textomat2(x : cube'hwtex_nearest(2), y : cube'unchecked, pos : ivec2)
    y[pos[0], pos[1], 0..1] = x[pos[0], pos[1], 0..1]
end

function [] = __kernel__ __copy_textomat3(x : cube'hwtex_nearest(4), y : cube'unchecked, pos : ivec2)
    y[pos[0], pos[1], 0..2] = x[pos[0], pos[1], 0..2]
end

function [] = __kernel__ __copy_textomat4(x : cube'hwtex_nearest(4), y : cube'unchecked, pos : ivec2)
    y[pos[0], pos[1], 0..3] = x[pos[0], pos[1], 0..3]
end

%==============================================================================
% Matrix construction
%==============================================================================

% Constant matrix (for the function __zeros and __ones).
function y:cube'unchecked'nocopy = __constantmtx(M : int, N : int, P : int, val : scalar) concealed
    y = uninit(M,N,P)
    #pragma force_parallel
    for m = 0..size(y,0)-1
        for n = 0..size(y,1)-1
            for k = 0..size(y,2)-1
                y[m,n,k] = val
            end
        end
    end
end

function y:cube'unchecked'nocopy = __constantmtx(dims : vec, val : scalar) concealed
    y = uninit(dims)
    parallel_do(numel(y),y,val, __kernel__ (y : vec, val : scalar, pos : int) -> y[pos] = val)
end

% Constant complex-valued matrix (for the function __czeros and __cones).
function y:ccube'unchecked'nocopy = __cconstantmtx(M : int, N : int, P : int, val : cscalar) concealed
    y = cuninit(M,N,P)
    parallel_do(numel(y),y,val, __kernel__ (y : cvec, val : cscalar, pos : int) -> y[pos] = val)
end

function y:ccube'unchecked'nocopy = __cconstantmtx(dims : vec, val : cscalar) concealed
    y = cuninit(dims)
    parallel_do(numel(y),y,val, __kernel__ (y : cvec, val : cscalar, pos : int) -> y[pos] = val)
end


% For the following functions, the runtime library automatically handles special
% call cases (e.g. when a vector is passed)...
reduction (m : scalar)                          -> zeros(m)        = __constantmtx(1,m,1,0)
reduction (m : scalar, n : scalar)              -> zeros(m, n)     = __constantmtx(m,n,1,0)
reduction (m : scalar, n : scalar, p : scalar)  -> zeros(m, n, p)  = __constantmtx(m,n,p,0)
reduction (m : vec)                             -> zeros(m)        = __constantmtx(m,0)
reduction (m : scalar)                          -> ones(m)         = __constantmtx(1,m,1,1)
reduction (m : scalar, n : scalar)              -> ones(m, n)      = __constantmtx(m,n,1,1)
reduction (m : scalar, n : scalar, p : scalar)  -> ones(m, n, p)   = __constantmtx(m,n,p,1)
reduction (m : vec)                             -> ones(m)         = __constantmtx(m,1)
reduction (m : scalar)                          -> czeros(m)       = __cconstantmtx(1,m,1,0i)
reduction (m : scalar, n : scalar)              -> czeros(m, n)    = __cconstantmtx(m,n,1,0i)
reduction (m : scalar, n : scalar, p : scalar)  -> czeros(m, n, p) = __cconstantmtx(m,n,p,0i)
reduction (m : vec)                             -> czeros(m)       = __cconstantmtx(m,0i)
reduction (m : scalar)                          -> cones(m)        = __cconstantmtx(1,m,1,1i)
reduction (m : scalar, n : scalar)              -> cones(m, n)     = __cconstantmtx(m,n,1,1i)
reduction (m : scalar, n : scalar, p : scalar)  -> cones(m, n, p)  = __cconstantmtx(m,n,p,1i)
reduction (m : vec)                             -> cones(m)        = __cconstantmtx(m,1i)
reduction (m : scalar, v : scalar)              -> __const(m, v)   = __constantmtx(1,m,1,v)
reduction (m : scalar, n : scalar, v : scalar)  -> __const(m, n, v)= __constantmtx(m,n,1,v)
reduction (m : scalar, n : scalar, p : scalar, v : scalar)  -> __const(m, n, p, v)  = __constantmtx(m,n,p,v)
reduction (m : vec, v : scalar)  -> __const(m, v) = __constantmtx(m,v)
reduction (m : scalar, v : cscalar)             -> __cconst(m, v)        = __cconstantmtx(1,m,1,v)
reduction (m : scalar, n : scalar, v : cscalar) -> __cconst(m, n, v)     = __cconstantmtx(m,n,1,v)
reduction (m : scalar, n : scalar, p : scalar, v : cscalar)  -> __cconst(m, n, p, v)  = __cconstantmtx(m,n,p,v)
reduction (m : vec, v : cscalar)  -> __cconst(m, v) = __cconstantmtx(m,v)

% Identity matrix
function y:'unchecked'nocopy = __identitymtx(N : int) concealed
    y = uninit(N,N)
    for m=0..N-1
        for n=0..N-1
            y[m,n] = m==n
        end
    end
end

reduction (n : int) -> eye(n) = __identitymtx(n)

% Sequences
function y:'unchecked'nocopy = r_seq[T](a : T, b : T) concealed
    count = b - a + 1
    y = uninit[T](count)
    for m = 0..count-1
        y[m] = a + m
    end
end

function y:'unchecked = r_seq2[T](a : T, b : T, step : T) concealed
    count = floor((b - a)/step) + 1
    y = uninit[T](count)
    for m = 0..count-1
        y[m] = a + m * step
    end
end

function y:'unchecked'nocopy = r_linspace(a : scalar, b : scalar, count : int) concealed
    step = (b-a)/(count-1)
    y = uninit(count)
    for m = 0..count-1
        y[m] = a + step * m
    end
end


%==============================================================================
% Operators
%==============================================================================

% Negation
operator (a : cube{:})  -> -a = -a
operator (a : ccube{:}) -> -a = -a

% Point-wise addition of matrices
operator (a : cube{:}, b : cube{:})   -> a + b = a + b
operator (a : ccube{:}, b : ccube{:}) -> a + b = a + b
operator (a : cube{:}, b : ccube{:})  -> a + b = a + b
operator (a : ccube{:}, b : cube{:})  -> a + b = b + a

operator (a : cube{:}, b : cube{:})   -> (a += b) = (a += b)
operator (a : ccube{:}, b : ccube{:}) -> (a += b) = (a += b)
operator (a : ccube{:}, b : cube{:})  -> (a += b) = (a += b)
operator (a : cube{:}, b : scalar) -> (a += b) = (a += b)
operator (a : ccube{:}, b : scalar) -> (a += b) = (a += b)
operator (a : ccube{:}, b : cscalar) -> (a += b) = (a += b)

% Point-wise subtraction of matrices
operator (a : cube{:}, b : cube{:})   -> a - b = a - b
operator (a : ccube{:}, b : ccube{:}) -> a - b = a - b
operator (a : cube{:}, b : ccube{:})  -> a - b = a - b
operator (a : ccube{:}, b : cube{:})  -> a - b = a - b

operator (a : cube{:}, b : cube{:})   -> (a -= b) = (a -= b)
operator (a : ccube{:}, b : ccube{:}) -> (a -= b) = (a -= b)
operator (a : ccube{:}, b : cube{:})  -> (a -= b) = (a -= b)
operator (a : cube{:}, b : scalar) -> (a -= b) = (a -= b)
operator (a : ccube{:}, b : scalar) -> (a -= b) = (a -= b)
operator (a : ccube{:}, b : cscalar) -> (a -= b) = (a -= b)

% Point-wise multiplication of matrices
operator (a : cube{:}, b : cube{:})   -> a .* b = a .* b
operator (a : ccube{:}, b : ccube{:}) -> a .* b = a .* b
operator (a : cube{:}, b : ccube{:})  -> a .* b = b .* a
operator (a : ccube{:}, b : cube{:})  -> a .* b = a .* b

operator (a : cube{:}, b : cube{:})   -> (a .*= b) = (a .*= b)
operator (a : ccube{:}, b : ccube{:}) -> (a .*= b) = (a .*= b)
operator (a : ccube{:}, b : cube{:})  -> (a .*= b) = (a .*= b)
operator (a : cube{:}, b : scalar) -> (a .*= b) = (a .*= b)
operator (a : ccube{:}, b : scalar) -> (a .*= b) = (a .*= b)
operator (a : ccube{:}, b : cscalar) -> (a .*= b) = (a .*= b)

% Point-wise division of matrices
operator (a : cube{:}, b : cube{:})   -> a ./ b = a ./ b
operator (a : ccube{:}, b : ccube{:}) -> a ./ b = a ./ b
operator (a : cube{:}, b : ccube{:})  -> a ./ b = a ./ b
operator (a : ccube{:}, b : cube{:})  -> a ./ b = a ./ b

operator (a : cube{:}, b : cube{:})   -> (a ./= b) = (a ./= b)
operator (a : ccube{:}, b : ccube{:}) -> (a ./= b) = (a ./= b)
operator (a : ccube{:}, b : cube{:})  -> (a ./= b) = (a ./= b)
operator (a : cube{:}, b : scalar) -> (a ./= b) = (a ./= b)
operator (a : ccube{:}, b : scalar) -> (a ./= b) = (a ./= b)
operator (a : ccube{:}, b : cscalar) -> (a ./= b) = (a ./= b)

% Point-wise power of matrices
operator (a : cube{:}, b : cube{:})   -> a .^ b = a .^ b
operator (a : ccube{:}, b : ccube{:}) -> a .^ b = a .^ b
operator (a : cube{:}, b : ccube{:})  -> a .^ b = a .^ b
operator (a : ccube{:}, b : cube{:})  -> a .^ b = a .^ b

operator (a : cube{:}, b : cube{:})   -> (a .^= b) = (a .^= b)
operator (a : ccube{:}, b : ccube{:}) -> (a .^= b) = (a .^= b)
operator (a : ccube{:}, b : cube{:})  -> (a .^= b) = (a .^= b)
operator (a : cube{:}, b : scalar) -> (a .^= b) = (a .^= b)
operator (a : ccube{:}, b : scalar) -> (a .^= b) = (a .^= b)
operator (a : ccube{:}, b : cscalar) -> (a .^= b) = (a .^= b)

% Adding a scalar number
operator (a : cube{:}, b : scalar)   -> a + b = a + b
operator (a : ccube{:}, b : cscalar) -> a + b = a + b
operator (a : ccube{:}, b : scalar)  -> a + b = a + b
operator (a : cube{:}, b : cscalar)  -> a + b = a + b
operator (a : scalar, b : cube{:})   -> a + b = a + b
operator (a : cscalar, b : ccube{:}) -> a + b = a + b
operator (a : scalar, b : ccube{:})  -> a + b = a + b
operator (a : cscalar, b : cube{:})  -> a + b = a + b

%% Subtracting a scalar number
operator (a : cube{:}, b : scalar)   -> a - b = a - b
operator (a : ccube{:}, b : cscalar) -> a - b = a - b
operator (a : ccube{:}, b : scalar)  -> a - b = a - b
operator (a : cube{:}, b : cscalar)  -> a - b = a - b

% Subtracting a scalar number (in reverse)
operator (a : scalar, b : cube{:})   -> a - b = a - b
operator (a : cscalar, b : ccube{:}) -> a - b = a - b
operator (a : scalar, b : ccube{:})  -> a - b = a - b
operator (a : cscalar, b : cube{:})  -> a - b = a - b

% Multiplication with a scalar number
operator (a : cube{:}, b : scalar)   -> a * b = a .* b
operator (a : ccube{:}, b : cscalar) -> a * b = a .* b
operator (a : ccube{:}, b : scalar)  -> a * b = a .* b
operator (a : cube{:}, b : cscalar)  -> a * b = a .* b
operator (a : cube{:}, b : scalar)   -> a .* b = a .* b
operator (a : ccube{:}, b : cscalar) -> a .* b = a .* b
operator (a : ccube{:}, b : scalar)  -> a .* b = a .* b
operator (a : cube{:}, b : cscalar)  -> a .* b = a .* b

operator (a : scalar, b : cube{:})   -> a * b = a .* b
operator (a : cscalar, b : ccube{:}) -> a * b = a .* b
operator (a : scalar, b : ccube{:})  -> a * b = a .* b
operator (a : cscalar, b : cube{:})  -> a * b = a .* b
operator (a : scalar, b : cube{:})   -> a .* b = a .* b
operator (a : cscalar, b : ccube{:}) -> a .* b = a .* b
operator (a : scalar, b : ccube{:})  -> a .* b = a .* b
operator (a : cscalar, b : cube{:})  -> a .* b = a .* b

operator (a : cube{:}, b : scalar) -> (a *= b) = (a *= b)
operator (a : ccube{:}, b : scalar) -> (a *= b) = (a *= b)
operator (a : ccube{:}, b : cscalar) -> (a *= b) = (a *= b)

% Division by a scalar number
operator (a : cube{:}, b : scalar)   -> a / b = a / b
operator (a : ccube{:}, b : cscalar) -> a / b = a / b
operator (a : ccube{:}, b : scalar)  -> a / b = a / b
operator (a : cube{:}, b : cscalar)  -> a / b = a / b
operator (a : cube{:}, b : scalar)   -> a ./ b = a / b
operator (a : ccube{:}, b : cscalar) -> a ./ b = a / b
operator (a : ccube{:}, b : scalar)  -> a ./ b = a / b
operator (a : cube{:}, b : cscalar)  -> a ./ b = a / b

operator (a : cube{:}, b : scalar) -> (a /= b) = (a /= b)
operator (a : ccube{:}, b : scalar) -> (a /= b) = (a /= b)
operator (a : ccube{:}, b : cscalar) -> (a /= b) = (a /= b)

% Reverse pointwise division scalar number (in reverse)
operator (a : scalar, b : cube{:})   -> a ./ b = a ./ b
operator (a : cscalar, b : ccube{:}) -> a ./ b = a ./ b
operator (a : scalar, b : ccube{:})  -> a ./ b = a ./ b
operator (a : cscalar, b : cube{:})  -> a ./ b = a ./ b

% Power of a scalar number
operator (a : cube{:}, b : scalar)   -> a .^ b = a .^ b
operator (a : ccube{:}, b : cscalar) -> a .^ b = a .^ b
operator (a : ccube{:}, b : scalar)  -> a .^ b = a .^ b
operator (a : cube{:}, b : cscalar)  -> a .^ b = a .^ b

% Scalar number to a matrix power (pointwise)
operator (a : scalar, b : cube{:})   -> a .^ b = a .^ b
operator (a : cscalar, b : ccube{:}) -> a .^ b = a .^ b
operator (a : scalar, b : ccube{:})  -> a .^ b = a .^ b
operator (a : cscalar, b : cube{:})  -> a .^ b = a .^ b

%% Less
operator (a : cube{:}, b : cube{:})   -> a < b = a < b
operator (a : cube{:}, b : scalar) -> a < b = a < b
operator (a : scalar, b : cube{:}) -> a < b = a < b
operator (a : cube{:}, b : cube{:})   -> a > b = b < a
operator (a : cube{:}, b : scalar) -> a > b = b < a
operator (a : scalar, b : cube{:}) -> a > b = b < a

% Less or equal
operator (a : cube{:}, b : cube{:})   -> a <= b = a <= b
operator (a : cube{:}, b : scalar) -> a <= b = a <= b
operator (a : scalar, b : cube{:}) -> a <= b = a <= b
operator (a : cube{:}, b : cube{:})   -> a >= b = b <= a
operator (a : cube{:}, b : scalar) -> a >= b = b <= a
operator (a : scalar, b : cube{:}) -> a >= b = b <= a

% Equal
operator (a : cube{:}, b : cube{:})     -> a == b = a == b
operator (a : cube{:}, b : scalar)   -> a == b = a == b
operator (a : scalar, b : cube{:})   -> a == b = a == b
operator (a : ccube{:}, b : ccube{:})   -> a == b = a == b
operator (a : ccube{:}, b : cscalar) -> a == b = a == b
operator (a : cscalar, b : ccube{:}) -> a == b = a == b
%
%% Not equal
operator (a : cube{:}, b : cube{:})     -> a != b = a != b
operator (a : cube{:}, b : scalar)   -> a != b = a != b
operator (a : scalar, b : cube{:})   -> a != b = a != b
operator (a : ccube{:}, b : ccube{:})   -> a != b = a != b
operator (a : ccube{:}, b : cscalar) -> a != b = a != b
operator (a : cscalar, b : ccube{:}) -> a != b = a != b

% Inverse
operator (a : cube{:})  -> !a = !a

% Logical AND
operator (a : cube{:}, b : cube{:})     -> a && b = a && b
operator (a : cube{:}, b : scalar)   -> a && b = a && b
operator (a : scalar, b : cube{:})   -> a && b = a && b

% Logical OR
operator (a : cube{:}, b : cube{:})     -> a || b = a || b
operator (a : cube{:}, b : scalar)   -> a || b = a || b
operator (a : scalar, b : cube{:})   -> a || b = a || b

% Bitwise AND
operator (a : cube{:}, b : cube{:})     -> and(a, b) = and(a, b)
operator (a : cube{:}, b : scalar)   -> and(a, b) = and(a, b)
operator (a : scalar, b : cube{:})   -> and(a, b) = and(a, b)

% Bitwise OR
operator (a : cube{:}, b : cube{:})     -> or(a, b) = or(a, b)
operator (a : cube{:}, b : scalar)   -> or(a, b) = or(a, b)
operator (a : scalar, b : cube{:})   -> or(a, b) = or(a, b)

% Bitwise XOR
operator (a : cube{:}, b : cube{:})     -> xor(a, b) = xor(a, b)
operator (a : cube{:}, b : scalar)   -> xor(a, b) = xor(a, b)
operator (a : scalar, b : cube{:})   -> xor(a, b) = xor(a, b)

% Bitwise left shift
operator (a : cube{:}, b : cube{:})     -> shl(a, b) = shl(a, b)
operator (a : cube{:}, b : scalar)   -> shl(a, b) = shl(a, b)
operator (a : scalar, b : cube{:})   -> shl(a, b) = shl(a, b)

% Bitwise right shift
operator (a : cube{:}, b : cube{:})     -> shr(a, b) = shr(a, b)
operator (a : cube{:}, b : scalar)   -> shr(a, b) = shr(a, b)
operator (a : scalar, b : cube{:})   -> shr(a, b) = shr(a, b)

%==============================================================================
% Matrix functions
%==============================================================================

function y : 'unchecked = __transpose[T](x : mat[T]'unchecked) concealed
    y = mat[T](size(x,1),size(x,0))

    #pragma force_parallel
    for m = 0..size(y,0)-1
        for n = 0..size(y,1)-1
            y[m,n] = x[n,m]
        end
    end
end

reduction (x : mat)  -> transpose(x) = __transpose(x)
reduction (x : cmat) -> transpose(x) = __transpose(x)

function y : 'unchecked = __herm_transpose(x : cmat) concealed
    y = cuninit(size(x,1),size(x,0))

    #pragma force_parallel
    for m = 0..size(y,0)-1
        for n = 0..size(y,1)-1
            y[m,n] = conj(x[n,m])
        end
    end
end

%reduction (x : mat) -> herm_transpose(x) = __transpose(x)
%reduction (x : cmat) -> herm_transpose(x) = __herm_transpose(x)

%==============================================================================
% Matrix multiplication
% TODO - we may have to replace this by a more efficient algorithm
%==============================================================================

function y = __mmultiply[T](a : mat[T]'unchecked, b : mat[T]'unchecked) concealed
    assert(size(a,1)==size(b,0),"Matrix multiplication: dimensions do not match:",size(a,0..1),"x",size(b,0..1),"!")
    y0 = uninit[T](size(a,0),size(b,1))
    for m = 0..size(y0,0)-1
        for n = 0..size(y0,1)-1
            result = cast(0.0, T)
            for k = 0..size(a,1)-1
                result += a[m,k] * b[k,n]
            end
            y0[m,n] = result
        end
    end
    if numel(y0)==1 % The result is a scalar number
        y = y0[0]
    else
        y = y0
    endif
end

reduction (a : mat, b : mat)     -> a * b = __mmultiply(a, b)
reduction (a : cmat, b : cmat)   -> a * b = __mmultiply(a, b)

%==============================================================================
% Mathematical functions on matrices
%==============================================================================

operator (a : cube{:}) -> abs(a) = abs(a)
operator (a : cube{:}[int]) -> abs(a) = abs(a)
operator (a : cube{:}) -> ceil(a) = ceil(a)
operator (a : cube{:}) -> round(a) = round(a)
operator (a : cube{:}) -> cos(a) = cos(a)
operator (a : cube{:}) -> sin(a) = sin(a)
operator (a : cube{:}) -> exp(a) = exp(a)
operator (a : cube{:}) -> exp2(a) = exp2(a)
operator (a : cube{:}) -> floor(a) = floor(a)
operator (a : cube{:}, b : cube{:}) -> mod(a, b) = mod(a, b)
operator (a : cube{:}, b : scalar) -> mod(a, b) = mod(a, b)
operator (a : cube{:}) -> frac(a) = frac(a)
operator (a : cube{:}) -> log(a) = log(a)
operator (a : cube{:}) -> log2(a) = log2(a)
operator (a : cube{:}) -> log10(a) = log10(a)
operator (a : cube{:}, b : cube{:}) -> min(a, b) = min(a, b)
operator (a : cube{:}, b : cube{:}) -> max(a, b) = max(a, b)
operator (a : cube{:}, b : cube{:}) -> pow(a, b) = pow(a, b)
operator (a : cube{:}, b : scalar) -> min(a, b) = min(a, b)
operator (a : cube{:}, b : scalar) -> max(a, b) = max(a, b)
operator (a : scalar, b : cube{:}) -> min(a, b) = min(a, b)
operator (a : scalar, b : cube{:}) -> max(a, b) = max(a, b)
operator (a : cube{:}, b : scalar) -> pow(a, b) = pow(a, b)
operator (a : scalar, b : cube{:}) -> pow(a, b) = pow(a, b)
operator (a : cube{:}) -> saturate(a) = saturate(a)
operator (a : cube{:}) -> angle(a) = 0*a
operator (a : cube{:}) -> sign(a) = sign(a)
operator (a : cube{:}) -> sqrt(a) = sqrt(a)
operator (a : cube{:}) -> cosh(a) = cosh(a)
operator (a : cube{:}) -> sinh(a) = sinh(a)
operator (a : cube{:}) -> tanh(a) = tanh(a)
operator (a : cube{:}) -> isinf(a) = isinf(a)
operator (a : cube{:}) -> isfinite(a) = isfinite(a)
operator (a : cube{:}) -> isnan(a) = isnan(a)
operator (a : cube{:}, b : cube{:}) -> periodize(a, b) = periodize(a, b)
operator (a : cube{:}, b : cube{:}) -> mirror_ext(a, b) = mirror_ext(a, b)
operator (a : cube{:}, b : cube{:}) -> clamp(a, b) = clamp(a, b)
operator (a : cube{:}, b : scalar) -> periodize(a, b) = periodize(a, b)
operator (a : cube{:}, b : scalar) -> mirror_ext(a, b) = mirror_ext(a, b)
operator (a : cube{:}, b : scalar) -> clamp(a, b) = clamp(a, b)
operator (a : cube{:}) -> tan(a) = tan(a)
operator (a : cube{:}) -> acos(a) = acos(a)
operator (a : cube{:}) -> asin(a) = asin(a)
operator (a : cube{:}) -> atan(a) = atan(a)
operator (a : cube{:}, b : cube{:}) -> atan2(a,b) = atan2(a,b)
operator (a : cube{:}) -> acosh(a) = acosh(a)
operator (a : cube{:}) -> asinh(a) = asinh(a)
operator (a : cube{:}) -> atanh(a) = atanh(a)
operator (a : cube{:}) -> rcp_rn(a) = rcp_rn(a)
operator (a : cube{:}) -> rcp_ru(a) = rcp_ru(a)
operator (a : cube{:}) -> rcp_rz(a) = rcp_rz(a)
operator (a : cube{:}, b : cube{:}) -> div_approx(a, b) = div_approx(a, b)


% TODO: conversion functions int, float, scalar needed?

% Complex-valued functions (note: type inference will be used for determining
% the type of the return value)
operator (a : ccube{:}) -> abs(a) = abs(a)
operator (a : ccube{:}) -> angle(a) = angle(a)
operator (a : ccube{:}) -> cos(a) = cos(a)
operator (a : ccube{:}) -> sin(a) = sin(a)
operator (a : ccube{:}) -> exp(a) = exp(a)
operator (a : ccube{:}) -> exp2(a) = exp2(a)
operator (a : ccube{:}) -> log(a) = log(a)
operator (a : ccube{:}) -> log2(a) = log2(a)
operator (a : ccube{:}) -> log10(a) = log10(a)
operator (a : ccube{:}, b : ccube{:}) -> pow(a, b) = pow(a, b)
%operator (a : ccube{:}) -> sign(a) = sign(a)
operator (a : ccube{:}) -> sqrt(a) = sqrt(a)
operator (a : ccube{:}) -> tan(a) = tan(a)
operator (a : ccube{:}) -> asin(a) = asin(a)
operator (a : ccube{:}) -> acos(a) = acos(a)
operator (a : ccube{:}) -> atan(a) = atan(a)
operator (a : ccube{:}) -> ceil(a) = ceil(a)
operator (a : ccube{:}) -> round(a) = round(a)
operator (a : ccube{:}) -> floor(a) = floor(a)
operator (a : ccube{:}) -> frac(a) = frac(a)
operator (a : ccube{:}) -> saturate(a) = saturate(a)
operator (a : ccube{:}) -> isinf(a) = isinf(a)
operator (a : ccube{:}) -> isfinite(a) = isfinite(a)
operator (a : ccube{:}) -> isnan(a) = isnan(a)
operator (a : ccube{:}) -> real(a) = real(a)
operator (a : ccube{:}) -> imag(a) = imag(a)
operator (a : ccube{:}) -> conj(a) = conj(a)
reduction (x : cube{:}) -> conj(x) = x
operator (a : cube{:}, b : cube{:}) -> complex(a, b) = complex(a, b)
operator (a : cube{:}) -> complex(a) = complex(a, 0.0)


function y = r_replicatemtx[T](x : cube{:}[T], cnt : vec) concealed
    if numel(cnt) < ndims(x)
        error "(repmat): the replication vector should have at least ", ndims(x), " components!"
    endif
    n = numel(cnt)
    x_dim = size(x,0..n-1)
    y = uninit[T](x_dim .* cnt)
    rep = prod(cnt)

    function [] = __kernel__ kernel (x : vec[T]'unchecked, y : vec[T]'unchecked, x_dim, cnt, rep : int, pos : int)
        y_dim = x_dim .* cnt
        x_pos = ind2pos(x_dim, pos)
        val = x[pos]
        for j=0..rep-1
            y_pos = x_pos + x_dim .* ind2pos(cnt, j)
            y[pos2ind(y_dim, y_pos)] = val
        end
    end

    match n with
    | 1 -> parallel_do(numel(x), x, y, x_dim, cnt[0], rep, $specialize(kernel,type(x_dim,int) && type(cnt,int)))
    | 2 -> parallel_do(numel(x), x, y, x_dim, cnt, rep, $specialize(kernel,type(x_dim,ivec2) && type(cnt,ivec2)))
    | 3 -> parallel_do(numel(x), x, y, x_dim, cnt, rep, $specialize(kernel,type(x_dim,ivec3) && type(cnt,ivec3)))
    | 4 -> parallel_do(numel(x), x, y, x_dim, cnt, rep, $specialize(kernel,type(x_dim,ivec4) && type(cnt,ivec4)))
    | 5 -> parallel_do(numel(x), x, y, x_dim, cnt, rep, $specialize(kernel,type(x_dim,ivec5) && type(cnt,ivec5)))
    | 6 -> parallel_do(numel(x), x, y, x_dim, cnt, rep, $specialize(kernel,type(x_dim,ivec6) && type(cnt,ivec6)))
    | _ -> error("(repmat): no more than 6 dimensions are currently supported!")
    end
end

reduction (x : cube{:}, c : int)      -> repmat(x, c) = r_replicatemtx(x, [c])
reduction (x : cube{:}, c : vec)      -> repmat(x, c) = r_replicatemtx(x, c)
reduction (x : ccube{:}, c : int)     -> repmat(x, c) = r_replicatemtx(x, [c])
reduction (x : ccube{:}, c : vec)     -> repmat(x, c) = r_replicatemtx(x, c)


function y = __shuffledims[T](x : cube{:}[T], shuffle : vec)

    new_dim = size(x, shuffle)
    dim_stride = zeros(numel(shuffle))
    n = ndims(x)
    dim_stride[n-1] = 1
    {!interpreted for}
    for i = n-2..-1..0
        dim_stride[i] = dim_stride[i + 1] * size(x, i + 1)
    end
    dim_stride[n..numel(shuffle)-1] = dim_stride[0] * size(x, 0)
    new_stride = zeros(numel(shuffle))
    {!interpreted for}
    for i = 0..numel(shuffle)-1
        new_stride[i] = dim_stride[shuffle[i]]
    end

    function [] = __kernel__ kernel (x : vec[T]'unchecked, y : vec[T]'unchecked, x_dim, y_stride, step : int, pos : int)
        for p = pos..step..numel(x)-1
            x_pos = ind2pos(x_dim, p)
            y[p] = x[dotprod(x_pos, y_stride)]
        end
    end

    y = uninit[T](new_dim)
    cnt = min(nextpow2(numel(y)), 16384)
    match numel(new_dim) with
    | 1 -> y = reshape(x, numel(x))
    | 2 -> parallel_do(cnt,x,y,new_dim,new_stride,cnt,$specialize(kernel, type(x_dim, ivec2) && type(y_stride, ivec2)))
    | 3 -> parallel_do(cnt,x,y,new_dim,new_stride,cnt,$specialize(kernel, type(x_dim, ivec3) && type(y_stride, ivec3)))
    | 4 -> parallel_do(cnt,x,y,new_dim,new_stride,cnt,$specialize(kernel, type(x_dim, ivec4) && type(y_stride, ivec4)))
    | 5 -> parallel_do(cnt,x,y,new_dim,new_stride,cnt,$specialize(kernel, type(x_dim, ivec5) && type(y_stride, ivec5)))
    | 6 -> parallel_do(cnt,x,y,new_dim,new_stride,cnt,$specialize(kernel, type(x_dim, ivec6) && type(y_stride, ivec6)))
    end
end


reduction (x : cube{:},  shuf : vec)  -> shuffledims(x, shuf) = __shuffledims(x, shuf)
reduction (x : ccube{:}, shuf : vec)  -> shuffledims(x, shuf) = __shuffledims(x, shuf)

%==============================================================================
% Binary operations (to be used in generic definitions)
%==============================================================================

op_minus       = __device__ (x, y) -> -x
op_add         = __device__ (x, y) -> x + y
op_subtract    = __device__ (x, y) -> x - y
op_mul         = __device__ (x, y) -> x * y
op_int_divide  = __device__ (x, y) -> int(x / y)  % Return type = int
op_divide      = __device__ (x, y) -> x / y       % Return type = scalar
op_pow         = __device__ (x, y) -> x ^ y
op_less        = __device__ (x, y) -> x < y
op_leq         = __device__ (x, y) -> x <= y
op_greater     = __device__ (x, y) -> x > y
op_geq         = __device__ (x, y) -> x >= y
op_eq          = __device__ (x, y) -> x == y
op_neq         = __device__ (x, y) -> x != y
op_negate      = __device__ (x, y) -> !x
op_and         = __device__ (x, y) -> and(x,y)
op_or          = __device__ (x, y) -> or(x,y)
op_xor         = __device__ (x, y) -> xor(x,y)
op_shl         = __device__ (x, y) -> shl(x,y) % left bitshift x << y
op_shr         = __device__ (x, y) -> shr(x,y) % right bitshift x >> y
op_min         = __device__ (x, y) -> min(x,y)
op_max         = __device__ (x, y) -> max(x,y)


%==============================================================================
% Matrix sum, product, minimum and maximum
%==============================================================================

% Sum of all elements in the specified cube.
% Note: we use Kahan's summation algorithm, to be more robust against numeric precision errors
function y = __sum[T](x : cube{:}[T]) concealed
    % This should be the CPU version
%    function z : T = __kernel__ serial_kernel(x : vec[T]'unchecked)
%        c = cast(0, T)
%        total = cast(0, T)
%        for n=0..numel(x)-1
%            y = x[n] - c
%            t = total + y
%            c = (t - total) - y
%            total = t
%        end
%        z = total
%    end
%
%    y = serial_do(1,x,serial_kernel)
    ux = reshape(x, numel(x))
    y = cast(0.0, T)
    if numel(x)>0
        #pragma force_parallel
        for m=0..numel(x)-1
            y += ux[m]
        end
    endif
end

function y = __prod[T](x : cube{:}[T]) concealed
    ux = reshape(x, numel(x))
    y = cast(1.0, T)
    if numel(x)>0
        #pragma force_parallel
        for m=0..numel(x)-1
            y *= ux[m]
        end
    endif
end

function y : scalar = __min[T](x : cube{:}[T]) concealed
    ux = reshape(x, numel(x))
    y = maxvalue(T)
    if numel(x)>0
        #pragma force_parallel
        for m=0..numel(x)-1
            y __= ux[m]
        end
    endif
end

function y : scalar = __max[T](x : cube{:}[T]) concealed
    ux = reshape(x, numel(x))
    y = minvalue(T)
    if numel(x)>0
        #pragma force_parallel
        for m=0..numel(x)-1
            y ^^= ux[m]
        end
    endif
end

reduction (x : cube{:}) -> sum(x)  = __sum(x)
reduction (x : ccube{:}) -> sum(x) = __sum(x)
reduction (x : cube{:}[int]) -> sum(x) = __sum(x)
reduction (x : cube{:}) -> prod(x) = __prod(x)
reduction (x : ccube{:}) -> prod(x) = __prod(x)
reduction (x : cube{:}) -> min(x) = __min(x)
reduction (x : cube{:}) -> max(x) = __max(x)
reduction (x : cube{:}[int]) -> min(x) = __min(x)
reduction (x : cube{:}[int]) -> max(x) = __max(x)
reduction (x : cube{:}) -> mean(x) = __sum(x) / numel(x)
reduction (x : ccube{:}) -> mean(x) = __sum(x) / numel(x)

function y : cube{:}[T] = __dimensionop[T](x : cube{:}[T], dim : int, z0 : T,
    gpu_kernel : [__kernel__ (vec[T], vec[T], T, int,int,int) -> ()],
    cpu_kernel : [__kernel__ (vec[T], vec[T], T, int,int,int) -> ()]) concealed

    rec_dim = size(x,0..dim-1)
    dominant_dim = size(x,dim)
    u = max(dim+1,ndims(x))
    sup_dim = size(x,dim+1..u)
    new_dims = size(x, 0..u) % the new dimensions
    new_dims[dim] = 1 % set the dominant dimension to 1
    type cube_n : cube{:}[T]
    y = cube_n(new_dims)


    match schedule(new_dims, x, y, prod(rec_dim), dominant_dim, prod(sup_dim), {gpu_kernel,cpu_kernel}) with
    | 0 -> % GPU version of the kernel
        max_threads = prod(max_block_size(gpu_kernel,[64,64]))
        block_size_x = int(max(1, (dominant_dim + max_threads - 1) / max_threads))
        n_threads = nextpow2(dominant_dim / block_size_x)
        block_size_y = int(max_threads / n_threads)
        blocks_y = int((prod(rec_dim) * prod(sup_dim) + block_size_y - 1) / block_size_y)
        blocks_z = 1
        max_grid_size = 65535

        while blocks_y > max_grid_size && and(blocks_y, 1) == 0
            blocks_y /= 2
            blocks_z *= 2
        end
        while blocks_y > max_grid_size && mod(blocks_y, 3) == 0
            blocks_y /= 3
            blocks_z *= 3
        end
        while blocks_y > max_grid_size && mod(blocks_y, 5) == 0
            blocks_y /= 5
            blocks_z *= 5
        end

        parallel_do([[blocks_z, blocks_y*block_size_y, n_threads],[1, block_size_y, n_threads]],x,y,z0,prod(rec_dim), dominant_dim, prod(sup_dim),gpu_kernel)
    | 1 -> % CPU version of the kernel
        parallel_do(prod(new_dims),x,y,z0,prod(rec_dim), dominant_dim, prod(sup_dim),cpu_kernel)
    end
end

% Sum of all elements in the specified cube, along the given dimension
function y : cube{:}[T] = r_aggregator_dim[T](x : cube{:}[T], dim : int, z0 : T, aggregator : [__device__ (T, T) -> T]) concealed
    % kernel for summing a matrix along the rows...
    function [] = __kernel__ gpu_kernel(x : vec[T]'unchecked, y : vec[T]'unchecked, z0 : T, x_rows : int, x_cols : int, x_stride : int,
        aggregator : [__device__ (T, T) -> T], blkdim : ivec3, blkpos : ivec3, blkcnt : ivec3, blkidx : ivec3)

        {!kernel target="gpu"}
        bins = shared[T](1024)
        i = blkpos[2]
        m = (blkidx[1]*blkdim[1]+blkpos[1]) + blkidx[0]*blkdim[1]*blkcnt[1]
        pos = ind2pos([x_rows, x_stride], m)
        if pos[0] >= x_rows || pos[1] >= x_stride
            return
        endif

        % step 1 - parallel sum
        val = z0
        offset = pos[0] * x_cols * x_stride + pos[1]
        for n=i * x_stride..blkdim[2]*x_stride..x_cols*x_stride-1
            val = aggregator(val, x[offset + n])
        end
        i = blkpos[1]*blkdim[2] + blkpos[2] % note: blkpos[2]==0
        bins[i] = val

        % step 2 - reduction
        syncthreads
        bit = 1
        while bit<blkdim[2]
            index=2*bit*blkpos[2]
            if index+bit<blkdim[2]
                bins[index] = aggregator(bins[index], bins[index+bit])
            endif
            syncthreads
            bit *= 2
        end
        if blkpos[2]==0
            y[pos[0] * x_stride + pos[1]] = bins[blkpos[1]*blkdim[2]]
        endif
    end

    function [] = __kernel__ cpu_kernel(x : vec[T]'unchecked, y : vec[T]'unchecked, z0 : T, x_rows : int, x_cols : int, x_stride : int,
        aggregator : [__device__ (T, T) -> T], pos : int)

        {!kernel target="cpu"}
        pos2 = ind2pos([x_rows, x_stride], pos)
        sz = [x_rows, x_cols, x_stride]
        val = z0

        for i=0..x_cols-1
            j=pos2ind(sz, [pos2[0],i,pos2[1]])
            val = aggregator(val, x[j])
        end

        y[pos] = val
    end

    y = __dimensionop(x, dim, z0, $specialize(gpu_kernel, aggregator==aggregator),
                              $specialize(cpu_kernel, aggregator==aggregator))
end



reduction (x : cube{:}, dim  : scalar)  -> sum(x, dim) = r_aggregator_dim(x, dim, 0.0, op_add)
reduction (x : ccube{:}, dim : scalar)  -> sum(x, dim) = r_aggregator_dim(x, dim, 0.0j, op_add)
reduction (x : cube{:}, dim  : scalar)  -> prod(x, dim) = r_aggregator_dim(x, dim, 1.0, op_mul)
reduction (x : ccube{:}, dim : scalar)  -> prod(x, dim) = r_aggregator_dim(x, dim, 1.0 + 0.0j, op_mul)
reduction (x : cube{:}, dim  : scalar)  -> mindim(x, dim) = r_aggregator_dim(x, dim, maxvalue(scalar), op_min)
reduction (x : cube{:}, dim  : scalar)  -> maxdim(x, dim) = r_aggregator_dim(x, dim, minvalue(scalar), op_max)
reduction (x : cube{:}, dim : scalar)   -> mean(x, dim) = r_aggregator_dim(x, dim, 0.0, op_add) / size(x, dim)

function y = r_recop[T](x : cube{:}[T], dim : int, z0 : T, aggregator : [__device__ (T, T) -> T])
    function [] = __kernel__ gpu_kernel(x : vec[T]'unchecked, y : vec[T],z0 : T, aggregator : [__device__ (T, T) -> T],
        block_size : int, y_rows : int, y_cols : int, y_stride : int, blkpos : ivec3, blkdim : ivec3, blkidx : ivec3, blkcnt : ivec3)

        {!kernel target="gpu"}
        block_sum = shared[T](1024)
        m = (blkidx[1]*blkdim[1] + blkpos[1]) + blkidx[0]*(blkdim[1]*blkcnt[1])
        pos = ind2pos([y_rows, y_stride], m)
        if pos[0] >= y_rows || pos[1] >= y_stride
            return
        endif

        % first, compute the cumulative sum within each block
        offset = pos[0]*y_cols*y_stride + pos[1]
        sum = z0
        count = min(block_size, y_cols - blkpos[2] * block_size)
        for i=0..count-1
            n = (blkpos[2] * block_size + i) * y_stride + offset
            y[n] = sum = aggregator(sum, x[n])
        end

        j = blkpos[1]*blkdim[2]+blkpos[2]  % note: blkpos[0]==0
        block_sum[j] = sum
        syncthreads

        % second, cumulate the block sums (there are blkdim[1] entries)
        % up-sweep
        %	 1 1 1 1 1 1 1 1
        %	   2   2   2   2
        %		   4       4
        %	 1 2 1 4 1 2 1 8
        [bit, i] = [1, blkpos[2]]
        while bit < blkdim[2]
            if and(blkpos[2],bit*2-1)==0
                block_sum[j+bit] = aggregator(block_sum[j], block_sum[j+bit])
            endif
            syncthreads
            j += bit
            bit *= 2
        end

        % down-sweep
        %   1 2 1 4 1 2 1 0     8
        %	      0       4
        %	  0   2   4   6
        %	0 1 2 3 4 5 6 7
        if blkpos[2] == 0
            block_sum[(blkpos[1]+1)*blkdim[2]-1] = z0
        endif
        bit /= 2
        while bit >= 1
            j -= bit
            if and(blkpos[2],bit*2-1)==0
                sum = aggregator(block_sum[j], block_sum[j+bit])
                block_sum[j] = block_sum[j+bit]
                block_sum[j+bit] = sum
            endif
            syncthreads
            bit /= 2
        end
        % accumulate the block sums
        sum = block_sum[blkpos[1]*blkdim[2]+blkpos[2]]
        for i=0..count-1
            n = (blkpos[2] * block_size + i) * y_stride + offset
            y[n] = aggregator(y[n], sum)
        end
    end

    function [] = __kernel__ cpu_kernel(x : vec[T]'unchecked, y : vec[T]'unchecked, x_rows : int, x_cols : int, x_stride : int,
        aggregator : [__device__ (T, T) -> T], pos : int)

        {!kernel target="cpu"}
        pos2 = ind2pos([x_rows, x_stride], pos)
        sz = [x_rows, x_cols, x_stride]
        val = z0

        for i=0..x_cols-1
            p = [pos2[0],i,pos2[1]]
            j=pos2ind(sz, p)
            val = aggregator(val, x[j])
            y[j] = val
        end
    end

    rec_dim = size(x,0..dim-1)
    dominant_dim = size(x,dim)
    u = max(dim+1,ndims(x))
    sup_dim = size(x,dim+1..u)
    new_dims = size(x,0..u) % the new dimensions
    new_dims[dim] = 1

    y = uninit[T](size(x))

    gpu_kernel_specialized = $specialize(gpu_kernel, z0==z0 && aggregator==aggregator)
    cpu_kernel_specialized = $specialize(cpu_kernel, z0==z0 && aggregator==aggregator)
 
    match schedule(size(x), x, y, prod(rec_dim), dominant_dim, prod(sup_dim), {gpu_kernel_specialized,cpu_kernel_specialized}) with
    | 0 -> % GPU version of the kernel
        max_threads = prod(max_block_size(gpu_kernel_specialized,[64,64]))
        block_size_x = int(max(1, (dominant_dim + max_threads - 1) / max_threads))
        n_threads = nextpow2(dominant_dim / block_size_x)
        block_size_y = int(max_threads / n_threads)
        blocks_y = int((prod(rec_dim) * prod(sup_dim) + block_size_y - 1) / block_size_y)
        blocks_z = 1
        max_grid_size = 65535

        while blocks_y > max_grid_size && and(blocks_y, 1) == 0
            blocks_y /= 2
            blocks_z *= 2
        end
        while blocks_y > max_grid_size && mod(blocks_y, 3) == 0
            blocks_y /= 3
            blocks_z *= 3
        end
        while blocks_y > max_grid_size && mod(blocks_y, 5) == 0
            blocks_y /= 5
            blocks_z *= 5
        end
        parallel_do([[blocks_z, blocks_y*block_size_y, n_threads],[1, block_size_y, n_threads]],
            x,y,block_size_x, prod(rec_dim), dominant_dim, prod(sup_dim), gpu_kernel_specialized)

    | 1 -> % CPU version of the kernel
        parallel_do(prod(new_dims),x,y,prod(rec_dim), dominant_dim, prod(sup_dim),cpu_kernel_specialized)
    end
end


reduction (x : cube{:}, dim : scalar) -> cumsum(x, dim) = r_recop(x, dim, 0.0, op_add)
reduction (x : cube{:})               -> cumsum(x) = cumsum(x, numel(size(x))-1)
reduction (x : cube{:}, dim : scalar) -> cumprod(x, dim) = r_recop(x, dim, 1.0, op_mul)
reduction (x : cube{:})               -> cumprod(x) = cumprod(x, numel(size(x))-1)
reduction (x : cube{:}, dim : scalar) -> cummin(x, dim) = r_recop(x, dim, maxvalue(scalar), op_min)
reduction (x : cube{:})               -> cummin(x) = cummin(x, numel(size(x))-1)
reduction (x : cube{:}, dim : scalar) -> cummax(x, dim) = r_recop(x, dim, minvalue(scalar), op_max)
reduction (x : cube{:})               -> cummax(x) = cummax(x, numel(size(x))-1)


%==============================================================================
% Helper function for checking indices. This function is called implicitly
% by the run-time system. DO NOT change the name of this function.
%==============================================================================
function y = __checkindices_base(indices : vec'unchecked, max_ind : int)
    for m = 0..numel(indices)-1
        if indices[m]<0 || indices[m] >= max_ind
            y = 1
        endif
    end
end

reduction (ind, max_ind) -> __checkindices(ind, max_ind) = __checkindices_base(ind, max_ind)

%==============================================================================
% Slice accessors for matrices and vectors
% Note: it may be possible that one of the indices is a scalar number. In this
% case, the runtime library will upcast the scalar number to a vector of length
% 1, so that the accessors below can still be used.
%==============================================================================

% Vector slice accessors
function y = vector_slice_get[T](x : vec[T]'unchecked, indices : vec'unchecked)
    sz = size(x)
    {!interpreted for}
    for i = 0..numel(sz)-1
        if sz[i] > 1
            sz[i] = numel(indices)
        endif
    end
    y = uninit[T](sz)
    for m = 0..numel(indices)-1
        y[m] = x[indices[m]]
    end
end

function z = vector_slice_put[T](y : vec[T]'unchecked, x : vec[T]'unchecked, indices : vec'unchecked)
    for m = 0..numel(indices)-1
        y[indices[m]] = x[m]
    end
    z = y
end

function z = vector_slice_put_const[T](y : vec[T]'unchecked, x : T, indices : vec'unchecked)
    for m = 0..numel(indices)-1
        y[indices[m]] = x
    end
    z = y
end

% Note: an alternative would be to define the reductions
reduction (x:vec, y:vec)                      -> x[y] = vector_slice_get(x, y)
reduction (x:vec, y:vec, z:vec)               -> (x[y] = z) = vector_slice_put(x,z,y)
reduction (x:vec, y:vec, z:scalar)            -> (x[y] = z) = vector_slice_put_const(x,z,y)
reduction (x:vec, z:vec)                      -> (x[:] = z) = vector_slice_put(x,z,0..numel(x)-1)
reduction (x:vec, z:scalar)                   -> (x[:] = z) = vector_slice_put_const(x,z,0..numel(x)-1)

reduction (x:cvec, y:vec)                     -> x[y] = vector_slice_get(x, y)
reduction (x:cvec, y:vec, z:cvec)             -> (x[y] = z) = vector_slice_put(x,z,y)
reduction (x:cvec, y:vec, z:cscalar)          -> (x[y] = z) = vector_slice_put_const(x,z,y)
reduction (x:cvec, z:cvec)                    -> (x[:] = z) = vector_slice_put(x,z,0..numel(x)-1)
reduction (x:cvec, z:cscalar)                 -> (x[:] = z) = vector_slice_put_const(x,z,0..numel(x)-1)

reduction (x:vec[??], y:vec)                  -> x[y] = vector_slice_get(x, y)
reduction (x:vec[??], y:vec, z:vec[??])       -> (x[y] = z) = vector_slice_put(x,z,y)
reduction (x:vec[??], y:vec, z)               -> (x[y] = z) = vector_slice_put_const(x,z,y)
reduction (x:vec[??], z:vec[??])              -> (x[:] = z) = vector_slice_put(x,z,0..numel(x)-1)
reduction (x:vec[??], z)                      -> (x[:] = z) = vector_slice_put_const(x,z,0..numel(x)-1)

% Matrix slice accessors
function y = matrix_slice_get[T](x : mat[T]'unchecked, ind1 : vec'unchecked, ind2 : vec'unchecked)
    y = uninit[T](numel(ind1),numel(ind2))
    for m=0..numel(ind1)-1
        for n=0..numel(ind2)-1
            y[m,n] = x[ind1[m],ind2[n]]
        end
    end
end

function z = matrix_slice_put[T](y : mat[T]'unchecked, x : mat[T]'unchecked, ind1 : vec'unchecked, ind2 : vec'unchecked)
    x = reshape(x, [numel(ind1), numel(ind2)])
    %assert(size(x,0..1)==[numel(ind1),numel(ind2)], "Dimensions do not match for assignment!")
    for m=0..numel(ind1)-1
        for n=0..numel(ind2)-1
            y[ind1[m],ind2[n]] = x[m,n]
        end
    end
    z = y
end

function z = matrix_slice_put_const[T](y : mat[T]'unchecked, x : T, ind1 : vec'unchecked, ind2 : vec'unchecked)
    for m=0..numel(ind1)-1
        for n=0..numel(ind2)-1
            y[ind1[m],ind2[n]] = x
        end
    end
    z = y
end

reduction (x:mat, p1:vec, p2:vec)            -> x[p1,p2]        = matrix_slice_get(x,p1,p2)
reduction (x:mat, p1:vec, p2:vec, z:mat)     -> (x[p1,p2] = z)  = matrix_slice_put(x,z,p1,p2)
reduction (x:mat, p1:vec, p2:vec, z:scalar)  -> (x[p1,p2] = z)  = matrix_slice_put_const(x,z,p1,p2)
reduction (x:mat, p2:vec)                    -> x[:,p2]         = matrix_slice_get(x,0..size(x,0)-1,p2)
reduction (x:mat, p1:vec)                    -> x[p1,:]         = matrix_slice_get(x,p1,0..size(x,1)-1)
reduction (x:mat, p2:vec, z:mat)             -> (x[:,p2] = z)   = matrix_slice_put(x,z,0..size(x,0)-1,p2)
reduction (x:mat, p1:vec, z:mat)             -> (x[p1,:] = z)   = matrix_slice_put(x,z,p1,0..size(x,1)-1)
reduction (x:mat, p2:vec, z:scalar)          -> (x[:,p2] = z)   = matrix_slice_put_const(x,z,0..size(x,0)-1,p2)
reduction (x:mat, p1:vec, z:scalar)          -> (x[p1,:] = z)   = matrix_slice_put_const(x,z,p1,0..size(x,1)-1)
reduction (x:mat, z:mat)                     -> (x[:,:] = z)    = matrix_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1)
reduction (x:mat, z:scalar)                  -> (x[:,:] = z)    = matrix_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1)

reduction (x:cmat, p1:vec, p2:vec)            -> x[p1,p2]       = matrix_slice_get(x,p1,p2)
reduction (x:cmat, p1:vec, p2:vec, z:cmat)    -> (x[p1,p2] = z) = matrix_slice_put(x,z,p1,p2)
reduction (x:cmat, p1:vec, p2:vec, z:cscalar) -> (x[p1,p2] = z) = matrix_slice_put_const(x,z,p1,p2)
reduction (x:cmat, p2:vec)                    -> x[:,p2]        = matrix_slice_get(x,0..size(x,0)-1,p2)
reduction (x:cmat, p1:vec)                    -> x[p1,:]        = matrix_slice_get(x,p1,0..size(x,1)-1)
reduction (x:cmat, p2:vec, z:cmat)            -> (x[:,p2] = z)  = matrix_slice_put(x,z,0..size(x,0)-1,p2)
reduction (x:cmat, p1:vec, z:cmat)            -> (x[p1,:] = z)  = matrix_slice_put(x,z,p1,0..size(x,1)-1)
reduction (x:cmat, p2:vec, z:cscalar)         -> (x[:,p2] = z)  = matrix_slice_put_const(x,z,0..size(x,0)-1,p2)
reduction (x:cmat, p1:vec, z:cscalar)         -> (x[p1,:] = z)  = matrix_slice_put_const(x,z,p1,0..size(x,1)-1)
reduction (x:cmat, z:cmat)                    -> (x[:,:] = z)   = matrix_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1)
reduction (x:cmat, z:cscalar)                 -> (x[:,:] = z)   = matrix_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1)

reduction (x:mat[??], p1:vec, p2:vec)        -> x[p1,p2]        = matrix_slice_get(x,p1,p2)
reduction (x:mat[??], p1:vec, p2:vec, z:mat[??]) -> (x[p1,p2] = z) = matrix_slice_put(x,z,p1,p2)
reduction (x:mat[??], p1:vec, p2:vec, z)     -> (x[p1,p2] = z) = matrix_slice_put_const(x,z,p1,p2)
reduction (x:mat[??], p2:vec)                -> x[:,p2]         = matrix_slice_get(x,0..size(x,0)-1,p2)
reduction (x:mat[??], p1:vec)                -> x[p1,:]         = matrix_slice_get(x,p1,0..size(x,1)-1)
reduction (x:mat[??], p2:vec, z:mat[??])     -> (x[:,p2] = z)   = matrix_slice_put(x,z,0..size(x,0)-1,p2)
reduction (x:mat[??], p1:vec, z:mat[??])     -> (x[p1,:] = z)   = matrix_slice_put(x,z,p1,0..size(x,1)-1)
reduction (x:mat[??], p2:vec, z)             -> (x[:,p2] = z)   = matrix_slice_put_const(x,z,0..size(x,0)-1,p2)
reduction (x:mat[??], p1:vec, z)             -> (x[p1,:] = z)   = matrix_slice_put_const(x,z,p1,0..size(x,1)-1)
reduction (x:mat[??], z:mat[??])             -> (x[:,:] = z)    = matrix_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1)
reduction (x:mat[??], z)                     -> (x[:,:] = z)    = matrix_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1)

% Cube slice accessors
function y = cube_slice_get[T](x : cube[T]'unchecked, ind1 : vec'unchecked, ind2 : vec'unchecked, ind3 : vec'unchecked)
    y = uninit[T](numel(ind1),numel(ind2),numel(ind3))
    for m=0..numel(ind1)-1
        for n=0..numel(ind2)-1
            for k=0..numel(ind3)-1
                y[m,n,k] = x[ind1[m],ind2[n],ind3[k]]
            end
        end
    end
end

function z = cube_slice_put[T](y : cube[T]'unchecked, x : cube[T]'unchecked, ind1 : vec'unchecked, ind2 : vec'unchecked, ind3 : vec'unchecked)
    x = reshape(x, [numel(ind1), numel(ind2), numel(ind3)])
    for m=0..numel(ind1)-1
        for n=0..numel(ind2)-1
            for k=0..numel(ind3)-1
                y[ind1[m],ind2[n],ind3[k]] = x[m,n,k]
            end
        end
    end
    z = y
end

function z = cube_slice_put_const[T](y : cube[T]'unchecked, x : T, ind1 : vec'unchecked, ind2 : vec'unchecked, ind3 : vec'unchecked)
    for m=0..numel(ind1)-1
        for n=0..numel(ind2)-1
            for k=0..numel(ind3)-1
                y[ind1[m],ind2[n],ind3[k]] = x
            end
        end
    end
    z = y
end

reduction (x:cube, p1:vec, p2:vec, p3:vec)           -> x[p1,p2,p3] = cube_slice_get(x,p1,p2,p3)
reduction (x:cube, p1:vec, p2:vec, p3:vec, z:cube)   -> (x[p1,p2,p3] = z) = cube_slice_put(x,z,p1,p2,p3)
reduction (x:cube, p1:vec, p2:vec, p3:vec, z:scalar) -> (x[p1,p2,p3] = z) = cube_slice_put_const(x,z,p1,p2,p3)
reduction (x:cube,         p2:vec, p3:vec)           -> x[:,p2,p3]  = cube_slice_get(x,0..size(x,0)-1,p2,p3)
reduction (x:cube, p1:vec,         p3:vec)           -> x[p1,:,p3]  = cube_slice_get(x,p1,0..size(x,1)-1,p3)
reduction (x:cube, p1:vec, p2:vec        )           -> x[p1,p2,:]  = cube_slice_get(x,p1,p2,0..size(x,2)-1)
reduction (x:cube,                 p3:vec)           -> x[:,:,p3]   = cube_slice_get(x,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:cube, p1:vec                )           -> x[p1,:,:]   = cube_slice_get(x,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube,         p2:vec        )           -> x[:,p2,:]   = cube_slice_get(x,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:cube,         p2:vec, p3:vec, z:cube)   -> (x[:,p2,p3] = z) = cube_slice_put(x,z,0..size(x,0)-1,p2,p3)
reduction (x:cube, p1:vec,         p3:vec, z:cube)   -> (x[p1,:,p3] = z) = cube_slice_put(x,z,p1,0..size(x,1)-1,p3)
reduction (x:cube, p1:vec, p2:vec,         z:cube)   -> (x[p1,p2,:] = z) = cube_slice_put(x,z,p1,p2,0..size(x,2)-1)
reduction (x:cube,                 p3:vec, z:cube)   -> (x[:,:,p3]  = z) = cube_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:cube, p1:vec,                 z:cube)   -> (x[p1,:,:]  = z) = cube_slice_put(x,z,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube,         p2:vec,         z:cube)   -> (x[:,p2,:]  = z) = cube_slice_put(x,z,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:cube,         p2:vec, p3:vec, z:scalar) -> (x[:,p2,p3] = z) = cube_slice_put_const(x,z,0..size(x,0)-1,p2,p3)
reduction (x:cube, p1:vec,         p3:vec, z:scalar) -> (x[p1,:,p3] = z) = cube_slice_put_const(x,z,p1,0..size(x,1)-1,p3)
reduction (x:cube, p1:vec, p2:vec,         z:scalar) -> (x[p1,p2,:] = z) = cube_slice_put_const(x,z,p1,p2,0..size(x,2)-1)
reduction (x:cube,                 p3:vec, z:scalar) -> (x[:,:,p3]  = z) = cube_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:cube, p1:vec,                 z:scalar) -> (x[p1,:,:]  = z) = cube_slice_put_const(x,z,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube,         p2:vec,         z:scalar) -> (x[:,p2,:]  = z) = cube_slice_put_const(x,z,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:cube, z:cube)   -> (x[:,:,:] = z)    = cube_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube, z:scalar) -> (x[:,:,:] = z)    = cube_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1,0..size(x,2)-1)

reduction (x:ccube, p1:vec, p2:vec, p3:vec)            -> x[p1,p2,p3] = cube_slice_get(x,p1,p2,p3)
reduction (x:ccube, p1:vec, p2:vec, p3:vec, z:ccube)   -> (x[p1,p2,p3] = z) = cube_slice_put(x,z,p1,p2,p3)
reduction (x:ccube, p1:vec, p2:vec, p3:vec, z:cscalar) -> (x[p1,p2,p3] = z) = cube_slice_put_const(x,z,p1,p2,p3)
reduction (x:ccube,         p2:vec, p3:vec)            -> x[:,p2,p3]  = cube_slice_get(x,0..size(x,0)-1,p2,p3)
reduction (x:ccube, p1:vec,         p3:vec)            -> x[p1,:,p3]  = cube_slice_get(x,p1,0..size(x,1)-1,p3)
reduction (x:ccube, p1:vec, p2:vec        )            -> x[p1,p2,:]  = cube_slice_get(x,p1,p2,0..size(x,2)-1)
reduction (x:ccube,                 p3:vec)            -> x[:,:,p3]   = cube_slice_get(x,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:ccube, p1:vec                )            -> x[p1,:,:]   = cube_slice_get(x,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:ccube,         p2:vec        )            -> x[:,p2,:]   = cube_slice_get(x,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:ccube,         p2:vec, p3:vec, z:ccube)   -> (x[:,p2,p3] = z) = cube_slice_put(x,z,0..size(x,0)-1,p2,p3)
reduction (x:ccube, p1:vec,         p3:vec, z:ccube)   -> (x[p1,:,p3] = z) = cube_slice_put(x,z,p1,0..size(x,1)-1,p3)
reduction (x:ccube, p1:vec, p2:vec,         z:ccube)   -> (x[p1,p2,:] = z) = cube_slice_put(x,z,p1,p2,0..size(x,2)-1)
reduction (x:ccube,                 p3:vec, z:ccube)   -> (x[:,:,p3]  = z) = cube_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:ccube, p1:vec,                 z:ccube)   -> (x[p1,:,:]  = z) = cube_slice_put(x,z,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:ccube,         p2:vec,         z:ccube)   -> (x[:,p2,:]  = z) = cube_slice_put(x,z,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:ccube,         p2:vec, p3:vec, z:cscalar) -> (x[:,p2,p3] = z) = cube_slice_put_const(x,z,0..size(x,0)-1,p2,p3)
reduction (x:ccube, p1:vec,         p3:vec, z:cscalar) -> (x[p1,:,p3] = z) = cube_slice_put_const(x,z,p1,0..size(x,1)-1,p3)
reduction (x:ccube, p1:vec, p2:vec,         z:cscalar) -> (x[p1,p2,:] = z) = cube_slice_put_const(x,z,p1,p2,0..size(x,2)-1)
reduction (x:ccube,                 p3:vec, z:cscalar) -> (x[:,:,p3]  = z) = cube_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:ccube, p1:vec,                 z:cscalar) -> (x[p1,:,:]  = z) = cube_slice_put_const(x,z,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:ccube,         p2:vec,         z:cscalar) -> (x[:,p2,:]  = z) = cube_slice_put_const(x,z,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:ccube, z:ccube)   -> (x[:,:,:] = z)    = cube_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:ccube, z:cscalar) -> (x[:,:,:] = z)    = cube_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1,0..size(x,2)-1)

reduction (x:cube[??], p1:vec, p2:vec, p3:vec)       -> x[p1,p2,p3] = cube_slice_get(x,p1,p2,p3)
reduction (x:cube[??], p1:vec, p2:vec, p3:vec, z:cube[??]) -> (x[p1,p2,p3] = z) = cube_slice_put(x,z,p1,p2,p3)
reduction (x:cube[??], p1:vec, p2:vec, p3:vec, z)    -> (x[p1,p2,p3] = z) = cube_slice_put_const(x,z,p1,p2,p3)
reduction (x:cube[??],         p2:vec, p3:vec)       -> x[:,p2,p3]  = cube_slice_get(x,0..size(x,0)-1,p2,p3)
reduction (x:cube[??], p1:vec,         p3:vec)       -> x[p1,:,p3]  = cube_slice_get(x,p1,0..size(x,1)-1,p3)
reduction (x:cube[??], p1:vec, p2:vec        )       -> x[p1,p2,:]  = cube_slice_get(x,p1,p2,0..size(x,2)-1)
reduction (x:cube[??],                 p3:vec)       -> x[:,:,p3]   = cube_slice_get(x,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:cube[??], p1:vec                )       -> x[p1,:,:]   = cube_slice_get(x,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube[??],         p2:vec        )       -> x[:,p2,:]   = cube_slice_get(x,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:cube[??],         p2:vec, p3:vec, z:cube[??])   -> (x[:,p2,p3] = z) = cube_slice_put(x,z,0..size(x,0)-1,p2,p3)
reduction (x:cube[??], p1:vec,         p3:vec, z:cube[??])   -> (x[p1,:,p3] = z) = cube_slice_put(x,z,p1,0..size(x,1)-1,p3)
reduction (x:cube[??], p1:vec, p2:vec,         z:cube[??])   -> (x[p1,p2,:] = z) = cube_slice_put(x,z,p1,p2,0..size(x,2)-1)
reduction (x:cube[??],                 p3:vec, z:cube[??])   -> (x[:,:,p3]  = z) = cube_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:cube[??], p1:vec,                 z:cube[??])   -> (x[p1,:,:]  = z) = cube_slice_put(x,z,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube[??],         p2:vec,         z:cube[??])   -> (x[:,p2,:]  = z) = cube_slice_put(x,z,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:cube[??],         p2:vec, p3:vec, z) -> (x[:,p2,p3] = z) = cube_slice_put_const(x,z,0..size(x,0)-1,p2,p3)
reduction (x:cube[??], p1:vec,         p3:vec, z) -> (x[p1,:,p3] = z) = cube_slice_put_const(x,z,p1,0..size(x,1)-1,p3)
reduction (x:cube[??], p1:vec, p2:vec,         z) -> (x[p1,p2,:] = z) = cube_slice_put_const(x,z,p1,p2,0..size(x,2)-1)
reduction (x:cube[??],                 p3:vec, z) -> (x[:,:,p3]  = z) = cube_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1,p3)
reduction (x:cube[??], p1:vec,                 z) -> (x[p1,:,:]  = z) = cube_slice_put_const(x,z,p1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube[??],         p2:vec,         z) -> (x[:,p2,:]  = z) = cube_slice_put_const(x,z,0..size(x,0)-1,p2,0..size(x,2)-1)
reduction (x:cube[??], z:cube[??])   -> (x[:,:,:] = z)    = cube_slice_put(x,z,0..size(x,0)-1,0..size(x,1)-1,0..size(x,2)-1)
reduction (x:cube[??], z)            -> (x[:,:,:] = z)    = cube_slice_put_const(x,z,0..size(x,0)-1,0..size(x,1)-1,0..size(x,2)-1)

function z = ncube_slice_put_scalar[T](y : cube{:}[T]'unchecked, x : T, ind : vec[vec'unchecked])
    dims = uninit(numel(ind))
    n = numel(dims)
    {!interpreted for}
    for i=0..n-1
        dims[i] = numel(ind[i])
    end

    function [] = __kernel__ kernel(y : vec[T]'unchecked, x : T, x_dims : vec, y_dims : vec, ind : vec[vec'unchecked], pos : int)
        y_idx = 0
        y_mul = 1
        p = pos
        for m=numel(x_dims)-1..-1..0
            u = mod(p,int(x_dims[m]))
            p = int(p/x_dims[m])
            y_idx += ind[m][u] * y_mul
            y_mul *= y_dims[m]
        end
        y[y_idx] = x
    end

    parallel_do(prod(dims),y,x,dims,size(y,0..n-1),ind,kernel)
    z = y
end

function z = ncube_slice_put[T](y : cube{:}[T]'unchecked, x : cube{:}[T]'unchecked, ind : vec[vec'unchecked])
    dims = uninit(numel(ind))
    n = numel(dims)
    {!interpreted for}
    for i=0..n-1
        dims[i] = numel(ind[i])
    end

    if ndims(x) < n
        x_dims = uninit(n)
        j = 0
        {!interpreted for}
        for i=0..n-1
            if dims[i]==1
                x_dims[i] = 1
            else
                x_dims[i] = size(x,j)
                j += 1
            endif
        end
    else
        x_dims = size(x,0..n-1)
    endif

    function [] = __kernel__ kernel(y : vec[T]'unchecked, x : vec[T]'unchecked, ind_dims : vec, x_dims : vec, y_dims : vec, ind : vec[vec'unchecked], pos : int)
        y_idx = 0; y_mul = 1
        x_idx = 0; x_mul = 1
        p = pos
        for m=numel(ind_dims)-1..-1..0
            u = mod(p,int(ind_dims[m]))
            p = int(p/ind_dims[m])
            y_idx += ind[m][u] * y_mul
            y_mul *= y_dims[m]
            x_idx += u * x_mul
            x_mul *= x_dims[m]
        end
        y[y_idx] = x[x_idx]
    end

    parallel_do(prod(dims),y,x,dims,x_dims,size(y,0..n-1),ind,kernel)
    z = y
end

function z = ncube_slice_get[T](y : cube{:}[T]'unchecked, ind : vec[vec'unchecked])
    dims = uninit(numel(ind))
    n = numel(dims)
    {!interpreted for}
    for i=0..n-1
        dims[i] = numel(ind[i])
    end
    z = uninit(dims)

    function [] = __kernel__ kernel(y : vec[T]'unchecked, z : vec[T]'unchecked, ind_dims : vec, y_dims : vec, ind : vec[vec'unchecked], pos : int)
        y_idx = 0; y_mul = 1
        p = pos
        for m=numel(ind_dims)-1..-1..0
            u = mod(p,int(ind_dims[m]))
            p = int(p/ind_dims[m])
            y_idx += ind[m][u] * y_mul
            y_mul *= y_dims[m]
        end
        z[pos] = y[y_idx]
    end

    parallel_do(prod(dims),y,z,dims,size(y,0..n-1),ind,kernel)
end

reduction (x : cube{:}, ind : vec[vec], y : scalar)  -> (x[ind] = y) = ncube_slice_put_scalar(x, y, ind)
reduction (x : cube{:}, ind : vec[vec], y : cube{:}) -> (x[ind] = y) = ncube_slice_put(x, y, ind)
reduction (x : cube{:}, ind : vec[vec])              -> (x[ind])     = ncube_slice_get(x, ind)

reduction (x : ccube{:}, ind : vec[vec], y : cscalar)  -> (x[ind] = y) = ncube_slice_put_scalar(x, y, ind)
reduction (x : ccube{:}, ind : vec[vec], y : ccube{:}) -> (x[ind] = y) = ncube_slice_put(x, y, ind)
reduction (x : ccube{:}, ind : vec[vec])               -> (x[ind])     = ncube_slice_get(x, ind)

reduction (x : cube{:}[??], ind : vec[vec], y)               -> (x[ind] = y) = ncube_slice_put_scalar(x, y, ind)
reduction (x : cube{:}[??], ind : vec[vec], y : cube{:}[??]) -> (x[ind] = y) = ncube_slice_put(x, y, ind)
reduction (x : cube{:}[??], ind : vec[vec])                  -> (x[ind])     = ncube_slice_get(x, ind)

reduction (a : scalar, b : scalar)  -> a..b = r_seq(a, b)
reduction (a : scalar, b : scalar, c : scalar)  -> a..b..c = r_seq2(a, c, b)
reduction (a : scalar, b : scalar, cnt : int)  -> linspace(a,b,cnt) = r_linspace(a,b,cnt)

%==============================================================================
% Random generators (NOTE: requires dynamic kernel memory!!!)
%==============================================================================

type mt_rand_data : mutable class
    matrix_a : uint32
    mask_b : uint32
    mask_c : uint32
    seed : uint32
end

RAND_THREADS = 4096
rand_table = vec[mt_rand_data](RAND_THREADS)

% initialization
function [] = mtrand_init()
    file_name = "MersenneTwister.dat"  % TODO - need to embed as a resource
    f = fopen(file_name, "rb")
    data : vec[uint32]'unchecked = fread(f,RAND_THREADS*4,"uint32")
    fclose(f)
    #pragma force_parallel
    for i = 0..RAND_THREADS-1
        %rand_table[i] = mt_rand_data(seed:=mod((i+1) * 929256287, 1396212421), matrix_a:=034518974h, mask_b:=04E55F2E6h, mask_c:=02524562h)
        rand_table[i] = mt_rand_data(matrix_a:=data[4*i],mask_b:=data[4*i+1],
                                     mask_c:=data[4*i+2],seed:=mod(i * 929256287, 1396212421))
    end
end

% Function: mt_randseed
%
% Initialize the random generator using the specified seed function
function [] = mtrand_seed(seed : scalar)
    iseed = int(seed)

    #pragma force_parallel
    for i = 0..RAND_THREADS-1
        rand_table[i].seed = mod((i + iseed) * 929256287, 1396212421)
    end
end

% Function: mtrand_genN
%
% Implements the Mersenne-Twister algorithm for drawing a series of random values.
% The function generates N random values and writes it to the destination vector
% at the specified offset and with the specified stride.
%
function [] = __device__ mtrand_genN(ds_MT : vec[mt_rand_data]'unchecked, outvals : vec'unchecked, offset : int, stride : int, N : int) concealed

    % generator constants
    MT_MM = 9
    MT_NN = 19
    MT_WMASK    : uint32 = 0FFFFFFFFh
    MT_UMASK    : uint32 = 0FFFFFFFEh
    MT_LMASK    : uint32 = 1h
    MT_SHIFT0 = 12
    MT_SHIFTB = 7
    MT_SHIFTC = 15
    MT_SHIFT1 = 18

    % dynamic memory version (not supported by OpenCL back-ends)
    %mt = vec[uint32](MT_NN)
    % static memory version
    mt = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]

    matrix_a = ds_MT[offset].matrix_a
    mask_b = ds_MT[offset].mask_b
    mask_c = ds_MT[offset].mask_c
    mt[0] = ds_MT[offset].seed

    for state = 1..MT_NN-1
        tmp = xor(cast(mt[state - 1], uint32), shr(cast(mt[state-1], uint32), cast(30, uint32)))
        mt[state] = and(cast(1812433253, uint32) * tmp + cast(state, uint32), MT_WMASK)
    end

    state = 0
    state1 = 0
    stateM = 0
    mti1 = cast(mt[0], uint32)

    for n = 0..N-1
        state1 += 1
        stateM += MT_MM
        if state1 >= MT_NN
            state1 -= MT_NN
        endif
        if stateM >= MT_NN
            stateM -= MT_NN
        endif
        mti  : uint32 = mti1
        mti1 : uint32 = cast(mt[state1],uint32)
        mtiM : uint32 = cast(mt[stateM],uint32)

        % MT recurrence
        x : uint32 = or(and(mti, MT_UMASK), and(mti1, MT_LMASK))
        x = xor(xor(mtiM, shr(x, 1)), and(x, 1) > 0 ? matrix_a : 0)

        mt[cast(state,int)] = x
        state = state1

        % Tempering transformation
        x = xor(x, shr(x, MT_SHIFT0))
        x = xor(x, and(shl(x, MT_SHIFTB), mask_b))
        x = xor(x, and(shl(x, MT_SHIFTC), mask_c))
        x = xor(x, shr(x, MT_SHIFT1))

        % Convert to (0, 1] scalar and write to global memory
        if offset + stride * n < numel(outvals)
            outvals[offset + stride * n] = (float(and(x,07FFFFFFFh)) / (080000000h))
        endif
    end

    ds_MT[offset].seed = mt[state]
end

% Function: rand
%
% Creates a uniformly distributed data vector of length a (values are in the range [0,1[)
%
% :function y = rand(a)
%
% Parameters:
% a - either a scalar value (creating a vector), or a vector of maximally length 3
%     (to create a matrix or cube)
%
function y = __rand(dims)
    function [] = __kernel__ kernel (rand_table : vec[mt_rand_data], y : vec, pos : int)
        tid = mod(pos, RAND_THREADS)
        mtrand_genN(rand_table, y, tid, RAND_THREADS, int((numel(y) - tid + RAND_THREADS - 1)/RAND_THREADS))
    end

    y = uninit(dims)
    parallel_do(min(numel(y), RAND_THREADS), rand_table, y, kernel)
end

% Function: box_muller
%
% Box Muller transform: generates two Gaussian random variables
% based on two uniform random variables
function y = __device__ box_muller(x : vec2)
    r   = x[0] <= 0 ? 0.0 : sqrt(-2 * log(x[0]))
    phi = 2*pi*x[1]
    y = r * [cos(phi), sin(phi)]
end

% Function: randn
%
% Creates a Gaussian distributed data vector of length a (values are in the range [0,1[)
%
% :function y = randn(a)
%
% Parameters:
% a - either a scalar value (creating a vector), or a vector of maximally length 3
%     (to create a matrix or cube)
%
function y = __randn(dims)
    function [] = __kernel__ kernel (rand_table : vec[mt_rand_data]'unchecked, y : vec'unchecked, pos : int)
        tid = mod(pos, RAND_THREADS)
        n = int((numel(y) - tid + RAND_THREADS - 1)/RAND_THREADS)
        mtrand_genN(rand_table, y, tid, RAND_THREADS, n)
    end

    function [] = __kernel__ apply_box_muller (y : vec, pos : int)
        k = pos * 2
        [y[k], y[k+1]] = box_muller([y[k], y[k+1]])
    end

    y = uninit(dims)
    parallel_do(numel(y)>RAND_THREADS ? RAND_THREADS : numel(y), rand_table, y, kernel)
    parallel_do(int((numel(y)+1)/2), y, apply_box_muller)
end

% rand and randn friends
% TODO: replace by variadic reductions
reduction (x : scalar) -> rand(x) = __rand([1,x])
reduction (x : scalar,y : scalar) -> rand(x,y) = __rand([x,y])
reduction (x : scalar,y : scalar,z : scalar) -> rand(x,y,z) = __rand([x,y,z])
reduction (x : scalar,y : scalar,z : scalar,w : scalar) -> rand(x,y,z,w) = __rand([x,y,z,w])
reduction (x : scalar,y : scalar,z : scalar,w : scalar, u : scalar) -> rand(x,y,z,w,u) = __rand([x,y,z,w,u])
reduction (x : vec) -> rand(x) = __rand(x)

reduction (x : scalar) -> randn(x) = __randn([1,x])
reduction (x : scalar,y : scalar) -> randn(x,y) = __randn([x,y])
reduction (x : scalar,y : scalar,z : scalar) -> randn(x,y,z) = __randn([x,y,z])
reduction (x : scalar,y : scalar,z : scalar,w : scalar) -> randn(x,y,z,w) = __randn([x,y,z,w])
reduction (x : scalar,y : scalar,z : scalar,w : scalar, u : scalar) -> randn(x,y,z,w,u) = __randn([x,y,z,w,u])
reduction (x : vec) -> randn(x) = __randn(x)

reduction () -> randseed() = mtrand_seed(0)
reduction (x : scalar) -> randseed(x) = mtrand_seed(x)

mtrand_init()  % Initialize the random generator!

%==============================================================================
% Fast Fourier Transforms
%==============================================================================
#pragma expression_optimizer(off)

function [] = __kernel__ butterflyN_dim2(x : ccube'unchecked,y : ccube'unchecked,omega : cmat'unchecked,i : int,ux : int,N : int,pos:ivec3)
    Nq = int(size(x,2)/N)
    p = mod(pos[2],ux) + int(pos[2]/(N*ux))*ux
    t = x[pos[0],pos[1],p+Nq*(N-1)]
    w = omega[i,pos[2]]
    for m = N-2..-1..0
        t = x[pos[0],pos[1],p+m*Nq] + t * w
    end
    y[pos] = t
end


function omega = __omega_mtx(N : int, q : vec, dir : int)
    omega = cuninit(numel(q), N)
    cq = uninit(numel(q)) % cumulative product
    cq[0] = q[0]
    for m=1..numel(q)-1
        cq[m] = cq[m-1] * q[m]
    end

    for m=0..numel(q)-1
        for n=0..N-1
            omega[m,n] = exp(2j*pi*n*dir/cq[m])
        end
    end
end

% Calculates the integer factorization of the number of dimensions
function y = factor(n : int)
    z = uninit(round(log2(n))) % upper bound
    i = 0
    {!interpreted for}
    while mod(n, 8) == 0 % Factor RADIX 8 terms
        n = int(n / (z[i] = 8))
        i += 1
    end
    {!interpreted for}
    while mod(n, 4) == 0 % Factor RADIX 4 terms
        n = int(n / (z[i] = 4))
        i += 1
    end
    if n>1
        {!interpreted for}
        for p = 2..n
            while mod(n, p) == 0
                n = int(n / (z[i] = p))
                i += 1
            end
            if n==1
                break
            endif
        end
    endif
    y = z[0.0..i-1.0]
end

function [y : ccube] = __fft3(x : ccube, dir, dims)

    scale = 1
    y = x

    if and(dims,1) % first dimension
        if ndims(y) == 1 && size(y,0) == 1
            y = shuffledims(y, [0,2,1])
        else
            y = shuffledims(y, [1,2,0])
        endif

        N = size(y,2)
        if N>1
            q = factor(N)
            omega = __omega_mtx(N,q,dir)

            z = cuninit(size(y))
            ux = 1
            {!interpreted for}
            for i=0..numel(q)-1
                match q[i] with
                | 2 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==2))
                | 3 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==3))
                | 4 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==4))
                | 5 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==5))
                | 8 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==8))
                | _ -> parallel_do(size(y),y,z,omega,i,ux,q[i],butterflyN_dim2)
                end
                %[z,y]={y,z} % switch the buffers
				tmp=y; y=z; z=tmp  % TODO: Fix for OpenCL does not support structs with pointers
                ux *= q[i]
            end
        endif

        if and(dims,2)==0 % trick to save a shuffle dims
            y = shuffledims(y,[2,0,1])
        endif
        scale *= N
    endif

    if and(dims,2) % second dimension
        y = shuffledims(y, and(dims,1) ? [2,1,0] : [0,2,1])

        N = size(y,2)
        if N>1
            q = factor(N)
            omega = __omega_mtx(N,q,dir)

            z = cuninit(size(y))
            ux = 1
            {!interpreted for}
            for i=0..numel(q)-1
                match q[i] with
                | 2 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==2))
                | 3 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==3))
                | 4 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==4))
                | 5 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==5))
                | 8 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==8))
                | _ -> parallel_do(size(y),y,z,omega,i,ux,q[i],butterflyN_dim2)
                end
                %[z,y]={y,z} % switch the buffers
                tmp=y; y=z; z=tmp  % TODO: Fix for OpenCL does not support structs with pointers
				ux *= q[i]
            end
        endif
        y = shuffledims(y, [0,2,1])
        scale *= N
    endif

    if and(dims,4) % third dimension
        N = size(y,2)
        if N > 1
            q = factor(N)
            omega = __omega_mtx(N,q,dir)

            z = cuninit(size(y))
            ux = 1
            {!interpreted for}
            for i=0..numel(q)-1
                match q[i] with
                | 2 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==2))
                | 3 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==3))
                | 4 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==4))
                | 5 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==5))
                | 8 -> parallel_do(size(y),y,z,omega,i,ux,$specialize(butterflyN_dim2, N==8))
                | _ -> parallel_do(size(y),y,z,omega,i,ux,q[i],butterflyN_dim2)
                end
                %[z,y]={y,z} % switch the buffers
                tmp=y; y=z; z=tmp  % TODO: Fix for OpenCL does not support structs with pointers
				ux *= q[i]
            end
            scale *= N
        endif
    endif

    y = y / sqrt(scale) % normalization
end

reduction (x : ccube) ->  $fft1(x) = __fft3(x,  1, size(x,2)==1 ? 1 : 2)
reduction (x : ccube) -> $ifft1(x) = __fft3(x, -1, size(x,2)==1 ? 1 : 2)
reduction (x : ccube) ->  $fft2(x) = __fft3(x,  1, 3)
reduction (x : ccube) -> $ifft2(x) = __fft3(x, -1, 3)
reduction (x : ccube) ->  $fft3(x) = __fft3(x,  1, 7)
reduction (x : ccube) -> $ifft3(x) = __fft3(x, -1, 7)
reduction (x : ccube) -> $irealfft1(x) = real(ifft1(x))
reduction (x : ccube) -> $irealfft2(x) = real(ifft2(x))
reduction (x : ccube) -> $irealfft3(x) = real(ifft3(x))

%
% FFTW / CuFFT helper functions
%
function [] = __fft1d_conjsymmetry(dst, src, w, h)
    ncoeff = int(w / 2) + 1 % number of non-redundant complex coefficients
    function [] = __kernel__ kernel(dst : cvec, src : cvec, w : int, h : int, ncoeff : int, pos : ivec2)
        [m,n] = pos
        dst_idx = m*w+n
        if n<ncoeff
            dst[dst_idx] = src[ncoeff * m + n]
        else % apply the conjugate symmetry relationship
            dst[dst_idx] = conj(src[ncoeff * m + w-n])
        endif
    end
    parallel_do(size(dst), dst, src, w, h, ncoeff, kernel)
end
reduction (dst, src, w, h) -> fft1d_conjsymmetry(dst, src, w, h) = __fft1d_conjsymmetry(dst, src, w, h)

function [] = __ifft1d_conjsymmetry(dst, src, w, h)
    ncoeff = int(w / 2) + 1 % number of non-redundant complex coefficients
    function [] = __kernel__ kernel(dst : cvec, src : cvec, w : int, h : int, ncoeff : int, pos : ivec2)
        [m,n] = pos
        src_idx = m*w+n
        if n<ncoeff
            dst[ncoeff * m + n] = src[src_idx]
        endif
        if n == ncoeff-2
            dst[ncoeff*m + ncoeff-1] = src[src_idx+1]
        endif
    end
    parallel_do([size(dst,0), int(size(dst,1)/2)], dst, src, w, h, ncoeff, kernel)
end
reduction (dst, src, w, h) -> ifft1d_conjsymmetry(dst, src, w, h) = __ifft1d_conjsymmetry(dst, src, w, h)


function [] = __fft2d_conjsymmetry(dst, src, w, h)
    ncoeff = int(w / 2) + 1 % number of non-redundant complex coefficients
    function [] = __kernel__ kernel(dst : cvec, src : cvec, w : int, h : int, ncoeff : int, pos : ivec2)
        [m,n] = pos
        dst_idx = m*w+n
        if n<ncoeff
            dst[dst_idx] = src[ncoeff * m + n]
        else % apply the conjugate symmetry relationship
            n = w-n
            if m>0
                m = h-m
            endif
            dst[dst_idx] = conj(src[ncoeff * m + n])
        endif
    end
    parallel_do(size(dst), dst, src, w, h, ncoeff, kernel)
end
reduction (dst, src, w, h) -> fft2d_conjsymmetry(dst, src, w, h) = __fft2d_conjsymmetry(dst, src, w, h)

function [] = __ifft2d_conjsymmetry(dst, src, w, h)
    ncoeff = int(w / 2) + 1 % number of non-redundant complex coefficients
    function [] = __kernel__ kernel(dst : cvec, src : cvec, w : int, h : int, ncoeff : int, pos : ivec2)
        [m,n] = pos
        src_idx = m*w+n
        if n<ncoeff
            dst[ncoeff * m + n] = src[src_idx]
        endif
        if n == ncoeff-2
            dst[ncoeff*m + ncoeff-1] = src[src_idx+1]
        endif
    end
    parallel_do([size(dst,0), int(size(dst,1)/2)], dst, src, w, h, ncoeff, kernel)
end
reduction (dst, src, w, h) -> ifft2d_conjsymmetry(dst, src, w, h) = __ifft2d_conjsymmetry(dst, src, w, h)

function [] = __fft2d_conjsymmetry_multislice(dst, src, nslices, w, h)
    ncoeff = int(w / 2) + 1 % number of non-redundant complex coefficients
    function [] = __kernel__ kernel(dst : cvec, src : cvec, w : int, h : int, ncoeff : int, nslices : int, pos : ivec3)
        [m,n,k] = pos
        dst_idx = nslices * (m*w+n) + k
        if n<ncoeff
            dst[dst_idx] = src[nslices * (m*ncoeff+n) + k]
        else % apply the conjugate symmetry relationship
            n = w-n
            if m>0
                m = h-m
            endif
            dst[dst_idx] = conj(src[nslices * (m*ncoeff+n) + k])
        endif
    end
    parallel_do([size(dst,0),size(dst,1),nslices], dst, src, w, h, ncoeff, nslices, kernel)
end
reduction (dst, src, nslices, w, h) -> fft2d_conjsymmetry_multislice(dst, src, nslices, w, h) = __fft2d_conjsymmetry_multislice(dst, src, nslices, w, h)



function [] = __fft3d_conjsymmetry(dst, src, dim1, dim2, dim3)
    ncoeff = int(dim3 / 2) + 1 % number of non-redundant complex coefficients
    function [] = __kernel__ kernel(dst : cvec, src : cvec, dims : ivec3, ncoeff : int, pos : ivec3)
        [m,n,k] = pos
        dst_idx = (m*dims[1]+n)*dims[2]+k
        if k<ncoeff
            dst[dst_idx] = src[(m*dims[1]+n)*ncoeff+k]
        else % apply the conjugate symmetry relationship
            k = dims[2]-k
            if m>0
                m = dims[0]-m
            endif
            if n>0
                n = dims[1]-n
            endif
            dst[dst_idx] = conj(src[(m*dims[1]+n)*ncoeff+k])
        endif
    end
    parallel_do(size(dst), dst, src, [dim1,dim2,dim3], ncoeff, kernel)
end
reduction (dst, src, dim1, dim2, dim3) -> fft3d_conjsymmetry(dst, src, dim1, dim2, dim3) = __fft3d_conjsymmetry(dst, src, dim1, dim2, dim3)


function [] = __ifft3d_conjsymmetry(dst, src, dim1, dim2, dim3)
    ncoeff = int(dim3 / 2) + 1 % number of non-redundant complex coefficients
    function [] = __kernel__ kernel(dst : cvec, src : cvec, dims : ivec3, ncoeff : int, pos : ivec3)
        [m,n,k] = pos
        src_idx = (m*dims[1]+n)*dims[2]+k
        if k<ncoeff
            dst[(m*dims[1]+n)*ncoeff+k] = src[src_idx]
        endif
        if k == ncoeff-2
            dst[(m*dims[1]+n)*ncoeff+ncoeff-1] = src[src_idx + 1]
        endif
    end
    parallel_do([size(dst,0), size(dst,1), int(size(dst,2)/2)], dst, src, [dim1,dim2,dim3], ncoeff, kernel)
end
reduction (dst, src, dim1, dim2, dim3) -> ifft3d_conjsymmetry(dst, src, dim1, dim2, dim3) = __ifft3d_conjsymmetry(dst, src, dim1, dim2, dim3)


function [] = __fft1d_conjsymmetry_batched_multislice(dst, src, w : int, nbatches : int, nslices : int)
    ncoeff = int(w / 2)+1
    stride = int(numel(dst)/(w*nslices))
    function [] = __kernel__ kernel(dst : cvec, src : cvec, pos : ivec3)
        [batch,n,k] = pos
        k += batch*stride
        dst_idx = n*nslices+k
        if n < ncoeff
            dst[dst_idx] = src[n * nslices + k]
        else
            dst[dst_idx] = conj(src[(w - n) * nslices + k])
        endif
    end
	
	parallel_do([nbatches,w,nslices],dst,src,kernel)
end

reduction (dst, src, w, nbatches, nslices) -> fft1d_conjsymmetry_batched_multislice(dst, src, w, nbatches, nslices) = __fft1d_conjsymmetry_batched_multislice(dst, src, w, nbatches, nslices)

function [] = __fft2d_conjsymmetry_batched_multislice(dst, src, h : int, w : int, nbatches : int, nslices : int)
    ncoeff = int(w / 2)+1
    stride = int(h*w*nslices)
    function [] = __kernel__ kernel(dst : cvec, src : cvec, pos : ivec4)
        [batch,m,n,k] = pos
        k += batch*stride
        dst_idx = (m * w + n)*nslices+k
        if n < ncoeff
            dst[dst_idx] = src[(m * ncoeff + n) * nslices + k]
        else
			n = w - n
			if m > 0
				m = h - m
			endif
            dst[dst_idx] = conj(src[(m * ncoeff + n) * nslices + k])
        endif
    end
	
	parallel_do([nbatches,h,w,nslices],dst,src,kernel)
end

reduction (dst, src, h, w, nbatches, nslices) -> fft2d_conjsymmetry_batched_multislice(dst, src, h, w, nbatches, nslices) = __fft2d_conjsymmetry_batched_multislice(dst, src, h, w, nbatches, nslices)

function [] = __fft3d_conjsymmetry_batched_multislice(dst, src, d : int, h : int, w : int, nbatches : int, nslices : int)
    ncoeff = int(w / 2)+1
    stride = int(d*h*w*nslices)
    function [] = __kernel__ kernel(dst : cvec, src : cvec, pos : ivec5)
        [batch,r,m,n,k] = pos
        k += batch*stride
        dst_idx = ((r * h + m) * w + n)*nslices+k
        if n < ncoeff
            dst[dst_idx] = src[((r * h + m) * ncoeff + n) * nslices + k]
        else
			n = w - n
			if m > 0
				m = h - m
			endif
			if r > 0
				r = d - r
			endif
            dst[dst_idx] = conj(src[((r * h + m) * ncoeff + n) * nslices + k])
        endif
    end
	
	parallel_do([nbatches,d,h,w,nslices],dst,src,kernel)
end

reduction (dst, src, d, h, w, nbatches, nslices) -> fft3d_conjsymmetry_batched_multislice(dst, src, d, h, w, nbatches, nslices) = __fft3d_conjsymmetry_batched_multislice(dst, src, d, h, w, nbatches, nslices)


%==============================================================================
% OpenGL/Hardware texture unit helper functions
%==============================================================================
function [] = __kernel__ __cvt_float2half(dst : vec[scalar'half]'unchecked, src : vec[scalar'single]'unchecked, pos : int)
    dst[pos] = src[pos]
end

function [] = __kernel__ __cvt_double2half(dst : vec[scalar'half]'unchecked, src : vec[scalar'double]'unchecked, pos : int)
    dst[pos] = src[pos]
end

function [] = __kernel__ __cvt_double2float(dst : vec[scalar'single]'unchecked, src : vec[scalar'double]'unchecked, pos : int)
    dst[pos] = src[pos]
end

%==============================================================================
% Int-float conversion functions
%==============================================================================
function [y : cube{:}] = __float(x : cube{:}[int])
    function [] = __kernel__ kernel(x : vec[int], y : vec, pos : int)
        y[pos] = float(x[pos])
    end
    y = uninit(size(x))
    parallel_do(numel(y),x,y,kernel)
end

function [y : cube{:}[int]] = __int(x : cube{:})
    function [] = __kernel__ kernel(x : vec, y : vec[int], pos : int)
        y[pos] = int(x[pos])
    end
    y = uninit[int](size(x))
    parallel_do(numel(y),x,y,kernel)
end

reduction (x : cube{:}[int]) -> float(x) = __float(x)
reduction (x : cube{:})      -> int(x) = __int(x)

%==============================================================================
% Accuracy profiler helper functions
%==============================================================================
operator (a : cube{:}[int], b : cube{:}[int]) -> (a - b) = (a - b)
operator (a : cube{:}[int], b : cube{:}[int]) -> min(a, b) = min(a, b)
operator (a : cube{:}[int], b : cube{:}[int]) -> max(a, b) = max(a, b)
operator (a : cube{:}[int], b : scalar)    -> a.^b = a.^b

% Restore treat_vecX_as_vec
{!reduction treat_vecX_as_vec=true}