% Type: Concurrent_set
% 
% A concurrent set implementation: a set for which nodes can be 
% added/removed in parallel, based on Harris' algorithm:
%
% Lock-free linkedlist implementation of Harris' algorithm
%  "A Pragmatic Implementation of Non-Blocking Linked Lists" 
%   T. Harris, p. 300-314, DISC 2001.
%
% 
{!author name="Bart Goossens"}
{!doc category="Data structures"}

is_marked_ref    = i -> and(i, 040000000h) > 0
unset_mark       = i -> and(i, not(040000000h))
set_mark         = i -> or(i, 040000000h)

type Concurrent_set[T] : mutable class
    items       : vec[T]
    next        : vec[int]
    head        : int
    tail        : int
    count       : int
    next_free   : int
endtype

% Function: Concurrent_set
% 
% Constructs a new concurrent set of type T
% 
% Usage:
%   : function [s : ^Concurrent_set[T]] = Concurrent_set[T](max_capacity : int = 32768)
% 
% Parameters:
% max_capacity - the maximum capacity of the set (i.e., the maximal number of elements that it can contain)
% 
% Returns:
% s - a new concurrent set
function [s : ^Concurrent_set[T]] = Concurrent_set[T](max_capacity : int = 32768)
    s = Concurrent_set[T](items := vec[T](max_capacity), 
                        next := vec[int](max_capacity), 
                        head:=1, tail:=2, count:=0, next_free:=3)
    for i=0..numel(s.items)-2
        s.next[i] = (i < numel(s.items) - 1) ? i + 2 : 0
    endfor
    s.items[0] = minvalue(T)
    s.items[1] = maxvalue(T)
    s.next[s.tail-1] = 0
    s.next[s.head-1] = s.tail
endfunction

% Function: Concurrent_set.search
% 
% Searches the concurrent set for the specified value
% 
% Usage:
%   : function [left_node : int, right_node : int] = Concurrent_set.search(val : T)
% 
% Parameters:
% val - the value to look for
% 
% Returns:
% left_node - the left node (node before the one that contains val)
% right_node - the right node (the node that contains val, or something >= val)
function [left_node : int, right_node : int] = __device__ search[T](self : ^Concurrent_set[T], val : T)
    left_node_next = left_node = right_node = 0
    while true
        t = self.head
        t_next = self.next[self.head - 1]
        while is_marked_ref(t_next) || self.items[t-1] < val
            if !is_marked_ref(t_next)
                left_node = t
                left_node_next = t_next
            endif
            t = unset_mark(t_next)
            if t == self.tail
                break
            endif
            t_next = self.next[t - 1]
        endwhile
        right_node = t
        if left_node_next == right_node
            if !is_marked_ref(self.next[right_node-1])
                return
            endif
        else
            if left_node_next == (self.next[left_node-1] <- (left_node_next, right_node))
                % TODO: dispose left_node_next
                if !is_marked_ref(self.next[right_node-1])
                    return
                endif
            endif
        endif
    endwhile
endfunction

% Function: Concurrent_set.newnode
% 
% Creates a new node. The node is taken from the free list,
% using an atomic CAS algorithm.
% 
% Usage:
%   : function [out : int] = Concurrent_set.newnode(val)
% 
% Parameters:
% val - the value stored in the node
% 
% Returns:
% out - the created node.
function ptr : int = __device__ newnode[T](self : ^Concurrent_set[T], val : T)
    repeat
        ptr = self.next_free
        if ptr == 0
            error "parallel_set: memory allocator exhausted!"
        endif
        next = self.next[ptr - 1]
        check = (self.next_free <- (ptr, next))
    until ptr == check
    self.items[ptr - 1] = val
endfunction

function result : int = __device__ contains[T](self : ^Concurrent_set[T], val : T)
    result = 0
    iterator = unset_mark(self.next[self.head-1])
    while iterator != self.tail
        if !is_marked_ref(self.next[iterator-1]) && self.items[iterator-1] >= val
            % either we found it, or found the first larger element
            result = self.items[iterator-1] == val
            return
        endif
        iterator = unset_mark(self.next[iterator-1])
    endwhile
endfunction

% Function: Concurrent_set.add
% 
% Adds a new value to the concurrent set
% 
% Usage:
%   : function [success] = Concurrent_set.add(val : T)
% 
% Parameters:
% val - the value to add
% 
% Returns:
% success - true, if successful (i.e., the node did not exist yet), otherwise false.
function success = __device__ add[T](self : ^Concurrent_set[T], val : T)
    new_elem = self.newnode(val)
    success = false
    while true
        [left : int, right : int] = self.search(val)
        % Item already existing
        if right != self.tail && self.items[right-1] == val
            return
        endif
        self.next[new_elem-1] = right
        if right == (self.next[left-1] <- (right, new_elem))
            self.count += 1
            success = true
            return
        endif
    endwhile
endfunction

% Function: Concurrent_set.remove
% 
% Removes the specified node from the concurrent set
% 
% Usage:
%   : function [success] = Concurrent_set.remove(val)
% 
% Parameters:
% val - 
% 
% Returns:
% success - true, if the node was successfully removed, 
% otherwise false (e.g., the node did not exist yet).
function success = __device__ remove[T](self : ^Concurrent_set[T], val : T)
    while true
        [left : int, right : int] = self.search(val)
        % Check if we have found our node
        if right == self.tail || self.items[right-1] != val
            success = false
            return
        endif
        right_succ = self.next[right-1]
        if !is_marked_ref(right_succ)
            if right_succ == (self.next[right-1] <- (right_succ, set_mark(right_succ)))
                self.count -= 1
                success = true
                return
            endif
        endif
    endwhile
endfunction

function [] = __kernel__ test(set : ^Concurrent_set[int], pos : int)   
    %if set.count < 30000 
        set.add(pos + 10)
    %else
    %    set.remove(mod(pos,30000) + 10)
    %endif
endfunction

function y = __device__ test2(set : ^Concurrent_set[int])
    y = set.contains(1000)
endfunction

function [] = main()

    set = Concurrent_set[int](32768)
    print set    
    tic()
    parallel_do(30000, set, test)
    toc()
    print set.count
    
    print sum(set.items == 1000)
    print test2(set)
    print set.contains(1000)

endfunction