%==============================================================================
% cublas.q
% Helper reductions for automatic acceleration using cuBLAS
% Bart Goossens, june 2016
%==============================================================================
{!author name="Bart Goossens"}
{!doc category="Linear algebra/BLAS"}

if blas_version() == 0
    error "The cuBLAS library is currently not available."
endif

reduction (x : [cube{n}|ccube{n}]) _
    -> sum(abs(x)) = blas_asum(x) where $ftype("__host__")
reduction (x : [cube{n}|ccube{n}]) _
    -> sqrt(sum(abs(x).^2)) = blas_nrm2(x) where $ftype("__host__")
reduction (x : cube{n}) _
    -> sqrt(sum(x.^2)) = blas_nrm2(x) where $ftype("__host__")
reduction (x : vec, y : vec, a : scalar) _   
    -> a*x + y = (z_1 = copy(y); blas_axpy(x,z_1,a);z_1) where $ftype("__host__")
reduction (x : cube{n}, y : cube{n}, a : scalar) _
    -> a*x + y = (z_1 = copy(y); blas_axpy(x,z_1,a);z_1) where $ftype("__host__")
reduction (x : cube{n}, y : cube{n}) _
    -> sum(x .* y) = blas_dot(x, y) where $ftype("__host__")
reduction (x : ccube{n}, y : ccube{n}) _
    -> sum(x .* y) = blas_dot(x, y) where $ftype("__host__")
    
% blas_gemm/1
reduction (A : cube{n}, B : cube{n}, C : cube{n}, a : scalar, b : scalar) _
    -> a*A*B+b*C = blas_gemm(A,B,copy(C),false,false,a,b) where $ftype("__host__")
reduction (A : cube{n}, B : cube{n}, C : cube{n}, a : scalar, b : scalar) _
    -> a*transpose(A)*B+b*C = blas_gemm(A,B,copy(C),true,false,a,b) where $ftype("__host__")
reduction (A : cube{n}, B : cube{n}, C : cube{n}, a : scalar, b : scalar) _
    -> a*A*transpose(B)+b*C = blas_gemm(A,B,copy(C),false,true,a,b) where $ftype("__host__")
reduction (A : cube{n}, B : cube{n}, C : cube{n}, a : scalar, b : scalar) _
    -> a*transpose(A)*transpose(B)+b*C = blas_gemm(A,B,copy(C),true,true,a,b) where $ftype("__host__")

reduction (A : ccube{n}, B : ccube{n}, C : ccube{n}, a : cscalar, b : cscalar) _
    -> a*A*B+b*C = blas_gemm(A,B,copy(C),false,false,a,b) where $ftype("__host__")
reduction (A : ccube{n}, B : ccube{n}, C : ccube{n}, a : cscalar, b : cscalar) _
    -> a*transpose(A)*B+b*C = blas_gemm(A,B,copy(C),true,false,a,b) where $ftype("__host__")
reduction (A : ccube{n}, B : ccube{n}, C : ccube{n}, a : cscalar, b : cscalar) _
    -> a*A*transpose(B)+b*C = blas_gemm(A,B,copy(C),false,true,a,b) where $ftype("__host__")
reduction (A : ccube{n}, B : ccube{n}, C : ccube{n}, a : cscalar, b : cscalar) _
    -> a*transpose(A)*transpose(B)+b*C = blas_gemm(A,B,copy(C),true,true,a,b) where $ftype("__host__")

% blas_gemm/2
reduction (A : mat, B : mat, a : scalar) _
    -> a*A*B = blas_gemm(A,B,uninit(size(A,0),size(B,1)),false,false,a,0) where $ftype("__host__")
reduction (A : mat, B : mat, a : scalar) _
    -> a*transpose(A)*B = blas_gemm(A,B,uninit(size(A,1),size(B,1)),true,false,a,0) where $ftype("__host__")
reduction (A : mat, B : mat, a : scalar) _
    -> a*A*transpose(B) = blas_gemm(A,B,uninit(size(A,0),size(B,0)),false,true,a,0) where $ftype("__host__")
reduction (A : mat, B : mat, a : scalar) _
    -> a*transpose(A)*transpose(B) = blas_gemm(A,B,uninit(size(A,1),size(B,0)),true,true,a,0) where $ftype("__host__")

reduction (A : mat, B : mat) _
    -> transpose(A)*B = blas_gemm(A,B,uninit(size(A,1),size(B,0)),true,false,1,0) where $ftype("__host__")
reduction (A : mat, B : mat) _
    -> A*transpose(B) = blas_gemm(A,B,uninit(size(A,0),size(B,0)),false,true,1,0) where $ftype("__host__")
reduction (A : mat, B : mat) _
    -> transpose(A)*transpose(B) = blas_gemm(A,B,uninit(size(A,1),size(B,0)),true,true,1,0) where $ftype("__host__")

reduction (A : cmat, B : cmat, a : cscalar) _
    -> a*A*B = blas_gemm(A,B,uninit(size(A,0),size(B,1)),false,false,a,0) where $ftype("__host__")
reduction (A : cmat, B : cmat, a : cscalar) _
    -> a*transpose(A)*B = blas_gemm(A,B,uninit(size(A,1),size(B,1)),true,false,a,0) where $ftype("__host__")
reduction (A : cmat, B : cmat, a : cscalar) _
    -> a*A*transpose(B) = blas_gemm(A,B,uninit(size(A,0),size(B,0)),false,true,a,0) where $ftype("__host__")
reduction (A : cmat, B : cmat, a : cscalar) _
    -> a*transpose(A)*transpose(B) = blas_gemm(A,B,uninit(size(A,1),size(B,0)),true,true,a,0) where $ftype("__host__")

reduction (A : cmat, B : cmat) _
    -> transpose(A)*B = blas_gemm(A,B,uninit(size(A,1),size(B,0)),true,false,1,0) where $ftype("__host__")
reduction (A : cmat, B : cmat) _
    -> A*transpose(B) = blas_gemm(A,B,uninit(size(A,0),size(B,0)),false,true,1,0) where $ftype("__host__")
reduction (A : cmat, B : cmat) _
    -> transpose(A)*transpose(B) = blas_gemm(A,B,uninit(size(A,1),size(B,0)),true,true,1,0) where $ftype("__host__")


% blas_geam
reduction (A : cube{n}, B : cube{n}, a : scalar, b : scalar) _
    -> a*A+b*B = blas_geam(A,B,uninit(size(A)),false,false,a,b) where $ftype("__host__")
reduction (A : mat, B : mat, a : scalar, b : scalar) _
    -> a*A+b*transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,a,b) where $ftype("__host__")
reduction (A : mat, B : mat, a : scalar, b : scalar) _
    -> a*transpose(A)+b*B = blas_geam(A,B,uninit(size(B)),true,false,a,b) where $ftype("__host__")

reduction (A : cube{n}, B : cube{n}, b : scalar) _
    -> A+b*B = blas_geam(A,B,uninit(size(A)),false,false,1,b) where $ftype("__host__")
reduction (A : mat, B : mat, b : scalar) _
    -> A+b*transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,1,b) where $ftype("__host__")
reduction (A : mat, B : mat, b : scalar) _
    -> transpose(A)+b*B = blas_geam(A,B,uninit(size(B)),true,false,1,b) where $ftype("__host__")
    
reduction (A : cube{n}, B : cube{n}, a : scalar) _
    -> a*A+B = blas_geam(A,B,uninit(size(A)),false,false,a,1) where $ftype("__host__")
reduction (A : mat, B : mat, a : scalar) _
    -> a*A+transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,a,1) where $ftype("__host__")
reduction (A : mat, B : mat, a : scalar) _
    -> a*transpose(A)+B = blas_geam(A,B,uninit(size(B)),true,false,a,1) where $ftype("__host__")
            
reduction (A : mat, B : mat) _
    -> A+transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,1,1) where $ftype("__host__")
                 
reduction (A : ccube{n}, B : ccube{n}, a : cscalar, b : cscalar) _
    -> a*A+b*B = blas_geam(A,B,uninit(size(A)),false,false,a,b) where $ftype("__host__")
reduction (A : cmat, B : cmat, a : cscalar, b : cscalar) _
    -> a*A+b*transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,a,b) where $ftype("__host__")
reduction (A : cmat, B : cmat, a : cscalar, b : cscalar) _
    -> a*transpose(A)+b*B = blas_geam(A,B,uninit(size(B)),true,false,a,b) where $ftype("__host__")

reduction (A : ccube{n}, B : ccube{n}, b : cscalar) _
    -> A+b*B = blas_geam(A,B,uninit(size(A)),false,false,1,b) where $ftype("__host__")
reduction (A : cmat, B : cmat, b : cscalar) _
    -> A+b*transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,1,b) where $ftype("__host__")
reduction (A : cmat, B : cmat, b : cscalar) _
    -> transpose(A)+b*B = blas_geam(A,B,uninit(size(B)),true,false,1,b) where $ftype("__host__")
    
reduction (A : ccube{n}, B : ccube{n}, a : cscalar) _
    -> a*A+B = blas_geam(A,B,uninit(size(A)),false,false,a,1) where $ftype("__host__")
reduction (A : cmat, B : cmat, a : cscalar) _
    -> a*A+transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,a,1) where $ftype("__host__")
reduction (A : cmat, B : cmat, a : cscalar) _
    -> a*transpose(A)+B = blas_geam(A,B,uninit(size(B)),true,false,a,1) where $ftype("__host__")
            
reduction (A : cmat, B : cmat) _
    -> A+transpose(B) = blas_geam(A,B,uninit(size(A)),false,true,1,1) where $ftype("__host__")
    
function [] = main()

    x = 1..63
    b = ones(size(x))
    y = 2.0*b + x
    print y
    
    A = randn(4,4)*100
    B = randn(4,4)*100
    At = transpose(A)

    print round(At*B)
    print round(transpose(A)*B)
    print round(2*A+4*transpose(B))
    print round(A + transpose(B))

endfunction
