import "linalg.q"

{!author name="Bart Goossens"}
{!doc category="Statistics"}

% Section: curve_fitting.q
% Contains several curve fitting functions, such as linear regression, robust linear regression.

type robust_function : [__device__ (scalar, scalar) -> scalar]

% Function: weight_function_constant
% 
% Returns the constant weight 1
% 
% Usage:
%   : function [w : scalar] = __device__ weight_function_constant(x : scalar, d : scalar)
% 
% Parameters:
% x - ignored
% d - ignored
% 
% Returns:
% w - the constant weight 1
function w = __device__ weight_function_constant(x : scalar, d : scalar)
    w = 1.0
endfunction

% Function: weight_function_L1
% 
% L1 weight function, defined as `x->1/max(eps,abs(x))`
% 
% Usage:
%   : function [w : scalar] = __device__ weight_function_L1(x : scalar, d : scalar)
% 
% Parameters:
% x - positive input value
% d - unused robust function parameter
% 
% Returns:
% w - the L1 weight
function w = __device__ weight_function_L1(x : scalar, d : scalar)
    w = 1.0 / max(eps, abs(x))
endfunction

% Function: weight_function_L12
% 
% L12 weight function, defined as `x->1/sqrt(1+0.5*x^2)`
% 
% Usage:
%   : function [w : scalar] = __device__ weight_function_L12(x : scalar, d : scalar)
% 
% Parameters:
% x - positive input value
% d - unused robust function parameter
% 
% Returns:
% w - the L12 weight
function w = __device__ weight_function_L12(x : scalar, d : scalar)
    w = 1.0 / sqrt(1 + 0.5 * x ^ 2)
endfunction

% Function: weight_function_huber
% 
% Huber robust function, defined as `x-> x<d ? 1.0 : d/x`
% 
% Usage:
%   : function [w : scalar] = __device__ weight_function_huber(x : scalar, d : scalar)
% 
% Parameters:
% x - positive input value
% d - Huber weight parameter
% 
% Returns:
% w - the Huber weight for x
function w = __device__ weight_function_huber(x : scalar, d : scalar)
    d = d <= 0 ? 1.345 : d
    w = x < d ? 1.0 : d / x
endfunction

% Function: weight_function_fair
% 
% Fair robust function, defined as `x->1/(1+x/d)`
% 
% Usage:
%   : function [w : scalar] = __device__ weight_function_fair(x : scalar, d : scalar)
% 
% Parameters:
% x - positive input value
% d - Fair weight parameter
% 
% Returns:
% w - the Fair weight for x
function w = __device__ weight_function_fair(x : scalar, d : scalar)
    d = d == 0 ? 1/1.4 : 1/d
    w = 1/(1+x*d)
endfunction

% Function: weight_function_welsch
% 
% Welsch robust function, defined as `x->exp(-x^2/d^2)`
% 
% Usage:
%   : function [w : scalar] = __device__ weight_function_welsch(x : scalar, d : scalar)
% 
% Parameters:
% x - positive input value
% d - Welsch weight parameter
% 
% Returns:
% w - the Welsch weight for x
function w = __device__ weight_function_welsch(x : scalar, d : scalar)
    d = d == 0 ? 1/2.9846 : 1/d
    w = exp(-x^2*d^2)
endfunction

% Function: eval_poly
% 
% Evaluate the univariate polynomial `sum(A.*(x.^(0..N-1)))`
% 
% Usage:
%   : function [y : vec] = eval_poly(x : vec'const'static, A : vec'const'static)
% 
% Parameters:
% x - a vector of x values
% A - a vector containing the the coefficients of the polynomial
% 
% Returns:
% y - the evaluated polynomial for each x value
function y : vec = eval_poly(x : vec, A : vec)
    N = numel(A)-1
    X = uninit(numel(x),N+1)
    
    {!parallel for}
    for m=0..numel(x)-1
        x0 = x[m]
        X[m,0] = 1.0
        X[m,1..N] = x0.^(1..N)
    endfor
    
    y = transpose(X * A)
endfunction
  
% Function: linear_regression
% 
% Calculates a linear regression on the first column of p (x-values) and the second column of p
% (y-values).
% 
% Usage:
%   : function [A : vec'col] = linear_regression(p : mat(:,2), w : vec, N : int)
% 
% Parameters:
% p - matrix with a couple (x,y) on each row
% w - weight for each sample
% N - the order of the polynomial to be fitted (1=linear, 2=quadratic, ...)
% 
% Returns:
% A - a row vector with the coefficients of the fitted univariate polynomial
function A : vec'row = linear_regression(p : mat(:,2), N : int, w : vec = [])
    assert(size(p,0)==numel(w), "The number of weights does not match the data!")
    if numel(w)==0
        w = ones(size(p,0))
    endif

    X = uninit(size(p,0),N+1)
    {!parallel for}
    for m=0..size(p,0)-1
        X[m,0] = w[m]
        X[m,1..N] = w[m] * p[m,0].^(1..N)
    endfor
    
    A = inv(transpose(X)*X)*transpose(X)*p[:,1]
endfunction

% Function: linear_robust_regression
% 
% Calculates a linear robust regression on the first column of p (x-values) and the second column of p
% (y-values). This type of regression is robust against outliers
% 
% Usage:
%   : function A = linear_robust_regression(p : mat(:,2), N : int, weightfn : robust_function, 
%   :    param : scalar = 0.0, max_iter : int = 10)
% 
% Parameters:
% p - matrix with a couple (x,y) on each row
% N - the order of the polynomial to be fitted (1=linear, 2=quadratic, ...)
% weightfn - the robust function to be used (<weight_function_L1>, <weight_function_L12>,
%   <weight_function_huber>, <weight_function_fair>, <weight_function_welsch>)
% param - a parameter for the robust function
% max_iter - the maximum number of iterations
% 
% Returns:
% A - a row vector with the coefficients of the fitted univariate polynomial
function A = linear_robust_regression(p : mat(:,2), N : int, weightfn : robust_function, 
    param : scalar = 0.0, max_iter : int = 100)

    X = uninit(size(p,0),N+1)
    A = linear_regression(p, N, ones(size(p,0))) % initial solution
    err = maxvalue(scalar)
    
    for iter=1..max_iter
        y : vec = eval_poly(p[:,0], A)
        last_err = err
        err = 0.0
    
        {!parallel for}
        for m=0..size(p,0)-1
            w = weightfn(abs(y[m] - p[m,1]), param)
            err += w * (y[m] - p[m,1])^2
            x0 = p[m,0]
            X[m,0] = w
            X[m,1..N] = w * x0.^(1..N)
        endfor
        
        print "iter ",iter, " err ",err
        
        A = inv(transpose(X)*X)*transpose(X)*p[:,1]
        
        if err>0.999*last_err
            break % convergence
        endif
    endfor
endfunction

% Function: plot_regression_result
% 
% Visualizes the result of the regression
% 
% Usage:
%   : function [] = plot_regression_result(p : mat(:,2), A : vec)
% 
% Parameters:
% p - matrix with a couple (x,y) on each row
% A - a vector containing the the coefficients of the polynomial
function [] = plot_regression_result(p : mat(:,2), A : vec)
    y : vec = eval_poly(p[:,0], A)
    plot([y,transpose(p[:,1])])
endfunction

function [] = main()
    x = linspace(0,1,256)
    y = 2*x.^2+0.4+0.5*randn(size(x))
    y[10] = 12
    y[20] = 15
    y[60] = 8
    p : mat(:,2) = transpose([x,y])
    w = ones(size(p,0))
    plot(x,y)
    A1 = linear_regression(p, 2, w)
    A2 = linear_robust_regression(p, 2, weight_function_huber)
    print "linear regression: ", A1
    print "linear robust regression: ",A2
    plot_regression_result(p,A1),title("Linear regression")
    plot_regression_result(p,A2),title("Robust regression")
endfunction