%==========================================================
% Quasar library for deep learning training and processing
% March 2016 - Simon Donné
%==========================================================
import "Quasar.UI.dll"
import "Quasar.DNN.dll"
import "imresize.q"

{!author name="Simon Donné"}
{!doc category="Machine Learning"}

type Data : cube
type Parameter : cube{4}
type Parameters : vec[Parameter]
type ForwardCNNFunction  : [(Data,Parameters)->(Data)] %(input, parameters) -> (output)
type BackwardCNNFunction : [(Data,Data,Parameters,Data)->(Data,Parameters)] %(input,output,parameters,dzdoutput)->(dzdinput,dzdparameters)
type DeepLearningLayer : class
    name : string
    forward_pass  : ForwardCNNFunction
    backward_pass : BackwardCNNFunction
    parameters : Parameters
    dropout : scalar
endtype
type DeepLearningLayers : vec[^DeepLearningLayer]
type DeepLearningLossFunction : [(Data,Data)->(Data,Data)]
type DeepLearningSettings : mutable class
    rate:scalar
    momentum_first_order:scalar
    momentum_second_order:scalar
    plot_period:int
endtype
type DeepLearningNetwork : mutable class
    loss : DeepLearningLossFunction
    loss_name : string
    layers : DeepLearningLayers
    settings : ^DeepLearningSettings
    nr_layers : int
endtype

function [] = main()
    %toy example: sharpen a linearly upsampled image
    %if you want to use the read_from_file and save_to_file, please make sure the filepath is set correctly (the folder needs to exist!)
    %it also includes some code to display the current filter weights
    gt = zeros(200,200)
    parallel_do(size(gt),__kernel__ (pos)->gt[pos]=mod((pos[0]<pos[1])+(199-pos[0]>pos[1]),2))
    cim = imresize(imresize(gt,1/4,"linear"),4,"linear")
    mask = zeros(200,200)
    mask[7..192,7..192] = 1
    
%    imshow(gt)
%    imshow(cim)
    
    save_to_file = false
    read_from_file = true
    filepath = "/home/sdonn/code/Machine Learning/quasar_machines"
    network_name = "sharpening"
    
    network = initDeepLearningNetwork()
    if read_from_file && network.exists_on_file(filepath,network_name)
        network.load_from_file(filepath,network_name)
    else    
        network.add_layer(get_gabor_convolve_layer(radius=5,c_in=1,c_out_factor=13))
        network.add_layer(get_convolve_layer(radius=0,c_in=13,c_out=4))
        network.add_layer(get_RELU_layer())
        network.add_layer(lyr=get_convolve_layer(radius=2,c_in=4,c_out=1))
            lyr.dropout = 0.0
        network.set_loss_function("MSE")
        network.settings.rate = 1e-3
        network.settings.plot_period = 100
    endif
    
    for loop=1..1
        network.train(cim,gt,nr_iterations=300,mask)
        
        if save_to_file
            network.save_to_file(filepath,network_name)
        endif
    endfor
    
    network.layers[0].visualize_convolve_layer()
endfunction

% Function: DeepLearningFrameworkVersion
% Returns the current version of the deep learning framework.
% The version number should be used as follows:
%   - changes behind the comma should be backwards compatible (i.e. a v1.6 network can be read in by the v1.7 framework, but not necessarily vice versa -- although an effort will be made)
%   - changes in the integer part of the version number are not compatible at all (i.e. a v2.0 is not legible by either v1.0 or v3.0)
% As such, when you add layer types, please increment the number behind the comma
%
% : function [version:scalar] = DeepLearningFrameworkVersion()
%
function [version:scalar] = DeepLearningFrameworkVersion()
    version = 2.00
endfunction

% Function: DeepLearningNetwork.load_from_file
% Loads a pre-trained network from the hard disk, with checks for the framework version
% Must be called on a pre-initialized DeepLearningNetwork
%
% : function [] = load_from_file(self:^DeepLearningNetwork,filepath:string,network_name:string)
%
% See also:
%    <initDeepLearningNetwork>,<DeepLearningFrameworkVersion>,<save_to_file>
%
function [] = load_from_file(self:^DeepLearningNetwork,filepath:string,network_name:string)
    %test for legible error in case the path does not exist yet
    dir(strcat(filepath,"/*"))
    
    %load the meta file: framework version, loss function info, layer info, training settings
    metafilename = sprintf("%s/%s.meta",filepath,network_name)
    fh = fopen(metafilename,"rt")
    
    %sanity check on the text file
    if fgets(fh)!="Quasar deep learning network"
        error("This is not a quasar deep learning network!")
    endif
    
    %version sanity check
    that_version = sscanf(fgets(fh),"%f")[0]
    this_version = DeepLearningFrameworkVersion()
    if int(that_version) > int(this_version)
        error("The network was created with a newer version of the framework that is not supported.")
    elseif that_version > this_version + 1e-3
        printf("WARNING> The network was created with a newer version of the framework. Trying to load it, but some features might be unsupported.\n")
    elseif int(that_version) < int(this_version)
        error("The network was created with a version of the framework that is no longer supported.")
    endif
    
    %get the network parameters
    loss_name = fgets(fh)
    self.set_loss_function(loss_name)
    self.nr_layers = sscanf(fgets(fh),"%d")[0]
    self.layers = DeepLearningLayers(self.nr_layers*2)
    for layer = 0..self.nr_layers-1
        layer_name = fgets(fh)
        match layer_name with
        | "dilation"            ->  self.layers[layer] = get_dilation_layer()
        | "subsample_NN"        ->  self.layers[layer] = get_subsample_NN_layer()
        | "erosion"             ->  self.layers[layer] = get_erosion_layer()
        | "convolution"         ->  self.layers[layer] = get_convolve_layer()
        | "biased_convolution"  ->  self.layers[layer] = get_biased_convolve_layer()
        | "fully connected"     ->  self.layers[layer] = get_fully_connected_layer()
        | "sigmoid"             ->  self.layers[layer] = get_sigmoid_layer()
        | "tanh"                ->  self.layers[layer] = get_tanh_layer()
        | "ReLU"                ->  self.layers[layer] = get_RELU_layer()
        | "leaky ReLU"          ->  self.layers[layer] = get_leaky_RELU_layer()
        | _ -> error(sprintf("Unknown layer type: %s",layer_name))
        endmatch
    endfor
    self.settings.rate = sscanf(fgets(fh),"%f")[0]
    self.settings.momentum_first_order = sscanf(fgets(fh),"%f")[0]
    self.settings.momentum_second_order = sscanf(fgets(fh),"%f")[0]
    self.settings.plot_period = sscanf(fgets(fh),"%d")[0]
    
    %read in each of the layers' parameters
    for layer = 0..self.nr_layers-1
        layerfilename = sprintf("%s/%s.layer%d",filepath,network_name,layer)
        [tparam] = load(layerfilename)
        self.layers[layer].parameters[:]=tparam
    endfor
endfunction

% Function: DeepLearningNetwork.exists_on_file
% Checks whether a given network exists in a certain location. Does not perform sanity checks on the network itself (yet).
%
% : function [] = exists_on_file(self:^DeepLearningNetwork,filepath:string,network_name:string)
%
% See also:
%    <initDeepLearningNetwork>,<load_from_file>,<save_to_file>
%
function [exists:int]= exists_on_file(self:^DeepLearningNetwork,filepath:string,network_name:string)
    exists = exist(sprintf("%s/%s.meta",filepath,network_name), "file")
endfunction

% Function: DeepLearningNetwork.save_to_file
% Save a network to the hard disk. Must be called on a pre-initialized DeepLearningNetwork.
%
% : function [] = save_to_file(self:^DeepLearningNetwork,filepath:string,network_name:string)
%
% See also:
%    <initDeepLearningNetwork>,<DeepLearningFrameworkVersion>,<load_from_file>
%
function [] = save_to_file(self:^DeepLearningNetwork,filepath:string,network_name:string)
    %test for legible error in case the path does not exist yet
    dir(strcat(filepath,"/*"))
    
    %save the meta file: framework version, loss function info, layer info, training settings
    metafilename = sprintf("%s/%s.meta",filepath,network_name)
    fh = fopen(metafilename,"wt")
    fprintf(fh,"Quasar deep learning network\n")
    fprintf(fh,"%1.2f\n",DeepLearningFrameworkVersion())
    fprintf(fh,"%s\n",self.loss_name)
    fprintf(fh,"%d\n",self.nr_layers)
    for layer = 0..self.nr_layers-1
        fprintf(fh,"%s\n",self.layers[layer].name)
    endfor
    fprintf(fh,"%f\n",self.settings.rate)
    fprintf(fh,"%f\n",self.settings.momentum_first_order)
    fprintf(fh,"%f\n",self.settings.momentum_second_order)
    fprintf(fh,"%d\n",self.settings.plot_period)
    fclose(fh)
    
    %for each layer, save a separate (binary) file with its parameter values
    for layer = 0..self.nr_layers-1
        layerfilename = sprintf("%s/%s.layer%d",filepath,network_name,layer)
        save(layerfilename,copy(self.layers[layer].parameters))
    endfor
endfunction

% Function: DeepLearningNetwork.train_multiple
% Trains a pre-constructed network given a set of inputs and groundtruths. At each iteration, the subsequent input-groundtruth pair will be taken, while preserving the momentum.
% The optional arguments (in order):
%   - the number of gradient descent iterations (integer, defaults to 100)
%   - the masks: only groundtruth pixels for which masks[img][pos] is set to 1 are counted in the error criterion (same dimensions as groundtruth, defaults to all 1's)
%   - the visualization: boolean value to indicate whether the GUI is shown during optimization. The refresh rate is controlled using DeepLearningNetwork.settings.plot_period
%
% The DeepLearningNetwork contains a settings field, comprising
%   - network.settings.rate: step size for the gradient descent (defaults to 1e-3, which is plausible for groundtruth in [0,1])
%   - network.settings.momentum: momentum for the stochastic gradient descent (in [0,1], higher will mean more regularized steps - defaults to 0.85)
%   - network.settings.plot_period: refresh rate for the UI when visualization is turned on
%
% The training method is so-called Stochastic Gradient Descent, which uses a momentum term, a shrinkage term, and standard gradient descent.
%   The momentum is a network-wide setting, whereas the shrinkage terms are set per-parameter (each layer for its own, in layer.shrink_rates).
%
% : function [] = train_multiple(self:^DeepLearningNetwork,input:Data,groundtruth:Data,nr_iterations:int=100,mask:Data=zeros(0,0,0,0),visualization:int=true)
%
% See also:
%    <initDeepLearningNetwork>,<train>
%
function [] = train_multiple(self:^DeepLearningNetwork,inputs:vec[Data],groundtruths:vec[Data],nr_iterations:int=100,masks:vec[Data]=vec[Data](0),visualization:int=true)
    %learning with stochastic gradient descent: parameters
    rate = self.settings.rate
    momentum_first_order = self.settings.momentum_first_order
    momentum_second_order = self.settings.momentum_second_order
    plot_period = self.settings.plot_period
    if plot_period == 0 || !has_display()
        visualization = false
    endif
    
    nr_images = numel(inputs)
    assert(numel(groundtruths) == nr_images)
    
    summasks = zeros(nr_images)
    pos_counts = zeros(nr_images)
    neg_counts = zeros(nr_images)
    if numel(masks) == 0
        masks = vec[Data](nr_images)
        for img = 0..nr_images-1
            masks[img] = ones(size(groundtruths[img]))
        endfor
    endif
    for img = 0..nr_images-1
        if numel(masks[img]) < numel(groundtruths[img])
            masks[img] = ones(size(groundtruths[img]))
        endif
        summasks[img] = sum(masks[img])
        pos_counts[img] = max(sum(groundtruths[img] >= 1),1e-9)
        neg_counts[img] = max(sum(groundtruths[img] <= 0),1e-9)
    endfor
    
    
    %preparing the visualization, if necessary
    if visualization
        frm = form("Deep learning network training")
        hgr_plt = frm.add_horizontallayout()
        dsp_plt = hgr_plt.add_display()
        dsp_hst = hgr_plt.add_display()
        dsp_out = frm.add_display()
    endif
    
    %optimization variables
    first_order_moments = vec[Parameters](self.nr_layers)
    second_order_moments = vec[Parameters](self.nr_layers)
    for i = 0..self.nr_layers-1
        nr_params = numel(self.layers[i].parameters)
        first_order_moments[i] = Parameters(nr_params)
        for j = 0..nr_params-1
            first_order_moments[i][j] = zeros(size(self.layers[i].parameters[j]))
        endfor
        second_order_moments[i] = Parameters(nr_params)
        for j = 0..nr_params-1
            second_order_moments[i][j] = zeros(size(self.layers[i].parameters[j]))
        endfor
    endfor

    errors = zeros(nr_images)
    error_log = zeros(nr_iterations)
    for iteration = 0..nr_iterations-1
%        img_idx = mod(floor((iteration+1)/plot_period),nr_images)
        img_idx = mod(iteration,nr_images)
        input = inputs[img_idx]
        groundtruth = groundtruths[img_idx]
        mask = masks[img_idx]
        summask = summasks[img_idx]
        pos_count = pos_counts[img_idx]
        neg_count = neg_counts[img_idx]
        
        tic()
        %the forward pass
        [layer_output,input_masks] = self.process_full(input)
        output = layer_output[self.nr_layers-1]

        %the current derivative with respect to the output
        [z,dzdoutput] = self.loss(output,groundtruth)
        %apply an ignore mask around the borders of the groundtruth
        z = z .* mask
        dzdoutput = dzdoutput .* mask
        %save the error for fitting
        this_error = sum(z)/summask
        if this_error > 0.7
            stop
        endif
        errors[img_idx] = this_error
        error_log[iteration] = sum(errors)/max(sum(errors>0),1e-9)
        
        %for each layer, backward propagation, updating the momentum and taking the step
        unbiased_rate = rate * sqrt(1-momentum_second_order.^(iteration+1)) / (1-momentum_first_order.^(iteration+1)) %see Adam paper
        optimization_epsilon = 1e-8
        for i = self.nr_layers-1..-1..0
            [dzdoutput,dzdparameters] = self.layers[i].backward_pass(i==0?input:layer_output[i-1],layer_output[i],self.layers[i].parameters,dzdoutput)
            %we apply the input mask to the dzdoutput, as these inputs were not actually available!
            dzdoutput = dzdoutput .* input_masks[i]
            nr_params = numel(self.layers[i].parameters)
            for j = 0..nr_params-1
                grad:Parameter = dzdparameters[j]
                first_order_moments[i][j] = first_order_moments[i][j]*momentum_first_order + (1-momentum_first_order)*grad
                    second_order_moments[i][j] = second_order_moments[i][j]*momentum_second_order + (1-momentum_second_order)*grad.^2
                    self.layers[i].parameters[j] = self.layers[i].parameters[j] - unbiased_rate * first_order_moments[i][j] ./ (sqrt(second_order_moments[i][j])+optimization_epsilon)
            endfor
        endfor
        
        if visualization && (mod(iteration+1,plot_period) == 0 || iteration == 0)
            output = self.process(input)
            printf("CNN Adam iteration %d: %f ms\n",iteration+1,toc(""))
            if self.loss_name == "binary"
                false_positive = (output > 0.5) && (groundtruth <= 0)
                false_negative = (output < 0.5) && (groundtruth >= 1)
                 true_positive = (output <=0.5) && (groundtruth <= 0)
                 true_negative = (output >=0.5) && (groundtruth >= 1)
                error_image = uninit(size(mask,0),size(mask,1),3)
                error_image[:,:,0] = (false_positive || false_negative) .* mask
                error_image[:,:,1] = ( true_positive || true_negative ) .* mask
                error_image[:,:,2] = groundtruth .* mask
                dsp_out.remove_all()
                dsp_out.imshow(error_image)
            else 
                dsp_out.imshow(output,[min(groundtruth),max(groundtruth)])
            endif
            
            %the plot output: error functions
            dsp_plt.plot(1..iteration+1,error_log[0..iteration],"b")%data fit
            xlim([1,nr_iterations])
            ylim([0,max(error_log[iteration]*1.5,1e-9)])
            title("Data fit error")
            hold("off")
            
            %the current score distribution of positive and negative pixels
            %we formulate a histogram between -1 and 2 with 32 bins - only applicable if the groundtruth is in that same range
            if self.loss_name == "binary"
                hist_length = 128
                hist_steps = -1..(3/(hist_length-1))..2
                pos_hist = zeros(hist_length)
                neg_hist = zeros(hist_length)
                function [] = __kernel__ fill_hist(output:Data,hist_length:int,groundtruth:Data,pos_hist:vec,neg_hist:vec,pos:ivec3)
                    val = output[pos]
                    idx:int = int(round((val+1)/3*(hist_length-1)))
                    idx = clamp(idx,hist_length-1)
                    if groundtruth[pos] >= 1
                        pos_hist[idx] += 1.0
                    elseif groundtruth[pos] <= 0
                        neg_hist[idx] += 1.0
                    endif
                endfunction
                parallel_do(size(output),output,hist_length,groundtruth,pos_hist,neg_hist,fill_hist)
                dsp_hst.plot([0.0,0.0],[0.0,1.0],"b")
                hold("on")
                plot([1.0,1.0],[0.0,1.0],"b")
                plot(hist_steps,neg_hist/neg_count,"r")
                plot(hist_steps,pos_hist/pos_count,"g")
                xlim([-1.0,2.0])
                title("Output distributions (red background, green foreground)")
                hold("off")
            endif
        else
            toc("")
        endif
    endfor
    
    if visualization
        frm.close()
    endif
endfunction

% Function: DeepLearningNetwork.train
% Trains a pre-constructed network given the input and groundtruth.
% The optional arguments (in order):
%   - the number of gradient descent iterations (integer, defaults to 100)
%   - the mask: only groundtruth pixels for which mask[pos] is set to 1 are counted in the error criterion (same dimensions as groundtruth, defaults to all 1's)
%   - the visualization: boolean value to indicate whether the GUI is shown during optimization. The refresh rate is controlled using DeepLearningNetwork.settings.plot_period
%
% The DeepLearningNetwork contains a settings field, comprising
%   - network.settings.rate: step size for the gradient descent (defaults to 1e-3, which is plausible for groundtruth in [0,1])
%   - network.settings.momentum: momentum for the stochastic gradient descent (in [0,1], higher will mean more regularized steps - defaults to 0.85)
%   - network.settings.plot_period: refresh rate for the UI when visualization is turned on
%
% The training method is so-called Stochastic Gradient Descent, which uses a momentum term, a shrinkage term, and standard gradient descent.
%   The momentum is a network-wide setting, whereas the shrinkage terms are set per-parameter (each layer for its own, in layer.shrink_rates).
%
% : function [] = train(self:^DeepLearningNetwork,input:Data,groundtruth:Data,nr_iterations:int=100,mask:Data=zeros(0,0,0),visualization:int=true)
%
% See also:
%    <initDeepLearningNetwork>
%
function [] = train(self:^DeepLearningNetwork,input:Data,groundtruth:Data,nr_iterations:int=100,mask:Data=zeros(0,0,0),visualization:int=true)
    inputs=vec[Data](1)
    inputs[0] = input
    groundtruths=vec[Data](1)
    groundtruths[0] = groundtruth
    masks=vec[Data](1)
    masks[0] = mask
    self.train_multiple(inputs,groundtruths,nr_iterations,masks,visualization)
endfunction

% Function: deeplearning_binary_loss
% Internal function. Binary loss function. Please refer to <set_loss_function> instead.
%
% See also:
%    <set_loss_function>
%
function [z:Data,dzdoutput:Data] = deeplearning_binary_loss(output:Data,groundtruth:Data)
    pos_count = max(sum(groundtruth >= 1),1e-9)
    neg_count = max(sum(groundtruth <= 0),1e-9)
    z = max(output.*(1-groundtruth),0)/neg_count + max((1 - output).*groundtruth,0)/pos_count
    z = z*numel(groundtruth)
    dzdoutput = (output > 0).*(1-groundtruth)/neg_count - (output < 1).*(groundtruth)/pos_count
    dzdoutput = dzdoutput*numel(groundtruth)
endfunction

% Function: deeplearning_MSE_loss
% Internal function. MSE loss function. Please refer to <set_loss_function> instead.
%
% See also:
%    <set_loss_function>
%
function [z:Data,dzdoutput:Data] = deeplearning_MSE_loss(output:Data,groundtruth:Data)
    z = (output-groundtruth).^2
    dzdoutput = 2*(output-groundtruth)
endfunction

% Function: deeplearning_MAE_loss
% Internal function. MAE loss function. Please refer to <set_loss_function> instead.
%
% See also:
%    <set_loss_function>
%
function [z:Data,dzdoutput:Data] = deeplearning_MAE_loss(output:Data,groundtruth:Data)
    z = abs(output-groundtruth)
    dzdoutput = sign(output-groundtruth)
endfunction

% Function: deeplearning_empty_loss
% Internal function. Default (unset) loss function. Please refer to <set_loss_function> instead.
% Will throw an error when called, to indicate that a loss function should be set.
%
% See also:
%    <set_loss_function>
%
function [z:Data,dzdoutput:Data] = deeplearning_empty_loss(output:Data,groundtruth:Data)
    z = zeros(size(groundtruth))
    dzdoutput = zeros(size(output))
    error("Please pick a loss function for your network using DeepLearningNetwork.set_loss_function()!")
endfunction

% Function: DeepLearningNetwork.process_full
% Internal function. Full propagation function, which also returns each intermediate result. Please refer to <process> instead.
%
% See also:
%    <process>
%
function [output:vec[Data],input_masks:vec[Data]] = process_full(self:^DeepLearningNetwork,input:Data)
    output = vec[Data](self.nr_layers)
    input_masks = vec[Data](self.nr_layers)
    input_masks[0] = rand(size(input)) >= self.layers[0].dropout
    output[0] = self.layers[0].forward_pass(input.*input_masks[0], self.layers[0].parameters) / (1-self.layers[0].dropout)
    for i = 1..self.nr_layers-1
        input_masks[i] = rand(size(output[i-1])) >= self.layers[i].dropout
        output[i] = self.layers[i].forward_pass(output[i-1].*input_masks[i], self.layers[i].parameters) / (1-self.layers[i].dropout)
    endfor
endfunction

% Function: DeepLearningNetwork.process
% Network propagation function, which returns the output of the entire network.
%
%   function output:Data = process(self:^DeepLearningNetwork,input:Data)
%
% See also:
%    <train>
%
function [output:Data] = process(self:^DeepLearningNetwork,input:Data)
    output = self.layers[0].forward_pass(input, self.layers[0].parameters)
    for i = 1..self.nr_layers-1
        output = self.layers[i].forward_pass(output, self.layers[i].parameters)
    endfor
endfunction

% Function: DeepLearningLayer.forward_pass
% Internal function for the processing performed by a single layer. Please refer to <process> instead.
%
% See also:
%    <process>
%
function [output:Data] = forward_pass(self:^DeepLearningLayer,input:Data)
    output = self.forward_pass(input,copy(self.parameters))
endfunction

% Function: deeplearning_dilation_forward
% Internal function for the processing performed by a dilation layer. Please refer to <get_dilation_layer> instead.
%
% See also:
%    <get_dilation_layer>
%
function [output:Data] = deeplearning_dilation_forward(input:Data,parameters:Parameters)
    rho = int(parameters[0][0])
    output = zeros(size(input))
    function [] = __kernel__ forward_dilation(input,rho,output,pos:ivec3)
        bestval = input[pos]
        for dx = -rho..rho
        for dy = -rho..rho
            bestval = max(bestval,input[pos[0]+dx,pos[1]+dy,pos[2]])
        endfor
        endfor
        output[pos] = bestval
    endfunction
    parallel_do(size(output),input,rho,output,forward_dilation)
endfunction

% Function: deeplearning_dilation_backward
% Internal function for the processing performed by a dilation layer. Please refer to <get_dilation_layer> instead.
%
% See also:
%    <get_dilation_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_dilation_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    rho = int(parameters[0][0])
    dzdinput = zeros(size(input))
    %backpropagation is pretty easy: simply find the pixel responsible for the resulting value, and put the dzdoutput there
    function [] = __kernel__ backward_dilation(input,rho,dzdoutput,dzdinput,pos:ivec3)
        bestpos = [pos[0]-rho,pos[1]-rho,pos[2]]
        bestval = input[bestpos]
        for dx = -rho..rho
        for dy = -rho..rho
            thispos = [pos[0]+dx,pos[1]+dy,pos[2]]
            thisval = input[thispos]
            if thisval > bestval
                bestval = thisval
                bestpos = thispos
            endif
        endfor
        endfor
        dzdinput[bestpos] += dzdoutput[pos]
    endfunction
    parallel_do(size(dzdinput),input,rho,dzdoutput,dzdinput,backward_dilation)
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_dilation_layer
% Function for acquiring a pre-constructed dilation layer, optionally with a specified radius (defaults to 1, so a 3x3 dilation).
% To be used in conjunction with <add_layer>.
% If you are looking for the maxpooling layer, simply concatenate a dilation layer with a subsample_NN layer (<get_subsample_NN_layer>).
%
% : function [dilate_layer:^DeepLearningLayer] = get_dilation_layer(radius:int=1)
%
% See also:
%    <add_layer>
%
function [dilate_layer:^DeepLearningLayer] = get_dilation_layer(radius:int=1)
    init_param = Parameters(1)
    init_param[0] = [radius]
    dilate_layer = DeepLearningLayer(name:="dilation",forward_pass:=deeplearning_dilation_forward,backward_pass:=deeplearning_dilation_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_NMS_forward
% Internal function for the processing performed by a non-maximum suppression layer (radius 1). Please refer to <get_NMS_layer> instead.
%
% See also:
%    <get_NMS_layer>
%
function [output:Data] = deeplearning_NMS_forward(input:Data,parameters:Parameters)
    output = zeros(size(input))
    function [] = __kernel__ forward_NMS(input,output,pos:ivec3)
        if input[pos] >= input[pos+[1,0,0]] && input[pos] >= input[pos-[1,0,0]] && input[pos] >= input[pos+[0,1,0]] && input[pos] >= input[pos-[0,1,0]]
            output[pos] = input[pos]
        endif
    endfunction
    parallel_do(size(output),input,output,forward_NMS)
endfunction

% Function: deeplearning_NMS_backward
% Internal function for the processing performed by a non-maximum suppression layer. Please refer to <get_dilation_layer> instead.
%
% See also:
%    <get_dilation_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_NMS_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    dzdinput = zeros(size(input))
    %backpropagation is pretty easy: simply find the pixel responsible for the resulting value, and put the dzdoutput there
    function [] = __kernel__ backward_NMS(input,dzdoutput,dzdinput,pos:ivec3)
        if input[pos] >= input[pos+[1,0,0]] && input[pos] >= input[pos-[1,0,0]] && input[pos] >= input[pos+[0,1,0]] && input[pos] >= input[pos-[0,1,0]]
            dzdinput[pos] = dzdoutput[pos]
        endif
    endfunction
    parallel_do(size(dzdinput),input,dzdoutput,dzdinput,backward_NMS)
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_NMS_layer
% Function for acquiring a a non-maximum suppression layer. Every channel is treated separately.
% To be used in conjunction with <add_layer>.
%
% : function [NMS_layer:^DeepLearningLayer] = get_NMS_layer(radius:int=1)
%
% See also:
%    <add_layer>
%
function [NMS_layer:^DeepLearningLayer] = get_NMS_layer()
    init_param = Parameters(1)
    init_param[0] = [0.0]
    
    NMS_layer = DeepLearningLayer(name:="NMS",forward_pass:=deeplearning_NMS_forward,backward_pass:=deeplearning_NMS_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_erosion_forward
% Internal function for the processing performed by an erosion layer. Please refer to <get_erosion_layer> instead.
%
% See also:
%    <get_erosion_layer>
%
function [output:Data] = deeplearning_erosion_forward(input:Data,parameters:Parameters)
    rho = int(parameters[0][0])
    output = zeros(size(input))
    function [] = __kernel__ forward_erosion(input,rho,output,pos:ivec3)
        bestval = input[pos]
        for dx = -rho..rho
        for dy = -rho..rho
            bestval = min(bestval,input[pos[0]+dx,pos[1]+dy,pos[2]])
        endfor
        endfor
        output[pos] = bestval
    endfunction
    parallel_do(size(output),input,rho,output,forward_erosion)
endfunction

% Function: deeplearning_erosion_backward
% Internal function for the processing performed by an erosion layer. Please refer to <get_erosion_layer> instead.
%
% See also:
%    <get_erosion_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_erosion_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    rho = int(parameters[0][0])
    dzdinput = zeros(size(input))
    %backpropagation is pretty easy: simply find the pixel responsible for the resulting value, and put the dzdoutput there
    function [] = __kernel__ backward_erosion(input,rho,dzdoutput,dzdinput,pos:ivec3)
        bestpos = [pos[0]-rho,pos[1]-rho,pos[2]]
        bestval = input[bestpos]
        for dx = -rho..rho
        for dy = -rho..rho
            thispos = [pos[0]+dx,pos[1]+dy,pos[2]]
            thisval = input[thispos]
            if thisval < bestval
                bestval = thisval
                bestpos = thispos
            endif
        endfor
        endfor
        dzdinput[bestpos] += dzdoutput[pos]
    endfunction
    parallel_do(size(dzdinput),input,rho,dzdoutput,dzdinput,backward_erosion)
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_erosion_layer
% Function for acquiring a pre-constructed erosion layer, optionally with a specified radius (defaults to 1, so a 3x3 dilation).
% To be used in conjunction with <add_layer>.
%
% : function [erode_layer:^DeepLearningLayer] = get_erosion_layer(radius:int=1)
%
% See also:
%    <add_layer>
%
function [erode_layer:^DeepLearningLayer] = get_erosion_layer(radius:int=1)
    init_param = Parameters(1)
    init_param[0] = [radius]
    
    erode_layer = DeepLearningLayer(name:="erosion",forward_pass:=deeplearning_erosion_forward,backward_pass:=deeplearning_erosion_backward,parameters:=init_param,dropout:=0.0)
endfunction

%Function: deeplearning_fully_connected_forward
% Internal function for the processing performed by a fully connected layer. Please refer to <get_fully_connected_layer> instead.
%
% See also:
%    <get_fully_connected_layer>
%
function [output:Data] = deeplearning_fully_connected_forward(input:Data,parameters:Parameters)
    filter = parameters[0]
    channels_in  = size(filter,2)
    channels_out = size(filter,3)
    
    output = zeros(1,1,channels_out)
    
    if channels_out > 1
        for c = 0..channels_out-1
            output[c] = sum(input.*filter[:,:,:,c])
        endfor
    else
        output = [sum(input .* filter[:,:,:,0])]
    endif
endfunction

% Function: deeplearning_fully_connected_backward
% Internal function for the processing performed by a fully connected layer. Please refer to <get_fully_connected_layer> instead.
%
% See also:
%    <get_fully_connected_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_fully_connected_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    w = parameters[0]
    channels_in  = size(w,2)
    channels_out = size(w,3)
    
    if channels_out > 1
        dzdinput = zeros(size(input))
        for c = 0..channels_out-1
            dzdinput = dzdinput + w[:,:,:,c]*dzdoutput[c]
        endfor
    else
        dzdinput = w[:,:,:,0]*dzdoutput[0]
    endif
    
    dzdw  = uninit(size(w))
    if channels_out > 1
        for c = 0..channels_out-1
            dzdw[:,:,:,c] = dzdoutput[c]*input
        endfor
    else
        dzdw = dzdoutput[0] * input[:,:,:]
    endif
    
    dzdparameters = Parameters(1)
    dzdparameters[0] = dzdw
endfunction

% Function: get_convolve_layer
% Function for acquiring a fully connected layer.
% To be used in conjunction with <add_layer>.
%
% : function [fully_connected_layer:^DeepLearningLayer] = get_fully_connected_layer(M:int,N:int,channels_in:int,channels_out:int)
%
% See also:
%    <add_layer>
%
function [fully_connected_layer:^DeepLearningLayer] = get_fully_connected_layer(M:int=0,N:int=0,channels_in:int=0,channels_out:int=0)
    init_param = Parameters(1)
    w:Parameter = randn(M,N,channels_in,channels_out)/10
    w = w - mean(w)
    init_param[0] = w
    
    fully_connected_layer = DeepLearningLayer(name:="fully connected",forward_pass:=deeplearning_fully_connected_forward,backward_pass:=deeplearning_fully_connected_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_convolve_forward
% Internal function for the processing performed by a convolution layer. Please refer to <get_convolve_layer> instead.
%
% See also:
%    <get_convolve_layer>
%
function [output:Data] = deeplearning_convolve_forward(input:Data,parameters:Parameters)
    w = parameters[0]
    channels_in  = size(w,2)
    channels_out = size(w,3)
    radius = int((size(w,0)-1)/2)
    
    if cudnn_version() == 0
    
        function [] = __kernel__ imfilter_kernel_nonsep(result,pos : ivec3)
            sum = 0.0
            for dy=-radius..radius
            for dx=-radius..radius
            for c_in = 0..channels_in-1
                sum += w[dy+radius,dx+radius,c_in,pos[2]] * input[pos[0]+dy,pos[1]+dx,c_in]
            endfor
            endfor
            endfor
        
            result[pos] = sum
        endfunction
        output = uninit(size(input,0),size(input,1),channels_out)
        parallel_do(size(output,0..2),output,imfilter_kernel_nonsep)
    
    else
        % Use the cuDNN implementation
        conv_desc = new(qdnn_conv_descriptor)
        conv_desc.pad_w = radius
        conv_desc.pad_h = radius
        input_c = cudnn_nchw_tensor(input)
        output_c = uninit(1,channels_out,size(input,0),size(input,1))
        w_c = shuffledims(w,[3,2,0,1])
        cudnn_convolution_forward(1.0, input_c, 0.0, output_c, w_c, conv_desc)
        output = cudnn_hwcn_tensor(output_c)
    endif
endfunction

% Function: deeplearning_convolve_backward
% Internal function for the processing performed by a convolution layer. Please refer to <get_convolve_layer> instead.
%
% See also:
%    <get_convolve_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_convolve_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    w = parameters[0]
    channels_in  = size(w,2)
    channels_out = size(w,3)
    radius = int((size(w,0)-1)/2)
    
    if cudnn_version() == 0
        dzdinput = uninit(size(input))
        dzdw = zeros(size(w))
        function [] = __kernel__ backward_convolution_1(dzdoutput:Data,dzdinput:Data,w:Parameter,radius:int,pos:ivec3)
            %find all pixels that contributed to this location and propagate the derivative
            summ = 0.0
            for dy = -radius..radius
            for dx = -radius..radius
            for c_out  = 0..channels_out-1
                summ += w[dy+radius,dx+radius,pos[2],c_out]*dzdoutput[pos[0]+dy,pos[1]+dx,c_out]
                dzdw[dy+radius,dx+radius,pos[2],c_out] += dzdoutput[pos[0],pos[1],c_out]*input[pos[0]+dy,pos[1]+dx,pos[2]]
            endfor
            endfor
            endfor
            dzdinput[pos] = summ
        endfunction
        parallel_do(size(input),dzdoutput,dzdinput,w,radius,backward_convolution_1)

        dzdparameters = Parameters(1)
        dzdparameters[0] = dzdw

    else
    
        conv_desc = new(qdnn_conv_descriptor)
        conv_desc.pad_w = radius
        conv_desc.pad_h = radius
        input_c = cudnn_nchw_tensor(input)
        dzdoutput_c = cudnn_nchw_tensor(dzdoutput)
        dzdinput = uninit(size(input_c))        
        dzdw = shuffledims(w,[3,2,0,1])        
        w_c = shuffledims(w,[3,2,0,1])
        cudnn_convolution_backward_filter(1.0,input_c,dzdoutput_c,0.0,dzdw,conv_desc)
        cudnn_convolution_backward_data(1.0,w_c,dzdoutput_c,0.0,dzdinput,conv_desc)
        dzdinput = cudnn_hwcn_tensor(dzdinput)
        dzdparameters = Parameters(1)
        dzdparameters[0] = shuffledims(dzdw,[2,3,1,0])       
    endif
    
endfunction

% Function: get_convolve_layer
% Function for acquiring a pre-constructed convolution layer, optionally with a specified radius (defaults to 1, so a 3x3 dilation), and number of input/output channels (defaults to 3/1)
% To be used in conjunction with <add_layer>.
%
% : function [convolve_layer:^DeepLearningLayer] = get_convolve_layer(radius:int=2,channels_in:int=3,channels_out:int=1)
%
% See also:
%    <add_layer>
%
function [convolve_layer:^DeepLearningLayer] = get_convolve_layer(radius:int=2,channels_in:int=3,channels_out:int=1)
    init_param = Parameters(1)
    w:Parameter = randn(2*radius+1,2*radius+1,channels_in,channels_out)/10
    w = w - mean(w)
    init_param[0] = w
    
    convolve_layer = DeepLearningLayer(name:="convolution",forward_pass:=deeplearning_convolve_forward,backward_pass:=deeplearning_convolve_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_biased_convolve_forward
% Internal function for the processing performed by a biased convolution layer. Please refer to <get_biased_convolve_layer> instead.
%
% See also:
%    <get_biased_convolve_layer>
%
function [output:Data] = deeplearning_biased_convolve_forward(input:Data,parameters:Parameters)
    filters = parameters[0]
    biases  = parameters[1]
    channels_in  = size(filters,2)
    channels_out = size(filters,3)
    
    if cudnn_version() == 0
        center = int((size(filters,0..1)-1)/2)
        function [] = __kernel__ imfilter_kernel_nonsep(result,pos : ivec3)
            sum = 0.0
            for m=0..size(filters,0)-1
            for n=0..size(filters,1)-1
            for c_in = 0..channels_in-1
                sum += filters[m,n,c_in,pos[2]] * input[clamp(pos[0] + m-center[0],size(input,0)-1),clamp(pos[1]+n-center[1],size(input,1)-1),c_in]
            endfor
            endfor
            endfor
        
            result[pos] = sum + biases[pos[2],0,0,0]
        endfunction
        output = uninit(size(input,0),size(input,1),channels_out)
        parallel_do(size(output,0..2),output,imfilter_kernel_nonsep)
    else
        % Use the cuDNN implementation
        radius = int((size(filters,0)-1)/2)
        conv_desc = new(qdnn_conv_descriptor)
        conv_desc.pad_w = radius
        conv_desc.pad_h = radius
        input_c = cudnn_nchw_tensor(input)
        output_c = uninit(1,channels_out,size(input,0),size(input,1))
        filters_c = shuffledims(filters,[3,2,0,1])
        cudnn_convolution_forward(1.0, input_c, 0.0, output_c, filters_c, conv_desc)
        output = cudnn_hwcn_tensor(output_c)    
        parallel_do(size(output,0..2),output,__kernel__ (output:Data, pos: ivec3) -> output[pos] = output[pos] + biases[pos[2],0,0,0])
    endif
endfunction

% Function: deeplearning_biased_convolve_backward
% Internal function for the processing performed by a biased convolution layer. Please refer to <get_biased_convolve_layer> instead.
%
% See also:
%    <get_biased_convolve_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_biased_convolve_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    ws = parameters[0]
    bs = parameters[1]
    channels_in  = size(ws,2)
    channels_out = size(ws,3)
    radius = int((size(ws,0)-1)/2)

    if cudnn_version() == 0
        dzdinput = zeros(size(input))
        dzdw  = zeros(size(ws))
        function [] = __kernel__ backward_convolution_1(dzdoutput:Data,dzdinput:Data,w:Parameter,radius:int,pos:ivec3)
            %find all pixels that contributed to this location and propagate the derivative
            summ = 0.0
            for dx = -radius..radius
            for dy = -radius..radius
            for c_out  = 0..channels_out-1
                summ += w[dx+radius,dy+radius,pos[2],c_out]*dzdoutput[pos[0]+dx,pos[1]+dy,c_out]
                dzdw[dy+radius,dx+radius,pos[2],c_out] += dzdoutput[pos[0],pos[1],c_out]*input[pos[0]+dy,pos[1]+dx,pos[2]]
            endfor
            endfor
            endfor
            dzdinput[pos] = summ
        endfunction
        parallel_do(size(input),dzdoutput,dzdinput,ws,radius,backward_convolution_1)
    
        dzdb = squeeze(sum(sum(dzdoutput,0),1))
        dzdparameters = Parameters(2)
        dzdparameters[0] = dzdw
        dzdparameters[1] = dzdb        
    else
    
        conv_desc = new(qdnn_conv_descriptor)
        conv_desc.pad_w = radius
        conv_desc.pad_h = radius
        input_c = cudnn_nchw_tensor(input)
        dzdoutput_c = cudnn_nchw_tensor(dzdoutput)
        dzdinput = uninit(size(input_c))        
        dzdw = shuffledims(ws,[3,2,0,1])        
        w_c = shuffledims(ws,[3,2,0,1])
        dzdb = zeros(channels_out)
        cudnn_convolution_backward_filter(1.0,input_c,dzdoutput_c,0.0,dzdw,conv_desc)
        cudnn_convolution_backward_data(1.0,w_c,dzdoutput_c,0.0,dzdinput,conv_desc)
        cudnn_convolution_backward_bias(1.0,dzdoutput_c,0.0,dzdb)
        dzdinput = cudnn_hwcn_tensor(dzdinput)
        dzdparameters = Parameters(2)
        dzdparameters[0] = shuffledims(dzdw,[2,3,1,0])   
        dzdparameters[1] = transpose(dzdb)
    endif
    

endfunction

% Function: get_biased_convolve_layer
% Function for acquiring a pre-constructed biased convolution layer, optionally with a specified radius (defaults to 1, so a 3x3 kernel), and number of input/output channels (defaults to 3/1)
% To be used in conjunction with <add_layer>.
%
% : function [biased_convolve_layer:^DeepLearningLayer] = get_biased_convolve_layer(radius:int=2,channels_in:int=3,channels_out:int=1)
%
% See also:
%    <add_layer>
%
function [biased_convolve_layer:^DeepLearningLayer] = get_biased_convolve_layer(radius:int=2,channels_in:int=3,channels_out:int=1)
    init_param = Parameters(2)
    w:Parameter = randn(2*radius+1,2*radius+1,channels_in,channels_out)/100
    init_param[0] = w - mean(w)
    init_param[1] = zeros(channels_out,1,1,1)
    
    biased_convolve_layer = DeepLearningLayer(name:="biased_convolution",forward_pass:=deeplearning_biased_convolve_forward,backward_pass:=deeplearning_biased_convolve_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: get_gabor_convolve_layer
% Function for acquiring a pre-constructed biased convolution layer, initialized with a gabor filter bank.
% The number of input channels is (obviously) free to choose, but the number of output channels is fixed at either 5 or 13 because of lazy programming.
% To be used in conjunction with <add_layer>.
%
% : function [gabor_convolve_layer:^DeepLearningLayer] = get_5_channel_gabor_convolve_layer(radius:int=5,channels_in:int=3, channels_out:int=13)
%
function [gabor_convolve_layer:^DeepLearningLayer] = get_gabor_convolve_layer(radius:int=5,channels_in:int=1,channels_out_factor:int=5)
    if channels_out_factor != 5 && channels_out_factor != 13
        error("Only gabor filter bank with 5 or 13 output factor is currently available")
    endif
    init_param = Parameters(2)
    %now initialize those filters!
    channels_out = channels_out_factor*channels_in
    w:Parameter = zeros(2*radius+1,2*radius+1,channels_in,channels_out)
    for i = 0..channels_in-1
        %bandwidth 0
        w[:,:,i,i*channels_out_factor+0 ] = gabor_kernel(radius,0,0)
        %bandwidth 1
        w[:,:,i,i*channels_out_factor+1 ] = gabor_kernel(radius,     0,1)
        w[:,:,i,i*channels_out_factor+2 ] = gabor_kernel(radius,  pi/4,1)
        w[:,:,i,i*channels_out_factor+3 ] = gabor_kernel(radius,  pi/2,1)
        w[:,:,i,i*channels_out_factor+4 ] = gabor_kernel(radius,3*pi/4,1)
        
        if channels_out > 5
        %bandwidth 2
        w[:,:,i,i*channels_out_factor+5 ] = gabor_kernel(radius,     0,2)
        w[:,:,i,i*channels_out_factor+6 ] = gabor_kernel(radius,  pi/8,2)
        w[:,:,i,i*channels_out_factor+7 ] = gabor_kernel(radius,  pi/4,2)
        w[:,:,i,i*channels_out_factor+8 ] = gabor_kernel(radius,3*pi/8,2)
        w[:,:,i,i*channels_out_factor+9 ] = gabor_kernel(radius,  pi/2,2)
        w[:,:,i,i*channels_out_factor+10]= gabor_kernel(radius,5*pi/8,2)
        w[:,:,i,i*channels_out_factor+11]= gabor_kernel(radius,3*pi/4,2)
        w[:,:,i,i*channels_out_factor+12]= gabor_kernel(radius,7*pi/8,2)
        endif
    endfor
    init_param[0] = w
    init_param[1] = zeros(channels_out,1,1,1)
    
    gabor_convolve_layer = DeepLearningLayer(name:="biased_convolution",forward_pass:=deeplearning_biased_convolve_forward,backward_pass:=deeplearning_biased_convolve_backward,parameters:=init_param,dropout:=0.0)
endfunction

function [filter:mat] = gabor_kernel(radius:int,orientation:scalar,bandwidth:scalar)
    gbw = radius/3
    A = [[cos(orientation),-sin(orientation)],[sin(orientation),cos(orientation)]] * [[1/gbw^2,0],[0,1/gbw^2]] * [[cos(orientation),sin(orientation)],[-sin(orientation),cos(orientation)]]
    k0 = bandwidth*[[cos(orientation)],[sin(orientation)]]
    xcs = ones(2*radius+1,1)*(-radius..radius)
    ycs = transpose(-radius..radius)*ones(1,2*radius+1)
    x : mat = vertcat(xcs[:],ycs[:])
    %filter_values = 1/2/pi/gbw^2 * exp(-(diag(transpose(x)*A*x))/2) .* exp(1i*transpose(k0)*x)
    filter_values = 1/2/pi/gbw^2 * exp(-(diag(transpose(x)*A*x))/2) .* exp(1i*complex(transpose(k0)*x))
    filter = real(reshape(filter_values,size(xcs)))
endfunction

% Function: visualize_convolve_layer
% Function for visualizing the filterbank of this convolutional layer - does not include the biases. Only valid for convolution and biased_convolution layers
%
% : function [] = visualize_convolve_layer(self:^DeepLearningLayer)
%
function [] = visualize_convolve_layer(self:^DeepLearningLayer)
    if self.name != "convolution" && self.name != "biased_convolution"
        error("This function call is only allowed for convolution layers!")
    endif
    
    %display the filter estimates
    [h,w,ci,co] = size(self.parameters[0],0..3)
    coy = floor(sqrt(co))
    if mod(co,coy)!=0
        coy = ceil(sqrt(co))
    endif
    if mod(co,coy)!=0
        coy = floor(co^(1/3))
    endif
    if mod(co,coy)!=0
        coy = ceil(co^(1/3))
    endif
    if mod(co,coy)!=0
        coy = 1
    endif
    cox = int(co / coy) %always integer due to the preceding code, but merits an explicit cast
    filter_image = zeros((h+2)*coy,(w+2)*cox,ci)
    for fx = 0..cox-1
    for fy = 0..coy-1
        tmpimg = self.parameters[0][0..h-1,0..w-1,0..ci-1,fx*coy+fy]
        tmpimg = (tmpimg-min(tmpimg))/(max(tmpimg)-min(tmpimg))
        filter_image[fy*(h+2)+(1..h),fx*(w+2)+(1..w),0..ci-1] = tmpimg
    endfor
    endfor
    imshow(imresize(filter_image,8,"nearest"))
endfunction

% Function: deeplearning_subsample_NN_forward
% Internal function for the processing performed by a nearest-neighbour subsampling layer. Please refer to <get_subsample_NN_layer> instead.
%
% See also:
%    <get_subsample_NN_layer>
%
function [output:Data] = deeplearning_subsample_NN_forward(input:Data,parameters:Parameters)
    factor = parameters[0][0]
    
    size_in = size(input,0..1)
    channels = size(input,2)
    size_out = floor(size_in/factor)
    
    output = uninit(size_out[0],size_out[1],channels)
    function [] = __kernel__ subsample_NN(pos:ivec2)
        ups = int((pos+0.5)*factor)
        for c = 0..channels-1
            output[pos[0],pos[1],c] = input[ups[0],ups[1],c]
        endfor
    endfunction
    parallel_do(size_out,subsample_NN)
endfunction

% Function: deeplearning_subsample_NN_backward
% Internal function for the processing performed by a nearest-neighbour subsampling layer. Please refer to <get_subsample_NN_layer> instead.
%
% See also:
%    <get_subsample_NN_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_subsample_NN_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    factor = parameters[0][0]
    
    channels = size(input,2)
    dzdinput = zeros(size(input))
    %backpropagation is pretty easy: simply find the pixel responsible for the resulting value, and put the dzdoutput there
    function [] = __kernel__ backward_subsample_NN(pos:ivec2)
        ups = int((pos+0.5)*factor)
        for c = 0..channels-1
            dzdinput[ups[0],ups[1],c] = dzdoutput[pos[0],pos[1],c]
        endfor
    endfunction
    parallel_do(size(dzdoutput,0..1),backward_subsample_NN)
    
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_subsample_NN_layer
% Function for acquiring a pre-constructed nearest-neighbour subsampling layer, optionally with a specified factor (defaults to 2).
% To be used in conjunction with <add_layer>.
%
% : function [subsample_layer:^DeepLearningLayer] = get_subsample_NN_layer(factor:int=2)
%
% See also:
%    <add_layer>
%
function [subsample_layer:^DeepLearningLayer] = get_subsample_NN_layer(factor:int=2)
    init_param = Parameters(1)
    init_param[0] = [factor]
    
    subsample_layer = DeepLearningLayer(name:="subsample_NN",forward_pass:=deeplearning_subsample_NN_forward,backward_pass:=deeplearning_subsample_NN_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_upsample_NN_forward
% Internal function for the processing performed by a nearest-neighbour upsampling layer. Please refer to <get_upsample_NN_layer> instead.
%
% See also:
%    <get_upsample_NN_layer>
%
function [output:Data] = deeplearning_upsample_NN_forward(input:Data,parameters:Parameters)
    factor = parameters[0][0]
    
    size_in = size(input,0..1)
    channels = size(input,2)
    size_out = floor(size_in*factor)
    
    output = uninit(size_out[0],size_out[1],channels)
    function [] = __kernel__ upsample_NN(pos:ivec2)
        ups = int(pos/factor)
        for c = 0..channels-1
            output[pos[0],pos[1],c] = input[ups[0],ups[1],c]
        endfor
    endfunction
    parallel_do(size_out,upsample_NN)
endfunction

% Function: deeplearning_upsample_NN_backward
% Internal function for the processing performed by a nearest-neighbour upsampling layer. Please refer to <get_upsample_NN_layer> instead.
%
% See also:
%    <get_upsample_NN_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_upsample_NN_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    factor = parameters[0][0]
    
    channels = size(input,2)
    dzdinput = zeros(size(input))
    %backpropagation is pretty easy: simply find the pixel responsible for the resulting value, and put the dzdoutput there
    function [] = __kernel__ backward_upsample_NN(pos:ivec2)
        ups = int(pos/factor)
        for c = 0..channels-1
            dzdinput[ups[0],ups[1],c] += dzdoutput[pos[0],pos[1],c]
        endfor
    endfunction
    parallel_do(size(dzdoutput,0..1),backward_upsample_NN)
    
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_upsample_NN_layer
% Function for acquiring a pre-constructed nearest-neighbour upsampling layer, optionally with a specified factor (defaults to 2).
% To be used in conjunction with <add_layer>.
%
% : function [upsample_layer:^DeepLearningLayer] = get_upsample_NN_layer(factor:int=2)
%
% See also:
%    <add_layer>
%
function [upsample_layer:^DeepLearningLayer] = get_upsample_NN_layer(factor:int=2)
    init_param = Parameters(1)
    init_param[0] = [factor]
    
    upsample_layer = DeepLearningLayer(name:="upsample_NN",forward_pass:=deeplearning_upsample_NN_forward,backward_pass:=deeplearning_upsample_NN_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_sigmoid_forward
% Internal function for the processing performed by a sigmoid activation layer. Please refer to <get_sigmoid_layer> instead.
%
% See also:
%    <get_upsample_NN_layer>
%
function [output:Data] = deeplearning_sigmoid_forward(input:Data,parameters:Parameters)
    output = 1./(1+exp(-input))
endfunction

% Function: deeplearning_sigmoid_backward
% Internal function for the processing performed by a sigmoid activation layer. Please refer to <get_sigmoid_layer> instead.
%
% See also:
%    <get_upsample_NN_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_sigmoid_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    dzdinput = dzdoutput.*output.*(1-output)
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_sigmoid_layer
% Function for acquiring a pre-constructed sigmoid activation layer.
% To be used in conjunction with <add_layer>.
%
% : function [sigmoid_layer:^DeepLearningLayer] = get_sigmoid_layer()
%
% See also:
%    <add_layer>
%
function [sigmoid_layer:^DeepLearningLayer] = get_sigmoid_layer()
    init_param = Parameters(1)
    init_param[0] = [0.0]
    
    sigmoid_layer = DeepLearningLayer(name:="sigmoid",forward_pass:=deeplearning_sigmoid_forward,backward_pass:=deeplearning_sigmoid_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_tanh_forward
% Internal function for the processing performed by a hyperbolic tangent activation layer. Please refer to <get_tanh_layer> instead.
%
% See also:
%    <get_tanh_layer>
%
function [output:Data] = deeplearning_tanh_forward(input:Data,parameters:Parameters)
    output = (exp(input)-exp(-input))./(exp(input)+exp(-input))
endfunction

% Function: deeplearning_tanh_backward
% Internal function for the processing performed by a hyperbolic tangent activation layer. Please refer to <get_tanh_layer> instead.
%
% See also:
%    <get_tanh_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_tanh_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    dzdinput = dzdoutput.*(1-output.^2)
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_tanh_layer
% Function for acquiring a pre-constructed hyperbolic tangent activation layer.
% To be used in conjunction with <add_layer>.
%
% : function [sigmoid_layer:^DeepLearningLayer] = get_tanh_layer()
%
% See also:
%    <add_layer>
%
function [tanh_layer:^DeepLearningLayer] = get_tanh_layer()
    init_param = Parameters(1)
    init_param[0] = [0.0]
    
    tanh_layer = DeepLearningLayer(name:="tanh",forward_pass:=deeplearning_tanh_forward,backward_pass:=deeplearning_tanh_backward,parameters:=init_param,dropout:=0.0)
endfunction

% Function: deeplearning_RELU_forward
% Internal function for the processing performed by a rectified linear unit (ReLU) activation layer. Please refer to <get_RELU_layer> instead.
%
% See also:
%    <get_RELU_layer>
%
function [output:Data] = deeplearning_RELU_forward(input:Data,parameters:Parameters)
    output = max(input,0)
endfunction

% Function: deeplearning_RELU_backward
% Internal function for the processing performed by a rectified linear unit (ReLU) activation layer. Please refer to <get_RELU_layer> instead.
%
% See also:
%    <get_RELU_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_RELU_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    dzdinput = dzdoutput.*(input>0)
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_RELU_layer
% Function for acquiring a pre-constructed rectified linear unit (ReLU) activation layer.
% To be used in conjunction with <add_layer>.
%
% : function [RELU_layer:^DeepLearningLayer] = get_RELU_layer()
%
% See also:
%    <add_layer>
%
function [RELU_layer:^DeepLearningLayer] = get_RELU_layer()
    init_param = Parameters(1)
    init_param[0] = [0.0]
    
    RELU_layer = DeepLearningLayer(name:="ReLU",forward_pass:=deeplearning_RELU_forward,backward_pass:=deeplearning_RELU_backward,parameters:=init_param,dropout:=0.0)
endfunction 

% Function: deeplearning_leaky_RELU_forward
% Internal function for the processing performed by a leaky rectified linear unit (ReLU) activation layer. Please refer to <get_leaky_RELU_layer> instead.
%
% See also:
%    <get_leaky_RELU_layer>
%
function [output:Data] = deeplearning_leaky_RELU_forward(input:Data,parameters:Parameters)
    leakage = parameters[0][0]
    output = max(input,input*leakage)
endfunction

% Function: deeplearning_leaky_RELU_backward
% Internal function for the processing performed by a leaky rectified linear unit (ReLU) activation layer. Please refer to <get_leaky_RELU_layer> instead.
%
% See also:
%    <get_leaky_RELU_layer>
%
function [dzdinput:Data,dzdparameters:Parameters] = deeplearning_leaky_RELU_backward(input:Data,output:Data,parameters:Parameters,dzdoutput:Data)
    leakage = parameters[0][0]
    dzdinput = dzdoutput.*((input>0)*(1-leakage)+leakage)
    dzdparameters = Parameters(1)
    dzdparameters[0] = [0.0]
endfunction

% Function: get_leaky_RELU_layer
% Function for acquiring a pre-constructed leaky rectified linear unit (ReLU) activation layer.
% This leaky ReLU will not train its leakage parameter - it is set statically.
% To be used in conjunction with <add_layer>.
%
% : function [leaky_RELU_layer:^DeepLearningLayer] = get_leaky_RELU_layer()
%
% See also:
%    <add_layer>
%
function [leaky_RELU_layer:^DeepLearningLayer] = get_leaky_RELU_layer(leakage:scalar=0.0)
    init_param = Parameters(1)
    init_param[0] = [leakage]
    
    leaky_RELU_layer = DeepLearningLayer(name:="leaky ReLU",forward_pass:=deeplearning_leaky_RELU_forward,backward_pass:=deeplearning_leaky_RELU_backward,parameters:=init_param,dropout:=0.0)
endfunction 

% Function: initDeepLearningNetwork
% Function for acquiring a correctly initialized DeepLearningNetwork.
% The maximum number of layers the network can have is (currently) fixed at initialization time, and handed as an optional parameter.
% After initialization, the loss function still needs to be set with <set_loss_function>, and layers need to be added with <add_layer>.
%
% :  function [network:DeepLearningNetwork] = initDeepLearningNetwork(max_layers:int=20)
%
% See also:
%    <add_layer>,<set_loss_function>,<train>,<process>,<save_to_file>,<load_from_file>
%
function [network:DeepLearningNetwork] = initDeepLearningNetwork(max_layers:int=20)
    default_settings = DeepLearningSettings(rate:=1e-3,momentum_first_order:= 0.9,momentum_second_order:=0.999,plot_period:= 10)
    network = DeepLearningNetwork(loss:=deeplearning_empty_loss,loss_name:="empty",layers:=DeepLearningLayers(max_layers),settings:=default_settings,nr_layers:=0)
endfunction

% Function: set_loss_function
% Function for setting the loss function used in an initialized DeepLearningNetwork.
% Currently, alternatives are either binary classification ("binary") or MSE to the groundtruth ("MSE")
%
% : function [] = set_loss_function(self:^DeepLearningNetwork,name:string)
%
% See also:
%    <initDeepLearningNetwork>
%
function [] = set_loss_function(self:^DeepLearningNetwork,name:string)
    self.loss_name = name
    match name with
    | "MSE" -> loss_info = deeplearning_MSE_loss
    | "MAE" -> loss_info = deeplearning_MAE_loss
    | "binary" -> loss_info = deeplearning_binary_loss
    |_ -> error(sprintf("The loss function %s is currently not implemented",name))
    endmatch
    self.loss = loss_info
endfunction

% Function: add_layer
% Function for adding a new layer to an initialized DeepLearningNetwork.
%
% :  function [] = add_layer(self:^DeepLearningNetwork,layer:^DeepLearningLayer)
%
% Currently, the following layers are implemented:
%   - <get_dilation_layer>
%   - <get_erosion_layer>
%   - <get_convolve_layer>
%   - <get_fully_connected_layer>
%   - <get_subsample_NN_layer>
%   - <get_upsample_NN_layer>
%   - <get_sigmoid_layer>
%   - <get_tanh_layer>
%   - <get_RELU_layer>
%
% See also:
%    <initDeepLearningNetwork>
%
function [] = add_layer(self:^DeepLearningNetwork,layer:^DeepLearningLayer)
    self.layers[self.nr_layers] = layer
    self.nr_layers = self.nr_layers + 1
endfunction
