import "system.q"
import "denoise_blsgsm.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Restoration/Wavelets"}

% Function: denoise_band_vectorprobshrink
%
% Denoising of a wavelet subband using the vector-Probshrink method.
%
% B. Goossens, A. Pizurica and W. Philips, "Removal of Correlated Noise by Modeling the 
% Signal of Interest in the Wavelet Domain," in IEEE Transactions on Image Processing, 
% vol 18 (6), p.1153--1165, june 2009.
%
% Parameters:
% x - subband to denoise (both input & output)
% wnd - the size of the local window
% C_n - noise covariance matrix
% sigma_n - the noise standard deviation
% T - threshold defining the signal of interest
%
function [] = denoise_band_vectorprobshrink(x, wnd, C_n, sigma_n, T : scalar)

    % The central coefficient in the window
    d = prod(wnd)
    c = floor(d/2)

    % Prior distribution for z & the number of mixture components
    K : int'const = 4
    [z, p_z] = bkf_prior(K, 0.7) % TODO - estimate tau from the image

    if isscalar(sigma_n)
        C_n = C_n * sigma_n.^2
    else
        sigma2_max = max(sigma_n.^2)
        C_n = C_n * sigma2_max
    endif
    C_y = compute_covmtx_spat_stationary(x, wnd)
    C_x = correct_eigenvalues(C_y - C_n)

    % Symmetric matrix square root of the noise covariance
    [S, S_inv] = symsqrt_and_inv(C_n)

    % Compute S^{-1}C_{x}S^{-T} -> SVD Q Lambda Q^T
    [Q, LambdaE, Q2] = svd(S_inv*C_x*transpose(S_inv))
    Lambda = diag(LambdaE)

    % Correct eigenvalues if necessary
    epsilon = 1e-2
    Lambda = max(epsilon*ones(d), diag(LambdaE))

    % Compute U=(SQ)^{-1}; this transform will decorrelate the observed noisy samples
	% The matrix inverse can be replaced by a multiplication
    U = transpose(Q) * S_inv

    % Compute the probability that the signal-of-interest is absent (p(H_0)),
    % which is constant for the whole subband
    p_H0 = 0
    p_H0_comp = zeros(1,K)
    {!interpreted for}
    for k=0..K-1
        p_H0_comp[k] = exp(0.5*sum(log(T) - log(T + Lambda * z[k])))
        p_H0 = p_H0 + p_H0_comp[k] * p_z[k]
    endfor

    % Compute normalisation factors for the densities - use logarithm again
    % to avoid numerical underflows in later steps
    rho = zeros(1,K)
    tau = zeros(1,K)
    rho_gn = zeros(K,d)
    tau_gn = zeros(K,d)
    {!interpreted for}
    for k=0..K-1
        r = z[k] * Lambda

        rho[k] = -0.5*sum(log(1+r)) + log(p_z[k]) + 4*d
        tau[k] =  0.5*sum(log((T+r)./(T+((1+T)*r)))) + log(p_z[k]) + log(p_H0_comp[k]) + 4*d
        rho_gn[k,:] = -0.5./(1+r)
        tau_gn[k,:] = -0.5*(T+r)./(T+((1+T)*r))
    endfor

    if isscalar(sigma_n)
        % Estimator for stationary noise
        function [] = __kernel__ probshrink_estimator(x : mat'circular, y : mat'unchecked, U : mat'unchecked, 
            rho_gn : mat'unchecked, tau_gn : mat'unchecked, rho : vec'unchecked(K), tau : vec'unchecked(K), _
            wnd : ivec2, pos : ivec2, blkdim : ivec2, blkpos : ivec2)
            
            d = wnd[0]*wnd[1]
            c = pos - int(wnd/2)
            K = numel(rho)
            assert(d<=9)
            v = shared(blkdim[0],blkdim[1],d)

            % decorrelate the observed coefficient vector
            for m=0..d-1
                sum = 0.0
                for n=0..d-1
                    p = [n/wnd[1],mod(n,wnd[1])] + c
                    sum += U[m,n] * x[p]
                endfor
                v[blkpos[0],blkpos[1],m] = sum
            endfor

            % compute the likelihood p(y|z_k)
            p_y = 0.0
            p_yH0 = 0.0
            
            for k=0..K-1
                sum1 = rho[k] % related to f(y)
                sum2 = tau[k] % related to f(y|H_0)
                for i=0..d-1
                    val = v[blkpos[0],blkpos[1],i]
                    val *= val               
                    sum1 += val*rho_gn[k,i]
                    sum2 += val*tau_gn[k,i]
                endfor
                p_y += exp(sum1)
                p_yH0 += exp(sum2)
            endfor
            
            p_H0_y = p_yH0/p_y
            if p_H0_y > 1.0 || isnan(p_H0_y)
                p_H0_y = 1.0        
            endif
            y[pos] = x[pos] * (1.0-p_H0_y)  % shrinkage!        
        endfunction
        
    else
        function [] = __kernel__ probshrink_estimator_nonstat(x : mat'circular, y : mat'unchecked, U : mat'unchecked, 
            V_n : mat'unchecked, T : scalar, z : vec, Lambda : vec, rho : vec'unchecked, tau : vec'unchecked, wnd : ivec2, _
            vscale : ivec2, sigma2_max : scalar, pos : ivec2, blkdim : ivec2, blkpos : ivec2)
            
            d = wnd[0]*wnd[1]
            c = pos - int(wnd/2)
            K = numel(rho)
            assert(d<=9)
            v = shared(blkdim[0],blkdim[1],d)
            v_sq = max(V_n[pos.*vscale].^2 / sigma2_max,1e-4)

            % decorrelate the observed coefficient vector
            for m=0..d-1
                sum = 0.0
                for n=0..d-1
                    p = [n/wnd[1],mod(n,wnd[1])] + c
                    sum += U[m,n] * x[p]
                endfor
                v[blkpos[0],blkpos[1],m] = sum
            endfor

            % compute the likelihood p(y|z_k)
            p_y = 0.0
            p_yH0 = 0.0
            
            for k=0..K-1
                sum1 = rho[k] % related to f(y)
                sum2 = tau[k] % related to f(y|H_0)
                for i=0..d-1
                    val = v[blkpos[0],blkpos[1],i]
                    val *= val         
                    r = z[k] * Lambda[i]      
                    rho_gn = -0.5./(v_sq+r)
                    tau_gn = -0.5*(v_sq*T+r)./(v_sq*(v_sq*T+((1+T)*r)))
                    sum1 += val*rho_gn
                    sum2 += val*tau_gn
                endfor
                p_y += exp(sum1)
                p_yH0 += exp(sum2)
            endfor
            
            p_H0_y = p_yH0/p_y
            if p_H0_y > 1.0 || isnan(p_H0_y)
                p_H0_y = 1.0        
            endif
            y[pos] = x[pos] * (1.0-p_H0_y)  % shrinkage!        
        endfunction
    endif
    
    y = copy(x)
    
    if isscalar(sigma_n)
        parallel_do([size(x,0..2),max_block_size([16,32,1])],y,x,U,rho_gn,tau_gn,rho,tau,wnd,probshrink_estimator)
    else
        vscale = size(sigma_n,0..1) ./ size(x,0..1) % subsample in case the dimensions don't match
        parallel_do([size(x,0..2),max_block_size([16,32,1])],y,x,U,sigma_n,T,z,Lambda,rho,tau,wnd,vscale, _
            sigma2_max,probshrink_estimator_nonstat)    
    endif
endfunction

% Function: denoise_image_vectorprobshrink
%
% Denoises an image using the vector Probshrink method. Both grayscale and RGB color images are
% supported.
%
% :function y = denoise_image_vectorprobshrink(img_noisy, sigma, sparsity_tf="dtcwt", J=5, K=8)
%
% Parameters:
% img_noisy - noisy input image
% sigma - the noise standard deviation
% sparsity_tf - the transform used for achieving sparsity
%      "shearlet": the bandlimited discrete shearlet transform
%      "wavelet": the orthogonal wavelet transform
%      "dtcwt": the dual-tree complex wavelet transform
%      "steerable pyramid": steerable pyramids
% J - the number of analysis scales for the transform
% K - the number of analysis orientations for the transform (shearlet and steerable
%     pyramids only)
function y = denoise_image_vectorprobshrink(img_noisy, sigma, sparsity_tf="dtcwt", J=5, K=8, T=2.64)

    sz = size(img_noisy)
    match sparsity_tf with
    | "shearlet" -> 
            [S, S_H] = build_energyweighted_dst2d(sz, K, J)
    | "wavelet"  -> 
            [S, S_H] = build_dwt(db8, J)
    | "dtcwt"    -> 
            [S, S_H] = build_dtcwt(filtercoeff_farras, filtercoeff_dualfilt, J)
    | "steerable pyramid"      -> 
            [S, S_H] = build_stp(sz, K, J)
    | _          -> 
            error("Sparsity transform unknown!")
    endmatch

    w = S(img_noisy)
    w2 = S(randn(size(img_noisy)))

    wnd = [3,3]
    nc = size(img_noisy, 2)

    bands = lincell(w)
    bands2 = lincell(w2)

    for j=0..numel(bands)-5
        % Grayscale images
        if nc == 1
            C_n = compute_covmtx_spat_stationary(bands2[j], wnd) 
            denoise_band_vectorprobshrink(bands[j], wnd, C_n, sigma, T)
        elseif nc == 3
        % Color images
            for k=0..2
                C_n = compute_covmtx_spat_stationary(bands2[j][:,:,k], wnd)
                w_slice = bands[j][:,:,k]
                denoise_band_vectorprobshrink(w_slice, wnd, C_n, sigma, T)
                bands[j][:,:,k] = w_slice
            endfor
        endif
    endfor
    y = S_H(w)
    
endfunction

%
% The main function - for testing the code
%
function [] = main()

    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))

    img = imread("lena_big.tif")[:,:,1]

    %sigma = 25
    sigma = 40 * transpose(0..size(img,0)-1) * ones(1,size(img,1)) / size(img,0)
    img_noisy = img + sigma .* randn(size(img))
    J = 4 % number of scales
    K = 8 % number of orientations
    %T = 2.64
    T = 8.1 % Set T a bit higher in case of nonstationary noise (parameter tuning)

    tic()
    %[S, S_H] = build_dst2d([size(img,0),size(img,1),1], K, J)
    [S, S_H] = build_dtcwt(filtercoeff_farras, filtercoeff_dualfilt, J)
    %[S, S_H] = build_improved_dtcwt(filtercoeff_farras, filtercoeff_dualfilt, J)
    w = S(img_noisy)

    w2 = S(randn(size(img_noisy)))
    bands = lincell(w)
    bands2 = lincell(w2)

    wnd = [3,3]
    nc = size(img_noisy, 2)

    for j=0..numel(bands)-5
        % Grayscale images
        if nc == 1
            C_n = compute_covmtx_spat_stationary(bands2[j], wnd) 
            denoise_band_vectorprobshrink(bands[j], wnd, C_n, sigma, T)
        elseif nc == 3
        % Color images
            for k=0..2
                C_n = compute_covmtx_spat_stationary(bands2[j][:,:,k], wnd)
                w_slice = bands[j][:,:,k]
                denoise_band_vectorprobshrink(w_slice, wnd, C_n, sigma, T)
                bands[j][:,:,k] = w_slice
            endfor
        endif
    endfor

    y = S_H(w)
    toc()
    
    print "psnr=",psnr(y,img)," dB"

    imshow (img_noisy); title("Noisy image")  
    imshow (y); title("Denoised image")

endfunction