% dst.q
% Feb 2011 - Bart Goossens
import "system.q"
import "dwt.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Multiresolution"}

% Section: dst.q
%

% Function: pseudo_polargrid
% Constructs of a pseudo-polar grid
%
% :[r, theta] = pseudo_polargrid(M : int, N : int)
% 
% Parameters:
% M - the height of the grid (# samples in the y-direction)
% N - the width of the grid (# samples in the x-direction)
% r - the radial components
% theta - the angular components
function [r, theta] = pseudo_polargrid(M : int, N : int)
    [xr, yr] = meshgrid((1..N)-N/2-1,((1..M)-M/2-1)*(N/M))
    r = max(abs(xr), abs(yr))
    theta = -atan2(yr,xr)

    tri = (z : mat) -> z.*(z<pi/2) + (pi-z).*(z>=pi/2)
    theta += sin(2*tri(2*theta-pi/4)-pi/2)/14; 
endfunction

% Function: wedge_filter
% Computes shearlet wedge filters
%
% :function G = wedge_filter(sz,K,bw)
%
% Parameters:
% sz - the size of the input image
% K - the number of orientations
% bw - the transition bandwidth (pi/32..pi/2)
function G = wedge_filter(sz,K:scalar,bw:scalar)
    if numel(sz)==2
        [M,N,C] = [sz[0],sz[1],1]
    else
        [M,N,C] = sz[0..2]
    endif
    [r, theta] = pseudo_polargrid(M,N)
    v = x -> 3*x.^2-2*x.^3

    ntaps = 1024
    phi = (0..ntaps-1)*(pi*K/ntaps)
    h = (phi < 2*bw) .* sin(v(phi/(2*bw))*pi/2) + _
                    (phi >= 2*bw) .* (phi < pi) + _
                    (phi >= pi) .* (phi <= pi+2*bw) .* cos(v((phi-pi)/(2*bw))*pi/2)
    phi = phi/K
    G = vec[cube](K)
    U = 0
    for k=0..K-2
        theta0 = mod((3*pi)/4 + (k+1)*(pi/K) + 0*bw/K + theta, pi*ones(M,N))
        G[k] = interp1(phi, h, theta0)
        % tricks to end up with a realvalued impulse response
        G[k][0..M-1,0] = 1/sqrt(K)
        G[k][0,0..N-1] = 1/sqrt(K) 
        G[k] = repmat(fftshift2(G[k]),[1,1,C])
        U += G[k].^2        
    endfor  
    G[K-1]=sqrt(max(0,1 - U))     
endfunction

% Function: wedge_filter_set
% Computes shearlet wedge filter for different scales. Used when the number of
% orientations depends on the scale
%
% :function G = wedge_filter(sz,K,bw)
%
% Parameters:
% sz - the size of the input image
% K - a vector with the number of orientations per scale
% bw - the transition bandwidth (pi/32..pi/2)
function G = wedge_filter_set(sz,K:vec,bw:scalar)

    G = cell(numel(K))
    for j=0..numel(K)-1
        G[j] = wedge_filter(sz,K[j],bw)
    endfor

endfunction

function H=radial_filter_cascade(H0, G0, J, sz)
    % cascade algorithm for subsequent scales
    N = numel(H0)
    H=cell(J+1)
    H[0] = copy(G0)
    lo = copy(H0)
    for j=1..J-1
        H[j] = copy(lo)
        r=0..N/(2^j)-1
        H[j][r] = H[j][r] .* G0[0..(2^j)..N-1]        
        lo[r]   = lo[r]   .* H0[0..(2^j)..N-1]
        lo[N/(2^j)..N-1] = 0 
    endfor
    H[J]=lo

    % convert to 2D frequency responses
    [r, theta] = pseudo_polargrid(sz[0],sz[1])
    r = (r/(sz[1]/2))*N
    for j=0..J
        H[j] = repmat(fftshift2(interp1(0..N-1,H[j],r)),[1,1,sz[2]])
    endfor
endfunction

% Function: meyer_radial_wavelet_filters
% Computes Meyer radial wavelet filters used in the implementation
% of the discrete shearlet transform.
%
% :function H=meyer_radial_wavelet_filters(sz,J)
%
% Parameters:
% sz - the size of the input images
% J - the number of scales of the transform
function H=meyer_radial_wavelet_filters(sz,J)

    N=256    
    omega=(0..N-1)*(pi/N)
    v = x -> 3*x.^2-2*x.^3
    
    H0 = (omega < pi/2) + _
         cos((pi/2)*v(2*(omega/pi)-1)).*((pi/2)<=omega).*(omega<=pi)
    G0 = sin((pi/2)*v(2*(omega/pi)-1)).*((pi/2)<=omega).*(omega<=pi) + _
         (omega > pi)

    H = radial_filter_cascade(H0, G0, J, sz)
endfunction

% Function: radial_wavelet_filters
% Computes shearlet radial filters based on compactly supported
% wavelets. The function calculates the DTFT of the filters in order
% to obtain the frequency response, which is transformed to 2D using
% the pseudopolar grid.
%
% :function H=meyer_radial_wavelet_filters(sz,J)
%
% Parameters:
% w - the wavelet to be used (db1, db2, ...)
% sz - the size of the input images
% J - the number of scales of the transform
function H=radial_wavelet_filters(w,sz,J)
    N=256    
    % we perform a frequency correction, for easier comparison to Meyer wavelets
    omega=min((0..N-1)*(pi/N), pi)

    H0 = complex(zeros(N))
    G0 = complex(zeros(N))
    for k=0..numel(w)-1
        z = exp(1j*omega*k)
        H0 = H0 + w[k] * z
        G0 = G0 + w[k] * z * (-1)^k
    endfor
    H0 = H0 / sqrt(2)
    G0 = G0 / sqrt(2)
    H = radial_filter_cascade(abs(H0), abs(G0), J, sz)
endfunction

% Function: subsample_fft
%
% Computes a subsampling operation in FFT domain
%
% :subsample_fft(x, D)
%
% Parameters:
% x - DFT input coefficients
% y - DFT output coefficients
% D - subsampling factor
function y = subsample_fft(x, D)
    sz : ivec3 = [size(x,0)/D,size(x,1)/D,size(x,2)]
    function [] = __kernel__ subsample_fft_kernel(y : ccube, x : ccube, D : int, sz : ivec3, pos : ivec3)
        y_sum = 0i
        for m=0..D-1
            for n=0..D-1
                y_sum += x[m*sz[0]+pos[0],n*sz[1]+pos[1],pos[2]]
            endfor
        endfor
        y[pos] = y_sum/float(D)
    endfunction
    y = complex(zeros(sz))
    parallel_do(sz,y,x,D,sz,subsample_fft_kernel)
%    r0 = 0..sz[0]-1
%    r1 = 0..sz[1]-1
%    y = 0
%    for m=0..D-1
%        for n=0..D-1
%            y += x[m*sz[0]+r0,n*sz[1]+r1,:]/D^2
%        end
%    end
endfunction

% Function: upsample_fft
%
% Computes an upsampling operation in FFT domain
%
% :function y = upsample_fft(x, D)
%
% Parameters:
% x - DFT input coefficients
% y - DFT output coefficients
% D - upsampling factor
function y : ccube = upsample_fft(x : ccube, D)
    y = repmat(x*D, [D,D,1])    
endfunction

function y : cube = upsample_fft(x : cube, D)
    y = repmat(x*D, [D,D,1])   
endfunction

% Function: pyramid_fft
% 
% Pyramid decomposition in FFT domain
%
% :function w = pyramid_fft(x, af, df)
%
% Parameters:
% x - DFT coefficients of the input image
% af - DFT coefficients of the analysis/synthesis filters
% df - the decimation factors for each filter
function w = pyramid_fft(x, af, df)
    w = cell(numel(af))

    % simply multiply the DFT of the input image with the 
    % frequency response of the filter response for scale j
    for j=0..numel(af)-1
        w[j] = x .* af[j]
        if df[j]>1
            w[j] = subsample_fft(w[j],df[j])
        endif
    endfor
endfunction

% Function: ipyramid_fft
%
% Pyramid reconstruction in FFT domain
%
% :function y = ipyramid_fft(w, af, df)
%
% Parameters:
% w - FFT pyramid coefficients
% af - DFT coefficients of the analysis/synthesis filters
% df - the upsampling factors for each filter
function y = ipyramid_fft(w, af, df)
    y = complex(0)
    for j=0..numel(af)-1
        % upsample if necessary
        if df[j]==1
            y += w[j] .* af[j]
        else
            y += upsample_fft(w[j], df[j]) .* af[j]
        endif
    endfor
endfunction

% Function: shearlet_analysis
% 
% Computes the discrete shearlet transform
%
% :function w = shearlet_analysis(x, J, K, af, G, decimate=1)
%
% Parameters:
% x  - input image (real-valued or complex-valued)
% J  - number of scales
% K  - number of orientations
% af - analysis wavelet filters
% G - analysis shearing filters
% decimate - perform radial decimation (to reduce the redunancy without
%      loss of information).
function w = shearlet_analysis(x, J, K, af, G, decimate=1)
    sz = size(x)
    if decimate
        df = [1,2,4,8,16,32,64]  % decimation factors    
    else
        df = ones(J+1)
    endif
    if any(frac(size(x,0..1)/(2*df[J-1])) != 0)
        error "Image dimensions are not divisible by ",2*df[J-1], "! Reduce the number of scales."
    endif

    if !iscomplex(x)
        w_pyr = pyramid_fft(fft2(x),af,df)

        w = cell(J+1) % cell array for the shearlet subbands    

        for j=0..J-1
            r0 = 0..df[j]..sz[0]-1
            r1 = 0..df[j]..sz[1]-1
            if isscalar(K)
                w[j] = cell(K)
                if df[j]==1
                    for k=0..cast(K,int)-1
                        w[j][k]=real(ifft2(G[k].*w_pyr[j]))
                    endfor
                else
                    for k=0..cast(K,int)-1
                        w[j][k]=real(ifft2(G[k][r0,r1,:].*w_pyr[j]))
                    endfor
                endif
            else
                w[j] = cell(K[j])
                if df[j]==1
                    for k=0..K[j]-1
                        w[j][k]=real(ifft2(G[j][k].*w_pyr[j]))
                    endfor
                else
                    for k=0..K[j]-1
                        w[j][k]=real(ifft2(G[j][k][r0,r1,:].*w_pyr[j]))
                    endfor
                endif
            endif
        endfor
        % scaling coefficients
        w[J]=real(ifft2(w_pyr[J]))
    else
        w_pyr = pyramid_fft(fft2(x),af,df)
        w = cell(J+1) % cell array for the shearlet subbands    

        for j=0..J-1
            w[j] = cell(K)
            r0 = 0..df[j]..sz[0]-1
            r1 = 0..df[j]..sz[1]-1
            nK = isscalar(K) ? K : K[j]
            if df[j]==1
                for k=0..nK-1
                    w[j][k]=ifft2(G[k].*w_pyr[j])
                endfor
            else
                for k=0..nK-1
                    w[j][k]=ifft2(G[k][r0,r1,:].*w_pyr[j])
                endfor
            endif
        endfor
        % scaling coefficients
        w[J]=ifft2(w_pyr[J])
    endif
endfunction

% Function: shearlet_synthesis
%
% Computes the inverse discrete shearlet transform
%
% :function y = shearlet_synthesis(w, J, K, af, G, decimate=1)
%
% w  - shearlet/scaling coefficients
% J  - number of scales
% K  - number of orientations
% af - synthesis wavelet filters
% G - synthesis shearing filters
% decimate - shearlet coefficients are computed with radial decimation 
function y : [cube|ccube] = shearlet_synthesis(w, J, K, af, G, decimate=1)
    if decimate
        df = [1,2,4,8,16,32,64]  % decimation factors    
    else
        df = ones(J+1)
    endif
    sz = size(w[0][0])
    w_pyr = cell(J+1)

    % scaling coefficients
    w_pyr[J] = fft2(w[J])

    for j=0..J-1
        band = complex(0)
        r0 = 0..df[j]..sz[0]-1
        r1 = 0..df[j]..sz[1]-1
        
        if isscalar(K)
            if df[j]==1
                for k=0..cast(K,int)-1
                    band += fft2(w[j][k]) .* G[k]
                endfor
            else
                for k=0..cast(K,int)-1
                    band += fft2(w[j][k]) .* G[k][r0,r1,:]
                endfor
            endif
        else
            if df[j]==1
                for k=0..K[j]-1
                    band += fft2(w[j][k]) .* G[j][k]
                endfor
            else
                for k=0..K[j]-1
                    band += fft2(w[j][k]) .* G[j][k][r0,r1,:]
                endfor
            endif
        endif
        w_pyr[j] = band
    endfor
    if !iscomplex(w[J])
        y = real(ifft2(ipyramid_fft(w_pyr,af,df)))
    else
        y = ifft2(ipyramid_fft(w_pyr,af,df))
    endif
endfunction

% Function: shearlet_compute_energy_per_subband
%
% Computes the energy per subband, useful for subband normalization.
%
% :function e = shearlet_compute_energy_per_subband(J, K, af, G)
%
% Parameters:
% J  - number of scales
% K  - number of orientations
% af - synthesis wavelet filters
% G - synthesis shearing filters
function e = shearlet_compute_energy_per_subband(J, K, af, G)
    df = [1,1,2,4,8,16,32,64]
    sz = size(G[0])
    e = cell(J+1)
    e_sum = 0
    for j=0..J-1
        r0 = 0..df[j]..sz[0]-1
        r1 = 0..df[j]..sz[1]-1
        if df[j]==1
            hi = af[j] .* G[0]
        else
            hi = af[j][r0,r1,:] .* G[0][r0,r1,:]            
        endif
        e[j] = cell(K)
        e[j][0] = sqrt(mean(abs(hi).^2))
        for k=1..K-1
            e[j][k]=e[j][0]
        endfor
        e_sum += K*sum(e[j][0].^2)
    endfor
    r0 = 0..df[J-1]..sz[0]-1
    r1 = 0..df[J-1]..sz[1]-1
    e[J] = sqrt(mean(abs(af[j][r0,r1,:]).^2))
    e_sum += e[J].^2
    %print "Sum=", e_sum
endfunction

%========================================

function [] = main()

    img = imread("lena_big.tif")
    %img = imread("lighthouse.png")
    img = img[:,:,1]
   % img = img[0..255,0..255,:]
   %img = 128+128*randn(512,512)

    bw = pi/2
    K=[8,4,4,4]
    J=4

    sz = size(img,0..2)
    
    print "Begin filter computation"
    % computation of the filters
    tic()
    G = wedge_filter_set(sz,K,bw)
    %H = meyer_radial_wavelet_filters(sz,J)
    H = radial_wavelet_filters(db8,sz,J)
    toc()

    print "End filter computation"
    %imshow(fftshift2(abs(H[0])),[])
    %imshow(fftshift2(abs(H[1])),[])
    %imshow(fftshift2(abs(H[2])),[])

    %imshow(fftshift2(abs(G[0])),[])
    %return
    %en = shearlet_compute_energy_per_subband(J,K,H,G)
    
    tic()
    for k=0..0
        w = shearlet_analysis(img,J,K,H,G,0)
        img_out = shearlet_synthesis(w,J,K,H,G,0)
    endfor
    toc()

    %imshow(w[2][4],[])
    %imshow(fftshift2(abs(H[1])),[])
    imshow(img_out,[])

%    for j=0..J-1
%        w = shearlet_analysis(zeros(size(img)),J,K,H,G)
%        sz = size(w[j][1])
%        w[j][1][sz[0]/2,sz[1]/2] = 1
%        img_out = shearlet_synthesis(w,J,K,H,G)
%        %imshow(img_out,[])
%        imshow(fftshift2(abs(fft2(img_out))),[])
%    end       

    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    img_psnr = psnr(img,img_out)
    print "psnr=", img_psnr, "dB"
    assert(img_psnr >= 130, "PSNR goal was not reached!")
    print "abs err=", max(abs(img-img_out))/255
    print "max freq err=", max(abs(fft2(img-img_out)))/sqrt(numel(img))
    imshow(abs(fftshift2(fft2(img-img_out))))

endfunction
