% Section: dtcwt_improveddirectionalselectivity.q
%
% Implementation of the DT-CWT with improved directional selectivity
% for the first scale according to the technique described in:
%   B. Goossens, A. Pizurica and W. Philips, "A Filter Design Technique for 
%   Improving the Directional Selectivity of the First Scale of the Dual-Tree 
%   Complex Wavelet Transform," in Proc. of the IEEE Int. Conf. Image Processing 
%   (ICIP2009), Nov. 7-11, Cairo, Egypt
%   This technique is implemented here by applying an allpass filter to the
%   first-scale coefficients (before the complex phase modulation step)
%
import "system.q"
import "dtcwt.q"
import "imfilter.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Multiresolution"}

% Function: dtcwt_fswavelet
%
% Computation of first scale wavelets for the dual-tree complex
% wavelet transform
%
% :y = dtcwt_fswavelet(x = db2)
%
function y = dtcwt_fswavelet(x = db2)
    n = numel(x)
    y = zeros(2,n+2)
    y[1,1..n] = x
    y[0,0..n-1] = x 
end

% Function: dtcwt_fracshifting_response
%
% Calculates the impulse response of a fractional delay allpass filter
%
% Parameters:
% M - the number of taps
% alpha - fractional delay parameter
%
% Returns:
% q  - fractional delay filter kernel
% q_origin - origin of the fractional delay filter kernel 
function [q, q_origin] = dtcwt_fracshifting_response(M : int, alpha : scalar)
    m = -1..M-2
    q = alpha.^m * (1-alpha.^2)
    q[0] = -alpha
    q_origin = 1
end

% Function: allpass_filter_band
%
% Applies a 1D allpass filter to the specified wavelet subband (with
% circular boundary extension)
function [] = allpass_filter_band(x, dim, q, q_origin, band_yx, band_sz)
    r1 = band_yx[0]+(0..band_sz[0]-1)
    r2 = band_yx[1]+(0..band_sz[1]-1)
    if dim == 1 % filtering along the rows (horizontally)
        x[r1,r2,:] = imfilter(x[r1,r2,:], q, [0, q_origin], "circular")
    else
        x[r1,r2,:] = imfilter(x[r1,r2,:], transpose(q), [q_origin, 0], "circular")
    endif
end

% Function: dtcwt2d_fulloriented
% Dual-tree complex wavelet transform with improved directional selectivity
%
% :function bands = dtcwt2d_fulloriented(x, w1, w2, J, q, q_origin)
%
% Parameters:
% x  - input image
% w1 - filters for the first scale
% w2 - filters for the other scales
% J  - number of scales
% q  - fractional delay filter kernel
% q_origin - origin of the fractional delay filter kernel
function bands = dtcwt2d_fulloriented(x, w1 : vec[wavelet_params], w2 : vec[wavelet_params], J, q, q_origin)
    b = cell(2,2)
    [M,N] = size(x,0..1)/2
    q = fliplr(q)
    q_origin = numel(q)-1-q_origin

    % four DWTs in parallel
    for m=0..1
        for n=0..1
            b[m,n] = copy(x)
            % first scale
            dwt2d_untiled_rowcol(b[m,n], size(x,0..2), w1[m], w1[n], 1)

            % allpass filtering (after decimation)
            if n==1
                allpass_filter_band(b[m,n], 1, q, q_origin, [0,N], [M,N]) % band 1 - LH
            endif
            if m==1
                allpass_filter_band(b[m,n], 0, q, q_origin, [M,0], [M,N]) % band 2 - HL
            endif
            if n==1
                allpass_filter_band(b[m,n], 1, q, q_origin, [M,N], [M,N]) % band 3 - HH
            endif
            if m==1
                allpass_filter_band(b[m,n], 0, q, q_origin, [M,N], [M,N]) % band 3 - HH
            endif

            % other scales
            dwt2d_untiled_rowcol(b[m,n], [size(x,0)/2,size(x,1)/2,size(x,2)],w2[m], w2[n], J-1)
        end
    end
    % complex phase modulation
    b = dtcwt2d_complexphasemod(b, J)

    % extract subbands = TODO - this step significantly affects performance...
    bands = cell(J+1)
    sz = size(x,0..2)
    bands[J] = cell(2,2)
    for j=0..J-1
        sz = [sz[0]/2,sz[1]/2,sz[2]]
        bands[j] = cell(2,2)
        for m=0..1
            for n=0..1
                y = b[m,n]
                bands[j][m,n] = cell(3)
                bands[j][m,n][0] = y[sz[0]..2*sz[0]-1,0..sz[1]-1,:]
                bands[j][m,n][1] = y[0..sz[0]-1,sz[1]..2*sz[1]-1,:]
                bands[j][m,n][2] = y[sz[0]..2*sz[0]-1,sz[1]..2*sz[1]-1,:]       
                if j==J-1
                    bands[J][m,n] = y[0..sz[0]-1,0..sz[1]-1,:]
                endif
            end
        end
    end
end

% Function: idtcwt2d_fulloriented
% Inverse dual-tree complex wavelet transform with improved directional selectivity
%
% :function y = idtcwt2d_fulloriented(bands, w1, w2, J, q, q_origin)
%
% Parameters:
% x  - input image
% w1 - filters for the first scale
% w2 - filters for the other scales
% J  - number of scales
% q  - fractional delay filter kernel
% q_origin - origin of the fractional delay filter kernel
function y = idtcwt2d_fulloriented(bands, w1 : vec[wavelet_params], w2 : vec[wavelet_params], J, q, q_origin)
    % generate bands
    b = cell(2,2)
    for m=0..1
        for n=0..1
            sz = size(bands[J][m,n],0..2)            
            y = zeros([sz[0]*2^J, sz[1]*2^J, sz[2]])
            y[0..sz[0]-1,0..sz[1]-1,:] = bands[J][m,n] 
            for j=J-1..-1..0
                y[sz[0]..2*sz[0]-1,0..sz[1]-1,:] = bands[j][m,n][0]
                y[0..sz[0]-1,sz[1]..2*sz[1]-1,:] = bands[j][m,n][1]
                y[sz[0]..2*sz[0]-1,sz[1]..2*sz[1]-1,:] = bands[j][m,n][2]
                sz = [sz[0]*2,sz[1]*2,sz[2]]
            end
            b[m,n] = y
        end
    end

    % complex phase modulation
    b = dtcwt2d_complexphasemod(b, J)
    [M,N] = size(b[0,0],0..1)/2

    y = 0
    % four DWTs in parallel
    for m=0..1
        for n=0..1
            u = copy(b[m,n])

            % allpass filtering (after decimation)
            if n==1
                allpass_filter_band(u, 1, q, q_origin, [0,N], [M,N]) % band 1 - LH
            endif
            if m==1
                allpass_filter_band(u, 0, q, q_origin, [M,0], [M,N]) % band 2 - HL
            endif
            if n==1
                allpass_filter_band(u, 1, q, q_origin, [M,N], [M,N]) % band 3 - HH
            endif
            if m==1
                allpass_filter_band(u, 0, q, q_origin, [M,N], [M,N]) % band 3 - HH
            endif

            % scales J..2
            idwt2d_untiled_rowcol(u, [size(u,0)/2^J,size(u,1)/2^J,size(u,2)], w2[m], w2[n], J-1)

            % first scale
            idwt2d_untiled_rowcol(u, [size(u,0)/2,size(u,1)/2,size(u,2)], w1[m], w1[n], 1)
            y = y + 0.25 * u
        end
    end
end

function [] = main()

    x = zeros(256,256)
    %x = imread("lena_big.tif")
    J = 1   % number of scales
    w1 = complex_orthogonal_wavelet(dtcwt_fswavelet(sym2))  % wavelet for the first scale
    w2 = complex_orthogonal_wavelet(filtercoeff_selcw[3,1]) % wavelet for the other scales

    [q, q_origin] = dtcwt_fracshifting_response(10, 0.564717683715766)
    print q
    %x[0,:] = 1
    %allpass_filter_band(x, 0, q, q_origin, [0,0], size(x))
    %colormap("gray")
    %imshow(x,[])
    %return

    %print dtcwt_fracshifting_response(16, 0.5647)
    %return
    w = dtcwt2d_fulloriented(x, w1, w2, J, q, q_origin)

    band = w[0][1,1][1]
    band[size(band,0)/2,size(band,1)/2]=1
    y = idtcwt2d_fulloriented(w, w1, w2, J, q, q_origin)
    band[size(band,0)/2,size(band,1)/2]=0

    imshow(y,[])
    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    print psnr(x,y)
    assert(type(y,"cube"))
    y = log(1+abs(fftshift2(fft2(y)))[:,:,0])
    assert(type(y,"mat"))
    imshow(y,[])
end