%============================================================
% dwt.q
% Definition & implementation of discrete wavelet transforms
%============================================================

{!author name="Bart Goossens"}
{!doc category="Image Processing/Multiresolution"}

% Section: dwt.q
% Computation of 1D and 2D discrete wavelet transforms

% Constants: Wavelets
%
% haar - Analysis scaling coefficients for the haar wavelet
% db2 - Analysis scaling coefficients for Daubechies' wavelet of length 4 (2 vanishing moments)
% db3 - Analysis scaling coefficients for Daubechies' wavelet of length 6 (3 vanishing moments)
% db4 - Analysis scaling coefficients for Daubechies' wavelet of length 8 (4 vanishing moments)
% db5 - Analysis scaling coefficients for Daubechies' wavelet of length 10 (5 vanishing moments)
% db6 - Analysis scaling coefficients for Daubechies' wavelet of length 12 (6 vanishing moments)
% db7 - Analysis scaling coefficients for Daubechies' wavelet of length 14 (7 vanishing moments)
% db8 - Analysis scaling coefficients for Daubechies' wavelet of length 16 (8 vanishing moments)
% db10 - Analysis scaling coefficients for Daubechies' wavelet of length 20 (10 vanishing moments)
% db12 - Analysis scaling coefficients for Daubechies' wavelet of length 24 (12 vanishing moments)
% db14 - Analysis scaling coefficients for Daubechies' wavelet of length 28 (14 vanishing moments)
% db16 - Analysis scaling coefficients for Daubechies' wavelet of length 32 (16 vanishing moments)
% sym2 - Analysis scaling coefficients for Daubechies' symmlet of length 4 (2 vanishing moments)
% sym3 - Analysis scaling coefficients for Daubechies' symmlet of length 6 (3 vanishing moments)
% sym4 - Analysis scaling coefficients for Daubechies' symmlet of length 8 (4 vanishing moments)
% sym5 - Analysis scaling coefficients for Daubechies' symmlet of length 10 (5 vanishing moments)
% sym6 - Analysis scaling coefficients for Daubechies' symmlet of length 12 (6 vanishing moments)
% sym7 - Analysis scaling coefficients for Daubechies' symmlet of length 14 (7 vanishing moments)
% sym8 - Analysis scaling coefficients for Daubechies' symmlet of length 16 (8 vanishing moments)
% sym10 - Analysis scaling coefficients for Daubechies' symmlet of length 20 (10 vanishing moments)
% sym12 - Analysis scaling coefficients for Daubechies' symmlet of length 24 (12 vanishing moments)
% sym14 - Analysis scaling coefficients for Daubechies' symmlet of length 28 (14 vanishing moments)
% sym16 - Analysis scaling coefficients for Daubechies' symmlet of length 32 (16 vanishing moments)
% coif1 - Coiflet of length 1
% coif2 - Coiflet of length 2
% coif3 - Coiflet of length 3
% coif4 - Coiflet of length 4
% coif5 - Coiflet of length 5
% all_wavelets - A cell array containing all wavelets defined in this module
% all_wavelet_names - A cell array containing the names of all wavelets defined in this module
%
haar = [0.707106781187,0.707106781187]
db2 = [-0.129409522551,0.224143868042,0.836516303737,0.482962913145]
db3 = [0.035226291882,-0.085441273882,-0.135011020010,0.459877502119,0.806891509313,0.332670552951]
db4 = [-0.010597401785,0.032883011667,0.030841381836,-0.187034811719,-0.027983769417,0.630880767930,0.714846570553,0.230377813309]
db5 = [0.003335725285,-0.012580751999,-0.006241490213,0.077571493840,-0.032244869585,-0.242294887066,0.138428145901,0.724308528439,0.603829269797,0.160102397974]
db6 = [-0.001077301085,0.004777257511,0.000553842201,-0.031582039318,0.027522865530,0.097501605587,-0.129766867567,-0.226264693965,0.315250351709,0.751133908022,0.494623890398,0.111540743350]
db7 = [0.000353713800,-0.001801640704,0.000429577973,0.012550998556,-0.016574541631,-0.038029936935,0.080612609151,0.071309219267,-0.224036184994,-0.143906003929,0.469782287405,0.729132090847,0.396539319482,0.077852054085]
db8 = [-0.000117476784,0.000675449406,-0.000391740373,-0.004870352993,0.008746094047,0.013981027917,-0.044088253931,-0.017369301002,0.128747426620,0.000472484574,-0.284015542962,-0.015829105256,0.585354683655,0.675630736298,0.312871590914,0.054415842243]
db10 = [-0.000013264203,0.000093588670,-0.000116466855,-0.000685856695,0.001992405295,0.001395351747,-0.010733175483,0.003606553567,0.033212674059,-0.029457536822,-0.071394147166,0.093057364604,0.127369340336,-0.195946274377,-0.249846424326,0.281172343660,0.688459039453,0.527201188931,0.188176800078,0.026670057901]
db12 = [-0.000001529072,0.000012776952,-0.000024241546,-0.000088504109,0.000388653063,0.000006545128,-0.002179503619,0.002248607241,0.006711499009,-0.012840825198,-0.012218649070,0.041546277495,0.010849130256,-0.096432120097,0.005359569674,0.182478605928,-0.023779257256,-0.316178453753,-0.044763885654,0.515886478428,0.657198722579,0.377355135214,0.109566272821,0.013112257957]
db14 = [-0.000000178714,0.000001724995,-0.000004389705,-0.000010337209,0.000068755043,-0.000041777246,-0.000386831947,0.000708021154,0.001061691086,-0.003849638868,-0.000746218989,0.012789493266,-0.005615049530,-0.030185351540,0.026981408308,0.055237126259,-0.071548955504,-0.086748411568,0.139989016584,0.138395213865,-0.218033529993,-0.271688552279,0.218670687759,0.631187849105,0.554305617941,0.254850267793,0.062364758849,0.006461153460]
db16 = [-0.000000021093,0.000000230878,-0.000000736366,-0.000001043571,0.000011336609,-0.000013945669,-0.000061035966,0.000174787245,0.000114241520,-0.000941021749,0.000407896981,0.003128023381,-0.003644279621,-0.006990014563,0.013993768860,0.010297659641,-0.036888397692,-0.007588974369,0.075924236044,-0.006239722752,-0.132388305564,0.027340263753,0.211190693947,-0.027918208133,-0.327063310528,-0.089751089402,0.440290256886,0.637356332084,0.430312722846,0.165064283489,0.034907714324,0.003189220925]
sym2 = [-0.129409522551,0.224143868042,0.836516303737,0.482962913145]
sym3 = [0.035226291882,-0.085441273882,-0.135011020010,0.459877502119,0.806891509313,0.332670552951]
sym4 = [-0.075765714789,-0.029635527646,0.497618667632,0.803738751806,0.297857795605,-0.099219543577,-0.012603967262,0.032223100604]
sym5 = [0.027333068345,0.029519490926,-0.039134249302,0.199397533977,0.723407690402,0.633978963458,0.016602105765,-0.175328089908,-0.021101834025,0.019538882735]
sym6 = [0.015404109327,0.003490712084,-0.117990111148,-0.048311742586,0.491055941927,0.787641141030,0.337929421728,-0.072637522786,-0.021060292512,0.044724901771,0.001767711864,-0.007800708325]
sym7 = [0.002681814568,-0.001047384889,-0.012636303403,0.030515513166,0.067892693501,-0.049552834937,0.017441255087,0.536101917092,0.767764317003,0.288629631752,-0.140047240443,-0.107808237704,0.004010244872,0.010268176709]
sym8 = [-0.003382415951,-0.000542132332,0.031695087811,0.007607487325,-0.143294238351,-0.061273359068,0.481359651258,0.777185751701,0.364441894835,-0.051945838108,-0.027219029917,0.049137179674,0.003808752014,-0.014952258337,-0.000302920515,0.001889950333]
sym10 = [0.000770159809,0.000095632671,-0.008641299277,-0.001465382581,0.045927239231,0.011609893904,-0.159494278885,-0.070880535783,0.471690666938,0.769510037021,0.383826761067,-0.035536740474,-0.031990056882,0.049994972077,0.005764912034,-0.020354939812,-0.000804358932,0.004593173585,0.000057036084,-0.000459329421]
sym12 = [0.000111967194,-0.000011353928,-0.001349755756,0.000180214090,0.007414965518,-0.001408909244,-0.024220722675,0.007553780612,0.049179318300,-0.035848830737,-0.022162306170,0.398885972390,0.763479097784,0.462741031219,-0.078332622316,-0.170370697239,0.015301740622,0.057804179446,-0.002604391031,-0.014589836449,0.000307647796,0.002350297614,-0.000018158079,-0.000179066587]
sym14 = [-0.000025879090,0.000011210866,0.000398435673,-0.000062865425,-0.002579441726,0.000366476574,0.010037693718,-0.002753774791,-0.029196217764,0.004280520499,0.037433088363,-0.057634498351,-0.035318112115,0.393201521962,0.759976241961,0.475335762634,-0.058111823318,-0.159997411147,0.025898587531,0.069827616362,-0.002365048837,-0.019439314264,0.001013141987,0.004532677472,-0.000073214214,-0.000605760182,0.000019329017,0.000044618978]
sym16 = [0.000006230007,-0.000003113556,-0.000109431479,0.000028078582,0.000852354711,-0.000108445622,-0.003880912253,0.000718211979,0.012666731660,-0.003126517172,-0.031051202844,0.004869274405,0.032333091611,-0.066983049071,-0.034574228418,0.397122933620,0.756524987877,0.475342806013,-0.054040601387,-0.159592192186,0.030721139063,0.078037852904,-0.003510275068,-0.024952758046,0.001359844742,0.006937761131,-0.000222116476,-0.001338720607,0.000036565925,0.000165456796,-0.000005396483,-0.000010797982]
coif1 = [-0.015655728135,-0.072732619513,0.384864846864,0.852572020212,0.337897662458,-0.072732619513]
coif2 = [-0.000720549445,-0.001823208871,0.005611434819,0.023680171946,-0.059434418646,-0.076488599078,0.417005184422,0.812723635446,0.386110066821,-0.067372554722,-0.041464936782,0.016387336464]
coif3 = [-0.000034599773,-0.000070983303,0.000466216960,0.001117518771,-0.002574517689,-0.009007976137,0.015880544864,0.034555027573,-0.082301927107,-0.071799821619,0.428483476378,0.793777222626,0.405176902410,-0.061123390003,-0.065771911282,0.023452696142,0.007782596427,-0.003793512864]
coif4 = [-0.000001784985,-0.000003259680,0.000031229876,0.000062339034,-0.000259974552,-0.000589020756,0.001266561929,0.003751436157,-0.005658286687,-0.015211731528,0.025082261845,0.039334427123,-0.096220442034,-0.066627474263,0.434386056491,0.782238930920,0.415308407030,-0.056077313317,-0.081266699681,0.026682300156,0.016068943965,-0.007346166328,-0.001629492013,0.000892313669]
coif5 = [-0.000000095177,-0.000000167443,0.000002063762,0.000003734655,-0.000021315027,-0.000041340432,0.000140541150,0.000302259582,-0.000638131343,-0.001662863702,0.002433373213,0.006764185448,-0.009164231162,-0.019761778943,0.032683574267,0.041289208750,-0.105574208703,-0.062035963963,0.437991626172,0.774289603653,0.421566206691,-0.052043163176,-0.091920010560,0.028168028971,0.023408156786,-0.010131117520,-0.004159358781,0.002178236358,0.000358589688,-0.000212080840]

all_wavelets = {haar,db2,db3,db4,db5,db6,db7,db8,db10,db12,db14,db16,sym2,sym3,sym4,sym5, _
                sym6,sym7,sym8,sym10,sym12,sym14,sym16,coif1,coif2,coif3,coif4,coif5}
all_wavelet_names = {"haar","db2","db3","db4","db5","db6","db7","db8","db10","db12","db14","db16","sym2","sym3","sym4","sym5", _
                     "sym6","sym7","sym8","sym10","sym12","sym14","sym16","coif1","coif2","coif3","coif4","coif5"}

type wavelet_params : dynamic class
    filter_coeff : mat(4,:)
    ctd : int
    ctr : int
endtype

% Forward transform along the columns...
function [] = __kernel__ dwt_dim0(x : cube'unchecked, y : cube'unchecked, wc : mat'unchecked'const, ctd : int, n : int, pos : ivec3)
    K = 16*n + ctd
    a = 0.0
    b = 0.0
    tilepos = int((2*pos[0])/n)
    j0 = tilepos*n
    for k=0..(size(wc,1)-1)
        j = j0+mod(2*pos[0]+k+K,n)
        u = x[j,pos[1],pos[2]]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[j0+mod(pos[0],int(n/2)),    pos[1],pos[2]]=a
    y[j0+int(n/2)+mod(pos[0],int(n/2)),pos[1],pos[2]]=b
endfunction

% Forward transform along the rows...
function [] = __kernel__ dwt_dim1(x : cube'unchecked, y : cube'unchecked, wc : mat'unchecked'const, ctd : int, n : int, pos : ivec3)
    K = 16*n + ctd
    a = 0.0
    b = 0.0
    tilepos = int((2*pos[1])/n)
    j0 = tilepos*n
    for k=0..(size(wc,1)-1)
        j = j0+mod(2*pos[1]+k+K,n)
        u = x[pos[0],j,pos[2]]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[pos[0],j0+mod(pos[1],int(n/2)),    pos[2]]=a
    y[pos[0],j0+int(n/2)+mod(pos[1],int(n/2)),pos[2]]=b
endfunction

% Forward transform along the third dimension...
function [] = __kernel__ dwt_dim2(x : cube'unchecked, y : cube'unchecked, wc : mat'unchecked, ctd : int, n : int, pos : ivec3)
    K = 16*n + ctd
    a = 0.0
    b = 0.0
    tilepos = int((2*pos[2])/n)
    j0 = tilepos*n
    for k=0..(size(wc,1)-1)
        j = j0+mod(2*pos[2]+k+K,n)
        u = x[pos[0],pos[1],j]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[pos[0],pos[1],j0+mod(pos[2],int(n/2))]=a
    y[pos[0],pos[1],j0+int(n/2)+mod(pos[2],int(n/2))]=b
endfunction

% Inverse transform along the columns...
function [] = __kernel__ idwt_dim0(x : cube'unchecked, y : cube'unchecked, wc : mat'unchecked'const, ctr : int, n : int, pos : ivec3)
    K = 16*n + ctr
    a = 0.0
    b = 0.0
    j0 = (pos[0]-mod(pos[0],n))*2
    for k=0..(int(size(wc,1)/2)-1)
        j = j0+mod(pos[0]+k+K,n)
        u = x[j,pos[1],pos[2]]
        v = x[n+j,pos[1],pos[2]]
        a = a + wc[2,2*k  ] * u + wc[3,2*k  ] * v
        b = b + wc[2,2*k+1] * u + wc[3,2*k+1] * v
    endfor
    y[pos[0]*2  ,pos[1],pos[2]] = b
    y[pos[0]*2+1,pos[1],pos[2]] = a
endfunction

% Inverse transform along the rows... 
function [] = __kernel__ idwt_dim1(x : cube'unchecked, y : cube'unchecked, wc : mat'unchecked'const, ctr : int, n : int, pos : ivec3)
    K = 16*n + ctr
    a = 0.0
    b = 0.0
    j0 = (pos[1]-mod(pos[1],n))*2
    for k=0..(int(size(wc,1)/2)-1)
        j = j0 + mod(pos[1]+k+K,n)
        u = x[pos[0],j,pos[2]]
        v = x[pos[0],n+j,pos[2]]

        a = a + wc[2,2*k  ] * u + wc[3,2*k  ] * v
        b = b + wc[2,2*k+1] * u + wc[3,2*k+1] * v
    endfor
    y[pos[0],pos[1]*2  ,pos[2]] = b
    y[pos[0],pos[1]*2+1,pos[2]] = a
endfunction

% Inverse transform along the third dimension... 
function [] = __kernel__ idwt_dim2(x : cube'unchecked, y : cube'unchecked, wc : mat'unchecked, ctr : int, n : int, pos : ivec3)
    K = 16*n + ctr
    a = 0.0
    b = 0.0
    j0 = (pos[2]-mod(pos[2],n))*2
    for k=0..(int(size(wc,1)/2)-1)
        j = j0 + mod(pos[2]+k+K,n)
        u = x[pos[0],pos[1],j]
        v = x[pos[0],pos[1],n+j]

        a = a + wc[2,2*k  ] * u + wc[3,2*k  ] * v
        b = b + wc[2,2*k+1] * u + wc[3,2*k+1] * v
    endfor
    y[pos[0],pos[1],pos[2]*2  ] = b
    y[pos[0],pos[1],pos[2]*2+1] = a
endfunction

% Function: orthogonal_wavelet
%
% Computes filter coefficients for an orthogonal wavelet
%
% :function [p : wavelet_params] = orthogonal_wavelet(w)
%
% Parameters:
% w - orthogonal wavelet (see <Wavelets>)
% p.filter_coeff - matrix with wavelet filter coefficients
% o row 0 - analysis scaling filter coefficients
% o row 1 - analysis wavelet filter coefficients
% o row 2 - synthesis scaling filter coefficients
% o row 3 - synthesis wavelet filter coefficients
% p.ctd - origin of the analysis filter
% p.ctr - origin of the synthesis filter
function [p : wavelet_params] = orthogonal_wavelet(w : vec)
    n = numel(w)
    wc = zeros(4,n)
    % lpd, hpd, lpr, hpr    
    wc[0,:] = w 
    wc[1,:] = w[n-1..(-1)..0] .* (-1).^(0..n-1)
    wc[2,:] = w[n-1..(-1)..0]
    wc[3,:] = wc[1,n-1..(-1)..0]

    if mod(n/2,2)==0
       ctd = -floor(n/2)
    else
       ctd = -floor(n/2)-1
    endif
    ctr = 1-floor(n/4)
    p = wavelet_params(filter_coeff:=wc,ctd:=ctd,ctr:=ctr)
endfunction

% Function: dwt2d_untiled
%
% Untiled inplace version of the <dwt2d>
%
% :function [] = dwt2d_untiled(x, w : wavelet_params, J)
%
% Parameters:
% x - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
function [] = dwt2d_untiled(x, w : wavelet_params, J)

    sz = size(x,0..2)
    if mod(sz[0], 2^J) != 0 || mod(sz[1], 2^J) != 0
        error "dwt2d: input matrix should have dimensions that are a multiple of ",2^J
    endif

    z = uninit(size(x))
    for j=1..J
        parallel_do([sz[0],sz[1]/2,sz[2]],x,z,w.filter_coeff,w.ctd,sz[1],dwt_dim1)
        parallel_do([sz[0]/2,sz[1],sz[2]],z,x,w.filter_coeff,w.ctd,sz[0],dwt_dim0)

        sz = [sz[0]/2,sz[1]/2,sz[2]]
    endfor
endfunction

% Function: idwt2d_untiled
%
% Untiled inplace version of the <idwt2d>
%
% :function [] = idwt2d_untiled(y, w : wavelet_params, J)
%
% Parameters:
% y - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
function [] = idwt2d_untiled(y, w : wavelet_params, J)
    sz = [size(y,0)/2.^J, size(y,1)/2.^J, size(y,2)]

    z = uninit(size(y))
    for j=1..J                
        parallel_do([sz[0]*2,sz[1],sz[2]],y,z,w.filter_coeff,w.ctr,sz[1],idwt_dim1)
        parallel_do([sz[0],sz[1]*2,sz[2]],z,y,w.filter_coeff,w.ctr,sz[0],idwt_dim0)       

        sz = [sz[0]*2,sz[1]*2,sz[2]]
    endfor
endfunction

% Function: dwt3d_untiled
%
% Untiled inplace version of the <dwt3d>
%
% :function [] = dwt3d_untiled(x, w : wavelet_params, J)
%
% Parameters:
% x - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
function [] = dwt3d_untiled(x, w : wavelet_params, J)

    sz = size(x,0..2)
    if mod(sz[0], 2^J) != 0 || mod(sz[1], 2^J) != 0 || mod(sz[2], 2^J) != 0
        error "dwt3d: input matrix should have dimensions that are a multiple of ",2^J
    endif

    z = uninit(size(x))
    y = uninit(size(x))    
    for j=1..J
        parallel_do([sz[0],sz[1],sz[2]/2],x,y,w.filter_coeff,w.ctd,sz[2],dwt_dim2)
        parallel_do([sz[0],sz[1]/2,sz[2]],y,z,w.filter_coeff,w.ctd,sz[1],dwt_dim1)
        parallel_do([sz[0]/2,sz[1],sz[2]],z,x,w.filter_coeff,w.ctd,sz[0],dwt_dim0)
        sz /= 2
    endfor
endfunction

% Function: idwt3d_untiled
%
% Untiled inplace version of the <idwt3d>
%
% :function [] = idwt3d_untiled(y, w : wavelet_params, J)
%
% Parameters:
% y - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
function [] = idwt3d_untiled(y, w : wavelet_params, J)
    sz = size(y,0..2)/2.^J

    z = uninit(size(y))
    x = uninit(size(y))
    for j=1..J                
        parallel_do([sz[0]*2,sz[1]*2,sz[2]],y,x,w.filter_coeff,w.ctr,sz[2],idwt_dim2)
        parallel_do([sz[0]*2,sz[1],sz[2]*2],x,z,w.filter_coeff,w.ctr,sz[1],idwt_dim1)
        parallel_do([sz[0],sz[1]*2,sz[2]*2],z,y,w.filter_coeff,w.ctr,sz[0],idwt_dim0)       
        sz *= 2
    endfor
endfunction

% Function: pdwt2d_untiled
%
% Untiled inplace version of the 2d wavelet packet transform
%
% :function [] = pdwt2d_untiled(x, w : wavelet_params, J, rev=0)
%
% Parameters:
% x - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% rev - time reversal the wavelet filter coefficients
function [] = pdwt2d_untiled(x, w : wavelet_params, J, rev=0)

    sz = size(x,0..2)
    sz0 = copy(sz)
    if mod(sz[0], 2^J) != 0 || mod(sz[1], 2^J) != 0
        error "dwt2d: input matrix should have dimensions that are a multiple of ",2^J
    endif
    wc = w.filter_coeff
    if rev
        wc = copy(wc)
        wc[1,:] = wc[1, size(wc,1)-1..-1..0]
    endif

    z = uninit(size(x))
    for j=1..J
        parallel_do([sz0[0],sz0[1]/2,sz0[2]],x,z,wc,w.ctd,sz[1],dwt_dim1)
        parallel_do([sz0[0]/2,sz0[1],sz0[2]],z,x,wc,w.ctd,sz[0],dwt_dim0)

        sz = [sz[0]/2,sz[1]/2,sz[2]]
    endfor
endfunction

% Function: ipdwt2d_untiled
%
% Untiled inplace version of the 2d inverse wavelet packet transform
%
% :function [] = ipdwt2d_untiled(y, w : wavelet_params, J)
%
% Parameters:
% y - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
function [] = ipdwt2d_untiled(y, w : wavelet_params, J)
    sz = [size(y,0)/2.^J, size(y,1)/2.^J, size(y,2)]
    sz0 = size(y,0..2)

    z = uninit(size(y))
    for j=1..J                
        parallel_do([sz0[0],sz0[1]/2,sz0[2]],y,z,w.filter_coeff,w.ctr,sz[1],idwt_dim1)
        parallel_do([sz0[0]/2,sz0[1],sz0[2]],z,y,w.filter_coeff,w.ctr,sz[0],idwt_dim0)       

        sz = [sz[0]*2,sz[1]*2,sz[2]]
    endfor
endfunction


% version with different row/column filters
function [] = dwt2d_untiled_rowcol(x, sz, w1 : wavelet_params, w2 : wavelet_params, J)
    z = uninit(size(x))
    for j=1..J
        parallel_do([sz[0],sz[1]/2,sz[2]],x,z,w2.filter_coeff,w2.ctd,sz[1],dwt_dim1)        
        parallel_do([sz[0]/2,sz[1],sz[2]],z,x,w1.filter_coeff,w1.ctd,sz[0],dwt_dim0)       
        sz = [sz[0]/2,sz[1]/2,sz[2]]
    endfor
endfunction

function [] = idwt2d_untiled_rowcol(x, sz, w1 : wavelet_params, w2 : wavelet_params, J)
    if sz[0] > size(x,0)/2.^J || sz[1] > size(x,1)/2.^J 
        error "(idwt2d_untiled_rowcol) invalid dimensions!"
    endif   

    z = uninit(size(x))
    for j=1..J                
        parallel_do([sz[0]*2,sz[1],sz[2]],x,z,w2.filter_coeff,w2.ctr,sz[1],idwt_dim1)
        parallel_do([sz[0],sz[1]*2,sz[2]],z,x,w1.filter_coeff,w1.ctr,sz[0],idwt_dim0)       
        sz = [sz[0]*2,sz[1]*2,sz[2]]
    endfor
endfunction

% 3D version with different row/column/slice filters
function [] = dwt3d_untiled_rowcol(x, sz, w1 : wavelet_params, w2 : wavelet_params, w3 : wavelet_params, J)
    z = uninit(size(x))
    y = uninit(size(x))
    for j=1..J
        parallel_do([sz[0],sz[1],sz[2]/2],x,y,w3.filter_coeff,w3.ctd,sz[2],dwt_dim2)
        parallel_do([sz[0],sz[1]/2,sz[2]],y,z,w2.filter_coeff,w2.ctd,sz[1],dwt_dim1)
        parallel_do([sz[0]/2,sz[1],sz[2]],z,x,w1.filter_coeff,w1.ctd,sz[0],dwt_dim0)
        sz /= 2
    endfor
endfunction

function [] = idwt3d_untiled_rowcol(x, sz, w1 : wavelet_params, w2 : wavelet_params, w3 : wavelet_params, J)
    if sz[0] > size(x,0)/2.^J || sz[1] > size(x,1)/2.^J || sz[2] > size(x,2)/2.^J
        error "(idwt3d_untiled_rowcol) invalid dimensions!"
    endif   

    z = uninit(size(x))
    y = uninit(size(x))
    for j=1..J                
        parallel_do([sz[0]*2,sz[1]*2,sz[2]],x,y,w3.filter_coeff,w3.ctr,sz[2],idwt_dim2)
        parallel_do([sz[0]*2,sz[1],sz[2]*2],y,z,w2.filter_coeff,w2.ctr,sz[1],idwt_dim1)
        parallel_do([sz[0],sz[1]*2,sz[2]*2],z,x,w1.filter_coeff,w1.ctr,sz[0],idwt_dim0)   
        sz *= 2
    endfor
endfunction

% rev : reverse the highpass coefficients
function [] = pdwt2d_untiled_rowcol(x, sz, w1 : wavelet_params, w2 : wavelet_params, J, rev=0)
    sz0 = size(x,0..2)
    col_wc = w1.filter_coeff
    row_wc = w2.filter_coeff
    if rev
        col_wc = copy(col_wc)
        row_wc = copy(row_wc)
        col_wc[1,:] = col_wc[1, size(col_wc,1)-1..-1..0]
        row_wc[1,:] = row_wc[1, size(row_wc,1)-1..-1..0]
    endif

    z = uninit(size(x))
    for j=1..J
        parallel_do([sz0[0],sz0[1]/2,sz0[2]],x,z,row_wc,w2.ctd,sz[1],dwt_dim1)
        parallel_do([sz0[0]/2,sz0[1],sz0[2]],z,x,col_wc,w1.ctd,sz[0],dwt_dim0)
        sz = [sz[0]/2,sz[1]/2,sz[2]]
    endfor
endfunction

function [] = ipdwt2d_untiled_rowcol(x, sz, w1 : wavelet_params, w2 : wavelet_params, J)
    %sz = [size(x,0)/2.^J, size(x,1)/2.^J, size(x,2)]
    sz0 = size(x,0..2)

    z = uninit(size(x))
    for j=1..J                
        parallel_do([sz0[0],sz0[1]/2,sz0[2]],x,z,w2.filter_coeff,w2.ctr,sz[1],idwt_dim1)
        parallel_do([sz0[0]/2,sz0[1],sz0[2]],z,x,w1.filter_coeff,w1.ctr,sz[0],idwt_dim0)       
        sz = [sz[0]*2,sz[1]*2,sz[2]]
    endfor
endfunction

% Function: dwt2d
%
% Computes the 2D DWT.
%
% :function bands = dwt2d(x, w : wavelet_params, J)
%
% Parameters:
% x - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% bands - an output cell array containing the wavelet subbands
% o bands[0..J-1][0..2]: wavelet subbands for orientations HL,LH,HH
% o bands[J]: scaling coefficients
%
% See also:
%    <idwt2d>, <dwt1d>, <dwt3d>
function bands = dwt2d(x, w : wavelet_params, J)
    sz = size(x,0..2)
    if mod(sz[0], 2^J) != 0 || mod(sz[1], 2^J) != 0
        error "dwt2d: input matrix should have dimensions that are a multiple of ",2^J
    endif
    bands = cell(J+1)
    y = copy(x)
    z = uninit(size(x))
    for j=0..J-1
        parallel_do([sz[0],sz[1]/2,sz[2]],y,z,w.filter_coeff,w.ctd,sz[1],dwt_dim1)
        parallel_do([sz[0]/2,sz[1],sz[2]],z,y,w.filter_coeff,w.ctd,sz[0],dwt_dim0)
        sz = [sz[0]/2,sz[1]/2,sz[2]]

        bands[j] = cell(3)
        bands[j][0] = y[sz[0]..2*sz[0]-1,0..sz[1]-1,:]
        bands[j][1] = y[0..sz[0]-1,sz[1]..2*sz[1]-1,:]
        bands[j][2] = y[sz[0]..2*sz[0]-1,sz[1]..2*sz[1]-1,:]       
    endfor
    bands[J] = y[0..sz[0]-1,0..sz[1]-1,:]
endfunction

function [] = idwt2d_inplace(bands, y, w : wavelet_params, J)
    sz = size(bands[J],0..2)
    y[0..sz[0]-1,0..sz[1]-1,:] = bands[J]    
    for j=J-1..-1..0
        y[sz[0]..2*sz[0]-1,0..sz[1]-1,:] = bands[j][0]
        y[0..sz[0]-1,sz[1]..2*sz[1]-1,:] = bands[j][1]
        y[sz[0]..2*sz[0]-1,sz[1]..2*sz[1]-1,:] = bands[j][2]
        sz = [sz[0]*2,sz[1]*2,sz[2]]
    endfor
    idwt2d_untiled(y, w, J)
endfunction

% Function: idwt2d
%
% Computes the inverse 2D DWT.
%
% :function y = idwt2d(bands, w : wavelet_params, J)
% 
% Parameters:
% bands - cell array containing the wavelet subbands
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% y - the output data
%
% See also:
%    <dwt2d>, <dwt1d>, <dwt3d>
function y : cube = idwt2d(bands, w : wavelet_params, J)
    sz = size(bands[J],0..2)
    y = uninit([sz[0]*2^J, sz[1]*2^J, sz[2]])
    idwt2d_inplace(bands, y, w, J)
endfunction


% Function: dwt3d
%
% Computes the 3D DWT.
%
% :function bands = dwt3d(x, w : wavelet_params, J)
%
% Parameters:
% x - the input data
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% bands - an output cell array containing the wavelet subbands
% o bands[0..J-1][0..7]: wavelet subbands for different 3D orientations
% o bands[J]: scaling coefficients
%
% See also:
%    <idwt3d>, <dwt2d>, <dwt1d>
function bands = dwt3d(x, w : wavelet_params, J)
    sz = size(x,0..2)
    if mod(sz[0], 2^J) != 0 || mod(sz[1], 2^J) != 0 || mod(sz[2], 2^J) != 0
        error "dwt3d: input matrix should have dimensions that are a multiple of ",2^J
    endif
    bands = cell(J+1)
    y = copy(x)
    z = uninit(size(x))
    v = uninit(size(x))
    for j=0..J-1
        parallel_do([sz[0],sz[1],sz[2]/2],y,v,w.filter_coeff,w.ctd,sz[2],dwt_dim2)
        parallel_do([sz[0],sz[1]/2,sz[2]],v,z,w.filter_coeff,w.ctd,sz[1],dwt_dim1)
        parallel_do([sz[0]/2,sz[1],sz[2]],z,y,w.filter_coeff,w.ctd,sz[0],dwt_dim0)
        sz /= 2
        
        r0L=0..sz[0]-1
        r1L=0..sz[1]-1
        r2L=0..sz[2]-1
        r0R=sz[0]..2*sz[0]-1
        r1R=sz[1]..2*sz[1]-1
        r2R=sz[2]..2*sz[2]-1        

        bands[j] = cell(7)
        bands[j][0] = y[r0R,r1L,r2L]
        bands[j][1] = y[r0L,r1R,r2L]
        bands[j][2] = y[r0R,r1R,r2L]       
        bands[j][3] = y[r0L,r1L,r2R]
        bands[j][4] = y[r0R,r1L,r2R]
        bands[j][5] = y[r0L,r1R,r2R]
        bands[j][6] = y[r0R,r1R,r2R]
    endfor
    bands[J] = y[0..sz[0]-1,0..sz[1]-1,0..sz[2]-1]

endfunction

function [] = idwt3d_inplace(bands, y, w : wavelet_params, J)
    sz = size(bands[J],0..2)
    y[0..sz[0]-1,0..sz[1]-1,0..sz[2]-1] = bands[J]    
    for j=J-1..-1..0
        r0L=0..sz[0]-1
        r1L=0..sz[1]-1
        r2L=0..sz[2]-1
        r0R=sz[0]..2*sz[0]-1
        r1R=sz[1]..2*sz[1]-1
        r2R=sz[2]..2*sz[2]-1        

        y[r0R,r1L,r2L] = bands[j][0]
        y[r0L,r1R,r2L] = bands[j][1]
        y[r0R,r1R,r2L] = bands[j][2]       
        y[r0L,r1L,r2R] = bands[j][3]
        y[r0R,r1L,r2R] = bands[j][4]
        y[r0L,r1R,r2R] = bands[j][5]
        y[r0R,r1R,r2R] = bands[j][6]
        sz *= 2
    endfor
    idwt3d_untiled(y, w, J)
endfunction

% Function: idwt3d
%
% Computes the inverse 3D DWT.
%
% :function y = idwt3d(bands, w : wavelet_params, J)
% 
% Parameters:
% bands - cell array containing the wavelet subbands
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% y - the output data
%
% See also:
%    <dwt3d>, <dwt2d>, <dwt1d>
function y : cube = idwt3d(bands, w : wavelet_params, J)
    sz = size(bands[J],0..2)
    y = uninit(sz*2^J)
    idwt3d_inplace(bands, y, w, J)
endfunction

% Function: dwt1d
%
% Computes the 1D DWT along the rows of the input matrix x
%
% :function bands = dwt1d(x, w : wavelet_params, J)
%
%
% Parameters:
% x - the input signal/image
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% bands - cell array containing the wavelet subbands
%
% See also:
%    <idwt1d>, <dwt2d>
function bands = dwt1d(x, w : wavelet_params, J)
    sz = size(x,0..2)
    if mod(sz[1], 2^J) != 0
        error "dwt1d: input matrix should have dimensions that are a multiple of ",2^J
    endif
    bands = cell(J+1)
    y = copy(x)
    z = uninit(size(x))
    for j=0..J-1
        parallel_do([sz[0],sz[1]/2,sz[2]],y,z,w.filter_coeff,w.ctd,sz[1],dwt_dim1)
        sz = [sz[0],sz[1]/2,sz[2]]

        bands[j] = y[:,sz[1]..2*sz[1]-1,:]
    endfor
    bands[J] = y[:,0..sz[1]-1,:]
endfunction

function [] = idwt1d_untiled(y, w : wavelet_params, J)
    sz = [size(y,0), size(y,1)/2.^J, size(y,2)]

    z = uninit(size(y))
    for j=1..J                
        parallel_do([sz[0],sz[1]*2,sz[2]],z,y,w.filter_coeff,w.ctr,sz[0],idwt_dim0)       

        sz = [sz[0],sz[1]*2,sz[2]]
    endfor
endfunction

% Function: idwt1d_inplace
%
% Computes an inplace inverse 1D DWT.
%
% :function [] = idwt1d(bands, y, w : wavelet_params, J)
% 
% Parameters:
% bands - cell array containing the wavelet subbands
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% y - the output data
%
% See also:
%    <dwt1d>, <dwt2d>
function [] = idwt1d_inplace(bands, y, w : wavelet_params, J)
    sz = size(bands[J],0..2)
    y[:,0..sz[1]-1,:] = bands[J]    
    for j=J-1..-1..0
        y[:,sz[1]..2*sz[1]-1,:] = bands[j]
        sz = [sz[0],sz[1]*2,sz[2]]
    endfor
    idwt1d_untiled(y, w, J)
endfunction

% Function: idwt1d
%
% Computes the inverse 1D DWT.
%
% :function y = idwt1d(bands, w : wavelet_params, J)
% 
% Parameters:
% bands - cell array containing the wavelet subbands
% w - the filters being used (see <orthogonal_wavelet>)
% J - the number of scales
% y - the output data
%
% See also:
%    <dwt1d>, <dwt2d>
function y = idwt1d(bands, w : wavelet_params, J)
    sz = size(bands[J],0..2)
    y = uninit([sz[0], sz[1]*2^J, sz[2]])
    idwt1d_inplace(bands, y, w, J)
endfunction

function [] = main()

    nr = x -> ((x - min(x))/(max(x)-min(x)))*255
    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    %nr = x -> (1 + x ./ max(abs(x))) * 127.5

    img = imread("lena_big.tif")
    %img = imread("lighthouse.png")
    %img = img[:,:,1]
    J = 5
    y = copy(img)
    print min(img),",",max(img)

%    dwt2d_untiled(y, db4, 1)
%    imshow(y,[])
%    return

    %img = idwt2d(dwt2d(img, db4, J), db4, J)
    
    w = orthogonal_wavelet(db8)

    tic()
    for k=1..100%000
    y = copy(img)
    %dwt2d_untiled(y, db4, J)
    %idwt2d_untiled(y, db4, J) % warm-up
    %w = dwt2d(img, db4, J)
    %y = idwt2d(w, db4, J)
    dwt2d_untiled_rowcol(y, size(y,0..2), w, w, J)
    idwt2d_untiled_rowcol(y, [size(y,0)/2^J,size(y,1)/2^J,size(y,2)], w, w, J)
    endfor
    toc()
    syncthreads(host)
    %t=imshow(y,[])
    imshow(y)
    %print "psnr=", psnr(img,y), "dB"
    %stop
endfunction
