{!author name="Bart Goossens"}
{!doc category="Image Processing/Filters"}

% Function: fast_bilateral_filter
%
% Bilateral filter for images - using the fast histogram based method
%
% :function [] = fast_bilateral_filter(x : cube, y : cube, r : int, h : scalar, _
%    bins : int=256, bin_size : scalar = 4)
%
% Parameters:
% x - the input image
% y - the output image (allocated to have the same size as the input image)
% r - the radius of the bilateral filter (r = 4 corresponds to a 9x9 window)
% h - bandwidth parameter of the bilateral filter
% bins - the number of histogram bins (the higher, the more memory usage and the higher
%   the computational cost)
% bin_size - the size of each bin (for bins=256 and a 8-bit image, the bin_size is 1, 
%     for bins=256 and a 16-bit image, the bin_size is 256)
%
% Notes:
% * the spatial distance term is currently being ignored for efficiency reasons
% * the method is an implementation of the following paper:
%      http://www.shellandslate.com/download/fastmedian_5506.pdf
%   to deal somewhat with the spatial distance term, the fast bilateral filter
%   can be applied iteratively.
function [] = fast_bilateral_filter(x : cube, y : cube, r : int, h : scalar, _
    bins : int=256, bin_size : scalar = 4)

    % add_hist_row : adds one single row to the histogram
    function [] = __device__ add_hist_row(x : cube'mirror'hwtex_const, hist_im : cube{4}[int]'unchecked, row : int, r : int, N : int, sgn : int, pos : ivec3, blkdim : int)
        Nhalf = int(N / 2)

        for p = pos[1]..blkdim..size(x,1)+r-1
            n = mod(p, N)
            p0 = p - n

            if n < Nhalf
                for k = -Nhalf+n..-1
                    hist_im[int(x[row, p0 + k, pos[2]]),         pos[0], pos[2], p] += sgn
                    hist_im[int(x[row, p0 + 2*r+1 + k, pos[2]]), pos[0], pos[2], p] -= sgn
                endfor
            elseif n == Nhalf
                % this is the most labour-intensive loop
                for k = 0..2*r
                    hist_im[int(x[row, p0 + k, pos[2]]),         pos[0], pos[2], p] += sgn
                endfor
            else
                for k = 0..n-Nhalf-1
                    hist_im[int(x[row, p0 + k, pos[2]]),         pos[0], pos[2], p] -= sgn
                    hist_im[int(x[row, p0 + 2*r+1 + k, pos[2]]), pos[0], pos[2], p] += sgn
                endfor
            endif
        endfor
    endfunction

    % compute_median : computes the median, based on the local histograms
    function med : int = __device__ compute_median(hist_im : cube{4}'unchecked, p : int, N : int, idx_median : int, b : int, c : int)
        Nhalf = int(N/2)
        n = mod(p, N)
        p0 = p - n

        sum = 0
        i = 0
        while sum < idx_median && i < size(hist_im,0)
            if n == Nhalf
                val = hist_im[i, b, c, p]
            else
                val = hist_im[i, b, c, p] + hist_im[i, b, c, p0 + Nhalf]
            endif
            sum += int(val)
            i += 1
        endwhile
        med = i-1
    endfunction
        
    % compute_filter_output : computes the bilateral filter output, based on the local histograms
    function [y : vec3] = __device__ compute_filter_output_RGB(hist_im : cube{4}[int]'unchecked, _
        w : vec'unchecked, input : ivec3, p : int, N : int, b : int)
        
        n = mod(p, N)
        p0 = p - n
        Nhalf = int(N/2)

        y = [0.0, 0.0, 0.0]
        weight_sum = [0.0, 0.0, 0.0]
        bins = size(hist_im,0)
        for k=1..bins-1
            if n == Nhalf
                val = hist_im[k, b, 0..2, p]
            else
                val = hist_im[k, b, 0..2, p] + hist_im[k, b, 0..2, p0 + Nhalf]
            endif
            weight = [w[input[0] - k + bins] * val[0], _
                      w[input[1] - k + bins] * val[1], _
                      w[input[2] - k + bins] * val[2]]
            y += weight .* [k, k, k]
            weight_sum += weight
        endfor

        % the separable histogram step
        y = [y[0] * weight_sum[1] * weight_sum[2], _
             weight_sum[0] * y[1] * weight_sum[2], _
             weight_sum[0] * weight_sum[1] * y[2]]
        weight_sum = weight_sum[0] * weight_sum[1] * weight_sum[2] * [1.0,1.0,1.0]
        if min(weight_sum) < 1e-4
            y = float(input)
        else
            y ./= weight_sum
        endif
    endfunction

    function [y : scalar] = __device__ compute_filter_output_grayscale(hist_im : cube{4}[int], _
        w : vec'unchecked, input : scalar, p : int, N : int, b : int)
        
        n = mod(p, N)
        p0 = p - n
        Nhalf = int(N/2)

        y = 0.0
        weight_sum = 0.0
        bins = size(hist_im,0)
        for k=1..bins-1
            if n == Nhalf
                val = hist_im[k, b, 0, p]
            else
                val = hist_im[k, b, 0, p] + hist_im[k, b, 0, p0 + Nhalf]
            endif
            weight = w[input - k + bins] * val
            y += weight .* k
            weight_sum += weight
        endfor
        if weight_sum < 1e-4
            y = input
        else
            y = y / weight_sum
        endif
    endfunction
       
    % ARGUMENTS:
    % x : input image
    % y : bilateral filtered output image
    % hist_im : temporary memory for the histograms (one per column of the input image)
    % w : weighting function [-256..255] - 512 entries
    % r : the radius of the bilateral filter (r = 4 corresponds to a 9x9 window)
    % N : the number of columns processed simultaneously
    function [] = __kernel__ bilateral_filter_kernel(x : cube'unchecked'const'hwtex_const, _
        y : cube'unchecked, hist_im : cube{4}[int], w : vec, r : int, N : int, K : int, blkdim : ivec3, pos : ivec3)


        % Vertical step size
        step_y = int((size(x,0) + K*blkdim[0] - 1) ./ (K*blkdim[0])) % dim[0]=K*blkdim[0]

        % Index of the median
        idx_median = int((2*r+1)^2/2)

        % Initialization - add the first r rows
        for m = -r..r-1
            add_hist_row(x, hist_im, pos[0]*step_y + m, r, N, 1, pos, blkdim[1])
        endfor

        memfence
        syncthreads

        for m = pos[0]*step_y..min((pos[0]+1)*step_y, size(x,0))-1
            % Running phase - add row (r), subtract row (r-1)
            add_hist_row(x, hist_im, m+r, r, N, 1, pos, blkdim[1])
            add_hist_row(x, hist_im, m-r-1, r, N, -1, pos, blkdim[1])
            memfence
            syncthreads

            % At this time, we have all local histograms at row m at our disposal, so we can
            % directly proceed by computing the median
            for p = pos[1]..blkdim[1]..size(x,1)-1
                if pos[2] == 0 && size(y,2) == 3
                    y[m, p, 0..2] = compute_filter_output_RGB(hist_im, w, int(x[m, p, 0..2]), p, N, pos[0])
                else
                    y[m, p, 0] = compute_filter_output_grayscale(hist_im, w, int(x[m, p, 0]), p, N, pos[0])
                endif
            endfor
            
            memfence
            syncthreads
        endfor
    endfunction
    
    nextpow2 = x -> 2^ceil(log2(x))
    N = floor((r)/2)*2+1
    K = 64
    w = exp(-(10^h)/(r+1)^2 * (-bins..bins-1).^2/(bins/256)^2)  % weighting function

    % Compute the block size of the filter
    blk_size = max_block_size(bilateral_filter_kernel, [K, size(x,1), size(x,2)])
    assert(blk_size[2] == size(x,2), "fast_bilateral_filter: block dimensioning problem");

    % First - make sure the input is between 0 and bins-1
    x = saturate(x/(bins*bin_size-1))*(bins-1)    
    
    % Apply the filter...
    type cube4 : cube{4}
    hist_im = cube4[int](bins, blk_size[0]*K, blk_size[2], size(x,1) + nextpow2(N))
    B_size = [K,1,1].*blk_size
    parallel_do([B_size,blk_size],x,y,hist_im,w,r,N,K,bilateral_filter_kernel)
    y[:,:,:] = max(0,min(bins-1, y))
    if bin_size!=1
        y *= (bins*bin_size-1)/(bins-1)
    endif
endfunction

function [] = main()
    x = repmat(imread("lighthouse.png")[:,:,1],[2,4,1])
    x += 10*randn(size(x))
    y = zeros(size(x))
    tic()
    for k=0..9
        fast_bilateral_filter(x, y, 20, 1, 256, 4)
    endfor
   toc()
    print min(y),"-",max(y)
    imshow(y,[0,255])
endfunction