% ==============================================
% Functional image processing library for Quasar
% Oct 2012-May 2013, Bart Goossens
% ==============================================

{!author name="Bart Goossens"}
{!doc category="Image Processing/Functional"}

% Notes: 
% * Due to the support for generics, inlining and function pointer avoidance
%   compilation techniques, a new programming approach is possible for defining
%   linear filters (and other image processing operations).
% * Because the relationships are functional, we use a little bit of Haskell-
%   like notation here and there.

% Define an image as a function from a three-dimensional
% input coordinate to an output color vector (e.g. RGB)
type image : [__device__ ivec2'const -> vec3]

% Image processing algorithms are then simply device functions that "work on"
% images. 
type algorithm : [__device__ (image, ivec2'const) -> vec3]

% The application of a processing algorithm to an image is then again an image
type application : [(image, algorithm) -> image]

% A color processing function
type color_function : [__device__ vec3'const -> vec3]

% A rasterizer function
type rasterizer : [(image, ivec3'const) -> cube(:,:,3)]

% Function for converting to an image & define it using a reduction, so that
% we can easily convert
make_image : [cube -> image] =
    im_data -> __device__ (pos) -> im_data[pos[0], pos[1], 0..2]

reduction (x : cube) -> image(x) = make_image(x)

% Defining an image by circular boundary extension
make_image_ext_circular  : [cube'circular -> image] =
    im_data -> __device__ (pos) -> im_data[pos[0], pos[1], 0..2]

% Algorithm that applies a point-wise operation
pointwise_op : [color_function -> algorithm] =
    fn -> __device__ (x, pos) -> fn(x(pos))

% Translation algorithm
translate : [vec2 -> algorithm] =
    shift -> __device__ (x, pos) -> x(pos + int(shift))

% A horizontal mean filter, that takes an image as input. Note that the type
% of mean_filter_hor is 'algorithm'.
%mean_filter_hor : [int -> algorithm]
function algo : algorithm = mean_filter_hor(N : int)
    norm_factor = 1.0 / N % normalization factor
    function y : vec3 = __device__ algo(input : image, pos : ivec2)
        y = [0., 0., 0.]
        for n=-int(N/2)..int(N/2)
            y = y + input(pos + [0,n])
        end
        y = y * norm_factor
    endfunction
endfunction

% A vertical mean filter, that takes an image as input. Note that the type
% of mean_filter_ver is 'algorithm'.
%mean_filter_ver : [int -> algorithm]
function algo : algorithm = mean_filter_ver(N : int)
    norm_factor = 1.0 / N % normalization factor
    function y : vec3 = __device__ algo(input : image, pos : ivec2)
        y = [0., 0., 0.]
        for n=-int(N/2)..int(N/2)
            y = y + input(pos + [n,0])
        endfor
        y = y * norm_factor
    endfunction
endfunction

% A simple laplacian filter
laplacian_filter : algorithm =
    __device__ (input, pos) ->
        4*input(pos) - input(pos+[0,1]) - input(pos+[0,-1]) - input(pos+[1,0]) - input(pos+[-1,0])

% Computes the specified image (convert it back to its pixel representation)
compute : rasterizer =
    (x, sz) -> (y = zeros(sz); parallel_do(sz, __kernel__ (pos : ivec2) -> y[pos[0],pos[1],0..2] = x(pos)); y)

% Computes the specified image, using shared memory
% 'border' specifies the size of the block overlap.
function [apply : application] = apply_sharedmem(border : int'const, sz : ivec3)
    function [y_img : image] = apply(x : image, p : algorithm)
        y = zeros(sz)
        function [] = __kernel__ kernel(pos : ivec2, blkdim : ivec2, blkpos : ivec2)
            assert(border<=4)
            mem : cube'unchecked = shared(blkdim[0]+border*2+1, blkdim[1]+border*2+1, 3)
            % read input data
            q = pos - [border, border]
            mem[blkpos[0], blkpos[1], 0..2] = x(q)
            % deal with the image borders
            if blkpos[0] <= border*2
                mem[blkpos[0]+blkdim[0],blkpos[1],0..2] = x(q+[blkdim[0],0])
            endif
            if blkpos[1] <= border*2
                mem[blkpos[0],blkpos[1]+blkdim[1],0..2] = x(q+[0,blkdim[1]])
                if blkpos[0] <= border*2
                    mem[blkpos[0]+blkdim[0],blkpos[1]+blkdim[1],0..2] = x(q+blkdim)
                endif
            endif
            syncthreads
            % apply the fuction
            y[pos[0],pos[1],0..2] = p(((p : ivec2) -> mem[p[0], p[1], 0..2]), 
                                      blkpos + [border, border])
        endfunction
        parallel_do(sz, kernel)
        y_img = make_image(y) % Convert back to an image...
    endfunction
endfunction

% Apply the specified processing algorithm to the image
% (using deferred processing)
apply_deferred : application =
    (x, p) -> __device__ (pos) -> p(x, pos)

% Immediate computation of an intermediate result
apply_immediate : [vec3 -> application] = 
    sz -> (x, p) -> make_image(compute(apply_deferred(x, p), sz))

% Apply immediate, using column-dominant (instead of row-dominant)
% processing optimization
apply_immediate_coldominant : [vec3 -> application] =
    sz -> (x, p) ->
    (z_data = zeros(sz[1], sz[0], 3); 
     % 1. compute the transposed version of x
     parallel_do(sz, __kernel__ (pos : ivec2) -> z_data[pos[1],pos[0],0..2] = x(pos));
     % 2. construct an image that swaps the input coordinates
     z = __device__ (pos : ivec2) -> z_data[pos[1],pos[0],0..2];
     % 3. apply the algorithm to the "double transposed" input image
     apply_immediate(sz)(x, p)
    )


% Pipeline assignment (eager evaluation)
reduction (y : cube, x : image) -> (y := x) = y = compute(x, size(y))

% Pipeline operations
symbolic immediate, deferred, sharedmem, immediate_coldom
reduction (x : cube, y) -> (x |> y) = make_image(x) |> y
reduction (x : image, a : algorithm, sz : ivec3) -> (x |> immediate(sz, a)) = apply_immediate(sz)(x, a)
reduction (x : image, a : algorithm) -> (x |> deferred(a)) = apply_deferred(x, a)
reduction (x : image, y, a : algorithm, sz : ivec3) -> (x |> sharedmem(a, sz, y)) = apply_sharedmem(int(y/2), sz)(x, a)
reduction (x : image, a : algorithm, sz : ivec3) -> (x |> immediate_coldom(a, sz, y)) = apply_immediate_coldominant(sz)(x, a)

function [] = main()
    im = imread("lena_big.tif")
    im_out = zeros(384,384,3)
    sz = size(im_out)
    tic()
    p = im |> deferred(translate([64, 64])) _
           |> deferred(mean_filter_ver(7)) _
           |> sharedmem(mean_filter_hor(7), sz, 7)
    im_out := p
    toc()
    imshow (im_out)
endfunction