%==============================================================================
% floattypes.q
% Quasar library for extended scalaring datatype support
% (C) Gepura, 2018
%==============================================================================
% By default, 'scalar' represents floating point numbers with a precision that
% is configurable when the program starts. In case it is necessary to 
% explicitly specify the floating point precision, the type modifiers 'half,
% 'single and 'double can be used for respectively 16, 32 and 64-bit floating
% point formats. This module also defines the necessary operators to work with
% these data types.
%==============================================================================

% ==== Type definitions =======

% Half-precision (16-bit) floating point numbers
type hvec   : vec[scalar'half]
type hmat   : mat[scalar'half]
type hcube  : cube[scalar'half]
type hcubeN : cube{:}[scalar'half]

% Single-precision (32-bit) floating point numbers
type fvec   : vec[scalar'single]
type fmat   : mat[scalar'single]
type fcube  : cube[scalar'single]
type fcubeN : cube{:}[scalar'single]

% Double-precision (64-bit) floating point numbers
type dvec   : vec[scalar'double]
type dmat   : mat[scalar'double]
type dcube  : cube[scalar'double]
type dcubeN : cube{:}[scalar'double]

% Half precision SIMD vector of length 2
type hvec2  : hvec(2)

function [im_out] = scalar_to_half(im : cube{:})
    im_out = hcubeN(size(im))
    function [] = __kernel__ kernel(im_out : hvec, im : vec, pos : int)
        {!kernel_tiling dims=2; mode="local"}
        im_out[pos] = im[pos]
    endfunction
    parallel_do(numel(im_out), im_out, im, kernel)
endfunction

function [im_out] = half_to_scalar(im : hcubeN)
    im_out = uninit(size(im))
    function [] = __kernel__ kernel(im_out : vec, im : hvec, pos : int)
        {!kernel_tiling dims=2; mode="local"}
        im_out[pos] = im[pos]
    endfunction
    parallel_do(numel(im_out), im_out, im, kernel)
endfunction

reduction (im : cube{:}) -> half(im) = scalar_to_half(im)
reduction (im : hcubeN) -> float(im) = half_to_scalar(im)

operator (a : hcubeN) -> -a = -a
operator (a : hcubeN, b : scalar) -> a + b = a + b
operator (a : scalar, b : hcubeN) -> a + b = a + b
operator (a : hcubeN, b : hcubeN) -> a + b = a + b
operator (a : hcubeN, b : scalar) -> a - b = a - b
operator (a : scalar, b : hcubeN) -> a - b = a - b
operator (a : hcubeN, b : hcubeN) -> a - b = a - b
operator (a : hcubeN, b : scalar) -> a * b = a * b
operator (a : hcubeN, b : scalar) -> a .* b = a .* b
operator (a : scalar, b : hcubeN) -> a * b = a * b
operator (a : scalar, b : hcubeN) -> a .* b = a .* b
operator (a : hcubeN, b : hcubeN) -> a .* b = a .* b
operator (a : hcubeN, b : scalar) -> a ./ b = a ./ b
operator (a : scalar, b : hcubeN) -> a ./ b = a ./ b
operator (a : hcubeN, b : hcubeN) -> a ./ b = a ./ b
operator (a : hcubeN, b : scalar) -> a / b = a / b
operator (a : hcubeN, b : scalar) -> a .^ b = a .^ b
operator (a : scalar, b : hcubeN) -> a .^ b = a .^ b
operator (a : hcubeN, b : hcubeN) -> a .^ b = a .^ b

operator (a : hcubeN) -> !a = !a
operator (a : hcubeN, b : scalar) -> a < b = a < b
operator (a : scalar, b : hcubeN) -> a < b = a < b
operator (a : hcubeN, b : hcubeN) -> a < b = a < b
operator (a : hcubeN, b : scalar) -> a > b = a > b
operator (a : scalar, b : hcubeN) -> a > b = a > b
operator (a : hcubeN, b : hcubeN) -> a > b = a > b
operator (a : hcubeN, b : scalar) -> a <= b = a <= b
operator (a : scalar, b : hcubeN) -> a <= b = a <= b
operator (a : hcubeN, b : hcubeN) -> a <= b = a <= b
operator (a : hcubeN, b : scalar) -> a >= b = a >= b
operator (a : scalar, b : hcubeN) -> a >= b = a >= b
operator (a : hcubeN, b : hcubeN) -> a >= b = a >= b
operator (a : hcubeN, b : scalar) -> a == b = a == b
operator (a : scalar, b : hcubeN) -> a == b = a == b
operator (a : hcubeN, b : hcubeN) -> a == b = a == b
operator (a : hcubeN, b : scalar) -> a != b = a != b
operator (a : scalar, b : hcubeN) -> a != b = a != b
operator (a : hcubeN, b : hcubeN) -> a != b = a != b

operator (a : hcubeN, b : scalar) -> min(a, b) = min(a, b)
operator (a : scalar, b : hcubeN) -> min(a, b) = min(a, b)
operator (a : hcubeN, b : hcubeN) -> min(a, b) = min(a, b)
operator (a : hcubeN, b : scalar) -> max(a, b) = min(a, b)
operator (a : scalar, b : hcubeN) -> max(a, b) = max(a, b)
operator (a : hcubeN, b : hcubeN) -> max(a, b) = max(a, b)


 
function [im_out] = scalar_to_single(im : cube{:})
    im_out = fcubeN(size(im))
    function [] = __kernel__ kernel(im_out : fvec, im : vec, pos : int)
        im_out[pos] = im[pos]
    endfunction
    parallel_do(numel(im_out), im_out, im, kernel)
endfunction

function [im_out] = single_to_scalar(im : fcubeN)
    im_out = uninit(size(im))
    function [] = __kernel__ kernel(im_out : vec, im : fvec, pos : int)
        im_out[pos] = im[pos]
    endfunction
    parallel_do(numel(im_out), im_out, im, kernel)
endfunction

function [im_out] = scalar_to_double(im : cube{:})
    im_out = dcubeN(size(im))
    function [] = __kernel__ kernel(im_out : dvec, im : vec, pos : int)
        im_out[pos] = im[pos]
    endfunction
    parallel_do(numel(im_out), im_out, im, kernel)
endfunction

function [im_out] = double_to_scalar(im : dcubeN)
    im_out = uninit(size(im))
    function [] = __kernel__ kernel(im_out : vec, im : dvec, pos : int)
        im_out[pos] = im[pos]
    endfunction
    parallel_do(numel(im_out), im_out, im, kernel)
endfunction

reduction (im : cube{:}) -> single(im) = scalar_to_single(im)
reduction (im : fcubeN) -> float(im) = single_to_scalar(im)
reduction (im : cube{:}) -> double(im) = scalar_to_double(im)
reduction (im : dcubeN) -> float(im) = double_to_scalar(im)


function [] = main()
    im = half(imread("lena_big.tif"))
    
    im = (im - min(im))./(max(im) - min(im))
    imshow(float(im), [0, 1])
    
    im2 = double(imread("lena_big.tif")/255)
    
    imshow(float(im2), [0, 1])
    
endfunction