% Implementation of Gaussian and Laplacian pyramids
% 
import "imfilter.q"

% Function: gaussian_pyramid
% 
% Calculates the Gaussian pyramid of the specified input image. Gaussian pyramids are commonly used
% in computer vision for e.g., object, edge or corner detection.
% 
% Usage:
%   : function [y : vec[cube]] = gaussian_pyramid(x : [mat|cube], sigma : scalar'const, 
%           num_scales : int, boundary_mode : string'const = "zero")
% 
% Parameters:
% x - input image
% sigma - sigma parameter of the Gaussian filter
% num_scales - the number of scales of the Gaussian pyramid.
% boundary_mode - the boundary mode to be used for the Gaussian pyramid
% 
% Returns:
% y - a vector containing the subbands of the Gaussian pyramid
% 
% See also:
% <imfilter>
function y = gaussian_pyramid(x : [mat|cube], sigma : scalar, num_scales : int, boundary_mode : string = "zero")

    y = vec[cube](num_scales)
    for i = 0..num_scales-1
        y[i] = x
        % Filter
        x = gaussian_filter(x, sigma, ceil(3*sigma), boundary_mode)
        
        % Subsample
        x = x[0..2..size(x,0)-1, 0..2..size(x,1)-1, :]
    endfor

endfunction

% Function: laplacian_pyramid
% 
% Calculates the Laplacian pyramid of the specified input image. Decomposes the image into 
% different "layers" of detail, but in a reversible way. 
% 
% Usage:
%   : function [y : vec[cube]] = laplacian_pyramid(x : [mat|cube], sigma : scalar'const, 
%       num_scales : int, boundary_mode : string'const = "zero")
% 
% Parameters:
% x - input image
% sigma - sigma parameter of the Gaussian filter used during the decomposition
% num_scales - the number of scales of the Laplacian pyramid.
% boundary_mode - the boundary mode to be used for the Gaussian pyramid
% 
% Returns:
% y - a vector containing the subbands of the Laplacian pyramid
% 
% See also:
% <imfilter>
function y = laplacian_pyramid(x : [mat|cube], sigma : scalar, num_scales : int, boundary_mode : string = "zero")

    y = vec[cube](num_scales)
    for i = 0..num_scales-2
        % Filter
        z = gaussian_filter(x, sigma, 0, boundary_mode)
        
        % Subsample
        z = z[0..2..size(x,0)-1, 0..2..size(x,1)-1, :]
        
        % Upsample again & filter
        w = zeros(size(x))
        w[0..2..size(x,0)-1, 0..2..size(x,1)-1, :] = 4*z
        w = gaussian_filter(w, sigma, ceil(3*sigma), boundary_mode)
        
        y[i] = x - w % Difference image
        x = z % Proceed with subsampled image
    endfor
    y[num_scales-1] = x

endfunction

% Function: laplacian_reconstruction
% 
% Inverts the Laplacian pyramid transform - reconstructs the original input image out of its
% Laplacian pyramid subbands. This is mostly useful when performing some processing of the subbands.
% 
% Usage:
%   : function [y : cube] = laplacian_reconstruction(x : vec[cube]'const, sigma : scalar'const, 
%       boundary_mode : string'const = "zero")
% 
% Parameters:
% x - a vector containing the subbands of the Laplacing pyramid
% sigma - sigma parameter of the Gaussian filter used during the decomposition (note: the value
%   of this parameter should match the value used for laplacing_pyramid)
% boundary_mode - the boundary mode to be used for the Gaussian pyramid
% 
% Returns:
% y - the reconstructed image
function y = laplacian_reconstruction(x : vec[cube], sigma : scalar, boundary_mode : string = "zero")

    num_scales = numel(x)
    y = x[num_scales-1]
    for i = num_scales-2..-1..0
        % upsample 'y' and filter
        w = zeros(size(x[i]))
        w[0..2..size(w,0)-1, 0..2..size(w,1)-1, :] = 4*y
        w = gaussian_filter(w, sigma, ceil(3*sigma), boundary_mode)
        
        y = x[i] + w % new output image
    endfor

endfunction


function [] = main()

    im = imread("lighthouse.png")
    sigma = 2

    im_pyr = laplacian_pyramid(im, sigma, 3)
    
    for i=0..numel(im_pyr)-1
        imshow(im_pyr[i])
        title(sprintf("Pyramid level %d", i+1))
    endfor
    
    im_out = laplacian_reconstruction(im_pyr, sigma)

    psnr = (x, y) -> 20*log10(255/mean((x - y).^2))

    imshow(im_out)
    title(sprintf("Laplacian reconstruction: PSNR=%f dB", psnr(im, im_out)))
    
    % Now we will perform a simple sharpening operation
    im_pyr[0] = im_pyr[0] * 2 % increase by 100%
    
    im_sharpened = laplacian_reconstruction(im_pyr, sigma)
    imshow(im_sharpened,[0,255])
    title("Sharpened image")

endfunction
   
