%==============================================================================================
% Quasar implementation of guided image filtering
% March 2015 - Hiep Luong and Simon Donné
% 
% Implementation of the following paper:
% He, K., Sun, J., & Tang, X. (2010, September). Guided image filtering. In European conference 
% on computer vision (pp. 1-14). Springer, Berlin, Heidelberg.
%==============================================================================================
import "colortransform.q"

{!author name="Hiep Luong, Simon Donné"}
{!doc category="Image Processing/Filters"}

type cube4 : cube{4}

function [] = main()
    img_depth = imread("kinect_depth.png")
    img_depth = (img_depth[:,:,1] + 256 * img_depth[:,:,2])/ 256 
    img_depth = img_depth + 0 * randn(size(img_depth))
    img_rgb = imread("kinect_color.png")
    
    ry = 180..size(img_depth,0)-1
    rx = 0..419
    img_depth = img_depth[ry,rx]
    img_rgb = img_rgb[ry,rx,:]
    imshow(img_depth)
    imshow(img_rgb)
    
    threshold = 1e-8
    result = copy(img_depth)
    img_rgb_gray = rgb2gray(img_rgb)
    for it=0..130
        result += 0.01 * (guided_filter(result, img_rgb_gray, 5, threshold) - result)
    endfor
    %result = guided_filter(img_depth, img_rgb_gray, 3, threshold*1000)
    imshow(result,[-40,161])

    

endfunction

function [] = main1()
    image = imread("lena_big.tif")
    iterations = 100
    threshold = 10^2
    
    x = rand(512,512)
    tic()
    for i = 1..iterations
        result = guided_filter(rgb2gray(image),rgb2gray(image),5,threshold)
    endfor
    toc(sprintf("%d gray-scale guided filterings",iterations))
    imshow(result)
    
    x = rand(512,512,3)
    tic()
    for i = 1..iterations
        guided_filter(x,image,5,threshold)
    endfor
    toc(sprintf("%d three-channel guided filterings",iterations))
endfunction

% Function: guided_filter
% Performs guided filtering according to the paper "Guided Image Filtering" by Kaiming He.
% the input image can be either grayscale or colour (with an arbitrary number of channels),
% but the guide image must be either grayscale (same guide image for all of the channels),
% or contain the exact same number of channels as the input image
%
% : function output = guided_filter(input, guide, radius, treshold)
%
% Parameters:
%   input - the image to filter (gray-scale or colour-image)
%   guide - the guide to the filter (gray-scale or colour-image)
%   radius - the radius of the filter
%   treshold - the treshold for the filtering
%   output - the output image
%
function [output:[mat|cube]] = guided_filter(input:[mat|cube],guide:[mat|cube],radius:int, threshold:scalar)
    
    assert( (size(input,2)==size(guide,2)) || (1==size(guide,2)) )
    
    function [output:mat] = guided_filter_mat(input:mat,guide:mat,radius:int, threshold:scalar)
        filter_area = (2*radius+1)^2
        threshold = threshold*filter_area^2
        
        [M,N] = size(input,0..1)
        
        tmp = uninit(M,N,4)
        ab = uninit(M,N,2)
        tmp_ab = uninit(M,N,2)
        output = uninit(M,N)
        
        %accumulators
        function [] = __kernel__ joint_box_filter_hor(g : mat'mirror, x : mat'mirror, tmp : cube'unchecked, r : int, pos : vec2)
            %{!kernel_transform enable="localwindow"}
            %{!specialize args=r == 3 || r==5 || r == 7}
            sums:vec4=[0.0,0.0,0.0,0.0]
            for i=-r..r
                t_g = g[pos + [0,i]]
                t_x = x[pos + [0,i]]
                sums = sums + [t_g,t_x,t_g*t_x,t_g*t_g]
            endfor
            tmp[pos[0],pos[1],0..3] = sums
        endfunction
        function [] = __kernel__ joint_box_filter_ver(tmp:cube'mirror, ab : cube'unchecked, r : int, N : scalar, e : scalar, pos : vec2)
            %{!kernel_transform enable="localwindow"; max_block_numel=256}
            %{!specialize args=r == 3 || r==5 || r == 7}
            sums:vec4 = [0.0,0.0,0.0,0.0]
            for i=-r..r
                sums = sums + tmp[pos[0]+i,pos[1],0..3]
            endfor
            tmp_a = (sums[2]*N - sums[0]*sums[1])/(sums[3]*N - sums[0]*sums[0] + e)
            ab[pos[0],pos[1],0..1] = [tmp_a,(sums[1] - tmp_a*sums[0])/N]
        endfunction
        %distributors
        function [] = __kernel__ joint_box_filter_hor2(ab : cube'mirror, tmp_ab:cube'unchecked, r : int, pos : vec2)
            %{!kernel_transform enable="localwindow"}
            %{!specialize args=r == 3 || r==5 || r == 7}
            s_ab = [0.0,0.0]
            for i=-r..r
                s_ab = s_ab + ab[pos[0],pos[1]+i,0..1]
            endfor
            tmp_ab[pos[0],pos[1],0..1] = s_ab
        endfunction
        function [] = __kernel__ joint_box_filter_ver2(tmp_ab : cube'mirror, g : mat, y : mat'unchecked, r : int, N : scalar, pos : vec2)
            %{!kernel_transform enable="localwindow"}
            %{!specialize args=r == 3 || r==5 || r == 7}
            s_ab = [0.0,0.0]
            for i=-r..r
                s_ab = s_ab + tmp_ab[pos[0]+i,pos[1],0..1]
            endfor
            y[pos] = (s_ab[0]*g[pos] + s_ab[1])/N
        endfunction

        parallel_do (size(input), guide, input, tmp, radius, joint_box_filter_hor)
        parallel_do (size(input), tmp, ab, radius, filter_area, threshold, joint_box_filter_ver)
        parallel_do (size(guide), ab, tmp_ab, radius, joint_box_filter_hor2)
        parallel_do (size(guide), tmp_ab, guide, output, radius, filter_area, joint_box_filter_ver2)
    endfunction
    function [output:cube] = guided_filter_cube(input:cube,guide:cube,radius:int, threshold:scalar)
        filter_area = (2*radius+1)^2
        threshold = threshold*filter_area^2
        
        [M,N,C] = size(input,0..2)
        
        tmp = uninit(M,N,C,4)
        ab = uninit(M,N,C,2)
        tmp_ab = uninit(M,N,C,2)
        output = uninit(M,N,C)
        
        %accumulators
        function [] = __kernel__ joint_box_filter_hor(g : cube'mirror, x : cube'mirror, tmp : cube4'unchecked, r : int, pos : vec3)
            %{!kernel_transform enable="localwindow"}
            %{!specialize args=r == 3 || r==5 || r == 7}
            sums:vec4=[0.0,0.0,0.0,0.0]
            for i=-r..r
                t_g = g[pos + [0,i,0]]
                t_x = x[pos + [0,i,0]]
                sums = sums + [t_g,t_x,t_g*t_x,t_g*t_g]
            endfor
            tmp[pos[0],pos[1],pos[2],0..3] = sums
        endfunction
        function [] = __kernel__ joint_box_filter_ver(tmp:cube4'mirror, ab : cube4'unchecked, r : int, N : scalar, e : scalar, pos : vec3)
            %{!kernel_transform enable="localwindow"}
            %{!specialize args=r == 3 || r==5 || r == 7}
            sums:vec4 = [0.0,0.0,0.0,0.0]
            for i=-r..r
                sums = sums + tmp[pos[0]+i,pos[1],pos[2],0..3]
            endfor
            tmp_a = (sums[2]*N - sums[0]*sums[1])/(sums[3]*N - sums[0]*sums[0] + e)
            ab[pos[0],pos[1],pos[2],0..1] = [tmp_a,(sums[1] - tmp_a*sums[0])/N]
        endfunction
        %distributors
        function [] = __kernel__ joint_box_filter_hor2(ab : cube4'mirror, tmp_ab:cube4'unchecked, r : int, pos : vec3)
            %{!kernel_transform enable="localwindow"}
            %{!specialize args=r == 3 || r==5 || r == 7}
            s_ab = [0.0,0.0]
            for i=-r..r
                s_ab = s_ab + ab[pos[0],pos[1]+i,pos[2],0..1]
            endfor
            tmp_ab[pos[0],pos[1],pos[2],0..1] = s_ab
        endfunction
        function [] = __kernel__ joint_box_filter_ver2(tmp_ab : cube4'mirror, g : cube, y : cube'unchecked, r : int, N : scalar, pos : vec3)
            %{!kernel_transform enable="localwindow"}
            %{!specialize args=r == 3 || r==5 || r == 7}
            s_ab = [0.0,0.0]
            for i=-r..r
                s_ab = s_ab + tmp_ab[pos[0]+i,pos[1],pos[2],0..1]
            endfor
            y[pos] = (s_ab[0]*g[pos] + s_ab[1])/N
        endfunction

        parallel_do (size(input), guide, input, tmp, radius, joint_box_filter_hor)
        parallel_do (size(input), tmp, ab, radius, filter_area, threshold, joint_box_filter_ver)
        parallel_do (size(guide), ab, tmp_ab, radius, joint_box_filter_hor2)
        parallel_do (size(guide), tmp_ab, guide, output, radius, filter_area, joint_box_filter_ver2)
    endfunction
    
    if size(input,2) >1
        output = guided_filter_cube(input,guide, radius, threshold)
    else
        output = guided_filter_mat(input,guide, radius, threshold)
    endif
endfunction
