% Type: Hashmap
%
% A simple, non-concurrent implementation of a hashmap, mainly intended
% to be used from device functions
import "list.q"

type Hashmap_entry[K, T] : mutable class
    key : K
    value : T
endtype

type Hashmap_bucketlist[K, T] : vec[^List[Hashmap_entry[K,T]]]

type Hashmap[K, T] : mutable class
    buckets : Hashmap_bucketlist[K, T]
    hashkeygenerator : [K -> int]
endtype

% Function: Hashmap
% 
% Constructor for the hashmap class
% 
% Usage:
%   : function [y : hashmap[K,T]] = Hashmap[K, T](hashkeygenerator : [K->int])
% 
% Parameters:
% hashkeygenerator - function that generates a hashcode (integer) for a given key
% 
% Returns:
% y - a new hashmap 
function y = Hashmap[K,T](hashkeygenerator : [K -> int])
    y = Hashmap[K,T](buckets:=Hashmap_bucketlist[K,T](127), _ 
                     hashkeygenerator:=hashkeygenerator)
endfunction

% Function: Hashmap[K,T].add
% 
% Adds a new entry to this hash map
% 
% Usage:
%   : function [] = Hashmap[K,T].add(key : K, value : T)
% 
% Parameters:
% key - the key to add
% value - the value to add
function [] = add[K,T](self : ^Hashmap[K,T], key : K, value : T)
    index = mod(self.hashkeygenerator(key), numel(self.buckets))
    if (self.buckets)[index] == nullptr
        (self.buckets)[index] = List[Hashmap_entry[K,T]]()
    endif
    (self.buckets)[index].add(Hashmap_entry[K,T](key:=key,value:=value))
endfunction

% Function: Hashmap[K,T].get
% 
% Hashmap lookup function, based on a given key
% 
% Usage:
%   : function [y] = Hashmap[K,T].get[K, T](key : K)
% 
% Parameters:
% key - the key to lookup
% 
% Returns:
% y - the value of the hashmap for entry 'key'
function y = get[K,T](self : ^Hashmap[K,T], key : K)
    index = mod(self.hashkeygenerator(key), numel(self.buckets))
    % TODO: static typing is necessary, the type inference engine
    % cannot deduce this itself...
    %lookup = t -> t.key == key
    lookup : [Hashmap_entry[K,T]->int] = t -> t.key == key
    q : int = self.buckets[index].find_first_if(lookup)
    if q<0
        error "The hashmap does not contain an item with key ", key
    endif
    y = self.buckets[index].at(q).value
endfunction

% Function: Hashmap[K,T].remove
% 
% Remotes an element from a hashmap
% 
% Usage:
%   : function [] = Hashmap[K,T].remove[K, T](self : ^hashmap[K,T], key)
% 
% Parameters:
% key - the key to remove
function [] = Remove[K,T](self : ^Hashmap[K,T], key : K)
    index = mod(self.hashkeygenerator(key), numel(self.buckets))
    lookup : [Hashmap_entry[K,T]->int] = t -> t.key == key
    q : int = (self.buckets)[index].find_first_if(lookup)
    if q >= 0
        self.buckets[index].remove_at(q)
    endif
endfunction

function [] = main()

    hashkeygenerator = (i : int) -> i
    
    map = Hashmap[int, scalar](hashkeygenerator)
    map.add(4, 1.234)
    map.add(3, 0.156)
    
    tic()
    for k=0..10000
        map.add(k, rand())
    endfor
    toc("Adding 10000 elements")
    
    print map.get(4)
    print map.get(3)
    
    hashkeygenerator2 : [scalar -> int] = i -> round(i)
    
    map2 = Hashmap[scalar, cscalar](hashkeygenerator2)
    map2.add(2.0, 1+2i)
    map2.add(2.888, 2-4i)
    print map2.get(2.0) %TODO: does not work with 2.1

endfunction
