%==============================================================================
% hough.q
% Hough transform in Quasar
% Simon Donné, April 2014
%==============================================================================
import "cannyedge.q"

{!author name="Simon Donné"}
{!doc category="Image Processing/Features"}

% Function: hough
%
% Implementation of the Hough transform for line detection.
%
% Usage:
%
% :function [y:mat] = hough(x:mat, thetas:vec, b_range:ivec2)
%
% Parameters:
% x - the input image, a single-channel image
% thetas - the angles to evaluate (vector of values)
% b_range - the interval of intersections with the top row to evaluate
% y - the resulting hough transform
% 
function [htf:mat] = hough(img:mat, thetas:vec, b_range:ivec2)
    n_theta = numel(thetas)
    n_b = (b_range[1]-b_range[0]+1)
    htf = zeros(n_theta, n_b)
    function [] = __kernel__ fillhtf(pos:ivec3)
        theta = thetas[pos[2]]
        b = round(-pos[0]*tan(theta) + pos[1])
        idx = (b - b_range[0])
        idxdown = int(idx)
        idxfrac = frac(idx)
        if(idx >= 0 && idx < n_b)
            htf[pos[2],idxdown] += img[pos[0],pos[1]]*(1-idxfrac)
            htf[pos[2],idxdown+1] += img[pos[0],pos[1]]*(idxfrac)
        endif
    end
    parallel_do([size(img,0), size(img,1), n_theta], fillhtf)
end

function [] = main()
    image = imread("lena_big.tif")[150..179,300..349,1]/255
    image = cannyedge(image, 1, 0.01, 0.05)
    %remove the edges of this image: canny is screwed up
    image = image[1..size(image, 0)-2, 1..size(image, 1)-2]
            
    houghtf = hough(image, -pi/2..0.01..pi/2, [0,1]*size(image, 1))
    imshow(image)
    imshow(houghtf)
end