{!author name="Bart Goossens"}
{!doc category="Image Processing/Image Transforms"}

import "linalg.q"

% Section: imaffinetransform.q
% This file contains functions implementing affine spatial image transforms


% Function: iminterpolate_nearest
% 
% Performs nearest neighbor interpolation on an image
% 
% Usage:
%   : function [y : scalar] = __device__ iminterpolate_nearest(z0 : cube'const, x : vec3'const)
% 
% Parameters:
% z0 - the image to be interpolated
% x - the sample position (with fractional component)
% 
% Returns:
% y - the interpolated sample of z0 at position x
function y = __device__ iminterpolate_nearest(z0 : cube, x : vec3)
    index = int(x[0..1])        
    y = z0[index[0],index[1],int(x[2])]
endfunction

% Function: iminterpolate_bilinear
% 
% Performs bilinear interpolation on an image
% 
% Usage:
%   : function [y : scalar] = __device__ iminterpolate_bilinear(z0 : cube'const, x : vec3'const)
% 
% Parameters:
% z0 - the image to be interpolated
% x - the sample position (with fractional component)
% 
% Returns:
% y - the interpolated sample of z0 at position x
function y = __device__ iminterpolate_bilinear(z0 : cube, x : vec3)
    p = x[0..1]
    index = int(p)        
    index2 = index+1
    fr = frac(p)
    y = z0[index[0],index[1],int(x[2])] * (1-fr[0]) * (1-fr[1]) + _
        z0[index[0],index2[1],int(x[2])] * (1 - fr[0]) * fr[1] + _
        z0[index2[0],index[1],int(x[2])] * fr[0] * (1 - fr[1]) + _
        z0[index2[0],index2[1],int(x[2])] * fr[0] * fr[1]
endfunction

% Function: iminterpolate_sinc_blackmanharris
% 
% Performs sinc interpolation on an image with a blackman harris window
% 
% Usage:
%   : function [y : scalar] = __device__ iminterpolation_sinc_blackmanharris[N:int](z0 : cube'const, x : vec3'const)
% 
% Parameters:
% z0 - the image to be interpolated
% x - the sample position (with fractional component)
% N - half the size of the window (e.g., 4)
% 
% Returns:
% y - the interpolated sample of z0 at position x
function q = __device__ iminterpolation_sinc_blackmanharris[N:int](img_in:cube, p:vec3)
    weight_x = zeros(2*N)
    weight_y = zeros(2*N)
    block = zeros(2*N)
    
    fraction = p[0..1] - floor(p[0..1]) + N
    position = int(floor(p[0..1])) - N
    
    ppi = pi/4
    
    % construction of weight matrix
    for i = 0..2*N-1
        pos_kernel = float(i) - fraction[0]
        coefficient = ppi*pos_kernel
        weight_x[i] = sinc(pos_kernel)*(0.35875 + 0.48829*cos(coefficient) + 0.14128*cos(2*coefficient) + 0.01168*cos(3*coefficient))
        
        pos_kernel = float(i) - fraction[1]
        coefficient = ppi*pos_kernel
        weight_y[i] = sinc(pos_kernel)*(0.35875 + 0.48829*cos(coefficient) + 0.14128*cos(2*coefficient) + 0.01168*cos(3*coefficient))
    endfor
    
    % separable resampling
    for i = 0..2*N-1
        for j = 0..2*N-1
            block[j] = block[j] + img_in[position[0] + i,position[1] + j,int(p[2])]*weight_y[i]
        endfor
    endfor
    
    q = dotprod(block, weight_x)
endfunction

% Function: imaffinetransform
% 
% Performs an affine spatial transform to the specified image
% 
% Usage:
%   : function [im_out : cube] = imaffinetransform(im_in : cube, A : mat(3,3), size_mode = "pad", bound_ext = "zero", interp_mode = "bilinear")
% 
% Parameters:
% im_in - the input image to be transformed
% A - a 3x3 affine transform matrix (scaling, rotation, translation, shear)
% size_mode - `"pad"`: the input image is padded, `"keep"`: retains the input image dimensions
%      `"twice"`: extends to twice the size
% bound_ext - `"zero"`: uses zero boundary extension, `"mirror"`: uses mirror extension
% interp_mode - selects the interpolation mode
% * `"nearest"`: use nearest neighbor interpolation
% * `"bilinear"`: use bilinear interpolation
% * `"hwtex_bilinear"`: use bilinear interpolation of the GPU hardware texturing unit
% * `"sinc_bh8"`: uses sinc interpolation with Blackman Harris window of length 16
% * `"sinc_bh4"`: uses sinc interpolation with Blackman Harris window of length 8
% 
% Returns:
% im_out - the transformed output matrix
%
% See also:
%    <make_translation_matrix>, <make_scaling_matrix>, <make_rotation_matrix>,
%    <make_fliphorizontally_matrix>, <make_flipvertically_matrix>
function im_out : cube = imaffinetransform(im_in : cube, A : mat(3,3), size_mode = "pad", bound_ext = "zero", interp_mode = "bilinear")

    % transform the boundary points
    p1 = transpose(A*transpose([0,0,1]))
    p2 = transpose(A*transpose([0,size(im_in,1)-1,1]))
    p3 = transpose(A*transpose([size(im_in,0)-1,0,1]))
    p4 = transpose(A*transpose([size(im_in,0)-1,size(im_in,1)-1,1]))
    
    % homogeneous coordinates
    p1 = p1 / p1[2]
    p2 = p2 / p2[2]
    p3 = p3 / p3[2]
    p4 = p4 / p4[2]
    
    % determine the dimensions of the output image
    match size_mode with 
    | "pad" ->
        pA = mindim([p1,p2,p3,p4],0)
        pB = maxdim([p1,p2,p3,p4],0)
        print pA, ";", pB
        sz = ceil((pB - pA)[0..1])
        offset = -0.5*(sz - size(im_in,0..1))
    | "keep" ->
        sz = size(im_in,0..1)
        offset = [0.,0.]
    | "twice" ->
        sz = size(im_in,0..1)*2
        offset = -0.25*sz
    | _ ->
        error "(imaffinetransform) Unknown size mode: ", size_mode        
    endmatch
    im_out = zeros(sz[0], sz[1], size(im_in,2))
    
    Ainv = inv(A)

    function [] = __kernel__ kernel(z0 : cube, y : cube, Ainv : mat, offset : vec2, fn : [__device__ (cube, vec3) -> scalar], pos : ivec3)
        q = pos[0..1] + offset
        p = [Ainv[0,0] * q[0] + Ainv[0,1] * q[1] + Ainv[0,2],
             Ainv[1,0] * q[0] + Ainv[1,1] * q[1] + Ainv[1,2] ,
             Ainv[2,0]        + Ainv[2,1]        + Ainv[2,2]]
             
        y[pos] = fn(z0, [p[0]/p[2],p[1]/p[2],pos[2]])
    endfunction
    
    function [] = __kernel__ kernel_hwtex_safe(z0 : cube'hwtex_linear'safe, y : cube, Ainv : mat, offset : vec2, pos : ivec3)
        q = pos[0..1] + offset
        p = [Ainv[0,0] * q[0] + Ainv[0,1] * q[1] + Ainv[0,2],
             Ainv[1,0] * q[0] + Ainv[1,1] * q[1] + Ainv[1,2] ,
             Ainv[2,0]        + Ainv[2,1]        + Ainv[2,2]]
             
        y[pos] = z0[p[0]/p[2],p[1]/p[2],pos[2]]
    endfunction
    
    function [] = __kernel__ kernel_hwtex_mirror(z0 : cube'hwtex_linear'mirror, y : cube, Ainv : mat, offset : vec2, pos : ivec3)
        q = pos[0..1] + offset
        p = [Ainv[0,0] * q[0] + Ainv[0,1] * q[1] + Ainv[0,2],
             Ainv[1,0] * q[0] + Ainv[1,1] * q[1] + Ainv[1,2] ,
             Ainv[2,0]        + Ainv[2,1]        + Ainv[2,2]]
             
        y[pos] = z0[p[0]/p[2],p[1]/p[2],pos[2]]
    endfunction
    
    match interp_mode with
    | "bilinear" ->
        match bound_ext with
        | "zero"   -> parallel_do(size(im_out), im_in, im_out, Ainv, offset, iminterpolate_bilinear, kernel)
        | "mirror" -> parallel_do(size(im_out), im_in, im_out, Ainv, offset, $specialize(iminterpolate_bilinear, type(z0,"cube'mirror")), kernel)
        endmatch
    | "nearest" ->
        parallel_do(size(im_out), im_in, im_out, Ainv, offset, iminterpolate_nearest, kernel)
    | "hwtex_bilinear" ->
        match bound_ext with
        | "zero"   -> parallel_do(size(im_out), im_in, im_out, Ainv, offset, kernel_hwtex_safe)
        | "mirror" -> parallel_do(size(im_out), im_in, im_out, Ainv, offset, kernel_hwtex_mirror)
        endmatch
    | "sinc_bh8" ->
        parallel_do(size(im_out), im_in, im_out, Ainv, offset, iminterpolation_sinc_blackmanharris[8], kernel)
    | "sinc_bh4" ->
        parallel_do(size(im_out), im_in, im_out, Ainv, offset, iminterpolation_sinc_blackmanharris[4], kernel)
    | _ ->
        error "(imaffinetransform) Unknown interpolation mode: ", interp_mode
    endmatch
    
endfunction

% Function: make_translation_matrix
% 
% Constructs a 3x3 affine translation matrix
% 
% Usage:
%   : function [Y : mat] = make_translation_matrix(X : vec2)
% 
% Parameters:
% X - the translation vector (a vector of length 2)
% 
% Returns:
% Y - a 3x3 affine translation matrix
function Y = make_translation_matrix(X : vec2)
    Y = [[1,0,X[0]],[0,1,X[1]],[0,0,1]]
endfunction

% Function: make_scaling_matrix
% 
% Constructs a 3x3 affine scaling matrix
% 
% Usage:
%   : function [Y : mat] = make_scaling_matrix(X : vec2)
% 
% Parameters:
% X - the scaling factors along each dimension (a vector of length 2)
% 
% Returns:
% Y - a 3x3 affine scaling matrix
function Y = make_scaling_matrix(X : vec2)
    Y = [[X[0],0,0],[0,X[1],0],[0,0,1]]
endfunction

% Function: make_fliphorizontally_matrix
% 
% Constructs a matrix to flip the image horizontally
% 
% Usage:
%   : function [Y : mat] = make_fliphorizontally_matrix()
% 
% Returns:
% Y - a 3x3 affine matrix to flip the image horizontally
function Y = make_fliphorizontally_matrix()
    Y = [[1,0,0],[0,-1,0],[0,0,1]]
endfunction

% Function: make_flipvertically_matrix
% 
% Constructs a matrix to flip the image vertically
% 
% Usage:
%   : function [Y : mat] = make_flipvertically_matrix()
% 
% Returns:
% Y - a 3x3 affine matrix to flip the image vertically
function Y = make_flipvertically_matrix()
    Y = [[-1,0,0],[0,1,0],[0,0,1]]
endfunction

% Function: make_rotation_matrix
% 
% Constructs a 3x3 affine rotation matrix
% 
% Usage:
%   : function [Y : mat] = make_rotation_matrix(theta : scalar)
% 
% Parameters:
% theta - the rotation angle (in radians)
% 
% Returns:
% Y - a 3x3 affine rotation matrix
function Y = make_rotation_matrix(theta : scalar)
    Y = [[cos(theta), -sin(theta), 0],
         [sin(theta),  cos(theta), 0],
         [0, 0, 1]]
endfunction

% Function: imrotate
% 
% Rotates the input image using the specified angle (in radians)
% 
% Usage:
%   : function [im_out : cube] = imrotate(im_in, theta, size_mode = "pad", bound_ext = "zero", interp_mode = "bilinear")
% 
% Parameters:
% im_in - the input image
% theta - the angle to rotate the image (in radians)
% size_mode - `"pad"`: the input image is padded, `"keep"`: retains the input image dimensions
% bound_ext - `"zero"`: uses zero boundary extension, `"mirror"`: uses mirror extension
% interp_mode - `"nearest"`: use nearest neighbor interpolation, `"bilinear"`: use bilinear interpolation
% 
% Returns:
% im_out - the rotated output image
function im_out = imrotate(im_in, theta, size_mode = "pad", bound_ext = "zero", interp_mode = "bilinear")

    sz = size(im_in,0..1)
    A =  make_translation_matrix(0.5*sz) *
         make_rotation_matrix(theta) *
         make_translation_matrix(-0.5*sz)
    im_out = imaffinetransform(im_in, A, size_mode, bound_ext, interp_mode)
    
endfunction

% Function: imscale
% 
% Scales the input image horizontally and vertically
% 
% Usage:
%   : function [im_out : cube] = imscale(im_in, sc : vec2, size_mode = "pad", bound_ext = "zero", interp_mode = "bilinear")
% 
% Parameters:
% im_in - the input image
% sc - the scaling factor (either a scalar, or a vector of length 2, where each component indicates the scaling factor in the specified dimension)
% size_mode - `"pad"`: the input image is padded, `"keep"`: retains the input image dimensions
% bound_ext - `"zero"`: uses zero boundary extension, `"mirror"`: uses mirror extension
% interp_mode - `"nearest"`: use nearest neighbor interpolation, `"bilinear"`: use bilinear interpolation
% 
% Returns:
% im_out - the scaled output image
function im_out = imscale(im_in, sc, size_mode = "pad", bound_ext = "zero", interp_mode = "bilinear")

    if type(sc, scalar)
        sc = [sc, sc]
    endif

    sz = size(im_in,0..1)
    A =  make_translation_matrix(0.5*sz) *
         make_scaling_matrix(sc) *
         make_translation_matrix(-0.5*sz)
    im_out = imaffinetransform(im_in, A, size_mode, bound_ext, interp_mode)
    
endfunction

function [] = main()

    im = imread("lena_big.tif")
    theta = pi/30
    
    sz = size(im,0..1)
    A =  make_translation_matrix(0.5*sz) *
         make_rotation_matrix(theta) *
         make_flipvertically_matrix() *
         make_translation_matrix(-0.5*sz)
    tic()
    im_out = imaffinetransform(im, A, "pad", "mirror", "sinc_bh8")
    toc()
    imshow(im_out)
    
    im_out = imscale(im, [1.2, 1.2]/2)
    imshow(im_out)

endfunction