%==============================================================================
% imfilter.q
% Image filtering functions
%==============================================================================
import "system.q"
import "imboundary.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Filters"}

% Section: imfilter.q
% This file contains very useful functions for linear image filtering, as well
% as routines to check whether a given filter mask is separable.


% Function: is_filtermask_separable
%
% Checks whether a given kernel is separable, and if so, returns the separable components
%
% : function [result, mask_hor, mask_ver] = is_filtermask_separable(kernel)
%
% Parameters:
%   kernel - the filter mask to check
%   result - a boolean value indicating whether the mask is separable
%   mask_hor - the horizontal filter mask (only when the filter is separable)
%   mask_ver - the vertical filter mask (only when the filter is separable)
%
% See also:
%   <imfilter>
%
function [result, mask_hor, mask_ver] = is_filtermask_separable(kernel : mat)
    if ndims(kernel)==1 % row or column filter
        result = true
        if size(kernel,0)==1
            mask_hor = kernel
            mask_ver = ones(1)
        else
            mask_hor = ones(1)
            mask_ver = kernel
        endif
        return
    endif 
    nz = find(kernel != 0, "first")
    if size(nz,0) == 0  % kernel is zero, hence separable
        result = true  
        mask_hor = zeros(1, size(kernel, 1))
        mask_ver = zeros(size(kernel, 0), 1)
        return
    endif
    nz = nz[0,:] % the first match is sufficient for decomposing...
    sep_kernel = kernel[:,nz[1]] * kernel[nz[0],:] / kernel[nz[0],nz[1]]
    if sum(abs(sep_kernel - kernel))<1e-7
        mask_ver = kernel[:,nz[1]]
        mask_hor = kernel[nz[0],:] / kernel[nz[0],nz[1]]
        result = true
    else        
        mask_hor = mask_ver = [0]
        result = false
    endif
endfunction

% Function: imfilter_separable
%
% Applies a 2D separable impulse response (FIR) filter to an image
% : function y = imfilter_separable(x, vmask, hmask, center = 0, boundary_mode = "zero")
%
% Parameters:
% x         - the input image
% vmask     - the vertical filter mask
% hmask     - the horizontal filter mask
% center    - a 2D vector containing the center of the filter mask
% boundary_mode - the boundary extension mode to be used for filtering
%    o "zero": the image is extended with zeros outside its boundaries
%    o "mirror": values outside the image are obtained by mirroring
%    o "circular": values outside the image are obtained by circularly 
%                   shifting (resulting in a periodic extension)
%    o "clamped": values outside of the image are obtained from the 
%                   borders instead.
% Notes:
% - The actual 2D mask can be calculated by the formula:
% : mask = transpose(vmask) * hmask.
% - The function skip the separability check, and may slightly be faster
%   than imfilter for separable filters
%
% See also:
%    <periodize>, <mirror_ext>, <is_filtermask_separable>, <imfilter>
%
function y = imfilter_separable(x, vmask, hmask, center = [0,0], boundary_mode = "zero")

    % Separable filter
    % NOTE - this is a specialized algorithm that efficiently makes use of the shared memory
    function [] = __kernel__ imfilter_kernel_shared_hor(x : cube, y : cube, fc : vec'hwconst'unchecked, n : int, N : int, get : boundextfunction, blkidx : vec3, blkpos : vec3, blkdim : vec3)
        pos = blkidx .* [blkdim[0]*4, blkdim[1], blkdim[2]] + blkpos
        {!specialize args=N==1 || N==2 || N==3 || N==4 || N==5 || N==6 || N==7 || N==8 || N==12 || N==16}
        % store in shared memory
        x_sh = shared(4, 128)
        y_sh = shared(4, 128)
        z_sh = shared(4, 128)
        u_sh = shared(4, 128)
        x_sh[blkpos[0..1]] = get(x,pos+[0,-n,0])
        y_sh[blkpos[0..1]] = get(x,pos+[blkdim[0],-n,0])
        z_sh[blkpos[0..1]] = get(x,pos+[2*blkdim[0],-n,0])
        u_sh[blkpos[0..1]] = get(x,pos+[3*blkdim[0],-n,0])
        if blkpos[1] <= N
            x_sh[blkpos[0], blkpos[1] + blkdim[1]] = get(x,pos+[0,          blkdim[1]-n,0])
            y_sh[blkpos[0], blkpos[1] + blkdim[1]] = get(x,pos+[blkdim[0],  blkdim[1]-n,0])
            z_sh[blkpos[0], blkpos[1] + blkdim[1]] = get(x,pos+[2*blkdim[0],blkdim[1]-n,0])
            u_sh[blkpos[0], blkpos[1] + blkdim[1]] = get(x,pos+[3*blkdim[0],blkdim[1]-n,0])
        endif
        syncthreads
        
        sum = [0.,0.,0.,0.]
        for i=0..N-1
            sum += [x_sh[blkpos[0], blkpos[1] + i], _
                    y_sh[blkpos[0], blkpos[1] + i], _
                    z_sh[blkpos[0], blkpos[1] + i], _
                    u_sh[blkpos[0], blkpos[1] + i]] * fc[i]
        endfor
        y[pos] = sum[0]
        y[pos+[blkdim[0],0,0]] = sum[1]
        y[pos+[2*blkdim[0],0,0]] = sum[2]
        y[pos+[3*blkdim[0],0,0]] = sum[3]
    endfunction

    function [] = __kernel__ imfilter_kernel_shared_ver(x : cube, y : cube, fc : vec'hwconst'unchecked, n : int, N : int, get : boundextfunction, blkidx : vec3, blkpos : vec3, blkdim : vec3)
        pos = blkidx .* [blkdim[0], blkdim[1]*4, blkdim[2]] + blkpos
        {!specialize args=N==1 || N==2 || N==3 || N==4 || N==5 || N==6 || N==7 || N==8 || N==12 || N==16}
        
        % store in shared memory
        x_sh = shared(128, 4)
        y_sh = shared(128, 4)
        z_sh = shared(128, 4)
        u_sh = shared(128, 4)
        x_sh[blkpos[0..1]] = get(x,pos+[-n,0,0])
        y_sh[blkpos[0..1]] = get(x,pos+[-n,blkdim[1],0])
        z_sh[blkpos[0..1]] = get(x,pos+[-n,2*blkdim[1],0])
        u_sh[blkpos[0..1]] = get(x,pos+[-n,3*blkdim[1],0])
        if blkpos[0] < N
            x_sh[blkpos[0] + blkdim[0], blkpos[1]] = get(x,pos+[blkdim[0]-n,0,0])
            y_sh[blkpos[0] + blkdim[0], blkpos[1]] = get(x,pos+[blkdim[0]-n,blkdim[1],0])
            z_sh[blkpos[0] + blkdim[0], blkpos[1]] = get(x,pos+[blkdim[0]-n,2*blkdim[1],0])
            u_sh[blkpos[0] + blkdim[0], blkpos[1]] = get(x,pos+[blkdim[0]-n,3*blkdim[1],0])
        endif
        syncthreads
        
        sum = [0.,0.,0.,0.]
        for i=0..N-1
            sum += [x_sh[blkpos[0] + i, blkpos[1]], _
                    y_sh[blkpos[0] + i, blkpos[1]], _
                    z_sh[blkpos[0] + i, blkpos[1]], _
                    u_sh[blkpos[0] + i, blkpos[1]]] * fc[i]
        endfor
        y[pos] = sum[0]
        y[pos+[0,blkdim[1],0]] = sum[1]
        y[pos+[0,2*blkdim[1],0]] = sum[2]
        y[pos+[0,3*blkdim[1],0]] = sum[3]
    endfunction
    
    % 1D Horizontal filter kernel
    function [] = __kernel__ imfilter_kernel_hor(y : cube'unchecked, _
            x : cube'hwtex_const, mask : vec'unchecked'hwconst, center : int, get : boundextfunction, pos : ivec3)
        sum = 0.0
        for m=0..numel(mask)-1
            sum += mask[m] * get(x,pos + [0,m-center,0])
        endfor
        y[pos] = sum
    endfunction

    % 1D Vertical filter kernel
    function [] = __kernel__ imfilter_kernel_ver(y : cube'unchecked, _
            x : cube'hwtex_const, mask : vec'unchecked'hwconst, center : int, get : boundextfunction, pos : ivec3)
        sum = 0.0
        for m=0..numel(mask)-1
            sum += mask[m] * get(x,pos + [m-center,0,0])
        endfor
        y[pos] = sum
    endfunction
    
    y = uninit(size(x))
    z = uninit(size(x))

    % If the size of the filter mask becomes larger than the block size,
    % it is more efficient to use the direct separable approach, without
    % shared memory.
    % The factor 2 is to allow the overlap size to be of the same order as the block size
    if numel(hmask)>512 || numel(vmask)>64 || gpu_count() == 0
        % Select the correct kernel, depending on the boundary mode
        match boundary_mode with
        | "zero"     -> parallel_do(size(x,0..2),z,x,hmask,center[1],boundext_zeros,imfilter_kernel_hor)
                        parallel_do(size(x,0..2),y,z,vmask,center[0],boundext_zeros,imfilter_kernel_ver)
        | "mirror"   -> parallel_do(size(x,0..2),z,x,hmask,center[1],boundext_mirror,imfilter_kernel_hor)
                        parallel_do(size(x,0..2),y,z,vmask,center[0],boundext_mirror,imfilter_kernel_ver)
        | "circular" -> parallel_do(size(x,0..2),z,x,hmask,center[1],boundext_circular,imfilter_kernel_hor)
                        parallel_do(size(x,0..2),y,z,vmask,center[0],boundext_circular,imfilter_kernel_ver)
        | "clamped"  -> parallel_do(size(x,0..2),z,x,hmask,center[1],boundext_clamped,imfilter_kernel_hor)
                        parallel_do(size(x,0..2),y,z,vmask,center[0],boundext_clamped,imfilter_kernel_ver)
        | _ -> error ("Unknown boundary extension mode!")
        endmatch
    else
        [y1,x1] = int((size(y,0..1)+15)/4)
        [y2,x2] = int((size(y,0..1)+63)/64)*64
        match boundary_mode with
        | "zero"     -> parallel_do ([[y1,x2,size(y,2)],[4,64,1]], x, z, hmask,center[1],numel(hmask),boundext_zeros,imfilter_kernel_shared_hor)
                        parallel_do ([[y2,x1,size(y,2)],[64,4,1]], z, y, vmask,center[0],numel(vmask),boundext_zeros,imfilter_kernel_shared_ver) 
        | "mirror"   -> parallel_do ([[y1,x2,size(y,2)],[4,64,1]], x, z, hmask,center[1],numel(hmask),boundext_mirror,imfilter_kernel_shared_hor)
                        parallel_do ([[y2,x1,size(y,2)],[64,4,1]], z, y, vmask,center[0],numel(vmask),boundext_mirror,imfilter_kernel_shared_ver) 
        | "circular" -> parallel_do ([[y1,x2,size(y,2)],[4,64,1]], x, z, hmask,center[1],numel(hmask),boundext_circular,imfilter_kernel_shared_hor)
                        parallel_do ([[y2,x1,size(y,2)],[64,4,1]], z, y, vmask,center[0],numel(vmask),boundext_circular,imfilter_kernel_shared_ver) 
        | "clamped"  -> parallel_do ([[y1,x2,size(y,2)],[4,64,1]], x, z, hmask,center[1],numel(hmask),boundext_clamped,imfilter_kernel_shared_hor)
                        parallel_do ([[y2,x1,size(y,2)],[64,4,1]], z, y, vmask,center[0],numel(vmask),boundext_clamped,imfilter_kernel_shared_ver) 
        | _ -> error ("Unknown boundary extension mode!")
        endmatch
    endif
endfunction

% Function: imfilter
%
% Applies a 2D finite impulse response (FIR) filter to an image.
%
% : function y = imfilter(x, mask, center = 0, boundary_mode = "zero")
%
% Parameters:
% x         - the input image
% mask      - the filter mask (i.e. filter impulse response)
% center    - a 2D vector containing the center of the filter mask
% boundary_mode - the boundary extension mode to be used for filtering
%    o "zero": the image is extended with zeros outside its boundaries
%    o "mirror": values outside the image are obtained by mirroring
%    o "circular": values outside the image are obtained by circularly 
%                   shifting (resulting in a periodic extension)
%
% Notes:
% The function automatically checks whether the 2D filter mask is 
% separable. If this is the case, a more efficient separable implementation
% is used.
%
% See also:
%    <periodize>, <mirror_ext>, <is_filtermask_separable>, <imfilter_separable>
%
function y = imfilter(x : [mat | cube], mask, center = 0, boundary_mode = "zero")

    % Non separable filter kernel
    function [] = __kernel__ imfilter_kernel_nonsep(y : cube'unchecked, _
            x : cube, get : boundextfunction, mask : mat'unchecked'const'hwconst, center : ivec2, pos : ivec3)
        sum = 0.0
        for m=0..size(mask,0)-1
            for n=0..size(mask,1)-1
                sum += mask[m,n] * get(x,pos + [m-center[0],n-center[1],0])
            endfor
        endfor
        y[pos] = sum
    endfunction

    if isscalar(center)
        center = [0,0]
    endif        

    % Try to decompose the mask in horizontal & vertical components
    [is_separable, mask_hor, mask_ver] = is_filtermask_separable(mask)
    if is_separable 
        y = imfilter_separable(x, mask_ver, mask_hor, center, boundary_mode)       
    else
        y = uninit(size(x))
        % Select the correct kernel, depending on the boundary mode
        match boundary_mode with
        | "zero"     -> parallel_do(size(x),y,x,boundext_zeros,mask,center,imfilter_kernel_nonsep)
        | "mirror"   -> parallel_do(size(x),y,x,boundext_mirror,mask,center,imfilter_kernel_nonsep)
        | "circular" -> parallel_do(size(x),y,x,boundext_circular,mask,center,imfilter_kernel_nonsep)
        | "clamped"  -> parallel_do(size(x),y,x,boundext_clamped,mask,center,imfilter_kernel_nonsep)
        | _ -> error ("Unknown boundary extension mode!")
        endmatch
    endif 
endfunction

% Function: hor_sobel
%
% Extracts the horizontal edges of an image using a horizontal sobel filter
%
% :function y = hor_sobel(x)
function y = hor_sobel(x : [mat | cube])
    y = imfilter_separable(x, [1,0,-1], [1,2,1], [1, 1])
endfunction

% Function: ver_sobel
%
% Extracts the vertical edges of an image using a vertical sobel filter
%
% :function y = ver_sobel(x)
function y = ver_sobel(x : [mat | cube])
    y = imfilter_separable(x, [1,2,1], [1,0,-1], [1, 1])
endfunction

% Function: sobel
%
% Find the edge magnitude using the Sobel filters.
%
% :function y = sobel(x)
function y = sobel(x : [mat | cube])
    y = sqrt(hor_sobel(x).^2 + ver_sobel(x).^2)
endfunction

% Function: hor_scharr
%
% Extracts the horizontal edges of an image using a horizontal scharr filter
%
% :function y = hor_scharr(x)
function y = hor_scharr(x : [mat | cube])
    y = imfilter_separable(x, [1,0,-1], [3,10,3]/16, [1, 1])
endfunction

% Function: ver_scharr
%
% Extracts the vertical edges of an image using a vertical scharr filter
%
% :function y = ver_scharr(x)
function y = ver_scharr(x : [mat | cube])
    y = imfilter_separable(x, [3,10,3]/16, [1,0,-1], [1, 1])
endfunction

% Function: scharr
%
% Find the edge magnitude using the scharr filters.
%
% :function y = scharr(x)
function y = scharr(x : [mat | cube])
    y = sqrt(hor_scharr(x).^2 + ver_scharr(x).^2)
endfunction

% Function: hor_prewith
%
% Extracts the horizontal edges of an image using a horizontal prewith filter
%
% :function y = hor_prewith(x)
function y = hor_prewith(x : [mat | cube])
    y = imfilter_separable(x, [1,0,-1], [1,1,1], [1, 1])
endfunction

% Function: ver_prewith
%
% Extracts the vertical edges of an image using a vertical prewith filter
%
% :function y = ver_prewith(x)
function y = ver_prewith(x : [mat | cube])
    y = imfilter_separable(x, [1,1,1], [1,0,-1], [1, 1])
endfunction

% Function: prewith
%
% Find the edge magnitude using the prewith filters.
%
% :function y = prewith(x)
function y = prewith(x : [mat | cube])
    y = sqrt(hor_prewith(x).^2 + ver_prewith(x).^2)
endfunction

% Function: gaussian_filter
%
% Applies a Gaussian filter to the image
%
% :function y=gaussian_filter(x,sigma,n=7)
%
% Parameters:
% x - the input image
% sigma - the sigma parameter of the gaussian filter
% n - the number of filter taps (divided by 2). The filter support is -n..n.
% 
function y=gaussian_filter(x : [mat | cube],sigma : scalar,n : scalar = 0, boundary_mode : string = "zero")
    % Compute the filter kernel
    if n==0
        n = floor(1.5*sigma)
    endif
    fc = exp(-0.5/sigma^2*(-n..n).^2)
    fc = fc / sum(fc)
    y = imfilter_separable(x,fc,fc,[n,n],boundary_mode)
endfunction


% Function: gaussianfilter_kernel_2d
% 
% Generates an isotropic or anisotropic kernel for a 2D Gaussian filter
% 
% Usage:
%   : function [kernel : mat] = gaussianfilter_kernel_2d(sz : vec2, sigma : ivec2, theta = 0)
% 
% Parameters:
% sz - the size of the filter mask. The origin is at (sz/2)
% sigma - the standard deviation of the Gaussian function along each dimension. When the two
%   components of sigma differ, an anisotropic kernel is being created
% theta - 
% 
% Returns:
% kernel - 
function [kernel:mat] = gaussianfilter_kernel_2d(sz:vec2,sigma:ivec2,theta:scalar=0)
    [M,N] = sz[0..1]
    if theta == 0
        fx = exp(-0.5/sigma[1]^2*((0..N-1)-N/2).^2)
        fx = fx / sum(fx)
        fy = exp(-0.5/sigma[0]^2*((0..M-1)-M/2).^2)
        fy = fy / sum(fy)
        kernel = transpose(fy)*fx
    else        
        [x,y] = meshgrid((0..N-1)-N/2, (0..M-1)-M/2)
        kernel = exp(-0.5*((x*sin(theta)+y*cos(theta))/sigma[0]).^2-0.5*((x*cos(theta)-y*sin(theta))/sigma[1]).^2 )
    endif
endfunction

% Function: im_fftfreqresponse
% 
% Calculates the DFT domain frequency response corresponding to a given filter mask,
% using the Fast Fourier transform (FFT). Once calculated, several images can quickly
% be filtered in the DFT domain using the same filter H.
% 
% Usage:
%   : function [H : ccube] = im_fftfreqresponse(h : mat, sz : ivec3, center = 0)
% 
% Parameters:
% h - the filter mask (in the image domain)
% sz - the size of the resulting DFT filter
% center - the center of the filter mask
% 
% Returns:
% H - the DFT domain representation of the filter mask
function H = im_fftfreqresponse(h : mat, sz : ivec3, center = 0)

    if isscalar(center)
        center = [0,0]
    endif    
    
    % Create the filter mask
    h_ext : cube = zeros(sz)
    translate : ivec2 = center
    parallel_do([size(h,0),size(h,1),size(h_ext,2)], h, __kernel__ (h : mat, pos : ivec3) -> 
        h_ext[periodize(pos[0]-translate[0],size(h_ext,0)),
              periodize(pos[1]-translate[1],size(h_ext,1)),pos[2]] = h[pos[0..1]])
    H = fft2(h_ext)
                  
endfunction

% Function: im_fftfilter
% 
% Performs an image filtering in the DFT domain. Due to the nature of the DFT,
% only circular boundary extension is supported by this function.
% 
% Usage:
%   : function [y] = im_fftfilter(x : [mat|cube], H : ccube)
% 
% Parameters:
% x - the input image
% H - the DFT domain representation of the filter (see <im_fftfreqresponse>)
% 
% Returns:
% y - the filtered output image
function y = im_fftfilter(x : [mat | cube], H : ccube)
    y = real(ifft2(fft2(x) .* H))
endfunction

% Function: im_fftfilter_bymask
% 
% Performs an image filtering in the DFT domain, provided a mask in the image domain.
% This function combines the functions <im_fftfreqresponse> and <im_fftfilter>. In 
% addition, the function also implements several boundary extension modes, leading
% to a result that is similar to <imfilter>. This function is more efficient than
% <imfilter> when the filter mask is large (e.g. 30x30).
% 
% Usage:
%   : function [y] = im_fftfilter_bymask(x : [mat|cube], h : mat, center = 0)
% 
% Parameters:
% x - the input image
% h - the input image filter mask
% center - a 2D vector, containing the center of the filter mask (default: [0,0])
% boundary_mode - the boundary extension mode to be used for filtering
%    o "zero": the image is extended with zeros outside its boundaries
%    o "mirror": values outside the image are obtained by mirroring
%    o "circular": values outside the image are obtained by circularly 
%                   shifting (resulting in a periodic extension)
% Returns:
% y - the filtered output image
%
% See also:
%    <imfilter>
function y = im_fftfilter_bymask(x : [mat | cube], h : mat, center : ivec2, boundary_mode="circular")
    if boundary_mode=="circular"
        H = im_fftfreqresponse(h, size(x,0..2), center)
        y = real(ifft2(fft2(x) .* H))
    else
        % need to perform some boundary extension
        x_ext = zeros(size(x,0..2) + [size(h,0),size(h,1),0])
    
        %kernel = __kernel__ (x : cube, translate : vec3, get : boundextfunction, pos : ivec3) ->                
        %        x_ext[pos] = get(x, pos - translate)
        
        kernel = __kernel__ (x : cube, translate : ivec3, get : boundextfunction, pos : ivec3) ->                
               x_ext[pos] = get(x, pos - translate)
    
        match boundary_mode with
        | "zero"     -> parallel_do(size(x_ext),x,[center[0],center[1],0],boundext_zeros, kernel)
        | "mirror"   -> parallel_do(size(x_ext),x,[center[0],center[1],0],boundext_mirror, kernel)
        | "circular" -> parallel_do(size(x_ext),x,[center[0],center[1],0],boundext_circular, kernel)
        | "clamped"  -> parallel_do(size(x_ext),x,[center[0],center[1],0],boundext_clamped, kernel)
        | _ -> error ("Unknown boundary extension mode!")
        endmatch
        
        H = im_fftfreqresponse(h, size(x_ext,0..2), center)
        y = real(ifft2(fft2(x_ext) .* H))[center[0] + (0..size(x,0)-1), center[1] + (0..size(x,1)-1),:]
        
    endif
endfunction

% Function: boxfilter
% 
% Calculates a box filter on an image, using an O(1) complexity algorithm
% 
% Usage:
%   : function [im_out : cube] = boxfilter(im_in : cube'const, wnd : vec'const(2), normalize : int = 0)
% 
% Parameters:
% im_in - the input image
% half_wnd - half the size of the window (the actual window is `-half_wnd[0]..halfwnd[0]` x 
%    `-half_wnd[1]..half_wnd[1]`
% normalize - if 1, the output of the box filter is normalized (to be a mean filter)
% 
% Returns:
% im_out - the output image
% 
% Notes:
% - this function uses zero boundary extension
function im_out:cube = boxfilter(im_in:cube, half_wnd : vec(2), normalize : int = 0)
    if all(half_wnd == [3, 3])
        function [] = __kernel__ filter3x3_kernel_auto(im_in:cube,im_out:cube,pos:ivec3,blkpos:ivec3,blkdim:ivec3)        
            {!kernel_transform enable="localwindow"}
            im_out[pos] = (im_in[pos+[-1,-1,0]]+im_in[pos+[-1,0,0]]+im_in[pos+[-1,1,0]] + _
                           im_in[pos+[ 0,-1,0]]+im_in[pos         ]+im_in[pos+[0,1,0]] + _
                           im_in[pos+[ 1,-1,0]]+im_in[pos+[ 1,0,0]]+im_in[pos+[1,1,0]]) * (1.0/9)
        endfunction

        parallel_do(size(im_out,0..2),im_in,im_out,filter3x3_kernel_auto)
    else

        {!function_transform enable="kernel_generator"; threshold = 0}
        [M,N,C] = size(im_in,0..2)
        im_out = zeros(M,N,C)
        [r,s] = half_wnd
                
        %cumulative sum over Y axis 
        im_cum = cumsum(im_in,0)
        %difference over Y axis
        im_out[0..r,:,:] = im_cum[1+r-1..2*r,:,:]
        im_out[r+1..M-r-1,:,:] = im_cum[2*r+1..M-1,:,:] - im_cum[0..M-2*r-2,:,:]
        im_out[M-r..M-1,:,:] = repmat(im_cum[M-1,:,:], [r, 1, 1]) - im_cum[M-2*r-1..M-r-2,:,:]
            
        %Acumulative sum over X axis
        im_cum = cumsum(im_out,1)
        
        %difference over Y axis
        im_out[:,0..s,:] = im_cum[:, 1+s-1..2*s,:]
        im_out[:,s+1..N-s-1,:] = im_cum[:,2*s+1..N-1,:]-im_cum[:,0..N-2*s-2,:]
        im_out[:,N-s..N-1,:] = repmat(im_cum[:,N-1,:], [1, s, 1]) - im_cum[:, N-2*s-1..N-s-2,:]
        
        if normalize
            im_out *= (1.0 / prod(2*half_wnd+1))
        endif
            
    endif

endfunction

function [] = main()
    
    im = imread("lena_big.tif") 
    sz = [20,20]
    kernel = [[0,-1,0],[-1,4,-1],[0,-1,0]]
    kernel = gaussianfilter_kernel_2d(sz,[10,6],pi/2)
    kernel = kernel/sum(kernel)
    runs = 200
    tic()
    for it=1..runs
        im1 = im_fftfilter_bymask(im, kernel, sz/2, "circular")
    endfor
    toc("im_fftfilter_bymask")
    tic()
    for it=1..runs
        im2 = imfilter(im, kernel, sz/2, "circular")    
    endfor
    toc("imfilter")
    imshow(im1); title("im1")
    imshow(im2); title("im2")
    imshow(im1-im2)
    print mean(abs(im1-im2))
    return
   
    %kernel = [[0,-1,0],[-1,4,-1],[0,-1,0]]
    kernel = transpose([1,2,2,2,1])*[1,2,2,2,1]/64
    

    tic()
    for k=1..runs
        y = imfilter(im, kernel, [2,2], "zero")
    endfor
    toc()

    tic()
    for k=1..runs
        y = imfilter_separable(im, [1,2,2,2,1]/8, [1,2,2,2,1]/8, [2,2], "mirror")
        %y = gaussian_filter(im, 1, boundary_mode:="mirror")
    endfor
    toc()

    imshow(y,[0, 255])    

    %A = zeros(256,256)
    %A[126..4..134,126..4..134] = 1
    %print find(A)

    %[result, mask_hor, mask_ver] = is_filtermask_separable(transpose([1,2,1])*[1,2,1]/4)
    [result, mask_hor, mask_ver] = is_filtermask_separable([1,2,1]/4)
    print result, ",", mask_hor, ",", mask_ver

endfunction
