%==============================================================================
% imhist.q
% Computation of image histograms & histogram egalization
% Bart Goossens, 2012
%==============================================================================
import "system.q"
import "colortransform.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Histograms"}

% Section: imhist.q
% This file contains routines for computing 1D and 2D histograms of images and
% performing histogram equalization.

% Function: imhist
% Computes the histogram of an image
%
% : function y = imhist(x, bins=256)
%
% Parameters:
% x - the input image
% bins - either 1) the number of bins to be used, or 2) a vector containing
%        the boundaries of the bins.
% y - the output histogram (has size C x num_bins), where C is the number of
%     color components of x, i.e. C = size(x,2)
function [y:cube] = imhist(x:cube, bins=256)

    if isscalar(bins)
        num_bins : int = bins
        bins=[0, round(256/bins)]
    else
        num_bins = numel(bins)
    endif

    minval : scalar = bins[0]
    denom  : scalar = 1 / (bins[1] - bins[0])

    if num_bins <= 256
        y = zeros(size(x, 2), num_bins)
        for m=0..size(x,0)-1
            for n=0..size(x,1)-1
                for k=0..size(x,2)-1
                    {!kernel_transform enable="sharedmemcaching"}
                    {!kernel_arg name=y; type="inout"; access="shared"; op="+="; cache_slices=y[:,:,0]; numel=768}
                    val = int((x[m,n,k]-minval)*denom)
                    if val>=0 && val<num_bins
                        y[k,val,0] += 1
                    endif
                end
            end
        end
    else
        nThreads = 4
        y = zeros(size(x, 2), num_bins, nThreads)
        for m=0..size(x,0)-1
            for n=0..size(x,1)-1
                for k=0..size(x,2)-1
                    val2 = int((x[m,n,k]-minval)*denom)
                    thread_idx = mod(pos2ind(size(x,0..2), [m,n,k]),nThreads)
                    if val2>=0 && val2<num_bins
                        y[k,val2,thread_idx] += 1
                    endif
                end
            end
        end
        y = sum(y, 2)
    endif
end

% Function: imjointhist
% Computes the joint histogram of two images
%
% : function h : mat = imjointhist(x, y, bins_x = 256, bins_y = 256)
%
% Parameters:
% x - input image #1
% y - input image #2
% bins_x - either 1) the number of bins to be used for x, or 2) a 
%		 linearly spaced vector containing the bins of x.
% bins_y - either 1) the number of bins to be used for y, or 2) a 
%		 linearly spaced vector containing the bins of y.
% h - a 2D joint histogram
function h : mat = imjointhist(x, y, bins_x = 256, bins_y = 256)
    if isscalar(bins_x)
        bins_x=0..floor(bins_x)-1
    endif
    if isscalar(bins_y)
        bins_y=0..floor(bins_y)-1
    endif
    h = zeros(numel(bins_y),numel(bins_x))
    if any(size(x) != size(y))
        error "(imjointhist) x and y should have the same dimensions!"
    endif
    function [] = __kernel__ imjointhist_kernel(x : cube'unchecked, y : cube'unchecked, h : mat'unchecked, _
            x_min : scalar, y_min : scalar, x_step : scalar, y_step : scalar, pos : ivec3)
        
        xind = int((x[pos]-x_min)*x_step)
        yind = int((y[pos]-y_min)*y_step)
        if xind>=0 && xind<size(h,1) && yind>=0 && yind<size(h,0)
            h[yind,xind] += 1.0
        endif
    end

    parallel_do(size(x),x,y,h,bins_x[0],bins_y[0], _
        1/(bins_x[1]-bins_x[0]),1/(bins_y[1]-bins_y[0]),imjointhist_kernel)
end

% Function: imequalizehist
%
% Performs histogram equalization
% 
% : function y = imequalizehist(x, mode = "yuv")
%
% Parameters:
% x - input image
% mode - the histogram equalization mode
%   o "rgb" - use histogram equalization in RGB space, where each color component is
%             processed individually.
%   o "grayscale" - histogram equalization of grayscale images. In case an RGB image is 
%             specified, the image is first converted to grayscale.
%   o "yuv" - performs histogram equalization on the Y-channel of the input image.
function y = imequalizehist(x, mode = "yuv")
    is_rgb = size(x,2)==3
    if !is_rgb
        mode = "grayscale"
    elseif is_rgb && mode == "grayscale"
        x = rgb2gray(x)
    endif

    n = 256
    p = linspace(min(x),max(x),n) 
    imh = imhist(x,p) / prod(size(x,0..1)) * 255 % Compute normalized histogram

    match mode with
    | ("rgb", "grayscale") ->
        f = cumsum(imh)
    | "yuv" ->
        lumh = [0.2126, 0.7152, 0.0722] * imh[:,:,0]   % RGB image - compute luminance histogram    
        f = cumsum(lumh)
    | _ -> 
        error "(imequalizehist): invalid mode!"
    end   

    % Normalize f so that it has the same range as the inputn
    f = p[0] + (p[n-1] - p[0]) * f / n 

    % Now - map the values onto the respective output. 
    match mode with
    | "grayscale" ->
        y = interp1(p,f,x,"linear")
    | "rgb" ->
        y = zeros(size(x))
        y[:,:,0] = interp1(p,f[0,:],x[:,:,0],"linear")
        y[:,:,1] = interp1(p,f[1,:],x[:,:,1],"linear")
        y[:,:,2] = interp1(p,f[2,:],x[:,:,2],"linear")
    | "yuv" ->
        [U,Ut] = get_colortransform_from_RGB("YUV")
        y = colortransform(U, x)
        y[:,:,0] = interp1(p,f,y[:,:,0],"linear") % Only apply to the Y channel
        y = colortransform(Ut, y)                
    end
end

function [] = main()
    im = imread("lena_big.tif")%[:,:,0]
    tic()
    for k=0..1999
    y = imhist(im)
    end    
    toc()
    plot(y)
    print sum(y), ",", numel(im)

    return
    tic()
    h = imjointhist(im[:,:,0],im[:,:,2],0..255,0..255)
    toc()
    imshow(h,[])
    return
    %plot(y)
    title("Histogram")
    tic()
    imeq = imequalizehist(im, "rgb")
    toc()
    imshow(imeq)
    %plot(imhist(imeq))
end
