
{!author name="Bart Goossens, Simon Donné"}
{!doc category="Image Processing/Filters"}

% Function: immedfilt
%
% Median filter for images - using the Fast median filtering algorithm
%
% Parameters:
% x - input image
% r - the radius of the median filter (r = 4 corresponds to a 9x9 window)
% bins - the number of histogram bins (the higher, the more memory usage and the higher
%   the computational cost)
% bin_size - the size of each bin (for bins=256 and a 8-bit image, the bin_size is 1, 
%     for bins=256 and a 16-bit image, the bin_size is 256)
%
% :function y = immedfilt(x : cube, r : int, bins : int=256, bin_size : scalar = 1)
function y = immedfilt(x : cube, r : int, bins : int=256, bin_size : scalar = 1)
    if r==1
        function [] = __kernel__ medfilt_opt3x3(y:cube, x:cube'mirror, pos:ivec3)
            %oriented to http://graphics.cs.williams.edu/papers/MedianShaderX6/
            t:vec9 = uninit(9)
            t[0] = x[pos[0]-1, pos[1]-1, pos[2]]
            t[1] = x[pos[0]  , pos[1]-1, pos[2]]
            t[2] = x[pos[0]+1, pos[1]-1, pos[2]]
            t[3] = x[pos[0]-1, pos[1]  , pos[2]]
            t[4] = x[pos[0]  , pos[1]  , pos[2]]
            t[5] = x[pos[0]+1, pos[1]  , pos[2]]
            t[6] = x[pos[0]-1, pos[1]+1, pos[2]]
            t[7] = x[pos[0]  , pos[1]+1, pos[2]]
            t[8] = x[pos[0]+1, pos[1]+1, pos[2]]
            
            locations0 = [0,1,2,0,0,4,3,1,3,1,1,4,2,2,4,2,3,4,3,3]
            locations1 = [3,4,5,1,2,5,5,2,4,3,6,6,6,3,7,4,7,8,8,4]
            for idx = 0..19
                i = locations0[idx]
                j = locations1[idx]
                tmp = t[i]
                t[i] = min(t[i],t[j])
                t[j] = max(tmp, t[j])
            endfor
            y[pos] = t[4]
        endfunction
        y = uninit(size(x))
        parallel_do(size(x), y, x, medfilt_opt3x3)
    else
    
        % add_hist_row : adds one single row to the histogram
        function [] = __device__ add_hist_row(x : cube'mirror, hist_im : cube{4}'unchecked, row : int, r : int, N : int, sgn : int, pos : ivec3, blkdim : int)
            Nhalf = int(N / 2)

            for p = pos[1]..blkdim..size(x,1)+r-1
                n = mod(p, N)
                p0 = p - n

                if n < Nhalf
                    for k = -Nhalf+n..-1
                        hist_im[int(x[row, p0 + k, pos[2]]),         pos[0], pos[2], p] += sgn
                        hist_im[int(x[row, p0 + 2*r+1 + k, pos[2]]), pos[0], pos[2], p] -= sgn
                    endfor
                elseif n == Nhalf
                    % this is the most labour-intensive loop
                    for k = 0..2*r
                        hist_im[int(x[row, p0 + k, pos[2]]),         pos[0], pos[2], p] += sgn
                    endfor
                else
                    for k = 0..n-Nhalf-1
                        hist_im[int(x[row, p0 + k, pos[2]]),         pos[0], pos[2], p] -= sgn
                        hist_im[int(x[row, p0 + 2*r+1 + k, pos[2]]), pos[0], pos[2], p] += sgn
                    endfor
                endif
            endfor
        endfunction

        % compute_median : computes the median, based on the local histograms
        function med : int = __device__ compute_median(hist_im : cube{4}'unchecked, p : int, N : int, idx_median : int, b : int, c : int)
            Nhalf = int(N/2)
            n = mod(p, N)
            p0 = p - n

            sum = 0
            i = 0
            while sum < idx_median && i < size(hist_im,0)
                if n == Nhalf
                    val = hist_im[i, b, c, p]
                else
                    val = hist_im[i, b, c, p] + hist_im[i, b, c, p0 + Nhalf]
                endif
                sum += int(val)
                i += 1
            endwhile
            med = i-1
        endfunction
       
        % ARGUMENTS:
        % x : input image
        % y : median filtered output image
        % hist_im : temporary memory for the histograms (one per column of the input image)
        % r : the radius of the median filter (r = 4 corresponds to a 9x9 window)
        % N : the number of columns processed simultaneously
        % bins : the number of histogram bins
        function [] = __kernel__ median_filter_kernel(x : cube, y : cube'unchecked, hist_im : cube{4}'unchecked, _
            r : int, N : int, K : int, blkdim : ivec3, pos : ivec3)

            % Vertical step size
            step_y = int((size(x,0) + K*blkdim[0] - 1) ./ (K*blkdim[0])) % dim[0]=K*blkdim[0]

            % Index of the median
            idx_median = int((2*r+1)^2/2)

            % Initialization - add the first r rows
            for m = -r..r-1
                add_hist_row(x, hist_im, pos[0]*step_y + m, r, N, 1, pos, blkdim[1])
            endfor

            memfence
            syncthreads

            for m = pos[0]*step_y..min((pos[0]+1)*step_y, size(x,0))-1
                % Running phase - add row (r), subtract row (r-1)
                add_hist_row(x, hist_im, m+r, r, N, 1, pos, blkdim[1])
                add_hist_row(x, hist_im, m-r-1, r, N, -1, pos, blkdim[1])
                memfence
                syncthreads

                % At this time, we have all local histograms at row m at our disposal, so we can
                % directly proceed by computing the median
                for p = pos[1]..blkdim[1]..size(x,1)-1
                    y[m, p, pos[2]] = compute_median(hist_im, p, N, idx_median, pos[0], pos[2])
                endfor
                
                memfence
                syncthreads
            endfor
        endfunction

        nextpow2 = x -> 2^ceil(log2(x))   
        N = floor(r/2)*2+1
        K = 64

        % Compute the block size of the filter
        blk_size = max_block_size(median_filter_kernel, [K, size(x,1), size(x,2)])
        if blk_size[0]>64
            blk_size[0]=64
        endif
        
        y = zeros(size(x))
        
        if bin_size != 1
            x = x * (1/bin_size)
        endif
        hist_im = zeros(bins, blk_size[0]*K, blk_size[2], size(x,1) + nextpow2(N))
        B_size = [K,1,1].*blk_size
        parallel_do([B_size,blk_size],x,y,hist_im,r,N,K,median_filter_kernel)
        if bin_size != 1
            y *= bin_size
        endif
    endif
endfunction

import "Quasar.UI.dll"
function [] = main()

    x = repmat(imread("lighthouse.png"),[1,1,1])
    x = x[0..383,0..250-1,:]
    %x = zeros(512,512,3)
    %x = imread("lena_big.tif")
    tic()
    for runs=1..100
        y = immedfilt(x, 18)
    endfor
    toc()
    imshow(y - x, [-128,128])
imshow(y)
%   frm = form("Window")
%   frm.width = 800
%   disp1 = frm.add_display()        
%    tic()
%    while !frm.closed()
%        y = immedfilt(x+10*randn(size(x)), 20)
%        disp1.imshow(y,[0,255])
%        pause(0)
%    endwhile
%    toc()
%    imshow(y)
endfunction