%==========================================================
% immorphology.q
% Quasar library for mathematical morphology
%
% (C) Gepura 2013-2018
%==========================================================

op_min = __device__ (x,y) -> min(x,y)
op_max = __device__ (x,y) -> max(x,y)

{!author name="Bart Goossens"}
{!doc category="Image Processing/Mathematical morphology"}

% Function: strel
%
% Implementation of some much-used structure elements for morphology.
% 
% :function [element:mat] = strel(name:string,radius:int)
%
% Parameters:
% name      -   name of the element (currently "disk" or "square")
% radius    -   the size of the resulting element will be [2*radius, 2*radius]
% 
function [element:mat] = strel(name:string,radius:int)
    disk_expr = __kernel__ (element:mat,pos:ivec2) -> element[pos] = sqrt(sum((pos-radius).^2)) <= radius
    sqr_expr  = __kernel__ (element:mat,pos:ivec2) -> element[pos] = 1
    
    element = uninit(2*radius+1,2*radius+1)
    match name with
    | "disk" -> parallel_do(size(element),element,disk_expr)
    | "square" -> parallel_do(size(element),element,sqr_expr)
    | _ -> error("This structure element is not currently implemented!")
    end
end

% Function: morph_filter
%
% Generic prototype for ANY morphological filter in this module
% 
% :function y = morph_filter[T](x : cube[T], mask : mat, ctr : ivec2, 
% :                         init : T, morph_op : [__device__ (T, T) -> T])
%
% Parameters:
% x    - the input image
% mask - the morphology mask (should be binary!)
% ctr  - the center of the mask
% morph_op - the operation (e.g. minimum or maximum)
% init - the value to be used when the mask is empty
%
% Remarks:
% - Gray scale morphology is performed. However, when the input is binary,
%   binary morphology is correctly obtained.
% - For color images, the filter processes each channel individually.
function y = morph_filter[T](x : cube[T], mask : mat, ctr : ivec2, init : T, morph_op : [__device__ (T, T) -> T])
    function [] = __kernel__ kernel (x : cube[T]'mirror, y, mask, ctr, init, morph_op, pos)
        res = init
        for m=0..size(mask,0)-1
            for n=0..size(mask,1)-1
                if mask[m,n] != 0
                    res = morph_op(res,x[pos+[m-ctr[0],n-ctr[1],0]])
                endif
            end
        end
        y[pos] = res
    end

    y = cube[T](size(x))
    parallel_do(size(x),x,y,mask,ctr,init,morph_op,kernel)
end


% Function: imerode
%
% Mathematical erosion computes for every pixel (i,j) the minimum 
% over all pixels in the window centered at (i,j), according to 
% the given mask.
%
% :y = imerode(x : cube, mask : mat, ctr : ivec2)
%
% Parameters:
% x - the input image
% mask - the mask to use
% ctr - the coordinates of the center of the mask (typically floor(size(mask)/2))
function y = imerode(x : cube, mask : mat, ctr : ivec2 = [-1,-1])
    if ctr[0] < 0 || ctr[1] < 0
        ctr = round((size(mask,0..1)-1)/2)
    endif
    y = morph_filter(x, mask, ctr, 1e37, op_min)

end

% Function: imdilate
%
% Mathematical dilation computes for every pixel (i,j) the maximum
% over all pixels in the window centered at (i,j), according to 
% the given mask.
%
% :y = imdilate(x : cube, mask : mat, ctr : ivec2)
%
% Parameters:
% x - the input image
% mask - the mask to use
% ctr - the coordinates of the center of the mask (typically floor(size(mask)/2))
function y = imdilate(x : cube, mask : mat, ctr : ivec2 = [-1,-1])
    if ctr[0] < 0 || ctr[1] < 0
        ctr = round((size(mask,0..1)-1)/2)
    endif
    y = morph_filter(x, mask, ctr, -1e37, op_max)

end

% Function: imopen
%
% The morphological opening on an image is defined as an erosion followed by
% a dilation. Opening can remove small bright intensities and connect
% small dark regions. 
%
% :function y = imopen(x : cube, mask : mat, ctr : ivec2)
%
% Parameters:
% x - the input image
% mask - the mask to use
% ctr - the coordinates of the center of the mask (typically floor(size(mask)/2))
function y = imopen(x : cube, mask : mat, ctr : ivec2 = [-1,-1])
    y = imdilate(imerode(x, mask, ctr), mask, ctr)

end

% Function: imclose
% The morphological closing on an image is defined as an dilation followed by
% a erosion. Opening can remove small dark intensities and connect
% small bright regions. 
%
% :function y = imclose(x : cube, mask : mat, ctr : ivec2)
%
% Parameters:
% x - the input image
% mask - the mask to use
% ctr - the coordinates of the center of the mask (typically floor(size(mask)/2))
function y = imclose(x : cube, mask : mat, ctr : ivec2 = [-1,-1])
    y = imerode(imdilate(x, mask, ctr), mask, ctr)

end


function [] = main()
    im=imread("lena_big.tif")

    tic()
    y = copy(im)
    for k=0..4
        y = imclose(imopen(y,ones(3,3),[1,1]),ones(3,3),[1,1])
    end
    toc()
    %y = morph_filter(im, ones(5,5),[2,2], 1e37, )

    imshow(y,[])

end