%==============================================================================
% imresize.q
% 2D Image resizing functions
% Ported from the C++ version of Hiep Luong.
% Bart Goossens, 2012
%==============================================================================
import "system.q"
import "imfilter.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Interpolation"}


% Function: imresize
%
% Performs image resizing using the specified interpolation kernel
%
% :function y = imresize(x, sz, interpolation = "sinc", boundary_mode = "mirror", points = 4)
%
% Parameters:
% x  - the input image
% sz - the size of the output image (or magnification factor)
% interpolation - the interpolation method being used: "nearest", "linear", "cubic", "sinc", "htu_nearest", "htu_linear"
% boundary_mode - the boundary extension mode: "zero", "mirror", "circular"
% points - half the number of points used for sinc resampling.
%
% Notes:
% interpolation modes "htu_nearest" and "htu_linear" enable the hardware texturing units of the GPU,
% yielding a dramatic speedup (3x-4x) compared to "nearest" and "linear". Because these methods have
% reduced precision, both hardware and non-hardware textured versions can be used.
function y:[mat|cube] = imresize(x:[mat|cube], sz, interpolation = "sinc", boundary_mode = "mirror", points : int = 4)

    % only a magnification factor was specified
    if isscalar(sz)
        sz = [round(size(x,0)*sz),round(size(x,1)*sz),size(x,2)]
    endif

    % no interpolation required
    if all(sz[0..1] == size(x,0..1))
        y = x
        return
    endif

    y = zeros(sz[0],sz[1],size(x,2))
    scalef : vec = size(x, 0..2) ./ [sz[0],sz[1],size(x,2)]

    match interpolation with
    %
    % Nearest neighbor interpolation
    %
    | "nearest" ->
        if all(scalef == round(scalef)) % Rescale by integer factor

            parallel_do(size(y),x,y,scalef, __kernel__ _
                (x : cube'unchecked, y : cube'unchecked, scalef : ivec3, pos : ivec3) -> y[pos] = x[pos .* scalef])

        else

            parallel_do(size(y),x,y,scalef, __kernel__ _
                (x : cube'unchecked, y : cube'unchecked, scalef : vec3, pos : ivec3) -> y[pos] = x[int(pos .* scalef)])

        endif

    %
    % Nearest neighbor interpolation (GPU hardware texturing unit)
    %
    | "htu_nearest" ->
        parallel_do(size(y),x,y,scalef, __kernel__ _
            (x : cube'hwtex_nearest, y : cube'unchecked, scalef : vec3, pos : ivec3) -> y[pos] = x[pos .* scalef])

    %
    % (Bi)linear interpolation (GPU hardware texturing unit)
    %
    | "htu_linear" ->
        parallel_do(size(y),x,y,scalef, __kernel__ _
            (x : cube'hwtex_linear, y : cube'unchecked, scalef : vec3, pos : ivec3) -> y[pos] = x[pos .* scalef])

    %
    % (Bi)linear interpolation
    %
    | ("linear", "bilinear") ->

        % linear interpolation function
        function y = __device__ interpolate_bilinear_zero_ext(x : cube, pos : vec3)
            f = frac(pos)
            i = int(pos)
            y = x[i] * (1.0 - f[0]) * (1.0 - f[1]) + _
                x[i+[0,1,0]] * (1.0 - f[0]) * f[1] + _
                x[i+[1,0,0]] * f[0] * (1.0 - f[1]) + _
                x[i+[1,1,0]] * f[0] * f[1]
        endfunction

        function y = __device__ interpolate_bilinear_mirror_ext(x : cube'mirror, pos : vec3)
            f = frac(pos)
            i = int(pos)
            y = x[i] * (1.0 - f[0]) * (1.0 - f[1]) + _
                x[i+[0,1,0]] * (1.0 - f[0]) * f[1] + _
                x[i+[1,0,0]] * f[0] * (1.0 - f[1]) + _
                x[i+[1,1,0]] * f[0] * f[1]
        endfunction

        function y = __device__ interpolate_bilinear_circular_ext(x : cube'circular, pos : vec3)
            f = frac(pos)
            i = int(pos)
            y = x[i] * (1.0 - f[0]) * (1.0 - f[1]) + _
                x[i+[0,1,0]] * (1.0 - f[0]) * f[1] + _
                x[i+[1,0,0]] * f[0] * (1.0 - f[1]) + _
                x[i+[1,1,0]] * f[0] * f[1]
        endfunction

        match boundary_mode with
        | "zero"        -> parallel_do(size(y),x,y,scalef, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, scalef : vec3, pos : ivec3) -> _
                            y[pos] = interpolate_bilinear_zero_ext(x, pos .* scalef))
        | "mirror"      -> parallel_do(size(y),x,y,scalef, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, scalef : vec3, pos : ivec3) -> _
                            y[pos] = interpolate_bilinear_mirror_ext(x, pos .* scalef))
        | "circular"    -> parallel_do(size(y),x,y,scalef, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, scalef : vec3, pos : ivec3) -> _
                            y[pos] = interpolate_bilinear_circular_ext(x, pos .* scalef))
        | _ -> error ("Unknown boundary extension mode!")
        endmatch

    %
    % (Bi)cubic interpolation
    %
    | ("cubic", "bicubic") ->

        pos_y = scalef[0]*0.5 + scalef[0] * (0..sz[0]-1)
        pos_x = scalef[1]*0.5 + scalef[1] * (0..sz[1]-1)
        img_temp = zeros(size(x,0),sz[1],size(x,2))

        % Helper functions for computing weights
        compute_weight = __device__ (delta : scalar) -> _
            (1.0/6.0)*(max(0.0,1.0-delta)^3 - 4.0*max(0.0,-delta)^3 + 6.0*max(0.0,-1.0-delta)^3 - 4.0*max(0.0,-2.0-delta)^3)
        compute_weight_vec = __device__ (delta : scalar) -> _
            [compute_weight(delta), _
             compute_weight(delta - 1.0), _
             compute_weight(delta - 2.0), _
             compute_weight(delta - 3.0)]

        % Cubic interpolation functions
        function y : scalar = __device__ interpolate_cubic_hor_zero_ext(x : cube, i : ivec3, pos : scalar)
            i = [i[0], int(pos), i[2]]
            w = compute_weight_vec(frac(pos))
            y = dotprod(w, [x[i-[0,1,0]], x[i], x[i+[0,1,0]], x[i+2*[0,2,0]]])
        endfunction

        function y : scalar = __device__ interpolate_cubic_ver_zero_ext(x : cube, i : ivec3, pos : scalar)
            i = [int(pos), i[1], i[2]]
            w = compute_weight_vec(frac(pos))
            y = dotprod(w, [x[i-[1,0,0]], x[i], x[i+[1,0,0]], x[i+[2,0,0]]])
        endfunction

        function y : scalar = __device__ interpolate_cubic_hor_mirror_ext(x : cube'unchecked, i : ivec3, pos : scalar)
            i = [i[0], int(pos), i[2]]
            w = compute_weight_vec(frac(pos))
            y = dotprod(w, [x[i[0],mirror_ext(i[1]-1,size(x,1)), i[2]], _
                            x[i], _
                            x[i[0],mirror_ext(i[1]+1,size(x,1)), i[2]], _
                            x[i[0],mirror_ext(i[1]+2,size(x,1)), i[2]]])
        endfunction

        function y : scalar = __device__ interpolate_cubic_ver_mirror_ext(x : cube'unchecked, i : ivec3, pos : scalar)
            i = [int(pos), i[1], i[2]]
            w = compute_weight_vec(frac(pos))
            y = dotprod(w, [x[mirror_ext(i[0]-1,size(x,0)), i[1], i[2]], _
                            x[i], _
                            x[mirror_ext(i[0]+1,size(x,0)), i[1], i[2]], _
                            x[mirror_ext(i[0]+2,size(x,0)), i[1], i[2]]])
        endfunction

        function y : scalar = __device__ interpolate_cubic_hor_circular_ext(x : cube'unchecked, i : ivec3, pos : scalar)
            i = [i[0], int(pos), i[2]]
            w = compute_weight_vec(frac(pos))
            y = dotprod(w, [x[i[0],periodize(i[1]-1,size(x,1)), i[2]], _
                            x[i], _
                            x[i[0],periodize(i[1]+1,size(x,1)), i[2]], _
                            x[i[0],periodize(i[1]+2,size(x,1)), i[2]]])
        endfunction

        function y : scalar = __device__ interpolate_cubic_ver_circular_ext(x : cube'unchecked, i : ivec3, pos : scalar)
            i = [int(pos), i[1], i[2]]
            w = compute_weight_vec(frac(pos))
            y = dotprod(w, [x[periodize(i[0]-1,size(x,0)), i[1], i[2]], _
                            x[i], _
                            x[periodize(i[0]+1,size(x,0)), i[1], i[2]], _
                            x[periodize(i[0]+2,size(x,0)), i[1], i[2]]])
        endfunction

        match boundary_mode with
        | "zero"        -> parallel_do(size(img_temp),x,img_temp,pos_x, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_x : vec'unchecked, pos : ivec3) -> _
                            y[pos] = interpolate_cubic_hor_zero_ext(x, pos, pos_x[pos[1]]))
                           parallel_do(size(y),img_temp,y,pos_y, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_y : vec'unchecked, pos : ivec3) -> _
                            y[pos] = interpolate_cubic_ver_zero_ext(x, pos, pos_y[pos[0]]))
        | "mirror"      -> parallel_do(size(img_temp),x,img_temp,pos_x, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_x : vec'unchecked, pos : ivec3) -> _
                            y[pos] = interpolate_cubic_hor_mirror_ext(x, pos, pos_x[pos[1]]))
                           parallel_do(size(y),img_temp,y,pos_y, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_y : vec'unchecked, pos : ivec3) -> _
                            y[pos] = interpolate_cubic_ver_mirror_ext(x, pos, pos_y[pos[0]]))
        | "circular"    -> parallel_do(size(img_temp),x,img_temp,pos_x, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_x : vec'unchecked, pos : ivec3) -> _
                            y[pos] = interpolate_cubic_hor_circular_ext(x, pos, pos_x[pos[1]]))
                           parallel_do(size(y),img_temp,y,pos_y, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_y : vec'unchecked, pos : ivec3) -> _
                            y[pos] = interpolate_cubic_ver_circular_ext(x, pos, pos_y[pos[0]]))
        | _ -> error ("Unknown boundary extension mode!")
        endmatch

    %
    % Sinc interpolation
    %
    | "sinc" ->

        pos_y = scalef[0]*0.5 + scalef[0] * (0..sz[0]-1)
        pos_x = scalef[1]*0.5 + scalef[1] * (0..sz[1]-1)
        img_temp = zeros(size(x,0),sz[1],size(x,2))

        if scalef[1] > 1.0 % downsampling in the horizontal direction - apply ideal lowpass filter first
            pulse = sinc((-points..points) / scalef[1])
            x = imfilter(x, pulse/sum(pulse), [0, points+1], boundary_mode)
        endif
        if scalef[0] > 1.0 % downsampling in the vertical direction - apply ideal lowpass filter first
            pulse = sinc((-points..points) / scalef[0])
            x = imfilter(x, transpose(pulse/sum(pulse)), [points+1, 0], boundary_mode)
        endif

        % sinc helper function
        sinc_fn = __device__ (x : scalar) -> (x == 0.0) ? 1.0 : sin(pi*x)/(pi*x)

        % windowing functions (t = -pi..pi)
        window_blackmanharris = __device__ (t : scalar) -> 0.35875 + 0.48829*cos(t) + 0.14128*cos(2*t) + 0.01168*cos(3*t)
        window_lanzcos = __device__ (t : scalar) -> sinc_fn(t)

        % sinc resampling with ZERO boundary extension
        function y = __device__ interpolate_sinc_hor_zero_ext(x : cube, i : ivec3, pos : scalar, npoints : int)
            i = [i[0], int(pos), i[2]]
            f = frac(pos)
            y = 0.0
            for n=-npoints+1..npoints
                w = window_blackmanharris((float(n) - f) * pi / npoints) * sinc_fn(float(n) - f)
                y += w * x[i + [0,n,0]]
            endfor
        endfunction

        function y = __device__ interpolate_sinc_ver_zero_ext(x : cube, i : ivec3, pos : scalar, npoints : int)
            i = [int(pos), i[1], i[2]]
            f = frac(pos)
            y = 0.0
            for n=-npoints+1..npoints
                w = window_blackmanharris((float(n) - f) * pi / npoints) * sinc_fn(float(n) - f)
                y += w * x[i + [n,0,0]]
            endfor
        endfunction

        % sinc resampling with MIRROR boundary extension
        function y = __device__ interpolate_sinc_hor_mirror_ext(x : cube'unchecked, i : ivec3, pos : scalar, npoints : int)
            i = [i[0], int(pos), i[2]]
            f = frac(pos)
            y = 0.0
            for n=-npoints+1..npoints
                w = window_blackmanharris((float(n) - f) * pi / npoints) * sinc_fn(float(n) - f)
                y += w * x[i[0], mirror_ext(i[1] + n, size(x,1)), i[2]]
            endfor
        endfunction

        function y = __device__ interpolate_sinc_ver_mirror_ext(x : cube'unchecked, i : ivec3, pos : scalar, npoints : int)
            i = [int(pos), i[1], i[2]]
            f = frac(pos)
            y = 0.0
            for n=-npoints+1..npoints
                w = window_blackmanharris((float(n) - f) * pi / npoints) * sinc_fn(float(n) - f)
                y += w * x[mirror_ext(i[0] + n, size(x,0)), i[1], i[2]]
            endfor
        endfunction

        % sinc resampling with CIRCULAR boundary extension
        function y = __device__ interpolate_sinc_hor_circular_ext(x : cube'unchecked, i : ivec3, pos : scalar, npoints : int)
            i = [i[0], int(pos), i[2]]
            f = frac(pos)
            y = 0.0
            for n=-npoints+1..npoints
                w = window_blackmanharris((float(n) - f) * pi / npoints) * sinc_fn(float(n) - f)
                y += w * x[i[0], periodize(i[1] + n, size(x,1)), i[2]]
            endfor
        endfunction

        function y = __device__ interpolate_sinc_ver_circular_ext(x : cube'unchecked, i : ivec3, pos : scalar, npoints : int)
            i = [int(pos), i[1], i[2]]
            f = frac(pos)
            y = 0.0
            for n=-npoints+1..npoints
                w = window_blackmanharris((float(n) - f) * pi / npoints) * sinc_fn(float(n) - f)
                y += w * x[periodize(i[0] + n, size(x,0)), i[1], i[2]]
            endfor
        endfunction

        match boundary_mode with
        | "zero"        -> parallel_do(size(img_temp),x,img_temp,pos_x,points, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_x : vec'unchecked, pos : ivec3, npoints : int) -> _
                            y[pos] = interpolate_sinc_hor_zero_ext(x, pos, pos_x[pos[1]], npoints))
                           parallel_do(size(y),img_temp,y,pos_y,points, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_y : vec'unchecked, pos : ivec3, npoints : int) -> _
                            y[pos] = interpolate_sinc_ver_zero_ext(x, pos, pos_y[pos[0]], npoints))
        | "mirror"      -> parallel_do(size(img_temp),x,img_temp,pos_x,points, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_x : vec'unchecked, pos : ivec3, npoints : int) -> _
                            y[pos] = interpolate_sinc_hor_mirror_ext(x, pos, pos_x[pos[1]], npoints))
                           parallel_do(size(y),img_temp,y,pos_y,points, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_y : vec'unchecked, pos : ivec3, npoints : int) -> _
                            y[pos] = interpolate_sinc_ver_mirror_ext(x, pos, pos_y[pos[0]], npoints))
        | "circular"    -> parallel_do(size(img_temp),x,img_temp,pos_x,points, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_x : vec'unchecked, pos : ivec3, npoints : int) -> _
                            y[pos] = interpolate_sinc_hor_circular_ext(x, pos, pos_x[pos[1]], npoints))
                           parallel_do(size(y),img_temp,y,pos_y,points, __kernel__ _
                            (x : cube'unchecked, y : cube'unchecked, pos_y : vec'unchecked, pos : ivec3, npoints : int) -> _
                            y[pos] = interpolate_sinc_ver_circular_ext(x, pos, pos_y[pos[0]], npoints))
        | _ -> error ("Unknown boundary extension mode!")
        endmatch

    | _ -> error("(imresize) Unknown interpolation method!")
    endmatch

endfunction

function [] = main()

    im = imread("lena_big.tif")

    tic()
    for k=0..1000
    y = imresize(im,[1024,1024],"linear")
    endfor
    toc("linear")


    tic()
    for k=0..1000
    y = imresize(im,[1024,1024],"htu_linear")
    endfor
    toc("HTU linear")
    imshow(y,[])

endfunction
