%==============================================================================
% inttypes.q
% Quasar library for integer datatype support
% (C) Gepura, Oct. 2012-2018
%==============================================================================
% This module defines binary operations and conversion functions for 
% integer matrices: 
%      (vec..mat..cube)[int]
%      (vec..mat..cube)[int8]
%      (vec..mat..cube)[int16]
%      (vec..mat..cube)[int32]
%      (vec..mat..cube)[int64]
%      (vec..mat..cube)[uint8]
%      (vec..mat..cube)[uint16]
%      (vec..mat..cube)[uint32]
%      (vec..mat..cube)[uint64]
% Note that no cross-operations are allowed (for example adding a matrix
% of type mat[int64] to a matrix of type mat[int8]). If this functionality
% is required, it is still possible this implementation using reductions,
% in a manner similar as performed in this module.
%==============================================================================
% NOTES: - I use a trick using device functions, to reduce the size of the 
%          resulting binary.
%        - Arithmetic using uint32, int64 and uint64 is (currently) depends
%          on the scalar data type. In 32-bit floating point (FP) mode, the
%          values are correct in the [-2^24..2^24-1] range, while in 64-bit FP 
%          mode the values are correct when between [-2^53,2^53-1]. This is
%          because Quasar currently only supports 2 "basic" primitive types
%          for performing arithmetic operations: 'int' and 'scalar'.
%        - for the types uint32, int64 and uint64, the binary bitwise functions
%          and, or, xor, shl, shr are not implemented.
% OTHER ISSUES: - the transpose function is currently not accelerated...
%==============================================================================

% Definition of the integer vector, matrix and cube[int] types
type ivec : vec[int]
type imat : mat[int]
type icube : cube[int]

% Special purpose SIMD types
type i8vec4 : vec[int8](4)
type u8vec4 : vec[uint8](4)
type i16vec4 : vec[int16](4)
type u16vec4 : vec[uint16](4)

%==============================================================================
% Operator definitions
%==============================================================================

type int_bin_op : [__device__ (int, int) -> int]

int_minus       = __device__ (x : int, y : int) -> -x
int_add         = __device__ (x : int, y : int) -> x + y
int_subtract    = __device__ (x : int, y : int) -> x - y
int_mul         = __device__ (x : int, y : int) -> x * y
% note: conversion int(.) needed, otherwise Quasar thinks the result is scalar
int_divide      = __device__ (x : int, y : int) -> int(x / y) 
int_pow         = __device__ (x : int, y : int) -> x ^ y
int_less        = __device__ (x : int, y : int) -> x < y
int_leq         = __device__ (x : int, y : int) -> x <= y
int_greater     = __device__ (x : int, y : int) -> x > y
int_geq         = __device__ (x : int, y : int) -> x >= y
int_eq          = __device__ (x : int, y : int) -> x == y
int_neq         = __device__ (x : int, y : int) -> x != y
int_negate      = __device__ (x : int, y : int) -> !x
int_and         = __device__ (x : int, y : int) -> and(x,y)
int_or          = __device__ (x : int, y : int) -> or(x,y)
int_xor         = __device__ (x : int, y : int) -> xor(x,y)
int_shl         = __device__ (x : int, y : int) -> shl(x,y) % left bitshift x << y
int_shr         = __device__ (x : int, y : int) -> shr(x,y) % right bitshift x >> y
int_min         = __device__ (x : int, y : int) -> min(x,y)
int_max         = __device__ (x : int, y : int) -> max(x,y)

type scalar_bin_op : [__device__ (scalar, scalar) -> scalar]

scalar_minus       = __device__ (x : scalar, y : scalar) -> -x
scalar_add         = __device__ (x : scalar, y : scalar) -> x + y
scalar_subtract    = __device__ (x : scalar, y : scalar) -> x - y
scalar_mul         = __device__ (x : scalar, y : scalar) -> x * y
scalar_divide      = __device__ (x : scalar, y : scalar) -> x / y
scalar_pow         = __device__ (x : scalar, y : scalar) -> x ^ y
scalar_less        = __device__ (x : scalar, y : scalar) -> x < y
scalar_leq         = __device__ (x : scalar, y : scalar) -> x <= y
scalar_greater     = __device__ (x : scalar, y : scalar) -> x > y
scalar_geq         = __device__ (x : scalar, y : scalar) -> x >= y
scalar_eq          = __device__ (x : scalar, y : scalar) -> x == y
scalar_neq         = __device__ (x : scalar, y : scalar) -> x != y
scalar_negate      = __device__ (x : scalar, y : scalar) -> !x
scalar_min         = __device__ (x : scalar, y : scalar) -> min(x,y)
scalar_max         = __device__ (x : scalar, y : scalar) -> max(x,y)


%==============================================================================
% Integer type implementation
%==============================================================================

% A binary operation on integers
function z : cube[int]'unchecked = int_binaryop(x : cube[int]'unchecked, y : cube[int]'unchecked, op : int_bin_op)
    z = cube[int](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[int]'unchecked = int_binaryop(x : cube[int]'unchecked, y : scalar, op : int_bin_op)
    z = cube[int](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[int]'unchecked = int_binaryop(x : scalar, y : cube[int]'unchecked, op : int_bin_op)
    z = cube[int](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end


% Negation
reduction (a : cube[int]) -> -a = int_binaryop(a, a, int_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[int], b : cube[int]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[int], b : cube[int]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[int], b : cube[int]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[int], b : cube[int]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a / b = int_binaryop(a, b, int_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[int], b : cube[int])    -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[int])       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
% Less
reduction (a : cube[int], b : cube[int]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
% Greater
reduction (a : cube[int], b : cube[int]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[int], b : cube[int]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[int], b : cube[int]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
% Equal
reduction (a : cube[int], b : cube[int]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[int], b : cube[int]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[int]) -> !a = int_binaryop(a, a, int_negate) where $ftype("__host__")

% And
reduction (a : cube[int], b : cube[int]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
% Or
reduction (a : cube[int], b : cube[int]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
% Xor
reduction (a : cube[int], b : cube[int]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
% Shl
reduction (a : cube[int], b : cube[int]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
% Shr
reduction (a : cube[int], b : cube[int]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
% Min
reduction (a : cube[int], b : cube[int]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
% Max
reduction (a : cube[int], b : cube[int]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int], b : scalar)       -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")

%==============================================================================
% Int-to-float conversion
%==============================================================================
function y = int2float(x : cube[int]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y = float2int(x : cube'unchecked) 
    y = cube[int](size(x))  
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = int(x[m,n,p])
            end
        end
    end
end

reduction (x : cube[int])   -> float(x) = int2float(x) where $ftype("__host__")
reduction (x : cube)        -> int(x) = float2int(x) where $ftype("__host__")
reduction (x : cube[int])   -> int(x) = x % already int!

%==============================================================================
%
% The following code has been copy-pasted a lot - we could improve this by using
% "generic" parameter types. But for the moment, it will also work this way...
%

%==============================================================================
% cube[int8] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[int8]'unchecked = int_binaryop(x : cube[int8]'unchecked, y : cube[int8]'unchecked, op : int_bin_op)
    z = cube[int8](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[int8]'unchecked = int_binaryop(x : cube[int8]'unchecked, y : scalar, op : int_bin_op)
    z = cube[int8](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[int8]'unchecked = int_binaryop(x : scalar, y : cube[int8]'unchecked, op : int_bin_op)
    z = cube[int8](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[int8]) -> -a = int_binaryop(a, a, int_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[int8], b : cube[int8]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)        -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : scalar,        b : cube[int8]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[int8], b : cube[int8]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)        -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : scalar,        b : cube[int8]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[int8], b : cube[int8]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[int8], b : cube[int8]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a / b = int_binaryop(a, b, int_divide) where $ftype("__host__")

% Point-wise power of matrices
reduction (a : cube[int8], b : cube[int8]) -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[int8])       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
% Less
reduction (a : cube[int8], b : cube[int8]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
% Greater
reduction (a : cube[int8], b : cube[int8]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[int8], b : cube[int8]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[int8], b : cube[int8]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
% Equal
reduction (a : cube[int8], b : cube[int8]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[int8], b : cube[int8]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[int8]) -> !a = int_binaryop(a, a, int_negate) where $ftype("__host__")

% And
reduction (a : cube[int8], b : cube[int8]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
% Or
reduction (a : cube[int8], b : cube[int8]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
% Xor
reduction (a : cube[int8], b : cube[int8]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
% Shl
reduction (a : cube[int8], b : cube[int8]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
% Shr
reduction (a : cube[int8], b : cube[int8]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
% Min
reduction (a : cube[int8], b : cube[int8]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
% Max
reduction (a : cube[int8], b : cube[int8]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int8], b : scalar)       -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int8]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")


function y : cube'unchecked = int8_to_float(x : cube[int8]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y : cube[int8]'unchecked = float_to_int8(x : cube'unchecked) 
    y = cube[int8](size(x))  
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end
        end
    end
end

reduction (x : cube[int8])   -> float(x) = int8_to_float(x) where $ftype("__host__")
reduction (x : cube) -> int8(x) = float_to_int8(x) where $ftype("__host__")

%==============================================================================
% cube[uint8] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[uint8]'unchecked = int_binaryop(x : cube[uint8]'unchecked, y : cube[uint8]'unchecked, op : int_bin_op)
    z = cube[uint8](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[uint8]'unchecked = int_binaryop(x : cube[uint8]'unchecked, y : scalar, op : int_bin_op)
    z = cube[uint8](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[uint8]'unchecked = int_binaryop(x : scalar, y : cube[uint8]'unchecked, op : int_bin_op)
    z = cube[uint8](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[uint8]) -> -a = int_binaryop(a, a, int_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[uint8], b : cube[uint8]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[uint8], b : cube[uint8]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[uint8], b : cube[uint8]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[uint8], b : cube[uint8]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a / b = int_binaryop(a, b, int_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[uint8], b : cube[uint8]) -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[uint8])       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
% Less
reduction (a : cube[uint8], b : cube[uint8]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
% Greater
reduction (a : cube[uint8], b : cube[uint8]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[uint8], b : cube[uint8]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[uint8], b : cube[uint8]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
% Equal
reduction (a : cube[uint8], b : cube[uint8]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[uint8], b : cube[uint8]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[uint8]) -> !a = int_binaryop(a, a, int_negate) where $ftype("__host__")

% And
reduction (a : cube[uint8], b : cube[uint8]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
% Or
reduction (a : cube[uint8], b : cube[uint8]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
% Xor
reduction (a : cube[uint8], b : cube[uint8]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
% Shl
reduction (a : cube[uint8], b : cube[uint8]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
% Shr
reduction (a : cube[uint8], b : cube[uint8]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
% Min
reduction (a : cube[uint8], b : cube[uint8]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
% Max
reduction (a : cube[uint8], b : cube[uint8]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[uint8], b : scalar)       -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint8]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")


function y : cube'unchecked = uint8_to_float(x : cube[uint8]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y : cube[uint8]'unchecked = float_to_uint8(x : cube'unchecked) 
    y = cube[uint8](size(x))  
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end
        end
    end
end

reduction (x : cube[uint8])   -> float(x) = uint8_to_float(x) where $ftype("__host__")
reduction (x : cube)          -> uint8(x) = float_to_uint8(x) where $ftype("__host__")

%==============================================================================
% cube[int16] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[int16]'unchecked = int_binaryop(x : cube[int16]'unchecked, y : cube[int16]'unchecked, op : int_bin_op)
    z = cube[int16](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[int16]'unchecked = int_binaryop(x : cube[int16]'unchecked, y : scalar, op : int_bin_op)
    z = cube[int16](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[int16]'unchecked = int_binaryop(x : scalar, y : cube[int16]'unchecked, op : int_bin_op)
    z = cube[int16](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[int16]) -> -a = int_binaryop(a, a, int_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[int16], b : cube[int16]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[int16], b : cube[int16]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[int16], b : cube[int16]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[int16], b : cube[int16]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a / b = int_binaryop(a, b, int_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[int16], b : cube[int16]) -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[int16])       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
% Less
reduction (a : cube[int16], b : cube[int16]) -> a < b = int_binaryop(a, b, int_less)  where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
% Greater
reduction (a : cube[int16], b : cube[int16]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[int16], b : cube[int16]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[int16], b : cube[int16]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
% Equal
reduction (a : cube[int16], b : cube[int16]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[int16], b : cube[int16]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[int16]) -> !a = int_binaryop(a, a, int_negate) where $ftype("__host__")

% And
reduction (a : cube[int16], b : cube[int16]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
% Or
reduction (a : cube[int16], b : cube[int16]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
% Xor
reduction (a : cube[int16], b : cube[int16]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
% Shl
reduction (a : cube[int16], b : cube[int16]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
% Shr
reduction (a : cube[int16], b : cube[int16]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
% Min
reduction (a : cube[int16], b : cube[int16]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
% Max
reduction (a : cube[int16], b : cube[int16]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int16], b : scalar)       -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int16]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")


function y : cube'unchecked = int16_to_float(x : cube[int16]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y : cube[int16]'unchecked = float_to_int16(x : cube'unchecked) 
    y = cube[int16](size(x)) 
    {!parallel for} 
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end
        end
    end
end

reduction (x : cube[int16])   -> float(x) = int16_to_float(x) where $ftype("__host__")
reduction (x : cube)          -> int16(x) = float_to_int16(x) where $ftype("__host__")

%==============================================================================
% cube[uint16] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[uint16]'unchecked = int_binaryop(x : cube[uint16]'unchecked, y : cube[uint16]'unchecked, op : int_bin_op)
    z = cube[uint16](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[uint16]'unchecked = int_binaryop(x : cube[uint16]'unchecked, y : scalar, op : int_bin_op)
    z = cube[uint16](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[uint16]'unchecked = int_binaryop(x : scalar, y : cube[uint16]'unchecked, op : int_bin_op)
    z = cube[uint16](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[uint16]) -> -a = int_binaryop(a, a, int_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[uint16], b : cube[uint16]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[uint16], b : cube[uint16]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[uint16], b : cube[uint16]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[uint16], b : cube[uint16]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a / b = int_binaryop(a, b, int_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[uint16], b : cube[uint16]) -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[uint16])       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
% Less
reduction (a : cube[uint16], b : cube[uint16]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
% Greater
reduction (a : cube[uint16], b : cube[uint16]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[uint16], b : cube[uint16]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[uint16], b : cube[uint16]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
% Equal
reduction (a : cube[uint16], b : cube[uint16]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[uint16], b : cube[uint16]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[uint16]) -> !a = int_binaryop(a, a, int_negate) where $ftype("__host__")

% And
reduction (a : cube[uint16], b : cube[uint16]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
% Or
reduction (a : cube[uint16], b : cube[uint16]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
% Xor
reduction (a : cube[uint16], b : cube[uint16]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
% Shl
reduction (a : cube[uint16], b : cube[uint16]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
% Shr
reduction (a : cube[uint16], b : cube[uint16]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
% Min
reduction (a : cube[uint16], b : cube[uint16]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
% Max
reduction (a : cube[uint16], b : cube[uint16]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[uint16], b : scalar)       -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint16]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")


function y : cube'unchecked = uint16_to_float(x : cube[uint16]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y : cube[uint16]'unchecked = float_to_uint16(x : cube'unchecked) 
    y = cube[uint16](size(x)) 
    {!parallel for} 
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end
        end
    end
end

reduction (x : cube[uint16])   -> float(x) = uint16_to_float(x) where $ftype("__host__")
reduction (x : cube)           -> uint16(x) = float_to_uint16(x) where $ftype("__host__")

%==============================================================================
% cube[int32] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[int32]'unchecked = int_binaryop(x : cube[int32]'unchecked, y : cube[int32]'unchecked, op : int_bin_op)
    z = cube[int32](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[int32]'unchecked = int_binaryop(x : cube[int32]'unchecked, y : scalar, op : int_bin_op) 
    z = cube[int32](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[int32]'unchecked = int_binaryop(x : scalar, y : cube[int32]'unchecked, op : int_bin_op)
    z = cube[int32](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[int32]) -> -a = int_binaryop(a, a, int_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[int32], b : cube[int32]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a + b = int_binaryop(a, b, int_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[int32], b : cube[int32]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a - b = int_binaryop(a, b, int_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[int32], b : cube[int32]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a .* b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a * b = int_binaryop(a, b, int_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[int32], b : cube[int32]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a ./ b = int_binaryop(a, b, int_divide) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a / b = int_binaryop(a, b, int_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[int32], b : cube[int32]) -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[int32])       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a .^ b = int_binaryop(a, b, int_pow) where $ftype("__host__")
% Less
reduction (a : cube[int32], b : cube[int32]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a < b = int_binaryop(a, b, int_less) where $ftype("__host__")
% Greater
reduction (a : cube[int32], b : cube[int32]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a > b = int_binaryop(a, b, int_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[int32], b : cube[int32]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a <= b = int_binaryop(a, b, int_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[int32], b : cube[int32]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a >= b = int_binaryop(a, b, int_geq) where $ftype("__host__")
% Equal
reduction (a : cube[int32], b : cube[int32]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a == b = int_binaryop(a, b, int_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[int32], b : cube[int32]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> a != b = int_binaryop(a, b, int_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[int32]) -> !a = int_binaryop(a, a, int_negate) where $ftype("__host__")

% And
reduction (a : cube[int32], b : cube[int32]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> and(a,b) = int_binaryop(a, b, int_and) where $ftype("__host__")
% Or
reduction (a : cube[int32], b : cube[int32]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> or(a,b) = int_binaryop(a, b, int_or) where $ftype("__host__")
% Xor
reduction (a : cube[int32], b : cube[int32]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> xor(a,b) = int_binaryop(a, b, int_xor) where $ftype("__host__")
% Shl
reduction (a : cube[int32], b : cube[int32]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> shl(a,b) = int_binaryop(a, b, int_shl) where $ftype("__host__")
% Shr
reduction (a : cube[int32], b : cube[int32]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> shr(a,b) = int_binaryop(a, b, int_shr) where $ftype("__host__")
% Min
reduction (a : cube[int32], b : cube[int32]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> min(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
% Max
reduction (a : cube[int32], b : cube[int32]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : cube[int32], b : scalar)       -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int32]) -> max(a,b) = int_binaryop(a, b, int_min) where $ftype("__host__")


function y : cube'unchecked = int32_to_float(x : cube[int32]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y : cube[int32]'unchecked = float_to_int32(x : cube'unchecked) 
    y = cube[int32](size(x))
    {!parallel for}  
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end
        end
    end
end

reduction (x : cube[int32])   -> float(x) = int32_to_float(x) where $ftype("__host__")
reduction (x : cube)          -> int32(x) = float_to_int32(x) where $ftype("__host__")

%==============================================================================
% cube[uint32] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[uint32]'unchecked = scalar_binaryop(x : cube[uint32]'unchecked, y : cube[uint32]'unchecked, op : scalar_bin_op)
    z = cube[uint32](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[uint32]'unchecked = scalar_binaryop(x : cube[uint32]'unchecked, y : scalar, op : scalar_bin_op)
    z = cube[uint32](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[uint32]'unchecked = scalar_binaryop(x : scalar, y : cube[uint32]'unchecked, op : scalar_bin_op)
    z = cube[uint32](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[uint32]) -> -a = scalar_binaryop(a, a, scalar_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[uint32], b : cube[uint32]) -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[uint32], b : cube[uint32]) -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[uint32], b : cube[uint32]) -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a * b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a * b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[uint32], b : cube[uint32]) -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a / b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[uint32], b : cube[uint32]) -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[uint32])       -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
% Less
reduction (a : cube[uint32], b : cube[uint32]) -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
% Greater
reduction (a : cube[uint32], b : cube[uint32]) -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[uint32], b : cube[uint32]) -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[uint32], b : cube[uint32]) -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
% Equal
reduction (a : cube[uint32], b : cube[uint32]) -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[uint32], b : cube[uint32]) -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[uint32]) -> !a = scalar_binaryop(a, a, scalar_negate) where $ftype("__host__")

%% And
%reduction (a : cube[uint32], b : cube[uint32]) -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%reduction (a : cube[uint32], b : scalar)       -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint32]) -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%% Or
%reduction (a : cube[uint32], b : cube[uint32]) -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%reduction (a : cube[uint32], b : scalar)       -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint32]) -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%% Xor
%reduction (a : cube[uint32], b : cube[uint32]) -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%reduction (a : cube[uint32], b : scalar)       -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint32]) -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%% Shl
%reduction (a : cube[uint32], b : cube[uint32]) -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%reduction (a : cube[uint32], b : scalar)       -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint32]) -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%% Shr
%reduction (a : cube[uint32], b : cube[uint32]) -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
%reduction (a : cube[uint32], b : scalar)       -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint32]) -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
% Min
reduction (a : cube[uint32], b : cube[uint32]) -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
% Max
reduction (a : cube[uint32], b : cube[uint32]) -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : cube[uint32], b : scalar)       -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint32]) -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")


function y : cube'unchecked = uint32_to_float(x : cube[uint32]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y : cube[uint32]'unchecked = float_to_uint32(x : cube'unchecked) 
    y = cube[uint32](size(x))  
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end
        end
    end
end

reduction (x : cube[uint32])   -> float(x) = uint32_to_float(x) where $ftype("__host__")
reduction (x : cube)           -> uint32(x) = float_to_uint32(x) where $ftype("__host__")

%==============================================================================
% cube[int64] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[int64]'unchecked = scalar_binaryop(x : cube[int64]'unchecked, y : cube[int64]'unchecked, op : scalar_bin_op)
    z = cube[int64](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[int64]'unchecked = scalar_binaryop(x : cube[int64]'unchecked, y : scalar, op : scalar_bin_op)
    z = cube[int64](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[int64]'unchecked = scalar_binaryop(x : scalar, y : cube[int64]'unchecked, op : scalar_bin_op)
    z = cube[int64](size(y))
    {!parallel for}
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[int64]) -> -a = scalar_binaryop(a, a, scalar_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[int64], b : cube[int64]) -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[int64], b : cube[int64]) -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[int64], b : cube[int64]) -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a * b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a * b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[int64], b : cube[int64]) -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a / b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[int64], b : cube[int64]) -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[int64])       -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
% Less
reduction (a : cube[int64], b : cube[int64]) -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
% Greater
reduction (a : cube[int64], b : cube[int64]) -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[int64], b : cube[int64]) -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[int64], b : cube[int64]) -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
% Equal
reduction (a : cube[int64], b : cube[int64]) -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[int64], b : cube[int64]) -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[int64]) -> !a = scalar_binaryop(a, a, scalar_negate) where $ftype("__host__")

%% And
%reduction (a : cube[int64], b : cube[int64]) -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%reduction (a : cube[int64], b : scalar)       -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%reduction (a : scalar,       b : cube[int64]) -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%% Or
%reduction (a : cube[int64], b : cube[int64]) -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%reduction (a : cube[int64], b : scalar)       -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%reduction (a : scalar,       b : cube[int64]) -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%% Xor
%reduction (a : cube[int64], b : cube[int64]) -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%reduction (a : cube[int64], b : scalar)       -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%reduction (a : scalar,       b : cube[int64]) -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%% Shl
%reduction (a : cube[int64], b : cube[int64]) -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%reduction (a : cube[int64], b : scalar)       -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%reduction (a : scalar,       b : cube[int64]) -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%% Shr
%reduction (a : cube[int64], b : cube[int64]) -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
%reduction (a : cube[int64], b : scalar)       -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
%reduction (a : scalar,       b : cube[int64]) -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
% Min
reduction (a : cube[int64], b : cube[int64]) -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
% Max
reduction (a : cube[int64], b : cube[int64]) -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : cube[int64], b : scalar)       -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[int64]) -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")


function y : cube'unchecked = int64_to_float(x : cube[int64]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end
        end
    end
end

function y : cube[int64]'unchecked = float_to_int64(x : cube'unchecked) 
    y = cube[int64](size(x))  
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end
        end
    end
end

reduction (x : cube[int64])   -> float(x) = int64_to_float(x) where $ftype("__host__")
reduction (x : cube)          -> int64(x) = float_to_int64(x) where $ftype("__host__")

%==============================================================================
% cube[uint64] type implementation
%==============================================================================

% A binary operation on integers
function z : cube[uint64]'unchecked = scalar_binaryop(x : cube[uint64]'unchecked, y : cube[uint64]'unchecked, op : scalar_bin_op)
    z = cube[uint64](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y[m,n,p])
            end
        end
    end
end

function z : cube[uint64]'unchecked = scalar_binaryop(x : cube[uint64]'unchecked, y : scalar, op : scalar_bin_op)
    z = cube[uint64](size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                z[m,n,p] = op(x[m,n,p],y)
            end
        end
    end
end

function z : cube[uint64]'unchecked = scalar_binaryop(x : scalar, y : cube[uint64]'unchecked, op : scalar_bin_op)
    z = cube[uint64](size(y))
    for m=0..size(y,0)-1
        for n=0..size(y,1)-1
            for p=0..size(y,2)-1
                z[m,n,p] = op(x,y[m,n,p])
            end
        end
    end
end

% Negation
reduction (a : cube[uint64]) -> -a = scalar_binaryop(a, a, scalar_minus) where $ftype("__host__")
% Point-wise addition of matrices
reduction (a : cube[uint64], b : cube[uint64]) -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a + b = scalar_binaryop(a, b, scalar_add) where $ftype("__host__")
% Point-wise subtraction of matrices
reduction (a : cube[uint64], b : cube[uint64]) -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a - b = scalar_binaryop(a, b, scalar_subtract) where $ftype("__host__")
% Point-wise multiplication of matrices
reduction (a : cube[uint64], b : cube[uint64]) -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a .* b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a * b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a * b = scalar_binaryop(a, b, scalar_mul) where $ftype("__host__")
% Point-wise division of matrices
reduction (a : cube[uint64], b : cube[uint64]) -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a ./ b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a / b = scalar_binaryop(a, b, scalar_divide) where $ftype("__host__")
% Point-wise power of matrices
reduction (a : cube[uint64], b : cube[uint64]) -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
reduction (a : scalar, b : cube[uint64])       -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a .^ b = scalar_binaryop(a, b, scalar_pow) where $ftype("__host__")
% Less
reduction (a : cube[uint64], b : cube[uint64]) -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a < b = scalar_binaryop(a, b, scalar_less) where $ftype("__host__")
% Greater
reduction (a : cube[uint64], b : cube[uint64]) -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a > b = scalar_binaryop(a, b, scalar_greater) where $ftype("__host__")
% Less or equal
reduction (a : cube[uint64], b : cube[uint64]) -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a <= b = scalar_binaryop(a, b, scalar_leq) where $ftype("__host__")
% Greater or equal
reduction (a : cube[uint64], b : cube[uint64]) -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a >= b = scalar_binaryop(a, b, scalar_geq) where $ftype("__host__")
% Equal
reduction (a : cube[uint64], b : cube[uint64]) -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a == b = scalar_binaryop(a, b, scalar_eq) where $ftype("__host__")
% Not equal
reduction (a : cube[uint64], b : cube[uint64]) -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> a != b = scalar_binaryop(a, b, scalar_neq) where $ftype("__host__")
% Inverse
reduction (a : cube[uint64]) -> !a = scalar_binaryop(a, a, scalar_negate) where $ftype("__host__")

%% And
%reduction (a : cube[uint64], b : cube[uint64]) -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%reduction (a : cube[uint64], b : scalar)       -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint64]) -> and(a,b) = scalar_binaryop(a, b, scalar_and) where $ftype("__host__")
%% Or
%reduction (a : cube[uint64], b : cube[uint64]) -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%reduction (a : cube[uint64], b : scalar)       -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint64]) -> or(a,b) = scalar_binaryop(a, b, scalar_or) where $ftype("__host__")
%% Xor
%reduction (a : cube[uint64], b : cube[uint64]) -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%reduction (a : cube[uint64], b : scalar)       -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint64]) -> xor(a,b) = scalar_binaryop(a, b, scalar_xor) where $ftype("__host__")
%% Shl
%reduction (a : cube[uint64], b : cube[uint64]) -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%reduction (a : cube[uint64], b : scalar)       -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint64]) -> shl(a,b) = scalar_binaryop(a, b, scalar_shl) where $ftype("__host__")
%% Shr
%reduction (a : cube[uint64], b : cube[uint64]) -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
%reduction (a : cube[uint64], b : scalar)       -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
%reduction (a : scalar,       b : cube[uint64]) -> shr(a,b) = scalar_binaryop(a, b, scalar_shr) where $ftype("__host__")
% Min
reduction (a : cube[uint64], b : cube[uint64]) -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> min(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
% Max
reduction (a : cube[uint64], b : cube[uint64]) -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : cube[uint64], b : scalar)       -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")
reduction (a : scalar,       b : cube[uint64]) -> max(a,b) = scalar_binaryop(a, b, scalar_min) where $ftype("__host__")


function y : cube'unchecked = uint64_to_float(x : cube[uint64]'unchecked)
    y = uninit(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = float(x[m,n,p])
            end                      
        end
    end
end

function y : cube[uint64]'unchecked = float_to_uint64(x : cube'unchecked) 
    y = cube[uint64](size(x))  
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            for p=0..size(x,2)-1
                y[m,n,p] = x[m,n,p]
            end           
        end
    end
end

reduction (x : cube[uint64])   -> float(x) = uint64_to_float(x) where $ftype("__host__")
reduction (x : cube) -> uint64(x) = float_to_uint64(x) where $ftype("__host__")
