%==============================================================================
% joint_bilateral_upsampling.q
% Joint bilateral upsampling as in "Joint Bilateral Upsampling" (Kopf et al., 2007)
% Simon Donné, April 2014
%==============================================================================

import "imresize.q"
import "colortransform.q"

{!author name="Simon Donné"}
{!doc category="Image Processing/Interpolation"}


%Joint Bilateral Upsampling (Kopf et al., 2007)
%   scales up an image using a guidance image
%       useful for pyramidic schemes, e.g. in stereo matching or optical flow
%       when upsampling from one level to the next
%
%   the options unknowns decides the handling of 0-values: if unknowns is set,
%       zero values will be seen as unknowns and ignored
%
% Function: bilateral_upsample
%
% Performs upsampling of a low-resolution image using a guide image of the required resolution.
%   An implementation of "Joint Bilateral Upsampling" (Kopf et al., 2007).
%
% :function img:cube = bilateral_upsample(lowres:cube, highres:cube, unknowns:int, d:int, alpha:scalar, beta:scalar)
%
% Parameters:
% lowres - The image to upsample. An arbitrary number of channels is allowed.
% highres - The guide image. Its size dictates the size of the upscaled image, and it should be in the Lab colour space.
% unknowns - How should 0-entries be handled? If this flag is set, they will be ignored in calculations - assumed to be "unknown" (such as gaps in depth maps and such).
% d - window size for the bilateral guide filter
% alpha - parameter for the guide filter (colour distance)
% beta - parameter for the guide filter (spatial distance)
function [img:cube] = bilateral_upsample(lowres:cube, highres:cube, unknowns:int, d:int, alpha:scalar, beta:scalar)
    [M,N] = size(highres, 0..1)
    [lM, lN] = size(lowres, 0..1)
    img = zeros(M,N, size(lowres,2))
    
    function [] = __kernel__ JBU(lowresk:cube'clamped, highresk:cube'clamped, ys:scalar, xs:scalar, d:int, pos:ivec3)        
        coeffsum = 0.0
        valsum = 0.0
        
        for yo = -d..d
        for xo = -d..d
            qqd:ivec4 = [pos[0]+yo,pos[1]+xo, int(round((pos[0] + yo)/ys)),int(round((pos[1]+xo)/xs))] %TODO instead of round, interpolate?
            modi = (unknowns && (lowresk[qqd[2], qqd[3],pos[2]]==0)) || (qqd[0] < 0) || (qqd[0] >= M) || (qqd[1] < 0) || (qqd[1] >= N)
            if(modi == 0)
                cdist = (highresk[qqd[0],qqd[1],0] - highresk[pos[0], pos[1],0])^2 _
                            + (highresk[qqd[0],qqd[1],1] - highresk[pos[0], pos[1],1])^2 _
                            + (highresk[qqd[0],qqd[1],2] - highresk[pos[0], pos[1],2])^2
                thiscoeff = exp(-(cdist/alpha)-((xo^2 + yo^2)/beta))
                coeffsum = coeffsum + thiscoeff
                valsum = valsum + thiscoeff * lowresk[qqd[2], qqd[3],pos[2]]
            endif
        end
        end
        
        img[pos] = valsum / max(coeffsum, 1e-9)
    end
    parallel_do(size(img), lowres, highres, M/lM, N/lN, d, JBU)
end

function [] = main()
    %less impressive test scenario
    img_orig = imread("lena_big.tif")
    [M,N] = size(img_orig, 0..1)
    imshow(img_orig)
    resize_factor = 2
    img_small = imresize(img_orig, [int(M/resize_factor),int(N/resize_factor)])
    %add some AWGN
    img_small = img_small + rand(size(img_small))*5
    imshow(img_small)
    
    img_lab = rgb2lab(img_orig)
    tic()
    bilup=bilateral_upsample(img_small, img_lab, 1, 7, 20, 25)
    toc() 
    imshow(bilup)
    
    imshow(imresize(img_small, [M,N], "linear"))
    

    %test code on stereo matching datasets:
%    %test data: stereo matching ground truth
%    highres=imread("/scratch/sdonn/SM/Cones/im2.ppm")
%    [M,N] = size(highres, 0..1)
%    lowresc=imread("/scratch/sdonn/SM/Cones/disp2.png")[:,:,0]/4
%    
%    resize_factor = 2
%    
%    %anti-aliased linear downsampling taking into account unknowns:
%    lowres = imresize(lowresc, [int(M/resize_factor),int(N/resize_factor)])
%    
%    %add some noise in the estimated locations of the image (non-0 valued)
%    lowres = lowres + (lowres!=0) .*rand(size(lowres, 0..1)) * 1
%    
%    imshow(lowres)
%    
%    labhighres = rgb2lab(highres)
%    tic()
%    bilup =bilateral_upsample(lowres, labhighres, 1, 7, 20, 25)
%    toc() 
%    imshow(bilup)
%    
%    imshow(imresize(lowres, [M,N], "linear"))
end
