import "inttypes.q"
import "matrix_inverse_device.q"

% Section: kmeans.q
% 
% K-means clustering algorithms
% 

function [centroids : mat(:,N)] = kmeans_initialization[N : int](data : mat(:,N), K : int, method : string = "forgy")
    M = size(data,0) % number of data samples
    index = int(rand(K)*M)
    
    match method with
    | "forgy" ->
        centroids = uninit(K,N)
       
        {!parallel for}
        for k=0..K-1
            centroids[k,:] = data[index[k],:]
        endfor
    | "random" -> % random partition method
    
        minval = mindim(data,0)
        maxval = maxdim(data,0)
        centroids = (repmat(0.5 * (minval + maxval), [K,1]) + _
                     randn(K,N)) .* repmat(maxval - minval,[K,1])
    
    endmatch

endfunction

function y = __device__ index_of_max[N:int](x : vec(N))
    y = 0
    v = x[0]
    for n=1..numel(x)-1
        {!unroll}
        w = x[n]
        if w>v
            v=w
            y=n
        endif
    endfor
endfunction


function [costfn] = kmeans_clustering_it[N:int,K:int](data : mat(:,N), centroids : mat(K,N), membership : vec[int])
    M = size(data,0) % number of data samples
    assert(numel(membership) == M, "membership vector has invalid length")
    
    % Performance results RTX 2080 (200 iterations)
    %        Kernel Fusion     | Shared Mem promotion    | Execution time
    % -------------------------|-------------------------|-----------------------------------
    %           Off            |     Off                 |  1493.382 ms
    %           On             |     Off                 |  1589.914 ms
    %           Off            |     On                  |  268.87 ms
    %           On             |     On                  |  171.79 ms
    
    
    %{!kernel_fusion strategy="manual"}
    {!function_transform enable="sharedmempromotion"}
    
    % update the membership
    costfn = 0.0
    {!parallel for}
    for m=0..M-1
        sample = data[m,:]
        d = zeros(K) % dynamic kernel memory avoided by type parameter
    
        % for each cluster
        for k=0..K-1
            {!unroll}
            d[k] = sum((centroids[k,:] - sample).^2)
        endfor
        
        membership[m] = index_of_max(-d)
        costfn += min(d)
    endfor
    
    % update the cluster centers
    count  = ones(K)
    
    {!parallel for}
    for k=0..M-1

        mship = membership[k]
        count[mship] += 1
        centroids[mship,:] += data[k,:]

    endfor
    
    % normalization
    centroids ./= repmat(transpose(max(1,count)),[1,N])
    %print costfn
endfunction

function y = __device__ gaussian_dist_eval[N:int](mu : vec(N), C : mat(N,N), x : vec(N))
    y = det(2*pi*C)^(-0.5) * exp(-0.5*(x-mu)*inv(C)*transpose(x-mu))
endfunction

function y = __device__ gaussian_dist_inv_eval[N:int](mu : vec(N), C : mat(N,N), x : vec(N))
    y = det(2*pi*C)^(-0.5) * exp(-0.5*(x-mu)*C*transpose(x-mu))
endfunction


function [costfn] = gaussian_mixture_clustering_it[N:int,K:int](data : mat(:,N), mu : mat(K,N), _
    weight : vec(K), cov : cube(K,N,N), membership : mat(:,K))
    
    M = size(data,0) % number of data samples
    assert(size(membership,0) == M, "membership vector has invalid length")

    % E-step: update the membership
    costfn = 0.0
    Nk = zeros(K)
    new_mu = zeros(K,N)
    new_cov = zeros(K,N,N)
    
    {!parallel for}
    for m=0..M-1
        
        % TODO handling too much shared memory - work in steps k=0..K0..K-1
        mu_sh : 'shared = mu
        cov_inv_sh : 'shared = uninit(size(cov))
        
        {!begin_grid_group}
        for k=0..K-1
            cov_inv_sh[k,:,:] = inv(cov[k,:,:])
        endfor
        
        % store cov[:,:,:], mu[:,:] in shared memory
        sample = data[m,:]
        
        % M step
        p_k = uninit(K)
        p_k_tot = 0.0
        for k=0..K-1
            p_k[k] = gaussian_dist_inv_eval[N](mu_sh[k,:], cov_inv_sh[k,:,:], sample) * weight[k]
            p_k_tot += p_k[k]
        endfor
        p_k = p_k/max(1e-10,p_k_tot)
        membership[m,:] = p_k
        
        % E step
        Nk += p_k       
        
        for k=0..K-1
            new_mu[k,:] += p_k[k] * sample
            new_cov[k,:,:] += p_k[k] .* (transpose(sample - mu_sh[k,:]) * (sample - mu_sh[k,:]))
        endfor
                
        costfn += sum(p_k)
        {!end_grid_group}
    endfor
    
    % Write the results to the output matrices
    Nk = max(Nk,1)
    weight[:] = Nk/M
    mu[:,:] = new_mu./repmat(transpose(Nk),[1,N])
    cov[:,:,:] = new_cov./repmat(transpose(Nk),[1,N,N])
endfunction

function [mu : mat(:,N), cov : cube(:,N,N), weights : vec(:)] = _
    gmem_initialization[N:int](data : mat(:,N), K : int)
    
    M = size(data,0) % number of data samples
    index = int(rand(K)*M)    

    mu = uninit(K,N)
    cov = uninit(K,N,N)
   
    {!parallel for}
    for k=0..K-1
        mu[k,:] = data[index[k],:]
        cov[k,:,:] = eye(N)
    endfor
    
    weights = ones(K)/K

endfunction



function [] = main()

    img = imread("lena_big.tif")
    
    N : 'const = 3 % data dimensionality
    K : 'const = 16 % number of clusters
    data = reshape(img, [size(img,0)*size(img,1),N])
    centroids : mat(K,N) = kmeans_initialization(data,K)
    membership = vec[int](size(data,0))    
   
    
    tic()
    costfn = maxvalue(scalar)
    for iter=0..199
        oldcostfn = costfn
        costfn = kmeans_clustering_it(data, centroids, membership)
        if 0&&iter > 10 && costfn > 0.999999*oldcostfn
            break % converged
        endif
    endfor
    toc(sprintf("K means clustering %d iterations", iter))
    print "iter=",iter, " costfn=",costfn
    print "centroids: ", centroids
    print "sum(membership)=", sum(membership)
    
    img_kmeans = uninit(size(img))
    {!parallel for}
    for m=0..size(img,0)-1
        for n=0..size(img,0)-1
            img_kmeans[m,n,:] = centroids[membership[m*size(img,0)+n],:]
        endfor
    endfor
    imshow(img_kmeans,[0,255]),title("kmeans clustering")

    return

    [mu : mat(K,N), cov : cube(K,N,N), weights : vec(K)] = gmem_initialization(data, K)
    
    print type(mu),",",type(cov),",",type(weights)
    membership2 = zeros(size(data,0),numel(weights))
    tic()
    costfn = maxvalue(scalar)
    for iter=0..3
        oldcostfn = costfn
        costfn = gaussian_mixture_clustering_it(data, mu, weights, cov, membership2)
        if costfn > 0.999*oldcostfn 
            break % converged
        endif
        print "iter=",iter, " costfn=",costfn
    endfor
    toc("EM clustering")
    print "weights: ", weights

    img_em = uninit(size(img))
                
    {!parallel for}
    for m=0..size(img,0)-1
        for n=0..size(img,0)-1
            img_em[m,n,:] = mu[index_of_max(membership2[m*size(img,0)+n,:]),:]
        endfor
    endfor
    
            
    
    imshow(img_em,[0,255]),title("em clustering")

endfunction
