%============================================
% Quasar library for linear algebra routines
% Oct 2012 - Bart Goossens
%============================================
import "system.q"

{!author name="Bart Goossens"}
{!doc category="Linear algebra"}

% Section: linalg.q
% Quasar library for linear algebra routines. 

% Function: utriangm
% 
% Constructs an upper triangular matrix based on a vector containing the
% upper triangular matrix elements
% 
% Usage:
%   : function [Y : mat] = __device__ utriangm[N](X : vec'const'static'unchecked((N*(N+1)/2)))
% 
% Parameters:
% X - input vector of length N*(N-1)
% 
% Returns:
% Y - N x N upper triangular matrix
function Y = __device__ utriangm[N](X : vec'unchecked(N*(N+1)/2))
    Y = zeros(N,N)
    i = 0
    for m=0..N-1
        for n=m..N-1
            Y[m,n] = X[i]
            i += 1
        endfor
    endfor
endfunction

function Y = __device__ utriangv[N](X : mat(N,N))
    Y = zeros(N*(N+1)/2)
    i = 0
    for m=0..N-1
        for n=m..N-1
            Y[i] = X[m,n]
            i += 1
        endfor
    endfor
endfunction

% Function: utriang_vprod
% 
% Calculates the upper triangular part of the product of two vectors A*B^T
% 
% Usage:
%   : function [Y] = __device__ utriang_vprod[N](A : vec'col(N), B : vec'col'unchecked(N))
% 
% Parameters:
% A - Input vector A
% B - Input vector B
% 
% Returns:
% Y - Vector containing upper triangular elements of the product A*B^T
function Y = __device__ utriang_vprod[N](A : vec'col'unchecked(N), B : vec'col'unchecked(N))
    Y = uninit(N*(N+1)/2)
    i = 0
    for m=0..N-1
        for n=m..N-1
            Y[i] = A[m] * B[n]
            i += 1
        endfor
    endfor
endfunction

% Function: utriang_symext
% 
% Performs symmetric extension of an upper triangular matrixx
% 
% Usage:
%   : function [Y : mat] = __device__ utriang_symext[N](X : mat'const(N,N))
% 
% Parameters:
% X - an upper triangular matrix
% 
% Returns:
% Y - a symmetric matrix
function Y = __device__ utriang_symext[N](X : mat(N,N))
    Y = uninit(N,N)
    for m=0..N-1
        for n=0..N-1
            Y[m,n]=n<m ? X[n,m] : X[m,n]
        endfor
    endfor
endfunction

% Function: inv
% Computes the inverse of a squared matrix
%
% : function y : mat = inv(x : mat)
%
function y : mat = inv(x : mat)

    [rows,cols] = size(x,0..1)
    if ndims(x) > 2 || rows != cols
        error "(inv) Matrix must be square!"
    endif

    A = zeros(rows, 2*cols)

    % A serial function
    function [] = __kernel__ serial_kernel (A : mat'unchecked, x : mat'unchecked, rows : int, cols : int)
        row = 0
        temp = 0.0
        for i=0..rows-1
            for j=0..cols-1
                A[i,j] = x[i,j]
                A[i,j+cols] = (i == j) ? 1 : 0
            endfor
        endfor

        for j=0..cols-1
            if A[j,j]==0
                % Search non-zero element in column
                row = j
                for i=j..rows-1
                    if abs(A[i,j])>0
                        row = i
                    endif
                endfor
                if row == j                    
                    error "(inv) There is no existing inverse matrix!"
                else
                    % Swap row i and j
                    for k=0..size(A,1)-1
                        temp = A[row,k]
                        A[row,k] = A[j,k]
                        A[j,k] = temp
                    endfor
                endif
            endif

            % Row operations
            for i=0..rows-1
                temp = A[i,j]
                if temp != 0.0 && i != j
                    for k=0..size(A,1)-1
                        A[i,k] = A[i,k]*A[j,j] - temp*A[j,k]
                    endfor
                endif
            endfor
            
            % Rescale all rows
            for i=0..rows-1
                temp = A[i,i]
                if temp != 0.0
                    for j2=0..size(A,1)-1
                        A[i,j2] = A[i,j2] / temp
                    endfor
                endif
            endfor
        endfor
    endfunction

    serial_do(1, A, x, rows, cols, serial_kernel)

    y = A[:, cols..2*cols-1]            
endfunction

function I  = __device__ inv(X : mat)
    N = size(X,0)
    assert(size(X,0)==size(X,1), "(inv) Matrix must be square!")

    Y = copy(X)
    I = eye(N)

    for i=0..N-1
        Yii = Y[i,i]
        %{!parallel for}
        for j=0..N-1
            if j != i
                I[i,j] = I[i,j] / Yii
                Y[i,j] = Y[i,j] / Yii
            endif
        endfor
        I[i,i] = I[i,i] / Yii
        Y[i,i] = 1.0

        %{!parallel for}
        for m=0..N-1
            for n=0..N-1
                if m != i
                    Yii = Y[m, i]
                    I[m,n] = I[m,n] - I[i, n] * Yii
                    if n != i
                        Y[m,n] = Y[m,n] - Y[i, n] * Yii
                    endif
                endif
            endfor
        endfor

        for n=0..N-1
            Y[n,i] = 0.0
        endfor
    endfor
endfunction


% Function: cholesky_device
% 
% Performs Cholesky decomposition of the positive definite matrix Sigma, storing the
% result in Y. At the end of the process, the following relationship holds:
%
% :Sigma = Y*transpose(Y)
%
% where Y is a lower triangular matrix
% 
% Usage:
%   : function [] = __device__ cholesky_device(Sigma : mat'const, Y : mat, p : vec)
% 
% Parameters:
% Sigma - The positive definite input matrix
% Y - The output matrix
% p - A temporary vector with length the same as the rows/colums of Sigma
function [] = __device__ cholesky_device(Sigma : mat, Y : mat, p : vec)
    % % Algorithm ported from Numerical Recipes in C
    assert(size(Sigma,0) == size(Sigma,1), "Square matrix expected!")
    assert(size(Y) == size(Sigma), "Matrices should be the same size")
    assert(size(Sigma,0) == numel(p))
    n = size(Sigma,0)
    
    for i=0..n-1
        for j=0..n-1
            Y[i,j] = Sigma[i,j]
        endfor
    endfor
    
    for i=0..n-1
        for j=i..n-1
            sum = Y[i,j]
            for k=i-1..-1..0
                sum -= Y[i,k]*Y[j,k]
            endfor
            if i==j
                if sum <= 0
                    error "Sigma, with rounding errors, is not positive definite" 
                endif
                p[i]=sqrt(sum)
            else
                Y[j,i]=sum/p[i]
            endif
        endfor
    endfor
    for i=0..n-1
        for j=0..n-1
            if i<j
                Y[i,j]=0.0
            endif
        endfor
    endfor
    for i=0..n-1
        Y[i,i] = p[i]
    endfor
endfunction


% Function: cholesky
% 
% Performs Cholesky decomposition of the positive definite matrix X,:
%
% :X = Y*transpose(Y)
% 
% where Y is a lower triangular matrix
%
% Usage:
%   : function [Y : mat] = cholesky(X : mat)
%   : function [Y : mat] = __device__ cholesky(X : mat)
% 
% Parameters:
% X - A positive definite input matrix
% 
% Returns:
% Y - A lower triangular matrix matrix Y, satisfying X = Y*transpose(Y)
function [Y : mat] = cholesky(X : mat)
    Y = uninit(size(X))
    p = uninit(size(X,0))
    cholesky_device(X, Y, p)
endfunction

function [Y : mat] = __device__ cholesky(X : mat)
    Y = uninit(size(X))
    p = uninit(size(X,0))
    cholesky_device(X, Y, p)
endfunction


% Algorithm ported from Numerical Recipes in C
% Given a matrix a, this routine replaces it by the LU decomposition of a rowwise
% permutation of itself.  a and n are input.  a is output, arranged as in equation 
% (2.3.14) above indx[1..n]  is  an  output  vector  that  records  the  row  
% permutation  effected  by  the  partial pivoting; d is output as �1 depending on 
% whether the number of row interchanges was even or odd, respectively.  This 
% routine is used in combination with lubksb to solve linear equations or  invert a 
% matrix.
% Note: indx should have length 'size(a,0)'
function [d : scalar, singular : int] = __device__ ludcmp(a : mat'unchecked, indx : vec[int]'unchecked, vv : vec)
    n = size(a,0)
    d = 1.0
    singular = 0
    imax = 0

    % No row interchanges yet.
    for i=0..n-1  % Loop over rows to get the implicit scaling information.	
        big=0.0
        for j=0..n-1
            big = max(big, abs(a[i,j]))
        endfor
        if big==0.0
            singular = 1
            return
        endif
        vv[i] = 1.0/big % No nonzero largest element - save the scaling
    endfor

    for j=0..n-1  % This is the loop over columns of Crout's method.
        for i=0..j-1  % This is equation (2.3.12) except for i = j.
            sum = a[i,j]
            for k=0..i-1
                sum -= a[i,k]*a[k,j]
            endfor
            a[i,j] = sum
        endfor

        big = 0.0  % Initialize for the search for largest pivot element.
        for i=j..n-1  % This is i = j of equation (2.3.12) and i = j +1...N of equation (2.3.13).
            sum = a[i,j]
            for k=0..j-1
                sum -= a[i,k]*a[k,j]
            endfor
            a[i,j] = sum
            dum = abs(vv[i]*sum)
            if dum >= big % Is the figure of merit for the pivot better than the best so far?
                big = dum
                imax = i
            endif
        endfor

        if j != imax  % Do we need to interchange rows?	
            for k=0..n-1
                dum2 = a[imax,k]
                a[imax,k]=a[j,k]
                a[j,k] = dum2
            endfor
            d = -d % ...and change the parity of d.
            vv[imax] = vv[j] % Also interchange the scale factor.
        endif

        indx[j] = imax
		% If the  pivot element is zero the matrix is singular (at least to the precision of the
		% algorithm). For some applications on singular matrices, it is desirable to substitute eps for zero
        if a[j,j] == 0.0
            a[j,j] = eps
        endif

        if j != n  %  Now, finally, divide by the pivot element.
            dum2 = 1.0/a[j,j]
            for i=j+1..n-1
                a[i,j] = a[i,j] * dum2
            endfor
        endif
    endfor
endfunction


% Algorithm ported from Numerical Recipes in C
% Solves the set of n linear equations A�X = B.  Here a[1..n][1..n] is input, not as the matrix
% A but rather as its LU decomposition, determined by the routine ludcmp.  indx[1..n] is input
% as the permutation vector returned by ludcmp.  b[1..n] is input as the right-hand side vector
% B, and returns with the solution vector X.  a, n, and indx are not modified by this routine
% and can be left in place for successive calls with di?erent right-hand sides b.  This routine takes
% into account the possibility that b will begin with many zero elements, so it is efficient for use
% in  matrix  inversion.
%
function [] = __device__ lubksb(a : mat'const'unchecked, indx : vec[int]'const'unchecked, b : vec'unchecked)
    ii=-1
    n = size(a,0)
    % When ii is set to a positive value, it will become the index of the first nonvanishing 
	% element of b.  We now do the  forward substitution,  equation  (2.3.6). The only new wrinkle 
	% is to unscramble the permutation as we go.
    for i=0..n-1
        ip = indx[i]
        sum = b[ip]
        b[ip] = b[i]
        if ii>=0
            for j=ii..i-1
                sum -= a[i,j]*b[j]
            endfor
        elseif abs(sum) != 0.0 % A nonzero element was encountered, so from now on we
            ii = i             % will have to do the sums in the loop above.
        endif
        b[i] = sum
    endfor

    for i=n-1..-1..0 % Now we do the backsubstitution, equation (2.3.7).
        sum = b[i]
        for j=i+1..n-1
            sum -= a[i,j]*b[j]
        endfor
        b[i] = sum/a[i,i] % Store a component of the solution vector X.
    endfor
endfunction

% Function: lsolve
% Solve the linear system of equations A.X = B (where X is a vector)
%
% : function [X : vec] = lsolve(A : mat, B : vec)
% : function [X : vec] = __device__ lsolve(A : mat, B : vec)
%
function [X : vec] = lsolve(A : mat, B : vec)
    assert(size(A,0)==size(A,1), "(lsolve) Square matrix expected for A!")
    indx = vec[int](size(A,0))
    C = copy(A) % Because we do not want to overwrite A
    X = copy(B) % Because we do not want to overwrite B
    [d,singular] = ludcmp(C, indx, zeros(size(A,0)))
    if singular
        error "(lsolve) Singular matrix detected!"
    endif
    lubksb(C, indx, X)
endfunction

function [X : vec] = __device__ lsolve(A : mat, B : vec)
    assert(size(A,0)==size(A,1), "(lsolve) Square matrix expected for A!")
    indx = vec[int](size(A,0))
    C = copy(A) % Because we do not want to overwrite A
    X = copy(B) % Because we do not want to overwrite B
    [d, singular] = ludcmp(C, indx, zeros(size(A,0)))
    if singular
        error "(lsolve) Singular matrix detected!"
    endif
    lubksb(C, indx, X)
endfunction

% Computation of a Singular Value Decomposition - code from Hiep Luong
% SVD according to:
% : a = p*d*q.transpose()
% where d is a diagonal matrix!
function [] = __device__ svd_core(a : mat'const'unchecked, p : mat'unchecked, d : mat'unchecked, q : mat'unchecked, r : vec'unchecked)

    % Helper functions
    function y = __device__ SIGN(a : scalar, b : scalar)
        y = b >= 0.0 ? abs(a) : -abs(a)
    endfunction

    function y = __device__ radius(u : scalar, v : scalar)
        u = abs(u)
        v = abs(v)
        if u > v
            w = v/u
            y = u*sqrt(1.0 + w*w)
        elseif v != 0.0
            w = u/v
            y = v*sqrt(1.0 + w*w)
        else
            y = 0.0
        endif
    endfunction

    m = size(a,0)
    n = size(a,1)
    l = 0
    nm = 0
    nm1 = n - 1
    mm1 = m - 1
    g = 0.0
    anorm = 0.0
    scale = 0.0

    % Copy the upper triangular matrix of 'a' to 'p'
    for i=0..size(p,0)-1
        for j=0..size(p,1)-1
            if i<size(a,0) && j<size(a,1)
                p[i,j] = a[i,j]
            else
                p[i,j] = 0.0
            endif
        endfor
    endfor
         
    % Householder reduction to bidigonal form
    for i=0..n-1
        l=i+1
        r[i] = scale*g
        g = 0.0
        scale = 0.0
        s = 0.0
        if i < m
            for k = i..m-1
                scale += abs(p[k,i])
            endfor
            if scale != 0.0
                for k=i..m-1
                    p[k,i]=p[k,i] / scale
                    s += p[k,i]*p[k,i]
                endfor
                f = p[i,i]
                g = -SIGN(sqrt(s),f)
                h = f*g-s
                p[i,i] = f-g
                if i != nm1
                    for j=l..n-1
                        s = 0.0
                        for k=i..m-1
                            s += p[k,i]*p[k,j]
                        endfor
                        f = s/h
                        for k=i..m-1
                            p[k,j] = p[k,j] + f*p[k,i]
                        endfor
                    endfor
                endif
                for k=i..m-1
                    p[k,i]=p[k,i]*scale
                endfor
            endif
        endif
        d[i,i] = scale*g
        g = 0.0
        s = 0.0
        scale = 0.0
        if i<m && i != nm1
            for k=l..n-1
                scale += abs(p[i,k])
            endfor
            if scale != 0
                for k=l..n-1
                    p[i,k]=p[i,k]/scale
                    s += p[i,k]*p[i,k]
                endfor
                f = p[i,l]
                g = -SIGN(sqrt(s),f)
                h = f*g-s
                p[i,l]=f-g
                for k=l..n-1
                    r[k] = p[i,k]/h
                endfor
                if i != mm1
                    for j=l..m-1
                        s = 0.0
                        for k=l..n-1
                            s += p[j,k]*p[i,k]
                        endfor
                        for k=l..n-1
                            p[j,k] = p[j,k] + s*r[k]
                        endfor
                    endfor
                endif
                for k=l..n-1
                    p[i,k] = p[i,k]*scale
                endfor
            endif
        endif
        x = abs(d[i,i]) + abs(r[i])
        anorm = max(anorm,x)
    endfor
        
    % Accumulation of right-hand transformations
    for i=n-1..-1..0
        if i<nm1
            if abs(g)!=0.0
                for j=l..n-1
                    q[j,i] = (p[i,j]/p[i,l])/g
                endfor
                for j=l..n-1
                    s = 0.0
                    for k=l..n-1
                        s += p[i,k]*q[k,j]
                    endfor
                    for k=l..n-1
                        q[k,j]=q[k,j]+s*q[k,i]
                    endfor
                endfor
            endif
            for j=l..n-1
                q[i,j]=0.0
                q[j,i]=0.0
            endfor
        endif
        q[i,i]=1
        g=r[i]
        l=i
    endfor

    % Accumulation of left-hand transformations
    for i=min(mm1,nm1)..-1..0
        l=i+1
        g=d[i,i]
        for j=l..n-1
            p[i,j]=0
        endfor
        if abs(g) != 0.0
            g=1.0/g
            for j=l..n-1
                s=0.0
                for k=l..m-1
                    s += p[k,i]*p[k,j]
                endfor
                f = (s/p[i,i])*g
                for k=i..m-1
                    p[k,j]=p[k,j]+f*p[k,i]
                endfor
            endfor
            for j=i..m-1
                p[j,i]=p[j,i]*g
            endfor
        else
            for j=i..m-1
                p[j,i]=0.0
            endfor
        endif
        p[i,i]=p[i,i]+1.0
    endfor
                    
    % diagonalization of the bidiagonal form
    for k=nm1..-1..0
        % loop over singular values
        for its=0..29
            % loop over allowed iterations
            flag = 1
            for l=k..-1..1
                % test for splitting
                nm=l-1 % note that r(l) is always zero
                if abs(r[l])+anorm==anorm
                    flag=0
                    break
                endif
                if abs(d[nm,nm])+anorm==anorm
                    break
                endif
            endfor
            if flag != 0.0
                c = 0.   % cancellation of r(l),if l>1
                s = 1.
                for i=l..k
                    f=s*r[i]
                    r[i]=r[i]*c
                    if abs(f)+anorm==anorm
                        break
                    endif
                    g = d[i,i]
                    h = radius(f,g)
                    d[i,i]=h
                    h=1.0/h
                    c=g*h
                    s=-f*h
                    for j=0..m-1
                        y=p[j,nm]
                        z=p[j,i]
                        p[j,nm]=y*c+z*s
                        p[j,i]=z*c-y*s
                    endfor
                endfor
            endif
            z = d[k,k]
            if l==k 
                if z<0.0 % convergence
                    d[k,k]=-z
                    for j=0..n-1
                        q[j,k]=-q[j,k]
                    endfor
                endif
                break
            endif
            if its==29
                return
            endif
            x = d[l,l]       % shift from bottom 2-by-2 minor 
            nm = k - 1
            y = d[nm,nm]
            g = r[nm]
            h = r[k]
            f = ((y - z)*(y + z) + (g - h)*(g + h))/(2.0*h*y)
            g = radius(f,1.)
            % next QR transformation
            f = ((x - z)*(x + z) + h*((y/(f + SIGN(g,f))) - h))/x
            c = 1.0
            s = 1.0
            for j=l..nm
                i = j + 1
                g = r[i]
                y = d[i,i]
                h = s*g
                g = c*g
                z = radius(f,h)
                r[j] = z
                c = f/z
                s = h/z
                f = x*c + g*s
                g = g*c - x*s
                h = y*s
                y = y*c
                for jj=0..n-1
                    x = q[jj,j]
                    z = q[jj,i]
                    q[jj,j] = x*c + z*s
                    q[jj,i] = z*c - x*s
                endfor
                z = radius(f,h)
                d[j,j] = z       % rotation can be arbitrary id z=0 
                if (z != 0)
                    z = 1.0/z
                    c = f*z
                    s = h*z
                endif
                f = c*g + s*y
                x = c*y - s*g
                for jj=0..m-1
                    y = p[jj,j]
                    z = p[jj,i]
                    p[jj,j] = y*c + z*s
                    p[jj,i] = z*c - y*s
                endfor
            endfor
            r[l] = 0
            r[k] = f
            d[k,k] = x
        endfor
    endfor
endfunction

% Function: svd
% Computation of a singular value decomposition (SVD).
%
% : function [p:mat,d:mat,q:mat] = svd(a:mat)
% : function [p:mat,d:mat,q:mat] = __device__ svd(a:mat)
%
% The relationship between the parameters is as follows:
% : a = p*d*transpose(q)
% where d is a diagonal matrix!
%
function [p:mat,d:mat,q:mat] = svd(a:mat)
    p = zeros(size(a,[0,1]))
    d = zeros(size(a,[1,1]))
    q = zeros(size(a,[1,1]))
    r = zeros(size(a, 1))
    svd_core(a,p,d,q,r)    
endfunction

function [p:mat,d:mat,q:mat] = __device__ svd(a:mat)
    p = zeros(size(a,[0,1]))
    d = zeros(size(a,[1,1]))
    q = zeros(size(a,[1,1]))
    r = zeros(size(a, 1))
    svd_core(a,p,d,q,r)    
endfunction


% Function: det
% Determinant of a real-valued matrix
%
% : function y : scalar = det(A : mat)
%
% See also:
%    <logdet>
%
function y : scalar = det(A : mat)
    assert(size(A,0)==size(A,1), "(det) Square matrix expected!")
    indx = vec[int](size(A,0))
    Lambda = copy(A) % Because we do not want to overwrite A
    [d, singular] = ludcmp(Lambda, indx, zeros(size(A,0)))
    if singular
        y = 0.0
    else
        y = d * prod(diag(Lambda))
    endif
endfunction


% Function: logdet
%
% Logarithm of the determinant of a real-valued matrix
%
% : function y : scalar = logdet(A : mat)
%
% The logdet evaluates the logarithm of the determinant of A by
% computing the sum of the logarithms of the eigenvalues of A, this results
% in some better numerical accuracy, especially in case the condition number 
% of A is very high.
%
% : logdet = (A : mat) -> log(det(A))
%
% See also:
%    <det>
%
function y : scalar = logdet(A : mat)
    assert(size(A,0)==size(A,1), "(logdet) Square matrix expected!")
    indx = vec[int](size(A,0))
    Lambda = copy(A) % Because we do not want to overwrite A
    [_, singular] = ludcmp(Lambda, indx, zeros(size(A,0)))
    if singular
        y = log(0)
    else
        y = 0.0
        {!parallel for}
        for i=0..size(Lambda,0)-1
            y += log(max(eps,Lambda[i,i]))
        endfor
    endif
endfunction

% Function: symsqrt
% Computes the symmetric square root of a positive definite matrix
%
% : function y : mat = symsqrt(C : mat)
function y : mat = symsqrt(C : mat)
    [U, Lambda, V] = svd(C)
    y = U*diagm(sqrt(diag(Lambda)))*transpose(V)
endfunction

% Function: pinv
% Computes the pseudo-inverse of a matrix
%
% : function y : mat = pinv(x : mat)
function y : mat = pinv(x : mat)   
    if size(x,0) == size(x,1) % square matrices
        [U, D, V] = svd(x)
        d : vec = diag(D)
        x_tol = eps * max(abs(D))
        {!parallel for}
        for i=0..numel(d)-1
            if abs(d[i])<x_tol
                d[i] = 0
            else
                d[i] = 1.0/d[i]
            endif
        endfor
        y = V*diagm(d)*transpose(U)
    % non-square matrices - commutes with transpose so we pick the fastest option.
    elseif size(x,0)*2 < size(x,1)
        y = transpose(x) * pinv(x*transpose(x))
    else
        y = pinv(transpose(x)*x) * transpose(x)
    endif
endfunction

% Function: toeplitz
%
% Computes a square toeplitz matrix
%
% : function y : mat = toeplitz(n : int, x : vec)
%
% Parameters:
%   n - the number of columns/rows of the output matrix
%   x - an odd-sized vector containing the off-diagonal elements
%
% Example:
% : toeplitz(5, [1,2,3]) =
% :     [[2,3,0,0,0],
% :      [1,2,3,0,0],
% :      [0,1,2,3,0],
% :      [0,0,1,2,3],
% :      [0,0,0,1,2]]
%
function y : mat = toeplitz(n : int, x : vec)
    y = uninit(n,n)
    center = int(numel(x)/2)
    {!parallel for}
    for i=0..n-1
        for j=0..n-1
            y[i,j] = x[i-j+center]
        endfor
    endfor
endfunction


% Function: polyval
%
% Evaluates a polynomial with coefficients `p` in points `x`
%
% : function y = polyval(p : vec, x)
%
% Parameters:
% p - the coefficients of the polynomial
% x - the x-values for which to evaluate the polynomial
%
% See also:
% <wpolyfit>, <polyfit>
function y = polyval(p : vec, x)
    function [] = __kernel__ polyval_kernel(y : cube'unchecked, x : cube'unchecked, p : vec'unchecked'const, pos : ivec3)
        y_sum = 0.0
        for i=0..numel(p)-1
            y_sum = y_sum + p[i] * x[pos]^i
        endfor
        y[pos] = y_sum
    endfunction
    if isscalar(x)
        y = sum(p .* x .^ transpose(0..numel(p)-1))
    else
        y = zeros(size(x))
        parallel_do(size(x),y,x,p,polyval_kernel)
    endif
endfunction

% Function: wpolyfit
% Performs a weighted univariate polynomial fitting
%
% : function [beta : vec] = wpolyfit(x : vec, y : vec, w : vec, order : int)
%
% Parameters:
% x - the input samples (x-coordinates)
% y - the output samples (y-coordinates)
% w - a weight per sample (between 0 and 1)
% order - the order of the polynomial to be fitted (1=linear, 2=quadratic, ...)
% 
% Returns:
% beta - the coefficients of the fitted polynomial
% 
% Remarks:
% - this function has been superseded by the function <linear_regression>.
%
% See also:
% <polyval>, <polyfit>, <linear_regression>
function [beta : vec] = wpolyfit(x : vec, y : vec, w : vec, order : int)
    if numel(x)!=numel(y) || numel(x)!=numel(w)
        error "(wpolyfit) All input dimensions should have the same size!"
    endif
    n = numel(x)
    % reshape to a vector
    y = reshape(y, [n, 1])
    x1 = reshape(x, [n, 1])*ones(1,order+1)
    w = reshape(w, [n, 1])
    % update the x-coefficients:
    parallel_do(size(x1),x1,w,__kernel__ (x : mat'unchecked, w : vec'unchecked, pos : vec2) -> _
        x[pos] = sqrt(w[pos[0]]) * (x[pos] ^ pos[1]))
    lambda=1e-2
    A = inv(transpose(x1)*x1+lambda*eye(order+1))
    beta = (A * transpose(x1)) * (sqrt(w) .* y)
endfunction

% Function: polyfit
% Performs a univariate polynomial fitting
%
% : function [beta : vec] = polyfit(x : vec, y : vec, order : int)
%
% Parameters:
% x - the input samples (x-coordinates)
% y - the output samples (y-coordinates)
% order - the order of the polynomial to be fitted (1=linear, 2=quadratic, ...)
% 
% Returns:
% beta - the coefficients of the fitted polynomial
% 
% Remarks:
% - this function has been superceded by the function <linear_regression>.
%
% See also:
% <polyval>, <wpolyfit>, <linear_regression>
polyfit = (x : vec, y : vec, order : int) -> wpolyfit(x, y, ones(size(x)), order)


function [] = __device__ jacobi_rotate(a : mat, i : int, j : int, k : int, l : int, g : scalar, h : scalar, s : scalar, tau : scalar)
    [g, h] = [a[i-1,j-1], a[k-1,l-1]]
    [a[i-1,j-1], a[k-1,l-1]] = [g-s*(h+g*tau), h+s*(g-h*tau)]
endfunction
    
% Function: jacobi
%
% Computes all eigenvalues and eigenvectors of a real symmetric matrix a. On
% output, the elements of a above the diagonal are destroyed. 
%
% Parameters:
% a - the input matrix (a real symmetric matrix)
% d - the eigenvalues of a (output matrix)
% v - a matrix whose columns contain, on output, the normalized eigenvectors of a
% nrot - the number of Jacobi rotations that were required.
function [nrot : int] = __device__ jacobi(a : mat'unchecked, n : int, d : vec'unchecked, v : mat'unchecked)
    b = zeros(n)
    z = zeros(n)       
    % initialize v to the identity matrix
    for ip=1..n
        for iq=1..n
            v[ip-1,iq-1]=0.0
        endfor
        v[ip-1,ip-1]=1.0
    endfor
    
    % initialize b and d to the diagonal
    for ip=1..n
        b[ip-1]=d[ip-1]=a[ip-1,ip-1]            
        z[ip-1]=0.0
    endfor
    nrot=0
    for i=1..50
        sm=0.0
        % sum of off-diagonal elements
        for ip=1..n-1
            for iq=ip+1..n
                sm += abs(a[ip-1,iq-1])
            endfor
        endfor
        % The normal return, which relies on quadratic convergence to machine underflow
        if abs(sm) < eps
            return
        endif
        
        if i<4 % on the first three sweeps
            tresh=0.2*sm/(n*n)
        else
            tresh=0.0 % thereafter
        endif
        for ip=1..n-1
            for iq=ip+1..n
                g=100.0*abs(a[ip-1,iq-1])
                % After four sweeps, skip the rotation if the off-diagonal element is small
                if i>4 && (abs(d[ip-1])+g) == abs(d[ip-1]) _
                       && (abs(d[iq-1])+g) == abs(d[iq-1])
                    a[ip-1,iq-1]=0.0
                elseif abs(a[ip-1,iq-1]) > tresh
                    h=d[iq-1]-d[ip-1]
                    if (abs(h)+g) == abs(h)
                        t=a[ip-1,iq-1]/h
                    else
                        theta=0.5*h/a[ip-1,iq-1]
                        t=1.0/(abs(theta)+sqrt(1.0+theta*theta))
                        if theta < 0.0
                            t=-t
                        endif
                    endif
                    c=1.0/sqrt(1+t*t)
                    s=t*c
                    tau=s/(1.0+c)
                    h=t*a[ip-1,iq-1]
                    z[ip-1] = z[ip-1] - h
                    z[iq-1] = z[iq-1] + h
                    d[ip-1] = d[ip-1] - h
                    d[iq-1] = d[iq-1] + h
                    a[ip-1,iq-1] = 0.0
                    % Case of rotations 1 = j < p.
                    for j=1..ip-1
                        jacobi_rotate(a,j,ip,j,iq,g,h,s,tau)
                    endfor
                    % Case of rotations p < j < q.
                    for j=ip+1..iq-1
                        jacobi_rotate(a,ip,j,j,iq,g,h,s,tau)
                    endfor
                    % Case of rotations q < j = n.
                    for j=iq+1..n
                        jacobi_rotate(a,ip,j,iq,j,g,h,s,tau)
                    endfor
                    for j=1..n
                        jacobi_rotate(v,j,ip,j,iq,g,h,s,tau)
                    endfor
                    nrot+=1
                endif
            endfor
        endfor
        % Update d with the sum and reinitialize z
        for ip=1..n
            b[ip-1] = b[ip-1] + z[ip-1]
            d[ip-1] = b[ip-1]
            z[ip-1] = 0.0
        endfor
    endfor
    error "Too many iterations in routine jacobi!"
endfunction
    
% Function: symeig
% 
% Calculates the eigenvalues and eigenvectors of a positive definite
% matrix, using the Jacobi algorithm
% 
% Usage:
%   : function [v : mat, lambda : vec] = __device__ symeig(x : mat)
%   : function [v : mat, lambda : vec] = symeig(x : mat)
% 
% Parameters:
% x - the input matrix
% 
% Returns:
% v - a matrix with eigenvectors
% lambda - a matrix with eigenvalues
%
% Note: the function can be called from both host and device functions.
function [v : mat, lambda : vec] = __device__ symeig(x : mat)
    assert(size(x,0)==size(x,1)) %, "symeig - square matrix expected!"

    v = eye(size(x,0))
    lambda = zeros(size(x,0))
    _ = jacobi(copy(x), size(x,0), lambda, v)
endfunction

function [v, lambda] = symeig(x : mat)
    assert(size(x,0)==size(x,1), "symeig - square matrix expected!")      

    v = zeros(size(x))
    lambda = zeros(size(x,0))
    _ = jacobi(copy(x), size(x,0), lambda, v)
endfunction

% Function: rank
% 
% Computes the rank of a matrix, i.e. the number of linearly independent columns of the matrix
% 
% Usage:
%   : function [r : int] = rank(A : mat)
% 
% Parameters:
% A - input matrix
% 
% Returns:
% r - rank of A
function r:int = rank(A:mat)
    [_,d,_] = svd(A)
    s = diag(d)
    tol = max(size(A)) * eps
    r = sum(s > tol)
endfunction

function y = __device__ det(A:mat'unchecked(2,2))
    y = A[0,0] * A[1,1] - A[0,1] * A[1,0];
endfunction

function U = __device__ inv(A:mat'unchecked(2,2))
    d = det(A)
    if d == 0
        error "Matrix not invertible"
    endif
    inv_det = 1/d
    U = [[A[1,1], -A[0,1]], [-A[1,0], A[0,0]]] * inv_det
endfunction

function y = __device__ det(A:mat'unchecked(3,3))
    y = A[0, 0] * (A[1, 1] * A[2, 2] - A[2, 1] * A[1, 2]) -
        A[0, 1] * (A[1, 0] * A[2, 2] - A[1, 2] * A[2, 0]) +
        A[0, 2] * (A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]);
endfunction

function U = __device__ inv(A:mat'unchecked(3,3))
    d = det(A)
    if d == 0
        error "Matrix not invertible"
    endif
    invdet = 1 / d

    U = uninit(3,3)
    U[0, 0] = (A[1, 1] * A[2, 2] - A[2, 1] * A[1, 2]) * invdet
    U[0, 1] = (A[0, 2] * A[2, 1] - A[0, 1] * A[2, 2]) * invdet
    U[0, 2] = (A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]) * invdet
    U[1, 0] = (A[1, 2] * A[2, 0] - A[1, 0] * A[2, 2]) * invdet
    U[1, 1] = (A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]) * invdet
    U[1, 2] = (A[1, 0] * A[0, 2] - A[0, 0] * A[1, 2]) * invdet
    U[2, 0] = (A[1, 0] * A[2, 1] - A[2, 0] * A[1, 1]) * invdet
    U[2, 1] = (A[2, 0] * A[0, 1] - A[0, 0] * A[2, 1]) * invdet
    U[2, 2] = (A[0, 0] * A[1, 1] - A[1, 0] * A[0, 1]) * invdet
endfunction

function U = __device__ inv(A:mat'unchecked(4,4))
    U = uninit(4,4)
    U[0,0] = A[1,1]  * A[2,2] * A[3,3] -
             A[1,1]  * A[2,3] * A[3,2] -
             A[2,1]  * A[1,2]  * A[3,3] +
             A[2,1]  * A[1,3]  * A[3,2] +
             A[3,1] * A[1,2]  * A[2,3] -
             A[3,1] * A[1,3]  * A[2,2];

    U[1,0] = -A[1,0]  * A[2,2] * A[3,3] +
              A[1,0]  * A[2,3] * A[3,2] +
              A[2,0]  * A[1,2]  * A[3,3] -
              A[2,0]  * A[1,3]  * A[3,2] -
              A[3,0] * A[1,2]  * A[2,3] +
              A[3,0] * A[1,3]  * A[2,2];

    U[2,0] = A[1,0]  * A[2,1] * A[3,3] -
             A[1,0]  * A[2,3] * A[3,1] -
             A[2,0]  * A[1,1] * A[3,3] +
             A[2,0]  * A[1,3] * A[3,1] +
             A[3,0] * A[1,1] * A[2,3] -
             A[3,0] * A[1,3] * A[2,1];

    U[3,0] = -A[1,0]  * A[2,1] * A[3,2] +
               A[1,0]  * A[2,2] * A[3,1] +
               A[2,0]  * A[1,1] * A[3,2] -
               A[2,0]  * A[1,2] * A[3,1] -
               A[3,0] * A[1,1] * A[2,2] +
               A[3,0] * A[1,2] * A[2,1];

    U[0,1] = -A[0,1]  * A[2,2] * A[3,3] +
              A[0,1]  * A[2,3] * A[3,2] +
              A[2,1]  * A[0,2] * A[3,3] -
              A[2,1]  * A[0,3] * A[3,2] -
              A[3,1] * A[0,2] * A[2,3] +
              A[3,1] * A[0,3] * A[2,2];

    U[1,1] = A[0,0]  * A[2,2] * A[3,3] -
             A[0,0]  * A[2,3] * A[3,2] -
             A[2,0]  * A[0,2] * A[3,3] +
             A[2,0]  * A[0,3] * A[3,2] +
             A[3,0] * A[0,2] * A[2,3] -
             A[3,0] * A[0,3] * A[2,2];

    U[2,1] = -A[0,0]  * A[2,1] * A[3,3] +
              A[0,0]  * A[2,3] * A[3,1] +
              A[2,0]  * A[0,1] * A[3,3] -
              A[2,0]  * A[0,3] * A[3,1] -
              A[3,0] * A[0,1] * A[2,3] +
              A[3,0] * A[0,3] * A[2,1];

    U[3,1] = A[0,0]  * A[2,1] * A[3,2] -
              A[0,0]  * A[2,2] * A[3,1] -
              A[2,0]  * A[0,1] * A[3,2] +
              A[2,0]  * A[0,2] * A[3,1] +
              A[3,0] * A[0,1] * A[2,2] -
              A[3,0] * A[0,2] * A[2,1];

    U[0,2] = A[0,1]  * A[1,2] * A[3,3] -
             A[0,1]  * A[1,3] * A[3,2] -
             A[1,1]  * A[0,2] * A[3,3] +
             A[1,1]  * A[0,3] * A[3,2] +
             A[3,1] * A[0,2] * A[1,3] -
             A[3,1] * A[0,3] * A[1,2];

    U[1,2] = -A[0,0]  * A[1,2] * A[3,3] +
              A[0,0]  * A[1,3] * A[3,2] +
              A[1,0]  * A[0,2] * A[3,3] -
              A[1,0]  * A[0,3] * A[3,2] -
              A[3,0] * A[0,2] * A[1,3] +
              A[3,0] * A[0,3] * A[1,2];

    U[2,2] = A[0,0]  * A[1,1] * A[3,3] -
              A[0,0]  * A[1,3] * A[3,1] -
              A[1,0]  * A[0,1] * A[3,3] +
              A[1,0]  * A[0,3] * A[3,1] +
              A[3,0] * A[0,1] * A[1,3] -
              A[3,0] * A[0,3] * A[1,1];

    U[3,2] = -A[0,0]  * A[1,1] * A[3,2] +
               A[0,0]  * A[1,2] * A[3,1] +
               A[1,0]  * A[0,1] * A[3,2] -
               A[1,0]  * A[0,2] * A[3,1] -
               A[3,0] * A[0,1] * A[1,2] +
               A[3,0] * A[0,2] * A[1,1];

    U[0,3] = -A[0,1] * A[1,2] * A[2,3] +
              A[0,1] * A[1,3] * A[2,2] +
              A[1,1] * A[0,2] * A[2,3] -
              A[1,1] * A[0,3] * A[2,2] -
              A[2,1] * A[0,2] * A[1,3] +
              A[2,1] * A[0,3] * A[1,2];

    U[1,3] = A[0,0] * A[1,2] * A[2,3] -
             A[0,0] * A[1,3] * A[2,2] -
             A[1,0] * A[0,2] * A[2,3] +
             A[1,0] * A[0,3] * A[2,2] +
             A[2,0] * A[0,2] * A[1,3] -
             A[2,0] * A[0,3] * A[1,2];

    U[2,3] = -A[0,0] * A[1,1] * A[2,3] +
               A[0,0] * A[1,3] * A[2,1] +
               A[1,0] * A[0,1] * A[2,3] -
               A[1,0] * A[0,3] * A[2,1] -
               A[2,0] * A[0,1] * A[1,3] +
               A[2,0] * A[0,3] * A[1,1];

    U[3,3] = A[0,0] * A[1,1] * A[2,2] -
              A[0,0] * A[1,2] * A[2,1] -
              A[1,0] * A[0,1] * A[2,2] +
              A[1,0] * A[0,2] * A[2,1] +
              A[2,0] * A[0,1] * A[1,2] -
              A[2,0] * A[0,2] * A[1,1];

    det = A[0,0] * U[0,0] + A[0,1] * U[1,0] + A[0,2] * U[2,0] + A[0,3] * U[3,0]
    if det == 0
        error "Matrix not invertible"
    endif
    U *= 1.0 / det;
endfunction

function U = __device__ cholesky(A:mat'unchecked(4,4))
    U = zeros(4,4)
    U[0,0] = sqrt(A[0,0])
    U[1,0] = A[1,0] / U[0,0]
    U[1,1] = sqrt(A[1,1] - U[1,0]^2)
    U[2,0] = A[2,0] / U[0,0]
    U[2,1] = (A[2,1] - U[2,0]*U[1,0]) / U[1,1]
    U[2,2] = sqrt(A[2,2] - (U[2,0]^2 + U[2,1]^2))
    U[3,0] = A[3,0] / U[0,0]
    U[3,1] = (A[3,1] - U[3,0]*U[1,0]) / U[1,1]
    U[3,2] = (A[3,2] - (U[3,0]*U[2,0] + U[3,1]*U[2,1])) / U[2,2]
    U[3,3] = sqrt(A[3,3] - (U[3,0]^2 + U[3,1]^2 + U[3,2]^2)) 
endfunction

function U = __device__ cholesky(A:mat'unchecked(3,3))
    U = zeros(3,3)
    U[0,0] = sqrt(A[0,0])
    U[1,0] = A[1,0] / U[0,0]
    U[1,1] = sqrt(A[1,1] - U[1,0]^2)
    U[2,0] = A[2,0] / U[0,0]
    U[2,1] = (A[2,1] - U[2,0]*U[1,0]) / U[1,1]
    U[2,2] = sqrt(A[2,2] - (U[2,0]^2 + U[2,1]^2))
endfunction

function U = __device__ cholesky(A:mat'unchecked(2,2))
    U = [[sqrt(A[0,0]), 0], _
         [A[1,0] / sqrt(A[0,0]), sqrt(A[1,1] - (A[1,0] / sqrt(A[0,0]))^2)]]
endfunction

function [] = main()
    print "det(zeros(3,3))=", det(zeros(3,3))

    A = [[2, 1, 1, 1], [1, 2, 1, 1], [1, 1, 2, 1], [1, 1, 1, 2]]
    print inv(A)
    print logdet(A), ",", log(det(A))

    %print det(A)
    [U,L,V] = svd(A)
    print U,",",L,",",V
    print "SVD: ", U*L*transpose(V)
    
    [U,L2] = symeig(A)
    print U,",",L2
    print "symeig: ", U*diagm(L2)*transpose(U)


    A = [[2, 1, 0, 0],
         [1, 2, 1, 0],
         [0, 1, 2, 1],
         [0, 0, 1, 2]]
    [U,L,V] = svd(A)
    print U,",",L,",",V
    B = [1,2,3,4]
    print det(A)
    print log(det(A)),",",logdet(A)


    print toeplitz(8,[1,-2,4,-2,1])

    A = randn(16,16)
    A = A * transpose(A)
    tic()
    for k=0..99
        X = pinv(A)
    endfor
    toc()
    print "det(X)=",det(A)
    print "logdet(X)=",logdet(A)

    % polynomial fitting test
    x = 0..255
    y = 40+2*(x/255).^2*255+randn(size(x))*20
    p = polyfit(x,y,3)
    print p
    plot(vertcat(y,polyval(p,x)))
    %print polyval(p,0)
endfunction
