
{!author name="Bart Goossens"}
{!doc category="Data structures"}

% Type: List[T]
%
% Generic list container. A list contains a variable number
% of elements. Adding an item to the list has complexity O(1),
% while lookup and removal has complexity O(n), with n the 
% number of items in the list (count).
% This list implementation is intended to be used mostly
% from host functions (e.g., adding, removing). Once filled,
% the data can be accessed from kernel/device functions.
%
type List[T] : mutable class
    items       : vec[T]    
    count       : int
    capacity    : int
end

% Function: List
% 
% Constructor for a list object
% 
% Usage:
%   : function [y : List[T]] = List[T](capacity : int = 8)
% 
% Parameters:
% capacity - the initial capacity of the list (the maximally expected number
% of elements that the list can contain. Note, if the capacity is exceeded,
% more space for the list is allocated)
% 
% Returns:
% y - a new list object
function y : List[T] = List[T](capacity : int = 8)
    y = List[T](items:=vec[T](capacity), capacity:=capacity, count:=0)
end

%function [] = __device__ vec_slice_copy(x : vec, y : vec, a : int, b : int)
%    for n=a..b-1
%        y[n] = x[n]
%    end
%end
%
%reduction (y : vec, x : vec, a : [int|scalar], b : [int|scalar]) -> _
%    (y[a..b] = x) = vec_slice_copy(x,y,a,b) where $ftype("__device__")


% Function: List.set_capacity
% 
% Sets the capacity for the list
% 
% Usage:
%   : function [] = List.set_capacity[T](self, count)
% 
% Parameters:
% self - the list object
% count - the new capacity (the maximally expected number
% of elements that the list can contain)
function [] = set_capacity[T](self : ^List[T], count)
    old = self.items
    self.capacity = count
    self.items = vec[T](self.capacity) % Attempt to use a variable as a function!
    self.items[0..numel(old)-1] = old
end

% Function: List.add
% 
% Adds a new element to the end of the list
% 
% Usage:
%   : function [] = List.add[T](self, item)
% 
% Parameters:
% self - the list object
% item - the item to add
function [] = add[T](self : ^List[T], item : T)
    if self.count >= self.capacity
        self.set_capacity(self.capacity * 2)
    endif
    index = (self.count+=1)
    %print "add: ", self.count
    self.items[index-1] = item
end

% Function: List.remove_at
% 
% Removes an element from the list at the specified position
% 
% Usage:
%   : function [] = List.remove_at[T](self, index : int)
% 
% Parameters:
% self - the list object
% index - the 0-based index of the item to remove
function [] = remove_at[T](self : ^List[T], index : int)
    {!serial for}
    for m=index..self.count-2
        self.items[m] = self.items[m+1]
    end
    self.count-=1
end

% Function: List.count_if
% 
% Counts the number of items in a list for which a given predicate
% is satisfied
% 
% Usage:
%   : function [cnt] = List.count_if[T](self, predicate : [__device__ T->int])
% 
% Parameters:
% self - the list object
% predicate - a predicate function returning true or false, depending on the input value
% 
% Returns:
% cnt - the returned count.
function cnt = count_if[T](self : ^List[T], predicate : [T -> int])
    cnt = 0
    for m=0..self.count-1
        cnt += predicate(self.items[m]) != 0 ? 1 : 0
    end
end

% Function: List.at
% 
% Returns the value of the element at position index
% 
% Usage:
%   : function [y] = List.at[T](self, index : int)
% 
% Parameters:
% self - the list object
% index - a 0-based index in the list
% 
% Returns:
% y - the element at position index
function y = at[T](self : ^List[T], index : int)
    y = self.items[index]
end

% Function: List.clear
% 
% Clears the content of this list
% 
% Usage:
%   : function [] = List.clear[T](self)
% 
% Parameters:
% self - the list object
function [] = clear[T](self : ^List[T])
    self.count = 0
end


% Function: List.to_string
% 
% Converts the list to a string representation
% 
% Usage:
%   : function [] = to_string()
function s = to_string[T](self : ^List[T])
    s = "["
    for i=0..self.count-1
        s = sprintf("%s %f", s, self.items[i])
    end
    s = strcat(s, "]")
end

% Function: List.find_first
% 
% Find the first occurence of the specified item in the list
% 
% Usage:
%   : function [] = List.find_first[T](item : T)
% 
% Parameters:
% item - the item to look for
function index = find_first[T](self : ^List[T], item : T)
    for m=0..self.count-1
        if self.items[m] == item
            index = m
            return
        endif
    end
    index = -1
end

% Function: List.find_last
% 
% Find the last occurence of the specified item in the list
% 
% Usage:
%   : function [] = List.find_last[T](item : T)
% 
% Parameters:
% item - the item to look for
function index = find_last[T](self : ^List[T], item : T)
    for m=self.count-1..-1..0
        if self.items[m] == item
            index = m
            return
        endif
    end
    index = -1
end

% Function: List.find_first_if
% 
% Find the first occurence for which the given predicate holds
% 
% Usage:
%   : function [] = List.find_first[T](item : T, pred : [T -> int])
% 
% Parameters:
% item - the item to look for
function index = find_first_if[T](self : ^List[T], pred : [T -> int])
    for m=0..self.count-1
        if pred(self.items[m])
            index = m
            return
        endif
    end
    index = -1
end

% Function: List.find_last_if
% 
% Find the last occurence of the specified item in the list
% 
% Usage:
%   : function [] = List.find_last[T](item : T)
% 
% Parameters:
% item - the item to look for
function index = find_last_if[T](self : ^List[T], pred : [T -> int])
    for m=self.count-1..-1..0
        if pred(self.items[m])
            index = m
            return
        endif
    end
    index = -1
end

% Reduction to get/set the n-th element of the list
% 
%reduction (T, self : List[T], index : int) -> self[index] = self.items[index]
%reduction (T, self : List[T], index : int, value : T) -> (self[index] = value) = self.items[index] = value

function [] = main()
    A : ^List[scalar] = List[scalar](10)
    A.add(1)
    A.add(2)
    A.add(3)
    print A
    %print $dbg(list[scalar])
    
    larger_than_1 = __device__ (i : scalar) -> int(i > 1)    
    print A.count_if(larger_than_1)    

    B : ^List[cscalar] = List[cscalar](10)
    print $dbg(B)
    B.add(1i)
    B.add(2i)
    B.add(3i)
    %print $dbg(B)
    print "B.count=",B.count
    B.remove_at(1)
    print "B.count=",B.count
    print B.at(B.count-1)    
    
    %C : ^list[vec] = create_list[vec](10)
    %C.add([1,2,3,4])
    %C.add([5,6,7])
    %print C
    %print A[2]
    
    D : ^List[uint8] = List[uint8](100)
    D.add(4)
    D.add(8)
    print D.at(0)
    print D.find_first(2), ",", D.find_first(8)
    print D.items[D.find_first_if(x -> int(x > 5))]
    %print D.items[D.find_first_if((x : uint8) -> (x > 5))]
    print D.to_string()
end