%REQUIRES_DISPLAY
{!author name="Martin Dimitrievski"}
{!doc category="Image Processing/Mathematical morphology"}

import "colortransform.q"
import "Quasar.UI.dll"

% euclidean distance helper functions
function out = __device__ EDT_f(x : scalar, i : scalar, g_i : scalar)
    out = (x-i)*(x-i) + g_i*g_i
end

function out = __device__ EDT_Sep(i : scalar, u : scalar, g_i : scalar, g_u : scalar)
    out = floor((u*u - i*i + g_u*g_u - g_i*g_i)/(2*(u-i)))
end

% manhattan distance helper functions
function out = MDT_f(x : scalar, i : scalar, g_i : scalar)
    out = abs(x-i) + g_i
end

function out = MDT_Sep(i : scalar, u : scalar, g_i : scalar, g_u : scalar)
    if (g_u >= (g_i + u - i))
        out = 1/0
    elseif (g_i > (g_u + u - i))
        out = -1/0
    else
        out = floor((g_u - g_i + u + i)/2)
    endif
end

% chessboard distance helper functions
function out = CDT_f(x : scalar, i : scalar, g_i : scalar)
    out = max(abs(x-i), g_i)
end

function out = CDT_Sep(i : scalar, u : scalar, g_i : scalar, g_u : scalar)
    if (g_i <= g_u)
        out = max(i+g_u, floor((i+u)/2))
    else
        out = min(u-g_i, floor((i+u)/2))
    endif
end

% Function: distance_transform_meijster
%
% Computes the distance transform of a binary image B, using 
% the algorithm by
%
% Meijster, Arnold, Jos BTM Roerdink, and Wim H. Hesselink. 
% "A general algorithm for computing distance transforms in 
% linear time." Mathematical Morphology and its applications 
% to image and signal processing. Springer, Boston, MA, 2002. %
% p331-340.
%
%
% Parameters:
% B - input binary image
% out - the resulting distance map 
%
% :function out = distance_transform_meijster(B : mat)

function out = distance_transform_meijster(B : mat)
    
    %b = reshape(B,[numel(B),1])
    b = transpose(B[:])
    
    [n,m] = size(B)
    g = zeros(size(b))
    
    % First phase
    {!parallel for}
    for x=0..m-1
        if b[x+0*m] == 1
            g[x+0*m] = 0
        else
            g[x+0*m] = m+n
        endif
        % Scan 1
        for y=1..n-1
            if b[x+y*m] == 1
                g[x+y*m] = 0
            else
                g[x+y*m] = 1 + g[x+(y-1)*m]
            endif
        end %endfor scan1
        % Scan 2
        %        #pragma force_serial
        for y=(n-2)..-1..0
            if g[x+(y+1)*m] < g[x+y*m]
                g[x+y*m] = 1 + g[x+(y+1)*m]
            endif
        end %endfor scan2
    end
    
    % Second phase
    dt = zeros(size(b))
    s = zeros(m,n)
    t = zeros(m,n)
    q = 0
    w = 0
        
    {!parallel for}
    for y=0..n-1 
        q = 0;
        s[0,y] = 0.0;
        t[0,y] = 0.0;
        
        % Scan 3
        for u=1..m-1
            while q >= 0 && (EDT_f(t[q,y], s[q,y], g[s[q,y]+y*m]) > EDT_f(t[q,y], u, g[u+y*m]))
                q=q-1
                if q<0
                    break
                endif
            end
            if (q < 0)
                q = 0
                s[0,y] = u
            else
                w = 1 + EDT_Sep(s[q,y], u, g[s[q,y]+y*m], g[u+y*m])
                if (w < m)
                    q=q+1
                    s[q,y] = u
                    t[q,y] = w
                endif
            endif
        end %endfor scan3
        
        % Scan 4
        for uu=m-1..-1..1
            d = EDT_f(uu, s[q,y], g[s[q,y]+y*m])
            % if the distance metric is euclidean distance
            d = floor(sqrt(d))
            %endif euclidean distance
            dt[uu+y*m] = d
            if (uu == t[q,y])
                q=q-1
            endif
        end %endfor scan4
        
    end %endfor Second phase
    
    
    out = reshape(dt,[n,m])
    
end

function [] = main()
    
    img = imread("lena_big.tif")
    img = img[:,:,1]
    img = img>128
    
    out = distance_transform_meijster(img)
    
    out[out>50]=50
    
    frm = form("Quasar Meijster transform demo")
    frm.width = 800
    disp1 = frm.add_display()
    disp2 = frm.add_display()
    frm.show()
    g=disp1.imshow(img*255)
    h=disp2.imshow(out*255/50)
    g.connect(h)
    
end
