% Generic interface to multiresolution transforms
import "system.q"
import "dwt.q"
import "dtcwt.q"
import "dst.q"
import "udwt.q"
import "steerable_pyramids.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Multiresolution"}

% Section: multirestransforms.q
% This library contains several multiresolution transforms, and allows
% image processing algorithms to use the same interface to each of these
% transforms. This has the benefit that transforms can easily be 
% changed/combined. The technique is always the same: first "build" your 
% transform, as follows
% : [W, W_T] = build_XXXX(parameters)
% where
% W - the forward transformation function
% W_T - the backward transformation function
% For orthogonal transforms, we have that W * W_T = I or, equivalently that 
% W = transpose(W_T).
%

% Function: lincell
% Linearize a cell array, i.e. convert nested cell arrays
% to one 1D cell array.
%
% : function y = lincell(x)
%
% Notes:
% this function only involves pointer copies, which are very fast
function y = lincell(x)
    y = {}
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            if iscell(x[m,n])
                y = cat(y, lincell(x[m,n]))
            else
                xcell = cell(1)
                xcell[0] = x[m,n]
                y = cat(y, xcell)
            endif
        end
    end
end

% Function: build_dwt
% Constructs a 2D decimated wavelet transform
%
% : function [W, W_T] = build_dwt(w, J)
%
% Parameters:
% w - the wavelet filters to be used for the construction (e.g. db2,sym8,...)
% J - the number of scales
% W - the forward wavelet transform
% W_T - the inverse wavelet transform
function [W, W_T] = build_dwt(w, J)
    wc = orthogonal_wavelet(w)
    W = x -> dwt2d(x, wc, J)
    W_T = x -> idwt2d(x, wc, J)
end

% Function: build_udwt
% 
% Constructs a 2D undecimated wavelet transform
% 
% Usage:
%   : function [W : [??->vec[??]], W_T : [??->??]] = build_udwt(w, J)
% 
% Parameters:
% w - the wavelet filters to be used for the construction (e.g., db2, sym4, ...)
% J - the number of scales
% 
% Returns:
% W - the forward wavelet transform
% W_T - the inverse wavelet transform
function [W, W_T] = build_udwt(w, J)
    wc = orthogonal_wavelet(w)
    W = x -> udwt2d(x, wc, J)
    W_T = x -> iudwt2d(x, wc, J)
end

% Function: build_dtcwt
% Construct a 2D DT-CWT transform
%
% : function [W, W_T] = build_dtcwt(w1, w2, J)
%
% Parameters:
% w1 - the wavelet filters to be used for the first scale of the DT-CWT
% w2 - the wavelet filters to be used for the other scales
% J - the number of scales
% W - the forward complex wavelet transform
% W_T - the inverse complex wavelet transform
function [W, W_T] = build_dtcwt(cw1, cw2, J)
    w1 = complex_orthogonal_wavelet(cw1)
    w2 = complex_orthogonal_wavelet(cw2)
    W = x -> dtcwt2d(x, w1, w2, J)
    W_T = x -> idtcwt2d(x, w1, w2, J)
end

% Function: build_stp
% Builds a steerable pyramid transform
%
% : function [S, S_H] = build_stp(sz, K, J)
%
% Parameters:
% sz - size of the images that have to be processed
% K - the number of orientations of the transform
% J - the number of scales of the transform
function [S, S_H] = build_stp(sz, K, J)
    [r, theta] = polargrid(sz)
    G = steerable_filters(fftshift2(theta),sz,K)
    H = stp_meyer_radial_wavelet_filters(r,sz,J)    
    S = x -> steerable_pyramid_analysis(x,J,K,H,G)
    S_H = x -> steerable_pyramid_synthesis(x,J,K,H,G)
end

% Function: build_dst2d
% Build a 2D shearlet transform
%
% : function [S, S_H] = build_dst2d(sz, K, J)
%
% Parameters:
% sz - size of the images that have to be processed
% K - the number of orientations of the transform
% J - the number of scales of the transform
% S - the forward shearlet transform
% S_H - the adjoint shearlet transform
%
% Notes:
% In this implementation, the number of orientations is the same
% for all scales - so a scalar number needs to be specified for K.
function [S, S_H] = build_dst2d(sz, K, J)
    bw = pi/2
    % Precomputation of the shearlet filters...
    G = wedge_filter(sz,K, bw)
    H = meyer_radial_wavelet_filters(sz,J)
    S = x -> shearlet_analysis(x,J,K,H,G,1)
    S_H = x -> shearlet_synthesis(x,J,K,H,G,1)
end

% Function: build_energyweighted_dst2d
% Builds an energy-weighted (non-Parseval) 2D shearlet transform
% (in which the variance of white noise is preserved in the subbands)
%
% : function [S, S_H] = build_energyweighted_dst2d(sz, K, J)
%
% Parameters:
% sz - size of the images that have to be processed
% K - the number of orientations of the transform
% J - the number of scales of the transform
% S - the forward shearlet transform
% S_H - the backward shearlet transform
%
% Notes: 
% S_back is actually not the adjoint of S; this is due to the 
% normalization is being done.
function [S, S_back] = build_energyweighted_dst2d(sz, K, J)
    bw = pi/2
    G = wedge_filter(sz,K, bw)
    H = meyer_radial_wavelet_filters(sz,J)
    % subband energies
    e = shearlet_compute_energy_per_subband(J,K,H,G)
    print e
    S = x -> shearlet_analysis(x,J,K,H,G,1) ./ e
    S_back = x -> shearlet_synthesis(x .* e,J,K,H,G,1)
end

function [] = main()
    %tic()
    %for k=0..13000
    %c = lincell({{1,2},{3,{4}}})
    %end
    %toc()

    %print type(ones(4,4),"cell")
    %print iscell(ones(4,4))
    %print iscell(cell(1,4))
    %return

    %print cat({1,2},{3,4})
    %print lincell({{1,2},{3,{ones(4,4)}}})
    print lincell({[1,2],{3,4}})

    %im=imread("lena_big.tif")
    %[S, S_H] = build_dst2d(size(im), 4, 3)
    %y = S_H(S(im))

    %imshow([0,0])
end
