%REQUIRES_DISPLAY
%=================================================
% optical_flow_farneback.q
% Implementation of "Two-frame motion estimation based on polynomial expansion" (Farnebäck et al., 2003)
%   Ported from OpenCV 2.3.1 (C++), then parallellized.
%
% Simon Donné, April 2014
%=================================================

import "optical_flow_visualization.q"

import "imresize.q"
import "colortransform.q"
import "Quasar.Video.dll"
import "Quasar.UI.dll"
import "inttypes.q"

{!author name="Simon Donné"}
{!doc category="Optical Flow"}

function [ig11:scalar, ig03:scalar, ig33:scalar, ig55:scalar]  = __device__ optical_flow_farneback_prepare_gaussian(n:int, sigma:scalar, g:vec, xg:vec, xxg:vec)
    g[:] = exp(-(-n..n).^2/2/sigma^2)
    s = 1./sum(g)
    g[:] = g[:]*s
    xg[:] = (-n..n).*g[:]
    xxg[:] = (-n..n).^2.*g[:]
    G = zeros(6,6)
    
    for y = -n..n
    for x = -n..n
        G[0,0] += g[y+n]*g[x+n]
        G[1,1] += g[y+n]*g[x+n]*x*x
        G[3,3] += g[y+n]*g[x+n]*x*x*x*x
        G[5,5] += g[y+n]*g[x+n]*x*x*y*y
    end
    end
    
    G[2,2] = G[0,3] = G[0,4] = G[3,0] = G[4,0] = G[1,1]
    G[4,4] = G[3,3]
    G[3,4] = G[4,3] = G[5,5]
    
    %ideally, this call:
%    invG = inv(G)
    %sadly, no static CUDA linking in 4.2
    
    A = zeros(6, 12)

    % A serial function
    function [] = __device__ serial_inverse (A : mat'unchecked, x : mat'unchecked, rows : int, cols : int)
        row = 0
        temp = 0.0
        for i=0..rows-1
            for j=0..cols-1
                A[i,j] = x[i,j]
                A[i,j+cols] = (i == j) ? 1 : 0
            end
        end

        for j=0..cols-1
            if A[j,j]==0
                % Search non-zero element in column
                row = j
                for i=j..rows-1
                    if abs(A[i,j])>0
                        row = i
                    endif
                end
                if row == j                    
                    error "(inv) There is no existing inverse matrix!"
                else
                    % Swap row i and j
                    for k=0..size(A,1)-1
                        temp = A[row,k]
                        A[row,k] = A[j,k]
                        A[j,k] = temp
                    end
                endif
            endif

            % Row operations
            for i=0..rows-1
                temp = A[i,j]
                if temp != 0.0 && i != j
                    for k=0..size(A,1)-1
                        A[i,k] = A[i,k]*A[j,j] - temp*A[j,k]
                    end
                endif
            end

            % Rescale all rows
            for i=0..rows-1
                temp = A[i,i]
                for j2=0..size(A,1)-1
                    A[i,j2] = A[i,j2] / temp
                end
            end
        end
    end
    serial_inverse(A, G, 6, 6)

    invG = A[:, 6..11]            

    ig11 = invG[1,1]
    ig03 = invG[0,3]
    ig33 = invG[3,3]
    ig55 = invG[5,5]
end

function [] = optical_flow_farneback_poly_exp_parallel(image:mat, n:int, sigma:scalar, R:cube)
    [M,N] = size(image, 0..1)
    
    g   = uninit(2*n+1)
    xg  = uninit(2*n+1)
    xxg = uninit(2*n+1)
    [ig11, ig03, ig33, ig55] = optical_flow_farneback_prepare_gaussian(n, sigma, g, xg, xxg)
    
    function [] = __kernel__ getprecalcs(out : cube'unchecked, in : mat'clamped, mask0:vec, mask1:vec, mask2:vec, center : int, pos : ivec2)
        sums:vec3 = [0.0,0.0,0.0]
        for m=0..numel(mask0)-1
            sums[0] = sums[0] + mask0[m] * in[pos + [m-center,0]]
            sums[1] = sums[1] + mask1[m] * in[pos + [m-center,0]]
            sums[2] = sums[2] + mask2[m] * in[pos + [m-center,0]]
        end
        out[pos[0],pos[1], 0] = sums[0]
        out[pos[0],pos[1], 1] = sums[1]
        out[pos[0],pos[1], 2] = sums[2]
    end
    
    rows = uninit(M,N,3)
    parallel_do(size(image), rows, image, g, xg, xxg, n, getprecalcs)
    
    function [] = __kernel__ actualcalcs(out:cube'unchecked, rowsk:cube'clamped, pos:ivec2)
        b1 = rowsk[pos[0],pos[1],0]*g[n]
        b2 = 0.0
        b3 = rowsk[pos[0],pos[1],1]*g[n]
        b4 = 0.0
        b5 = rowsk[pos[0],pos[1],2]*g[n]
        b6 = 0.0
        for k = 1..n
            b1 = b1 + (rowsk[pos[0],pos[1]+k,0] + rowsk[pos[0],pos[1]-k,0])*g[n+k]
            b4 = b4 + (rowsk[pos[0],pos[1]+k,0] + rowsk[pos[0],pos[1]-k,0])*xxg[n+k]
            b2 = b2 + (rowsk[pos[0],pos[1]+k,0] - rowsk[pos[0],pos[1]-k,0])*xg[n+k]
            b3 = b3 + (rowsk[pos[0],pos[1]+k,1] + rowsk[pos[0],pos[1]-k,1])*g[n+k]
            b6 = b6 + (rowsk[pos[0],pos[1]+k,1] - rowsk[pos[0],pos[1]-k,1])*xg[n+k]
            b5 = b5 + (rowsk[pos[0],pos[1]+k,2] + rowsk[pos[0],pos[1]-k,2])*g[n+k]
        end
        
        out[pos[0],pos[1],1] = b2*ig11
        out[pos[0],pos[1],0] = (b3*ig11)
        out[pos[0],pos[1],3] = (b1*ig03 + b4*ig33)
        out[pos[0],pos[1],2] = (b1*ig03 + b5*ig33)
        out[pos[0],pos[1],4] = (b6*ig55)
    end
    parallel_do(size(image), R, rows, actualcalcs)
end

function [] = optical_flow_farneback_update_matrices_parallel(R_prev:cube, R_next:cube, flow:cube, matrix:cube)
    border_size = 5
    border = [0.14, 0.14, 0.4472, 0.4472, 0.4472]
    [M,N] = size(flow, 0..1)
    
    function [] = __kernel__ fillMatrix(matrixk:cube, pos:ivec2)
        y = pos[0]
        x = pos[1]
        dx = flow[y,x,0]        
        dy = flow[y,x,1]
        fx = x + dx
        fy = y + dy
        
        x1 = int(max(floor(fx), 0))
        y1 = int(max(floor(fy), 0))
        
        fx = fx - x1
        fy = fy - y1
        
        a00 = (1-fx)*(1-fy)
        a01 = fx*(1-fy)
        a10 = (1-fx)*fy
        a11 = fx*fy
        
        r2 = a00*R_next[y1,x1,0] + a01*R_next[y1,x1+1,0] + a10*R_next[y1+1,x1,0] + a11*R_next[y1+1,x1+1,0]
        r3 = a00*R_next[y1,x1,1] + a01*R_next[y1,x1+1,1] + a10*R_next[y1+1,x1,1] + a11*R_next[y1+1,x1+1,1]
        r4 = a00*R_next[y1,x1,2] + a01*R_next[y1,x1+1,2] + a10*R_next[y1+1,x1,2] + a11*R_next[y1+1,x1+1,2]
        r5 = a00*R_next[y1,x1,3] + a01*R_next[y1,x1+1,3] + a10*R_next[y1+1,x1,3] + a11*R_next[y1+1,x1+1,3]
        r6 = a00*R_next[y1,x1,4] + a01*R_next[y1,x1+1,4] + a10*R_next[y1+1,x1,4] + a11*R_next[y1+1,x1+1,4]

        r2 = (R_prev[y,x,0] - r2)/2
        r3 = (R_prev[y,x,1] - r3)/2
        r4 = (R_prev[y,x,2] + r4)/2
        r5 = (R_prev[y,x,3] + r5)/2
        r6 = (R_prev[y,x,4] + r6)/4
        
        r2 += r4*dy + r6*dx
        r3 += r6*dy + r5*dx
        
        scale = (x < border_size ? border[x] : 1)*_
                    (x >= N - border_size ? border[N - x - 1] : 1)*_
                    (y < border_size ? border[y] : 1)*_
                    (y >= M - border_size ? border[M - y - 1] : 1)

                r2 *= scale; r3 *= scale; r4 *= scale;
                r5 *= scale; r6 *= scale;
                
        matrixk[y,x,0] = r4*r4 + r6*r6
        matrixk[y,x,1] = (r4 + r5)*r6
        matrixk[y,x,2] = r5*r5 + r6*r6
        matrixk[y,x,3] = r4*r2 + r6*r3
        matrixk[y,x,4] = r6*r2 + r5*r3
    end
    parallel_do([M,N], matrix, fillMatrix)
end

function [] = optical_flow_farneback_update_flow_parallel(flow:cube, matrices:cube, block_size:int)
    [M,N] = size(flow, 0..1)
    
    matrices_tmp = zeros(size(matrices))
    %note: destructive a.k.a. in-place!
    function [] = __kernel__ intimage_part1(data:cube, pos:ivec3)
        matrices_tmp[pos] = matrices_tmp[pos[0], pos[1]-1, pos[2]] + data[pos]
    end
    parallel_do([M,N,5], matrices, intimage_part1)
    function [] = __kernel__ intimage_part2(data:cube, pos:ivec3)
        data[pos] = data[pos[0]-1, pos[1], pos[2]] + matrices_tmp[pos]
    end
    parallel_do([M,N,5], matrices, intimage_part2)

    function [] = __kernel__ update_flow_smart(flowk:cube, intimagek:cube'clamped, block_size:int, pos:ivec2)
        sums : vec5 = [0.0,0.0,0.0,0.0,0.0]
        m = int(block_size/2)
        %what fraction of the window is actually seen
        y_before = max(m - pos[0],0)
        x_before = max(m - pos[1],0)
        y_after  = max(pos[0] + m - M + 1,0)
        x_after  = max(pos[1] + m - N + 1,0)
        
        seen = ((block_size-y_before-y_after)*(block_size-x_before-x_after)) / block_size^2
        iseen = seen==0 ? 0.0 : 1/seen
        
        for k = 0..4
            sums[k] = (intimagek[pos[0]+m, pos[1]+m, k] - intimagek[pos[0]+m, pos[1]-m-1, k] - intimagek[pos[0]-m-1, pos[1]+m, k] + intimagek[pos[0]-m-1, pos[1]-m-1, k]) * iseen
        end 
                    
        det = (sums[0]*sums[2] - sums[1]*sums[1]+1e-3*block_size^4)
        idet = det==0 ? 0.0 : 1/det
        f1 = (sums[0]*sums[4]-sums[1]*sums[3])*idet
        f2 = (sums[2]*sums[3]-sums[1]*sums[4])*idet
        
        flowk[pos[0],pos[1],0] = isnan(f1) || !isfinite(f1) ? 0.0 : f1
        flowk[pos[0],pos[1],1] = isnan(f2) || !isfinite(f1) ? 0.0 : f2
    end
    parallel_do([M,N], flow, matrices, block_size, update_flow_smart)
    assert(sum(isnan(flow))==0)
end

% Function: optical_flow_farneback_parallel
%
% :function [flow:cube] = optical_flow_farneback_parallel(prev_frame:mat, next_frame:mat, pyramid_factor:scalar, window_size:int, its:int, poly_n:int, poly_sigma:scalar)
%
% Implementation of "Two-frame motion estimation based on polynomial expansion" (Farnebäck et al., 2003)
%
% Parameters:
%   - prev_frame, next_frame: the input images
%   - pyramid_factor: scaling factor for the pyramidization
%   - window_size: window size for the optical flow calculation. Lower size mean better edge preservation, but more noisy estimates.
%   - its: The amount of iterations at each level of the pyramid.
%   - poly_n, poly_sigma: parameters for the gaussian blur when estimating the polynomial coefficients.
%
function [flow:cube] = optical_flow_farneback_parallel(prev_frame:mat, next_frame:mat, pyramid_factor:scalar, window_size:int, its:int, poly_n:int, poly_sigma:scalar)
    [M0,N0] = size(prev_frame, 0..1)
    %get the number of pyramid levels, lowest level should have a size larger than 32x32
    min_size = 32
    pyramid_levels = min(_
    ceil(log(min_size/M0)/log(pyramid_factor)), _
    ceil(log(min_size/N0)/log(pyramid_factor)))
    
    for level = (pyramid_levels-1)..-1..0
        scale = pyramid_factor^level
        sigma = (1/scale - 1)*0.5
        smooth_sz = max(round(sigma*5),3)
        
        M = round(M0*scale)
        N = round(N0*scale)
        
        if level == pyramid_levels-1
            flow = zeros(M, N, 2)
        else
            flow = imresize(flow, [M, N])/pyramid_factor
        endif
        
        if(sigma != 0)
            fimg:mat = gaussian_filter(prev_frame,sigma,smooth_sz, boundary_mode = "mirror")
        else
            fimg:mat = prev_frame
        endif
        I = imresize(fimg, [M,N], "cubic")
        R_prev = uninit(M,N,5)
        optical_flow_farneback_poly_exp_parallel(I, poly_n, poly_sigma, R_prev)
        if(sigma != 0)
            fimg:mat  = gaussian_filter(next_frame,sigma,smooth_sz, boundary_mode = "mirror")
        else
            fimg:mat = prev_frame
        endif
        I = imresize(fimg, [M,N], "cubic")
        R_next = uninit(M,N,5)
        optical_flow_farneback_poly_exp_parallel(I, poly_n, poly_sigma, R_next)
        
        border_size = 5
        
        matrices = uninit(M,N,border_size)
        
        for i = 1..its
            optical_flow_farneback_update_matrices_parallel(R_prev, R_next, flow, matrices)
            optical_flow_farneback_update_flow_parallel(flow, matrices, window_size)
        end        
    end
end

function [] = main()
    function [] = example_single_image_pair()     
        function Iw:mat = linear_warp(I:mat, u0:mat, v0:mat)
            Iw = uninit(size(I))
            function [] = __kernel__ linear_warp_kernel(out:mat, in:mat'safe, offsetx:mat, offsety:mat, pos:ivec2)
                xidx = floor(pos[1]+offsetx[pos])
                xfrac = frac(pos[1]+offsetx[pos])
                yidx = floor(pos[0]+offsety[pos])
                yfrac = frac(pos[0]+offsety[pos])
                out[pos]=xfrac*yfrac*in[yidx+1, xidx+1]+(1-xfrac)*(1-yfrac)*in[yidx, xidx]+(1-xfrac)*yfrac*in[yidx+1,xidx] + xfrac*(1-yfrac)*in[yidx, xidx+1]
            end
            parallel_do(size(Iw), Iw, I, u0, v0, linear_warp_kernel)
        end
        pyramid_factor = 1/2
        window_size = 15
        its = 10
        poly_n = 5
        poly_sigma = 1.5
        
        ILg = imread("/scratch/sdonn/OF/grayscale/Walking/frame08.png")[:,:,0]
        IRg = imread("/scratch/sdonn/OF/grayscale/Walking/frame11.png")[:,:,0]
        
        tic()
        for i = 1..10
        flow_p = optical_flow_farneback_parallel(ILg, IRg, pyramid_factor, window_size, its, poly_n, poly_sigma)
        end
        toc("Parallel implementation x10")
        
        imshow(show_flow(-flow_p[:,:,1], -flow_p[:,:,0], 20))
        imshow(linear_warp(IRg,flow_p[:,:,0], flow_p[:,:,1]))
        show_flow_demo([80,80])
    end
    function [] = example_video_processing() 
        pyramid_factor = 1/4
        window_size = 15
        its = 3
        poly_n = 5
        poly_sigma = 0.7
        
        %read in an entire video sequence
        stream = vidopen("flowers.avi")
        
        frames = vec[mat](floor(stream.duration_sec * stream.avg_frame_rate))

        tic()
        if (vidreadframe(stream))
            i=0
            new_frame = rgb2gray(float(stream.rgb_data))
            
            frm = form()
            dsp = frm.add_display()
			dsp.sync_framerate(25)
        
            show_flow_demo([80,80])
            
            while vidreadframe(stream) && !frm.closed()
                old_frame = new_frame
                new_frame = rgb2gray(float(stream.rgb_data))
                flow_p = optical_flow_farneback_parallel(new_frame, old_frame, pyramid_factor, window_size, its, poly_n, poly_sigma)
                dsp.imshow(show_flow(flow_p[:,:,1], flow_p[:,:,0], 4))    
                i+=1
            end
        else
            error "There was a problem reading this video (apparently there are no frames)"
        endif
        time = toc()        
        
        printf("Average FPS: %f", 1000*(i-1)/time)
    end
    
    %example_single_image_pair()
    example_video_processing()
end
