%=================================================
% optical_flow_visualization.q
% Visualization methods for optical flow
% Simon Donné, April 2014
%=================================================

import "colortransform.q"

{!author name="Simon Donné"}
{!doc category="Optical Flow"}

% Function: show_flow
%
% :function [img:cube] = show_flow(dx:mat, dy:mat, scale:scalar)
%
% Visualizes optical flow fields where the hue represents the angle and saturation represents the magnitude.
%
% Parameters:
%   - dx, dy: optical flow estimates.
%   - scale: magnitudes are clamped to this value, and saturation is linear between 0 and scale.
%
function [img:cube] = show_flow(dx:mat, dy:mat, scale:scalar) 
    angles=atan2(dy,dx)
    magnitudes=sqrt(dx.^2 + dy.^2)
    angles[angles<0] = angles + 2*pi
    climg = zeros(size(dx,0), size(dx,1), 3)
    climg[:,:,0] = angles/2/pi
    climg[:,:,1] = min(magnitudes,scale) / scale
    climg[:,:,2] = min(magnitudes,scale) /scale
    img = hsv2rgb(climg)
end

% Function: show_flow_demo
%
% :function [img:cube] = show_flow_demo(sz:ivec2)
%
% Demonstration image for optical flow visualization, given a size. Negative estimates indicate a movement to the left and up
%
function [] = show_flow_demo(sz:ivec2)
    u = ones(2*sz[1]+1, 1)*(-sz[0]..sz[0])
    v = transpose(-sz[1]..sz[1])*ones(1,2*sz[0]+1)
    imshow(show_flow(u,v,max(sz)))
end

function [] = main()
    show_flow_demo([80,80])
end