%==============================================================================
% Parallel building blocks for Quasar
% (C) 2014 Bart Goossens
% Ghent University, Dept. TELIN-IPI-iMinds
%==============================================================================
% In this library, a number of highly specialized parallel algorithms are
% implemented. These algorithms are typically easy to implement in serial code,
% however often very complex in a parallel algorithm. To take advantage of 
% the computing power of both CPU and GPU, most of the algorithms are 
% implemented here both in serial/in parallel. The run-time scheduler may pick
% the best version of the algorithm depending on the available resources.
%==============================================================================
import "system.q"

{!author name="Bart Goossens"}
{!doc category="General"}

% New algorithms:
% find_index_max - finds the index of the FIRST maximal value of the given cube.
% find_index_max_per_row - The index of the maximum for every row individually
% find_indices_nonzero - find all non-zero indices of a given vector
% find_indices_nonzero_per_row - gives access to the individual rows
% select_threshold - Selects all values that are larger than a given threshold
% distinct - Selects all distinct values (note: performs a sorting operation)
% distinct_map - Returns a vector with all distinct values and the mapping 
%     onto this vector
% apply_map - Applies a non-linear mapping to a matrix (similar to interp1)
% cummax - Cumulative maximum
% replace_mask - Replaces values of the data cube by the specified constant, 
%     according to the given mask, i.e. implementing logical indexing.
% replace_by_index - Replaces values of the data cube by a certain value,
%     but only at the specified indices.

% Function: delinearize_index
% 
% De-linearizes an index (converts from a 1D index to a higher-
% dimensional index) given the size of the matrix. Note: only up to three
% dimensions are currently supported
% 
% Usage:
%   : function [y] = delinearize_index(ind : [int|vec], sz : vec)
% 
% Parameters:
% ind - the index (or indices) to convert
% sz - the size of the input data
% 
% Returns:
% y - an n-D coordinate
function y = delinearize_index(ind : [int|vec], sz : vec)
    dims = sum(sz > 1) % TODO: really need a built-in function!
    
    if isscalar(ind)  
        y = ind2pos(sz, ind)
    else
        match dims with
        | 1 ->
            y = ind
        | 2 ->
            y = uninit(numel(ind), 2)
            parallel_do(numel(ind), ind, y, sz, __kernel__ (x : vec, y : mat, sz : ivec2, pos : int) ->
                y[pos,0..1] = ind2pos(sz, x[pos]))
        | 3 ->
            y = uninit(numel(ind), 3)
            parallel_do(numel(ind), ind, y, sz, __kernel__ (x : vec, y : mat, sz : ivec3, pos : int) ->
                y[pos,0..2] = ind2pos(sz, x[pos]))
        | 4 ->
            y = uninit(numel(ind), 4)
            parallel_do(numel(ind), ind, y, sz, __kernel__ (x : vec, y : mat, sz : ivec4, pos : int) ->
                y[pos,0..3] = ind2pos(sz, x[pos]))
        | 5 ->
            y = uninit(numel(ind), 5)
            parallel_do(numel(ind), ind, y, sz, __kernel__ (x : vec, y : mat, sz : ivec5, pos : int) ->
                y[pos,0..4] = ind2pos(sz, x[pos]))
        | _ ->
            error "delinearize_index: currently only up to 5 dimensions supported!"
        endmatch
    endif
endfunction

% Function: apply_fn
% 
% Applies a mapping function onto the input data.
% 
% Usage:
%   : function [y] = apply_map(x : cube, fn : [__device__ scalar -> scalar])
% 
% Parameters:
% x - the input image
% map - the mapping function to be applied
% 
% Returns:
% y - the mapped output image
function y = apply_fn(x : cube, fn : [__device__ scalar -> scalar])
    y = uninit(size(x))
    parallel_do(size(y),x,y,fn,__kernel__ (pos : ivec3, x : cube'unchecked, y : cube, _
        fn : [__device__ scalar -> scalar]) -> y[pos] = fn(x[pos]))
endfunction

% Function: apply_map
% 
% Applies a mapping function onto the input data.
% 
% Usage:
%   : function [y] = apply_map(x : cube, map : vec)
% 
% Parameters:
% x - the input image
% map - the mapping function to be applied
% 
% Returns:
% y - the mapped output image
function y = apply_map(x : cube, map : vec)
    y = apply_fn(x, __device__ (x : scalar) -> map[x])
endfunction

% Function: find_index_max
% 
% finds the index of the FIRST occuring maximal value of the given cube.
% The cube is traversed in zigzag order (as in raster scanning). 
% 
% Usage:
%   : function [y : int] = find_index_max(x : cube{n})
%   : function [y : int] = __device__ find_index_max[T](x : vec[T])
%   : function [y : int] = __device__ find_index_max[T](x : mat[T])
%   : function [y : int] = __device__ find_index_max[T](x : cube[T])
% 
% Parameters:
% x - the input matrix
% 
% Returns:
% y - the index of the first maximal value in x (as an integer). 
% 
% Notes:
% In order to obtain a coordinate, you need to call the function <ind2pos>
% on the result.
%
% See also:
% <find_index_min>
function y = find_index_max(x : cube{n})
    function [] = __kernel__ parallel_kernel(x : vec'unchecked, sz : int, xv : vec, 
        yv : vec, blkdim : int, blkpos : int, blkidx : int, pos : int)
        
        {!kernel target="gpu"}        
        bins : vec'unchecked = shared(blkdim)
        indices : vec[int]'unchecked = shared[int](blkdim)

        ind = pos
        val = x[ind]
        nblocks = int((numel(x) + sz-1) / sz)

        for n=1..nblocks-1
            i=pos+n*sz
            if i<numel(x) && val < x[i]
                val = x[i]
                ind = i
            endif
        endfor
        indices[blkpos] = ind
        bins[blkpos] = val
        syncthreads
        bit = 1
        while bit<blkdim
            if and(blkpos,2*bit-1) == 0 && blkpos+bit<blkdim
                [t1,t2] = [bins[blkpos], bins[blkpos+bit]]
                if t1<t2 || (t1 == t2 && indices[blkpos] > indices[blkpos+bit])
                    bins[blkpos] = t2
                    indices[blkpos] = indices[blkpos+bit]
                endif
            endif
            syncthreads
            bit *= 2
        endwhile
        if blkpos == 0      
            xv[blkidx] = indices[0]
            yv[blkidx] = bins[0]
        endif
    endfunction
    
    function y = __kernel__ serial_kernel(x : vec)
        {!kernel target="cpu"}
        y = 0
        maxval = x[0]
        for i=1..numel(x)-1
            if maxval<x[i]
                maxval=x[i]
                y = i
            endif
        endfor
    endfunction
    
    N = numel(x)
    if N==1
        y = 0
        return
    endif
    match schedule(N, x, {parallel_kernel, serial_kernel}) with
    | 0 ->
        [block_cnt, block_size] = opt_block_cnt(parallel_kernel, 0, N)
        xv = zeros(prod(block_cnt))
        {!alloc mode="gpu";type="pinnedhostmem"}
        yv = uninit(prod(block_cnt))
        sz = prod(block_size .* block_cnt)
        parallel_do([block_size.*block_cnt, block_size], x, sz, xv, yv, parallel_kernel)
        % We still need to find the minimum for each block. This is achieved using the
        % serial kernel. To avoid a separate cuMemcpyD2H operation, we use pinned host 
        % memory and write from the previous kernel directly to the system memory.
        y = xv[serial_do(1, yv, serial_kernel)]
    | 1 ->
        y = serial_do(1, x, serial_kernel)
    endmatch
endfunction

% Function: find_index_min
% 
% finds the index of the FIRST occuring minimal value of the given cube.
% The cube is traversed in zigzag order (as in raster scanning). 
% 
% Usage:
%   : function [y : int] = find_index_min(x : cube{n})
%   : function [y : int] = __device__ find_index_min[T](x : vec[T])
%   : function [y : int] = __device__ find_index_min[T](x : mat[T])
%   : function [y : int] = __device__ find_index_min[T](x : cube[T])
% 
% Parameters:
% x - the input matrix
% 
% Returns:
% y - the index of the first minimal value in x (as an integer). 
% 
% Notes:
% In order to obtain a coordinate, you need to call the function <ind2pos>
% on the result.
%
% See also:
% <find_index_max>
function y = find_index_min(x : cube{n})
    y = find_index_max(-x)
endfunction

function y : int = __device__ find_index_max[T](x : vec[T])
    y = 0
    maxval = x[0]
    for i=1..numel(x)-1
        x_i = x[i]
        if maxval<x_i
            maxval=x_i
            y = i
        endif
    endfor
endfunction

function y : int = __device__ find_index_max[T](x : mat[T])
    y = 0
    maxval = x[0,0]
    for m=1..numel(x)-1
        i = ind2pos(size(x), m)
        x_i = x[i]
        if maxval<x_i
            maxval=x_i
            y = m
        endif
    endfor
endfunction

function y : int = __device__ find_index_max[T](x : cube[T])
    y = 0
    maxval = x[0,0,0]
    for m=1..numel(x)-1
        i = ind2pos(size(x), m)
        x_i = x[i]
        if maxval<x_i
            maxval=x_i
            y = m
        endif
    endfor
endfunction

function y : int = __device__ find_index_min[T](x : vec[T])
    y = 0
    minval = x[0]
    for i=1..numel(x)-1
        x_i = x[i]
        if minval>x_i
            minval=x_i
            y = i
        endif
    endfor
endfunction

function y : int = __device__ find_index_min[T](x : mat[T])
    y = 0
    minval = x[0,0]
    for m=1..numel(x)-1
        i = ind2pos(size(x), m)
        x_i = x[i]
        if minval>x_i
            minval=x_i
            y = m
        endif
    endfor
endfunction

function y : int = __device__ find_index_min[T](x : cube[T])
    y = 0
    minval = x[0,0,0]
    for m=1..numel(x)-1
        i = ind2pos(size(x), m)
        x_i = x[i]
        if minval>x_i
            minval=x_i
            y = m
        endif
    endfor
endfunction

% Function: find_indices_nonzero
% 
% finds the indices of all non-zero values of the given cube.
% 
% Usage:
%   : function [y : int] = find_indices_nonzero(x : cube)
% 
% Parameters:
% x - the input matrix
% 
% Returns:
% y - a vector with indices of the non-zero elements of x. 
% 
% Notes:
% * In order to obtain a coordinate, you need to call the function 
% <delinearize_index> on the result.
% * There is no guarantee that the resulting index vector is sorted
function y = find_indices_nonzero(x : cube)   
    
    function [] = __kernel__ swizzle(x : vec, y : mat'unchecked'nocopy, sz : int, num_sm : int, pos : int)
        {!kernel target="gpu"}
        blkdim = int(sz/num_sm)
        p = mod(pos,sz)
        idx = int(pos/sz)
        mp = int(p/blkdim)
        y[mp,mod(p,blkdim)+idx*blkdim] = x[pos] != 0 ? 1 : 0
    endfunction
    
    function [] = __kernel__ parallel_find(x : vec'unchecked, y : vec'unchecked, _
        cnt : vec'unchecked, sz : int, pos : int, blkidx : int)
        {!kernel target="gpu"}
        nblocks = int((numel(x) + sz - 1) ./ sz)

        counter : vec[int]'unchecked = shared[int](64)
        counter[blkidx] = blkidx == 0 ? 0 : int(cnt[blkidx-1]) % reset
        syncthreads

        for m=0..nblocks-1
            p = pos + m .* sz

            if p<numel(x) && x[p] != 0
                % allocate one entry in the output
                idx = (counter[blkidx] += 1)
                y[idx-1] = float(p)
            endif
        endfor
    endfunction
    
    function [] = __kernel__ singlethreaded_find(x : vec'unchecked, y : vec'unchecked)
        idx = 0
        for m=0..numel(x)-1
            if x[m] != 0
                y[idx] = m
                idx += 1
            endif
        endfor
    endfunction
    
    match schedule(numel(x),x, {parallel_find, singlethreaded_find}) with    
    | 0 ->
        blk_size = prod(max_block_size(parallel_find, prevpow2(numel(x))))
        num_sm = min([64,ceil(numel(x)/blk_size), prod(opt_block_cnt(parallel_find, blk_size, numel(x)))])
        sz = blk_size * num_sm
        
        x2 = uninit(num_sm,int(ceil(numel(x)/sz))*blk_size)
        parallel_do(numel(x2), x, x2, sz, num_sm, swizzle)
        x2 = sum(x2,1)
        x2 = cumsum(transpose(x2))
        nz = max(0,x2[num_sm-1])
        y = uninit(int(nz))    
        
        parallel_do([[sz],[blk_size]],x,y,x2,sz,parallel_find)
    | 1 ->
        nz = sum(x != 0)
        y = uninit(nz)
        serial_do(1,x,y,singlethreaded_find)
    endmatch
endfunction
    
% Function: select
% 
% Selects all values with the specified filter.
% 
% Usage:
%   : function [y : vec] = select(x : cube, filter : [__device__ scalar -> int])
% 
% Parameters:
% x - the input data
% T - the threshold to apply
% 
% Returns:
% y - the filtered set of input values
function y = select(x : cube, filter : [__device__ scalar -> scalar])       
    function [] = __kernel__ swizzle(x : vec, y : mat, sz : int, num_sm : int, pos : int)
        blkdim = int(sz/num_sm)
        p = mod(pos,sz)
        idx = int(pos/sz)
        mp = int(p/blkdim)
        y[mp,mod(p,blkdim)+idx*blkdim] = x[pos]
    endfunction
    
    function [] = __kernel__ parallel_find(x : vec'unchecked, y : vec'unchecked, cnt : vec'unchecked, 
        filter : [__device__ scalar -> scalar], sz : int, pos : int, blkidx : int)
        
        counter : vec[int]'unchecked = shared[int](64)
        counter[blkidx] = blkidx==0 ? 0 : int(cnt[blkidx-1]) % reset
        syncthreads

        for p = pos..sz..numel(x)-1
            if p<numel(x) && filter(x[p])
                % allocate one entry in the output
                idx = (counter[blkidx] += 1)
                y[idx-1] = x[p]
            endif
        endfor
    endfunction
    
    function [] = __kernel__ singlethreaded_find(x : vec'unchecked, y : vec'unchecked, 
        filter : [__device__ scalar -> scalar])
        idx = 0
        for m=0..numel(x)-1
            val = x[m]
            if filter(val)
                y[idx] = val
                idx += 1
            endif
        endfor
    endfunction
    
    match schedule(numel(x),x, {parallel_find, singlethreaded_find}) with
    | 0 ->
        blk_size = prod(max_block_size(parallel_find, prevpow2(numel(x))))
        num_sm = min(64,prod(opt_block_cnt(parallel_find, blk_size, numel(x))))
        sz = blk_size * num_sm
        
        x2 = uninit(num_sm,int(ceil(numel(x)/sz))*blk_size)
        parallel_do(numel(x2), x, x2, sz, num_sm, swizzle)
        x2 = sum(x2,1)
        x2 = cumsum(transpose(x2))
        nz = max(0,ceil(x2[num_sm-1]))
        y = uninit(nz)    
        
        parallel_do([[sz],[blk_size]],x,y,x2,filter,sz,parallel_find)
    | 1 ->
        nz = sum(apply_fn(x, filter))
        y = uninit(nz)
        serial_do(1,x,y,filter,singlethreaded_find)
    endmatch
endfunction

% Function: select_and_map
% 
% Selects all values with the specified filter and maps them onto a new domain.
% 
% Usage:
%   : function [y : vec] = select_and_map(x : cube, filter : [__device__(scalar) -> int], 
%   :                                     map : [__device__(int) -> scalar])
% 
% Parameters:
% x - the input data
% filter - returns 1 if the specified input value is selected, otherwise 0
% map - maps the input index onto an output value (which will be written to y)
% 
% Returns:
% y - the filtered and set of mapped values
function y = select_and_map(x : cube, filter : [__device__(scalar) -> scalar], map : [__device__(int) -> scalar])
    
    function [] = __kernel__ swizzle(x : vec, y : mat, sz : int, num_sm : int, pos : int)
        blkdim = int(sz/num_sm)
        p = mod(pos,sz)
        idx = int(pos/sz)
        mp = int(p/blkdim)
        y[mp,mod(p,blkdim)+idx*blkdim] = filter(x[pos])
    endfunction
    
    function [] = __kernel__ parallel_find(x : vec'unchecked, y : vec'unchecked, cnt : vec'unchecked, 
        sz : int, filter : [__device__(scalar) -> scalar], map : [__device__(int) -> scalar], 
        pos : int, blkidx : int)
        
        counter : vec[int]'unchecked = shared[int](64)
        counter[blkidx] = blkidx==0 ? 0 : int(cnt[blkidx-1]) % reset
        syncthreads

        for p = pos..sz..numel(x)-1
            if p<numel(x) 
                if filter(x[p])
                    % allocate one entry in the output
                    idx = (counter[blkidx] += 1)
                    y[idx-1] = map(p)
                endif
            endif
        endfor
    endfunction
    
    function [] = __kernel__ singlethreaded_find(x : vec'unchecked, y : vec'unchecked, 
        filter : [__device__(scalar) -> scalar], map : [__device__(int) -> scalar])
        idx = 0
        for m=0..numel(x)-1
            if filter(x[m])
                y[idx] = map(m)
                idx += 1
            endif
        endfor
    endfunction
    
    match schedule(numel(x),x, {parallel_find, singlethreaded_find}) with
    | 0 ->
        blk_size = prod(max_block_size(parallel_find, prevpow2(numel(x))))
        num_sm = min(64,prod(opt_block_cnt(parallel_find, blk_size, numel(x))))
        sz = blk_size * num_sm
        
        x2 = uninit(num_sm,int(ceil(numel(x)/sz))*blk_size)
        parallel_do(numel(x2), x, x2, sz, num_sm, swizzle)
        x2 = sum(x2,1)
        x2 = cumsum(transpose(x2))
        nz = max(0,ceil(x2[num_sm-1]))
        y = uninit(nz)    
        
        parallel_do([[sz],[blk_size]],x,y,x2,sz,filter,map,parallel_find)
    | 1 ->
        nz = sum(apply_fn(x, filter))
        y = uninit(nz)
        serial_do(1,x,y,filter,map,singlethreaded_find)
    endmatch
endfunction
    
% Function: select_threshold
% 
% Selects all values that are larger than a given threshold
% useful for implementing 'logical' indexing, e.g. in Matlab: x(x > T).
% 
% Usage:
%   : function [y] = select_threshold(x : cube, T : scalar)
% 
% Parameters:
% x - the input data
% T - the threshold to apply
% 
% Returns:
% y - the filtered set if input values, with the threshold applied
function y = select_threshold(x : cube, T : scalar)       
    y = select(x, __device__ (x : scalar) -> x >= T)
endfunction

% Function: distinct
% 
% Selects all distinct values in a given vector/matrix/cube
% 
% Usage:
%   : function [y] = distinct(x : cube)
% 
% Parameters:
% x - the input data
% 
% Returns:
% y - a vector of all the distinct values
function y : vec = distinct(x : cube)
    z = reshape(x,numel(x))
    sort(z)    
    dz = uninit(numel(x))
    parallel_do(numel(x),z,__kernel__ (z : vec, pos : int) -> 
        dz[pos] = pos == 0 || z[pos] != z[pos-1])   % TODO dz = gives an error!
    
    y = z[find_indices_nonzero(dz)]
endfunction


% Function: distinct_map
% 
% Selects all distinct values in a given vector/matrix/cube + 
% the mapping function of the original set onto y. This function
% is useful for e.g. connected components algorithms for performing
% label normalization. The connected component algorithm may output
% labels in the range [0,MN-1], where MN are the dimensions of the
% image. With this function, each integer in the range 0..K-1 will
% correspond to a unique label (where K is the total number of labels)
% 
% Usage:
%   : function [y, map] = distinct_map(x : cube)
% 
% Parameters:
% x - the input data (>=0)
% 
% Returns:
% y - a vector of all the distinct values
% map - a mapping function of x onto the index of y.
function [y, map] = distinct_map(x : cube)
    z = reshape(x,numel(x))
    sort(z)
    dz = uninit(numel(x))
    % Detect all the transitions
    parallel_do(numel(x),z,__kernel__ (z : vec, pos : int) -> 
        dz[pos] = pos == 0 || z[pos] != z[pos-1])   % TODO dz = gives an error!
    
    ind : vec = find_indices_nonzero(dz)
    y = z[ind] % the set of unique labels
    map = zeros(max(y)+1) % construct a mapping
    parallel_do(numel(ind),y, __kernel__ (y : vec, pos : int) -> 
        map[y[pos]] = pos)
endfunction

% Function: replace_mask
% 
% Interpolates between two input data cubes based on a given mask (with
% values between 0 - 1). This function can be used for alpha matting as well.
% 
% Usage:
%   : function [y] = replace_mask(x_old : cube, x_new : [cube|scalar], mask : cube)
% 
% Parameters:
% x_old - the old input data
% x_new - the new input data
% mask - is one at places where the value needs to be changed
% 
% Returns:
% y - output data
function y = replace_mask(x_old : cube, x_new : [cube|scalar], mask : cube)
    if isscalar(x_new)
        y = x_old + (x_new - x_old) .* mask
    else
        y = x_old + (cast(x_new, cube) - x_old) .* mask
    endif
endfunction

% Function: replace_by_index
% 
% Replaces values of the data cube by a certain value, but only at the 
% specified indices.
% 
% Usage:
%   : function [y] = replace_by_index(x_old : cube, x_new : [cube|scalar], indices : vec)
% 
% Parameters:
% x_old - the old input data (a cube or a scalar value)
% x_new - the new input data
% indices - the set of positions that need to be changed
% 
% Returns:
% y - output data
function y = replace_by_index(x_old : cube, x_new : [cube|scalar], indices : vec)
    y = copy(x_old)
    if isscalar(x_new)
        parallel_do(numel(indices), y, x_new, indices, 
            __kernel__ (y : vec, x_new : scalar, indices : vec, pos : int) ->
                (ind = indices[pos], y[ind] = x_new))    
    else       
        parallel_do(numel(indices), y, x_new, indices, 
            __kernel__ (y : vec, x_new : vec, indices : vec, pos : int) ->
                (ind = indices[pos], y[ind] = x_new[ind]))
    endif
endfunction

function y : cube{n} = dimensionop(x : cube{n}, dim : int, 
    gpu_kernel : [__kernel__ (vec, vec,int,int,int) -> ()],
    cpu_kernel : [__kernel__ (vec, vec,int,int,int) -> ()]) concealed

    rec_dim = size(x,0..dim-1)
    dominant_dim = size(x,dim)
    u = max(dim+1,ndims(x))
    sup_dim = size(x,dim+1..u)
    new_dims = size(x, 0..u) % the new dimensions
    new_dims[dim] = 1 % set the dominant dimension to 1
    type cube_n : cube{n}
    y = cube_n(new_dims)
        
    
    match schedule(new_dims, x, y, prod(rec_dim), dominant_dim, prod(sup_dim), {gpu_kernel,cpu_kernel}) with
    | 0 -> % GPU version of the kernel
        max_threads = prod(max_block_size(gpu_kernel,[64,64]))
        block_size_x = int(max(1, (dominant_dim + max_threads - 1) / max_threads))
        n_threads = nextpow2(dominant_dim / block_size_x)
        block_size_y = int(max_threads / n_threads)    
        blocks_y = int((prod(rec_dim) * prod(sup_dim) + block_size_y - 1) / block_size_y)
        blocks_z = 1
        max_grid_size = 65535
            
        while blocks_y > max_grid_size && and(blocks_y, 1) == 0
            blocks_y /= 2
            blocks_z *= 2
        endwhile
        while blocks_y > max_grid_size && mod(blocks_y, 3) == 0
            blocks_y /= 3
            blocks_z *= 3
        endwhile
        while blocks_y > max_grid_size && mod(blocks_y, 5) == 0
            blocks_y /= 5
            blocks_z *= 5
        endwhile
        
        parallel_do([[blocks_z, blocks_y*block_size_y, n_threads],[1, block_size_y, n_threads]], _
            x,y,prod(rec_dim), dominant_dim, prod(sup_dim),gpu_kernel)
    | 1 -> % CPU version of the kernel
        parallel_do(prod(new_dims),x,y,prod(rec_dim), dominant_dim, prod(sup_dim),cpu_kernel)
    endmatch
endfunction
    
% Function: aggregate
% 
% Performs an aggregation operation to a given data cube, along the dim-th dimension,
% using a user-defined aggregation function
% 
% Usage:
%   : function [y : cube{n}] = aggregate(x : cube{n}, dim : int, z0, aggregator : [__device__ (scalar,scalar)->scalar])
% 
% Parameters:
% x - the input data cube
% dim - base-0 dimension to apply the aggregation
% z0 - the initial value to be used for the aggregation
% aggregator - the function used for aggregation
% 
% Returns:
% y - the aggregated data cube
%
% Remarks:
% * the function <squeeze> may be useful to eliminate singleton dimensions resulting from the
%   aggregation
function y : cube{n} = aggregate(x : cube{n}, dim : int, z0 : scalar, _
    aggregator : [__device__ (scalar, scalar) -> scalar]) concealed

    % kernel for summing a matrix along the rows...
    function [] = __kernel__ gpu_kernel(x : vec'unchecked, y : vec'unchecked, x_rows : int, _
        x_cols : int, x_stride : int, aggregator : [__device__ (scalar, scalar) -> scalar], _
        blkdim : ivec3, blkpos : ivec3, blkcnt : ivec3, blkidx : ivec3)    
    
        {!kernel target="gpu"}
        bins = shared(1024)
        i = blkpos[2]
        m = (blkidx[1]*blkdim[1]+blkpos[1]) + blkidx[0]*blkdim[1]*blkcnt[1]
        pos = ind2pos([x_rows, x_stride], m)
        if pos[0] >= x_rows || pos[1] >= x_stride
            return
        endif
               
        % step 1 - parallel sum
        val = z0
        offset = pos[0] * x_cols * x_stride + pos[1]
        for n=i * x_stride..blkdim[2]*x_stride..x_cols*x_stride-1
            val = aggregator(val, x[offset + n])
        endfor
        i = blkpos[1]*blkdim[2] + blkpos[2] % note: blkpos[2]==0
        bins[i] = val
        
        % step 2 - reduction
        syncthreads
        bit = 1
        while bit<blkdim[2]
            if and(blkpos[2],2*bit-1) == 0 && blkpos[2]+bit<blkdim[2]
                bins[i] = aggregator(bins[i], bins[i+bit])
            endif
            syncthreads            
            bit *= 2
        endwhile
        if blkpos[2]==0
            y[pos[0] * x_stride + pos[1]] = bins[blkpos[1]*blkdim[2]]
        endif
    endfunction
    
    function [] = __kernel__ cpu_kernel(x : vec'unchecked, y : vec'unchecked, x_rows : int, _
        x_cols : int, x_stride : int, aggregator : [__device__ (scalar, scalar) -> scalar], pos : int)
        
        {!kernel target="cpu"}
        pos2 = ind2pos([x_rows, x_stride], pos)
        sz = [x_rows, x_cols, x_stride]
        val = z0
        
        for i=0..x_cols-1
            j=pos2ind(sz, [pos2[0],i,pos2[1]])
            val = aggregator(val, x[j])
        endfor
        
        y[pos] = val
    endfunction
    
    y = dimensionop(x, dim, $specialize(gpu_kernel, aggregator==aggregator),
                            $specialize(cpu_kernel, aggregator==aggregator)) 
endfunction

% Function: count_nonzero
% 
% Counts the number of elements of `x` that are non-zero
% 
% Usage:
%   : function [y : scalar] = count_nonzero(x : cube{16}'const)
% 
% Parameters:
% x - the input matrix
% 
% Returns:
% y - the output matrix
function y = count_nonzero(x : cube{:})
    y = sum(x != 0)
endfunction

    
function [] = main()
    function [] = test_find_first()
        x = randn(65536*256)
        %x = zeros(65536)
        %x[332..3..36366] = 1.0
        N = 2
        tic()
        for k=0..N-1
            y2 = find(x, "first")
        endfor
        toc()
        tic()
        for k=0..N-1
            y2 = find_index_max(x)
        endfor
        toc()
        %print y, ",", x[y]
        print y2, ",", x[y2]
        print max(x)
        assert(x[y2] == max(x))
    endfunction

    function [] = test_find_indices_nonzero()
        x = randn(2^13+17)> 0
        x[150..3..169] = 0
        N = 10
        tic()
        for k=0..N-1
            u = find(x, "all")[:,1]
        endfor
        toc()
        print numel(u)
        %sort(u)
        %plot(u)
        tic()
        for k=0..N-1
            u = find_indices_nonzero(x)
            assert(numel(u) == sum(x))
        endfor
        toc()
        print numel(u)
        print "correct: ", sum(x > 0)
        
        %sort(u)
        %plot(u)
        
        %plot(select_threshold(randn(127,127), -1))
    endfunction
    
    function [] = test_distinct()
        x = floor(randn(65536))
        N = 2
        tic()
        for k=0..N-1
            y = distinct(x)
        endfor
        toc()
        print y    
    endfunction
    
    function [] = test_aggregation()
        im = imread("lena_big.tif")
        im_out = aggregate(im, 2, 0, (__device__ (x, y) -> max(x,y)))
        imshow(im_out)
    endfunction

    function [] = test_select_and_map()
        x = randn(4095)
        % Select positive elements of this array
        y = select_and_map(x,
                (__device__ (x : scalar) -> float(x > 0)),
                (__device__ (p : int) -> x[p]))
        print min(y)
        assert(min(y)>0)
        plot(x)
        plot(y)
    endfunction

    test_select_and_map()
    test_find_first()    
    test_find_indices_nonzero()
    test_distinct()
    test_aggregation()
    
endfunction