import "linalg.q"

{!author name="Bart Goossens"}
{!doc category="Statistics"}

% Section: pca.q
% Principal Component Analysis


% Function: cross_cov
% 
% Computes the cross covariance matrix between two data sets
% 
% Usage:
%   : function [C : mat(N,N)] = cross_cov[N](A : mat'const(:,N), B : mat'const(:,N), subtract_mean = 1)
% 
% Parameters:
% A - the first dataset (each row of the matrix is a vector of length N)
% B - the second dataset (each row of the matrix is a vector of length N)
% subtract_mean - indicates whether the mean needs to be subtract, before calculating the cross 
%    covariance
% 
% Returns:
% C - the cross-covariance between A and B
function C : mat(N,N) = cross_cov[N](A : mat(:,N), B : mat(:,N), subtract_mean = 1)
    
    % Number of data points
    M = size(A,0)
    assert(size(A,0)==size(B,0), "(cross_cov) number of data points in A and B must match")
    
    function [C : vec(N*(N-1))] = __kernel__ kernel(A : mat'unchecked(:,N), B : mat'unchecked(:,N), 
        subtract_mean : int, pos : int)
        Av = A[pos,:]
        Bv = B[pos,:]
        
        if subtract_mean
            Av -= mean(Av)
            Bv -= mean(Bv)
        endif

        % Uses parallel reduction algorithm for the fast calculation
        C += utriang_vprod[N](Av, Bv)
    endfunction
    
    C : mat(N,N) = utriangm[N](parallel_do(prod(size(A,0..1)),A,B,subtract_mean,kernel))/M

    % Symmetry extension        
    C = utriang_symext[N](C)
    
endfunction

% Function: cov
% 
% Computes the covariance matrix of the specified data set
% 
% Usage:
%   : function [C : mat(N,N)] = cov[N](A : mat'const'static(:,N), subtract_mean : ??'const = 1)
% 
% Parameters:
% A - the input dataset with each row of the matrix a vector of length N
% subtract_mean - indicates whether the mean needs to be subtract, before calculating the cross 
%    covariance
% 
% Returns:
% C - the covariance matrix of A
function C : mat(N,N) = cov[N](A : mat(:,N), subtract_mean = 1)
    
    % Number of data points
    M = size(A,0)
    
    function [C : vec(N*(N-1))] = __kernel__ kernel(A : mat'unchecked(:,N), subtract_mean : int, pos : int)
        Av : vec(N) = A[pos,:]

        if subtract_mean
            Av -= mean(Av)
        endif
        
        C += utriang_vprod(Av, Av)
    endfunction
    
    C : mat(N,N) = utriangm[N](parallel_do(prod(size(A,0..1)),A,subtract_mean,kernel))/M

    % Symmetry extension        
    C = utriang_symext(C)
    
endfunction

% Function: pca
% 
% Performs L2 Principal Component Analysis
% 
% Usage:
%   : function [U : mat, Lambda : vec] = pca[N](A : mat'static(:,N))
% 
% Parameters:
% A - the input dataset with each row of the matrix a vector of length N
% 
% Returns:
% U - the matrix with eigenvectors of the data covariance matrix (each row is an eigenvector)
% Lambda - the eigenvalues of the data covariance matrix
function [U, Lambda] = pca[N](A : mat(:,N))

    [U, Lambda] = symeig(cov[N](A, 1))

endfunction

% Function: l1pca
% 
% Performs L1 Principal Component Analysis. L1PCA is particularly suitable when the data contain
% outliers.
% 
% Usage:
%   : function [U : mat(N,N), Lambda : mat] = l1pca[N](A : mat'static(:,N))
% 
% Parameters:
% A - the input dataset with each row of the matrix a vector of length N
% 
% Returns:
% U - the matrix with eigenvectors of the data covariance matrix (each row is an eigenvector)
% Lambda - the eigenvalues of the data covariance matrix
function [U, Lambda] = l1pca[N](A : mat(:,N))

    % Subtract the mean (only needs to be done once)
    A = A - repmat(mean(A,1),[1,N])

    [U : mat(N,N), Lambda, V : mat(N,N)] = svd(cov[N](A, 0))

    function z : mat(N,N) = L1PCA_seed(A : mat(:,N), V : mat(N,N))
        epsilon = 1e-6 % epsilon for single precision floating point
        epsilon2 = 1e-2
        flag = 0
        norm = x -> sqrt(sum(x.^2))
        z = V ./ norm(V)
    
        while flag < 1
            % polarity check
            w = A * transpose(z)
            w2 : mat(:,N) = ones(size(w)) - 2.*(w<0)
            % flipping and maximization
            z2 : mat(N,N) = cross_cov[N](w2,A,0)
            z2 /= norm(z2)

            % convergence checks
            if sum(abs(z-z2)) < epsilon
                w3 = sum(sum(abs(w2),1) == 0)
                if w3 > 0
                    z2 = z2 + epsilon2.*(rand(size(z2))-0.5)
                    z2 /= norm(z2)
                else
                    flag = 1
                endif
            endif
            z = z2
        endwhile
    endfunction

    z = V
    for i=0..N-1
        z2 = L1PCA_seed(A, V)
        z[i,:] = z2[i,:]
        A = A - A * (transpose(z2)*z2)
    endfor
    U = z*3
endfunction


function [] = main()
    im = imread("lena_big.tif")
    imA : mat(:,3) = reshape(im, [prod(size(im,0..1)),size(im,2)])
    runs = 1000
    tic()
    for k=0..runs-1
        A = cov(imA)
    endfor
    toc()
    print A
    tic()
    [U, Lambda] = pca(imA)
    toc("L2 PCA")
    print U
    print Lambda
    
    tic()
    [U, Lambda] = l1pca(imA)
    toc("L1 PCA")
    print U
    print Lambda
    
endfunction