%============================================
% Quasar procrustes analysis
% Sep 2015 - Simon Donné
%===========================================
import "system.q"
import "linalg.q"

{!author name="Simon Donné"}
{!doc category="Linear algebra"}

function [] = main()
    N  = 50
    angle = pi/8
    transform = [[cos(angle),-sin(angle), 0.5],_
                 [sin(angle), cos(angle),-0.5],_
                 [         0,          0, 1.2]]
    u  = [[rand(2,N) * 2],[ones(1,N)]]
    ut = transform * u
    ut = ut ./ repmat(ut[2,:],[3,1])
    
    scatter(u [0,:],u [1,:],"r")
    hold("on")
    scatter(ut[0,:],ut[1,:],"g")
    for n = 0..N-1
        plot([u[0,n],ut[0,n]],[u[1,n],ut[1,n]],"k")
    end
    
    [tf,Z] = procrustes(u[0..1,:],ut[0..1,:])
    
    scatter(Z[0,:],Z[1,:],"b")
end

% Function: procrustes
% Estimates a linear transformation (translation, rotation, scaling) to match Y with X
%   
% : function [tf,Z] = procrustes(X : mat, Y : mat, reflection : int = 1)
%   
%   X and Y: DxN input matrices (2-dimensional arrays of equal size)
%   
%   reflection: indicator of whether to allow reflection. Reflection is only allowed if explicitly indicated:
%       reflection = -1: force reflection
%       reflection =  0: allow reflection
%       reflection =  1:   bar reflection
%   
%   tf: a (D+1)x(D+1) homography that can be applied to the points in homogeneous representation
%   
%   Z: DxN output matrix of the transformed points from Y
function [tf:mat,Z:mat] = procrustes(X:mat,Y:mat,reflection:int=1)
    if any(size(X) != size(Y))
        error("Procrustes analysis requires arguments to have the same size.")
    endif
    if numel(size(X)) > 2
        error("Procrustes analysis requires two-dimensional arguments.")
    endif
    [D,N] = size(X,0..1)
    
    %center both datasets at the origin: subtract their mean
    mean_X = mean(X,1)
    mean_Y = mean(Y,1)
    X0 = X - repmat(mean_X,[1,N])
    Y0 = Y - repmat(mean_Y,[1,N])
    
    %scale to unit norm
    norm_X = sqrt(sum(X0.^2))
    norm_Y = sqrt(sum(Y0.^2))
    X0 = X0/norm_X
    Y0 = Y0/norm_Y
    
    %get the rotation matrix for Y
    A = X0*transpose(Y0)
    [U,S,V] = svd(A)
    T = V*transpose(U)
    
    %force the reflection state to whatever was decided if necessary
    if sign(det(T))*reflection < 0
        S[N-1,N-1] = -S[N-1,N-1]
        V[:,N-1] = -V[:,N-1]
        T = V*transpose(U)
    endif
    
    %the scaling
    trace_S = sum(S)
    b = trace_S * norm_X / norm_Y
    
    %generate the outputs
    Z = norm_X*trace_S*transpose(T)*Y0 + repmat(mean_X,[1,N])
    c = mean_X - b*transpose(T)*mean_Y
    
    tf = zeros(D+1,D+1)
    tf[0..D-1,0..D-1] = T
    tf[0..D-1,D] = -T*c
    tf[D,D] = b
end