% Type: Queue
%
% Generic implementation of a FIFO queue. Allows queuing and dequeuing
% elements in a first-in first-out order.
% 
% This queue implementation is intended to be used mostly
% from host functions (e.g., adding, removing). Once filled,
% the data can be accessed from kernel/device functions.
{!author name="Bart Goossens"}
{!doc category="Data structures"}

type Queue[T] : mutable class
    items       : vec[T]    
    first_index : int
    end_index   : int
    capacity    : int
endtype

% Function: Queue
% 
% Constructor for a queue object
% 
% Usage:
%   : function [y : Queue[T]] = Queue[T](capacity : int)
% 
% Parameters:
% capacity - the initial capacity of the queue (the maximally expected number
% of elements that the queue can contain. Note, if the capacity is exceeded,
% more space for the queue is allocated)
% 
% Returns:
% y - a new list object
function y : ^Queue[T] = Queue[T](capacity : int = 10)
    y = Queue[T](items:=vec[T](capacity), capacity:=capacity, first_index:=0, end_index:=0)
endfunction

% Function: Queue.count
% 
% Returns the number of elements in this queue
% 
% Usage:
%   : function [cnt : int] = Queue[T].count()
% 
% Returns:
% cnt - the number of elements in this queue
function count : int = count[T](self : ^Queue[T])
    count = self.end_index - self.first_index;
    if count < 0
        count += self.capacity
    endif
endfunction

% Function: Queue.enqueue
% 
% Enqueues an element to the queue. 
% 
% Usage:
%   : function [] = Queue[T].enqueue[T](item : T)
% 
% Parameters:
% item - the item to add to the queue
function [] = enqueue[T](self : ^Queue[T], item : T)
    if self.count() >= self.capacity
        %self.set_capacity(self.capacity * 2)
        error "Queue.enqueue: the maximum capacity has been reached!"
    endif
    index = self.end_index
    self.end_index = mod(self.end_index+1, self.capacity)
    self.items[index] = item
endfunction

% Function: Queue.dequeue
% 
% Dequeues an element from the queue
% 
% Usage:
%   : function y = Queue[T].dequeue()
function y = dequeue[T](self : ^Queue[T])
    if self.end_index == self.first_index
        error "Queue.dequeue: queue empty!"
    endif    
    y = self.items[self.first_index]
    self.first_index = mod(self.first_index+1, self.capacity)
endfunction

% Function: Queue.clear
% 
% Clears the content of this queue
% 
% Usage:
%   : function [] = __device__ Queue[T].clear()
% 
% Parameters:
% self - the queue object to clear
function [] = clear[T](self : ^Queue[T])
    self.first_index = self.end_index = 0
endfunction

% Function: Queue.at
% 
% Gets the element of a queue with the specified index
% 
% Usage:
%   : function [out] = Queue[T].at(index : int)
% 
% Parameters:
% index - an index between 0 and Queue[T].count()-1
% 
% Returns:
% out - the element at position index.
function y = at[T](self : ^Queue[T], index : int)
    if index < 0 || index >= self.count()
        error "Queue.at: index out of bounds: ",index," must be between ",0," and ",self.count()-1
    endif
    index = mod(index + self.first_index, self.capacity)
    y = self.items[index]
endfunction

% Function: Queue.to_string
% 
% Converts the Queue to a string representation
% 
% Usage:
%   : function [out] = Queue[T].to_string()
% 
% Returns:
% out - a string representation for the queue
function [s] = to_string[T](self : ^Queue[T])
    s = "["
    for i=0..self.count()-1
        if i>0
            s=s+","
        endif
        s=s+sprintf("%g",self.at(i))
    endfor
    s = s+ "]"
endfunction

function [] = main()
    s = Queue[int](32)
    for k=0..9
        s.enqueue(1+k)
    endfor
    
    print s.dequeue()
    print s.dequeue()
    print s.to_string()
    
endfunction