% Quasar special function library
% This class contains physical constants and special functions not found
% All physical constants are in CGS units.
% Notes: 
% - These special functions do not necessarily use the fastest or most 
%   accurate algorithms.
% - Some special functions are already built-in in Quasar and are not defined 
%   here:
%       asinh, acosh, atanh, erf, erfc, erfinv, erfcinv, lgamma, tgamma

{!author name="Bart Goossens"}
{!doc category="Mathematics/Special functions"}

MAXLOG = 7.09782712893383996732E2
MINLOG = -7.451332191019412076235E2
MAXGAM = 171.624376956302725
SQTPI = 2.50662827463100050242E0
SQRTH = 7.07106781186547524401e-1
LOGPI = 1.14472988584940017414

BOLTZMAN = 1.3807e-16       % Boltzman Constant. Units erg/deg(K) 
ECHARGE = 4.8032e-10        % Elementary Charge. Units statcoulomb 
EMASS = 9.1095e-28          % Electron Mass. Units g 
PMASS = 1.6726e-24          % Proton Mass. Units g 
GRAV = 6.6720e-08           % Gravitational Constant. Units dyne-cm^2/g^2
PLANCK = 6.6262e-27         % Planck constant. Units erg-sec 
LIGHTSPEED = 2.9979e10      % Speed of Light in a Vacuum. Units cm/sec 
STEFANBOLTZ = 5.6703e-5     % Stefan-Boltzman Constant. Units erg/cm^2-sec-deg^4 
AVOGADRO = 6.0220e23        % Avogadro Number. Units  1/mol 
GASCONSTANT = 8.3144e07     % Gas Constant. Units erg/deg-mol 
GRAVACC = 980.67            % Gravitational Acceleration at the Earths surface. Units cm/sec^2 
SOLARMASS = 1.99e33         % Solar Mass. Units g 
SOLARRADIUS = 6.96e10       % Solar Radius. Units cm
SOLARLUM = 3.90e33          % Solar Luminosity. Units erg/sec
SOLARFLUX = 6.41e10         % Solar Flux. Units erg/cm^2-sec
AU = 1.50e13                % Astronomical Unit (radius of the Earth's orbit). Units cm

% Function: bessel_j0
% Returns the Bessel function of order 0 of the specified number.
%
% : function [y : scalar] = __device__ bessel_j0(x : scalar)
function [y : scalar] = __device__ bessel_j0(x : scalar)
    ax = abs(x)
    if ax < 8.0
        y = x*x
        ans1 = 57568490574.0 + y*(-13362590354.0 + y*(651619640.7 + y*(-11214424.18 + y*(77392.33017 + y*(-184.9052456)))))
        ans2 = 57568490411.0 + y*(1029532985.0 + y*(9494680.718 + y*(59272.64853 + y*(267.8532712 + y*1.0))))
        y = ans1/ans2

    else
        z = 8.0/ax
        y = z*z
        xx = ax - 0.785398164
        ans1 = 1.0 + y*(-0.1098628627e-2 + y*(0.2734510407e-4 + y*(-0.2073370639e-5 + y*0.2093887211e-6)))
        ans2 = -0.1562499995e-1 + y*(0.1430488765e-3 + y*(-0.6911147651e-5 + y*(0.7621095161e-6 - y*0.934935152e-7)))
        y = sqrt(0.636619772/ax)*(cos(xx)*ans1 - z*sin(xx)*ans2)
    endif
end

% Function: bessel_j1
% Returns the Bessel function of order 1 of the specified number.
%
% : function [y : scalar] = __device__ bessel_j1(x : scalar)
function [y : scalar] = __device__ bessel_j1(x : scalar)  
    ax = abs(x)
    if ax < 8.0
        y = x*x
        ans1 = x*(72362614232.0 + y*(-7895059235.0 + y*(242396853.1 + y*(-2972611.439 + y*(15704.48260 + y*(-30.16036606))))))
        ans2 = 144725228442.0 + y*(2300535178.0 + y*(18583304.74 + y*(99447.43394 + y*(376.9991397 + y*1.0))))
        y = ans1/ans2
    else
        z = 8.0/ax
        xx = ax - 2.356194491
        y = z*z
        ans1 = 1.0 + y*(0.183105e-2 + y*(-0.3516396496e-4 + y*(0.2457520174e-5 + y*(-0.240337019e-6))))
        ans2 = 0.04687499995 + y*(-0.2002690873e-3 + y*(0.8449199096e-5 + y*(-0.88228987e-6 + y*0.105787412e-6)))
        ans = sqrt(0.636619772/ax)* (cos(xx)*ans1 - z*sin(xx)*ans2)
        if (x < 0.0) 
            ans = -ans
        endif
        y = ans
    endif
end

% Function: bessel_jn
% Returns the Bessel function of order n of the specified number.
%
% : function [y : scalar] = __device__ bessel_jn(x : scalar)
function [y : scalar] = __device__ bessel_jn(n : int, x : scalar)
    ACC = 40.0
    BIGNO = 1e10
    BIGNI = 1e-10

    if n == 0 
        y=bessel_j0(x)
        return
    endif
    if n == 1
        y=bessel_j1(x)
        return
    endif

    ax = abs(x)
    if ax == 0.0
        y = 0.0
    elseif ax > float(n)
        tox = 2.0/ax
        bjm = bessel_j0(ax)
        bj = bessel_j1(ax)
        for j = 1..n-1
            bjp = j*tox*bj - bjm
            bjm = bj
            bj = bjp
        end
        ans = bj
    else
        tox = 2.0/ax
        m = 2*int((n + int(sqrt(ACC*n)))/2)
        jsum = 0
        bjp = ans = sum = 0.0
        bj = 1.0
        for j = m..-1..1
            bjm = j*tox*bj - bjp
            bjp = bj
            bj = bjm
            if abs(bj) > BIGNO
                bj *= BIGNI
                bjp *= BIGNI
                ans *= BIGNI
                sum *= BIGNI
            endif
            if jsum 
                sum += bj
            endif
            jsum = !jsum
            if j == n 
                ans = bjp
            endif
        end
        sum = 2.0*sum - bj
        ans /= sum
    endif
    y = x < 0.0 && mod(n,2) == 1 ? -ans : ans
end

% Function: bessel_y0
% Returns the Bessel function of the second kind, of order 0 of the specified number.
%
% : function [y : scalar] = __device__ bessel_y0(x : scalar)
function [y : scalar] = __device__ bessel_y0(x : scalar)
    if (x < 8.0)
        y = x*x
        ans1 = -2957821389.0 + y*(7062834065.0 + y*(-512359803.6 + y*(10879881.29 + y*(-86327.92757 + y*228.4622733))))
        ans2 = 40076544269.0 + y*(745249964.8 + y*(7189466.438 + y*(47447.26470 + y*(226.1030244 + y*1.0))))
        y = (ans1/ans2) + 0.636619772*bessel_j0(x)*log(x)
    else
        z = 8.0/x
        y = z*z
        xx = x - 0.785398164

        ans1 = 1.0 + y*(-0.1098628627e-2 + y*(0.2734510407e-4 + y*(-0.2073370639e-5 + y*0.2093887211e-6)))
        ans2 = -0.1562499995e-1 + y*(0.1430488765e-3 + y*(-0.6911147651e-5 + y*(0.7621095161e-6 + y*(-0.934945152e-7))))
        y = sqrt(0.636619772/x)*(sin(xx)*ans1 + z*cos(xx)*ans2)
    endif
end

% Function: bessel_y1
% Returns the Bessel function of the second kind, of order 1 of the specified number.
%
% : function [y : scalar] = __device__ bessel_y1(x : scalar)
function [y : scalar] = __device__ bessel_y1(x : scalar)
    if (x < 8.0)
        y = x*x
        ans1 = x*(-0.4900604943e13 + y*(0.1275274390e13 + y*(-0.5153438139e11 + y*(0.7349264551e9 + y*(-0.4237922726e7 + y*0.8511937935e4)))))
        ans2 = 0.2499580570e14 + y*(0.4244419664e12 + y*(0.3733650367e10 + y*(0.2245904002e8 + y*(0.1020426050e6 + y*(0.3549632885e3 + y)))))
        y = (ans1/ans2) + 0.636619772*(bessel_j1(x)*log(x) - 1.0/x)
    else
        z = 8.0/x
        y = z*z
        xx = x - 2.356194491
        ans1 = 1.0 + y*(0.183105e-2 + y*(-0.3516396496e-4 + y*(0.2457520174e-5 + y*(-0.240337019e-6))))
        ans2 = 0.04687499995 + y*(-0.2002690873e-3 + y*(0.8449199096e-5 + y*(-0.88228987e-6 + y*0.105787412e-6)))
        y = sqrt(0.636619772/x)*(sin(xx)*ans1 + z*cos(xx)*ans2)
    endif
end

    
% Function: bessel_yn
% Returns the Bessel function of the second kind, of order n of the specified number.
%
% : function [y : scalar] = __device__ bessel_yn(x : scalar)
function [y : scalar] = __device__ bessel_yn(n : int, x : scalar)
    if n == 0 
        y = bessel_y0(x)
        return
    endif
    if n == 1
        y = bessel_y1(x)
        return
    endif
    tox = 2.0/x
    by = bessel_y1(x)
    bym = bessel_y0(x)
    for j = 1..n-1
        byp = j*tox*by - bym
        bym = by
        by = byp
    end
    y = by
end

% Function: igam
% Returns the incomplete gamma function.
%
% : function [y : scalar] = __device__ igam(a : scalar, x : scalar)
function [y : scalar] = __device__ igam(a : scalar, x : scalar)
    if x <= 0 || a <= 0
        y = 0.0
        return
    endif

    if x > 1.0 && x > a
        y = 1.0 - igamc(a, x)
        return
    endif

    % Compute  x**a * exp(-x) / gamma(a) 
    ax = a*log(x) - x - lgamma(a)
    if ax < -MAXLOG
        y = 0.0
        return
    endif

    ax = exp(ax)

    % power series
    r = a
    c = 1.0
    ans = 1.0    

    while 1
        r += 1.0
        c *= x/r
        ans += c
        if c/ans <= eps
            break
        endif
    end
    y = ans*ax/a
end

% Function: igamc
% Returns the complemented incomplete gamma function.
%
% : function [y : scalar] = __device__ igamc(a : scalar, x : scalar)
function [y : scalar] = __device__ igamc(a : scalar, x : scalar)
    big = 4.503599627370496e15
    biginv = 2.22044604925031308085e-16    

    if x <= 0 || a <= 0
        y = 1.0
        return
    endif

    if x < 1.0 || x < a 
        y = 1.0 - igam(a, x)
        return
    endif

    ax = a*log(x) - x - lgamma(a)
    if ax < -MAXLOG
        y = 0.0
        return
    endif

    ax = exp(ax)

    % continued fraction
    y = 1.0 - a
    z = x + y + 1.0
    c = 0.0
    pkm2 = 1.0
    qkm2 = x
    pkm1 = x + 1.0
    qkm1 = z*x
    ans = pkm1/qkm1
    t = 1+eps

    while t>eps
        c += 1.0
        y += 1.0
        z += 2.0
        yc = y*c
        pk = pkm1*z - pkm2*yc
        qk = qkm1*z - qkm2*yc
        if qk != 0
            r = pk/qk
            t = abs((ans - r)/r)
            ans = r
        else
            t = 1.0
        endif

        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk
        if abs(pk) > big
            pkm2 *= biginv
            pkm1 *= biginv
            qkm2 *= biginv
            qkm1 *= biginv
        endif
    end
    y = ans*ax
end

% Function: stirf
% Return the gamma function computed by Stirling's formula.
%
% : function [y : scalar] = __device__ stirf(x : scalar)
function [y : scalar] = __device__ stirf(x : scalar)
    STIR = [7.87311395793093628397E-4, -2.29549961613378126380E-4, -2.68132617805781232825E-3, _   
            3.47222221605458667310E-3, 8.33333333333482257126E-2]
    MAXSTIR = 143.01608
    w = 1.0/x
    y = exp(x)

    w = 1.0 + w*polevl(w, STIR, 4)

    if x > MAXSTIR    
        % Avoid overflow in pow
        v = x ^ (0.5*x - 0.25)
        y = v*(v/y)
    else
        y = (x ^ (x - 0.5))/y
    endif
    y = SQTPI*y*w
end

% Function: gamma
% Returns the gamma function of the specified number.
%
% : function [y : scalar] = __device__ gamma(x : scalar)
function [y : scalar] = __device__ gamma(x : scalar)
    P = [1.60119522476751861407E-4, 1.19135147006586384913E-3, 1.04213797561761569935E-2, _
         4.76367800457137231464E-2, 2.07448227648435975150E-1, 4.94214826801497100753E-1, _
         9.99999999999999996796E-1]
    Q = [-2.31581873324120129819E-5, 5.39605580493303397842E-4, -4.45641913851797240494E-3, _
         1.18139785222060435552E-2, 3.58236398605498653373E-2, -2.34591795718243348568E-1, _
         7.14304917030273074085E-2, 1.00000000000000000320E0]

    q = abs(x)
    if q > 33.0
        if x < 0.0        
            p = floor(q)
            if p == q
                y = nan
                return
            endif
            z = q - p
            if z > 0.5
                p += 1.0
                z = q - p
            endif
            z = q*sin(pi*z)
            if z == 0.0 
                y = nan
                return
            endif
            z = abs(z)
            z = pi/(z*stirf(q))
            y = -z        
            return
        else        
            y = stirf(x)
            return
        endif
    endif

    z = 1.0
    while x >= 3.0
        x -= 1.0;
        z *= x;
    end

    while x < 0.0
        if x == 0.0
            y = nan % gamma: singular
            return
        elseif (x > -1.0E-9)
            y = (z/((1.0 + 0.5772156649015329*x)*x))
            return
        endif
        z /= x
        x += 1.0
    end

    while x < 2.0
        if x == 0.0
            y = nan % gamma: singular
            return
        elseif (x < 1.0E-9)        
            y = (z/((1.0 + 0.5772156649015329*x)*x))
            return
        endif
        z /= x
        x += 1.0
    end

    if x == 2.0 || x == 3.0
        y = z
        return
    endif

    x -= 2.0
    p = polevl(x, P, 6)
    q = polevl(x, Q, 7)
    y = z*p/q
end

% Function: chisq
% Returns the area under the left hand tail (from 0 to x)
% of the Chi square probability density function with v degrees of freedom.
%
% : function [y : scalar] = __device__ chisq(df : scalar, x : scalar)
function [y : scalar] = __device__ chisq(df : scalar, x : scalar)
    if x < 0.0 || df < 1.0
        y = 0.0
    else
        y = igam(df/2.0, x/2.0)
    endif
end

% Function: chisqc
% Returns the area under the right hand tail (from x to infinity)
% of the Chi square probability density function with v degrees of freedom:
%
% : function [y : scalar] = __device__ chisqc(df : scalar, x : scalar)
function [y : scalar] = __device__ chisqc(df : scalar, x : scalar)
    if x < 0.0 || df < 1.0
        y = 0.0
    else
        y = igamc(df/2.0, x/2.0)
    endif
end
    
% Function: poisson
% Returns the sum of the first k terms of the Poisson distribution.
%
% : function [y : scalar] = __device__ poisson(k : int, x : scalar)
function [y : scalar] = __device__ poisson(k : int, x : scalar)
    if k < 0 || x < 0
        y = 0.0
    else
        y = igamc(float(k + 1), x)
    endif
end

% Function: poissonc
% Returns the sum of the terms k+1 to infinity of the Poisson distribution.
%
% : function [y : scalar] = __device__ poissonc(k : int, x : scalar)
function [y : scalar] = __device__ poissonc(k : int, x : scalar)
    if k < 0 || x < 0
        y = 0.0
    else
        y = igam(float(k + 1), x)
    endif
end

% Function: cumnormal
% Returns the area under the Gaussian probability density function, integrated from minus infinity to a.
%
% : function [y : scalar] = __device__ cumnormal(a : scalar)
function [y : scalar] = __device__ cumnormal(a : scalar)
    x = a*SQRTH
    z = abs(x)

    if z < SQRTH
        y = 0.5 + 0.5*erf(x)
    else
        y = 0.5*erfc(z)
        if x > 0
            y = 1.0 - y
        endif
    endif   
end

% Function: polevl
% Evaluates polynomial of degree N
%
% : function [ans : scalar] = __device__ polevl(x : scalar, coef : vec, N : int)
function [ans : scalar] = __device__ polevl(x : scalar, coef : vec, N : int)
    ans = coef[0]
    for i=1..N
        ans = ans*x + coef[i]
    end
end

% Function: p1evl
% Evaluates polynomial of degree N with assumtion that coef[N] = 1.0
%
% : function [ans : scalar] = __device__ p1evl(x : scalar, coef : vec, N : int)
function [ans : scalar] = __device__ p1evl(x : scalar, coef : vec, N : int)
    ans = x + coef[0]
    for i=1..N
        ans = ans*x + coef[i]
    end
end

% Function: ibeta
% Returns the incomplete beta function evaluated from zero to xx.
%
% : function [ans : scalar] = __device__ ibeta(aa : scalar, bb : scalar, xx : scalar)
function [ans : scalar] = __device__ ibeta(aa : scalar, bb : scalar, xx : scalar)
    if aa <= 0.0 || bb <= 0.0 
        ans = nan
    elseif xx <= 0.0 || xx >= 1.0
        ans = 0.0
    elseif xx ==1.0
        ans = 1.0
    else
        ans = nan
    else
        flag = 0
        if (bb*xx) <= 1.0 && xx <= 0.95
            ans = pseries(aa, bb, xx)
            return
        endif
        w = 1.0 - xx
        % Reverse a and b if x is greater than the mean. */
        if (xx > (aa/(aa + bb)))
            flag = 1
            a = bb
            b = aa
            xc = xx
            x = w
        else
            a = aa
            b = bb
            xc = w
            x = xx
        endif    
        if flag && (b*x) <= 1.0 && x <= 0.95
            t = pseries(a, b, x)
            if t <= eps
                ans = 1.0 - eps
            else 
                ans = 1.0 - t
            endif
            return
        endif
        % Choose expansion for better convergence.
        y = x*(a + b - 2.0) - (a - 1.0)
        if y < 0.0
            w = incbcf(a, b, x)
        else
            w = incbd(a, b, x)/xc
        endif

        y = a*log(x)
        t = b*log(xc)
        if (a + b) < MAXGAM && abs(y) < MAXLOG && abs(t) < MAXLOG
            t = xc ^ b
            t *= x ^ a
            t /= a
            t *= w
            t *= gamma(a + b)/(gamma(a)*gamma(b))
            if flag
                if t <= eps
                    t = 1.0 - eps
                else 
                    t = 1.0 - t
                endif
            endif
            ans = t
            return
        endif
        % Resort to logarithms.
        y += t + lgamma(a + b) - lgamma(a) - lgamma(b)
        y += log(w/a)
        if y < MINLOG
            t = 0.0
        else
            t = exp(y)
        endif

        if flag
            if t <= eps
                t = 1.0 - eps
            else 
                t = 1.0 - t
            endif
        endif
        ans = t
    endif
end

% Function: incbcf
% Returns the continued fraction expansion #1 for incomplete beta integral.
%
% : function [ans : scalar] = __device__ incbcf(a : scalar, b : scalar, x : scalar)
function [ans : scalar] = __device__ incbcf(a : scalar, b : scalar, x : scalar)
    big = 4.503599627370496e15
    biginv = 2.22044604925031308085e-16

    k1 = a
    k2 = a + b
    k3 = a
    k4 = a + 1.0
    k5 = 1.0
    k6 = b - 1.0
    k7 = k4
    k8 = a + 2.0

    pkm2 = 0.0
    qkm2 = 1.0
    pkm1 = 1.0
    qkm1 = 1.0
    ans = 1.0
    r = 1.0
    
    thresh = 3.0*eps
    for n=0..300
        xk = -(x*k1*k2)/(k3*k4)
        pk = pkm1 + pkm2*xk
        qk = qkm1 + qkm2*xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        xk = (x*k5*k6)/(k7*k8)
        pk = pkm1 + pkm2*xk
        qk = qkm1 + qkm2*xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        if qk != 0
            r = pk/qk
        endif
        if r != 0
            t = abs((ans - r)/r)
            ans = r
        else
            t = 1.0
        endif

        if t < thresh
            break
        endif

        k1 += 1.0
        k2 += 1.0
        k3 += 2.0
        k4 += 2.0
        k5 += 1.0
        k6 -= 1.0
        k7 += 2.0
        k8 += 2.0

        if abs(qk) + abs(pk) > big
            pkm2 *= biginv
            pkm1 *= biginv
            qkm2 *= biginv
            qkm1 *= biginv
        endif
        if abs(qk) < biginv || abs(pk) < biginv
            pkm2 *= big
            pkm1 *= big
            qkm2 *= big
            qkm1 *= big
        endif
    end
end

% Function: incbd
% Returns the continued fraction expansion #2 for incomplete beta integral.
%
% : function [ans : scalar] = incbd(a : scalar, b : scalar, x : scalar)
function [ans : scalar] = __device__ incbd(a : scalar, b : scalar, x : scalar)
    big = 4.503599627370496e15
    biginv = 2.22044604925031308085e-16

    k1 = a
    k2 = b - 1.0
    k3 = a
    k4 = a + 1.0
    k5 = 1.0
    k6 = a + b
    k7 = a + 1.0
    
    k8 = a + 2.0

    pkm2 = 0.0
    qkm2 = 1.0
    pkm1 = 1.0
    qkm1 = 1.0
    z = x/(1.0 - x)
    ans = 1.0
    r = 1.0
    n = 0
    thresh = 3.0*eps

    for n=1..300
        xk = -(z*k1*k2)/(k3*k4)
        pk = pkm1 + pkm2*xk
        qk = qkm1 + qkm2*xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        xk = (z*k5*k6)/(k7*k8)
        pk = pkm1 + pkm2*xk
        qk = qkm1 + qkm2*xk
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk

        if qk != 0 
            r = pk/qk
        endif
        if r != 0
            t = abs((ans - r)/r)
            ans = r        
        else
            t = 1.0
        endif

        if t < thresh
            break
        endif

        k1 += 1.0
        k2 -= 1.0
        k3 += 2.0
        k4 += 2.0
        k5 += 1.0
        k6 += 1.0
        k7 += 2.0
        k8 += 2.0

        if abs(qk) + abs(pk) > big
            pkm2 *= biginv
            pkm1 *= biginv
            qkm2 *= biginv
            qkm1 *= biginv
        endif
        if abs(qk) < biginv || abs(pk) < biginv
            pkm2 *= big
            pkm1 *= big
            qkm2 *= big
            qkm1 *= big
        endif
    end
end

% Function: pseries
% Returns the power series for incomplete beta integral. Use when b*x is small and x not too close to 1.
%
% : function [s : scalar] = __device__ pseries(a : scalar, b : scalar, x : scalar)
function [s : scalar] = __device__ pseries(a : scalar, b : scalar, x : scalar)
    ai = 1.0/a
    u = (1.0 - b)*x
    v = u/(a + 1.0)
    t1 = v
    t = u
    n = 2.0
    s = 0.0
    z = eps*ai
    while abs(v) > z
        u = (n - b)*x/n
        t *= u
        v = t/(a + n)
        s += v
        n += 1.0
    end
    s += t1
    s += ai

    u = a*log(x)
    if (a + b) < MAXGAM && abs(u) < MAXLOG
        t = gamma(a + b)/(gamma(a)*gamma(b))
        s = s*t*x^a    
    else
        t = lgamma(a + b) - lgamma(a) - lgamma(b) + u + log(s)
        if t < MINLOG
            s = 0.0
        else
            s = exp(t)
        endif
    endif
end
