% Type: Stack
%
% Generic implementation of a stack. Allows to push elements
% to the top of the stack and retrieve them (pop).
% 
% This stack implementation is intended to be used mostly
% from host functions (e.g., adding, removing). Once filled,
% the data can be accessed from kernel/device functions.
{!author name="Bart Goossens"}
{!doc category="Data structures"}

type Stack[T] : mutable class
    items       : vec[T]    
    count       : int
    capacity    : int
endtype

% Function: Stack
% 
% Constructor for a stack object
% 
% Usage:
%   : function [y : Stack[T]] = Stack[T](capacity : int)
% 
% Parameters:
% capacity - the initial capacity of the stack (the maximally expected number
% of elements that the stack can contain. Note, if the capacity is exceeded,
% more space for the stack is allocated)
% 
% Returns:
% y - a new list object
function y : Stack[T] = Stack[T](capacity : int = 10)
    y = Stack[T](items:=vec[T](capacity), capacity:=capacity, count:=0)
endfunction

% Function: push
% 
% Pushes an element to the stack. The function uses an atomic operation,
% so it is (partially) thread-safe
% 
% Usage:
%   : function [] = Stack[T].push[T](item : T)
% 
% Parameters:
% item - the item to add to the stack
function [] = push[T](self : ^Stack[T], item : T)
    if self.count >= self.capacity
        self.set_capacity(self.capacity * 2)
    endif
    index = (self.count+=1)
    self.items[index-1] = item
endfunction

function [] = __device__ push[T](self : ^Stack[T], item : T)
    if self.count >= self.capacity
        error "Max capacity reached"
    endif
    index = (self.count+=1)
    self.items[index-1] = item
endfunction

% Function: pop
% 
% Pops an element from the stack
% 
% Usage:
%   : function [] = Stack[T].pop()
function y = pop[T](self : ^Stack[T])
    index = (self.count-=1)
    y = self.items[index]
endfunction

% Function: stack.clear
% 
% Clears the content of this stack
% 
% Usage:
%   : function [] = __device__ Stack[T].clear()
% 
% Parameters:
% self - the stack object
function [] = clear[T](self : ^Stack[T])
    self.count = 0
endfunction

% Function: Stack.to_string
% 
% Converts the Stack to a string representation
% 
% Usage:
%   : function [out] = Stack[T].to_string()
% 
% Returns:
% out - a string representation for the stack
function [s] = to_string[T](self : ^Stack[T])
    s = "["
    for i=0..self.count-1
        if i>0
            s=s+","
        endif
        s=s+sprintf("%g",self.items[i])
    endfor
    s = s+ "]"
endfunction

function [] = main()
    s = Stack[int]()
    s.push(4)
    s.push(5)
    print s.to_string()
    generate = () -> s.pop()
    print generate()
    print generate()
endfunction