% Steerable pyramid transform
import "system.q"
import "dst.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Multiresolution"}

% construction of polar grid
function [r, theta] = polargrid(sz : vec)
    [x, y] = meshgrid((1..sz[1])-sz[1]/2-1,((1..sz[0])-sz[0]/2-1)*(sz[1]/sz[0]))

    r = sqrt(x.^2+y.^2)
    theta = atan2(-y,x)    
end

% design of the steerable filters
function G = steerable_filters(theta, sz, K : int)
    norm_const = factorial(K-1)/sqrt(K*factorial(2*K-2))
    G = cell(K)
    for k=0..K-1
		theta0 = 3*pi/4 + pi*(k-1)/K + pi/(2*K)
        G[k] = pow(2*abs(cos(theta-theta0)),(K-1)*ones(size(theta)))*norm_const
        G[k] = repmat(G[k],[1,1,sz[2]])
    end
end

function H=stp_meyer_radial_wavelet_filters(r,sz,J)

    N=256    
    omega=(0..N-1)*(pi/N)
    v = x -> 3*x.^2-2*x.^3
    
    H0 = (omega < pi/2) + _
         cos((pi/2)*v(2*(omega/pi)-1)).*((pi/2)<=omega).*(omega<=pi)
    G0 = sin((pi/2)*v(2*(omega/pi)-1)).*((pi/2)<=omega).*(omega<=pi) + _
         (omega > pi)

    % cascade algorithm for subsequent scales
    H=cell(J+1)
    H[0] = copy(G0)
    lo = copy(H0)
    for j=1..J-1
        H[j] = copy(lo)
        n=0..N/(2^j)-1
        H[j][n] = H[j][n] .* G0[0..(2^j)..N-1]        
        lo[n]   = lo[n]   .* H0[0..(2^j)..N-1]
        lo[N/(2^j)..N-1] = 0                
    end
    H[J]=lo

    % convert to 2D frequency responses
    r = (r/(sz[0]/2+2))*N
    for j=0..J
        H[j] = repmat(fftshift2(interp1(0..N-1,H[j],r)),[1,1,sz[2]])
    end
end

% Function: steerable_pyramid_analysis
%
% Steerable pyramid transform - analysis
%
% Parameters:
% x  - input image
% J  - number of scales
% K  - number of orientations
% af - analysis wavelet filters
% G - analysis steerable filters
%
% :function w = steerable_pyramid_analysis(x, J, K, af, G)
function w = steerable_pyramid_analysis(x, J, K, af, G)

    sz = size(x,0..2)
    df = [1,1,2,4,8,16]  % decimation factors    
    w_pyr = pyramid_fft(fft2(x),af,df)
    w = cell(J+1) % cell array for the shearlet subbands

    for j=0..J-1
        w[j] = cell(K)
        r0 = 0..df[j]..sz[0]-1
        r1 = 0..df[j]..sz[1]-1
        if df[j]==1
            for k=0..K-1
                w[j][k]=(irealfft2(G[k].*w_pyr[j]))
            end
        else
            for k=0..K-1
                w[j][k]=(irealfft2(G[k][r0,r1,:].*w_pyr[j]))
            end
        endif
    end
    % scaling coefficients
    w[J]=(irealfft2(w_pyr[J]))
end

% Function: steerable_pyramid_synthesis
%
% steerable pyramid transform - synthesis
%
% Parameters:
% x  - input image
% J  - number of scales
% K  - number of orientations
% af - synthesis wavelet filters
% G - synthesis steerable filters
%
% :function y = steerable_pyramid_synthesis(w, J, K, af, G)
function y = steerable_pyramid_synthesis(w, J, K, af, G)
    df = [1,1,2,4,8,16]  % decimation factors    
    sz = size(w[0][0],0..2)
    w_pyr = cell(J+1)

    % scaling coefficients
    w_pyr[J] = fft2(w[J])

    for j=0..J-1
        band = complex(0)
        r0 = 0..df[j]..sz[0]-1
        r1 = 0..df[j]..sz[1]-1
        if df[j]==1
            for k=0..K-1            
                band = band + fft2(w[j][k]) .* G[k]
            end
        else
            for k=0..K-1            
                band = band + fft2(w[j][k]) .* G[k][r0,r1,:]
            end
        endif
        w_pyr[j] = band
    end
    y = (irealfft2(ipyramid_fft(w_pyr,af,df)))
end

function [] = main()
    img = imread("lena_big.tif")
    img = img[:,:,1]

    sz = size(img,0..2)
    bw = pi/2
    K=8
    J=4

    % computation of the filters
    tic()
    [r, theta] = polargrid(sz)
    G = steerable_filters(fftshift2(theta), sz, K)
    H = stp_meyer_radial_wavelet_filters(r, sz, J)
    toc()

    tic()    
    for k=0..0
        w = steerable_pyramid_analysis(img,J,K,H,G)
        img_out = steerable_pyramid_synthesis(w,J,K,H,G)
    end
    toc()
    imshow(img_out,[])

    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    print "psnr=", psnr(img,img_out), "dB"
end
