%===========================
% String library for Quasar
% Bart Goossens, Sept. 2015
%===========================
import "system.q"

{!author name="Bart Goossens"}
{!doc category="General"}

% Function: string.indexof
% 
% Returns the index of the first occurence of a character within a string
% 
% Usage:
%   : function [y] = string.indexof(self : string, c : string)
% 
% Parameters:
% self - the string to search through
% c - the character to be found
% 
% Returns:
% y - the index of the found character, or -1 if no occurence was found.
function y = indexof(self : string, c : string)
    if numel(c)>1
        error "(indexof) Second argument should be a one-character string!"
    endif
    y = find(self[:] == c[0], "first")
    y = size(y,0) == 0 ? -1 : y[1]
end

% Function: string.lastindexof
% 
% Returns the index of the last occurence of a character within a string
% 
% Usage:
%   : function [y] = string.lastindexof(self : string, c : string)
% 
% Parameters:
% self - the string to search through
% c - the character to be found
% 
% Returns:
% y - the index of the found character, or -1 if no occurence was found.
function y = lastindexof(self : string, c : string)
    if numel(c)>1
        error "(lastindexof) Second argument should be a one-character string!"
    endif
    y = find(self[:] == c[0], "last")
    y = size(y,0) == 0 ? -1 : y[1]
end


% Function: string.substr
% 
% Returns a part of the string
% 
% Usage:
%   : function [y] = string.substr(self : string, b:int, e:int=-1)
% 
% Parameters:
% self - the string to start from
% b - the index of the first character
% e - the index of the first character NOT to be in the string
% 
% Returns:
% y - the required substring
function y = substr(self : string, beginning:int, ending:int = -1)
    if ending<0
        ending = numel(self)
    endif
    if numel(self) <= beginning || beginning < 0 || numel(self)<ending
        error "(substr) Indices out of range!"
    endif
    if beginning == ending-1
        y = tounicode([self[beginning]])
    else
        y = tounicode(self[beginning..(ending-1)])
    endif
end

% Function: string.toupper
% 
% Converts the specified string to upper case
% 
% Usage:
%   : function [y] = string.toupper(self : string)
% 
% Parameters:
% self - the string to convert
% 
% Returns:
% y - the string converted to upper case
function y = toupper(self : string)
    A = self[:]
    A -= (A >= 97 && A <= 122) * 32
    y = tounicode(A)
end

% Function: string.tolower
% 
% Converts the specified string to lower case
% 
% Usage:
%   : function [y] = string.tolower(self : string)
% 
% Parameters:
% self - the string to convert
% 
% Returns:
% y - the string converted to lower case
function y = tolower(self : string)
    A = self[:]
    A += (A >= 65 && A <= 90) * 32
    y = tounicode(A)
end

% Function: string.reverse
% 
% Returns a string in which the order of the characters is reversed
% 
% Usage:
%   : function [y] = string.reverse(self : string)
% 
% Parameters:
% self - the string to reverse
% 
% Returns:
% y - the reversed string
function y = reverse(self : string)
    y = tounicode(self[numel(self)-1..-1..0])
end

% Function: string.findstring
% 
% Finds all occurences of the specified substring within a string. Returns a vector
% with the indices of the occurences.
% 
% Usage:
%   : function [ind] = string.findstring(self : string, search_for : string)
% 
% Parameters:
% self - the string to search through
% search_for - the substring to be found
% 
% Returns:
% ind - a vector with the indices of the occurences
function ind:vec = findstring(self : string, search_for : string)
    X = [self[:]]
    S = [search_for[:]]
    if numel(S) > numel(X)
        ind = zeros(0)
        return
    endif
    ind = zeros(numel(X))
    j = 0
    {!interpreted for}
    for i=0..numel(X)-numel(S)
        if all(X[i+(0..numel(S)-1)] == S)
            ind[j] = i
            j += 1
        endif
    end
    if j > 0
        ind = ind[0..j-1]
    else
        ind = zeros(0)
    endif
end

% Function: string.replace
% 
% Replaces all occurences of a substring within a string by a replacement stringh
% 
% Usage:
%   : function [y] = string.replace(self : string, search_for : string, replace_by : string)
% 
% Parameters:
% self - the string to be changed
% search_for - the search string
% replace_by - the replacement string
% 
% Returns:
% y - an output string in which all occurences of `search_for` are replaced by `replace_by`
function y = replace(self : string, search_for : string, replace_by : string)
    X = self[:] 
    S = search_for[:]
    R = replace_by[:]
    ind = self.findstring(search_for)
        
    if numel(ind)>0
        Y = zeros(numel(X) + numel(ind) * (numel(R) - numel(S)))
        [j,k] = [ind[0],ind[0]]      
        if ind[0] > 0
            Y[0..j-1] = X[0..k-1]
        endif
        for i=0..numel(ind)-1
            Y[j+(0..numel(R)-1)] = R
            j += numel(R)
            k += numel(S)
            if i<numel(ind)-2
                n = ind[i+1]-ind[i]-numel(S)
                Y[j+(0..n-1)] = X[k+(0..n-1)]
                j += n
                k += n
            endif
        end
        n = numel(X)-ind[numel(ind)-1]-numel(S)
        Y[j+(0..n-1)] = X[k+(0..n-1)]        
        y = tounicode(Y)
    else
        y = self
    endif
end

% Function: string.split
% 
% Splits a string according to the specified delimiter. The delimiter may contain
% one or multiple characters. In case of multiple characters, the function considers
% the delimiter to be a substring. For example "1, 2".split(", ") returns {"1","2"}
% 
% Usage:
%   : function [y : vec[string]] = string.split(self : string, delim : string)
% 
% Parameters:
% self - the string to split
% delim - the delimiter string to be used for splitting
% 
% Returns:
% y - a vector with the splitted strings
% 
% Notes:
% The reverse operation is <vec[string].join>
function y : vec[string] = split(self : string, delim : string)
    X = self[:]
    ind = self.findstring(delim)
    y = vec[string](numel(ind)+2)
    if numel(ind) > 0
        i = 0
        if ind[0]>0
            y[i] = tounicode(X[0..ind[0]-1])
            i += 1
        endif
        for j=1..numel(ind)-1
            y[i] = tounicode(X[ind[j-1]+numel(delim)..ind[j]-1])
            i += 1
        end
        n = numel(X)-ind[numel(ind)-1]-numel(delim)
        if n > 0
            y[i] = tounicode(X[ind[numel(ind)-1]+numel(delim)+(0..n-1)])
            i += 1
        endif
        y = y[0..i-1]
    else
        y[0] = self
    endif
end

% Function: vec[string].join
% 
% Joins different strings using the specified delimiter
% 
% Usage:
%   : function [y : string] = vec[string].join(self : vec[string], delim : string)
% 
% Parameters:
% self - a vector of strings to be joined
% delim - the delimiter string to be used for joining
% 
% Returns:
% y - the joined string
% 
% Notes:
% The reverse operation is <string.split>
function y : string = join(self : vec[string], delim : string)
    n = 0
    {!interpreted for}
    for i=0..numel(self)-1
        n += numel(self[i])
    end    
    Y = zeros(n+(numel(self)-1)*numel(delim))
    D = delim[:]
    j = 0
    {!interpreted for}
    for i=0..numel(self)-1
        Y[j+(0..numel(self[i])-1)] = self[i][:]
        j += numel(self[i])
        if i<numel(self)-1
            Y[j+(0..numel(delim)-1)] = D
            j += numel(delim)
        endif
    end
    y = tounicode(Y)
end

% Function: string.trim
% 
% Trims the specified characters (usually whitespace) from the input string, both on
% the left side as well as on the right side.
% 
% Usage:
%   : function [y] = string.trim(self : string, trimchars : string)
% 
% Parameters:
% self - the string to be trimmed
% trimchars - the set of characters to be trimmed. May be a string with multiple characters.
% 
% Returns:
% y - the trimmed string
function y = trim(self : string, trimchars : string)
    X = self[:]
    T = [trimchars[:]]
    for i=0..numel(X)-1
        if numel(find(T == X[i], "first")) == 0
            break
        endif
    end
    for j=numel(X)-1..-1..i
        if numel(find(T == X[j], "first")) == 0
            break
        endif        
    end
    y = tounicode(X[i..j])
end

function [] = main()
    print cat("ab","cd","ef")
    print vertcat("Line 1", "Line 2")
    print size("test")
    print "description = {".findstring("=")
    print "test".indexof("s"), ",", "test".lastindexof("t")   
    print "azerty".toupper(), " IS GREAT".tolower()
    print "anagram".reverse()
    print "acrabababas".replace("ba", "x")
    print "<","  hello  ".trim(" "), ">"
    print "12,45,there".split(",")
    print {"12","45"}.join(":")
    print "hi there".substr(3)
end