%REQUIRES_DISPLAY
import "colormap.q"
import "Quasar.UI.dll"

type color4 : class
    r : scalar
    g : scalar
    b : scalar
    a : scalar    
end

type vertex : class
    position : vec3
    normal : vec3
    color : color4
end

type vertex_buffer : vec[vertex]

function [] = surf3d(x : mat, y : mat, z : mat)

    % Fill the vertex buffer with vertices. Note: for each quad, we will use 6 vertices...
    % Spherical color mapping
    function [] = __kernel__ fill_spherical(vert : vertex_buffer, x : mat'unchecked, y : mat'unchecked, z : mat'unchecked, scale : vec3, cmap : mat'unchecked, range : vec2, pos : ivec2)
		idx = 6*(pos[0] * (size(z,1)-1) + pos[1])
		
		% Determine the corner coordinates of each quad
		v1 = [x[pos],       y[pos],         z[pos]]       .* scale
		v2 = [x[pos+[0,1]], y[pos+[0,1]],   z[pos+[0,1]]] .* scale
		v3 = [x[pos+[1,1]], y[pos+[1,1]],   z[pos+[1,1]]] .* scale
		v4 = [x[pos+[1,0]], y[pos+[1,0]],   z[pos+[1,0]]] .* scale
        
        norm = (v : vec3) -> sqrt(sum(v.^2))

		% Compute the colors of each vertex
        c1 = cmap[norm(v1) / range[1]*(size(cmap,0)-1), 0..2]
        c2 = cmap[norm(v2) / range[1]*(size(cmap,0)-1), 0..2]
        c3 = cmap[norm(v3) / range[1]*(size(cmap,0)-1), 0..2]
        c4 = cmap[norm(v4) / range[1]*(size(cmap,0)-1), 0..2]

		% Compute the normals
		n1 = [0.,0.,1.]
		n2 = [0.,0.,1.]
		n3 = [0.,0.,1.]
		n4 = [0.,0.,1.]

		% Generate the vertex list
		vert[idx+0] = vertex(position:=v1,normal:=n1,color:=color4(r:=c1[0],g:=c1[1],b:=c1[2],a:=1.)) % Triangle 1 = [v1,v2,v3]
		vert[idx+1] = vertex(position:=v2,normal:=n2,color:=color4(r:=c2[0],g:=c2[1],b:=c2[2],a:=1.))
		vert[idx+2] = vertex(position:=v3,normal:=n3,color:=color4(r:=c3[0],g:=c3[1],b:=c3[2],a:=1.))
		vert[idx+3] = vertex(position:=v3,normal:=n3,color:=color4(r:=c3[0],g:=c3[1],b:=c3[2],a:=1.)) % Triangle 2 = [v3,v4,v1]
		vert[idx+4] = vertex(position:=v4,normal:=n4,color:=color4(r:=c4[0],g:=c4[1],b:=c4[2],a:=1.))
		vert[idx+5] = vertex(position:=v1,normal:=n1,color:=color4(r:=c1[0],g:=c1[1],b:=c1[2],a:=1.))
	end

    % Custom color mapping
    function [] = __kernel__ fill_custom(vert : vertex_buffer, x : mat'unchecked, y : mat'unchecked, z : mat'unchecked, scale : vec3, c : mat'unchecked, cmap : mat, pos : ivec2)
		idx = 6*(pos[0] * (size(z,1)-1) + pos[1])
		
		% Determine the corner coordinates of each quad
		v1 = [x[pos],       y[pos],         z[pos]]       .* scale
		v2 = [x[pos+[0,1]], y[pos+[0,1]],   z[pos+[0,1]]] .* scale
		v3 = [x[pos+[1,1]], y[pos+[1,1]],   z[pos+[1,1]]] .* scale
		v4 = [x[pos+[1,0]], y[pos+[1,0]],   z[pos+[1,0]]] .* scale      

		% Compute the colors of each vertex
        c1 = cmap[c[pos],       0..2]
        c2 = cmap[c[pos+[0,1]], 0..2]
        c3 = cmap[c[pos+[1,1]], 0..2]
        c4 = cmap[c[pos+[1,0]], 0..2]

		% Compute the normals
		n1 = [0.,0.,1.]
		n2 = [0.,0.,1.]
		n3 = [0.,0.,1.]
		n4 = [0.,0.,1.]

		% Generate the vertex list
		vert[idx+0] = vertex(position:=v1,normal:=n1,color:=color4(r:=c1[0],g:=c1[1],b:=c1[2],a:=1.)) % Triangle 1 = [v1,v2,v3]
		vert[idx+1] = vertex(position:=v2,normal:=n2,color:=color4(r:=c2[0],g:=c2[1],b:=c2[2],a:=1.))
		vert[idx+2] = vertex(position:=v3,normal:=n3,color:=color4(r:=c3[0],g:=c3[1],b:=c3[2],a:=1.))
		vert[idx+3] = vertex(position:=v3,normal:=n3,color:=color4(r:=c3[0],g:=c3[1],b:=c3[2],a:=1.)) % Triangle 2 = [v3,v4,v1]
		vert[idx+4] = vertex(position:=v4,normal:=n4,color:=color4(r:=c4[0],g:=c4[1],b:=c4[2],a:=1.))
		vert[idx+5] = vertex(position:=v1,normal:=n1,color:=color4(r:=c1[0],g:=c1[1],b:=c1[2],a:=1.))	    
    end
    
    function [] = __kernel__ fill_grid_hor(vert : vertex_buffer, x : mat'unchecked, y : mat'unchecked, z : mat'unchecked, scale : vec3, grid_size : int, pos : ivec2)
        idx = 2*(pos[0] * (size(z,1)-1) + pos[1])
        color = color4(r:=0.75,g:=0.75,b:=0.75,a:=1.)
        
        p = [pos[0]*(size(x,0)-2)/grid_size,pos[1]]
        v1 = [x[p],       y[p],         z[p]]       .* scale + [0.,0.,0.001]
        v2 = [x[p+[0,1]], y[p+[0,1]],   z[p+[0,1]]] .* scale + [0.,0.,0.001]
        
        vert[idx+0] = vertex(position:=v1,normal:=[0.,0.,0.],color:=color)
        vert[idx+1] = vertex(position:=v2,normal:=[0.,0.,0.],color:=color)
    end
    
    function [] = __kernel__ fill_grid_vert(vert : vertex_buffer, x : mat'unchecked, y : mat'unchecked, z : mat'unchecked, scale : vec3, grid_size:int, pos : ivec2)
        idx = 2*(pos[0] * grid_size + pos[1])
        color = color4(r:=0.75,g:=0.75,b:=0.75,a:=1.)
        
        p = [pos[0],pos[1]*(size(x,1)-2)/grid_size]
        v1 = [x[p],       y[p],         z[p]]       .* scale + [0.,0.,0.001]
        v2 = [x[p+[1,0]], y[p+[1,0]],   z[p+[1,0]]] .* scale + [0.,0.,0.001]
        
        vert[idx+0] = vertex(position:=v1,normal:=[0.,0.,0.],color:=color)
        vert[idx+1] = vertex(position:=v2,normal:=[0.,0.,0.],color:=color)
    end
            
    % Calculate the local gradient
    function y = gradient(x : mat)        
        y = zeros(size(x))
        parallel_do(size(y),__kernel__ (pos : ivec2) -> y[pos] = sqrt((x[pos+[0,1]]-x[pos])^2+(x[pos+[1,0]]-x[pos])^2))
    end

    % Convert the color map to a 256x3 matrix
    map = colormap("hsv")
    map = squeeze(cmap(linspace(1, 0, 256), map, [0, 1]))/255.0

    % Calculate the scale
    scale = [0.5/max(abs(x)), 0.5/max(abs(y)), 0.5/max(abs(z))]

    % Allocate the vertex buffer
    vert = vertex_buffer(6*(size(z,0)-1)*(size(z,1)-1))
    % Fill the vertex buffer
    %parallel_do(size(z,0..1)-1,vert,x,y,z,scale,map,[0,max(sqrt(x.^2+y.^2+z.^2))],fill_spherical)
    grad = gradient(z)
    parallel_do(size(z,0..1)-1,vert,x,y,z,scale,grad/max(grad)*255,map,fill_custom)
    
    % Grid lines
    grid_size = 40
    grid_hor = vertex_buffer(2*grid_size*(size(z,1)-1))
    grid_ver = vertex_buffer(2*grid_size*(size(z,0)-1))
    parallel_do([grid_size, size(z,1)-1], grid_hor,x,y,z,scale,grid_size,fill_grid_hor)
    parallel_do([size(z,0)-1,grid_size], grid_ver,x,y,z,scale,grid_size,fill_grid_vert)

    axes = new(qvectorlayer3d)
    axes.setcolor("black")
    %axes.drawline([0, 0, 0], [0.6, 0, 0])
    %axes.drawline([0, 0, 0], [0.0, 0.6, 0])
    %axes.drawline([0, 0, 0], [0.0, 0.0, 0.6])
    %axes.print([0.65, 0, 0], "x")
    %axes.print([0, 0.65, 0], "y")
    %axes.print([0, 0, 0.65], "z")
    axes.drawbox([-0.5,-0.5,-0.5],[0.5,0.5,0.5])
    axes.print([0,0.65,-0.5],"x")
    axes.print([-0.65,0,-0.5],"y")
    axes.print([0.65,0.5,0],"z")

    % axis labels
    for k=linspace(-0.5,0.5,10)
        axes.print([k,0.56,-0.50], sprintf("%.03f",k*0.5/scale[0]))
        axes.print([-0.56,k,-0.50], sprintf("%.03f",k*0.5/scale[1]))
        axes.print([0.56,0.56,k], sprintf("%.03f",k*0.5/scale[2]))
        axes.drawline([k,0.5,-0.5],[k,0.52,-0.5])
        axes.drawline([-0.5,k,-0.5],[-0.52,k,-0.5])
        axes.drawline([0.5,0.5,k],[0.52,0.50,k])
    end
    
    % box
    axes.setcolor("lightgray")
    for k=linspace(-0.5,0.5,10)
        if k>-0.5 && k<0.5
            axes.drawline([k,-0.5,-0.5],[k,0.5,-0.5])
            axes.drawline([-0.5,k,-0.5],[0.5,k,-0.5])
                        
            axes.drawline([0.5,-0.5,k],[0.5,0.5,k])
            axes.drawline([0.5,k,-0.5],[0.5,k,0.5])
            
            axes.drawline([-0.5,-0.5,k],[0.5,-0.5,k])
            axes.drawline([k,-0.5,-0.5],[k,-0.5,0.5])
        endif
    end
    

    frm = form("Surface plot")
    frm.width = 800
    disp = frm.add_display()
    disp.width = 1024
    disp.height = 768
    renderer = disp.create_opengl_renderer()
    renderer.background_color = "white"
    renderer.enable_zbuffer = true
    renderer.draw_backfaces = true
    renderer.add(vert, "triangles")
    renderer.add(grid_hor, "lines")
    renderer.add(grid_ver, "lines")
    renderer.add(axes, "layer")
    renderer.pitch = 340
    frm.show()
    frm.wait()

end

function [] = main()

    [x, y] = meshgrid(linspace(-0.5,0.5,128),linspace(-0.5,0.5,128))
    a = 4
    z = exp(-4*sqrt(x.^2+y.^2)) .* cos(4*pi*x) %  -sin(a*pi*sqrt(x.^2+y.^2))./(a*pi*sqrt(x.^2+y.^2))

    surf3d(x,y,z)
end