%============================================
% Quasar system library
% 2011-2018 - Bart Goossens
%============================================

{!author name="Bart Goossens"}
{!doc category="General"}

% Section: system.q
% The system library for Quasar: contains functions that are not built-in,
% but that are very often used. 
%
% If the function you are looking for, cannot be found in system.q (or any
% of the other .q files located in the Library directory), please feel free to 
% implement this function, and send a copy of the code to 
% bart.goossens@telin.ugent.be. This way, I can make sure that the function
% is distributed automatically in later versions of Quasar.

%
% Group: type-checking functions
%
% Functions: type-checking functions
% isreal - returns 1 if x is real-valued, 0 otherwise
% iscomplex - returns 1 if x is complex-valued, 0 otherwise
% isscalar - returns 1 if x is scalar, 0 otherwise
% isvector - returns 1 if x is a vector (max. dimensionality 1), 0 otherwise
% ismatrix - returns 1 if x is a matrix (max. dimensionality 2), 0 otherwise
% iscube - returns 1 if x is a cube (max. dimensionality 3), 0 otherwise
% isfunction - returns 1 if x is a function, 0 otherwise
% isstring - returns 1 if x is a string, 0 otherwise
% isobject - returns 1 if x is an object, 0 otherwise
% iskernel - returns 1 if x is a kernel function, 0 otherwise
% iscell - returns 1 if x is a cell matrix, 0 otherwise
isreal      = x -> type(x, "scalar")  || type(x, "cube{:}")
iscomplex   = x -> type(x, "cscalar") || type(x, "ccube{:}")
isscalar	= x -> type(x, "scalar")  || type(x, "cscalar")
isvector    = x -> type(x, "vec")     || type(x, "cvec")
ismatrix	= x -> type(x, "mat")     || type(x, "cmat")
iscube      = x -> type(x, "cube")    || type(x, "ccube")
isncube     = x -> type(x, "cube{:}") || type(x, "ccube{:}")
isfunction	= x -> type(x, "function")
isstring	= x -> type(x, "string")
isobject    = x -> type(x, "object")
iskernel    = x -> type(x, "kernel_function")
iscell      = x -> type(x, "cell") && !isncube(x)

% Group: mathematical functions
%
% Functions: mathematical functions
% nextpow2 - returns the next integer power of two
% prevpow2 - returns the previous integer power of two
% std - computes the standard deviation of all elements in a matrix
% lerp - performs linear interpolation between a and b (with weight d)
nextpow2    = x -> exp2(ceil(log2(x)))
prevpow2    = x -> exp2(floor(log2(x)))
std         = x -> sqrt(sum((x - mean(x)).^2)/(numel(x)-1))
lerp        = (a, b, d) -> a + (b - a) * d

% Group: conversion functions
%
% Functions: conversion functions
% num2str - converts a number to string
% str2num - converts the string to a numerical value
num2str     = (x : scalar) -> sprintf("%f", x)
str2num     = (x : string) -> sscanf(x, "%f")[0]

% Group: general matrix functions
%
% Functions: general matrix functions
% isempty - returns 1 if the number of elements of x is 0 and 0 otherwise.
% ndims - returns the number of dimensions of x
% flipup - flips a two dimensional matrix upside-down (vertically)
% fliplr - flips a two dimensional matrix horizontally
% any - returns 1 if at least one element of x is non-zero, otherwise 0.
% all - returns 1 if all elements of x are non-zero, otherwise 0.
isempty		= x -> numel(x) == 0
flipud      = x -> x[size(x,0)-1..-1..0,:,:]
fliplr      = x -> x[:,size(x,1)-1..-1..0,:]
any         = x -> max(x != 0) == 1
all         = x -> numel(x) == 0 || min(x != 0) == 1
reverse     = x -> x[numel(x)-1..-1..0]

% Optimizations
reduction x         -> complex(zeros(x)) = czeros(x)
reduction (x,y)     -> complex(zeros(x,y)) = czeros(x,y)
reduction (x,y,z)   -> complex(zeros(x,y,z)) = czeros(x,y,z)
reduction x         -> complex(uninit(x)) = cuninit(x)
reduction (x,y)     -> complex(uninit(x,y)) = cuninit(x,y)
reduction (x,y,z)   -> complex(uninit(x,y,z)) = cuninit(x,y,z)
reduction x         -> complex(shared(x)) = cshared(x)
reduction (x,y)     -> complex(shared(x,y)) = cshared(x,y)
reduction (x,y,z)   -> complex(shared(x,y,z)) = cshared(x,y,z)
reduction (x) -> real(ifft1(x)) = irealfft1(x)
reduction (x) -> real(ifft2(x)) = irealfft2(x)
reduction (x) -> real(ifft3(x)) = irealfft3(x)
reduction (x) -> real(ifftn(x)) = irealfftn(x)
reduction (x,d) -> real(ifftn(x,d)) = irealfftn(x,d)
reduction x -> fft1(complex(x)) = fft1(x)
reduction x -> fft2(complex(x)) = fft2(x)
reduction x -> fft3(complex(x)) = fft3(x)
reduction x -> fftn(complex(x)) = fftn(x)
reduction (x,d) -> fftn(complex(x),d) = fftn(x,d)
reduction x -> fft1(ifft1(x)) = x
reduction x -> fft2(ifft2(x)) = x
reduction x -> ifft2(fft2(x)) = x
reduction x -> fft3(ifft3(x)) = x
reduction x -> ifft3(fft3(x)) = x
reduction x -> fftn(ifftn(x)) = x
reduction x -> ifftn(fftn(x)) = x
reduction (x,d) -> fftn(ifftn(x,d),d) = x
reduction (x,d) -> ifftn(fftn(x,d),d) = x
reduction x -> irealfft1(fft1(x)) = real(x)
reduction x -> irealfft2(fft2(x)) = real(x)
reduction x -> irealfft3(fft3(x)) = real(x)
reduction x -> irealfftn(fftn(x)) = real(x)
reduction (x,d) -> irealfftn(fftn(x,d),d) = real(x)
reduction (x:scalar) -> real(x) = x
reduction (x:mat) -> real(x) = x
reduction (x:cube) -> real(x) = x
reduction (x:scalar) -> imag(x) = 0
reduction (x:mat) -> imag(x) = zeros(size(x))
reduction (x:cube) -> imag(x) = zeros(size(x))
reduction (x:mat) -> transpose(transpose(x)) = x
reduction (x:cmat) -> transpose(transpose(x)) = x
reduction (x) -> x*conj(x) = abs(x)^2
reduction (x) -> x.*conj(x) = abs(x).^2

reduction (x : scalar)  -> x[:] = [x]
reduction (x : vec)     -> x[:] = x
reduction (x : cube{:}) -> x[:] = reshape(x,[1,numel(x)])
reduction (x : cube{:}) -> transpose(x[:]) = reshape(x,[numel(x),1])
reduction (x : cscalar) -> x[:] = [x]
reduction (x : cvec)    -> x[:] = x
reduction (x : ccube{:})-> x[:] = reshape(x,[1,numel(x)])
reduction (x : ccube{:})-> transpose(x[:]) = reshape(x,[numel(x),1])
reduction (x : cube{:}[??])-> x[:] = reshape(x,[1,numel(x)])
reduction (x, sz)       -> size(reshape(x, sz)) = sz
reduction (x, sz)       -> numel(reshape(x, sz)) = prod(sz)

% Logical indexing functions
type matrix_type : [vec|mat|cube|cvec|cmat|ccube]
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a > b] = c) = (x += (c - x) .* (a > b))
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a < b] = c) = (x += (c - x) .* (a < b))
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a <= b] = c) = (x += (c - x) .* (a <= b))
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a >= b] = c) = (x += (c - x) .* (a >= b))
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a == b] = c) = (x += (c - x) .* (a == b))
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a != b] = c) = (x += (c - x) .* (a != b))
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a && b] = c) = (x += (c - x) .* (a && b))
reduction (x : matrix_type, a : matrix_type, b, c) -> (x[a || b] = c) = (x += (c - x) .* (a || b))


%==================
% system functions
%==================

% Function: diagm
%
% Creates a diagonal matrix from a one dimensional vector. The non-diagonal
% elements are set to zero.
%
% : function y : mat = diagm(x : vec)
% : function y : mat = __device__ diagm(x : vec)
%
% Parameters:
%   x - a vector containing the diagonal elements
%
% See also:
%   <diag>
%
function y = diagm[T](x : vec[T]'static)
    N = numel(x)
    y = uninit[T](N,N)
    {!parallel for}
    for m=0..N-1
        for n=0..N-1
            y[m,n] = (m == n) ? x[m] : 0.0
        endfor
    endfor
endfunction

function y = __device__ diagm[T](x : vec[T]'static)
    N = numel(x)
    y = uninit[T](N,N)
    {!parallel for}
    for m=0..N-1
        for n=0..N-1
            y[m,n] = (m == n) ? x[m] : 0.0
        endfor
    endfor
endfunction

function y = __device__ diag(x)
    assert(size(x,0)==size(x,1), "diag: square matrix expected!")
    N = size(x,0)
    y = vec(N)
    {!parallel for}
    for m=0..N-1
        y[m] = x[m,m]
    endfor
endfunction

% Function: diagm
%
% Creates a diagonal matrix from a one dimension
%
% : function y : cmat = diagm(x : cvec)
% : function y : cmat = __device__ diagm(x : cvec)
%
% Parameters:
%   x - a vector containing the diagonal elements
%
% See also:
%   <diag>
%
function y : cmat = diagm(x : cvec)
    N = numel(x)
    y = cuninit(N,N)
    {!parallel for}
    for m=0..N-1
        for n=0..N-1
            y[m,n] = (m == n) ? x[m] : complex(0.0)
        endfor
    endfor
endfunction

reduction (x : [vec|cvec]) -> diag(x) = diagm(x)

% Function: diag
%
% Extracts the diagonal elements of a square real-valued matrix
%
% : function y : vec = diag(x : mat)
%
% Parameters:
%   x - a square matrix to extract the diagonal elements from.
%
% See also:
%   <diagm>
function y : vec = diag(x : mat)
    assert(size(x,0)==size(x,1), "diag: square matrix expected (size=",size(x),")!")
    N = size(x,0)
    y = vec(N)
    {!parallel for}
    for m=0..N-1
        y[m] = x[m,m]
    endfor
endfunction

% Function: diag
%
% Extracts the diagonal elements of a square complex-valued matrix
%
% : function y : cvec = diag(x : cmat)
%
% Parameters:
%   x - a square matrix to extract the diagonal elements from.
%
% See also:
%   <diagm>
function y : cvec = diag(x : cmat)
    assert(size(x,0)==size(x,1), "diag: square matrix expected (size=",size(x),")!")
    N = size(x,0)
    y = complex(zeros(N))
    {!parallel for}
    for m=0..N-1
        y[m] = x[m,m]
    endfor
endfunction

% Function: trace
%
% Computes the trace of a matrix (i.e. the sum of the diagonal elements)
%
% : function y = trace(x)
%
% Parameters:
%   x - a square input matrix
%
reduction (x : mat)  -> trace(x) = sum(diag(x))
reduction (x : cmat) -> trace(x) = sum(diag(x))

% Function: circshift
%
% Applies circular (periodic) shifting of a given vector,
%   according to the relationship y[p] = x[p+d]
%
% : function y : vec = circshift(x : vec, d : ivec1)
%
% Parameters:
%   x - an input vector
%   d - an integer displacement
%
function y : vec = circshift(x : vec, d : ivec1)
    y = uninit(size(x))
    parallel_do(numel(x),x,y,d,__kernel__ (x : vec'circular, y : vec'unchecked, d : ivec1, pos : ivec1) -> y[pos] = x[pos+d]) 
endfunction

% Function: circshift
%
% Applies circular (periodic) shifting of a given matrix,
%   according to the relationship y[p] = x[p+d]
%
% : function y : mat = circshift(x : mat, d : ivec2)
%
% Parameters:
%   x - an input matrix
%   d - an integer displacement vector of length 2
%
function y : mat = circshift(x : mat, d : ivec2)
    y = uninit(size(x))
    parallel_do(size(x),x,y,d,__kernel__ (x : mat'circular, y : mat'unchecked, d : ivec2, pos : ivec2) -> y[pos] = x[pos+d]) 
endfunction

% Function: circshift
%
% Applies circular (periodic) shifting of a given cube,
%   according to the relationship y[p] = x[p+d]
%
% : function y : cube = circshift(x : cube, d : ivec3)
%
% Parameters:
%   x - an input cube
%   d - an integer displacement vector of length 3
%
function y : cube = circshift(x : cube, d : ivec3)
    y = uninit(size(x))
    parallel_do(size(x),x,y,d,__kernel__ (x : cube'circular, y : cube'unchecked, d : ivec3, pos : ivec3) -> y[pos] = x[pos+d]) 
endfunction

% Function: circshift
%
% Applies circular (periodic) shifting of a given complex-valued vector,
%   according to the relationship y[p] = x[p+d]
%
% : function y : cvec = circshift(x : cvec, d : ivec1)
%
% Parameters:
%   x - an input vector
%   d - an integer displacement
%
function y : cvec = circshift(x : cvec, d : ivec1)
    y = complex(uninit(size(x)))
    parallel_do(numel(x),x,y,d,__kernel__ (x : cvec'circular, y : cvec'unchecked, d : ivec1, pos : ivec1) -> y[pos] = x[pos+d]) 
endfunction

% Function: circshift
%
% Applies circular (periodic) shifting of a given complex-valued matrix,
%   according to the relationship y[p] = x[p+d]
%
% : function y : cmat = circshift(x : cmat, d : ivec2)
%
% Parameters:
%   x - an input matrix
%   d - an integer displacement vector of length 2
%
function y : cmat = circshift(x : cmat, d : ivec2)
    y = complex(uninit(size(x)))
    parallel_do(size(x),x,y,d,__kernel__ (x : cmat'circular, y : cmat'unchecked, d : ivec2, pos : ivec2) -> y[pos] = x[pos+d]) 
endfunction

% Function: circshift
%
% Applies circular (periodic) shifting of a given complex-valued cube,
%   according to the relationship y[p] = x[p+d]
%
% : function y : ccube = circshift(x : ccube, d : ivec3)
%
% Parameters:
%   x - an input cube
%   d - an integer displacement vector of length 3
%
function y : ccube = circshift(x : ccube, d : ivec3)
    y = complex(uninit(size(x)))
    parallel_do(size(x),x,y,d,__kernel__ (x : ccube'circular, y : ccube'unchecked, d : ivec3, pos : ivec3) -> y[pos] = x[pos+d]) 
endfunction


% Function: fftshift1
%
% Moves the DC components of a 1D discrete Fourier transform (DFT)
% to the center of the data. For higher dimensional data - it is 
% presumed that the FFT data is stored in the last dimension
%
% : function y = fftshift1(x)
%
% Parameters:
%   x - the input data
%
% See also:
%   <ifftshift1>
%
function y = fftshift1(x)
    dim = [0,0,0]
    if size(x,2)==1
        if size(x,1)==1
            dim[0] = floor(size(x,0)/2)
        else
            dim[1] = floor(size(x,1)/2)
        endif
    else
        dim[2] = floor(size(x,2)/2)
    endif
    y = circshift(x, dim)
endfunction

% Function: ifftshift1
%
% The inverse operation of fftshift1
%
% : function y = ifftshift1(x)
%
% Parameters:
%   x - the input data
%
% See also:
%   <fftshift1>
%
function y = ifftshift1(x)
    dim = [0,0,0]
    if size(x,2)==1
        if size(x,1)==1
            dim[0] = ceil(size(x,0)/2)
        else
            dim[1] = ceil(size(x,1)/2)
        endif
    else
        dim[2] = ceil(size(x,2)/2)
    endif
    y = circshift(x, dim)
endfunction

% Function: fftshift2
%
% Moves the DC components of a 2D discrete Fourier transform (DFT)
% to the center of the data
%
% : function y = fftshift2(x)
%
% Parameters:
%   x - the input data
%
% See also:
%   <ifftshift2>
%
function y = fftshift2(x)
    y = circshift(x, [floor(size(x,0)/2),floor(size(x,1)/2),0])
endfunction

% Function: ifftshift2
%
% The inverse operation of fftshift2
%
% : function y = ifftshift2(x)
%
% Parameters:
%   x - the input data
%
% See also:
%   <fftshift2>
%
function y = ifftshift2(x)
    y = circshift(x, [ceil(size(x,0)/2),ceil(size(x,1)/2),0])
endfunction

% Function: fftshift3
%
% Moves the DC components of a 3D discrete Fourier transform (DFT)
% to the center of the data.
%
% : function y = fftshift3(x)
%
% Parameters:
%   x - the input data
%
% See also:
%   <ifftshift3>
%
function y = fftshift3(x)
    y = circshift(x, [int(size(x,0)/2), int(size(x,1)/2), int(size(x,2)/2)])
endfunction

% Function: ifftshift3
%
% The inverse operation of fftshift3
%
% : function y = ifftshift3(x)
%
% Parameters:
%   x - the input data
%
% See also:
%   <fftshift3>
%
function y = ifftshift3(x)
    y = circshift(x, [ceil(size(x,0)/2), ceil(size(x,1)/2), ceil(size(x,2)/2)])
endfunction


% Function: binsearch1d
%
% A 1D binary search algorithm: given a sorted vector x0, find the index of
% the element that is closest to 'value'.
%
% : function [mid : int] = __device__ binsearch1d(x0 : vec'unchecked, value : scalar)
%
% Parameters:
%   x0 - a sorted vector
%
function [mid : int] = __device__ binsearch1d(x0 : vec'unchecked, value : scalar)
    n = numel(x0)
    index = min(max(value, x0[0]), x0[n-1])

    % Do a binary search
    left = 0
    right = n-2
    while right >= left
        mid = int((left + right)/2)   % compute mid-point
        if x0[mid] <= index && index < x0[mid+1]
            % found!
            break
        elseif x0[mid+1] <= index
            left = mid+1
        else %if x0[mid] > index
            right = mid
        endif
    endwhile
endfunction

% Function: interp1_function
%
% Get a device function that can be used for one-dimensional interpolation
%
% : function fn : [__device__ scalar->scalar] = interp1_function(
% :   x0 : vec'unchecked, y0 : vec'unchecked, interpolation : string = "linear", uniform = 1)
%
% Parameters:
%   x0 - x-values of the input vector
%   y0 - y-values of the input vector
%   interpolation : the interpolation method to use
%      o  "nearest" - nearest neighbor interpolation
%      o  "linear"  - linear interpolation
%      o  "cubic"   - cubic interpolation
%   uniform : specifies whether the input data is uniformly spaced (uniform=1) 
%       or sorted (uniform=0). In case of sorted input data, a binary search
%       algorithm will be used.
%
% Remarks:
% The function returns a device function that can subsequently be used in other
% kernel functions. Function closures are used to store the data [x0, y0]
% Note that x0 does not have to be uniformly-spaced. In case of other spacings,
% a binary search algorithm is employed for the look-up.
%
function fn : [__device__ scalar->scalar] = interp1_function( _
    x0 : vec, y0 : vec'unchecked, interpolation : string = "linear", uniform = 1)

    % assume that x0 is an evenly spaced vector!
    assert(numel(x0) >= 2, "interp1: at least 2 elements are required for x0!")
    assert(numel(x0) == numel(y0), "interp1: x0 and y0 should be vectors of the same size!")
    n = int(numel(y0))

    % x0 is a uniformly spaced input vector
    if uniform == true
        step = 1./(x0[1] - x0[0])
        start = x0[0]        

        match interpolation with
        | "nearest" -> fn = __device__ (x : scalar) -> ( _
                                index = int((x - start) * step); _
                                index = min(max(index, 0), n-1); _
                                y0[index])
        | "linear"  -> fn = __device__ (x : scalar) -> ( _
                                p = (x - start) * step; _
                                p = min(max(p, 0.), n-1.); _
                                index = int(p); _
                                fr = frac(p); _
                                y0[index] * (1.0-fr) + y0[min(index+1,int(n)-1)] * fr)
        | "cubic"   -> fn = __device__ (x : scalar) -> ( _
                                p = (x - start) * step; _
                                p = min(max(p, 0.), n-1.); _
                                index = int(p); _
                                fr = frac(p); _
                                ym1 = y0[max(index-1,0)]; 
                                y   = y0[index]; 
                                y1  = y0[min(index+1,n-1)]; 
                                y2  = y0[min(index+2,n-2)]; _
                                a = (3 * (y-y1) - ym1 + y2) * 0.5; _
                                b = 2*y1 + ym1 - (5*y + y2) * 0.5; _
                                c = (y1 - ym1) * 0.5; _
                                y + (((a * fr) + b) * fr + c) * fr)
        | _ -> error("[interp1_function] Invalid interpolation method")
        endmatch
    % assume x0 is sorted...
    % TODO - cubic interpolation
    else
        match interpolation with
        | "nearest"     -> fn = __device__ (x : scalar) -> ( _
                                index = binsearch1d(x0, x); _
                                y0[index])
        | "linear"      -> fn = __device__ (x : scalar) -> ( _
                                index = binsearch1d(x0, x); _
                                alpha = (x - x0[index])/(x0[index+1] - x0[index]); _
                                y0[index] * (1 - alpha) + y0[index+1] * alpha)        
        | _ -> error("[interp1_function] Invalid interpolation method")
        endmatch
    endif  
endfunction

% Function: interp1
%
% Performs one-dimensional interpolation
%
% : function y = interp1(x0 : vec,y0 : vec,x,interpolation="linear",uniform=1)
%
% Parameters:
%   x0 - x-values of the input vector
%   y0 - y-values of the input vector
%   x - the x-values to interpolate
%   interpolation : the interpolation method to use
%      o  "nearest" - nearest neighbor interpolation
%      o  "linear"  - linear interpolation
%      o  "cubic"   - cubic interpolation
%   uniform : specifies whether the input data is uniformly spaced (uniform=1) 
%       or sorted (uniform=0). In case of sorted input data, a binary search
%       algorithm will be used.
% Returns:
%   y - the interpolated vector
function y = interp1(x0 : vec,y0 : vec,x,interpolation="linear",uniform=1)

    y = uninit(size(x))
    
    % assume that x0 is an evenly spaced vector!
    assert(numel(x0) >= 2, "interp1: at least 2 elements are required for x0!")
    assert(numel(x0) == numel(y0), "interp1: x0 and y0 should be vectors of the same size!")
    n = int(numel(y0))

    step = 1./(x0[1] - x0[0])
    start = x0[0]        

    if uniform
        n = int(numel(y0))
        match interpolation with
        | "nearest" ->
            parallel_do(size(x),x,y, __kernel__ (x : cube'unchecked, y : cube'unchecked, pos : ivec3) -> _
                                y[pos] = y0[min(max(int((x[pos] - start) * step), 0), n-1)])
        | "linear" ->
            parallel_do(size(x),x,y, __kernel__ (x : cube'unchecked, y : cube'unchecked, pos : ivec3) -> _
                                (p = min(max((x[pos] - start) * step, 0.), n-1.); _
                                index = int(p); _
                                fr = frac(p); _
                                y[pos] = y0[index] * (1.0-fr) + y0[min(index+1,int(n)-1)] * fr))
        | "cubic" ->
            parallel_do(size(x),x,y, __kernel__ (x : cube'unchecked, y : cube'unchecked, _
                pos : ivec3) -> (p = (x[pos] - start) * step; _
                                p = min(max(p, 0.), n-1.); _
                                index = int(p); _
                                fr = frac(p); _
                                ym1 = y0[max(index-1,0)]; 
                                yy   = y0[index]; 
                                y1  = y0[min(index+1,n-1)]; 
                                y2  = y0[min(index+2,n-2)]; _
                                a = (3 * (yy-y1) - ym1 + y2) * 0.5; _
                                b = 2*y1 + ym1 - (5*yy + y2) * 0.5; _
                                c = (y1 - ym1) * 0.5; _
                                y[pos] = yy + (((a * fr) + b) * fr + c) * fr))
        | _ -> error("[interp1] Invalid interpolation method")
        endmatch
    else
        match interpolation with
        | "nearest" ->
            parallel_do(size(x),x,y, __kernel__ (x : cube'unchecked, y : cube'unchecked, _
                pos : ivec3) -> (index = binsearch1d(x0, x[pos]); _
                                y[pos] = y0[index]))
        | "linear" ->
            parallel_do(size(x),x,y, __kernel__ (x : cube'unchecked, y : cube'unchecked, _
                pos : ivec3) -> (index = binsearch1d(x0, x[pos]); _
                                alpha = (x[pos] - x0[index])/(x0[index+1] - x0[index]); _
                                y[pos] = y0[index] * (1 - alpha) + y0[index+1] * alpha))
        | _ -> error("[interp1] Invalid interpolation method")
        endmatch
    endif    

endfunction

% Function: interp2_function
%
% Get a device function that can be used for two-dimensional interpolation
%
% : function fn : [__device__ vec2->scalar] = interp2_function(
%    x0 : vec'unchecked, y0 : vec'unchecked, z0 : mat'unchecked,
%    interpolation : string = "linear")
%
% Parameters:
%   x0 - x-values of the input vector (along the 2nd dimension)
%   y0 - y-values of the input vector (along the 1st dimension)
%   interpolation : the interpolation method to use
%      o  "nearest" - nearest neighbor interpolation
%      o  "linear"  - linear interpolation
%      o  "cubic"   - cubic interpolation
% Returns:
%   y - the interpolated image
%
% Remarks:
% The function returns a device function that can subsequently be used in other
% kernel functions. Function closures are used to store the data [x0, y0]
%
function fn : [__device__ vec2->scalar] = interp2_function( _
    x0 : vec, y0 : vec, z0 : mat'unchecked, _
    interpolation : string = "linear")
    
    % assume that x0 is an evenly spaced vector!
    assert(max(numel(x0),numel(y0)) >= 2, _
        "interp2: at least 2 elements are required for x0 and y0!")
    assert(size(z0,0..1) == [numel(y0), numel(x0)], _
        "interp2: x0 and y0 should be vectors of the same size!")
    n = size(z0,0..1)
    
    % x0 and y0 are uniformly spaced input vectors
    step = [1./(y0[1] - y0[0]), 1./(x0[1] - x0[0])]
    start = [y0[0], x0[0]]
    assert(type(step, "vec2"))
    assert(type(start, "vec2"))

    match interpolation with
    | "nearest" -> fn = __device__ (x : vec2) -> ( _
                            index = int((x - start) .* step); _
                            index = min(max(index, [0, 0]), int(n)-1); _
                            z0[index[0],index[1]])
    | "linear"  -> fn = __device__ (x : vec2) -> ( _
                            p = (x - start) .* step; _
                            p = min(max(p, [0., 0.]), float(n)-1.); _
                            index = int(p); _
                            index2 = min(index+1,n-1); _
                            fr = frac(p); _
                            z0[index] * (1-fr[0]) * (1-fr[1]) + _
                            z0[index[0],index2[1]] * (1 - fr[0]) * fr[1] + _
                            z0[index2[0],index[1]] * fr[0] * (1 - fr[1]) + _
                            z0[index2] * fr[0] * fr[1])                           
    | _ -> error("[interp2_function] Invalid interpolation method")
    endmatch
endfunction

% Function: interp2
%
% Performs two-dimensional interpolation
%
% : function z : cube = interp2(x0:vec,y0:vec,z0:mat,x:cube,y:cube,interpolation="linear")
%
% Parameters:
%   x0 - vector of x-values 
%   y0 - vector of y-values
%   z0 - matrix with z-values - the points (x0,y0,z0) define a surface
%   x - the x-values to interpolate
%   y - the y-values to interpolate
%   interpolation : the interpolation method to use
%      o  "nearest" - nearest neighbor interpolation
%      o  "linear"  - linear interpolation
%      o  "cubic"   - cubic interpolation
%
function z : cube = interp2(x0:vec,y0:vec,z0:mat,x:cube,y:cube,interpolation="linear")
    assert(size(x)==size(y), "(interp2) x and y should have the same size!")
    
    z = uninit(size(x))
    step = [1./(y0[1] - y0[0]), 1./(x0[1] - x0[0])]
    start = [y0[0], x0[0]]
    n = size(z,0..1)
    
    match interpolation with
    | "nearest" ->
        parallel_do(size(x),x,y,z,__kernel__ (x : cube'unchecked, y : cube'unchecked, z : cube'unchecked, pos : ivec3) -> _
                            (index = int(([x[pos], y[pos]] - start) .* step); _
                            index = min(max(index, [0, 0]), int(n)-1); _
                            z[pos] = z0[index[0],index[1]]))
    | "linear" ->
        parallel_do(size(x),x,y,z,__kernel__ (x : cube'unchecked, y : cube'unchecked, z : cube'unchecked, pos : ivec3) -> _
                            (p = ([x[pos], y[pos]] - start) .* step; _
                            p = min(max(p, [0., 0.]), float(n)-1.); _
                            index = int(p); _
                            index2 = min(index+1,n-1); _
                            fr = frac(p); _
                            z[pos] = z0[index] * (1-fr[0]) * (1-fr[1]) + _
                            z0[index[0],index2[1]] * (1 - fr[0]) * fr[1] + _
                            z0[index2[0],index[1]] * fr[0] * (1 - fr[1]) + _
                            z0[index2] * fr[0] * fr[1]))
    | _ -> error("[interp2] Invalid interpolation method")
    endmatch
endfunction

% Function: invert_function
%
% Computes the inverse of a function
%
% : function [x0,y] = invert_function(x : vec)  
%
function [x0,y : vec] = invert_function(x : vec)    
    y0 = 0..numel(x)-1                    % output range
    x0 = linspace(min(x),max(x),numel(x)) % input range
    y = interp1(x,y0,x0,"linear",0)
endfunction

% Function: meshgrid
%
% Generate matrices with coordinates for a 2D Cartesian grid
%
% : function [x : mat, y : mat] = meshgrid(range_x : vec, range_y : vec)
% : function [x, y] = __device__ meshgrid(range_x, range_y)
%
% Note: the matrix indexing is as follows: Y x X
%
% Parameters:
%   range_x - values for the x-axis of the Cartesian grid
%   range_y - values for the y-axis of the Cartesian grid
%   x - output matrix with x-coordinates
%   y - output matrix with y-coordinates
%
function [x : mat, y : mat] = meshgrid(range_x : vec, range_y : vec)
    x = ones(numel(range_y),1) * range_x
    y = transpose(range_y) * ones(1,numel(range_x))
endfunction

function [x, y] = __device__ meshgrid(range_x : vec, range_y : vec)
    x = ones(numel(range_y),1) * range_x
    y = transpose(range_y) * ones(1,numel(range_x))
endfunction

% Function: meshgrid
%
% Generate matrices with coordinates for a 3D Cartesian grid
%
% : function [x : cube, y : cube, z : cube] = meshgrid(
% :      range_x : vec, range_y : vec, range_z : vec)
%
% Parameters:
%   range_x - values for the x-axis of the Cartesian grid
%   range_y - values for the y-axis of the Cartesian grid
%   range_z - values for the z-axis of the Cartesian grid
%   x - output matrix with x-coordinates
%   y - output matrix with y-coordinates
%   z - output matrix with z-coordinates
%
function [x : cube, y : cube, z : cube] = meshgrid( _
    range_x : vec, range_y : vec, range_z : vec)
    [M,N,P] = [numel(range_x),numel(range_y),numel(range_z)]
    x = uninit(M,N,P)
    y = uninit(M,N,P)
    z = uninit(M,N,P)
    parallel_do([M,N,P],__kernel__ (pos:ivec3) -> ( _
        x[pos] = range_x[pos[0]]; _
        y[pos] = range_y[pos[1]]; _
        z[pos] = range_z[pos[2]]))
endfunction

% Function: swap
% Swaps the values of two variables
%
% : function [y1, y2] = swap(x1, x2)
function [y1, y2] = swap(x1, x2)
    [y1, y2] = [x2, x1]
endfunction

% Note - in the following, the kernel function generator will be used to expand the definitions

% Function: isnan
% Check which elements of the matrix are NaN
% The isnan function is available as a built-in function inside
% __kernel__ or __device__ functions. The following definition is
% necessary to ensure that 'isnan' can also be used in regular
% functions.
%
% : function y = isnan(x)
%
operator (x : cube) -> isnan(x) = isnan(x)
operator (x : ccube) -> isnan(x) = isnan(x)

% Function: isinf
% Check which elements of the matrix are infinite
% The isnan function is available as a built-in function inside
% __kernel__ or __device__ functions. The following definition is
% necessary to ensure that 'isinf' can also be used in regular
% functions.
%
% : function y = isinf(x)
%
operator (x : cube) -> isinf(x) = isinf(x)
operator (x : ccube) -> isinf(x) = isinf(x)

% Function: isfinite
% Check which elements of the matrix are finite
% The isnan function is available as a built-in function inside
% __kernel__ or __device__ functions. The following definition is
% necessary to ensure that 'isfinite' can also be used in regular
% functions.
%
% : function y = isfinite(x)
%
operator (x : cube) -> isfinite(x) = isfinite(x)
operator (x : ccube) -> isfinite(x) = isfinite(x)

% Function: sort
% inplace sort along the rows (currently: ascending sort). 
% Internally uses a parallel bitonic sorting algorithm (with complexity O(N)^2 * N / T,
% where T is the number of parallel threads).
%
% : function [] = sort(x)
% : function [] = __device__ sort(x)
%
% Parameters:
% x - the data values to sort
%
function [] = sort(x)

    if numel(x) == 0
        return
    endif

    % Notes: the idea of the implementation of the sort function is to separate the
    % computational stages that can be performed in shared memory from the stages that
    % need to be done in global memory. This is done for the bitonic sorting algorithm,
    % which is a parallel sorting network. When the number of columns is sufficiently
    % small, the complete processing can take advantage of the shared memory. The 
    % algorithm is then very fast. When the number of colums is large as well, the first
    % log2(shared_mem) stages are performed using the shared memory processing, the 
    % remaining stages are done in global memory. This way, we retain fast processing 
    % while avoiding restrictions on the maximum data dimensions. 

    % Kernel function that performs bitonic sort with shared memory
    function [] = __kernel__ bitsort_shared(x : mat, y : mat, n : int, blkdim : ivec2, blkpos : ivec2, pos : ivec2)

        % copy the row to the shared memory...
        assert(blkdim[0]*n <= 2048)
        s = shared(blkdim[0], n)
        delta = int(pos[1]/blkdim[1])*n

        for l = 0..blkdim[1]..n-1
            tid = blkpos[1] + l
            iid = tid + delta
            if iid < size(x,1)
                s[blkpos[0], tid] = x[pos[0], iid]
            else
                s[blkpos[0], tid] = maxvalue(scalar) % maximum floating point value
            endif
        endfor
        syncthreads

        % parallel bitonic sort
        k = 2
        while k <= n
            % bitonic merge
            j = int(k / 2)
            while j > 0
                for l = 0..blkdim[1]..n-1
                    tid = blkpos[1] + l % thread id
                    ixj = xor(tid, j)
                    if tid < ixj
                        if and(tid, k) == 0
                            v = [blkpos[0], tid]
                            w = [blkpos[0], ixj]
                        else
                            v = [blkpos[0], ixj]
                            w = [blkpos[0], tid]
                        endif
                        if s[v] > s[w]
                            %[s[v], s[w]] = [s[w], s[v]]
                            tmp = s[v]; s[v] = s[w]; s[w] = tmp
                        endif
                    endif
                endfor
                syncthreads
                j /= 2
            endwhile
            k *= 2
        endwhile

        % Copy back the results
        for l = 0..blkdim[1]..n-1
            tid = blkpos[1] + l
            iid = tid + delta
            if iid < size(y,1)
                y[pos[0], iid] = s[blkpos[0], tid]
            endif
        endfor
    endfunction

    % Kernel function that performs bitonic sort without shared memory
    function [] = __kernel__ bitsort(s : mat, n : int, k : int, j : int, pos : ivec2)
        % parallel bitonic sort
        i = pos[1] % thread id
        ixj = xor(i, j)
        if i < ixj
            if and(i, k) == 0
                v = [pos[0], i]
                w = [pos[0], ixj]
            else
                v = [pos[0], ixj]
                w = [pos[0], i]
            endif
            if s[v] > s[w] 
                tmp = s[v]; s[v] = s[w]; s[w] = tmp
            endif
        endif
    endfunction

    n = nextpow2(size(x,1))  % bitonic sort only handles powers of two...
    blocks = ceil(n/2048)    % calculate the number of blocks that can be performed in parallel
    max_n = floor(n/blocks)  % the threshold for the number of columns that can be stored in
                             % shared memory
    sz = max_block_size(bitsort_shared, [size(x,0),max_n,1])

    if max_n == n
        % we can compute the complete operation in shared memory!!!
        parallel_do([[size(x,0),sz[1],1],sz],x,x,max_n,bitsort_shared)
    else
        s = uninit(size(x,0),n)
        
        % sorting inside each block
        parallel_do([[size(x,0),sz[1]*blocks,1],sz],x,s,max_n,bitsort_shared)        
        % additional steps cannot be done in shared memory
        k = max_n        
        while k <= n
            j = int(k/2)            
            while j > 0
                parallel_do([size(x,0),n],s,n,k,j,bitsort)
                j = int(j/2)
            endwhile
            k *= 2
        endwhile
        % copy the results back
        x[:,0..numel(x)-1] = s[:,0..numel(x)-1]
        %x[:,:] = s[:,0..numel(x)-1]
    endif
endfunction

function [] = __device__ sort(A : mat)
    [M,N] = size(A,0..1)
    
    % Sort each row of the data matrix
    for k=0..M-1
        % Build binary heap from all items
        for i=0..numel(A)-1
            index = i
            item = A[k,i]
            while index > 0 && A[k,int((index-1)/2)] < item
                top = int((index-1)/2)
                A[k,index] = A[k,top]
                index = top
            endwhile
        endfor
        
        for i=numel(A)-1..-1..1
            % Delete max and place it as last
            last = A[k,i]
            A[k,i] = A[k,0]
            index = 0
            % The last one positioned in the heap
            while index * 2 + 1 < i
                left = index * 2 + 1
                right = left + 1
                if right < i && A[k,left] < A[k,right]
                    if last > A[k,right]
                        break
                    endif
                    A[k,index] = A[k,right]
                    index = right
                else
                    if last > A[k,left]
                        break
                    endif
                    A[k,index] = A[k,left]
                    index = left
                endif
            endwhile
            A[k,index] = last
        endfor
    endfor
endfunction

% Function: sort2
% a non-inplace sort along the rows that also returns the indices (ascending sort)
% Internally uses a parallel bitonic sorting algorithm (with complexity O(N)^2 * N / T,
% where T is the number of parallel threads). When a matrix with multiple rows is 
% provided, the function sorts along each row independently and in parallel.
%
% : function [y, ind] = sort2(x)
% : function [y, ind] = __device__ sort2(x)
%
% Parameters:
% x - the data values to sort
% y - output values of the sorted data
% ind - contains the resulting indices
%
function [y, ind] = sort2(x)

    if numel(x) == 0
        y = []
        ind = []
        return
    endif
    
    % Notes: the idea of the implementation of the sort function is to separate the
    % computational stages that can be performed in shared memory from the stages that
    % need to be done in global memory. This is done for the bitonic sorting algorithm,
    % which is a parallel sorting network. When the number of columns is sufficiently
    % small, the complete processing can take advantage of the shared memory. The 
    % algorithm is then very fast. When the number of colums is large as well, the first
    % log2(shared_mem) stages are performed using the shared memory processing, the 
    % remaining stages are done in global memory. This way, we retain fast processing 
    % while avoiding restrictions on the maximum data dimensions. 

    % Kernel function that performs bitonic sort with shared memory
    function [] = __kernel__ bitsort_ind_shared(x : mat, y : mat, ind : mat, n : int, blkdim : ivec2, blkpos : ivec2, pos : ivec2)

        % copy the row to the shared memory...
        assert(blkdim[0]*n <= 2048)
        s = shared(blkdim[0], n)
        t = shared(blkdim[0], n)
        delta = int(pos[1]/blkdim[1])*n

        for l = 0..blkdim[1]..n-1
            tid = blkpos[1] + l
            iid = tid + delta
            if iid < size(x,1)
                s[blkpos[0], tid] = x[pos[0], iid]
                t[blkpos[0], tid] = iid
            else
                s[blkpos[0], tid] = maxvalue(scalar) % maximum floating point value
            endif
        endfor
        syncthreads

        % parallel bitonic sort
        k = 2
        while k <= n
            % bitonic merge
            j = int(k / 2)
            while j > 0
                for l = 0..blkdim[1]..n-1
                    tid = blkpos[1] + l % thread id
                    ixj = xor(tid, j)
                    if tid < ixj
                        if and(tid, k) == 0
                            v = [blkpos[0], tid]
                            w = [blkpos[0], ixj]
                        else
                            v = [blkpos[0], ixj]
                            w = [blkpos[0], tid]
                        endif
                        if s[v] > s[w]
                            %[s[v], s[w]] = [s[w], s[v]]
                            %[t[v], t[w]] = [t[w], t[v]]
                            tmp = s[v]; s[v] = s[w]; s[w] = tmp
                            tmp = t[v]; t[v] = t[w]; t[w] = tmp
                        endif
                    endif
                endfor
                syncthreads
                j /= 2
            endwhile
            k *= 2
        endwhile

        % Copy back the results
        for l = 0..blkdim[1]..n-1
            tid = blkpos[1] + l
            iid = tid + delta
            if iid < size(y,1)
                y[pos[0], iid] = s[blkpos[0], tid]
                ind[pos[0], iid] = t[blkpos[0], tid]
            endif
        endfor
    endfunction

    % Kernel function that performs bitonic sort without shared memory
    function [] = __kernel__ bitsort_ind(s : mat, t : mat, n : int, k : int, j : int, pos : ivec2)
        % parallel bitonic sort

        % v = blkpos[1] + (0..blkdim[1]..n-1)
        % w = xor(blkpos[1], j) + (0..blkdim[1]..n-1)
        
        i = pos[1] % thread id
        ixj = xor(i, j)
        if i < ixj
            if and(i, k) == 0
                v = [pos[0], i]
                w = [pos[0], ixj]
            else
                v = [pos[0], ixj]
                w = [pos[0], i]
            endif
            if s[v] > s[w] 
                tmp = s[v]; s[v] = s[w]; s[w] = tmp
                tmp = t[v]; t[v] = t[w]; t[w] = tmp  
            endif
        endif
    endfunction

    n = nextpow2(size(x,1))  % bitonic sort only handles powers of two...
    blocks = ceil(n/2048)    % calculate the number of blocks that can be performed in parallel
    max_n = floor(n/blocks)  % the threshold for the number of columns that can be stored in
                             % shared memory
    sz = max_block_size(bitsort_ind_shared, [size(x,0),max_n,1])

    if max_n == n
        y = uninit(size(x))
        ind = uninit(size(x))
    
        % we can compute the complete operation in shared memory!!!
        parallel_do([[size(x,0),sz[1],1],sz],x,y,ind,max_n,bitsort_ind_shared)
    else
        s = uninit(size(x,0),n)
        sind = uninit(size(x,0),n)
        
        % sorting inside each block
        parallel_do([[size(x,0),sz[1]*blocks,1],sz],x,s,sind,max_n,bitsort_ind_shared)        
        % additional steps cannot be done in shared memory
        k = max_n        
        while k <= n
            j = int(k/2)            
            while j > 0
                parallel_do([size(x,0),n],s,sind,n,k,j,bitsort_ind)
                j = int(j/2)
            endwhile
            k *= 2
        endwhile
        % copy the results back
        y = s[:,0..numel(x)-1]
        ind = sind[:,0..numel(x)-1]
    endif
endfunction


function [A : mat, ind : mat] = __device__ sort2(x : mat)
    A = copy(x)
    [M,N] = size(x,0..1)
    ind = uninit(M,N)
    
    if numel(x) == 0
        A = copy(x)
        ind = copy(x)
        return
    endif
    
    % Sort each row of the data matrix
    for k=0..M-1
        ind[k,:] = 0..N-1
        
        % Build binary heap from all items
        for i=0..numel(A)-1
            index = i
            item = A[k,i]
            while index > 0 && A[k,int((index-1)/2)] < item
                top = int((index-1)/2)
                [A[k,index], ind[k,index]] = [A[k,top], ind[k,top]]
                index = top
            endwhile
        endfor
        
        for i=numel(A)-1..-1..1
            % Delete max and place it as last
            [last, ind_last] = [A[k,i], ind[k,i]]
            
            [A[k,i], ind[k,i]] = [A[k,0], ind[k,0]]
            index = 0
            % The last one positioned in the heap
            while index * 2 + 1 < i
                left = index * 2 + 1
                right = left + 1
                if right < i && A[k,left] < A[k,right]
                    if last > A[k,right]
                        break
                    endif
                    [A[k,index], ind[k,index]] = [A[k,right], ind[k,right]]
                    index = right
                else
                    if last > A[k,left]
                        break
                    endif
                    [A[k,index], ind[k,index]] = [A[k,left], ind[k,left]]
                    index = left
                endif
            endwhile
            [A[k,index], ind[k,index]] = [last, ind_last]
        endfor
    endfor
endfunction

% Function: median
%
% Computes the median along the rows of a given matrix.
%
% : function y : vec = median(x : mat)
% : function y : vec = __device__ median(x : mat)
%
% Parameters:
% x - a matrix with values to compute the median from
% y - the output vector
%
function y : vec = median(x : mat)

    x_sorted = copy(x)
    sort(x_sorted)
    n = size(x,1)

    % even number of elements
    if mod(n,2) == 0
        y = 0.5 * (x_sorted[:,int(n/2)-1] + x_sorted[:,int(n/2)])
    else
        y = x_sorted[:,floor(n/2)]
    endif
endfunction

function y = __device__ median(x : mat)
    x_sorted = copy(x)
    sort(x_sorted)
    [M,N] = size(x,0..1)
    n = int(N/2)

    % For each row of the matrix
    for k=0..M-1
        % even number of elements
        if mod(N,2) == 0
            y = 0.5 * (x_sorted[k,n-1] + x_sorted[k,n])
        else
            y = x_sorted[k,n]
        endif
    endfor
endfunction

% Function: cat
% Horizontal concatenation function: works for matrix, complex matrices and cell matrices
%
% : function y = cat(...args : vec[??])
%
function y = cat(...args : vec[??])
    a = args[0]
    if numel(args)==1
        y = a
        return
    endif
    
    [args_ok1, args_ok2] = [true, true]
    args_real = isreal(a)
    n = uninit(numel(args)+1)    
    n[0] = 0
    n[1] = size(a,1)
    [m,_,p] = size(a,0..2)
    for k=1..numel(args)-1
        args_ok1 = args_ok1 && size(a,0) == size(args[k],0) && size(a,2) == size(args[k],2)
        args_ok2 = args_ok2 && (numel(args[k]) == 0 || type(a, type(args[k]), "ignore_size") || type(args[k], type(a), "ignore_size"))
        args_real = args_real && isreal(args[k])
        n[k+1] = size(args[k],1)
    endfor
    n = cumsum(n)
    N = n[numel(n)-1]
    assert(args_ok1, "Wrong dimensions for function 'cat'")
    %assert(args_ok2, "'cat': arguments should be of the same type!")

    if isstring(a)
        y = a
        for k=1..numel(args)-1
            y = sprintf("%s%s",y, args[k])
        endfor
        return
    elseif iscell(a)
        y = cell(m,N,p)
    elseif args_real
        y = zeros(m,N,p)
    else
        y = czeros(m,N,p)
    endif

    for k=0..numel(args)-1
        if n[k+1]-n[k]>0
            y[:,n[k]..n[k+1]-1,:] = args[k]
        endif
    endfor
endfunction

% Function: vertcat
% Vertical concatenation function: works for matrix, complex matrices and cell matrices
%
% : function y = vertcat(...args : vec[??])
%
%function y = vertcat(a, b)
%    assert(size(a,1) == size(b,1) && size(a,2) == size(b,2), _
%            "Wrong dimensions for function 'vertcat'")
%    assert(numel(a) == 0 || type(a, type(b)) || type(b, type(a)), "'vertcat': arguments should be of the same type!")
%
%    [m,n,k] = size(a,0..2)
%    m2 = size(b,0)
%    if iscell(a)
%        y = cell(m+m2,n,k)
%    elseif isreal(a) && isreal(b)
%        y = zeros(m+m2,n,k)
%    else
%        y = czeros(m+m2,n,k)
%    endif
%    y[0..m-1,:,:] = a
%    y[m..m+m2-1,:,:] = b   
%endfunction

function y = vertcat(...args : vec[??])
    a = args[0]
    if numel(args)==1
        y = a
        return
    endif
    
    [args_ok1, args_ok2] = [true, true]
    args_real = isreal(a)
    m = uninit(numel(args)+1)    
    m[0] = 0
    m[1] = size(a,0)
    [_,n,p] = size(a,0..2)
    for k=1..numel(args)-1
        args_ok1 = args_ok1 && size(a,1) == size(args[k],1) && size(a,2) == size(args[k],2)
        args_ok2 = args_ok2 && ((numel(args[k]) == 0 _
            || type(a, type(args[k]), "ignore_size") || type(args[k], type(a), "ignore_size")))
        args_real = args_real && isreal(args[k])
        m[k+1] = size(args[k],0)
    endfor
    m = cumsum(m)
    M = m[numel(m)-1]
    assert(args_ok1, "Wrong dimensions for function 'vertcat'")
    assert(args_ok2, "'vertcat': arguments should be of the same type!")

    if isstring(a)
        y = a
        for k=1..numel(args)-1
            y = sprintf("%s\n%s",y, args[k])
        endfor
        return
    elseif iscell(a)
        y = cell(M,n,p)
    elseif args_real
        y = zeros(M,n,p)
    else
        y = czeros(M,n,p)
    endif

    for k=0..numel(args)-1
        if m[k+1]-m[k]>0
            y[m[k]..m[k+1]-1,:,:] = args[k]
        endif
    endfor
endfunction

% Function: find
%
% Finds the location of the non-zero elements of x.
% The result is a list of positions [[x0, y0, z0], [x1, y1, z1], ...].
%
% : function y : mat = find(x : cube, mode : string = "all", dim : int = 0)
%
% Parameters:
% x - a data cube
% y - a matrix with coordinates of the positions of size Nx3
%       where N is the number of elements of x != 1, i.e. sum(x!=1)
% mode - the search mode
%   o "all": lists all occurences of elements != 1
%   o "first": returns the first occurence of an element != 1
%   o "last": returns the last occurence of an element != 1
%
function y : mat(:,3) = find(x : cube, mode : string = "all", dim : int = 1)
    match mode with
    | "all" ->
        function [] = __kernel__ swizzle(x : vec, y : mat'unchecked, sz : int, num_sm : int, pos : int)
            {!kernel target="gpu"}
            blkdim = int(sz/num_sm)
            p = mod(pos,sz)
            idx = int(pos/sz)
            mp = int(p/blkdim)
            y[mp,mod(p,blkdim)+idx*blkdim] = x[pos] != 0 ? 1 : 0
        endfunction
        
        function [] = __kernel__ parallel_find(x : vec'unchecked, y : mat'unchecked(:,3), cnt : vec, sz : int, x_sz : ivec3, pos : int, blkidx : int)
            {!kernel target="gpu"}
            nblocks = int((numel(x) + sz - 1) ./ sz)

            counter : vec[int]'unchecked = shared[int](64)
            counter[blkidx] = int(cnt[blkidx-1]) % reset
            syncthreads

            for m=0..nblocks-1
                p = pos + m .* sz

                if p<numel(x) && x[p] != 0
                    % allocate one entry in the output
                    idx = (counter[blkidx] += 1)
                    y[idx-1,0..2] = ind2pos(x_sz, p)
                endif
            endfor
        endfunction
        
        function [] = __kernel__ singlethreaded_find(x : vec'unchecked, y : mat'unchecked(:,3), x_sz : ivec3)
            idx = 0
            for m=0..numel(x)-1
                if x[m] != 0
                    y[idx,0..2] = ind2pos(x_sz,m)
                    idx += 1
                endif
            endfor
        endfunction
        
        match schedule(numel(x),x, {parallel_find, singlethreaded_find}) with    
        | 0 ->
            blk_size = prod(max_block_size(parallel_find, prevpow2(numel(x))))
            num_sm = min(64,prod(opt_block_cnt(parallel_find, blk_size, numel(x))))
            sz = blk_size * num_sm
            
            x2 = uninit(num_sm,int(ceil(numel(x)/sz))*blk_size)
            parallel_do(numel(x2), x, x2, sz, num_sm, swizzle)
            x2 = sum(x2,1)
            x2 = cumsum(transpose(x2))
            nz = max(0,x2[num_sm-1])
            y = uninit(nz,3)    
            
            parallel_do([[sz],[blk_size]],x,y,x2,sz,size(x,0..2),parallel_find)
        | 1 ->
            nz = sum(x != 0)
            y = uninit(nz,3)
            serial_do(1,x,y,size(x,0..2),singlethreaded_find)
        endmatch
    | "first" ->  % find the first occurence along dimension dim
        assert(!(dim < 0 || dim>2),
                 "(find) invalid input dimensions: should be a vector/matrix!")        
        assert(dim == 1, "(find) only dimension 1 is currently supported...")
        
        function y : int = __kernel__ parallel_find_first(x : vec'unchecked, blkpos : int, blkdim : int)
            nblocks = int((numel(x) + blkdim - 1) / blkdim)
            y = numel(x)
            
            % find the first occurrences within each block            
            for m=0..nblocks-1
                p = blkpos + m * blkdim
                if p < numel(x) && x[p] != 0 % result found
                    y __= p
                    break
                endif
            endfor
        endfunction

        rv = parallel_do(max_block_size(parallel_find_first, nextpow2(numel(x))), x, parallel_find_first)
        if rv>=numel(x) % not found
            y = []
        else
            % decode the index...
            y = ind2pos(size(x,0..2), rv)
        endif

    | "last" ->  % find the last occurence along dimension dim
        assert(!(dim < 0 || dim>2),
                 "(find) invalid input dimensions: should be a vector/matrix!")        
        assert(dim == 1, "(find) only dimension 1 is currently supported...")
        
        function y : int = __kernel__ parallel_find_last(x : vec'unchecked, blkpos : int, blkdim : int)
            nblocks = int((numel(x) + blkdim - 1) / blkdim)
                       
            y = -1
            syncthreads
            
            % find the first occurrences within each block            
            for m=nblocks-1..-1..0
                p = blkpos + m * blkdim
                if p < numel(x) && x[p] != 0 % result found
                    y ^^= p
                    break
                endif
            endfor
        endfunction

        rv = parallel_do(max_block_size(parallel_find_last, nextpow2(numel(x))), x, parallel_find_last)
        if rv<0 % not found
            y = []
        else
            % decode the index...
            y = ind2pos(size(x,0..2), rv)
        endif
    | _ ->
        error "(find) unrecognized mode!"
    endmatch
endfunction

% Function: sinc
% The normalized sinc function, defined as sin(pi*x)/(pi*x)
%
% : sinc = __device__ (x : scalar) -> (x == 0.0) ? 1.0 : sin(pi*x)/(pi*x)
%
sinc = __device__ (x : scalar) -> (x == 0.0) ? 1.0 : sin(pi*x)/(pi*x)

function [y : cube] = sinc_cube(x : cube)
    y = zeros(size(x))
    parallel_do(size(y),x,y,__kernel__ (x : cube'unchecked, y : cube'unchecked, pos : ivec3) -> y[pos] = sinc(x[pos]))
endfunction
reduction (x : cube) -> sinc(x) = sinc_cube(x)

% Function: herm_transpose
% Hermitian transpose of a matrix
%
% : function y : scalar = herm_transpose(x : scalar)
% : function y : cscalar = herm_transpose(x : cscalar)
% : function y : mat = herm_transpose(x : mat)
% : function y : cmat = herm_transpose(x : cmat)
%
% See also:
%    <transpose>
reduction (x : scalar) -> herm_transpose(x) = x
reduction (x : cscalar) -> herm_transpose(x) = conj(x)
reduction (x : mat) -> herm_transpose(x) = transpose(x)
reduction (x : cmat) -> herm_transpose(x) = conj(transpose(x))


% Function: kron
% Computes the Kronecker product of two multi-dimensional matrices
%
% : function y = kron(a : cube{:}, b : cube{:})
% : function y = __device__ kron(a : vec, b : vec)
% : function y = __device__ kron(a : mat, b : mat)
% : function y = __device__ kron(a : cube, b : cube)
%
function z = kron(x : cube{:}, y : cube{:})

    if ndims(x) + ndims(y) > 4
        error("kron can currently handle up to dimension 4!")
    endif

    function [] = __kernel__ kernel(x : vec, y : vec, z : vec, x_dim : ivec4, y_dim : ivec4, pos : int, blkdim : int, blkcnt : int)
        stepsize = blkdim * blkcnt
        total = prod(x_dim)
        rep = prod(y_dim)
        z_dim = x_dim .* y_dim
        
        for i=pos..stepsize..total-1
            x_pos = ind2pos(x_dim, i)
            val = x[i]
            for j=0..rep-1
                y_pos = ind2pos(y_dim, j)
                z_pos = x_pos + x_dim .* y_pos
                z[pos2ind(z_dim, z_pos)] = val * y[pos2ind(y_dim, y_pos)]
            endfor
        endfor        
    endfunction

    d = max(numel(size(x)),numel(size(y))) % dimensionality of the result
    z = uninit(size(x,0..d-1).*size(y,0..d-1))
    nextpow2 = n -> int(2^ceil(log2(n)))    
    parallel_do(min(nextpow2(numel(x)), 16384), x, y, z, size(x,0..3), size(y,0..3), kernel)   

endfunction

function z : cube{N} = __device__ kron[N](x : cube{N}, y : cube{N})
    z = uninit(size(x,0..N-1).*size(y,0..N-1))
    for i=0..numel(z)-1
        m=ind2pos(size(x,0..N-1), int(i/numel(y)))
        n=ind2pos(size(y,0..N-1), mod(i,numel(y)))
        z[ind2pos(size(z,0..N-1),i)] = x[m]*y[n]
    endfor
endfunction

% Function: periodize
% Computes x modulo a: but in a sense that the result is in the range [0, a[, 
% even for negative numbers
%
% : function y : cube = periodize(x : cube, a : scalar)
% : function y : cube = periodize(x : cube, a : cube)
%
operator (x : cube, a : scalar) -> periodize(x, a) = periodize(x, a)
operator (x : cube, a : cube)    -> periodize(x, a) = periodize(x, a)
%
% Function: mirror_ext
% Performs mirror reflection of the input coordinate
%
% : function y : cube = mirror_ext(x : cube, a : scalar)
% : function y : cube = mirror_ext(x : cube, a : cube)
%
operator (x : cube, a : scalar) -> mirror_ext(x, a) = mirror_ext(x, a)
operator (x : cube, a : cube)    -> mirror_ext(x, a) = mirror_ext(x, a)

% Function: nanmean
%
% Calculates the mean of all matrix elements in x, after removing
% all NaN values.
%
% Parameters:
% x - The input data
% y - The mean of the data (calculated by ignoring all NaN values).
function y = nanmean(x : cube{:})
    function [total : scalar, count : int] = __kernel__ kernel(x : vec, pos : int)
    val = x[pos]
    if !isnan(val)
        total += val
        count += 1
    endif
    endfunction

    [total, count] = parallel_do(numel(x),x,kernel)
    y = total/count
endfunction
function y = __device__ nanmean(x)
    total = 0.0
    count = 0
    for k=0..numel(x)-1
        val = x[ind2pos(size(x), k)]
        if !isnan(val)
            total += val
            count += 1
        endif
    endfor
    y = total/count
endfunction

% Function: nanmin
%
% Calculates the minimum of all matrix elements in x, after removing
% all NaN values.
%
% Parameters:
% x - The input data
% y - The minimum of the data (calculated by ignoring all NaN values).
function y = nanmin(x : cube{:})
    function [count : int] = __kernel__ kernel(x : vec, retval : vec'unchecked, pos : int)
        val = x[pos]
        if !isnan(val)
            retval[0] __= val
            count += 1
        endif
    endfunction

    retval = [maxvalue(scalar)]
    count = parallel_do(numel(x),x,retval,kernel)
    y = retval[0]
endfunction
function minval = __device__ nanmin(x)
    minval = maxvalue(scalar)
    for k=0..numel(x)-1
        val = x[ind2pos(size(x), k)]
        if !isnan(val)
            minval __= val
        endif
    endfor
endfunction

% Function: nanmax
%
% Calculates the maximum of all matrix elements in x, after removing
% all NaN values.
%
% Parameters:
% x - The input data
% y - The maximum of the data (calculated by ignoring all NaN values).
function y = nanmax(x : cube)
    function [count : int] = __kernel__ kernel(x : vec, retval : vec'unchecked, pos : int)
        val = x[pos]
        if !isnan(val)
            retval[0] ^^= val
            count += 1
        endif
    endfunction

    retval = [minvalue(scalar)]
    count = parallel_do(numel(x),x,retval,kernel)
    y = retval[0]
endfunction

function maxval = __device__ nanmax(x)
    maxval = maxvalue(scalar)
    for k=0..numel(x)-1
        val = x[ind2pos(size(x), k)]
        if !isnan(val)
            maxval __= val
        endif
    endfor
endfunction



function y = __device__ catv(...args : vec[vec])
    a = args[0]
    if numel(args)==1
        y = a
        return
    endif
    
    n = uninit(numel(args)+1)
    n[0] = 0
    for k=0..numel(args)-1
        n[k+1] = numel(args[k])
    endfor
    n2 = cumsum(n)
    N = n2[numel(n)-1]
    y = uninit(N)

    for k=0..numel(args)-1
        if n[k+1]-n[k]>0
            y[n[k]..n[k+1]-1] = args[k]
        endif
    endfor
endfunction

function y = __device__ catm(...args : vec[mat])
    a = args[0]
    if numel(args)==1
        y = a
        return
    endif
    
    [args_ok1] = [true]
    n = uninit(numel(args)+1)
    n[0] = 0
    n[1] = size(a,1)
    [m,_] = size(a,0..1)
    for k=1..numel(args)-1
        args_ok1 = args_ok1 && size(a,0) == size(args[k],0)
        n[k+1] = size(args[k],1)
    endfor
    n = cumsum(n)
    N = n[numel(n)-1]
    assert(args_ok1, "Wrong dimensions for function 'cat'")

    y = uninit(m,N)
    for k=0..numel(args)-1
        if n[k+1]-n[k]>0
            y[:,n[k]..n[k+1]-1] = args[k]
        endif
    endfor
endfunction


%======================================
% Some test routines are placed here...
%======================================

function [] = main()

    print cat([1,2],[3,4,5],[7,8,9])
    print vertcat([1,2,1],[[3,4,5],[5,6,7]],[7,8,9])

    x = linspace(0,pi,256)
    y = linspace(0,pi,256)    
    plot(interp1(x,y,linspace(0,4,100),"nearest"))


    print isinf([1/0, 2])
    print periodize(-1*ones(4,4),4)

    x = cell(2,2)
    y = 5
    print iscell(x),",",iscell(y)

    A = eye(9)
    A[0,0] = 2
    B = diag(A)
    print diag(A), ",", trace(A)
    print diagm(diag(A))
    print isinf(1./A)
    print interp1([0,1,2],[0,1,2],[0.5,1.5],"linear")

    [x, y] = swap(2, 3)
    print x,",",y

    % sorting test
    vals = 10..-1..0
    [y,ind] = sort2(vals)
    print y
    print ind

    print herm_transpose([1+1i,2+2i])


endfunction