%
% tridiag.q
% Quasar implementation of the Thomas algorithm for solving tridiagonal matrix problems
%
{!author name="Simon Donné"}
{!doc category="Linear algebra"}

% Function: solve_tridiagonal_matrix
%
% Solve the tridiagonal matrix problem using the simplified Thomas algorithm.
% This means that the diagonal elements are assumed to be nonzero.
% It solves the system Mx = d, where M is a tridiagonal matrix.
% :function [x : vec] = __device__ solve_tridiagonal_matrix(a:vec,b:vec,c:vec,d:vec)
%
% Parameters:
% a - the subdiagonal (the first element is ignored)
% b - the diagonal (assumed to be non-zero)
% c - the superdiagonal (the last element is ignored)
% d - the constant elements of the system
%
function [x:vec] = __device__ solve_tridiagonal_matrix(a:vec,b:vec,c:vec,d:vec)
    N = numel(a)
    
    %forward pass
    c[0] = c[0]/b[0]
    for n = 1..N-1
        c[n] = c[n]/(b[n] - a[n]*c[n-1])
    end
    
    d[0] = d[0]/b[0]
    for n = 1..N-1
        d[n] = (d[n]-a[n]*d[n-1])/(b[n]-a[n]*c[n-1])
    end
    
    %backward substitution
    x = zeros(N)
    x[N-1] = d[N-1]
    for n = N-2..-1..0
        x[n] = d[n] -c[n]*x[n+1]
    end
end

function [] = main()
    %using the tridiagonal algorithm to solve the following system:
    %
    %   [1   1] [x1]   [2]
    %   [     ]*[  ] = [ ]
    %   [1  -1] [x2]   [0]
    %
    subdiagonal = [0,1]
    diagonal = [1,-1]
    superdiagonal = [1, 0]
    f = [2,0]
    
    tic()
    x = solve_tridiagonal_matrix(subdiagonal, diagonal, superdiagonal, f)+0
    toc()
    
    print x
end