% Undecimated wavelet transforms
% March 18, 2016
import "dwt.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Multiresolution"}

% Function: udwt1d_dim0
% 
% Calculates a 1D undecimated wavelet transform (along the
% columns of the input matrix)
% 
% Usage:
%   : [h : cube, g : cube] = udwt1d_dim0(x : cube, w : mat, j : int, ctd : int)
% 
% Parameters:
% x - the input matrix/signal
% w - the wavelet filters to use
% j - the current scale (0=finest scale)
% ctd - the central coefficient index of the wavelet analysis filter
%
% Return values:
% h - the lowpass filtered subband
% g - the highpass filtered subband
function [h : cube'unchecked, g : cube'unchecked] = udwt1d_dim0(x : cube'unchecked, w : mat'unchecked, j : int, ctd : int)
    n = size(x, 0)
    h = zeros(size(x))
    g = zeros(size(x))
    k_max = size(w,1)-1
    scale = shl(1, j)
    K=10*n+ctd
    
    for p=0..size(x,1)-1
        for i=0..n-1
            for c=0..size(x,2)-1
                a = b = 0.0
                i2 = mod(i+scale*K, n)
                for k=0..k_max
                    a += w[0,k] * x[i2,p,c]
                    b += w[1,k] * x[i2,p,c]
                    i2 += scale
                    if i2 >= n
                        i2 -= n
                    endif
                endfor
                h[i,p,c] = (1.0/sqrt(2.0)) * a
                g[i,p,c] = (1.0/sqrt(2.0)) * b
            endfor
        endfor
    endfor
endfunction


% Function: udwt1d_dim1
% 
% Calculates a 1D undecimated wavelet transform (along the
% rows of the input matrix)
% 
% Usage:
%   : [h : cube, g : cube] = udwt1d_dim1(x : cube, w : mat, j : int, ctd : int)
% 
% Parameters:
% x - the input matrix/signal
% w - the wavelet filters to use
% j - the current scale (0=finest scale)
% ctd - the central coefficient index of the wavelet analysis filter
%
% Return values:
% h - the lowpass filtered subband
% g - the highpass filtered subband
function [h : cube'unchecked, g : cube'unchecked] = udwt1d_dim1(x : cube'unchecked, w : mat'unchecked, j : int, ctd : int)
    n = size(x, 1)
    h = zeros(size(x))
    g = zeros(size(x))
    k_max = size(w,1)-1
    step = shl(1, j)
    K=10*n+ctd
    
    for p=0..size(x,0)-1
        for i=0..n-1
            for c=0..size(x, 2)-1
                a = b = 0.0
                i2 = mod(i+step*K, n)
                for k=0..k_max
                    a += w[0,k] * x[p, i2, c]
                    b += w[1,k] * x[p, i2, c]
                    i2 += step
                    if i2 >= n
                        i2 -= n
                    endif
                endfor
                h[p,i,c] = (1.0/sqrt(2.0)) * a
                g[p,i,c] = (1.0/sqrt(2.0)) * b
            endfor
        endfor
    endfor
endfunction

% Function: iudwt1d_dim0
% 
% Computes the inverse undecimated 1D wavelet transform along the columns of the
% input matrix
% 
% Usage:
%   : function y = iudwt1d_dim0(h : cube, g : cube, w : mat, j : int, ctr : int)
% 
% Parameters:
% h - input lowpass subband
% g - input highpass subband
% w - the wavelet filters to use
% j - the current scale (0=finest scale)
% ctr - the central coefficient index of the synthesis filter
% 
% Returns:
% y - the synthesized matrix
function y : cube'unchecked = iudwt1d_dim0(h : cube'unchecked, g : cube'unchecked, w : mat'unchecked, j : int, ctr : int)
    y = zeros(size(h))
    n = size(h, 0)
    K = 10*n + ctr*2-1
    k_max = size(w,1)-1
    step = shl(1,j)

    for p=0..size(h,1)-1
        for i=0..n-1
            for c=0..size(h, 2)-1
                a = 0.0
                i2 = mod(i + step*K, n)
                for k=0..k_max
                    a +=  w[2,k] * h[i2, p, c] + w[3, k] * g[i2, p, c]
                    i2 += step
                    if i2 >= n
                        i2 -= n
                    endif
                endfor
                y[i, p, c] = (1.0/sqrt(2.0)) * a
            endfor
        endfor
    endfor
endfunction

% Function: iudwt1d_dim1
% 
% Computes the inverse undecimated 1D wavelet transform along the rows of the
% input matrix
% 
% Usage:
%   : function [y : mat] = iudwt1d_dim1(h : mat, g : mat, w : mat, j : int, ctr : int)
% 
% Parameters:
% h - input lowpass subband
% g - input highpass subband
% w - the wavelet filters to use
% j - the current scale (0=finest scale)
% ctr - the central coefficient index of the synthesis filter
% 
% Returns:
% y - the synthesized matrix
function y : cube'unchecked = iudwt1d_dim1(h : cube'unchecked, g : cube'unchecked, w : mat'unchecked, j : int, ctr : int)
    y = zeros(size(h))
    n = size(h, 1)
    K = 10*n + ctr*2-1
    k_max = size(w,1)-1
    step = shl(1,j)
    for p=0..size(h,0)-1
        for i=0..n-1
            for c=0..size(h,2)-1
                a = 0.0
                i2 = mod(i + step*K, n)
                for k=0..k_max
                    a += w[2,k] * h[p, i2, c] + w[3, k] * g[p, i2, c]
                    i2 += step
                    if i2 >= n
                        i2 -= n
                    endif
                endfor
                y[p, i, c] = (1.0/sqrt(2.0)) * a
            endfor
        endfor
    endfor
endfunction

% Function: udwt2d
% 
% Calculates the undecimated 2D wavelet transform of an input image
% 
% Usage:
%   : function [y : vec[??]] = udwt2d(x : cube, w : wavelet_params, J : int)
% 
% Parameters:
% x - the input image
% w - the wavelet filters (e.g. obtained using `orthogonal_wavelet`)
% J - the number of scales
% 
% Returns:
% y - the wavelet subbands y[j][k],j=0..J-1,k=0..2 
%     the scaling coefficients y[J]
function y = udwt2d(x : cube, w : wavelet_params, J : int)
    sz = size(x,0..2)
    if mod(sz[0], 2^J) != 0 || mod(sz[1], 2^J) != 0
        error "dwt2d: input matrix should have dimensions that are a multiple of ",2^J
    endif
    
    lo = x
    y = cell(J+1)
    for j=0..J-1
        bands = cell(3)
        [lo, hi] = udwt1d_dim0(lo, w.filter_coeff, j, w.ctd)
        [lo, bands[0]] = udwt1d_dim1(lo, w.filter_coeff, j, w.ctd)
        [bands[1], bands[2]] = udwt1d_dim1(hi, w.filter_coeff, j, w.ctd)
        y[j] = bands
    endfor
    y[J] = lo
endfunction

% Function: iudwt2d
% 
% Calculates an inverse undecimated 2D wavelet transform 
% 
% Usage:
%   : function [lo] = iudwt2d(x : vec[??], w : wavelet_params, J : int)
% 
% Parameters:
% x - the wavelet subbands
% w - the wavelet filters
% J - the number of scales
% 
% Returns:
% lo - the reconstructed image
function lo = iudwt2d(x : vec[??], w : wavelet_params, J : int)
    lo = x[J]
    for j=J-1..-1..0
        bands = x[j]
        hi = iudwt1d_dim1(bands[1], bands[2], w.filter_coeff, j, w.ctr)
        lo = iudwt1d_dim1(lo, bands[0], w.filter_coeff, j, w.ctr)
        lo = iudwt1d_dim0(lo, hi, w.filter_coeff, j, w.ctr)
    endfor
endfunction

function [] = main()

    im = imread("lena_big.tif")
    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))

    w = orthogonal_wavelet(db4)
    J = 3
    
%    [wc,ctd,ctr] = orthogonal_wavelet(w)
%    [lo, hi] = udwt1d_dim0(im, wc, 0, ctd)
%    im2 = iudwt1d_dim0(lo, hi, wc, 0, ctr)
%    imshow(im2)
%    print psnr(im,im2)
%    return
    
    bands = udwt2d(im, w, 3)
    im2 = iudwt2d(bands, w, 3)
    print "psnr=", psnr(im,im2), "dB"
    imshow(im2)

endfunction