%=================================================
% volumeraycasting.q
% Real-time visualization of 3D volumes in Quasar
% March 2013, Bart Goossens
%=================================================
% Note: use the function volumeraycasting for rendering.
import "system.q"
import "inttypes.q"

{!author name="Bart Goossens"}
{!doc category="Visualization"}

type ray : class
    src : vec3
    dir : vec3
endtype

type ray_segment : mutable class
    near : scalar
    far : scalar
    near_intensity : scalar
endtype

type box : class
    min : vec3
    max : vec3
endtype

type line_segment : mutable class
    near : vec3
    far : vec3
    near_dim : int
    is_intersecting : int
endtype

% Function: box_contains
%
% Checks whether the specified box contains a given point.
%
function y : int = __device__ box_contains(b : box, p : vec3)
    y = b.min[0] <= p[0] && _
        b.min[1] <= p[1] && _
        b.min[2] <= p[2] && _
        b.max[0] >= p[0] && _
        b.max[1] >= p[1] && _
        b.max[2] >= p[2]
        
endfunction

% Function: intersectswith
%
% Checks whether a ray intersects with a given box (and also
% computes the intersection)
%
function [ls:line_segment] = __device__ intersectswith(b : box, r : ray)

    v1 = [r.dir[0]<0.0 ? b.max[0] : b.min[0], _
          r.dir[1]<0.0 ? b.max[1] : b.min[1], _
          r.dir[2]<0.0 ? b.max[2] : b.min[2]]
    w1 = [r.dir[0]<0.0 ? b.min[0] : b.max[0], _
          r.dir[1]<0.0 ? b.min[1] : b.max[1], _
          r.dir[2]<0.0 ? b.min[2] : b.max[2]]               
    ls=line_segment(near:=[0.,0.,0.],far:=[0.,0.,0.],near_dim:=0,is_intersecting:=false)
    
    
    tmin = (v1 - r.src)./r.dir
    tmax = (w1 - r.src)./r.dir
    
    if tmin[0]>tmax[1] || tmin[1]>tmax[0]
        return
    endif
    tmin[0] = max(tmin[0],tmin[1])
    tmax[0] = min(tmax[0],tmax[1])
    if tmin[0] > tmax[2] || tmin[2] > tmax[0]
        return
    endif
    tmin[0] = max(tmin[0],tmin[2])
    tmax[0] = min(tmax[0],tmax[2])
    
    ls.near_dim = tmin[0] == tmin[1] ? 1 : tmin[0] == tmin[2] ? 2 : 0        
    ls.near = r.src + tmin[0] * r.dir            
    ls.far  = r.src + tmax[0] * r.dir + 0
    ls.is_intersecting = true
endfunction

% Function: ray_rotate_yz
%
% Rotation of a ray in the YZ-plane
function [r_out : ray] = __device__ ray_rotate_yz(r : ray, p : vec3, theta : scalar)
    [cs,sn] = [cos(theta),sin(theta)]
    ray_dir = [r.dir[0], _
               r.dir[1] * cs - r.dir[2] * sn, _               
               r.dir[1] * sn + r.dir[2] * cs]
    ray_src = r.src - p
    ray_src = p + [ray_src[0], _
                   ray_src[1] * cs - ray_src[2] * sn, _
                   ray_src[1] * sn + ray_src[2] * cs+0]
    r_out = ray(src:=ray_src, dir:=ray_dir)
endfunction

% Function: ray_scale
%
% 3D scaling of a ray
function [r_out : ray] = __device__ ray_scale(r : ray, scale : vec3)
    ray_dir = r.dir .* scale
    ray_dir = ray_dir / sqrt(sum(ray_dir.^2.0))
    r_out = ray(src:=r.src .* scale, dir:=ray_dir)
endfunction

function y = __device__ normalize(x : vec3)
    y = x * rsqrt(sum(x.^2))
endfunction

function y = __device__ reflect(incident : vec3, normal : vec3)
    y = incident - 2.0 * dotprod(incident, normal) * normal
endfunction

% Function: raytracer_nearest
%
% Volumetric ray tracer implementation, using nearest neighbor interpolation
% Note: we use CUDA hardware textures for speeding up things
%
% Parameters:
%   im_out - the output image
%   voxels - a 3D cube with the voxel data
%   voxelsf - Prefiltered 3D cube using a 2x2x2-box filter
%   camera_pos - The 3D position of the camera.
%   scale - a scaling factor used for rendering
%   theta - the rotation angle in the YZ-plane
function [] = __kernel__ raytracer_nearest( _
    im_out : cube'unchecked(:,:,3), _
    voxels : cube'hwtex_nearest, _
    voxelsf : cube'hwtex_nearest, _
    voxelsize : vec3, _
    camera_pos : vec3, _
    scale : scalar, _
    theta: scalar, _
    pos : ivec2)
    
    % Compute the ray direction and origin
    p = (float(pos) - 0.5*float(size(im_out, 0..1))) * scale
    ray_dir = [p[0],p[1],0]-camera_pos
    ray_dir = ray_dir / sqrt(sum(ray_dir.^2.0)) % normalization - needed for lighting
    
    % Ray trace the main cube
    ray_ = ray(src:=camera_pos, dir:=ray_dir)  
    offset = int([0.5*size(voxels,0), 0.5*size(voxels,1), 0].*voxelsize)
    ray_ = ray_rotate_yz(ray_, [0.0, 0.0, 0.5*size(voxels,2).*voxelsize[2]], theta)
    ray_ = ray_scale(ray_, 1./voxelsize)
    
    data_cube = box(min:=-float(offset), max:=float(size(voxels)-offset))
    is = intersectswith(data_cube, ray_)    
                
    % We are outside the data cube...
    if !is.is_intersecting
        im_out[...pos,0..2] = 0.0
        return
    endif   

    % The light direction    
    light_dir = [-1.0,1.0,-1.0]/3            

    % Render the surface    
    i = 0
    data_cube = box(min:=data_cube.min-1.0, max:=data_cube.max+1.0)
    dst_val = 0.0
    dst_alpha = 0.0
    step_size = 1.0
    gradient = [1.,1.,1.]
    
    % ray dithering
    steps = int(0.3333 * sum(abs((is.far - is.near) ./ ray_.dir))/step_size)
       
    for i=0..steps
        q = is.near + step_size * (i)*ray_.dir
        qi = offset + int(floor(q))
        if voxelsf[qi] > 0.0
            qo = q + offset
            src_alpha = voxels[qo]
            
            % compute the gradient
            new_gradient = [voxels[qo+[1.,0.,0.]],voxels[qo+[0.,1.,0.]],voxels[qo+[0.,0.,1.]]] - src_alpha
            norm = dotprod(new_gradient,new_gradient)
            if norm > 0.01
                gradient += 0.1 * (new_gradient / norm - gradient)
            endif            
            
            % lighting 
            src_val = (0.5 + 0.5 * dotprod(light_dir, float(gradient))) * src_alpha
            
            % compositing    
            dst_val   += src_val * (1 - dst_alpha)  * src_alpha
            dst_alpha += (1 - dst_alpha) * src_alpha
            if (1 - dst_alpha) < 1e-4 % epsilon
                break               
            endif                 
        endif
    endfor
    im_out[...pos,0..2] = dst_val

endfunction

% Function: raytracer_trilinear
%
% Volumetric ray tracer implementation, using trilinear interpolation
% Note: we use CUDA hardware textures for speeding up things
%
% Parameters:
%   im_out - the output image
%   voxels - a 3D cube with the voxel data
%   voxelsf - Prefiltered 3D cube using a 2x2x2-box filter
%   camera_pos - The 3D position of the camera.
%   scale - a scaling factor used for rendering
%   theta - the rotation angle in the YZ-plane
function [] = __kernel__ raytracer_trilinear( _
    im_out : cube'unchecked(:,:,3), _
    voxels : cube'hwtex_linear, _
    voxelsize : vec3, _
    camera_pos : vec3, _
    scale : scalar, _
    theta: scalar, _
    pos : ivec2)
    
    % Compute the ray direction and origin
    p = (float(pos) - 0.5*float(size(im_out,0..1))) * scale
    ray_dir = [p[0],p[1],0]-camera_pos
    ray_dir = ray_dir / sqrt(sum(ray_dir.^2.0)) % normalization - needed for lighting
    
    % Ray trace the main cube
    ray_ = ray(src:=camera_pos, dir:=ray_dir)
    offset = int([0.5*size(voxels,0), 0.5*size(voxels,1), 0].*voxelsize)
    ray_ = ray_rotate_yz(ray_, [0.0, 0.0, 0.5*size(voxels,2).*voxelsize[2]], theta)
    ray_ = ray_scale(ray_, 1./voxelsize)
    
    data_cube = box(min:=-float(offset), max:=float(size(voxels)-offset))
    is = intersectswith(data_cube, ray_)    
                
    % We are outside the data cube...
    if !is.is_intersecting
        im_out[...pos,0..2] = 0.0
        return
    endif   

    % The light direction    
    light_dir = [-1.0,1.0,-1.0]/3         

    % Render the surface    
    i = 0
    data_cube = box(min:=data_cube.min-1.0, max:=data_cube.max+1.0)
    dst_val = 0.0
    dst_alpha = 0.0
    step_size = 0.25
    gradient = [1.,1.,1.]
    
    % ray dithering
    %rnd = frac(sin(dotprod(float(pos),[12.9898,78.233])) * 43758.5453)
    steps = int(0.333 * sum(abs((is.far - is.near) ./ ray_.dir))/step_size)
       
    for i=0..steps
        q = is.near + step_size * (i)*ray_.dir
        % qi = offset + int(floor(q))
        if true
            qo = q + offset
            src_alpha = voxels[qo]
            
            % compute the gradient
            new_gradient = [voxels[qo+[1.,0.,0.]],voxels[qo+[0.,1.,0.]],voxels[qo+[0.,0.,1.]]] - src_alpha
            norm = dotprod(new_gradient,new_gradient)
            if norm > 0.01
                gradient += 0.1 * (new_gradient / sqrt(norm) - gradient)
            endif            
            
            % lighting 
            src_val = (1.0 + 0.2 * dotprod(light_dir, float(gradient))) * src_alpha
            
            % compositing    
            dst_val   += src_val * (1 - dst_alpha)  * src_alpha
            dst_alpha += (1 - dst_alpha) * src_alpha
            if (1 - dst_alpha) < 1e-4 % epsilon
                break               
            endif                 
        endif
    endfor
    im_out[...pos,0..2] = dst_val

endfunction

% Function: raytracer_trilinear_pyramid
%
% Volumetric ray tracer implementation, using trilinear interpolation
% Note: we use CUDA hardware textures for speeding up things
%
% Parameters:
%   im_out - the output image
%   voxels - a 3D cube with the voxel data
%   voxelsf - Prefiltered 3D cube using a 2x2x2-box filter
%   camera_pos - The 3D position of the camera.
%   scale - a scaling factor used for rendering
%   theta - the rotation angle in the YZ-plane
function [] = __kernel__ raytracer_trilinear_pyramid( _
    im_out : cube'unchecked(:,:,3), _
    voxels : cube'hwtex_linear, _
    voxelsize : vec3, _
    pyr : vec[cube], _
    camera_pos : vec3, _
    scale : scalar, _
    theta: scalar, _
    pos : ivec2)
    
    % Compute the ray direction and origin
    p = (float(pos) - 0.5*float(size(im_out, 0..1))) * scale
    ray_dir = [p[0],p[1],0]-camera_pos
    ray_dir = ray_dir / sqrt(sum(ray_dir.^2.0)) % normalization - needed for lighting
    
    % Ray trace the main cube
    ray_ = ray(src:=camera_pos, dir:=ray_dir)    
    offset = int([0.5*size(voxels,0), 0.5*size(voxels,1), 0].*voxelsize)
    ray_ = ray_rotate_yz(ray_, [0.0, 0.0, 0.5*size(voxels,2).*voxelsize[2]], theta)
    ray_ = ray_scale(ray_, 1./voxelsize)
    
    data_cube = box(min:=-float(offset)+1, max:=float(size(voxels)-1-offset))
    is = intersectswith(data_cube, ray_)    
                
    % We are outside the data cube...
    if !is.is_intersecting
        im_out[...pos,0..2] = 0.0
        return
    endif   

    % The light direction    
    light_dir = normalize([-1.0,1.0,-1.0])
    light1_dir = normalize([1.0,-1.0,-1.0])
    light1 = [-10.0, 10.0, 10.0]
    light2_dir = normalize([1.0,-1.0,1.0])
    
    % Render the surface    
    i = 0    
    dst_val = [0.0,0.0,0.0]
    step_size = 0.25
    gradient = [1.,1.,1.]
    
    % ray dithering
    steps = int(0.333 * sum(abs((is.far - is.near) ./ ray_.dir))/step_size)
       
    mr_level = 0  % multiresolution level
    mr_stepsize = 1 % multiresolution depth
    q = is.near
    qi = offset + int(q)
    
    while i<=steps        
        % if high-level cell contains a one, drop a level
        if mr_level > 0 && pyr[mr_level-1][int(qi/mr_stepsize)]>0
            mr_level -= 1
            mr_stepsize = int(mr_stepsize/2)
            i -= mr_stepsize
            continue
        endif
                
        qo = q + offset
        src_alpha = voxels[qo]
        
        if src_alpha > 0                        
            % compute the gradient
            new_gradient = [voxels[qo+[1.,0.,0.]],voxels[qo+[0.,1.,0.]],voxels[qo+[0.,0.,1.]]] - src_alpha
            norm = dotprod(new_gradient,new_gradient)
            if norm > 0.01
                gradient += 0.1 * (new_gradient / sqrt(norm) - gradient)
            endif            
            normal = gradient
            
            light_dir1 = normalize(light1 - q)
            diffuse = max(dotprod(light_dir1, normal), 0.0)
            diffuse2 = max(dotprod(light_dir, normal), 0.0)
            diffuse3 = max(dotprod(light2_dir, normal), 0.0)
            specular = max(dotprod(reflect(-light_dir, normal), ray_.dir), 0.0)
            temp = (0.3 + 0.3 * (diffuse + diffuse2 + diffuse3))
            dst_val += temp * [0.75, 0.6, 1.0] % blue
            dst_val += 0.4 * pow(specular, 1.5)% shininess
            break
  
        endif
        % jump to higher level
        while mr_level < numel(pyr) && pyr[mr_level][int(qi/(2*mr_stepsize))] == 0
            mr_level += 1
            mr_stepsize *= 2   
        endwhile
                
        % Step forward
        i += mr_stepsize
        q = is.near + step_size * (i)*ray_.dir
        qi = offset + int(q)        
    endwhile
    im_out[...pos,0..2] = dst_val

endfunction

% Function: voxelfilter
%
% A 3D 2x2x2 box filter, used to speed up the volumetric rendering
%
% Parameters:
%   voxelsf - the filtered 3D volume
%   voxels - the original 3D volume
%
function [] = __kernel__ voxelfilter(voxelsf : cube'unchecked, voxels : cube'hwtex_nearest, pos : ivec3)
    voxelsf[pos] = voxels[pos]+voxels[pos+[0,0,1]]+voxels[pos+[0,0,-1]]+ _
                   voxels[pos+[0,1,0]]+voxels[pos+[0,1,1]]+voxels[pos+[0,-1,0]]+voxels[pos+[0,-1,-1]] + _
                   voxels[pos+[1,0,0]]+voxels[pos+[1,0,1]]+voxels[pos+[-1,0,0]]+voxels[pos+[-1,0,-1]] + _
                   voxels[pos+[1,1,0]]+voxels[pos+[1,1,1]]+voxels[pos+[-1,-1,0]]+voxels[pos+[-1,-1,-1]]
endfunction

% Function: build_vrc_pyramid
%
% Build a pyramid for speeding up the volume ray casting
%
function pyr : vec[cube] = build_vrc_pyramid(voxels : cube, levels : int)
    
    function [] = __kernel__ compute_pyramid_level(voxelsf : cube'unchecked, voxels : cube, pos : ivec3)
        p = 2*pos
        %voxelsf[pos] = voxels[p]+voxels[p+[0,0,1]]+voxels[p+[0,0,-1]]+ _
        %               voxels[p+[0,1,0]]+voxels[p+[0,1,1]]+voxels[p+[0,-1,0]]+voxels[p+[0,-1,-1]] + _
        %               voxels[p+[1,0,0]]+voxels[p+[1,0,1]]+voxels[p+[-1,0,0]]+voxels[p+[-1,0,-1]] + _
        %               voxels[p+[1,1,0]]+voxels[p+[1,1,1]]+voxels[p+[-1,-1,0]]+voxels[p+[-1,-1,-1]]
        s = 0.0
        % TODO - this loop can be improved (separable filtering)
        for k=-1..2
            for l=-1..2
                for q=-1..2
                    s += voxels[p+[k,l,q]]
                endfor
            endfor
        endfor
        voxelsf[pos] = s
    endfunction

    type vec_cube : vec[cube]
    pyr = vec_cube(levels)
    last = voxels
    for i=0..levels-1
        pyr[i] = uninit(int((size(voxels) + 2^(i+1) - 1)  / 2^(i+1)))
        
        parallel_do(size(pyr[i]),pyr[i],last,compute_pyramid_level)
        last = pyr[i]
    endfor

endfunction

% Function: volumeraycasting
%
% Real-time visualization of 3D volumes (the main function)
%
% Parameters:
%
% voxels - the 3D volume to visualize
% voxelsize - the size of an individual voxel (isotropic voxels: use [1,1,1])
% interpolation_method - the interpolation technique to use ("nearest" or "trilinear")
function [] = volumeraycasting(voxels, voxelsize, interpolation_method : string = "trilinear")

    hold("off")
    im_out = zeros(768,1024,3)
    scale = 0.5/2
    camera_pos = [-15*15,5.5,-60*10]
    theta = 0.0
    
    voxelsf = zeros(size(voxels))    
    parallel_do(size(voxels),voxelsf,voxels,voxelfilter)       
    
    if interpolation_method == "trilinear"
        levels = 5
        pyr = build_vrc_pyramid(voxels,levels)
    endif
    
    repeat    
        theta = theta + 0.01  
        match interpolation_method with
        | "nearest" ->
            parallel_do(size(im_out,0..1),im_out,voxels,voxelsf,voxelsize,camera_pos,scale,theta,raytracer_nearest)
        | "trilinear" ->
            parallel_do(size(im_out,0..1),im_out,voxels,voxelsize,pyr,camera_pos,scale,theta,raytracer_trilinear_pyramid)
        | _ ->
            error "Undefined interpolation method: ", interpolation_method
        endmatch
        fig = imshow(im_out,[0,2])
        if fig != null
            fig.enable_3dcontrol()
            theta = fig.theta
            scale = 0.5/2 * fig.scale
        endif
           
    until !hold("on")
    
endfunction

% Some demo code...
function [] = main()

    %volumeraycasting(voxels, [1.0,1.0,1.0], "trilinear")
    %return

    function y = __device__ mandlebulbe(x : vec3)
        epsilon : scalar'const = 1e-5
        z = x*0.1
        c = x*0.1
        r = 0.0
        for count = 0..999
            z2 = z .* z
            r = sqrt(sum(z.^2))
            if r>2.0
                break
            endif
            
            planeXY = sqrt(z2[0] + z2[1])
            r += epsilon
            
            sinPhi = z[1] / planeXY
            cosPhi = z[0] / planeXY
            sinThe = planeXY / r
            cosThe = z[2] / r
            
            % First cascade level
            sinPhi = 2.0 * sinPhi * cosPhi
            cosPhi = 2.0 * cosPhi * cosPhi - 1.0
            sinThe = 2.0 * sinThe * cosThe
            cosThe = 2.0 * cosThe * cosThe - 1.0

            % Second cascade level
            sinPhi = 2.0 * sinPhi * cosPhi
            cosPhi = 2.0 * cosPhi * cosPhi - 1.0
            sinThe = 2.0 * sinThe * cosThe
            cosThe = 2.0 * cosThe * cosThe - 1.0

            % Third cascade level
            sinPhi = 2.0 * sinPhi * cosPhi
            cosPhi = 2.0 * cosPhi * cosPhi - 1.0
            sinThe = 2.0 * sinThe * cosThe
            cosThe = 2.0 * cosThe * cosThe - 1.0

            rPow = r^8
            
            z = [sinThe * cosPhi, sinThe * sinPhi, cosThe]
            z *= rPow
            z += c
        endfor
        y = r - 2.0
    endfunction

    voxels = zeros(512,512,512)
    
    function [] = __kernel__ mandlebulbe_kernel(voxels : cube, pos : vec(3))
        r = (pos - 0.5 * size(voxels,0..2)) ./ (0.5 * size(voxels,0..2))
        voxels[pos] = mandlebulbe(r * 10.0) < 0.0
    endfunction
    parallel_do(size(voxels), voxels, mandlebulbe_kernel)
            
    print size(voxels)
    
    im_out = zeros(1024,2048,3)  
    scale = 0.5
    camera_pos = [-15*15,5.5,-60*10]
    voxelsize = [1.0, 1.0, 1.0]
    
    voxelsf = zeros(size(voxels))    
    parallel_do(size(voxels),voxelsf,voxels,voxelfilter)
       
    levels = 5
    pyr = build_vrc_pyramid(voxels,levels)
    
    %voxels = float_to_uint8(voxels*255)    
    %voxelsf = float_to_uint8(voxels*255)    
    
    theta = pi/2
    repeat
        %R = 36-36*cos(mod(theta*4,2*pi))
        %voxelsf = zeros(size(voxels))
        %voxels = (sqrt(x.^2+y.^2+z.^2) < 128) .* (sqrt(y.^2+z.^2) > R) .* (sqrt(x.^2+y.^2) > R) .* (sqrt(x.^2+z.^2) > R)
        %tic()
        %parallel_do(size(voxels),voxelsf,voxels,voxelfilter)
        %toc()
    
        theta = theta + 0.01  
        parallel_do(size(im_out,0..1),im_out,voxels,voxelsize,pyr,camera_pos,scale,theta,raytracer_trilinear_pyramid)
        %parallel_do(size(im_out,0..1),im_out,voxels,voxelsize,camera_pos,scale,theta,raytracer_trilinear)
        %parallel_do(size(im_out,0..1),im_out,voxels,voxelsf,voxelsize,camera_pos,scale,theta,raytracer_nearest)
        fig = imshow(im_out,[0,1])
        if fig != null
            fig.enable_3dcontrol()
            theta = fig.theta
            scale = 0.5 * fig.scale
        endif
    until !hold("on")
    
endfunction