%
% Wavelet and Complex-wavelet based demosaicing
% Bart Goossens, Jan Aelterman, 2012-2013
% Ghent University, Dept. TELIN/IPI/iMinds
% This is a Quasar-port of the demosaicing method from the following paper
%
% J. Aelterman, B. Goossens, A. Pizurica and W. Philips, "Computationally efficient Locally Adaptive 
% Demosaicing of Color Filter Array Images using the Dual-Tree Complex Wavelet Packet Transform, " 
% PLOS One, Vol. 8 No. 5, p.1-18, 2013.
%
import "system.q"
import "bayer_mosaic.q"
import "dwt.q"
import "dtcwt.q"
import "imfilter.q"

% conversion to 3d bands
ch = __device__ (x : ivec2, y : int) -> [x[0],x[1],y]

function [] = __kernel__ wavelet_demosaic_bands(w1 : mat'unchecked, w2 : mat'unchecked, w3 : cube'unchecked, _
    signHHLL : int, signLHLLr : int, signHLLLr : int, signLHLLb : int, signHLLLb : int, _
    pos : ivec2)

    m = int(size(w1,0)/4)
    n = int(size(w1,1)/4)

    % Band indices
    LH = [0, 2*n]
    HL = [2*m, 0]
    HH = [2*m, 2*n]
    LLLL = pos
    LLLH = LLLL + [0, n]
    LLHL = LLLL + [m, 0]
    LLHH = LLLL + [m, n]
    LHLL = LH + LLLL
    LHLH = LH + LLLH
    LHHL = LH + LLHL
    LHHH = LH + LLHH
    HLLL = HL + LLLL
    HLLH = HL + LLLH
    HLHL = HL + LLHL
    HLHH = HL + LLHH
    HHLL = HH + LLLL
    HHLH = HH + LLLH
    HHHL = HH + LLHL
    HHHH = HH + LLHH

    % Green subbands
    w3[ch(HHLH,1)] = w1[HHLH]
    w3[ch(HHHL,1)] = w1[HHHL]
    w3[ch(HHHH,1)] = w1[HHHH]
    w3[ch(HHLL,1)] = 0.0

    w3[ch(LHLH,1)] = w1[LHLH]
    w3[ch(LHHL,1)] = w1[LHHL]
    w3[ch(LHHH,1)] = w1[LHHH]
    w3[ch(LHLL,1)] = 0.0

    w3[ch(HLLH,1)] = w1[HLLH]
    w3[ch(HLHL,1)] = w1[HLHL]
    w3[ch(HLHH,1)] = w1[HLHH]
    w3[ch(HLLL,1)] = 0.0

    w3[ch(LLLH,1)] = w1[LLLH]
    w3[ch(LLHL,1)] = w1[LLHL]
    w3[ch(LLHH,1)] = w1[LLHH]
    w3[ch(LLLL,1)] = w2[LLLL] - signHHLL * w2[HHLL]

    % Red subbands
    w3[ch(HHLH,0)] = w3[ch(HHLH,1)]
    w3[ch(HHHL,0)] = w3[ch(HHHL,1)]
    w3[ch(HHHH,0)] = w3[ch(HHHH,1)]
    w3[ch(HHLL,0)] = w3[ch(HHLL,1)]

    w3[ch(LHLH,0)] = w3[ch(LHLH,1)]
    w3[ch(LHHL,0)] = w3[ch(LHHL,1)]
    w3[ch(LHHH,0)] = w3[ch(LHHH,1)]
    w3[ch(LHLL,0)] = w3[ch(LHLL,1)]

    w3[ch(HLLH,0)] = w3[ch(HLLH,1)]
    w3[ch(HLHL,0)] = w3[ch(HLHL,1)]
    w3[ch(HLHH,0)] = w3[ch(HLHH,1)]
    w3[ch(HLLL,0)] = w3[ch(HLLL,1)]

    w3[ch(LLLH,0)] = w3[ch(LLLH,1)]
    w3[ch(LLHL,0)] = w3[ch(LLHL,1)]
    w3[ch(LLHH,0)] = w3[ch(LLHH,1)]
    w3[ch(LLLL,0)] = (2*signHHLL*w2[HHLL] + signLHLLr*w2[LHLL] + signHLLLr*w2[HLLL]) + w3[ch(LLLL,1)]

    % Blue subbands
    w3[ch(HHLH,2)] = w3[ch(HHLH,1)]
    w3[ch(HHHL,2)] = w3[ch(HHHL,1)]
    w3[ch(HHHH,2)] = w3[ch(HHHH,1)]
    w3[ch(HHLL,2)] = w3[ch(HHLL,1)]

    w3[ch(LHLH,2)] = w3[ch(LHLH,1)]
    w3[ch(LHHL,2)] = w3[ch(LHHL,1)]
    w3[ch(LHHH,2)] = w3[ch(LHHH,1)]
    w3[ch(LHLL,2)] = w3[ch(LHLL,1)]

    w3[ch(HLLH,2)] = w3[ch(HLLH,1)]
    w3[ch(HLHL,2)] = w3[ch(HLHL,1)]
    w3[ch(HLHH,2)] = w3[ch(HLHH,1)]
    w3[ch(HLLL,2)] = w3[ch(HLLL,1)]

    w3[ch(LLLH,2)] = w3[ch(LLLH,1)]
    w3[ch(LLHL,2)] = w3[ch(LLHL,1)]
    w3[ch(LLHH,2)] = w3[ch(LLHH,1)]
    w3[ch(LLLL,2)] = (2*signHHLL*w2[HHLL] + signLHLLb*w2[LHLL] + signHLLLb*w2[HLLL]) + w3[ch(LLLL,1)]

end

function [signHHLL,signLHLLr,signHLLLr,signLHLLb,signHLLLb] = wavelet_demosaicing_signs(pattern : mat)
    % determine signs...
    if pattern[0,0]==1 || pattern[0,0]==3 % green
        signHHLL=-1
        if pattern[1,0] != 0 % blue
            signLHLLr=-1
            signHLLLr=+1
            signLHLLb=+1
            signHLLLb=-1
        else
            signLHLLr=+1
            signHLLLr=-1
            signLHLLb=-1
            signHLLLb=+1
        endif
    else
        signHHLL=1
        if pattern[1,1]==0 % red
            signLHLLr=-1
            signHLLLr=-1
        else
            signLHLLr=1
            signHLLLr=1
        endif
        if pattern[1,1]==2 % blue
            signLHLLb=-1
            signHLLLb=-1
        else
            signLHLLb=1
            signHLLLb=1
        endif
    endif
end

% wavelet demosaicing
function [y : cube] = wavelet_demosaicing(x : mat, pattern : mat, wcoeff)
    % apply a two-level untiled packet wavelet transform
    w1 = copy(x)
    w2 = copy(x)
      
    %wcoeff = wcoeff[numel(wcoeff)-1..-1..0]
    pdwt2d_untiled(w1, wcoeff, 2, false)
    pdwt2d_untiled(w2, wcoeff, 2, true)

    [signHHLL,signLHLLr,signHLLLr,signLHLLb,signHLLLb] = wavelet_demosaicing_signs(pattern)

    y = zeros(size(w1,0),size(w1,1),3)
    parallel_do(size(w1,0..1)/4,w1,w2,y,signHHLL,signLHLLb,signHLLLb,signLHLLr,signHLLLr,wavelet_demosaic_bands)
        
    % compute inverse wavelet packet transform
    ipdwt2d_untiled(y, wcoeff, 2)
end

% wavelet demosaicing using cycle spinning
function [y : cube] = wavelet_demosaicing_spinning(x : mat, pattern : mat, wcoeff)
    y = zeros(size(x,0),size(x,1),3)
    y1 = zeros(size(x,0),size(x,1),3)
    for m=0..3
        for n=0..3
            w1 = circshift(x, [m,n,0])
            w2 = copy(w1)
            pdwt2d_untiled(w1, wcoeff, 2, false)
            pdwt2d_untiled(w2, wcoeff, 2, true)

            shifted_pattern = circshift(pattern, [m,n])
            [signHHLL,signLHLLr,signHLLLr,signLHLLb,signHLLLb] = wavelet_demosaicing_signs(shifted_pattern)

            parallel_do(size(w1,0..1)/4,w1,w2,y1,signHHLL,signLHLLb,signHLLLb,signLHLLr,signHLLLr,wavelet_demosaic_bands)

            % compute inverse wavelet packet transform
            ipdwt2d_untiled(y1, wcoeff, 2)
            y = y + circshift(y1, -[m,n,0])/16
        end
    end
end

% Copy the XXLL bands, i.e., HHLL, LHLL, HLLL, which were obtained using the reversed
% highpass filter coefficients, to dst. This allows to save a lot of memory, since
% the original HHLL, LHLL, HLLL bands in dst are no longer used.
function [] = __kernel__ copy_XXLL_bands(dst : mat'unchecked, src : mat'unchecked, pos : ivec2)
    m = int(size(src,0)/4)
    n = int(size(src,1)/4)

    % Band indices
    LH = [0, 2*n]
    HL = [2*m, 0]
    HH = [2*m, 2*n]
    LLLL = pos
    LHLL = LH + LLLL
    HLLL = HL + LLLL
    HHLL = HH + LLLL
    dst[HHLL] = src[HHLL]
    dst[LHLL] = src[LHLL]
    dst[HLLL] = src[HLLL]
end

function [] = __kernel__ cplxwavelet_demosaic_bands(w1 : mat'unchecked, w3 : cube'unchecked, _
    w_LHLL_sum : mat'unchecked,w_HLLL_sum : mat'unchecked, _
    signHHLL : int, signLHLLr : int, signHLLLr : int, signLHLLg : int, signHLLLg : int, _
    signLHLLb : int, signHLLLb : int, pos : ivec2)

    m = int(size(w1,0)/4)
    n = int(size(w1,1)/4)

    % Band indices
    LH = [0, 2*n]
    HL = [2*m, 0]
    HH = [2*m, 2*n]
    LLLL = pos
    LLLH = LLLL + [0, n]
    LLHL = LLLL + [m, 0]
    LLHH = LLLL + [m, n]
    LHLL = LH + LLLL
    LHLH = LH + LLLH
    LHHL = LH + LLHL
    LHHH = LH + LLHH
    HLLL = HL + LLLL
    HLLH = HL + LLLH
    HLHL = HL + LLHL
    HLHH = HL + LLHH
    HHLL = HH + LLLL
    HHLH = HH + LLLH
    HHHL = HH + LLHL
    HHHH = HH + LLHH

    % Make a decision of edge orientation
    C_m = 14.  % cost of a miss detection
    C_u = 5.  % cost of an unsure detection
    log_C_u_C_m = log(C_u/C_m)
    w_LHLL = w_LHLL_sum[pos]
    w_HLLL = w_HLLL_sum[pos]
    b_max = max(w_LHLL, w_HLLL)/4
    b_min = min(w_LHLL, w_HLLL)/4
    if (w_LHLL/b_max < log_C_u_C_m - w_HLLL/b_min || _
        w_HLLL/b_max < log_C_u_C_m - w_LHLL/b_max)
        map = 0.5 % unsure detection
    else
        map = float(w_LHLL > w_HLLL)
    endif
    % border: unsure detection
    if max(pos)<=1 || min(pos-HH)>=-1
        map = 0.5
    endif

    % Green subbands
    w3[ch(HHLH,1)] = w1[HHLH]
    w3[ch(HHHL,1)] = w1[HHHL]
    w3[ch(HHHH,1)] = w1[HHHH]
    w3[ch(HHLL,1)] = 0.0

    w3[ch(LHLH,1)] = w1[LHLH]
    w3[ch(LHHL,1)] = w1[LHHL]
    w3[ch(LHHH,1)] = w1[LHHH]
    w3[ch(LHLL,1)] = map == 1.0 ? (signLHLLg * w1[LHLL] - signHLLLg*w1[HLLL]) : 0.0 % reconstruction of high frequencies

    w3[ch(HLLH,1)] = w1[HLLH]
    w3[ch(HLHL,1)] = w1[HLHL]
    w3[ch(HLHH,1)] = w1[HLHH]
    w3[ch(HLLL,1)] = map == 0.0 ? (signHLLLg * w1[HLLL] - signLHLLg*w1[LHLL]) : 0.0 % reconstruction of high frequencies

    w3[ch(LLLH,1)] = w1[LLLH]
    w3[ch(LLHL,1)] = w1[LLHL]
    w3[ch(LLHH,1)] = w1[LLHH]
    w3[ch(LLLL,1)] = w1[LLLL] - signHHLL * w1[HHLL]

    % Red subbands
    w3[ch(HHLH,0)] = w3[ch(HHLH,1)]
    w3[ch(HHHL,0)] = w3[ch(HHHL,1)]
    w3[ch(HHHH,0)] = w3[ch(HHHH,1)]
    w3[ch(HHLL,0)] = w3[ch(HHLL,1)]

    w3[ch(LHLH,0)] = w3[ch(LHLH,1)]
    w3[ch(LHHL,0)] = w3[ch(LHHL,1)]
    w3[ch(LHHH,0)] = w3[ch(LHHH,1)]
    w3[ch(LHLL,0)] = w3[ch(LHLL,1)]

    w3[ch(HLLH,0)] = w3[ch(HLLH,1)]
    w3[ch(HLHL,0)] = w3[ch(HLHL,1)]
    w3[ch(HLHH,0)] = w3[ch(HLHH,1)]
    w3[ch(HLLL,0)] = w3[ch(HLLL,1)]

    w3[ch(LLLH,0)] = w3[ch(LLLH,1)]
    w3[ch(LLHL,0)] = w3[ch(LLHL,1)]
    w3[ch(LLHH,0)] = w3[ch(LLHH,1)]
    w3[ch(LLLL,0)] = 2*(signHHLL*w1[HHLL] + signLHLLr*w1[LHLL]*(1-map) + signHLLLr*w1[HLLL]*map) + w3[ch(LLLL,1)]

    % Blue subbands
    w3[ch(HHLH,2)] = w3[ch(HHLH,1)]
    w3[ch(HHHL,2)] = w3[ch(HHHL,1)]
    w3[ch(HHHH,2)] = w3[ch(HHHH,1)]
    w3[ch(HHLL,2)] = w3[ch(HHLL,1)]

    w3[ch(LHLH,2)] = w3[ch(LHLH,1)]
    w3[ch(LHHL,2)] = w3[ch(LHHL,1)]
    w3[ch(LHHH,2)] = w3[ch(LHHH,1)]
    w3[ch(LHLL,2)] = w3[ch(LHLL,1)]

    w3[ch(HLLH,2)] = w3[ch(HLLH,1)]
    w3[ch(HLHL,2)] = w3[ch(HLHL,1)]
    w3[ch(HLHH,2)] = w3[ch(HLHH,1)]
    w3[ch(HLLL,2)] = w3[ch(HLLL,1)]

    w3[ch(LLLH,2)] = w3[ch(LLLH,1)]
    w3[ch(LLHL,2)] = w3[ch(LLHL,1)]
    w3[ch(LLHH,2)] = w3[ch(LLHH,1)]
    w3[ch(LLLL,2)] = 2*(signHHLL*w1[HHLL] + signLHLLb*w1[LHLL]*(1-map) + signHLLLb*w1[HLLL]*map) + w3[ch(LLLL,1)]

end

% complex wavelet demosaicing
function [y : cube] = cplxwavelet_demosaicing(x : mat, pattern : mat, wcoeff1 : vec[wavelet_params](2), wcoeff2 : vec[wavelet_params](2))
    w1 = pdtcwt(x, wcoeff1, wcoeff2, 2, false, false)
    w2 = pdtcwt(x, wcoeff1, wcoeff2, 2, false, true)
    for m=0..1
        for n=0..1
            parallel_do(size(w1[m,n],0..1)/4,w1[m,n],w2[m,n],copy_XXLL_bands)
        end
    end
    w2 = 0 % Release the memory used by w2

    y1 = cell(2,2)

    M = size(x,0)/2
    N = size(x,1)/2

    % step 1 - estimate the dominant orientation
    w_LHLL_sum = 0
    w_HLLL_sum = 0

    for m=0..1
        for n=0..1
            w_LHLL = w1[m,n][(0..M/2-1),N+(0..N/2-1)]
            w_HLLL = w1[m,n][M+(0..M/2-1),(0..N/2-1)]

            % decompose one more scale...
            pdwt2d_untiled_rowcol(w_LHLL, size(w_LHLL,0..2), wcoeff2[m], wcoeff2[n], 1)
            pdwt2d_untiled_rowcol(w_HLLL, size(w_HLLL,0..2), wcoeff2[m], wcoeff2[n], 1)
            % set the LL coefficients to zero (effectively creating a bandpass filter)
            w_LHLL[0..M/4-1,0..N/4-1] = 0
            w_HLLL[0..M/4-1,0..N/4-1] = 0
            % reconstruct this scale
            ipdwt2d_untiled_rowcol(w_LHLL, size(w_LHLL,0..2)/2, wcoeff2[m], wcoeff2[n], 1)
            ipdwt2d_untiled_rowcol(w_HLLL, size(w_HLLL,0..2)/2, wcoeff2[m], wcoeff2[n], 1)

            w_LHLL_sum = w_LHLL_sum + abs(w_LHLL)
            w_HLLL_sum = w_HLLL_sum + abs(w_HLLL)
        end
    end

    for m=0..1
        for n=0..1
            [signHHLL,signLHLLr,signHLLLr,signLHLLb,signHLLLb] = wavelet_demosaicing_signs(pattern)
            signLHLLg = signLHLLb %(-1)^(m+n)
            signHLLLg = signHLLLb % (-1)^(m+n)
            y1[m,n] = zeros(size(x,0),size(x,1),3)
            parallel_do(size(w1[m,n],0..1)/4,w1[m,n],y1[m,n],w_LHLL_sum,w_HLLL_sum, _
                signHHLL,signLHLLb,signHLLLb,signLHLLg,signHLLLg,signLHLLr,signHLLLr,cplxwavelet_demosaic_bands)
        end
    end

    w_LHLL_sum = 0 % release memory used by w_LHLL_sum
    w_HLLL_sum = 0 % release memory used by w_HLLL_sum
    w1 = 0 % release memory used by w1

    y = pidtcwt(y1, wcoeff1, wcoeff2, 2, false)
end

function [] = main()
    %pattern = [[0,1],[3,2]]
    pattern = [[0,1],[3,2]]  %*default
    %pattern = [[1,0],[2,3]]
    %pattern = [[1,2],[0,3]]
    im = imread("lighthouse.png")    
    %im = imread("F:\Temp\left0000.bmp")
    %im = imread("F:\Data\phaseone\N_D3X-ISO100-S8_crop3.png")
    %im = imread("lena_big.tif")    
    im_bayer = bayer_subsampling(im,pattern)
    im_mosaic = bayer_mosaicimage(im_bayer,pattern)   
    %im_mosaic = imread("C:\Users\bgoossen\Pictures\hiep.png")[:,:,0]
    
    %pattern = [[3,2],[0,1]]  
    %im_recon = wavelet_demosaicing(im_mosaic, pattern, db2)
    
    w1 = complex_orthogonal_wavelet(filtercoeff_farras)
    w2 = complex_orthogonal_wavelet(filtercoeff_selcw[4,3])
    tic()
    im_recon = cplxwavelet_demosaicing(im_mosaic, pattern, w1, w2)
    toc()
    %imshow(im_mosaic)
    imshow(im_recon)
    im=0
    return
    
    tic()
    %im_recon = wavelet_demosaicing(im_mosaic, pattern, db2)
    for k=0..9
    %im_recon = wavelet_demosaicing_spinning(im_mosaic, pattern, db2)
    im_recon = cplxwavelet_demosaicing(im_mosaic, pattern, w1, w2)
    end
    toc()
    
    
    imshow(im)
    imshow(im_recon)
    imshow(im_recon-im,[])
    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    print "psnr=", psnr(im,im_recon), "dB"
end