import "wavelet_demosaicing.q"
import "denoise_vectorprobshrink.q"
import "covariance.q"

% conversion to 3d bands
ch = __device__ (x : ivec2, y : int) -> [x[0],x[1],y]

% Function: denoisaic_band_vectorprobshrink
%
% ProbShrink denoisaicing with the complex wavelet packet transform!
%
% Parameters:
%    w - wavelet subband to apply joint denoising and demosaicing
%    y - denoised and demosaiced complex RGB output data
%    sigma_n - the standard deviation of the noise, per RGB color channel (TO DO - make position dependent)
%    T - signal of interest threshold
%    A - demosaicing matrices for the three possible interpolation choices (horizontal, vertical, unsure)
%    map - the selected interpolation method (horizontal, vertical or unsure)
%    sub - the subsampling factor for map (because the interpolation map is stored in a
%          higher resolution than the wavelet coefficients.
%
% Notes:
%    The algorithm does not take a local neighborhood into account, i.e. the
%    algorithm performs poisewise coefficient processing. However, inter-
%    luminance/chrominance correlations are considered, leading to 6x6 covariance
%    matrices.
%
function [] = denoisaic_band_vectorprobshrink(w : cube(:,:,8), y : cube, sigma_n : vec4, _
    T : scalar, A : cube, map : mat, sub : int)

    sz = size(w,0..1)
    C_w = covariance_8x8(w)

    d = 6 % number of dimensions
    D_n = diagm([sigma_n[0], sigma_n[1], sigma_n[2], sigma_n[3], _
                sigma_n[0], sigma_n[1], sigma_n[2], sigma_n[3]].^2)
    Var_x = correct_eigenvalues(C_w - D_n)

    C_n0 = A[:,:,0] * D_n * transpose(A[:,:,0])
    C_n1 = A[:,:,1] * D_n * transpose(A[:,:,1])
    C_n2 = A[:,:,2] * D_n * transpose(A[:,:,2])

    C_x0 = A[:,:,0] * Var_x * transpose(A[:,:,0])
    C_x1 = A[:,:,1] * Var_x * transpose(A[:,:,1])
    C_x2 = A[:,:,2] * Var_x * transpose(A[:,:,2])

    % Prior distribution for z & the number of mixture components
    K = 4
    [z, p_z] = bkf_prior(K, 1.0) % TODO - estimate tau from the image

    % Symmetric matrix square root of the noise covariance
    [S0, S_inv0] = symsqrt_and_inv(C_n0)
    [S1, S_inv1] = symsqrt_and_inv(C_n1)
    [S2, S_inv2] = symsqrt_and_inv(C_n2)

    % Compute S^{-1}C_{x}S^{-T} -> SVD Q Lambda Q^T
    [Q0, Lambda0, Q3] = svd(S_inv0*C_x0*transpose(S_inv0))
    [Q1, Lambda1, Q3] = svd(S_inv1*C_x1*transpose(S_inv1))
    [Q2, Lambda2, Q3] = svd(S_inv2*C_x2*transpose(S_inv2))
    Lambda = {diag(Lambda0),diag(Lambda1),diag(Lambda2)}

    % Compute U=(SQ)^{-1}*A; this transform will decorrelate the observed noisy samples
	% The matrix inverse can be replaced by a multiplication
    U = zeros(6,8,3)
    U[:,:,0] = transpose(Q0) * S_inv0 * A[:,:,0]
    U[:,:,1] = transpose(Q1) * S_inv1 * A[:,:,1]
    U[:,:,2] = transpose(Q2) * S_inv2 * A[:,:,2]

    % Compute the probability that the signal-of-interest is absent (p(H_0)),
    % which is constant for the whole subband
    p_H0 = zeros(3,1)
    p_H0_comp = zeros(3,K)
    for k=0..K-1
        p_H0_comp[0,k] = exp(0.5*sum(log(T) - log(T + Lambda0 * z[k])))
        p_H0_comp[1,k] = exp(0.5*sum(log(T) - log(T + Lambda1 * z[k])))
        p_H0_comp[2,k] = exp(0.5*sum(log(T) - log(T + Lambda2 * z[k])))
        p_H0 += p_H0_comp[:,k] * p_z[k]
    end

    % Compute normalisation factors for the densities - use logarithm again
    % to avoid numerical underflows in later steps
    rho = zeros(K,3)
    tau = zeros(K,3)
    rho_gn = zeros(K,d,3)
    tau_gn = zeros(K,d,3)
    for k=0..K-1
        for m=0..2
            r = z[k] * max(Lambda[m],1e-4)
            rho[k,m] = -0.5*sum(log(1+r)) + log(p_z[k]) + 4*d
            tau[k,m] =  0.5*sum(log((T+r)./(T+((1+T)*r)))) + log(p_z[k]) + log(p_H0_comp[m,k]) + 4*d
            rho_gn[k,:,m] = -0.5./(1+r)
            tau_gn[k,:,m] = -0.5*(T+r)./(T+((1+T)*r))
        end
    end

    function [] = __kernel__ denoisaic_kernel(x : cube'unchecked, y : cube'unchecked, U : cube'unchecked, _
        A : cube'unchecked, rho_gn : cube'unchecked, tau_gn : cube'unchecked, _
        rho : mat'unchecked, tau : mat'unchecked, map : mat'unchecked, sub : int, _
        pos : ivec2, blkdim : ivec2, blkpos : ivec2)
        
        K = size(rho,0)
        assert(blkdim[0]*blkdim[1]*6 <= 1024*6)
        v = shared(blkdim[0],blkdim[1],6)
        Dm = int(map[pos[0]*sub+sub/2,pos[1]*sub+sub/2]) % The selected interpolation method

        % decorrelate the observed coefficient vector
        for m=0..5
            sum1 = 0.0
            sum2 = 0.0
            for n=0..7
                sum1 += U[m,n,Dm] * x[pos[0],pos[1],n]
                sum2 += A[m,n,Dm] * x[pos[0],pos[1],n]
            end
            v[blkpos[0],blkpos[1],m] = sum1
            y[pos[0],pos[1],m] = sum2
        end

        % compute the likelihood p(y|z_k)
        p_y = 0.0
        p_yH0 = 0.0
        
        for k=0..K-1
            sum1 = rho[k,Dm] % related to f(y)
            sum2 = tau[k,Dm] % related to f(y|H_0)
            for i=0..5
                val = v[blkpos[0],blkpos[1],i]
                val *= val               
                sum1 += val*rho_gn[k,i,Dm]
                sum2 += val*tau_gn[k,i,Dm]
            end
            p_y += exp(sum1)
            p_yH0 += exp(sum2)
        end
        
        p_H0_y = p_yH0/p_y
        if p_H0_y > 1.0 || isnan(p_H0_y)
            p_H0_y = 1.0        
        endif
        for m=0..5
            y[pos[0],pos[1],m] *= (1.0-p_H0_y)  % shrinkage!
        end
    end
        
    parallel_do([sz[0],sz[1],1],w,y,U,A,rho_gn,tau_gn,rho,tau,map,sub,denoisaic_kernel)
end

% Function: denoisaic_band_blsgsm
%
% Denoisaicing of a subband using the BLS-GSM method
%
% Parameters:
%    w - wavelet subband to apply joint denoising and demosaicing
%    y - denoised and demosaiced complex RGB output data
%    sigma_n - the standard deviation of the noise, per RGB color channel (TO DO - make position dependent)
%    A - demosaicing matrices for the three possible interpolation choices (horizontal, vertical, unsure)
%    map - the selected interpolation method (horizontal, vertical or unsure)
%    sub - the subsampling factor for map (because the interpolation map is stored in a
%          higher resolution than the wavelet coefficients.
%
% Notes:
%    The algorithm does not take a local neighborhood into account, i.e. the
%    algorithm performs poisewise coefficient processing. However, inter-
%    luminance/chrominance correlations are considered, leading to 6x6 covariance
%    matrices.
%
function [] = denoisaic_band_blsgsm(w : cube(:,:,8), y : cube, sigma_n : vec4, _
    A : cube, map : mat, sub : int)

    sz = size(w,0..1)
    C_w = covariance_8x8(w)
    
    d : int'const = 6 % number of dimensions
    D_n = diagm([sigma_n[0], sigma_n[1], sigma_n[2], sigma_n[3], _
                sigma_n[0], sigma_n[1], sigma_n[2], sigma_n[3]].^2)
    Var_x = correct_eigenvalues(C_w - D_n)
    
    C_n0 = A[:,:,0] * D_n * transpose(A[:,:,0])
    C_n1 = A[:,:,1] * D_n * transpose(A[:,:,1])
    C_n2 = A[:,:,2] * D_n * transpose(A[:,:,2])

    C_x0 = A[:,:,0] * Var_x * transpose(A[:,:,0])
    C_x1 = A[:,:,1] * Var_x * transpose(A[:,:,1])
    C_x2 = A[:,:,2] * Var_x * transpose(A[:,:,2])

    % Prior distribution for z & the number of mixture components
    K : int'const = 4
    [z, p_z] = bkf_prior(K, 1.0) % TODO - estimate tau from the image    

    % Symmetric matrix square root of the noise covariance
    [S0, S_inv0] = symsqrt_and_inv(C_n0)
    [S1, S_inv1] = symsqrt_and_inv(C_n1)
    [S2, S_inv2] = symsqrt_and_inv(C_n2)

    % Compute S^{-1}C_{x}S^{-T} -> SVD Q Lambda Q^T
    [Q0, Lambda0, Q3] = svd(S_inv0*C_x0*transpose(S_inv0))
    [Q1, Lambda1, Q3] = svd(S_inv1*C_x1*transpose(S_inv1))
    [Q2, Lambda2, Q3] = svd(S_inv2*C_x2*transpose(S_inv2))
    Lambda = [diag(Lambda0),diag(Lambda1),diag(Lambda2)]

    % Compute U=(SQ)^{-1}*A; this transform will decorrelate the observed noisy samples
    % The matrix inverse can be replaced by a multiplication
    U = zeros(6,8,3)
    U[:,:,0] = transpose(Q0) * S_inv0 * A[:,:,0]
    U[:,:,1] = transpose(Q1) * S_inv1 * A[:,:,1]
    U[:,:,2] = transpose(Q2) * S_inv2 * A[:,:,2]

    % Compute U^(-1)=SQ
    U_inv = zeros(6,6,3)
    U_inv[:,:,0] = S0*Q0
    U_inv[:,:,1] = S1*Q1
    U_inv[:,:,2] = S2*Q2

    % Correct eigenvalues if necessary
    epsilon = 1e-4
    Lambda = max(epsilon*ones(3,d), Lambda)

    % Logarithm of the determinant of C_z=z C_x + C_n
    v_ld = zeros(K,3)
    for k=0..K-1
        v_ld[k,0] = 4*d + logdet(z[k] * C_x0 + C_n0)
        v_ld[k,1] = 4*d + logdet(z[k] * C_x1 + C_n1)
        v_ld[k,2] = 4*d + logdet(z[k] * C_x2 + C_n2)
    end

    % Estimator for stationary noise
    function [] = __kernel__ denoisaic_kernel(x : cube'unchecked, y : cube'unchecked, U : cube'unchecked(6,8,3), 
        U_inv : cube'unchecked(6,6,3), v_ld : mat'unchecked(K,3), Lambda : mat'unchecked(3,d), _
        z : vec'unchecked(K), p_z : vec'unchecked(K), pos : ivec2, blkdim : ivec2, blkpos : ivec2)
        
        K = numel(v_ld)
        assert(blkdim[0]*blkdim[1]*6 <= 1024*6)
        v = shared(blkdim[0],blkdim[1],6)
        Dm = int(map[pos[0]*sub+sub/2,pos[1]*sub+sub/2]) % The selected interpolation method
        
        p_yz = shared(K,blkdim[0],blkdim[1])

        % decorrelate the observed coefficient vector
        for m=0..5
            sum = 0.0
            for n=0..7
                sum += U[m,n,Dm] * x[pos[0],pos[1],n]
            end
            v[blkpos[0],blkpos[1],m] = sum
        end

        % compute the likelihood p(y|z_k)
        p_y = 0.0
        for k=0..K-1
            sum = v_ld[k,Dm]
            for i=0..5
                val = v[blkpos[0],blkpos[1],i]
                sum += val*val/(z[k]*Lambda[Dm,i]+1)
            end
            val = p_z[k]*exp(-0.5*sum)
            p_yz[k,blkpos[0],blkpos[1]] = val
            p_y += val
        end

        % estimate the noise-free signal
        y_est = zeros(6)
        if p_y > 0            
            for k=0..K-1
                val = 0.0
                for m=0..5
                    for i=0..5
                        r = z[k] * Lambda[Dm,i]
                        val += U_inv[m,n,Dm] * r/(r+1)*v[blkpos[0],blkpos[1],i]
                    end
                    y_est[m] += val * p_yz[k,blkpos[0],blkpos[1]]/p_y
                end
            end           
        endif
        y[pos[0],pos[1],0..5] = y_est
    end
    
    parallel_do([[sz[0],sz[1],1],max_block_size(denoisaic_kernel,[16,32,1])], _
        w,y,U,U_inv,v_ld,Lambda,z,p_z,denoisaic_kernel)    
end


%
% Dedicated denoisaicing algorithm for the LHLL and HLLL bands
%
function [] = denoisaic_LHLL_HLLL_band_vectorprobshrink(w : cube'unchecked(:,:,4), y : cube'unchecked, sigma_n : vec4, _
    T : scalar, A : cube'unchecked, map : mat'unchecked)

    sz = size(w,0..1)
    C_w = covariance_4x4(w)

    d = 4 % number of dimensions
    D_n = diagm(sigma_n[1].^2 * [1,1,1,1])
    Var_x = correct_eigenvalues(C_w - D_n)

    C_n0 = A[:,:,0] * D_n * transpose(A[:,:,0])
    C_n1 = A[:,:,1] * D_n * transpose(A[:,:,1])
    C_n2 = A[:,:,2] * D_n * transpose(A[:,:,2])

    C_x0 = A[:,:,0] * Var_x * transpose(A[:,:,0])
    C_x1 = A[:,:,1] * Var_x * transpose(A[:,:,1])
    C_x2 = A[:,:,2] * Var_x * transpose(A[:,:,2])

    % Prior distribution for z & the number of mixture components
    K = 4
    [z, p_z] = bkf_prior(K, 1.0) % TODO - estimate tau from the image

    % Symmetric matrix square root of the noise covariance
    [S0, S_inv0] = symsqrt_and_inv(C_n0)
    [S1, S_inv1] = symsqrt_and_inv(C_n1)
    [S2, S_inv2] = symsqrt_and_inv(C_n2)

    % Compute S^{-1}C_{x}S^{-T} -> SVD Q Lambda Q^T
    [Q0, Lambda0, Q3] = svd(S_inv0*C_x0*transpose(S_inv0))
    [Q1, Lambda1, Q3] = svd(S_inv1*C_x1*transpose(S_inv1))
    [Q2, Lambda2, Q3] = svd(S_inv2*C_x2*transpose(S_inv2))
    Lambda = {diag(Lambda0),diag(Lambda1),diag(Lambda2)}

    % Compute U=(SQ)^{-1}*A; this transform will decorrelate the observed noisy samples
	% The matrix inverse can be replaced by a multiplication
    U = zeros(4,4,3)
    U[:,:,0] = transpose(Q0) * S_inv0 * A[:,:,0]
    U[:,:,1] = transpose(Q1) * S_inv1 * A[:,:,1]
    U[:,:,2] = transpose(Q2) * S_inv2 * A[:,:,2]

    % Compute the probability that the signal-of-interest is absent (p(H_0)),
    % which is constant for the whole subband
    p_H0 = zeros(3,1)
    p_H0_comp = zeros(3,K)
    for k=0..K-1
        p_H0_comp[0,k] = exp(0.5*sum(log(T) - log(T + Lambda0 * z[k])))
        p_H0_comp[1,k] = exp(0.5*sum(log(T) - log(T + Lambda1 * z[k])))
        p_H0_comp[2,k] = exp(0.5*sum(log(T) - log(T + Lambda2 * z[k])))
        p_H0 += p_H0_comp[:,k] * p_z[k]
    end

    % Compute normalisation factors for the densities - use logarithm again
    % to avoid numerical underflows in later steps
    rho = zeros(K,3)
    tau = zeros(K,3)
    rho_gn = zeros(K,d,3)
    tau_gn = zeros(K,d,3)
    for k=0..K-1
        for m=0..2
            r = z[k] * max(Lambda[m],1e-4)
            rho[k,m] = -0.5*sum(log(1+r)) + log(p_z[k]) + 4*d
            tau[k,m] =  0.5*sum(log((T+r)./(T+((1+T)*r)))) + log(p_z[k]) + log(p_H0_comp[m,k]) + 4*d
            rho_gn[k,:,m] = -0.5./(1+r)
            tau_gn[k,:,m] = -0.5*(T+r)./(T+((1+T)*r))
        end
    end

    function [] = __kernel__ denoisaic_kernel(x : cube'unchecked, y : cube'unchecked, U : cube'unchecked, _
        A : cube'unchecked, rho_gn : cube'unchecked, tau_gn : cube'unchecked, _
        rho : mat'unchecked, tau : mat'unchecked, map : mat'unchecked, _
        pos : ivec2, blkdim : ivec2, blkpos : ivec2)
        
        K = size(rho,0)
        assert(blkdim[0] * blkdim[1] * 4 <= 1024 * 4)
        v = shared(blkdim[0],blkdim[1],4)
        Dm = int(map[pos]) % The selected interpolation method

        % decorrelate the observed coefficient vector
        for m=0..3
            sum1 = 0.0
            sum2 = 0.0
            for n=0..3
                sum1 += U[m,n,Dm] * x[pos[0],pos[1],n]
                sum2 += A[m,n,Dm] * x[pos[0],pos[1],n]
            end
            v[blkpos[0],blkpos[1],m] = sum1
            y[pos[0],pos[1],m] = sum2
        end

        % compute the likelihood p(y|z_k)
        p_y = 0.0
        p_yH0 = 0.0
        
        for k=0..K-1
            sum1 = rho[k,Dm] % related to f(y)
            sum2 = tau[k,Dm] % related to f(y|H_0)
            for i=0..3
                val = v[blkpos[0],blkpos[1],i]
                val *= val               
                sum1 += val*rho_gn[k,i,Dm]
                sum2 += val*tau_gn[k,i,Dm]
            end
            p_y += exp(sum1)
            p_yH0 += exp(sum2)
        end
        
        p_H0_y = p_yH0/p_y
        if p_H0_y > 1.0 || isnan(p_H0_y)
            p_H0_y = 1.0        
        endif

        % demosaicing & shrinkage
        y[pos[0],pos[1],0..3] = y[pos[0],pos[1],0..3] * (1.0-p_H0_y)
    end
        
    parallel_do([sz[0],sz[1],1],w,y,U,A,rho_gn,tau_gn,rho,tau,map,denoisaic_kernel)
end

function [] = __kernel__ estimate_edge_direction( _
    w_LHLL_sum : mat'unchecked, w_HLLL_sum : mat'unchecked, map : mat'unchecked, pos : ivec2)
    % Make a decision of edge orientation
    C_m = 14.  % cost of a miss detection
    C_u = 5.  % cost of an unsure detection

    log_C_u_C_m = log(C_u/C_m)
    w_LHLL = w_LHLL_sum[pos]
    w_HLLL = w_HLLL_sum[pos]
    b_max = max(w_LHLL, w_HLLL)/4
    b_min = min(w_LHLL, w_HLLL)/4
    if (w_LHLL/b_max < log_C_u_C_m - w_HLLL/b_min || _
        w_HLLL/b_max < log_C_u_C_m - w_LHLL/b_max)
        mapv = 0.5 % unsure detection
    else
        mapv = float(w_LHLL > w_HLLL)
    endif
    % border: unsure detection
    if max(pos)<=3 || min(pos-size(map))>=-3
        mapv = 0.5
    endif
    map[pos] = mapv*2
end

function [] = __kernel__ cplxwavelet_demux_bands(w1 : mat'unchecked, w3 : cube'unchecked, _
    signHHLL : int, signLHLLr : int, signHLLLr : int, signLHLLg : int, signHLLLg : int, _
    signLHLLb : int, signHLLLb : int, pos : ivec2)

    m = int(size(w1,0)/4)
    n = int(size(w1,1)/4)

    % Band indices
    LH = [0, 2*n]
    HL = [2*m, 0]
    HH = [2*m, 2*n]
    LLLL = pos
    LLLH = LLLL + [0, n]
    LLHL = LLLL + [m, 0]
    LLHH = LLLL + [m, n]
    LHLL = LH + LLLL
    LHLH = LH + LLLH
    LHHL = LH + LLHL
    LHHH = LH + LLHH
    HLLL = HL + LLLL
    HLLH = HL + LLLH
    HLHL = HL + LLHL
    HLHH = HL + LLHH
    HHLL = HH + LLLL
    HHLH = HH + LLLH
    HHHL = HH + LLHL
    HHHH = HH + LLHH
    
    % Luminance
    w3[HHLH[0],HHLH[1],0..2] = w1[HHLH]
    w3[HHHL[0],HHHL[1],0..2] = w1[HHHL]
    w3[HHHH[0],HHHH[1],0..2] = w1[HHHH]
    w3[HHLL[0],HHLL[1],0..2] = 0.

    w3[LHLH[0],LHLH[1],0..2] = w1[LHLH]
    w3[LHHL[0],LHHL[1],0..2] = w1[LHHL]
    w3[LHHH[0],LHHH[1],0..2] = w1[LHHH]
    w3[LHLL[0],LHLL[1],0..2] = 0.

    w3[HLLH[0],HLLH[1],0..2] = w1[HLLH]
    w3[HLHL[0],HLHL[1],0..2] = w1[HLHL]
    w3[HLHH[0],HLHH[1],0..2] = w1[HLHH]
    w3[HLLL[0],HLLL[1],0..2] = 0.

    w3[LLLH[0],LLLH[1],0..2] = w1[LLLH]
    w3[LLHL[0],LLHL[1],0..2] = w1[LLHL]
    w3[LLHH[0],LLHH[1],0..2] = w1[LLHH]
end

%
% Extract a band of wavelet coefficients out of the wavelet packet transform coefficients
% The output is a matrix with 8 components (LLLL/LHLL/HLLL,HHLL coefficients and real
% and imaginary parts)
%
function [] = __kernel__ extract_bands(y : cube'unchecked, w1 : mat'unchecked, w2 : mat'unchecked, offset : ivec2, pos : ivec2)
    m = int(size(w1,0)/2)
    n = int(size(w1,1)/2)
    LLLL = [0, 0]
    LHLL = [0, n]
    HLLL = [m, 0]
    HHLL = [m, n]
    y[pos[0],pos[1],0..7] = [w1[offset + pos + LLLL], w1[offset + pos + LHLL], w1[offset + pos + HLLL], w1[offset + pos + HHLL], _
                             w2[offset + pos + LLLL], w2[offset + pos + LHLL], w2[offset + pos + HLLL], w2[offset + pos + HHLL]]
end

%
% Put back processed subbands in the output wavelet packet transform band images (w1, w2)
% Also performs the complex phase modulation (starting from complex wavelet coefficients)
% The input subband has 6 components: RGB_r RGB_i. The first trio are the real components,
% the second trio are the imaginary components
%
function [] = __kernel__ restore_bands(x : cube'unchecked, w1 : cube'unchecked, w2 : cube'unchecked, offset : ivec2, pos : ivec2)
    p = [pos[0]+offset[0],pos[1]+offset[1]]
    RGB_r = x[pos[0],pos[1],0..2]
    RGB_i = x[pos[0],pos[1],3..5]
    w1[p[0],p[1],0..2] = 0.5 * (RGB_r + RGB_i)
    w2[p[0],p[1],0..2] = 0.5 * (RGB_r - RGB_i)
end

%
% Kernel for reconstructing the lowpass part of LLLL - uses the interpolation 
% formula according to the orientation map
%
function [] = __kernel__ reconstruct_ll(x : cube'unchecked, y : cube'unchecked, A : cube'unchecked, _
    map : mat'unchecked, sub : int, pos:ivec3)
    n = int(map[pos[0]*sub,pos[1]*sub]) % Retrieve the interpolation formula...
    y[pos] = dotprod(A[pos[2],0..7,n], x[pos[0],pos[1],0..7])
end

function [] = __kernel__ extract_LHLL_HLLL_bands(y : cube'unchecked, w1 : mat'unchecked, w2 : mat'unchecked, pos : ivec2)
    m = int(size(w1,0)/2)
    n = int(size(w1,1)/2)
    LHLL = [0, n]
    HLLL = [m, 0]
    y[pos[0],pos[1],0..3] = [w1[pos + LHLL], w1[pos + HLLL], w2[pos + LHLL], w2[pos + HLLL]]
end

function [] = __kernel__ restore_LHLL_HLLL_bands(x : cube'unchecked, w1 : cube'unchecked, w2 : cube'unchecked, pos : ivec2)
    m = int(size(w1,0)/2)
    n = int(size(w1,1)/2)
    LHLL = [0, n]
    HLLL = [m, 0]
    w1[pos[0] + LHLL[0], pos[1] + LHLL[1], 0..2] = x[pos[0],pos[1],0]
    w1[pos[0] + HLLL[0], pos[1] + HLLL[1], 0..2] = x[pos[0],pos[1],1]
    w2[pos[0] + LHLL[0], pos[1] + LHLL[1], 0..2] = x[pos[0],pos[1],2]
    w2[pos[0] + HLLL[0], pos[1] + HLLL[1], 0..2] = x[pos[0],pos[1],3]
end
    
% complex wavelet demosaicing
function [y : cube] = cplxwavelet_denoisaicing(x : mat, pattern : mat, wcoeff1 : vec[wavelet_params](2), wcoeff2 : vec[wavelet_params](2), num_scales : int, sigma_n : vec4)
    w1 = pdtcwt(x, wcoeff1, wcoeff2, 2, false, false)
    w2 = pdtcwt(x, wcoeff1, wcoeff2, 2, false, true)
    for m=0..1
        for n=0..1
            parallel_do(size(w1[m,n],0..1)/4,w1[m,n],w2[m,n],copy_XXLL_bands)
        end
    end
    w2 = 0 % Release the memory used by w2

    y1 = cell(2,2)
    M = int(size(x,0)/2)
    N = int(size(x,1)/2)

    % Check the input dimensions
    if any(mod(size(x,0..1),2^num_scales)!=0)
        error "(cplxwavelet_denoisaicing) Image dimensions should be a multiple of ",2^num_scales,"!"
    endif

    % denoising parameters
    wnd = [3,3] % size of the local window for the luminance bands
    C_n = eye(9) % due to orthonormal transform    
    T = 2.5
    T2 = 1.0
    J = num_scales % number of scales

    %===== estimate the dominant orientation & denoise =====
    % the pure luminance subbands
    w_LHLL_sum = 0
    w_HLLL_sum = 0    
    pair = [[0,3],[2,1]] % Contains pairs of real and imaginary band indices

    for m=0..1
        for n=0..1
            w_LHLL = w1[m,n][(0..M/2-1),N+(0..N/2-1)]
            w_HLLL = w1[m,n][M+(0..M/2-1),(0..N/2-1)]

            % decompose one more scale...
            pdwt2d_untiled_rowcol(w_LHLL, size(w_LHLL,0..2), wcoeff2[m], wcoeff2[n], 1)
            pdwt2d_untiled_rowcol(w_HLLL, size(w_HLLL,0..2), wcoeff2[m], wcoeff2[n], 1)
            % set the LL coefficients to zero (effectively creating a bandpass filter)
            w_LHLL[0..M/4-1,0..N/4-1] = 0
            w_HLLL[0..M/4-1,0..N/4-1] = 0
            % reconstruct this scale
            ipdwt2d_untiled_rowcol(w_LHLL, size(w_LHLL,0..2)/2, wcoeff2[m], wcoeff2[n], 1)
            ipdwt2d_untiled_rowcol(w_HLLL, size(w_HLLL,0..2)/2, wcoeff2[m], wcoeff2[n], 1)

            w_LHLL_sum = w_LHLL_sum + abs(w_LHLL)
            w_HLLL_sum = w_HLLL_sum + abs(w_HLLL)
        end
    end

    % Denoising of the chrominance-free luminance bands (all bands except HHLL, LHLL, HLLL, LLLL)
    if min(sigma_n) > 0.1
        dtcwt2d_complexphasemod(w1, 2)
        for m=0..1
            for n=0..1
                for k=[1,3,4,5,6,7,9,11,12,13,14,15]
                    [pm, pn] = [floor(k/4), mod(k,4)]
                    rm = pm*(M/2) + (0..M/2-1)
                    rn = pn*(N/2) + (0..N/2-1)
                    band = w1[m,n][rm,rn]
                    % use the sigma from the Green channel, as this corresponds to luminance
                    denoise_band_vectorprobshrink(band, wnd, C_n, sigma_n[1]*1.2, T)
                    w1[m,n][rm,rn] = band
                end
            end
        end
        dtcwt2d_complexphasemod(w1, 2)
    endif

    % Get the required demosaicing signs (depend on the Bayer pattern)
    [signHHLL,signLHLLr,signHLLLr,signLHLLb,signHLLLb] = wavelet_demosaicing_signs(pattern)
    signLHLLg = signLHLLb
    signHLLLg = signHLLLb

    map = zeros(size(w_LHLL_sum))
    parallel_do(size(map),w_LHLL_sum,w_HLLL_sum,map,estimate_edge_direction)
    w_LHLL_sum = 0 % release memory used by w_LHLL_sum
    w_HLLL_sum = 0 % release memory used by w_HLLL_sum

    %imshow(map,[0,1])
    
    %===== reconstruct the luminance values =====
    % Demultiplex the various luminance bands (except HHLL, LHLL, HLLL and LLLL, which also
    % contain chroma information). Note that the luminance bands have been denoised in the previous step
    for m=0..1
        for n=0..1
            y1[m,n] = zeros(size(x,0),size(x,1),3)

            parallel_do(size(w1[m,n],0..1)/4,w1[m,n],y1[m,n], _
                signHHLL,signLHLLb,signHLLLb,signLHLLg,signHLLLg,signLHLLr,signHLLLr,cplxwavelet_demux_bands)
        end
    end

    %===== reconstruct the chrominance values =====
    % 1 - decompose the HHLL, LHLL, HLLL and LLLL subbands into additional scales...
    for m=0..1
        for n=0..1
            rm = 0..M/2-1
            rn = 0..N/2-1
            w_LLLL = w1[m,n][rm,rn]
            w_LHLL = w1[m,n][rm,N+rn]
            w_HLLL = w1[m,n][M+rm,rn]
            w_HHLL = w1[m,n][M+rm,N+rn]
            dwt2d_untiled_rowcol(w_LLLL,[M/2,N/2,1], wcoeff2[m], wcoeff2[n], J-2)
            dwt2d_untiled_rowcol(w_LHLL,[M/2,N/2,1], wcoeff2[m], wcoeff2[n], J-2)
            dwt2d_untiled_rowcol(w_HLLL,[M/2,N/2,1], wcoeff2[m], wcoeff2[n], J-2)
            dwt2d_untiled_rowcol(w_HHLL,[M/2,N/2,1], wcoeff2[m], wcoeff2[n], J-2)
            w1[m,n][rm,rn] = w_LLLL
            w1[m,n][rm,N+rn] = w_LHLL
            w1[m,n][M+rm,rn] = w_HLLL
            w1[m,n][M+rm,N+rn] = w_HHLL
        end
    end

    %
    % 2 - compute the matrices A_m - these are the demosaicing rules for the
    %                                LLLL/LHLL/HLLL,HHLL luminance coefficients
    % (m=0 : vertical edge, m=1 : horizontal edge, m=1/2 : not sure)
    %
    A_def = m -> kron([[1,1],[1,-1]], [[1,2*(1-m)*signLHLLb,2*m*signHLLLb,signHHLL], _
                                       [1,0,0,-signHHLL], _
                                       [1,2*(1-m)*signLHLLr,2*m*signHLLLr,signHHLL]])
    A=zeros(6,8,3)
    A[:,:,0] = A_def(0)
    A[:,:,1] = A_def(0.5)
    A[:,:,2] = A_def(1)

    %
    % 3 - denoise every subband of LLLL
    % A more sophisticated denoising technique is used here since
    % 1/ the noise components are correlated due to the demosaicing rule
    % 2/ the demosaicing rule depends on the local edge orientation -
    %    Our algorithm also takes this into account.
    %
    for j=2..J-1
        for m=0..1
            Mj = M/(2^j)
            Nj = N/(2^j)
            rm = 0..Mj-1
            rn = 0..Nj-1

            for k=0..2
                offset={[0,Nj],[Mj,0],[Mj,Nj]}[k]
                band_in = zeros(Mj, Nj, 8)
                band_out = zeros(Mj, Nj, 6) % complex RGB data
                parallel_do([Mj,Nj],band_in,w1[int(pair[m,0]/2),mod(pair[m,0],2)],w1[int(pair[m,1]/2),mod(pair[m,1],2)],offset,extract_bands)

                if min(sigma_n)>0.1
                    denoisaic_band_vectorprobshrink(band_in, band_out, sigma_n, T2, A, map,2^(j-1))
                    %denoisaic_band_blsgsm(band_in, band_out, sigma_n, A, map,2^(j-1))
                else % noise level is too low - it does not make sense to do denoising
                    parallel_do([Mj,Nj,6],band_in,band_out,A,map,2^(j-1),reconstruct_ll)
                endif

                % output = [RGB_LLLL_real RGB_LLLL_imag]
                parallel_do([Mj,Nj],band_out,y1[int(pair[m,0]/2),mod(pair[m,0],2)],y1[int(pair[m,1]/2),mod(pair[m,1],2)],offset,restore_bands)
            end
        end
    end

    % 4 - copy the lowpass part
    for m=0..1
        Mj = M/(2^(J-1))
        Nj = N/(2^(J-1))
        band_in = zeros(Mj, Nj, 8)
        band_out = zeros(Mj, Nj, 6)
        parallel_do([Mj,Nj],band_in,w1[int(pair[m,0]/2),mod(pair[m,0],2)],w1[int(pair[m,1]/2),mod(pair[m,1],2)],[0,0],extract_bands)
        parallel_do([Mj,Nj,6],band_in,band_out,A,map,2^(J-2),reconstruct_ll)
        parallel_do([Mj,Nj],band_out,y1[int(pair[m,0]/2),mod(pair[m,0],2)],y1[int(pair[m,1]/2),mod(pair[m,1],2)],[0,0],restore_bands)
    end

    %imshow(y1[0,0][0..M/2-1,0..N/2-1,:],[]); title("y1[0,0] - Lowpass(1)")

    % 5 - reconstruct LLLL, LHLL and HLLL subbands...
    for m=0..1
        for n=0..1
            rm = 0..M/2-1
            rn = 0..N/2-1

            % 1) LLLL
            sz = [M/2^(J-1),N/2^(J-1),3]
            w_LLLL = y1[m,n][rm,rn,:]
            idwt2d_untiled_rowcol(w_LLLL, sz, wcoeff2[m], wcoeff2[n], J-2)
            y1[m,n][rm,rn,:] = w_LLLL
            w_LLLL = 0 % Release memory

            % 2) LHLL
            sz = [M/2^(J-1),N/2^(J-1),1]
            w_LHLL = w1[m,n][rm,N+rn]
            idwt2d_untiled_rowcol(w_LHLL, sz, wcoeff2[m], wcoeff2[n], J-2)
            w1[m,n][rm,N+rn] = w_LHLL
            w_LHLL = 0 

            % 3) HLLL
            w_HLLL = w1[m,n][M+rm,rn]
            idwt2d_untiled_rowcol(w_HLLL, sz, wcoeff2[m], wcoeff2[n], J-2)
            w1[m,n][M+rm,rn] = w_HLLL
        end
    end

    % 6 - reconstruction of the high frequencies LHLL and HLLL
    %     this requires a new demosaicing rule...
    A_def = m -> kron([[1,1],[1,-1]], [[m*signLHLLg, -m*signHLLLg],[-(1-m)*signLHLLg,(1-m)*signHLLLg]])
    A=zeros(4,4,3)
    A[:,:,0] = A_def(0)
    A[:,:,1] = A_def(0.5)
    A[:,:,2] = A_def(1)    

    for m=0..1
        [Mj,Nj] = [M,N]/2
        band_in4 = zeros(Mj, Nj, 4)
        band_out4 = zeros(Mj, Nj, 4)
        parallel_do([Mj,Nj],band_in4,w1[int(pair[m,0]/2),mod(pair[m,0],2)],w1[int(pair[m,1]/2),mod(pair[m,1],2)],extract_LHLL_HLLL_bands)
        denoisaic_LHLL_HLLL_band_vectorprobshrink(band_in4, band_out4, sigma_n, T2, A, map)
        parallel_do([Mj,Nj],band_out4,y1[int(pair[m,0]/2),mod(pair[m,0],2)],y1[int(pair[m,1]/2),mod(pair[m,1],2)],restore_LHLL_HLLL_bands)
    end

    % 7 - note that the HHLL subbands are 0 (chroma is here)

    %imshow(w1[0,0],[]); title("w1[0,0] - Luminance")
    %imshow(y1[0,0][0..M/2-1,0..N/2-1,:],[]); title("y1[0,0] - Lowpass")
    %imshow(y1[0,0],[]); title("y1[0,0]")

    w1 = 0 % release memory used by w1

    y = pidtcwt(y1, wcoeff1, wcoeff2, 2, false)
end

function [] = test_crosscorrelation()
    wcoeff1 = complex_orthogonal_wavelet(filtercoeff_farras)
    wcoeff2 = complex_orthogonal_wavelet(filtercoeff_selcw[4,3])

    x = fft2(randn(128,128))
    x = irealfft2(x ./ abs(x))*128
    J = 2
    w = pdtcwt(x, wcoeff1, wcoeff2, J, false, false)
    dtcwt2d_complexphasemod(w, J)
    [M,N] = size(x,0..1)
    pair = [[0,3],[2,1]]
    for m=0..1
        for j=0..0
            Mj = M/(2^J)
            Nj = N/(2^J)
            rm = 0..Mj-1
            rn = 0..Nj-1

            for k=0..2
                offset={[0,Nj],[Mj,0],[Mj,Nj]}[k]
                band1 = w[pair[m,0]][offset[0]+rm,offset[1]+rn]
                band2 = w[pair[m,1]][offset[0]+rm,offset[1]+rn]
                n = numel(band1)
                C = zeros(2,2)
                C[0,0] = sum(band1.^2)/n
                C[0,1] = C[1,0] = sum(band1.*band2)/n
                C[1,1] = sum(band2.^2)/n
                printf("m=%d j=%d k=%d C=%f", m,j,k,C)
            end
        end
    end       
end

function [] = main()

    %pattern = [[0,1],[3,2]]
    %pattern = [[2,3],[1,0]]
    %pattern = [[1,0],[2,3]]
    %pattern = [[1,0],[2,3]]
    pattern = [[2,1],[3,0]]

    psnr = (x, y) -> 10*log10(255^2./mean((x-y).^2))
    sigma = 10
    im = imread("lighthouse.png")

%    #pragma expression_optimizer(off)    
%    [x,y] = meshgrid(0..size(im,1)-1, 0..size(im,0)-1)
%    blob = exp(-(0.001 *((x-60).^2+(y-45).^2).^4))
%    im[:,:,0] = im[:,:,0] + 128*blob
%    im[:,:,2] = im[:,:,2] + 128*blob
%    
%    im[101,61,[0,2]] = 255
    %imshow(im,[])
    %return
    
    
    %im = imread("F:\Temp\left0000.bmp")
    %im = im[0..1071,:,:]

    %im = randn(1072,1920,3)
    %im = imread("lena_big.tif")
    %im = imread("c:\temp\data\kodim04.png")
    %im = imread("c:\bart\work\kodim04.png")
    im_noisy = im  + sigma * randn(size(im))
    im_bayer = bayer_subsampling(im_noisy,pattern)

    im_mosaic = bayer_mosaicimage(im_bayer,pattern)   
    
    w1 = complex_orthogonal_wavelet(filtercoeff_farras)
    w2 = complex_orthogonal_wavelet(filtercoeff_dualfilt)

    tic()
    im_recon = cplxwavelet_demosaicing(im_mosaic, pattern, w1, w2)
    toc()
    
    imshow(im_recon);title(sprintf("Demosaiced: %f dB",psnr(im,im_recon)))
    tic()    
    im_recon = cplxwavelet_denoisaicing(im_mosaic, pattern, w1, w2,4,sigma*[1,1/2,1,1/2])
    toc()
    %imshow(im)
    imshow(im_recon);title(sprintf("Denoisaiced: %f dB",psnr(im,im_recon)))
    %imshow(im_recon-im,[])
    print "psnr=", psnr(im,im_recon), "dB"
end