{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Advanced sort"}
{!title name="A single-kernel generic sorting algorithm"}

% Section: Overview
%
% Demonstrates the use of grid groups in combination with grid synchronization.
% Grid synchronization is "emulated", therefore it does not require a TCC-mode
% driver. In addition, it works with CUDA versions older than v9.0 (although
% this is of not much relevance currently).
% The algorithm implements a single kernel bitonic sorting algorithm, using 
% shared memory as much as possible.
% 
import "system.q"

function [] = fast_sort[T](x : mat[T])

    function [] = __kernel__ bitsort(s : mat[T]'unchecked, x : mat[T]'unchecked, _
        log2_n : int, pos : ivec2)
        
        % Load data in shared memory with the specified stride
        B : int'const = 512
        s_sh : vec[T]'unchecked = shared[T](B)
     
        % Due to the bitonic sort, this algorithm requires that the number of blocks
        % is a power of two. This can be achieved by setting the following code attribute:
        {!kernel target="gpu"; num_blocks_pow_of_2=true}          
        {!begin_grid_group}
        
        % Read input
        if pos[1] < size(x,1)
            s[pos] = x[pos]
        else
            s[pos] = maxvalue(T)
        endif

        syncthreads grid
        

        for k=1..log2_n
            for log2j_base = k-1..-9..0  % -9 = log2_cols = log2(B)-1
                log2j_end = max(0, log2j_base-8)
                
                {!begin_grid_group}              
                % Rotate bits
                src_index = shl(and(pos[1],shl(1,log2_n-log2j_end)-1),log2j_end) + _
                            shr(pos[1],log2_n-log2j_end)
                i = and(pos[1], B-1)
                b = and(src_index, shl(1, k))>0

                s_sh[i] = s[pos[0], src_index]
                syncthreads block
                             
                for log2j=log2j_base-log2j_end..-1..0
                    ixj = xor(i, shl(1,log2j))
                    if i < ixj 
                        v = (ixj - i) * b + i
                        w = (i - ixj) * b + ixj

                        if s_sh[v] > s_sh[w]
                            tmp = s_sh[v]; s_sh[v] = s_sh[w]; s_sh[w] = tmp
                        endif
                    endif
                    syncthreads block
                endfor
                
                s[pos[0], src_index] = s_sh[i]

                syncthreads grid
                
            endfor
        endfor
        
        syncthreads grid
        
        % Write output
        if pos[1] < size(x,1)
            x[pos] = s[pos]
        endif
        
    endfunction
    
    n = max(512,nextpow2(size(x,1)))  % bitonic sort only handles powers of two...
    s = uninit[T](size(x,0), n)
      
    parallel_do([[size(x,0),n,1],[1,512,1]],s,x,round(log2(n)),bitsort)
endfunction

function [ind, y] = fast_sort_ind[T](x : mat[T])

    B : int'const = 512
    log2_B : int'const = 9
    {!function_transform enable="kernel_generator"; threshold = 0}
        
    function [] = __kernel__ bitsort(s : mat[T]'unchecked, t : mat'unchecked, x : mat[T]'unchecked, 
                                     y : mat[T]'unchecked, 
        log2_n : int, pos : ivec2)
        % parallel bitonic sort taking maximal use of shared memory
        
        % Load data in shared memory
        s_sh : vec[T]'unchecked = shared[T](B)
        t_sh : vec'unchecked = shared(B)
               
        {!kernel target="gpu"; num_blocks_pow_of_2=true}          
        {!begin_grid_group}
        
        % Read input
        if pos[1] < size(x,1)
            s[pos] = x[pos]
            t[pos] = pos[1]
        else
            s[pos] = maxvalue(T)
        endif

        syncthreads grid

        for k=1..log2_n
            for log2j_base = k-1..-log2_B..0  % -9 = log2_cols = log2(B)-1
                log2j_end = max(0, log2j_base-8)
                
                {!begin_grid_group}              
                % Rotate bits
                src_index = shl(and(pos[1],shl(1,log2_n-log2j_end)-1),log2j_end) + _
                            shr(pos[1],log2_n-log2j_end)
                i = and(pos[1], B-1)
                b = and(src_index, shl(1, k))>0

                s_sh[i] = s[pos[0], src_index]
                t_sh[i] = t[pos[0], src_index]
                syncthreads block
                             
                for log2j=log2j_base-log2j_end..-1..0
                    ixj = xor(i, shl(1,log2j))
                    if i < ixj 
                        v = (ixj - i) * b + i
                        w = (i - ixj) * b + ixj

                        if s_sh[v] > s_sh[w]
                            tmp = s_sh[v]; s_sh[v] = s_sh[w]; s_sh[w] = tmp
                            tmp2 = t_sh[v]; t_sh[v] = t_sh[w]; t_sh[w] = tmp2
                        endif
                    endif
                    syncthreads block
                endfor
                
                s[pos[0], src_index] = s_sh[i]
                t[pos[0], src_index] = t_sh[i]
                
                syncthreads grid
            endfor
        endfor
        
        syncthreads grid
        % Write output
        if pos[1] < size(x,1)
            y[pos] = s[pos]
        endif
        
    endfunction
    
    n = max(B,nextpow2(size(x,1)))  % bitonic sort only handles powers of two...
    s = uninit[T](size(x,0), n)
    t = uninit(size(x,0), n)
    y = uninit[T](size(x))
    parallel_do([[size(x,0),n,1],[1,B,1]],s,t,x,y,round(log2(n)),bitsort)
    ind = t[:,0..size(x,1)-1]
endfunction

function [] = main()


    vals = rand(1,981574)
    runs = 100

    tic()
    {!interpreted for}
    for k=0..runs-1
    vals_sorted = copy(vals)
    fast_sort(vals_sorted)
    endfor
    t2=toc()
    
    printf("sorted %d keys/second\n", numel(vals)/(t2*1e-3)*runs)

    tic()
    {!interpreted for}
    for k=0..runs-1
    [_, vals_sorted] = fast_sort_ind(vals)
    endfor
    t3=toc()

    plot(float(vals_sorted[0,:]))

endfunction