{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Grid synchronization"}
{!title name="Demonstration of a grid synchronization"}

% Section: Overview
% 
% In the past, kernel code with out-of-block dependencies had to be split into multiple kernel 
% launches, due to the lack of inter-block communication and synchronization methods. However, 
% there is an overhead involved with kernel launches, both due to the CUDA driver/OS kernel as
% well as the Quasar runtime system. Since CUDA 9, cooperative kernel launches are available,
% but only on GPUs in Tesla Compute Cluster (TCC) mode. On Geforce GTX/RTX GPUs, if we would
% query for the device capability for cooperative kernel launches, we obtain 0 (hence the device
% cannot take advantage of grid synchronization). As alternative, Quasar provides an emulation
% of grid synchronization via atomic CAS instructions and via special runtime system support.
% It suffices to place "syncthreads(grid)" as a barrier in the kernel code. 

function y = gaussian_filter_separable(x : cube, fc : vec, n : int)
    function[] = __kernel__ gaussian_filter_hor(x: cube'mirror, y : cube'unchecked'nocopy, fc: vec'unchecked'hwconst, n : int, pos : ivec3)
        sum = 0.
        for i = 0..numel(fc) - 1
            sum = sum + x[pos + [0, i - n, 0]] * fc[i]
        endfor
        y[pos] = sum  
    endfunction

    function[] = __kernel__ gaussian_filter_ver(x: cube'mirror, y : cube'unchecked'nocopy, fc: vec'unchecked'hwconst, n : int, pos : ivec3)
        sum = 0.
        for i = 0..numel(fc) - 1
            sum = sum + x[pos + [i - n, 0, 0]] * fc[i]
        endfor
        y[pos] = sum          
    endfunction

    y = uninit(size(x))
    tmp = uninit(size(x))
    parallel_do([size(y,0..2),[8,32,1]], x, tmp, fc, n, gaussian_filter_hor)
    parallel_do([size(y,0..2),[8,32,1]], tmp, y, fc, n, gaussian_filter_ver)
endfunction


function y = gaussian_filter_fused2(x : cube, fc : vec, n : int)
    function[] = __kernel__ kernel(x: cube'mirror, y : cube'unchecked'nocopy, tmp : cube'mirror, fc: vec'unchecked'hwconst, n : int, pos : ivec3)
        {!kernel target="gpu"}
        {!cuda_launch_bounds max_threads_per_block=512}
        {!kernel_transform enable="localwindow"}
        assert(numel(fc)<=30)
        sum = 0.
        for i = 0..numel(fc) - 1
            sum = sum + x[pos + [0, i - n, 0]] * fc[i]
        endfor
        tmp[pos] = sum
        syncthreads(grid)  % === grid synchronization ===
        sum = 0.
        for i = 0..numel(fc) - 1
            sum = sum + tmp[pos + [i - n, 0, 0]] * fc[i]
        endfor
        y[pos] = sum
    endfunction

    y = uninit(size(x))
    tmp = uninit(size(x))
    parallel_do(size(y), x, y, tmp, fc, n, kernel)
endfunction


function [] = main()
    factor = 7
    sigma = 25

    % computation of the filter coefficients
    n = 2 * factor
    fc = exp(-(-n..n-1).^ 2 / sigma)
    fc /= sum(fc)
    
    im = imread("lighthouse.png","rgb")
    runs = 1000
    tic()
    for run=0..runs-1
        im_out = gaussian_filter_separable(im, fc, n)
    endfor
    toc("unfused kernel")
    
    tic()
    for run=0..runs-1
        im_out3 = gaussian_filter_fused2(im, fc, n)
    endfor
    toc("fused kernel (syncthreads(grid))")
    
    print "MSE: ", mean(abs(im_out - im_out3))
endfunction