{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Half precision float"}
{!title name="Half precision floating point filters"}

% Section: Overview
%
% Demonstrates how to develop a filter with input and output 
% in half-precision (16-bit) floating point format. The advantage
% of this representation is that it reduces the bandwidth by a 
% factor of 2x compared to the 32-bit floating point format. This
% might result in a speed-up, but only when the floating point 
% numbers are transferred in pairs (hvec2 data type) or larger
% evenly sized vectors. This can be achieved by using the 
% vectorization features of Quasar.
%
type half   : scalar'half
type hvec   : vec[half]
type hvec2   : vec[half](2)
type hmat   : mat[half]
type hcube  : cube[half]
type hcubeN : cube{:}[half]

% Configures the stride of the filter.
stride : int'const = 1

function [im_out : hmat] = im_to_half(im : mat)

    im_out = hmat(size(im))
    for m=0..size(im_out,0)-1
        for n=0..size(im_out,1)-1
            {!kernel target="gpu"}
            {!kernel_transform enable="simdprocessing"; target="gpu"; simd_width=2} 
            im_out[m,n] = im[m,n]
        endfor
    endfor

endfunction


im : mat'hwtex_const'unchecked = repmat(imread("lena_big.tif")[:,:,1],[4,4])

% Convert the image to half precision floating point format
im_half : hmat'hwtex_const'unchecked = im_to_half(im)

% Allocate output images
im_out : 'unchecked = hmat(size(im))
im_out1 : 'unchecked = mat(size(im))
im_out2 : 'unchecked = hmat(size(im))

runs = 1000
calc_gflops = time_msec -> 8*numel(im) * runs / (1e-3*time_msec)*1e-9

K : int'const = 8

%
% 1. Reference 32-bit floating point filter
% Note: !interpreted for ensures a kernel is only generated for the inner loop.
% 
tic()
{!interpreted for}
for k=1..runs
    {!parallel for}
    for m=0..size(im,0)-1
        for n=0..size(im,1)-1
            r = 0.0
            for x=0..K-1          
                r += im[m,n+stride*x] 
            endfor     
            im_out1[m,n] = r/(2*K+1)
        endfor
    endfor
endfor
toc("32-bit floating point")


%
% 2. Automatic SIMD processing
% 
tic()
{!interpreted for}
for k=1..runs
    {!parallel for}
    for m=0..size(im,0)-1
        for n=0..size(im,1)-1
            {!kernel_transform enable="simdprocessing"; target="gpu"; simd_width=2} 
            r = 0.0
            for x=0..K-1          
                r += im_half[m,n+stride*x] 
            endfor     
            im_out[m,n] = r/(2*K+1)            
        endfor
    endfor
endfor
toc("16-bit floating point (default)")


%
% 3. Manual SIMD processing
% 
tic()
{!interpreted for}
for k=0..runs
    {!parallel for}
    for m=0..size(im,0)-1
        for n=0..2..size(im,1)-1
            % By default, Quasar will upcast the 16-bit pair to 32-bit. To prevent this,
            % the following code attribute can be used. Note that this may lead to
            % clipping problems, so it can only be used when it is sure that the
            % result stays within range.
            {!simd_noupcast enable=im_half}
            r2 = hvec(2)
            for x=0..7            
                r2 = r2 + im_half[m,(n+stride*x)+(0..1)] 
            endfor
            im_out[m,n+(0..1)] = r2/(2*K+1)
        endfor
    endfor
endfor
syncthreads(host)
time=toc("16-bit floating point SIMD")
print calc_gflops(time), " GFLOPS/s"

% Display the result
imshow(im_out,[0,255])
