{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Hardware texturing demo"}
{!title name="Shows how to use the GPU Hardware texturing from Quasar"}

% Section: Overview
%
% Illustrates the use of the CUDA hardware texture unit (HTI) in Quasar. It is as simple as
% annotating the function parameter types with:
% - hwtex_nearest (nearest neighbor interpolation, 2D/3D grayscale)
% - hwtex_linear (linear interpolation, 2D/3D grayscale)
% - hwtex_nearest(4) (nearest neighbor interpolation, 2D/3D RGBA)
% - hwtex_linear(4) (nearest neighbor interpolation, 2D/3D RGBA)
% There is one special case, it is possible to map an RGB image onto the HTU
% 
% Notes:
% - hwconst indicates the GPU hardware constant memory.

import "inttypes.q"
import "imhist.q"

% 1/  2D interpolation of scalar matrices

% Kernel function, not using hardware textures
function [] = __kernel__ interpolate_nonhwtex (y : mat, x : mat, scale : scalar, pos : ivec2)
    
    scaled_pos = scale * pos
    f = frac(scaled_pos)
    i = int(floor(scaled_pos))
    
    y[pos] = (1 - f[0]) * (1 - f[1]) * x[i[0], i[1]] +
                  f[0]  * (1 - f[1]) * x[i[0]+1, i[1]] +
             (1 - f[0]) *      f[1]  * x[i[0], i[1]+1] +
                  f[0]  *      f[1]  * x[i[0]+1, i[1]+1]        
    
endfunction

% Kernel function, using hardware textures
function [] = __kernel__ interpolate_hwtex (y : mat, x : mat'hwtex_linear, scale : scalar, pos : ivec2)        
    y[pos] = x[scale * pos]
endfunction

% 2/  3D interpolation of scalar matrices

function [] = __kernel__ interpolate_nonhwtex3d (y : cube, x : cube, scale : vec3, pos : ivec3)
    
    scaled_pos = scale .* pos
    f = frac(scaled_pos)
    i = int(floor(scaled_pos))
        
    y[pos] = (1 - f[0]) * (1 - f[1]) * (1 - f[2]) * x[i] +
                  f[0]  * (1 - f[1]) * (1 - f[2]) * x[i+[1,0,0]] +
             (1 - f[0]) *      f[1]  * (1 - f[2]) * x[i+[0,1,0]] +
                  f[0]  *      f[1]  * (1 - f[2]) * x[i+[1,1,0]] +
             (1 - f[0]) * (1 - f[1]) *      f[2]  * x[i+[0,0,1]] +
                  f[0]  * (1 - f[1]) *      f[2]  * x[i+[1,0,1]] +
             (1 - f[0]) *      f[1]  *      f[2]  * x[i+[0,1,1]] +
                  f[0]  *      f[1]  *      f[2]  * x[i+[1,1,1]]
    
endfunction

% Kernel function, using hardware textures
function [] = __kernel__ interpolate_hwtex3d (y : cube, x : cube'hwtex_linear, scale : vec3, pos : ivec3)
    y[pos] = x[scale .* pos]
endfunction

% 3/ 2D interpolation of uint8 matrices

% Kernel function, not using hardware textures
function [] = __kernel__ interpolate_nonhwtex_uint8 (y : mat[uint8], x : mat[uint8], scale : scalar, pos : ivec2)
    
    scaled_pos = scale * pos
    f = frac(scaled_pos)
    i = int(floor(scaled_pos))
    
    y[pos] = (1 - f[0]) * (1 - f[1]) * x[i[0], i[1]] +
                  f[0]  * (1 - f[1]) * x[i[0]+1, i[1]] +
             (1 - f[0]) *      f[1]  * x[i[0], i[1]+1] +
                  f[0]  *      f[1]  * x[i[0]+1, i[1]+1]        
    
endfunction

% Kernel function, using hardware textures
function [] = __kernel__ interpolate_hwtex_uint8 (y : mat, x : mat[uint8]'hwtex_nearest, scale : scalar, pos : ivec2)        
    y[pos] = float(x[scale * pos])
endfunction

% Demo for single channel images (2D or 3D grayscale)
function [] = single_channel_demo()

    im = imread("lena_big.tif")
    x = im[:,:,1]
    y = zeros(size(x))
    scale = 1 %0.33
    runs = 1000
    
    tic()
    for i=0..runs-1
        parallel_do(size(x),y,x,scale,interpolate_nonhwtex)
    endfor
    toc("2D Interpolation without hardware texturing")

    y = zeros(size(x))
    tic()
    for i=0..runs-1
        parallel_do(size(x),y,x,scale,interpolate_hwtex)
    endfor
    toc("2D Interpolation with hardware texturing")
    imshow(y)
    title("2D Interpolation with hardware texturing")

    x = im
    y = zeros(size(x))

    tic()
    for i=0..runs-1
        parallel_do(size(x),y,x,[scale, scale, 1.0],interpolate_nonhwtex3d)
    endfor
    toc("3D Interpolation without hardware texturing")
   
    x = im
    y = zeros(size(x))
    tic()
    for i=0..runs-1
        parallel_do(size(x),y,x,[scale, scale, 1.0],interpolate_hwtex3d)
    endfor
    toc("3D Interpolation with hardware texturing")
    
    
    imshow(y)
    title("3D Interpolation with hardware texturing")
    
endfunction


sym8_lo = [-0.003382415951,-0.000542132332,0.031695087811, 0.007607487325,-0.143294238351,-0.061273359068,0.481359651258, 0.777185751701,0.364441894835,-0.051945838108,-0.027219029917,0.049137179674,0.003808752014,-0.014952258337,-0.000302920515,0.001889950333]
sym8_hi = [-0.003382415951, 0.000542132332,0.031695087811,-0.007607487325,-0.143294238351, 0.061273359068,0.481359651258,-0.777185751701,0.364441894835, 0.051945838108,-0.027219029917,-0.049137179674,0.003808752014,0.014952258337,-0.000302920515,-0.001889950333]
sym8 = [sym8_lo, sym8_hi]

% Direct implementation, using global memory
function [] = __kernel__ dwt_dim0_direct(x : cube, y : cube, wc : mat'hwconst, ctd : int, n : int, pos : ivec3)
    K = 16*n + ctd
    a = 0.0
    b = 0.0
    tilepos = int((2*pos[0])/n)
    j0 = tilepos*n
    for k=0..(16-1)
        j = j0+mod(2*pos[0]+k+K,n)
        u = x[j,pos[1],pos[2]]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[j0+mod(pos[0],int(n/2)),    pos[1],pos[2]]=a
    y[j0+int(n/2)+mod(pos[0],int(n/2)),pos[1],pos[2]]=b
endfunction

% vec3 implementation
function [] = __kernel__ dwt_dim0_vec3(x : cube, y : cube, wc : mat'hwconst, ctd : int, n : int, pos : ivec2)
    K = 16*n + ctd
    a = [0.0,0.0,0.0]
    b = [0.0,0.0,0.0]
    tilepos = int((2*pos[0])/n)
    j0 = tilepos*n
    for k=0..(16-1)
        j = j0+mod(2*pos[0]+k+K,n)
        u = x[j,pos[1],0..2]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[j0+mod(pos[0],int(n/2)),    pos[1],0..2]=a
    y[j0+int(n/2)+mod(pos[0],int(n/2)),pos[1],0..2]=b
endfunction

% hwtex implementation
function [] = __kernel__ dwt_dim0_hwtex(x : cube'hwtex_nearest, y : cube'hwtex_nearest'nocopy, wc : mat'hwconst, ctd : int, n : int, pos : ivec3)
    K = 16*n + ctd
    a = 0.0
    b = 0.0
    tilepos = int((2*pos[0])/n)
    j0 = tilepos*n
    for k=0..(16-1)
        j = j0+mod(2*pos[0]+k+K,n)
        u = x[j,pos[1],pos[2]]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[j0+mod(pos[0],int(n/2)),    pos[1],pos[2]]=a
    y[j0+int(n/2)+mod(pos[0],int(n/2)),pos[1],pos[2]]=b
endfunction

function [] = __kernel__ dwt_dim0_hwtex_v2(x : cube'hwtex_nearest, y : cube'hwtex_nearest'nocopy, wc : mat'hwconst, ctd : int, n : int, pos : ivec3)
    K = 16*n + ctd
    a = 0.0
    b = 0.0
    tilepos = int((2*pos[1])/n)
    j0 = tilepos*n
    for k=0..(16-1)
        j = j0+mod(2*pos[1]+k+K,n)
        u = x[pos[0],j,pos[2]]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[pos[0],j0+mod(pos[1],int(n/2)),    pos[2]]=a
    y[pos[0],j0+int(n/2)+mod(pos[1],int(n/2)),pos[2]]=b
endfunction

% hwtex_nearest(4) implementation
function [] = __kernel__ dwt_dim0_hwtex4(x : cube'hwtex_nearest(4), y : cube'hwtex_nearest(4)'nocopy, wc : mat'hwconst, ctd : int, n : int, pos : ivec2)
    K = 16*n + ctd
    a = [0.0,0.0,0.0,0.0]
    b = [0.0,0.0,0.0,0.0]
    tilepos = int((2*pos[0])/n)
    x
    j0 = tilepos*n
    for k=0..(16-1)
        j = j0+mod(2*pos[0]+k+K,n)
        u = x[j,pos[1],0..3]
        a = a + wc[0,k] * u
        b = b + wc[1,k] * u
    endfor
    y[j0+mod(pos[0],int(n/2)),    pos[1],0..3]=a
    y[j0+int(n/2)+mod(pos[0],int(n/2)),pos[1],0..3]=b
endfunction

% Demo for multi-channel (RGB/RGBA images)
function [] = multi_channel_demo()

    function y = toRGBA(x)
        y = zeros(size(x,0),size(x,1),4)
        y[:,:,:] = x
    endfunction

    im = imread("lena_big.tif", "rgba")
    im_out = uninit(size(im))
    num_runs = 1000

    % Direct implementation, using global memory
    tic()
    for n=0..num_runs-1
        parallel_do([size(im_out,0)/2,size(im_out,1),3],im,im_out,sym8,4,size(im_out,0), dwt_dim0_direct)
    endfor
    toc("Direct implementation")
    print "Check sum: ",mean(im_out)

    % Direct implementation, using fixed length vectors
    im_out = rand(size(im))
    tic()
    for n=0..num_runs-1
        parallel_do([size(im_out,0)/2,size(im_out,1)],im,im_out,sym8,4,size(im_out,0), dwt_dim0_vec3)
    endfor
    toc("vec3 implementation")
    print "Check sum: ",mean(im_out)

    % hwtex_nearest implementation
    im_out = rand(size(im))
    tic()
    for n=0..num_runs-1
        parallel_do([size(im_out,0)/2,size(im_out,1),4],im,im_out,sym8,4,size(im_out,0), dwt_dim0_hwtex)
    endfor
    toc("hwtex_nearest implementation")
    print "Check sum: ",mean(im_out[:,:,0..2])
    imshow(im_out)


    % hwtex_nearest(4) implementation
    im2 = copy(im)
    im_out = rand(size(im))
    tic()
    for n=0..num_runs-1
        parallel_do([size(im_out,0)/2,size(im_out,1)],im2,im_out,sym8,4,size(im_out,0), dwt_dim0_hwtex4)
    endfor
    toc("hwtex_nearest(4) implementation")
    print "Check sum: ",mean(im_out)
    imshow(im_out)


endfunction

function [] = main()
    single_channel_demo()
    multi_channel_demo()
endfunction