{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Kernel fusion"}
{!title name="Demonstrates how kernel fusion can be used to speed up Matlab-style code"}

% Section: Overview
%
% In a vast majority of cases, matrix slicing code is best converted to a single kernel. By default,
% and to enable line-by-line debugging of code, the Quasar compiler will convert each line into a 
% kernel function launch. Due to the many device memory read/writes, this approach has a 
% tremendeous performance penalty compared to a single kernel in which the temporary data is stored
% in the registers. The kernel fusion is a simple tool that allows several kernels to be fused
% into a single kernel, thereby optimizing the memory accesses. This example demonstrates that
% Matlab-style code can be written but has the same performance as a manually written kernel 
% (in this example, written in the form of a for-loop).
% 
% Of course, one could always write single kernels from the first time, however, there are cases
% that either library functions are used or that fusing the kernels manually is simply too much
% work, or too error-prone.
% 
% The kernel fusion approach taken in Quasar goes further than simply "copy-pasting" the kernels
% together. Instead, it looks for opportunities to reuse temporary buffers, to statically reschedule
% kernel functions, to jointly optimize shared memory etc.
% 

function y = process_manual(x : cube)

    y = uninit(size(x))
    for [m,n] = 0..size(x,0..1)-1
        [xR,xG,xB] = x[m,n,0..2]
        [yR,yG,yB] = [xR-xG,xG,xB-xG]
        y[m,n,0..2] = [yR,yG,yB]
    endfor

endfunction

function y = process[kernel_fusion](x : cube)
    if kernel_fusion
        {!kernel_fusion}
    endif
    % High-level Quasar code, the goal is here to use the kernel fusion
    % transform to fuse all of these operations in one single kernel.
    % The resulting kernel function should contain:
    %
    % y[pos[0],pos[1],0]=(x[pos[0],pos[1],0]-x[pos[0],pos[1],1]) 
    % y[pos[0],pos[1],1]=x[pos[0],pos[1],1] 
    % y[pos[0],pos[1],2]=(x[pos[0],pos[1],2]-x[pos[0],pos[1],1]) 
    %
    % This offers a vast performance improvement compared to the single
    % kernels launched sequentially
    xR = x[:,:,0]
    xG = x[:,:,1]
    xB = x[:,:,2]

    yR = xR - xG
    yG = xG
    yB = xB - xG
    
    y = uninit(size(x))    
    y[:,:,0] = yR
    y[:,:,1] = yG
    y[:,:,2] = yB

endfunction

im = imread("lena_big.tif")

runs = 500
tic()
for k=1..runs
    process_manual(im)
endfor
toc("process_manual")

tic()
for k=1..runs
    process[false](im)
endfor
toc("process (no kernel fusion)")

tic()
for k=1..runs
    process[true](im)
endfor
toc("process (kernel fusion)")
