%RUN_DISABLED
{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Memory coalescing"}
{!title name="Memory coalescing demo"}

% Section: Overview
% 
% The purpose of this demonstration is to show the huge effect of memory coalescing on the
% performance of kernel functions. Memory coalescing is an efficient memory transfer mechanism
% from the device memory of the GPU to the registers, that requires all threads within a hwarp
% to access consecutive locations in memory. If the addresses satisfy this criterion,
% at runtime, the memory controller can issue a 128-byte transfer (which fills an entire cache line)
% at once. Note that for this purpose, the start address also needs to be aligned).
% In the non coalesced example below, the data is accessed in a column major ordering. In this
% case the memory coalescing conditions are not satisfied, causing the kernel to have a reduced
% performance (compared to kernel_coalesced).
% In Quasar, kernels with uncoalesced memory accesses can be corrected automatically, under some
% conditions (more specifically, a kernel dimension reordering must be possible to restore
% coalescing). This is achieved by adding the {!kernel_transform enable="memcoalescing"} to the
% kernel function body.
% 
% Results for a Geforce RTX 2080 GPU:
% No memory coalescing: 1317.036 ms
% Memory coalescing: 895.8915 ms

function [] = main()
    % Note: extend the input image, because today's GPUs are very fast
    im = repmat(imread("lena_big.tif")[:,:,1],[4,4,1])
    im_out = zeros(size(im))
    [M,N] = size(im,0..1)

    function [] = __kernel__ kernel_nocoalesced(im : mat'unchecked, im_out : mat'unchecked, pos : ivec2)
        %Uncomment to perform coalescing automatically.
        %{!kernel_transform enable="memcoalescing"}
        im_out[pos[1],pos[0]] = im[pos[1],pos[0]]
    endfunction

    function [] = __kernel__ kernel_coalesced(im : mat'unchecked, im_out : mat'unchecked, pos : ivec2)
        im_out[pos] = im[pos]
    endfunction

    runs = 10000

    tic()
    for k=0..runs-1
    parallel_do(size(im_out,[1,0]),im,im_out,kernel_nocoalesced)
    endfor
    toc("No memory coalescing")
    
    tic()
    for k=0..runs-1
    parallel_do(size(im_out,[0,1]),im,im_out,kernel_coalesced)
    endfor
    toc("Memory coalescing")
    
    imshow(im)
endfunction
