{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Multi-GPU processing"}
{!title name="Demonstrates how to perform multi-GPU processing with manual scheduling"}

% Section: Overview
%
% This particular sample requires using the Hyperion engine (v2) runtime devices, because the
% v1 runtime simply does not support multi-GPU. The program performs a sequence of forward and 
% adjoint parallel beam projections, in which the workload is divided among the available GPUs
% in the system. The example has been tested with up to 6 Geforce RTX 2080 Ti GPUs and is able
% to attain a utilization of around 90% on each GPU (this is due to the synchronization between 
% the GPUs). For performance reasons, it is best to connect each GPU pair with an NVLink bridge.
% For example, for four GPUs, two NVLink bridges are required.

% Computation of the peak-signal-to-noise ratio
psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))

num_gpus = gpu_count()

% Function: forward_parabeam_projection
%
% Computes a forward parallel beam projection (image domain -> sinogram domain)
%
% Parameters:
% x - input image 
% parageom - geometry settings
%    parageom.angles: projection angles in radians
%    parageom.offsets: detector offsets
%
% Notes:
% In this version, nearest neighbor interpolation is used (for performance reasons).
function y = forward_parabeam_projection(x, parageom)

    function [] = __kernel__ process(x : mat'unchecked, y : mat'unchecked, angles : vec, _
        offsets : vec, pos : ivec2)

        d = offsets[pos[0]] % offset along the detector array
        theta = angles[pos[1]]
        v = [cos(theta), sin(theta)]
        w = [-sin(theta), cos(theta)]
        ctr = size(x)/2 + d * w
        sum = 0.0

        lambda1 = ([size(x,0)-2.0, 0.0]-ctr[0])/v[0]
        lambda2 = ([size(x,1)-2.0, 0.0]-ctr[1])/v[1]
        lambda = [max(min(lambda1),min(lambda2)),min(max(lambda1),max(lambda2))]

        for i=lambda[0]..lambda[1]
            p = ctr + i * v            
            sum += x[p]
        end
        y[pos] = sum
    endfunction

    K = numel(parageom.angles)/num_gpus
    y = cell(num_gpus)
    for k=0..num_gpus-1
        {!sched gpu_index=k}
        y[k] = zeros(numel(parageom.offsets),K)
    endfor

    % Split the output data into tiles and individually generate each tile
    for k=0..num_gpus-1
        {!sched gpu_index=k}
        parallel_do(size(y[k]),x,y[k],parageom.angles[k*K..(k*K+K-1)],parageom.offsets,process)
    endfor
    
    {!sched gpu_index=0}

endfunction

% Function: adjoint_parabeam_projection
%
% Computes the adjoint of the parallel beam projection (sinogram domain -> image domain).
% The adjoint is not the same as the backward projection, however a reconstruction can be
% obtained using e.g. the conjugate gradient algorithm.
%
% Parameters:
% x - input sinogram
% parageom - geometry settings
% o  parageom.angles: projection angles in radians
% o  parageom.offsets: detector offsets
%
% Notes:
% In this version, nearest neighbor interpolation is used (for performance reasons).
function y = adjoint_parabeam_projection(x, parageom, sz)

    % note : x = sinogram space, y = image space
    function [] = __kernel__ process(x : mat'unchecked, y : mat'unchecked, angles : vec, _
        offsets : vec, pos : ivec2)

        d = offsets[pos[0]] % offset along the detector array
        theta = angles[pos[1]]
        v = [cos(theta), sin(theta)]
        w = [-sin(theta), cos(theta)]
        ctr = (size(y)/2) + d * w
        sum = x[pos]

        lambda1 = ([size(y,0)-2.0, 0.0]-ctr[0])/v[0]
        lambda2 = ([size(y,1)-2.0, 0.0]-ctr[1])/v[1]
        lambda = [max(min(lambda1),min(lambda2)),min(max(lambda1),max(lambda2))]
        
        % synchronization issue to solve
        for i=lambda[0]..lambda[1]
            p = int(ctr + i * v)
            y[p] += sum  % this is an atomic operation            
        endfor
    endfunction
    
    K = numel(parageom.angles)/num_gpus
    
    % Individually backproject each tile
    yp = cell(num_gpus)
    for k=0..num_gpus-1
        {!sched gpu_index=k}
        yp[k] = zeros(sz)
        parallel_do(size(x[k]),x[k],yp[k],parageom.angles[k*K..(k*K+K-1)],parageom.offsets,process)
    endfor
    
    % Accumulate the result
    {!sched gpu_index=0}
    for k=1..num_gpus-1
        yp[0] += yp[k]
    endfor
    y = yp[0]
  
endfunction


% main function - testing only...
function [] = main()
    im = imread("SheppLogan.png")
    im = im[:,:,0]

    parageom = object()
    {!alloc mode="gpu"}
    parageom.angles = linspace(0,2*pi,1024)
    {!alloc mode="gpu"}
    parageom.offsets = -511..512 % -255..256
    
    tic()
    for k=0..9999
    y = forward_parabeam_projection(im, parageom)
    z = adjoint_parabeam_projection(y, parageom, size(im))
    endfor
    toc()

    get_psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    ytot = zeros(size(y[0],0),2*size(y[0],1))
    ytot[:,0..size(y[0],1)-1] = y[0]
    ytot[:,size(y[0],1)..2*size(y[0],1)-1] = y[1]
    
    imshow(ytot,[]); title("Sinogram data")
    imshow(z,[]); title("backprojected data")

endfunction
