{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Shared memory designator demo"}
{!title name="Shows how to use shared memory designators from Quasar"}

% Because most of the kernel bottlenecks are caused by the extensive 
% high latency device memory reads/writes, Quasar offers a "convenience" 
% feature for more easily taking advantage of the shared memory. This is
% called a shared memory designator. The idea is that a local variable 
% type can be annotated with 'shared. Then the compiler will generate 
% code to distribute operations (such as assignments) over
% all threads within a block. The advantage is that the user no longer
% needs to design the thread mapping, taking memory coalescing conditions
% and SIMD advantages into account. This is done automatically.
%
% This sample shows how to implement a 2D separable linear filter with 
% a single kernel. The input as well as the intermediate result are 
% stored in shared memory.
% 
% Note that shared memory can be seen as a sort of "software" controlled
% memory, in the sense that it programmatically defines which data is 
% loaded into the L1 cache (which is for many GPU architectures unified
% with the shared memory).

function img_out = fast_filter(img_in : cube(:,:,3), fc : vec'hwconst)
    c : int'const = 3
    function [] = __kernel__ kernel(img_out : cube(:,:,3), pos : ivec2)
        
        p = pos - mod(pos, 16)
        s1 : 'shared = img_in[p[0]+(-c..16+c),p[1]+(-c..16+c),:]
        s2 : 'shared = uninit(16+2*c+1,16,3)
   
        % work shared along the threads
        for m=0..16+2*c  % <=== load a few extra rows
            for n=0..15
                total = zeros(3)
                for k=0..numel(fc)-1
                    total += fc[k] * s1[m,n+k,:]
                endfor
                
                s2[m,n,:] = total
            endfor
        endfor
        
        total1 = zeros(3)
        [m1,n1] = pos-p     % <==== subtract offset
        for k=0..numel(fc)-1
            total1 += fc[k] * s2[m1+k,n1,:]
        endfor
        img_out[pos[0],pos[1],:] = total1

    endfunction
    img_out = uninit(size(img_in))
    parallel_do([size(img_in,0..1),[16,16]],img_out, kernel)
endfunction

n = 3
fc = exp(-0.05*(-n..n).^2/4)
fc = fc / sum(fc)
N = 1000

img_in = imread("lena_big.tif")

tic()
for k=1..N
    img_out = fast_filter(img_in, fc)
endfor
toc("Implementation using shared memory")
imshow(img_out)
