{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/tensor core kernel programming demo"}
{!title name="Shows how to program the NVIDIA tensor cores from Quasar"}

% Section: Overview
%
% This example demonstrates the use of NVIDIA tensor cores for
% accelerating convolutions and matrix multiplications.
% In this sample, we show how horizontal and vertical convolutions
% can be implemented using tensor cores. Note that the tensor cores
% rely on mixed 32-bit/16-bit float precision, so the results are
% slightly less accurate.
%
% The program should print the following output:
% Horizontal FP32: 906.0566 ms
% Horizontal TC: 96.807 ms
% 4.00492858886719-254.898834228516
% Vertical FP32: 3093.8572 ms
% Vertical TC: 97.6296 ms
% error hor=0.0457806686560313 error ver=0.0457932402690252
% Horizontal/vertical TC: 272.8147 ms
% 184.490234580468 GFLOP
%
% Note: requires the Volta (or newer) GPU architecture

import "floattypes.q"

type half : scalar'half

function A = generate_toeplitz_blocks(filtercoeff : vec, N : int, B : int, c : int)

    A = zeros(B,(2*N+1)*B)

    for m=0..size(A,0)-1
        for n=0..size(A,1)-1
            A[m, n] = filtercoeff[n-N*B-m+c]
        endfor
    endfor

endfunction

function y = convolve_hor_nontc_f32[B : int, N : int](x : mat, A : mat)

    y = zeros(size(x))
    {!parallel for}
    for m=0..size(x,0)-1
        for n=0..B..size(x,1)-1

            % Matrix multiplication
            result = zeros(B)
            for k = -N..N
                result += x[m,n-k*B+(0..B-1)] * A[0..B-1, (N+k)*B+(0..B-1)]
            endfor

            y[m,n+(0..B-1)] = result

        endfor
    endfor

endfunction

function y = convolve_ver_nontc_f32[B : int, N : int](x : mat, A : mat)

    y = zeros(size(x))
    {!parallel for}
    for m=0..B..size(x,0)-1
        for n=0..size(x,1)-1

            % Matrix multiplication
            result = zeros(B)
            for k = -N..N
                result += A[0..B-1, (N+k)*B+(0..B-1)]*x[m+k*B+(0..B-1),n]
            endfor

            y[m+(0..B-1),n] = result

        endfor
    endfor

endfunction

function y = convolve_hor_tc_f16[B : int, N : int](x : mat[half], A : mat[half])
    warp_size : int'const = 32
    WMMA_M : int'const = B
    WMMA_N : int'const = B
    WMMA_K : int'const = B
    warps_per_block : int'const = 16

    function [] = __kernel__ kernel[B : int, N : int](x : mat[scalar'half], y : mat, A : mat[half], pos : ivec2, tc : tensor_core)
        {!kernel target="nvidia_cuda"}

        A_sh : 'shared = A[0..B-1, 0..(2*N+1)*B-1]
        A_sh2 : 'shared = uninit(B, B)
        A_sh2[:,:] = A[0..B-1, 0..B-1]

        a_frag = tc.alloc("matrix_a", WMMA_M, WMMA_N, WMMA_K, typename(scalar'half), "row_major")
        b_frag = tc.alloc("matrix_b", WMMA_M, WMMA_N, WMMA_K, typename(scalar'half), "row_major")
        acc_frag = tc.alloc("accumulator", WMMA_M, WMMA_N, WMMA_K, typename(scalar'single), "Default")

        {!begin_grid_group}
        m = pos[0]
        n = int(pos[1]/warp_size)

        acc_frag.set(0.0)

        for k=-N..N
            if n-k>=0 && B*(n-k+1)<=size(x,1)
                a_frag.load_matrix(x, m * B + (0..B-1), (n-k)*B + (0..B-1))
                b_frag.load_matrix(A_sh, (0..B-1), (N+k) * B + (0..B-1))

                acc_frag.multiply_accumulate_matrix(acc_frag, a_frag, b_frag)
            endif
        endfor

        acc_frag.store_matrix(y, m * B + (0..B-1), n * B + (0..B-1), "mem_row_major")
        {!end_grid_group}

    endfunction


    y = uninit(size(x))

    parallel_do([[int(size(x,0)/B), int(size(x,1)/B)*warp_size,1], _
                 [warps_per_block, warp_size, 1]],x,y,A,kernel[B,N])

endfunction

function y = convolve_ver_tc_f16[B : int, N : int](x : mat[half], A : mat[half])
    warp_size : int'const = 32
    WMMA_M : int'const = B
    WMMA_N : int'const = B
    WMMA_K : int'const = B
    warps_per_block : int'const = 16

    function [] = __kernel__ kernel[B : int, N : int](x : mat[scalar'half], y : mat, A : mat[half], pos : ivec2, tc : tensor_core)
        {!kernel target="nvidia_cuda"}

        A_sh : 'shared = A[0..B-1, 0..(2*N+1)*B-1]

        a_frag = tc.alloc("matrix_a", WMMA_M, WMMA_N, WMMA_K, typename(scalar'half), "row_major")
        b_frag = tc.alloc("matrix_b", WMMA_M, WMMA_N, WMMA_K, typename(scalar'half), "row_major")
        acc_frag = tc.alloc("accumulator", WMMA_M, WMMA_N, WMMA_K, typename(scalar'single), "Default")

        {!begin_grid_group}
        m = pos[0]
        n = int(pos[1]/warp_size)

        acc_frag.set(0.0)

        for k=-N..N
            if m+k>=0 && B*(m+k+1)<=size(x,0)
                a_frag.load_matrix(A_sh, (0..B-1), (N+k) * B + (0..B-1))
                b_frag.load_matrix(x, (m+k) * B + (0..B-1), n*B + (0..B-1))


                acc_frag.multiply_accumulate_matrix(acc_frag, a_frag, b_frag)
            endif
        endfor

        acc_frag.store_matrix(y, m * B + (0..B-1), n * B + (0..B-1), "mem_row_major")
        {!end_grid_group}

    endfunction


    y = uninit(size(x))

    parallel_do([[int(size(x,0)/B), int(size(x,1)/B)*warp_size,1], _
                 [warps_per_block, warp_size, 1]],x,y,A,kernel[B,N])

endfunction


function y = convolve_hv_tc_f16[B : int, N : int](x : mat[half], A : mat[half]) % seperable horizontal/vertical
    warp_size : int'const = 32
    WMMA_M : int'const = B
    WMMA_N : int'const = B
    WMMA_K : int'const = B
    warps_per_block : int'const = 8

    function [] = __kernel__ kernel[B : int, N : int](x : mat[scalar'half], y : mat, z : mat, A : mat[half], pos : ivec2, tc : tensor_core)
        {!kernel target="nvidia_cuda"}

        A_sh : 'shared = A[0..B-1, 0..(2*N+1)*B-1]
        Z_sh : 'shared = uninit_T(typename(scalar'half), B*(warps_per_block+2*N), B)

        a_frag = tc.alloc("matrix_a", WMMA_M, WMMA_N, WMMA_K, typename(scalar'half), "row_major")
        b_frag = tc.alloc("matrix_b", WMMA_M, WMMA_N, WMMA_K, typename(scalar'half), "row_major")
        acc_frag = tc.alloc("accumulator", WMMA_M, WMMA_N, WMMA_K, typename(scalar'single), "Default")

        {!begin_grid_group}
        m = pos[0]
        n = int(pos[1]/warp_size)

        acc_frag.set(0.0)

        for k=-N..N
            if n-k>=0 && B*(n-k+1)<=size(x,1)
                a_frag.load_matrix(x, m * B + (0..B-1), (n-k)*B + (0..B-1))
                b_frag.load_matrix(A_sh, (0..B-1), (N+k) * B + (0..B-1))

                acc_frag.multiply_accumulate_matrix(acc_frag, a_frag, b_frag)
            endif
        endfor

        acc_frag.store_matrix(z, m * B + (0..B-1), n * B + (0..B-1), "mem_row_major")
        {!end_grid_group}

        syncthreads(grid)

        {!begin_grid_group}
        m = pos[0]
        n = int(pos[1]/warp_size)
        mw = mod(m, warps_per_block) * B

        acc_frag.set(0.0)
        Z_sh[(0..(warps_per_block+2*N)*B-1),:] = z[(m-N) * B - mw + (0..(warps_per_block+2*N)*B-1), n*B + (0..B-1)]

        for k=-N..N
            if m+k>=0 && B*(m+k+1)<=size(x,0)

                a_frag.load_matrix(A_sh, (0..B-1), (N+k) * B + (0..B-1))
                b_frag.load_matrix(Z_sh, mw+(N+k)*B+(0..B-1), 0..B-1)

                acc_frag.multiply_accumulate_matrix(acc_frag, a_frag, b_frag)
            endif
        endfor

        acc_frag.store_matrix(y, m * B + (0..B-1), n * B + (0..B-1), "mem_row_major")
        {!end_grid_group}

    endfunction

    z = uninit(size(x))
    y = uninit(size(x))

    parallel_do([[int(size(x,0)/B), int(size(x,1)/B)*warp_size,1], _
                 [warps_per_block, warp_size, 1]],x,y,z,A,kernel[B,N])

endfunction

cuda_arch = gpu_feature("cuda_arch")
if cuda_arch < 700
    print "Tensor cores are not supported for compute capability ", cuda_arch
    return
endif

filtercoeff = exp(-0.25*(-3..3).^2)
filtercoeff = filtercoeff / sum(filtercoeff)
print filtercoeff

B : int'const = 16
N : int'const = 1
A = generate_toeplitz_blocks(filtercoeff, N, B, 3)
imshow(A)

x = repmat(imread("lighthouse.png")[:,:,1],[4,4])
flop = 2*numel(x)*16

runs = 1000
tic()
for run=1..runs
    y1 = convolve_hor_nontc_f32[B,N](x, A)
endfor
toc("Horizontal FP32")

xh = scalar_to_half(x)
Ah = scalar_to_half(A)

tic()
for run=1..runs
    y2 = convolve_hor_tc_f16[B,N](xh, Ah)
endfor
toc("Horizontal TC")

print min(y2), "-", max(y2)
tic()
for run=1..runs
    y3 = convolve_ver_nontc_f32[B,N](x, A)
endfor
toc("Vertical FP32")

xh = scalar_to_half(x)
Ah = scalar_to_half(A)

tic()
for run=1..runs
    y4 = convolve_ver_tc_f16[B,N](xh, Ah)
endfor
toc("Vertical TC")

imshow(y2,[0,255]),title("Horizontal TC filtering")
imshow(y4,[0,255]),title("Vertical TC filtering")

print "error hor=", mean(abs(y1-y2)), " error ver=", mean(abs(y3-y4))

tic()
for run=1..runs
    y5 = convolve_hv_tc_f16[B,N](xh, Ah)
endfor
time=toc("Horizontal/vertical TC")

print flop*runs/(time*1e-3)/1e9, " GFLOP"

imshow(y5,[0,255]),title("Separable TC filtering")
