{!author name="Bart Goossens"}
{!doc category="Samples/Advanced CUDA/Warp shuffling demo"}
{!title name="Shows how to use CUDA warp shuffling from Quasar"}

% Section: Overview
%
% Warp shuffling, introduced in the kepler architecture, are an efficient method for interthread 
% communication within a warp (similar to x64 SIMD shuffling instructions). Quasar's 
% implementation uses cooperative groups from CUDA 9 (which means that it doesn't work with older
% CUDA versions). To use warp shuffling, a special kernel function parameter coalesced_threads 
% needs to be defined, of type thread_block. This class provides several warp shuffling operations.
% In addition, it might sometimes be useful to perform synchronization across one warp. This is 
% achieved using
% 
% syncthreads(warp)
%
% In general, warp shuffling code is generated automatically by Quasar for e.g., parallel
% reductions. Therefore it is only useful to use warp shuffling in other algorithms in which
% interwarp communication is feasible.
% 

function y : scalar = __kernel__ reduce_sum(coalesced_threads : thread_block, x : vec'unchecked, pos : int, blkdim : int, blkcnt : int)
    {!kernel target="nvidia_cuda"}
    total = 0.0
    
    % Grid strided-loop
    for index=pos..blkdim*blkcnt..numel(x)-1    
        total += x[index]
    endfor
    
    total += coalesced_threads.shfl_down(total, 16)
    total += coalesced_threads.shfl_down(total, 8)
    total += coalesced_threads.shfl_down(total, 4)
    total += coalesced_threads.shfl_down(total, 2)
    total += coalesced_threads.shfl_down(total, 1)

    % Atomics for accumulation
    if coalesced_threads.thread_idx==0
        y += total
    endif
    
endfunction

% Combined shared memory/warp shuffle
function y : scalar = __kernel__ reduce_sum_shared(coalesced_threads : thread_block, x : vec'unchecked, blkpos : int, pos : int, blkdim : int, blkcnt : int)
    {!kernel target="nvidia_cuda"}
    total = 0.0
    
    % Grid strided loop
    for index=pos..blkdim*blkcnt..numel(x)-1    
        val = x[index]                
        total += val
    endfor
    
    total += coalesced_threads.shfl_down(total, 16)
    total += coalesced_threads.shfl_down(total, 8)
    total += coalesced_threads.shfl_down(total, 4)
    total += coalesced_threads.shfl_down(total, 2)
    total += coalesced_threads.shfl_down(total, 1)
    
    % Atomics for accumulation
    accum = shared(16)
    
    if coalesced_threads.thread_idx==0
        accum[int(blkpos/32)]=total
    endif
    
    syncthreads block
    
    if blkpos < 16
        total = accum[blkpos]
    endif
    
    if blkpos < 32
        syncthreads warp
        total += coalesced_threads.shfl_down(total, 8)
        total += coalesced_threads.shfl_down(total, 4)
        total += coalesced_threads.shfl_down(total, 2)
        total += coalesced_threads.shfl_down(total, 1)
    endif
    
    if blkpos==0
        y += total
    endif

endfunction

function y : scalar = __kernel__ auto_sum(x : vec'unchecked, pos : int)
    y += x[pos]
endfunction

x = rand(32*65536)

runs = 5000
tic()
{!interpreted for}
for k=1..runs
    %y = sum(x)
    y = parallel_do(numel(x),x,auto_sum)
endfor
syncthreads host
toc("Automatic parallel reduction sum")
print y

grid_size = 64
block_size = 512
grid_size = prod(opt_block_cnt(reduce_sum,block_size))

tic()
for k=1..runs    
    y = parallel_do([[block_size*grid_size], [block_size]],x,reduce_sum)
endfor
toc("Warp shuffling operations")
print y


tic()
for k=1..runs    
    y = parallel_do([[block_size*grid_size], [block_size]],x,reduce_sum_shared)
endfor
toc("Warp shuffling operations (hybrid shared memory)")
print y
