% Deblurring via automatic calculation of the adjoint operation
import "Quasar.CompMath.dll"

% Parallel-beam reconstruction using the conjugate gradient method, with
% automatic calculation of the adjoint.
% Sample uses Algorithmic Differentiation library. See documentation for the explanation.
{!title name="Conjugate gradient-based deblurring using automatic adjoint calculation"}
{!author name="Bart Goossens"}
{!doc category="Samples/Algorithmic Differentiation"}


function [y : mat] = blur_filter(x : mat, f : vec)
    %N = int(numel(f)/2)
    y = zeros(size(x))
    x1 : mat = zeros(size(y))

    % Horizontal filter
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            sum = 0.0
            for k=0..numel(f)-1
                sum += x[m,k+(n)] * f[k]
            endfor
            x1[m,n] = sum
        endfor
    endfor

    % Vertical filter
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            sum = 0.0
            for k=0..numel(f)-1
                sum += x1[k+(m),n] * f[k]
            endfor
            y[m,n] = sum
        endfor
    endfor
endfunction


function x = conjugate_gradients(A, y, max_iter=20, tol=1e-15)
    x = 0
    A_H = $unapply($adjoint(A(y), y), y)
    dotprod = (v, w) -> sum(sum(v .* w))
    r = A_H(y)
    p = r
    new_err = dotprod(r, r)

    for it=1..max_iter  
        err=new_err
        print "CG iteration ",it," err=",err
        A_p = A_H(A(p))
        alpha = err / dotprod(p, A_p)
        x = x + alpha * p
        r = r - alpha * A_p
        new_err = dotprod(r, r)
        if (new_err/numel(x) < tol)
            break
        endif
        beta = new_err / err
        p = r + beta * p
    endfor
endfunction

function [] = main()

    f = [1,2,3,2,1]
    f = f / sum(f)

    im = imread("lena_big.tif")[:,:,1]

    % Blur and add noise
    im_blurred = blur_filter(im, f) + 2*randn(size(im))

    % L2 restoration (not even Tikhonov!)
    im_restored = conjugate_gradients(((x : mat) -> blur_filter(x, f)), im_blurred)

    imshow(im_blurred); title("Blurred image")
    imshow(im_restored); title("Restored image")

    psnr = (x, y) -> 10*log10(255^2/mean((x - y).^2))
    print "psnr_in=", psnr(im_blurred, im), " dB"
    print "psnr_out=", psnr(im_restored, im), " dB"

endfunction