% Gradient descent for a two layer convolutional neural network (image denoising application)
% Sample uses Algorithmic Differentiation library. See documentation for the explanation.
{!title name="Two layer convolutional neural network using algorithmic differentiation"}
{!author name="Bart Goossens"}
{!doc category="Samples/Algorithmic Differentiation"}

import "Quasar.CompMath.dll" 
import "Quasar.DNN.dll"

type weighttype : vec[{cube{4}, cube{4}, vec}]

function y = __device__ pointwise_relu(x : scalar, alpha : scalar)
    y = x < 0 ? alpha * x : x
endfunction

reduction (x, alpha) -> $diff(pointwise_relu(x, alpha), x) = x < 0 ? alpha : 1
reduction (x, alpha) -> $adjoint(pointwise_relu(x, alpha), x) = pointwise_relu(x, alpha) % pointwise operation

function [y : cube{4}] = conv(x : cube{4}, f : cube{4})    
    [K,L] = int(size(f,2..3)/2) % half filter support size
    conv_desc = new(qdnn_conv_descriptor)
    conv_desc.pad_w = L
    conv_desc.pad_h = K
    input_c = x
    output_c = uninit(size(x,0),size(f,0),size(x,2),size(x,3))
    w_c = f
    cudnn_convolution_forward(1.0, input_c, 0.0, output_c, w_c, conv_desc)   
    y = output_c
endfunction

function [df : cube{4}] = conv_diff_f(x : cube{4}, f : cube{4}, y : cube{4})
    [K,L] = int(size(f,2..3)/2) % half filter support size
    conv_desc = new(qdnn_conv_descriptor)
    conv_desc.pad_w = L
    conv_desc.pad_h = K
    input_c = x
    dzdoutput_c = y   
    dzdw = uninit(size(f))
    cudnn_convolution_backward_filter(1.0,input_c,dzdoutput_c,0.0,dzdw,conv_desc)
    df = dzdw
endfunction

function [dx : cube{4}] = conv_diff_x(x : cube{4}, f : cube{4}, y : cube{4})
    [K,L] = int(size(f,2..3)/2)
    conv_desc = new(qdnn_conv_descriptor)
    conv_desc.pad_w = L
    conv_desc.pad_h = K
    input_c = copy(x)
    dzdoutput_c = copy(y)
    dzdinput = uninit(size(input_c))        
    cudnn_convolution_backward_data(1.0,f,dzdoutput_c,0.0,dzdinput,conv_desc)
    dx = dzdinput
endfunction

reduction (x, f, y) -> $diffmul(conv(x, f), f, y) = conv_diff_f(x, f, y)
reduction (x, f, y) -> $diffmul(conv(x, f), x, y) = conv_diff_x(x, f, y)
reduction (x, f, y) -> $adjoint(conv_diff_f(x, f, y), y) = conv_diff_f(x, f, y)
reduction (x, f, y) -> $adjoint(conv_diff_x(x, f, y), y) = conv_diff_x(x, f, y)


function y = relu(x : cube{4}, alpha : scalar)
    y = zeros(size(x))
    {!parallel for}
    for [m,n,p,q]=0..size(x,0..3)-1
        y[m,n,p,q] = pointwise_relu(x[m,n,p,q], alpha)
    endfor
endfunction

function y = average(x : cube{4}, w : vec)
    y = zeros(size(x,0),1,size(x,2),size(x,3))
    {!parallel for}
    for [q,m,n]=0..size(x,[0,2,3])-1
        sum = 0.0
        for p=0..size(x,1)-1
            sum += x[q,p,m,n] * w[p]
        endfor
        y[q,0,m,n] = sum
    endfor
endfunction

function y = forward_network(x : cube{4}, f : weighttype)   
    alpha = 1.0

    % Split into steps 
    c : cube{4} = conv(x, f[0])
    r : cube{4} = relu(c, alpha)
    c2 : cube{4} = conv(r, f[1])
    r2 : cube{4} = relu(c2, alpha)
    y = average(r2, f[2])
endfunction

function C = mse_data_fidelity(u : cube{4}, y : cube{4})
    C = 0.0
    for [m,n,p,q]=0..size(u,0..3)-1
        C += (u[m,n,p,q] - y[m,n,p,q])^2
    endfor
endfunction

function C = cost_function(u : cube{4}, y : cube{4}, f : weighttype)
    x = forward_network(y, f)
    C = mse_data_fidelity(u, x)
endfunction

function y = cellsum(x)
    if type(x, "cube{n}") || type(x, "ccube{n}")
        y = sum(x)
    elseif type(x, "scalar") || type(x, "cscalar")
        y = x
    else
        y = 0.0
        for i=0..numel(x)-1
            y += cellsum(x[i])
        endfor
    endif
endfunction

% gradient_descent is a generic function in Quasar (like C++ templates) because it 
% requires access to cost_fn. This is: as long as cost_fn cannot be determined at
% compile-time, the function cannot be used from user-code.
function y = gradient_descent(cost_fn, x, psnr, step_size=0.1, max_iter=50, epsilon=1e-10)
    % Maple-like syntax for obtaining the gradient
    gradient = $unapply($diff(cost_fn(x), x), x)

    % Initial solution
    y = x

    {!interpreted for}
    for k=0..max_iter
        alpha = step_size*0.01 + min(k/50,1)*step_size*0.99
        y = y - alpha * gradient(y)

        % Stop condition
        %if cellsum(abs(y-y_old)) < epsilon
        %    break
        %endif
        printf("Iteration %d, cost: %g, psnr: %g\n",k+1,cost_fn(y),psnr(y))
    endfor
endfunction

function y = nonlinear_cg(cost_fn, x, psnr, max_iter=5, epsilon=1e-20)    
    gradient = $unapply($diff(cost_fn(x), x), x)
    innerprod = (x, y) -> cellsum(x.*y)
    
    function alpha = line_search(cost_fn, y, s)
        alpha = 1.0
        orig_val = cost_fn(y)        
        while cost_fn(y - alpha * s) >= orig_val
            alpha *= 0.5
            if alpha < 1e-16                
                break
            endif
        endwhile
    endfunction

    % Initial solution
    y = x
    
    % Steepest decent solution
    Delta = gradient(y)
    s = Delta
    alpha = line_search(cost_fn, y, s)
    y -= alpha * s

    for k=0..max_iter
        printf("Iteration %d, cost: %g, psnr: %g\n",k+1,cost_fn(y),psnr(y))
        
        % Calculate the steepest direction
        Delta_last = Delta
        Delta = gradient(y)
        
        % Fletcher-Reeves update
        beta = innerprod(Delta,Delta)/innerprod(Delta_last,Delta_last)
    
        % Calculate the conjugate direction
        s_last = copy(s)
        s = Delta + beta * s_last
        
        % Perform a line search
        alpha = line_search(cost_fn, y, s)
    
        % Update the solution
        y_old = y
        y = y_old - alpha * s

        % Stop condition
        %if cellsum(abs(y-y_old)) < epsilon
        %    break
        %endif
    endfor

endfunction

function y = randcoeff(K,M,N,P)
    y = randn(K,M,N,P)
    y = y/sqrt(sum(y.^2))*10  
endfunction

function [] = main()

    % Filter coefficients to estimate
    % Dimensions: Output Input Height Width
    f = { randcoeff(4,1,9,9)/4, randcoeff(4,4,9,9)/4, ones(4)/4}
    
    im = imread("lena_big.tif")[:,:,1]
    im_noisy = im + 20*randn(size(im))
    
    im_c = shuffledims(im,[3,2,0,1])
    im_noisy_c = shuffledims(im_noisy,[3,2,0,1])    
    
    psnr = (x, y) -> 10*log10(255^2/mean((x - y).^2))
    my_psnr = f -> psnr(forward_network(im_noisy_c, f), im_c)

    f_new = gradient_descent(((f : weighttype) -> cost_function(im_c, im_noisy_c, f)), f, my_psnr, 5e-12,400)

    im_restored = squeeze(forward_network(im_noisy_c, f_new))

    imshow(im_noisy,[0,255]),title("Noisy image")
    imshow(im_restored,[0,255]),title("Denoised image")

    print "psnr_in=", psnr(im_noisy, im), " dB"
    print "psnr_out=", psnr(im_restored, im), " dB"

endfunction