% Newton's algorithm for L1-based regularization
% Sample uses Algorithmic Differentiation library. See documentation for the explanation.
import "Quasar.CompMath.dll"

{!title name="Total variation with Newton-based L1 regularization"}
{!author name="Bart Goossens"}
{!history date="April 22, 2015", desc="Initial implementation"}
{!doc category="Samples/Algorithmic Differentiation"}

reduction x -> $diff(abs(x),x) = sign(x)
reduction x -> $diff(sign(x),x) = 2*delta(x)
reduction x -> $diff(copy(x),x) = ones(size(x))

% C = sum((g - b).^2) + lambda*|dg/dx| + lambda*|dg/dy|

delta = __device__ (x) -> x==0 ? 1.0 : 0.0

function C : scalar = costfunc(g : mat, b : mat, lambda : scalar)
    C = 0.0
    for m=0..size(g,0)-1
        for n=0..size(g,1)-1
            C += (g[m,n] - b[m,n]).^2 % data fidelity
            C += lambda*(g[m,n-1] - g[m,n]).^2 % horizontal partial derivative
            C += lambda*(g[m-1,n] - g[m,n]).^2 % vertical partial derivative
        endfor
    endfor
endfunction

function y = newton(cost_fn : [?? -> scalar], x : cube, max_iter=45, epsilon=1e-6)    
    dx=zeros(size(x))
    % Maple-like syntax for obtaining the gradient and the Hessian in point x 
    gradient = $unapply($diff(cost_fn(x), x), x)
    % The hessian is actually the Jacobian of the gradient
    hessian = $unapply($diffmul(gradient(x), x, dx), x, dx)
    hessian_t = $unapply($adjoint(hessian(x, dx), dx), x, dx)

    max_inner_iter = 1
    step_size = 0.1
    tol = 1e-6
    lambda= 1e-4 % regularization

    % Initial solution
    y = x

    for k=0..max_iter
        % Solve H f(y) = x
        %   => H^T H f(y) = H^T grad(y)
        u = 0
        g = gradient(y)
        r = hessian_t(y,g)
        p = r
        new_err = dotprod(r, r)
        print "iteration ",k," err=",new_err

        for it=1..max_inner_iter  
            err=new_err
            %print "CG iteration ",it," err=",err
            A_p = lambda*p + hessian_t(y, hessian(y, p))
            alpha = err / dotprod(p, A_p)
            u = u + alpha * p
            r = r - alpha * A_p
            new_err = dotprod(r, r)
            if (new_err/numel(u) < tol)
                break
            endif
            beta = new_err / err
            p = r + beta * p
        endfor

        y_old = y
        y = y_old - step_size * u

        % Stop condition
        if sum(abs(y-y_old)) < epsilon
            break
        endif
    endfor

endfunction

function [] = main()
    
    im = imread("lena_big.tif")[:,:,1]
    im_noisy = im + 25 * randn(size(im))
    lambda = 25

    % Call the gradient descent algorithm to derive the cost function
    % The explicit lambda syntax is required for closure variable binding. This
    % way, the variables im_noisy and lambda, which are defined locally, can be
    % passed to the cost function. 
    im_den = newton(((x : mat) -> costfunc(x, im_noisy, lambda)), im_noisy)

    imshow(im_den,[0,255])

    psnr = (x, y) -> 10*log10(255^2/mean((x - y).^2))
    print "psnr_in=", psnr(im_noisy, im), " dB"
    print "psnr_out=", psnr(im_den, im), " dB"

endfunction