import "Quasar.CompMath.dll"

% Parallel-beam reconstruction using the conjugate gradient method, with
% automatic calculation of the adjoint.
% Sample uses Algorithmic Differentiation library. See documentation for the explanation.
{!title name="Parallel beam projection inversion (Radon Transform) using automatic adjoint calculation"}
{!author name="Bart Goossens"}
{!doc category="Samples/Algorithmic Differentiation"}

function y = parallel_beam(x : mat, angles : vec, offsets : vec)
    y = zeros(size(x))
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            d = offsets[m] % offset along the detector array
            theta = angles[n]
            v = [cos(theta), sin(theta)]
            w = [-sin(theta), cos(theta)]
            ctr = size(x)/2 + d * w

            lambda1 = ([size(x,0)-2.0, 0.0]-ctr[0])/v[0]
            lambda2 = ([size(x,1)-2.0, 0.0]-ctr[1])/v[1]
            lambda = [max(min(lambda1),min(lambda2)),min(max(lambda1),max(lambda2))]
            sum = 0.0

            % Note: the automatic adjoint algorithm requires a lower and upper bound
            % for this for-loop.
            assert(lambda[0]>=-512 && lambda[1] <= 512)
            for i=lambda[0]..lambda[1]
                p = ctr + i * v            
                sum += x[int(p)]
            endfor
            y[m,n] = sum
        endfor
    endfor
endfunction

function x = conjugate_gradients(A, y, max_iter=20, tol=1e-15)
    x = 0
    A_H = $unapply($adjoint(A(y), y), y)
    dotprod = (v, w) -> sum(sum(v .* w))
    r = A_H(y)
    p = r
    new_err = dotprod(r, r)
    
    %{!kernel_fusion}

    for it=1..max_iter  
        err=new_err
        print "CG iteration ",it," err=",err
        A_p = A_H(A(p))
        alpha = err / dotprod(p, A_p)
        x = x + alpha * p
        r = r - alpha * A_p
        new_err = dotprod(r, r)
        if (new_err/numel(x) < tol)
            break
        endif
        beta = new_err / err
        p = r + beta * p
    endfor
endfunction

function [] = main()
    im = imread("SheppLogan.png")
    im = im[:,:,0]

    angles = linspace(0,2*pi,256)
    offsets = -255..256

    % Forward projection
    y = parallel_beam(im, angles, offsets)

    % Inversion
    im_recon = conjugate_gradients(((x : mat) -> parallel_beam(x, angles, offsets)), y)

    imshow(im_recon)

    psnr = (x, y) -> 10*log10(255^2/mean((x - y).^2))
    print "psnr_out=", psnr(im_recon, im), " dB"


endfunction