import "Quasar.CompMath.dll"
import "linalg.q"

{!title name="Photometric registration using differentiable programming"}
{!author name="Bart Goossens"}
{!doc category="Samples/Algorithmic Differentiation"}

% Section: overview
% 
% In photometric registration, the goal is to map intensities of a reference
% image onto intensities of a current image. Even though the mapping is often 
% performed by a parametric function, here we will use a non-parametric function
% that is learned from the data. A smoothness constraint is added as regularization
% to ensure that the learned function is sufficiently smooth.
% 
% As the resulting loss function is quadratic, we will use two solvers to find
% the solution: 1) the method of conjugate gradients, and 2) a linear solver based
% on the Moore-Penrose pseudo-inverse. For the latter, the hessian of the loss
% function is calculated.

function C = loss_func(a : vec'clamped(256), x : cube, y : cube)   
    C = 0.0
    for [m,n,p]=0..size(x,0..2)-1
        {!function_transform enable="sharedmempromotion"}
        C += (x[m,n,p] - a[y[m,n,p]]).^2 
    endfor
    % Smoothness constraint
    lambda = 0.5*numel(x)
    for i=1..254
        C += lambda * (a[i-1] - 2*a[i] + a[i+1]).^2
    endfor
endfunction

function y = mapping_forward(a : vec'clamped(256), x : cube)
    y = zeros(size(x))
    for [m,n,p]=0..size(x,0..2)-1
        {!function_transform enable="sharedmempromotion"}
        y[m,n,p] = a[x[m,n,p]]
    endfor
endfunction

mapping_backward = (a : vec'clamped(256), x : cube, da : cube) -> _
    $diffmul(mapping_forward(a, x), a, da)
    
system_grad = (a : vec'clamped(256), x : cube, y : cube) -> _
    $diff(loss_func(a, x, y), a)
    
system_matrix = (a : vec'clamped(256), x : cube, y : cube) -> _
    $hessian(loss_func(a, x, y), a)
    

function y = gradient_descent(cost_fn, x, step_size=1e-15, max_iter=250, epsilon=1e-8)
    % Maple-like syntax for obtaining the gradient
    gradient = $unapply($diff(cost_fn(x), x), x)

    % Initial solution
    y = x

    for k=0..max_iter
        y_old = y
        y = y_old - step_size * gradient(y)
        if sum(isnan(y))>0
            print "NaN detected at iteration ",k
            break
        endif
        
        if mod(k,100)==0
            print "iter ",k, " cost_fn=", cost_fn(y)
        endif

        % Stop condition
        if sum(abs(y-y_old)) < epsilon
            break
        endif
    endfor
endfunction

function z = linear_cg(cost_fn, y0, max_iter=20, tol=1e-15)
    % Maple-like syntax for obtaining the gradient
    gradient = $unapply($diff(cost_fn(y0), y0), y0)

    z = y0
    dotprod = (v, w) -> sum(v .* w)
    
    D = -gradient(zeros(size(y0)))
    r = -gradient(z)  % b - a*x
    p = r
    new_err = dotprod(r, r)

    for it=1..max_iter  
        err=new_err
        print "CG iteration ",it," err=",err
        A_p = gradient(p) + D 
        alpha = err / dotprod(p, A_p)
        z = z + alpha * p
        z = min(max(z, 0.1), 5000)
        r = r - alpha * A_p
        new_err = dotprod(r, r)
        if (new_err/numel(z) < tol)
            break
        endif
        beta = new_err / err
        p = r + beta * p
    endfor
endfunction

reduction (a : cube) -> a[:] = reshape(a,numel(a))

function [] = main()
    im1 = imread("lighthouse.png")
    im1 = im1[0..8..size(im1,0)-1,0..8..size(im1,1)-1,:]
    im1 += 10*randn(size(im1))
    im1 = min(max(im1,10),240)
    im2 = 255 * (im1 / 255).^0.1 + 20 + 10*randn(size(im1))   

    if sum(isnan(im2))>0
        print "im2 contains NaNs"
        return
    endif
        
    scatter(im1[:],im2[:])%.set_renderer("opengl")
    hold("on")

    a = 0..255 
    % 1. Gradient descent method
    tic()
        a_out = linear_cg(((a : vec(256)) -> loss_func(a, im2, im1)), a, 500)
    toc("Conjugate gradient method")
    plot(a_out,"b")
    
    % Solve as a linear system - uses the Jacobian
    tic()
    A = system_matrix(a, im2, im1)
    B = -system_grad(zeros(size(a)), im2, im1)
    B = cast( transpose(inv(A)*transpose(B)), typename(vec(256)))
    
    toc("Solve as linear system")
    plot(B,"g")
    legend("Conjugate gradients", "Pseudo-inverse")
    
endfunction