{!author name="Bart Goossens"}
{!doc category="Samples/Basic/Rotate scale and translate"}
{!title name="Rotating, scaling and translating an input image"}

% Section: Overview
%
% This example demonstrates how to apply spatial transformation 
% (rotation, scale and translation) to an image, with linear interpolation, 
% and how this can be achieved using `__kernel__` and `__device__` functions.

% Function: rotatescaletranslate
%
% Rotates, scales and translates an input image
%
% Parameters:
% img_in - the input image
% img_out - the output image (does not necessarily need to have the same size
%     as the input image)
% theta - the rotation angle, in radials
% s - the scaling factor
% tx - the horizontal translation component
% ty - the vertical translation component

function [] = rotatescaletranslate(img_in, img_out, theta, s, tx, ty)
    % Device function for performing linear interpolation
    function [q:vec3] = __device__ linear_interpolate(img_in:cube'unchecked, p:vec2)
        i = floor(p)
        f = frac(p)
        if min(i) >= 0 && max(i-size(img_in,0..1)) < -1
            q = img_in[i[0],i[1],0..2]     * (1 - f[0]) * (1 - f[1]) + _
                img_in[i[0],i[1]+1,0..2]   * (1 - f[0]) * f[1] + _
                img_in[i[0]+1,i[1],0..2]   * f[0] * (1 - f[1]) + _
                img_in[i[0]+1,i[1]+1,0..2] * f[0] * f[1]
        else
            q = [0.,0.,0.]
        endif
    end

    function [] = __kernel__ tf_kernel(img_out : cube'unchecked, _
		img_in:cube'unchecked, A:mat2x2'unchecked, t:vec2, pos:ivec2)
        center = size(img_in,0..1)/2
        p = float(pos - center)
        p = [A[0,0]*p[0] + A[0,1]*p[1], A[1,0]*p[0] + A[1,1]*p[1]] + center + t
        img_out[pos[0],pos[1],0..2] = linear_interpolate(img_in, p)        
    end

    degrees_to_radians = theta -> theta*pi/180
    theta = degrees_to_radians(theta)
    A = [[cos(theta), -sin(theta)],
         [sin(theta),  cos(theta)]] * 2^s
    
    parallel_do(size(img_out,0..1),img_out,img_in,A,-[ty,tx],tf_kernel)
end

function [] = main()
    img_in = imread("lena_big.tif")
    img_out = zeros(size(img_in))
    rotatescaletranslate(img_in,img_out,45,1,0.5,0.0)
    imshow(img_out)
end
