import "Quasar.DNN.dll"

{!author name="Bart Goossens"}
{!doc category="Samples/Convolution/CUDNN"}
{!title name="Demonstrates how to use cuDNN functions in Quasar"}

% Section: overview
%
% This sample illustrates how to use the NVidia CUDNN library,
% which is highly optimized for performing forward and backward
% convolution.
%
% Data representation:
%
% CUDNN works with tensors, which are either
% - cube{4} objects of size N x C x H x W
% - cube{5} objects of size N x D x C x H x W
% where
% - N : number of images
% - C : number of image features (or color components)
% - H : height of the image
% - W : width of the image
%
% Note that in Quasar, the default format is H x W x C.
%
% Essentially, one single image can be converted to 
% this format using the function 
% shuffledims(X/255, [3, 2, 0, 1])
% but for convenience, the function cudnn_nchw_tensor() 
% performs exactly the same 
% operation. cudnn_hwcn_tensor() can be used for 
% converting back to the H x W x C x N format
%
% The CUDNN library can be accessed by importing Quasar.DNN.dll.
% This library also contains other functions, such as softmax,
% pooling etc. See the documentation for more information.
%

% Check whether CUDNN is installed/available on the
% current system
if cudnn_version() == 0
    print "cuDNN not supported for this machine."
    return    
endif

im = cudnn_nchw_tensor(imread("lena_big.tif")/255)
im_out = zeros(1, 3, 512, 512)

dY = copy(im)
dX = zeros(size(im_out))

pooling_desc = new(qdnn_pooling_descriptor)
sz = cudnn_pooling_getforwardoutputdim(im, pooling_desc)

% filter dimensions
% P x C x H x W
% - P is the number of output channels
% - C is the number of input channels
% - H is the filter height
% - W is the filter width
filter_coeff = zeros(3,3,5,5)
for m=0..2
    filter_coeff[:,m,:,:] = 1
end
filter_coeff = filter_coeff/numel(filter_coeff)*3

print size(im)

% Note: when possible, please cache the qdnn_conv_descriptor object!
conv_desc = new(qdnn_conv_descriptor)
conv_desc.pad_w = 2  % specifies horizontal center of the filter mask
conv_desc.pad_h = 2  % specifies vertical center of the filter mask
print conv_desc

num_runs = 100
tic()
for run=0..num_runs-1
    cudnn_convolution_forward(1.0, im, 0.0, im_out, filter_coeff, conv_desc)
end
toc("forward convolution")

tic()
print size(im), ";", size(im_out), ";", size(filter_coeff)
for run=0..num_runs-1
    cudnn_convolution_backward_filter(1.0, im, im_out, 0.0, filter_coeff, conv_desc)
end
toc("backward convolution filter")

tic()
for run=0..num_runs-1
    cudnn_convolution_backward_data(1.0, filter_coeff, im_out, 0.0, im, conv_desc)
end
toc("backward convolution data")
imshow(cudnn_hwcn_tensor(im_out))

B = zeros(1,3)
tic()
for run=0..num_runs-1
    cudnn_convolution_backward_bias(1.0, im_out, 0.0, B)
end
toc("backward convolution")

imshow(cudnn_hwcn_tensor(im), [])
