%
% filter3x3.q
%
% Bart Goossens, Aug 2012
%
{!author name="Bart Goossens"}
{!doc category="Samples/Convolution/3x3 taps filter"}
{!title name="Different implementations of a 3x3 filter"}

% Section: overview
%
% Example that shows the different options to implement a 3x3 filter: 
% 1) Non-separable implementation
% 2) Separable implementation, two stages
% 3) Separable implementation, one stage, using shared memory.
%
% In general, the approach using shared memory is the fastest 
% (up to 30% faster on most GPUs). Note that the out of bounds 
% checking compilation option needs to be turned off in order to have this benefit. 
%
% Also important is that the upper bounds for using shared memory
% are specified. This can be done using the assertion system. The
% compiler is then able to compute the maximal amount of shared memory
% that will be needed by the kernel function.
%
x = imread("lena_big.tif")
y = zeros(size(x))
z = zeros(size(x))
cnt = 999 % number of runs - for accurate time measurement we need to
          % average over a large number of runs.

% 1) Non-separable implementation
tic()
for k=0..cnt-1
    parallel_do(size(y),x,y,__kernel__ (x:cube,y:cube,pos:ivec3) -> _
        y[pos] = (x[pos+[-1,-1,0]]+x[pos+[-1,0,0]]+x[pos+[-1,1,0]] + _
                  x[pos+[ 0,-1,0]]+x[pos         ]+x[pos+[0,1,0]] + _
                  x[pos+[ 1,-1,0]]+x[pos+[ 1,0,0]]+x[pos+[1,1,0]]) * (1.0/9))
end
toc("Non-separable implementation")

% 2) Separable implementation, two stages
tic()
for k=0..cnt-1
    parallel_do(size(y),x,y,__kernel__ (x:cube,y:cube,pos:ivec3) -> y[pos] = x[pos+[-1,0,0]]+x[pos]+x[pos+[1,0,0]])
    parallel_do(size(x),y,z,__kernel__ (x:cube,y:cube,pos:ivec3) -> y[pos] = x[pos+[0,-1,0]]+x[pos]+x[pos+[0,1,0]]*(1.0/9))
end
toc("Separable implementation")

% 3) Separable implementation, one stage, using shared memory.

function [] = __kernel__ filter3x3_kernel_separable(x:cube,y:cube,pos:ivec3,blkpos:ivec3,blkdim:ivec3)        
    [M,N,P] = blkdim+[2,0,0]
    assert(M<=10 && N<=16 && P<=3) % specify upper bounds for the amount of shared memory
    vals = shared(M, N, P) % shared memory      

    sum = 0.
    for i=pos[1]-1..pos[1]+1   % step 1 - horizontal filter
        sum += x[pos[0],i,blkpos[2]]
    end
    vals[blkpos] = sum  % store the result

    if blkpos[0]<2 % filter two extra rows (needed for vertical filtering
        sum = 0.
        for i=pos[1]-1..pos[1]+1
            sum += x[pos[0]+blkdim[0],i,blkpos[2]]
        end
        vals[blkpos+[blkdim[0],0,0]] = sum 
    endif

    syncthreads
        
    sum = 0.
    for i=blkpos[0]..blkpos[0]+2   % step 2 - vertical filter
        sum += vals[i,blkpos[1],blkpos[2]]
    end
    y[pos] = sum*(1.0/9)
end

tic()
for k=0..cnt-1
    parallel_do([size(y),[8,16,3]],x,y,filter3x3_kernel_separable)
end
toc("Separable implementation")

% 4) Automatically generated separable implementation, using shared memory
% 
function [] = __kernel__ filter3x3_kernel_auto(x:cube,y:cube,pos:ivec3,blkpos:ivec3,blkdim:ivec3)        
    {!kernel_transform enable="localwindow"}
    y[pos] = (x[pos+[-1,-1,0]]+x[pos+[-1,0,0]]+x[pos+[-1,1,0]] + _
              x[pos+[ 0,-1,0]]+x[pos         ]+x[pos+[0,1,0]] + _
              x[pos+[ 1,-1,0]]+x[pos+[ 1,0,0]]+x[pos+[1,1,0]]) * (1.0/9)
end

tic()
for k=0..cnt-1
    parallel_do(size(y,0..2),x,y,filter3x3_kernel_auto)
end
toc("Local windowing transform")


imshow(y,[])
