%
% gaussian_filter.q
%
% Bart Goossens, Aug 2012
%
{!author name="Bart Goossens"}
{!doc category="Samples/Convolution/Gaussian filter"}
{!title name="2D Gaussian filter sample implementation"}

% Section: Overview
%
% This example illustrates how to implement a separable Gaussian filter in Quasar.
% We specify `'mirror` as a boundary extension method for the convolution. 
% Different boundary extension methods are possible, see manual.
% We use the local windowing kernel transform to automatically cache the input
% image in the shared memory of the GPU. This is achieved using the line:
% 
% :{!kernel_transform enable="localwindow"}
%
% Note that for the CPU, this directive has no effect.
function y = gaussian_filter(x : cube(:,:,3), sigma)
    function [] = __kernel__ gaussian_filter_hor(x : cube'mirror, y : cube'unchecked, fc : vec'unchecked'const, n : int, pos : vec3)
        {!kernel_transform enable="localwindow"}
        assert(numel(fc)<=11) % Upper bound required for calculating the amount of shared memory
        sum = 0.
        for i=0..numel(fc)-1
            sum = sum + x[pos + [0,i-n,0]] * fc[i]
        end
        y[pos] = sum
    end

    function [] = __kernel__ gaussian_filter_ver(x : cube'mirror, y : cube'unchecked, fc : vec'unchecked'const, n : int, pos : vec3)
        {!kernel_transform enable="localwindow"}
        assert(numel(fc)<=11) % Upper bound required for calculating the amount of shared memory
        sum = 0.
        for i=0..numel(fc)-1
            sum = sum + x[pos + [i-n,0,0]] * fc[i]
        end
        y[pos] = sum
    end

    % computation of the filter coefficients
    n = 4
    fc = exp(-0.5*(-n..n).^2/sigma)
    fc = fc / sum(fc)

    y = zeros(size(x))
    tmp = zeros(size(x))
    parallel_do (size(y), x, tmp, fc, n, gaussian_filter_hor)
    parallel_do (size(y), tmp, y, fc, n, gaussian_filter_ver)
end

% Place the main function with test code here, so that the function
% 'gauss_filter' can still be used from other Quasar programs, without
% running the test code.
function [] = main()
    img_in = imread("lena_big.tif")    
    num_runs = 1000
    
    tic()
    for i=1..num_runs
        img_out = gaussian_filter(img_in, 4)
    end
    toc()
    imshow(img_out)
end