% imagedomain_vs_fft.q
%
% Bart Goossens, Nov 2012
%
{!author name="Bart Goossens"}
{!doc category="Samples/Convolution/Image domain filter vs FFT filter"}
{!title name="Comparison of filtering in image domain versus FFT domain "}

% Section: Overview
%
% This example investigates whether it is more efficient to implement a 
% separable filter with a given support size in the image domain OR
% in the FFT domain.
% Note that the results maybe computation engine specific.
%
import "bregman_utils.q"

function fc = compute_mask(sigma = 20, n)
    fc = exp(-0.5*(-n..n).^2/sigma)
    fc = fc / sum(fc)
end

function y = gaussian_filter(x, fc, n)
    function [] = __kernel__ gaussian_filter_hor(x : cube'circular, y : cube'unchecked, fc : vec'unchecked'const, n : int, pos : vec3)
        sum = 0.
        for i=0..numel(fc)-1
            sum = sum + x[pos + [0,i-n,0]] * fc[i]
        end
        y[pos] = sum
    end

    function [] = __kernel__ gaussian_filter_ver(x : cube'circular, y : cube'unchecked, fc : vec'unchecked'const, n : int, pos : vec3)
        sum = 0.
        for i=0..numel(fc)-1
            sum = sum + x[pos + [i-n,0,0]] * fc[i]
        end
        y[pos] = sum
    end

    y = uninit(size(x))
    tmp = uninit(size(x))
    parallel_do (size(y), x, tmp, fc, n, gaussian_filter_hor)
    parallel_do (size(y), tmp, y, fc, n, gaussian_filter_ver)
end

function [] = main()

    img_in = imread("lena_big.tif")[:,:,1]
    sigma = 20
    runs = 10
    range_n = 3..31
    time = zeros(2,numel(range_n))

    for k=0..numel(range_n)-1
        n = range_n[k]
        fc = compute_mask(sigma, n)
        H = fft2(kernelshift(transpose(fc)*fc, size(img_in,0), size(img_in,1)))

        tic()
        for l=0..runs-1
            img_out = gaussian_filter(img_in, fc, n)
        end
        time[0,k]=toc()

        tic()
        for l=0..runs-1
            img_out = real(ifft2(fft2(img_in) .* H))
        end
        time[1,k]=toc()
    end

    plot(2*range_n+1, time)
    title("Image domain filtering vs. FFT domain filtering")
    xlabel("filter support size")
    ylabel("computation time (s)")
end
