%REQUIRES_DISPLAY
% particle_system.q
% Aug 2014
{!author name="Bart Goossens"}
{!doc category="Samples/Graphics/OpenGL particle system"}
{!title name="OpenGL particle system demonstration"}

% Section: Overview
%
% This sample demonstrates how OpenGL vertex buffers can be used and updated from Quasar.
% It suffices to define the right data structures (with position, normal).
% The data structures will be automatically mapped onto OpenGL vertex formats. 

import "Quasar.UI.dll"

type vertex : class
    position : vec3
    old_position : vec3
    color : vec3
end

norm : [__device__ (vec3)->(scalar)] = __device__ (x:vec3) -> sqrt(sum(x.^2))
normalize : [__device__ (vec3)->(vec3)] = __device__ (x:vec3) -> x / norm(x)

function [v : vec[vertex]] = build_pointcloud(coord : mat)
    v = vec[vertex](size(coord,0))
    
    % Fill the vertex buffer with the vertices
    {!parallel for}
    for k = 0..numel(v)-1
        r = sum(abs(coord[k,0..2]))
        u = (0.25+coord[k,1])*2
        p = [coord[k,0], coord[k, 1], coord[k, 2]]
        speed = normalize(p)
        v[k] = vertex(old_position:=p, 
                      position:=p + speed,
                      color:=[min(1,1-r+u), u, r])
    end
end   

function [] = update_pointcloud(v : vec[vertex], t : scalar, delta_t : scalar)

    attractor = 0.2*[cos(2*t),sin(sqrt(3)*cos(t)),cos(sqrt(2)*sin(t))]

    {!parallel for}
    for k = 0..numel(v)-1
        a = attractor - v[k].position
        a = normalize(a)
        
        old_position = v[k].old_position        
        new_position = 2 * v[k].position - old_position + 10*a*delta_t^2.0
        
        % Keep all particles inside
        if norm(new_position) > 2
            new_position = 2*normalize(new_position)
        endif
        
        if norm(new_position)<0.3
            new_position = new_position + (new_position-old_position)*0.01
        endif
        
        v[k] = vertex(old_position := v[k].position, 
                      position := new_position,
                      color := v[k].color)
        
    end

end

function [] = main()

    frm = form("OpenGL point-cloud visualization demo")
    frm.width = 800
    disp = frm.add_display()
    disp.sync_framerate(40) % Render at 40 FPS
    renderer = disp.create_opengl_renderer()          
    layer = new(qvectorlayer3d)
    renderer.add(layer, "layer")    
    frm.show()

    % Generate a set of random coordinates
    points = build_pointcloud(0.5*(rand(4000000,3)-0.5))
    layer.setpointsize(0.01)
    layer.drawpoints(points)
    
    t = 0.0
    delta_t = 0.005
    while !frm.closed()
        t += delta_t
        update_pointcloud(points, t, delta_t)
        layer.update()
        pause(10)
    end

end