%==============================================================================
% Bregman_deblurring_and_denoising.q
%==============================================================================
{!author name="Bart Goossens"}
{!doc category="Samples/Image Processing/Bregman deblurring and denoising"}
{!title name="Shows how to implement an Bregman deblurring and denoising filter"}

% Section: overview
% 
% Deconvolution & denoising algorithm for RGB data, using Bregman
% Total variation (TV) and shearlets.

import "system.q"
import "dst.q"
import "bregman_utils.q"
import "imfilter.q"
import "colortransform.q"
import "multirestransforms.q"
import "linalg.q"

#pragma expression_optimizer(on)

function result = bregman_deblurring_denoising(img_blurred, sparsity_tf = "shearlet", _
    num_scales = 3, sigma_blur = 0.5, mu = 1, _
    lambda_tv = 2, lambda_shearlet = 0.25, outer_iterations = 5, inner_iterations = 3)

    gamma = 1e-2  % stabilization constant
    sz = size(img_blurred)

    % Compute the inter-band covariance matrix
    if size(img_blurred,2)>1
        img_f = imfilter(img_blurred, [[0,-1,0],[-1,4,-1],[0,-1,0]])
        C = zeros(size(img_f,2)*[1,1])
        for m=0..size(C,0)-1
            for n=m..size(C,1)-1
                C[m,n] = C[n,m] = sum(img_f[:,:,m] .* img_f[:,:,n])        
            end
        end
        img_f = 0 % Release the memory in use
        [U,Lambda,V] = svd(C) % Then U is the decorrelating transform
        Ut = transpose(U)
    else
        U = 1
        Ut = 1
    endif

    % Compute the Gaussian blurring kernel
    hdims = 15
    [m,n] = meshgrid(-hdims..hdims,-hdims..hdims);
    kernel = exp(-0.5*(abs(m)+abs(n))/sigma_blur)
    kernel = kernel / sum(kernel)
    kernelt = circshift(flipud(fliplr(kernel)),[1,1])

    img_orig = copy(img_blurred)

    % Setup the sparsity transform
    bw = pi/32
    K = 4
    J = num_scales
    match sparsity_tf with
    | "shearlet" -> 
            [S, S_H] = build_energyweighted_dst2d(sz, K, J)
    | "wavelet"  -> 
            [S, S_H] = build_dwt(db8, J)
    | "dtcwt"    -> 
            [S, S_H] = build_dtcwt(filtercoeff_farras, filtercoeff_dualfilt, J)
    | "steerable pyramid"      -> 
            [S, S_H] = build_stp(sz, K, J)
    | "none"     ->
            S = S_H = (x) -> 0
    | _          -> 
            error("Sparsity transform unknown!")
    end

    dx = zeros(sz)
    dy = zeros(sz)
    bx = zeros(sz)
    by = zeros(sz)
    result = copy(img_blurred)

    ds = S(dx)
    bs = copy(ds)
    hfft = fft2(kernelshift(kernel,sz[0],sz[1]))
    htfft = fft2(kernelshift(kernel,sz[0],sz[1]))

    extend_dim = x -> repmat(x,[1,1,sz[2]])

    % total variation part
    fidelity = real(ifft2(mu * extend_dim(htfft) .* fft2(img_blurred)))
    laplacian = kernelshift([[0,-1,0],[-1,4,-1],[0,-1,0]],sz[0],sz[1])

    uk = mu*(htfft.*hfft) + lambda_tv*fft2(laplacian) + (lambda_shearlet + gamma)
    uk = extend_dim(uk) % extend the number of components if necessary...

    for outer=1..outer_iterations
        for inner=1..inner_iterations
            f = fft2(fidelity + colortransform(Ut, lambda_tv * Dxt(dx - bx) + lambda_tv * Dyt(dy - by) + _
                lambda_shearlet * S_H(ds - bs)) + gamma * result)
            result = real(ifft2(f ./ uk))

            result2 = colortransform(U, result)
            gx = Dx(result2)
            gy = Dy(result2)
            gs = S(result2)
            dx = shrink(gx + bx,1/lambda_tv)
            dy = shrink(gy + by,1/lambda_tv)
            ds = shrink(gs + bs,1/lambda_shearlet)

            bx = bx + gx - dx
            by = by + gy - dy
            bs = bs + gs - ds
        end

        img_blurred += img_orig - real(ifft2(extend_dim(hfft) .* fft2(result)))
        fidelity = real(ifft2(mu * extend_dim(htfft) .* fft2(img_blurred)))
    end
end

function [] = main()

    img_orig = imread("lena_big.tif")
    img_orig = imfilter(img_orig, [[0,1,0],[1,2,1],[0,1,0]]/6,[1,1])
    img_orig += randn(size(img_orig))

    imshow(copy(img_orig))
    title("Original image")

    tic()
    result = bregman_deblurring_denoising(img_orig, "shearlet")
    toc()

    imshow(result,[0,255])
    title("Deblurred image")

%    print "psnr=", psnr(img_orig,img),"dB->", psnr(result,img), "dB"
end