{!author name="Bart Goossens"}
{!doc category="Samples/Image Processing/Wavelet transform thresholding"}
{!title name="Wavelet transform domain thresholding"}

% Section: overview
%
% Example of image denoising: soft-thresholding using the orthogonal wavelet transform. 

import "dwt.q"

% shrinkage of a subband
function y = shrink(x, T)
    function [] = __kernel__ process(x : cube'unchecked, y : cube'unchecked, T : scalar, pos : ivec3)
        xi = x[pos]
        y[pos] = sign(xi).*max(abs(xi)-T,0.)
    endfunction
    
    y = zeros(size(x))
    parallel_do(size(x),x,y,T,process)        
endfunction

% shrinkage of all wavelet subbands except the lowpass band
function [] = shrink_inplace(x, T, J)
    function [] = __kernel__ process(x : cube'unchecked, T : scalar, B : ivec2,  pos : ivec3)
        if pos[0] > B[0] || pos[1] > B[1]  % Outside the lowpass wavelet subband
            xi = x[pos]
            x[pos] = sign(xi).*max(abs(xi)-T,0.)
        endif
    endfunction
    
    parallel_do(size(x),x,T,size(x,0..1)/2^J,process)        
endfunction

function y : cube = threshold_cellvector(x, J, wavelet, T)
    % Apply the 2D wavelet transform to the input image
    w = dwt2d(x, wavelet, J)
    % Shrink each wavelet subband (the wavelet subbands are stored in cell vectors)
    for j=0..J-1
        for k=0..2
            w[j][k] = shrink(w[j][k], T) 
        endfor
    endfor
    % Calculate the inverse 2D wavelet transform
    y = idwt2d(w, wavelet, J)
endfunction

% Untiled implementation, which is expected to be slightly faster than 'threshold_cellvector',
% because cell vectors do not have to be created
function y = threshold_untiled(x, J, wavelet, T)
    y = copy(x)
    % Apply the 2D wavelet transform to the input image
    dwt2d_untiled(y, wavelet, J)
    shrink_inplace(y, T, J)    
    % Calculate the inverse 2D wavelet transform
    idwt2d_untiled(y, wavelet, J)
endfunction


function [] = main()

    x = imread("lena_big.tif")
    % add noise
    x += 25*randn(size(x))
    J = 3
    T = 25
    runs = 10
    w = orthogonal_wavelet(db2)
    
    tic()
    for run=1..runs
        y = threshold_cellvector(x,J,w,T)
    endfor
    toc("Cell matrix implementation")
    
    tic()
    for run=1..runs
        y = threshold_untiled(x,J,w,T)
    endfor
    toc("Untiled implementation")
    
    imshow(y)

endfunction
