%
% inplace_haar.q
%
% Bart Goossens, Aug 2012
%
{!author name="Bart Goossens"}
{!doc category="Samples/Image Processing/Inplace haar wavelet transform"}
{!title name="Inplace haar wavelet transformation demonstration"}

% Section: Overview
%
% The following code demonstrates an inplace Haar wavelet transform, implemented
% using the lifting scheme (but without normalization). The forward and backward 
% transform respectively use the 2 x 2  transform matrices:
% :   A      = [1/2  1/2]
% :            [  1   -1]
% and
% :   A^(-1) = [  1  1/2]
% :          = [  1 -1/2]
% The main advantages of the Haar wavelet transform in the context of Quasar 
% programs, is that the transform is very fast (takes less than 2 ms to compute 
% for a 512 x 512 x 3  input image on a NVidia Geforce 435M using the 
% CUDA computation engine). Moreover, for integer input data within the range [0,255], 
% this unnormalized transform does not suffer from floating point rounding errors, 
% hence the reconstruction (backward transform applied after the forward transform)
% is exact.
%
import "system.q"

% Forward haar transform
function [] = haar_fw(x, num_scales)
    function [] = __kernel__ hor_haar_fw_kernel(x : cube'unchecked, y : cube'unchecked, j : int, pos : ivec3)
        n = size(x,1)/2^(j+1)
        if mod(pos[1],2)==0
            [a, b] = [x[pos], x[pos+[0,1,0]]]
            y[pos[0],pos[1]/2,pos[2]]=0.5*(a+b)
            y[pos[0],pos[1]/2+n,pos[2]]=a-b
        endif
    end
    function [] = __kernel__ ver_haar_fw_kernel(x : cube'unchecked, y : cube'unchecked, j : int, pos : ivec3)
        m = size(x,0)/2^(j+1)
        if mod(pos[0],2)==0
            [a, b] = [x[pos], x[pos+[1,0,0]]]
            y[pos[0]/2,pos[1],pos[2]]=0.5*(a+b)
            y[pos[0]/2+m,pos[1],pos[2]]=a-b
        endif
    end

    tmp = zeros(size(x))
    for j=0..num_scales-1
        sz = [size(x,0)/2^j,size(x,1)/2^j,size(x,2)]
        parallel_do(sz,x,tmp,j,hor_haar_fw_kernel)
        parallel_do(sz,tmp,x,j,ver_haar_fw_kernel)
    end
end

% Inverse haar transform
function [] = haar_bw(x, num_scales)
    function [] = __kernel__ hor_haar_bw_kernel(x : cube'unchecked, y : cube'unchecked, j : int, pos : ivec3)
        n = size(x,1)/2^(j+1)
        if mod(pos[1],2)==0
            [a, b] = [x[pos[0],pos[1]/2,pos[2]], x[pos[0],pos[1]/2+n,pos[2]]]
            y[pos]=a+0.5*b
            y[pos+[0,1,0]]=a-0.5*b
        endif
    end
    function [] = __kernel__ ver_haar_bw_kernel(x : cube'unchecked, y : cube'unchecked, j : int, pos : ivec3)
        m = size(x,0)/2^(j+1)
        if mod(pos[0],2)==0
            [a, b] = [x[pos[0]/2,pos[1],pos[2]], x[pos[0]/2+m,pos[1],pos[2]]]
            y[pos]=a+0.5*b
            y[pos+[1,0,0]]=a-0.5*b
        endif
    end

    tmp = zeros(size(x))
    for j=num_scales-1..-1..0
        sz = [size(x,0)/2^j,size(x,1)/2^j,size(x,2)]
        parallel_do(sz,x,tmp,j,hor_haar_bw_kernel)
        parallel_do(sz,tmp,x,j,ver_haar_bw_kernel)
    end
end

function [] = main()
    x = imread("lena_big.tif")
    y = copy(x)
    J = 4
    tic()
    for k=0..99
    haar_fw(y, J)
    haar_bw(y, J)
    end
    toc()
    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    print "psnr=",psnr(x,y),"dB"
    imshow(y)
end