%REQUIRES_DISPLAY
{!author name="Bart Goossens"}
{!doc category="Samples/Image Processing/Lens blurring"}
{!title name="Real-time lens blurring filter"}

%
% Section: Overview
% 
% This script demonstrates how to perform artificial lens blurring in Quasar,
% using a depth image and an RGB image

import "Quasar.UI.dll"
import "system.q"


% Function: gaussian_filter
% 
% Performs a space-variant Gaussian filtering operation
% 
% Usage:
%   : function [y] = gaussian_filter(x, sigma)
% 
% Parameters:
% x - the input image
% sigma - the sigma values of the blur kernel, for 
%         every position in the image
% 
% Returns:
% y - the output image
function y = gaussian_filter(x, sigma)
    function [] = __kernel__ gaussian_filter_hor(x : cube'mirror, y : cube'unchecked, sigma : mat'unchecked, n : int, pos : vec2)
        s = sigma[pos]
        if s == 0  % sigma=0 corresponds to no filtering
            y[pos[0],pos[1],0..2] = x[pos[0],pos[1],0..2]
            return
        endif       
        sum = [0.,0.,0.,0.]
        for i=-n..n % Convolution loop
            w = exp(-(i)^2/s^2)   % Calculate weight
            sum[0..2] = sum[0..2] + x[pos[0],pos[1]+i,0..2] * w  % Accumulate RGB
            sum[3] = sum[3] + w % Accumulate weights
        end
        y[pos[0],pos[1],0..2] = sum[0..2] / sum[3]
    end
    function [] = __kernel__ gaussian_filter_ver(x : cube'mirror, y : cube'unchecked, sigma : mat'unchecked, n : int, pos : vec2)
        s = sigma[pos]
        if s == 0   % sigma=0 corresponds to no filtering
            y[pos[0],pos[1],0..2] = x[pos[0],pos[1],0..2]
            return
        endif
        sum = [0.,0.,0.,0.]
        for i=-n..n  % Convolution loop
            w = exp(-(i)^2/s^2)    % Calculate weight
            sum[0..2] = sum[0..2] + x[pos[0]+i,pos[1],0..2] * w   % Accumulate RGB
            sum[3] = sum[3] + w   % Accumulate weights
        end
        y[pos[0],pos[1],0..2] = sum[0..2] / sum[3]
    end

    % computation of the filter coefficients    
    y = uninit(size(x))
    tmp = uninit(size(x))
    ntaps = 15
    parallel_do (size(y), x, tmp, sigma, ntaps, gaussian_filter_hor)
    parallel_do (size(y), tmp, y, sigma, ntaps, gaussian_filter_ver)
end

function [] = main()
    im_rgb = imread("col_tarot.jpg")
    im_d : cube = imread("dmap_tarot.jpg")
    
    show_input_images = false

    if show_input_images
        imshow(im_rgb);title("RGB image")
        imshow(im_d); title("Depth image")
    endif

    im_d = gaussian_filter(im_d, 15*ones(size(im_d,0..1)))
    frm = form("Lens defocusing demo")
    disp = frm.add_display()    
    disp.sync_framerate(40)    

    state = object()    
    state.focus_point = size(im_rgb,0..1) / 2
    state.new_focus_point = state.focus_point

    function [] = update(p)
        state.new_focus_point = p
    end

    frm.show()
    h = disp.imshow(im_rgb,[0,255])
    h.onSelectPoint.add(update)       
    [x, y] = meshgrid(0..size(im_d,1)-1,0..size(im_d,0)-1)
    
    while !frm.closed()
        state.focus_point += 0.25 * (state.new_focus_point - state.focus_point)
        sigma = (
            0.0025*(im_d[state.focus_point[0], state.focus_point[1],0] - im_d[:,:,0]).^2 +
            0.000001*(state.focus_point[0] - y).^2 +
            0.000001*(state.focus_point[1] - x).^2)
        
        im_defocused = gaussian_filter(im_rgb, sigma)
        disp.imshow(im_defocused,[0,255])           
        pause(0)
    end
    
end
