{!author name="Bart Goossens"}
{!doc category="Samples/Image Processing/Primal dual deblurring and denoising"}
{!title name="Primal-dual convex optimization algorithm for deblurring and denoising"}

% Section: Overview
%
% Deconvolution & denoising algorithm for RGB data, using the Primal-Dual
% optimization method. Three different sparsity transforms are available:
% * Shearlets
% * Wavelets
% * Complex wavelets
% * Steerable pyramids
%
import "system.q"
import "dst.q"
import "bregman_utils.q"
import "imfilter.q"
import "colortransform.q"
import "multirestransforms.q"
import "inttypes.q"

function result_primal = primaldual_deblurring_denoising(img_blurred, sparsity_tf = "shearlet", _
    color_transform = "PCA", _
    num_scales = 3, sigma_blur = 1, mu = 1, _
    lambda_shearlet = 10, pd_iterations = 50)

    gamma = 1e-2  % stabilization constant
    sz = size(img_blurred)

    % Compute the color transform
    %[U, Ut] = get_colortransform_from_RGB(color_transform, img_blurred)
    U = eye(3)
    Ut = eye(3)

    % Compute the Gaussian blurring kernel
    hdims = 15
    [m,n] = meshgrid(-hdims..hdims,-hdims..hdims);
    kernel = exp(-0.5*(m.^2+n.^2)/sigma_blur^2)
    kernel = kernel / sum(kernel)
    kernelt = circshift(flipud(fliplr(kernel)),[1,1])

    img_orig = copy(img_blurred)

    % Setup the sparsity transform
    bw = pi/32
    K = 4
    J = num_scales
    match sparsity_tf with
    | "shearlet" -> 
            [S, S_H] = build_energyweighted_dst2d(sz, K, J)
    | "wavelet"  -> 
            [S, S_H] = build_dwt(db8, J)
    | "dtcwt"    -> 
            [S, S_H] = build_dtcwt(filtercoeff_farras, filtercoeff_dualfilt, J)
    | "steerable pyramid"      -> 
            [S, S_H] = build_stp(sz, K, J)
    | _          -> 
            error("Sparsity transform unknown!")
    end

    result_primal = copy(img_blurred)
    result_dual = S(result_primal)
    result_predict = copy(result_primal)

    hfft = fft2(kernelshift(kernel,sz[0],sz[1]))
    htfft = conj(fft2(kernelshift(kernel,sz[0],sz[1])))

    extend_dim = x -> repmat(x,[1,1,sz[2]])

    % Primal dual
    t = 10
    s = 1/t
    theta = 0.0
    term_nominator = t*lambda_shearlet*(fft2(img_blurred).*extend_dim(htfft))
    term_denominator = extend_dim(t*lambda_shearlet*htfft.*hfft + 1)

    print "Processing..."

    for outer=1..pd_iterations
        spredict = S(result_predict)
        expr_dual = result_dual + s*spredict
        result_dual = expr_dual ./ max(1, abs(expr_dual))

        result_primal2 = result_primal
        expr_primal = result_primal - t*S_H(result_dual)
        result_primal = real(ifft2((term_nominator + fft2(expr_primal)) ./ term_denominator)) % monotone operator
        result_predict = result_primal + theta*(result_primal - result_primal2)
        pause(0)
    end    
end

function [] = main()

    img_orig = imread("lena_big.tif")
    img_orig = imfilter(img_orig, [[0,1,0],[1,2,1],[0,1,0]]/6,[1,1])
    img_orig += randn(size(img_orig)) * 1

    imshow(img_orig)
    title("Original image")

    tic()
    result = primaldual_deblurring_denoising(img_orig, "shearlet")
    toc()

    imshow(result,[0,255])
    title("Primal-dual deblurred image")

%    print "psnr=", psnr(img_orig,img),"dB->", psnr(result,img), "dB"
end