%REQUIRES_DISPLAY
{!author name="Bart Goossens"}
{!doc category="Samples/Image Processing/RAW digital camera images"}
{!title name="Sample reconstruction of RAW digital camera images"}

% Section: Overview
%
% Demonstration of the Quasar LibRaw library for
% reading and reconstructing RAW digital camera images. The current
% version of libraw bundled with Quasar is version 0.15.
%
% Note - this demo version simply uses bilinear interpolation,
% which is obviously not the best you can do.
%
import "LibRawNet.dll"
import "Quasar.UI.dll"
import "imhist.q"

% Function: demosaic_bilinear
%
% Bilinear demosaicing for Bayer CFAs 
%
% Parameters:
% im - image in "Bayer format" (i.e. 4 color channels, R1,G1,B1,G2)
%====================================================================
function im_out : cube = demosaic_bilinear(im : cube, bayer_pattern)
    function [] = __kernel__ mosaic(im_out : cube'unchecked, im_in : cube'unchecked, i_pattern : ivec4, o_pattern : ivec4, pos : ivec2)
        x = im_in[pos[0],pos[1],0..3]        
        im_out[2*pos[0],2*pos[1],o_pattern[0]]     = x[i_pattern[0]]
        im_out[2*pos[0],2*pos[1]+1,o_pattern[1]]   = x[i_pattern[1]]
        im_out[2*pos[0]+1,2*pos[1],o_pattern[2]]   = x[i_pattern[2]]
        im_out[2*pos[0]+1,2*pos[1]+1,o_pattern[3]] = x[i_pattern[3]]
    end
    function [] = __kernel__ demos_bilinear_step1(im_out : cube, o_pattern : ivec4, pos : ivec3)
        m = and(pos[0],1)
        n = and(pos[1],1)
        v = o_pattern[and(m+1,1)*2+n] == pos[2]
        h = o_pattern[m*2+and(n+1,1)] == pos[2]
        if o_pattern[m*2+n] != pos[2]
            nb_up = im_out[pos-[1,0,0]]
            nb_down = im_out[pos+[1,0,0]]
            nb_left = im_out[pos-[0,1,0]]
            nb_right = im_out[pos+[0,1,0]]
            if h && v
                im_out[pos] = (nb_up + nb_down + nb_left + nb_right) * 0.25
            elseif v
                im_out[pos] = (nb_up + nb_down) * 0.5
            elseif h
                im_out[pos] = (nb_left + nb_right) * 0.5                        
            endif
        endif        
    end
    function [] = __kernel__ demos_bilinear_step2(im_out : cube, o_pattern : ivec4, pos : ivec3)
        m = and(pos[0],1)
        n = and(pos[1],1)
        v = o_pattern[and(m+1,1)*2+n] == pos[2]
        h = o_pattern[m*2+and(n+1,1)] == pos[2]
        if o_pattern[m*2+n] != pos[2]
            if !h && !v
                nb_up = im_out[pos-[1,0,0]]
                nb_down = im_out[pos+[1,0,0]]
                nb_left = im_out[pos-[0,1,0]]
                nb_right = im_out[pos+[0,1,0]]
                im_out[pos] = (nb_up + nb_down + nb_left + nb_right) * 0.25
            endif
        endif
    end
    [rows,cols,C] = size(im)
    im_out = uninit(2*rows,2*cols,3)
    i_pattern = reshape(bayer_pattern,4)
    o_pattern = copy(i_pattern)
    o_pattern = o_pattern - 2 * (o_pattern == 3)
    parallel_do([rows,cols],im_out,im,i_pattern,o_pattern,mosaic)
    parallel_do(size(im_out),im_out,o_pattern,demos_bilinear_step1)
    parallel_do(size(im_out),im_out,o_pattern,demos_bilinear_step2)    
end

% Function: gamma_curve
%
% Computation of the gamma correction curve
%
% Parameters:
% pwr - gamma power (default: 0.4)
% ts - reciprocal of the gamma power (default: 2.2)
% imax - maximum intensity `white point�
% mode - 1: forward gamma correction
%       0: backward gamma correction
%====================================================================
function C = gamma_curve(pwr=0.4, ts=2.2, imax, mode=1)
    g = zeros(6)
    bnd = [0,0]
    g[0] = pwr
    g[1] = ts
    bnd[g[1] >= 1] = 1
    if (g[1] && (g[1]-1)*(g[0]-1) <= 0)
        for i=0..47
            g[2] = (bnd[0] + bnd[1])/2
            if g[0]!=0
                bnd[(pow(g[2]/g[1],-g[0]) - 1)/g[0] - 1/g[2] > -1] = g[2]
            else
                bnd[(g[2]/exp(1-1/g[2])) < g[1]] = g[2]
            endif
        end
        g[3] = g[2] / g[1]
        if g[0]!=0
            g[4] = g[2] * (1/g[0] - 1)
        endif
    endif
    if g[0]!=0
        g[5] = 1 / (g[1]*sqrt(g[3])/2 - g[4]*(1 - g[3]) + (1 - g[3]^(1+g[0]))*(1 + g[4])/(1 + g[0])) - 1
    else     
        g[5] = 1 / (g[1]*sqrt(g[3])/2 + 1 - g[2] - g[3] - g[2]*g[3]*(log(g[3]) - 1)) - 1
    endif
    function [] = __kernel__ gamma_curve_kernel(curve : vec'unchecked, g : vec'unchecked'const, mode : int, imax : int, pos : int)
        r = float(pos)/float(imax)
        if pos < imax
            val = mode ? (r < g[3] ? r*g[1] : (g[0] ? pow(r,g[0])*(1+g[4])-g[4]    : log(r)*g[2]+1)) : _
                         (r < g[2] ? r/g[1] : (g[0] ? pow((r+g[4])/(1+g[4]),1/g[0]) : exp((r-1)/g[2])))
            curve[pos] = val * 65536
        else
            curve[pos] = 65535
        endif
    end        
    print g
    C = zeros(1,65536)
    parallel_do(numel(C),C,g,mode,imax,gamma_curve_kernel)
end

% Function: dcr_reconstruct
%
% Simple reconstruction of RAW digital camera images
%
% Parameters:
% im - the RAW image to be reconstructed
%====================================================================
function im_out = dcr_reconstruct(im)

    im_out = im.rawdata

    %==================
    % Black subtraction
    %==================
    function [] = __kernel__ subtract_black(im : cube'unchecked, black : vec4, pos : ivec2) 
        x = im[pos[0],pos[1],0..3] - black
        im[pos[0],pos[1],0..3] = max([0.,0.,0.,0.], x)
    end
    im.color.maximum = im.color.maximum - im.color.black
    parallel_do(size(im_out,0..1),im_out,im.color.black + im.color.cblack,subtract_black)

    %==============
    % Scaling
    %==============
    % 1) premultiplication and conversion to the range [0..1]
    function [] = __kernel__ scale_colors(im : cube'unchecked, scale_mul : vec4, pos : ivec2) 
        im[pos[0],pos[1],0..3] = im[pos[0],pos[1],0..3] .* scale_mul
    end
    scale_mul = im.color.pre_mul / max(im.color.pre_mul) * (1.0/im.color.maximum)
    scale_mul[3] = scale_mul[1]
    parallel_do(size(im_out,0..1),im_out,scale_mul,scale_colors)

    %=====================
    % Exposure correction
    %=====================
    exp_shift = 1
    if exp_shift != 1
        %im_out = exposure_correction(im_out, exp_shift)
    endif

    %==============
    % Demosaicking
    %==============
    im_out = 0 % release memory
    im_out = demosaic_bilinear(im.rawdata, im.color.bayer)

    %======================================
    % Color correction using rgb_cam matrix
    %======================================
    function [] = __kernel__ color_transform(im : cube'unchecked, rgb_cam : mat'unchecked'const, pos : ivec2) 
        x = im[pos[0],pos[1],0..2]
        y = [rgb_cam[0,0] * x[0] + rgb_cam[0,1] * x[1] + rgb_cam[0,2] * x[2], _
             rgb_cam[1,0] * x[0] + rgb_cam[1,1] * x[1] + rgb_cam[1,2] * x[2], _
             rgb_cam[2,0] * x[0] + rgb_cam[2,1] * x[1] + rgb_cam[2,2] * x[2]]
        im[pos[0],pos[1],0..2] = saturate(y)*65535
        if size(im,2)==4
            im[pos[0],pos[1],3] = saturate(im[pos[0],pos[1],3]) * 65535
        endif
    end
    parallel_do(size(im_out,0..1),im_out,im.color.rgb_cam,color_transform)
   
    %======================================
    % Gamma correction
    %======================================
    do_gamma_correction = false
    if do_gamma_correction
        % compute white point as the 99th percentile
        bins = 0..65535
        im_his = imhist(im_out, bins)        
        percentile=0.95*prod(size(im_out,0..1))
        ipos = find(cumsum(im_his)>percentile,"first",1)[0,1]
        t_white = max(8192,bins[ipos])
        print "white point: ", t_white

        gamm = [0.4,2.2] % gamma curve
        curve = gamma_curve(gamm[0],gamm[1],t_white)

        % apply the gamma curve
        function [] = __kernel__ apply_gamma(im : cube'unchecked, curve : vec'unchecked, pos : ivec2)
            im[pos[0],pos[1],0] = curve[int(im[pos[0],pos[1],0])]
            im[pos[0],pos[1],1] = curve[int(im[pos[0],pos[1],1])]
            im[pos[0],pos[1],2] = curve[int(im[pos[0],pos[1],2])]
        end
        parallel_do(size(im_out,0..1),im_out,curve,apply_gamma)
    endif

    im_out *= 1.0/256  % normalize to the range 0-255 (input range = 0-65535)
end

function [] = main()

    % Enter a .Nef file here
    filename = uigetfile("RAW camera images (*.nef;*.raw;*.cr2;*.rw2)|*.nef;*.raw;*.cr2;*.rw2", 
                         "Select a RAW input image")    
    if numel(filename)==0
        error "Please select an input file!"
    endif
    im = readraw(filename)

    print "Camera model: ", im.idata.make, " ", im.idata.model
    print "Aperture: ",im.other.aperture,", focal length: ",im.other.focal_len, ", ISO: ",im.other.iso_speed, ", shutter speed: ", im.other.shutter
    print "Image dimensions: ", size(im.rawdata,0..1)*2

    print im.color.bayer
	
    tic()
    im = dcr_reconstruct(im)
    toc()

    imshow(im);title("Reconstructed Digital Image")
end