// This sample demonstrates the interoperability of Quasar with CUDA, so
// that existing CUDA applications can use routines programmed in CUDA.
// Everything is designed in such a way that transfers between CPU and GPU
// are avoided.
#include "quasar_dsl.h"
#include "cuda.h"

static void checkResult(CUresult result)
{
	if (result != CUDA_SUCCESS)
	{
		tprintf(_T("CUDA operation failed with error code %d\n"), result);
		exit(-1);
	}
}

void sample11(void)
{
	using namespace quasar;
	using quasar::ref;

	try
	{
		ref<IQuasarHost> host = IQuasarHost::Create(_T("cuda0.device.xml"), false);

		Function linspace(_T("linspace(scalar,scalar,scalar)"));
		Function ones(_T("ones()"));
		Function transpose(_T("transpose()"));
		Function imshow(_T("imshow(cube)"));
		const int N = 512;

		QValue im = ones(N,1) * linspace(0,255,N);

		// Obtain the current CUDA context & device name
		CUcontext ctx;
		checkResult(cuCtxGetCurrent(&ctx));
		tprintf(_T("CUDA context: %x\n"), ctx);

		CUdevice device;
		checkResult(cuCtxGetDevice(&device));

		char deviceName[80];
		checkResult(cuDeviceGetName(deviceName, sizeof(deviceName), device));
		printf("CUDA device name: %s\n", deviceName);

		// Now we attempt to obtain the device pointer for im
		CUdeviceptr devicePtr;
		LockResults result = host->Lock(im, TYPE_SCALAR, LOCK_READ, MEMRESOURCE_SINGLE_CUDA, (void**) &devicePtr);

		switch (result)
		{
		// Called in case we want to obtain a CUDA pointer when running Quasar in
		// CPU mode.
		case LOCKRESULT_RES_NOT_AVAILABLE:
			wprintf(_T("Lock failed - the requested resource is not available\n"));
			break;
		case LOCKRESULT_OUT_OF_MEM:
			wprintf(_T("Lock failed - insufficient memory resources\n"));
			break;
		case LOCKRESULT_INVALID:
			wprintf(_T("Lock failed - invalid request\n"));
			break;
		}

		tprintf(_T("CUDA device pointer: %x\n"), devicePtr);

		//======================================================================
		// 1. Accessing Quasar matrices from CUDA
		//======================================================================
		// To check whether the device pointer is valid, we copy a number of values to
		// system memory
		scalar vals[8];
		checkResult(cuMemcpyDtoH(vals, devicePtr, sizeof(scalar)*8));
		for (int i = 0; i < 8; i ++)
			tprintf(_T("%g "), vals[i]);
		tprintf(_T("\n"));

		// Always need to unlock!!!
		host->Unlock(im, LOCK_READ, MEMRESOURCE_SINGLE_CUDA);

		//======================================================================
		// 2. Accessing CUDA data from Quasar
		//======================================================================
		const int P = 256;
		const int Q = 256;
		scalar data[P*Q];
		for (int m = 0; m < P; m++)
			for (int n = 0; n < Q; n++)
			{
				data[m*Q+n]=cos(4*m*n*M_PI/(P*Q));
			}

		CUdeviceptr devicePtr2;
		checkResult(cuMemAlloc(&devicePtr2, sizeof(scalar) * P*Q));
		checkResult(cuMemcpyHtoD(devicePtr2, data, sizeof(scalar) * P*Q));

		// Now, we want to construct a Quasar matrix with the specified device
		// pointer.
		QValue im2 = host->CreateMatrix<scalar>(P,Q);
		host->Lock(im2, TYPE_SCALAR, LOCK_WRITE, MEMRESOURCE_SINGLE_CUDA, (void**) &devicePtr);
		checkResult(cuMemcpyDtoD(devicePtr, devicePtr2, sizeof(scalar) * P*Q));
		//checkResult(cuMemcpyHtoD(devicePtr, data, sizeof(scalar) * P*Q));
		host->Unlock(im2, LOCK_WRITE, MEMRESOURCE_SINGLE_CUDA);

		checkResult(cuMemFree(devicePtr2));

		imshow(im);
		imshow(im2);

		host->RunApp();
	}
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
}
