// This sample demonstrates the interoperability of Quasar with OpenCL, so
// that existing OpenCL applications can use routines programmed in OpenCL.
// Everything is designed in such a way that transfers between CPU and GPU
// are avoided.
#include "quasar_dsl.h"
#include "CL/OpenCL.h"

static void checkResult(cl_int result)
{
	if (result != CL_SUCCESS)
	{
		wprintf(_T("OpenCL operation failed with error code %d\n"), result);
		exit(-1);
	}
}

void sample12(void)
{
	using namespace quasar;
	using quasar::ref;

	try
		{

		// Need a Hyperion device configuration file for OpenCL
		ref<IQuasarHost> host = IQuasarHost::Create(_T("opencl0.device.xml"), false);

		Function linspace(_T("linspace(scalar,scalar,scalar)"));
		Function ones(_T("ones()"));
		Function transpose(_T("transpose()"));
		Function imshow(_T("imshow(cube)"));
		const int N = 512;

		QValue im = ones(N,1) * linspace(0,255,N);

		// Obtain access to the OpenCL context & command queue
		cl_context context;
		cl_command_queue cmdQueue;
		host->QueryProperty(OPENCL_CURRENT_CONTEXT, 0, (void**)&context, sizeof(context));
		host->QueryProperty(OPENCL_COMMANDQUEUE, 0, (void**)&cmdQueue, sizeof(cmdQueue));

		tprintf(_T("OpenCL context: %x\n"), context);
		tprintf(_T("OpenCL command queue: %x\n"), cmdQueue);

		// Now we attempt to obtain the device pointer for im
		cl_mem devicePtr;
		LockResults result = host->Lock(im, TYPE_SCALAR, LOCK_READ, MEMRESOURCE_SINGLE_OPENCL, (void**) &devicePtr);

		switch (result)
		{
		// Called in case we want to obtain a OpenCL pointer when running Quasar in
		// CPU mode.
		case LOCKRESULT_RES_NOT_AVAILABLE:
			tprintf(_T("Lock failed - the requested resource is not available\n"));
			break;
		case LOCKRESULT_OUT_OF_MEM:
			tprintf(_T("Lock failed - insufficient memory resources\n"));
			break;
		case LOCKRESULT_INVALID:
			tprintf(_T("Lock failed - invalid request\n"));
			break;
		}

		tprintf(_T("OpenCL device pointer: %x\n"), devicePtr);

		//======================================================================
		// 1. Accessing Quasar matrices from OpenCL
		//======================================================================
		// To check whether the device pointer is valid, we copy a number of values to
		// system memory
		scalar vals[8];
		checkResult(clEnqueueReadBuffer(cmdQueue, devicePtr, true, 0, sizeof(scalar)*8, vals, 0, NULL, NULL));

		for (int i = 0; i < 8; i ++)
			tprintf(_T("%g "), vals[i]);
		tprintf(_T("\n"));

		// Always need to unlock!!!
		host->Unlock(im, LOCK_READ, MEMRESOURCE_SINGLE_OPENCL);

		//======================================================================
		// 2. Accessing OpenCL data from Quasar
		//======================================================================
		const int P = 256;
		const int Q = 256;
		scalar data[P*Q];
		for (int m = 0; m < P; m++)
			for (int n = 0; n < Q; n++)
			{
				data[m*Q+n]=cos(4*m*n*M_PI/(P*Q));
			}

		cl_int errCode;
		cl_mem devicePtr2 = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
			sizeof(scalar) * P * Q, data, &errCode);
		checkResult(errCode);


		// Now, we want to construct a Quasar matrix with the specified device
		// pointer.
		QValue im2 = host->CreateMatrix<scalar>(P,Q);
		host->Lock(im2, TYPE_SCALAR, LOCK_WRITE, MEMRESOURCE_SINGLE_OPENCL, (void**) &devicePtr);
		checkResult(clEnqueueCopyBuffer(cmdQueue, devicePtr2, devicePtr, 0, 0, sizeof(scalar) * P*Q, 0, NULL, NULL));
		host->Unlock(im2, LOCK_WRITE, MEMRESOURCE_SINGLE_OPENCL);

		checkResult(clReleaseMemObject(devicePtr2));

		imshow(im);
		imshow(im2);

		host->RunApp();
	}
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
}
