//
// Sample demonstrates how native pointers can be wrapped so that they
// can be passed to Quasar functions.
// (C) Gepura 2019-2020
//
#include "quasar_dsl.h"
#include "quasar_dsl_class_ext.h"


using namespace quasar;

static Type nativePointerType;

// Note: needs to implement IUnknown, so that is why we derive from IUnknown
class NativePointer : public ObjBase
{
	void *ptr;

public:
	NativePointer(void *ptr) :ObjBase((qvalue_t &)nativePointerType), ptr(ptr) 
	{
	}

public:
	// Class must have a wrap() method, because some of the macros
	// in quasar_dsl_class_ext.h depend on it. The wrap function wraps
	// the object in a qvalue_t object.
	qvalue_t wrap()
	{
		qvalue_t qv;
		qv.type = TYPE_UNTYPEDOBJECT;
		qv.private_obj = this;
		return qv;
	}

	// Allow casting to a pointer value
	void * ptr_val() const { return ptr; }

	static NativePointer *from_qvalue(const qvalue_t &qv)
	{
		return static_cast<NativePointer *>(qv.private_obj);
	}

	virtual ~NativePointer()
	{
		tprintf(_T("NativePointer destructor called\n"));
	}

	static void register_type()
	{
		// Define the type so that it can be used from Quasar. Note that using this
		// mechanism, objects of the type are subject to automatic memory management
		// e.g. transfer between CPU and GPU etc. The objects can also be inspected
		// in the debugger, serialized to file, ...
		BEGIN_ABSTRACT_TYPE(builder, NativePointer, _T("native_pointer"), TYPEFLAGS_DYNAMIC_CLASS);
		nativePointerType = builder.CreateType();
	}
	static void unregister_type()
	{
		nativePointerType = Type();
	}
};

// Needs to be defined within the Quasar namespace
namespace quasar
{
	DECLARE_STRUCT_TYPE(NativePointer); // Generates type information for our struct
	IMPLEMENT_TYPE(TYPE_UNTYPEDOBJECT, NativePointer, _T("native_pointer"))
}

void sample22(void)
{
	using quasar::ref;

	try
	{
		ref<IQuasarHost> host = IQuasarHost::Create(_T("cpu"));
		
		// We need to register the type before we can use it
		NativePointer::register_type();

		// Create a typed object.
		QValue obj = (new NativePointer((void*)0x10203040))->wrap();

		Function print(_T("print(...)"));
		Function type_(_T("type(...)"));

		// Check the type
		Type obj_type = type_(obj);

		if (host->TypeCompareToSpecification(obj_type, nativePointerType))
		{
			print(obj);
			print(obj_type);

			// Now we are sure about the type, we extract the pointer
			tprintf(_T("Pointer: %p\n"), NativePointer::from_qvalue(obj)->ptr_val());
		}
		else
		{
			tprintf(_T("Something went wrong and our object has the wrong type..."));
		}

		// Release the type before the "host" object goes out of scope
		NativePointer::unregister_type();
	}
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
}
