//
// Sample demonstrates how native pointers can be wrapped so that they
// can be passed to Quasar functions.
// (C) Gepura 2019
//
#include "quasar_dsl.h"
#include "quasar_dsl_class_ext.h"

using namespace quasar;

static Type nativeClassType;

// Note: needs to implement IUnknown, so that is why we derive from IUnknown
class NativeClass : public ObjBase
{
public:
	int intField;
	QValue stringField;

public:
	NativeClass() : ObjBase((qvalue_t &)nativeClassType)
	{
		intField = 1;
		stringField = _T("Sample string field");
	}

public:
	void do_something()
	{
		string_t string_val;
		IQuasarHost::GetInstance()->GetString(stringField, string_val);
		tprintf(_T("The value of intField is %d!\n"), intField);
		tprintf(_T("The value of stringField is %ls!\n"), (LPCTSTR)string_val);
	}

	static NativeClass *from_qvalue(const qvalue_t &qv)
	{
		_ASSERT(((QValue)qv).HasType(nativeClassType));
		return static_cast<NativeClass *>(qv.private_obj);
	}

	static void register_type()
	{
		// Define the type so that it can be used from Quasar. Note that using this
		// mechanism, objects of the type are subject to automatic memory management
		// e.g. transfer between CPU and GPU etc. The objects can also be inspected
		// in the debugger, serialized to file, ...
		BEGIN_ABSTRACT_TYPE(builder, NativeClass, _T("native_class"), TYPEFLAGS_DYNAMIC_CLASS);
		ADD_FIELD(builder, intField);
		ADD_FIELD(builder, stringField);
		ADD_METHOD(builder, do_something);
		nativeClassType = builder.CreateType();
	}

	static void unregister_type()
	{
		nativeClassType = Type();
	}

	virtual ~NativeClass()
	{
		tprintf(_T("NativeClass destructor called\n"));
	}
};

// Needs to be defined within the Quasar namespace
namespace quasar
{
	DECLARE_STRUCT_TYPE(NativeClass); // Generates type information for our struct
	IMPLEMENT_TYPE(TYPE_UNTYPEDOBJECT, NativeClass, _T("native_class"))
}

void sample23(void)
{
	using quasar::ref;

	try
	{
		ref<IQuasarHost> host = IQuasarHost::Create(_T("cpu"));

		NativeClass::register_type();

		// Create a typed object.
		QValue obj = (new NativeClass)->wrap();

		
		Function print(_T("print(...)"));
		Function type_(_T("type(...)"));
		Function refcount(_T("$refcount(...)"));
		print(obj);
		print(type_(obj));
		print(refcount(obj));

		// Call a method on the object
		Method(nativeClassType, _T("do_something()"))(obj);

		// Dynamic access to a field
		obj.SetField(_T("intField"), 12345);

		print(obj.GetField(_T("intField")));
		tprintf(_T("Direct getter: %d\n"), NativeClass::from_qvalue(obj)->intField);

		print(refcount(obj));

		// Unregister the type before the "host" object goes out of scope
		NativeClass::unregister_type();
	}
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
}
