// Note : quasar_dsl.h should be placed first!
#include "quasar_dsl.h"
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/legacy/legacy.hpp>
#include <opencv2/highgui/highgui.hpp>

using namespace cv;
using namespace quasar;

// Converts an image to the Quasar format. Note that
// the function makes a copy of the matrix.
QValue ConvertImage(IplImage *img)
{
    IQuasarHost *host = IQuasarHost::GetInstance();
    QValue qv =
        host->CreateCube<uint8_t>(img->height, img->width, img->nChannels);
    CubeBase<uint8_t> mtx = host->LockCube<uint8_t>(qv, LOCK_WRITE);
    int samplesPerRow = mtx.dim2 * mtx.dim3;

    if (strcmp(img->channelSeq, "RGB") == 0 ||
        strncmp(img->channelSeq, "GRAY", 4) == 0)
    {
        for (int m = 0; m < mtx.dim1; m++)
        {
            memcpy(&mtx.data[m * samplesPerRow],
                   &img->imageData[m * img->widthStep], samplesPerRow);
        }
    }
    else if (strcmp(img->channelSeq, "BGR") == 0)
    {
        int C = img->nChannels;
        // Need to convert from BGR to RGB
        for (int m = 0; m < mtx.dim1; m++)
        {
            uint8_t *dstPtr = &mtx.data[m * samplesPerRow];
            const uint8_t *srcPtr =
                (const uint8_t *)&img->imageData[m * img->widthStep];

            for (int n = 0; n < mtx.dim2; n++)
            {
                dstPtr[n * C + 0] = srcPtr[n * C + 2];
                dstPtr[n * C + 1] = srcPtr[n * C + 1];
                dstPtr[n * C + 2] = srcPtr[n * C + 0];
            }
        }
    }
    else
    {
        // Unsupported image format
        _ASSERT(false);
    }

    host->UnlockCube(qv, LOCK_WRITE);
    return qv;
}

void sample4(void)
{
    using quasar::ref;

    try
    {
		ref<IQuasarHost> host = IQuasarHost::Create(_T("cuda"));

        IplImage *img = cvLoadImage("c:\\temp\\crowd.jpg");
        IplImage *img_8uc1 = cvCreateImage(cvGetSize(img), 8, 1);
        IplImage *img_edge = cvCreateImage(cvGetSize(img), 8, 1);
        IplImage *img_8uc3 = cvCreateImage(cvGetSize(img), 8, 3);

        cvCvtColor(img, img_8uc1, CV_BGR2GRAY);

        cvThreshold(img_8uc1, img_edge, 128, 255, CV_THRESH_BINARY);

        CvMemStorage *storage = cvCreateMemStorage();
        CvSeq *first_contour = NULL;

        int Nc = cvFindContours(img_edge, storage, &first_contour,
                                sizeof(CvContour), CV_RETR_LIST);

        int n = 0;
        wprintf(_T("Total Contours Detected: %d\n"), Nc);
        CvScalar red = CV_RGB(250, 0, 0);
        CvScalar blue = CV_RGB(0, 0, 250);

        for (CvSeq *c = first_contour; c != NULL; c = c->h_next)
        {
            cvCvtColor(img_8uc1, img_8uc3, CV_GRAY2BGR);
            cvDrawContours(img_8uc3, c,
                           red,  // Red
                           blue, // Blue
                           1,    // Vary max_level and compare results
                           2, 8);
            break;
        }

        // ======== Transition to Quasar =======
        QValue qv = ConvertImage(img_8uc3);

        Function imshow(_T("imshow(cube)"));
        imshow(qv);


        host->RunApp();
        cvReleaseImage(&img_8uc1);
        cvReleaseImage(&img_8uc3);
        cvReleaseImage(&img_edge);
    }
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
    catch (cv::Exception ex)
    {
        std::cout << ex.msg << std::endl;
    }
}
