// Note : quasar_dsl.h should be placed first!
#include "quasar_dsl.h"
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/legacy/legacy.hpp>
#include <opencv2/objdetect/objdetect.hpp>
#include <opencv2/highgui/highgui.hpp>

using namespace cv;
using namespace quasar;

static CascadeClassifier face_cascade;

Mat convert_image(const qvalue_t &qv)
{
    IQuasarHost *host = IQuasarHost::GetInstance();
    int C = qv.dim3;

    Mat img(qv.dim1, qv.dim2, C == 3 ? CV_8UC3 : CV_8UC1);

    CubeBase<uint8_t> data = host->LockCube<uint8_t>(qv, LOCK_READ);

    // Need to convert from BGR to RGB
    for (int m = 0; m < img.rows; m++)
    {
        // memcpy(&img.data[m * img.step.p[0]],
        //	   &data.data[m * qv.cubeVal.dim2 * ch],
        //	   qv.cubeVal.dim2 * ch);
        uint8_t *dstPtr = &img.data[m * img.step.p[0]];
        const uint8_t *srcPtr = (const uint8_t *)&data.data[m * qv.dim2 * C];

        for (int n = 0; n < img.cols; n++)
        {
            dstPtr[n * C + 0] = srcPtr[n * C + 2];
            dstPtr[n * C + 1] = srcPtr[n * C + 1];
            dstPtr[n * C + 2] = srcPtr[n * C + 0];
        }
    }

    host->UnlockCube(qv, LOCK_READ);
    return img;
}

void facedetection(void *closure,
				   const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
{
    IQuasarHost *host = IQuasarHost::GetInstance();
    Mat captureFrame = convert_image(argsIn[0]);
    Mat grayscaleFrame;

    // convert captured image to gray scale and equalize
    cvtColor(captureFrame, grayscaleFrame, CV_BGR2GRAY);
    equalizeHist(grayscaleFrame, grayscaleFrame);

    // create a vector array to store the face found
    std::vector<Rect> faces;

    // find faces and store them in the vector array
    // face_cascade.detectMultiScale(grayscaleFrame, faces, 1.1, 3,
    // CV_HAAR_FIND_BIGGEST_OBJECT|CV_HAAR_SCALE_IMAGE, Size(30,30));
    face_cascade.detectMultiScale(grayscaleFrame, faces, 1.1, 3,
                                  CV_HAAR_SCALE_IMAGE, Size(150, 150));

    qvalue_t qv = host->CreateMatrix<scalar>((int)faces.size(), 4);
    MatrixBase<scalar> mtx = host->LockMatrix<scalar>(qv, LOCK_WRITE);

    for (int i = 0; i < faces.size(); i++)
    {
        mtx.data[4 * i + 0] = faces[i].x;
        mtx.data[4 * i + 1] = faces[i].y;
        mtx.data[4 * i + 2] = faces[i].x + faces[i].width;
        mtx.data[4 * i + 3] = faces[i].y + faces[i].height;
    }

    host->UnlockMatrix(qv, LOCK_WRITE);
    argsOut[0] = qv;
}

void sample5(void)
{
    // Mat captureFrame(cvLoadImage( "c:\\temp\\crowd.jpg" ));
    // Mat grayscaleFrame;

    // Initialize OpenCV face detection
    char xmlFile[_MAX_PATH];
    size_t retSize;
    getenv_s(&retSize, xmlFile, "OPENCV_DIR");
    strcat_s(
        xmlFile,
        "/../../../sources/data/haarcascades/haarcascade_frontalface_alt2.xml");

    // Create the cascade classifier object used for the face detection
    face_cascade.load(xmlFile);

	try
	{
		// Initialize Quasar
		using quasar::ref;
		ref<IQuasarHost> host = IQuasarHost::Create(_T("cuda"));
		// host->EnableProfiling(PROFILE_MEMLEAKS); // Mem-leaks will be printed
		// when the program stops
		// host->EnableProfiling(PROFILE_EXECUTIONTIME);

		// Register our function
		Type lambdaType(_T("[cube[uint8]->mat]"));
		Function lambda = Function::CreateLambda(lambdaType, NULL, facedetection);
		host->WriteVariable(_T("facedetection"), lambda);

		// Load the Quasar program
		wprintf(_T("Compiling code...\n"));
		LPCTSTR errorMsg = NULL;
		if (!host->LoadSourceModule(
				_T("../../Quasar/CppHost/Samples/sample5_opencv_facedetection.q"),
				&errorMsg))
		{
			tprintf(_T("Error while compiling program:\n%s\n"), errorMsg);
			return;
		}

		// Call the main function
		Function run(_T("run()"));
		run();

		host->RunApp();
	}
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
}
