import "Quasar.Video.dll"
import "Quasar.UI.dll"
import "inttypes.q"

% Function will be externally defined in the C++ program. However
% we need to declare it here so that the Quasar compiler knows
% about it.
facedetection : [cube[uint8] -> cube]

type vertexnormal : class
    position : vec3
    normal : vec3
end

% Draws a cone on the given 3D vector layer. The cone has a bottom plane with
% center [0,0,0], height 'height' and N tracks.
function [] = draw_cone(layer : qvectorlayer3d, radius, height, N)

    theta = linspace(0,2*pi,N)
    p = vec[vertexnormal](N+1)
    p[0] = vertexnormal(position:=[0,0,height], normal:=[0,0,1])    
    for i=0..N-1
        p[1+i] = vertexnormal(position:=[radius*cos(theta[i]),radius*sin(theta[i]),0.0],
                              normal:=[cos(theta[i]),sin(theta[i]),0])        
    end    
    layer.filltrianglefan(p)

end

% Draws a cube with dimensions [0,0,0]-[1,1,1] onto the given 3D vector layer.
function [] = draw_cube(layer : qvectorlayer3d)
    indices = `0,1,2,0,2,3, % Front
               5,4,7,5,7,6, % Back
               4,0,3,4,3,7, % Left
               1,5,6,1,6,2, % Right
               3,2,6,3,6,7, % Top
               4,5,1,4,1,0' % Bottom
    layer.translatetransform([-0.5,-0.5,0])
    p = vec[vertexnormal](8)
    p[0] = vertexnormal(position:=[0,0,1],normal:=[0,0,1])
    p[1] = vertexnormal(position:=[1,0,1],normal:=[1,0,1])
    p[2] = vertexnormal(position:=[1,1,1],normal:=[1,1,1])
    p[3] = vertexnormal(position:=[0,1,1],normal:=[0,1,1])
    p[4] = vertexnormal(position:=[0,0,0],normal:=[0,0,0])
    p[5] = vertexnormal(position:=[1,0,0],normal:=[1,0,0])
    p[6] = vertexnormal(position:=[1,1,0],normal:=[1,1,0])
    p[7] = vertexnormal(position:=[0,1,0],normal:=[0,1,0])
    layer.fillindexedtriangles(p, indices)
    layer.translatetransform([0.5,0.5,0])
end

% Draws a sphere with radius 1 onto the given 3D vector layer.
% Use a sufficient number of tracks and slices (e.g. 20) to make
% sure that the triangle mesh approximates the sphere.
function [] = draw_sphere(layer,num_tracks : int,num_slices : int)

    z = linspace(-1,1,num_slices)
    theta = linspace(0,2*pi,num_tracks)
    
    norm = __device__ (x : vec3) -> dotprod(x,x)

    #pragma loop_parallelizer(off)
    for i=0..num_slices-2
        R1=sqrt(1-z[i]^2)
        R2=sqrt(1-z[i+1]^2)
        p = mat[vertexnormal](num_tracks,2)
        for j=0..num_tracks-1
            q1 = [R1*cos(theta[j]),R1*sin(theta[j]),z[i]]
            q2 = [R2*cos(theta[j]),R2*sin(theta[j]),z[i+1]]
            p[j,0] = vertexnormal(position:=q1, normal:=q1/sqrt(norm(q1)))
            p[j,1] = vertexnormal(position:=q2, normal:=q2/sqrt(norm(q2)))
        end
        layer.setcolor("orange")
        layer.filltrianglestrip(p)
        layer.setcolor("gray")
        layer.drawlines(p)
    end

end

function [] = draw_stilllife(gl : qgldisplay, layer : qvectorlayer3d)

    % Setup the scene lights
    light = gl.lights[0]
    light.position = [2,2,2]
    light.ambientcolor = [0.2,0.2,0.2,1]
    light.diffusecolor = [1,1,1,1]
    light.specularcolor = [1,1,1,1]
    light.enabled = true
    layer.enable_lighting = true       
    
    % bottom plane
    layer.setcolor("purple")
	layer.translatetransform([0,0,-0.5])
    layer.drawbox([-0.5,-0.5,0],[0.5,0.5,0.5])
    layer.fillrect([0,0,0],[1,0,0],[0,1,0])
    
    layer.setcolor("red")
    layer.pushtransform()
    layer.translatetransform([0.2,0.2,0])
    draw_cone(layer, 0.2, 0.6, 20)
    layer.poptransform()

    layer.setcolor("yellow")
    layer.pushtransform()
    layer.translatetransform([-0.2,-0.2,0])
    layer.scaletransform(0.3*[1,1,1])
    draw_cube(layer)
    layer.poptransform()

    layer.pushtransform()
    layer.scaletransform(0.2*[1,1,1])
    layer.translatetransform([1,-1,1])
    draw_sphere(layer,20,20)
    layer.poptransform()
    layer.enable_lighting = false

end


function [] = run()
    cams = vidcamlist()
    if numel(cams)==0
        error "Unfortunately, we could not find a webcam in your system."
    endif
    % select the first webcam
    cam = cams[0]

	stream = vidopen(sprintf("cam:video=""%s"",video_size=640x480,frame_rate=30",cam)) % Opens the specified video file for playing
    vidstate = object()
    [vidstate.is_playing, vidstate.allow_seeking, vidstate.show_next_frame] = [true, true, true]
    print "Video stream information: ",stream

	frm = form(sprintf("Real-time face detection - %s", cam))
	disp = frm.add_display()
    [disp.width, disp.height] = [1536, 1024]
	frm.center()
	frm.show()
	[frm.width, frm.height] = [stream.frame_width + 80, stream.frame_height + 120]

    layer = new(qvectorlayer3d)            
    renderer = disp.create_opengl_renderer()
    renderer.background_color = "black"
    renderer.enable_zbuffer = true
    renderer.draw_backfaces = true 
    renderer.show_coords = false       
    renderer.pitch = 340

    draw_stilllife(renderer, layer)
    
    
	while !frm.closed()
        if vidstate.is_playing || vidstate.show_next_frame            
            if vidstate.is_playing ? !vidreadframe(stream) : false
                vidseek(stream,0) % Play in a loop
            endif   
			A = facedetection(stream.rgb_data)    
			
			if size(A,0)>0
				renderer.roll = -90*asin(0.5*(A[0]+A[2])/size(stream.rgb_data,1))
				renderer.pitch = 280+0.25*(A[1]+A[3]-0.5*size(stream.rgb_data,0))
			endif
			     
			layer2 = new(qvectorlayer)
			layer2.translatetransform([10, 10])
			layer2.scaletransform([0.5,0.5])
            layer2.drawimage([0,0], float(stream.rgb_data)) 
			layer2.setpencolor([1,0,0,0])
			layer2.drawrect([0,0],size(stream.rgb_data,[1,0]))
			layer2.setpencolor([0,1,0,0])
			for i=0..size(A,0)-1
				layer2.drawrect(A[i,0..1],A[i,2..3])
			end

			renderer.clear()
			renderer.add(layer, "layer")
			renderer.add(layer2, "layer")
                    	
            if frm.closed()
                break
            endif

            vidstate.allow_seeking = false
            vidstate.allow_seeking = true
            vidstate.show_next_frame = false
        endif 
        pause(50)
    end
end