#include "quasar_dsl.h"

using namespace quasar;

struct point
{
    int x;
    int y;
};

namespace quasar
{
    DECLARE_STRUCT_TYPE(point); // Generates type information for our struct
    IMPLEMENT_TYPE(TYPE_TYPEDOBJECT, point, _T("point"))
}

void sample7(void)
{
	using quasar::ref;

	try
	{
		ref<IQuasarHost> host = IQuasarHost::Create(_T("cpu"));
		Function object(_T("object()"));
		Function print(_T("print()"));

		// Create an untyped (Python-like) object, that has properties that can be
		// assigned at run-time.
		QValue obj = object();
		obj.SetField(_T("x"), 2.5);
		obj.Field(_T("x")) += 2.5;
		obj.Field(_T("x")).InplaceMin(1.5);
		print(obj);

		// Define the type so that it can be used from Quasar. Note that using this
		// mechanism, objects of the type are subject to automatic memory management
		// e.g. transfer between CPU and GPU etc. The objects can also be inspected
		// in the debugger, serialized to file, ...
		TypeBuilder builder(_T("sample7"), _T("point"), TYPEFLAGS_MUTABLE_CLASS);
		builder.AddField(_T("x"), Type(_T("int")));
		builder.AddField(_T("y"), Type(_T("int")));
		Type type = builder.CreateType();

		// Create a typed object.
		QValue obj2 = host->CreateTypedObject(type);
		// The "slow" way of setting properties
		obj2.SetField(_T("x"), 4.5);
		obj2.SetField(_T("y"), 7.5);

		// A more efficient way for directly getting access to the fields
		point *p = host->LockObject<point>(obj2);
		tprintf(_T("p.x = %d\n"), p->x);
		tprintf(_T("p.y = %d\n"), p->y);
		host->UnlockObject(obj2);
		print(obj2);

		// Create a typed object - a one-line approach
		NamedParam params[] = { NamedParam(_T("x"), 4.5), NamedParam(_T("y"), 7.5) };
		QValue obj3 = QValue::CreateObject(type, params);
		print(obj3);

		// Create a generic type
		QValue mtx = QValue::CreateVector<point>(4);
		VectorBase<point> mtxVals = host->LockVector<point>(mtx);
		for (int i = 0; i < 4; i++)
		{
			mtxVals.data[i].x = i;
			mtxVals.data[i].y = 4 - i;
		}
		host->UnlockVector(mtx);

		// TODO-
		//((point *)mtx(0, BoundaryAccessMode::Default))->x = 1;

		print(mtx);
	}
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
}
