// Comparison of the Quasar automatic vectorizer vs. straightforward
// C++ code.
#include "quasar_dsl.h"

void cpp_mean_filter(quasar::Matrix x, quasar::Matrix y)
{
    using namespace quasar;
    for (int m = 0; m < x.dim1; m++)
        for (int n = 0; n < x.dim2; n++)
        {
            /*y.data[m * y.dim2 + n] =
                    (x.data[m * x.dim2 + n-2] +
                     x.data[m * x.dim2 + n-1] +
                     x.data[m * x.dim2 + n] +
                     x.data[m * x.dim2 + n+1] +
                     x.data[m * x.dim2 + n+2])/5.0;*/
            matrix_set_at<Unchecked, scalar>(
                y, make_int2(m, n),
                (matrix_get_at<Unchecked, scalar>(x, make_int2(m, n - 2)) +
                 matrix_get_at<Unchecked, scalar>(x, make_int2(m, n - 1)) +
                 matrix_get_at<Unchecked, scalar>(x, make_int2(m, n)) +
                 matrix_get_at<Unchecked, scalar>(x, make_int2(m, n + 1)) +
                 matrix_get_at<Unchecked, scalar>(x, make_int2(m, n + 2))) /
                    5.0);
        }
}

void sample8(void)
{
    using namespace quasar;

    // Inline definition of a Quasar program, similar to OpenGL/DirectX shader
    // language
    LPCTSTR sourceProgram =
        _T("function [] = mean_filter(im : mat, im_out : mat'unchecked)\n")
        _T("	  #pragma force_parallel\n")
        _T("    for m=0..size(im,0)-1\n")
        _T("        for n=0..size(im,1)-1\n")
        _T("            {! loop_tiling dims=[1,8]; mode=\"local\"}\n")
        _T("            r = 0\n")
        _T("            for x=-2..2\n")
        _T("                r += im[m,n+x]\n")
        _T("            end\n")
        _T("            im_out[m,n] = r/5.0\n")
        _T("        end\n")
        _T("    end\n")
        _T("end");

    LPCTSTR errorMsg = NULL;

    try
    {
		quasar::ref<IQuasarHost> host = IQuasarHost::Create(_T("cpu"));

		// Export Quasar functions to C++
		Function tic(_T("tic()"));
		Function toc(_T("toc()"));
		Function imread(_T("imread(string)"));
		Function imshow(_T("imshow(cube)"));

		tprintf(_T("Compiling code...\n"));
		tic();
		if (!host->LoadModuleFromSource(_T("sample8"), sourceProgram, &errorMsg))
		{
			tprintf(_T("Error while compiling program:\n%s\n"), errorMsg);
			return;
		}
		toc();

		// The function mean_filter is now available
		Function mean_filter(_T("mean_filter(mat,mat)"));


		// Load an image
		QValue img = imread(_T("lena_big.tif"));
		tprintf(_T("The image has dimensions %dx%dx%d\n"), size(img, 0), size(img, 1),
				size(img, 2));

		// img = img[:,:,1]
		img = img(Range(), Range(), QValue(1));

		QValue img_out = QValue::CreateMatrix<scalar>(size(img, 0), size(img, 1));

		tic();
		for (int i = 0; i < 999; i++)
		{
			Matrix img_mtx = host->LockMatrix<scalar>(img);
			Matrix img_out_mtx = host->LockMatrix<scalar>(img_out);
			cpp_mean_filter(img_mtx, img_out_mtx);
			host->UnlockMatrix(img_out);
			host->UnlockMatrix(img);
		}
		toc(_T("C++ based filter"));

		tic();
		for (int i = 0; i < 999; i++)
		{
			mean_filter(img, img_out);
		}
		toc(_T("Quasar filter"));


		imshow(img_out);

		/*	scalar fcoeff[] = { 1.0/9, 2.0/9, 3.0/9, 2.0/9, 1.0/9 };
				const int N = 2;
				QValue img_out = gaussian_filter(img, fcoeff, N);

				// Show the result
				QValue p1 = imshow(img);
				QValue p2 = imshow(img_out);

				Type qimshow = type(p1);
				Method connect(qimshow, _T("connect(??)");
				connect(p1, p2);  // Call the function p1.connect(p2)!
				*/
		host->RunApp();
	}
	catch (exception_t ex)
	{
		tprintf(_T("An error occurred:\nSource: %s\nMessage: %s\nStack trace: %s"),
			ex.source.get_buf(),
			ex.message.get_buf(),
			ex.stack_trace.get_buf());
	}
}
