//
// cpp_filter3x3.cu
// 
// Example CUDA module that can be used from Quasar. The module defines
// the kernel "filter3x3", for two different targets: CUDA and CPU.
// The kernel function parameters are parsed automatically by Quasar -
// as long as primitive datatypes or datatypes from the quasar namespace 
// are used, the kernel function can be mapped to Quasar.
//
// Note: definition of the closure parameter is mandatory, although 
// currently, the closure variables cannot be bound to Quasar variables.
//
// (c) 2013 Bart Goossens
//
#include "quasar.h"

using namespace quasar;

#ifdef TARGET_CUDA

	/* A 3x3 separable filter, using shared memory*/
	extern "C" __global__ void filter3x3(void* closure, Cube x, Cube y)
	{
		int3 pos = make_int3(blockIdx.y * blockDim.y + threadIdx.y, blockIdx.x *
            blockDim.x + threadIdx.x, blockIdx.z * blockDim.z + threadIdx.z);
		int3 blkpos = make_int3(threadIdx.y, threadIdx.x, threadIdx.z);
		int3 blkdim = make_int3(blockDim.y, blockDim.x, blockDim.z);

		/* Allocate temporary shared memory */
		shmem _shmem;
		shmem_init(&_shmem);
		Cube tmp = shmem_alloc<scalar>(&_shmem,blkdim.x+2,blkdim.y+2,blkdim.z);

		if (blkpos.x < 2)
			cube_set_at(tmp, blkpos + make_int3(blkdim.x, 0, 0), 
                cube_get_at_circ(x, pos + make_int3(blkdim.x, 0, 0)));

		if (blkpos.y < 2)
			cube_set_at(tmp, blkpos + make_int3(0, blkdim.y, 0), 
                cube_get_at_circ(x, pos + make_int3(0, blkdim.y, 0)));

		if (blkpos.x < 2 && blkpos.y < 2)
			cube_set_at(tmp, blkpos + make_int3(blkdim.x, blkdim.y, 0), 
                cube_get_at_circ(x, pos + make_int3(blkdim.x, blkdim.y, 0)));

		cube_set_at(tmp, blkpos, cube_get_at(x, pos));

		__syncthreads();
		scalar total = 0.0;
		for (int m = 0; m <= 2; m++)
			for (int n = 0; n <= 2; n++)
			{
				total += cube_get_at_safe(tmp, 
                    make_int3(blkpos.x+m, blkpos.y+n, blkpos.z));
			}		

		cube_set_at(y, pos, scalar(1.0/9) * total);
	}


#else
#include <omp.h>

	/* A 3x3 separable filter - OpenMP-based implementation on the CPU */
	extern "C" EXPORT void ENTRY(filter3x3)(void* closure, Cube x, Cube y, 
        int3 _gridDim, int3 _blockDim, int _nThreads)
	{
		omp_set_num_threads(_nThreads);
		#pragma omp parallel for
		for (int m=0; m<_gridDim.x * _blockDim.x; m++)
			for (int n=0; n<_gridDim.y * _blockDim.y; n++)
				for (int k=0; k<_gridDim.z * _blockDim.z; k++)
		{
			int3 pos = make_int3(m,n,k);
			cube_set_at(y, pos, scalar(1.0/9) * (cube_get_at(x, pos) +
											cube_get_at_circ(x, pos + make_int3(-1,-1,0)) +
											cube_get_at_circ(x, pos + make_int3(-1, 0,0)) +
											cube_get_at_circ(x, pos + make_int3(-1, 1,0)) +
											cube_get_at_circ(x, pos + make_int3( 0,-1,0)) +
											cube_get_at_circ(x, pos + make_int3( 0, 1,0)) +
											cube_get_at_circ(x, pos + make_int3( 1,-1,0)) +
											cube_get_at_circ(x, pos + make_int3( 1, 1,0)) +
											cube_get_at_circ(x, pos + make_int3( 1, 1,0))));
		}
	}
	 
#endif