//
// cpp_sample2.cpp
// 
// Example C++ module that can be used from Quasar, demonstrating the use
// of the Quasar DSL (which permits passing variables of any type, including
// dynamic types, such as cell matrices from Quasar to C++). Additionally,
// the Quasar DSL allows directly allocating managed host memory and in fact
// gives access to the full functions of the Quasar Runtime system.
//
// Note that in order to use the Quasar DSL interface, it is required to
// provide the 'module_init' and 'module_term' exports, as in the definitions
// below.
//
// (c) 2016 Bart Goossens
//
#include "quasar_dsl.h"
#include <utility>

using namespace quasar;

static IQuasarHost* host;

// A CPU device function that uses a Quasar callback in order to allocate a matrix
extern "C" EXPORT __device__ void ENTRY(generate_matrix)(void *closure, int M, int N, QValue *retVal)
{
	//QValue A = QValue::CreateMatrix<scalar>(M, N);
	//Matrix Adata = host->LockMatrix<scalar>(A);
	//host->UnlockMatrix(A);

	// Alternative version, using cooperative memory management
	cooperative<Matrix> A = zeros_coop<scalar>(M, N);
	for (int m = 0; m < A.dim1; m ++)
		for (int n = 0; n < A.dim2; n ++)
		{
			A.data[m * A.dim2 + n] = m + n;
		}
	*retVal = A;
}

// A CPU device function that processes a cell matrix
// Note: due to the C ABI interface, it is required to pass QValues using the 
// qvalue_t struct.
extern "C" EXPORT __device__ void ENTRY(multiply_cell)(void *closure, qvalue_t _cell)
{
	QValue cell(_cell);
	for (int m = 0; m < size(cell,0); m++)
		for (int n = 0; n < size(cell,1); n++)
			for (int k = 0; k < size(cell,2);  k++)
	{
		cell(m, n, k) = (QValue)cell(m, m, k) * 2.0;
	}
}

// Function: module_init
//
// Function called when the module is loaded. You can place all
// initialization code here.
//
// Parameters:
// - hostprivParams: private host parameters. Just pass to IQuasarHost::Create
// - moduleDetails: private module parameters. Just pass to IQuasarHost::Create
// - flags: private flags for creating the host connection. Just pass to IQuasarHost::Create
extern "C" EXPORT __device__ void module_init(
	LPCVOID hostprivParams, 
	LPCVOID moduleDetails, 
	int flags)
{
	host = IQuasarHost::Connect(hostprivParams, moduleDetails, flags);
}

// Function called when the module is unloaded
extern "C" EXPORT __device__ void module_term()
{
	// Release the host connection
	if (host)
	{
		printf("module_term called\n");
		delete host;
		printf("delete host successful IQuasarHost=%p!\n", IQuasarHost::GetInstance());
		host = NULL;
	}
}
