// lib_read_dataset.cpp
//
// Example Quasar C++ library - exposes the read_dataset() function which
// can be called from Quasar. The function loads a text file (in .csv format)
// as a data matrix.
// 
// Authors: Mohsen Nourazar, Bart Goossens
//

#ifdef _MSC_VEC
#define _CRT_SECURE_NO_WARNINGS  // We use the unsafe CRT functions to be 
								 // compatible with GCC
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <utility>
#include <wchar.h>
#include "quasar.h"
#include "quasar_dsl.h"
#include "quasar_dsl_class_ext.h"

using namespace quasar;

static IQuasarHost* host = nullptr;

/*
 * Helper function to call the Quasar "error" function
 * */
static void error(LPCTSTR error_msg)
{
	Function err(_T("error(...)"));
	err(error_msg);
}

/*
 * read_dataset
 * 
 * public function exposed to Quasar. 
 * 
 * Usage:
 * [npoints, nfeatures, Qmat] = read_dataset(qfilepath)
 * 
 * Arguments:
 * - qfilepath (string): file path of the file to read
 * 
 * Remarks:
 * - the input file is a space or comma-separated value list
 *   (e.g. a CSV file), each row is on a different line.
 * 
 * 
 * Returns: 
 * - npoints
 * - nfeatures
 * - a feature matrix (npoints x nfeatures)
 * */
QValue read_dataset(QValue qfilepath)
{	
	// Protection if read_dataset is called with no module_init()
    if(host == nullptr)
	{
		error(_T("Quasar host not initialized correctly!"));
        return std::make_tuple(0,0,QValue());
    }
	
	// Convert the QValue to a string variable
    string_t wfilepath;
    host->GetString(qfilepath, wfilepath);

	// Empty string was passed
    if(wfilepath.get_length() == 0)
        return std::make_tuple(0,0,QValue());

    char filename[256];
    size_t sz = wcstombs(filename, wfilepath.get_buf(), 256);
    if(sz == 0)
	{
		error(_T("No filename specified!"));
        return std::make_tuple(0,0,QValue());
	}
	
    int i = 0, j = 0;
    char line[1024];
    FILE *infile;
    int nfeatures = 0, npoints = 0;

    if ((infile = fopen(filename, "r")) == nullptr)
	{
		error(_T("Could not open input file!"));
		return std::make_tuple(0,0,QValue());
	}
	
	// Count the number of points that are in the file
	while (fgets(line, 1024, infile) != nullptr)
	{
		if (strtok(line, " \t\n") != 0)
			npoints++;			
	}

	rewind(infile);

	// Count the number of features
	while (fgets(line, 1024, infile) != nullptr) 
	{
		if (strtok(line, " \t\n") != 0) {
			/* ignore the id (first attribute): nfeatures = 1; */
			while (strtok(nullptr, " ,\t\n") != nullptr) nfeatures++;
			break;
		}
	}        

	// allocate space for features[] and read attributes of all objects
	float  *buf = (float*) malloc(npoints*nfeatures*sizeof(float));

	// Read the dataset
	rewind(infile);
	i = 0;
	while (fgets(line, 1024, infile) != nullptr)
	{
		if (strtok(line, " \t\n") == nullptr)
			continue;            
		for (j=0; j<nfeatures; j++)
		{
			buf[i] = (float)atof(strtok(nullptr, " ,\t\n"));  
			i++;             
		} 
				   
	}
	fclose(infile);	

	// Create an output matrix
	QValue Qmat = QValue::CreateMatrix<scalar>(npoints, nfeatures);
	Matrix mat = host->LockMatrix<scalar>(Qmat);
	memcpy(mat.data, buf, npoints*nfeatures*sizeof(float)); /* now features holds 2-dimensional array of features */
	host->UnlockMatrix(Qmat);
	free(buf);

	return std::make_tuple(npoints, nfeatures, Qmat);
}


// Function: module_init
//
// Function called when the module is loaded. You can place all
// initialization code here.
//
// Parameters:
// - hostprivParams: private host parameters. Just pass to IQuasarHost::Create
// - moduleDetails: private module parameters. Just pass to IQuasarHost::Create
// - flags: private flags for creating the host connection. Just pass to IQuasarHost::Create
extern "C" EXPORT __device__ void module_init(
	LPCVOID hostprivParams, 
	LPCVOID moduleDetails, 
	int flags)
{
	host = IQuasarHost::Connect(hostprivParams, moduleDetails, flags);
	
	// Register the function so that it can be used from Quasar. Note that
	// the function signature of the function will be extracted automatically.
	// In case unsupported parameter types are used, an error will result.
	// See documentation for a list of supported parameter types.
	ADD_FUNCTION(read_dataset);
}

// Function called when the module is unloaded
extern "C" EXPORT __device__ void module_term()
{
	// Release the host connection
	if (host)
	{
		delete host;
		host = nullptr;
	}
}
