#include "quasar_dsl.h"
#include "DeckLinkAPI_h.h"
#include "DeckLinkDevice.h"

using namespace quasar;

class ScreenPreviewCallback;

// Static objects (static objects are bad, here they are used for 
// simplicity)
static IQuasarHost* host;
static DeckLinkDeviceDiscovery * discovery;
static DeckLinkDevice * device;
static Function* error;
static ScreenPreviewCallback *screenPreviewCallback;

class ScreenPreviewCallback : public IDeckLinkScreenPreviewCallback
{
private:
	ULONG m_cRef;
	HANDLE frameEvent;
	IDeckLinkVideoFrame *frame;
	QValue buffer;

public:
	ScreenPreviewCallback();
	virtual ~ScreenPreviewCallback();
	ULONG AddRef();
	ULONG Release();
	HRESULT QueryInterface(REFIID  riid, LPVOID * ppvObj);
	HRESULT STDMETHODCALLTYPE ScreenPreviewCallback::DrawFrame(
		IDeckLinkVideoFrame *theFrame);
	void ReadFrame(qvalue_t &buffer);
};

ScreenPreviewCallback::ScreenPreviewCallback() : m_cRef(1)
{
	frameEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	frame = NULL;
}

ScreenPreviewCallback::~ScreenPreviewCallback()
{
	CloseHandle(frameEvent);
	if (frame != NULL)
	{
		frame->Release();
		frame = NULL;
	}
}

HRESULT STDMETHODCALLTYPE ScreenPreviewCallback::DrawFrame(
	IDeckLinkVideoFrame *theFrame)
{
	if (frame != NULL)
	{
		frame->Release();
		frame = NULL;
	}
	frame = theFrame;
	theFrame->AddRef();
	PulseEvent(frameEvent);

	return S_OK;
}

HRESULT ScreenPreviewCallback::QueryInterface(REFIID  riid, LPVOID * ppvObj)
{
	// Always set out parameter to NULL, validating it first.
	if (!ppvObj)
		return E_INVALIDARG;
	*ppvObj = NULL;
	if (riid == IID_IUnknown)
	{
		// Increment the reference count and return the pointer.
		*ppvObj = (LPVOID)this;
		AddRef();
		return NOERROR;
	}
	return E_NOINTERFACE;
}

ULONG ScreenPreviewCallback::AddRef()
{
	InterlockedIncrement(&m_cRef);
	return m_cRef;
}
ULONG ScreenPreviewCallback::Release()
{
	// Decrement the object's internal counter.
	ULONG ulRefCount = InterlockedDecrement(&m_cRef);
	if (0 == m_cRef)
	{
		delete this;
	}
	return ulRefCount;
}

void ScreenPreviewCallback::ReadFrame(qvalue_t &bufferOut)
{
	WaitForSingleObject(frameEvent, INFINITE);

	int width = (int)frame->GetWidth();
	int height = (int)frame->GetHeight();

	if (((qvalue_t) buffer).type == TYPE_VOID ||
		size(buffer, 0) != height ||
		size(buffer, 1) != width)
	{
		buffer = host->CreateMatrix<unsigned short>(height, width);
	}

	unsigned char *buf = NULL;
	if (frame->GetBytes((void **)&buf) == S_OK)
	{
		auto cube = host->LockMatrix<unsigned char>(buffer, quasar::LOCK_WRITE, MEMRESOURCE_CPU);
		int pitch = (int)frame->GetRowBytes();
		int row_bytes = width * sizeof(unsigned short);

		for (int row = 0; row < height; row++)
		{
			memcpy(cube.data + row * row_bytes,
				buf + pitch * row,
				row_bytes);
		}

		host->UnlockMatrix(buffer, quasar::LOCK_WRITE, MEMRESOURCE_CPU);
	}

	bufferOut = (qvalue_t) this->buffer;
}


//
// Returns the decklink devices
//
extern "C" EXPORT __device__ void ENTRY(decklink_get_devices)(void *closure, QValue *devices)
{
	QValue result = host->CreateVector<qvalue_t>(discovery->Count());

	for (int index = 0; index < discovery->Count(); index++)
	{
		DeckLinkDevice * dd = new DeckLinkDevice(discovery->GetDeckLink(index));
		dd->Init();
		result(index) = QValue(dd->GetDeviceName().c_str());
		dd->Release();
	}

	*devices = result;
}

//
// Checks if a device is opened
//
static void decklink_checkdevice()
{
	if (device == NULL)
	{
		(*error)(_T("No device opened: use decklink_open() first to open a device"));
		return;
	}
}

//
// Returns the display modes for the specified device
//
extern "C" EXPORT __device__ void ENTRY(decklink_get_displaymodes)(void *closure, QValue *displayModes)
{
	decklink_checkdevice();

	std::vector<std::wstring> display_modes;
	device->GetDisplayModeNames(display_modes);
	QValue result = host->CreateVector<qvalue_t>(display_modes.size());

	for (int index = 0; index < display_modes.size(); index++)
	{
		result(index) = QValue(display_modes.at(index).c_str());
	}
	*displayModes = result;
}

//
// Closes the specified decklink device
//
extern "C" EXPORT __device__ void ENTRY(decklink_close)(void *closure)
{
	if (device != NULL)
	{
		if (device->IsCapturing())
			device->StopCapture();
		device->Release();
		device = NULL;
	}
}


//
// Opens the specified decklink device
//
extern "C" EXPORT __device__ void ENTRY(decklink_open)(void *closure, int device_index)
{
	ENTRY(decklink_close)(NULL);

	if (device_index < 0 || device_index >= discovery->Count())
	{
		(*error)(_T("decklink_open: invalid index!"));
		return;
	}

	device = new DeckLinkDevice(discovery->GetDeckLink(device_index));
	device->Init();
}


//
// Starts capturing on the specified decklink device
//
extern "C" EXPORT __device__ void ENTRY(decklink_startcapture)(void *closure, int display_mode_index)
{
	decklink_checkdevice();

	screenPreviewCallback = new ScreenPreviewCallback();
	device->StartCapture(display_mode_index, screenPreviewCallback, false);
	screenPreviewCallback->Release();
}

//
// Stops capturing on the specified decklink device
//
extern "C" EXPORT __device__ void ENTRY(decklink_stopcapture)(void *closure)
{
	decklink_checkdevice();
	device->StopCapture();
	screenPreviewCallback = NULL;
}

//
// Reads a frame for the decklink device
//
extern "C" EXPORT __device__ void ENTRY(decklink_readframe)(void *closure, QValue *pbuffer)
{
	decklink_checkdevice();
	if (!device->IsCapturing() || screenPreviewCallback == NULL)
	{
		(*error)(_T("decklink_readframe: device is currently not capturing!"));
		return;
	}

	qvalue_t buffer;
	screenPreviewCallback->ReadFrame(buffer);
	*pbuffer = buffer;
	pbuffer->Retain();
}


// Function: module_init
//
// Function called when the module is loaded. You can place all
// initialization code here.
//
// Parameters:
// - hostprivParams: private host parameters. Just pass to IQuasarHost::Create
// - moduleDetails: private module parameters. Just pass to IQuasarHost::Create
// - flags: private flags for creating the host connection. Just pass to IQuasarHost::Create
extern "C" EXPORT __device__ void module_init(
	LPCVOID hostprivParams,
	LPCVOID moduleDetails,
	int flags)
{
	host = IQuasarHost::Connect(hostprivParams, moduleDetails, flags);

	error = new Function(_T("error(...)"));

	CoInitialize(NULL);

	discovery = new DeckLinkDeviceDiscovery();
	discovery->Enable();

	// Wait for the devices to be enumerated
	if (!discovery->WaitForDevices())
	{
		wprintf(L"Warning: no DeckLink devices discovered!\n");
	}
}

// Function called when the module is unloaded
extern "C" EXPORT __device__ void module_term()
{
	if (device != NULL)
	{
		device->StopCapture();
		device->Release();
		device = NULL;
	}
	if (discovery != NULL)
	{
		discovery->Release();
		discovery = NULL;
	}
	if (error != NULL)
	{
		delete error;
		error = NULL;
	}
	// Release the host connection
	if (host)
	{
		delete host;
		host = NULL;
	}

	CoUninitialize();
}
