﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using Quasar.TypeSystem;
using Quasar.Computations;
using Quasar.Computations.Math;
using DirectShowLib;

namespace DirectShowCam
{
    /// <summary>
    /// Quasar library for directshow video source access. This C# source file
    /// also demonstrates how interfacing with other software components can be
    /// done.
    /// </summary>
    public class DirectShowCam
    {
        // Private fields to store the computation engine and evaluation stack
		internal static IComputationEngine ce;
		internal static IEvaluationStack es;

		// Function called when the library is initialized
        public static void run(IComputationEngine ce, IEvaluationStack es)
        {
            DirectShowCam.ce = ce;
            DirectShowCam.es = es;
        }

        /// <summary>
        /// Function that lists the cameras in the system. Returns a cell vector
        /// with camera info.
        /// </summary>
        [System.ComponentModel.Description(
            "Function: dx_listcam\n\n" +
            "Returns a cell array with information about the cameras installed in the system."),
         QObject.MemberVisible]       
        [return: QObject.ArgTypeAttribute("vec[??]")]
        public static Matrix dx_listcam()
        {
            var videoInputDevices = new List<DsDevice>(DsDevice.GetDevicesOfCat(FilterCategory.VideoInputDevice));

            var deviceInfo = new QValue[videoInputDevices.Count];

            for (int i = 0; i < videoInputDevices.Count; i++)
            {
                var camInfo = new QObject();                
                camInfo["index"] = new QValue(i);
                camInfo["name"] = new QValue(videoInputDevices[i].Name);
                camInfo["device_path"] = new QValue(videoInputDevices[i].DevicePath);
                deviceInfo[i] = camInfo;
            }

            return ce.MatrixFactory.New(AllocationFlags.None, deviceInfo, 1, deviceInfo.Length);
        }        

        /// <summary>
        /// Start capturing, calling the specified callback function
        /// </summary>
        [System.ComponentModel.Description(
            "Function: dx_capture\n\n" +
            "Starts capturing from the specified camera.\n\n"+
            ": function [] = dx_capture(camIndex : int, frameCaptureCallback : [?? -> ()])\n\n" + 
            "Parameters:\n" +
            "   cam_index - the index of the camera\n"+
            "   frame_capture_callback - the function to call when a new frame is captured\n"),
         QObject.MemberVisible]
        [return: QObject.ArgType("qdirectshowcam")]
        public static QObject dx_capture(QValue camIndex, QValue frameCaptureCallback)
        {
            var videoInputDevices = new List<DsDevice>(DsDevice.GetDevicesOfCat(FilterCategory.VideoInputDevice));            
            int[] ids = null;

            switch (camIndex.type)
            {
                case QValue.Types.QVT_SCALAR:
                    int id = camIndex.ToInt();
                    if (id < 0 || id >= videoInputDevices.Count)
                    {
                        es.Push(string.Format("(dx_capture) Camera index should be in the range (0-{0})!", videoInputDevices.Count - 1));
                        ce.FunctionCall("error", 1);
                    }
                    ids = new[] { id };
                    break;
                case QValue.Types.QVT_MATRIX:
                    Matrix idMatrix = (Matrix)camIndex.value;
                    idMatrix.BeginRead();
                    ids = (int[])GenericTypes.Cast(idMatrix.raw, typeof(int));
                    break;
                default:
                    es.Push("(dx_capture) Camera index should be scalar!");
                    ce.FunctionCall("error", 1);
                    break;
            }

            // Setup the callback function
            if (frameCaptureCallback.type != QValue.Types.QVT_LAMBDAEXPRESSION)
            {
                es.Push("(dx_capture) Callback function expected!");
                ce.FunctionCall("error", 1);
            }

            QLambdaExpression frameCaptureCallbackFunction = (QLambdaExpression)frameCaptureCallback.value;
            FunctionTypeInfo fti = frameCaptureCallbackFunction.TypeInfo as FunctionTypeInfo;
            if (fti.inArgTypes.Length > 2)
            {
                es.Push("(dx_capture) Lambda expression should have only at most two input arguments!");
                ce.FunctionCall("error", 1);
            }

            if (ce.FltType != typeof(float))
            {
                es.Push("(dx_capture) Video real-time processing is only supported for single precision floating point data (for performance reasons)!");
                ce.FunctionCall("error", 1);
            }

            DirectShowCamBase cam;

            if (ids.Length == 1)
                cam = new DirectShowCamBase(ids[0]);
            else
                cam = new DirectShowCamBase(ids);

            // Allocate a frame buffer
            Matrix frame_buffer; 
            Matrix frame_buffers;


            frame_buffer = ce.MatrixFactory.Alloc(AllocationFlags.None, typeof(byte), cam.Height, cam.Width, 3);

            frame_buffers = ce.MatrixFactory.New(AllocationFlags.None, GenericTypes.NewArray(frame_buffer.GetType(), ids.Length), 1, ids.Length);
            frame_buffers.BeginWrite();
            frame_buffers.SetAt(0, frame_buffer);
            for (int i = 1; i < ids.Length; i++)
            {
                Matrix fb2 = ce.MatrixFactory.Alloc(AllocationFlags.None, typeof(byte), cam.Height, cam.Width, 3);
                frame_buffers.SetAt(i, fb2);
                ce.Release(fb2);
            }

            cam["frame_buffer"] = frame_buffer;
            cam["frame_buffers"] = frame_buffers;
            cam["frame_width"] = new QValue(cam.Width);
            cam["frame_height"] = new QValue(cam.Height);
            cam["frame_rate"] = new QValue(cam.FramesPerSecond);
            ce.Release(frame_buffer); // Release, because a reference is already stored in "cam".
            ce.Release(frame_buffers);

            // Initialize the frame buffer for writing...
            frame_buffer.BeginWrite();
           
            cam.OnVideoFrameCaptured += delegate(object sender, VideoFrameCapturedEventArgs args)
            {
                int height = cam.Height;
                int width = cam.Width;
                int pitch = 3 * width;
                int srcCamIndex = args.CamIndex;
                Matrix fb = (Matrix) frame_buffers.GetAt(srcCamIndex);               

                // Note that the OnVideoFrameCaptured may be called from a DirectX thread,
                // therefore, we need to perform these actions on the main thread!
                Gtk.Application.Invoke(delegate
                {
                    // Already stopped capturing...
                    if (cam.State == DirectShowCamBase.PlaybackState.Stopped)
                        return;

                    fb.BeginWrite();

                    // Turn the frame upside down & do the necessary conversions
                    unsafe
                    {
                        fixed (byte* frameBufferPointer = fb.raw_ptr<byte>())
                        {
                            for (int row = 0; row < height; row++)
                            {
                                byte* pSrc = (byte*)args.DataPointer.ToPointer() + (height - 1 - row) * pitch;
                                byte* pDst = (byte*)frameBufferPointer + row * pitch;
                                CopyMemory(pDst, pSrc, pitch);
                            }
                        }
                    }


                    frameCaptureCallbackFunction.Invoke(cam, srcCamIndex);
                });
            };
            return cam;
        }
        
        // Function called when the library is de-initialized
        static void cleanup()
        {
        }

        // Windows only!
        [System.Runtime.InteropServices.DllImport("kernel32.dll")]
        private static extern unsafe uint CopyMemory(
            void* Destination,
            void* Source,
            int Length);
    }
}
