﻿using System;
using System.IO;
using System.Xml;
using System.Collections.Generic;
using System.Threading;
using System.Runtime.InteropServices;
using System.Linq;
using System.Text;
using DirectShowLib;
using WindowsMediaLib;
using Quasar.Computations;

namespace DirectShowCam
{
    /// <summary>
    /// This class exposes a VideoTexture2D property for use by Direct3D components and applications that 
    /// consume Texture2D objects.  This class uses the 3rd Party DirectShow.NET library.
    /// </summary>
    public class VideoCaptureSource : QObject, ISampleGrabberCB
    {
        protected DsROTEntry rotEntry;
        protected bool hasVideo;
        protected bool hasAudio;
        protected bool gpuProcessing;
        private VideoType vidType;
        private DsDevice videoDevice;
        private int width;
        private int height;
        private int source = 0;
        private int framesPerSecond = 30;

        /// <summary>
        /// This event is raised whenever the current frame of the VideoTexture changes.  Use this event to push
        /// new data to subcribers, as an alternative to polling by calling the VideoTexture2D property.
        /// </summary>
        public event EventHandler<VideoFrameCapturedEventArgs> OnVideoFrameCaptured;
        public event EventHandler OnVideoCaptureStopped;

        /// <summary>
        /// The width of the source video stream (pixels).
        /// </summary>
        public int Width
        {
            get { return width; }
        }

        /// <summary>
        /// The height of the source video stream (pixels).
        /// </summary>
        public int Height
        {
            get { return height; }
        }

        /// <summary>
        /// The index of the DirectX Video Device source.  Default is 0, for the first device found, or this value can
        /// be set in the constructor.
        /// </summary>
        public int Source
        {
            get { return source; }
        }

        /// <summary>
        /// The DirectShow Device that is providing the video.
        /// </summary>
        public DsDevice VideoDevice
        {
            get { return videoDevice; }
        }

        private string videoFile;
        /// <summary>
        /// The video file to use as a source for the VideoTexture.
        /// </summary>
        public string VideoFile
        {
            get { return videoFile; }
            set { videoFile = value; }
        }

        private bool isLooping = true;
        /// <summary>
        /// Should the video clip loop continuously? (Default = True)
        /// </summary>
        public bool IsLooping
        {
            get { return isLooping; }
            set { isLooping = value; }
        }

        private uint timeout = 500;
        /// <summary>
        /// Timeout for Graph Events, in milliseconds. (Default = 500)
        /// </summary>
        public uint Timeout
        {
            get { return timeout; }
            set { timeout = value; }
        }

        /// <summary>
        /// The playback rate for video clips. (Default = 30)
        /// </summary>
        public int FramesPerSecond
        {
            get { return framesPerSecond; }
            set { framesPerSecond = value; }
        }

        private double duration = 0;
        /// <summary>
        /// The duration of the video sequence (in seconds)
        /// </summary>
        public double Duration
        {
            get { return duration; }
        }

        // Current state of the graph, which can be changed in multiple threads
        private volatile PlaybackState playbackState = PlaybackState.Stopped;

        #region Constructors
        /// <summary>
        /// Graphics Device based constructor. This constructor assumes that the first DirectShow video device that is found 
        /// will be used as the source of the video.
        /// </summary>
        /// <param name="graphicsDevice">The GraphicsDevice where the video texture will be rendered.  Typically, this is the
        /// GraphicsDevice of the Game or Application that uses the VideoTexture.</param>
        public VideoCaptureSource() : this(0) { }
        /// <summary>
        /// A constructor that specifies a particular DirectShow Video Device to use as the source of the video.
        /// </summary>
        /// <param name="videoSource">The DirectShow video device to use.  This integer represents the index that is
        /// assigned to the device as the available video devices are enumerated.</param>
        /// <param name="graphicsDevice">The GraphicsDevice where the video texture will be rendered.  Typically, this is the
        /// GraphicsDevice of the Game or Application that uses the VideoTexture.</param>
        public VideoCaptureSource(int videoSource)
        {
            this.source = videoSource;
            this.vidType = VideoType.Live;
            // 1. Obtain the video device.
            this.videoDevice = SelectWebcam(this.source);
            // 2. Setup the Capture Graph.
            SetupGraph(this.videoDevice);
            // 3. Start the video streaming process.
            Start();
        }
        /// <summary>
        /// A constructor that specifies a particular video file to use as the source of the video.
        /// </summary>
        /// <param name="videoFile">An .avi, .mpeg, .wmv video file.</param>
        /// <param name="graphicsDevice">The GraphicsDevice where the video texture will be rendered.  Typically, this is the
        /// GraphicsDevice of the Game or Application that uses the VideoTexture.</param>
        public VideoCaptureSource(string videoFile)
        {
            this.videoFile = videoFile;

            // 1. Determine the type of video file.
            if (videoFile.EndsWith(".avi", true, System.Globalization.CultureInfo.CurrentUICulture))
            {
                this.vidType = VideoType.AVI;
            }
            else if (videoFile.EndsWith(".mpeg", true, System.Globalization.CultureInfo.CurrentUICulture))
            {
                this.vidType = VideoType.MPEG;
            }
            else if (videoFile.EndsWith(".mp4", true, System.Globalization.CultureInfo.CurrentUICulture))
            {
                this.vidType = VideoType.MPEG4;
            }
            else if (videoFile.EndsWith(".flv", true, System.Globalization.CultureInfo.CurrentUICulture))
            {
                this.vidType = VideoType.FLV;
            }
            else if (videoFile.EndsWith(".mts", true, System.Globalization.CultureInfo.CurrentUICulture) ||
                     videoFile.EndsWith(".m2ts", true, System.Globalization.CultureInfo.CurrentUICulture))
            {
                this.vidType = VideoType.MTS;
            }
            else if (videoFile.EndsWith(".wmv", true, System.Globalization.CultureInfo.CurrentUICulture))
            {
                this.vidType = VideoType.WMV;
            }
            else
            {
                this.vidType = VideoType.Unknown;
            }
            // 2. Setup the Capture Graph.
            SetupGraph(this.videoFile); // Works for .avi, .mpeg,.wmv files. 
            // 3. Start the video streaming process.
            Start();
        }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Shut down the DirectShowNetVideoPlayer.
        /// </summary>
        public override void Dispose()
        {
            Stop();

            lock (this)
            {
                // Stop the eventwatcher thread
                if (this.eventWatcherThread != null)
                {
                    finishThread = true;
                    manualResetEvent.Set();
                    threadFinishedEvent.WaitOne();
                    this.eventWatcherThread = null;
                }
            }

            if (this.mediaSeeking != null)
            {
                this.mediaSeeking = null;
            }

            if (this.mediaEventEx != null)
            {
                this.mediaEventEx = null;
            }

            if (this.mediaControl != null)
            {
                this.mediaControl.Stop();
                Marshal.ReleaseComObject(this.mediaControl);
                this.mediaControl = null;
            }
            if (this.basicAudio != null)
            {
                this.basicAudio = null;
            }

            if (this.manualResetEvent != null)
            {
                this.manualResetEvent.Close();
            }
            if (this.threadFinishedEvent != null)
            {
                this.threadFinishedEvent.Close();
            }

            if (rotEntry != null)
            {
                rotEntry.Dispose();
                rotEntry = null;
            }
            
        }

        #endregion IDisposable Members

        #region Fields and Working Methods that require DirectShowLib.

        private DirectShowLib.IMediaControl mediaControl = null;
        private DirectShowLib.IMediaEventEx mediaEventEx = null;
        private DirectShowLib.IMediaSeeking mediaSeeking = null;
        private DirectShowLib.IBasicAudio basicAudio = null;

        private DsDevice SelectWebcam(int deviceToUse)
        {
            DsDevice dsDevice = null;
            try
            {
                Console.WriteLine("Trying to select video input device " + deviceToUse.ToString());

                dsDevice = null;

                DsDevice[] videoInputDevices = DirectShowLib.DsDevice.GetDevicesOfCat(FilterCategory.VideoInputDevice);

                // See if there are any Video Input Devices attached and active, and if so, put them into the array...
                if (videoInputDevices.Length == 0)
                {
                    throw new Exception("No video input devices found!");
                }

                if (deviceToUse < videoInputDevices.Length)	// last available deviceToUse
                {
                    dsDevice = videoInputDevices[deviceToUse] as DsDevice;	// cast using as					
                }

                // and perform a final check to make sure the device is selected.
                if (dsDevice == null)
                {
                    throw new Exception("No video input devices selected!");
                }

                Console.WriteLine("Success!");
                return dsDevice;
            }
            catch (Exception ex)
            {
                Console.WriteLine("Failure...");
                Console.WriteLine("Video Input Selection Exception: " + ex.Message);
                throw new Exception("Video Input Selection Exception", ex);
            }
        }

        private void SetupGraph(DsDevice dev)
        {
            try
            {
                int hr;

                // 1. Start building the graph, using FilterGraph and CaptureGraphBuilder2
                DirectShowLib.IFilterGraph2 graphBuilder = (DirectShowLib.IFilterGraph2)new FilterGraph();
                ICaptureGraphBuilder2 builder = (ICaptureGraphBuilder2)new CaptureGraphBuilder2();
                hr = builder.SetFiltergraph(graphBuilder);
                DsError.ThrowExceptionForHR(hr);

                // 2. Add the source filter for video device input.
                DirectShowLib.IBaseFilter sourceFilter = null;
                hr = graphBuilder.AddSourceFilterForMoniker(dev.Mon, null, "Video Input Filter", out sourceFilter);
                DsError.ThrowExceptionForHR(hr);

                // 3. Get the SampleGrabber interface, configure it, and add it to the graph.
                DirectShowLib.ISampleGrabber sampGrabber = (DirectShowLib.ISampleGrabber)new SampleGrabber();
                ConfigureSampleGrabber(sampGrabber);
                hr = graphBuilder.AddFilter((DirectShowLib.IBaseFilter)sampGrabber, "SampleGrabber");
                DsError.ThrowExceptionForHR(hr);

                // 4. Add the null renderer (since we don't want to render in a seperate window.)
                DirectShowLib.IBaseFilter nullRenderer = (DirectShowLib.IBaseFilter)new NullRenderer();
                hr = graphBuilder.AddFilter(nullRenderer, "Null Renderer");
                DsError.ThrowExceptionForHR(hr);

                // 5. Configure the render stream.
                hr = builder.RenderStream(PinCategory.Capture, MediaType.Video, sourceFilter, (DirectShowLib.IBaseFilter)sampGrabber, nullRenderer);
                DsError.ThrowExceptionForHR(hr);

                // 6. Now that everthing is configured and set up, save the width, height, stride information for use later.
                SaveSizeInfo(sampGrabber);

                // 7. Obtain the interfaces that we will use to control the execution of the filter graph.
                this.mediaControl = graphBuilder as DirectShowLib.IMediaControl;
                this.mediaEventEx = graphBuilder as DirectShowLib.IMediaEventEx;
                this.mediaSeeking = graphBuilder as DirectShowLib.IMediaSeeking;
                this.basicAudio = graphBuilder as DirectShowLib.IBasicAudio;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                throw ex;
            }
        }

        /// <summary>
        /// This method sets up the DirectShow filter graph and obtains the interfaces necessary to control playback
        /// for VideoTextures created from video files.  This method works for .avi, .mpeg, and .wmv files.  
        /// </summary>
        /// <param name="videoFile">The .avi, .mpeg, or .wmv video file.</param>
        private void SetupGraph(string videoFile)
        {
            try
            {
                int hr;


                // 1. Start building the graph, using FilterGraph and CaptureGraphBuilder2
                DirectShowLib.IGraphBuilder graphBuilder = (DirectShowLib.IGraphBuilder)new FilterGraph();
                DirectShowLib.ICaptureGraphBuilder2 builder = (DirectShowLib.ICaptureGraphBuilder2)new CaptureGraphBuilder2();
                hr = builder.SetFiltergraph(graphBuilder);
                DsError.ThrowExceptionForHR(hr);

                // 2. Add the source filter for the video file input.
                DirectShowLib.IBaseFilter sourceFilter = null;
                hr = graphBuilder.AddSourceFilter(videoFile, videoFile, out sourceFilter);
                DsError.ThrowExceptionForHR(hr);

                // 3. Get the SampleGrabber interface, configure it, and add it to the graph.
                DirectShowLib.ISampleGrabber sampGrabber = (DirectShowLib.ISampleGrabber)new SampleGrabber();
                ConfigureSampleGrabber(sampGrabber);
                hr = graphBuilder.AddFilter((DirectShowLib.IBaseFilter)sampGrabber, "SampleGrabber");
                DsError.ThrowExceptionForHR(hr);

                // 4. Add the null renderer (since we don't want to render in a seperate window.)
                DirectShowLib.IBaseFilter nullRenderer = (DirectShowLib.IBaseFilter)new NullRenderer();
                hr = graphBuilder.AddFilter(nullRenderer, "Null Renderer");
                DsError.ThrowExceptionForHR(hr);                

                // 5. Render the stream.  The way the stream is rendered depends on its type.
                int hr1 = builder.RenderStream(null, vidType == VideoType.WMV ? new DsGuid(MediaType.Video) : null, sourceFilter, (DirectShowLib.IBaseFilter)sampGrabber, nullRenderer);                
                int hr2 = builder.RenderStream(null, MediaType.Audio, sourceFilter, null, null);

                hasVideo = hr1 == 0;
                hasAudio = hr2 == 0;

                if (hr1 != 0 && hr2 != 0)
                {
                    DsError.ThrowExceptionForHR(hr1);
                }

                rotEntry = new DsROTEntry(graphBuilder);                

                //DsError.ThrowExceptionForHR(hr);

                // 6. Now that everthing is configured and set up, save the width, height, stride information for use later.
                if (hasVideo)
                {
                    SaveSizeInfo(sampGrabber);
                }

                // 7. Obtain the interfaces that we will use to control the execution of the filter graph.
                this.mediaControl = graphBuilder as DirectShowLib.IMediaControl;
                this.mediaEventEx = graphBuilder as DirectShowLib.IMediaEventEx;
                this.mediaSeeking = graphBuilder as DirectShowLib.IMediaSeeking;
                this.basicAudio = graphBuilder as DirectShowLib.IBasicAudio;

                // 8. Get the duration of the sequence
                long current;
                hr = mediaSeeking.GetStopPosition(out current);
                DsError.ThrowExceptionForHR(hr);
                duration = (double)current / 1e7; // in 100 nanoseconds
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                throw ex;
            }
        }

        /// <summary>
        /// Creates a compressor for the specified device
        /// </summary>
        private IBaseFilter CreateCompressor(DsDevice device)
        {
            Guid grf = typeof(IBaseFilter).GUID;
            object o;
            device.Mon.BindToObject(null, null, ref grf, out o);
            return o as IBaseFilter;
        }

        internal const EventCode EC_VideoFileComplete = (EventCode)0x8000;

        public delegate void ProgressDelegate(VideoCaptureSource player, double sampleTime, double duration, long bytesWritten, ref bool cancel);

        /// <summary>
        /// Save the video to the specified output file
        /// </summary>
        /// <param name="outputFileName"></param>
        public void Save(VideoType outputFormatType,
                         string outputFileName, 
                         VideoCompressor videoCompressor,
                         AudioCompressor audioCompressor,
                         ProgressDelegate onProgress)
        {
            // Stop playback of the current file
            Stop();

            // Create a new graph for performing the conversion...
            try
            {
                int hr;

                // 1. Start building the graph, using FilterGraph and CaptureGraphBuilder2 & assign the file
                DirectShowLib.IGraphBuilder graphBuilder = (DirectShowLib.IGraphBuilder)new FilterGraph();
                DirectShowLib.ICaptureGraphBuilder2 builder = (DirectShowLib.ICaptureGraphBuilder2)new CaptureGraphBuilder2();
                hr = builder.SetFiltergraph(graphBuilder);
                DsError.ThrowExceptionForHR(hr);

                IBaseFilter pMux;
                IFileSinkFilter pFilter;
                switch (outputFormatType)
                {
                    case VideoType.AVI:
                        hr = builder.SetOutputFileName(MediaSubType.Avi, outputFileName, out pMux, out pFilter);
                        break;
                    case VideoType.WMV:
                        {
                            hr = builder.SetOutputFileName(MediaSubType.Asf, outputFileName, out pMux, out pFilter);

                            WindowsMediaLib.IConfigAsfWriter lConfig = pMux as WindowsMediaLib.IConfigAsfWriter;
                            ConfigureProfile(lConfig, videoCompressor, audioCompressor);

                            // Windows Media 8 Best Quality based VBR for Broadband.
                            // See c:\windows\WMSysPr9.prx for GUIDs
                            //Guid cat = new Guid("{048439BA-309C-440e-9CB4-3DCCA3756423}");

                            //lConfig.ConfigureFilterUsingProfileGuid(cat);                           
                        }
                        break;

                    default:
                        Marshal.ReleaseComObject(graphBuilder);
                        throw new NotSupportedException("Unsupported file format!");                        
                }
                DsError.ThrowExceptionForHR(hr);
                Marshal.ReleaseComObject(pFilter); // Not needed

                // 2. Add the source filter for the video file input.
                DirectShowLib.IBaseFilter sourceFilter = null;
                hr = graphBuilder.AddSourceFilter(videoFile, videoFile, out sourceFilter);
                DsError.ThrowExceptionForHR(hr);

                // 3. Get the SampleGrabber interface, configure it, and add it to the graph.
                DirectShowLib.ISampleGrabber sampGrabber = (DirectShowLib.ISampleGrabber)new SampleGrabber();
                ConfigureSampleGrabber(sampGrabber);
                hr = graphBuilder.AddFilter((DirectShowLib.IBaseFilter)sampGrabber, "SampleGrabber");
                DsError.ThrowExceptionForHR(hr);

                // 4. Create the video compressor
                DirectShowLib.IBaseFilter videoCompressorFilter = pMux;
                if (videoCompressor != null && videoCompressor.device != null)
                {
                    videoCompressorFilter = CreateCompressor(videoCompressor.device);
                    hr = graphBuilder.AddFilter(videoCompressorFilter, "Video compressor");
                    DsError.ThrowExceptionForHR(hr);
                }

                // 5. Create the audio compressor
                DirectShowLib.IBaseFilter audioCompressorFilter = pMux;
                if (audioCompressor != null && audioCompressor.device != null)
                {
                    audioCompressorFilter = CreateCompressor(audioCompressor.device);
                    hr = graphBuilder.AddFilter(audioCompressorFilter, "Audio compressor");
                    DsError.ThrowExceptionForHR(hr);
                }


                // 6. Render the stream.  The way the stream is rendered depends on its type.
                hr = builder.RenderStream(null, null, sourceFilter, null, (DirectShowLib.IBaseFilter)sampGrabber);
                DsError.ThrowExceptionForHR(hr);

                hr = builder.RenderStream(null, null, (DirectShowLib.IBaseFilter)sampGrabber, null, videoCompressorFilter);
                DsError.ThrowExceptionForHR(hr);
                
                hr = builder.RenderStream(null, MediaType.Audio, sourceFilter, null, audioCompressorFilter);
                bool audioPresent = hr == 0;

                if (videoCompressorFilter != pMux)
                {
                    hr = builder.RenderStream(null, null, videoCompressorFilter, null, pMux);
                    DsError.ThrowExceptionForHR(hr);
                }

                if (audioCompressorFilter != pMux && audioPresent)
                {
                    hr = builder.RenderStream(null, null, audioCompressorFilter, null, pMux);
                    DsError.ThrowExceptionForHR(hr);
                }


                rotEntry = new DsROTEntry(graphBuilder);

                // 7. Now that everthing is configured and set up, save the width, height, stride information for use later.
                SaveSizeInfo(sampGrabber);

                // 8. Obtain the interfaces that we will use to control the execution of the filter graph.
                DirectShowLib.IMediaControl mediaControl = graphBuilder as DirectShowLib.IMediaControl;
                DirectShowLib.IMediaEventEx mediaEventEx = graphBuilder as DirectShowLib.IMediaEventEx;
                DirectShowLib.IMediaSeeking mediaSeeking = graphBuilder as DirectShowLib.IMediaSeeking;
                DirectShowLib.IBasicAudio basicAudio = graphBuilder as DirectShowLib.IBasicAudio;
                
              
                IntPtr graphEvent;
                hr = mediaEventEx.GetEventHandle(out graphEvent);
                DsError.ThrowExceptionForHR(hr);
                ManualResetEvent manualResetEvent = new ManualResetEvent(false);
                manualResetEvent.SafeWaitHandle = new Microsoft.Win32.SafeHandles.SafeWaitHandle(graphEvent, false);

                hr = mediaControl.Run();
                DsError.ThrowExceptionForHR(hr);

                bool finished = false;
                while (!finished)
                {
                    manualResetEvent.WaitOne(50); // Wait 50 msec

                    IntPtr eventParam1, eventParam2;
                    EventCode eventCode;                    

                    // Update progress
                    if (onProgress != null)
                    {
                        onProgress(this, m_SampleTime, duration, new FileInfo(outputFileName).Length, ref finished);
                    }

                    // Process all queued events
                    while (mediaEventEx.GetEvent(out eventCode, out eventParam1, out eventParam2, 0) == 0)
                    {
                        switch (eventCode)
                        {
                            case EventCode.Complete:
                            case EC_VideoFileComplete:

                                finished = true;
                                break;
                            default:
                                break;
                        }

                        // Free event parameter memory.
                        hr = this.mediaEventEx.FreeEventParams(eventCode, eventParam1, eventParam2);
                        DsError.ThrowExceptionForHR(hr);
                    }
                }

                hr = mediaControl.Stop();
                DsError.ThrowExceptionForHR(hr);

                manualResetEvent.Close();

                Marshal.ReleaseComObject(graphBuilder);

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                throw ex;
            }
        }

        private void ConfigureProfile(WindowsMediaLib.IConfigAsfWriter lConfig, VideoCompressor videoCompressor, AudioCompressor audioCompressor)
        {
            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.Load(GetType().Assembly.GetManifestResourceStream("DirectShowCam.WMVProfile.xml"));

            if (audioCompressor != null)
            {
                XmlNode audioConf = xmlDoc.SelectSingleNode(@"profile/streamconfig[@streamname=""Audio1""]");
                audioConf.Attributes["bitrate"].Value = audioCompressor.bitRate.ToString();
            }

            if (videoCompressor != null)
            {
                XmlNode videoConf = xmlDoc.SelectSingleNode(@"profile/streamconfig[@streamname=""Video2""]");
                videoConf.Attributes["bitrate"].Value = videoCompressor.bitRate.ToString();
                XmlNode videoQualityConf = videoConf.SelectSingleNode("videomediaprops");
                videoQualityConf.Attributes["quality"].Value = videoCompressor.quality.ToString();
                videoQualityConf.Attributes["maxkeyframespacing"].Value = Math.Floor (1e7 * videoCompressor.secsPerKeyFrame).ToString();
                XmlNode videoInfoHdr = videoConf.SelectSingleNode("wmmediatype/videoinfoheader");
                videoInfoHdr.Attributes["dwbitrate"].Value = videoCompressor.bitRate.ToString();
                videoInfoHdr.Attributes["avgtimeperframe"].Value = Math.Floor(1e7 / videoCompressor.fps).ToString();
                videoInfoHdr.SelectSingleNode("rcsource").Attributes["right"].Value = videoCompressor.frameWidth.ToString();
                videoInfoHdr.SelectSingleNode("rcsource").Attributes["bottom"].Value = videoCompressor.frameHeight.ToString();
                videoInfoHdr.SelectSingleNode("rctarget").Attributes["right"].Value = videoCompressor.frameWidth.ToString();
                videoInfoHdr.SelectSingleNode("rctarget").Attributes["bottom"].Value = videoCompressor.frameHeight.ToString();
                videoInfoHdr.SelectSingleNode("bitmapinfoheader").Attributes["biwidth"].Value = videoCompressor.frameWidth.ToString();
                videoInfoHdr.SelectSingleNode("bitmapinfoheader").Attributes["biheight"].Value = videoCompressor.frameHeight.ToString();
            }


            string xmlStream = xmlDoc.OuterXml;

            IWMProfileManager profileMan;
            IWMProfile profile;
            WMUtils.WMCreateProfileManager(out profileMan);
            profileMan.LoadProfileByData(xmlStream, out profile);
            StringBuilder profileName = new StringBuilder(100);
            int profileNameLength = 100;
            profile.GetName(profileName, ref profileNameLength);
            lConfig.ConfigureFilterUsingProfile(profile);           

            Marshal.ReleaseComObject(profileMan);

            /*IWMProfileManager pWMProfileManager = null; 
            WMUtils.WMCreateProfileManager(out pWMProfileManager); 
            var prof2 = pWMProfileManager as IWMProfileManager2; 
            prof2.SetSystemProfileVersion(WMVersion.V8_0); 
            int pcProfiles; 
            prof2.GetSystemProfileCount(out pcProfiles); 
            for (int index = 0; index < pcProfiles; index++) 
            { 
                IWMProfile profile = null; 
                prof2.LoadSystemProfile(index, out profile); 
                int namelength = 100; 
                StringBuilder strbuilder = new StringBuilder(namelength); 
                profile.GetName(strbuilder, ref namelength); 
                System.Diagnostics.Debug.WriteLine(strbuilder.ToString()); 
                Guid codecguid; 
                ((IWMProfile2)profile).GetProfileID(out codecguid); 
                if (strbuilder.ToString() == "Windows Media Video 8 for Local Area Network (768 Kbps)") 
                { 
                    cat = codecguid; 
                } 
                Marshal.ReleaseComObject(profile); 
            } 
            lConfig.ConfigureFilterUsingProfileGuid(cat);*/
        }

        private void ConfigureSampleGrabber(ISampleGrabber sampGrabber)
        {
            AMMediaType media;
            int hr;

            // Set the media type to Video/RBG24
            media = new AMMediaType();
            media.majorType = MediaType.Video;
            media.subType = MediaSubType.RGB32;
            media.formatType = FormatType.VideoInfo;
            hr = sampGrabber.SetMediaType(media);
            DsError.ThrowExceptionForHR(hr);

            //DsUtils.FreeAMMediaType(media);
            //media = null;

            // Configure the samplegrabber
            hr = sampGrabber.SetCallback(this, 1);
            DsError.ThrowExceptionForHR(hr);
        }

        private void SaveSizeInfo(ISampleGrabber sampGrabber)
        {
            int hr;

            // Get the media type from the SampleGrabber
            AMMediaType media = new AMMediaType();
            hr = sampGrabber.GetConnectedMediaType(media);
            DsError.ThrowExceptionForHR(hr);

            if ((media.formatType != FormatType.VideoInfo) || (media.formatPtr == IntPtr.Zero))
            {
                throw new NotSupportedException("Unknown Grabber Media Format");
            }

            // Grab the size info
            VideoInfoHeader videoInfoHeader = (VideoInfoHeader)Marshal.PtrToStructure(media.formatPtr, typeof(VideoInfoHeader));
            this.width = videoInfoHeader.BmiHeader.Width;
            this.height = videoInfoHeader.BmiHeader.Height;
            this.framesPerSecond = (int) (1e7 / (double)videoInfoHeader.AvgTimePerFrame);            

            DsUtils.FreeAMMediaType(media);
            media = null;
        }

        /// <summary>
        /// Start the Video Playback.
        /// </summary>
        [QObject.MemberVisible]
        public void Start()
        {
            if (this.playbackState == PlaybackState.Stopped)
            {
                Run();
                // Use one or the other of the following techniques (ManualEventWatcher is prefered.)
                //ManualEventWatcher(); // Watches for graph events on a seperate thread, and calls HandleGraphEvents() when triggered.
                //TimerBasedEventPolling(); // Calls the HandleGraphEvents() on a timer.
            }
        }

        /// <summary>
        /// Run the Video Playback.
        /// </summary>
        public void Run()
        {
            if (this.playbackState == PlaybackState.Stopped || this.playbackState == PlaybackState.Paused)
            {
                if (this.mediaControl != null)
                {
                    int hr = this.mediaControl.Run();
                    DsError.ThrowExceptionForHR(hr);
                    this.playbackState = PlaybackState.Running;                    
                }
            }
        }

        /// <summary>
        /// Pause the Video Playback.
        /// </summary>
        [QObject.MemberVisible]
        public void Pause()
        {
            if (this.playbackState == PlaybackState.Running)
            {
                if (this.mediaControl != null)
                {
                    // Calling pause by itself does not seem to do anything. (VideoTexture keeps streaming for some reason).
                    // Calling pause, then stop when ready, results in an image visible and frozen on the screen. (Which is what we want...)
                    int hr = this.mediaControl.Pause();
                    DsError.ThrowExceptionForHR(hr);
                    hr = this.mediaControl.StopWhenReady();
                    DsError.ThrowExceptionForHR(hr);
                    this.playbackState = PlaybackState.Paused;
                }
            }
        }

        /// <summary>
        /// Halt the Video Playback.  
        /// </summary>
        [QObject.MemberVisible]
        public void Stop()
        {
            if (this.playbackState == PlaybackState.Running || this.playbackState == PlaybackState.Paused)
            {
                if (this.mediaControl != null)
                {
                    int hr = this.mediaControl.StopWhenReady();
                    DsError.ThrowExceptionForHR(hr);
                    this.playbackState = PlaybackState.Stopped;
                    if (OnVideoCaptureStopped != null)
                    {
                        OnVideoCaptureStopped(this, EventArgs.Empty);
                    }
                }
            }
        }

        public const int E_NOTIMPL = -2147467263;

        [QObject.MemberVisible]
        public double Position
        {
            get
            {
                long current;
                if (mediaSeeking == null) return 0;
                int hr = mediaSeeking.GetCurrentPosition(out current);
                DsError.ThrowExceptionForHR(hr);
                return (double)current / 1e7; // in 100 nanoseconds
            }
            set
            {
                long current = (long)(value * 1e7);
                long stop = (long)(duration * 1e7);
                bool playing = this.State == PlaybackState.Running;
                // We need to stop the playback - otherwise the program hangs...
                lock (this)
                {
                    if (playing) Stop();
                }
                lock (this)
                {
                    int hr = mediaSeeking.SetPositions(current, AMSeekingSeekingFlags.AbsolutePositioning | AMSeekingSeekingFlags.SeekToKeyFrame | 
                                                        AMSeekingSeekingFlags.NoFlush,
                                                       stop, AMSeekingSeekingFlags.NoPositioning);
                    if (hr != E_NOTIMPL)
                        DsError.ThrowExceptionForHR(hr);
                }
                lock (this)
                {
                    if (playing) Start();
                }
            }
        }

        /// <summary>
        /// Audio volume (in 100ths of decibels) -> from -10000 to 0
        /// </summary>
        [QObject.MemberVisible]
        public int AudioVolume
        {
            get
            {
                int volume;
                if (basicAudio == null)
                    return 0;
                int hr = basicAudio.get_Volume(out volume);
                DsError.ThrowExceptionForHR(hr);
                return volume;
            }
            set
            {
                if (basicAudio != null)
                {
                    int hr = basicAudio.put_Volume(value);
                    //DsError.ThrowExceptionForHR(hr);
                }
            }
        }

        public PlaybackState State { get { return playbackState; } }

        #region ManualEventWatcher Technique

        private IntPtr graphEvent = IntPtr.Zero;
        private ManualResetEvent manualResetEvent;
        private ManualResetEvent threadFinishedEvent;
        private Thread eventWatcherThread;

        private void ManualEventWatcher()
        {
            if (this.mediaEventEx != null)
            {
                int hr = this.mediaEventEx.GetEventHandle(out this.graphEvent);
                DsError.ThrowExceptionForHR(hr);
                this.threadFinishedEvent = new ManualResetEvent(false);
                this.manualResetEvent = new ManualResetEvent(false);
                this.manualResetEvent.SafeWaitHandle = new Microsoft.Win32.SafeHandles.SafeWaitHandle(this.graphEvent, true);
                this.finishThread = false;

                // Create a new thread to wait for events
                this.eventWatcherThread = new Thread(new ThreadStart(this.WatchForGraphEvents));
                this.eventWatcherThread.Name = "Event Watcher Thread";
                this.eventWatcherThread.Start();
            }
        }

        private bool finishThread = false;
        private void WatchForGraphEvents()
        {
            while (true)
            {
                this.manualResetEvent.WaitOne(-1, true);
                if (finishThread) break;

                lock (this)
                {
                    HandleGraphEvents();
                }
            }
            threadFinishedEvent.Set();
        }

        #endregion ManualEventWatcher Technique

        #region TimerBasedEventPolling Technique

        private System.Timers.Timer eventTimer;

        private void TimerBasedEventPolling()
        {
            this.eventTimer = new System.Timers.Timer(100);
            this.eventTimer.Elapsed += new System.Timers.ElapsedEventHandler(eventTimer_Elapsed);
            this.eventTimer.Enabled = true;
        }

        private void eventTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            HandleGraphEvents();
        }

        #endregion TimerBasedEventPolling Technique

        private void HandleGraphEvents()
        {
            if (this.mediaEventEx != null)
            {
                int hr = 0;
                IntPtr eventParam1, eventParam2;
                EventCode eventCode;

                // Print a console message for debugging purposes
                Console.WriteLine("Calling MediaEventEx.GetEvent()");

                // Process all queued events
                while (this.mediaEventEx.GetEvent(out eventCode, out eventParam1, out eventParam2, 0) == 0)
                {
                    switch (eventCode)
                    {
                        case EventCode.Complete:
                            if (this.isLooping)
                            {
                                DsLong pos = new DsLong(0);
                                // Reset to first frame of the clip
                                hr = this.mediaSeeking.SetPositions(pos, AMSeekingSeekingFlags.AbsolutePositioning, null, AMSeekingSeekingFlags.NoPositioning);
                                if (hr < 0) // SetPositions has failed.  Just try stopping and starting again.
                                {
                                    Stop();
                                    Run();
                                }
                            }
                            break;
                        default:
                            break;
                    }

                    // Free event parameter memory.
                    hr = this.mediaEventEx.FreeEventParams(eventCode, eventParam1, eventParam2);
                    DsError.ThrowExceptionForHR(hr);
                }
            }
           
        }

        #region ISampleGrabberCB Members
        protected double m_SampleTime;
        protected bool m_SaveMode;
        /// <summary>
        /// Buffer Callback method from the  DirectShow.NET ISampleGrabberCB interface.  This method is called
        /// when a new frame is grabbed by the SampleGrabber.
        /// </summary>
        /// <param name="SampleTime">The sample time.</param>
        /// <param name="pBuffer">A pointer to the image buffer that contains the grabbed sample.</param>
        /// <param name="BufferLen">The length of the image buffer containing the grabbed sample.</param>
        /// <returns>0 = success.</returns>
        int ISampleGrabberCB.BufferCB(double SampleTime, IntPtr pBuffer, int BufferLen)
        {
            if (finishThread) return -1;

            lock (this)
            {
                // notify subscribers, if any.
                if (OnVideoFrameCaptured != null)
                    OnVideoFrameCaptured(this, new VideoFrameCapturedEventArgs(this, pBuffer, BufferLen, SampleTime));
            }

            return 0;
        }
        /// <summary>
        /// Sample CallBack method from the ISampleGrabberCB interface (DirectShow.NET).  Not used.
        /// </summary>
        /// <param name="SampleTime"></param>
        /// <param name="pSample"></param>
        /// <returns></returns>
        /// 
        int ISampleGrabberCB.SampleCB(double SampleTime, DirectShowLib.IMediaSample pSample)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        #endregion

        #endregion Working Methods that require DirectShowLib

        public enum PlaybackState
        {
            Stopped,
            Paused,
            Running,
            Exiting
        }

        public enum VideoType
        {
            AVI,
            MPEG,
            MPEG4,
            FLV,
            MTS,
            WMV,
            Live,
            Unknown
        }
    }

    /// <summary>
    /// An argument to be passed by the NewVideoTexture Event.
    /// </summary>
    public class VideoFrameCapturedEventArgs : EventArgs
    {
        public double SampleTime { get; private set; }
        public VideoCaptureSource Source { get; private set; }
        public IntPtr DataPointer { get; private set; }
        public int Bytes { get; private set; }

        /// <summary>
        /// Constructor.
        /// </summary>
        public VideoFrameCapturedEventArgs(VideoCaptureSource source, IntPtr dataPointer, int bytes, double sampleTime)
        {
            this.Source = source;
            this.DataPointer = dataPointer;
            this.Bytes = bytes;
            this.SampleTime = sampleTime;
        }
    }
}
