using System;
using System.IO;
using System.Linq;
using Quasar.Computations;
using Quasar.Computations.Math;
using BitMiracle.LibTiff.Classic;
using System.Reflection;
using System.ComponentModel;

namespace TIFFIO
{
	public static class TIFFIO
	{
		/// <summary>
		/// The computation engine to be used.
		/// </summary>
		static IComputationEngine ce;
		/// <summary>
		/// The evaluation stack, which is used for passing arguments to
		/// internal Quasar functions and for intermediate results of computation
		/// </summary>
		static IEvaluationStack es;

		/// <summary>
		/// The @run function is called by Quasar at initialization of the library
		/// (i.e., when executing import "TIFFIO.dll"). Here, we just keep
		/// track of the computation engine and the evaluation stack.
		/// </summary>
		public static void @run(IComputationEngine ce, IEvaluationStack es)
		{
			TIFFIO.ce = ce;
			TIFFIO.es = es;

			// Initialization code can be put here
		}

		/// <summary>
		/// The @main function is called by Quasar when the library is executed
		/// in standalone mode. For example, through the commandline:
		///
		/// Quasar.exe -debug -gpu TIFFIO.dll
		/// (although I have to check if this use case is currently implemented)
		/// </summary>
		public static void @main(QValue args)
		{
			// Main function - called when the DLL is executed directly.
		}

		/// <summary>
		/// The @cleanup function is called by Quasar usually when the library
		/// is released (for example when the program is terminated, or when the
		/// debug session in Redshift is finished).
		///
		/// At this stage, unmanaged resources can be freed.
		/// </summary>
		public static void @cleanup()
		{
		}

        [Category("File system"),
         Description("Function: tiffread\n" +
            "Reads a TIFF file. By default, all slices are read. The function supports special bit depths, such as 16-bit tiff.\n" +
            "\n" +
            ": function y = tiffread(filename)\n" +
            ": function y = tiffread(filename, slice_range)\n" +
            "\n" +
            "Parameters:\n" +
            "filename - the TIFF file to read\n" +
            "slice_range - selects a ranges of slices to be loaded from the TIFF file\n" +
            "\n" +
            "Returns:\n" +
            "an object with properties `data` (in case of single-slice TIFF files) and `slices` (in case of multi-slice TIFF files)." +
            "Slices is a cell vector for which every element is an image.\n\n" +
            "Example: \n" +
            "import \"TIFFIO.dll\"" +
            "A = tiffread(\"image.tif\")\n" +
            "print numel(A.slices)\n" +
            "imshow(A.slices[0])\n")]
        public static QValue tiffread(QValue fileName)
        {
            return tiffread(fileName, QValue.SliceIndexAll);
        }

        /// <summary>
        /// Loads the specified TIFF file (a specific range of slices)
        /// </summary>
        public static QValue tiffread(QValue fileName, QValue sliceRange)
		{
            int[] multiSliceRange = null;
			// Type checking
			if (fileName.type != QValue.Types.QVT_STRING)
				throw new ArgumentException("(tiffread) Argument 'filename' should be a string!");

            string searchPath = string.Empty;
            string file_name = fileName.ToString();
            // Scan if we can find the file
            if (!Quasar.RuntimeSystem.LocateFile(ref file_name, ref searchPath))
            {
                throw new ApplicationException(string.Format("(tiffread) error while loading bitmap: cannot find file '{0}', " +
                    "despite my efforts by looking in the following dirs:\n{1}!", Path.GetFileName(file_name),
                    searchPath));
            }

            using (Tiff image = Tiff.Open(file_name, "r"))
            {
                if (image == null)
                    throw new InvalidOperationException(string.Format("Could not open TIFF file '{0}'!", file_name));
                int width = image.GetFieldDefaulted(TiffTag.IMAGEWIDTH)[0].ToInt();
                int height = image.GetFieldDefaulted(TiffTag.IMAGELENGTH)[0].ToInt();
                int depth = image.GetFieldDefaulted(TiffTag.IMAGEDEPTH)[0].ToInt();
                int bitsPerSample = image.GetFieldDefaulted(TiffTag.BITSPERSAMPLE)[0].ToInt();
                int samplesPerPixel = image.GetFieldDefaulted(TiffTag.SAMPLESPERPIXEL)[0].ToInt();

                // Create a strcture with the TAG labels
                QObject result = new QObject();

                // Determine the slice ranges
                int dircount = image.NumberOfDirectories();

                switch (sliceRange.type)
                {
                    case QValue.Types.QVT_SCALAR: // Read a single slice
                        multiSliceRange = new int[] { sliceRange.ToInt() };
                        if (multiSliceRange[0] < 0 || multiSliceRange[0] >= dircount)
                            throw new ArgumentException(
                                string.Format("(tiffread) slice index must be between 0 and {1} (I got {2})!", dircount - 1, multiSliceRange[0]));
                        break;
                    case QValue.Types.QVT_MATRIX:
                        Matrix range = (Matrix)sliceRange.value;
                        if (range.numdims != 1)
                            goto default;

                        // The input array may be in float[], in double[] or even in int[], so
                        // we use a generic casting function, because we want the data in int[] format
                        multiSliceRange = (int []) GenericTypes.Cast(range.raw, typeof(int));
                        break;
                    case QValue.Types.QVT_VOID:
                        if (sliceRange.Equals(QValue.SliceIndexAll)) // All slices
                        {
                            if (dircount == 1) // There is only one slice
                                multiSliceRange = new int[] { 1 };
                            else
                                multiSliceRange = Enumerable.Range(0, dircount).ToArray();
                        }
                        break;
                    default:
                        throw new ArgumentException("tiffread) slice index must be either a vector or a scalar!");
                }

                foreach (string name in Enum.GetNames(typeof(TiffTag)))
                {
                    TiffTag tag = (TiffTag)Enum.Parse(typeof(TiffTag), name);

                    if (tag != TiffTag.IGNORE &&
                        image.GetFieldDefaulted(tag) != null)
                    {
                        result[name.ToLower()] = ToQValue(image.GetFieldDefaulted(tag)[0]);
                    }
                }

                // Allocate a cell array for holding the output matrices
                Matrix sliceMtx = ce.MatrixFactory.New(AllocationFlags.None,
                    GenericTypes.NewArray(ce.TypeEnvironment.matrixType, multiSliceRange.Length),
                    new ElemMaxDims(3), 1, multiSliceRange.Length);
                result["slices"] = sliceMtx;

                for (int index = 0; index < multiSliceRange.Length; index++)
                {
                    image.SetDirectory((short) multiSliceRange[index]);

                    FieldValue[] value = image.GetField(TiffTag.ROWSPERSTRIP);
                    if (value == null)
                    {
                        // ROWSPERSTRIP is not set
                        image.SetField(TiffTag.ROWSPERSTRIP, height);
                    }

                    // Allocate the output image
                    Matrix imageMtx = ce.MatrixFactory.Alloc(AllocationFlags.ApplyUserAttributes, ce.FltType,
                        height,
                        width,
                        samplesPerPixel);
                    imageMtx.BeginWrite();

                    // data always points to the first slice
                    if (index == 0)
                    {
                        result["data"] = imageMtx;
                    }

                    // store the slice into the slice matrix
                    sliceMtx.SetAt(index, imageMtx);

                    bool useStripBasedAPI = true;

                    // STRIP-BASED API
                    if (useStripBasedAPI)
                    {
                        int offset = 0;
                        byte[] scanLine = new byte[image.ScanlineSize()];
                        ushort[] usBuf = null;
                        uint[] uiBuf = null;

                        var compression = (Compression)image.GetFieldDefaulted(TiffTag.COMPRESSION)[0].Value;
                        if (compression != Compression.NONE &&
                            !image.IsCodecConfigured(compression))
                        {
                            throw new InvalidOperationException(string.Format("The TIFF compression '{0}' is currently not supported!",
                                compression.ToString()));
                        }

                        switch (bitsPerSample)
                        {
                            case 16:
                                usBuf = new ushort[scanLine.Length / 2];
                                break;
                            case 32:
                                uiBuf = new uint[scanLine.Length / 4];
                                break;
                        }

                        for (int row = 0; row < height; row++)
                        {
                            image.ReadScanline(scanLine, row);

                            switch (bitsPerSample)
                            {
                                case 8: // 8-bit input
                                    GenericTypes.Convert(scanLine, 0, imageMtx.raw, offset, scanLine.Length);
                                    offset += scanLine.Length;
                                    break;
                                case 16: // 16-bit input
                                    Buffer.BlockCopy(scanLine, 0, usBuf, 0, scanLine.Length);
                                    GenericTypes.Convert(usBuf, 0, imageMtx.raw, offset, usBuf.Length);
                                    offset += usBuf.Length;
                                    break;
                                case 32: // 32-bit input
                                    Buffer.BlockCopy(scanLine, 0, uiBuf, 0, scanLine.Length);
                                    GenericTypes.Convert(uiBuf, 0, imageMtx.raw, offset, uiBuf.Length);
                                    offset += uiBuf.Length;
                                    break;
                                default:
                                    throw new InvalidOperationException(string.Format(
                                        "The number of bits/sample {0} is not supported!", bitsPerSample));
                            }
                        }
                    }

                    // TILE-BASED API
                    if (!useStripBasedAPI)
                    {
                        int tileWidth = image.GetField(TiffTag.TILEWIDTH)[0].ToInt();
                        int tileLength = image.GetField(TiffTag.TILELENGTH)[0].ToInt();

                        byte[] buf = new byte[image.TileSize()];
                        ushort[] usBuf = null;
                        uint[] uiBuf = null;

                        switch (bitsPerSample)
                        {
                            case 8: break;
                            case 16: usBuf = new ushort[buf.Length / 2]; break;
                            case 32: uiBuf = new uint[buf.Length / 4]; break;
                        }

                        for (int y = 0; y < height; y += tileLength)
                        {
                            for (int x = 0; x < width; x += tileWidth)
                            {
                                image.ReadTile(buf, 0, x, y, 0, 0);

                                // TODO
                            }
                        }
                    }
                    CoreLibrary.Release(imageMtx);
                }
                CoreLibrary.Release(sliceMtx); // We are not holding a reference anymore
                return result;
            } // Using
		}

        /// <summary>
        /// Saves the specified TIFF file
        /// </summary>
        [QObject.MemberVisible,
         Category("File system"),
         Description("Function: tiffwrite\n" +
            "Creates a TIFF file and writes data to it.\n" +
            "\n" +
            ": function y = tiffwrite(filename, data)\n" +
            "\n" +
            "Parameters:\n" +
            "filename - the TIFF file to write\n" +
            "data - an object (as returned by `tiffread`), with a `data` or `slices` field. Note that additional parameters, such as `bits_per_sample` may need to be set.\n" +
            "\n" +
            "Example: \n" +
            "import \"TIFFIO.dll\"" +
            "im = imread(\"image.tif\")" +
            "s = vec[mat](size(im, 2))" +
            "for i = 0..size(im, 2) - 1" +
            "    s[i] = im[:,:, i]" +
            "end" +
            "tiffwrite(\"image_slices.tif\",{slices:=s, bitspersample:=8})")]
        public static void tiffwrite(string fileName, QValue qValue)
        {
            QObject data;
            switch (qValue.type)
            {
                case QValue.Types.QVT_UNTYPEDOBJECT:
                    data = (QObject) qValue.value;
                    data.Addref();
                    break;
                case QValue.Types.QVT_MATRIX:
                    data = new QObject();
                    data["data"] = qValue;
                    break;
                default:
                    throw new ArgumentException("(tiffwrite) Matrix or object expected as second argument!");
            }

            try
            {

                using (Tiff image = Tiff.Open(fileName, "w"))
                {
                    if (image == null)
                        throw new InvalidOperationException(string.Format("Could not open TIFF file '{0}'!", fileName));

                    // Set the TIFF fields
                    foreach (string name in Enum.GetNames(typeof(TiffTag)))
                    {
                        QValue qvalue;
                        TiffTag tag = (TiffTag)Enum.Parse(typeof(TiffTag), name);
                        if (tag != TiffTag.IGNORE &&
                            data.TryGetValue(name.ToLower(), out qvalue))
                        {
                            image.SetField(tag, FromQValue(qvalue, image.GetFieldDefaulted(tag)[0]));
                        }
                    }

                    Matrix sliceMtx = null;
                    Matrix imageMtx = null;
                    QValue qv;
                    if (data.TryGetValue("slices", out qv))
                    {
                        sliceMtx = (Matrix)qv.value;
                        sliceMtx.Addref();
                    }
                    if (data.TryGetValue("data", out qv))
                    {
                        imageMtx = (Matrix)qv.value;
                        // No Addref here!!
                    }

                    try
                    {
                        // Slice matrix not created yet
                        if (sliceMtx == null)
                        {
                            Array elems = Array.CreateInstance(imageMtx.GetType(), 1);
                            elems.SetValue(imageMtx, 0);
                            imageMtx.Addref();
                            sliceMtx = ce.MatrixFactory.New(AllocationFlags.None, elems, 1, 1);
                        }

                        var compression = (Compression)image.GetFieldDefaulted(TiffTag.COMPRESSION)[0].Value;
                        if (compression != Compression.NONE &&
                            !image.IsCodecConfigured(compression))
                        {
                            throw new InvalidOperationException(string.Format("The TIFF compression '{0}' is currently not supported!",
                                compression.ToString()));
                        }

                        // Use the strip-based API
                        for (int index = 0; index < sliceMtx.numel; index++)
                        {
                            int offset = 0;
                            ushort[] usBuf = null;
                            uint[] uiBuf = null;

                            if (index != 0)
                            {
                                // save previous directory data
                                image.WriteDirectory();

                                // create new directory and make it current
                                image.CreateDirectory();
                                image.SetDirectory((short)index);
                            }

                            imageMtx = (Matrix)sliceMtx.GetAt(index);

                            int width = GetOrSetValue(image, TiffTag.IMAGEWIDTH, () => imageMtx.dim2);
                            int height = GetOrSetValue(image, TiffTag.IMAGELENGTH, () => imageMtx.dim1);
                            int channels = GetOrSetValue(image, TiffTag.SAMPLESPERPIXEL, () => imageMtx.dim3);
                            int bitsPerSample = GetOrSetValue(image, TiffTag.BITSPERSAMPLE, () => 8);
                            image.SetField(TiffTag.PLANARCONFIG, PlanarConfig.CONTIG);
                            if (sliceMtx.numel > 0)
                                image.SetField(TiffTag.SUBFILETYPE, FileType.PAGE);
                            image.CheckpointDirectory();

                            byte[] scanLine = new byte[image.ScanlineSize()];

                            if (imageMtx.dim1 != height ||
                                imageMtx.dim2 != width ||
                                imageMtx.dim3 != channels)
                            {
                                throw new InvalidOperationException(
                                    string.Format("Matrix dimensions for slice {0} are incorrect (expected: {1}x{2}x{3}, got: {4}x{5}x{6})!",
                                    index,
                                    height,
                                    width,
                                    channels,
                                    imageMtx.dim1,
                                    imageMtx.dim2,
                                    imageMtx.dim3));
                            }

                            switch (bitsPerSample)
                            {
                                case 16:
                                    usBuf = new ushort[scanLine.Length / 2];
                                    break;
                                case 32:
                                    uiBuf = new uint[scanLine.Length / 4];
                                    break;
                            }

                            for (int row = 0; row < height; row++)
                            {
                                switch (bitsPerSample)
                                {
                                    case 8: // 8-bit input
                                        GenericTypes.Convert(imageMtx.raw, offset, scanLine, 0, scanLine.Length);
                                        offset += scanLine.Length;
                                        break;
                                    case 16: // 16-bit input
                                        GenericTypes.Convert(imageMtx.raw, offset, usBuf, 0, usBuf.Length);
                                        Buffer.BlockCopy(usBuf, 0, scanLine, 0, scanLine.Length);
                                        offset += usBuf.Length;
                                        break;
                                    case 32: // 32-bit input
                                        GenericTypes.Convert(imageMtx.raw, offset, uiBuf, 0, uiBuf.Length);
                                        Buffer.BlockCopy(uiBuf, 0, scanLine, 0, scanLine.Length);
                                        offset += uiBuf.Length;
                                        break;
                                    default:
                                        throw new InvalidOperationException(string.Format(
                                            "The number of bits/sample {0} is not supported!", bitsPerSample));
                                }
                                image.WriteScanline(scanLine, row);
                            }

                            imageMtx.Release();
                        }
                    }
                    finally
                    {
                        if (sliceMtx != null) sliceMtx.Release();

                    }
                }
            }
            finally
            {
                data.Release();
            }
        }

        private static int GetOrSetValue(Tiff image, TiffTag tag, Func<int> defaultFunc)
        {
            FieldValue[] values = image.GetField(tag);
            if (values == null ||
                values.Length == 0 ||
                (tag == TiffTag.IMAGELENGTH && values[0].ToInt() == 0))
            {
                int value = defaultFunc();
                image.SetField(tag, value);
                return value;
            }
            return values[0].ToInt();
        }

        /// <summary>
        /// Converts a TIFF field to QValue
        /// </summary>
        private static QValue ToQValue(FieldValue fieldValue)
        {
            var typeCode = Type.GetTypeCode(fieldValue.Value.GetType());

            if (fieldValue.Value is Compression)
            {
                return new QValue(((Compression)fieldValue.Value).ToString());
            }
            else if (typeCode == TypeCode.Int16 ||
                typeCode == TypeCode.Int32 ||
                typeCode == TypeCode.Int64 ||
                typeCode == TypeCode.UInt16 ||
                typeCode == TypeCode.UInt32 ||
                typeCode == TypeCode.UInt64)
            {
                return new QValue(fieldValue.ToInt());
            }
            else if (typeCode == TypeCode.Double ||
                     typeCode == TypeCode.Single)
            {
                return new QValue(fieldValue.ToDouble());
            }
            else if (fieldValue.Value is string)
            {
                return new QValue(fieldValue.ToString());
            }
            return QObject.Null;
        }

        private static object FromQValue(QValue qv, FieldValue defaultValue)
        {
            var destType = defaultValue.Value.GetType();
            var typeCode = Type.GetTypeCode(destType);

            if (defaultValue.Value is Compression)
            {
                return Enum.Parse(typeof(Compression), (string) qv.value);
            }
            else if (typeCode == TypeCode.Int16 ||
                typeCode == TypeCode.Int32 ||
                typeCode == TypeCode.Int64 ||
                typeCode == TypeCode.UInt16 ||
                typeCode == TypeCode.UInt32 ||
                typeCode == TypeCode.UInt64)
            {
                return Convert.ChangeType(qv.value, destType);
            }
            else if (typeCode == TypeCode.Double ||
                     typeCode == TypeCode.Single)
            {
                return Convert.ChangeType(qv.value, destType);
            }
            else if (defaultValue.Value is string)
            {
                return (string) qv.value;
            }
            return null;
        }
	}
}

