%
% mandelbrot.q
%
% Bart Goossens, Aug 2012
{!author name="Bart Goossens"}
{!doc category="Samples/Misc/Mandelbrot fractal"}
{!title name="Mandelbrot fractal demonstration"}

% Section: Overview
%
% Computation of the mandelbrot fractal, using complex arithmetic.
% Note - the best zoom is obtained when running the program in
% DOUBLE floating point precision mode (-double parameter).
% 
% More information:
%
% In Quasar, the internal representation of scalar numbers "scalar"
% (or complex scalar numbers "cscalar"), is not specified at the 
% code level. This allows the floating point representation to be 
% changed on a global level. By default, Quasar will use single 
% precision floating point numbers.
% However, it is possible to compile and run the programs using 
% double precision as well, by passing the -double command line 
% option to Quasar, e.g.:
% 
% :./Quasar.exe -debug -double script.q
% 
% Overview of floating point representations
% :                      || IEEE 32-bit (single precision) |   IEEE 64-bit (double precision)   
% :----------------------++------------------------------------+------------------------------------
% :      Significand     ||              23 bits           |              52 bits               
% :       Exponent       ||               8 bits           |              11 bits               
% :  Minimum pos. value  ||      1.17549435E-38            |  2.225073858507201\times10^{-308}  
% :  Maximum pos. value  ||      3.40282347E-38            |  1.797693134862316\times10^{308}   
% :  Exact integer repr. || -2^24+1 to 2^24-1 (16,777,215) |       -2^{53}-1 to 2^{53}-1        
% 
% For performance reasons, it is recommended to use single 
% precision. Note that some older GPUs have limited double 
% precision FP support. CUDA devices before compute capability 1.3 
% even do not have double precision FP support. Moreover, using 
% double precision FP numbers doubles the memory bandwidth. 
% Consequently, programs using double precision may run up to 2x 
% slower than programs with single precision FP. However, there are 
% some reasons to enable double precision in Quasar programs:
% 
% * When numerical accuracy is an issue: remark that results 
%   obtained using double-precision arithmetic may differ from the 
%   same operations obtained using single-precision arithmetic, due 
%   to the greater precision and due to rounding errors. Therefore, 
%   it is important to compare and express the results within a 
%   certain tolerance rather than expecting them to be exact. 
%   Moreover, GPU devices typically flush numbers smaller than the 
%   minimum representable value (in absolute sense) to zero. 
%   Correspondingly, by using double precision FP numbers it may be 
%   possible to reduce some of the error introduced by underflow, 
%   as the minimal representable value is of the order 10^{-308} 
%   for double, while 10^{-38} for single precision. 
% 
% * For comparing the results of the algorithms to MATLAB/C++ 
%   implementations using double precision FP numbers.
% 
% Often it is useful to check whether the program is not suffering 
% from floating point inaccuracies. This can simply be done by 
% running the program once in double precision mode.
% 
% Note: NVidia GPU's GTX 260, 275, 280, 285, 295 chips (with 
% compute capability 1.3) have a low performance in double 
% precision computations (about 1/8 of single precision 
% performance). Devices of the NVidia Fermi architecture (compute 
% capability 2.0+) have 1/2 the performance of single precision 
% operations.
% 
% Finally, recall that FP math is not associative, i.e. the sum (A+B)+C
%  is not guaranteed to be equal to A+(B+C). When parallelizing 
% computations, the order of the operations is often changed (and 
% may be even unspecified), leading to results which may differ 
% each time the technique runs even with the same input data. This 
% limitation is not inherent to Quasar, but applies to all 
% approaches that perform parallel computations using floating 
% point numbers. 

import "Quasar.UI.dll"
import "colormap.q"

function [] = __kernel__ mandelbrot_fractal(im : mat'unchecked, log2_s : scalar, t : cscalar, theta : scalar, num_it : int, pos : ivec2)
    p = (float(pos) ./ size(im,0..1))-0.5
    %c = t*2^(-0.5*log2_s)+2^(0.5*log2_s)*complex(p[1],p[0])*exp(1j*theta)
    c = complex(p[1],p[0])*exp(1j*theta)
    z = 0i
    N = 2.0
    r = max(0,-log2_s)
    for n=1..num_it
        if abs(z)>N
            break
        endif
        
        %z = (2^r*z*z + 2^(0.5 * log2_s-r)* c)
        z = (2^r*z*z + 2^(log2_s-r)* c + 2^(-r)*t)
        % update r
        %r_new = max(0,log2(abs(z))-10)
        %z = 2^(r-r_new)*z
        %r = r_new
    end
    % The first log should be log2
    im[pos] = n-log2((log2(abs(z)+r)/log(N)))
end

x = zeros(768,768)
%colormap("cmy")
colormap("hotter")
%colormap("hsv")
state = object()
state.t = complex(-1.42)
state.theta = 0
state.log2_s = log2(10)
state.auto_zoom = true
max_order = 20

% A simple GUI...
frm = form("Quasar - Fractal demonstration")
slider = frm.add_slider("Navigation point",1.42,0,2)
slider.onchange.add(() -> state.t = complex(-slider.value))
slider = frm.add_slider("Rotation",0,0,2*pi)
slider.onchange.add(() -> state.theta = slider.value)
slider = frm.add_slider("Zoom",0,0,2+max_order)
slider.onchange.add(() -> state.log2_s = (2-slider.value))
checkbox = frm.add_checkbox("Auto zoom",true)
checkbox.onchange.add(() -> state.auto_zoom = checkbox.value)

frm.width = 840
frm.move(0,0)
disp = frm.add_display()
disp.set_aspect(1.0)
frm.show()
log2_s = log2(10)
log2_f = log2(1.005)
num_it = 1024
tf = x -> x/256
%tf = x -> periodize(0.25*log(x+1.0),1.0)

tic()
frames = 0
while (!frm.closed())
    if state.auto_zoom
        if state.log2_s < log2(1/10^max_order)
            log2_f = -log2_f
        elseif state.log2_s > 100
            log2_f = -log2_f
        endif
        state.log2_s = state.log2_s - log2_f
        slider.value = 2-state.log2_s
    endif
    parallel_do(size(x),x,state.log2_s,state.t,state.theta,num_it,mandelbrot_fractal)  
    disp.imshow(cmap(tf(x), current_color_map.data,[0,1]),[0,255])   
    pause(0)
    frames += 1
end
time = toc()
print frames/(time)*1000, " FPS"
