{!author name="Bart Goossens"}
{!doc category="Samples/Parallelization/Find index of the minimum of an array"}
{!title name="Find the index of the minimum of an array (argmin)"}

% Section: Overview
%
% Calculating the index of the minimum of an array is a task that is quite
% simple on a sequential CPU. A parallel algorithm for this task is a little
% bit more comprehensive. The implementation below is based on a sequential
% search combined with a parallel reduction algorithm. Instead of only 
% calculating the minimum of the array, we simultaneously keep track of the 
% index for which this minimum is attained.
% 
% Note: this implementation can also be found in the parallelblocks.q library
% under the name find_index_max. Note that
% find_index_min(x) = find_index_max(-x)

function y = find_index_min(x : cube{n})

    % We use a specific algorithm for GPU
    function [] = __kernel__ parallel_kernel(x : vec'unchecked, sz : int, xv : vec, 
        yv : vec, blkdim : int, blkpos : int, blkidx : int, pos : int)
        
        {!kernel target="gpu"}        
        bins : vec'unchecked = shared(blkdim)
        indices : vec[int]'unchecked = shared[int](blkdim)

        ind = pos
        val = x[ind]
        nblocks = int((numel(x) + sz-1) / sz)

        for n=1..nblocks-1
            i=pos+n*sz
            if i<numel(x) && val > x[i]
                val = x[i]
                ind = i
            endif
        endfor
        indices[blkpos] = ind
        bins[blkpos] = val
        syncthreads
        bit = 1
        while bit<blkdim
            if and(blkpos,2*bit-1) == 0 && blkpos+bit<blkdim
                [t1,t2] = [bins[blkpos], bins[blkpos+bit]]
                if t1>t2 || (t1 == t2 && indices[blkpos] > indices[blkpos+bit])
                    bins[blkpos] = t2
                    indices[blkpos] = indices[blkpos+bit]
                endif
            endif
            syncthreads
            bit *= 2
        endwhile
        if blkpos == 0      
            xv[blkidx] = indices[0]
            yv[blkidx] = bins[0]
        endif
    endfunction
    
    % And another algorithm for CPU
    function y = __kernel__ serial_kernel(x : vec)
        {!kernel target="cpu"}
        y = 0
        minval = x[0]
        for i=1..numel(x)-1
            if minval>x[i]
                minval=x[i]
                y = i
            endif
        endfor
    endfunction
    
    N = numel(x)
    % Invoke the runtime scheduler to determine whether to run the parallel
    % or the serial kernel
    match schedule(N, x, {parallel_kernel, serial_kernel}) with
    | 0 ->        
        % Now we calculate the best suited block size and block count for 
        % executing this kernel. opt_block_cnt uses both properties of
        % the kernel as well as characteristics of the GPU to determine these
        % two numbers.
        [block_cnt, block_size] = opt_block_cnt(parallel_kernel, 0, N)
        xv = zeros(prod(block_cnt))
        {!alloc mode="gpu";type="pinnedhostmem"}
        yv = uninit(prod(block_cnt))
        sz = prod(block_size .* block_cnt)
        parallel_do([block_size.*block_cnt, block_size], x, sz, xv, yv, parallel_kernel)
        % We still need to find the minimum for each block. This is achieved using the
        % serial kernel. To avoid a separate cuMemcpyD2H operation, we use pinned host 
        % memory and write from the previous kernel directly to the system memory.
        y = xv[serial_do(1, yv, serial_kernel)]
    | 1 ->
        y = serial_do(1, x, serial_kernel)
    endmatch
endfunction

x = randn(65536*256)
N = 1000
tic()
for k=0..N-1
    y = find_index_min(x)
endfor
toc()

assert(x[y] == min(x), "result incorrect")
