{!author name="Bart Goossens"}
{!doc category="Samples/Parallelization/Find Volume Minimum point"}
{!title name="Sample implementation of non-local minimum suppression and minimum point calculation"}

% Finding the minimum point of a volume
% 
import "system.q"
import "parallelblocks.q"
import "imfilter.q"

% Function: suppress_non_local_maxima
% 
% Suppresses non local maxima of the specified matrix
% 
% Usage:
%   : function [y : mat] = suppress_non_local_maxima(x : mat'const, half_wnd : int)
% 
% Parameters:
% x - the input matrix
% half_wnd - half the size of the window. The exact window size is (2H+1 x 2H+1)
% 
% Returns:
% y - the output image containing the remaining local maxima
function y = suppress_non_local_maxima(x : mat, half_wnd : int)

    z = uninit(size(x))
    y = uninit(size(x))
    
    % Vertical filter
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            
            max_val = minvalue(scalar)
            for k=-half_wnd..half_wnd
                max_val = max(max_val, x[m,n + k])
            endfor
            
            z[m,n] = max_val
        endfor
    endfor
    
    % Horizontal filter
    for m=0..size(x,0)-1
        for n=0..size(x,1)-1
            
            max_val = minvalue(scalar)
            for k=-half_wnd..half_wnd
                max_val = max(max_val, z[m + k,n])
            endfor
            
            y[m,n] = max_val == x[m,n] ? x[m,n] : 0.0
        endfor
    endfor

endfunction

% Function: find_minimum_points
% 
% Finds the coordinates of the voxels with minimum z-coordinate
% 
% Usage:
%   : function [points] = find_minimum_points(voxels : cube'const)
% 
% Parameters:
% voxels - an input voxel cube
% 
% Returns:
% points - a matrix of size Nx3 containing the coordinates
function points = find_minimum_points(voxels : cube)
    activity = squeeze(sum(sum(voxels,1),0))
    slice_index = find(activity, "first")
    slice = voxels[:,:,slice_index[0]]
    mask = cat((0..5), (4..-1..0))
    slice = imfilter(slice, transpose(mask)*mask)
    
    imshow(slice);title("Before local maximum suppression")
    slice = suppress_non_local_maxima(slice,3)
    imshow(slice);title("After local maximum suppression")
    result = find_indices_nonzero(slice!=0)
    points = zeros(numel(result),3)
    for i=0..numel(result)-1
        p = ind2pos(size(slice), result[i])
        points[i,0..2] = [p[0], p[1], slice_index[0,1]]
    endfor
endfunction

function [] = main()
    N = 128
    [x,y,z] = meshgrid(-N..N-1,-N..N-1,-N..N-1)

    voxels = ((x-32).^2 + (y-64).^2 - 4* z < 0) + 
             ((x+20).^2 + (y+80).^2 - 4* z < 0)

    print find_minimum_points(voxels)

endfunction
