{!author name="Bart Goossens"}
{!doc category="Samples/Parallelization/Integral image"}
{!title name="Integral image implementation via automatic parallelization"}

% Section: Overview
%
% This demo program showcases how to calculate an integral image, via the
% automatic parallelization techniques in Quasar. In essence, the integral
% image can be calculated simply by two cumulative sums: one in the first 
% dimension, and a second one in the second dimension. This is readily 
% achieved using the cumsum() function which has a second argument to 
% specify the dimension. Here, we want to achieve the same by writing the
% for-loops by hand. Because the cumulative sum is sequential, it seems not
% obvious how to parallelize such loop. Luckily the Quasar compiler can
% recognize this pattern and maps it automatically onto a parallel prefix
% sum algorithm (an algorithm that has been designed for this purpose).
% The added flexibility is that the loops can be more complex, they can
% contain additional statements/calculations. As long as the pattern is
% not broken, the whole loop will parallelize and still be mapped onto the
% parallel prefix sum algorithm.
% 
function [] = integral_image_gray(x : mat, y : mat)
    z = zeros(size(x))

    {!parallel for}
    for [m,n]=0..size(y,0..1)-1
        z[m,n] = z[m,n-1] + x[m,n]
    endfor
    
    {!parallel for}
    for [m,n]=0..size(y,0..1)-1
        y[m,n] = y[m-1,n] + z[m,n]
    endfor
endfunction

function [] = integral_image_rgb(x : cube, y : cube)
    z = zeros(size(x))

    {!parallel for}
    for [m,n,p]=0..size(y,0..2)-1
        z[m,n,p] = z[m,n-1,p] + x[m,n,p]
    endfor
    
    {!parallel for}
    for [m,n,p]=0..size(y,0..2)-1
        y[m,n,p] = y[m-1,n,p] + z[m,n,p]
    endfor
endfunction

function [] = main()

    im = imread("lighthouse.png")
    im_gray = 0.2126 * im[:,:,0] + 0.7152 * im[:,:,1] + 0.0722 * im[:,:,2]
    im_int = uninit(size(im_gray))

    %
    % Grayscale
    %    
    runs = 1000
    tic()
    for run=1..runs
        integral_image_gray(im_gray, im_int)
    endfor
    toc("integral_image_gray")

    tic()
    for run=1..runs
        im_int_cumsum = cumsum(cumsum(im_gray, 0), 1)
    endfor
    toc("cumsum_gray")

    assert(mean(abs(im_int - im_int_cumsum)) < 1, "Error in calculations")

    %
    % RGB
    % 
    im_int = uninit(size(im))
    tic()
    for run=1..runs
        integral_image_rgb(im, im_int)
    endfor
    toc("integral_image_rgb")

    tic()
    for run=1..runs
        im_int_cumsum = cumsum(cumsum(im, 0), 1)
    endfor
    toc("cumsum_rgb")

    assert(mean(abs(im_int - im_int_cumsum)) < 1, "Error in calculations")

endfunction
