{!author name="Bart Goossens"}
{!doc category="Samples/Parallelization/4x4 matrix parallel reduction"}
{!title name="Parallel reduction algorithm for small matrices"}

% Section: Overview
%
% In some algorithms, it is useful to aggregate small matrices that
% are calculated in each iteration of a big for-loop. The following example
% shows a loop that, once parallelized, is mapped onto a parallel reduction
% algorithm and implemented using a combination of shared memory and 
% warp shuffling operations. By using warp shuffling, the amount of shared
% memory can vastly be reduced (by a factor 32, the warp size), compared to
% a parallel reduction algorithm that uses no warp shuffling. This is of
% importance when the matrix size increases. For 4x4 matrices and 32-bit
% floating point numbers, 1024 bytes of shared memory are used.
% 
% When the purpose is to only calculate a sum of matrices, the partial
% parallel reduction can be used (see other example), however, if the
% algorithm performs additional computations and in case each thread needs
% to perform its own specific function, it might be useful to use this 
% approach.


function A = sum_matrices[K,M](data : cube'unchecked(:,K,M), N : int)
    A = zeros(K, M)
    {!parallel for}
    for m=0..N-1
        val = data[m,:,:]
        A += val
    endfor
endfunction

function [] = main()
    N = 80000
    data = rand(N, 4, 4)
    runs = 1000

    tic()
        for run=0..runs-1
            A = sum_matrices(data, N)
        endfor
    toc()
    
    print A
    print squeeze(sum(data, 0))
    
    % Note due to parallelization, the order of the floating point operations is not guaranteed,
    % so different ways of calculating the same result may have different errors
    assert(sum(abs(A-squeeze(sum(data, 0))))<1, "Result incorrect")

endfunction