{!author name="Bart Goossens"}
{!doc category="Samples/Parallelization/Parallel histogram"}
{!title name="Parallel calculation of an image histogram"}


%
% Section: Overview
% 
% This sample illustrates how a histogram can be calculated in parallel

% Direct histogram implementation
function y = hist_direct(im : cube)
    y = zeros(256)
    
    for m=0..size(im,0)-1
        for n=0..size(im,1)-1
            for k=0..size(im,2)-1                
                v = im[m,n,k]
                y[v] += 1
            endfor
        endfor
    endfor

endfunction

% Optimized histogram implementation
% we use the sharedmemcaching transform, to store the intermediate results in shared memory. This 
% is performed in combination of a kernel tiling pass, which splits the calculation in tiles of
% 128x128. 
% 
function y : vec'unchecked = hist_optimized_sharedmemcaching(im)
    y = zeros(256)
    
    {!parallel for}
    for m=0..size(im,0)-1
        for n=0..size(im,1)-1
            for k=0..size(im,2)-1
                % Enable the shared memory caching transform
                {!kernel_transform enable="sharedmemcaching"}
                % Cache the variable y in shared memory, for both reading/writing (inout) and for
                % atomic operation +=.
                {!kernel_arg name=y; type="inout"; access="shared"; op="+="; cache_slices=y[:]; numel=256}
                {!kernel_tiling dims=[128,128,1]; mode="global"; target="gpu"}                
                v = int(im[m,n,k])
                y[v] += 1
            endfor
        endfor
    endfor

endfunction

%
% Optimized implementation using shared memory designators (fully automatic)
%
function y : vec'unchecked(256) = hist_optimized_sharedmemdesignators1(im : cube)
    y = zeros(256)
    {!function_transform enable="sharedmempromotion"}
    
    {!parallel for}
    for m=0..size(im,0)-1
        for n=0..size(im,1)-1
            for k=0..size(im,2)-1
                v = int(im[m,n,k])
                y[v] += 1
            endfor
        endfor
    endfor

endfunction

%
% Optimized implementation using shared memory designators (semi-automatic)
% We use the 'shared designator to indicate that the variable needs to be stored
% in shared memory. All vector/matrix operations on this variable will be
% parallelized and distributed over the available threads. Next, we use
% grid groups to separate position-dependent parts of the algorithm from
% position-independent parts of the algorithm. For example, the initialization
% of the shared memory and the write back only needs to be performed once per
% group of blocks. The runtime will launch the GPU kernel with the maximal
% number of concurrent blocks (and active warps) that the GPU supports.
%
function y : vec'unchecked(256) = hist_optimized_sharedmemdesignators2(im : cube)
    y = zeros(256)
    
    {!parallel for}
    for m=0..size(im,0)-1
        for n=0..size(im,1)-1
            for k=0..size(im,2)-1
                y_sh : 'shared = zeros(256)
                {!begin_grid_group}
                v = int(im[m,n,k])
                y_sh[v] += 1
                {!end_grid_group}
                syncthreads
                y[:] += y_sh
            endfor
        endfor
    endfor

endfunction

function y : vec'unchecked(256) = hist_optimized_sharedmemdesignators3(im : cube)
    y = zeros(256)
    
    {!parallel for}
    for m=0..size(im,0)-1
        for n=0..size(im,1)-1
            for k=0..size(im,2)-1
                y_sh : 'shared = zeros(256,3)
                {!begin_grid_group}
                v = int(im[m,n,k])
                y_sh[v,k] += 1
                {!end_grid_group}
                syncthreads
                %y[:] += y_sh[:,0] + y_sh[:,1] + y_sh[:,2] % sum(y_sh, 1)
                y[:] += sum(y_sh,1)
            endfor
        endfor
    endfor

endfunction

function [] = main()

    im = imread("lena_big.tif")
    num_runs = 1000
    
    tic()
    for run=0..num_runs-1
        y = hist_direct(im)
    endfor
    toc("Direct implementation")
    assert(sum(y)==numel(im), "The result of the direct implementation is incorrect!")
    
    tic()
    for run=0..num_runs-1
        y = hist_optimized_sharedmemcaching(im)
    endfor
    toc("Optimized algorithm (shared mem caching)")
    assert(sum(y)==numel(im), "The result of the optimized algorithm is incorrect!")

    tic()
    for run=0..num_runs-1
        y = hist_optimized_sharedmemdesignators1(im)
    endfor
    toc("Optimized algorithm (shared mem designators 1)")
    assert(sum(y)==numel(im), "The result of the optimized algorithm is incorrect!")
    
    tic()
    for run=0..num_runs-1
        y = hist_optimized_sharedmemdesignators3(im)
    endfor
    toc("Optimized algorithm (shared mem designators 2)")
    print sum(y),",",numel(im)
    assert(sum(y)==numel(im), "The result of the optimized algorithm is incorrect!")
endfunction
