#
# This sample project shows how to implement custom PyTorch 
# neural network layers in Quasar. It is an adjusted version
# of the PyTorch custom layer Relu example
#
# Sample requires the following Python packages:
# - pytorch
# - pyquasar
# - pycuda

import torch
import pycuda.autoinit
import pycuda.driver as cuda
import pyquasar as q
import numpy as np
from pytorch_helpers import from_tensor, to_tensor

quasar_src = """
function y = relu_forward(x : cube{4})
    {!function_transform enable="kernel_generator"; threshold=0}
    y = x .* (x >= 0)
endfunction

function y = relu_backward(x : cube{4})
    {!function_transform enable="kernel_generator"; threshold=0}
    y = x .* (x >= 0)
endfunction
"""
target_device = "cuda"  # Set to CPU to run on the CPU
#target_device = "cpu"

# Initialize Quasar
q.init(target_device,load_compiler=True)

# Initialize PyTorch
dtype = torch.float32
device = torch.device(target_device) # Uncomment this to run on GPU


class MyReLU(torch.autograd.Function):
    """
    We can implement our own custom autograd Functions by subclassing
    torch.autograd.Function and implementing the forward and backward passes
    which operate on Tensors.
    """

    @staticmethod
    def forward(ctx, input):
        """
        In the forward pass we receive a Tensor containing the input and return
        a Tensor containing the output. ctx is a context object that can be used
        to stash information for backward computation. You can cache arbitrary
        objects for use in the backward pass using the ctx.save_for_backward method.
        """
        ctx.save_for_backward(input)
        return to_tensor(q.relu_forward(from_tensor(input)), device)
        #return input.clamp(min=0)

    @staticmethod
    def backward(ctx, grad_output):
        """
        In the backward pass we receive a Tensor containing the gradient of the loss
        with respect to the output, and we need to compute the gradient of the loss
        with respect to the input.
        """
        input, = ctx.saved_tensors
        #grad_input = grad_output.clone()
        #grad_input[input < 0] = 0
        grad_input = to_tensor(q.relu_backward(from_tensor(grad_output)), device)
        return grad_input


A = q.eye(32)
print(np.array(to_tensor(A, device).cpu()))
print(from_tensor(torch.ones([4,4], dtype=torch.float32, device=device)))

# Compile the Quasar code
q.compile_string("layer", quasar_src)

# N is batch size; D_in is input dimension;
# H is hidden dimension; D_out is output dimension.
N, D_in, H, D_out = 64, 1000, 100, 10

# Create random Tensors to hold input and outputs.
x = torch.randn(N, D_in, device=device, dtype=dtype)
y = torch.randn(N, D_out, device=device, dtype=dtype)

# Create random Tensors for weights.
w1 = torch.randn(D_in, H, device=device, dtype=dtype, requires_grad=True)
w2 = torch.randn(H, D_out, device=device, dtype=dtype, requires_grad=True)

learning_rate = 1e-6
for t in range(500):
    # To apply our Function, we use Function.apply method. We alias this as 'relu'.
    relu = MyReLU.apply

    # Forward pass: compute predicted y using operations; we compute
    # ReLU using our custom autograd operation.
    y_pred = relu(x.mm(w1)).mm(w2)

    # Compute and print loss
    loss = (y_pred - y).pow(2).sum()
    if t % 100 == 99:
        print(t, loss.item())

    # Use autograd to compute the backward pass.
    loss.backward()

    # Update weights using gradient descent
    with torch.no_grad():
        w1 -= learning_rate * w1.grad
        w2 -= learning_rate * w2.grad

        # Manually zero the gradients after updating weights
        w1.grad.zero_()
        w2.grad.zero_()

# Destroys the Quasar host
q.destroy()
