# pytorch_helpers.py 
#
# Helper functions for interacting with PyTorch from Quasar. Conversion of 
# Quasar matrices to PyTorch tensors and back.

import torch
import pycuda.autoinit
import pycuda.driver as cuda
import pyquasar as q
import numpy as np

def to_tensor(A, device):
    """
    Converts a Quasar matrix to PyTorch Tensor
    """
    assert q.type(A, "[scalar|cube{:}]"), "A : cube{:} or A : scalar expected"
    
	# input is scalar, convert to vector of length 1
	# TODO - may be done more efficiently by directly invoking the PyTorch functions
    if q.type(A, "scalar"): 
        tmp = A				
        A = q.zeros(1)
        A[0] = tmp

    if device.type=="cpu":
        tensor = torch.tensor((), dtype=torch.float32, device="cpu").new_zeros(tuple(q.size(A).to_py().astype(int)))
        storage = tensor.storage().cpu()
        ptr = A.lock(q.LOCK_READ, q.MEMRESOURCE_CPU)
        q.memcpy(storage.data_ptr(), ptr, storage.size() * storage.element_size())
        A.unlock(q.LOCK_READ, q.MEMRESOURCE_CPU)
    else:
        tensor = torch.tensor((), dtype=torch.float32, device=device).new_zeros(tuple(q.size(A).to_py().astype(int)))
        storage = tensor.storage().cuda()
        ptr = A.lock(q.LOCK_READ, q.MEMRESOURCE_SINGLE_CUDA)
        cuda.memcpy_dtod(storage.data_ptr(), ptr, storage.size() * storage.element_size())
        A.unlock(q.LOCK_READ, q.MEMRESOURCE_SINGLE_CUDA)

    return tensor

def from_tensor(A):
    """
    Converts a PyTorch tensor to a Quasar matrix
    """
    assert torch.is_tensor(A) and A.dtype == torch.float32, "torch.FloatTensor expected!"

    mtx = q.uninit(list(A.size()))
    storage = A.storage()
    if storage.is_cuda:
        ptr = mtx.lock(q.LOCK_WRITE, q.MEMRESOURCE_SINGLE_CUDA)
        cuda.memcpy_dtod(ptr, storage.data_ptr(), storage.size() * storage.element_size()) 
        mtx.unlock(q.LOCK_WRITE, q.MEMRESOURCE_SINGLE_CUDA)
    else:
        storage = storage.cpu()
        ptr = mtx.lock(q.LOCK_WRITE, q.MEMRESOURCE_CPU)
        q.memcpy(ptr, storage.data_ptr(), storage.size() * storage.element_size())
        mtx.unlock(q.LOCK_WRITE, q.MEMRESOURCE_CPU)
        
    if q.numel(mtx)==1:
        return mtx[0]  # convert to scalar
    else:
        return mtx



