{!author name="Bart Goossens"}
{!doc category="Samples/Quasar tricks/Filter specialization"}
{!title name="A function specialization technique to optimize a filter implementation"}

% Section: Overview
%
% Demonstration of using specialization to improve the performance of 
% filtering operations, using the GPU hardware texture units.
% Note that there are some limitations for this technique:
% * The matrix dimensions must be a multiple of 32
% * Not all boundary extension methods are supported (only: `'clamp`, `'circular` and `'safe`)
%
function [] = __kernel__ filter_kernel(x : mat, y : mat'unchecked, a : vec, dir : ivec2, pos : ivec2)
    offset = int(numel(a)/2)
    total = 0.

    for k=0..numel(a)-1
        total += x[pos + (k - offset).* dir] * a[k]
    end

    y[pos] = total
end

im = imread("lena_big.tif")[:,:,1]
im_out = zeros(size(im))

a = [1, 2, 3, 2, 1] / 9

% First, we run the filter directly, by passing all parameters
% to the kernel function.
tic()
for k=0..999
    parallel_do(size(im),im,im_out,a,[0,1],filter_kernel)
end
toc("Unoptimized approach")

% Second, we change our kernel function statically, to use the
% hardware texturing units. Note: this only has effect in
% single precision floating point mode. 
% Additionally, we use hardware constant memory for 'a'.
tic()
for k=0..999
    parallel_do(size(im),im,im_out,a,[0,1],$specialize(filter_kernel,type(x,"mat'hwtex_nearest") && type(a,"vec'hwconst")))
end
toc("With hardware texturing")

% Third way: we generate a kernel function for which the parameters
% are substituted + using hardware texturing units
tic()
for k=0..999
    parallel_do(size(im),im,im_out,$specialize(filter_kernel,type(x,"mat'hwtex_nearest") && a==[1,2,3,2,1]/9 && dir==[0,1]))
end
toc("With hardware texturing, specialized")



imshow(im_out)
