{!author name="Bart Goossens"}
{!doc category="Samples/Quasar tricks/Expression optimization"}
{!title name="Demonstration of the automatic kernel generator to optimize expressions"}

% Section: Overview
%
% This example illustrates how the Quasar optimizer can be used to significantly
% boost the performance of certain operations. In this case, on a NVidia Geforce 
% 435M card, a speed-up factor of 15 is obtained for matrix slice assignments. 
% This is because of the expensive sequence allocations 0..511 and ':', which 
% require a matrix to be allocated and initialized each time. The optimization 
% circumvents that, by recognizing operations of the form 
% `y[a..b,:,:] = x[c..d,:,:]`.
% In future versions, this will be done completely automatically by the
% expression optimizer (#pragma expression_optimizer(on)).
%
% The example demonstrates two Quasar optimization techniques:
% 1) the automatic loop parallelizer
% 2) the advanced use of reductions
im = imread("lena_big.tif")
x = im[0..63,:,:]
count = 256

N = size(x, 0)

#pragma loop_parallelizer(on)

function [] = slice_copier(x : cube, y : cube'unchecked, a:int, b:int, c:int, d:int)
    {!parallel for}
    for m=a..b
        for n=0..size(x,1)-1
            for k=0..size(x,2)-1
                y[m,n,k] = x[m-a+c,n,k]
            end
        end
    end
end

#pragma loop_parallelizer(off)

% slice copier
reduction (x:cube, y:cube, a:scalar, b:scalar, c:scalar, d:scalar) -> (y[a..b,:,:] = x[c..d,:,:]) = slice_copier(x,y,a,b,c,d)

% without reductions
tic()
#pragma expression_optimizer(off)
for k=0..999
    y = zeros(count,size(x,1),size(x,2))
    y[0..N/2-1,:,:] = x[0..N/2-1,:,:]
    y[count-N/2..count-1,:,:] = x[N/2..N-1,:,:]
end
toc()

% with automatic kernel generator
tic()
#pragma expression_optimizer(on)
r1=0..N/2-1
r2=count-N/2..count-1
r3=0..511
r4=0..2
for k=0..999
    y = zeros(count,size(x,1),size(x,2))
    y[r1,r3,r4] = x[r1,r3,r4]
    y[r2,r3,r4] = x[r1,r3,r4]
end
toc()


% with manual reductions (same operation as automatic kernel generator)
#pragma expression_optimizer(off)
tic()
for k=0..999
    y = zeros(count,size(x,1),size(x,2))
    y[0..N/2-1,:,:] = x[0..N/2-1,:,:]
    y[count-N/2..count-1,:,:] = x[N/2..N-1,:,:]
end
toc()


imshow(y)