//
// Example C++ module that can be used from Quasar, demonstrating the use
// of the Quasar DSL (which permits passing variables of any type, including
// dynamic types, such as cell matrices from Quasar to C++). Additionally,
// the Quasar DSL allows directly allocating managed host memory and in fact
// gives access to the full functions of the Quasar Runtime system.
//
// Note that in order to use the Quasar DSL interface, it is required to
// provide the 'module_init' and 'module_term' exports, as in the definitions
// below.
//
// See External Interface Reference for an explanation of how to use the
// Quasar C++ host API.
// 
// You can compile this C++ library directly from Quasar Redshift.
// To use this library, simply import from Quasar:
//
// import "my_library.64.so"
//
// Note that this is not really platform-independent. To import in a platform-
// independent way:
//
// import "my_library.${BITNESS}.${NATIVEEXT}"
//
#include "quasar_dsl.h"
#include "quasar_dsl_class_ext.h"
#include <utility>

using namespace quasar;

static IQuasarHost* host; // Pointer to the Quasar host
static CoreLib* coreLib;  // Pointer to the Quasar core library (allows to easily
						  // call Quasar functions)

// A CPU device function that uses a Quasar callback in order to allocate a matrix
QValue generate_matrix(int M, int N)
{
	QValue A = QValue::CreateMatrix<scalar>(M, N);
	Matrix Adata = host->LockMatrix<scalar>(A);
	
	// Alternative version, using cooperative memory management
	//cooperative<Matrix> A = zeros_coop<scalar>(M, N);
	for (int m = 0; m < Adata.dim1; m ++)
		for (int n = 0; n < Adata.dim2; n ++)
		{
			Adata.data[m * Adata.dim2 + n] = m + n;
		}
	host->UnlockMatrix(A);
	return A;
}

// A CPU device function that processes a cell matrix
// Note: due to the C ABI interface, it is required to pass QValues using the 
// qvalue_t struct.
void multiply_cell(QValue cell)
{	
	// Example of an error generated by the module
	if (! (int) coreLib->type(cell, Type(_T("cube[??]"))))
	{
		coreLib->error_(_T("multiply_cell: object should be a cell matrix"));
	}
	if ( (int) coreLib->ndims(cell) > 3)
	{
		coreLib->error_(_T("multiply_cell: max dimensionality is 3"));
	}
	
	for (int m = 0; m < size(cell,0); m++)
		for (int n = 0; n < size(cell,1); n++)
			for (int k = 0; k < size(cell,2);  k++)
	{
		cell(m, n, k) = (QValue)cell(m, m, k) * 2.0;
	}
}

// Function: module_init
//
// Function called when the module is loaded. You can place all
// initialization code here.
//
// Parameters:
// - hostprivParams: private host parameters. Just pass to IQuasarHost::Create
// - moduleDetails: private module parameters. Just pass to IQuasarHost::Create
// - flags: private flags for creating the host connection. Just pass to IQuasarHost::Create
extern "C" EXPORT __device__ void module_init(
	LPCVOID hostprivParams, 
	LPCVOID moduleDetails, 
	int flags)
{
	// Connect to the Quasar host
	host = IQuasarHost::Connect(hostprivParams, moduleDetails, flags);

	// Create a new interface to the core library
	coreLib = new CoreLib();

	// Register the functions, so that they can be called from Quasar
	ADD_FUNCTION(generate_matrix);
	ADD_FUNCTION(multiply_cell);
}

// Function called when the module is unloaded
extern "C" EXPORT __device__ void module_term()
{
	if (coreLib)
	{
		delete coreLib;
		coreLib = nullptr;
	}
	
	// Release the host connection
	if (host)
	{
		delete host;
		host = nullptr;
	}
}
