% Differentiable programming project - MNIST classification
% Shows how a neural network can be implemented, completely from scratch,
% using differentiable programming. See the algorithmic differentiation manual
% for more explanation.

import "Quasar.CompMath.dll" 
import "Quasar.DNN.dll"
import "Quasar.UI.dll"
import "inttypes.q"

type Data : cube{4}

% Type: NetworkParameters
% 
% The structure containing our network parameters. Indicating the size is optional.
% 
type NetworkParameters : mutable class
    A : mat(28*28, 128)
    B : mat(128, 10)
endtype

% Function: init_params
% 
% Initializes the network parameters
% 
% Usage:
%   : function [p : ^NetworkParameters] = init_params()
% 
function [p : ^NetworkParameters] = init_params()
    epsilon = 0.12
    p = NetworkParameters( _
        A:=rand(28*28, 128)*epsilon*2 - epsilon, _
        B:=rand(128, 10)*epsilon*2 - epsilon _
    )
endfunction

% Function: scalar_relu
% 
% Scalar RELU function
% 
% Usage:
%   : function [y : scalar] = __device__ scalar_relu(x : scalar'const, alpha : scalar'const)
% 
% Parameters:
% x - input weight 
% alpha - value used to multiply with negative x
% 
% Returns:
% y - output weight value, after appying the relu
function y = __device__ scalar_relu(x : scalar, alpha : scalar)
    y = x < 0 ? alpha * x : x
endfunction

reduction (x, alpha) -> $diff(scalar_relu(x, alpha), x) = x < 0 ? alpha : 1
reduction (x, alpha, dx) -> $diffmul(scalar_relu(x, alpha), x, dx) = (x >= 0).*dx.*(1-alpha) + dx.*alpha

reduction (x, alpha) -> $adjoint(scalar_relu(x, alpha), x) = scalar_relu(x, alpha) % pointwise operation

% Function: relu
% 
% Applies the RELU to an entire weight matrix
% 
% Usage:
%   : relu(x : cube{4}, alpha : scalar)
% 
% Parameters:
% x - input weights
% alpha - value used to multiply with negative x
% 
% Returns:
% y - output weights
function y = relu(x : cube{4}, alpha : scalar)
    y = uninit(size(x))
    {!parallel for}
    for [m,n,p,q]=0..size(x,0..3)-1
        y[m,n,p,q] = scalar_relu(x[m,n,p,q], alpha)
    endfor
endfunction

% Function: logsoftmax
% 
% Calculates the logarithm of the softmax function: for every point i from the dataset, we
% calculate:
%     log(exp(x[i, m])/sum(exp(i, :)))
% Due to the logarithm, we can apply a scaling trick that originates from Hidden Markov models,
% which significantly improves numerical stability.
% 
% Usage:
%   : function [y : mat] = logsoftmax(x : mat'const)
% 
% Parameters:
% x - input weight matrix
% 
% Returns:
% y - output weight matrix
function y : mat = logsoftmax(x : mat)
    {!ad_support convert_matrix_ops_to_loops=true}
    y = zeros(size(x))
    
    {!parallel for}
    for m=0..size(x,0)-1
        total = 1e-8
        x_max = max(x[m, :])
        for n=0..size(x,1)-1
            total += exp(x[m,n] - x_max)
        endfor
        for n=0..size(x,1)-1
            y[m,n] = x[m,n] - log(total) - x_max
        endfor
    endfor

endfunction

% Function: logsoftmax_deriv
% 
% Calculates the derivative of the softmax function
% 
% Usage:
%   : function [x_deriv : mat] = logsoftmax_deriv(x : mat'const, dx : mat'const)
% 
function x_deriv:mat=logsoftmax_deriv(x:mat,dx:mat) 
      x_deriv=zeros(size(x)) 
      
      {!parallel for}
      for m=0..size(x,0)-1 
          total0_accum0=1e-8
          total0_deriv_accum0=0.0
          x_max = max(x[m, :])
          for n=0..size(x,1)-1
              total0_accum0+=exp(x[m,n] - x_max) 
              total0_deriv_accum0-=dx[m,n]
          endfor           
          for n=0..size(x,1)-1 
              x_deriv[m,n]=dx[m,n] + exp(x[m,n]-x_max)*total0_deriv_accum0/total0_accum0
          endfor 
      endfor 
endfunction 

reduction (x, dx) -> $diffmul(logsoftmax(x), x, dx) = logsoftmax_deriv(x, dx)  
reduction (x, dx) -> $adjoint(logsoftmax_deriv(x, dx), dx) = logsoftmax_deriv(x, dx)  

% Function: dropout
% 
% Implements a dropout layer
% 
% Usage:
%   : function y = dropout(x : mat, U : mat)
% 
% Parameters:
% x - input weights
% U - random values used for dropout
% 
% Returns:
% y - output weights
function y = dropout(x : mat, U : mat)
    y = zeros(size(x))
    for [m,n]=0..size(x,0..1)-1
        y[m,n] = U[m,n] > 0.0 ? x[m,n] : 0.0
    endfor
endfunction

% Function: forward_network
% 
% Implements our forward network
% 
% Usage:
%   : function [y : mat(:, 10)] = forward_network(x : cube(:,28,28), p : ^NetworkParameters)
% 
% Parameters:
% x - input data
% p - network parameters
% 
% Returns:
% y - output data
function [y : mat(:, 10)] = forward_network(x : cube(:,28,28), p : ^NetworkParameters)

    A : mat(:, 128) = rand(size(x,0), 128) - 0.05

    {!ad_support convert_matrix_ops_to_loops=true}
    alpha = 0.0
    x1 : mat(:, 28*28) = reshape(x, [size(x,0), prod(size(x,1..2))])
    x2 : mat(:, 128) = x1 * p.A
    x3 = relu(x2, alpha)
    x4 = dropout(x3, A)
    x5 : mat(:, 10) = x4 * p.B
    y = logsoftmax(x5)

endfunction

% Function: cost_function_crossentropy
% 
% Implements the logarithm of the cross entropy. Needs to be used in combination with the 
% logsoftmax function, because the function assumes that the logarithm is already taken
% 
% Usage:
%   : function C = cost_function_crossentropy(x : mat(:,10), y : vec, p : ^NetworkParameters)
% 
% Parameters:
% x - input dataset
% y - groundtruth data values (category of each sample)
% p - network parameters
% 
% Returns:
% C - the cost value
function C = cost_function_crossentropy(x : mat(:,10), y : vec, p : ^NetworkParameters)

    C = 0.0
    % cross-entropy
    for m = 0..size(x,0)-1
        C -= x[m, int(y[m])]
    endfor

endfunction

% Function: argmax
% 
% Calculates argmax along the rows of the vector
% 
% Usage:
%   : function [y : vec] = argmax(x : mat'const)
% 
% Parameters:
% x - input weights
% 
% Returns:
% y - indices of the maximum weight along each row of the input matrix
function y = argmax(x : mat)
    y = zeros(size(x,0))
    for m=0..size(x, 0)-1
        best = 0
        for k=1..size(x,1)-1
            if x[m,k]>x[m,best]
                best = k
            endif
        endfor
        y[m] = best
    endfor
endfunction

% Function: train_adam
% 
% Implements ADAM's training algorithm with batches of 32 samples
% 
% Usage:
%   : function [] = train_adam(cost_fn : [(cube{4},cube{4},^NetworkParameters)->scalar], _
%         forward_network : [(cube{4},^NetworkParameters)->cube{4}], _
%         gradient : [(cube{4},cube{4},^NetworkParameters)->^NetworkParameters], _
%         inputs : cube{4}, groundtruths : cube{4}, valid_inputs : cube{4}, _
%         valid_groundtruths : cube{4}'const, params : ^NetworkParameters, rate : ??'const = 0.001, _
%         momentum_first_order : ??'const = 0.9, momentum_second_order : ??'const = 0.999, _
%         visualization : ??'const = false)
% 
% Parameters:
% cost_fn - cost function to minimize
% forward_network - forward network function
% gradient - gradient of the cost function
% inputs - training dataset
% groundtruths - groundtruths for the training dataset
% valid_inputs - validation dataset
% valid_groundtruths - groundtruths for the validation dataset
% params - network parameters (both input and output)
% rate - initial learning rate
% momentum_first_order - momentum 1st order parameter
% momentum_second_order - momentum 2nd order parameter
% visualization - enables visualization
function params_out = train_adam( _
    cost_fn : [(Data, Data, ^NetworkParameters) -> scalar], _
    forward_network : [(Data, ^NetworkParameters) -> Data], _
    gradient : [(Data, Data, ^NetworkParameters) -> (^NetworkParameters)], _
    inputs : Data, _
    groundtruths : Data, _
    valid_inputs : Data, _
    valid_groundtruths : Data, _
    params : ^NetworkParameters, _
    rate = 1e-3, _
    momentum_first_order = 0.9, _
    momentum_second_order = 0.999, _
    visualization = false)
     
    nr_images = size(inputs,0)
    assert(size(groundtruths,0) == nr_images)
         
    %preparing the visualization, if necessary
    if visualization
        frm = form("Deep learning network training")
        hgr_plt = frm.add_horizontallayout()
        dsp_plt_err = hgr_plt.add_display()
        dsp_plt_acc = hgr_plt.add_display()
    endif
    
    first_order_moments = 0*params
    second_order_moments = 0*params
    batch_size : 'const = 32
    nr_epochs = 2
    nr_iterations = int(nr_images/batch_size)
    optimization_epsilon = 1e-3
    
    accuracies = zeros(nr_epochs)
    valid_accuracies = zeros(nr_epochs)
    cost = zeros(nr_epochs)
    
    for epoch = 0..nr_epochs-1
        tic()
        for iteration = 0..nr_iterations-1
            img_idx = zeros(batch_size)
            for i=0..batch_size-1
                img_idx[i] = mod(pos2ind([32, int(nr_images/32)], [i, int(iteration)*batch_size]), nr_images)
            endfor
            input = inputs[img_idx, :, :, :]
            groundtruth = groundtruths[img_idx, :]
            
            % the forward pass
            output = forward_network(input, params)
            
            % calculate the gradient
            grad = gradient(input, groundtruth, params)
            
            %for each layer, backward propagation, updating the momentum and taking the step
            unbiased_rate = rate * sqrt(1-momentum_second_order.^(iteration+epoch*nr_iterations+1)) / _
                (1-momentum_first_order.^(iteration+epoch*nr_iterations+1)) %see Adam paper
                    
            % updates the moments and the parameters (uses cell arithmetic)
            first_order_moments = first_order_moments * momentum_first_order + grad .* (1 - momentum_first_order)
            second_order_moments = second_order_moments * momentum_second_order + grad.^2 .* (1 - momentum_second_order)
            params = params - unbiased_rate * first_order_moments ./ (sqrt(second_order_moments) + optimization_epsilon)
        endfor
        
        printf("epoch %d: %f ms  ",epoch,toc(""))

        % Calculate the accuracy over the whole training and validation datasets
        accuracies[epoch] = mean(argmax(forward_network(inputs, params)) == groundtruths) 
        valid_accuracies[epoch] = mean(argmax(forward_network(valid_inputs, params)) == valid_groundtruths) 
        cost[epoch] = cost_fn(forward_network(inputs, params), groundtruths, params)
        printf("accuracy: %f, valid_acc: %f\n", accuracies[epoch], valid_accuracies[epoch])
                        
        %the plot output: error functions
        dsp_plt_err.plot(1..epoch+1,cost[0..epoch],"b")
        xlim([1,nr_epochs])
        ylim([0,max(max(cost*1.5),1e-9)])
        title("Cost function")
        dsp_plt_acc.plot(1..epoch+1,accuracies[0..epoch],"b", 1..epoch+1,valid_accuracies[0..epoch],"g")
        xlim([1,nr_epochs])
        ylim([0,1])
        title("Accuracy")
    endfor
    params_out = params
endfunction

% Function: calc_inference_accuracy
% 
% Implements the inference function for our network; calculates the accuracy
% 
% Usage:
%   : function [accuracy] = inference(x_test, y_test, params : ^NetworkParameters)
% 
% Parameters:
% x_test - input dataset
% y_test - input groundtruths
% params - network parameters
% 
% Returns:
% accuracy - accuracy of the network on the input dataset
function accuracy = calc_inference_accuracy(x_test, y_test, params : ^NetworkParameters)
    output = forward_network(x_test, params)
    accuracy = mean(argmax(output) == y_test) 
endfunction

function [] = main()

    [x_train8, y_train8, x_test8, y_test8] = load("mnist.qd")
    x_train = float(x_train8)/255
    x_test = float(x_test8)/255
    y_train = transpose(float(y_train8))
    y_test = transpose(float(y_test8))
    
    
    gradient = (x : cube(:,28,28), y : Data, p : ^NetworkParameters) -> $diff(cost_function_crossentropy(forward_network(x, p), y, p), p)
    params = init_params()
    
    params = train_adam( _
        cost_function_crossentropy, _
        forward_network, _
        gradient, _
        x_train, _
        y_train, _
        x_test, _
        y_test, _
        params, _
        rate := 1e-3, _
        visualization := true)    
        
     printf("Accuracy: %.02f%%\n", calc_inference_accuracy(x_test, y_test, params)*100)

endfunction