//=============================================================================
// Quasar OpenCL runtime backend
// (C) 2011-2014 Bart Goossens
//=============================================================================
#ifndef __QUASAR_CL__
#define __QUASAR_CL__

#define INLINE inline
#define NULL 0

//++++++++ Main scalar definition ++++++++++
#ifdef DBL_SCALAR
#pragma OPENCL EXTENSION cl_khr_fp64: enable
typedef double scalar;
#else
typedef float scalar;
#endif

// Macros to avoid compiler warnings due to cl_khr_fp64 not being enabled
#ifdef DBL_SCALAR
#define ONE_HALF 0.5
#else
#define ONE_HALF 0.5f
#endif

typedef struct complex_t
{
  scalar x;
  scalar y;
} complex_t;

//=============================================================================
// Constants
//=============================================================================
#ifndef M_PI
#   ifdef DBL_SCALAR
#   define M_PI    3.14159265358979323846
#   else
#   define M_PI    3.14159265358979323846f
#   endif
#endif

//=============================================================================
// Helper functions
//=============================================================================

// Integer power function (exponentiation by squaring)
int powii(int base, int a) { 
    int res = 1;
    while (a) {
        if (a & 1)
            res *= base;
        a >>= 1;
        base *= base;
    }
    return res;
}

//=============================================================================
// Atomic operations
//=============================================================================

// Most of the atomic operations are not supported by OpenCL 1.2, therefore we rely
// on the atomic_cmpxchg function.
#define atomicCAS32 atomic_cmpxchg
#define atomicCAS64 atomic_cmpxchg

// Since we don't have generic address spaces as in CUDA, we must support atomic 
// operations for both the global and local address spaces.

// Integers - operator
#define DECL_ATOMIC_OP(name, addrspace, X)	\
    int atomic_##name(addrspace int* pd, int d) {  \
        union { \
            int i; \
            int d; \
        } Old; \
        union { \
            int i; \
            int d; \
        } New; \
        for(;;) { \
            Old.i = *pd;   \
            New.d = Old.d X d;  \
            if(atomicCAS32( pd, Old.i, New.i) == Old.i) \
                return New.d; \
        } \
    }
// Integers - function
#define DECL_ATOMIC_OP_FN(name, addrspace, fn)	\
    int atomic_##name(addrspace int* pd, int d) {  \
        union { \
            int i; \
            int d; \
        } Old; \
        union { \
            int i; \
            int d; \
        } New; \
        for(;;) { \
            Old.i = *(addrspace int*)pd;   \
            New.d = fn(Old.d, d);  \
            if(atomicCAS32(pd, Old.i, New.i) == Old.i) \
                return New.d; \
        } \
    }
// Float32 - operator
#define DECL_FLOAT_ATOMIC_OP(name, addrspace, X)	\
    float fatomic_##name(addrspace float* pd, float d) {  \
        union { \
            int i; \
            float d; \
        } Old; \
        union { \
            int i; \
            float d; \
        } New; \
        for(;;) { \
            Old.i = *(addrspace int*)pd;   \
            New.d = Old.d X d;  \
            if(atomicCAS32( (addrspace int *)pd, Old.i, New.i) == Old.i) \
                return New.d; \
        } \
    }
// Float32 - function
#define DECL_FLOAT_ATOMIC_OP_FN(name, addrspace, fn)	\
    float fatomic_##name(addrspace float* pd, float d) {  \
        union { \
            int i; \
            float d; \
        } Old; \
        union { \
            int i; \
            float d; \
        } New; \
        for(;;) { \
            Old.i = *(addrspace int*)pd;   \
            New.d = fn(Old.d, d);  \
            if(atomicCAS32( (addrspace int *)pd, Old.i, New.i) == Old.i) \
                return New.d; \
        } \
    }
    
// Float64 - operator
#define DECL_DOUBLE_ATOMIC_OP(name, addrspace, X)	\
    double datomic_##name(addrspace double* pd, double d) {  \
        union { \
            long long i; \
            double d; \
        } Old; \
        union { \
            long long i; \
            double d; \
        } New; \
        for(;;) { \
            Old.i = *(addrspace long long*)pd;   \
            New.d = Old.d X d;  \
            if(atomicCAS64( (addrspace long long *)pd, Old.i, New.i) == Old.i) \
                return New.d; \
        } \
    }
// Float64 - function
#define DECL_DOUBLE_ATOMIC_OP_FN(name, addrspace, fn)	\
    double datomic_##name(addrspace double* pd, double d) {  \
        union { \
            long long i; \
            double d; \
        } Old; \
        union { \
            long long i; \
            double d; \
        } New; \
        for(;;) { \
            Old.i = *(addrspace long long*)pd;   \
            New.d = fn(Old.d, d);  \
            if(atomicCAS64( (addrspace long long *)pd, Old.i, New.i) == Old.i) \
                return New.d; \
        } \
    }   

#define atomic_add_local atomic_add

// Integer atomic operations (most are supported by OpenCL 1.2):
// atomic_add, atomic_min, atomic_max, atomic_and, atomic_or, atomic_xor
DECL_ATOMIC_OP(mul, __global, *)	
DECL_ATOMIC_OP(div, __global, /)
DECL_ATOMIC_OP_FN(pow, __global, powii)
DECL_ATOMIC_OP(mul_local, __local, *)	
DECL_ATOMIC_OP(div_local, __local, /)
DECL_ATOMIC_OP_FN(pow_local, __local, powii)

// Single precision atomic operators
DECL_FLOAT_ATOMIC_OP(add, __global, +)
DECL_FLOAT_ATOMIC_OP(mul, __global, *)
DECL_FLOAT_ATOMIC_OP(div, __global, /)
DECL_FLOAT_ATOMIC_OP_FN(pow, __global, pow)
DECL_FLOAT_ATOMIC_OP_FN(min, __global, min)
DECL_FLOAT_ATOMIC_OP_FN(max, __global, max)

DECL_FLOAT_ATOMIC_OP(add_local, __local, +)
DECL_FLOAT_ATOMIC_OP(mul_local, __local, *)
DECL_FLOAT_ATOMIC_OP(div_local, __local, /)
DECL_FLOAT_ATOMIC_OP_FN(pow_local, __local, pow)
DECL_FLOAT_ATOMIC_OP_FN(min_local, __local, min)
DECL_FLOAT_ATOMIC_OP_FN(max_local, __local, max)

// Double precision atomic operators
#ifdef DBL_SCALAR
	DECL_DOUBLE_ATOMIC_OP(add, __global, +)
	DECL_DOUBLE_ATOMIC_OP(mul, __global, +)
	DECL_DOUBLE_ATOMIC_OP(div, __global, +)
	DECL_DOUBLE_ATOMIC_OP_FN(pow, __global, pow)
	DECL_DOUBLE_ATOMIC_OP_FN(min, __global, min)
	DECL_DOUBLE_ATOMIC_OP_FN(max, __global, max)
	
	DECL_DOUBLE_ATOMIC_OP(add_local, __local, +)
	DECL_DOUBLE_ATOMIC_OP(mul_local, __local, +)
	DECL_DOUBLE_ATOMIC_OP(div_local, __local, +)
	DECL_DOUBLE_ATOMIC_OP_FN(pow_local, __local, pow)
	DECL_DOUBLE_ATOMIC_OP_FN(min_local, __local, min)
	DECL_DOUBLE_ATOMIC_OP_FN(max_local, __local, max)	
#endif

//=============================================================================
// Complex number arithmetic
//=============================================================================

complex_t make_complex(scalar x, scalar y)
{
    complex_t z = { x, y };
    return z;
}

scalar real(complex_t a)
{
    return a.x;
}

scalar imag(complex_t a)
{
    return a.y;
}

// Functions for complex numbers
complex_t cplx_negate(complex_t a)
{
    return make_complex(-a.x, -a.y);
}

complex_t cplx_add(complex_t a, complex_t b)
{
    return make_complex(a.x + b.x, a.y + b.y);
}

complex_t cplx_addcr(complex_t a, scalar b)
{
    return make_complex(a.x + b, a.y);
}

complex_t cplx_addrc(scalar a, complex_t b)
{
    return make_complex(a + b.x, b.y);
}

complex_t cplx_sub(complex_t a, complex_t b)
{
    return make_complex(a.x - b.x, a.y - b.y);
}

complex_t cplx_subcr(complex_t a, scalar b)
{
    return make_complex(a.x - b, a.y);
}

complex_t cplx_subrc(scalar a, complex_t b)
{
    return make_complex(a - b.x, -b.y);
}

complex_t cplx_mul(complex_t a, complex_t b)
{
    return make_complex(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
}

complex_t cplx_mulcr(complex_t a, scalar b)
{
    return make_complex(a.x * b, a.y * b);
}

complex_t cplx_mulrc(scalar a, complex_t b)
{
    return make_complex(a * b.x, a * b.y);
}

complex_t cplx_div(complex_t a, complex_t b)
{
    if (b.x < b.y)
    {   // overflow-safe implementation
        scalar t = b.x / b.y;
        scalar d = b.y * (1 + t * t);
        return cplx_mul(a, make_complex(t / d, (scalar)(-1) / d));
    }
    else
    {
        scalar t = b.y / b.x;
        scalar d = b.x * (1 + t * t);
        return cplx_mul(a, make_complex(1 / d, -t / d));
    }
}

complex_t cplx_divcr(complex_t a, scalar b)
{
    return make_complex(a.x / b, a.y / b);
}

complex_t cplx_divrc(scalar a, complex_t b)
{
    return cplx_div(make_complex(a, 0), b);
}

int cplx_equal(complex_t a, complex_t b)
{
    return a.x == b.x && a.y == b.y;
}

int cplx_equalcr(complex_t a, scalar b)
{
    return a.x == b && a.y == 0;
}

int cplx_equalrc(scalar a, complex_t b)
{
    return a == b.x && 0 == b.y;
}

int cplx_notequal(complex_t a, complex_t b)
{
    return a.x != b.x || a.y != b.y;
}

int cplx_notequalcr(complex_t a, scalar b)
{
    return a.x != b || a.y != 0;
}

int cplx_notequalrc(scalar a, complex_t b)
{
    return a != b.x || 0 != b.y;
}

complex_t cplx_pow(complex_t a, complex_t b)
{
	scalar rm = a.x * a.x + a.y * a.y;
	if (rm == 0) return make_complex(0,0);
	scalar r = log(rm);
	scalar t = atan2(a.y, a.x);
	scalar ra = exp(ONE_HALF * b.x * r - b.y * t);
	scalar rb = ONE_HALF * b.y * r + b.x * t;
	return make_complex(ra*cos(rb),ra*sin(rb));
}

complex_t cplx_powrc(scalar a, complex_t b)
{
	scalar r = log(a);
	scalar t = sign(a)*M_PI;
	scalar ra = exp(b.x * r - b.y * t);
	scalar rb = b.y * r + b.x * t;
	return make_complex(ra*cos(rb),ra*sin(rb));
}

complex_t cplx_powcr(complex_t a, scalar b)
{
	scalar r = log(a.x * a.x + a.y * a.y);
	scalar t = atan2(a.y, a.x);
	scalar ra = exp(ONE_HALF * b * r);
	scalar rb = b * t;
	return make_complex(ra*cos(rb),ra*sin(rb));
}

//=============================================================================
// Built-in vector functions
//=============================================================================
float vsumf(__global float *a, int count)
{
    float sum = 0;
    for (int i = 0; i < count; i ++)
        sum += a[i];
    return sum;
}
#ifdef DBL_SCALAR
double vsumd(__global double *a, int count)
{
    double sum = 0;
    for (int i = 0; i < count; i ++)
        sum += a[i];
    return sum;
}
#endif
float vprodf(__global float *a, int count)
{
    float prod = 1;
    for (int i = 0; i < count; i ++)
        prod += a[i];
    return prod;
}
#ifdef DBL_SCALAR
double vprodd(__global double *a, int count)
{
    double prod = 1;
    for (int i = 0; i < count; i ++)
        prod *= a[i];
    return prod;
}
#endif
float vminf(__global float *a, int count)
{
    float m = a[0];
    for (int i = 1; i < count; i ++)
        m = min(m, a[i]);
    return m;
}
#ifdef DBL_SCALAR
double vmind(__global double *a, int count)
{
    double m = a[0];
    for (int i = 1; i < count; i ++)
        m = min(m, a[i]);
    return m;
}
#endif
float vmaxf(__global float *a, int count)
{
    float m = a[0];
    for (int i = 1; i < count; i ++)
        m = max(m, a[i]);
    return m;
}
#ifdef DBL_SCALAR
double vmaxd(__global double *a, int count)
{
    double m = a[0];
    for (int i = 1; i < count; i ++)
        m = max(m, a[i]);
    return m;
}
#endif
float vmeanf(__global float *a, int count) { return vsumf(a,count)/count; }
#ifdef DBL_SCALAR
double vmeand(__global double *a, int count) { return vsumd(a,count)/count; }
#endif
float vdotprodf(__global float *a, int count_a, __global float *b, int count_b)
{
    //assert(count_a==count_b);
    float sum = 0;
    for (int i = 0; i < count_a; i ++)
        sum += a[i] * b[i];
    return sum;
}
#ifdef DBL_SCALAR
double vdotprodd(__global double *a, int count_a, __global double *b, int count_b)
{
    //assert(count_a==count_b);
    double sum = 0;
    for (int i = 0; i < count_a; i ++)
        sum += a[i] * b[i];
    return sum;
}
#endif

#endif

// The following approximation functions currently have no OpenCL mapping, we replace them
// with the slower but more accurate versions
scalar rcp_rn(scalar a) { return (scalar)1.0 / a; }
scalar rcp_ru(scalar a) { return (scalar)1.0 / a; }
scalar rcp_rz(scalar a) { return (scalar)1.0 / a; }
scalar div_approx(scalar a, scalar b) { return a / b; }