#ifndef __QUASAR_NO_EXPORTS__
#define __QUASAR_NO_EXPORTS__
#endif
#include "quasar_dsl.h"

namespace quasar
{
	QUASARHOSTAPI QValue QValue::TypedNullptr = QValue(TYPE_TYPEDOBJECT, NULL);
	QUASARHOSTAPI QValue QValue::UntypedNullptr = QValue(TYPE_UNTYPEDOBJECT, NULL);
	QUASARHOSTAPI QValue QValue::Default = QValue(TYPE_VOID, NULL);

	IMPLEMENT_TYPE(TYPE_VOID, QValue, _T("??"));
	IMPLEMENT_TYPE(TYPE_LAMBDAEXPR, Function, _T("lambda_expr"));
	IMPLEMENT_TYPE(TYPE_TYPEINFO, Type, _T("type"));

	/*------------------------------------------------------------------------
	 * QValue::CreateObject
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI QValue QValue::CreateObject(const Type& type)
	{
		qvalue_t typeData = type;
		QValue result;
		IQuasarHost *host = IQuasarHost::GetInstance();
		switch (type.GetDataType())
		{
		case TYPE_TYPEDOBJECT:
			result = host->CreateTypedObject(type);
			break;
		case TYPE_UNTYPEDOBJECT:
			result = host->CreateUntypedObject(type);
			break;
		default:
			_ASSERT(false); // Invalid type
			break;
		}
		return result;
	}

	/*------------------------------------------------------------------------
	 * Range constructor
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI QValue Range(scalar minVal, scalar maxVal, scalar step)
	{
		IQuasarHost *instance = IQuasarHost::GetInstance();
		IComputationEngine *engine = instance->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		stack->PushValue(qvalue_t::fromScalar(minVal));
		stack->PushValue(qvalue_t::fromScalar(step));
		stack->PushValue(qvalue_t::fromScalar(maxVal));
		engine->Process(OP_DOTDOTDOT);
		return stack->PopValue();
	}

	/*------------------------------------------------------------------------
	 * Range constructor (with integers)
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI QValue Range(int minVal, int maxVal, int step)
	{
		IQuasarHost *instance = IQuasarHost::GetInstance();
		IComputationEngine *engine = instance->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		stack->PushValue(qvalue_t::fromInt(minVal));
		stack->PushValue(qvalue_t::fromInt(step));
		stack->PushValue(qvalue_t::fromInt(maxVal));
		engine->Process(OP_DOTDOTDOT);
		return stack->PopValue();
	}

	/*------------------------------------------------------------------------
	 * Range constructor (with QValue)
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI QValue Range(QValue minVal, QValue maxVal)
	{
		IQuasarHost *instance = IQuasarHost::GetInstance();
		IComputationEngine *engine = instance->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		stack->PushValue(minVal);
		stack->PushValue(maxVal);
		engine->Process(OP_DOTDOT);
		return stack->PopValue();
	}

	/*------------------------------------------------------------------------
	 * Range constructor (with QValue)
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI QValue Range(QValue minVal, QValue maxVal, QValue step)
	{
		IQuasarHost *instance = IQuasarHost::GetInstance();
		IComputationEngine *engine = instance->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		stack->PushValue(minVal);
		stack->PushValue(step);
		stack->PushValue(maxVal);
		engine->Process(OP_DOTDOTDOT);
		return stack->PopValue();
	}

	/*------------------------------------------------------------------------
	 * GetType function: returns the type of this QValue object
	 *------------------------------------------------------------------------*/
	Type QValue::GetType() const
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		return host->GetType(value);
	}

	/*------------------------------------------------------------------------
	 * HasType function: checks if the given QValue has the specified type
	 *------------------------------------------------------------------------*/
	bool QValue::HasType(Type type) const {
		IQuasarHost *host = IQuasarHost::GetInstance();
		Type myType = GetType();
		return host->TypeCompareToSpecification(myType, type);
	}

	/*------------------------------------------------------------------------
	 * size function
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI int size(QValue val, int dim)
	{
		int len = 1;
		switch (val->type)
		{
		case TYPE_VEC:
		case TYPE_MAT:
		case TYPE_CUBE:
		case TYPE_NCUBE:
		case TYPE_CVEC:
		case TYPE_CMAT:
		case TYPE_CCUBE:
		case TYPE_NCCUBE:
			if (dim < 3)
			{
				len = (dim == 0) ? val->dim1 :
					(dim == 1) ? val->dim2 :
					(dim == 2) ? val->dim3 : 0;
			}
			else // Higher dimensional matrix
			{
				int dims[16];
				int ndims = 16;
				IQuasarHost::GetInstance()->GetNDims(val, &ndims, dims);
				len = (dim < ndims) ? dims[dim] : 0;
			}
			break;
		}
		return len;
	}

	/*------------------------------------------------------------------------
	 * size function
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI QValue size(QValue val)
	{
		IQuasarHost *instance = IQuasarHost::GetInstance();
		IComputationEngine *engine = instance->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		stack->PushValue(val);
		engine->FunctionCall(_T("size"), 1);
		return stack->PopValue();
	}

	/*------------------------------------------------------------------------
	 * size function
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI QValue size(QValue val, QValue dim)
	{
		IQuasarHost *instance = IQuasarHost::GetInstance();
		IComputationEngine *engine = instance->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		stack->PushValue(val);
		stack->PushValue(dim);
		engine->FunctionCall(_T("size"), 2);
		return stack->PopValue();
	}

	/*------------------------------------------------------------------------
	 * numel function
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI int numel(QValue val)
	{
		if (val->type == TYPE_NCUBE || val->type == TYPE_NCCUBE)
		{
			int dims[16];
			int ndims = 16;
			int p = 1;
			IQuasarHost::GetInstance()->GetNDims(val, &ndims, dims);
			for (int i = 0; i < ndims; i++)
				p *= dims[i];
			return p;
		}
		else
		{
			return val->dim1*val->dim2*val->dim3;
		}
	}

	/*------------------------------------------------------------------------
	 * type function
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI Type type(const QValue &val)
	{
		return IQuasarHost::GetInstance()->GetType(val);
	}

	#ifdef CPLUSPLUS11
	/*------------------------------------------------------------------------
	 * QValue::SwitchSelector
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI int QValue::SwitchSelector(std::initializer_list<QValue> list, QValue value)
	{
		int index = 0;
		for (auto item = list.begin(); item != list.end(); item++, index ++)
			if (*item == value)
				return index;
		return (int) list.size(); // Default case
	}
	#endif

	QUASARHOSTAPI void Type::ObtainTypeParameters()
	{
		genericTypeParameters.clear();

		string_t moduleName;
		string_t typeName;
		ClassInfoFlags classInfoFlags;
		int numResources = 0;
		int numGenericTypeParams = 32, numInstanceTypeParams = 32;
		qhandle genericTypeParams[32], instanceTypeParams[32];
		qhandle baseClass = nullptr;
		IQuasarHost::GetInstance()->TypeGetClassInfo(value.private_obj, moduleName, typeName, classInfoFlags, nullptr, &numResources,
			nullptr, &baseClass, &numGenericTypeParams, genericTypeParams, &numInstanceTypeParams, instanceTypeParams);

		qvalue_t type;
		type.type = TYPE_TYPEINFO;
		type.private_obj = baseClass;
		if (baseClass != nullptr)
			this->baseClass.push_back(Type(type));

		for (int i = 0; i < numGenericTypeParams; i ++)
		{
			type.private_obj = genericTypeParameters[i];
			genericTypeParameters.push_back(Type(type));
		};
		for (int i = 0; i < numInstanceTypeParams; i ++)
		{
			type.private_obj = instanceTypeParams[i];
			instanceTypeParameters.push_back(Type(type));
		}
		genericTypeParametersInitialized = true;
	}


	/*------------------------------------------------------------------------
	 * Function::copy_args
	 * helper function used by Function::CreateLambda
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void Function::copy_args(QValue *args, const qvalue_t *argsIn, int numArgs)
	{
		for (int i = 0; i < numArgs; i++)
		{
			args[i] = argsIn[i];
			args[i].Retain();
		}
	}

	/*------------------------------------------------------------------------
	 * RuntimeControl::RuntimeControl
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI RuntimeControl::RuntimeControl()
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		qvalue_t obj;
		if (!host->GetRuntimeControl(obj))
			(QValue&)*this = UntypedNullptr;
		else
			value = obj;
	}

	#ifdef CPLUSPLUS11
	/*------------------------------------------------------------------------
	 * IRuntimeReductionEngine::Add
	 *------------------------------------------------------------------------*/
	bool IRuntimeReductionEngine::Add(
		std::initializer_list<VariableDecl> variables,
		std::initializer_list<Symbol> searchExprSymbols, bool isVariadic, qvalue_t action, qvalue_t whereClause)
	{
		Expression searchExpr(searchExprSymbols, isVariadic);
		return Add(variables.begin(), (int)variables.size(), &searchExpr, &action, &whereClause);
	}
	#endif

}
