//=============================================================================
// Quasar Domain-specific Language (DSL) for C++
// (C) 2011-2020 Gepura
//=============================================================================
#ifndef __QUASAR_DSL_H__
#define __QUASAR_DSL_H__

#include "quasar_host.h"

#ifdef CPLUSPLUS11
#include <functional>
#include <initializer_list>
#include <type_traits>
#include <tuple>
#endif
#include <vector>

namespace quasar
{
	// Forward declarations
	class QValue;
	class Function;
	class Type;
	class TypeBuilder;
	class NamedParam;
	class RuntimeControl;

#ifdef CPLUSPLUS14
	namespace utility
	{
		template<int... Is>
		struct seq { };

		template<int N, int... Is>
		struct gen_seq : gen_seq<N - 1, N - 1, Is...> { };

		template<int... Is>
		struct gen_seq<0, Is...> : seq<Is...> { };

		template<size_t, class T>
		using T_ = T;

		template<class T, size_t... Is>
		auto tuple_of_length(std::index_sequence<Is...>) { return std::tuple<T_<Is, T>...>{}; }

		template<class T, size_t N>
		auto tuple_of_length() { return tuple_of_length<T>(std::make_index_sequence<N>{}); }
	}

#endif

	template<typename>
	struct dereference;

	template<typename T>
	struct dereference<T*>
	{
		typedef T type;
	};

	// Provides automatic locking of objects
	template <typename T>
	class auto_lock : public T
	{
	private:
		qvalue_t obj;
		LockingModes lockingMode;
		MemResources memResource;

		auto_lock(T t, qvalue_t obj, LockingModes lockingMode = LOCK_READWRITE,
			MemResources memResource = MEMRESOURCE_CPU)
		{
			(T &)*this = t;
			this->lockingMode = lockingMode;
			this->memResource = memResource;
			this->obj = obj;
		}
	public:
		template <typename R>
		static auto_lock lock_vector(qvalue_t obj, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
			return auto_lock(IQuasarHost::GetInstance()->LockVector<R>(obj, lockingMode, memResource), obj, lockingMode, memResource);
		}
		template <typename R>
		static auto_lock lock_matrix(qvalue_t obj, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
			return auto_lock(IQuasarHost::GetInstance()->LockMatrix<R>(obj, lockingMode, memResource), obj, lockingMode, memResource);
		}
		template <typename R>
		static auto_lock lock_cube(qvalue_t obj, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
			return auto_lock(IQuasarHost::GetInstance()->LockCube<R>(obj, lockingMode, memResource), obj, lockingMode, memResource);
		}
		template <typename R, int N>
		static auto_lock lock_ncube(qvalue_t obj, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
			return auto_lock(IQuasarHost::GetInstance()->LockNCube<R, N>(obj, lockingMode, memResource), obj, lockingMode, memResource);
		}
		static auto_lock lock_object(qvalue_t obj, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
			return auto_lock(IQuasarHost::GetInstance()->LockObject<dereference<T>::type>(obj, lockingMode, memResource), obj, lockingMode, memResource);
		}
		~auto_lock()
		{
			IQuasarHost::GetInstance()->Unlock(obj, lockingMode, memResource);
		}
	};

	// Object-oriented wrapper for the raw qvalue_t data type
	class QValue
	{
		friend class Function;
		friend class Reduction;

	public:
		class FieldRef;

	protected:
		qvalue_t value;
		QValue(enum DataTypes type, qhandle private_obj)
		{
			value.type = type;
			value.private_obj = private_obj;
		}

	public:
		QValue()
		{
			value.type = TYPE_VOID;
			value.private_obj = NULL;
		}
		QValue(float scalarVal) { value = qvalue_t::fromScalar(scalarVal); }
		QValue(double scalarVal) { value = qvalue_t::fromScalar(scalarVal); }
		QValue(cscalar complexVal)
		{
			value = qvalue_t::fromCScalar(complexVal);
		}
		QValue(int intVal) { value = qvalue_t::fromInt(intVal); }
		QValue(size_t sizeVal) { value = qvalue_t::fromLong((long long)sizeVal); }
		QValue(bool blVal) { value = qvalue_t::fromInt(blVal ? 1 : 0); }
		QValue(long long longVal) { value = qvalue_t::fromLong(longVal); }
		static QUASARHOSTAPI QValue TypedNullptr;
		static QUASARHOSTAPI QValue UntypedNullptr;
		static QUASARHOSTAPI QValue Default;

	#ifdef CPLUSPLUS14
		template <typename ...T, int ...S>
		QValue(std::tuple<T...> tuple, utility::seq<S...>) // Tuple constructor
		{
			const int N = sizeof...(T);
			QValue args[N] = { std::get<S>(tuple)... };
			value = CreateCellMatrix_helper(args, N);
		}

		template <typename ...T>
		QValue(std::tuple<T...> tuple) : QValue(tuple, utility::gen_seq<sizeof...(T)>())
		{
		}

	#endif

		// Helper class for element-wise operations
		class ElemWiseOp
		{
			qvalue_t value;

		public:
			ElemWiseOp(QValue other):value(other.value)
			{
				Retain();
			}
			ElemWiseOp(const ElemWiseOp &other):value(other.value)
			{
				Retain();
			}
			~ElemWiseOp() { IQuasarHost::GetInstance()->ReleaseRef(value); }
			void Retain() { IQuasarHost::GetInstance()->AddRef(value); }

			operator QValue();
		};

		template <int N> class Elem
		{
			qvalue_t mtx;
			qvalue_t index[N];
			BoundaryAccessMode boundaryAccessMode;

		public:
			Elem(const qvalue_t &value, qvalue_t index[N],
				BoundaryAccessMode boundaryAccessMode)
			{
				this->mtx = value;
				this->boundaryAccessMode = boundaryAccessMode;
				IQuasarHost* host = IQuasarHost::GetInstance();
				for (int i = 0; i < N; i++)
				{
					this->index[i] = index[i];
					host->AddRef(this->index[i]);
				}
				host->AddRef(mtx);
			}
			Elem(const Elem &elem)
			{
				this->mtx = elem.mtx;
				this->boundaryAccessMode = elem.boundaryAccessMode;
				IQuasarHost* host = IQuasarHost::GetInstance();
				for (int i = 0; i < N; i++)
				{
					this->index[i] = elem.index[i];
					host->AddRef(this->index[i]);
				}
				host->AddRef(mtx);
			}
			~Elem()
			{
				IQuasarHost* host = IQuasarHost::GetInstance();
				host->ReleaseRef(mtx);
				for (int i = 0; i < N; i++)
					host->DeleteValue(index[i]);
			}

			Elem& operator=(const qvalue_t &newValue);
			Elem& operator=(int value) { operator=(qvalue_t::fromInt(value)); return *this; }
			Elem& operator=(scalar value) { operator=(qvalue_t::fromScalar(value)); return *this; }
			Elem& operator=(cscalar value) { operator=(qvalue_t::fromCScalar(value)); return *this; }
			Elem& operator=(QValue newValue) { operator=(newValue.value); return *this; }
			Elem& operator=(string_t newValue) { operator=((QValue)newValue); return *this; }
			Elem& operator=(scalar1 v) { return operator=((QValue)v); }
			Elem& operator=(scalar2 v) { return operator=((QValue)v); }
			Elem& operator=(scalar3 v) { return operator=((QValue)v); }
			Elem& operator=(scalar4 v) { return operator=((QValue)v); }
			Elem& operator=(cscalar1 v) { return operator=((QValue)v); }
			Elem& operator=(cscalar2 v) { return operator=((QValue)v); }
			Elem& operator=(cscalar3 v) { return operator=((QValue)v); }
			Elem& operator=(cscalar4 v) { return operator=((QValue)v); }
			Elem& operator=(int1 v) { return operator=((QValue)v); }
			Elem& operator=(int2 v) { return operator=((QValue)v); }
			Elem& operator=(int3 v) { return operator=((QValue)v); }
			Elem& operator=(int4 v) { return operator=((QValue)v); }
			template <int K>
			Elem& operator=(scalarN<K> v) { return operator=((QValue)v); }
			template <int K>
			Elem& operator=(cscalarN<K> v) { return operator=((QValue)v); }
			template <int K>
			Elem& operator=(intN<K> v) { return operator=((QValue)v); }
			template <int K>
			Elem& operator=(const Elem<K>& v) { return operator=((QValue)v); }
			Elem& operator=(const Elem& v) { return operator=((QValue)v); }


			template <class R>
			Elem &operator=(const R &other)
			{
#ifdef CPLUSPLUS11
				static_assert(typeinfo_t<R>::type == TYPE_TYPEDOBJECT ||
							  typeinfo_t<R>::type == TYPE_LAMBDAEXPR ||
							  typeinfo_t<R>::type == TYPE_TYPEINFO, "operator=: Typed object expected!");
#endif
				QValue newValue = QValue::FromObj(other);
				return operator=(newValue.value);
			}

		private:
			Elem& InplaceOp(OperatorTypes opType, QValue b);

		public:
			Elem& operator+=(QValue b) { return InplaceOp(OP_ADD, b); }
			Elem& operator-=(QValue b) { return InplaceOp(OP_SUB, b); }
			Elem& operator*=(QValue b) { return InplaceOp(OP_MULTIPLY, b); }
			Elem& operator/=(QValue b) { return InplaceOp(OP_DIVIDE, b); }

			Elem& InplaceAdd(QValue b) { return InplaceOp(OP_ADD, b); }
			Elem& InplaceSubtract(QValue b) { return InplaceOp(OP_SUB, b); }
			Elem& InplaceMultiply(QValue b) { return InplaceOp(OP_MULTIPLY, b); }
			Elem& InplaceDivide(QValue b) { return InplaceOp(OP_DIVIDE, b); }
			Elem& InplacePow(QValue b) { return InplaceOp(OP_POW, b); }
			Elem& InplaceXor(QValue b) { return InplaceOp(OP_LOG_XOR, b); }
			Elem& InplaceOr(QValue b) { return InplaceOp(OP_LOG_OR, b); }
			Elem& InplaceAnd(QValue b) { return InplaceOp(OP_LOG_AND, b); }
			Elem& InplaceMin(QValue b) { return InplaceOp(OP_MIN, b); }
			Elem& InplaceMax(QValue b) { return InplaceOp(OP_MAX, b); }

			Elem& InplaceCAS(QValue refval, QValue newval)
			{
				*this = (*this == refval) ? newval : *this;
				return *this;
			}

			QValue InplaceExch(QValue newval)
			{
				QValue oldVal = *this;
				*this = newval;
				return oldVal;
			}

			// Map default operators onto QValue
			QValue operator+(QValue b) const { return operator QValue() + b; }
			QValue operator-(QValue b) const { return operator QValue() - b; }
			QValue operator*(QValue b) const { return operator QValue() * b; }
			QValue operator/(QValue b) const { return operator QValue() / b; }
			QValue operator-() const { return -operator QValue(); }
			QValue operator<(QValue b) const { return operator QValue() < b; }
			QValue operator<=(QValue b) const { return operator QValue() <= b; }
			QValue operator>(QValue b) const { return operator QValue() > b; }
			QValue operator>=(QValue b) const { return operator QValue() >= b; }

			template <int M> QValue operator+(Elem<M> b) const { return operator QValue() + (QValue)b; }
			template <int M> QValue operator-(Elem<M> b) const { return operator QValue() - (QValue)b; }
			template <int M> QValue operator*(Elem<M> b) const { return operator QValue() * (QValue)b; }
			template <int M> QValue operator/(Elem<M> b) const { return operator QValue() / (QValue)b; }
			template <int M> QValue operator<(Elem<M> b) const { return operator QValue() < (QValue)b; }
			template <int M> QValue operator<=(Elem<M> b) const { return operator QValue() <= (QValue)b; }
			template <int M> QValue operator>(Elem<M> b) const { return operator QValue() > (QValue)b; }
			template <int M> QValue operator>=(Elem<M> b) const { return operator QValue() >= (QValue)b; }
			template <int M> QValue operator==(Elem<M> b) const { return operator QValue() == (QValue)b; }
			template <int M> QValue operator!=(Elem<M> b) const { return operator QValue() != (QValue)b; }

			// Casting operators
			explicit operator float() const { return (QValue)*this; }
			explicit operator double() const { return (QValue)*this; }
			explicit operator int() const { return (QValue)*this; }
			explicit operator bool() const { return (QValue)*this; }
			explicit operator string_t() const { return (QValue)*this; }
			explicit operator cscalar() const { return (QValue)*this; }
			explicit operator int1() const { return (QValue)*this; }
			explicit operator int2() const { return (QValue)*this; }
			explicit operator int3() const { return (QValue)*this; }
			explicit operator int4() const { return (QValue)*this; }
			template <int N2>
			explicit operator intN<N2>() const { return (QValue)*this; }
			explicit operator scalar1() const { return (QValue)*this; }
			explicit operator scalar2() const { return (QValue)*this; }
			explicit operator scalar3() const { return (QValue)*this; }
			explicit operator scalar4() const { return (QValue)*this; }
			template <int N2>
			explicit operator scalarN<N2>() const { return (QValue)*this; }
			explicit operator cscalar1() const { return (QValue)*this; }
			explicit operator cscalar2() const { return (QValue)*this; }
			explicit operator cscalar3() const { return (QValue)*this; }
			explicit operator cscalar4() const { return (QValue)*this; }
			template <int N2>
			explicit operator cscalarN<N2>() const { return (QValue)*this; }

			ElemWiseOp ElemWise() const
			{
				QValue result = const_cast<Elem<N>*>(this)->operator QValue();
				return ElemWiseOp(result);
			}

			operator QValue() const;

			// Conversion function
			template <typename R>
			R as() const { return (R) operator QValue(); }

			Elem<1> operator()(const QValue &m)
			{
				qvalue_t index[1] = { m };
				return Elem<1>((QValue)* this, index, quasar::Default);
			}

			Elem<2> operator()(const QValue &m, const QValue &n)
			{
				qvalue_t index[2] = { m, n };
				return Elem<2>((QValue)* this, index, quasar::Default);
			}

			Elem<3> operator()(const QValue &m, const QValue &n,
				const QValue &k)
			{
				qvalue_t index[3] = { m, n, k };
				return Elem<3>((QValue)* this, index, quasar::Default);
			}

#ifdef CPLUSPLUS11
			template <typename...  Indices>
			Elem<sizeof...(Indices)> operator()(Indices... packedIndices)
			{
				constexpr int M = sizeof...(Indices);
				QValue indices[] = { packedIndices... };
				qvalue_t qv_indices[M];
				for (int i = 0; i < M; i++)
					qv_indices[i] = indices[i];
				return Elem<M>((QValue)*this, qv_indices, quasar::Default);
			}
#endif

			Elem<1> operator()(int m)
			{
				qvalue_t index[1] = { qvalue_t::fromScalar(m) };
				return Elem<1>((QValue)* this, index, quasar::Default);
			}

			Elem<2> operator()(int m, int n)
			{
				qvalue_t index[2] = { qvalue_t::fromScalar(m),
					qvalue_t::fromScalar(n) };
				return Elem<2>((QValue)* this, index, quasar::Default);
			}

			Elem<3> operator()(int m, int n, int k)
			{
				qvalue_t index[3] = { qvalue_t::fromScalar(m),
					qvalue_t::fromScalar(n),
					qvalue_t::fromScalar(k) };
				return Elem<3>((QValue)* this, index, quasar::Default);
			}

			QValue SetField(LPCTSTR fieldName, QValue value)
			{
				IQuasarHost::GetInstance()->SetField(operator QValue(), fieldName, value);
				return value;
			}

			QValue GetField(LPCTSTR fieldName)
			{
				qvalue_t y = qvalue_t::empty;
				IQuasarHost::GetInstance()->GetField(operator QValue(), fieldName, y);
				return y;
			}

			INLINE FieldRef Field(LPCTSTR fieldName);
		};

#ifdef _MSC_VER
		// Specialization for function calls with no arguments
		template<>
		class QValue::Elem<0>
		{
			qvalue_t mtx;

		public:
			Elem(const qvalue_t &value, qvalue_t *index,
				BoundaryAccessMode boundaryAccessMode)
			{
				mtx = value;
				IQuasarHost::GetInstance()->AddRef(mtx);
			}
			Elem(const Elem &elem)
			{
				this->mtx = elem.mtx;
				IQuasarHost::GetInstance()->AddRef(mtx);
			}
			~Elem()
			{
				IQuasarHost::GetInstance()->ReleaseRef(mtx);
			}
			operator QValue() const;
		};
#endif

		// Helper class for field operations
		class FieldRef
		{
			qvalue_t obj;
			string_t fieldName;
		public:
			FieldRef(const qvalue_t &obj, LPCTSTR fieldName)
			{
				this->obj = obj;
				this->fieldName = fieldName;
				Retain();
			}
			FieldRef(const FieldRef &field)
			{
				this->obj = field.obj;
				this->fieldName = field.fieldName;
				Retain();
			}
			~FieldRef()
			{
				IQuasarHost::GetInstance()->ReleaseRef(obj);
			}

		private:
			FieldRef& InplaceOp(OperatorTypes opType, QValue b);
			void Retain()
			{
				IQuasarHost::GetInstance()->AddRef(this->obj);
			}

		public:
			operator QValue() const;
			FieldRef& operator=(QValue b);
			FieldRef& operator+=(QValue b) { return InplaceOp(OP_ADD, b); }
			FieldRef& operator-=(QValue b) { return InplaceOp(OP_SUB, b); }
			FieldRef& operator*=(QValue b) { return InplaceOp(OP_MULTIPLY, b); }
			FieldRef& operator/=(QValue b) { return InplaceOp(OP_DIVIDE, b); }
			FieldRef& InplaceAdd(QValue b) { return InplaceOp(OP_ADD, b); }
			FieldRef& InplaceSubtract(QValue b) { return InplaceOp(OP_SUB, b); }
			FieldRef& InplaceMultiply(QValue b) { return InplaceOp(OP_MULTIPLY, b); }
			FieldRef& InplaceDivide(QValue b) { return InplaceOp(OP_DIVIDE, b); }
			FieldRef& InplacePow(QValue b) { return InplaceOp(OP_POW, b); }
			FieldRef& InplaceXor(QValue b) { return InplaceOp(OP_LOG_XOR, b); }
			FieldRef& InplaceOr(QValue b) { return InplaceOp(OP_LOG_OR, b); }
			FieldRef& InplaceAnd(QValue b) { return InplaceOp(OP_LOG_AND, b); }
			FieldRef& InplaceMin(QValue b) { return InplaceOp(OP_MIN, b); }
			FieldRef& InplaceMax(QValue b) { return InplaceOp(OP_MAX, b); }

			FieldRef& InplaceCAS(QValue refval, QValue newval)
			{
				*this = (operator QValue() == refval) ? newval : operator QValue();
				return *this;
			}

			QValue InplaceExch(QValue newval)
			{
				QValue oldVal = operator QValue();
				*this = newval;
				return oldVal;
			}

			Elem<0> operator()()
			{
				qvalue_t index[1];
				return Elem<0>(operator QValue(), index, quasar::Default);
			}

			Elem<1> operator()(QValue m, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[1] = { m };
				return Elem<1>(operator QValue(), index, accessMode);
			}

			Elem<2> operator()(QValue m, QValue n, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[2] = { m, n };
				return Elem<2>(operator QValue(), index, accessMode);
			}

			Elem<3> operator()(QValue m, QValue n, QValue k, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[3] = { m, n, k };
				return Elem<3>(operator QValue(), index, accessMode);
			}

			Elem<4> operator()(QValue m, QValue n, QValue k, QValue l, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[4] = { m, n, k, l };
				return Elem<4>(operator QValue(), index, accessMode);
			}

			Elem<5> operator()(QValue m, QValue n, QValue k, QValue l, QValue o, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[5] = { m, n, k, l, o };
				return Elem<5>(operator QValue(), index, accessMode);
			}
			Elem<6> operator()(QValue m, QValue n, QValue k, QValue l, QValue o, QValue p, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[6] = { m, n, k, l, o, p };
				return Elem<6>(operator QValue(), index, accessMode);
			}
			template <int N>
			Elem<N> operator()(QValue(&ind)[N], BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[N];
				for (int i = 0; i < N; i++)
					index[i] = ind[i];
				return Elem<N>(operator QValue(), index, accessMode);
			}

			Elem<1> operator()(int m, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[1] = { qvalue_t::fromScalar(m) };
				return Elem<1>(operator QValue(), index, accessMode);
			}

			Elem<2> operator()(int m, int n, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[2] = { qvalue_t::fromScalar(m),
					qvalue_t::fromScalar(n) };
				return Elem<2>(operator QValue(), index, accessMode);
			}

			Elem<3> operator()(int m, int n, int k, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[3] = { qvalue_t::fromScalar(m),
					qvalue_t::fromScalar(n),
					qvalue_t::fromScalar(k) };
				return Elem<3>(operator QValue(), index, accessMode);
			}

			Elem<4> operator()(int m, int n, int k, int l, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[4] = { qvalue_t::fromScalar(m),
					qvalue_t::fromScalar(n),
					qvalue_t::fromScalar(k),
					qvalue_t::fromScalar(l) };
				return Elem<4>(operator QValue(), index, accessMode);
			}

			Elem<5> operator()(int m, int n, int k, int l, int o, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[5] = { qvalue_t::fromScalar(m),
					qvalue_t::fromScalar(n),
					qvalue_t::fromScalar(k),
					qvalue_t::fromScalar(l),
					qvalue_t::fromScalar(o) };
				return Elem<5>(operator QValue(), index, accessMode);
			}

			Elem<6> operator()(int m, int n, int k, int l, int o, int p, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t index[6] = { qvalue_t::fromScalar(m),
					qvalue_t::fromScalar(n),
					qvalue_t::fromScalar(k),
					qvalue_t::fromScalar(l),
					qvalue_t::fromScalar(o),
					qvalue_t::fromScalar(p) };
				return Elem<6>(operator QValue(), index, accessMode);
			}

			template <int N>
			Elem<N> operator()(intN<N> index, BoundaryAccessMode accessMode = quasar::Default)
			{
				qvalue_t ind[N];
				for (int i = 0; i < N; i++)
					ind[i] = qvalue_t::fromScalar(index.el[i]);
				return Elem<N>(operator QValue(), index, accessMode);
			}

			ElemWiseOp ElemWise() const
			{
				return ElemWiseOp((QValue)*const_cast<FieldRef*>(this));
			}

			QValue SetField(LPCTSTR fieldName, QValue value)
			{
				IQuasarHost::GetInstance()->SetField(operator QValue(), fieldName, value);
				return value;
			}

			QValue GetField(LPCTSTR fieldName)
			{
				qvalue_t y = qvalue_t::empty;
				IQuasarHost::GetInstance()->GetField(operator QValue(), fieldName, y);
				return y;
			}

			FieldRef Field(LPCTSTR fieldName)
			{
				return FieldRef(operator QValue(), fieldName);
			}

			// Casting operators
			explicit operator float() const { return (QValue)*this; }
			explicit operator double() const { return (QValue)*this; }
			explicit operator int() const { return (QValue)*this; }
			explicit operator bool() const { return (QValue)*this; }
			explicit operator string_t() const { return (QValue)*this; }
			explicit operator cscalar() const { return (QValue)*this; }
			explicit operator int1() const { return (QValue)*this; }
			explicit operator int2() const { return (QValue)*this; }
			explicit operator int3() const { return (QValue)*this; }
			explicit operator int4() const { return (QValue)*this; }
			template <int N2>
			explicit operator intN<N2>() const { return (QValue)*this; }
			explicit operator scalar1() const { return (QValue)*this; }
			explicit operator scalar2() const { return (QValue)*this; }
			explicit operator scalar3() const { return (QValue)*this; }
			explicit operator scalar4() const { return (QValue)*this; }
			template <int N2>
			explicit operator scalarN<N2>() const { return (QValue)*this; }
			explicit operator cscalar1() const { return (QValue)*this; }
			explicit operator cscalar2() const { return (QValue)*this; }
			explicit operator cscalar3() const { return (QValue)*this; }
			explicit operator cscalar4() const { return (QValue)*this; }
			template <int N2>
			explicit operator cscalarN<N2>() const { return (QValue)*this; }
		};

		QValue(const qvalue_t &value) { this->value = value; }

		QValue(const QValue &other)
		{
			this->value = other.value;
			Retain();
		}

		QValue(LPCTSTR str)
		{
			value = IQuasarHost::GetInstance()->CreateString(str);
		}
		QValue(string_t str)
		{
			value = IQuasarHost::GetInstance()->CreateString(str.get_buf(), str.get_length());
		}
		template <typename T, int N> QValue(T(&vals)[N])
		{
			value = IQuasarHost::GetInstance()->CreateVector(vals);
		}

		template <class R>
		static QValue FromObj(const R val)
		{
#ifdef CPLUSPLUS11
			static_assert(typeinfo_t<R>::type == TYPE_TYPEDOBJECT, "QValue::FromObj: Typed object expected!");
#endif
			IQuasarHost * host = IQuasarHost::GetInstance();
			qvalue_t value = host->CreateTypedObject(host->GetTypeHandle<R>());
			R* ptr = host->LockObject<R>(value, LOCK_WRITE);
			*ptr = val;
			host->UnlockObject(value, LOCK_WRITE);
			return value;
		}

		// Construct a QValue from a fixed length vector
		QValue(scalar1 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<scalar(&)[1]>(val)); }
		QValue(scalar2 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<scalar(&)[2]>(val)); }
		QValue(scalar3 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<scalar(&)[3]>(val)); }
		QValue(scalar4 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<scalar(&)[4]>(val)); }
		QValue(cscalar1 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<cscalar(&)[1]>(val)); }
		QValue(cscalar2 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<cscalar(&)[2]>(val)); }
		QValue(cscalar3 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<cscalar(&)[3]>(val)); }
		QValue(cscalar4 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<cscalar(&)[4]>(val)); }
		QValue(int1 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<int(&)[1]>(val)); }
		QValue(int2 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<int(&)[2]>(val)); }
		QValue(int3 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<int(&)[3]>(val)); }
		QValue(int4 val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<int(&)[4]>(val)); }
		template <int N>
		QValue(scalarN<N> val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<scalar(&)[N]>(val)); }
		template <int N>
		QValue(cscalarN<N> val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<cscalar(&)[N]>(val)); }
		template <int N>
		QValue(intN<N> val) { value = IQuasarHost::GetInstance()->CreateVector(reinterpret_cast<int(&)[N]>(val)); }

		template <int N>
		QValue(Elem<N> val)
		{
			value = qvalue_t::empty;
			*this = val.operator QValue();
		}

		QValue(FieldRef fieldRef)
		{
			value = qvalue_t::empty;
			*this = fieldRef.operator QValue();
		}

		void Retain() { IQuasarHost::GetInstance()->AddRef(value); }

		qvalue_t *operator->() { return &value; }
		const qvalue_t *operator->() const { return &value; }
		operator const qvalue_t &() const { return value; }
		operator scalar() const
		{
			_ASSERT_EXPR(value.type == TYPE_INT || value.type == TYPE_SCALAR,
				"QValue::operator scalar() can only be called on scalar numbers");
			return value.type == TYPE_INT ? value.intVal : value.scalarVal;
		}
		operator cscalar() const
		{
			_ASSERT_EXPR(value.type == TYPE_INT || value.type == TYPE_SCALAR || value.type == TYPE_COMPLEXSCALAR,
				"QValue::operator cscalar() can only be called on scalar numbers");
			return value.type == TYPE_INT ? cscalar(value.intVal) :
				value.type == TYPE_SCALAR ? cscalar(value.scalarVal) : cscalar(value.realVal, value.imagVal);
		}
#ifdef DBL_SCALAR
		operator float() const { return operator scalar(); }
#else
		operator double() const { return operator scalar(); }
#endif
		operator int() const
		{
			_ASSERT_EXPR(value.type == TYPE_INT || value.type == TYPE_SCALAR,
				"QValue::operator int() can only be called on scalar numbers");
			return value.type == TYPE_INT ? value.intVal : (int)value.scalarVal;
		}
		operator bool() const
		{
			return this->operator int() != 0;
		}
		operator string_t() const
		{
			_ASSERT_EXPR(value.type == TYPE_STRING,
				"QValue::operator string_t() can only be called on strings");
			IQuasarHost* host = IQuasarHost::GetInstance();
			string_t strValue;
			host->GetString(value, strValue);
			return strValue;
		}
	private:
		template <class R> R read_ivecX() const
		{
			_ASSERT_EXPR(value.type == TYPE_VEC, "Fixed-length vector expected!");
			R retval;
			IQuasarHost* host = IQuasarHost::GetInstance();
			if (value.elem_type == TYPE_INT)
			{
				R *ptr = host->LockObject<R>(value, LOCK_READ);
				retval = *ptr;
				host->UnlockObject(value, LOCK_READ);
			}
			else if (value.elem_type == TYPE_SCALAR)
			{
				scalar *ptr = host->LockObject<scalar>(value, LOCK_READ);
				const int length = sizeof(R) / sizeof(int);
				for (int i = 0; i < length; i++)
					((int *)&retval)[i] = (int)ptr[i];
				host->UnlockObject(value, LOCK_READ);
			}
			else _ASSERT_EXPR(false, "Unsupported conversion operation!");
			return retval;
		}
		template <class R> R read_vecX() const
		{
			_ASSERT_EXPR(value.type == TYPE_VEC || value.type == TYPE_MAT ||
						 value.type == TYPE_CUBE || value.type == TYPE_NCUBE, "Fixed-length vector expected!");
			R retval;
			IQuasarHost* host = IQuasarHost::GetInstance();
			if (value.elem_type == TYPE_SCALAR)
			{
				R *ptr = host->LockObject<R>(value, LOCK_READ);
				retval = *ptr;
				host->UnlockObject(value, LOCK_READ);
			}
			else if (value.elem_type == TYPE_INT)
			{
				int *ptr = host->LockObject<int>(value, LOCK_READ);
				const int length = sizeof(R) / sizeof(int);
				for (int i = 0; i < length; i++)
					((scalar *)&retval)[i] = (scalar)ptr[i];
				host->UnlockObject(value, LOCK_READ);
			}
			else _ASSERT_EXPR(false, "Unsupported conversion operation!");
			return retval;
		}
		template <class R> R read_cvecX() const
		{
			_ASSERT_EXPR(value.type == TYPE_CVEC || value.type == TYPE_CMAT ||
				value.type == TYPE_CCUBE || value.type == TYPE_NCCUBE, "Fixed-length vector expected!");
			IQuasarHost* host = IQuasarHost::GetInstance();
			R *ptr = host->LockObject<R>(value, LOCK_READ);
			R retval = *ptr;
			host->UnlockObject(value, LOCK_READ);
			return retval;
		}
	public:
		operator int1() const { return read_ivecX<int1>(); }
		operator int2() const { return read_ivecX<int2>(); }
		operator int3() const { return read_ivecX<int3>(); }
		operator int4() const { return read_ivecX<int4>(); }
		template <int N>
		operator intN<N>() const { return read_ivecX<intN<N>>(); }

		operator scalar1() const { return read_vecX<scalar1>(); }
		operator scalar2() const { return read_vecX<scalar2>(); }
		operator scalar3() const { return read_vecX<scalar3>(); }
		operator scalar4() const { return read_vecX<scalar4>(); }
		template <int N>
		operator scalarN<N>() const { return read_vecX<scalarN<N>>(); }

		operator cscalar1() const { return read_cvecX<cscalar1>(); }
		operator cscalar2() const { return read_cvecX<cscalar2>(); }
		operator cscalar3() const { return read_cvecX<cscalar3>(); }
		operator cscalar4() const { return read_cvecX<cscalar4>(); }
		template <int N>
		operator cscalarN<N>() const { return read_cvecX<cscalarN<N>>(); }

		template <class R>
		explicit operator R() const
		{
			R retval = R();
			switch (value.type)
			{
			case TYPE_SCALAR:
			case TYPE_COMPLEXSCALAR:
				retval = *static_cast<R*>((void *)&value.scalarVal);
				break;
			case TYPE_INT:
				retval = *static_cast<R*>((void *)&value.intVal);
				break;
			case TYPE_TYPEDOBJECT:
				IQuasarHost* host = IQuasarHost::GetInstance();
				R *ptr = host->LockObject<R>(value, LOCK_READ);
				retval = *ptr;
				host->UnlockObject(value, LOCK_READ);
				break;
			}
			return retval;
		}
#ifdef CPLUSPLUS11
		qvalue_t &operator=(QValue && other)
		{
			ReleaseRef();
			value = other.value;
			other.value = qvalue_t::empty;
			return value;
		}
#endif

		qvalue_t &operator=(const QValue &other)
		{
			ReleaseRef();
			value = other.value;
			IQuasarHost::GetInstance()->AddRef(value);
			return value;
		}

		Elem<0> operator()() const
		{
			qvalue_t index[1] = { qvalue_t::empty };
			return Elem<0>(value, index, quasar::Default);
		}

		Elem<1> operator()(QValue m, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[1] = { m };
			return Elem<1>(value, index, accessMode);
		}

		Elem<2> operator()(QValue m, QValue n, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[2] = { m, n };
			return Elem<2>(value, index, accessMode);
		}

		Elem<3> operator()(QValue m, QValue n, QValue k, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[3] = { m, n, k };
			return Elem<3>(value, index, accessMode);
		}

		Elem<4> operator()(QValue m, QValue n, QValue k, QValue l, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[4] = { m, n, k, l };
			return Elem<4>(value, index, accessMode);
		}

		Elem<5> operator()(QValue m, QValue n, QValue k, QValue l, QValue o, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[5] = { m, n, k, l, o };
			return Elem<5>(value, index, accessMode);
		}
		Elem<6> operator()(QValue m, QValue n, QValue k, QValue l, QValue o, QValue p, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[6] = { m, n, k, l, o, p };
			return Elem<6>(value, index, accessMode);
		}
		template <int N>
		Elem<N> operator()(QValue(&ind)[N], BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[N];
			for (int i = 0; i < N; i++)
				index[i] = ind[i];
			return Elem<N>(value, index, accessMode);
		}

		Elem<1> operator()(int m, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[1] = { qvalue_t::fromScalar(m) };
			return Elem<1>(value, index, accessMode);
		}

		Elem<2> operator()(int m, int n, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[2] = { qvalue_t::fromScalar(m),
				qvalue_t::fromScalar(n) };
			return Elem<2>(value, index, accessMode);
		}

		Elem<3> operator()(int m, int n, int k, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[3] = { qvalue_t::fromScalar(m),
				qvalue_t::fromScalar(n),
				qvalue_t::fromScalar(k) };
			return Elem<3>(value, index, accessMode);
		}

		Elem<4> operator()(int m, int n, int k, int l, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[4] = { qvalue_t::fromScalar(m),
				qvalue_t::fromScalar(n),
				qvalue_t::fromScalar(k),
				qvalue_t::fromScalar(l) };
			return Elem<4>(value, index, accessMode);
		}

		Elem<5> operator()(int m, int n, int k, int l, int o, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[5] = { qvalue_t::fromScalar(m),
				qvalue_t::fromScalar(n),
				qvalue_t::fromScalar(k),
				qvalue_t::fromScalar(l),
				qvalue_t::fromScalar(o) };
			return Elem<5>(value, index, accessMode);
		}

		Elem<6> operator()(int m, int n, int k, int l, int o, int p, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t index[6] = { qvalue_t::fromScalar(m),
				qvalue_t::fromScalar(n),
				qvalue_t::fromScalar(k),
				qvalue_t::fromScalar(l),
				qvalue_t::fromScalar(o),
				qvalue_t::fromScalar(p) };
			return Elem<6>(value, index, accessMode);
		}

		template <int N>
		Elem<N> operator()(intN<N> index, BoundaryAccessMode accessMode = quasar::Default) const
		{
			qvalue_t ind[N];
			for (int i = 0;i<N;i++)
				ind[i] = qvalue_t::fromScalar(index.el[i]);
			return Elem<N>(value, index, accessMode);
		}

#ifdef CPLUSPLUS14
		// Call a function with multiple input and output values
		template <int N, typename ...T>
		decltype(utility::tuple_of_length<QValue, N>()) operator()(T... packedArgs);

		// Call a function with zero input values and multiple output values
		template <int N>
		decltype(utility::tuple_of_length<QValue, N>()) operator()();
#endif

		ElemWiseOp ElemWise() const {
			return ElemWiseOp(*const_cast<QValue *>(this));
		}

		void ReleaseRef() { IQuasarHost::GetInstance()->ReleaseRef(value); }
		// Note: the destructor is non-virtual to avoid frame pointer. Be aware
		// of possible memory leaks when a child object is released through a
		// base pointer. By design, all QValue derived instances should be such
		// that only the value is kept (and needs to be released).
		~QValue() {
			IQuasarHost* host = IQuasarHost::GetInstance();
			if (host) host->DeleteValue(value);
		}

        QValue SetField(LPCTSTR fieldName, QValue value)
        {
            IQuasarHost::GetInstance()->SetField(this->value, fieldName, value);
			return value;
        }

        QValue GetField(LPCTSTR fieldName)
        {
            qvalue_t y = qvalue_t::empty;
            IQuasarHost::GetInstance()->GetField(this->value, fieldName, y);
            return y;
        }

		FieldRef Field(LPCTSTR fieldName)
		{
			return FieldRef(this->value, fieldName);
		}

		template <typename T> static QValue CreateVector(int numel)
		{
			return IQuasarHost::GetInstance()->CreateVector<T>(numel);
		}

		template <typename T> static QValue CreateVector(int numel, const T *vals)
		{
			return IQuasarHost::GetInstance()->CreateVector<T>(numel, vals);
		}

#ifdef CPLUSPLUS11
		template <typename T>  static QValue CreateVector(std::initializer_list<T> vals)
		{
			return IQuasarHost::GetInstance()->CreateVector<T>((int)vals.size(), vals.begin());
		}
#endif

		// Creation of generic matrices
		template <typename T> static QValue CreateMatrix(int dim1, int dim2)
		{
			return IQuasarHost::GetInstance()->CreateMatrix<T>(dim1, dim2);
		}

		template <typename T>
		static QValue CreateMatrix(int dim1, int dim2, const T *vals)
		{
			return IQuasarHost::GetInstance()->CreateMatrix<T>(dim1, dim2,
				vals);
		}

		// Creation of matrices from slices
		static QValue CreateMatrix(const QValue *slices, int N)
		{
			IQuasarHost * host = IQuasarHost::GetInstance();
			IComputationEngine * engine = host->GetComputationEngine();
			IEvaluationStack * stack = engine->GetEvaluationStack();
			for (int i = 0; i < N; i++)
				stack->PushValue(slices[i]);
			engine->ConstructMatrix(N);
			return stack->PopValue();
		}

		template <int N>
		static QValue CreateMatrix(QValue(&slices)[N])
		{
			return CreateMatrix(slices, N);
		}

#ifdef CPLUSPLUS11
		static QValue CreateMatrix(std::initializer_list<QValue> list)
		{
			return CreateMatrix(list.begin(), (int)list.size());
		}
#endif

		// Creation of cell matrices from slices
		static qvalue_t CreateCellMatrix_helper(const QValue *slices, int N)
		{
			IQuasarHost * host = IQuasarHost::GetInstance();
			IComputationEngine * engine = host->GetComputationEngine();
			IEvaluationStack * stack = engine->GetEvaluationStack();
			for (int i = 0; i < N; i++)
				stack->PushValue(slices[i]);
			engine->ConstructCellMatrix(N);
			return stack->PopValue();
		}

		static QValue CreateCellMatrix(const QValue *slices, int N)
		{
			return (QValue)CreateCellMatrix_helper(slices, N);
		}

		template <int N>
		static QValue CreateCellMatrix(QValue(&slices)[N])
		{
			return CreateCellMatrix(slices, N);
		}

#ifdef CPLUSPLUS11
		static QValue CreateCellMatrix(std::initializer_list<QValue> list)
		{
			return CreateCellMatrix(list.begin(), (int)list.size());
		}
#endif

#ifdef CPLUSPLUS14
		template <typename ...T>
		static QValue CreateCellMatrix(std::tuple<T...> tuple)
		{
			return QValue(tuple);
		}
#endif

		// Creation of generic cubes
		template <typename T>
		static QValue CreateCube(int dim1, int dim2, int dim3)
		{
			return IQuasarHost::GetInstance()->CreateCube<T>(dim1, dim2, dim3);
		}

		// Creation of higher dimensional cubes
		template <typename T, int N>
		static QValue CreateNCube(intN<N> dims)
		{
			return IQuasarHost::GetInstance()->CreateNCube<T>(N, dims.el);
		}

		// Creation of objects
		static QUASARHOSTAPI QValue CreateObject(const Type& type);

		template <int N>
		static QValue CreateObject(const Type& type, NamedParam(&params)[N]);

	public: // Operators
		QValue operator+(QValue b) const { return BinaryOp(*this, b, OP_ADD); }
		QValue operator-(QValue b) const { return BinaryOp(*this, b, OP_SUB); }
		QValue operator*(QValue b) const { return BinaryOp(*this, b, OP_MULTIPLY); }
		QValue operator/(QValue b) const { return BinaryOp(*this, b, OP_DIVIDE); }
		QValue operator-() const { return UnaryOp(*this, OP_NEGATE); }
		QValue operator<(QValue b) const { return BinaryOp(*this, b, OP_LESS); }
		QValue operator<=(QValue b) const { return BinaryOp(*this, b, OP_LESSOREQ); }
		QValue operator>(QValue b) const { return BinaryOp(*this, b, OP_GREATER); }
		QValue operator>=(QValue b) const { return BinaryOp(*this, b, OP_GREATEROREQ); }
		QValue operator&&(QValue b) const { return BinaryOp(*this, b, OP_LOG_AND); }
		QValue operator||(QValue b) const { return BinaryOp(*this, b, OP_LOG_OR); }
		QValue operator^(QValue b) const { return BinaryOp(*this, b, OP_LOG_XOR); }
		QValue operator!() const { return UnaryOp(*this, OP_INVERT); }

		QValue operator+(int b) const { return operator+(QValue(b)); }
		QValue operator-(int b) const { return operator-(QValue(b)); }
		QValue operator*(int b) const { return operator*(QValue(b)); }
		QValue operator/(int b) const { return operator/(QValue(b)); }

		QValue operator+(float b) const { return operator+(QValue(b)); }
		QValue operator-(float b) const { return operator-(QValue(b)); }
		QValue operator*(float b) const { return operator*(QValue(b)); }
		QValue operator/(float b) const { return operator/(QValue(b)); }

		QValue operator+(double b) const { return operator+(QValue(b)); }
		QValue operator-(double b) const { return operator-(QValue(b)); }
		QValue operator*(double b) const { return operator*(QValue(b)); }
		QValue operator/(double b) const { return operator/(QValue(b)); }

		QValue& operator+=(QValue b) { return InplaceOp(OP_ADD, b); }
		QValue& operator-=(QValue b) { return InplaceOp(OP_SUB, b); }
		QValue& operator*=(QValue b) { return InplaceOp(OP_MULTIPLY, b); }
		QValue& operator/=(QValue b) { return InplaceOp(OP_DIVIDE, b); }

		QValue& operator+=(int b) { *this = *this + b; return *this; }
		QValue& operator-=(int b) { *this = *this - b; return *this; }
		QValue& operator*=(int b) { *this = *this * b; return *this; }
		QValue& operator/=(int b) { *this = *this / b; return *this; }

		QValue& operator+=(float b) { *this = *this + b; return *this; }
		QValue& operator-=(float b) { *this = *this - b; return *this; }
		QValue& operator*=(float b) { *this = *this * b; return *this; }
		QValue& operator/=(float b) { *this = *this / b; return *this; }

		QValue& operator+=(double b) { *this = *this + b; return *this; }
		QValue& operator-=(double b) { *this = *this - b; return *this; }
		QValue& operator*=(double b) { *this = *this * b; return *this; }
		QValue& operator/=(double b) { *this = *this / b; return *this; }

		QValue& InplaceAdd(QValue b) { return InplaceOp(OP_ADD_ASSIGN, b); }
		QValue& InplaceSubtract(QValue b) { return InplaceOp(OP_SUB_ASSIGN, b); }
		QValue& InplaceMultiply(QValue b) { return InplaceOp(OP_MULTIPLY_ASSIGN, b); }
		QValue& InplaceDivide(QValue b) { return InplaceOp(OP_DIVIDE_ASSIGN, b); }
		QValue& InplaceRDivide(QValue b) { return InplaceOp(OP_RDIVIDE_ASSIGN, b); }
		QValue& InplacePow(QValue b) { return InplaceOp(OP_POW_ASSIGN, b); }
		QValue& InplacePwMultiply(QValue b) { return InplaceOp(OP_PW_MULTIPLY_ASSIGN, b); }
		QValue& InplacePwDivide(QValue b) { return InplaceOp(OP_PW_DIVIDE_ASSIGN, b); }
		QValue& InplacePwPow(QValue b) { return InplaceOp(OP_PW_POW_ASSIGN, b); }
		QValue& InplaceXor(QValue b) { return InplaceOp(OP_XOR_ASSIGN, b); }
		QValue& InplaceOr(QValue b) { return InplaceOp(OP_OR_ASSIGN, b); }
		QValue& InplaceAnd(QValue b) { return InplaceOp(OP_AND_ASSIGN, b); }
		QValue& InplaceMin(QValue b) { return InplaceOp(OP_MIN_ASSIGN, b); }
		QValue& InplaceMax(QValue b) { return InplaceOp(OP_MAX_ASSIGN, b); }

		QValue& InplaceCAS(QValue refval, QValue newval)
		{
			*this = (*this == refval) ? newval : *this;
			return *this;
		}

		QValue InplaceExch(QValue newval)
		{
			QValue oldVal = *this;
			*this = newval;
			return oldVal;
		}

		QValue ElemwiseEquals(QValue b) {
			return BinaryOp(*this, b, OP_EQUAL);
		}

		QValue ElemwiseNotEquals(QValue b) {
			return BinaryOp(*this, b, OP_NOTEQUAL);
		}

		bool Equals(QValue b) const {
			return value.type == b.value.type &&
				value.private_obj == b.value.private_obj ||
				BinaryOp(*this, b, OP_EQUAL);
		}
		bool NEquals(QValue b) const {
			return (value.type != b.value.type ||
				value.private_obj != b.value.private_obj) &&
				BinaryOp(*this, b, OP_NOTEQUAL);
		}
		QValue operator==(QValue b) const {
			return BinaryOp(*this, b, OP_EQUAL);
		}
		QValue operator!=(QValue b) const {
			return BinaryOp(*this, b, OP_NOTEQUAL);
		}
		Type GetType() const;
		bool HasType(Type type) const;
		bool IsEmpty() const {
			return this->value.type == TYPE_VOID;
		}

		static QValue UnaryOp(QValue a, OperatorTypes op)
		{
			IEvaluationStack *stack = IQuasarHost::GetInstance()
				->GetComputationEngine()
				->GetEvaluationStack();

			stack->PushValue(a);
			stack->Process(op);
			return stack->PopValue();
		}
		static QValue BinaryOp(QValue a, QValue b, OperatorTypes op)
		{
			IEvaluationStack *stack = IQuasarHost::GetInstance()
				->GetComputationEngine()
				->GetEvaluationStack();

			stack->PushValue(a);
			stack->PushValue(b);
			stack->Process(op);
			return stack->PopValue();
		}

		QValue& InplaceOp(OperatorTypes opType, QValue b)
		{
			*this = BinaryOp(*this, b, opType);
			return *this;
		}


		template <int N>
		QValue InvokeMethod(LPCTSTR methodName, QValue (& args)[N])
		{
			IQuasarHost *host = IQuasarHost::GetInstance();
			qvalue_t t_args[N];
			qvalue_t out_arg = qvalue_t::empty;
			for (int i = 0; i < N; i++)
				t_args[i] = (qvalue_t)args[i];
			host->MethodCall(methodName, value, t_args, N, &out_arg, 1);
			return (QValue)out_arg;
		}

		QValue InvokeMethod(LPCTSTR methodName)
		{
			IQuasarHost *host = IQuasarHost::GetInstance();
			qvalue_t out_arg = qvalue_t::empty;
			host->MethodCall(methodName, value, NULL, 0, &out_arg, 1);
			return (QValue)out_arg;
		}

#ifdef CPLUSPLUS11
		template <typename...  Args>
		QValue InvokeMethod(LPCTSTR methodName, Args... packedArgs)
		{
			QValue args[] = { (QValue)packedArgs... };
			return InvokeMethod(methodName, args);
		}
#endif

#ifdef CPLUSPLUS11
		static QUASARHOSTAPI int SwitchSelector(std::initializer_list<QValue> list, QValue value);
#endif

#ifdef CPLUSPLUS14
	private:
		template <int N, typename R, int ...S>
		static R CreateTuple2(QValue(&args)[N], utility::seq<S...>)
		{
			return std::make_tuple(args[S]...);
		}
	public:
		// Unpacking function
		template <typename ...R>
		std::tuple<R...> Unpack()
		{
			_ASSERT(value.type == TYPE_VEC || value.type == TYPE_CVEC);
			const int N = sizeof...(R);
			QValue args[N];
			for (int i = 0; i < N; i++)
				args[i] = this->operator()(i, BoundaryAccessMode::Default);
			return CreateTuple2<N, std::tuple<R...>>(args, utility::gen_seq<N>());
		}
#endif

		static bool WriteHostVariable(IQuasarHost *host, LPCTSTR name, QValue value)
		{
			return host->WriteVariable(name, value);
		}
		static bool WriteHostVariable(LPCTSTR name, QValue value)
		{
			return WriteHostVariable(IQuasarHost::GetInstance(), name, value);
		}
		static bool ReadHostVariable(IQuasarHost *host, LPCTSTR name, QValue &value)
		{
			qvalue_t qv = qvalue_t::empty;
			bool result = host->ReadVariable(name, qv);
			value = QValue(qv);
			return result;
		}
		static bool ReadHostVariable(LPCTSTR name, QValue &value)
		{
			return ReadHostVariable(IQuasarHost::GetInstance(), name, value);
		}
	};

	INLINE QValue operator*(QValue::ElemWiseOp a, QValue b)
	{
		return QValue::BinaryOp(a, b, OP_PW_MULTIPLY);
	}

	INLINE QValue operator/(QValue::ElemWiseOp a, QValue b)
	{
		return QValue::BinaryOp(a, b, OP_PW_DIVIDE);
	}

	INLINE QValue operator+(float b, QValue c) { return QValue(b) + c; }
	INLINE QValue operator-(float b, QValue c) { return QValue(b) - c; }
	INLINE QValue operator*(float b, QValue c) { return QValue(b) * c; }
	INLINE QValue operator/(float b, QValue c) { return QValue(b) / c; }

	INLINE QValue operator+(double b, QValue c) { return QValue(b) + c; }
	INLINE QValue operator-(double b, QValue c) { return QValue(b) - c; }
	INLINE QValue operator*(double b, QValue c) { return QValue(b) * c; }
	INLINE QValue operator/(double b, QValue c) { return QValue(b) / c; }

	INLINE QValue operator+(cscalar b, QValue c) { return QValue(b) + c; }
	INLINE QValue operator-(cscalar b, QValue c) { return QValue(b) - c; }
	INLINE QValue operator*(cscalar b, QValue c) { return QValue(b) * c; }
	INLINE QValue operator/(cscalar b, QValue c) { return QValue(b) / c; }

	// Provides information about Quasar types
	class Type : public QValue
	{
		friend class TypeBuilder;

	protected:
		DataTypes dataType;
		int maxDims;
		int vectorLength;
		AccessModifiers accessModifier;
		bool isIntegerType;
		bool isPointer;
		bool genericTypeParametersInitialized;
		std::vector<Type> genericTypeParameters;
		std::vector<Type> instanceTypeParameters;
		std::vector<Type> baseClass;

	public:
		Type() : QValue(), genericTypeParametersInitialized(false) {} // Default constructor
		Type(const qvalue_t &value) : QValue(value), genericTypeParametersInitialized(false)
		{
			_ASSERT(value.type == TYPE_TYPEINFO);
			_ASSERT(value.private_obj != NULL);
			IQuasarHost::GetInstance()->TypeGetInfo(value, &dataType, &maxDims, &vectorLength, &accessModifier, &isIntegerType,
				&isPointer);
		}
		Type(const QValue &mvalue) : QValue(mvalue), genericTypeParametersInitialized(false)
		{
			_ASSERT(value.type == TYPE_TYPEINFO);
			IQuasarHost::GetInstance()->TypeGetInfo(value, &dataType, &maxDims, &vectorLength, &accessModifier, &isIntegerType,
				&isPointer);
		}
		Type(LPCTSTR typeName): genericTypeParametersInitialized(false)
		{
			value = IQuasarHost::GetInstance()->LookupType(typeName);
			_ASSERT(value.type == TYPE_TYPEINFO);
			_ASSERT(value.private_obj != NULL);
			IQuasarHost::GetInstance()->TypeGetInfo(value, &dataType, &maxDims, &vectorLength, &accessModifier, &isIntegerType,
				&isPointer);
		}

		Type(const Type &other) : QValue(other), dataType(other.dataType),
			maxDims(other.maxDims), vectorLength(other.vectorLength),
			accessModifier(other.accessModifier), isIntegerType(other.isIntegerType),
			isPointer(other.isPointer)
		{
		}
		Type& operator= (const Type &other)
		{
			IQuasarHost *host = IQuasarHost::GetInstance();
			host->AddRef(const_cast<qvalue_t &>(other.value));
			host->ReleaseRef(this->value);
			this->value = other.value;
			maxDims = other.maxDims;
			vectorLength = other.vectorLength;
			accessModifier = other.accessModifier;
			isIntegerType = other.isIntegerType;
			isPointer = other.isPointer;
			dataType = other.dataType;
			return *this;
		}
		string_t ToString()
		{
			return IQuasarHost::GetInstance()->TypeGetString(value);
		}
		operator qhandle() const { return value.private_obj; }
		static bool IsVectorType(DataTypes dataType) { return dataType == TYPE_VEC || dataType == TYPE_CVEC; }
		static bool IsMatrixType(DataTypes dataType) { return dataType == TYPE_MAT || dataType == TYPE_CMAT || IsVectorType(dataType); }
		static bool IsCubeType(DataTypes dataType) { return dataType == TYPE_CUBE || dataType == TYPE_CCUBE || IsMatrixType(dataType); }
		static bool IsNCubeType(DataTypes dataType) { return dataType == TYPE_NCUBE || dataType == TYPE_NCCUBE || IsCubeType(dataType); }
		DataTypes GetDataType() const { return dataType; }
		int GetMaxDims() const { return maxDims; }
		int GetVectorLength() const { return vectorLength; }
		AccessModifiers GetAccessModifier() const { return accessModifier; }
		bool IsIntegerType() const { return isIntegerType; }
		bool IsPointer() const { return isPointer; }

	private:
		void InternalCtor(const qvalue_t *inArgs, int nInArgs, qvalue_t *outArg)
		{
			*outArg = IQuasarHost::GetInstance()->CreateGeneric(value.private_obj, inArgs, nInArgs);
		}
	public:
		// operator() is used for constructing objects of this type
		QValue operator()(void)
		{
			qvalue_t retval = qvalue_t::empty;
			InternalCtor(NULL, 0, &retval);
			return retval;
		}
		QValue operator()(QValue a)
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a };
			InternalCtor(args, 1, &retval);
			return retval;
		}
		QValue operator()(QValue a, QValue b)
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b };
			InternalCtor(args, 2, &retval);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c)
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c };
			InternalCtor(args, 3, &retval);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c, QValue d)
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c, d };
			InternalCtor(args, 4, &retval);
			return retval;
		}

		template <int N>
		QValue operator()(QValue(&args)[N])
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t t_args[N];
			for (int index = 0; index < N; index++) t_args[index] = (qvalue_t)args[index];
			IQuasarHost *host = IQuasarHost::GetInstance();
			InternalCtor(t_args, N, &retval);
			for (int index = 0; index < N; index++)
				host->ReleaseRef(t_args[index]);
			return retval;
		}

		// Instantiation of generic types
		template <int N>
		Type Instantiate(Type(&args)[N])
		{
			qhandle type_params[N];
			for (int index = 0; index < N; index++) type_params[index] = args[index].value.private_obj;
			qvalue_t result;
			result.type = TYPE_TYPEINFO;
			result.private_obj = IQuasarHost::GetInstance()->InstantiateType(value.private_obj, type_params, N);
			return Type(result);
		}

#ifdef CPLUSPLUS11
		// Call a function with variable arguments and one return value
		template <typename ...T>
		QValue operator()(T... packedArgs)
		{
			QValue args[] = { (QValue)packedArgs... };
			return operator()(args);
		}
#endif

	private:
		static Type from_handle(qhandle handle)
		{
			qvalue_t result;
			result.type = TYPE_TYPEINFO;
			result.private_obj = handle;
			return Type(result);
		}

#ifdef CPLUSPLUS11
	public:
		// Type instantiation using the operator [{... type_list... }]
		Type operator[](std::initializer_list<Type> args)
		{
			_ASSERT_EXPR(args.size() <= 256, "Type instantiation: too many type parameters");
			qhandle type_params[256];
			for (size_t index = 0; index < args.size(); index++) type_params[index] = args.begin()[index].value.private_obj;
			return from_handle(IQuasarHost::GetInstance()->InstantiateType(value.private_obj, type_params, (int)args.size()));
		}
#endif

		// Construct a function type, for example [__device__ (scalar, mat)->scalar]
		static Type FunctionType(FunctionTypes functionType,
			std::initializer_list<Type> inArgTypes,
			std::initializer_list<Type> outArgTypes,
			FunctionTypeFlags flags)
		{
			_ASSERT_EXPR(inArgTypes.size() <= 256 && outArgTypes.size() <= 256, "FunctionType: Too many input/output arguments");
			qhandle in_type[256], out_type[256];
			for (size_t index = 0; index < inArgTypes.size(); index++)  in_type[index] = inArgTypes.begin()[index].value.private_obj;
			for (size_t index = 0; index < outArgTypes.size(); index++) out_type[index] = outArgTypes.begin()[index].value.private_obj;
			return from_handle(
				IQuasarHost::GetInstance()->CreateFunctionType(functionType,
					in_type, (int)inArgTypes.size(), out_type, (int)outArgTypes.size(), flags));
		}

		// Construct a type class, for example [int|scalar|vec|cvec]
		static Type TypeClass(std::initializer_list<Type> optionClasses)
		{
			_ASSERT_EXPR(optionClasses.size() <= 256, "TypeClass: Too many input/output arguments");
			qhandle option_types[256];
			for (size_t index = 0; index < optionClasses.size(); index++)  option_types[index] = optionClasses.begin()[index].value.private_obj;
			return from_handle(
				IQuasarHost::GetInstance()->CreateTypeClass(option_types, (int)optionClasses.size()));
		}

		// Constructs a type parameter
		static Type Parameter(LPCTSTR param_name)
		{
			return from_handle(IQuasarHost::GetInstance()->CreateTypeParameter(param_name, nullptr));
		}
		static Type Parameter(LPCTSTR param_name, Type param_type)
		{
			return from_handle(IQuasarHost::GetInstance()->CreateTypeParameter(param_name, param_type.value.private_obj));
		}

		// Constructs a pointer type
		Type PointerType()
		{
			return from_handle(IQuasarHost::GetInstance()->CreatePointerType(*this));
		}

		Type GetElementType()
		{
			return Type(IQuasarHost::GetInstance()->TypeGetElementType(value));
		}

		// Gets the type parameter with the specified index. For example, for 
		// the type specialization list[T] -> list[vec], the type parameter is 
		// vec (the corresponding generic type parameter is T)
		Type GetTypeParameter(int index)
		{
			if (!genericTypeParametersInitialized)
				ObtainTypeParameters();

			return instanceTypeParameters[index];
		}

		int GetNumTypeParameters()
		{
			if (!genericTypeParametersInitialized)
				ObtainTypeParameters();

			return (int) instanceTypeParameters.size();
		}

		Type GetGenericTypeParameter(int index)
		{
			if (!genericTypeParametersInitialized)
				ObtainTypeParameters();

			return genericTypeParameters[index];
		}

		int GetNumGenericTypeParameters()
		{
			if (!genericTypeParametersInitialized)
				ObtainTypeParameters();

			return (int)genericTypeParameters.size();
		}

		Type GetBaseClass()
		{
			if (!genericTypeParametersInitialized)
				ObtainTypeParameters();

			return baseClass.front();
		}

	private: 
		QUASARHOSTAPI void ObtainTypeParameters();
	};


	// Function wrapper - allows you to call functions defined in quasar
	class Function : public QValue
	{
		friend class Method;

		string_t functionSignature;
		int bindingToken;

	public:
		Function() : QValue(), bindingToken(0)
		{
		}

		Function(const qvalue_t &value) : QValue(value), bindingToken(0)
		{
			_ASSERT_EXPR(value.type == TYPE_LAMBDAEXPR, "Function: lambda expression expected!");
		}

        // Construct a function based on the given signature. For example:
        // Function imshow("imshow(cube,vec)");
        Function(LPCTSTR functionSignature) : bindingToken(0)
        {
            this->functionSignature = string_t(functionSignature);
        }

		Function(const QValue &other) : QValue(other), bindingToken(0) {
			_ASSERT_EXPR(value.type == TYPE_LAMBDAEXPR, "Function: lambda expression expected!");
		}

		operator const qvalue_t &() const
		{
			return this->value;
		}

		bool BindFunction()
		{
			IQuasarHost *host = IQuasarHost::GetInstance();

			if (value.private_obj != NULL &&
				(!functionSignature.get_length() || !host->IsRebindRequired(value, bindingToken)))
				return true;

			const int N = 256;
			TCHAR func[N];
            tcsncpy(func, functionSignature, N);
            func[N - 1] = '\0';
            TCHAR *t = tcschr(func, '(');
            if (t)
                *t = '\0';

			// Release the old function (if any)
			if (value.private_obj != nullptr)
				host->ReleaseRef(value);

			// First, check if there is a variable that defines this function
			if (!host->ReadVariable(func, value) ||
				value.type != TYPE_LAMBDAEXPR)
			{
				// Otherwise, check the reductions
				value = host->LookupFunction(functionSignature, bindingToken);

				_ASSERT(value.private_obj != NULL);
				// throw ex
			}
			_ASSERT(value.type == TYPE_LAMBDAEXPR);
			return value.private_obj != NULL;
		}

	private:
		void InternalCall(const qvalue_t *inArgs, int nInArgs, qvalue_t *outArgs, int nOutArgs) const
		{
			const_cast<Function *>(this)->BindFunction();
			auto host = IQuasarHost::GetInstance();
			host->FunctionCall(value, inArgs, nInArgs, outArgs, nOutArgs);
		}

	public:
		QValue operator()(void) const
		{
			qvalue_t retval = qvalue_t::empty;
			InternalCall(NULL, 0, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a };
			InternalCall(args, 1, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a, QValue b) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b };
			InternalCall(args, 2, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c };
			InternalCall(args, 3, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c, QValue d) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c, d };
			InternalCall(args, 4, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c, QValue d, QValue e) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c, d, e };
			InternalCall(args, 5, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c, QValue d, QValue e,
			QValue f) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c, d, e, f };
			InternalCall(args, 6, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c, QValue d, QValue e,
			QValue f, QValue g) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c, d, e, f, g };
			InternalCall(args, 7, &retval, 1);
			return retval;
		}
		QValue operator()(QValue a, QValue b, QValue c, QValue d, QValue e,
			QValue f, QValue g, QValue h) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = { a, b, c, d, e, f, g, h };
			InternalCall(args, 8, &retval, 1);
			return retval;
		}

		template <int N>
		QValue operator()(QValue(&args)[N]) const
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t t_args[N];
			const_cast<Function*>(this)->BindFunction();
			for (int index = 0; index < N; index++) t_args[index] = (qvalue_t)args[index];
			IQuasarHost *host = IQuasarHost::GetInstance();
			InternalCall(t_args, N, &retval, 1);
			return retval;
		}

#ifdef CPLUSPLUS11
		// Call a function with variable arguments and one return value
		template <typename ...T>
		QValue operator()(T... packedArgs) const
		{
			QValue args[] = { (QValue)packedArgs... };
			return operator()(args);
		}
#endif

#ifdef CPLUSPLUS14
		// Call a function with multiple input and output values
		template <int N, typename ...T>
		decltype(utility::tuple_of_length<QValue, N>()) operator()(T... packedArgs) const
		{
			const int M = sizeof...(T);
			typedef decltype(utility::tuple_of_length<QValue, N>()) R;
			QValue args[M] = { (QValue)packedArgs... };
			qvalue_t t_args[M];
			for (int index = 0; index < M; index++) t_args[index] = (qvalue_t)args[index];
			qvalue_t r_args[N];
			IQuasarHost* host = IQuasarHost::GetInstance();
			InternalCall(t_args, M, r_args, N);
			return CreateTuple<N, R>(r_args, utility::gen_seq<N>());
		}

		// Call a function with zero input values and multiple output values
		template <int N>
		decltype(utility::tuple_of_length<QValue, N>()) operator()() const
		{
			typedef decltype(utility::tuple_of_length<QValue, N>()) R;
			qvalue_t r_args[N];
			InternalCall(NULL, 0, r_args, N);
			return CreateTuple<N, R>(r_args, utility::gen_seq<N>());
		}
	#endif

		template <int N>
		static QValue DirectCall(IQuasarHost *host, LPCTSTR functionName, QValue (& args)[N])
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t t_args[N];
			for (int index = 0; index < N; index++) t_args[index] = (qvalue_t)args[index];
			host->FunctionCall(functionName, t_args, N, &retval, 1);
			for (int index = 0; index < N; index++)
				host->ReleaseRef(t_args[index]);
			return retval;
		}

		template <int N>
		static QValue DirectCall(LPCTSTR functionName, QValue(&args)[N])
		{
			IQuasarHost *host = IQuasarHost::GetInstance();
			return DirectCall(host, functionName, args);
		}
	#ifdef CPLUSPLUS11
		static QValue DirectCall(IQuasarHost *host, LPCTSTR functionName)
		{
			qvalue_t retval = qvalue_t::empty;
			host->FunctionCall(functionName, NULL, 0, &retval, 1);
			return retval;
		}
		static QValue DirectCall(LPCTSTR functionName)
		{
			return DirectCall(IQuasarHost::GetInstance(), functionName);
		}
		// One return value, N input values
		template <typename ...T>
		static QValue DirectCall(LPCTSTR functionName, T... packedArgs)
		{
			QValue args[] = { (QValue)packedArgs... };
			IQuasarHost *host = IQuasarHost::GetInstance();
			for (int i = 0; i < sizeof...(packedArgs); i++)
				host->AddRef((qvalue_t &)args[i]);
			return DirectCall(host, functionName, args);
		}
	#endif

	#ifdef CPLUSPLUS14
		// Multiple return values, zero input values
		template <int N>
		static decltype(utility::tuple_of_length<QValue, N>()) DirectCall(LPCTSTR functionName)
		{
			qvalue_t r_args[N];
			typedef decltype(utility::tuple_of_length<QValue, N>()) R;
			IQuasarHost::GetInstance()->FunctionCall(functionName, nullptr, 0, r_args, N);
			return CreateTuple<N, R>(r_args, utility::gen_seq<N>());
		}

		// Multiple return values, N input values
		template <int N, typename ...T>
		static decltype(utility::tuple_of_length<QValue, N>()) DirectCall(LPCTSTR functionName, T... packedArgs)
		{
			const int M = sizeof...(T);
			typedef decltype(utility::tuple_of_length<QValue, N>()) R;
			QValue args[M] = { (QValue)packedArgs... };
			qvalue_t t_args[M], r_args[N];
			for (int index = 0; index < M; index++)
				t_args[index] = (qvalue_t)args[index];

			IQuasarHost *host = IQuasarHost::GetInstance();
			host->FunctionCall(functionName, t_args, M, r_args, N);
			return CreateTuple<N, R>(r_args, utility::gen_seq<N>());
		}
	#endif

		static Function CreateLambda(Type functionType, void* closurePtr, LambdaDelegate* lambda)
		{
			return IQuasarHost::GetInstance()->CreateLambda((qhandle)functionType, closurePtr, lambda);
		}

	#ifdef CPLUSPLUS11
		static Function CreateVariadicLambda(Type type, QValue(*func)(const QValue* args, int N))
		{
			return CreateLambda(type, (void *)func,
				[](void* closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto func = (QValue(*)(const QValue* args, int N))closurePtr;
				QValue *args = new QValue[nArgsIn];
				for (int i = 0; i < nArgsIn; i++)
					args[i] = argsIn[i];
				argsOut[0] = (qvalue_t)func(args, nArgsIn);
				delete[] args;
			});
		}
	#endif

private:
	static QUASARHOSTAPI void copy_args(QValue *args, const qvalue_t *argsIn, int numArgs);

public:

#ifdef CPLUSPLUS14

	private:
		template<typename R, typename F, int... S>
		static R LambdaCallHelper(F&& func, QValue *args, utility::seq<S...>)
		{
			return (R)(std::forward<F>(func))(args[S]...);
		}

		// Helper function that expands an array into a variadic parameter pack
		template<typename R, typename F, int N>
		static R LambdaCallHelper(F&& func, QValue(&args)[N])
		{
			return LambdaCallHelper<R>(std::forward<F>(func), args, utility::gen_seq<N>());
		}

		template <typename T, int N, int ...S>
		static void ExpandTuple(qvalue_t(*args)[N], T tuple, utility::seq<S...>)
		{
			QValue t_args[] = { std::get<S>(tuple)... };
			for (int i = 0; i < sizeof...(S); i++)
			{
				t_args[i].Retain();
				(*args)[i] = (qvalue_t)t_args[i];
			}
		}

		template <typename T>
		static void ExpandTuple(qvalue_t*args, int N, T tuple)
		{
			// Tuple passed to a single output argument - convert automatically to
			// a cell matrix. This provides a mechanism for calling functions where
			// the caller does not know the number of return values.
			const int M = std::tuple_size<T>::value;
			if (M > 1 && N == 1)
			{
				QValue value = QValue::CreateCellMatrix(tuple);
				args[0] = (qvalue_t)value;
				value.Retain();
			}
			else
			{
				ExpandTuple((qvalue_t(*)[M]) args, tuple, utility::gen_seq<M>());
			}
		}

		template <int N, typename R, int ...S>
		static R CreateTuple(qvalue_t(&args)[N], utility::seq<S...>)
		{
			return std::make_tuple((QValue)args[S]...);
		}
	public:
		// Converter for lambda functions with variable number of arguments
		template <typename R, typename ...T>
		static Function CreateLambda(Type type, std::function<R(T...)> func)
		{
			auto closure = new decltype(func)(func);
			IQuasarHost* host = IQuasarHost::GetInstance();
			qvalue_t lambda = host->CreateLambda((qhandle)type, closure,
				[](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto func = *(std::function<R(T...)>*)closurePtr;
				_ASSERT_EXPR(nArgsIn == sizeof...(T) && nArgsOut == 1,
					"Invalid number of input/output arguments for function call!");
				QValue args[sizeof...(T)];
				copy_args(args, argsIn, sizeof...(T));
				QValue retVal = LambdaCallHelper<QValue>(func, args);
				retVal.Retain();
				argsOut[0] = (qvalue_t)retVal;
			});
			host->AddLambdaDestroyCallback(lambda, [](void* closurePtr, LambdaDelegate* lambda)
			{
				delete (std::function<R(T...)>*)closurePtr;
			});
			return lambda;
		}

		template <typename R>
		static Function CreateLambda(Type type, std::function<R()> func)
		{
			auto closure = new decltype(func)(func);
			IQuasarHost* host = IQuasarHost::GetInstance();
			qvalue_t lambda = host->CreateLambda((qhandle)type, closure,
				[](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto func = *(std::function<R()>*)closurePtr;
				_ASSERT_EXPR(nArgsIn == 0 && nArgsOut == 1,
					"Invalid number of input/output arguments for function call!");
				QValue retVal = func();
				retVal.Retain();
				argsOut[0] = (qvalue_t)retVal;
			});
			host->AddLambdaDestroyCallback(lambda, [](void* closurePtr, LambdaDelegate* lambda)
			{
				delete (std::function<R()>*)closurePtr;
			});
			return lambda;
		}

		// Zero inputs, variable number of outputs
		template <typename ...R>
		static Function CreateLambda(Type type, std::function<std::tuple<R...>()> func)
		{
			auto closure = new decltype(func)(func);
			IQuasarHost* host = IQuasarHost::GetInstance();
			qvalue_t lambda = host->CreateLambda((qhandle)type, closure,
				[](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto func = *(std::function<std::tuple<R...>()>*)closurePtr;
				_ASSERT_EXPR(nArgsIn == 0 && (nArgsOut == sizeof...(R) || nArgsOut == 1),
					"Invalid number of input/output arguments for function call!");
				const int N = sizeof...(R);
				ExpandTuple(argsOut, N, func());
			});
			host->AddLambdaDestroyCallback(lambda, [](void* closurePtr, LambdaDelegate* lambda)
			{
				delete (std::function<std::tuple<R...>()>*)closurePtr;
			});
			return lambda;
		}

		// Variable number of inputs, variable number of outputs
		template <typename ...T, typename ...R>
		static Function CreateLambda(Type type, std::function<std::tuple<R...>(T...)> func)
		{
			auto closure = new decltype(func)(func);
			IQuasarHost* host = IQuasarHost::GetInstance();
			qvalue_t lambda = host->CreateLambda((qhandle)type, closure,
				[](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto func = *(std::function<std::tuple<R...>(T...)>*)closurePtr;
				_ASSERT_EXPR(nArgsIn == sizeof...(T) && (nArgsOut == sizeof...(R) || nArgsOut == 1),
					"Invalid number of input/output arguments for function call!");
				QValue args[sizeof...(T)];
				copy_args(args, argsIn, sizeof...(T));
				const int N = sizeof...(R);
				ExpandTuple(argsOut, nArgsOut,
					LambdaCallHelper<std::tuple<R...> >(func, args));
			});
			host->AddLambdaDestroyCallback(lambda, [](void* closurePtr, LambdaDelegate* lambda)
			{
				delete (std::function<std::tuple<R...>(T...)>*)closurePtr;
			});
			return lambda;
		}

		template<typename T>
		struct memfun_type
		{
			using type = void;
		};

		template<typename Ret, typename Class, typename... Args>
		struct memfun_type<Ret(Class::*)(Args...) const>
		{
			using type = std::function<Ret(Args...)>;
		};

		template<typename F>
		static QValue CreateLambda(LPCTSTR typeName, F func)
		{
			typedef typename memfun_type<decltype(&F::operator())>::type FunctionType;
			return CreateLambda(Type(typeName), FunctionType(func));
		}
#endif

	};

	class Method : public QValue
	{
	private:
		Type type;
		int bindingToken;
		string_t functionSignature;

      public:
        Method(const Type &type, LPCTSTR functionSignature) : type(type), functionSignature(functionSignature)
        {
            
        }
		void BindMethod()
		{
			IQuasarHost *host = IQuasarHost::GetInstance();

			if (value.private_obj != NULL &&
				!host->IsRebindRequired(value, bindingToken))
				return;

			// Release the old function (if any)
			if (value.private_obj != nullptr)
				host->ReleaseRef(value);

			value = IQuasarHost::GetInstance()->LookupMethod(
				type, functionSignature, bindingToken);
			_ASSERT_EXPR((value.type == TYPE_METHOD || value.type == TYPE_LAMBDAEXPR) && value.private_obj != NULL,
				"Method constructor: function does not specify a value!");
		}
        QValue operator()(QValue target)
        {
            qvalue_t retval = qvalue_t::empty;
			BindMethod();
            IQuasarHost::GetInstance()->MethodCall(value, target, NULL, 0,
                                                   &retval, 1);
            return retval;
        }
        QValue operator()(QValue target, QValue a)
        {
            qvalue_t retval = qvalue_t::empty;
			qvalue_t args[] = {a};
			BindMethod();
			IQuasarHost::GetInstance()->MethodCall(
                value, target, args, 1, &retval, 1);
            return retval;
        }
        QValue operator()(QValue target, QValue a, QValue b)
        {
            qvalue_t retval = qvalue_t::empty;
            qvalue_t args[] = {a, b};
			BindMethod();
			IQuasarHost::GetInstance()->MethodCall(value, target, args, 2,
                                                   &retval, 1);
            return retval;
        }
        QValue operator()(QValue target, QValue a, QValue b, QValue c)
        {
            qvalue_t retval = qvalue_t::empty;
            qvalue_t args[] = {a, b, c};
			BindMethod();
			IQuasarHost::GetInstance()->MethodCall(value, target, args, 3,
                                                   &retval, 1);
            return retval;
        }
		template <int N>
		QValue operator()(QValue target, QValue(&args)[N])
		{
			qvalue_t retval = qvalue_t::empty;
			qvalue_t t_args[N];
			for (int index = 0; index < N; index++) t_args[index] = (qvalue_t)args[index];
			IQuasarHost *host = IQuasarHost::GetInstance();
			BindMethod();
			host->MethodCall(value, target, t_args, N, &retval, 1);
			return retval;
		}
#ifdef CPLUSPLUS11
		template <typename...  Args>
		QValue operator()(QValue target, Args... packedArgs)
		{
			QValue args[] = { (QValue)packedArgs... };
			return operator() (target, args);
		}
#endif

#ifdef CPLUSPLUS14
		// Call a function with multiple input and output values
		template <int N, typename ...T>
		decltype(utility::tuple_of_length<QValue, N>()) operator()(QValue target, T... packedArgs) const
		{
			const int M = sizeof...(T);
			typedef decltype(utility::tuple_of_length<QValue, N>()) R;
			QValue args[M] = { (QValue)packedArgs... };
			qvalue_t t_args[M];
			for (int index = 0; index < M; index++) t_args[index] = (qvalue_t)args[index];
			qvalue_t r_args[N];
			IQuasarHost* host = IQuasarHost::GetInstance();
			const_cast<Method*>(this)->BindMethod();
			host->MethodCall(value, target, t_args, M, r_args, N);
			return Function::CreateTuple<N, R>(r_args, utility::gen_seq<N>());
		}

		// Call a function with zero input values and multiple output values
		template <int N>
		decltype(utility::tuple_of_length<QValue, N>()) operator()(QValue target) const
		{
			typedef decltype(utility::tuple_of_length<QValue, N>()) R;
			qvalue_t r_args[N];
			IQuasarHost* host = IQuasarHost::GetInstance();
			const_cast<Method*>(this)->BindMethod();
			host->MethodCall(value, target, NULL, 0, r_args, N);
			return Function::CreateTuple<N, R>(r_args, utility::gen_seq<N>());
		}
#endif
	};


    class TypeBuilder : public QValue
    {
      public:
		TypeBuilder()
		{
			this->value = qvalue_t::empty;
		}
        TypeBuilder(LPCTSTR moduleName, LPCTSTR typeName, CreateTypeFlags flags = TYPEFLAGS_NONE)
        {
            this->value =
                IQuasarHost::GetInstance()->CreateType(moduleName, typeName, flags);
        }
        void AddField(LPCTSTR fieldName, Type fieldType)
        {
			_ASSERT_EXPR(fieldType->type == TYPE_TYPEINFO, "AddField: fieldType should be a type");
            IQuasarHost::GetInstance()->AddField(
                this->value.private_obj, fieldName, fieldType->private_obj);
        }
		void AddFieldWithOffset(LPCTSTR fieldName, Type fieldType, size_t offset)
		{
			_ASSERT_EXPR(fieldType->type == TYPE_TYPEINFO, "AddField: fieldType should be a type");
			IQuasarHost::GetInstance()->AddFieldWithOffset(
				this->value.private_obj, fieldName, fieldType->private_obj, offset);
		}
		void AddProperty(LPCTSTR propertyName, Type fieldType, QValue getAction, QValue setAction)
		{
			_ASSERT_EXPR(getAction->type == TYPE_METHOD || getAction->type == TYPE_LAMBDAEXPR,
				"getAction should be a value or lambda expression");
			_ASSERT_EXPR(setAction->type == TYPE_METHOD || setAction->type == TYPE_LAMBDAEXPR,
				"setAction should be a value or lambda expression");
			_ASSERT_EXPR(fieldType->type == TYPE_TYPEINFO, "AddField: fieldType should be a type");
			IQuasarHost::GetInstance()->AddProperty(
				this->value.private_obj, propertyName, fieldType->private_obj,
				getAction->private_obj, setAction->private_obj);
		}
		Type AddParameter(Type typeParam)
		{
			return Type(IQuasarHost::GetInstance()->AddParameter(this->value.private_obj,
				typeParam.value.private_obj));
		}
		Type SelfReference()
		{
			Retain();
			return Type(value);
		}
		void InheritType(const Type& baseType)
		{
			IQuasarHost::GetInstance()->InheritType(this->value.private_obj, baseType->private_obj);
		}
		string_t GetTypeName()
		{
			return IQuasarHost::GetInstance()->TypeGetString(this->value);
		}
		void SetConstructor(ConstructorDelegate *ctor)
		{
			IQuasarHost::GetInstance()->SetConstructor(this->value.private_obj, ctor);
		}
		void SetDynamicDispatcher(DynamicDispatchers disp, DynamicCallDelegate *call)
		{
			IQuasarHost::GetInstance()->SetDynamicDispatcher(this->value.private_obj, disp, call);
		}
		Type CreateType()
		{
			IQuasarHost::GetInstance()->FinalizeType(value.private_obj);
			IQuasarHost::GetInstance()->AddRef(value);
			return Type(value);
		}
	};

	// Slice selector
	inline QValue Range()
	{
		return QValue::Default;
	}

	// Construct a sequence
	QUASARHOSTAPI QValue Range(scalar minVal, scalar maxVal, scalar step);
	QUASARHOSTAPI QValue Range(int minVal, int maxVal, int step);
	QUASARHOSTAPI QValue Range(QValue minVal, QValue maxVal);
	QUASARHOSTAPI QValue Range(QValue minVal, QValue maxVal, QValue step);

	class NamedParam : public QValue
	{
	public:
		string_t Key;
		QValue Value;

		NamedParam(LPCTSTR key, const QValue value):
			Key(key, (int) tcslen(key)),
			Value(value)
		{
			this->value = IQuasarHost::GetInstance()->CreateNamedParam(this->Key, this->Value);
		}
	};

	// Some important builtin functions
	QUASARHOSTAPI int size(QValue val, int dim);
	QUASARHOSTAPI QValue size(QValue val);
	QUASARHOSTAPI QValue size(QValue val, QValue dim);
	QUASARHOSTAPI int numel(QValue val);
	QUASARHOSTAPI Type type(const QValue &val);

#ifdef CPLUSPLUS14
	// Call a function with multiple input and output values
	template <int N, typename ...T>
	decltype(utility::tuple_of_length<QValue, N>()) QValue::operator()(T... packedArgs)
	{
		_ASSERT(value.type == TYPE_LAMBDAEXPR);
		return ((Function)*this).operator()<N>(packedArgs...);
	}

	// Call a function with zero input values and multiple output values
	template <int N>
	decltype(utility::tuple_of_length<QValue, N>()) QValue::operator()()
	{
		_ASSERT(value.type == TYPE_LAMBDAEXPR);
		return ((Function)*this).operator()<N>();
	}
#endif

	template <int N> QValue::Elem<N>::operator QValue() const
	{
		_ASSERT(Type::IsNCubeType(mtx.type) || mtx.type == TYPE_LAMBDAEXPR || mtx.type == TYPE_TYPEINFO || mtx.type == TYPE_UNTYPEDOBJECT);
		switch (mtx.type)
		{
		case TYPE_LAMBDAEXPR:		// Call a function
		{
			QValue args[N];
			for (int i = 0; i < N; i++)
			{
				IQuasarHost::GetInstance()->AddRef(const_cast<qvalue_t &>(index[i]));
				args[i] = index[i];
			}
			IQuasarHost::GetInstance()->AddRef(const_cast<qvalue_t &>(mtx));
			return Function(mtx)(args);
		}
		case TYPE_TYPEINFO:
		{
			QValue args[N];
			for (int i = 0; i < N; i++)
			{
				IQuasarHost::GetInstance()->AddRef(const_cast<qvalue_t &>(index[i]));
				args[i] = index[i];
			}
			IQuasarHost::GetInstance()->AddRef(const_cast<qvalue_t &>(mtx));
			return Type(mtx)(args);
		}
		default:
		{
			IQuasarHost *host = IQuasarHost::GetInstance();
			IComputationEngine *engine = host->GetComputationEngine();
			IEvaluationStack *stack = engine->GetEvaluationStack();
			for (int i = 0; i < N; i++)
				stack->PushValue(index[i]);
			engine->ArrayGetAt(mtx, N, boundaryAccessMode);
			return stack->PopValue();
		}
		} // switch
	}

#if !defined(_MSC_VER)
	template <>
#endif
	INLINE QValue::Elem<0>::operator QValue() const
	{
		_ASSERT(mtx.type == TYPE_LAMBDAEXPR);
		return Function(mtx)();
	}

	template <int N>
	QValue::Elem<N>& QValue::Elem<N>::operator=(const qvalue_t &newValue)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		IComputationEngine *engine = host->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		_ASSERT(Type::IsNCubeType(mtx.type) || mtx.type == TYPE_UNTYPEDOBJECT);
		for (int i = 0; i < N; i++)
			stack->PushValue(index[i]);
		stack->PushValue(newValue);
		engine->ArraySetAt(mtx, N, boundaryAccessMode);
		host->ReleaseRef(stack->PopValue());
		return *this;
	}

	template <int N>
	QValue::Elem<N>& QValue::Elem<N>::InplaceOp(OperatorTypes opType, QValue b)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		IComputationEngine *engine = host->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		for (int i = 0; i < N; i++)
			stack->PushValue(index[i]);
		engine->ArrayGetAt(mtx, N, boundaryAccessMode);
		stack->PushValue(b.value);
		stack->Process(opType);
		QValue value = stack->PopValue();
		for (int i = 0; i < N; i++)
			stack->PushValue(index[i]);
		stack->PushValue(value);
		engine->ArraySetAt(mtx, N, boundaryAccessMode);
		host->ReleaseRef(stack->PopValue());
		return *this;
	}

	INLINE QValue::ElemWiseOp::operator QValue()
	{
		Retain();
		return value;
	}

	template <int N>
	QValue QValue::CreateObject(const Type& type, NamedParam(&params)[N])
	{
		QValue obj = CreateObject(type);
		for (int i = 0; i < N; i++)
			obj.SetField(params[i].Key, params[i].Value);
		return obj;
	}

	template <>
	INLINE QValue QValue::FromObj<Function>(const Function val)
	{
		return val;
	}

	template <>
	INLINE QValue QValue::FromObj<Type>(const Type val)
	{
		return val;
	}

	INLINE QValue::FieldRef::operator QValue() const
	{
		qvalue_t y = qvalue_t::empty;
		IQuasarHost::GetInstance()->GetField(this->obj, this->fieldName, y);
		return y;
	}
	INLINE QValue::FieldRef& QValue::FieldRef::operator=(QValue newValue)
	{
		IQuasarHost::GetInstance()->SetField(this->obj, this->fieldName, newValue.value);
		return *this;
	}

	INLINE QValue::FieldRef& QValue::FieldRef::InplaceOp(OperatorTypes opType, QValue b)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		IComputationEngine *engine = host->GetComputationEngine();
		IEvaluationStack *stack = engine->GetEvaluationStack();
		stack->PushValue(this->operator QValue());
		stack->PushValue(b.value);
		stack->Process(opType);
		*this = (QValue)stack->PopValue();
		return *this;
	}

	template <int N>
	INLINE QValue::FieldRef QValue::Elem<N>::Field(LPCTSTR fieldName)
	{
		return QValue::FieldRef(operator QValue(), fieldName);
	}

#define IMPL_QVALUE_OPERATORS(P, T,R)   \
	P QValue operator+(T a, R b) { return (QValue)a + (QValue)b; }	\
	P QValue operator-(T a, R b) { return (QValue)a - (QValue)b; }	\
	P QValue operator*(T a, R b) { return (QValue)a * (QValue)b; }	\
	P QValue operator/(T a, R b) { return (QValue)a / (QValue)b; }	\
	P QValue operator>(T a, R b) { return (QValue)a > (QValue)b; }	\
	P QValue operator<(T a, R b) { return (QValue)a < (QValue)b; }	\
	P QValue operator>=(T a, R b) { return (QValue)a >= (QValue)b; }	\
	P QValue operator<=(T a, R b) { return (QValue)a <= (QValue)b; }	\
	P QValue operator!=(T a, R b) { return (QValue)a != (QValue)b; }	\
	P QValue operator==(T a, R b) { return (QValue)a == (QValue)b; }

#define IMPL_QVALUE_ELEMWISE_OPERATORS(P, R)   \
	P QValue operator*(QValue::ElemWiseOp a, R b) { return a * (QValue)b; }	\
	P QValue operator/(QValue::ElemWiseOp a, R b) { return a / (QValue)b; }


	IMPL_QVALUE_OPERATORS(template <int N>, QValue::Elem<N>, QValue)
	IMPL_QVALUE_OPERATORS(template <int N>, QValue, QValue::Elem<N>)
	IMPL_QVALUE_OPERATORS(template <int N>, QValue::Elem<N>, scalar)
	IMPL_QVALUE_OPERATORS(template <int N>, scalar, QValue::Elem<N>)
	IMPL_QVALUE_OPERATORS(template <int N>, QValue::Elem<N>, cscalar)
	IMPL_QVALUE_OPERATORS(template <int N>, cscalar, QValue::Elem<N>)
	IMPL_QVALUE_OPERATORS(INLINE, QValue::FieldRef, QValue)
	IMPL_QVALUE_OPERATORS(INLINE, QValue, QValue::FieldRef)
	IMPL_QVALUE_OPERATORS(INLINE, QValue::FieldRef, scalar)
	IMPL_QVALUE_OPERATORS(INLINE, scalar, QValue::FieldRef)
	IMPL_QVALUE_OPERATORS(INLINE, QValue::FieldRef, cscalar)
	IMPL_QVALUE_OPERATORS(INLINE, cscalar, QValue::FieldRef)
	IMPL_QVALUE_ELEMWISE_OPERATORS(template <int N>, QValue::Elem<N>)
	IMPL_QVALUE_ELEMWISE_OPERATORS(INLINE, QValue::FieldRef)

#undef IMPL_QVALUE_OPERATORS
#undef IMPL_QVALUE_ELEMWISE_OPERATORS

	// Conversion function for creating uniform vecX values
	template <typename R, typename T>
	INLINE R make_uniform(T a) { return R(); }

	template <> INLINE scalar1 make_uniform(scalar a) { return make_scalar1(a); }
	template <> INLINE scalar2 make_uniform(scalar a) { return make_scalar2(a, a); }
	template <> INLINE scalar3 make_uniform(scalar a) { return make_scalar3(a, a, a); }
	template <> INLINE scalar4 make_uniform(scalar a) { return make_scalar4(a, a, a, a); }
	template <int N> INLINE scalarN<N> make_uniform(scalar a) {
		scalarN<N> ret;
		for (int i = 0; i < N; i++) ret.el[i] = a;
		return ret;
	}

	template <> INLINE int1 make_uniform(int a) { return make_int1(a); }
	template <> INLINE int2 make_uniform(int a) { return make_int2(a, a); }
	template <> INLINE int3 make_uniform(int a) { return make_int3(a, a, a); }
	template <> INLINE int4 make_uniform(int a) { return make_int4(a, a, a, a); }
	template <int N> INLINE intN<N> make_uniform(int a) {
		intN<N> ret;
		for (int i = 0; i < N; i++) ret.el[i] = a;
		return ret;
	}

	template <> INLINE cscalar1 make_uniform(cscalar a) { return make_cscalar1(a); }
	template <> INLINE cscalar2 make_uniform(cscalar a) { return make_cscalar2(a, a); }
	template <> INLINE cscalar3 make_uniform(cscalar a) { return make_cscalar3(a, a, a); }
	template <> INLINE cscalar4 make_uniform(cscalar a) { return make_cscalar4(a, a, a, a); }
	template <int N> INLINE cscalarN<N> make_uniform(cscalar a) {
		cscalarN<N> ret;
		for (int i = 0; i < N; i++) ret.el[i] = a;
		return ret;
	}

	// Provides cooperative memory resources (allocated by the Quasar host).
	// This class is used for implementing the cooperative memory model.
	template <typename T>
	class cooperative : public T
	{
	private:
		qvalue_t obj;
		LockingModes lockingMode;
		MemResources memResource;

	public:
		cooperative()
		{
			lockingMode = LOCK_READWRITE;
			memResource = MEMRESOURCE_CPU;
		}
		cooperative(T t, qvalue_t obj, LockingModes lockingMode = LOCK_READWRITE,
			MemResources memResource = MEMRESOURCE_CPU)
		{
			(T &)*this = t;
			this->lockingMode = lockingMode;
			this->memResource = memResource;
			this->obj = obj;
		}
		cooperative(const cooperative & b)
		{
			InitFrom(b);
		}
		~cooperative()
		{
			Dispose();
		}
		cooperative & operator= (const cooperative & b)
		{
			Dispose();
			InitFrom(b);
			return *this;
		}
		operator qvalue_t() const { return obj; }
		operator QValue() { IQuasarHost::GetInstance()->AddRef(obj); return obj; }
	private:
		void InitFrom(const cooperative & b)
		{
			(T &)*this = (T)b;
			this->lockingMode = b.lockingMode;
			this->memResource = b.memResource;
			this->obj = b.obj;
			void *pData;

			IQuasarHost *host = IQuasarHost::GetInstance();
			host->AddRef(obj);
			host->Lock(obj, TYPE_VOID, lockingMode, memResource, &pData);
			// Note - we can ignore pData here, since the pointer is already stored in T
		}
		void Dispose()
		{
			IQuasarHost *host = IQuasarHost::GetInstance();
			if (obj.private_obj)
			{
				host->Unlock(obj, lockingMode, memResource);
				host->ReleaseRef(obj);
				obj.private_obj = NULL;
			}
		}
	};

	// Cooperative memory allocation functions
	// 1. uninit_coop function
	template <typename R>
	static cooperative<VectorBase<R> > uninit_coop(int M, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		IQuasarHost *host = IQuasarHost::GetInstance();
		qvalue_t obj = host->CreateVector<R>(M);
		return cooperative<VectorBase<R> >(host->LockVector<R>(obj, lockingMode, memResource), obj, lockingMode, memResource);
	}
	template <typename R>
	static cooperative<MatrixBase<R> > uninit_coop(int M, int N, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		IQuasarHost *host = IQuasarHost::GetInstance();
		qvalue_t obj = host->CreateMatrix<R>(M, N);
		return cooperative<MatrixBase<R> >(host->LockMatrix<R>(obj, lockingMode, memResource), obj, lockingMode, memResource);
	}
	template <typename R>
	static cooperative<CubeBase<R> > uninit_coop(int M, int N, int P, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		IQuasarHost *host = IQuasarHost::GetInstance();
		qvalue_t obj = host->CreateCube<R>(M, N, P);
		return cooperative<CubeBase<R> >(host->LockCube<R>(obj, lockingMode, memResource), obj, lockingMode, memResource);
	}
	template <typename R, int N>
	static cooperative<NCubeBase<R, N> > uninit_coop(intN<N> dims, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		IQuasarHost *host = IQuasarHost::GetInstance();
		qvalue_t obj = host->CreateNCube<R, N>(dims);
		return cooperative<NCubeBase<R, N> >(host->LockNCube<R, N>(obj, lockingMode, memResource), obj, lockingMode, memResource);
	}
	// 2. zeros_coop function
	template <typename R>
	static cooperative<VectorBase<R> > zeros_coop(int M, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<VectorBase<R> > y = uninit_coop<R>(M, lockingMode, memResource);
		mem_fill<R>((R *)y.data, M, 0);
		return y;
	}
	template <typename R>
	static cooperative<MatrixBase<R> > zeros_coop(int M, int N, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<MatrixBase<R> > y = uninit_coop<R>(M, N, lockingMode, memResource);
		mem_fill<R>((R *)y.data, M*N, 0);
		return y;
	}
	template <typename R>
	static cooperative<CubeBase<R> > zeros_coop(int M, int N, int P, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<CubeBase<R> > y = uninit_coop<R>(M, N, P, lockingMode, memResource);
		mem_fill<R>((R *)y.data, M*N*P, 0);
		return y;
	}
	template <typename R, int N>
	static cooperative<NCubeBase<R, N> > zeros_coop(intN<N> dims, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<NCubeBase<R, N> > y = uninit_coop<R, N>(dims, lockingMode, memResource);
		mem_fill<R>((R *)y.data, prod(dims), 0);
		return y;
	}
	// 3. ones_coop function
	template <typename R>
	static cooperative<VectorBase<R> > ones_coop(int M, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<VectorBase<R> > y = uninit_coop<R>(M, lockingMode, memResource);
		mem_fill<R>((R *)y.data, M, 1);
		return y;
	}
	template <typename R>
	static cooperative<MatrixBase<R> > ones_coop(int M, int N, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<MatrixBase<R> > y = uninit_coop<R>(M, N, lockingMode, memResource);
		mem_fill<R>((R *)y.data, M*N, 1);
		return y;
	}
	template <typename R>
	static cooperative<CubeBase<R> > ones_coop(int M, int N, int P, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<CubeBase<R> > y = uninit_coop<R>(M, N, P, lockingMode, memResource);
		mem_fill<R>((R *)y.data, M*N*P, 1);
		return y;
	}
	template <typename R, int N>
	static cooperative<NCubeBase<R, N> > ones_coop(intN<N> dims, LockingModes lockingMode = LOCK_READWRITE, MemResources memResource = MEMRESOURCE_CPU) {
		cooperative<NCubeBase<R, N> > y = uninit_coop<R, N>(dims, lockingMode, memResource);
		mem_fill<R>((R *)y.data, prod(dims), 1);
		return y;
	}

	//
	// Symbol types (used by Quasar's postfix expression representation)
	//
	enum SymbolTypes
	{
		SYM_STRING,         // String symbol
		SYM_SCALAR,         // Scalar number
		SYM_COMPLEXSCALAR,  // Complex scalar number
		SYM_VARIABLE,       // Variable name
		SYM_TYPENAME,       // A type name
		SYM_OPERATOR,       // Application of an operator
		SYM_FUNCTION,       // Function call
		SYM_ARRAYINDEXER,   // Array indexing or construction
		SYM_ARRAY2NDLEVELINDEXER, // Second level array indexer
		SYM_CELLARRAYCONSTRUCTOR, // A cell array constructor
		SYM_REFERENCER,     // A referencer (for assignments)
		SYM_MEMBERREFERENCE,// A class member references
		SYM_MEMBERFUNCTION, // A class member function call
		SYM_FUNCTION2NDLEVELEVAL, // Second level function evaluation
		SYM_VARIABLEDECL,   // A variable declaration (var : Type)
		SYM_TYPECHECK,      // A type checking expression (A : Type)
		SYM_VARIABLEASSIGNMENT, // A variable assignment (CIL compiler only)
	};

	//
	// Variable declaration information
	//
	class VariableDecl
	{
	public:
		LPCTSTR typeName;
		LPCTSTR varName;

		VariableDecl(LPCTSTR varName, LPCTSTR typeName) :
			typeName(typeName),
			varName(varName)
		{}
	};

	//
	// Symbol (part of a postfix expression)
	//
	class Symbol
	{
	public:
		SymbolTypes symbolType;
		int numArgs;
		union
		{
			LPCTSTR stringVal;
			int intVal;
			scalar scalarVal;
			cscalar cscalarVal;
			OperatorTypes opVal;
			char reserved[16];
		};

		Symbol() {}
		Symbol(SymbolTypes symbolType, int numArgs, LPCTSTR stringVal) :
			symbolType(symbolType),
			numArgs(numArgs),
			stringVal(stringVal)
		{
		}
		Symbol(SymbolTypes symbolType, int numArgs, int intVal) :
			symbolType(symbolType),
			numArgs(numArgs),
			intVal(intVal)
		{
		}
		Symbol(scalar scalarVal) :
			symbolType(SYM_SCALAR),
			scalarVal(scalarVal)
		{
		}
		Symbol(cscalar cscalarVal) :
			symbolType(SYM_COMPLEXSCALAR),
			cscalarVal(cscalarVal)
		{
		}
		Symbol(OperatorTypes opType, int numArgs) :
			symbolType(SYM_OPERATOR),
			numArgs(numArgs),
			opVal(opType)
		{
		}
	};

	//
	// An expression, in postfix representation
	//
	class Expression
	{
	public:
		static const int MAX_SYMBOLS = 256;
	protected:
		Symbol symbols[MAX_SYMBOLS];
		int symbolCount;
		bool isVariadic;

	public:
		Expression(const Symbol* symbols, int symbolCount, bool isVariadic = false):isVariadic(isVariadic),symbolCount(symbolCount)
		{
			_ASSERT_EXPR(symbolCount <= MAX_SYMBOLS, "Expression: too many symbols");
			std::memcpy(this->symbols, symbols, symbolCount * sizeof(Symbol));
		}
		Expression(std::initializer_list<Symbol> list, bool isVariadic = false)
			:Expression(list.begin(), (int)list.size(), isVariadic)
		{
		}
		const Symbol* GetSymbols() const { return symbols; }
		const Symbol& GetSymbol(int index) const
		{
			_ASSERT_EXPR(index >= 0 && index < symbolCount, "Expression: index out of range!");
			return symbols[index];
		}
		int GetNumSymbols() const { return symbolCount; }
		bool IsVariadic() const { return isVariadic; }
	};

	class Reduction
	{
	public:
		string_t searchExpr;
		QValue action;
		QValue whereClause;

	public:
		Reduction(LPCTSTR searchExpr, QValue action)
			: searchExpr(searchExpr), action(action)
		{
			_ASSERT_EXPR(((qvalue_t&)action).type == TYPE_LAMBDAEXPR, "Action must be a lambda expression!");
			Register();
		}
		Reduction(LPCTSTR searchExpr, QValue action, QValue whereClause)
			: searchExpr(searchExpr), action(action), whereClause(whereClause)
		{
			_ASSERT_EXPR(((qvalue_t&)action).type == TYPE_LAMBDAEXPR, "Action must be a lambda expression!");
			_ASSERT_EXPR(((qvalue_t&)whereClause).type == TYPE_LAMBDAEXPR, "WhereClause must be a lambda expression!");
			Register();
		}

	private:
		void Register()
		{
			IRuntimeReductionEngine* rre =
				IQuasarHost::GetInstance()->GetComputationEngine()->GetRuntimeReductionEngine();
			bool retVal = rre->Add(searchExpr, action.value.private_obj, whereClause.value.private_obj);
			_ASSERT(retVal);
		}
		void Unregister()
		{
			IRuntimeReductionEngine* rre =
				IQuasarHost::GetInstance()->GetComputationEngine()->GetRuntimeReductionEngine();
			bool retVal = rre->Remove(searchExpr);
			_ASSERT(retVal);
		}
	};

	// RuntimeControl class: allows interacting directly with the runtime system,
	// allows controlling the memory manager / scheduler on a manual level.
	class RuntimeControl : public QValue
	{
		friend class QValue;

	public:
		QUASARHOSTAPI RuntimeControl();

		void Alloc(LPCTSTR mode)
		{
			QValue params[] = { NamedParam(_T("mode"), mode) };
			InvokeMethod(_T("alloc"), params);
		}
		void Transfer(QValue var, LPCTSTR target, int gpuIndex = 0)
		{
			QValue params[] = { NamedParam(_T("vars"), var), NamedParam(_T("target"), target), NamedParam(_T("gpu_index"), gpuIndex) };
			InvokeMethod(_T("transfer"), params);
		}
		void Schedule(LPCTSTR mode)
		{
			QValue params[] = { NamedParam(_T("mode"), mode) };
			InvokeMethod(_T("sched"), params);
		}
		void ScheduleGPU(int gpuIndex)
		{
			QValue params[] = { NamedParam(_T("gpu_index"), gpuIndex) };
			InvokeMethod(_T("sched"), params);
		}
		void ScheduleGPU()  { Schedule(_T("gpu")); }
		void ScheduleAuto() { Schedule(_T("auto")); }
		void ScheduleCPU()  { Schedule(_T("cpu")); }
	};

	// Access to the command line
	struct CommandLine
	{
		template <int N>
		static void GetArguments(QValue argsOut[N], int numMandatoryArgs, bool variadic)
		{
			qvalue_t args[N];
			IQuasarHost::GetInstance()->GetProgramCmdLineArguments(args, N, numMandatoryArgs, variadic);
			for (int i = 0; i < N; i++) argsOut[i] = args[i];
		}

#ifdef CPLUSPLUS14
	private:
		template <int N, typename R, int ...S>
		static R CreateTuple(qvalue_t(&args)[N], utility::seq<S...>)
		{
			return std::make_tuple((QValue)args[S]...);
		}
	public:
		template <typename ...T>
		static std::tuple<T...> GetArguments(int numMandatoryArgs, bool variadic)
		{
			const int N = sizeof...(T);
			qvalue_t args[N];
			IQuasarHost::GetInstance()->GetProgramCmdLineArguments(args, N, numMandatoryArgs, variadic);
			return CreateTuple<N, std::tuple<T...>>(args, utility::gen_seq<N>());

		}
#endif
	};

	// Quasar Core Library. Allows to easily call main Quasar functions by providing
	// static bindings to the library functions. Depending on the implementation, this
	// is basically a indirect function call.
	struct CoreLib
	{
		Function zeros, ones, uninit, uninit_T, czeros, cuninit, rand, randn, cell,
				 struct_, object, eye, shared, cshared, shared_T;
		Function size, numel, ndims;
		Function cast, complex, real, imag, transpose, copy, deepcopy, squeeze,
			repmat, reshape, type, maxvalue, minvalue, new_, tic, toc, pause,
			cell2mat, mat2cell, fromascii, fromunicode, toascii_,
			tounicode, shuffledims;
		Function sum, dotprod, cumsum, prod, cumprod, mean, linspace;
		Function parallel_do, serial_do, schedule;
		Function ind2pos, pos2ind, pos2indchk;
		Function imread, imwrite, fopen, fgetpos, fsetpos, fclose, dir;
		Function sprintf, sscanf, printf, fprintf, fprint, fread, fwrite, fscanf, fgets,
			feof, print_, error_, assert_, strcat, factorial, eval, save, load;
		Function fft1, fft2, fft3, ifft1, ifft2, ifft3, irealfft1, irealfft2, irealfft3;
		Function max_block_size, opt_block_size;
		Function abs, asin, acos, atan, atan2, ceil, round, cos, sin, exp, exp2,
			floor, mod, frac, log, log2, log10, lgamma, erf, erfc, max, min,
			pow, saturate, sign, sqrt, rsqrt, tan, angle, conj, float_, int_,
			isinf, isfinite, isnan, periodize, mirror_ext, clamp, sinh, cosh,
			tanh, asinh, aconsh, atanh, and_, or_, xor_, not_, shl, shr;
		Function imshow, plot, scatter, disp, sync_framerate, hold, title, xlabel,
			ylabel, xlim, ylim, legend;

		Type any_type, scalar, cscalar, iscalar, vec, mat, cube, cvec, cmat, ccube, string,
			function, device_function, kernel_function, lambda_expr, object_class, int8,
			int16, int32, uint8, uint16, uint32, int64, uint64,
			vec1, vec2, vec3, vec4, vec5, vec6, vec7, vec8, vec9, vec10, vec11, vec12, vec13, vec14, vec15, vec16,
			ivec1, ivec2, ivec3, ivec4, ivec5, ivec6, ivec7, ivec8, ivec9, ivec10, ivec11, ivec12, ivec13, ivec14, ivec15, ivec16,
			cvec1, cvec2, cvec3, cvec4, cvec5, cvec6, cvec7, cvec8, cvec9, cvec10, cvec11, cvec12, cvec13, cvec14, cvec15, cvec16;

		CoreLib() :
			zeros(_T("zeros(...)")), // function with variable arguments
			ones(_T("ones(...)")),
			eye(_T("eye(...)")),
			uninit(_T("uninit(...)")),
			uninit_T(_T("uninit_T(?\?,...)")),
			cuninit(_T("cuninit(...)")),
			czeros(_T("czeros(...)")),
			rand(_T("rand(...)")),
			randn(_T("randn(...)")),
			shared(_T("shared(...)")),
			cshared(_T("cshared(...)")),
			shared_T(_T("shared_T(...)")),
			cell(_T("cell(...)")),
			struct_(_T("struct()")),
			object(_T("object()")),
			size(_T("size(...)")),
			numel(_T("numel(?\?)")),
			ndims(_T("ndims(?\?)")),
			cast(_T("cast(?\?,?\?)")),
			complex(_T("complex(...)")),
			real(_T("real(?\?)")),
			imag(_T("imag(?\?)")),
			transpose(_T("transpose(?\?)")),
			copy(_T("copy(?\?)")),
			deepcopy(_T("deepcopy(?\?)")),
			squeeze(_T("squeeze(?\?)")),
			repmat(_T("repmat(?\?,?\?)")),
			reshape(_T("reshape(?\?,?\?)")),
			type(_T("type(...)")),
			maxvalue(_T("maxvalue(?\?)")),
			minvalue(_T("minvalue(?\?)")),
			new_(_T("new(?\?)")),
			tic(_T("tic()")),
			toc(_T("toc(...)")),
			pause(_T("pause(?\?)")),
			cell2mat(_T("cell2mat(?\?)")),
			mat2cell(_T("mat2cell(?\?)")),
			fromascii(_T("fromascii(?\?)")),
			fromunicode(_T("fromunicode(?\?)")),
			toascii_(_T("toascii(?\?)")), // toascii_ to avoid collision with toascii from <string.h>
			tounicode(_T("tounicode(?\?)")),
			shuffledims(_T("shuffledims(?\?,?\?)")),
			sum(_T("sum(...)")),
			dotprod(_T("dotprod(?\?)")),
			cumsum(_T("cumsum(...)")),
			prod(_T("prod(...)")),
			cumprod(_T("cumprod(...)")),
			mean(_T("mean(...)")),
			linspace(_T("linspace(?\?,?\?,?\?)")),
			parallel_do(_T("parallel_do(...)")),
			serial_do(_T("serial_do(...)")),
			schedule(_T("schedule(...)")),
			ind2pos(_T("ind2pos(?\?,?\?)")),
			pos2ind(_T("pos2ind(?\?,?\?)")),
			pos2indchk(_T("pos2indchk(?\?,?\?)")),
			imread(_T("imread(?\?)")),
			imwrite(_T("imwrite(?\?)")),
			fopen(_T("fopen(?\?,?\?)")),
			fgetpos(_T("fgetpos(?\?)")),
			fsetpos(_T("fsetpos(?\?,?\?)")),
			fclose(_T("fclose(?\?)")),
			dir(_T("dir(?\?)")),
			sprintf(_T("sprintf(...)")),
			sscanf(_T("sscanf(...)")),
			printf(_T("printf(...)")),
			fprintf(_T("fprintf(...)")),
			fprint(_T("fprint(...)")),
			fread(_T("fread(...)")),
			fwrite(_T("fwrite(...)")),
			fscanf(_T("fscanf(...)")),
			fgets(_T("fgets(?\?)")),
			feof(_T("feof(?\?)")),
			print_(_T("print(...)")),
			error_(_T("error(...)")),
			assert_(_T("assert(...)")),
			strcat(_T("strcat(...)")),
			factorial(_T("factorial(?\?)")),
			eval(_T("eval(?\?)")),
			save(_T("save(?\?)")),
			load(_T("load(?\?)")),
			fft1(_T("fft1(?\?)")),
			fft2(_T("fft2(?\?)")),
			fft3(_T("fft3(?\?)")),
			ifft1(_T("ifft1(?\?)")),
			ifft2(_T("ifft2(?\?)")),
			ifft3(_T("ifft3(?\?)")),
			irealfft1(_T("irealfft1(?\?)")),
			irealfft2(_T("irealfft2(?\?)")),
			irealfft3(_T("irealfft3(?\?)")),
			max_block_size(_T("max_block_size(...)")),
			opt_block_size(_T("opt_block_size(...)")),
			abs(_T("abs(?\?)")),
			asin(_T("asin(?\?)")),
			acos(_T("acos(?\?)")),
			atan(_T("atan(?\?)")),
			atan2(_T("atan2(?\?,?\?)")),
			ceil(_T("ceil(?\?)")),
			round(_T("round(?\?)")),
			cos(_T("cos(?\?)")),
			sin(_T("sin(?\?)")),
			exp(_T("exp(?\?)")),
			exp2(_T("exp2(?\?)")),
			floor(_T("floor(?\?)")),
			mod(_T("mod(?\?,?\?)")),
			frac(_T("frac(?\?)")),
			log(_T("log(?\?)")),
			log2(_T("log2(?\?)")),
			log10(_T("log10(?\?)")),
			lgamma(_T("lgamma(?\?)")),
			erf(_T("erf(?\?)")),
			erfc(_T("erfc(?\?)")),
			max(_T("max(...)")),
			min(_T("min(...)")),
			pow(_T("pow(?\?,?\?)")),
			saturate(_T("saturate(?\?)")),
			sign(_T("sign(?\?)")),
			sqrt(_T("sqrt(?\?)")),
			rsqrt(_T("rsqrt(?\?)")),
			tan(_T("tan(?\?)")),
			angle(_T("angle(?\?)")),
			conj(_T("conj(?\?)")),
			float_(_T("float(?\?)")),
			int_(_T("int(?\?)")),
			isinf(_T("isinf(?\?)")),
			isfinite(_T("isfinite(?\?)")),
			isnan(_T("isnan(?\?)")),
			periodize(_T("periodize(?\?,?\?)")),
			mirror_ext(_T("mirror_ext(?\?,?\?)")),
			clamp(_T("clamp(?\?,?\?)")),
			sinh(_T("sinh(?\?)")),
			cosh(_T("cosh(?\?)")),
			tanh(_T("tanh(?\?)")),
			asinh(_T("asinh(?\?)")),
			aconsh(_T("aconsh(?\?)")),
			atanh(_T("atanh(?\?)")),
			and_(_T("and(?\?,?\?)")),
			or_(_T("or(?\?,?\?)")),
			xor_(_T("xor(?\?,?\?)")),
			not_(_T("not(?\?)")),
			shl(_T("shl(?\?,?\?)")),
			shr(_T("shr(?\?,?\?)")),
			imshow(_T("imshow(...)")),
			plot(_T("plot(...)")),
			scatter(_T("scatter(...)")),
			disp(_T("disp(...)")),
			sync_framerate(_T("sync_framerate(...)")),
			hold(_T("hold(...)")),
			title(_T("title(...)")),
			xlabel(_T("xlabel(...)")),
			ylabel(_T("ylabel(...)")),
			xlim(_T("xlim(...)")),
			ylim(_T("ylim(...)")),
			legend(_T("legend(...)")),
			// Access to Quasar built-in types
			scalar(_T("scalar")),
			cscalar(_T("cscalar")),
			iscalar(_T("int")),
			vec(_T("vec")),
			mat(_T("mat")),
			cube(_T("cube")),
			cvec(_T("cvec")),
			cmat(_T("cmat")),
			ccube(_T("ccube")),
			string(_T("string")),
			function(_T("function")),
			device_function(_T("device_function")),
			kernel_function(_T("kernel_function")),
			lambda_expr(_T("lambda_expr")),
			object_class(_T("object")),
			any_type(_T("?\?")),
			int8(_T("int8")),
			int16(_T("int16")),
			int32(_T("int32")),
			uint8(_T("uint8")),
			uint16(_T("uint16")),
			uint32(_T("uint32")),
			int64(_T("int64")),
			uint64(_T("uint64")),
			 vec1(_T("vec1")),   vec2(_T("vec2")),     vec3(_T("vec3")),     vec4(_T("vec4")),     vec5(_T("vec5")),     vec6(_T("vec6")),     vec7(_T("vec7")),     vec8(_T("vec8")),
			 vec9(_T("vec9")),   vec10(_T("vec10")),   vec11(_T("vec11")),   vec12(_T("vec12")),   vec13(_T("vec13")),   vec14(_T("vec14")),   vec15(_T("vec15")),   vec16(_T("vec16")),
			ivec1(_T("ivec1")), ivec2(_T("ivec2")),   ivec3(_T("ivec3")),   ivec4(_T("ivec4")),   ivec5(_T("ivec5")),   ivec6(_T("ivec6")),   ivec7(_T("ivec7")),   ivec8(_T("ivec8")),
			ivec9(_T("ivec9")), ivec10(_T("ivec10")), ivec11(_T("ivec11")), ivec12(_T("ivec12")), ivec13(_T("ivec13")), ivec14(_T("ivec14")), ivec15(_T("ivec15")), ivec16(_T("ivec16")),
			cvec1(_T("cvec1")), cvec2(_T("cvec2")),   cvec3(_T("cvec3")),   cvec4(_T("cvec4")),   cvec5(_T("cvec5")),   cvec6(_T("cvec6")),   cvec7(_T("cvec7")),    cvec8(_T("cvec8")),
			cvec9(_T("cvec9")), cvec10(_T("cvec10")), cvec11(_T("cvec11")), cvec12(_T("cvec12")), cvec13(_T("cvec13")), cvec14(_T("cvec14")), cvec15(_T("cvec15")),  cvec16(_T("cvec16"))
		{}
	};

	DECLARE_TYPE(TYPE_VOID, QValue, _T("??"));
	DECLARE_TYPE(TYPE_LAMBDAEXPR, Function, _T("lambda_expr"));
	DECLARE_TYPE(TYPE_TYPEINFO, Type, _T("type"));

	INLINE void DbgBreak()
	{
#if defined(_MSC_VER)
		_CrtDbgBreak();
#elif defined(__GNUC__)
		__builtin_trap();
#endif
	}

} // namespace quasar

#endif
