#pragma once
//=============================================================================
// Quasar Domain-specific Language (DSL) for C++
// Extension for C++ class bindings
// (C) 2011-2020 Gepura
//=============================================================================

#include "quasar_dsl.h"
#include <sstream>
#include <type_traits>

namespace quasar
{
	namespace detail
	{
		// SFINAE helper routines for detecting whether a type has from_qvalue implemented
		template<class> struct sfinae_true : std::true_type {};
		template<class T, class A0>
		static sfinae_true<decltype(std::declval<T>().from_qvalue(std::declval<A0>()))> test_from_qvalue(int);
		template<class, class A0>
		static std::false_type test_from_qvalue(long);
		template<class T, class Arg>
		struct has_from_qvalue : decltype(test_from_qvalue<T, Arg>(0)){};

		template <typename T>
		static void TypeListToString(stringstream &output, T)
		{
			output << quasar::typeinfo_t<T>::typeName;
		}
		template <typename T, typename ...A>
		static void TypeListToString(stringstream &output, T t, A... cargs)
		{
			TypeListToString(output, t);
			output << _T(",");
			TypeListToString(output, cargs...);
		}

		static void TypeListToArray(LPCTSTR* array) {} // Empty argument list

		template <typename T>
		static void TypeListToArray(LPCTSTR* array, T)
		{
			*array = quasar::typeinfo_t<T>::typeName;
		}
		template <typename T, typename ...A>
		static void TypeListToArray(LPCTSTR* array, T t, A... cargs)
		{
			TypeListToArray(array, t);
			TypeListToArray(array + 1, cargs...);
		}

		// Single parameter conversion routine. Allows to override the
		// default QValue->R conversion (for example for quasar::runtime
		// specific types).
		template <typename R>
		static inline void Convert(QValue arg, R &out)
		{
			out = (R)arg; // Use default QValue conversion
		}
		// modern C++ boilerplate code for arg has from_qvalue method
		template <typename T, std::enable_if_t<has_from_qvalue<T, const qvalue_t &>() == true, int> = 0>
		static inline void Convert(QValue arg, ref<T> &out)
		{
			out = T::from_qvalue(arg);
		}
		template <typename T, std::enable_if_t<has_from_qvalue<T, const qvalue_t &>() == false, int> = 0>
		static inline void Convert(QValue arg, ref<T> &out)
		{
			out = ref<T>(static_cast<T*>(((qvalue_t&)arg).private_obj));
		}

		// Conversion to QValue
		template <typename T>
		static inline QValue Convert(T arg)
		{
			return (QValue)arg;
		}
		template <typename T>
		static inline QValue Convert(ref<T> arg)
		{
			qvalue_t qv = arg->wrap();
			return QValue(qv);
		}
		// Function to convert multiple QValue arguments to T...
		template <typename T>
		static inline void ConvertArgs(QValue* arg, int count, T& output)
		{
			Convert(*arg, output);
		}
		template <typename T, typename ...R>
		static inline void ConvertArgs(QValue* args, int count, T& output, R&... others)
		{
			ConvertArgs(args, count, output);
			if (count > 0)
				ConvertArgs(args + 1, count - 1, others...);
		}
		template <typename ...R, int ...I>
		static inline void ConvertRetVals(qvalue_t* args, std::tuple<R...>& val, utility::seq<I...>)
		{
			constexpr int N = sizeof...(R);
			QValue retVals[] = { Convert(std::get<I>(val))... };
			for (int i = 0; i < N; i++)
			{
				retVals[i].Retain();
				args[i] = retVals[i];
			}
		}

		// Function to perform the function calls
		template<typename R, typename...T, int... S>
		static R FuncCallHelper2(R(*func)(T...), QValue *args, utility::seq<S...>)
		{
			std::tuple<std::decay_t<T>...> cargs;
			ConvertArgs(args, sizeof...(T), std::get<S>(cargs)...);
			return (R)(*func)(std::get<S>(cargs)...);
		}

		// Function without parameters
		template<typename R, int... S>
		static R FuncCallHelper2(R(*func)(), QValue *args, utility::seq<S...>)
		{
			return (R)(*func)();
		}

		// Helper class for performing function calls that expands an array into a 
		// variadic parameter pack. Use functor to correctly handle functions returning 'void'.
		template<typename R, typename... T>
		struct FuncCallHelper
		{
			void operator ()(qvalue_t *argsOut, R(*func)(T...), QValue* args)
			{
				QValue retVal = Convert(FuncCallHelper2<R, T...>(func, args, utility::gen_seq<sizeof...(T)>()));
				retVal.Retain();
				argsOut[0] = retVal;
			}
		};

		template<typename... T>
		struct FuncCallHelper<void, T...>
		{
			void operator ()(qvalue_t *argsOut, void(*func)(T...), QValue *args)
			{
				constexpr int N = sizeof...(T);
				FuncCallHelper2<void, T...>(func, args, utility::gen_seq<N>());
				argsOut[0] = qvalue_t::empty;
			};
		};

		template<>
		struct FuncCallHelper<void>
		{
			void operator ()(qvalue_t *argsOut, void(*func)(), QValue *args)
			{
				FuncCallHelper2<void>(func, args, utility::gen_seq<0>());
				argsOut[0] = qvalue_t::empty;
			};
		};

		template<typename ...R, typename... T>
		struct FuncCallHelper<std::tuple<R...>, T...>
		{
			typedef std::tuple<R...> Rt;
			void operator ()(qvalue_t *argsOut, Rt(*func)(T...), QValue* args)
			{
				Rt retVal = FuncCallHelper2<Rt, T...>(func, args, utility::gen_seq<sizeof...(T)>());
				ConvertRetVals(argsOut, retVal, utility::gen_seq<sizeof...(R)>());
			}
		};


		// Function to perform the method calls
		template<typename R, typename C, typename...T, int... S>
		static R MethodCallHelper2(R(C::*func)(T...), C *pthis, QValue *args, utility::seq<S...>)
		{
			std::tuple<std::decay_t<T>...> cargs;
			ConvertArgs(args, sizeof...(T), std::get<S>(cargs)...);
			return (R)(pthis->*func)(std::get<S>(cargs)...);
		}

		// Idem but now for a const method
		template<typename R, typename C, typename...T, int... S>
		static R MethodCallHelper2(R(C::*func)(T...) const, C *pthis, QValue *args, utility::seq<S...>)
		{
			std::tuple<std::decay_t<T>...> cargs;
			ConvertArgs(args, sizeof...(T), std::get<S>(cargs)...);
			return (R)(pthis->*func)(std::get<S>(cargs)...);
		}

		// Method without parameters
		template<typename R, typename C, int... S>
		static R MethodCallHelper2(R(C::*func)(), C *pthis, QValue *args, utility::seq<S...>)
		{
			return (R)(pthis->*func)();
		}

		// Const method without parameters
		template<typename R, typename C, int... S>
		static R MethodCallHelper2(R(C::*func)() const, C *pthis, QValue *args, utility::seq<S...>)
		{
			return (R)(pthis->*func)();
		}

		// Helper class for method call that expands an array into a variadic parameter pack
		// Use functor to correctly handle functions returning 'void'.
		template<typename R, typename C, typename... T>
		struct MethodCallHelper
		{
			void operator ()(qvalue_t *argsOut, R(C::*func)(T...), C *pthis, QValue* args)
			{
				QValue retVal = Convert(MethodCallHelper2<R, C, T...>(func, pthis, args, utility::gen_seq<sizeof...(T)>()));
				retVal.Retain();
				argsOut[0] = retVal;
			}
			void operator ()(qvalue_t *argsOut, R(C::*func)(T...) const, C *pthis, QValue* args)
			{
				QValue retVal = Convert(MethodCallHelper2<R, C, T...>(func, pthis, args, utility::gen_seq<sizeof...(T)>()));
				retVal.Retain();
				argsOut[0] = retVal;
			}
		};

		template<typename C, typename... T>
		struct MethodCallHelper<void, C, T...>
		{
			void operator ()(qvalue_t *argsOut, void(C::*func)(T...), C *pthis, QValue *args)
			{
				constexpr int N = sizeof...(T);
				MethodCallHelper2<void, C, T...>(func, pthis, args, utility::gen_seq<N>());
				argsOut[0] = qvalue_t::empty;
			};
		};

		template<typename C>
		struct MethodCallHelper<void, C>
		{
			void operator ()(qvalue_t *argsOut, void(C::*func)(), C *pthis, QValue *args)
			{
				MethodCallHelper2<void, C>(func, pthis, args, utility::gen_seq<0>());
				argsOut[0] = qvalue_t::empty;
			};
		};

		template<typename ...R, typename C, typename... T>
		struct MethodCallHelper<std::tuple<R...>, C, T...>
		{
			typedef std::tuple<R...> Rt;
			void operator ()(qvalue_t *argsOut, Rt(C::*func)(T...), C *pthis, QValue* args)
			{
				Rt retVal = MethodCallHelper2<Rt, C, T...>(func, pthis, args, utility::gen_seq<sizeof...(T)>());
				ConvertRetVals(argsOut, retVal, utility::gen_seq<sizeof...(R)>());
			}
			void operator ()(qvalue_t *argsOut, Rt(C::*func)(T...) const, C *pthis, QValue* args)
			{
				Rt retVal = MethodCallHelper2<Rt, C, T...>(func, pthis, args, utility::gen_seq<sizeof...(T)>());
				ConvertRetVals(argsOut, retVal, utility::gen_seq<sizeof...(R)>());
			}
		};

		template<typename R, typename C>
		struct VariadicMethodCallHelper
		{
			qvalue_t operator ()(R(C::*func)(const qvalue_t*, int), C *pthis, const qvalue_t* inArgs, int nArgs)
			{
				QValue retVal = Convert((pthis->*func)(inArgs, nArgs));
				retVal.Retain();
				return (qvalue_t)retVal;
			}
		};
		template<typename C>
		struct VariadicMethodCallHelper<void, C>
		{
			qvalue_t operator ()(void(C::*func)(const qvalue_t*, int), C *pthis, const qvalue_t* inArgs, int nArgs)
			{
				(pthis->*func)(inArgs, nArgs);
				return qvalue_t::empty;
			}
		};

		template<typename R>
		struct VariadicFunctionCallHelper
		{
			qvalue_t operator ()(R(*func)(const qvalue_t*, int), const qvalue_t* inArgs, int nArgs)
			{
				QValue retVal = Convert((*func)(inArgs, nArgs));
				retVal.Retain();
				return (qvalue_t)retVal;
			}
		};
		template<>
		struct VariadicFunctionCallHelper<void>
		{
			qvalue_t operator ()(void(*func)(const qvalue_t*, int), const qvalue_t* inArgs, int nArgs)
			{
				(*func)(inArgs, nArgs);
				return qvalue_t::empty;
			}
		};
	}

	using namespace detail;

	/*
	* Helper class to perform reflection on C++ struct fields and methods,
	* allowing automatic binding.
	*/
	struct TypeBuilderExt
	{
		static LPCTSTR argNames[16];
	private:
		using stringstream = std::basic_stringstream<TCHAR, std::char_traits<TCHAR>, std::allocator<TCHAR>>;

#ifdef CPLUSPLUS14
		static void TypeListToString(stringstream &output) {} // Empty argument list

#endif
	private:
#ifdef CPLUSPLUS14
		static Expression PrepareAddMethod(TypeBuilder &tb, LPCTSTR methodName,
			std::vector<VariableDecl> &varDecl,
			string_t &className,
			string_t &typeString,
			const LPCTSTR *typeNames,
			const int N,
			LPCTSTR returnTypeName)
		{
			std::vector<Symbol> symbols;
			// Variable declarations
			varDecl.push_back(VariableDecl(_T("self"), className));
			for (int i = 0; i < N; i++)
				varDecl.push_back(VariableDecl(argNames[i], typeNames[i]));
			// Symbols
			symbols.push_back(Symbol(SYM_VARIABLE, 0, _T("self")));
			for (int i = 0; i < N; i++)
				symbols.push_back(Symbol(SYM_VARIABLE, 0, argNames[i]));
			symbols.push_back(Symbol(SYM_MEMBERFUNCTION, N, methodName));

			// Type string
			stringstream stream;
			stream << _T("[(");
			for (int i = 0; i < N; i++)
			{
				if (i > 0) stream << _T(",");
				stream << typeNames[i];
			}
			stream << _T(")->(");
			stream << returnTypeName;
			stream << _T(")]");
			typeString = (LPCTSTR)stream.str().c_str();
			return Expression(&symbols.front(), (int)symbols.size());
		}
		static Expression PrepareAddVariadicMethod(TypeBuilder &tb, LPCTSTR methodName,
			std::vector<VariableDecl> &varDecl,
			string_t &className,
			string_t &typeString,
			LPCTSTR returnTypeName)
		{
			std::vector<Symbol> symbols;
			// Variable declarations
			varDecl.push_back(VariableDecl(_T("self"), className));
			varDecl.push_back(VariableDecl(argNames[0], _T("?\?")));
			// Symbols
			symbols.push_back(Symbol(SYM_VARIABLE, 0, _T("self")));
			symbols.push_back(Symbol(SYM_VARIABLE, 0, argNames[0]));
			symbols.push_back(Symbol(SYM_MEMBERFUNCTION, 1, methodName));

			// Type string
			stringstream stream;
			stream << _T("[(...?\?)->(");
			stream << returnTypeName;
			stream << _T(")]");
			typeString = (LPCTSTR)stream.str().c_str();
			return Expression(&symbols.front(), (int)symbols.size(), true);
		}
		static Expression PrepareAddVariadicFunction(LPCTSTR funcName,
			std::vector<VariableDecl> &varDecl,
			string_t &typeString,
			LPCTSTR returnTypeName)
		{
			std::vector<Symbol> symbols;
			LPCTSTR argName = _T("args");
			// Variable declarations
			varDecl.push_back(VariableDecl(argName, _T("?\?")));
			// Symbols
			symbols.push_back(Symbol(SYM_VARIABLE, 0, argName));
			symbols.push_back(Symbol(SYM_FUNCTION, 1, funcName));

			// Type string
			stringstream stream;
			stream << _T("[(...?\?)->(");
			stream << returnTypeName;
			stream << _T(")]");
			typeString = (LPCTSTR)stream.str().c_str();
			return Expression(&symbols.front(), (int)symbols.size(), true);
		}
		static Expression PrepareAddFunction(LPCTSTR funcName,
			std::vector<VariableDecl> &varDecl,
			string_t &typeString,
			const LPCTSTR *typeNames,
			const int N,
			LPCTSTR returnTypeName)
		{
			std::vector<Symbol> symbols;
			// Variable declarations
			for (int i = 0; i < N; i++)
				varDecl.push_back(VariableDecl(argNames[i], typeNames[i]));
			// Symbols
			for (int i = 0; i < N; i++)
				symbols.push_back(Symbol(SYM_VARIABLE, 0, argNames[i]));
			symbols.push_back(Symbol(SYM_FUNCTION, N, funcName));

			// Type string
			stringstream stream;
			stream << _T("[(");
			for (int i = 0; i < N; i++)
			{
				if (i > 0) stream << _T(",");
				stream << typeNames[i];
			}
			stream << _T(")->(");
			stream << returnTypeName;
			stream << _T(")]");
			typeString = (LPCTSTR)stream.str().c_str();
			return Expression(&symbols.front(), (int)symbols.size());
		}

		static Expression PrepareAddOperator(OperatorTypes opType,
			std::vector<VariableDecl> &varDecl,
			string_t &typeString,
			const LPCTSTR *typeNames,
			const int N,
			LPCTSTR returnTypeName)
		{
			std::vector<Symbol> symbols;
			// Variable declarations
			for (int i = 0; i < N; i++)
				varDecl.push_back(VariableDecl(argNames[i], typeNames[i]));
			// Symbols
			for (int i = 0; i < N; i++)
				symbols.push_back(Symbol(SYM_VARIABLE, 0, argNames[i]));
			symbols.push_back(Symbol(SYM_OPERATOR, N, opType));

			// Type string
			stringstream stream;
			stream << _T("[(");
			for (int i = 0; i < N; i++)
			{
				if (i > 0) stream << _T(",");
				stream << typeNames[i];
			}
			stream << _T(")->(");
			stream << returnTypeName;
			stream << _T(")]");
			typeString = (LPCTSTR)stream.str().c_str();
			return Expression(&symbols.front(), (int)symbols.size());
		}

		static void RegisterMethod(LPCTSTR typeString, void *closurePtr, LambdaDelegate *wrapper,
			std::vector<VariableDecl> &varDecl, Expression &searchExpr, DataTypes dataType = TYPE_METHOD)
		{
			auto host = IQuasarHost::GetInstance();
			QValue action = host->CreateLambda(Type((LPCTSTR)typeString), closurePtr, wrapper);
			((qvalue_t&)action).type = dataType;
			bool result = host->GetComputationEngine()->GetRuntimeReductionEngine()->Add(&varDecl[0], (int)varDecl.size(), 
				&searchExpr, &(qvalue_t &) action, nullptr);
			// Reduction system does not support this reduction
			if (!result)
			{
				_ASSERT_EXPR(false, "Reduction system does not support value");
				delete closurePtr;
				action.ReleaseRef();
				return;
			}
			_ASSERT_EXPR(result, "error adding reduction!");
			host->AddLambdaDestroyCallback(action, [](void* closurePtr, LambdaDelegate* lambda)
			{
				delete closurePtr;
			});
		}
		static void PrepareAddProperty(TypeBuilder &tb, LPCTSTR propertyTypeName,
			string_t &getTypeString, string_t &setTypeString)
		{
			stringstream stream;
			stream << _T("[");
			stream << _T("?\?"); //(LPCTSTR)tb.GetTypeName();
			stream << _T("->");
			stream << propertyTypeName;
			stream << _T("]");
			getTypeString = (LPCTSTR)stream.str().c_str();

			stream = stringstream();
			stream << _T("[(");
			stream << _T("?\?"); //(LPCTSTR)tb.GetTypeName();
			stream << _T(",");
			stream << propertyTypeName;
			stream << _T(")->()]");
			setTypeString = (LPCTSTR)stream.str().c_str();
		}
		static void RegisterProperty(
			TypeBuilder &tb, LPCTSTR propertyName, LPCTSTR propertyType,
			LPCTSTR getTypeString, void* getClosurePtr, LambdaDelegate *getWrapper,
			LPCTSTR setTypeString, void* setClosurePtr, LambdaDelegate *setWrapper)
		{
			auto host = IQuasarHost::GetInstance();
			QValue getAction = host->CreateLambda(Type((LPCTSTR)getTypeString), getClosurePtr, getWrapper);
			QValue setAction = host->CreateLambda(Type((LPCTSTR)setTypeString), setClosurePtr, setWrapper);
			((qvalue_t&)getAction).type = TYPE_METHOD;
			((qvalue_t&)setAction).type = TYPE_METHOD;
			tb.AddProperty(propertyName, propertyType, getAction, setAction);
			host->AddLambdaDestroyCallback(getAction, [](void* closurePtr, LambdaDelegate* lambda)
			{
				delete closurePtr;
			});
			host->AddLambdaDestroyCallback(setAction, [](void* closurePtr, LambdaDelegate* lambda)
			{
				delete closurePtr;
			});
		}
#endif
	public:
#ifdef CPLUSPLUS14
		// Adds the method with the given member function reference. A wrapper
		// is automatically generated, so that the method can be called using
		// the LambdaDelegate calling interface
		template <typename R, typename... T, typename C>
		static void AddMethod(TypeBuilder& tb, LPCTSTR methodName, R(C::*memberFunc)(T...))
		{
			// (a0 : int, a1) -> c.get_length(a0)
			std::vector<VariableDecl> varDecl;
			std::vector<Symbol> symbols;
			constexpr const int N = sizeof...(T);
			LPCTSTR typeNames[N + 1];
			_ASSERT_EXPR(N < ARRAY_LENGTH(argNames, LPCTSTR), "Too many parameters");

			TypeListToArray(typeNames, T()...);
			string_t typeString;
			string_t className = tb.GetTypeName();
			Expression searchExpr = PrepareAddMethod(tb, methodName, varDecl,
				className, typeString, typeNames, N, typeinfo_t<R>::typeName);

			// Construct a LambdaDelegate with closure pointer that can be passed
			// to the Quasar runtime system. MethodCallHelper will allow our method
			// to be called natively (with automatic conversion from QValue to the
			// target type).
			void *closurePtr = new decltype(memberFunc)(memberFunc);
			LambdaDelegate* wrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto host = IQuasarHost::GetInstance();
				auto memberFunc = *(R(C::**)(T...))closurePtr; // Some C++ madness...
				_ASSERT_EXPR(nArgsIn - 1 == sizeof...(T) && nArgsOut == 1,
					"Invalid number of input/output arguments for value call!");
				QValue args[sizeof...(T)+1];
				const qvalue_t &pthis = argsIn[0];
				_ASSERT_EXPR(pthis.private_obj != nullptr, "Method call: this is null pointer");
				for (int i = 0; i < sizeof...(T); i++)
				{
					args[i] = argsIn[i + 1];
					args[i].Retain();
				}
				C *ptr = host->LockObject<C>(pthis);
				MethodCallHelper<R, C, T...>()(
					argsOut, memberFunc, ptr, args);
				host->UnlockObject(pthis);
		};

			// Define a reduction to call this method
			RegisterMethod(typeString, closurePtr, wrapper, varDecl, searchExpr);
		}

		// Method with multiple return values
		template <typename... R, typename... T, typename C>
		static void AddMethod(TypeBuilder& tb, LPCTSTR methodName, std::tuple<R...>(C::*memberFunc)(T...))
		{
			// (a0 : int, a1) -> c.get_length(a0)
			std::vector<VariableDecl> varDecl;
			std::vector<Symbol> symbols;
			constexpr const int N = sizeof...(T);
			LPCTSTR typeNames[N + 1];
			_ASSERT_EXPR(N < ARRAY_LENGTH(argNames, LPCTSTR), "Too many parameters");

			TypeListToArray(typeNames, T()...);
			string_t typeString;
			string_t className = tb.GetTypeName();
			Expression searchExpr = PrepareAddMethod(tb, methodName, varDecl,
				className, typeString, typeNames, N, _T("?\?")); // TODO : return types

			// Construct a LambdaDelegate with closure pointer that can be passed
			// to the Quasar runtime system. MethodCallHelper will allow our method
			// to be called natively (with automatic conversion from QValue to the
			// target type).
			void *closurePtr = new decltype(memberFunc)(memberFunc);
			LambdaDelegate* wrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto host = IQuasarHost::GetInstance();
				auto memberFunc = *(std::tuple<R...>(C::**)(T...))closurePtr; // Some C++ madness...
				_ASSERT_EXPR(nArgsIn - 1 == sizeof...(T),
					"Invalid number of input arguments for value call!");
				_ASSERT_EXPR(nArgsOut == sizeof...(R),
					"Invalid number of output arguments for value call!");
				QValue args[sizeof...(T) + 1];
				const qvalue_t &pthis = argsIn[0];
				for (int i = 0; i < sizeof...(T); i++)
				{
					args[i] = argsIn[i + 1];
					args[i].Retain();
				}
				C *ptr = host->LockObject<C>(pthis);
				MethodCallHelper<std::tuple<R...>, C, T...>()(
					argsOut, memberFunc, ptr, args);
				host->UnlockObject(pthis);
			};

			// Define a reduction to call this method
			RegisterMethod(typeString, closurePtr, wrapper, varDecl, searchExpr);
		}

		template <typename C, typename R>
		static void AddVariadicMethod(TypeBuilder &tb, LPCTSTR methodName, R(C::*memberFunc)(const qvalue_t*, int))
		{
			// (...a0) -> c.get_length(a0)
			std::vector<VariableDecl> varDecl;
			std::vector<Symbol> symbols;

			string_t typeString;
			string_t className = tb.GetTypeName();
			Expression searchExpr = PrepareAddVariadicMethod(tb, methodName, varDecl,
				className, typeString, typeinfo_t<R>::typeName);

			void *closurePtr = new decltype(memberFunc)(memberFunc);
			LambdaDelegate *wrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto host = IQuasarHost::GetInstance();
				auto memberFunc = *(R(C::**)(const qvalue_t*, int))closurePtr; // Some C++ madness...
				_ASSERT_EXPR(nArgsIn >= 1 && nArgsOut == 1,
					"Invalid number of input/output arguments for value call!");
				const qvalue_t &pthis = argsIn[0];
				_ASSERT_EXPR(pthis.private_obj != nullptr, "Method call: this is null pointer");
				C *ptr = host->LockObject<C>(pthis);
				argsOut[0] = VariadicMethodCallHelper<R, C>()(
					memberFunc, ptr, argsIn + 1, nArgsIn - 1);
				host->UnlockObject(pthis);
			};

			// Define a reduction to call this method
			RegisterMethod(typeString, closurePtr, wrapper, varDecl, searchExpr);
		}

		// Adds a property with the specified getter and setter. Wrapper are
		// automatically generated, so that the getters/setters can be called
		// using the LambdaDelegate calling interface
		template <typename C, typename R>
		static void AddProperty(TypeBuilder &tb, LPCTSTR propertyName,
			R(C::*getMethod)() const, void(C::*setMethod)(R))
		{
			void *getClosurePtr = new decltype(getMethod)(getMethod);
			void *setClosurePtr = new decltype(setMethod)(setMethod);
			_ASSERT_EXPR(((const qvalue_t &)tb).type == TYPE_TYPEINFO,
				"Type info expected!");

			string_t getTypeString, setTypeString;
			PrepareAddProperty(tb, typeinfo_t<R>::typeName, getTypeString, setTypeString);

			LambdaDelegate* getWrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				_ASSERT_EXPR(nArgsIn == 1 && nArgsOut == 1,
					"Invalid number of input/output arguments for property get!");
				auto memberFunc = *(R(C::**)() const)closurePtr; // Some C++ madness...
				const qvalue_t &pthis = argsIn[0];
				_ASSERT_EXPR(pthis.private_obj != nullptr, "Property get: this is null pointer");
				_ASSERT_EXPR(pthis.type == TYPE_UNTYPEDOBJECT,
					"Untyped object expected for property get!");
				C *ptr = static_cast<C *>(pthis.private_obj);
				MethodCallHelper<R, C>()(argsOut, memberFunc, ptr, nullptr);
			};

			LambdaDelegate* setWrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				_ASSERT_EXPR(nArgsIn == 2 && nArgsOut == 1,
					"Invalid number of input/output arguments for property set!");
				auto memberFunc = *(void(C::**)(R))closurePtr; // Some C++ madness...
				const qvalue_t &pthis = argsIn[0];
				_ASSERT_EXPR(pthis.private_obj != nullptr, "Property set: this is null pointer");
				QValue args[] = { argsIn[1] };
				//args[0].Retain(); //checked
				_ASSERT_EXPR(pthis.type == TYPE_UNTYPEDOBJECT,
					"Untyped object expected for property get!");
				C *ptr = static_cast<C *>(pthis.private_obj);
				MethodCallHelper<void, C, R>()(argsOut, memberFunc, ptr, args);
				argsOut[0] = qvalue_t::empty;
			};
			RegisterProperty(tb, propertyName, typeinfo_t<R>::typeName,
				getTypeString, getClosurePtr, getWrapper,
				setTypeString, setClosurePtr, setWrapper);
		}

		template <typename R, typename... T>
		static void RegisterFunction(LPCTSTR functionName, R(funcPtr)(T...))
		{
			std::vector<VariableDecl> varDecl;
			std::vector<Symbol> symbols;
			constexpr const int N = sizeof...(T);
			LPCTSTR typeNames[N];
			_ASSERT_EXPR(N < ARRAY_LENGTH(argNames, LPCTSTR), "Too many parameters");

			TypeListToArray(typeNames, T()...);
			string_t typeString;
			Expression searchExpr = PrepareAddFunction(functionName, varDecl,
				typeString, typeNames, N, typeinfo_t<R>::typeName);

			// Construct a LambdaDelegate with closure pointer that can be passed
			// to the Quasar runtime system. FuncCallHelper will allow our method
			// to be called natively (with automatic conversion from QValue to the
			// target type).
			void *closurePtr = new decltype(funcPtr)(funcPtr);
			LambdaDelegate* wrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto host = IQuasarHost::GetInstance();
				auto funcPtr = *(R(**)(T...))closurePtr; // Some C++ madness...
				_ASSERT_EXPR(nArgsIn == sizeof...(T) && nArgsOut == 1,
					"Invalid number of input/output arguments for function call!");
				QValue args[sizeof...(T)];
				for (int i = 0; i < sizeof...(T); i++)
				{
					args[i] = argsIn[i];
					args[i].Retain();
				}
#if defined(_WIN32) || defined(_WIN64)				
				FuncCallHelper<R, T...>()(argsOut, funcPtr, args);
#else// catch exceptions here and do not propagate them, since the managed code can not catch it in linux			
				try
				{
					FuncCallHelper<R, T...>()(argsOut, funcPtr, args);
				}
				catch(...)
				{
					//TODO
					// set the errorMsg, if it is not already set.
				}				
#endif
			};

			// Define a reduction to call this method
			RegisterMethod(typeString, closurePtr, wrapper, varDecl, searchExpr, TYPE_LAMBDAEXPR);
		}

		template <typename R>
		static void RegisterVariadicFunction(LPCTSTR functionName, R(funcPtr)(const qvalue_t*, int))
		{
			std::vector<VariableDecl> varDecl;
			std::vector<Symbol> symbols;

			string_t typeString;
			Expression searchExpr = PrepareAddVariadicFunction(functionName, varDecl,
				typeString, typeinfo_t<R>::typeName);

			void *closurePtr = new decltype(funcPtr)(funcPtr);
			LambdaDelegate* wrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto host = IQuasarHost::GetInstance();
				auto funcPtr = *(R(**)(const qvalue_t*, int))closurePtr; // Some C++ madness...
				_ASSERT_EXPR(nArgsOut == 1,
					"Invalid number of input/output arguments for function call!");
				argsOut[0] = VariadicFunctionCallHelper<R>()(funcPtr, argsIn, nArgsIn);
			};

			// Define a reduction to call this function
			RegisterMethod(typeString, closurePtr, wrapper, varDecl, searchExpr, TYPE_LAMBDAEXPR);
		}


		template <typename R, typename... T>
		static void RegisterOperator(OperatorTypes opType, R(funcPtr)(T...))
		{
			std::vector<VariableDecl> varDecl;
			std::vector<Symbol> symbols;
			constexpr const int N = sizeof...(T);
			LPCTSTR typeNames[N];
			_ASSERT_EXPR(N < ARRAY_LENGTH(argNames, LPCTSTR), "Too many parameters");

			TypeListToArray(typeNames, T()...);
			string_t typeString;
			Expression searchExpr = PrepareAddOperator(opType, varDecl,
				typeString, typeNames, N, typeinfo_t<R>::typeName);

			// Construct a LambdaDelegate with closure pointer that can be passed
			// to the Quasar runtime system. FuncCallHelper will allow our method
			// to be called natively (with automatic conversion from QValue to the
			// target type).
			void *closurePtr = new decltype(funcPtr)(funcPtr);
			LambdaDelegate* wrapper = [](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				auto host = IQuasarHost::GetInstance();
				auto funcPtr = *(R(**)(T...))closurePtr; // Some C++ madness...
				_ASSERT_EXPR(nArgsIn == sizeof...(T) && nArgsOut == 1,
					"Invalid number of input/output arguments for function call!");
				QValue args[sizeof...(T)];
				for (int i = 0; i < sizeof...(T); i++)
				{
					args[i] = argsIn[i];
					args[i].Retain();
				}
				FuncCallHelper<R, T...>()(argsOut, funcPtr, args);
			};

			// Define a reduction to call this method
			RegisterMethod(typeString, closurePtr, wrapper, varDecl, searchExpr, TYPE_LAMBDAEXPR);
		}
#endif
	};

	LPCTSTR SELECTANY TypeBuilderExt::argNames[16] = {
		_T("a"),_T("b"),_T("c"),_T("d"),_T("e"),_T("f"),_T("g"), _T("h"),
		_T("i"),_T("j"),_T("k"),_T("l"),_T("m"),_T("n"), _T("o"), _T("p") };

	#ifndef E_NOINTERFACE
	#define E_NOINTERFACE 0x80000004L
	#endif
	#ifndef S_OK
	#define S_OK 0
	#endif

	#ifndef STDAPI
	#if defined(WIN32) || defined(WIN64)
	#define STDAPI __stdcall
	#else
	#define STDAPI
	#endif
	#endif

	/*
	 * IUnknown interface definition
	 */
	#ifndef __IUnknown_INTERFACE_DEFINED__

	struct IID;
	typedef struct IID & REFIID;

	struct IUnknown
	{
		virtual long STDAPI QueryInterface(REFIID riid, void** ppvObject) = 0;
		virtual unsigned long STDAPI AddRef(void) = 0;
		virtual unsigned long STDAPI Release(void) = 0;
	}; 

	#endif

	/*
	 * Base object class
	 */
	class ObjBase : public IUnknown
	{
	private:
		int refCount;
		void *handle;

	public:
		ObjBase(qvalue_t &type) :refCount(0),handle(IQuasarHost::GetInstance()->RegisterInstance(this, (qvalue_t &)type)) {}
		virtual ~ObjBase() {}

		// Class must have a wrap() method, because some of the macros
		// in quasar_dsl_class_ext.h depend on it. The wrap function wraps
		// the object in a qvalue_t object.
		qvalue_t wrap()
		{
			qvalue_t qv;
			qv.type = TYPE_UNTYPEDOBJECT;
			qv.private_obj = handle;
			IQuasarHost::GetInstance()->AddRef(qv);
			return qv;
		}

		virtual long STDAPI QueryInterface(
			REFIID riid,
			void **ppvObject)
		{
			return E_NOINTERFACE;
		}

		virtual unsigned long STDAPI AddRef(void)
		{
			_atomicAdd(&refCount, 1);
			return refCount;
		}

		virtual unsigned long STDAPI Release(void)
		{
			if (_atomicAdd(&refCount, -1) == 0)
				delete this;
			return refCount;
		}
	};
} // namespace quasar

#define WFILE_HELPER(x) _T(x)
#define __WFILE__ WFILE_HELPER(__FILE__)

#ifdef CPLUSPLUS14
/*
* The following macros are useful for defining dynamic classes (aka untyped
* objects) that can be used from Quasar. The macros are best placed inside
* a static 'register_type' function within the class.
*/

/*
* Defines a new type.
* Parameters:
* - tb: variable name for the type builder (usually 'tb')
* - typeName: the C++ name of the type (without quotes)
* - extTypeName: the Quasar name of the type (with quotes; this may differ
*          from the C++ type name)
* - flags: TYPEFLAGS_DYNAMIC_CLASS, TYPEFLAGS_MUTABLE_CLASS or
*			TYPEFLAGS_IMMUTABLE_CLASS
*/
#define BEGIN_TYPE(tb, typeName, extTypeName, flags) \
	typedef typeName Self; \
	TypeBuilder tb(__WFILE__, extTypeName, flags); \
	tb.SetConstructor([](const qvalue_t* args, int nArgs) { return (new typeName())->wrap(); })

/*
 * Idem, but without constructor
 */
#define BEGIN_ABSTRACT_TYPE(tb, typeName, extTypeName, flags) \
	typedef typeName Self; \
	TypeBuilder tb(__WFILE__, extTypeName, flags);

/* Defines a new field within the class
* Parameters:
* - fieldName: reference to the class field (without quotes)
*/
#define ADD_FIELD(tb, fieldName) \
	tb.AddFieldWithOffset(_T(#fieldName), quasar::typeinfo_t<std::remove_pointer<decltype(fieldName)>::type>::typeName, offsetof(Self, fieldName))

/* Idem, but renamed field*/
#define ADD_FIELD_RENAME(tb, fieldName, newName) \
	tb.AddFieldWithOffset(_T(#newName), quasar::typeinfo_t<std::remove_pointer<decltype(fieldName)>::type>::typeName, offsetof(Self, fieldName))

/* Defines a new property within the class
* Parameters:
* - propertyName: name of the property (without quotes).
*    both methods get_##propertyName and set_##propertyName must exist.
*/
#define ADD_PROPERTY(tb, propertyName) \
	TypeBuilderExt::AddProperty(tb, _T(#propertyName), &get_##propertyName, &set_##propertyName)

/* Idem, but renamed property*/
#define ADD_PROPERTY_RENAME(tb, propertyName, newName) \
	TypeBuilderExt::AddProperty(tb, _T(#newName), &get_##propertyName, &set_##propertyName)

/* Defines a new method within the class
* Parameters:
* - methodName: name of the method (without quotes). The method type
*   and its parameters are captured using a variadic template technique
*/
#define ADD_METHOD(tb, methodName) \
	TypeBuilderExt::AddMethod(tb, _T(#methodName), &methodName)

/* Defines a new method within the class, but renames the C++ method name
* Parameters:
* - methodName: name of the method (without quotes).
* - newName: the Quasar name of the method
*/
#define ADD_METHOD_RENAME(tb, methodName, newName) \
	TypeBuilderExt::AddMethod(tb, _T(#newName), &methodName)

/* Idem, but works with overloaded methods */
#define ADD_OVERLOADED_METHOD(tb, methodName, signature) \
	TypeBuilderExt::AddMethod(tb, _T(#methodName), (signature)&methodName)


/* Defines a new variadic method within the class (with type signature
*   func(const qvalue_t* inArgs, int nArgs)
* Parameters:
* - methodName: name of the method (without quotes). The method type
*   and its parameters are captured using a variadic template technique
*/
#define ADD_VARIADIC_METHOD(tb, methodName) \
	TypeBuilderExt::AddVariadicMethod(tb, _T(#methodName), &methodName)

/* Idem, but renames the method */
#define ADD_VARIADIC_METHOD_RENAME(tb, methodName, newName) \
	TypeBuilderExt::AddVariadicMethod(tb, _T(#newName), &methodName)

/* Register a function with the reduction system */
#define ADD_FUNCTION(funcName) \
	TypeBuilderExt::RegisterFunction(_T(#funcName), &funcName)

/* Idem, but renames the function */
#define ADD_FUNCTION_RENAME(funcName, newName) \
	TypeBuilderExt::RegisterFunction(_T(#newName), &funcName)

/* Register a variadic function*/
#define ADD_VARIADIC_FUNCTION(funcName) \
	TypeBuilderExt::RegisterVariadicFunction(_T(#funcName), &funcName)

/* Idem, but renames the function name */
#define ADD_VARIADIC_FUNCTION_RENAME(funcName, newName) \
	TypeBuilderExt::RegisterVariadicFunction(_T(#newName), &funcName)

#define ADD_OPERATOR(funcName, opType) \
	TypeBuilderExt::RegisterOperator(opType, &funcName)


#endif
