//=============================================================================
// quasar_dynmem.h
// (C) 2013-2015 Gepura - imec - Ghent University
//=============================================================================
// Quasar dynamic memory management subsystem
// Based on a set of "allocator building blocks", more efficient allocators are
// constructed. The algorithms implement lock-free allocation mechanisms and
// hence should be quite fast, even when a large number of threads (>1024) are
// allocating memory concurrently.
//=============================================================================
#ifndef __QUASAR_DYNMEM_H__
#define __QUASAR_DYNMEM_H__
#include "quasar.h"

// The maximum parallelism in CUDA that we account for. Here, we count
// 8 multi-processors processing 1024 threads blocks. Note that the constant
// is mainly indicative and used to reduce memory conflicts
#define MAX_CUDA_PARALLEL 8192

namespace quasar
{
	// A generic atomicCAS function for pointers
#if defined(__LP64__)
	// 64-bit pointer mode
	INLINE __HOST_V__ __device__ void *
		atomicCASP(void *volatile *ptr, void *assumed, void *newval)
	{
		return (void *)_atomicCAS((long long *)ptr, (long long)assumed,
			(long long)newval);
	}
#else
	// 32-bit pointer mode
	INLINE __HOST_V__ __device__ void *
		atomicCASP(void *volatile *ptr, void *assumed, void *newval)
	{
		return (void *)_atomicCAS((unsigned int *)ptr, (int)assumed, (int)newval);
	}
#endif

    // Template function for calculating the logarithm basis 2
    template <size_t N> struct log2_
    {
        enum
        {
            value = 1 + log2_<N / 2>::value
        };
    };
    template <> struct log2_<1>
    {
        enum
        {
            value = 0
        };
    };
    template <> struct log2_<0>
    {
        enum
        {
            value = 0
        };
    };

    INLINE __device__ __HOST_V__ int CountBits(unsigned int v)
    {
		#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
			return __popc(v);
		#elif defined(__GNUG__)
			return __builtin_popcount(v);
		#elif defined(_MSC_VER)
			return __popcnt(v);
		#else
			v = v - ((v >> 1) & 0x55555555);
			v = (v & 0x33333333) + ((v >> 2) & 0x33333333);
			return ((v + (v >> 4) & 0xF0F0F0F) * 0x1010101) >> 24;
		#endif
    }

    INLINE __device__ __HOST_V__ int CountTrailingZeros(unsigned int x)
    {
		// See https://en.wikipedia.org/wiki/Find_first_set
		#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
			int val = __ffs(x);
			return val > 0 ? val - 1 : 32;
		#elif defined(__GNUG__)
			int val = __builtin_ffs(x);
			return val > 0 ? val - 1 : 32;
		#elif defined(_MSC_VER)
			unsigned int index;
			return _BitScanForward((unsigned long *)&index, x) ? index + 1 : 0;
		#else
		#error "CountTrailingZeros not implemented for this platform!"
		#endif
    }

	INLINE __device__ __HOST_V__ int CountTrailingOnes(unsigned int x)
	{
		// See https://en.wikipedia.org/wiki/Find_first_set
		#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
			int val = __ffs(~x);
			return val > 0 ? val - 1 : 32;
		#elif defined(__GNUG__)
			int val = __builtin_ffs(~x);
			return val > 0 ? val - 1 : 32;
		#elif defined(_MSC_VER)
			unsigned int index;
			return _BitScanForward((unsigned long *)&index, ~x) ? index + 1 : 0;
		#else
		#error "CountTrailingOnes not implemented for this platform!"
		#endif
	}

// Atomic dereferencing function - this function is needed in CUDA
// because atomic operations are only visible to other atomic ope-
// rations. The trick is then to perform a dereferencing using the
// atomic compare and exchange instruction.

	__HOST_V__ __device__ INLINE void *atomicDeref(void *volatile *address)
    {
	#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
		void *original = *address;
        void *assumed;
        do
        {
            assumed = original;
            original = atomicCASP(address, assumed, assumed);
        } while (original != assumed);
        return original;
	#else
		return *address;
	#endif
    }

    // Atomically assigning values to a pointer - same remarks as
    // in the above case apply.
    template <typename T> __HOST_V__ __device__ INLINE T *atomicSetP(T **dst, T *new_val)
    {
        volatile T *old_val;
        do
            old_val = *dst;
        while (atomicCASP((void *volatile *)dst, (void *)old_val,
                          (void *)new_val) != old_val);
        return (T *)old_val;
    }


    // On memory failure, a NULL pointer is returned
    INLINE __HOST_V__ __device__ void *alloc_fail() { return NULL; }

    // Memory alignment function
    template <int alignment> __HOST_V__ __device__ int align(size_t x)
    {
        return (x + alignment - 1) & ~(alignment - 1);
    }

    // Checks whether the address is aligned
    template <int alignment> __HOST_V__ __device__ bool is_aligned(void *ptr)
    {
        return ((size_t)ptr & (alignment - 1)) == 0;
    }

    template <bool> struct CompileTimeError;
    template <> struct CompileTimeError<true>
    {
    }; // specialized only for true
#define STATIC_CHECK(expr) (CompileTimeError<expr>())

    // Standard C memory allocator (Top allocator)
    // On GPUs we don't really want to use this function (it is generally too
    // slow)
    class Malloc
    {
      public:
		__HOST_V__ __device__ void *allocate(size_t size)
        {
            void *hdr = (void *)::malloc(size);
            if (hdr == NULL)
                return alloc_fail();
            return hdr;
        }
		__HOST_V__ __device__ void free(void *ptr) { ::free(ptr); }
		__HOST_V__ __device__ int get_size(void *ptr) { return -1; } // TODO
    };

    // An allocator that does nothing
    class NullAllocator
    {
      public:
		  __HOST_V__ __device__ void *allocate(size_t bytes)
        {
            Q_RAISE_ERROR("No allocator for object of the specified size!");
            return alloc_fail();
        }
		  __HOST_V__ __device__ void free(void *ptr) {}
        bool __HOST_V__ __device__ is_validpointer(void *ptr) { return false; }
		__HOST_V__ __device__ int get_size(void *ptr) { return -1; }
        __HOST_V__ __device__ int count_reserved() { return 0; }
		__HOST_V__ __device__ long long total_mem_used() { return 0; }
    };

    // Aligned memory allocator: allocates memory that is aligned on multiples
    // of the specified alignment constant
    template <int alignment, class Super> class AlignedAllocator : public Super
    {
    public:
		__HOST_V__ __device__ void *allocate(size_t size)
        {
            char *ptr = (char *)Super::allocate(size + alignment);
            if (ptr == NULL)
                return alloc_fail();
            char padding = (char)((size_t)ptr & (alignment - 1));
            if (padding == 0)
                padding = alignment; // Already aligned...
            ptr = (char *)((size_t)ptr + padding);
            ((char *)ptr)[-1] = padding;
            return ptr;
        }
		__HOST_V__ __device__ void free(void *ptr)
        {
            Super::free((char *)ptr - ((char *)ptr)[-1]);
        }
		__HOST_V__ __device__ int get_size(void *ptr) { return -1; } // TODO
        __HOST_V__ __device__ int count_reserved() { return 0; }
		__HOST_V__ __device__ long long total_mem_used() { return 0; } // TODO
		bool __HOST_V__ __device__ is_validpointer(void *ptr) { return true; } // TODO
    };


    // Implements thread safe (at least, if thread_safe=true) single linked
    // lists using a lock-free insertion and removal.
    template <typename NodeBase, bool thread_safe> class SingleLinkedList
    {
      public:
        struct Node : NodeBase
        {
            Node *next;
        };

      private:
        Node *first;

      public:
		__HOST_V__ __device__ SingleLinkedList() : first(NULL) {}
        __device__ void init() { first = NULL; }
        __device__ void addfirst(Node *node)
        {
            if (thread_safe == false)
            {
                node->next = first;
                first = node;
            }
            else
            {
                node->next = atomicSetP(&first, node);
            }
        }
        __device__ Node *get()
        {
            if (thread_safe)
                return (Node *)atomicDeref((void *volatile *)&first);
            else
                return (Node *)first;
        }
        __device__ Node *removefirst()
        {
            Node *node = get();
            if (node == NULL)
                return NULL;
            if (thread_safe == false)
                first = first->next;
            else
                node = atomicSetP(&first, (Node *)atomicDeref(
                                              (void *volatile *)&first->next));
            return (Node *)node;
        }
        __device__ bool tryremove(Node *node)
        {
            bool success = false;
            Node *currentFirst = get();
            if (currentFirst == node)
            {
                if (atomicCASP((void *volatile *)&first, currentFirst,
                               (Node *)atomicDeref(
                                   (void *volatile *)&currentFirst->next)) ==
                    currentFirst)
                {
                    return true;
                }
                currentFirst = first;
            }

            Node *next;
            for (Node *n = currentFirst; n != NULL; n = next)
            {
                if ((next = (Node *)atomicDeref((void *volatile *)&n->next)) ==
                    node)
                {
                    if (thread_safe == false)
                    {
                        next = n->next = node->next;
                        node->next = NULL;
                    }
                    else
                    {
						Node *p = next;
						do
						{
							next = p;
							p = (Node *)atomicCASP((void *volatile *)&n->next, next,
								atomicDeref((void *volatile *)&node->next));
						}
						while (p != next);

                        atomicSetP(&node->next, (Node *)NULL);
                    }
                    success = true;
                    break;
                }
            }
            return success;
        }
        __device__ void clear() { first = NULL; }
    };


    // ParallelZoneAllocator: my own parallel memory allocation
    // algorithm, implementing a lock-free approximate best fit
    // allocator using reservation bits. 32-bit reservation bits
    // are used, this leads to a restriction on the minimal and
    // maximal allocatable block sizes: the maximal block size
    // must be 32x the minimal block size.
    //
    // zone_size: the size of an individual zone (typically 1024 bytes)
    // zones: the number of zones.
    template <int zone_size, int zones, class Alternative>
    class ParallelZoneAllocator : public Alternative
    {
      public:
        enum
        {
            depth = log2_<32>::value,
            min_block_size = zone_size,
            max_block_size = zone_size * (1 << depth),
            total_size = max_block_size * zones, // The total amount of
            // memory that this allocator can store
            num_zones = zones
        };

      private:
        struct Header
        {
            int size;
        };
        unsigned int reserved[zones];
        char data[total_size];
        int last_index[MAX_CUDA_PARALLEL];

      public:
		__HOST_V__ __device__ ParallelZoneAllocator()
        {
            STATIC_CHECK(max_block_size <= INT_MAX);
#ifndef TARGET_CUDA
            memset(reserved, 0, sizeof(unsigned int) * num_zones);
#endif
        }

		__HOST_V__ __device__ void *allocate(size_t size)
        {
            int block_size = NextPow2((unsigned int)size + sizeof(Header));
            if (block_size > max_block_size)
                return Alternative::allocate(size);

            unsigned int width = (block_size + zone_size - 1) / zone_size;
            int log2_width = CountTrailingZeros(width);
            int tid = __threadidx();
            int bid = __blockidx();
            int nThreads = __threadcnt();
            int rid = (nThreads * bid + tid) % MAX_CUDA_PARALLEL;
            int i = last_index[rid]; // Start searching where we left off last
                                     // time...
            if (i <= 0)
            {
                // This will spread the threads across all chunks
                i = (tid * zones) / nThreads + bid;
            }
            i %= zones;
            int max_chunks = zones;
            unsigned int mask, orig, newOrig, d1, d2, d3, d4;
            int pos;
            newOrig = reserved[i];

        try_again:
            do
            {
                orig = newOrig;
                if (CountBits(orig) > 32 - (int)width)
                    goto next;
                else if (width == 32)
                {
                    mask = 0xffffffffu;
                    pos = 0;
                }
                else
                {
                    mask = 0;
                    d1 = orig | ((orig & 0x55555555u));
                    d2 = d1 | ((d1 & 0x33333333u) << 2);
                    d3 = d2 | ((d2 & 0x0f0f0f0fu) << 4);
                    d4 = d3 | ((d3 & 0x00ff00ffu) << 8);
                    if (log2_width < 4)
                        d4 = 0;
                    if (log2_width < 3)
                        d3 = 0;
                    if (log2_width < 2)
                        d2 = 0;
                    mask |= (1u << CountTrailingOnes(d4)) - 1;
                    mask |= (1u << CountTrailingOnes(d3 | mask)) - 1;
                    mask |= (1u << CountTrailingOnes(d2 | mask)) - 1;
                    mask |= (1u << CountTrailingOnes(d1 | mask)) - 1;
                    pos = CountTrailingOnes(orig | mask);
                    mask = (unsigned)(((1ull << width) - 1) << pos);

                    if ((pos > 32 - (int)width) | ((orig & mask) > 0))
                    { // Try the next bin...
                    next:
                        i = (i + 1) % zones;
                        if (--max_chunks <= 0)
                        {
                            return alloc_fail();
                        }
                        newOrig = reserved[i];
                        goto try_again;
                    }
                }
            } while ((newOrig = _atomicCAS(&reserved[i], orig, orig | mask)) !=
                orig);
            // Success!
            Header *hdr = (Header *)&data[(pos + (i << depth)) * zone_size];
            hdr->size = (int)size;   // Warning: conversion to 32-bit integer
            last_index[rid] = i + 1; // Keep track of the last position
            return hdr + 1;
        }

		__HOST_V__ __device__ void free(void *ptr)
        {
            Header *hdr = (Header *)ptr - 1;
            if (!is_internal_validpointer(hdr))
            {
                Alternative::free(ptr);
                return;
            }
            int block_size = NextPow2(hdr->size + sizeof(Header));
            unsigned int width = (block_size + zone_size - 1) / zone_size;
            int pos = PointerToPos(hdr);
            int i = pos >> depth;
            pos &= (1 << depth) - 1;
            unsigned int mask = (unsigned)(((1ull << width) - 1) << pos);

            _atomicAnd(&reserved[i], ~mask);
        }

        int __HOST_V__ __device__ get_size(void *ptr)
        {
            Header *hdr = (Header *)ptr - 1;
            if (!is_internal_validpointer(hdr))
                return Alternative::get_size(ptr);
            return hdr->size;
        }
		__HOST_V__ __device__ int get_overhead() { return sizeof(Header); }

        bool __HOST_V__ __device__ is_internal_validpointer(void *ptr)
        {
            return (char *)ptr >= &data[0] && (char *)ptr < &data[total_size];
        }

        bool __HOST_V__ __device__ is_validpointer(void *ptr)
        {
            return is_internal_validpointer(ptr) ||
                   Alternative::is_validpointer(ptr);
        }

        // Calculates the cell position for a given pointer
        INLINE __HOST_V__ __device__ int PointerToPos(const void *ptr)
        {
            return (int)(((char *)ptr - data) / zone_size);
        }

      public:
        // Diagnostic function
        __HOST_V__ __device__ int count_reserved(void)
        {
            int count = Alternative::count_reserved();
            for (int i = 0; i < zones; i++)
            {
                count += CountBits(reserved[i]);
            }
            return count;
        }
		__HOST_V__ __device__ long long total_mem_used()
        {
            long long total = Alternative::total_mem_used();
            int num_reserved_zones = 0;
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
			// Use parallel reduction algorithm
			int tid = __threadidx();
			int nThreads = __threadcnt();
			int val = 0;
			__shared__ int bins[2048];
			for (int i = tid; i < zones; i += nThreads)
				val += CountBits(reserved[i]);
			bins[tid] = val;
			__syncthreads();
			for (int bit = 1; bit < nThreads; bit <<= 1)
			{
				if ((tid & ((bit << 1) - 1)) == 0)
				{
					bins[tid] += bins[tid + bit];
				}
				__syncthreads();
			}
			num_reserved_zones = bins[0];
#else
			for (int i = 0; i < zones; i++)
			{
				num_reserved_zones += CountBits(reserved[i]);
			}
#endif
            return total + (long long)num_reserved_zones * zone_size;
        }
    };

    // ParallelChunkAllocator: allocates objects on a chunk. When
    // all current chunks are full, a new chunk is allocated. This
    // class is locking-free and thread-safe and can be used for
    // creating efficient parallel memory allocation algorithms.
    // The class keeps a single linked list of chunks.
    template <typename Chunk, class Super>
    class ParallelChunkAllocator : public Super
    {
        typedef SingleLinkedList<Chunk, true> ChunkList;
        ChunkList list;

      protected:
        struct Header
        {
            int owner;
            int size;
        };

      public:
		__HOST_V__ __device__ void *allocate(size_t bytes)
        {
            typename ChunkList::Node *chunk = list.get();
            Header *ptr = NULL;

            // Check all available chunks
            // WARNING: branch divergence when used on GPU
            for (;; )
            {
                // Add a new chunk, if necessary
                if (chunk == NULL)
                {
                    // chunk = new (Super::allocate(sizeof(typename
                    // ChunkList::Node))) typename ChunkList::Node();
                    chunk = (typename ChunkList::Node *)Super::allocate(
                        sizeof(typename ChunkList::Node));
                    if (chunk == NULL)
                    {
                        return alloc_fail(); // Running out of memory
                    }
                    int ref_cnt = Super::add_ref(chunk);
                    assert(ref_cnt == 1);
                    list.addfirst(chunk);
                }
                else if (!Super::take_ownership(chunk))
                {
					chunk = NULL; // At this point, the chunk becomes invalid.
                    continue;
                }

                // Attempt to allocate on the current chunk
                ptr = (Header *)chunk->allocate(bytes + sizeof(Header));
                if (ptr != NULL)
				{
					ptr->owner = (int)((char *)ptr - (char *)chunk);
					ptr->size = (int)(bytes + sizeof(Header)); // Warning: conversion to 32-bit int
					assert(ptr->owner >= 0);
					return ptr + 1;
				}

                // Unclaim ownership of the chunk
                release_chunk(chunk);

				// Go to the next chunk
				chunk = (typename ChunkList::Node *)atomicDeref((void *volatile *)&chunk->next);
            }
        }

        __HOST_V__ __device__ void free(void *ptr)
        {
            if (!is_validpointer(ptr))
                return;
            Header *header = (Header *)ptr - 1;
            typename ChunkList::Node *chunk =
                (typename ChunkList::Node *)((char *)header -
                                             (size_t)header->owner);
			assert(header->owner >= 0);
            chunk->free(ptr);
            release_chunk(chunk);
        }

		__HOST_V__ __device__ int get_size(void *ptr)
		{
			Header *header = (Header *)ptr - 1;
			return header->size - sizeof(Header);
		}

		__HOST_V__ __device__ int get_overhead()
		{
			return Super::get_overhead() + sizeof(Header);
		}

		__HOST_V__ __device__ void release_chunk(typename ChunkList::Node *chunk)
        {
            if (Super::release_ref(chunk) == 0)
            {
                list.tryremove(chunk);
                Super::free(chunk);
            }
        }

        bool __HOST_V__ __device__ is_validpointer(void *ptr)
        {
            return Super::is_validpointer(ptr);
        }

        // Diagnostic function
        __HOST_V__ __device__ int count_reserved(void) { return Super::count_reserved(); }
        __HOST_V__ __device__ int count_reserved_small(void)
        {
            int count = 0;
            for (typename ChunkList::Node *n = list.get(); n != NULL;
                 n = n->next)
                count += n->count_reserved();
            return count;
        }
		__HOST_V__ __device__ long long total_mem_used()
        {
            return Super::total_mem_used();
        }
    };


    // An allocator that adds a reference count to the allocated objects
    template <bool thread_safe, class Super>
    class ObjectAllocator : public Super
    {
      public:
        struct Header
        {
            int ref_count;
        };

		__HOST_V__ __device__ void *allocate(size_t bytes)
        {
            Header *obj = (Header *)Super::allocate(bytes + sizeof(Header));
            if (obj == NULL)
                return alloc_fail();
            obj->ref_count = 0;
            return obj + 1;
        }

		__HOST_V__ __device__ void *allocate_addref(size_t bytes)
        {
            Header *obj = (Header *)Super::allocate(bytes + sizeof(Header));
            if (obj == NULL)
                return alloc_fail();
            obj->ref_count = 1;
            return obj + 1;
        }

		__HOST_V__ __device__ void free(void *ptr)
        {
            Header *obj = (Header *)ptr - 1;
            Super::free(obj);
        }
		__HOST_V__ __device__ int ref_count(void *ptr)
        {
            Header *obj = (Header *)ptr - 1;
            return obj->ref_count;
        }
		__HOST_V__ __device__ int add_ref(void *ptr)
        {
            int cnt;
            Header *obj = (Header *)ptr - 1;
            if (thread_safe)
                cnt = _atomicAdd((int *)&obj->ref_count, 1);
            else
                cnt = ++obj->ref_count;
            return cnt;
        }
		__HOST_V__ __device__ int get_size(void *ptr)
        {
            Header *obj = (Header *)ptr - 1;
            return Super::get_size(obj) - sizeof(Header);
        }
		__HOST_V__ __device__ int get_overhead() { return Super::get_overhead() + sizeof(Header); }
		__HOST_V__ __device__ int release_ref(void *ptr)
        {
            int cnt;
            Header *obj = (Header *)ptr - 1;
            if (thread_safe)
                cnt = _atomicAdd((int *)&obj->ref_count, -1);
            else
                cnt = --obj->ref_count;
            return cnt;
        }
        // Attempts to atomically increase the reference count; if it was
        // already
        // zero, return false.
		__HOST_V__ __device__ bool take_ownership(void *ptr)
        {
			if (thread_safe)
			{
				Header *obj = (Header *)ptr - 1;
				int val;
				int original = obj->ref_count;
				if (!original)
					return false; // Object released

				do
				{
					val = original;
					original = _atomicCAS((unsigned int *)&obj->ref_count, val, val + 1);
					if (original != val && original == 0) // Someone else disposed the object
						return false;
				} while (original != val);
			}
			else
			{
				Header *obj = (Header *)ptr - 1;
				if (!obj->ref_count) return false;
				obj->ref_count++;
			}
            return true;
        }
        bool __HOST_V__ __device__ is_validpointer(void *ptr)
        {
            return Super::is_validpointer(ptr);
        }
		__HOST_V__ __device__ long long total_mem_used()
        {
            return Super::total_mem_used();
        }
    };

    // Allocator uses a parallel zone allocator for small objects and
    // the specified (Super) allocator for large objects.
    // threshold: the threshold to be used for the small object heap
    // min_block_size: the minimal block size of the super-allocator
    // (which is typically also a parallel zone allocator).
    // Due to the divergent branches the allocator should not be used
    // on a GPU.
    template <int threshold, int min_block_size, class Super, class Alternative>
    class ParallelHybridAllocator
        : public ObjectAllocator<
              true,
              ParallelChunkAllocator<
                  ParallelZoneAllocator<threshold / 32,
                                        (min_block_size * 31 / threshold) / 16,
                                        Alternative>,
                  Super> >
    {
      public:
        typedef ObjectAllocator<
            true,
            ParallelChunkAllocator<
                ParallelZoneAllocator<threshold / 32,
                                      (min_block_size * 31 / threshold) / 16,
                                      Alternative>,
                Super> > Base;
        enum
        {
            m_thres = threshold,
            sub_zone_size = threshold / 32,
            sub_zones = (min_block_size * 31 / threshold) / 16
        };
		struct Header
		{
			int size;
		};
		__HOST_V__ __device__ void *allocate(size_t bytes)
        {
            Header *ret;

			int o1 = Base::get_overhead();
			int o2 = Super::get_overhead();

            // Threshold exceeded - let the super allocator perform the allocation
            // WARNING: divergent branch if the number of bytes differs from
            // thread to thread
            if (get_selector(bytes) == 0)
                ret = (Header *) Super::allocate(bytes + sizeof(Header));
            // Otherwise, let the small object allocator do the allocation
            else
                ret = (Header *) Base::allocate(bytes  + sizeof(Header));
			if (ret == NULL) return alloc_fail();

			ret->size = bytes; // conversion to 32-bit int
            return ret + 1;
        }
		__HOST_V__ __device__ int get_selector(size_t bytes)
		{
			return bytes + Base::get_overhead() + sizeof(Header) < threshold;
		}
		__HOST_V__ __device__ int get_size(void *data)
        {
			Header* hdr = (Header*) data - 1;
			return hdr->size;
        }
		__HOST_V__ __device__ void free(void *data)
        {
			Header* hdr = (Header*) data - 1;
            // WARNING: divergent branch if the number of bytes differs from
            // thread to thread
			if (get_selector(hdr->size) == 0)
                Super::free(hdr);
            else
                Base::free(hdr);
        }
        bool __HOST_V__ __device__ is_validpointer(void *ptr)
        {
            return Super::is_validpointer(ptr) || Base::is_validpointer(ptr);
        }
		__HOST_V__ __device__ long long total_mem_used()
        {
            return Super::total_mem_used();
        }
    };

    template <int threshold, class SmallHeapAllocator, class LargeHeapAllocator>
    class ParallelHybridAllocator2
    {
        ObjectAllocator<true, SmallHeapAllocator> smallHeap;
        ObjectAllocator<true, LargeHeapAllocator> largeHeap;
        enum
        {
            overhead = 16,
            mod_thresh = threshold - overhead
        };
		__HOST_V__ __device__ void *allocate(size_t bytes)
        {
            void *ret;
            if (bytes >= mod_thresh)
                ret = largeHeap.allocate(bytes);
            else
                ret = smallHeap.allocate(bytes);
            return ret;
        }
		__HOST_V__ __device__ int get_size(void *data)
        {
            int size = largeHeap.get_size(data);
            if (size == -1)
                size = smallHeap.get_size(data);
            return size;
        }
		__HOST_V__ __device__ void free(void *data)
        {
            int size = get_size(data);
            if (size >= mod_thresh)
                largeHeap.free(data);
            else
                smallHeap.free(data);
        }
        bool __HOST_V__ __device__ is_validpointer(void *ptr)
        {
            return largeHeap.is_validpointer(ptr) ||
                   smallHeap.is_validpointer(ptr);
        }
		__HOST_V__ __device__ long long total_mem_used()
        {
            return largeHeap.total_mem_used() + smallHeap.total_mem_used();
        }
    };

    // The most important part - the specific dynamic memory configuration
    // to be used.
    class DynMemConfiguration
    {
    public:
		enum
		{
		// Total number of megabytes (in MB)
		#ifndef DYNAMIC_KERNEL_MEM_MB
			NUM_MB = 128,
		#else
			NUM_MB = DYNAMIC_KERNEL_MEM_MB,
		#endif

		// The maximal block size that can be allocated (in KB)
		#ifndef DYNAMIC_KERNEL_MAX_BS
			MAX_BS = 128,
		#else
			MAX_BS = DYNAMIC_KERNEL_MAX_BS,
		#endif
		};

        // The CPU allocator provides:
        // - reference counting
        // - hybrid allocation algorithm with a small object heap and a large
        // object heap
        // - large object heap: min_block_size = 2048, max_block_size = 32*2048
        // = 65536 (for MAX_BS=64)
        // - small object heap: min_block_size = 32, max_block_size = 1024
        // Allocator size: 67112964 bytes (32-bit)  or 67112968 bytes (64-bit).
		// (for NUM_MB = 64)
        typedef ParallelHybridAllocator<
            MAX_BS*32, MAX_BS*64,
            ObjectAllocator<true, ParallelZoneAllocator<MAX_BS*32, NUM_MB*1024/MAX_BS,
                                                        NullAllocator > >,
			NullAllocator >
            CPUAllocator;

        // The GPU allocator: a doubly nested parallel zone allocator. The
        // second parameter
        // controls the number of blocks. Total size = PARAM * 1024 * 2.
		typedef ObjectAllocator<
            true, ParallelZoneAllocator<MAX_BS, (NUM_MB*512*32)/MAX_BS,
                                        ParallelZoneAllocator<MAX_BS*32, NUM_MB*512/MAX_BS,
                                                              NullAllocator> > >
            GPUAllocator;

// typedef ObjectAllocator<true, ParallelZoneAllocator<1024, 16 * NUM_MB,
// NullAllocator> > GPUAllocator;

#ifdef TARGET_CUDA
        typedef GPUAllocator DefaultAllocator;
#else
        typedef GPUAllocator DefaultAllocator;
#endif
    };

} // namespace quasar

// Declare the dynamic memory symbol. The advantage of this approach is that it
// allows
// the complete dynamic memory block to be pre-allocated externally.
#ifdef TARGET_CUDA
__device__ quasar::DynMemConfiguration::DefaultAllocator *dynmem;
__device__ size_t dynmem_size =
    sizeof(quasar::DynMemConfiguration::DefaultAllocator);
__device__ int dynmem_maxblocksize =
	quasar::DynMemConfiguration::MAX_BS;
SELECTANY quasar::DynMemConfiguration::DefaultAllocator *dynmem_host;
SELECTANY size_t dynmem_size_host =
sizeof(quasar::DynMemConfiguration::DefaultAllocator);
SELECTANY int dynmem_maxblocksize_host =
quasar::DynMemConfiguration::MAX_BS;
#elif defined(__QUASAR_NO_EXPORTS__)
extern "C" {
	extern quasar::DynMemConfiguration::DefaultAllocator *dynmem;
	extern size_t dynmem_size;
	extern int dynmem_maxblocksize;
}
#elif defined(QUASARHOST_DLLIMPORT)
extern "C" {
	IMPORT quasar::DynMemConfiguration::DefaultAllocator *dynmem;
	IMPORT size_t dynmem_size;
	IMPORT int dynmem_maxblocksize;
}
#else
extern "C" {
	EXPORT SELECTANY quasar::DynMemConfiguration::DefaultAllocator *dynmem = NULL;
	EXPORT SELECTANY size_t dynmem_size =
		sizeof(quasar::DynMemConfiguration::DefaultAllocator);
	EXPORT SELECTANY int dynmem_maxblocksize =
		quasar::DynMemConfiguration::MAX_BS;
}
#endif

#if defined(TARGET_CUDA) && !defined(__CUDA_ARCH__) // host functions
#define get_dynmem() dynmem_host
#else
#define get_dynmem() dynmem
#endif

namespace quasar
{
    // uninit function

    template <typename T>
    INLINE __HOST_V__ __device__ VectorBase<T> uninit(const char *elemT, int dim1)
    {
        return VectorBase<T>((T *)get_dynmem()->allocate_addref(dim1 * sizeof(T)),
                             dim1);
    }

    template <typename T>
    INLINE __HOST_V__ __device__ MatrixBase<T> uninit(const char *elemT, int dim1,
                                           int dim2)
    {
        return MatrixBase<T>(
            (T *)get_dynmem()->allocate_addref(dim1 * dim2 * sizeof(T)), dim1, dim2);
    }

    template <typename T>
    INLINE __HOST_V__ __device__ CubeBase<T> uninit(const char *elemT, int dim1, int dim2,
                                         int dim3)
    {
        return CubeBase<T>(
            (T *)get_dynmem()->allocate_addref(dim1 * dim2 * dim3 * sizeof(T)), dim1,
            dim2, dim3);
    }

    // zeros function

    template <typename T>
    INLINE __HOST_V__ __device__ VectorBase<T> zeros(const char *elemT, int dim1)
    {
        VectorBase<T> v = uninit<T>(elemT, dim1);
        mem_fill<T>((T *)v.data, dim1, T());
        return v;
    }

    template <typename T>
    INLINE __HOST_V__ __device__ MatrixBase<T> zeros(const char *elemT, int dim1, int dim2)
    {
        MatrixBase<T> m = uninit<T>(elemT, dim1, dim2);
        mem_fill<T>((T *)m.data, dim1 * dim2, T());
        return m;
    }

    template <typename T>
    INLINE __HOST_V__ __device__ CubeBase<T> zeros(const char *elemT, int dim1, int dim2,
                                        int dim3)
    {
        CubeBase<T> c = uninit<T>(elemT, dim1, dim2, dim3);
        mem_fill<T>((T *)c.data, dim1 * dim2 * dim3, T());
        return c;
    }

    // ones function

    template <typename T>
    INLINE __HOST_V__ __device__ VectorBase<T> ones(const char *elemT, int dim1)
    {
        VectorBase<T> v = uninit<T>(elemT, dim1);
        mem_fill<T>((T *)v.data, dim1, 1);
        return v;
    }

    template <typename T>
    INLINE __HOST_V__ __device__ MatrixBase<T> ones(const char *elemT, int dim1, int dim2)
    {
        MatrixBase<T> m = uninit<T>(elemT, dim1, dim2);
        mem_fill<T>((T *)m.data, dim1 * dim2, 1);
        return m;
    }

    template <typename T>
    INLINE __HOST_V__ __device__ CubeBase<T> ones(const char *elemT, int dim1, int dim2,
                                       int dim3)
    {
        CubeBase<T> c = uninit<T>(elemT, dim1, dim2, dim3);
        mem_fill<T>((T *)c.data, dim1 * dim2 * dim3, 1);
        return c;
    }

// Adding references/releasing memory

#define IMPL_DYNMEM_FUNC(class_name)												\
    template <typename T> INLINE __HOST_V__ __device__ void class_name::addref()    \
    {																				\
        if (get_dynmem()->is_validpointer(data))									\
            get_dynmem()->add_ref(data);											\
    }																				\
    template <typename T> INLINE __HOST_V__ __device__ void class_name::release()   \
    {																				\
        if (get_dynmem()->is_validpointer(data) && get_dynmem()->release_ref(data) == 0)   \
        {																			\
            for (int i = 0; i < get_numel(); i++)									\
                data[i].release();													\
            get_dynmem()->free(data);												\
        }																			\
    }

    template <typename T> INLINE __HOST_V__ __device__ void VectorBase<T>::addref()
    {
        if (get_dynmem()->is_validpointer(data))
            get_dynmem()->add_ref(data);
    }
    template <typename T> INLINE __HOST_V__ __device__ void MatrixBase<T>::addref()
    {
        if (get_dynmem()->is_validpointer(data))
            get_dynmem()->add_ref(data);
    }
    template <typename T> INLINE __HOST_V__ __device__ void CubeBase<T>::addref()
    {
        if (get_dynmem()->is_validpointer(data))
            get_dynmem()->add_ref(data);
    }

    template <typename T> INLINE __HOST_V__ __device__ void VectorBase<T>::release()
    {
        if (get_dynmem()->is_validpointer(data) && get_dynmem()->release_ref(data) == 0)
            get_dynmem()->free(data);
    }
    template <typename T> INLINE __HOST_V__ __device__ void MatrixBase<T>::release()
    {
        if (get_dynmem()->is_validpointer(data) && get_dynmem()->release_ref(data) == 0)
            get_dynmem()->free(data);
    }
    template <typename T> INLINE __HOST_V__ __device__ void CubeBase<T>::release()
    {
        if (get_dynmem()->is_validpointer(data) && get_dynmem()->release_ref(data) == 0)
            get_dynmem()->free(data);
    }

    IMPL_DYNMEM_FUNC(VectorBase<VectorBase<T> >)
    IMPL_DYNMEM_FUNC(VectorBase<MatrixBase<T> >)
    IMPL_DYNMEM_FUNC(VectorBase<CubeBase<T> >)
    IMPL_DYNMEM_FUNC(MatrixBase<VectorBase<T> >)
    IMPL_DYNMEM_FUNC(MatrixBase<MatrixBase<T> >)
    IMPL_DYNMEM_FUNC(MatrixBase<CubeBase<T> >)
    IMPL_DYNMEM_FUNC(CubeBase<VectorBase<T> >)
    IMPL_DYNMEM_FUNC(CubeBase<MatrixBase<T> >)
    IMPL_DYNMEM_FUNC(CubeBase<CubeBase<T> >)

#undef IMPL_DYNMEM_FUNC

    // Smart object pointers.
    template <typename T> class object_ptr
    {
      protected:
        T *_ptr;

      public:
        INLINE __HOST_V__ __device__ object_ptr() : _ptr(NULL) {}
        INLINE __HOST_V__ __device__ object_ptr(T *ptr) : _ptr(ptr) {}
		INLINE __HOST_V__ __device__ object_ptr(long int nullPtr) : _ptr(NULL) {}
		INLINE __HOST_V__ __device__ object_ptr(int nullPtr) : _ptr(NULL) {}
		INLINE __HOST_V__ __device__ object_ptr(lref<T>* val) : _ptr(&(T &)*val) {}
		INLINE __HOST_V__ __device__ object_ptr(lref<T> val) : _ptr(&(T &)val) {}
        INLINE __HOST_V__ __device__ operator T *&() { return (T *&)_ptr; }
        INLINE __HOST_V__ __device__ T *operator->() { return (T *)_ptr; }
        INLINE __HOST_V__ __device__ operator T &() { return *_ptr; }
        INLINE __HOST_V__ __device__ void release()
        {
            if (_ptr && get_dynmem()->is_validpointer(_ptr) &&
                get_dynmem()->release_ref(_ptr) == 0)
            {
                _ptr->~T();
                get_dynmem()->free(_ptr);
            }
        }
        INLINE __HOST_V__ __device__ void addref()
        {
            if (_ptr && get_dynmem()->is_validpointer(_ptr))
                get_dynmem()->add_ref(_ptr);
        }
    };

    // Generic new operator (to initialize with a given value)
    template <typename R> __HOST_V__ __device__ object_ptr<R> qnew(R value)
    {
        object_ptr<R> ptr((R *)get_dynmem()->allocate_addref(sizeof(R)));
        *ptr = value;
        return ptr;
    }

    // Atomic operations on pointer types

    // NOTE: atomic operations on matrix references only work right now
    // when the matrices all have the same size. The workaround is to wrap
    // the matrices in objects() which can be atomically swapped.
    // NOTE: these functions should not be called directly from user code
    template <typename R>
    INLINE __HOST_V__ __device__ VectorBase<R> atomicSet(VectorBase<R> *tgt,
													   VectorBase<R> src)
    {
        void *current, *old = (void *)tgt->data;
        src.addref();
		src.addref();
        while ((current = atomicCASP((void *volatile *)&tgt->data, old,
                                     (void *)src.data)) != old)
            old = current;
        VectorBase<R>((R *)current, tgt->dim1).release();
        tgt->dim1 = src.dim1; // NOTE - is not safe!
        return src;
    }

    // Setting a pointer...
    template <typename R>
    INLINE __HOST_V__ __device__ MatrixBase<R> atomicSet(MatrixBase<R> *tgt,
													   MatrixBase<R> src)
    {
        void *current, *old = (void *)tgt->data;
        src.addref();
		src.addref();
        while ((current = atomicCASP((void *volatile *)&tgt->data, old,
                                     (void *)src.data)) != old)
            old = current;
        MatrixBase<R>((R *)current, tgt->dim1, tgt->dim2).release();
        tgt->dim1 = src.dim1; // NOTE - is not safe!
        tgt->dim2 = src.dim2; // NOTE - is not safe!
        return src;
    }

    template <typename R>
    INLINE __HOST_V__ __device__ CubeBase<R> atomicSet(CubeBase<R> *tgt, CubeBase<R> src)
    {
        void *current, *old = (void *)tgt->data;
        src.addref();
		src.addref();
        while ((current = atomicCASP((void *volatile *)&tgt->data, old,
                                     (void *)src.data)) != old)
            old = current;
        CubeBase<R>((R *)current, tgt->dim1, tgt->dim2, tgt->dim3).release();
        tgt->dim1 = src.dim1; // NOTE - is not safe!
        tgt->dim2 = src.dim2; // NOTE - is not safe!
        tgt->dim3 = src.dim3; // NOTE - is not safe!
        return src;
    }

    template <typename R>
    INLINE __HOST_V__ __device__ object_ptr<R> atomicSet(object_ptr<R> *tgt,
													   object_ptr<R> src)
    {
        void *current, *old = (R *)*tgt;
        src.addref();
		src.addref();
        while ((current = atomicCASP((void *volatile *)(R **)tgt, old,
                                     (void *)(R *)src)) != old)
            old = current;
        object_ptr<R>((R *)current).release();
        return src;
    }

	// Forward declarations for reshape function
	template <typename T>
	__device__ __HOST_V__ VectorBase<T> reshape(VectorBase<T>, int, const char *);
	template <typename T>
	__device__ __HOST_V__ MatrixBase<T> reshape(VectorBase<T>, int2, const char *);
	template <typename T>
	__device__ __HOST_V__ CubeBase<T> reshape(VectorBase<T>, int3, const char *);
	template <typename T>
	__device__ __HOST_V__ VectorBase<T> reshape(MatrixBase<T>, int, const char *);
	template <typename T>
	__device__ __HOST_V__ MatrixBase<T> reshape(MatrixBase<T>, int2, const char *);
	template <typename T>
	__device__ __HOST_V__ CubeBase<T> reshape(MatrixBase<T>, int3, const char *);
	template <typename T>
	__device__ __HOST_V__ VectorBase<T> reshape(CubeBase<T>, int, const char *);
	template <typename T>
	__device__ __HOST_V__ MatrixBase<T> reshape(CubeBase<T>, int2, const char *);
	template <typename T>
	__device__ __HOST_V__ CubeBase<T> reshape(CubeBase<T>, int3, const char *);

	// Automatic casting operations from vecX types to vec, involving
	// dynamic memory allocation
#define AUTOCAST_IMPL(R, T, Rstr)											    \
	INLINE VectorBase<R> __device__ __HOST_V__ autocast(T##1 a, ReturnType<VectorBase<R> >) { \
        VectorBase<R> b = uninit<R>(Rstr, 1);                                   \
        b.data[0] = a.x;                                                        \
        return b;                                                               \
    }	                                                                        \
    INLINE VectorBase<R> __device__ __HOST_V__ autocast(T##2 a, ReturnType<VectorBase<R> >) { \
        VectorBase<R> b = uninit<R>(Rstr, 2);                                   \
        b.data[0] = a.x; b.data[1] = a.y;                                       \
        return b;                                                               \
    }	                                                                        \
    INLINE VectorBase<R> __device__ __HOST_V__ autocast(T##3 a, ReturnType<VectorBase<R> >) { \
        VectorBase<R> b = uninit<R>(Rstr, 3);                                   \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z;                      \
        return b;                                                               \
    }	                                                                        \
    INLINE VectorBase<R> __device__ __HOST_V__ autocast(T##4 a, ReturnType<VectorBase<R> >) { \
        VectorBase<R> b = uninit<R>(Rstr, 4);                                   \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z; b.data[3] = a.w;     \
        return b;                                                               \
    }	                                                                        \
    template <int N>															\
	INLINE VectorBase<R> __device__ __HOST_V__ autocast(T##N<N> a, ReturnType<VectorBase<R> >)              \
	{                                                                           \
        VectorBase<R> b = uninit<R>(Rstr, N);                                   \
        for (int i = 0; i < N; i ++)                                            \
            b.data[i] = a.el[i];                                                \
        return b;                                                               \
    }	                                                                        \
	INLINE MatrixBase<R> __device__ __HOST_V__ autocast(T##1 a, int dim1, int dim2) { \
        MatrixBase<R> b = uninit<R>(Rstr, dim1, dim2);                          \
        b.data[0] = a.x;                                                        \
        return b;                                                               \
    }	                                                                        \
    INLINE MatrixBase<R> __device__ __HOST_V__ autocast(T##2 a, int dim1, int dim2) { \
        MatrixBase<R> b = uninit<R>(Rstr, dim1, dim2);                          \
        b.data[0] = a.x; b.data[1] = a.y;                                       \
        return b;                                                               \
    }	                                                                        \
    INLINE MatrixBase<R> __device__ __HOST_V__ autocast(T##3 a, int dim1, int dim2) { \
        MatrixBase<R> b = uninit<R>(Rstr, dim1, dim2);                          \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z;                      \
        return b;                                                               \
    }	                                                                        \
    INLINE MatrixBase<R> __device__ __HOST_V__ autocast(T##4 a, int dim1, int dim2) { \
        MatrixBase<R> b = uninit<R>(Rstr, dim1, dim2);                          \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z; b.data[3] = a.w;     \
        return b;                                                               \
    }	                                                                        \
    template <int N>															\
	INLINE MatrixBase<R> __device__ __HOST_V__ autocast(T##N<N> a, int dim1, int dim2) \
	{                                                                           \
        MatrixBase<R> b = uninit<R>(Rstr, dim1, dim2);                          \
        for (int i = 0; i < N; i ++)                                            \
            b.data[i] = a.el[i];                                                \
        return b;                                                               \
    }                                                                           \
	INLINE CubeBase<R> __device__ __HOST_V__ autocast(T##1 a, int dim1, int dim2, int dim3) { \
        CubeBase<R> b = uninit<R>(Rstr, dim1, dim2, dim3);                      \
        b.data[0] = a.x;                                                        \
        return b;                                                               \
    }	                                                                        \
    INLINE CubeBase<R> __device__ __HOST_V__ autocast(T##2 a, int dim1, int dim2, int dim3) { \
        CubeBase<R> b = uninit<R>(Rstr, dim1, dim2, dim3);                      \
        b.data[0] = a.x; b.data[1] = a.y;                                       \
        return b;                                                               \
    }	                                                                        \
    INLINE CubeBase<R> __device__ __HOST_V__ autocast(T##3 a, int dim1, int dim2, int dim3) { \
        CubeBase<R> b = uninit<R>(Rstr, dim1, dim2, dim3);                      \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z;                      \
        return b;                                                               \
    }	                                                                        \
    INLINE CubeBase<R> __device__ __HOST_V__ autocast(T##4 a, int dim1, int dim2, int dim3) { \
        CubeBase<R> b = uninit<R>(Rstr, dim1, dim2, dim3);                      \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z; b.data[3] = a.w;     \
        return b;                                                               \
    }	                                                                        \
    template <int N>															\
	INLINE CubeBase<R> __device__ __HOST_V__ autocast(T##N<N> a, int dim1, int dim2, int dim3) \
	{                                                                           \
        CubeBase<R> b = uninit<R>(Rstr, dim1, dim2, dim3);                      \
        for (int i = 0; i < N; i ++)                                            \
            b.data[i] = a.el[i];                                                \
        return b;                                                               \
    }                                                                           \
    INLINE MatrixBase<R> __device__ __HOST_V__ autocast(VectorBase<T> a, int dim1, int dim2) { \
        return reshape(a, make_int2(dim1,dim2), Rstr);                          \
    }                                                                           \
    INLINE CubeBase<R> __device__ __HOST_V__ autocast(VectorBase<T> a, int dim1, int dim2, int dim3) { \
        return reshape(a, make_int3(dim1,dim2,dim3), Rstr);                     \
    }																			\
	INLINE MatrixBase<R> __device__ __HOST_V__ autocast(VectorBase<T> a, ReturnType<MatrixBase<T> >) { \
        return reshape(a, make_int2(1,a.dim1), Rstr);                           \
    }                                                                           \
    INLINE CubeBase<R> __device__ __HOST_V__ autocast(VectorBase<T> a, ReturnType<CubeBase<T> >) {	\
        return reshape(a, make_int3(1,a.dim1, 1), Rstr);						\
    }																			\
	INLINE CubeBase<R> __device__ __HOST_V__ autocast(MatrixBase<T> a, ReturnType<CubeBase<T> >) {	\
        return reshape(a, make_int3(a.dim1, a.dim2, 1), Rstr);					\
    }																			\
	INLINE VectorBase<R> __device__ __HOST_V__ autocast(MatrixBase<T> a, ReturnType<VectorBase<T> >) {	\
        return reshape(a, a.dim1*a.dim2, Rstr);									\
    }																			\
	INLINE VectorBase<R> __device__ __HOST_V__ autocast(CubeBase<T> a, ReturnType<VectorBase<T> >) {	\
        return reshape(a, a.dim1*a.dim2*a.dim3, Rstr);							\
    }

	AUTOCAST_IMPL(scalar, scalar, "scalar");
	AUTOCAST_IMPL(cscalar, cscalar, "cscalar");
	AUTOCAST_IMPL(int, int, "int");
#undef AUTOCAST_IMPL

//=====================================
// Memory management utility functions
//=====================================
#if !defined(__QUASAR_NO_EXPORTS__) && !defined(HELIOS)
#if defined(TARGET_CUDA)
    extern "C" __global__ void dynmem_scrub(void **list, int count)
    {
        int pos = blockIdx.y * blockDim.y + threadIdx.y;
        if (pos < count)
            get_dynmem()->free(list[pos]);
    }
#else
    extern "C" EXPORT void dynmem_scrub(void **list, int count)
    {
        for (int i = 0; i < count; i++)
            get_dynmem()->free(list[i]);
    }
#endif

#ifdef TARGET_CUDA
    extern "C" __global__ void dynmem_totalused(long long *size)
    {
        // Will use a parallel algorithm!
        *size = get_dynmem()->total_mem_used();
    }
#else
    extern "C" EXPORT void dynmem_totalused(long long *size)
    {
        *size = get_dynmem()->total_mem_used();
    }
#endif
#endif // __QUASAR_NO_EXPORTS__

} // namespace quasar

#endif
