//
// quasar_host.cpp : Defines the bridge with the Quasar run-time
// environment and compilation system.
// (C) Gepura 2014-2020
//
#ifdef _MSC_VER
// See https://developercommunity.visualstudio.com/content/problem/93889/error-c2872-byte-ambiguous-symbol.html
// due to /std:c++17
#define _HAS_STD_BYTE 0
#endif

#ifndef __QUASAR_NO_EXPORTS__
#define __QUASAR_NO_EXPORTS__
#endif

#ifndef ENABLE_DYNAMIC_KERNEL_MEM
#define ENABLE_DYNAMIC_KERNEL_MEM
#endif

#include <stdarg.h>
#include "quasar.h"
#include "quasar_host.h"
#include "quasar_ncube.h"
#include "quasar_rt_bindings.h"
#include "quasar_type_impl.h"

#if defined(_WIN32) || defined(_WIN64)
#define WIN32_LEAN_AND_MEAN
#include <crtdbg.h>
#include <Windows.h>
#include <metahost.h>
#include <shellapi.h>
#define MS_DOTNET
#endif

#ifndef MS_DOTNET // Use the mono runtime as default
#define MONO_RUNTIME
#include <mono/jit/jit.h>
#include <mono/metadata/assembly.h>
#include <mono/metadata/mono-config.h>
#include <mono/metadata/mono-debug.h>
#include <mono/metadata/debug-helpers.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h> /* For mode constants */
#include <fcntl.h> /* For O_* constants */
#include <dlfcn.h>
// Note: due to naming conflict of macro with quasar namespace
#undef LOCK_WRITE
#undef LOCK_READ
#endif

// MSVC only
#ifdef _MSC_VER
#undef Assert
#pragma comment(lib, "mscoree.lib")
#pragma comment(lib, "shell32.lib")

// Import mscorlib.tlb (Microsoft Common Language Runtime Class Library).
#import "mscorlib.tlb" raw_interfaces_only high_property_prefixes("_get", "_put", "_putref") rename("ReportEvent", "InteropServices_ReportEvent") rename("or", "Or")
using namespace mscorlib;

#define Assert(EX,...) \
	  (void)((EX) || (dbgAssert (#EX, __FILE__, __LINE__, __VA_ARGS__),0))
#endif

#define RELEASE_SAFE(x)                                                        \
    if (x)                                                                     \
    {                                                                          \
        (x)->Release();                                                        \
        x = NULL;                                                              \
    }

#define DELETE_HANDLE(context, handle)                                         \
    if (handle)                                                                \
    {                                                                          \
        (context)->Delete(handle);                                             \
        handle = NULL;                                                         \
    }

// Symbols used in quasar.h, quasar_dynmem.h
extern "C"
{
	SELECTANY scalar *	shared_mem;
	SELECTANY int			shared_mem_dwords;
	SELECTANY quasar::DynMemConfiguration::DefaultAllocator *dynmem;
	SELECTANY size_t		dynmem_size;
	SELECTANY int			dynmem_maxblocksize;
}

//======================================================
// Begin interface
//======================================================

namespace quasar
{
	static LPCTSTR typeNames[TYPE_COUNT] = {
		_T("??"),          _T("scalar"), _T("cscalar"), _T("int"),   _T("vec"),    _T("mat"),
		_T("cube"),        _T("cvec"),   _T("cmat"),    _T("ccube"), _T("string"), _T("type"),
		_T("lambda_expr"), _T("int8"),   _T("int16"),   _T("int64"), _T("uint8"),  _T("uint16"),
		_T("uint32"),      _T("uint64"), NULL,       _T("object"),_T("cube{n}"),_T("ccube{n}"),
		NULL
	};

	struct QuasarContext;
	class ObjBaseImpl;
	class ComputationEngineImpl;
	class EvaluationStackImpl;
	class MatrixFactoryImpl;
	class TypeEnvironmentImpl;
	class RuntimeReductionEngineImpl;

	qvalue_t qvalue_t::empty;

	// Helper functions
	static void BeginReadCmdLineParams(LPCTSTR *&args, int &nArgs);
	static void EndReadCmdLineParams(LPCTSTR *args, int nArgs);

	/*------------------------------------------------------------------------
	 * QuasarContext struct
	 *
	 * The following struct defines the ABI with Quasar, and should not be
	 *  used from user code
	 *------------------------------------------------------------------------*/
	struct QuasarContext
	{
		TCHAR deviceName[512/sizeof(TCHAR)];		// Name of the computation device
		ScalarTypes scalarType;		//
		TCHAR errorMsg[3568/sizeof(TCHAR)];
		long long hostPtr;
		long long structSize;
		TCHAR runtimeName[512/sizeof(TCHAR)];

		bool(*FunctionExists)(LPCTSTR functionName);
		bool(*FunctionCall)(LPCTSTR functionName, LPCVOID argsIn, int nArgsIn,
			LPVOID argsOut, int nArgsOut);
		bool(*FunctionCallIndirect)(qhandle lambda, LPCVOID argsIn,
			int nArgsIn, LPVOID argsOut, int nArgsOut);
		bool(*Close)();
		qhandle(*ParseType)(LPCTSTR typeName);
		bool(*Delete)(qhandle handle);
		qhandle(*CreateNDMatrix)(qhandle elemType, const int *dims, int num_dims);
		qhandle(*CreateString)(LPCTSTR text, int length);
		qhandle(*CreateObject)(qhandle objType);
		qhandle(*CreateLambda)(qhandle functionTypeHandle, void* closurePtr,
			LambdaDelegate* lambda);
		void(*AddRef)(qhandle handle);
		bool(*Release)(qhandle handle);
		LockResults(*Lock)(qhandle handle, LockingModes lockingMode,
			MemResources memResource, void **pData);
		void(*Unlock)(qhandle handle, LockingModes lockingMode,
			MemResources memResource);
		void(*RunApp)();
		void(*DoEvents)();
		bool(*ReadVariable)(LPCTSTR varname, qvalue_t *value);
		bool(*WriteVariable)(LPCTSTR varname, const qvalue_t *value);
		bool(*GetField)(qhandle obj, LPCTSTR fieldName, qvalue_t *value);
		bool(*SetField)(qhandle obj, LPCTSTR fieldName, const qvalue_t *value);
		qhandle(*CreateType)(LPCTSTR moduleName, LPCTSTR typeName, CreateTypeFlags typeFlags);
		bool(*Type_AddField)(qhandle typeHandle, LPCTSTR fieldName,
			qhandle fieldType);
		qhandle(*Type_AddParameter)(qhandle typeHandle, qhandle paramTypeHandle);
		bool(*Type_Finalize)(qhandle typeHandle);
		bool(*LoadSourceModule)(LPCTSTR moduleName, LPCTSTR *errorMsg,
			LPCVOID reserved);
		bool(*LoadBinaryModule)(LPCTSTR moduleName, LPCTSTR *errorMsg,
			LPCVOID reserved);
		bool(*LoadModuleFromSource)(LPCTSTR moduleName, LPCTSTR sourceString,
			LPCTSTR *errorMsg, LPCVOID reserved);
		qhandle(*LookupFunction)(LPCTSTR functionSignature);
		qhandle(*GetType)(qhandle handle, DataTypes type);
		qhandle(*LookupMethod)(qhandle typeHandle, LPCTSTR functionSignature);
		bool(*MethodCall)(qhandle methodHandle, qhandle objHandle,
			DataTypes objType, LPCVOID argsIn, int nArgsIn,
			LPVOID argsOut, int nArgsOut);
		bool(*EnableProfiling)(ProfilingModes profilingMode,
			LPCTSTR outputFileName);
		qhandle(*GetComputationEngine)(LPCVOID reserved);
		bool(*ComputationEngine_GetName)(qhandle engine, LPTSTR name,
			int maxChars);
		qhandle(*ComputationEngine_GetEvaluationStack)(qhandle engine,
			LPCVOID reserved);
		qhandle(*ComputationEngine_GetMatrixFactory)(qhandle engine);
		qhandle(*ComputationEngine_GetTypeEnvironment)(qhandle engine);
		qhandle(*ComputationEngine_GetRuntimeReductionEngine)(qhandle engine);
		bool(*ComputationEngine_Process)(qhandle engine, OperatorTypes op);
		bool(*ComputationEngine_ConstructMatrix)(qhandle engine,
			int numRowsOrCols,
			int unusedFlags);
		bool(*ComputationEngine_ConstructCellMatrix)(qhandle engine,
			int length,
			int unusedFlags);
		bool(*ComputationEngine_FunctionCall)(qhandle engine,
			LPCTSTR functionName,
			int numArgs, int unusedFlags);
		bool(*ComputationEngine_FunctionExists)(qhandle engine,
			LPCTSTR functionName,
			int numArgs, int unusedFlags);
		bool(*ComputationEngine_ArrayGetAt)(qhandle engine, const qvalue_t *x,
			int numIndices,
			BoundaryAccessMode boundsMode);
		bool(*ComputationEngine_ArraySetAt)(qhandle engine, qvalue_t *x,
			int numIndices,
			BoundaryAccessMode boundsMode);
		bool(*ComputationEngine_Synchronize)(qhandle engine, int unusedFlags);
		bool(*Stack_PushMultiple)(qhandle stack, const qvalue_t *x, int cnt);
		bool(*Stack_PopMultiple)(qhandle stack, qvalue_t *x, int cnt);
		bool(*Stack_Clear)(qhandle stack);
		bool(*Stack_PushContext)(qhandle stack);
		bool(*Stack_PopContext)(qhandle stack);
		bool(*Stack_Process)(qhandle stack, OperatorTypes op);
		bool(*Stack_FunctionCall)(qhandle stack, LPCTSTR functionName,
			int numArgs, int unusedFlags);
		bool(*Stack_ArrayGetAt)(qhandle stack, const qvalue_t *x,
			int numIndices, BoundaryAccessMode boundsMode);
		bool(*Stack_ArraySetAt)(qhandle stack, qvalue_t *x, int numIndices,
			BoundaryAccessMode boundsMode);
		bool(*Stack_Evaluate)(qhandle stack, qvalue_t *x);
		bool(*Stack_GetCount)(qhandle stack);
		bool(*MatrixFactory_New)(qhandle factory, qvalue_t *y,
			AllocationFlags flags, LPCVOID elems, int64_t elemBytes, qhandle elemType,
			int nDims, const int *dims);
		ScalarTypes(*TypeEnvironment_GetScalarType)(qhandle typeEnv);
		bool(*RuntimeReductionEngine_Add)(qhandle rre, LPCTSTR searchExpr,
			qhandle action,
			qhandle whereClause,
			int unusedFlags);
		bool(*RuntimeReductionEngine_Remove)(qhandle rre, LPCTSTR searchExpr,
			int unusedFlags);
		bool(*QueryProperty)(HostProperties prop, int param,
			void** dst, size_t nBytes);
		bool(*UnloadModule)(LPCTSTR moduleName);
		bool(*ComputationEngine_ParallelDo)(qhandle engine, const qvalue_t *dims, const qvalue_t *function, int numArgs, qvalue_t *retval);
		bool(*ComputationEngine_SerialDo)(qhandle engine, const qvalue_t *dims, const qvalue_t *function, int numArgs, qvalue_t *retval);
		bool(*GetNDims)(qhandle matrix, int* dims, int* nDimsWritten);
		bool(*GetLastError)(LPTSTR *source, int *sourceLen, LPTSTR *errorMsg, int *errorMsgLen, LPTSTR *stackTrace, int *stackTraceLen);
		bool(*GetString)(qhandle strValue, LPTSTR* text, int* length);
		bool(*RegisterModule)(RegisterModuleOptions options, LPCTSTR moduleName, LPCTSTR fileName, RegisterModuleFlags flags);
		bool(*LoadFunction)(qvalue_t *func, qhandle closureArgs, LPCTSTR moduleName, LPCTSTR functionSignature, LPCTSTR funcName,
			FunctionFlags flags, PlatformTargets platformTargets, const LoadFunctionParams *params);
		bool(*Type_GetString)(qhandle typeValue, LPTSTR* text, int* length);
		bool(*Type_GetInfo)(qhandle typeValue, DataTypes *dataType, int *maxDims, int *vectorLength, AccessModifiers *accessModifier,
			bool *isIntegerType, bool *isPointer);
		bool(*Type_IsStatic)(qhandle typeValue, bool *result);
		bool(*Type_GetElementTypeAtPosition)(qhandle type, int dims, int* pos, qhandle *retVal);
		bool(*Type_CompareToSpecification)(qhandle type, qhandle specType, bool *result);
		qhandle(*CreateNamedParam)(LPCTSTR paramName, const qvalue_t *value);
		bool(*MethodCallDirect)(LPCTSTR methodName, qhandle objHandle,
			DataTypes objType, LPCVOID argsIn, int nArgsIn,
			LPVOID argsOut, int nArgsOut);
		bool(*GetMachineInfo)(const MachineInfo *machineInfo);
		bool(*FreeMachineInfo)(MachineInfo *machineInfo);
		bool(*OpenDevice)(LPCTSTR deviceId, const RuntimeSettings *runtimeSettings);
		bool(*CloseDevice)();
		bool(*CreateGeneric)(qhandle typeHandle, int nArgIn, const qvalue_t *inArgs, qvalue_t *outArg);
		bool(*AddLambdaDestroyCallback)(qhandle lambda, DisposeClosure* disposeDelegate);
		bool(*Type_AddFieldWithOffset)(qhandle typeHandle, LPCTSTR fieldName,
			qhandle fieldType, size_t offset);
		bool(*Type_AddProperty)(qhandle typeHandle, LPCTSTR propertyName, qhandle propertyType,
			qhandle getProperty, qhandle setProperty);
		bool(*Type_InheritType)(qhandle objType, qhandle baseType);
		bool(*Type_SetConstructor)(qhandle objType, ConstructorDelegate* constructor);
		bool(*Type_InstantiateType)(qhandle type, const qhandle* instance_type_params, int num_type_params, qhandle *retval);
		bool(*Type_GetElementType)(qhandle type, qhandle *retVal);
		bool(*CreateTypeClass)(const qhandle *elem_types, int elem_count, qhandle *retval);
		bool(*CreatePointerType)(qhandle type, qhandle *retval);
		bool(*CreateFunctionType)(FunctionTypes functionType, const qhandle* inputTypes, int numInputTypes,
			const qhandle* outputTypes, int numOutputTypes, FunctionTypeFlags flags, qhandle *retval);
		bool(*CreateTypeParameter)(LPCTSTR paramName, qhandle paramType, qhandle *retval);
		bool(*RuntimeReductionEngine_Add_v2)(qhandle rre, const VariableDecl* variables, int numVariables,
			const Symbol *searchExprSymbols, int numSearchExprSymbols, qhandle action, qhandle whereClause, int flags);
		bool(*Type_GetClassInfo)(LPCTSTR* moduleName, LPCTSTR* typeName, qhandle typeHandle, ClassInfoFlags *flags,
			MemResources* resources, int *numResources, size_t* size, qhandle *baseClass,
			int *numGenericTypeParams, qhandle *genericTypeParams, int *numInstanceTypeParams, qhandle *instanceTypeParams);
		bool(*EnumRegisteredTypes)(EnumRegisteredTypeDelegate *callback, void* userData);
		bool(*RuntimeReductionEngine_EnumFields)(qhandle type, int numOffsets, const MemResources *memResources,
			EnumFieldWithDetailsDelegate* callback, void* userData);
		bool(*RuntimeReductionEngine_EnumFunctions)(EnumFunctionsDelegate *callback, void* userData);
		bool(*RuntimeReductionEngine_EnumMethods)(qhandle type, EnumMethodsDelegate *callback, void* userData);

		// Advanced functions
		bool(*GetMemResourceForDevice)(MemResources *result, int deviceId);
		bool(*SetupKernelFunction)(qvalue_t *kernel, KernelLambdaDelegate cpu_kernel_adapter,
			KernelLambdaDelegate cuda_kernel_adapter, PrimeKernelLambdaDelegate cuda_prime_kernel_adapter, void *userData);
		bool(*SetupDeviceFunction)(qvalue_t *kernel, KernelLambdaDelegate cpu_device_adapter, void* func_ptr, void *userData);
		bool(*LaunchKernel)(qvalue_t *kernel, const void* params, size_t params_size, const void* closure_ptr, const qvalue_t *args,
			int3 grid_dim, int3 block_dim, int flags, int deviceId, void *stream);
		bool (*IsKernelInitialized)(qvalue_t *kernel);
		bool (*GetDataPtr)(qvalue_t *obj, MemResources memResource, void **ptr);
		bool(*GetLambdaFunctionPtr)(qvalue_t *lambda, int deviceId, void **ptr, void **closure);
		bool(*UnmarshalPointer)(qvalue_t *qv, const void *ptr, const void *data, size_t data_size, DataTypes data_type, int deviceId, int flags);
		bool(*UnmarshalTypedObject)(qvalue_t *qv, const void *ptr, size_t size, LPCTSTR typeName, int deviceId, int flags);
		bool(*RegisterInstance)(void *native_obj_ptr, qhandle type_handle, void**output_handle);
		bool(*RegisterModuleFromData)(RegisterModuleOptions options, LPCTSTR moduleName,
			LPCVOID data, size_t dataSize, RegisterModuleFlags flags);
		bool(*EnumFields)(const qvalue_t *obj, EnumFieldDelegate *callback, void *userData);
		bool(*Type_SetDynamicDispatcher)(qhandle typeHandle, DynamicDispatchers disp, DynamicCallDelegate *call);
		bool(*ParseNamedParam)(const qvalue_t *named_param, LPTSTR *text, int *textLen, qvalue_t *value);
		void(*FreeString)(LPCTSTR str);
		bool(*EnumVariables)(EnumVariableDelegate *callback, void *userData);
		bool(*GetRawSizeOfType)(qhandle typeHandle, MemResources memResource, size_t *size, size_t *alignment);
	};

	/*------------------------------------------------------------------------
	 * ObjBaseImpl class - implements reference counting and is the base
	 * class for all the host interfaces in this module.
	 *------------------------------------------------------------------------*/
	class ObjBaseImpl
	{
	protected:
		int refCount;

	public:
		ObjBaseImpl();
		void AddRef();
		void Release();
	};

	/*------------------------------------------------------------------------
	 * ObjBaseImpl::ObjBaseImpl
	 *------------------------------------------------------------------------*/
	ObjBaseImpl::ObjBaseImpl() { refCount = 0; }

	void ObjBaseImpl::AddRef() { refCount++; }

	void ObjBaseImpl::Release()
	{
		if (--refCount == 0)
			delete this;
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl class: implementation of the IComputationEngine
	 * interface
	 *------------------------------------------------------------------------*/
	class ComputationEngineImpl : public ObjBaseImpl, public IComputationEngine
	{
		QuasarContext *context;
		EvaluationStackImpl *evaluationStack;
		MatrixFactoryImpl *matrixFactory;
		TypeEnvironmentImpl *typeEnvironment;
		RuntimeReductionEngineImpl *rre;
		qhandle handle;

	public:
		ComputationEngineImpl(QuasarContext *context);
		virtual ~ComputationEngineImpl();
		bool GetName(LPTSTR name, int maxChars);
		IEvaluationStack *GetEvaluationStack();
		IMatrixFactory *GetMatrixFactory();
		ITypeEnvironment *GetTypeEnvironment();
		IRuntimeReductionEngine *GetRuntimeReductionEngine();
		void Process(OperatorTypes op);
		void ConstructMatrix(int numRowsOrCols);
		void ConstructCellMatrix(int length);
		bool FunctionCall(LPCTSTR functionName, int numArgs);
		bool FunctionExists(LPCTSTR functionName, int numArgs);
		void ArrayGetAt(const qvalue_t &x, int numIndices,
			BoundaryAccessMode boundsMode);
		void ArraySetAt(qvalue_t &x, int numIndices,
			BoundaryAccessMode boundsMode);
		void Synchronize();
		void ParallelDo(const qvalue_t& dims, const qvalue_t& function, int numArgs, qvalue_t &retval);
		void SerialDo(const qvalue_t& dims, const qvalue_t& function, int numArgs, qvalue_t &retval);
		void AddRef() { ObjBaseImpl::AddRef(); }
		void Release()
		{
			if (--refCount == 0)
				delete this;
		}
		qhandle GetHandle() { return handle; }
	};

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl class: implementation of the IEvaluationStack
	 * interface
	 *------------------------------------------------------------------------*/
	class EvaluationStackImpl : public ObjBaseImpl, public IEvaluationStack
	{
		QuasarContext *context;
		qhandle handle;

	public:
		EvaluationStackImpl(QuasarContext *context,
			ComputationEngineImpl *engine);
		virtual ~EvaluationStackImpl();
		qvalue_t PopValue();
		void PushValue(const qvalue_t &value);
		void PopMultiple(qvalue_t *vals, int count);
		void PushMultiple(const qvalue_t *vals, int count);
		void Clear();
		void PushContext();
		void PopContext();
		int GetCount();
		void Process(OperatorTypes op);
		void FunctionCall(LPCTSTR functionName, int numArgs);
		void ArrayGetAt(const qvalue_t &x, int numIndices,
			BoundaryAccessMode boundsMode);
		void ArraySetAt(qvalue_t &x, int numIndices,
			BoundaryAccessMode boundsMode);
		qvalue_t Evaluate(const qvalue_t &x);
		void AddRef() { ObjBaseImpl::AddRef(); }
		void Release()
		{
			if (--refCount == 0)
				delete this;
		}
	};

	/*------------------------------------------------------------------------
	 * MatrixFactoryImpl class: implementation of the IMatrixFactory
	 * interface
	 *------------------------------------------------------------------------*/
	class MatrixFactoryImpl : public ObjBaseImpl, public IMatrixFactory
	{
		QuasarContext *context;
		qhandle handle;

	public:
		MatrixFactoryImpl(QuasarContext *context,
			ComputationEngineImpl *engine);
		virtual ~MatrixFactoryImpl();
		qvalue_t New(AllocationFlags flags, LPCVOID elems, int64_t elemBytes, qhandle elemType, int nDims,
			const int *dims);
		void AddRef() { ObjBaseImpl::AddRef(); }
		void Release()
		{
			if (--refCount == 0)
				delete this;
		}
	};

	/*------------------------------------------------------------------------
	 * TypeEnvironmentImpl class: implementation of the ITypeEnvironment
	 * interface
	 *------------------------------------------------------------------------*/
	class TypeEnvironmentImpl : public ObjBaseImpl, public ITypeEnvironment
	{
		QuasarContext *context;
		qhandle handle;

	public:
		TypeEnvironmentImpl(QuasarContext *context,
			ComputationEngineImpl *engine);
		virtual ~TypeEnvironmentImpl();
		ScalarTypes GetScalarType();
		void AddRef() { ObjBaseImpl::AddRef(); }
		void Release()
		{
			if (--refCount == 0)
				delete this;
		}
	};

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl class: implementation of the
	 * IRuntimeReductionEngine interface
	 *------------------------------------------------------------------------*/
	class RuntimeReductionEngineImpl : public ObjBaseImpl,
		public IRuntimeReductionEngine
	{
		QuasarContext *context;
		qhandle handle;

	public:
		RuntimeReductionEngineImpl(QuasarContext *context,
			ComputationEngineImpl *engine);
		virtual ~RuntimeReductionEngineImpl();
		bool Add(LPCTSTR searchExpr, qhandle action,
			qhandle whereClause);
		bool Add(const VariableDecl* variables, int numVariables,
			Expression *searchExpr, const qvalue_t *action, const qvalue_t *whereClause);
		void EnumFields(qhandle type, int numOffsets, const MemResources *memResources,
			EnumFieldWithDetailsDelegate *callback, void* userData);
		void EnumFunctions(EnumFunctionsDelegate *callback, void* userData);
		void EnumMethods(qhandle type, EnumMethodsDelegate *callback, void* userData);
		bool Remove(LPCTSTR searchExpr);
		void AddRef() { ObjBaseImpl::AddRef(); }
		void Release()
		{
			if (--refCount == 0)
				delete this;
		}
	};

#ifndef MS_DOTNET
	typedef void ICLRMetaHost;
	typedef void ICLRRuntimeInfo;
	typedef void ICLRRuntimeHost;
#else // MONO_RUNTIME
	typedef void MonoDomain;
#endif

	/*------------------------------------------------------------------------
	 * QuasarHostImpl class: implementation of the IQuasarHost interface
	 *------------------------------------------------------------------------*/
	class QuasarHostImpl : public ObjBaseImpl, public IQuasarHost
	{
	private:
		QuasarContext context;
		ICLRMetaHost *pMetaHost;
		ICLRRuntimeInfo *pRuntimeInfo;
		ICLRRuntimeHost *pClrRuntimeHost;
		MonoDomain *monoDomain;
		ComputationEngineImpl *engine;
		qhandle typeHandles[TYPE_COUNT];
		bool isRefCountable[TYPE_COUNT];
		custom_exceptionhandler_t custom_exceptionhandler;
		static IQuasarHost *instance;
		string_t mainProgramName;
		int cmdLineParamOffset;
		std::vector<std::tuple<void*, HostDestroyCallbackDelegate*>> destroyCallbacks;

	public:
		QuasarHostImpl(const QuasarContext &context, ICLRMetaHost *pMetaHost,
			ICLRRuntimeInfo *pRuntimeInfo,
			ICLRRuntimeHost *pClrRuntimeHost,
			MonoDomain *monoDomain);
		void Initialized();
		void Uninitialize();
		bool FunctionExists(LPCTSTR functionName);
		bool FunctionCall(LPCTSTR functionName, const qvalue_t *inArgs,
			int nInArgs, qvalue_t *outArgs, int nOutArgs);
		bool FunctionCall(const qvalue_t &function, const qvalue_t *inArgs,
			int nInArgs, qvalue_t *outArgs, int nOutArgs);
		qvalue_t CreateVector(qhandle elemType, int numel, bool complex);
		qvalue_t CreateMatrix(qhandle elemType, int dim1, int dim2, bool complex);
		qvalue_t CreateCube(qhandle elemType, int dim1, int dim2, int dim3, bool complex);
		qvalue_t CreateNCube(qhandle elemType, int ndims, const int* dims, bool complex);
		qvalue_t CreateString(LPCTSTR text, size_t length);
		qvalue_t CreateTypedObject(qhandle objType);
		qvalue_t CreateUntypedObject(qhandle objType);
		qvalue_t CreateLambda(qhandle functionTypeHandle, void* closurePtr,
			LambdaDelegate* lambda);
		qvalue_t CreateNamedParam(LPCTSTR paramName, const qvalue_t &value);
		qvalue_t CreateGeneric(qhandle typeHandle, const qvalue_t *inArgs, int nInArgs);
		qvalue_t LookupFunction(LPCTSTR functionSignature);
		qvalue_t LookupFunction(LPCTSTR functionSignature, int &token);
		qvalue_t LookupType(LPCTSTR quasarTypeName);
		qvalue_t LookupMethod(const qvalue_t &type, LPCTSTR functionSignature);
		qvalue_t LookupMethod(const qvalue_t &type, LPCTSTR functionSignature, int &token);
		bool MethodCall(const qvalue_t &function, const qvalue_t &target,
			const qvalue_t *inArgs, int nInArgs, qvalue_t *outArgs,
			int nOutArgs);
		bool MethodCall(LPCTSTR methodName, const qvalue_t &target,
			const qvalue_t *inArgs, int nInArgs, qvalue_t *outArgs,
			int nOutArgs);
		void AddRef(qvalue_t &obj);
		void ReleaseRef(qvalue_t obj);
		void DeleteValue(qvalue_t &obj);
		void GetNDims(const qvalue_t &obj, int *nDimsWritten, int *dims);
		void GetString(const qvalue_t &obj, string_t &str);
		LockResults Lock(const qvalue_t &obj, DataTypes elemType,
			LockingModes lockingMode, MemResources memResource,
			void **pData);
		void Unlock(const qvalue_t &obj, LockingModes lockingMode,
			MemResources memResource);
		qhandle GetPrimitiveTypeHandle(DataTypes type, LPCTSTR typeName);
		virtual ~QuasarHostImpl();
		void RunApp() { context.RunApp(); }
		void DoEvents() { context.DoEvents(); }
		void EnumVariables(EnumVariableDelegate* callback, void *userData);
		bool ReadVariable(LPCTSTR varName, qvalue_t &val);
		bool WriteVariable(LPCTSTR varName, const qvalue_t &val);
		bool GetField(const qvalue_t &obj, LPCTSTR fieldName, qvalue_t &val);
		bool SetField(const qvalue_t &obj, LPCTSTR fieldName,
			const qvalue_t &val);
		bool LoadSourceModule(LPCTSTR moduleName, LPCTSTR *errorMsg);
		bool LoadBinaryModule(LPCTSTR binaryModuleName, LPCTSTR *errorMsg);
		bool LoadModuleFromSource(LPCTSTR moduleName, LPCTSTR sourceString,
			LPCTSTR *errorMsg);
		bool UnloadModule(LPCTSTR moduleName);
		LPCTSTR GetLastErrorMsg() { return context.errorMsg; }
		qvalue_t GetType(const qvalue_t &obj);
		qvalue_t CreateType(LPCTSTR moduleName, LPCTSTR typeName, CreateTypeFlags flags);
		bool AddField(qhandle objType, LPCTSTR fieldName, qhandle fieldType);
		qvalue_t AddParameter(qhandle objType, qhandle typeParam);
		void FinalizeType(qhandle typeHandle);
		bool EnableProfiling(ProfilingModes profilingMode,
			LPCTSTR outputFileName);
		bool QueryProperty(HostProperties prop, int param, void** dst, size_t nBytes);
		IEvaluationStack *CreateStack();
		IComputationEngine *GetComputationEngine();
		void AddRef() { ObjBaseImpl::AddRef(); }
		void Release()
		{
			if (--refCount == 0)
				delete this;
		}
		void RaiseException();
		void SetCustomExceptionHandler(custom_exceptionhandler_t handler);
		custom_exceptionhandler_t GetCustomExceptionHandler();
		bool RegisterModule(RegisterModuleOptions options, LPCTSTR moduleName,
			LPCTSTR fileName, RegisterModuleFlags flags);
		bool RegisterModuleFromData(RegisterModuleOptions options, LPCTSTR moduleName,
			LPCVOID data, size_t dataSize, RegisterModuleFlags flags);
		qvalue_t LoadFunction(qvalue_t closureArgs, LPCTSTR moduleName, LPCTSTR functionSignature, LPCTSTR funcName,
			FunctionFlags flags, PlatformTargets platformTargets, LoadFunctionParams &params);
		string_t TypeGetString(qvalue_t type);
		void TypeGetInfo(qvalue_t type, DataTypes *dataType, int *maxDims, int *vectorLength, AccessModifiers *accessModifier,
			bool *isIntegerType, bool *isPointer);
		bool TypeIsStatic(qvalue_t type);
		qvalue_t TypeGetElementType(qvalue_t type);
		qvalue_t TypeGetElementTypeAtPosition(qvalue_t type, int dims, int* pos);
		bool TypeCompareToSpecification(qvalue_t type, qvalue_t specType);
		bool AddFieldWithOffset(qhandle objType, LPCTSTR fieldName,
			qhandle fieldType, size_t offset);
		bool AddProperty(qhandle objType, LPCTSTR propertyName, qhandle fieldType,
			qhandle getProperty, qhandle setProperty);
		bool InheritType(qhandle objType, qhandle baseType);
		bool SetConstructor(qhandle objType, ConstructorDelegate* constructor);
		bool SetDynamicDispatcher(qhandle typeHandle, DynamicDispatchers disp, DynamicCallDelegate *call);
		qhandle CreateFunctionType(FunctionTypes functionType,
			const qhandle* inputTypes, int numInputTypes,
			const qhandle* outputTypes, int numOutputTypes, FunctionTypeFlags flags);
		qhandle CreateTypeClass(const qhandle *elemTypes, int elemTypeCount);
		qhandle InstantiateType(qhandle type, const qhandle* instance_type_params, int num_params);
		qhandle CreatePointerType(qhandle type);
		qhandle CreateTypeParameter(LPCTSTR paramName, qhandle paramType);
		void* RegisterInstance(void *native_obj_ptr, qvalue_t &type);
		MemResources GetMemResourceForDevice(int deviceId);
		void* GetDataPtr(qvalue_t *obj, MemResources memResource);
		void GetLambdaFunctionPtr(qvalue_t *lambda, int deviceId, void **ptr, void **closure);
		void UnmarshalPointer(qvalue_t *obj, LPCVOID ptr, const void *data, size_t data_size, DataTypes data_type, int deviceId, int flags);
		void UnmarshalTypedObject(qvalue_t *qv, const void *ptr, size_t size, LPCTSTR typeName, int deviceId, int flags);
		void SetupKernelFunction(
			qvalue_t &qv_kernel,
			std::function<KernelLambdaDelegate> cpu_kernel_adapter,
			std::function<KernelLambdaDelegate> cuda_kernel_adapter,
			std::function<PrimeKernelLambdaDelegate> cuda_prime_kernel_adapter);
		void SetupDeviceFunction(
			qvalue_t &qv_device,
			std::function<LambdaDelegate> cpu_device_adapter,
			void *cpu_device_func_ptr);
		bool IsKernelInitialized(qhandle kernelHandle);
		void LaunchKernel(qhandle kernel, LPCVOID params, size_t params_size,
			const void *closure_ptr, const qvalue_t * args, int3 gridDim, int3 blockDim,
			bool prime_kernel, int deviceId, qhandle _stream);
		void EnumRegisteredTypes(EnumRegisteredTypeDelegate *callback, void* userData);
		void EnumFields(const qvalue_t &obj, bool(*callback)(const qvalue_t *obj, LPCTSTR fieldName, qhandle fieldType, void* userData), void *userData);

		MachineInfo GetMachineInfo();
		void FreeMachineInfo(MachineInfo &machineInfo);
		void OpenDevice(LPCTSTR deviceId, const RuntimeSettings *runtimeSettings);
		void CloseDevice();
		void AddLambdaDestroyCallback(qvalue_t lambda, DisposeClosure* disposeDelegate);
		bool IsRebindRequired(qvalue_t lambda, int token);
		void SetMainProgram(LPCTSTR mainProgramName);
		bool IsMainProgram(LPCTSTR programName);
		void GetProgramCmdLineArguments(qvalue_t *argsOut, int numArgs, int numMandatoryArgs, bool variadic);
		void SetProgramCmdLineArgumentOffset(int offset);
		bool AddHostDestroyCallback(HostDestroyCallbackDelegate *callback, void *userData);
		void ParseNamedParam(const qvalue_t &named_param, string_t &name, qvalue_t &value);
		void FreeString(LPCTSTR strVal);
		void OnException(const exception_t &exception);
		size_t GetRawSizeOfType(qhandle typeHandle, MemResources memResource, size_t *alignment);
		void TypeGetClassInfo(qhandle typeHandle, string_t &moduleName, string_t &typeName, ClassInfoFlags &flags,
			MemResources* resources, int *numResources, size_t* size, qhandle *baseClass, int *numGenericTypeParams, qhandle *genericTypeParams,
			int *numInstanceTypeParams, qhandle *instanceTypeParams);
		static void ThrowException(string_t msg);

		static QuasarHostImpl *GetInstance() { return (QuasarHostImpl*) instance; }
		static void SetInstance(IQuasarHost *host) { instance = host; }
	};
}

//============================================================================
// Begin implementation
//============================================================================

namespace quasar
{
	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::ComputationEngineImpl
	 *------------------------------------------------------------------------*/
	ComputationEngineImpl::ComputationEngineImpl(QuasarContext *context)
	{
		this->context = context;
		this->handle = context->GetComputationEngine(NULL);
		this->evaluationStack = NULL;
		this->rre = NULL;
		this->matrixFactory = NULL;
		this->typeEnvironment = NULL;
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::GetName
	 *
	 * Retrieves the name of the computation engine.
	 *
	 * Parameters:
	 * name - a buffer to write to
	 * maxChars - the maximum number of characters to write
	 *------------------------------------------------------------------------*/
	bool ComputationEngineImpl::GetName(LPTSTR name, int maxChars)
	{
		return context->ComputationEngine_GetName(handle, name, maxChars);
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::Process
	 *
	 * Performs a calculation on the arguments passed to the evaluation stack�
	 *
	 * Parameters:
	 * op - the operation to be performed (e.g. OP_ADD, OP_SUBTRACT)
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::Process(OperatorTypes op)
	{
		if (!context->ComputationEngine_Process(handle, op))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::ConstructMatrix
	 *
	 * Constructs a vector/matrix based on arguments passed to the evaluation
	 * stack
	 * Parameters:
	 * numRowsOrCols - the number of elements pushed to the evaluation stack
	 *
	 * Notes:
	 * - if the evaluation stack contains scalar values, a vector will be
	 *   created
	 * - if the evaluation stack contains vectors, a matrix will be created
	 * - higher dimensional arrays are currently not supported. All elements
	 *   must have the same type, otherwise an error will result.
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::ConstructMatrix(int numRowsOrCols)
	{
		if (!context->ComputationEngine_ConstructMatrix(handle, numRowsOrCols, 0))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::ConstructCellMatrix
	 *
	 * Constructs a cell vector based on arguments passed to the evaluation
	 * stack.
	 *
	 * Parameters:
	 * length - the number of elements pushed to the evaluation stack
	 *
	 * Notes:
	 * - the function can currently only construct cell vectors (not matrices).
	 *   To construct a cell matrix, use CreateMatrix, CreateCube etc. and
	 *   fill up the elements manually using ArraySetAt.
	 * - if all NamedParams are passed (see CreateNamedParam), an untyped
	 *   object will be created with the fields set according to the named
	 *   parameters.
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::ConstructCellMatrix(int length)
	{
		if (!context->ComputationEngine_ConstructCellMatrix(handle, length, 0))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::FunctionCall
	 *
	 * Performs a function call to a function that is currently known to
	 * Quasar. This can be either a builtin function or a function that is
	 * defined via reductions.
	 *
	 * Parameters:
	 * functionName - the name of the function to call
	 * numArgs - the number of parameters to call the function
	 *------------------------------------------------------------------------*/
	bool ComputationEngineImpl::FunctionCall(LPCTSTR functionName, int numArgs)
	{
		if (!context->ComputationEngine_FunctionCall(handle, functionName, numArgs, 0))
			QuasarHostImpl::GetInstance()->RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::FunctionExists
	 *
	 * Checks whether the function with the given name and the specified
	 * number of parameters exists.
	 *
	 * Parameters:
	 * functionName - the name of the function to call
	 * numArgs - the number of parameters to call the function
	 *
	 * Notes:
	 * Even if the function exists, the function may not be callable because
	 * the parameter types may not match the signature of the function
	 *------------------------------------------------------------------------*/
	bool ComputationEngineImpl::FunctionExists(LPCTSTR functionName,
		int numArgs)
	{
		return context->ComputationEngine_FunctionExists(handle, functionName,
			numArgs, 0);
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::ArrayGetAt
	 *
	 * Gets the value of an array (vector, matrix, cube etc.) at the given
	 * position, where the array indices are pushed to the evaluation stack.
	 *
	 * Parameters:
	 * x - the array to obtain a value from
	 * numIndices - the number of indices of the array
	 * boundsMode - the boundary access mode of the array
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::ArrayGetAt(const qvalue_t &x, int numIndices,
		BoundaryAccessMode boundsMode)
	{
		if (!context->ComputationEngine_ArrayGetAt(handle, &x, numIndices, boundsMode))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::ArraySetAt
	 *
	 * Sets the value of an array (vector, matrix, cube etc.) at the given
	 * position, where both the value and the array indices are pushed
	 * to the evaluation stack.
	 *
	 * Parameters:
	 * x - the array to obtain a value from
	 * numIndices - the number of indices of the array
	 * boundsMode - the boundary access mode of the array
	 *
	 * Notes:
	 * The index/indices are pushed first, then the value of the array
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::ArraySetAt(qvalue_t &x, int numIndices,
		BoundaryAccessMode boundsMode)
	{
		if (!context->ComputationEngine_ArraySetAt(handle, &x, numIndices, boundsMode))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::Synchronize
	 *
	 * Performs a host synchronization operation. This will impose a barrier
	 * for all CPU threads but also all active GPU streams.
	 * In synchronous execution mode, this function will have no effect
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::Synchronize()
	{
		if (!context->ComputationEngine_Synchronize(handle, 0))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::ParallelDo
	 *
	 * Launches the given kernel function in parallel using the specified working
	 * dimensions. The parameters of the function are passed via the evaluation
	 * stack. The return value of the kernel function (if any) will be returned.
	 *
	 * Parameters:
	 * dims - a scalar/vector/matrix indicating the working dimensions of the
	 *    kernel. In case a 2x3 matrix, the first row indicates the working
	 *    dimensions and the second row contains the block dimensions.
	 * function - qvalue_t object corresponding to the kernel function to call
	 * numArgs - the number of parameters that were pushed to the evaluation
	 *    stack.
	 * retval - the return value of the function
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::ParallelDo(const qvalue_t& dims, const qvalue_t& function, int numArgs, qvalue_t &retval)
	{
		if (!context->ComputationEngine_ParallelDo(handle, &dims, &function, numArgs, &retval))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::SerialDo
	 *
	 * See description of ComputationEngineImpl::ParallelDo. The difference
	 * is that this function will run the kernel function serially (typically,
	 * on a CPU)
	 *------------------------------------------------------------------------*/
	void ComputationEngineImpl::SerialDo(const qvalue_t& dims, const qvalue_t& function, int numArgs, qvalue_t &retval)
	{
		if (!context->ComputationEngine_SerialDo(handle, &dims, &function, numArgs, &retval))
			QuasarHostImpl::GetInstance()->RaiseException();
	}


	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::GetEvaluationStack
	 *
	 * Gets the evaluation stack for this computation engine. The evaluation
	 * stack is used for passing values to the computation engine and returning
	 * their values. The underlying implementation may perform a deferred
	 * execution (i.e., the returned qvalue_t objects may be "futures").
	 *------------------------------------------------------------------------*/
	IEvaluationStack *ComputationEngineImpl::GetEvaluationStack()
	{
		if (evaluationStack == NULL)
		{
			evaluationStack = new EvaluationStackImpl(context, this);
			evaluationStack->AddRef();
		}
		return evaluationStack;
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::GetMatrixFactory
	 *
	 * Returns the matrix factory for this computation engine, which can be
	 * used to construct various matrices.
	 *------------------------------------------------------------------------*/
	IMatrixFactory *ComputationEngineImpl::GetMatrixFactory()
	{
		if (matrixFactory == NULL)
		{
			matrixFactory = new MatrixFactoryImpl(context, this);
			matrixFactory->AddRef();
		}
		return matrixFactory;
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::GetTypeEnvironment
	 *
	 * Returns the current type environment for this computation engine.
	 *------------------------------------------------------------------------*/
	ITypeEnvironment *ComputationEngineImpl::GetTypeEnvironment()
	{
		if (typeEnvironment == NULL)
		{
			typeEnvironment = new TypeEnvironmentImpl(context, this);
			typeEnvironment->AddRef();
		}
		return typeEnvironment;
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::GetRuntimeReductionEngine
	 *
	 * Returns the runtime reduction engine for this computation engine
	 *------------------------------------------------------------------------*/
	IRuntimeReductionEngine *ComputationEngineImpl::GetRuntimeReductionEngine()
	{
		if (rre == NULL)
		{
			rre = new RuntimeReductionEngineImpl(context, this);
			rre->AddRef();
		}
		return rre;
	}

	/*------------------------------------------------------------------------
	 * ComputationEngineImpl::~ComputationEngineImpl
	 *
	 * Computation engine destructor
	 *------------------------------------------------------------------------*/
	ComputationEngineImpl::~ComputationEngineImpl()
	{
		RELEASE_SAFE(evaluationStack);
		RELEASE_SAFE(matrixFactory);
		RELEASE_SAFE(typeEnvironment);
		RELEASE_SAFE(rre);
	}


	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::EvaluationStackImpl
	 *------------------------------------------------------------------------*/
	EvaluationStackImpl::EvaluationStackImpl(QuasarContext *context,
		ComputationEngineImpl *engine)
	{
		this->context = context;
		this->handle = context->ComputationEngine_GetEvaluationStack(
			engine->GetHandle(), NULL);
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::~EvaluationStackImpl
	 *------------------------------------------------------------------------*/
	EvaluationStackImpl::~EvaluationStackImpl()
	{
		DELETE_HANDLE(context, handle);
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::PopValue
	 *
	 * Pops one value from the evaluation stack. The returned value needs to
	 * be released by the caller
	 *------------------------------------------------------------------------*/
	qvalue_t EvaluationStackImpl::PopValue()
	{
		qvalue_t x;
		if (!context->Stack_PopMultiple(handle, &x, 1))
			QuasarHostImpl::GetInstance()->RaiseException();
		return x;
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::PushValue
	 *
	 * Pushes one value to the evaluation stack, transferring ownership to the
	 * evaluation stack.
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::PushValue(const qvalue_t &value)
	{
		if (!context->Stack_PushMultiple(handle, &value, 1))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::PopMultiple
	 *
	 * Same as PopValue, but now with multiple values at the time
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::PopMultiple(qvalue_t *vals, int count)
	{
		if (!context->Stack_PopMultiple(handle, vals, count))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::PushMultiple
	 *
	 * Same as PushValue, but now with multiple values at the time
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::PushMultiple(const qvalue_t *vals, int count)
	{
		if (!context->Stack_PushMultiple(handle, vals, count))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::Clear
	 *
	 * Clears the evaluation stack
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::Clear()
	{
		if (!context->Stack_Clear(handle))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::PushContext
	 *
	 * Saves the state of the evaluation stack, in order to be restored at a
	 * later time by PopContext
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::PushContext()
	{
		if (!context->Stack_PushContext(handle))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::PopContext
	 *
	 * Restores the state of the evaluation stack.
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::PopContext()
	{
		if (!context->Stack_PopContext(handle))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::Process
	 *
	 * Same as ComputationEngineImpl::Process
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::Process(OperatorTypes op)
	{
		if (!context->Stack_Process(handle, op))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::FunctionCall
	 *
	 * Same as ComputationEngineImpl::FunctionCall
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::FunctionCall(LPCTSTR functionName, int numArgs)
	{
		if (!context->Stack_FunctionCall(handle, functionName, numArgs, 0))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::ArrayGetAt
	 *
	 * Same as ComputationEngineImpl::ArrayGetAt
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::ArrayGetAt(const qvalue_t &x, int numIndices,
		BoundaryAccessMode boundsMode)
	{
		if (!context->Stack_ArrayGetAt(handle, &x, numIndices, boundsMode))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::ArraySetAt
	 *
	 * Same as ComputationEngineImpl::ArraySetAt
	 *------------------------------------------------------------------------*/
	void EvaluationStackImpl::ArraySetAt(qvalue_t &x, int numIndices,
		BoundaryAccessMode boundsMode)
	{
		if (!context->Stack_ArraySetAt(handle, &x, numIndices, boundsMode))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::Evaluate
	 *
	 * Performs evaluation on the given qvalue_t "future" and waits for the
	 * result if it is not yet available. Note that the implementation may not
	 * support such evaluation, in this case the input value is returned.
	 *------------------------------------------------------------------------*/
	qvalue_t EvaluationStackImpl::Evaluate(const qvalue_t &x)
	{
		qvalue_t a = x;
		if (!context->Stack_Evaluate(handle, &a))
			QuasarHostImpl::GetInstance()->RaiseException();
		return a;
	}

	/*------------------------------------------------------------------------
	 * EvaluationStackImpl::GetCount
	 *
	 * Returns the number of items currently pushed to the evaluation stack
	 *------------------------------------------------------------------------*/
	int EvaluationStackImpl::GetCount()
	{
		return context->Stack_GetCount(handle);
	}

	/*------------------------------------------------------------------------
	 * MatrixFactoryImpl::MatrixFactoryImpl
	 *------------------------------------------------------------------------*/
	MatrixFactoryImpl::MatrixFactoryImpl(QuasarContext *context,
		ComputationEngineImpl *engine)
	{
		this->context = context;
		this->handle =
			context->ComputationEngine_GetMatrixFactory(engine->GetHandle());
	}

	/*------------------------------------------------------------------------
	 * MatrixFactoryImpl::~MatrixFactoryImpl
	 *------------------------------------------------------------------------*/
	MatrixFactoryImpl::~MatrixFactoryImpl() { DELETE_HANDLE(context, handle); }

	qvalue_t MatrixFactoryImpl::New(AllocationFlags flags, LPCVOID elems, int64_t elemBytes, qhandle elemType,
		int nDims, const int *dims)
	{
		qvalue_t t;
		if (!context->MatrixFactory_New(handle, &t, flags, elems, elemBytes, elemType, nDims, dims))
			QuasarHostImpl::GetInstance()->RaiseException();
		return t;
	}

	/*------------------------------------------------------------------------
	 * TypeEnvironmentImpl::TypeEnvironmentImpl
	 *------------------------------------------------------------------------*/
	TypeEnvironmentImpl::TypeEnvironmentImpl(QuasarContext *context,
		ComputationEngineImpl *engine)
	{
		this->context = context;
		this->handle =
			context->ComputationEngine_GetTypeEnvironment(engine->GetHandle());
	}

	/*------------------------------------------------------------------------
	 * TypeEnvironmentImpl::~TypeEnvironmentImpl
	 *------------------------------------------------------------------------*/
	TypeEnvironmentImpl::~TypeEnvironmentImpl()
	{
		DELETE_HANDLE(context, handle);
	}

	/*------------------------------------------------------------------------
	 * TypeEnvironmentImpl::GetScalarType
	 *
	 * Returns the currently selected scalar type
	 *------------------------------------------------------------------------*/
	ScalarTypes TypeEnvironmentImpl::GetScalarType()
	{
		return context->TypeEnvironment_GetScalarType(handle);
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::RuntimeReductionEngineImpl
	 *------------------------------------------------------------------------*/
	RuntimeReductionEngineImpl::RuntimeReductionEngineImpl(
		QuasarContext *context, ComputationEngineImpl *engine)
	{
		this->context = context;
		this->handle =
			context->ComputationEngine_GetRuntimeReductionEngine(engine->GetHandle());
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::~RuntimeReductionEngineImpl
	 *------------------------------------------------------------------------*/
	RuntimeReductionEngineImpl::~RuntimeReductionEngineImpl()
	{
		DELETE_HANDLE(context, handle);
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::Add
	 *
	 * Adds the specified search expression (for example
	 * "(x, y) -> my_func(x,y)") to the runtime reduction system; when the
	 * pattern is matched and the where clause is satisfied, action will be
	 * called
	 *
	 * Parameters:
	 * searchExpr - expression to match
	 * action - action (Function) to execute
	 * whereClause - function that checks the conditions of the reduction
	 *          or NULL in case the whereClause==true.
	 *
	 * Returns:
	 * true if successful and the reduction was non-existing, otherwise false
	 *------------------------------------------------------------------------*/
	bool RuntimeReductionEngineImpl::Add(LPCTSTR searchExpr,
		qhandle action,
		qhandle whereClause)
	{
		if (!context->RuntimeReductionEngine_Add(handle, searchExpr, action, whereClause, 0))
		{
			QuasarHostImpl::GetInstance()->RaiseException();
		}
		return true;
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::Add
	 *------------------------------------------------------------------------*/
	bool RuntimeReductionEngineImpl::Add(const VariableDecl* variables, int numVariables,
		Expression *searchExpr, const qvalue_t *action, const qvalue_t *whereClause)
	{
		Assert(action->type == TYPE_LAMBDAEXPR ||
				action->type == TYPE_METHOD, "action: Method or lambda expression required");
		Assert(whereClause == nullptr || whereClause->private_obj == nullptr ||
				whereClause->type == TYPE_LAMBDAEXPR ||
				whereClause->type == TYPE_METHOD, "whereClause: Method or lambda expression required");
		if (!context->RuntimeReductionEngine_Add_v2(handle, variables, numVariables, searchExpr->GetSymbols(),
			searchExpr->GetNumSymbols(), action->private_obj, whereClause == nullptr ? nullptr : whereClause->private_obj,
			searchExpr->IsVariadic() ? 1 : 0))
		{
			QuasarHostImpl::GetInstance()->RaiseException();
		}
		return true;
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::Remove
	 *
	 * Removes a reduction that was previously added to the reduction engine.
	 *
	 * Returns:
	 * true if successful, otherwise false
	 *------------------------------------------------------------------------*/
	bool RuntimeReductionEngineImpl::Remove(LPCTSTR searchExpr)
	{
		return context->RuntimeReductionEngine_Remove(handle, searchExpr, 0);
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::EnumFields
	 *
	 * Enumerates all fields of a given type
	 *------------------------------------------------------------------------*/
	void RuntimeReductionEngineImpl::EnumFields(qhandle type, int numOffsets, const MemResources *memResources,
		EnumFieldWithDetailsDelegate *callback, void* userData)
	{
		if (!context->RuntimeReductionEngine_EnumFields(type, numOffsets, memResources, callback, userData))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::EnumFunctions
	 *
	 * Enumerates all functions currently registered in Quasar
	 *------------------------------------------------------------------------*/
	void RuntimeReductionEngineImpl::EnumFunctions(
		EnumFunctionsDelegate *callback, void* userData)
	{
		if (!context->RuntimeReductionEngine_EnumFunctions(callback, userData))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * RuntimeReductionEngineImpl::EnumMethods
	 *
	 * Enumerates all methods currently registered for a given type
	 *------------------------------------------------------------------------*/
	void RuntimeReductionEngineImpl::EnumMethods(qhandle type, EnumMethodsDelegate *callback, void* userData)
	{
		if (!context->RuntimeReductionEngine_EnumMethods(type, callback, userData))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	IQuasarHost *QuasarHostImpl::instance;

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::QuasarHostImpl
	 *------------------------------------------------------------------------*/
	QuasarHostImpl::QuasarHostImpl(const QuasarContext &context,
		ICLRMetaHost *pMetaHost,
		ICLRRuntimeInfo *pRuntimeInfo,
		ICLRRuntimeHost *pClrRuntimeHost,
		MonoDomain *monoDomain)
	{
		QuasarHostImpl::instance = this;
		this->context = context;
		this->pMetaHost = pMetaHost;
		this->pRuntimeInfo = pRuntimeInfo;
		this->pClrRuntimeHost = pClrRuntimeHost;
		this->monoDomain = monoDomain;
		this->custom_exceptionhandler = NULL;
		this->engine = NULL;
		this->cmdLineParamOffset = 1;

		if (tcslen(context.deviceName) > 0)
			Initialized();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::Initialized
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::Initialized()
	{
		for (int i = 0; i < TYPE_COUNT; i++)
		{
			typeHandles[i] =
				typeNames[i] ? context.ParseType(typeNames[i]) : NULL;
			isRefCountable[i] = false;
		}
		isRefCountable[TYPE_VEC] = true;
		isRefCountable[TYPE_MAT] = true;
		isRefCountable[TYPE_CUBE] = true;
		isRefCountable[TYPE_CVEC] = true;
		isRefCountable[TYPE_CMAT] = true;
		isRefCountable[TYPE_CCUBE] = true;
		isRefCountable[TYPE_NCUBE] = true;
		isRefCountable[TYPE_NCCUBE] = true;
		isRefCountable[TYPE_TYPEDOBJECT] = true;
		isRefCountable[TYPE_UNTYPEDOBJECT] = true;
		isRefCountable[TYPE_STRING] = true;
		isRefCountable[TYPE_TYPEINFO] = true;
		isRefCountable[TYPE_LAMBDAEXPR] = true;
		isRefCountable[TYPE_METHOD] = true;

		RELEASE_SAFE(engine);
		this->engine = new ComputationEngineImpl(&this->context);
		this->engine->AddRef();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::Uninitialize
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::Uninitialize()
	{
		for (int i = 0; i < TYPE_COUNT; i++)
			DELETE_HANDLE(&context, typeHandles[i]);
		RELEASE_SAFE(engine);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::~QuasarHostImpl
	 *------------------------------------------------------------------------*/
	QuasarHostImpl::~QuasarHostImpl()
	{
		for (auto& item : destroyCallbacks)
			std::get<1>(item)(std::get<0>(item), this);

		Uninitialize();
		if (context.Close)
		{
			context.Close();
			context.Close = NULL;
		}
#ifdef MS_DOTNET
		RELEASE_SAFE(pMetaHost);
		RELEASE_SAFE(pRuntimeInfo);
		RELEASE_SAFE(pClrRuntimeHost);
#else
		if (monoDomain != NULL)
		{
			mono_jit_cleanup(monoDomain);
			monoDomain = NULL;
		}
#endif
		QuasarHostImpl::instance = NULL;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::FunctionExists
	 *
	 * Checks whether any function with the given name exists, regardless
	 * of any number of parameters
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::FunctionExists(LPCTSTR functionName)
	{
		return context.FunctionExists(functionName);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::FunctionCall
	 *
	 * Direct calling interface, supporting multiple input/output arguments
	 *
	 * Parameters:
	 * inArgs - array of input arguments
	 * nInArgs - number of input arguments being passed
	 * outArgs - array of output arguments
	 * nOutArgs - number of output arguments expected
	 *
	 * Notes:
	 * It is the caller that decides upon the number of output arguments.
	 * This has the advantage that there will be no surprises. If the callee
	 * does not agree with the number of input/output parameters, an exception
	 * will be raised.
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::FunctionCall(LPCTSTR functionName,
		const qvalue_t *inArgs, int nInArgs,
		qvalue_t *outArgs, int nOutArgs)
	{
		bool retVal = context.FunctionCall(functionName, inArgs, nInArgs, outArgs, nOutArgs);

		if (!retVal)
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::FunctionCall
	 *
	 * Overload, allowing indirect function calls.
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::FunctionCall(const qvalue_t &value,
		const qvalue_t *inArgs, int nInArgs,
		qvalue_t *outArgs, int nOutArgs)
	{
		Assert(value.type == TYPE_LAMBDAEXPR ||
				value.type == TYPE_METHOD, "value: value or lambda expression required!");
		bool retVal = context.FunctionCallIndirect(value.private_obj, inArgs, nInArgs, outArgs, nOutArgs);

		if (!retVal)
			RaiseException();
		return true;
	}


	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateVector
	 *
	 * Creates a vector of the specified length and element type
	 *
	 * Parameters:
	 * elemType - type handle of the element type
	 * numel - number of elements
	 * complex - if true, a complex-valued vector will be created
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateVector(qhandle elemType, int numel, bool complex)
	{
		int dims[16] = { 1,numel };

		qvalue_t t;
		t.type = complex ? TYPE_CVEC : TYPE_VEC;
		t.ptrVal = NULL;
		t.dim1 = 1;
		t.dim2 = numel;
		t.dim3 = 1;
		t.private_obj = context.CreateNDMatrix(elemType, dims, 2);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateMatrix
	 *
	 * Creates a matrix of dimensions dim1 x dim2 of the specified element type
	 *
	 * Parameters:
	 * elemType - type handle of the element type
	 * dim1 - first dimension of the matrix
	 * dim2 - second dimension of the matrix
	 * complex - if true, a complex-valued matrix will be created
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateMatrix(qhandle elemType, int dim1, int dim2, bool complex)
	{
		int dims[16] = { dim1, dim2 };

		qvalue_t t;
		t.type = complex ? TYPE_CMAT : TYPE_MAT;
		t.ptrVal = NULL;
		t.dim1 = dim1;
		t.dim2 = dim2;
		t.dim3 = 1;
		t.private_obj = context.CreateNDMatrix(elemType, dims, 2);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateCube
	 *
	 * Creates a cube (3D array) of dimensions dim1 x dim2 of the specified
	 * element type
	 *
	 * Parameters:
	 * elemType - type handle of the element type
	 * dim1 - first dimension of the cube
	 * dim2 - second dimension of the cube
	 * dim3 - third dimension of the cube
	 * complex - if true, a complex-valued cube will be created
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateCube(qhandle elemType, int dim1, int dim2,
		int dim3, bool complex)
	{
		int dims[16] = { dim1, dim2, dim3 };

		qvalue_t t;
		t.type = complex ? TYPE_CCUBE : TYPE_CUBE;
		t.ptrVal = NULL;
		t.dim1 = dim1;
		t.dim2 = dim2;
		t.dim3 = dim3;
		t.private_obj = context.CreateNDMatrix(elemType, dims, 3);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateNCube
	 *
	 * Creates a hypercube (ND array) of the specified element type
	 *
	 * Parameters:
	 * elemType - type handle of the element type
	 * ndims - the number of dimensions (elements in 'dims')
	 * dims - array containing the dimensions of the matrix
	 * complex - if true, a complex-valued cube will be created
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateNCube(qhandle elemType, int ndims, const int* dims, bool complex)
	{
		Assert(ndims >= 1, "CreateNCube expects ndims>=1");
		if (ndims == 1)
			return CreateVector(elemType, dims[0], complex);
		qvalue_t t;
		t.type = complex ? TYPE_NCCUBE : TYPE_NCUBE;
		t.ptrVal = NULL;
		t.dim1 = dims[0];
		t.dim2 = ndims > 1 ? dims[1] : 1;
		t.dim3 = ndims > 2 ? dims[2] : 1;
		t.private_obj = context.CreateNDMatrix(elemType, dims, ndims);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateString
	 *
	 * Creates a qvalue_t object for the given string
	 *
	 * Parameters:
	 * text - LPCTSTR (const TCHAR *) representation of the string
	 * length - the length of the string (zero termination character is not
	 *   required).
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateString(LPCTSTR text, size_t length)
	{
		// string length not specified
		if (length == (size_t)-1)
			length = tcslen(text);
		Assert(0 <= length && length <= INT_MAX, "CreateString: invalid length");
		qvalue_t t;
		t.type = TYPE_STRING;
		t.private_obj = context.CreateString(text, (int)length);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateTypedObject
	 *
	 * Creates a typed object of the specified object type.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateTypedObject(qhandle objType)
	{
		qvalue_t t;
		t.type = TYPE_TYPEDOBJECT;
		t.private_obj = context.CreateObject(objType);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateUntypedObject
	 *
	 * Creates an untyped object of the specified object type. Untyped objects
	 * can have members added dynamically at runtime. Untyped objects cannot
	 * be passed to kernel functions.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateUntypedObject(qhandle objType)
	{
		qvalue_t t;
		t.type = TYPE_UNTYPEDOBJECT;
		t.private_obj = context.CreateObject(objType);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateLambda
	 *
	 * Creates a qvalue_t object for a lambda expression with closure pointer
	 * 'closurePtr'.
	 *
	 * Parameters:
	 * closurePtr - a data pointer to store in the qvalue_t object.
	 * lambda - the function to be called when the lambda expression is invoked
	 *
	 * Notes:
	 * in case `closurePtr` contains dynamically allocated data, you may use
	 * `IQuasarHost::AddLambdaDestroyCallback` to free the data.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateLambda(qhandle functionTypeHandle,
		void* closurePtr,
		LambdaDelegate* lambda)
	{
		qvalue_t t;
		t.type = TYPE_LAMBDAEXPR;
		t.private_obj = context.CreateLambda(functionTypeHandle, closurePtr, lambda);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateNamedParam
	 *
	 * Creates a qvalue_t object representing named parameter (param:=value).
	 * There are a limited number of functions that can deal with named
	 * parameters, for example IComputationEngine::ConstructCellMatrix
	 *
	 * Parameters:
	 * paramName - zero terminated string containing the parameter name
	 * value - the value to be associated with this parameter
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateNamedParam(LPCTSTR paramName, const qvalue_t &value)
	{
		qvalue_t t;
		t.type = TYPE_NAMEDPARAM;
		t.private_obj = context.CreateNamedParam(paramName, &value);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::ParseNamedParam
	 *
	 * Parses a named parameter, resulting in its name (and value)
	 *
	 * Parameters:
	 * paramName - zero terminated string containing the parameter name
	 * value - the value to be associated with this parameter
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::ParseNamedParam(const qvalue_t &named_param,
		string_t &paramName, qvalue_t &value)
	{
		Assert(named_param.type == TYPE_NAMEDPARAM, "ParseNamedParam: named parameter expected!");
		LPTSTR text;
		int length;
		if (!context.ParseNamedParam(&named_param, &text, &length, &value))
			RaiseException();
		paramName = string_t::unmarshal(text, length);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::FreeString
	 *
	 * Frees a string that was allocated by one of the host functions
	 * (for example TypeGetString)
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::FreeString(LPCTSTR strVal)
	{
		context.FreeString(strVal);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateGeneric
	 *
	 * Creates an instance of a type, with the specified type parameters
	 * For example `vec[cube](a,b,c)`
	 *
	 * Parameters:
	 * typeHandle - The type to be instantiated
	 * inArgs - array with input arguments
	 * nInArgs - number of input arguments being passed
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateGeneric(qhandle typeHandle, const qvalue_t *inArgs, int nInArgs)
	{
		qvalue_t t;
		if (!context.CreateGeneric(typeHandle, nInArgs, inArgs, &t))
			RaiseException();
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LookupFunction
	 *
	 * Looks up the function with the specified function signature.
	 * For example `print(...)`.
	 *
	 * Returns a qvalue_t object containing a function handle
	 *
	 * Parameters:
	 * functionSignature - a zero terminated string containing the function
	 *    signature. A function signature consists of the function name
	 *    with comma separated argument types in parentheses. Variadic
	 *    functions have '...' as last argument.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::LookupFunction(LPCTSTR functionSignature)
	{
		qvalue_t t;
		t.type = TYPE_LAMBDAEXPR;
		t.private_obj = context.LookupFunction(functionSignature);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LookupFunction
	 *
	 * Same as the above function, but returns a token. The token can be
	 * passed to the function `QuasarHostImpl::IsRebindRequired`. If rebinding
	 * is required, `LookupFunction` needs to be called again to obtain a more
	 * recent function handle.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::LookupFunction(LPCTSTR functionSignature, int &token)
	{
		token = 0; // Currently not supported by this runtime backend
		return LookupFunction(functionSignature);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LookupType
	 *
	 * Looks up the type with the specified name. Returns
	 *
	 * Parameters:
	 * quasarTypeName - a zero terminated string containing a Quasar type name
	 *			for example, 'vec', 'mat', 'lambda_expr', ...
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::LookupType(LPCTSTR quasarTypeName)
	{
		qvalue_t t;
		t.type = TYPE_TYPEINFO;
		t.private_obj = context.ParseType(quasarTypeName);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LookupMethod
	 *
	 * Looks up a method of the specified type with the given function
	 * signature.
	 *
	 * Parameters:
	 * functionSignature - a zero terminated string containing the function
	 *    signature. A function signature consists of the function name
	 *    with comma separated argument types in parentheses. Variadic
	 *    functions have '...' as last argument.
	 *
	 * Returns:
	 * a qvalue_t object containing a method handle
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::LookupMethod(const qvalue_t &type,
		LPCTSTR functionSignature)
	{
		Assert(type.type == TYPE_TYPEINFO, "LookupMethod: typeinfo expected for parameter 'type'");
		qvalue_t t;
		t.type = TYPE_METHOD;
		t.private_obj =
			context.LookupMethod(type.private_obj, functionSignature);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LookupMethod
	 *
	 * Same as the above function, but returns a token. The token can be
	 * passed to the function `QuasarHostImpl::IsRebindRequired`. If rebinding
	 * is required, `LookupMethod` needs to be called again to obtain a more
	 * recent function handle.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::LookupMethod(const qvalue_t &type, LPCTSTR functionSignature, int &token)
	{
		token = 0;
		return LookupMethod(type, functionSignature);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::MethodCall
	 *
	 * Calls a method on the given target object. The method is specified by
	 * a qvalue_t object containing a method handle
	 * (see QuasarHostImpl::LookupMethod).
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::MethodCall(const qvalue_t &function,
		const qvalue_t &target,
		const qvalue_t *inArgs, int nInArgs,
		qvalue_t *outArgs, int nOutArgs)
	{
		Assert(function.type == TYPE_METHOD, "function: value expected");
		Assert(target.type == TYPE_TYPEDOBJECT ||
			   target.type == TYPE_UNTYPEDOBJECT, "target: object expected");
		if (!context.MethodCall(function.private_obj, target.private_obj,
			target.type, inArgs, nInArgs, outArgs,
			nOutArgs))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::MethodCall
	 *
	 * Calls a method on the given target object.
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::MethodCall(LPCTSTR methodName,
		const qvalue_t &target,
		const qvalue_t *inArgs, int nInArgs,
		qvalue_t *outArgs, int nOutArgs)
	{
		Assert(target.type == TYPE_TYPEDOBJECT ||
			target.type == TYPE_UNTYPEDOBJECT, "target: object expected");
		if (!context.MethodCallDirect(methodName, target.private_obj,
			target.type, inArgs, nInArgs, outArgs,
			nOutArgs))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetType
	 *
	 * Returns a qvalue_t containing the type of the specified object
	 *
	 * Parameters:
	 * obj - the object to determine the type.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::GetType(const qvalue_t &obj)
	{
		qvalue_t t;
		t.type = TYPE_TYPEINFO;
		t.private_obj = context.GetType(obj.private_obj, obj.type);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateType
	 *
	 * Creates a user defined type.
	 *
	 * Parameters:
	 * moduleName - the name of the module that defines this type
	 * typeName - the name of the type
	 * typeFlags - additional flags (e.g. mutable, immutable, dynamic).
	 *
	 * Notes:
	 * Mutable and immutable will create a typed object.
	 * Dynamic creates an untyped object
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::CreateType(LPCTSTR moduleName, LPCTSTR typeName, CreateTypeFlags typeFlags)
	{
		qvalue_t t;
		t.type = TYPE_TYPEINFO;
		t.private_obj = context.CreateType(moduleName, typeName, typeFlags);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::AddField
	 *
	 * Adds a field with the specified name and type to a user defined type.
	 * The type should not be finalized yet (function FinalizeType).
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::AddField(qhandle objType, LPCTSTR fieldName,
		qhandle fieldType)
	{
		return context.Type_AddField(objType, fieldName, fieldType);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::AddParameter
	 *
	 * Adds a generic parameter to the specified type. This is useful for
	 * defining generic types, such as `my_container[T]`.
	 *
	 * Parameters:
	 * objType - type handle of the user defined object type
	 * typeParam - type parameter constructed using CreateTypeParameter
	 *
	 * Returns:
	 * the type handle of the generic parameter. The generic parameter
	 * can be used for the types of the fields of the type
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::AddParameter(qhandle objType, qhandle typeParam)
	{
		qvalue_t t;
		t.type = TYPE_TYPEINFO;
		t.private_obj = context.Type_AddParameter(objType, typeParam);
		return t;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::RegisterInstance
	 *
	 * Registers the specified native object with the Quasar Host
	 *------------------------------------------------------------------------*/
	void *QuasarHostImpl::RegisterInstance(void *native_obj_ptr, qvalue_t &type)
	{
		void *handle;
		Assert(type.type == TYPE_TYPEINFO &&
			type.private_obj != nullptr, "type: typeinfo expected");
		if (!context.RegisterInstance(native_obj_ptr, type.private_obj, &handle))
			RaiseException();
		return handle;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetMemResourceForDevice
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	MemResources QuasarHostImpl::GetMemResourceForDevice(int deviceId)
	{
		MemResources result;
		if (!context.GetMemResourceForDevice(&result, deviceId))
			RaiseException();
		return result;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetDataPtr
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	void* QuasarHostImpl::GetDataPtr(qvalue_t *obj, MemResources memResource)
	{
		void *ptr = nullptr;
		if (!context.GetDataPtr(obj, memResource, &ptr))
			RaiseException();
		return ptr;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetLambdaFunctionPtr
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::GetLambdaFunctionPtr(qvalue_t *lambda, int deviceId, void **ptr, void **closure)
	{
		if (!context.GetLambdaFunctionPtr(lambda, deviceId, ptr, closure))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::UnmarshalPointer
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::UnmarshalPointer(qvalue_t *obj, LPCVOID ptr, const void *data, size_t data_size, DataTypes data_type, int deviceId, int flags)
	{
		if (!context.UnmarshalPointer(obj, ptr, data, data_size, data_type, deviceId, flags))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::UnmarshalTypedObject
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::UnmarshalTypedObject(qvalue_t *qv, const void *ptr, size_t size, LPCTSTR typeName, int deviceId, int flags)
	{
		if (!context.UnmarshalTypedObject(qv, ptr, size, typeName, deviceId, flags))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetupKernelFunction
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::SetupKernelFunction(
		qvalue_t &qv_kernel,
		std::function<KernelLambdaDelegate> cpu_kernel_adapter,
		std::function<KernelLambdaDelegate> cuda_kernel_adapter,
		std::function<PrimeKernelLambdaDelegate> cuda_prime_kernel_adapter)
	{
		//context.SetupKernelFunction()
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetupDeviceFunction
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::SetupDeviceFunction(
		qvalue_t &qv_device,
		std::function<LambdaDelegate> cpu_device_adapter,
		void *cpu_device_func_ptr)
	{

	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::IsKernelInitialized
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::IsKernelInitialized(qhandle kernelHandle)
	{
		qvalue_t qvKernel;
		qvKernel.type = TYPE_LAMBDAEXPR;
		qvKernel.private_obj = kernelHandle;
		return context.IsKernelInitialized(&qvKernel);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LaunchKernel
	 *
	 * Advanced function
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::LaunchKernel(qhandle kernel, LPCVOID params, size_t params_size,
		const void *closure_ptr, const qvalue_t * args, int3 gridDim, int3 blockDim,
		bool prime_kernel, int deviceId, qhandle stream)
	{
		qvalue_t qvKernel;
		qvKernel.type = TYPE_LAMBDAEXPR;
		qvKernel.private_obj = kernel;
		if (!context.LaunchKernel(&qvKernel, params, params_size, closure_ptr, args, gridDim, blockDim, prime_kernel ? 1 : 0, deviceId, stream))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::FinalizeType
	 *
	 * Finalizes a type created with QuasarHostImpl::CreateType. Afterwards,
	 * no changes are possible.
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::FinalizeType(qhandle typeHandle)
	{
		if (!context.Type_Finalize(typeHandle))
			RaiseException();
	}


	/*------------------------------------------------------------------------
	 * QuasarHostImpl::AddRef
	 *
	 * Increments the reference count of the specified qvalue_t object
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::AddRef(qvalue_t &obj)
	{
		if (isRefCountable[obj.type])
		{
			context.AddRef(obj.private_obj);
		}
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::ReleaseRef
	 *
	 * Decrements the reference count of the specified qvalue_t object. If the
	 * reference count is 0, the object will be released.
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::ReleaseRef(qvalue_t obj)
	{
		if (isRefCountable[obj.type])
		{
			if (obj.private_obj != NULL && context.Release(obj.private_obj))
			{
				// The object has been released!
				obj.private_obj = NULL;
			}
		}
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::DeleteValue
	 *
	 * Function is now obsolete
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::DeleteValue(qvalue_t &obj)
	{
		if (obj.private_obj == NULL)
			return; // Already deleted

		if (isRefCountable[obj.type])
		{
			if (context.Release(obj.private_obj))
			{
				// The object has been released!
				obj.private_obj = NULL;
			}
		}
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetNDims
	 *
	 * Gets the dimensions of the specified array (vector/matrix/cube)
	 *
	 * Parameters:
	 * obj - the array to determine the dimensions
	 * nDimsWritten - (output) the number of dimensions of the array (the trailing ones are removed)
	 * dims - (output) array of the dimensions (size)
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::GetNDims(const qvalue_t &obj, int *nDimsWritten, int *dims)
	{
		if (isRefCountable[obj.type])
		{
			if (!context.GetNDims(obj.private_obj, dims, nDimsWritten))
				RaiseException();
		}
		else
			*nDimsWritten = 0;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetString
	 *
	 * Gets a string value associated with the given qvalue_t object. If the
	 * qvalue_t object does not represent a string, its value may be converted
	 * (for example from integer to string)
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::GetString(const qvalue_t &obj, string_t &ret)
	{
		if (obj.private_obj == NULL)
		{
			ret = string_t();
			return;
		}
		LPTSTR text;
		int length;
		if (!context.GetString(obj.private_obj, &text, &length))
			RaiseException();
		ret = string_t::unmarshal(text, length);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::TypeGetString
	 *
	 * Gets a textual representation of the given type handle.
	 *------------------------------------------------------------------------*/
	string_t QuasarHostImpl::TypeGetString(qvalue_t type)
	{
		Assert(type.type == TYPE_TYPEINFO &&
			   type.private_obj != NULL, "type: typeinfo expected");
		LPTSTR text;
		int length;
		if (!context.Type_GetString(type.private_obj, &text, &length))
			RaiseException();
		return string_t::unmarshal(text, length);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::TypeGetInfo
	 *
	 * Queries information about the specified type
	 *
	 * Parameters:
	 * type - qvalue_t type handle to obtain info.
	 * dataType - (output) the data type value of the type
	 * maxDims - (output) the maximal dimensionality of the type (in case of
	 *				an array
	 * vectorLength - (output) the length of the vector (in case of a vector)
	 * accessModifiers - access modifiers stored in the type
	 * isIntegerType - whether the type represents an integer (of any bitness)
	 * isPointer - whether the type represents a pointer type.
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::TypeGetInfo(qvalue_t type, DataTypes *dataType, int *maxDims, int *vectorLength, AccessModifiers *accessModifier,
		bool *isIntegerType, bool *isPointer)
	{
		Assert(type.type == TYPE_TYPEINFO &&
			type.private_obj != NULL, "type: typeinfo expected");
		if (!context.Type_GetInfo(type.private_obj, dataType, maxDims, vectorLength, accessModifier, isIntegerType, isPointer))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::TypeIsStatic
	 *
	 * Queries whether the specified type is a 'static' type. Static types
	 * are fully specified and their instances can be passed to kernel/device
	 * functions.
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::TypeIsStatic(qvalue_t type)
	{
		Assert(type.type == TYPE_TYPEINFO &&
			type.private_obj != NULL, "type: typeinfo expected");
		bool result = false;
		if (!context.Type_IsStatic(type.private_obj, &result))
			RaiseException();
		return result;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::TypeGetElementType
	 *
	 * Returns the element type of the specified type.
	 * For example vec[mat] -> mat
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::TypeGetElementType(qvalue_t type)
	{
		Assert(type.type == TYPE_TYPEINFO &&
			type.private_obj != NULL, "type: typeinfo expected");
		qvalue_t result;
		result.type = TYPE_TYPEINFO;
		//TODO
		//if (!context.Type_GetElementType(type.private_obj, &result.private_obj))
		//	RaiseException();
		return result;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::TypeGetElementTypeAtPosition
	 *
	 * Returns the element type of the specified type at the given position
	 * For example vec[{mat,cube}] and pos=1 -> cube
	 *
	 * This is useful for cell matrices with element types defined.
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::TypeGetElementTypeAtPosition(qvalue_t type, int dims, int* pos)
	{
		Assert(type.type == TYPE_TYPEINFO &&
			type.private_obj != NULL, "type: typeinfo expected");
		qvalue_t result;
		result.type = TYPE_TYPEINFO;
		if (!context.Type_GetElementTypeAtPosition(type.private_obj, dims, pos, &result.private_obj))
			RaiseException();
		return result;
	}
	/*------------------------------------------------------------------------
	 * QuasarHostImpl::TypeCompareToSpecification
	 *
	 * Checks if an object type has the given type specification.
	 *
	 * Parameters:
	 * type - qvalue_t object to check
	 * specType - qvalue_t containing type - the type specification to check
	 *			against
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::TypeCompareToSpecification(qvalue_t type, qvalue_t specType)
	{
		Assert(type.type == TYPE_TYPEINFO &&
			type.private_obj != NULL, "type: typeinfo expected");
		Assert(specType.type == TYPE_TYPEINFO &&
			specType.private_obj != NULL, "specType: typeinfo expected");

		bool result = false;
		if (!context.Type_CompareToSpecification(type.private_obj, specType.private_obj, &result))
			RaiseException();
		return result;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::AddFieldWithOffset
	 *
	 * Adds a field with a specific offset to a user defined type. The offset
	 * is usually obtained using the offsetof() macro.
	 * The type should not be finalized yet (function FinalizeType). Also see
	 * IQuasarHost::AddField.
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::AddFieldWithOffset(qhandle objType, LPCTSTR fieldName,
		qhandle fieldType, size_t offset)
	{
		if (!context.Type_AddFieldWithOffset(objType, fieldName, fieldType, offset))
			RaiseException();
		return true;
	}
	/*------------------------------------------------------------------------
	 * QuasarHostImpl::AddProperty
	 *
	 * Adds a property to the user defined type. A property consists of a get
	 * method and a set method. A property is similar to a field, with the
	 * difference that methods are used to get/set the value.
	 *
	 * Parameters:
	 * objType - handle of the user defined type
	 * propertyName - name of the property
	 * propertyType - handle of the type of the property (see LookupType)
	 * getProperty - a method (QLambdaDelegate) to get the value of the
	 *    property
	 * setProperty - a method (QLambdaDelegate) to set the value of the
	 *    property
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::AddProperty(qhandle objType, LPCTSTR propertyName, qhandle propertyType,
		qhandle getProperty, qhandle setProperty)
	{
		if (!context.Type_AddProperty(objType, propertyName, propertyType, getProperty, setProperty))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::InheritType
	 *
	 * Inherits all the fields, properties, methods from the base type.
	 *
	 * Parameters:
	 * objType - handle of the user defined type
	 * baseType - handle of the base type to inherit
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::InheritType(qhandle objType, qhandle baseType)
	{
		if (!context.Type_InheritType(objType, baseType))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetConstructor
	 *
	 * Sets a constructor to be called after an instance of the user-defined
	 * type is constructed.
	 *
	 * Parameters:
	 * objType - handle of the user defined type
	 * constructor - callback function to be called at object construction
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::SetConstructor(qhandle objType, ConstructorDelegate* constructor)
	{
		if (!context.Type_SetConstructor(objType, constructor))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetDynamicDispatcher
	 *
	 * Sets a dynamic method call handler, field get/set handler for an
	 * untyped object type. Function is mostly used for language interop.
	 *
	 * Parameters:
	 * objType - handle of the user defined type (dynamic class)
	 * disp - the type of dispatcher to set (method call, field get or field set)
	 * handler - callback function to be called for dynamic method calls
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::SetDynamicDispatcher(qhandle objType,
		DynamicDispatchers disp, DynamicCallDelegate *handler)
	{
		if (!context.Type_SetDynamicDispatcher(objType, disp, handler))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateFunctionType
	 *
	 * Creates a function type, based on the given set of input argument
	 * types and output argument types
	 *
	 * Parameters:
	 * functionType - type of the function (kernel, host or device)
	 * inputTypes - array of type handles of the input arguments
	 * numInputTypes - number of input arguments
	 * outputTypes - array of type handles of the output arguments
	 * numOutputTypes - number of output arguments
	 * flags - additional function flags (e.g. variadic)
	 *------------------------------------------------------------------------*/
	qhandle QuasarHostImpl::CreateFunctionType(FunctionTypes functionType,
		const qhandle* inputTypes, int numInputTypes,
		const qhandle* outputTypes, int numOutputTypes, FunctionTypeFlags flags)
	{
		qhandle handle;
		if (!context.CreateFunctionType(functionType, inputTypes, numInputTypes,
			outputTypes, numOutputTypes, flags, &handle))
			RaiseException();
		return handle;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateTypeClass
	 *
	 * Creates a type class, for example [int|cube]. An object belongs to
	 * a type class if it belongs to at leat one of its members.
	 *
	 * Parameters:
	 * elemTypes - array of type handles corresponding to the members of this
	 *    type class
	 * elemTypeCount - number of elements in the array 'elemTypes'.
	 *------------------------------------------------------------------------*/
	qhandle QuasarHostImpl::CreateTypeClass(const qhandle *elemTypes, int elemTypeCount)
	{
		qhandle handle;
		if (!context.CreateTypeClass(elemTypes, elemTypeCount, &handle))
			RaiseException();
		return handle;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::InstantiateType
	 *
	 * Creates an instance of a generic type, for example vec[cube]. Also works
	 * with user-defined generic types.
	 *
	 * Parameters:
	 * type - type handle of the generic type to be instantiated
	 * instance_type_params - the type parameters of the instance
	 * num_params - the number of type parameters (length of the array
	 *		instance_type_params)
	 *------------------------------------------------------------------------*/
	qhandle QuasarHostImpl::InstantiateType(qhandle type, const qhandle* instance_type_params, int num_params)
	{
		qhandle handle;
		if (!context.Type_InstantiateType(type, instance_type_params, num_params, &handle))
			RaiseException();
		return handle;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreatePointerType
	 *
	 * Creates a pointer type. Can only be used currently on mutable/immutable
	 * types. For example my_type -> ^my_type.
	 *
	 * Parameters:
	 * type - type handle of the type to create a pointer type from
	 *------------------------------------------------------------------------*/
	qhandle QuasarHostImpl::CreatePointerType(qhandle type)
	{
		qhandle handle;
		if (!context.CreatePointerType(type, &handle))
			RaiseException();
		return handle;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateTypeParameter
	 *
	 * Creates a type parameter that can be used for AddParameter.
	 *
	 * Parameters:
	 * paramName - name of the parameter
	 * paramType - the type of the parameter, or NULL in case there is no
	 *		type
	 *------------------------------------------------------------------------*/
	qhandle QuasarHostImpl::CreateTypeParameter(LPCTSTR paramName, qhandle paramType)
	{
		qhandle handle;
		if (!context.CreateTypeParameter(paramName, paramType, &handle))
			RaiseException();
		return NULL;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetMainProgram
	 *
	 * Sets the name of the main program
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::SetMainProgram(LPCTSTR mainProgramName)
	{
		this->mainProgramName = mainProgramName;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::IsMainProgram
	 *
	 * Checks whether the main program name == programName
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::IsMainProgram(LPCTSTR programName)
	{
		return mainProgramName == programName;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetProgramCmdLineArguments
	 * Reads the command line parameters when the program is executed natively
	 * and converts them to Quasar values, using the eval() function.
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::GetProgramCmdLineArguments(qvalue_t *argsOut, int numArgs, int numMandatoryArgs, bool variadic)
	{
		LPCTSTR *readArgs, *args;
		int nReadArgs, nArgs;

		BeginReadCmdLineParams(readArgs, nReadArgs);
		args = readArgs + cmdLineParamOffset;
		nArgs = nReadArgs - cmdLineParamOffset;

		// Check the number of arguments
		if (nArgs < numMandatoryArgs)
		{
			EndReadCmdLineParams(readArgs, nReadArgs);
			ThrowException(string_t::format(_T("Not enough parameters provided through the commandline (received: %d, required: %d)."),
				nArgs, numMandatoryArgs));
		}
		else if (nArgs > numArgs && !variadic)
		{
			EndReadCmdLineParams(readArgs, nReadArgs);
			ThrowException(string_t::format(_T("Too many parameters were passed through the commandline (received: %d, required: %d)."),
				nArgs, numArgs));
		}

		IEvaluationStack *stack = engine->GetEvaluationStack();

		int i, numVariadicArgs = 0;
		qvalue_t *variadicArgs = variadic ? new qvalue_t[nArgs - numArgs + 1] : NULL;
		for (i = 0; i < nArgs; i++)
		{
			string_t strVal = args[i];
			// Use the eval function for evaluating the command line parameters
			qvalue_t qstrVal = CreateString(strVal, strVal.get_length());
			stack->PushValue(qstrVal); // Reference will be destroyed by the consumer
			stack->FunctionCall(_T("eval"), 1);
			qvalue_t result = stack->PopValue();

			// Within the parameter set - write directly to the output
			if (i < numArgs || variadic && i < numArgs - 1)
				argsOut[i] = result;
			else
				variadicArgs[numVariadicArgs++] = result;
		}
		// Unspecified parameters we set to qvalue_t::empty
		for (i = nArgs; i < numArgs; i++)
			argsOut[i] = qvalue_t::empty;

		// Variadic function call - the last argument must be converted to a cell vector
		if (variadic)
		{
			for (i = 0; i < numVariadicArgs; i++)
				stack->PushValue(variadicArgs[i]);
			engine->ConstructCellMatrix(numVariadicArgs);
			argsOut[numArgs - 1] = stack->PopValue();
			delete[] variadicArgs;
		}

		EndReadCmdLineParams(readArgs, nReadArgs);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetProgramCmdLineArgumentOffset
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::SetProgramCmdLineArgumentOffset(int offset)
	{
		cmdLineParamOffset = offset;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CreateStack
	 *------------------------------------------------------------------------*/
	IEvaluationStack *QuasarHostImpl::CreateStack()
	{
		ThrowException(_T("CreateStack functionality is not implemented in this version."));
		return nullptr;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::QueryProperty
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::QueryProperty(HostProperties prop, int param, void** dst, size_t nBytes)
	{
		return context.QueryProperty(prop, param, dst, nBytes);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetPrimitiveTypeHandle
	 *------------------------------------------------------------------------*/
	qhandle QuasarHostImpl::GetPrimitiveTypeHandle(DataTypes type, LPCTSTR typeName)
	{
		if (typeName != nullptr)
		{
			switch (type)
			{
			case TYPE_NCUBE:
			case TYPE_NCCUBE:
				return context.ParseType(typeName); // Need the dimensionality
			}
		}
		return typeHandles[type];
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::Lock
	 *------------------------------------------------------------------------*/
	LockResults QuasarHostImpl::Lock(const qvalue_t &obj, DataTypes elemType,
		LockingModes lockingMode,
		MemResources memResource, void **pData)
	{
		return context.Lock(obj.private_obj, lockingMode, memResource, pData);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::Unlock
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::Unlock(const qvalue_t &obj, LockingModes lockingMode,
		MemResources memResource)
	{
		context.Unlock(obj.private_obj, lockingMode, memResource);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::EnumVariables
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::EnumVariables(EnumVariableDelegate* callback, void *userData)
	{
		if (!context.EnumVariables(callback, userData))
			QuasarHostImpl::GetInstance()->RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::ReadVariable
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::ReadVariable(LPCTSTR varName, qvalue_t &val)
	{
		return context.ReadVariable(varName, &val);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::WriteVariable
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::WriteVariable(LPCTSTR varName, const qvalue_t &val)
	{
		return context.WriteVariable(varName, &val);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetField
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::GetField(const qvalue_t &obj, LPCTSTR fieldName,
		qvalue_t &val)
	{
		Assert(obj.type == TYPE_UNTYPEDOBJECT || obj.type == TYPE_TYPEDOBJECT,
			"obj: object expected");
		if (!context.GetField(obj.private_obj, fieldName, &val))
		{
			QuasarHostImpl::GetInstance()->RaiseException();
			return false;
		}
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetField
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::SetField(const qvalue_t &obj, LPCTSTR fieldName,
		const qvalue_t &val)
	{
		Assert(obj.type == TYPE_UNTYPEDOBJECT || obj.type == TYPE_TYPEDOBJECT,
			"obj: object expected");
		if (!context.SetField(obj.private_obj, fieldName, &val))
			QuasarHostImpl::GetInstance()->RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LoadSourceModule
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::LoadSourceModule(LPCTSTR sourceModuleFileName,
		LPCTSTR *errorMsg)
	{
		LPCTSTR msg;
		bool result = context.LoadSourceModule(sourceModuleFileName, errorMsg, NULL);
		// If errorMsg == nullptr and the function failed, throw an exception
		if (errorMsg == nullptr && !result)
			QuasarHostImpl::GetInstance()->RaiseException();
		return result;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LoadBinaryModule
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::LoadBinaryModule(LPCTSTR binaryModuleName,
		LPCTSTR *errorMsg)
	{
		LPCTSTR msg;
		bool result = context.LoadBinaryModule(binaryModuleName, errorMsg != nullptr ? errorMsg : &msg, NULL);
		// If errorMsg == nullptr and the function failed, throw an exception
		if (errorMsg == nullptr && !result)
			QuasarHostImpl::GetInstance()->RaiseException();
		return result;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LoadModuleFromSource
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::LoadModuleFromSource(LPCTSTR moduleName,
		LPCTSTR sourceString,
		LPCTSTR *errorMsg)
	{
		return context.LoadModuleFromSource(moduleName, sourceString, errorMsg,
			NULL);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::UnloadModule
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::UnloadModule(LPCTSTR moduleName)
	{
		return context.UnloadModule(moduleName);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::EnableProfiling
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::EnableProfiling(ProfilingModes profilingMode,
		LPCTSTR outputFileName)
	{
		return context.EnableProfiling(profilingMode, outputFileName);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetComputationEngine
	 *------------------------------------------------------------------------*/
	IComputationEngine *QuasarHostImpl::GetComputationEngine()
	{
		return engine;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::RaiseException
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::RaiseException()
	{
		LPTSTR source, errorMsg, stackTrace;
		int sourceLen, errorMsgLen, stackTraceLen;
		if (context.GetLastError(&source, &sourceLen, &errorMsg, &errorMsgLen, &stackTrace, &stackTraceLen))
		{
			exception_t ex;
			ex.source = string_t::unmarshal(source, sourceLen);
			ex.message = string_t::unmarshal(errorMsg, errorMsgLen);
			ex.stack_trace = string_t::unmarshal(stackTrace, stackTraceLen);

			if (custom_exceptionhandler == NULL || !custom_exceptionhandler(ex))
				throw ex;
		}
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::SetCustomExceptionHandler
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::SetCustomExceptionHandler(custom_exceptionhandler_t handler)
	{
		this->custom_exceptionhandler = handler;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetCustomExceptionHandler
	 *------------------------------------------------------------------------*/
	custom_exceptionhandler_t QuasarHostImpl::GetCustomExceptionHandler()
	{
		return this->custom_exceptionhandler;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::RegisterModule
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::RegisterModule(RegisterModuleOptions options, LPCTSTR moduleName,
		LPCTSTR fileName, RegisterModuleFlags flags)
	{
		if (!context.RegisterModule(options, moduleName, fileName, flags))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::RegisterModuleFromData
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::RegisterModuleFromData(RegisterModuleOptions options, LPCTSTR moduleName,
		LPCVOID data, size_t dataSize, RegisterModuleFlags flags)
	{
		if (!context.RegisterModuleFromData(options, moduleName, data, dataSize, flags))
			RaiseException();
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::LoadFunction
	 *------------------------------------------------------------------------*/
	qvalue_t QuasarHostImpl::LoadFunction(qvalue_t closureArgs, LPCTSTR moduleName, LPCTSTR functionSignature, LPCTSTR funcName,
		FunctionFlags flags, PlatformTargets platformTargets, LoadFunctionParams &params)
	{
		qvalue_t result;
		if (!context.LoadFunction(&result, closureArgs.private_obj, moduleName, functionSignature, funcName, flags,
			platformTargets, &params))
			RaiseException();
		return result;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetMachineInfo
	 *------------------------------------------------------------------------*/
	MachineInfo QuasarHostImpl::GetMachineInfo()
	{
		MachineInfo m;
		if (!context.GetMachineInfo(&m)) RaiseException();
		return m;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::FreeMachineInfo
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::FreeMachineInfo(MachineInfo &machineInfo)
	{
		if (!context.FreeMachineInfo(&machineInfo))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::AddLambdaDestroyCallback
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::AddLambdaDestroyCallback(qvalue_t lambda, DisposeClosure* disposeDelegate)
	{
		Assert(lambda.type == TYPE_LAMBDAEXPR || lambda.type == TYPE_METHOD,
			"lambda - lambda expression or value expected");
		if (!context.AddLambdaDestroyCallback(lambda.private_obj, disposeDelegate))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::EnumRegisteredTypes
	 *
	 * Enumerates all types currently registered in Quasar
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::EnumRegisteredTypes(EnumRegisteredTypeDelegate *callback, void* userData)
	{
		if (!context.EnumRegisteredTypes(callback, userData))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::EnumFields
	 *
	 * Enumerates all types currently registered in Quasar
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::EnumFields(const qvalue_t &obj, EnumFieldDelegate *callback, void *userData)
	{
		if (!context.EnumFields(&obj, callback, userData))
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::GetRawSizeOfType
	 *
	 * Returns the raw size of a type in memory
	 *------------------------------------------------------------------------*/
	size_t QuasarHostImpl::GetRawSizeOfType(qhandle typeHandle, MemResources memResource, size_t *alignment)
	{
		size_t size;
		if (!context.GetRawSizeOfType(typeHandle, memResource, &size, alignment))
			RaiseException();
		return size;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::TypeGetClassInfo
	 *
	 * Gets class information for the specified type
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::TypeGetClassInfo(qhandle typeHandle, string_t &moduleName, string_t &typeName, ClassInfoFlags &flags,
		MemResources* resources, int *numResources, size_t* size, qhandle *baseClass, int* numGenericTypeParams, qhandle *genericTypeParams,
		int *numInstanceTypeParams, qhandle *instanceTypeParams)
	{
		LPCTSTR pmoduleName = nullptr;
		LPCTSTR ptypeName = nullptr;
		if (!context.Type_GetClassInfo(&pmoduleName, &ptypeName, typeHandle, &flags, resources, numResources,
			size, baseClass, numGenericTypeParams, genericTypeParams, numInstanceTypeParams, instanceTypeParams))
			RaiseException();
		moduleName = pmoduleName;
		typeName = ptypeName;
		context.FreeString(pmoduleName);
		context.FreeString(ptypeName);
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::IsRebindRequired
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::IsRebindRequired(qvalue_t lambda, int token)
	{
		return false; // Currently rebinding is not supported by this runtime implementation
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::OpenDevice
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::OpenDevice(LPCTSTR deviceId, const RuntimeSettings *runtimeSettings)
	{
		if (!context.OpenDevice(deviceId, runtimeSettings))
			RaiseException();

		Initialized();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::CloseDevice
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::CloseDevice()
	{
		Uninitialize();

		if (!context.CloseDevice())
			RaiseException();
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::AddHostDestroyCallback
	 *
	 * Registers a function to be called when the host is destroyed
	 *------------------------------------------------------------------------*/
	bool QuasarHostImpl::AddHostDestroyCallback(HostDestroyCallbackDelegate *callback, void *userData)
	{
		destroyCallbacks.push_back(std::make_tuple(userData, callback));
		return true;
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::OnException
	 *
	 * OnException should be called from within a shared library function,
	 * because native C++ exceptions cannot be passed directly to the C++ host
	 * (in particular in linux)
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::OnException(const exception_t &exception)
	{
		tprintf(_T("C++ exception caught: %ls\n"), (LPCTSTR)exception.message);
		Q_RAISE_ERROR("A C++ exception was caught!");
	}

	/*------------------------------------------------------------------------
	 * QuasarHostImpl::ThrowException
	 *------------------------------------------------------------------------*/
	void QuasarHostImpl::ThrowException(string_t msg)
	{
		exception_t ex;
		ex.source = _T("Quasar host");
		ex.message = msg;
		ex.stack_trace = string_t();
		throw ex;
	}

	//=============================================================================
	// quasar::bindings implementation
	//=============================================================================

	/*------------------------------------------------------------------------
	 * bindings::setup_kernel_function
	 *
	 * Sets up CPU and CUDA kernel functions
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::setup_kernel_function(
		qvalue_t qv_kernel,
		std::function<KernelLambdaDelegate> cpu_kernel_adapter,
		std::function<KernelLambdaDelegate> cuda_kernel_adapter,
		std::function<PrimeKernelLambdaDelegate> cuda_prime_kernel_adapter)
	{
		IQuasarHost::GetInstance()->SetupKernelFunction(qv_kernel, cpu_kernel_adapter,
			cuda_kernel_adapter, cuda_prime_kernel_adapter);
	}

	/*------------------------------------------------------------------------
	 * bindings::setup_device_function
	 *
	 * Sets up the CPU device implementation
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::setup_device_function(
		qvalue_t qv_device,
		std::function<LambdaDelegate> cpu_device_adapter,
		void *cpu_device_func_ptr)
	{
		IQuasarHost::GetInstance()->SetupDeviceFunction(qv_device, cpu_device_adapter,
			cpu_device_func_ptr);
	}

	/*------------------------------------------------------------------------
	 * bindings::is_kernel_initialized
	 *
	 * Checks whether the specified kernel is initialized (i.e. the CPU
	 * and GPU kernels are set)
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI bool bindings::is_kernel_initialized(compute_kernel *kernel)
	{
		return IQuasarHost::GetInstance()->IsKernelInitialized((qhandle)kernel);
	}

	/*------------------------------------------------------------------------
	 * bindings::launch_kernel
	 *
	 * Launches the specified kernel
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::launch_kernel(compute_kernel * kernel, LPCVOID params, size_t params_size,
		const void *closure_ptr, const qvalue_t * args, int3 gridDim, int3 blockDim,
		bool prime_kernel, int deviceId, execution_stream *stream)
	{
		IQuasarHost::GetInstance()->LaunchKernel((qhandle)kernel, params, params_size,
			closure_ptr, args, gridDim, blockDim, prime_kernel, deviceId, stream);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_marshal_fixed_length_array
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_marshal_fixed_length_array(int deviceId, const qvalue_t& in, void* dataOut, size_t bytes)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		MemResources memRes = host->GetMemResourceForDevice(deviceId);
		LPVOID ptr = host->GetDataPtr((qvalue_t*)&in, memRes);
		std::memcpy(dataOut, ptr, bytes);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_marshal_vector
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_marshal_vector(int deviceId, const qvalue_t& in, void*& data, int& dim1)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		MemResources memRes = host->GetMemResourceForDevice(deviceId);
		data = host->GetDataPtr((qvalue_t*)&in, memRes);

		int dims[16];
		int ndims = 16;
		IQuasarHost::GetInstance()->GetNDims((qvalue_t &)in, &ndims, dims);
		int prod = dims[0];
		for (int i = 1; i < ndims; i++) prod *= dims[i];
		dim1 = prod;
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_marshal_matrix
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_marshal_matrix(int deviceId, const qvalue_t& in, void*& data, int& dim1, int& dim2)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		MemResources memRes = host->GetMemResourceForDevice(deviceId);
		data = host->GetDataPtr((qvalue_t*)&in, memRes);

		int dims[16];
		int ndims = 16;
		IQuasarHost::GetInstance()->GetNDims((qvalue_t &)in, &ndims, dims);
		dim1 = dims[0];
		dim2 = dims[1];
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_marshal_cube
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_marshal_cube(int deviceId, const qvalue_t& in, void*& data, int& dim1, int& dim2, int& dim3)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		MemResources memRes = host->GetMemResourceForDevice(deviceId);
		data = host->GetDataPtr((qvalue_t*)&in, memRes);

		int dims[16];
		int ndims = 16;
		IQuasarHost::GetInstance()->GetNDims((qvalue_t &)in, &ndims, dims);
		dim1 = dims[0];
		dim2 = dims[1];
		dim3 = dims[2];
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_marshal_ncube
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_marshal_ncube(int deviceId, const qvalue_t& in, void*& data, int* dims, int N)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		MemResources memRes = host->GetMemResourceForDevice(deviceId);
		data = host->GetDataPtr((qvalue_t*)&in, memRes);
		IQuasarHost::GetInstance()->GetNDims((qvalue_t &)in, &N, dims);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_marshal_lambda
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_marshal_lambda(int devid, const qvalue_t& in, void*& func_ptr, void*& closure_ptr)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->GetLambdaFunctionPtr((qvalue_t *)&in, devid, &func_ptr, &closure_ptr);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_marshal_obj
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void* bindings::internal_marshal_obj(int devid, const qvalue_t& in)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		MemResources memRes = host->GetMemResourceForDevice(devid);
		return host->GetDataPtr((qvalue_t*)&in, memRes);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_vecx
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_vecx(int deviceId, qvalue_t& qv, const void* data, int dim1, size_t elemSize)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->LockVector<uint8_t>(qv, LOCK_WRITE, MEMRESOURCE_CPU);
		void *ptr = host->GetDataPtr(&qv, MEMRESOURCE_CPU);
		std::memcpy(ptr, data, dim1 * elemSize);
		host->UnlockVector(qv, LOCK_WRITE, MEMRESOURCE_CPU);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_vector
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_vector(int deviceId, qvalue_t& qv, const void* ptr, int dim1)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->UnmarshalPointer(&qv, ptr, &dim1, sizeof(int), TYPE_VEC, deviceId, 0);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_matrix
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_matrix(int deviceId, qvalue_t& qv, const void* ptr, int dim1, int dim2)
	{
		int dims[2] = { dim1, dim2 };
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->UnmarshalPointer(&qv, ptr, dims, 2*sizeof(int), TYPE_MAT, deviceId, 0);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_cube
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_cube(int deviceId, qvalue_t& qv, const void* ptr, int dim1, int dim2, int dim3)
	{
		int dims[3] = { dim1, dim2, dim3 };
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->UnmarshalPointer(&qv, ptr, dims, 3 * sizeof(int), TYPE_MAT, deviceId, 0);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_ncube
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_ncube(int deviceId, qvalue_t& qv, const void* ptr, const int *dims, int N)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->UnmarshalPointer(&qv, ptr, dims, N * sizeof(int), TYPE_NCUBE, deviceId, 0);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_lambda
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_lambda(int deviceId, qvalue_t& qv, const void* ptr, const void *data)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->UnmarshalPointer(&qv, ptr, &data, sizeof(void*), TYPE_LAMBDAEXPR, deviceId, 0);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_obj
	 * Unmarshalling of objects specified using a pointer
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_obj(int deviceId, qvalue_t& qv, void *ptr)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->UnmarshalPointer(&qv, ptr, nullptr, 0, TYPE_TYPEDOBJECT, deviceId, 0);
	}

	/*------------------------------------------------------------------------
	 * bindings::internal_unmarshal_typed_object
	 * Unmarshalling of objects specified by value
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void bindings::internal_unmarshal_typed_object(int deviceId,
		const void *ptr, size_t size, LPCTSTR typeName, qvalue_t &out)
	{
		IQuasarHost *host = IQuasarHost::GetInstance();
		host->UnmarshalTypedObject(&out, ptr, size, typeName, deviceId, 0);
	}

	//=============================================================================
	// String class implementation
	//=============================================================================

	/*------------------------------------------------------------------------
	 * string_t::set_string
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI void string_t::set_string(const TCHAR *chars, size_t length)
	{
		if (length == (size_t)-1)
			length = tcslen(chars);

#if defined(WIN32) || defined(_WIN32)
		TCHAR * ptr = (TCHAR *)CoTaskMemAlloc(sizeof(TCHAR) * (length + 1));
#else
		TCHAR * ptr = new TCHAR[length + 1];
#endif
		memcpy(ptr, chars, length * sizeof(TCHAR));
		ptr[length] = '\0'; // trailing zero, for C compatibility reasons
		this->length = length;
		this->chars = ptr;
	}

	/*------------------------------------------------------------------------
	 * string_t::string_t
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t::string_t(size_t length)
	{
#if defined(WIN32) || defined(_WIN32)
		TCHAR * ptr = (TCHAR *)CoTaskMemAlloc(sizeof(TCHAR) * (length + 1));
#else
		TCHAR * ptr = new TCHAR[length + 1];
#endif
		this->chars = ptr;
		this->length = length;
	}

	/*------------------------------------------------------------------------
	 * string_t::string_t
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t::string_t(TCHAR ch, size_t repeat) :string_t(repeat)
	{
		TCHAR * ptr = const_cast<TCHAR *>(chars);
		for (int i = 0; i < repeat; i++)
			ptr[i] = ch;
		ptr[repeat] = '\0'; // trailing zero, for C compatibility reasons
	}

	/*------------------------------------------------------------------------
	 * string_t::operator=
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t &string_t::operator=(const string_t &src)
	{
#if defined(WIN32) || defined(_WIN32)
		if (chars != nullptr) CoTaskMemFree((LPVOID)chars);
#else
		if (chars != nullptr) delete[] chars;
#endif
		set_string(src.chars, src.length);
		return *this;
	}

#ifdef CPLUSPLUS11
	/*------------------------------------------------------------------------
	 * string_t::operator=
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t &string_t::operator=(string_t && other)
	{
#if defined(WIN32) || defined(_WIN32)
		if (chars != nullptr) CoTaskMemFree((LPVOID)chars);
#else
		if (chars != nullptr) delete[] chars;
#endif
		*(void **)this = *(void **)&other; // copy vtable pointer (temp hack)
		chars = other.chars;
		length = other.length;
		other.chars = nullptr;
		other.length = 0;
		return *this;
	}
#endif

	/*------------------------------------------------------------------------
	 * string_t::format
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::format(const TCHAR *fmt, ...)
	{
		TCHAR buffer[1024];
		va_list argptr;
		va_start(argptr, fmt);
		int len = vstprintf(buffer, COUNT_OF(buffer), fmt, argptr);
		va_end(argptr);
		return string_t(buffer, len);
	}

	/*------------------------------------------------------------------------
	 * string_t::operator+
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::operator+(const string_t &b) const
	{
		string_t result(length + b.length);
		std::memcpy(const_cast<LPTSTR>(&result.chars[length]), &b[0], b.length * sizeof(TCHAR));
		return result;
	}

	/*------------------------------------------------------------------------
	 * string_t::unmarshal
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::unmarshal(const TCHAR *chars, size_t length)
	{
		string_t y(length);
		TCHAR *ptr = const_cast<TCHAR *>(y.chars);
		ptr[length] = '\0'; // trailing zero, for C compatibility reasons
		memcpy(ptr, chars, length * sizeof(TCHAR));
		return y;
	}

	/*------------------------------------------------------------------------
	 * string_t::operator<
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI bool string_t::operator< (const string_t &b) const
	{
		// Empty string
		if (chars == nullptr)
			return b.chars != nullptr;
		if (b.chars == nullptr)
			return false;
		return tstrcmp(chars, b.chars) < 0;
	}

	/*------------------------------------------------------------------------
	 * string_t::operator>
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI bool string_t::operator> (const string_t &b) const
	{
		// Empty string
		if (chars == nullptr)
			return false;
		if (b.chars == nullptr)
			return true;
		return tstrcmp(chars, b.chars) > 0;
	}

	/*------------------------------------------------------------------------
	 * string_t::operator==
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI bool string_t::operator== (const string_t &b) const
	{
		return length == b.length &&
			(length == 0 || tstrncmp(chars, b.chars, length) == 0);
	}

	/*------------------------------------------------------------------------
	 * string_t::operator== (LPCTSTR b)
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI bool string_t::operator== (LPCTSTR b) const
	{
		// The first condition accounts for the fact that chars==nullptr
		// for an empty string
		return length == 0 && *b == '\0' ||
			length > 0 && tstrcmp(chars, b) == 0;
	}

	/*------------------------------------------------------------------------
	 * string_t::get_hash_code
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI size_t string_t::get_hash_code() const
	{
		return 0;
	}

	/*------------------------------------------------------------------------
	 * string_t::~string_t
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t::~string_t()
	{
#if defined(WIN32) || defined(_WIN32)
		if (chars != nullptr) CoTaskMemFree((LPVOID)chars);
#else
		if (chars != nullptr) delete[] chars;
#endif

		chars = nullptr;
	}

	/*------------------------------------------------------------------------
	 * string_t::pad_left
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::pad_left(size_t count, TCHAR pad) const
	{
		if (length > count)
			return *this;

		string_t s(pad, count);
		std::memcpy(const_cast<TCHAR*>(s.chars) + (count - length), chars, length * sizeof(TCHAR));
		return s;
	}

	/*------------------------------------------------------------------------
	 * string_t::pad_right
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::pad_right(size_t count, TCHAR pad) const
	{
		if (length > count)
			return *this;

		string_t s(pad, count);
		std::memcpy(const_cast<TCHAR*>(s.chars), chars, length * sizeof(TCHAR));
		return s;
	}

	/*------------------------------------------------------------------------
	 * string_t::sub_string
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::sub_string(size_t start, size_t length) const
	{
		_ASSERT_EXPR(start <= this->length, "string_t::sub_string: index out of bounds!");
		length = min(length, this->length - start);
		return string_t(&chars[start], length);
	}

	/*------------------------------------------------------------------------
	 * string_t::is_whitespace
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI bool string_t::is_whitespace(char ch)
	{
		return ch == ' ' || ch == '\t';
	}

	/*------------------------------------------------------------------------
	 * string_t::trim
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::trim(TrimModes trim_mode) const
	{
		size_t start = 0;
		size_t end = length - 1;
		if (trim_mode == TRIM_LEFT || trim_mode == TRIM_BOTH)
		{
			while (start < length && string_t::is_whitespace(chars[start])) start ++;
		}
		if (trim_mode == TRIM_RIGHT || trim_mode == TRIM_BOTH)
		{
			while (end >= 0 && string_t::is_whitespace(chars[end])) end --;
		}
		return string_t(chars + start, end - start + 1);
	}

	/*------------------------------------------------------------------------
	 * string_t::change_case
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t string_t::change_case(StringCases string_case) const
	{
		string_t result(length);
		TCHAR *ptr = const_cast<TCHAR *>(result.chars);
		switch (string_case)
		{
		case LOWER:
			for (size_t i = 0; i < length; i ++)
				ptr[i] = totlower(chars[i]);
			break;
		case UPPER:
			for (size_t i = 0; i < length; i ++)
				ptr[i] = totupper(chars[i]);
			break;
		}
		return result;
	}

	/*------------------------------------------------------------------------
	 * Version::ToString
	 *------------------------------------------------------------------------*/
	QUASARHOSTAPI string_t Version::ToString() const
	{
		return string_t::format(_T("%d.%d.%d.%04d"), major, minor,
			revision < 0 ? 0 : revision, build < 0 ? 0 : build);
	}

	//=============================================================================
	//=============================================================================

#ifdef MS_DOTNET

	/*------------------------------------------------------------------------
	 * IQuasarHost::Create
	 *------------------------------------------------------------------------*/
	IQuasarHost *IQuasarHost::Create(LPCTSTR deviceName, bool loadCompiler)
	{
		HRESULT hr;

		// Support for �Quasar runtime system: load Helios.dll dynamically
		if (deviceName != nullptr &&
			tstrncmp(deviceName, _T("helios:"), 7) == 0)
		{
			HMODULE hmod = LoadLibrary(_T("Helios.dll"));
			if (hmod == NULL) // Could not load library
			{
				LPVOID lpMsgBuf = NULL;

				FormatMessage(
					FORMAT_MESSAGE_ALLOCATE_BUFFER |
					FORMAT_MESSAGE_FROM_SYSTEM |
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,
					GetLastError(),
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					(LPTSTR)&lpMsgBuf,
					0,
					NULL
				);
				string_t err_msg((LPCTSTR)lpMsgBuf);
				LocalFree(lpMsgBuf);
				QuasarHostImpl::ThrowException(err_msg);
			}
			typedef IQuasarHost *IQuasarHost_Create(LPCTSTR, bool);

			IQuasarHost_Create * addr = (IQuasarHost_Create *)GetProcAddress(hmod, "?Create@IQuasarHost@quasar@@SAPEAU12@PEB_W_N@Z");
			if (addr == NULL)
				QuasarHostImpl::ThrowException(_T("Entry IQuasarHost::Create could not be found in Helios.dll!"));

			IQuasarHost *host = (*addr)(deviceName + 7, loadCompiler);
			host->AddHostDestroyCallback([](void *closurePtr, IQuasarHost *host) {
				QuasarHostImpl::SetInstance(NULL);
			}, NULL);
			QuasarHostImpl::SetInstance(host);
			return host;
		}

		PCWSTR pszVersion =
			_T("v4.0.30319"); // The version of the .NET runtime to load
		PCWSTR pszAssemblyName = _T("mscorlib");
		PCWSTR pszTypeName = loadCompiler ? _T("Quasar.NativeInterfaceExt") : _T("Quasar.NativeInterface");
		PCWSTR pszMethodName = _T("InitHost");
		PCWSTR pszArgument = _T("");
		wchar_t pwzAssemblyPath[MAX_PATH];
		size_t returnSize;

		// Read the Quasar directory from the environment variables
		_wgetenv_s(&returnSize, pwzAssemblyPath, _T("QUASAR_PATH"));
		if (loadCompiler)
			lstrcat(pwzAssemblyPath, _T("\\Quasar.exe"));
		else
			lstrcat(pwzAssemblyPath, _T("\\Quasar.Runtime.dll"));

		ICLRMetaHost *pMetaHost = NULL;
		ICLRRuntimeInfo *pRuntimeInfo = NULL;

		// Instance already exists
		if (QuasarHostImpl::GetInstance() != NULL)
		{
			QuasarHostImpl::ThrowException(string_t::format(_T("Cannot create two Quasar host interfaces. Use ")
				_T("IQuasarHost.GetInstance() ")
				_T("to obtain a pointer to the current instance.\n")));
		}

		// ICorRuntimeHost and ICLRRuntimeHost are the two CLR hosting
		// interfaces
		// supported by CLR 4.0. Here we demo the ICLRRuntimeHost interface that
		// was provided in .NET v2.0 to support CLR 2.0 new features.
		// ICLRRuntimeHost does not support loading the .NET v1.x runtimes.
		ICLRRuntimeHost *pClrRuntimeHost = NULL;
		IUnknownPtr spAppDomainThunk = NULL;
		_AppDomainPtr spDefaultAppDomain = NULL;

		// The .NET assembly to load.
		bstr_t bstrAssemblyName(pszAssemblyName);
		_AssemblyPtr spAssembly = NULL;

		//
		// Load and start the .NET runtime.
		//
		tprintf(_T("Load and start the .NET runtime %s \n"), pszVersion);

		hr = CLRCreateInstance(CLSID_CLRMetaHost, IID_PPV_ARGS(&pMetaHost));
		if (FAILED(hr))
		{
			QuasarHostImpl::ThrowException(string_t::format(_T("CLRCreateInstance failed w/hr 0x%08lx\n"), hr));
		}

		// Get the ICLRRuntimeInfo corresponding to a particular CLR version. It
		// supersedes CorBindToRuntimeEx with STARTUP_LOADER_SAFEMODE.
		hr = pMetaHost->GetRuntime(pszVersion, IID_PPV_ARGS(&pRuntimeInfo));
		if (FAILED(hr))
		{
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("ICLRMetaHost::GetRuntime failed w/hr 0x%08lx\n"), hr));
		}

		// Check if the specified runtime can be loaded into the process. This
		// method will take into account other runtimes that may already be
		// loaded into the process and set pbLoadable to TRUE if this runtime
		// can be loaded in an in-process side-by-side fashion.
		BOOL fLoadable;
		hr = pRuntimeInfo->IsLoadable(&fLoadable);
		if (FAILED(hr))
		{
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("ICLRRuntimeInfo::IsLoadable failed w/hr 0x%08lx\n"), hr));
		}

		if (!fLoadable)
		{
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T(".NET runtime %s cannot be loaded\n"), pszVersion));
		}

		// Load the CLR into the current process and return a runtime interface
		// pointer. ICorRuntimeHost and ICLRRuntimeHost are the two CLR hosting
		// interfaces supported by CLR 4.0. Here we demo the ICLRRuntimeHost
		// interface that was provided in .NET v2.0 to support CLR 2.0 new
		// features. ICLRRuntimeHost does not support loading the .NET v1.x
		// runtimes.
		hr = pRuntimeInfo->GetInterface(CLSID_CLRRuntimeHost,
			IID_PPV_ARGS(&pClrRuntimeHost));
		if (FAILED(hr))
		{
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("ICLRRuntimeInfo::GetInterface failed w/hr 0x%08lx\n"), hr));
		}

		// Use ICLRRuntimeHost to get the CLR control interface.
		ICLRControl *pCLRControl = NULL;
		hr = pClrRuntimeHost->GetCLRControl(&pCLRControl);
		if (FAILED(hr))
		{
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("ICLRRuntimeHost::GetCLRControl failed w/hr 0x%08lx\n"), hr));
		}

		// Use ICLRControl to get a pointer to the failure policy
		// manager.
		ICLRPolicyManager* pCLRPolicyManager = NULL;
		hr = pCLRControl->GetCLRManager(IID_ICLRPolicyManager,
			(PVOID*)&pCLRPolicyManager);
		if (FAILED(hr))
		{
			RELEASE_SAFE(pCLRControl);
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("ICLRControl::GetCLRManager failed w/hr 0x%08lx\n"), hr));
		}

		hr = pCLRPolicyManager->SetUnhandledExceptionPolicy(eRuntimeDeterminedPolicy);
		if (FAILED(hr))
		{
			RELEASE_SAFE(pCLRPolicyManager);
			RELEASE_SAFE(pCLRControl);
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("ICLRPolicyManager::SetUnhandledExceptionPolicy failed w/hr 0x%08lx\n"), hr));
		}

		// Start the CLR.
		hr = pClrRuntimeHost->Start();
		if (FAILED(hr))
		{
			RELEASE_SAFE(pCLRPolicyManager);
			RELEASE_SAFE(pCLRControl);
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("CLR failed to start w/hr 0x%08lx\n"), hr));
		}

		HANDLE hMappedFileObject = NULL;
		LPVOID lpvSharedMem = NULL;

		hMappedFileObject =
			CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0,
				sizeof(QuasarContext), TEXT("quasar_init"));

		if (hMappedFileObject == NULL)
		{
			RELEASE_SAFE(pCLRPolicyManager);
			RELEASE_SAFE(pCLRControl);
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("Failed to create file mapping\n"), hr));
		}

		lpvSharedMem =
			MapViewOfFile(hMappedFileObject, FILE_MAP_READ | FILE_MAP_WRITE, 0,
				0, sizeof(QuasarContext));
		if (lpvSharedMem == NULL)
		{
			RELEASE_SAFE(pCLRPolicyManager);
			RELEASE_SAFE(pCLRControl);
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("Could not map view of file\n"), hr));
		}

		QuasarContext *params = (QuasarContext *)lpvSharedMem;
		memset(params, 0, sizeof(QuasarContext));
		if (deviceName == NULL)
			stprintf(params->deviceName, ARRAY_LENGTH(params->deviceName, TCHAR), _T(""));
		else
			stprintf(params->deviceName, ARRAY_LENGTH(params->deviceName, TCHAR), _T("%s"), deviceName);
#ifdef DBL_SCALAR
		params->scalarType = SCALARTYPE_DOUBLE;
#else
		params->scalarType = SCALARTYPE_SINGLE;
#endif
		stprintf(params->runtimeName, ARRAY_LENGTH(params->runtimeName, TCHAR), _T("Quasar .Net runtime"));
		params->structSize = sizeof(QuasarContext);
		params->hostPtr = 0;
		UnmapViewOfFile(lpvSharedMem);

		DWORD returnValue;
		hr = pClrRuntimeHost->ExecuteInDefaultAppDomain(
			pwzAssemblyPath, pszTypeName, pszMethodName, pszArgument,
			&returnValue);
		if (FAILED(hr))
		{
			RELEASE_SAFE(pCLRPolicyManager);
			RELEASE_SAFE(pCLRControl);
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("Failed to init Quasar w/hr 0x%08lx\n"), hr));
		}

		// Read back the parameters
		lpvSharedMem =
			MapViewOfFile(hMappedFileObject, FILE_MAP_READ | FILE_MAP_WRITE, 0,
				0, sizeof(QuasarContext));
		params = (QuasarContext *)lpvSharedMem;

		if (returnValue > 0 ||
			tcslen(params->errorMsg) > 0)
		{
			RELEASE_SAFE(pCLRPolicyManager);
			RELEASE_SAFE(pCLRControl);
			RELEASE_SAFE(pClrRuntimeHost);
			RELEASE_SAFE(pRuntimeInfo);
			RELEASE_SAFE(pMetaHost);
			QuasarHostImpl::ThrowException(string_t::format(_T("Could not load the Quasar runtime: %ls\n"), params->errorMsg));
		}

		QuasarHostImpl *hostImpl = new QuasarHostImpl(
			*params, pMetaHost, pRuntimeInfo, pClrRuntimeHost, NULL);
		UnmapViewOfFile(lpvSharedMem);
		CloseHandle(hMappedFileObject);

		return hostImpl;
	}

	/*------------------------------------------------------------------------
	 * BeginReadCmdLineParams
	 *------------------------------------------------------------------------*/
	void BeginReadCmdLineParams(LPCTSTR *&args, int &nArgs)
	{
		LPCTSTR cmdLine = GetCommandLine();
		args = (LPCTSTR *)CommandLineToArgvW(cmdLine, &nArgs);
		if (NULL == args)
		{
			QuasarHostImpl::ThrowException(_T("Could not access the command line parameters\n"));
		}
	}

	/*------------------------------------------------------------------------
	 * EndReadCmdLineParams
	 *------------------------------------------------------------------------*/
	void EndReadCmdLineParams(LPCTSTR *args, int nArgs)
	{
		LocalFree(args);
	}


#else // MONO_RUNTIME
	/*------------------------------------------------------------------------
	 * defined
	 *------------------------------------------------------------------------*/
#if defined(WIN32) || defined(_WIN32)
#define PATH_SEPARATOR "\\"
#else
#define PATH_SEPARATOR "/"
#endif

	static void fail_with(const char *fmt, ...)
	{
		va_list argptr;
		va_start(argptr, fmt);
		fprintf(stderr, "Failed to initialize the MONO runtime: ");
		vfprintf(stderr, fmt, argptr);
		fprintf(stderr, "\n");
		va_end(argptr);
		exit(1);
	}

	//
	// Parameter capture problem - the mono environment needs access to the
	// command line parameters. Using a GCC trick, we capture them here. later
	/*------------------------------------------------------------------------
	 * mono_runtime_set_main_args
	 *------------------------------------------------------------------------*/
	 // they can be passed to the runtime using mono_runtime_set_main_args().
	 //
	static int stored_argc;
	static char **stored_argv;
	__attribute__((constructor)) void capture_main_args(int argc, char **argv)
	{
		stored_argc = argc;
		stored_argv = argv;
	}

	/*------------------------------------------------------------------------
	 * BeginReadCmdLineParams
	 *------------------------------------------------------------------------*/
	static void BeginReadCmdLineParams(LPCTSTR *&args, int &nArgs)
	{
		args = (LPCTSTR *)stored_argv;
		nArgs = stored_argc;
		if (NULL == args)
		{
			QuasarHostImpl::ThrowException(_T("Could not access the command line parameters\n"));
		}
	}
	/*------------------------------------------------------------------------
	 * EndReadCmdLineParams
	 *------------------------------------------------------------------------*/
	static void EndReadCmdLineParams(LPCTSTR *args, int nArgs) {}

	IQuasarHost *IQuasarHost::Create(LPCTSTR deviceName, bool loadCompiler)
	{
		// Note: due to a bug in older versions of Mono, this requires Mono 5.0 or later!
		//
		// See https://github.com/mono/mono/blob/master/samples/embed/test-invoke.c
		// For some examples on how to use the MONO API.

		const char *assemblyName = loadCompiler ? "Quasar.exe" : "Quasar.Runtime.dll";
		const char *nameSpace = "Quasar";
		const char *typeName = loadCompiler ? "NativeInterfaceExt" : "NativeInterface";
		const char *methodName = "InitHost";
		QuasarHostImpl *hostImpl = NULL;

		// Support for µQuasar runtime system: load libhelios.so dynamically
		if (deviceName != nullptr &&
			tstrncmp(deviceName, _T("helios:"), 7) == 0)
		{
			void *lib = dlopen("libhelios.so", RTLD_LAZY);
			void *errPtr = dlerror();
			if (errPtr != nullptr)
			{
				QuasarHostImpl::ThrowException(string_t::format(_T("Device '%ls' could not be loaded: %s\n"), deviceName, (const char *)errPtr));
			}
			typedef IQuasarHost *IQuasarHost_Create(LPCTSTR, bool);
			IQuasarHost_Create *create = (IQuasarHost_Create *)dlsym(lib, "_ZN6quasar11IQuasarHost6CreateEPKwb");
			if (create == NULL)
				QuasarHostImpl::ThrowException(_T("Entry IQuasarHost::Create could not be found in Helios.dll!"));

			IQuasarHost *host = (*create)(deviceName + 7, loadCompiler);
			host->AddHostDestroyCallback([](void *closurePtr, IQuasarHost *host) {
				QuasarHostImpl::SetInstance(NULL);
			}, NULL);
			QuasarHostImpl::SetInstance(host);
			return host;
		}

		tprintf(_T("Loading MONO runtime...\n"));

		char *mono_path = getenv("MONO_PATH");

		if (mono_path == NULL)
		{
			fail_with("MONO_PATH environment variable needs to be set in order to locate mono!");
		}

		if (strstr(mono_path,"mono") == NULL)
		{
			fail_with("MONO_PATH environment variable needs to include \"mono\"!");
		}

		char mono_prefix_path[1024];
		int mono_path_len = (int)(strstr(mono_path,"mono") - mono_path + strlen("mono"));
		strncpy(mono_prefix_path, mono_path, mono_path_len);
		mono_prefix_path[mono_path_len] = '\0';

		char mono_etc_path[1024];
		char mono_lib_path[1024];

		sprintf(mono_etc_path, "%s" PATH_SEPARATOR "etc", mono_prefix_path);
		sprintf(mono_lib_path, "%s" PATH_SEPARATOR "lib", mono_prefix_path);

		// set lib & etc directory. It is necessay when uing the bundled mono.
		mono_set_dirs(mono_lib_path, mono_etc_path);

		//
		// Load the default Mono configuration file, this is needed
		// if you are planning on using the dllmaps defined on the
		// system configuration
		//
		mono_config_parse(NULL);

		//
		// mono_jit_init() creates a domain: each assembly is
		// loaded and run in a MonoDomain.
		MonoDomain *domain = mono_jit_init_version(assemblyName, "v4.0.30319");

		char *quasar_path = getenv("QUASAR_PATH");
		char assembly_path[1024];
		char config_path[1024];

		if (quasar_path == NULL)
		{
			fail_with("QUASAR_PATH environment variable needs to be set in order to locate Quasar!");
		}

		sprintf(assembly_path, "%s" PATH_SEPARATOR "%s",
			quasar_path,
			assemblyName);
		sprintf(config_path, "%s.config", assemblyName);

		// set the configuration file
		mono_domain_set_config(domain, quasar_path, config_path);

		// Enable debugging
		mono_debug_init(MONO_DEBUG_FORMAT_MONO);
		mono_debug_domain_create(domain);

		// Workaround for the environment command-line parameters bug
		mono_runtime_set_main_args(stored_argc, stored_argv);

		tprintf(_T("Loading Quasar runtime...\n"));
		MonoAssembly *assembly = mono_domain_assembly_open(domain, assembly_path);
		if (!assembly)
			fail_with("Could not load assembly: '%s'. Make sure the QUASAR_PATH "
				"environment variable is set!\n", assembly_path);

		// Invoke the method
		MonoImage *image = mono_assembly_get_image(assembly);
		if (!image)
			fail_with("Could not get image for assembly!");

		MonoClass *nativeInterfaceClass = mono_class_from_name(image, nameSpace, typeName);
		if (!nativeInterfaceClass)
			fail_with("Could not find class '%s.%s'!", nameSpace, typeName);

		// Need to execute the main function, because the EntryAssembly is set there...
		MonoException* exc;

		MonoMethod* main = mono_class_get_method_from_name(nativeInterfaceClass, "Main", -1);
		if (!main)
			fail_with("Could not find Main function in '%s.%s'!", nameSpace, typeName);
		MonoArray* pargs = mono_array_new(mono_domain_get(), mono_get_string_class(), 0);
		mono_runtime_exec_main(main, pargs, (MonoObject**)&exc);

		MonoMethod *initHost = mono_class_get_method_from_name(nativeInterfaceClass, methodName, 1);
		if (!initHost)
			fail_with("Could not find value '%s%s'!", typeName, methodName);

		// Setup memory mapped file
		int fd = open("/tmp/quasar_init", O_CREAT | O_RDWR, (mode_t)0700);
		if (fd == -1)
			fail_with("Could not create memory mapped file: %s!", strerror(errno));

		// set the size of the shared memory object
		int result = ftruncate(fd, sizeof(QuasarContext));
		if (result == -1)
			fail_with("Could not truncate file: %s!", strerror(errno));

		QuasarContext *params = (QuasarContext *)mmap(0, sizeof(QuasarContext),
			PROT_WRITE, MAP_SHARED, fd, 0);
		if (params == (QuasarContext *)-1)
			fail_with("mmap failed: %s!", strerror(errno));

		memset(params, 0, sizeof(QuasarContext));
		tprintf(_T("Device name: %ls\n"), deviceName);
		if (deviceName == NULL)
			stprintf(params->deviceName, COUNT_OF(params->deviceName), _T(""));
		else
			stprintf(params->deviceName, COUNT_OF(params->deviceName), _T("%ls"), deviceName);

#ifdef DBL_SCALAR
		params->scalarType = SCALARTYPE_DOUBLE;
#else
		params->scalarType = SCALARTYPE_SINGLE;
#endif
		munmap(params, sizeof(QuasarContext));
		close(fd);

		MonoString* value = mono_string_new(domain, "");
		void* args[1] = { value };
		MonoError error;

		tprintf(_T("Calling Quasar host...\n"));
		//mono_error_raise_exception (&error);
		MonoObject * pException = NULL;
		mono_runtime_invoke(initHost, NULL, args, &pException);

		if (pException != NULL)
		{
			fail_with("Exception caught: %s!",
				mono_string_to_utf8(mono_object_to_string(pException, NULL)));
		}

		fd = open("/tmp/quasar_init", O_RDWR, (mode_t)0700);
		if (fd == -1)
			fail_with("Could not open memory mapped file: %s!", strerror(errno));
		params = (QuasarContext *)mmap(0, sizeof(QuasarContext),
			PROT_READ, MAP_SHARED, fd, 0);

		if (tcslen(params->errorMsg) > 0)
		{
			tprintf(_T("Could not load the Quasar runtime: %ls\n"), params->errorMsg);
			goto Cleanup;
		}

		hostImpl = new QuasarHostImpl(*params, NULL, NULL, NULL, domain);

		munmap(params, sizeof(QuasarContext));
		close(fd);

		tprintf(_T("Host successfully created.\n"));

		return hostImpl;

	Cleanup:
		mono_jit_cleanup(domain);
		return NULL;
	}

#endif

	/*------------------------------------------------------------------------
	 * IQuasarHost::Connect
	 *------------------------------------------------------------------------*/
	 // Function called from initialization code in Quasar user modules (via the
	 // module_init function)
	IQuasarHost *IQuasarHost::Connect(LPCVOID hostprivParams,
		LPCVOID moduleDetails, int flags)
	{
		// Instance already exists
		if (QuasarHostImpl::GetInstance() != NULL)
		{
			printf("Cannot create two Quasar host interfaces. Use "
				"IQuasarHost.GetInstance() "
				"to obtain a pointer to the current instance.\n");
			return NULL;
		}

		QuasarContext *context = (QuasarContext *)hostprivParams;
		if (context->structSize < sizeof(QuasarContext))
		{
			printf("The Quasar host has an older version. Please update the host to be compatible with this module.\n");
			return NULL;
		}

		// Make sure the close function will not be called (otherwise
		// we may end up in an infinite recursion when terminating)!
		context->Close = NULL;

		// We can reuse the IQuasarHost object
		IQuasarHost *host;
		if (context->hostPtr != 0)
		{
			host = (IQuasarHost *)context->hostPtr;
			host->AddRef();
			QuasarHostImpl::SetInstance(host);
		}
		else
			host = new QuasarHostImpl(*(QuasarContext *)hostprivParams, NULL, NULL,
				NULL, NULL);
		return host;
	}

	/*------------------------------------------------------------------------
	 * IQuasarHost::GetInstance
	 *------------------------------------------------------------------------*/
	IQuasarHost *IQuasarHost::GetInstance(void)
	{
		return QuasarHostImpl::GetInstance();
	}

	/*------------------------------------------------------------------------
	 * IQuasarHost::GetRuntimeControl
	 *------------------------------------------------------------------------*/
	bool IQuasarHost::GetRuntimeControl(qvalue_t &obj)
	{
		return ReadVariable(_T("$rt"), obj);
	}

	/*------------------------------------------------------------------------
	 * IQuasarHost::Synchronize
	 *------------------------------------------------------------------------*/
	void IQuasarHost::Synchronize()
	{
		GetComputationEngine()->Synchronize();
	}

	namespace runtime
	{
		/*------------------------------------------------------------------------
		 * Debug assertions
		 *------------------------------------------------------------------------*/
		QUASARHOSTAPI void dbgAssert(const char *expr, const char *src_file, int line, const char *fmt, ...)
		{
			char buffer[1024];
			va_list argptr;
			va_start(argptr, fmt);
			// Visual C++: use 'safe' string formatting functions
#ifdef _MSC_VER
			int len = sprintf_s(buffer, "%s - line %d: %s\n\n", src_file, line, expr);
			len += vsprintf_s(buffer + len, sizeof(buffer) - len, fmt, argptr);
#else
			int len = sprintf(buffer, "%s - line %d: %s\n\n", src_file, line, expr);
			len += vsprintf(buffer + len, fmt, argptr);
#endif
			fprintf(stdout, "%s\n", buffer);
			va_end(argptr);

			_ASSERT_EXPR(false, buffer);
		}

		/*------------------------------------------------------------------------
		 * Debug warnings
		 *------------------------------------------------------------------------*/
		QUASARHOSTAPI void dbgWarning(const char *src_file, int line, const char *fmt, ...)
		{
			va_list argptr;
			va_start(argptr, fmt);
			// Visual C++: use 'safe' string formatting functions
#ifdef _MSC_VER
			printf_s("%s - line %d:\n", src_file, line);
			vprintf_s(fmt, argptr);
			printf_s("\n");
#else
			printf("%s - line %d:\n", src_file, line);
			vprintf(fmt, argptr);
			printf("\n");
#endif
			va_end(argptr);
		}

	} // namespace quasar::runtime

} // namespace quasar
