//=============================================================================
// Quasar Low-Level Host interface - access Quasar programs from C++
// (C) 2011-2020 Gepura
//=============================================================================
#ifndef __QUASAR_HOST_H__
#define __QUASAR_HOST_H__

#include "quasar.h"
#include "quasar_ncube.h"
#include "char16_string.h"

#if defined(_MSC_VER)
#include <crtdbg.h>
#else
#include <cassert>
#define _ASSERT(a) assert(a)
#define _ASSERT_EXPR(a,b) assert(a)
#endif

#ifdef CPLUSPLUS11
#include <initializer_list>
#include <functional>
#endif

typedef const void *LPCVOID;
typedef void *LPVOID;

#ifndef interface
#define __STRUCT__ struct
#define interface __STRUCT__
#endif

#define COUNT_OF(x) ((sizeof(x)/sizeof(0[x])) / ((size_t)(!(sizeof(x) % sizeof(0[x])))))

#ifdef QUASARHOST_DLLIMPORT
#define QUASARHOSTAPI __declspec(dllimport)
#elif defined(QUASARHOST_DLLEXPORT)
#define QUASARHOSTAPI __declspec(dllexport)
#else
#define QUASARHOSTAPI
#endif

namespace quasar
{
	// Some forward declaration
	class string_t;
	interface IComputationEngine;
	interface IEvaluationStack;
	interface IRuntimeReductionEngine;
	interface ITypeEnvironment;
	interface IQuasarHost;


	// Quasar internal data types
	enum DataTypes
	{
		TYPE_VOID,
		TYPE_SCALAR, // A scalar number (single or double precision, depending
					 // on whether DBL_SCALAR is defined)
		TYPE_COMPLEXSCALAR, // A complex-valued scalar number
		TYPE_INT,           // 32-bit integer
		TYPE_VEC,           // Variable length vector (of type 'vec')
		TYPE_MAT,           // Variable size matrix (of type 'mat')
		TYPE_CUBE,          // Variable size cube (of type 'cube')
		TYPE_CVEC,          // Variable length complex vector (of type 'cvec')
		TYPE_CMAT,          // Variable size complex matrix (of type 'cmat')
		TYPE_CCUBE,         // Variable size complex cube (of type 'ccube')
		TYPE_STRING,        // Variable length string
		TYPE_TYPEINFO,      // Type information
		TYPE_LAMBDAEXPR,    // Lambda expression
		TYPE_INT8,
		TYPE_INT16,
		TYPE_INT64,
		TYPE_UINT8,
		TYPE_UINT16,
		TYPE_UINT32,
		TYPE_UINT64,
		TYPE_TYPEDOBJECT,
		TYPE_UNTYPEDOBJECT,
		TYPE_NCUBE,			// Variable dimension & size cube
		TYPE_NCCUBE,		// Variable dimension & size complex cube
		TYPE_HALFSCALAR,		// 16-bit floating point value
		TYPE_SINGLESCALAR,	// 32-bit floating point value
		TYPE_DOUBLESCALAR,	// 64-bit floating point value
		TYPE_HALFCOMPLEXSCALAR,	// 32-bit half precision complex number
		TYPE_SINGLECOMPLEXSCALAR, // 64-bit single precision complex number
		TYPE_DOUBLECOMPLEXSCALAR, // 128-bit double precision complex number
		TYPE_METHOD,
		TYPE_NAMEDPARAM,

		TYPE_CHAR = 0x40,
		TYPE_LIST = 0x42,
		TYPE_LITE_OBJECT = 0x43,
		TYPE_CONST_STRING = 0x46,
		TYPE_UINT128 = 0x47,
		TYPE_INT128 = 0x48,
		TYPE_COUNT,

		TYPE_INT32 = TYPE_INT,
	};

	// Flags for locking arrays/matrices etc.
	enum LockingModes
	{
		LOCK_NONE = 0,
		LOCK_READ = 1,
		LOCK_WRITE = 2,
		LOCK_READWRITE = 3
	};

	// Memory resource flags (for identifying the resource that has to be
	// locked)
	enum MemResources
	{
		MEMRESOURCE_MANAGED = 0,
		MEMRESOURCE_CPU = 1,
		MEMRESOURCE_SINGLE_CUDA =
		2, // Selects the CUDA memory resource (if available)
		MEMRESOURCE_SINGLE_OPENCL =
		3, // Selects the OpenCL memory resource (if available)
		MEMRESOURCE_DEVICE0 = 0x10000,
		// Explicit resources
		MEMRESOURCE_DEVICE1 = MEMRESOURCE_DEVICE0 + 1,
		MEMRESOURCE_DEVICE2 = MEMRESOURCE_DEVICE0 + 2,
		MEMRESOURCE_DEVICE3 = MEMRESOURCE_DEVICE0 + 3,
		MEMRESOURCE_DEVICE4 = MEMRESOURCE_DEVICE0 + 4
	};

	// The results of a locking operation
	enum LockResults
	{
		LOCKRESULT_OK = 0,
		LOCKRESULT_RES_NOT_AVAILABLE = 1,
		LOCKRESULT_OUT_OF_MEM = 2,
		LOCKRESULT_INUSE = 3,
		LOCKRESULT_INVALID
	};

	// Flags for allocating memory
	enum AllocationFlags
	{
		ALLOCATIONFLAGS_NONE = 0,
		ALLOCATIONFLAGS_FORCEGPULOAD =
		2, // Forces the memory to be transferred to the (primary) GPU
		ALLOCATIONFLAGS_FORCECPULOAD =
		4 // Forces the memory to be transferred to the (primary) CPU
	};

	// Represents the data type used for 'scalar' in Quasar
	enum ScalarTypes
	{
		SCALARTYPE_SINGLE = 0, // 32-bit single precision floating point
		SCALARTYPE_DOUBLE = 1, // 64-bit double precision floating point
		SCALARTYPE_HALF = 2,   // 16-bit half precision floating point
	};

	// Numerical operations in Quasar
	enum OperatorTypes
	{
		OP_ADD = 0,         // addition
		OP_SUB = 1,         // subtraction
		OP_MULTIPLY = 2,    // multiplication (matrix multiplication)
		OP_DIVIDE = 3,      // divide
		OP_RDIVIDE = 4,     // right division
		OP_POW = 5,         // power
		OP_PW_MULTIPLY = 6, // point-wise multiplication
		OP_PW_DIVIDE = 7,   // point-wise division
		OP_PW_POW = 8,      // point-wise power
		OP_LOG_XOR = 9,     // logical XOR operation
		OP_LOG_OR = 10,      // logical OR
		OP_LOG_AND = 11,     // logical AND
		OP_MIN = 12,        // minimum
		OP_MAX = 13,        // maximum
		OP_LESS = 14,        // less than
		OP_LESSOREQ = 15,   // less than or equal
		OP_GREATER = 16,    // greater
		OP_GREATEROREQ = 17, // greater or equal
		OP_EQUAL = 18,       // equal
		OP_NOTEQUAL = 19,    // not equal
		OP_ASSIGN = 20,      // assignment
		OP_NEGATE = 21,      // negation (-)
		OP_INVERT = 22,      // logical inversion (!)
		OP_DOTDOT = 23,      // sequence (a..b)
		OP_DOTDOTDOT = 24,   // sequence with step (a..b..c)
		OP_SPREAD = 25,          // spread operator ... (variadic arguments)
		OP_LAMBDADEFSTART = 26,   // start of a lambda expression  ->
		OP_LAMBDADEFEND = 27,     // end of a lambda expression
		OP_ADD_ASSIGN = 28,         // inplace assignment +=
		OP_SUB_ASSIGN = 29,         // inplace subtraction -=
		OP_MULTIPLY_ASSIGN = 30,    // inplace multiplication *=
		OP_DIVIDE_ASSIGN = 31,      // inplace division /=
		OP_RDIVIDE_ASSIGN = 32,     // inplace left multiplication \=
		OP_POW_ASSIGN = 33,         // inplace power ^=
		OP_PW_MULTIPLY_ASSIGN = 34, // pointwise inplace multiplication .*=
		OP_PW_DIVIDE_ASSIGN = 35,   // pointwise inplace division ./=
		OP_PW_POW_ASSIGN = 36,      // pointwise inplace power .^=
		OP_XOR_ASSIGN = 37,         // pointwise inplace XOR ~=
		OP_OR_ASSIGN = 38,      // pointwise inplace OR |=
		OP_AND_ASSIGN = 39,     // pointwise inplace AND &=
		OP_MIN_ASSIGN = 40,     // pointwise inplace MIN __=
		OP_MAX_ASSIGN = 41,     // pointwise inplace MAX ^^=
		OP_COND_IF = 42,     // ternary operator - conditional IF ? :
		OP_SWITCH_CASE = 43, // switch case
		OP_DEF_ASSIGN = 44,      // reserved operator (:=)
		OP_PIPELINE_RIGHT = 45,  // reserved operator |>
		OP_PIPELINE_LEFT = 46,   // reserved operator <|
		OP_LAZYEXPR_START = 47,  // reserved
		OP_LAZYEXPR_END = 48,    // reserved
		OP_WHERE = 49,           // reserved operator 'where'
		OP_REMAINDER = 50,       // reserved
		OP_SHIFTLEFT = 51,       // reserved
		OP_SHIFTRIGHT = 52,      // reserved
		OP_COMPAREEXCHANGE = 53, // compare & exchange operator
		OP_EXCHANGE = 54,		 // exchange operator
		OP_COUNT				 // reserved

	};

	enum AccessModifiers
	{
		ACCESS_NONE = 0,            // default: "checked" outside kernel functions, "safe" inside kernel functions
		ACCESS_CIRCULAR = 1,        // Circular boundary extension (for matrices)
		ACCESS_MIRROR = 2,          // Mirror boundary extension (for matrices)
		ACCESS_CLAMPED = 4,         // Clamped boundary extension (for matrices)
		ACCESS_UNCHECKED = 8,       // Matrices: no bounds checking, Integers: no overflow checking
		ACCESS_CHECKED = 0x10,      // Matrices: bounds checking, Integers: overflow checking
		ACCESS_SAT = 0x20,          // Integer: saturated access
		ACCESS_CONST = 0x40,        // Matrices: denotes that their value cannot change
		ACCESS_HWTEX_NEAREST = 256, // Use a hardware texturing unit in nearest neigbor interpolation mode
		ACCESS_HWTEX_LINEAR = 512,  // Use a hardware texturing unit in linear interpolation mode
		ACCESS_HWTEX_2C = 0x10000,  // 2-component hardware texturing
		ACCESS_HWTEX_4C = 0x20000,  // 4-component hardware texturing
		ACCESS_HWTEX_CONST = 0x40000,// CUDA non-coherent texture cache using __ldg() instruction
		ACCESS_SAFE = 1024,         // Matrices: zero extension
		ACCESS_HWCONST = 2048,      // Use the device constant memory for this variable
		ACCESS_NOCOPY = 4096,       // Run-time optimization: don't copy the data to the device
	};

	enum ProfilingModes
	{
		PROFILE_NONE = 0,					   // No profiling
		PROFILE_EXECUTIONTIME = 2,			   // Profiles the execution time of functions (both CPU/GPU)
		PROFILE_MEMLEAKS = 3,				   // Detects memory leaks and gives more information on these leaks
		PROFILE_ACCURACY = 4,				   // Profiles the accuracy of the operations
	};

	enum HostProperties
	{
		OPENCL_CURRENT_CONTEXT = 0x1000,	   // The current OpenCL context
		OPENCL_COMMANDQUEUE = 0x1001		   // The current OpenCL command queue object
	};

	enum FunctionFlags
	{
		FF_NONE = 0,
		FF_USESHAREDMEMORY = 2,                // The function uses shared memory (__kernel__ only)
		FF_USESYNCHRONIZATION = 4,             // The function uses thread synchronization (__kernel__ only)
		FF_USEDYNAMICMEMORY = 8,               // The function uses dynamic memory (__kernel__ or __device__ functions)
		FF_CHECKERROR = 0x40,                  // Checks for errors when the function finishes (__kernel__ only)
		FF_UNIVARIATE = 0x1000,                // Flag used by kernel functions with one dimension
		FF_BIVARIATE = 0x2000,                 // Flag used by kernel functions with two dimensions
		FF_USEFUNCTIONPOINTERS = 0x20000,	   // Function pointers are being used internally
		FF_NOPRIMEFUNCTION = 0x100000,		   // Kernel function has no prime version (for irregular dimensions)
		FF_USECTGRIDSYNCHRONIZATION = 0x200000,	  // Kernel uses cooperative threading with grid synchronization
		FF_USECTMULTIGRIDSYNCHRONIZATION = 0x400000, // Kernel uses cooperative multi-grid threading
		FF_USEREGULARGRIDSYNCHRONIZATION = 0x800000, // Kernel uses "regular" grid synchronization
		FF_NUMBLOCKSPOWOFTWO = 0x1000000,	   // Kernel should be launched with number of blocks being a power of two
	};

	enum PlatformTargets
	{
		TARG_NONE = 0,
		TARG_CPU = 1,                          // Generic CPU target
		TARG_CUDA = 2,						   // CUDA target
		TARG_NVIDIA_CUDA = 2,                  // NVidia CUDA PTX target
		TARG_NVIDIA_OPENCL = 4,                // NVidia OpenCL PTX target
		TARG_AMD_OPENCL = 8,                   // AMD OpenCL target
		TARG_GENERIC_OPENCL = 16,              // Generic OpenCL target (vendor-independent)
		TARG_XILINX = 128,					   // Xilinx FPGA device
		TARG_DEFAULT = 0x7fffffff,

		// All NVidia targets
		TARG_NVIDIA = TARG_NVIDIA_CUDA | TARG_NVIDIA_OPENCL,
		// All AMD targets
		TARG_AMD = TARG_AMD_OPENCL,
		// All GPU targets
		TARG_GPU = TARG_NVIDIA_CUDA | TARG_NVIDIA_OPENCL | TARG_AMD_OPENCL | TARG_GENERIC_OPENCL,
		// All OpenCL targets
		TARG_OPENCL = TARG_NVIDIA_OPENCL | TARG_AMD_OPENCL | TARG_GENERIC_OPENCL,
	};

	enum RegisterModuleOptions
	{
		RM_CALLING_MODULE = 0,					// Register the current calling module
		RM_CPU_MODULE = 1,						// Register the specified CPU module
		RM_PTX_MODULE = 2,						// Register the specified PTX module
		RM_OPENCL_MODULE = 3					// Register the specified OpenCL module
	};

	enum RegisterModuleFlags
	{
		RMF_NONE = 0,
		RMF_LZMA_COMPRESSION = 1,				// Module is stored with LZMA compression
		RMF_LOAD_CPUMODULE = 2,					// Generate an exception when the module could not be found
		RMF_LOAD_CUDAMODULE = 2,				// Generate an exception when the module could not be found
	};

	enum CreateTypeFlags
	{
		TYPEFLAGS_NONE = 0,						// Default class
		TYPEFLAGS_MUTABLE_CLASS = 0,			// Mutable class
		TYPEFLAGS_IMMUTABLE_CLASS = 1,			// Immutable class
		TYPEFLAGS_DYNAMIC_CLASS = 2,			// Dynamic class
	};

	enum DynamicDispatchers
	{
		DD_METHODCALL = 0,
		DD_FIELDGET = 1,
		DD_FIELDSET = 2
	};

	enum FunctionTypes
	{
		FUNCTIONTYPE_HOST = 1,
		FUNCTIONTYPE_DEVICE = 5,
		FUNCTIONTYPE_KERNEL = 2
	};

	enum FunctionTypeFlags
	{
		FTF_NONE = 0,
		FTF_VARIADIC = 1
	};

	enum LoadFunctionParamsSet
	{
		LFP_NONE = 0,
		LFP_COMPLEXITY_SET = 1,
		LFP_SHAREDMEMSIZE_SET = 2
	};

	struct LoadFunctionParams
	{
		LoadFunctionParamsSet flags;
		int complexity;
		int sharedMemSize;
	};

	enum Platforms
	{
		PLATFORM_WINDOWS,
		PLATFORM_UNIX_LINUX,
		PLATFORM_MACOS
	};

	enum ComputationDeviceTypes
	{
		DEVTYPE_CPU,						// CPU device
		DEVTYPE_CUDA,						// CUDA device
		DEVTYPE_HYPERION,					// Hyperion (hybrid) device
		DEVTYPE_OPENCL						// OpenCL device
	};

	enum CUDAUnifiedMemoryModes
	{
		CUDAUNIFIED_OFF = 0,
		CUDAUNIFIED_GLOBAL = 1, // Use unified memory for everything
	};

	enum GLAntialiasingModes
	{
		GL_ANTIALIASING_OFF = 0,
		GL_ANTIALIASING2X = 2,
		GL_ANTIALIASING4X = 4,
		GL_ANTIALIASING8X = 8,
		GL_ANTIALIASING16X = 16,
	};

	enum GPUMemoryModel
	{
		GPUMM_SMALLFOOTPRINT,
		GPUMM_MEDIUMFOOTPRINT,
		GPUMM_LARGEFOOTPRINT
	};

	enum GPUSchedulingMode
	{
		GPUSM_MAXIMIZEPERFORMANCE,
		GPUSM_MAXIMIZESTABILITY
	};

	enum RuntimeSettingsFlags : unsigned long long
	{
		RSF_ENABLE_EXCEPTION_ON_NAN_SET = 2,
		RSF_ENABLE_EXCEPTION_OF_INF_SET = 4,
		RSF_MAX_CONCURRENT_KERNELS_SET = 8,
		RSF_CUDA_UNIFIEDMEMORYMODE_SET = 16,
		RSF_GPU_MEMORYMODEL_SET = 32,
		RSF_GPU_SCHEDULINGMODE_SET = 64,
		RSF_KERNEL_DYNAMICMEMORY_SET = 128,
		RSF_KERNEL_DYNAMICMEMORY_MAXBLOCKSIZE_SET = 0x100,
		RSF_USEHW16BITFLOATTEXTURES_SET = 0x200,
		RSF_ENABLEBOUNDCHECKS_SET = 0x400,
		RSF_ENABLE_OPENGL = 0x800,
		RSF_GLANTIALIASINGMODE_SET = 0x1000,
		RSF_CUDA_PINNEDHOSTMEMORY_SET = 0x2000,
		RSF_DISABLE_GTK_SET = 0x4000,
	};

	struct RuntimeSettings
	{
		int structSize;
		bool loadCompiler;
		ScalarTypes scalarType;
		bool enableConcurrentKernelExecution;
		RuntimeSettingsFlags flags; // The following properties require the corresponding
									// flags to be set
		bool enableExceptionOnNaN;
		bool enableExceptionOnInf;
		int maxConcurrentKernels;
		CUDAUnifiedMemoryModes cudaUnifiedMemoryMode;
		GPUMemoryModel gpuMemoryModel;
		GPUSchedulingMode gpuSchedulingMode;
		int kernelDynamicMemory;
		int kernelDynamicMemoryMaxBlockSize;
		bool useHw16bitFloatTextures;
		bool enableBoundsChecks;
		bool enableOpenGL;
		bool enableCudaPinnedHostMemory;
		GLAntialiasingModes glAntialiasingMode;
		bool disableGtk;
	};

	struct Version
	{
		int major;
		int minor;
		int build;
		int revision;

		QUASARHOSTAPI string_t ToString() const;
	};

	struct CPUInfo
	{
		int numCPUcores;
		TCHAR cpuName[512 / sizeof(TCHAR)];
		Platforms platform;
		Version platformVersion;
		int archBits; // x86 or x64
		long long totalMemory;
	};

	struct OpenCLPlatformInfo
	{
		TCHAR name[512 / sizeof(TCHAR)];
		TCHAR vendor[512 / sizeof(TCHAR)];
		TCHAR version[512 / sizeof(TCHAR)];
	};

	struct OpenCLInfo
	{
		int numPlatforms;
		const OpenCLPlatformInfo *const* platforms;
		bool installed;
	};

	struct CUDAInfo
	{
		Version driverVersion;
		Version runtimeVersion;
		int archBits;
		bool installed;
	};

	struct ComputationDeviceInfo
	{
		ComputationDeviceTypes target;
		int ordinal;
		TCHAR deviceId[1024 / sizeof(TCHAR)];
		TCHAR name[512 / sizeof(TCHAR)];
		TCHAR shortName[512 / sizeof(TCHAR)];
		TCHAR vendor[512 / sizeof(TCHAR)];
		TCHAR driverVersion[512 / sizeof(TCHAR)];
		long totalMemory;
		bool available;
		TCHAR deviceConfigFile[1024 / sizeof(TCHAR)];
	};

	struct MachineInfo
	{
		TCHAR hostName[512 / sizeof(TCHAR)];
		TCHAR domainName[512 / sizeof(TCHAR)];
		TCHAR ipAddress[512 / sizeof(TCHAR)];
		CPUInfo cpu;
		CUDAInfo cuda;
		OpenCLInfo opencl;
		int numComputationDevices;
		const ComputationDeviceInfo *const* computationDevices;
	};

	enum ClassInfoFlags
	{
		CIF_NONE = 0,
		CIF_REQUIRESMARSHALLING = 1,    // Marshalling required for this class
		CIF_BLITTABLE = 2,				// Class is blittable
		CIF_UNTYPEDOBJECT = 4,          // Instances of this type correspond to untyped (dynamic) objects
		CIF_IMMUTABLE = 32              // Class cannot be modified
	};

	//========================
	// Handling of enum flags
	//========================
#define DEFINE_ENUMOR(enum_t)	\
		INLINE enum_t operator| (enum_t a, enum_t b) { return \
			static_cast<enum_t>(static_cast<int>(a) | static_cast<int>(b));}

	DEFINE_ENUMOR(FunctionFlags);
	DEFINE_ENUMOR(PlatformTargets);
	DEFINE_ENUMOR(RegisterModuleFlags);
	DEFINE_ENUMOR(LoadFunctionParamsSet);
#undef DEFINE_ENUMOR

	class exception_t;

	//======================================================================
	// Quasar custom exception handler. When 'true' is returned, the
	// program exception is continued. When 'false' is returned, the
	// program is halted (by throwing the C++ exception)
	//======================================================================
	typedef bool(*custom_exceptionhandler_t)(const exception_t &ex);

	typedef void *qhandle; // internal quasar handle

	enum TrimModes
	{
		TRIM_BOTH,
		TRIM_LEFT,
		TRIM_RIGHT
	};

	enum StringCases
	{
		LOWER,
		UPPER
	};

	//
	// String type
	//
	class string_t
	{
	private:
		size_t length;
		const TCHAR *chars;
	public:
		string_t() :length(0), chars(NULL) {}
		string_t(const string_t &src) { set_string(src.chars, src.length); }
		string_t(const TCHAR *chars) { set_string(chars, tcslen(chars)); }
		string_t(const TCHAR *chars, size_t length) { set_string(chars, length); }
		QUASARHOSTAPI string_t(size_t length);
#ifdef CPLUSPLUS11
		string_t(string_t && other) : chars(other.chars), length(other.length) { other.chars = NULL; }
#endif
		QUASARHOSTAPI string_t(TCHAR ch, size_t repeat);
		QUASARHOSTAPI string_t &operator=(const string_t &src);
		static QUASARHOSTAPI string_t unmarshal(const TCHAR *chars, size_t length);
		static QUASARHOSTAPI string_t format(const TCHAR *fmt, ...);
		QUASARHOSTAPI virtual ~string_t();
		operator const TCHAR *() const { return chars; }
		LPCTSTR get_buf() const { return chars; }
		size_t get_length() const { return length; }
		const TCHAR& operator[](size_t index) const {
			_ASSERT_EXPR(index >= 0 && index < length, "String index out of bounds");
			return chars[index];
		}
		LPTSTR find_char(TCHAR chr, size_t start_index = 0) const { return tcschr(chars + start_index, chr); }
		QUASARHOSTAPI bool operator< (const string_t &b) const;
		QUASARHOSTAPI bool operator> (const string_t &b) const;
		QUASARHOSTAPI bool operator== (const string_t &b) const;
		QUASARHOSTAPI bool operator== (LPCTSTR b) const;
		bool operator!= (const string_t &b) const { return !this->operator==(b); }
		bool operator!= (LPCTSTR b) const { return !this->operator==(b); }
		QUASARHOSTAPI string_t operator+(const string_t &b) const;
		QUASARHOSTAPI string_t pad_left(size_t count, TCHAR pad = (TCHAR)' ') const;
		QUASARHOSTAPI string_t pad_right(size_t count, TCHAR pad = (TCHAR)' ') const;
		QUASARHOSTAPI string_t sub_string(size_t start, size_t length = (size_t)(-1)) const;
		QUASARHOSTAPI string_t trim(TrimModes trim_mode) const;
		QUASARHOSTAPI string_t change_case(StringCases string_case) const;
		QUASARHOSTAPI size_t get_hash_code() const;
		QUASARHOSTAPI static bool is_whitespace(char ch);
#ifdef CPLUSPLUS11
		QUASARHOSTAPI string_t &operator=(string_t && other);
#endif
	private:
		QUASARHOSTAPI void set_string(const TCHAR *chars, size_t length);
	};

	struct QUASARHOSTAPI qvalue_t;


	namespace runtime
	{
		class execution_stream;
		class compute_kernel;
	}

	//
	// Lambda expression with variable number of input/output arguments
	//
	typedef void LambdaDelegate(void* closurePtr,
		const qvalue_t *argsIn, int nArgsIn,
		qvalue_t *argsOut, int nArgsOut);

	//
	// Kernel lambda expression with variable number of input/output arguments
	//
	typedef void KernelLambdaDelegate(void* closurePtr,
		const qvalue_t *argsIn, int nArgsIn,
		qvalue_t *argsOut, int nArgsOut, int3, int3, int, int, runtime::execution_stream *);

	// Prime kernel lambda delegate (for irregular data dimensions)
	typedef void PrimeKernelLambdaDelegate(void* closurePtr,
		const qvalue_t *argsIn, int nArgsIn,
		qvalue_t *argsOut, int nArgsOut, int3, int3, int3, int, int, runtime::execution_stream *);

	//
	// Function to be called when the lambda expression is destroyed,
	// typically to release its closure arguments
	//
	typedef void DisposeClosure(void* closurePtr, LambdaDelegate* lambda);

	//
	// Constructor function, to construct objects
	//
	typedef qvalue_t ConstructorDelegate(const qvalue_t *inArgs, int nArgs);

	//
	// Callback called when the host is destroyed
	//
	typedef void HostDestroyCallbackDelegate(void *closurePtr, IQuasarHost *host);

	//
	// Other delegates
	//
	typedef bool EnumRegisteredTypeDelegate(qhandle typeHandle, void*);
	typedef bool EnumFieldDelegate(const qvalue_t *obj, LPCTSTR fieldName, qhandle fieldType, void* userData);
	typedef bool EnumFieldWithDetailsDelegate(qhandle, LPCTSTR fieldName, qhandle fieldType, int numOffsets, const size_t *offsets, void* userData);
	typedef bool EnumFunctionsDelegate(LPCTSTR functionName, int numInArgs, qhandle *inArgTypes, int numOutArgs, qhandle *outArgTypes, int flags, const qvalue_t *funcVal, int token, void* userData);
	typedef bool EnumMethodsDelegate(qhandle objTypeHandle, LPCTSTR functionName, int numInArgs, qhandle *inArgTypes, int numOutArgs, qhandle *outArgTypes, int flags, void* userData);
	typedef bool DynamicCallDelegate(qhandle objHandle, LPCTSTR methodName, const qvalue_t * inArgs, int nInArgs, qvalue_t * outArgs, int nOutArgs);
	typedef bool EnumVariableDelegate(LPCVOID futureContext, LPCTSTR variableName, const qvalue_t *variableValue, int variableFlags, void* userData);


	//
	// QValue types (24 bytes - 32-bit / 40 bytes - 64-bit)
	//
	struct QUASARHOSTAPI qvalue_t
	{
		enum DataTypes type; // The type of this QValue object
		enum DataTypes elem_type; // The element type of this QValue object (vectors/matrices etc)
		qhandle private_obj; // Internal object - don't use!
		union
		{
			// Scalar values
			scalar scalarVal;
			struct
			{
				scalar realVal;
				scalar imagVal;
			};

			// Integer values
			int intVal;
			int64_t longVal;
			uint64_t ulongVal;
			struct
			{
				uint64_t u128_lo;
				uint64_t u128_hi;
			};
			
			// Half precision floating point values
			uint16_t halfVal;
			struct
			{
				uint16_t hrealVal;
				uint16_t himagVal;
			};

			// Single precision floating point values
			float floatVal;
			struct
			{
				float frealVal;
				float fimagVal;
			};

			// Double precision floating point values
			double doubleVal;
			struct
			{
				double drealVal;
				double dimagVal;
			};

			// Vectors/matrices
			struct
			{
				void *ptrVal;
				int dim1;
				int dim2;
				int dim3;
			};

			// Strings
			struct
			{
				LPCTSTR strVal;
				size_t strLen;
			};
			// Const string
			const char *constStrVal;
			// Characters
			uint32_t charVal;

			// Functions
			struct
			{
				LambdaDelegate * funcPtrVal;
				LPCVOID funcClosure;
				bool directCallable; // Only if true, the above delegate & closure can be used
			};
			// Runtime type information
			struct
			{
				LPCVOID vtPtrVal;
				DataTypes dataType;
				DataTypes elemType;
			};
		};

		static qvalue_t fromInt(int v)
		{
			qvalue_t t;
			t.type = TYPE_INT;
			t.intVal = v;
			return t;
		}

		static qvalue_t fromLong(long long v)
		{
			qvalue_t t;
			t.type = TYPE_INT64;
			t.longVal = v;
			return t;
		}
		static qvalue_t fromULong(unsigned long long v)
		{
			qvalue_t t;
			t.type = TYPE_UINT64;
			t.longVal = v;
			return t;
		}

		static qvalue_t fromScalar(scalar v)
		{
			qvalue_t t;
			t.type = TYPE_SCALAR;
			t.scalarVal = v;
			return t;
		}
		static qvalue_t fromCScalar(cscalar v)
		{
			qvalue_t t;
			t.type = TYPE_COMPLEXSCALAR;
			t.realVal = v.x;
			t.imagVal = v.y;
			return t;
		}
		// Remove this constructor because unnecessary initialization shows up in profiling runs
#ifndef HELIOS
		qvalue_t() { type = TYPE_VOID; private_obj = NULL; }
#endif
		static qvalue_t empty;
	};

	//
	// Custom trait for storing type information about Quasar data types
	//
	template <typename T> struct typeinfo_t
	{
		static const DataTypes type = TYPE_COUNT;
		static LPCTSTR typeName;
	};

#define MINVAL_uint8_t	MINVAL_uint8
#define MAXVAL_uint8_t	MAXVAL_uint8
#define MINVAL_uint16_t MINVAL_uint16
#define MAXVAL_uint16_t MAXVAL_uint16
#define MINVAL_uint32_t MINVAL_uint32
#define MAXVAL_uint32_t MAXVAL_uint32
#define MINVAL_uint64_t MINVAL_uint64
#define MAXVAL_uint64_t MAXVAL_uint64
#define MINVAL_int8_t	MINVAL_int8
#define MAXVAL_int8_t	MAXVAL_int8
#define MINVAL_int16_t	MINVAL_int16
#define MAXVAL_int16_t	MAXVAL_int16
#define MINVAL_int32_t	MINVAL_int32
#define MAXVAL_int32_t	MAXVAL_int32
#define MINVAL_int64_t	MINVAL_int64
#define MAXVAL_int64_t	MAXVAL_int64

#define DECLARE_TYPE(quasarTypeCode, nativeType, quasarTypeName)               \
    template <> struct typeinfo_t<nativeType>                    \
    {                                                                          \
        static const DataTypes type = quasarTypeCode;                          \
        static LPCTSTR typeName;                                               \
    };
#define DECLARE_SCALAR_TYPE(quasarTypeCode, nativeType, quasarTypeName)        \
    template <> struct typeinfo_t<nativeType>                    \
    {                                                                          \
        static const DataTypes type = quasarTypeCode;                          \
        static LPCTSTR typeName;                                               \
		static nativeType minValue;											   \
		static nativeType maxValue;											   \
    };
#define IMPLEMENT_TYPE(quasarTypeCode, nativeType, quasarTypeName)             \
    LPCTSTR typeinfo_t<nativeType>::typeName = quasarTypeName;

#define IMPLEMENT_SCALAR_TYPE(quasarTypeCode, nativeType, quasarTypeName)      \
    LPCTSTR typeinfo_t<nativeType>::typeName = quasarTypeName;				   \
	nativeType typeinfo_t<nativeType>::minValue = MINVAL_##nativeType;		   \
	nativeType typeinfo_t<nativeType>::maxValue = MAXVAL_##nativeType;		   \

#define DECLARE_STRUCT_TYPE(nativeType)                                        \
    DECLARE_TYPE(TYPE_TYPEDOBJECT, nativeType, L#nativeType)

#define DECLARE_TYPEDOBJECT_TYPE(nativeType)                                   \
    DECLARE_STRUCT_TYPE(nativeType)

#define DECLARE_UNTYPEDOBJECT_TYPE(nativeType)                                 \
    DECLARE_TYPE(TYPE_UNTYPEDOBJECT, nativeType, L#nativeType)

#define IMPLEMENT_TYPE_NS(nsName, quasarTypeCode, nativeType, quasarTypeName)  \
	} namespace quasar {													   \
    LPCTSTR typeinfo_t<nsName::nativeType>::typeName = quasarTypeName;		   \
	} namespace nsName {

#define DECLARE_STRUCT_TYPE_NS(nsName, nativeType)                             \
	} namespace quasar {													   \
    DECLARE_TYPE(TYPE_TYPEDOBJECT, nsName::nativeType, L#nativeType)	       \
	} namespace nsName {


//>>Quasar type declarations
#ifndef DBL_SCALAR
	DECLARE_SCALAR_TYPE(TYPE_SCALAR, float, _T("scalar"));
	DECLARE_SCALAR_TYPE(TYPE_DOUBLESCALAR, double, _T("scalar'double"));
	DECLARE_TYPE(TYPE_COMPLEXSCALAR, cscalar_t<float>, _T("cscalar"));
	DECLARE_TYPE(TYPE_DOUBLECOMPLEXSCALAR, cscalar_t<double>, _T("cscalar"));
#else
	DECLARE_SCALAR_TYPE(TYPE_SINGLEFLOAT, float, _T("scalar'single"));
	DECLARE_SCALAR_TYPE(TYPE_SCALAR, double, _T("scalar"));
	DECLARE_TYPE(TYPE_COMPLEXSCALAR, cscalar_t<double>, _T("cscalar"));
	DECLARE_TYPE(TYPE_FLOATCOMPLEX, cscalar_t<float>, _T("cscalar"));
#endif
	DECLARE_SCALAR_TYPE(TYPE_HALFSCALAR, __half, _T("scalar'half"));
	DECLARE_TYPE(TYPE_HALFCOMPLEXSCALAR, cscalar_t<__half>, _T("scalar'half"));
	DECLARE_SCALAR_TYPE(TYPE_INT, int, _T("int"));
    DECLARE_TYPE(TYPE_VEC, Vector, _T("vec"));
    DECLARE_TYPE(TYPE_MAT, Matrix, _T("mat"));
    DECLARE_TYPE(TYPE_CUBE, Cube, _T("cube"));
    DECLARE_TYPE(TYPE_CVEC, CVector, _T("cvec"));
    DECLARE_TYPE(TYPE_CMAT, CMatrix, _T("cmat"));
    DECLARE_TYPE(TYPE_CCUBE, CCube, _T("ccube"));
    DECLARE_TYPE(TYPE_STRING, string_t, _T("string"));
	DECLARE_SCALAR_TYPE(TYPE_INT8, int8_t, _T("int8"));
	DECLARE_SCALAR_TYPE(TYPE_INT16, int16_t, _T("int16"));
	DECLARE_SCALAR_TYPE(TYPE_INT64, int64_t, _T("int64"));
	DECLARE_SCALAR_TYPE(TYPE_UINT8, uint8_t, _T("uint8"));
	DECLARE_SCALAR_TYPE(TYPE_UINT16, uint16_t, _T("uint16"));
	DECLARE_SCALAR_TYPE(TYPE_UINT32, uint32_t, _T("uint32"));
	DECLARE_SCALAR_TYPE(TYPE_UINT64, uint64_t, _T("uint64"));
    DECLARE_TYPE(TYPE_VOID, qvalue_t, _T("??"));

	DECLARE_TYPE(TYPE_NCUBE, NCube<4>, _T("cube{4}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<5>, _T("cube{5}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<6>, _T("cube{6}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<7>, _T("cube{7}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<8>, _T("cube{8}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<9>, _T("cube{9}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<10>, _T("cube{10}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<11>, _T("cube{11}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<12>, _T("cube{12}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<13>, _T("cube{13}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<14>, _T("cube{14}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<15>, _T("cube{15}"));
	DECLARE_TYPE(TYPE_NCUBE, NCube<16>, _T("cube{16}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<4>, _T("ccube{4}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<5>, _T("ccube{5}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<6>, _T("ccube{6}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<7>, _T("ccube{7}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<8>, _T("ccube{8}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<9>, _T("ccube{9}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<10>, _T("ccube{10}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<11>, _T("ccube{11}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<12>, _T("ccube{12}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<13>, _T("ccube{13}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<14>, _T("ccube{14}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<15>, _T("ccube{15}"));
	DECLARE_TYPE(TYPE_NCCUBE, NCCube<16>, _T("ccube{16}"));

	DECLARE_TYPE(TYPE_VEC, scalar1, _T("vec1"));
	DECLARE_TYPE(TYPE_VEC, scalar2, _T("vec2"));
	DECLARE_TYPE(TYPE_VEC, scalar3, _T("vec3"));
	DECLARE_TYPE(TYPE_VEC, scalar4, _T("vec4"));
	DECLARE_TYPE(TYPE_VEC, scalarN<5>, _T("vec5"));
	DECLARE_TYPE(TYPE_VEC, scalarN<6>, _T("vec6"));
	DECLARE_TYPE(TYPE_VEC, scalarN<7>, _T("vec7"));
	DECLARE_TYPE(TYPE_VEC, scalarN<8>, _T("vec8"));
	DECLARE_TYPE(TYPE_VEC, scalarN<9>, _T("vec9"));
	DECLARE_TYPE(TYPE_VEC, scalarN<10>, _T("vec10"));
	DECLARE_TYPE(TYPE_VEC, scalarN<11>, _T("vec11"));
	DECLARE_TYPE(TYPE_VEC, scalarN<12>, _T("vec12"));
	DECLARE_TYPE(TYPE_VEC, scalarN<13>, _T("vec13"));
	DECLARE_TYPE(TYPE_VEC, scalarN<14>, _T("vec14"));
	DECLARE_TYPE(TYPE_VEC, scalarN<15>, _T("vec15"));
	DECLARE_TYPE(TYPE_VEC, scalarN<16>, _T("vec16"));

	DECLARE_TYPE(TYPE_CVEC, cscalar1, _T("cvec1"));
	DECLARE_TYPE(TYPE_CVEC, cscalar2, _T("cvec2"));
	DECLARE_TYPE(TYPE_CVEC, cscalar3, _T("cvec3"));
	DECLARE_TYPE(TYPE_CVEC, cscalar4, _T("cvec4"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<5>, _T("cvec5"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<6>, _T("cvec6"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<7>, _T("cvec7"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<8>, _T("cvec8"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<9>, _T("cvec9"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<10>, _T("cvec10"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<11>, _T("cvec11"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<12>, _T("cvec12"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<13>, _T("cvec13"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<14>, _T("cvec14"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<15>, _T("cvec15"));
	DECLARE_TYPE(TYPE_CVEC, cscalarN<16>, _T("cvec16"));

	DECLARE_TYPE(TYPE_VEC, int1, _T("ivec1"));
	DECLARE_TYPE(TYPE_VEC, int2, _T("ivec2"));
	DECLARE_TYPE(TYPE_VEC, int3, _T("ivec3"));
	DECLARE_TYPE(TYPE_VEC, int4, _T("ivec4"));
	DECLARE_TYPE(TYPE_VEC, intN<5>, _T("ivec5"));
	DECLARE_TYPE(TYPE_VEC, intN<6>, _T("ivec6"));
	DECLARE_TYPE(TYPE_VEC, intN<7>, _T("ivec7"));
	DECLARE_TYPE(TYPE_VEC, intN<8>, _T("ivec8"));
	DECLARE_TYPE(TYPE_VEC, intN<9>, _T("ivec9"));
	DECLARE_TYPE(TYPE_VEC, intN<10>, _T("ivec10"));
	DECLARE_TYPE(TYPE_VEC, intN<11>, _T("ivec11"));
	DECLARE_TYPE(TYPE_VEC, intN<12>, _T("ivec12"));
	DECLARE_TYPE(TYPE_VEC, intN<13>, _T("ivec13"));
	DECLARE_TYPE(TYPE_VEC, intN<14>, _T("ivec14"));
	DECLARE_TYPE(TYPE_VEC, intN<15>, _T("ivec15"));
	DECLARE_TYPE(TYPE_VEC, intN<16>, _T("ivec16"));
//<<End Quasar type declarations

#define VERIFY_LOCK(x) (x)

	// Base class for reference countable objects
    interface IRefCountable
    {
        virtual void AddRef() = 0;
        virtual void Release() = 0;
    };

	class Expression;
	class VariableDecl;
	class Symbol;

    //======================================================================
    // IQuasarHost - allows you to communicate with the Quasar host. It is
    // possible to load Quasar modules, call functions, read and write
    // variable values. There are also some utility functions for message
    // loop handling etc.
    //======================================================================
    interface IQuasarHost : public IRefCountable
    {
		// Device API
		virtual MachineInfo GetMachineInfo() = 0;
		virtual void FreeMachineInfo(MachineInfo &machineInfo) = 0;
		virtual void OpenDevice(LPCTSTR deviceId, const RuntimeSettings *runtimeSettings) = 0;
		virtual void CloseDevice() = 0;

		// Module load/unload API
        virtual bool LoadSourceModule(LPCTSTR sourceModule,
                                      LPCTSTR * errorMsg) = 0;
        virtual bool LoadBinaryModule(LPCTSTR binaryModule,
                                      LPCTSTR * errorMsg) = 0;
        virtual bool LoadModuleFromSource(
            LPCTSTR moduleName, LPCTSTR sourceString, LPCTSTR * errorMsg) = 0;
		virtual bool UnloadModule(LPCTSTR moduleName) = 0;
		virtual bool RegisterModule(RegisterModuleOptions options, LPCTSTR moduleName,
			LPCTSTR fileName, RegisterModuleFlags flags) = 0;
		virtual bool RegisterModuleFromData(RegisterModuleOptions options, LPCTSTR moduleName,
			LPCVOID data, size_t dataSize, RegisterModuleFlags flags) = 0;

		// Function API
		virtual qvalue_t LoadFunction(qvalue_t closureArgs, LPCTSTR moduleName, LPCTSTR functionSignature, LPCTSTR funcName,
			FunctionFlags flags, PlatformTargets platformTargets, LoadFunctionParams &params) = 0;
		virtual bool FunctionExists(LPCTSTR functionName) = 0;
        virtual bool FunctionCall(LPCTSTR functionName, const qvalue_t *inArgs,
                                  int nInArgs, qvalue_t *outArgs,
                                  int nOutArgs) = 0;
        virtual bool FunctionCall(const qvalue_t &function,
                                  const qvalue_t *inArgs, int nInArgs,
                                  qvalue_t *outArgs, int nOutArgs) = 0;

		// Data structure API
        virtual qvalue_t CreateVector(qhandle elemType, int numel, bool complex) = 0;
        virtual qvalue_t CreateMatrix(qhandle elemType, int dim1, int dim2, bool complex) = 0;
        virtual qvalue_t CreateCube(qhandle elemType, int dim1, int dim2,
                                    int dim3, bool complex) = 0;
		virtual qvalue_t CreateNCube(qhandle elemType, int ndims, const int* dims, bool complex) = 0;
        virtual qvalue_t CreateTypedObject(qhandle objType) = 0;
        virtual qvalue_t CreateUntypedObject(qhandle objType) = 0;
        virtual qvalue_t CreateString(LPCTSTR text, size_t length = -1) = 0;
        virtual qvalue_t CreateLambda(qhandle functionTypeHandle,
									  void* closurePtr,
                                      LambdaDelegate* lambda) = 0;
		virtual qvalue_t CreateNamedParam(LPCTSTR paramName, const qvalue_t &value) = 0;
		virtual qvalue_t CreateGeneric(qhandle typeHandle, const qvalue_t *inArgs, int nInArgs) = 0;
		virtual void GetNDims(const qvalue_t &obj, int *nDimsWritten, int *dims) = 0;
		virtual void GetString(const qvalue_t &obj, string_t &str) = 0;
        virtual void AddRef(qvalue_t & obj) = 0;
        virtual void ReleaseRef(qvalue_t obj) = 0;
        virtual void DeleteValue(qvalue_t & obj) = 0;
        virtual LockResults Lock(const qvalue_t &obj, DataTypes elemType,
                                 LockingModes lockingMode,
                                 MemResources memResource, void **pData) = 0;
        virtual void Unlock(const qvalue_t &obj, LockingModes lockingMode,
                            MemResources memResource) = 0;
        virtual qhandle GetPrimitiveTypeHandle(DataTypes type, LPCTSTR typeName) = 0;

		// DSL API
        virtual void RunApp() = 0;
        virtual void DoEvents() = 0;
        virtual bool ReadVariable(LPCTSTR varName, qvalue_t & val) = 0;
        virtual bool WriteVariable(LPCTSTR varName, const qvalue_t &val) = 0;
        virtual bool GetField(const qvalue_t &obj, LPCTSTR fieldName,
                              qvalue_t &val) = 0;
        virtual bool SetField(const qvalue_t &obj, LPCTSTR fieldName,
                              const qvalue_t &val) = 0;
        virtual qvalue_t LookupFunction(LPCTSTR functionSignature) = 0;
        virtual qvalue_t LookupType(LPCTSTR quasarTypeName) = 0;
        virtual qvalue_t LookupMethod(const qvalue_t &type,
                                      LPCTSTR functionSignature) = 0;
        virtual bool MethodCall(const qvalue_t &function,
                                const qvalue_t &target, const qvalue_t *inArgs,
                                int nInArgs, qvalue_t *outArgs,
                                int nOutArgs) = 0;
		virtual bool MethodCall(LPCTSTR functionName,
								const qvalue_t &target, const qvalue_t *inArgs,
								int nInArgs, qvalue_t *outArgs,
								int nOutArgs) = 0;
        virtual qvalue_t GetType(const qvalue_t &obj) = 0;
        virtual qvalue_t CreateType(LPCTSTR moduleName, LPCTSTR typeName,
			CreateTypeFlags flags = TYPEFLAGS_NONE) = 0;
        virtual bool AddField(qhandle objType, LPCTSTR fieldName,
                              qhandle fieldType) = 0;
        virtual qvalue_t AddParameter(qhandle objType, qhandle typeParamHandle) = 0;

        virtual void FinalizeType(qhandle typeHandle) = 0;
        virtual bool EnableProfiling(ProfilingModes profilingMode,
                                     LPCTSTR outputFileName = NULL) = 0;
        virtual IEvaluationStack *CreateStack() = 0;
		virtual bool QueryProperty(HostProperties prop, int param, void** dst, size_t nBytes) = 0;
        virtual IComputationEngine *GetComputationEngine() = 0;
		virtual string_t TypeGetString(qvalue_t type) = 0;
		virtual void TypeGetInfo(qvalue_t type, DataTypes *dataType, int *maxDims, int *vectorLength, AccessModifiers *accessModifier,
			bool *isIntegerType, bool *isPointer) = 0;
		virtual bool TypeIsStatic(qvalue_t type) = 0;
		virtual qvalue_t TypeGetElementTypeAtPosition(qvalue_t type, int dims, int* pos) = 0;
		virtual bool TypeCompareToSpecification(qvalue_t type, qvalue_t specType) = 0;
		virtual void AddLambdaDestroyCallback(qvalue_t lambda, DisposeClosure* disposeDelegate) = 0;

		// Exception handling
		virtual void SetCustomExceptionHandler(custom_exceptionhandler_t handler) = 0;
		virtual custom_exceptionhandler_t GetCustomExceptionHandler() = 0;

		virtual void AddRef() = 0;
		virtual void Release() = 0;
		virtual ~IQuasarHost() {};
		virtual bool AddFieldWithOffset(qhandle objType, LPCTSTR fieldName,
			qhandle fieldType, size_t offset) = 0;
		virtual bool AddProperty(qhandle objType, LPCTSTR propertyName, qhandle fieldType,
			qhandle getProperty, qhandle setProperty) = 0;
		virtual bool InheritType(qhandle objType, qhandle baseType) = 0;
		virtual bool SetConstructor(qhandle objType, ConstructorDelegate* constructor) = 0;
		virtual qhandle CreateFunctionType(FunctionTypes functionType,
					const qhandle* inputTypes, int numInputTypes,
					const qhandle* outputTypes, int numOutputTypes, FunctionTypeFlags flags) = 0;
		virtual qhandle CreateTypeClass(const qhandle *elemTypes, int elemTypeCount) = 0;
		virtual qhandle InstantiateType(qhandle type, const qhandle* instance_type_params, int num_params) = 0;
		virtual qhandle CreatePointerType(qhandle type) = 0;
		virtual qhandle CreateTypeParameter(LPCTSTR paramName, qhandle paramType) = 0;
		virtual qvalue_t TypeGetElementType(qvalue_t type) = 0;
		virtual void SetMainProgram(LPCTSTR mainProgramName) = 0;
		virtual bool IsMainProgram(LPCTSTR programName) = 0;
		virtual bool IsRebindRequired(qvalue_t lambda, int token) = 0;
		virtual qvalue_t LookupFunction(LPCTSTR functionSignature, int &token) = 0;
		virtual qvalue_t LookupMethod(const qvalue_t &type, LPCTSTR functionSignature, int &token) = 0;
		virtual void GetProgramCmdLineArguments(qvalue_t *argsOut, int numArgs, int numMandatoryArgs, bool variadic) = 0;
		virtual void SetProgramCmdLineArgumentOffset(int offset) = 0;
		virtual void* RegisterInstance(void *native_obj_ptr, qvalue_t &type) = 0;
		virtual MemResources GetMemResourceForDevice(int deviceId) = 0;
		virtual void* GetDataPtr(qvalue_t *obj, MemResources memResource) = 0;
		virtual void GetLambdaFunctionPtr(qvalue_t *lambda, int deviceId, void **ptr, void **closure) = 0;
		virtual void UnmarshalPointer(qvalue_t *obj, LPCVOID ptr, const void *data, size_t data_size, DataTypes data_type, int deviceId, int flags) = 0;
		virtual void UnmarshalTypedObject(qvalue_t *qv, const void *ptr, size_t size, LPCTSTR typeName, int deviceId, int flags) = 0;

		virtual void EnumRegisteredTypes(EnumRegisteredTypeDelegate* callback, void* userData) = 0;
		virtual void EnumFields(const qvalue_t &obj, EnumFieldDelegate* callback, void *userData) = 0;
		virtual void EnumVariables(EnumVariableDelegate* callback, void *userData) = 0;
		virtual bool SetDynamicDispatcher(qhandle objType, DynamicDispatchers disp, DynamicCallDelegate *call) = 0;
		virtual bool AddHostDestroyCallback(HostDestroyCallbackDelegate *callback, void *userData) = 0;
		virtual void ParseNamedParam(const qvalue_t &named_param, string_t &name, qvalue_t &value) = 0;
		virtual void OnException(const exception_t &exception) = 0;
		virtual size_t GetRawSizeOfType(qhandle typeHandle, MemResources memResource, size_t *alignment = nullptr) = 0;
		virtual void TypeGetClassInfo(qhandle typeHandle, string_t &moduleName, string_t &typeName, ClassInfoFlags &flags,
			MemResources* resources, int *numResources, size_t* size, qhandle *baseClass, int *numGenericTypeParams, qhandle *genericTypeParams,
			int *numInstanceTypeParams, qhandle *instanceTypeParams) = 0;

        //=========================================================
        // Generic helper functions
        //=========================================================
        template <typename T> qhandle GetTypeHandle()
        {
            if CONSTEXPR(typeinfo_t<T>::type == TYPE_TYPEDOBJECT)
            {
                // If you get an error at this line, then you probably forgot to
                // define the macro DECLARE_STRUCT_TYPE() for your user-defined
                // type!
                return LookupType(typeinfo_t<T>::typeName).private_obj;
            }
			return GetPrimitiveTypeHandle(typeinfo_t<T>::type, typeinfo_t<T>::typeName);

		}

	private:
		template <typename T> bool IsComplex()
		{
			return typeinfo_t<T>::type == TYPE_COMPLEXSCALAR;
		}

	public:
        // Creation of generic vectors
        template <typename T> qvalue_t CreateVector(int numel)
        {
            return CreateVector(GetTypeHandle<T>(), numel, IsComplex<T>());
        }

        // Creation of generic matrices
        template <typename T> qvalue_t CreateMatrix(int dim1, int dim2)
        {
            return CreateMatrix(GetTypeHandle<T>(), dim1, dim2, IsComplex<T>());
        }

        // Creation of generic cubes
        template <typename T> qvalue_t CreateCube(int dim1, int dim2, int dim3)
        {
            return CreateCube(GetTypeHandle<T>(), dim1, dim2, dim3, IsComplex<T>());
        }

	    // Creation of generic n-cubes (higher dimensional cubes)
        template <typename T> qvalue_t CreateNCube(int ndims, const int *dims)
        {
            return CreateNCube(GetTypeHandle<T>(), ndims, dims, IsComplex<T>());
        }

		template <typename T, int N> qvalue_t CreateNCube(intN<N> dims)
		{
			return CreateNCube(GetTypeHandle<T>(), N, (int *) &dims, IsComplex<T>());
		}

        // Creation of initialized generic vectors
        template <typename T> qvalue_t CreateVector(int numel, const T *vals)
        {
            qvalue_t vec = CreateVector<T>(numel);
            VectorBase<T> v = LockVector<T>(vec);
            memcpy(v.data, vals, numel * sizeof(T));
            UnlockVector(vec);
            return vec;
        }
        template <typename T, int N> qvalue_t CreateVector(T(&vals)[N])
        {
            return CreateVector(N, vals);
        }

        template <typename T>
        qvalue_t CreateMatrix(int dim1, int dim2, const T *vals)
        {
            qvalue_t mat = CreateMatrix<T>(dim1, dim2);
            MatrixBase<T> m = LockMatrix<T>(mat);
            memcpy(m.data, vals, dim1 * dim2 * sizeof(T));
            UnlockMatrix(mat);
            return mat;
        }

		template <typename T>
		qvalue_t CreateCube(int dim1, int dim2, int dim3, const T *vals)
		{
			qvalue_t cube = CreateCube<T>(dim1, dim2, dim3);
			CubeBase<T> m = LockCube<T>(cube);
			memcpy(m.data, vals, dim1 * dim2 * dim3 * sizeof(T));
			UnlockCube(cube);
			return cube;
		}

        template <typename T>
        VectorBase<T> LockVector(const qvalue_t &obj,
                                 LockingModes lockingMode = LOCK_READWRITE,
                                 MemResources memResource = MEMRESOURCE_CPU)
        {
            LPVOID ptr = NULL;
            VERIFY_LOCK(
                Lock(obj, typeinfo_t<T>::type, lockingMode, memResource, &ptr));
            return VectorBase<T>((T *)ptr, obj.dim1*obj.dim2*obj.dim3);
        }
        void UnlockVector(const qvalue_t &obj,
                          LockingModes lockingMode = LOCK_READWRITE,
                          MemResources memResource = MEMRESOURCE_CPU)
        {
            Unlock(obj, lockingMode, memResource);
        }
        template <typename T>
        MatrixBase<T> LockMatrix(const qvalue_t &obj,
                                 LockingModes lockingMode = LOCK_READWRITE,
                                 MemResources memResource = MEMRESOURCE_CPU)
        {
            LPVOID ptr = NULL;
            VERIFY_LOCK(
                Lock(obj, typeinfo_t<T>::type, lockingMode, memResource, &ptr));
            return MatrixBase<T>((T *)ptr, obj.dim1, obj.dim2);
        }
        void UnlockMatrix(const qvalue_t &obj,
                          LockingModes lockingMode = LOCK_READWRITE,
                          MemResources memResource = MEMRESOURCE_CPU)
        {
            Unlock(obj, lockingMode, memResource);
        }
        template <typename T>
        CubeBase<T> LockCube(const qvalue_t &obj,
                             LockingModes lockingMode = LOCK_READWRITE,
                             MemResources memResource = MEMRESOURCE_CPU)
        {
            LPVOID ptr = NULL;
            VERIFY_LOCK(
                Lock(obj, typeinfo_t<T>::type, lockingMode, memResource, &ptr));
            return CubeBase<T>((T *)ptr, obj.dim1, obj.dim2, obj.dim3);
        }
        void UnlockCube(const qvalue_t &obj,
                        LockingModes lockingMode = LOCK_READWRITE,
                        MemResources memResource = MEMRESOURCE_CPU)
        {
            Unlock(obj, lockingMode, memResource);
        }
		template <typename T, int N>
        NCubeBase<T, N> LockNCube(const qvalue_t &obj,
                             LockingModes lockingMode = LOCK_READWRITE,
                             MemResources memResource = MEMRESOURCE_CPU)
        {
            LPVOID ptr = NULL;
			intN<N> dims;
			int ndims = N;
			// Get the dimensions
			GetNDims(obj, &ndims, (int *) &dims);
			_ASSERT_EXPR(ndims == N, "The number of dimensions do not match!");
            VERIFY_LOCK(
                Lock(obj, typeinfo_t<T>::type, lockingMode, memResource, &ptr));
            return NCubeBase<T, N>((T *)ptr, dims);
        }
        void UnlockNCube(const qvalue_t &obj,
                        LockingModes lockingMode = LOCK_READWRITE,
                        MemResources memResource = MEMRESOURCE_CPU)
        {
            Unlock(obj, lockingMode, memResource);
        }
        template <typename T>
        T *LockObject(const qvalue_t &obj,
                      LockingModes lockingMode = LOCK_READWRITE,
                      MemResources memResource = MEMRESOURCE_CPU)
        {
            LPVOID ptr = NULL;
			_ASSERT_EXPR(obj.type == TYPE_TYPEDOBJECT || obj.type == TYPE_UNTYPEDOBJECT ||
				obj.type == TYPE_VEC || obj.type == TYPE_MAT || obj.type == TYPE_CUBE || obj.type == TYPE_NCUBE ||
				obj.type == TYPE_CVEC || obj.type == TYPE_CMAT || obj.type == TYPE_CCUBE || obj.type == TYPE_NCCUBE,
				"LockObject can only be called on a typed object or a vector");
            VERIFY_LOCK(
                Lock(obj, TYPE_TYPEDOBJECT, lockingMode, memResource, &ptr));
            return (T *)ptr;
        }
        void UnlockObject(const qvalue_t &obj,
                          LockingModes lockingMode = LOCK_READWRITE,
                          MemResources memResource = MEMRESOURCE_CPU)
        {
            Unlock(obj, lockingMode, memResource);
        }
		virtual void Synchronize();
		virtual bool GetRuntimeControl(qvalue_t &obj);
		virtual void SetupKernelFunction(
			qvalue_t &qv_kernel,
			std::function<KernelLambdaDelegate> cpu_kernel_adapter,
			std::function<KernelLambdaDelegate> cuda_kernel_adapter,
			std::function<PrimeKernelLambdaDelegate> cuda_prime_kernel_adapter) = 0;
		virtual void SetupDeviceFunction(
			qvalue_t &qv_device,
			std::function<LambdaDelegate> cpu_device_adapter,
			void *cpu_device_func_ptr) = 0;
		virtual bool IsKernelInitialized(qhandle kernelHandle) = 0;
		virtual void LaunchKernel(qhandle kernel, LPCVOID params, size_t params_size,
			const void *closure_ptr, const qvalue_t * args, int3 gridDim, int3 blockDim,
			bool prime_kernel, int deviceId, qhandle _stream) = 0;
		virtual void SetupKernelFunction2(
			qvalue_t &qv_kernel,
			KernelLambdaDelegate* cpu_kernel_adapter,
			KernelLambdaDelegate* cuda_kernel_adapter,
			PrimeKernelLambdaDelegate* cuda_prime_kernel_adapter) { 
			SetupKernelFunction(qv_kernel, cpu_kernel_adapter, cuda_kernel_adapter, cuda_prime_kernel_adapter);
		}
		virtual void SetupDeviceFunction2(
			qvalue_t &qv_device,
			LambdaDelegate* cpu_device_adapter,
			void *cpu_device_func_ptr) {
			SetupDeviceFunction(qv_device, cpu_device_adapter, cpu_device_func_ptr);
		}


        static QUASARHOSTAPI IQuasarHost *Create(LPCTSTR deviceName = NULL, bool loadCompiler = true);
        static QUASARHOSTAPI IQuasarHost *Connect(LPCVOID hostprivParams,
                                    LPCVOID moduleDetails, int flags);
        static QUASARHOSTAPI IQuasarHost *GetInstance();
    };

    //======================================================================
    // IEvaluationStack - gives access to Quasar's internal evaluation stack
    //======================================================================
    interface IEvaluationStack : public IRefCountable
    {
        virtual qvalue_t PopValue() = 0;
        virtual void PushValue(const qvalue_t &value) = 0;
        virtual void PopMultiple(qvalue_t * vals, int count) = 0;
        virtual void PushMultiple(const qvalue_t *vals, int count) = 0;
        virtual void Clear() = 0;
        virtual void PushContext() = 0;
        virtual void PopContext() = 0;
        virtual void Process(OperatorTypes op) = 0;
        virtual void FunctionCall(LPCTSTR functionName, int numArgs) = 0;
        virtual void ArrayGetAt(const qvalue_t &x, int numIndices,
                                BoundaryAccessMode boundsMode) = 0;
        virtual void ArraySetAt(qvalue_t & x, int numIndices,
                                BoundaryAccessMode boundsMode) = 0;
        virtual qvalue_t Evaluate(const qvalue_t &x) = 0;
        virtual int GetCount() = 0;
    };

    //======================================================================
    // IMatrixFactory - provides helper routines for creating matrices
    //======================================================================
    interface IMatrixFactory : public IRefCountable
    {
        virtual qvalue_t New(AllocationFlags flags, LPCVOID elems, int64_t elemBytes,
							 qhandle elemType, int nDims, const int *dims) = 0;
    };

    //======================================================================
    // ITypeEnvironment - provides access to the type information for
    // certain types.
    //======================================================================
    interface ITypeEnvironment : public IRefCountable
    {
        virtual ScalarTypes GetScalarType() = 0;
    };

    //======================================================================
    // IRuntimeReductionEngine - provides access to the runtime reduction
    // engine. Allows you to define/undefine your own reductions that are
    // then applied at run-time.
    //======================================================================
    interface IRuntimeReductionEngine : public IRefCountable
    {
        virtual bool Add(LPCTSTR searchExpr, qhandle action,
						 qhandle whereClause) = 0;
		virtual bool Add(const VariableDecl* variables, int numVariables,
			Expression *searchExpr, const qvalue_t *action, const qvalue_t *whereClause) = 0;
        virtual bool Remove(LPCTSTR searchExpr) = 0;
		virtual void EnumFields(qhandle type, int numOffsets, const MemResources *memResources,
			EnumFieldWithDetailsDelegate* callback, void* userData) = 0;
		virtual void EnumFunctions(EnumFunctionsDelegate* callback, void* userData) = 0;
		virtual void EnumMethods(qhandle type, EnumMethodsDelegate * callback, void* userData) = 0;

	#ifdef CPLUSPLUS11
		bool Add(
			std::initializer_list<VariableDecl> variables,
			std::initializer_list<Symbol> searchExprSymbols, bool isVariadic, qvalue_t action, qvalue_t whereClause = qvalue_t::empty);
	#endif
    };

    //======================================================================
    // IComputationEngine - this interface represents the core of the Quasar
    // runtime system, by creating a uniform interface to the underlying
    // computation device
    //======================================================================
    interface IComputationEngine : public IRefCountable
    {
        virtual bool GetName(LPTSTR name, int maxChars) = 0;
        virtual IEvaluationStack *GetEvaluationStack() = 0;
        virtual IMatrixFactory *GetMatrixFactory() = 0;
        virtual ITypeEnvironment *GetTypeEnvironment() = 0;
        virtual IRuntimeReductionEngine *GetRuntimeReductionEngine() = 0;
        virtual void Process(OperatorTypes op) = 0;
        virtual void ConstructMatrix(int numRowsOrCols) = 0;
        virtual void ConstructCellMatrix(int length) = 0;
        virtual bool FunctionCall(LPCTSTR functionName, int numArgs) = 0;
        virtual bool FunctionExists(LPCTSTR functionName, int numArgs) = 0;
        virtual void ArrayGetAt(const qvalue_t &x, int numIndices,
                                BoundaryAccessMode boundsMode) = 0;
        virtual void ArraySetAt(qvalue_t & x, int numIndices,
                                BoundaryAccessMode boundsMode) = 0;
        virtual void Synchronize() = 0;
		virtual void ParallelDo(const qvalue_t& dims, const qvalue_t& function, int numArgs, qvalue_t &retval) = 0;
		virtual void SerialDo(const qvalue_t& dims, const qvalue_t& function, int numArgs, qvalue_t &retval) = 0;
    };

	//======================================================================
    // Quasar exception class
    //======================================================================
	class exception_t
	{
	public:
		string_t source;		// The source module that generated the exception
		string_t message;		// The error message
		string_t stack_trace;	// The internal stack trace of the error
	};

	//======================================================================
	// ref class
	//======================================================================
	// Smart reference pointer for IRefCountable classes (i.e. classes
	// that implement an AddRef() and a Release() method).
	template <typename T> class ref final
	{
	protected:
		T *obj;

	public:
		ref(T *obj) : obj(obj)
		{
			if (obj)
				obj->AddRef();
		}
		ref(const ref &other) : obj(other.obj)
		{
			if (obj)
				obj->AddRef();
		}
#ifdef CPLUSPLUS11
		ref(ref &&other):obj(other.obj)
		{
			other.obj = nullptr;
		}
#endif
		ref() : obj(nullptr) {}
		~ref()
		{
			if (obj != nullptr)
			{
				obj->Release();
				obj = NULL;
			}
		}
		// Takes over the ownership of an object. Typically used in combination
		// with the operator new. For example auto t = ref<T>(new T());
		static ref own(T * obj)
		{
			ref b; b.obj = obj; return b;
		}

		// Simplified way to construct objects.
		static ref<T> create()
		{
			return own(new T());
		}

		// Simplified way to construct objects (multiple parameter version).
#ifdef CPLUSPLUS14
		template <typename... Args>
		static ref<T> create(Args... args)
		{
			return own(new T(args...));
		}
#endif

		// Move constructor
#ifdef CPLUSPLUS11
		ref<T> &operator=(ref<T> && other)
		{
			if (obj != nullptr)
				obj->Release();
			obj = std::move(other.obj);
			other.obj = nullptr;
			return *this;
		}
#endif
		T *operator->() { return (T *)obj; }
		const T *operator->() const { return (const T *)obj; }
		bool operator==(T *other) const { return obj == other; }
		bool operator!=(T *other) const { return obj != other; }

		ref &operator=(const ref &other)
		{
			return *this = other.obj;
		}
		ref &operator=(T *other)
		{
			if (other != nullptr)
				other->AddRef();
			if (obj != nullptr)
			{
				obj->Release();
				obj = nullptr;
			}
			obj = other;
			return *this;
		}

		operator const T *() const { return (const T *)obj; }
		operator T *() { return (T *)obj; }

		// Performs a reference static cast
		template <typename R>
		ref<R> cast() const
		{
			R *other = static_cast<R*>(obj);
			return ref<R>(other);
		}
	};

	// So that ref<> has no influence on getting the typeinfo of a type
#ifdef CPLUSPLUS11
	template <typename T>
	struct typeinfo_t<ref<T>> : public typeinfo_t<T> {};
#endif

}

namespace std
{
	template <>
	struct hash<quasar::string_t>
	{
		size_t operator()(const quasar::string_t& t) const
		{
			return t.get_hash_code();
		}
	};
}

#endif
