//=============================================================================
// Quasar CUDA / OpenMP runtime backend
// (C) 2011-2020 Gepura
// Higher-dimensional extensions for the Quasar back-end
//=============================================================================
#include "quasar.h"

#ifndef QUASAR_NCUBE_H_INCLUDED
#define QUASAR_NCUBE_H_INCLUDED

namespace quasar
{
    template <typename T, int N> struct NCubeBase // Hyper-cube
    {
        T *data;
        intN<N> dims;
        INLINE __device__ __HOST_V__ int get_numel() const { return prod(dims); }

        INLINE __device__ __HOST_V__ NCubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ NCubeBase(T *data, intN<N> dims)
            : data(data), dims(dims)
        {
        }
        INLINE __device__ __HOST_V__ void addref()
        {
#ifdef ENABLE_DYNAMIC_KERNEL_MEM
            if (dynmem->is_validpointer(data))
                dynmem->add_ref(data);
#endif
        }
        INLINE __device__ __HOST_V__ void release()
        {
#ifdef ENABLE_DYNAMIC_KERNEL_MEM
            if (dynmem->is_validpointer(data) && dynmem->release_ref(data) == 0)
            {
                dynmem->free(data);
            }
#endif
        }
    };

    template <BoundaryAccessMode mode, typename T, int N>
    INLINE __device__ __HOST_V__ T ncube_get_at(const NCubeBase<T, N> &x, intN<N> pos)
    {
		int index = pos2indchk(x.dims, pos);
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, index >= 0, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, T())
		}
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, T())
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, index >= 0, T())
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if CONSTEXPR (mode == ConstZero)
		{
			if (index < 0)
				return T();
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if (index < 0)
			return T();

        return x.data[index];
    }

    template <BoundaryAccessMode mode, typename T, int N>
    INLINE __device__ __HOST_V__ T
    ncube_set_at(const NCubeBase<T, N> &x, intN<N> pos, T val)
    {
		int index = pos2indchk(x.dims, pos);
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, index >= 0, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, T())
		}
		else if (index < 0)
			return T();
        CHECK_NAN_OR_INF(x, val);
        return x.data[index] = val;
    }

	template <BoundaryAccessMode mode, typename T, int M, bool aligned, int N>
	INLINE __device__ __HOST_V__ typename VecType<T, M>::type ncube_get_vec_at(const NCubeBase<T, N> &x, intN<N> pos)
	{
		typedef typename VecType<T, M>::type R;	// Vector type
		typedef array<T, M> R_unaligned;	// Array type
		R retVal;

		int index = pos2indchk(x.dims, pos);
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, index >= 0, R())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, R())
		}
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, R())
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, index >= 0, R())
		}
		else if (index < 0)
			return R();

		if CONSTEXPR (mode < Const)
		{
			if CONSTEXPR (aligned)
				retVal = (R &)x.data[index];
			else
				(R_unaligned &)retVal = *((R_unaligned *)&x.data[index]);
		}
		else
		{
			if CONSTEXPR (aligned)
				retVal = LOAD_CONST(((R *)&x.data[index]));
			else
				(R_unaligned &)retVal = LOAD_CONST(((R_unaligned *)&x.data[index]));
		}
		return retVal;
	}

	template <BoundaryAccessMode mode, typename T, int M, bool aligned, int N>
    INLINE __device__ __HOST_V__ typename VecType<T, M>::type
    ncube_set_vec_at(const NCubeBase<T, N> &x, intN<N> pos,
		typename VecType<T, M>::type val)
    {
		typedef typename VecType<T, M>::type R;	// Vector type
		typedef array<T, M> R_unaligned;	// Array type

		int index = pos2indchk(x.dims, pos);
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, index >= 0, R())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, R())
		}
		else if (index < 0)
			return R();
        CHECK_NAN_OR_INF(x, val);

		if CONSTEXPR (aligned)
			(R &)x.data[index] = val;
		else
			(R_unaligned &)x.data[index] = (R_unaligned &)val;
		return val;
    }

    template <BoundaryAccessMode mode, typename T, int N>
    INLINE __device__ __HOST_V__ lref<T> ncube_get_ref_at(const NCubeBase<T, N> &x, intN<N> pos)
    {
		int index = pos2indchk(x.dims, pos);
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, index >= 0, lref<T>())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, lref<T>())
		}
		else if (index < 0)
			return lref<T>();
        return x.data[index];
    }

    // Atomic operations - 3 arguments
    template <BoundaryAccessMode mode, typename T, int N, T atomic_function(T *, T)>
    INLINE __device__ __HOST_V__ T ncube_at(const NCubeBase<T, N> &x, intN<N> pos, T val)
    {
        int index = pos2indchk(x.dims, pos);
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, index >= 0, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, T())
		}
		else if (index < 0)
			return T();
        CHECK_NAN_OR_INF(x, val);
        return atomic_function((T *)&x.data[index], val);
    }

	// Atomic operations - 4 arguments
	template <BoundaryAccessMode mode, typename T, int N, T atomic_function(T *, T, T)>
	INLINE __device__ __HOST_V__ T ncube_at(const NCubeBase<T, N> &x, intN<N> pos, T ref, T val)
	{
		int index = pos2indchk(x.dims, pos);
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, index >= 0, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0, T())
		}
		else if (index < 0)
			return T();
		CHECK_NAN_OR_INF(x, val);
		return atomic_function((T *)&x.data[index], ref, val);
	}

//===================================
// Specializations for cell matrices
//===================================
#ifdef ENABLE_DYNAMIC_KERNEL_MEM
#define DYNMEM_FUNC()                                                          \
    INLINE __device__ void addref()                                            \
    {                                                                          \
        if (dynmem->is_validpointer(data))                                     \
            dynmem->add_ref(data);                                             \
    }                                                                          \
    INLINE __device__ void release()                                           \
    {                                                                          \
        if (dynmem->is_validpointer(data) && dynmem->release_ref(data) == 0)   \
        {                                                                      \
            for (int i = 0; i < get_numel(); i++)                              \
                data[i].release();                                             \
            dynmem->free(data);                                                \
        }                                                                      \
    }
#else
#define DYNMEM_FUNC()                                                          \
    INLINE __device__ __HOST_V__ void addref() {}                              \
    INLINE __device__ __HOST_V__ void release() {}
#endif


    template <typename T, int N> struct VectorBase<NCubeBase<T, N> >
    {
        NCubeBase<T, N> *data;
        int dim1;

        INLINE __device__ __HOST_V__ VectorBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ VectorBase(NCubeBase<T, N> *data, int dim1)
            : data(data), dim1(dim1)
        {
        }
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1; }
        DYNMEM_FUNC()
    };

    template <typename T, int N> struct MatrixBase<NCubeBase<T, N> >
    {
        NCubeBase<T, N> *data;
        int dim1;
        int dim2;

        INLINE __device__ __HOST_V__ MatrixBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ MatrixBase(NCubeBase<T, N> *data, int dim1, int dim2)
            : data(data), dim1(dim1), dim2(dim2)
        {
        }
		INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2; }
        DYNMEM_FUNC()
    };

    template <typename T, int N> struct CubeBase<NCubeBase<T, N> >
    {
        NCubeBase<T, N> *data;
        int dim1;
        int dim2;
        int dim3;

        INLINE __device__ __HOST_V__ CubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__
        CubeBase(NCubeBase<T, N> *data, int dim1, int dim2, int dim3)
            : data(data), dim1(dim1), dim2(dim2), dim3(dim3)
        {
        }
		INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2 * dim3; }
        DYNMEM_FUNC()
    };

    template <typename T, int N> struct NCubeBase<VectorBase<T>, N>
    {
        VectorBase<T> *data;
        intN<N> dims;

        INLINE __device__ __HOST_V__ NCubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ NCubeBase(VectorBase<T> *data, intN<N> dims)
            : data(data), dims(dims)
        {
        }
		INLINE __device__ __HOST_V__ int get_numel() const { return prod(dims); }
        DYNMEM_FUNC()
    };

    template <typename T, int N> struct NCubeBase<MatrixBase<T>, N>
    {
        MatrixBase<T> *data;
        intN<N> dims;

        INLINE __device__ __HOST_V__ NCubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ NCubeBase(MatrixBase<T> *data, intN<N> dims)
            : data(data), dims(dims)
        {
        }
		INLINE __device__ __HOST_V__ int get_numel() const { return prod(dims); }
        DYNMEM_FUNC()
    };

    template <typename T, int N> struct NCubeBase<CubeBase<T>, N>
    {
        CubeBase<T> *data;
        intN<N> dims;

        INLINE __device__ __HOST_V__ NCubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ NCubeBase(CubeBase<T> *data, intN<N> dims)
            : data(data), dims(dims)
        {
        }
		INLINE __device__ __HOST_V__ int get_numel() const { return prod(dims); }
        DYNMEM_FUNC()
    };

    template <typename T, int M, int N> struct NCubeBase<NCubeBase<T, M>, N>
    {
        NCubeBase<T, M> *data;
        intN<N> dims;

        INLINE __device__ __HOST_V__ NCubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ NCubeBase(NCubeBase<T, M> *data, intN<N> dims)
            : data(data), dims(dims)
        {
        }
		INLINE __device__ __HOST_V__ int get_numel() const { return prod(dims); }
        DYNMEM_FUNC()
    };

#undef DYNMEM_FUNC

    template <typename T, int N>
    INLINE __device__ __HOST_V__ NCubeBase<T, N> make_ncube(T *data, intN<N> dims)
    {
        NCubeBase<T, N> c;
        c.dims = dims;
        c.data = data;
        return c;
    }

    //===================================
    // Shared memory allocation
    //===================================
    template <typename T, int N>
    INLINE __device__ __HOST_V__ NCubeBase<T, N> shmem_alloc(shmem *_shmem, intN<N> dims)
    {
        return NCubeBase<T, N>(shmem_allocate<T>(_shmem, prod(dims)), dims);
    }

    template <typename T>
    INLINE __device__ __HOST_V__ NCubeBase<T, 4> shmem_alloc(shmem *_shmem, int4 dims)
    {
        return shmem_alloc<T, 4>(_shmem, (intN<4>)dims);
    }

    template <typename T, int N>
    INLINE __device__ __HOST_V__ NCubeBase<T, N> shmem_zeros(shmem *_shmem, intN<N> dims)
    {
        int n = prod(dims);
        NCubeBase<T, N> c = make_ncube(shmem_allocate<T>(_shmem, n), dims);
        shmem_fill<T>(c.data, n, 0.0f);
        return c;
    }

    template <typename T>
    INLINE __device__ __HOST_V__ NCubeBase<T, 4> shmem_zeros(shmem *_shmem, int4 dims)
    {
        return shmem_zeros<T, 4>(_shmem, (intN<4>)dims);
    }

//===================================
// Dynamic memory allocation
//===================================
#ifdef ENABLE_DYNAMIC_KERNEL_MEM

    template <typename T, int N>
    INLINE __device__ __HOST_V__ NCubeBase<T, N> uninit(const char *elemT, intN<N> dims)
    {
        return NCubeBase<T, N>(
            (T *)dynmem->allocate_addref(prod(dims) * sizeof(T)), dims);
    }

    template <typename T>
    INLINE __device__ __HOST_V__ NCubeBase<T, 4> uninit(const char *elemT, int4 dims)
    {
        return uninit<T, 4>(elemT, (intN<4>)dims);
    }

    template <typename T, int N>
    INLINE __device__ __HOST_V__ NCubeBase<T, N> zeros(const char *elemT, intN<N> dims)
    {
        NCubeBase<T, N> c = uninit<T, N>(elemT, dims);
        mem_fill<T>((T *)c.data, prod(dims), T());
        return c;
    }

    template <typename T>
    INLINE __device__ __HOST_V__ NCubeBase<T, 4> zeros(const char *elemT, int4 dims)
    {
        return zeros<T, 4>(elemT, (intN<4>)dims);
    }

    template <typename T, int N>
    INLINE __device__ __HOST_V__ NCubeBase<T, N> ones(const char *elemT, intN<N> dims)
    {
        NCubeBase<T, N> c = uninit<T, N>(elemT, dims);
        mem_fill<T>((T *)c.data, prod(dims), 1);
        return c;
    }

    template <typename T>
    INLINE __device__ __HOST_V__ NCubeBase<T, 4> ones(const char *elemT, int4 dims)
    {
        return ones<T, 4>(elemT, (intN<4>)dims);
    }

	template <typename R, int N>
    INLINE __device__ __HOST_V__ NCubeBase<R, N> atomicSet(NCubeBase<R, N> *tgt, NCubeBase<R,N> src)
    {
        void *current, *old = (void *)tgt->data;
        src.addref();
		src.addref();
        while ((current = atomicCASP((void *volatile *)&tgt->data, old,
                                     (void *)src.data)) != old)
            old = current;
        NCubeBase<R, N>((R *)current, tgt->dims).release();
		for (int i = 0; i < N; i ++)
			tgt->dims.el[i] = src.dims.el[i]; // NOTE - is not safe!
        return src;
    }

	template <typename T, int N> __device__ __HOST_V__ NCubeBase<T, N> copy(NCubeBase<T, N> x, const char *elemT)
	{
		NCubeBase<T, N> y = uninit<T, N>(elemT, x.dims);
		int M = x.get_numel();
		for (int m = 0; m < M; m++)
			y.data[m] = x.data[m];
		return y;
	}

	// Reshape function (all possible combinations)
	template <typename T, int N>
	__device__ __HOST_V__ VectorBase<T> reshape(NCubeBase<T, N> a, int dims,
		const char *elemT)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (dims != prod(a.dims))
			Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
		return VectorBase<T>(copy(a, elemT).data, dims);
	}

	template <typename T, int N>
	__device__ __HOST_V__ MatrixBase<T> reshape(NCubeBase<T, N> a, int2 dims,
		const char *elemT)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (dims.x * dims.y != prod(a.dims))
			Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
		return MatrixBase<T>(copy(a, elemT).data, dims.x, dims.y);
	}

	template <typename T, int N>
	__device__ __HOST_V__ CubeBase<T> reshape(NCubeBase<T, N> a, int3 dims,
		const char *elemT)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (prod(dims) != prod(a.dims))
			Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
		return CubeBase<T>(copy(a, elemT).data, dims.x, dims.y, dims.z);
	}

	template <typename T, int N, int M>
	__device__ __HOST_V__ NCubeBase<T, M> reshape(NCubeBase<T, N> a, intN<M> dims,
		const char *elemT)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (prod(dims) != prod(a.dims))
			Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
		return NCubeBase<T, M>(copy(a, elemT).data, dims);
	}

	template <typename T, int N>
	__device__ __HOST_V__ NCubeBase<T, N> reshape(VectorBase<T> a, intN<N> dims,
		const char *elemT)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (prod(dims) != a.dim1)
			Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
		return NCubeBase<T, N>(copy(a, elemT).data, dims);
	}

	template <typename T, int N>
	__device__ __HOST_V__ NCubeBase<T, N> reshape(MatrixBase<T> a, intN<N> dims,
		const char *elemT)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (prod(dims) != a.dim1 * a.dim2)
			Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
		return NCubeBase<T, N>(copy(a, elemT).data, dims);
	}

	template <typename T, int N>
	__device__ __HOST_V__ NCubeBase<T, N> reshape(CubeBase<T> a, intN<N> dims,
		const char *elemT)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (prod(dims) != a.dim1 * a.dim2 * a.dim3)
			Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
		return NCubeBase<T, N>(copy(a, elemT).data, dims);
	}

	// NCube autocast implementation
#define AUTOCAST_IMPL(R, T, Rstr)											    \
    template <int N>															\
	INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(T##1 a, intN<N> dims) { \
        NCubeBase<R, N> b = uninit<R>(Rstr, dims);                              \
        b.data[0] = a.x;                                                        \
        return b;                                                               \
    }	                                                                        \
    template <int N>															\
    INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(T##2 a, intN<N> dims) { \
        NCubeBase<R, N> b = uninit<R>(Rstr, dims);                              \
        b.data[0] = a.x; b.data[1] = a.y;                                       \
        return b;                                                               \
    }	                                                                        \
    template <int N>															\
    INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(T##3 a, intN<N> dims) { \
        NCubeBase<R, N> b = uninit<R>(Rstr, dims);                              \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z;                      \
        return b;                                                               \
    }	                                                                        \
    template <int N>															\
    INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(T##4 a, intN<N> dims) { \
        NCubeBase<R, N> b = uninit<R>(Rstr, dims);                              \
        b.data[0] = a.x; b.data[1] = a.y; b.data[2] = a.z;                      \
        return b;                                                               \
    }	                                                                        \
    template <int N>															\
	INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(T##N<N> a, intN<N> dims) \
	{                                                                           \
        NCubeBase<R, N> b = uninit<R>(Rstr, dims);                              \
        for (int i = 0; i < N; i ++)                                            \
            b.data[i] = a.el[i];                                                \
        return b;                                                               \
    }                                                                           \
    template <int N>															\
    INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(VectorBase<T> a, intN<N> dims) { \
        return reshape(a, dims, Rstr);											\
    }																			\
    template <int N>															\
    INLINE VectorBase<T> __device__ __HOST_V__ autocast(NCubeBase<R, N> a, ReturnType<VectorBase<T> >) {	\
        return reshape(a, prod(a.dims), Rstr);									\
    }																			\
	template <int N>															\
    INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(VectorBase<T> a, ReturnType<NCubeBase<R, N> >) {	\
		intN<N> dims;															\
		dims.el[0] = a.dim1;                                                    \
		for (int i=1; i<N; i ++)												\
			dims.el[i] = 1;										                \
        return reshape(a, dims, Rstr);											\
    }																			\
    template <int N>															\
    INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(MatrixBase<T> a, ReturnType<NCubeBase<R, N> >) {	\
		intN<N> dims;															\
		dims.el[0] = a.dim1;                                                    \
		dims.el[1] = a.dim2;                                                    \
		for (int i=2; i<N; i ++)												\
			dims.el[i] = 1;										                \
        return reshape(a, dims, Rstr);											\
    }																			\
    template <int N>															\
    INLINE NCubeBase<R, N> __device__ __HOST_V__ autocast(CubeBase<T> a, ReturnType<NCubeBase<R, N> >) {	\
		intN<N> dims;															\
		dims.el[0] = a.dim1;                                                    \
		dims.el[1] = a.dim2;                                                    \
		dims.el[2] = a.dim3;                                                    \
		for (int i=3; i<N; i ++)												\
			dims.el[i] = 1;										                \
        return reshape(a, dims, Rstr);											\
    }

	AUTOCAST_IMPL(scalar, scalar, "scalar");
	AUTOCAST_IMPL(cscalar, cscalar, "cscalar");

#undef AUTOCAST_IMPL
#endif

	//
	// Dimension upconversion
	//
	template <typename T, int N>
	__device__ __HOST_V__ NCubeBase<T, N> static_autocast(VectorBase<T> x, ReturnType<NCubeBase<T, N> >)
	{
		NCubeBase<T, N> y;
		y.data = x.data;
		y.dims.el[0] = 0;
		y.dims.el[1] = x.dim1;
		for (int i = 2; i < N; i++)
			y.dims.el[i] = 1;
		return y;
	}

	template <typename T, int N>
	__device__ __HOST_V__ NCubeBase<T, N> static_autocast(MatrixBase<T> x, ReturnType<NCubeBase<T, N> >)
	{
		NCubeBase<T, N> y;
		y.data = x.data;
		y.dims.el[0] = x.dim1;
		y.dims.el[1] = x.dim2;
		for (int i = 2; i < N; i++)
			y.dims.el[i] = 1;
		return y;
	}

	template <typename T, int N>
	__device__ __HOST_V__ NCubeBase<T, N> static_autocast(CubeBase<T> x, ReturnType<NCubeBase<T, N> >)
	{
		NCubeBase<T, N> y;
		y.data = x.data;
		y.dims.el[0] = x.dim1;
		y.dims.el[1] = x.dim2;
		y.dims.el[2] = x.dim3;
		for (int i = 3; i < N; i++)
			y.dims.el[i] = 1;
		return y;
	}

	template <typename T, int M, int N>
	__device__ __HOST_V__ NCubeBase<T, N> static_autocast(NCubeBase<T, M> x, ReturnType<NCubeBase<T, N> >)
	{
		if (N >= M) // Upconversion
		{
			NCubeBase<T, N> y;
			y.data = x.data;
			for (int i = 0; i < N && i < M; i++)
				y.dims.el[i] = x.dims.el[i];
			for (int i = M; i < N; i++)
				y.dims.el[i] = 1;
			return y;
		}
		else // (N < M) - Downconversion
		{
#ifdef CHECK_UNSAFE_BOUNDS
			for (int i = N; i < M; i++)
				if (x.dims.el[i] > 1)
					Q_RAISE_ERROR("Cannot convert cubeN: the dimensions don't match!");
#endif
			NCubeBase<T, N> y;
			for (int i = 0; i < N; i++)
				y.dims.el[i] = x.dims.el[i];
			y.data = x.data;
			return y;
		}
	}

	//
	// Dimension downconversion
	//
	template <typename T, int N>
	__device__ __HOST_V__ VectorBase<T> static_autocast(NCubeBase<T, N> x, ReturnType<VectorBase<T> >)
	{
		return VectorBase<T>(x.data, prod(x.dims));
	}

	template <typename T, int N>
	__device__ __HOST_V__ MatrixBase<T> static_autocast(NCubeBase<T, N> x, ReturnType<MatrixBase<T> >)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		for (int i = 2; i < N; i++)
			if (x.dims.el[i] > 1)
				Q_RAISE_ERROR("Cannot convert cubeN to a matrix: the dimensions don't match!");
#endif
		return MatrixBase<T>(x.data, x.dims.el[0], x.dims.el[1]);
	}

	template <typename T, int N>
	__device__ __HOST_V__ CubeBase<T> static_autocast(NCubeBase<T, N> x, ReturnType<CubeBase<T> >)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		for (int i = 3; i < N; i++)
			if (x.dims.el[i] > 1)
				Q_RAISE_ERROR("Cannot convert cubeN to a cube: the dimensions don't match!");
#endif
		return CubeBase<T>(x.data, x.dims.el[0], x.dims.el[1], x.dims.el[2]);
	}

#define STATIC_AUTOCAST_IMPL(T)															\
    INLINE NCubeBase<T, 4> __device__ __HOST_V__ static_autocast(T##1 a, int4 dims) {	\
        CHECK_LENGTH(prod(dims)==1); return NCubeBase<T, 4>(&a.x, (intN<4> &)dims); }   \
    INLINE NCubeBase<T, 4> __device__ __HOST_V__ static_autocast(T##2 a, int4 dims) {	\
        CHECK_LENGTH(prod(dims)==2); return NCubeBase<T, 4>(&a.x, (intN<4> &)dims); }   \
    INLINE NCubeBase<T, 4> __device__ __HOST_V__ static_autocast(T##3 a, int4 dims) {	\
        CHECK_LENGTH(prod(dims)==3); return NCubeBase<T, 4>(&a.x, (intN<4> &)dims); }   \
    INLINE NCubeBase<T, 4> __device__ __HOST_V__ static_autocast(T##4 a, int4 dims) {	\
        CHECK_LENGTH(prod(dims)==4); return NCubeBase<T, 4>(&a.x, (intN<4> &)dims); }   \
    template <int M>																	\
	INLINE NCubeBase<T, 4> __device__ __HOST_V__ static_autocast(T##N<M> a, int4 dims) {	\
        CHECK_LENGTH(prod(dims)==M); return NCubeBase<T, 4>(a.el, (intN<4> &)dims); }	\
	INLINE NCubeBase<T, 4> __device__ __HOST_V__ static_autocast(VectorBase<T> a, int4 dims) { \
        CHECK_LENGTH(prod(dims)==a.dim1); return NCubeBase<T, 4>(a.data, (intN<4> &)dims); }	\
    template <int N>																	\
    INLINE NCubeBase<T, N> __device__ __HOST_V__ static_autocast(T##1 a, intN<N> dims) { \
        CHECK_LENGTH(prod(dims)==1); return NCubeBase<T, N>(&a.x, dims); }				\
    template <int N>																	\
    INLINE NCubeBase<T, N> __device__ __HOST_V__ static_autocast(T##2 a, intN<N> dims) { \
        CHECK_LENGTH(prod(dims)==2); return NCubeBase<T, N>(&a.x, dims); }				\
    template <int N>																	\
    INLINE NCubeBase<T, N> __device__ __HOST_V__ static_autocast(T##3 a, intN<N> dims) { \
        CHECK_LENGTH(prod(dims)==3); return NCubeBase<T, N>(&a.x, dims); }				\
    template <int N>																	\
    INLINE NCubeBase<T, N> __device__ __HOST_V__ static_autocast(T##4 a, intN<N> dims) { \
        CHECK_LENGTH(prod(dims)==4); return NCubeBase<T, N>(&a.x, dims); }				\
    template <int N, int M>																\
	INLINE NCubeBase<T, N> __device__ __HOST_V__ static_autocast(T##N<M> a, intN<N> dims) { \
        CHECK_LENGTH(prod(dims)==M); return NCubeBase<T, N>(a.el, dims); }				\
    template <int N>																	\
	INLINE NCubeBase<T, N> __device__ __HOST_V__ static_autocast(VectorBase<T> a, intN<N> dims) { \
        CHECK_LENGTH(prod(dims)==a.dim1); return NCubeBase<T, N>(a.data, dims); } 

	STATIC_AUTOCAST_IMPL(scalar)
	STATIC_AUTOCAST_IMPL(cscalar)
	STATIC_AUTOCAST_IMPL(int)

#undef STATIC_AUTOCAST_IMPL


	template <int N> struct NCube : public NCubeBase<scalar, N> {};
	template <int N> struct NCCube : public NCubeBase<cscalar, N> {};
} // namespace quasar

// QUASAR_NCUBE_H_INCLUDED
#endif
