#pragma once
#include "quasar_host.h"
#include "quasar_dsl.h"
#include "dbgassert.h"
#include <functional>
#include <type_traits>

// Quasar runtime bindings
//
// Generates the machinery to wrap arbitrary functions using generic
// std::function<QLambdaDelegate> so that the functions can easily
// be called from Quasar. The wrapping heavily relies on variadic
// template functions.
//
// This header file is also included by C++ code generated by the
// high-level code generator back-end. This is to avoid direct
// dependencies between the generated code and the implementation
// of the runtime system.

namespace quasar
{
	using quasar::runtime::dbgWarning;
	using quasar::runtime::dbgAssert;

	// Forward declarations
	using namespace quasar::runtime;

	class bindings
	{
	public:
		static QUASARHOSTAPI void setup_kernel_function(
			qvalue_t qv_kernel,
			std::function<KernelLambdaDelegate> cpu_kernel_adapter,
			std::function<KernelLambdaDelegate> cuda_kernel_adapter,
			std::function<PrimeKernelLambdaDelegate> cuda_prime_kernel_adapter);
		static QUASARHOSTAPI void setup_device_function(
			qvalue_t qv_device,
			std::function<LambdaDelegate> cpu_device_adapter, void *cpu_device_func_ptr);
		static QUASARHOSTAPI void launch_kernel(compute_kernel *kernel, LPCVOID params, size_t params_size,
			const void * closure_ptr, const qvalue_t * args,
			int3 gridDim, int3 blockDim, bool prime_kernel, int deviceId, execution_stream *stream);
		static QUASARHOSTAPI bool is_kernel_initialized(compute_kernel *kernel);

	public:
		// Loads a kernel function with the specified signature
		// Note that here, we use a template function, this allows us
		// to use C++ tricks in order to automatically generate stub code.
		template <typename Ret, typename C, typename ...T>
		static qvalue_t LoadKernelFunction(qvalue_t closureArgs, LPCTSTR moduleName, LPCTSTR functionSignature, LPCTSTR funcName,
			FunctionFlags flags, PlatformTargets platformTargets, LoadFunctionParams &params,
			Ret(*cpuKernelFunction)(C*, T...))
		{
			IQuasarHost * host = IQuasarHost::GetInstance();
			qvalue_t qv = host->LoadFunction(closureArgs, moduleName, functionSignature, funcName, flags, platformTargets, params);
			compute_kernel *kernel = static_cast<compute_kernel *>(qv.private_obj);

			if (!is_kernel_initialized(kernel))
			{
				setup_kernel_function(qv,
					cpu_build_kernel_call_adapter(cpuKernelFunction),
					gpu_build_kernel_call_adapter(kernel, cpuKernelFunction),
					gpu_build_prime_kernel_call_adapter(kernel, cpuKernelFunction));
			}
			return qv;
		}

		template <typename Ret, typename C, typename ...T>
		static qvalue_t LoadDeviceFunction(qvalue_t closureArgs, LPCTSTR moduleName, LPCTSTR functionSignature, LPCTSTR funcName,
			FunctionFlags flags, PlatformTargets platformTargets, LoadFunctionParams &params,
			Ret(*cpuKernelFunction)(C*, T...))
		{
			Assert(cpuKernelFunction != nullptr, "Null pointer specified!");

			IQuasarHost * host = IQuasarHost::GetInstance();
			qvalue_t qv = host->LoadFunction(closureArgs, moduleName, functionSignature, funcName, flags, platformTargets, params);

			setup_device_function(qv, cpu_build_call_adapter(cpuKernelFunction),
				reinterpret_cast<void *>(cpuKernelFunction));
			return qv;
		}

	private:
		/*=============================================================
			CUDA kernel binding
		  =============================================================*/
		template <int alignment>
		static inline size_t align(size_t size)
		{
			static_assert((alignment & (alignment - 1)) == 0, "Alignment should be a power of two!");
			return (size + alignment - 1) & ~(alignment - 1);
		}

		template <int N, typename R>
		static inline void gpu_arg_layout(uint8_t buf[N], size_t &params_size, int index, int cnt, R current)
		{
			if (index < cnt && params_size + sizeof(R) <= N)
			{
				constexpr int alignment = (int)std::alignment_of<R>::value;
				params_size = align<alignment>(params_size);
				*(R *)&buf[params_size] = current;
				params_size += sizeof(R);
			}
		}

	#ifdef CPLUSPLUS14
		template <int N, typename R, typename ...T>
		static inline void gpu_arg_layout(uint8_t buf[N], size_t &params_size, int index, int cnt, R current, T... cargs)
		{
			gpu_arg_layout<N>(buf, params_size, index, cnt, current);
			gpu_arg_layout<N>(buf, params_size, index + 1, cnt, cargs...);
		}

		template <int N, typename ...T, size_t ...I>
		static inline void gpu_kernel_call_adapter_helper(compute_kernel* kernel,
			void* closurePtr, int3 gridDim, int3 blockDim, int nThreads, int deviceId, const qvalue_t * args, execution_stream *stream, 
			std::tuple<std::decay_t<T>...>& cargs, std::index_sequence<I...>)
		{
			// Fill up the first N-3 elements
			marshal_arguments(deviceId, N - 3, args, std::get<I>(cargs)...);

			uint8_t params_base[1024];
			size_t params_size = 0;
			gpu_arg_layout<1024>(params_base, params_size, 0, N - 2, closurePtr, std::get<I>(cargs)...);

			// Launch the kernel function on the GPU
			launch_kernel(kernel, params_base, params_size, closurePtr, args, gridDim, blockDim, false, deviceId, stream);
		}

		template <int N, typename ...T, size_t ...I>
		static inline void gpu_prime_kernel_call_adapter_helper(compute_kernel* kernel,
			void* closurePtr, int3 dataDim, int3 gridDim, int3 blockDim, int nThreads, int deviceId, const qvalue_t * args, execution_stream *stream, 
			std::tuple<std::decay_t<T>...>& cargs, std::index_sequence<I...>)
		{
			// Fill up the first N-3 elements
			marshal_arguments(deviceId, N - 3, args, std::get<I>(cargs)...);

			uint8_t params_base[1040];
			size_t params_size = sizeof(int4);
			*(int3*)&params_base[0] = dataDim;
			gpu_arg_layout<1024>(&params_base[0],
				params_size, 0, N - 2, closurePtr, std::get<I>(cargs)...);

			// Launch the kernel function on the GPU
			launch_kernel(kernel, params_base, params_size, closurePtr, args, gridDim, blockDim, true, deviceId, stream);
		}

		// Type transform function to map CPU types onto the parameter to be passed to the GPU.
		// We convert lreferences & to pointers (so that eventually a device pointer will be passed).
		template <typename T>
		struct gpu_type_transform
		{
			typedef std::conditional_t<std::is_lvalue_reference<T>::value, void*, std::decay_t<T>> type;
		};

		// Builds a call adapter for kernel functions
		// Note: the kernel signature contains CPU parameters (int3, int3, int) that can safely be discarded
		template <typename Ret, typename C, typename ...T>
		static auto gpu_build_kernel_call_adapter(compute_kernel *kernel, Ret(*)(C*, T...))
		{
			std::function<KernelLambdaDelegate> del = [kernel](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut, int3 gridDim, int3 blockDim,
				int nThreads, int deviceId, execution_stream *stream)
			{
				const int N = sizeof...(T);
				Assert(N == nArgsIn + 3, "Invalid number of arguments for function!");
				std::tuple<typename gpu_type_transform<T>::type...> args;
				gpu_kernel_call_adapter_helper<N, typename gpu_type_transform<T>::type...>(kernel, closurePtr, gridDim, blockDim, nThreads, deviceId, argsIn,
					(execution_stream*)stream, args, std::make_index_sequence<N>{});
			};
			return del;
		}

		template <typename Ret, typename C, typename ...T>
		static auto gpu_build_prime_kernel_call_adapter(compute_kernel *kernel, Ret(*)(C*, T...))
		{
			std::function<PrimeKernelLambdaDelegate> del = [kernel](void *closurePtr,
				const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut, int3 dataDim, int3 gridDim, int3 blockDim,
				int nThreads, int deviceId, execution_stream *stream)
			{
				const int N = sizeof...(T);
				Assert(N == nArgsIn + 3, "Invalid number of arguments for function!");
				std::tuple<typename gpu_type_transform<T>::type...> args;
				gpu_prime_kernel_call_adapter_helper<N, typename gpu_type_transform<T>::type...>(kernel, closurePtr, dataDim, gridDim, blockDim, nThreads, deviceId, argsIn,
					stream, args, std::make_index_sequence<N>{});
			};
			return del;
		}

		/*=============================================================
		  CPU kernel binding
		  =============================================================*/
		template <typename ...T, size_t ...I>
		static void cpu_call_unmarshal_helper(int deviceId, int N, const qvalue_t* inArgs, qvalue_t* outArgs, std::tuple<std::decay_t<T>...>& cargs,
			std::index_sequence<I...>)
		{
			unmarshal_arguments(deviceId, 0, N, const_cast<qvalue_t*>(inArgs), outArgs - N, (std::tuple<T...>*)nullptr, std::get<I>(cargs)...);
		}

		template <typename Ret, typename C, typename ...T, size_t ...I>
		static inline Ret cpu_call_adapter_helper(int deviceId, Ret(*kernel)(C*, T...), int N, C* closurePtr, const qvalue_t * args, std::tuple<std::decay_t<T>...>& cargs,
			std::index_sequence<I...>)
		{
			marshal_arguments(deviceId, N, args, std::get<I>(cargs)...);
			return kernel(closurePtr, std::get<I>(cargs)...);
		}

		template <typename Ret, typename C, typename ...T, size_t ...I>
		static inline Ret cpu_kernel_call_adapter_helper(int deviceId, Ret(*kernel)(C*, T...), C* closurePtr, int3 gridDim, int3 blockDim,
			int nThreads, execution_stream * stream, const qvalue_t * args, std::tuple<std::decay_t<T>...>& cargs,
			std::index_sequence<I...>)
		{
			const int N = sizeof...(T);
			// Fill up the first N-3 elements
			marshal_arguments(deviceId, N - 3, args, std::get<I>(cargs)...);
			std::get<N - 3>(cargs) = gridDim;
			std::get<N - 2>(cargs) = blockDim;
			std::get<N - 1>(cargs) = nThreads;
			return kernel(closurePtr, std::get<I>(cargs)...);
		}

		// Builds a call adapter for device functions
		template <typename Ret, typename C, typename ...T>
		static inline auto cpu_build_call_adapter(Ret(*kernel)(C*, T...))
		{
			std::function<LambdaDelegate> del = [kernel](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				const int N = sizeof...(T);
				int deviceId = get_cpu_devid();

				Assert(N == nArgsIn && nArgsOut == 1, "Invalid number of arguments for function!");
				std::tuple<std::decay_t<T>...> args;
				Ret ret = cpu_call_adapter_helper(deviceId, kernel, N, (C*)closurePtr, argsIn, args, std::make_index_sequence<N>{});
				cpu_call_unmarshal_helper<T...>(deviceId, N, argsIn, nullptr, args, std::make_index_sequence<N>{});
				// Dispatch the result
				unmarshal(deviceId, ret, argsOut[0]);
			};
			return del;
		}

		// Builds a call adapter for device functions with multiple return values
		// The last nArgsIn arguments are the return values (passed using lreference &)
		template <typename C, typename ...T>
		static inline auto cpu_build_call_adapter(void(*kernel)(C*, T...))
		{
			std::function<LambdaDelegate> del = [kernel](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				const int N = sizeof...(T);
				int deviceId = get_cpu_devid();

				Assert(N == nArgsIn + nArgsOut || nArgsOut == 1 && N == nArgsIn, "Invalid number of arguments for function!");
				std::tuple<std::decay_t<T>...> args;
				cpu_call_adapter_helper<void>(deviceId, kernel, nArgsIn, (C*)closurePtr, argsIn, args, std::make_index_sequence<N>{});
				// Dispatch the result
				if (N == nArgsIn + nArgsOut)
					cpu_call_unmarshal_helper<T...>(deviceId, N - nArgsOut, argsIn, argsOut, args, std::make_index_sequence<N>{});
				else
					cpu_call_unmarshal_helper<T...>(deviceId, N, argsIn, nullptr, args, std::make_index_sequence<N>{});
			};
			return del;
		}

		template <typename Ret, typename C>
		static inline auto cpu_build_call_adapter(Ret(*kernel)(C*))
		{
			std::function<LambdaDelegate> del = [kernel](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				Assert(0 == nArgsIn && nArgsOut == 1, "Invalid number of arguments for function!");
				int deviceId = get_cpu_devid();
				Ret ret = kernel((C*)closurePtr);
				unmarshal(deviceId, ret, argsOut[0]);
			};
			return del;
		}

		template <typename C>
		static inline auto cpu_build_call_adapter(void(*kernel)(C*))
		{
			std::function<LambdaDelegate> del = [kernel](void *closurePtr, const qvalue_t *argsIn, int nArgsIn, qvalue_t *argsOut, int nArgsOut)
			{
				int deviceId = get_cpu_devid();
				Assert(0 == nArgsIn + nArgsOut || nArgsOut == 1 && 0 == nArgsIn, "Invalid number of arguments for function!");
				kernel((C*)closurePtr);
			};
			return del;
		}

		// Builds a call adapter for kernel functions
		template <typename Ret, typename C, typename ...T>
		static inline auto cpu_build_kernel_call_adapter(Ret(*kernel)(C*, T...))
		{
			std::function<KernelLambdaDelegate> del = [kernel](void *closurePtr, const qvalue_t *argsIn, int nArgsIn,
				qvalue_t *argsOut, int nArgsOut, int3 gridDim, int3 blockDim, int nThreads, int deviceId, execution_stream *stream)
			{
				const int N = sizeof...(T);
				Assert(N == nArgsIn + 3, "Invalid number of arguments for function!");
				std::tuple<std::decay_t<T>...> args;
				cpu_kernel_call_adapter_helper(deviceId, kernel, (C*)closurePtr, gridDim, blockDim, nThreads, stream, argsIn, args, std::make_index_sequence<N>{});
				cpu_call_unmarshal_helper<T...>(deviceId, N, argsIn, nullptr, args, std::make_index_sequence<N>{});
			};
			return del;
		}
	#endif		// CPLUSPLUS14

		/*=============================================================
		  Marshaller implementation
		  =============================================================*/
	public:
		static QUASARHOSTAPI void internal_marshal_fixed_length_array(int deviceId, const qvalue_t&, void* data, size_t bytes);
		static QUASARHOSTAPI void internal_marshal_vector(int deviceId, const qvalue_t&, void*& data, int& dim1);
		static QUASARHOSTAPI void internal_marshal_matrix(int deviceId, const qvalue_t&, void*& data, int& dim1, int& dim2);
		static QUASARHOSTAPI void internal_marshal_cube(int deviceId, const qvalue_t&, void*& data, int& dim1, int& dim2, int& dim3);
		static QUASARHOSTAPI void internal_marshal_ncube(int deviceId, const qvalue_t&, void*& data, int* dims, int N);
		static QUASARHOSTAPI void internal_marshal_lambda(int deviceId, const qvalue_t&, void*& func_ptr, void*& closure_ptr);
		static QUASARHOSTAPI void* internal_marshal_obj(int deviceId, const qvalue_t&);
		static QUASARHOSTAPI void internal_unmarshal_vecx(int deviceId, qvalue_t&, const void* data, int dim1, size_t elemSize);
		static QUASARHOSTAPI void internal_unmarshal_vector(int deviceId, qvalue_t&, const void* data, int dim1);
		static QUASARHOSTAPI void internal_unmarshal_matrix(int deviceId, qvalue_t&, const void* data, int dim1, int dim2);
		static QUASARHOSTAPI void internal_unmarshal_cube(int deviceId, qvalue_t&, const void* data, int dim1, int dim2, int dim3);
		static QUASARHOSTAPI void internal_unmarshal_ncube(int deviceId, qvalue_t&, const void* data, const int *dims, int N);
		static QUASARHOSTAPI void internal_unmarshal_lambda(int deviceId, qvalue_t&, const void* ptr, const void *data);
		static QUASARHOSTAPI void internal_unmarshal_obj(int deviceId, qvalue_t&, void *ptr);
		static QUASARHOSTAPI int get_cpu_devid();

	private:
		/*
		 * Hack to avoid reference counting mechanism
		 */
		static void Empty(QValue &v)
		{
			((qvalue_t &)v).type = TYPE_VOID;
		}

	public:
		 /*
		  * Generic marshalling function, used for scalar/cscalar as well as vecX types.
		  */
		template <typename T>
		static void marshal(int deviceId, qvalue_t in, T &out)
		{
			QValue v(in);
			out = (T)v;
			Empty(v);
		}

		template <typename T>
		static void marshal(int deviceId, qvalue_t in, VectorBase<T> &out)
		{
			internal_marshal_vector(deviceId, in, (void *&)out.data, out.dim1);
		}

		template <typename T>
		static void marshal(int deviceId, qvalue_t in, MatrixBase<T> &out)
		{
			internal_marshal_matrix(deviceId, in, (void *&)out.data, out.dim1, out.dim2);
		}

		template <typename T>
		static void marshal(int deviceId, qvalue_t in, CubeBase<T> &out)
		{
			internal_marshal_cube(deviceId, in, (void *&) out.data, out.dim1, out.dim2, out.dim3);
		}

		template <typename T, int N>
		static void marshal(int deviceId, qvalue_t in, NCubeBase<T, N> &out)
		{
			internal_marshal_ncube(deviceId, in, (void *&)out.data, out.dims.el, N);
		}

		template <typename C, typename F>
		static void marshal(int deviceId, qvalue_t in, function_ptr<C, F> &out)
		{
			internal_marshal_lambda(deviceId, in, (void*&)out.ptr, (void*&)out.data);
		}

		template <typename T>
		static void marshal(int deviceId, qvalue_t in, object_ptr<T> &out)
		{
			out = (T*)internal_marshal_obj(deviceId, in);
		}


		/*
		 * Generic unmarshalling function, used for scalar/cscalar as well as vecX types.
		 */
		template <typename T>
		static inline void unmarshal_default(const T &in, qvalue_t &out)
		{
			QValue v = in;
			out = (qvalue_t)v;
			v.Retain();
		}

		template <typename T>
		static inline void unmarshal_byref(int deviceId, const T &in, qvalue_t &out)
		{
		}

		template <int N>
		static inline void unmarshal_byref(int deviceId, const scalarN<N> &in, qvalue_t &out)
		{
			internal_unmarshal_vecx(deviceId, out, (void *)&in, N, sizeof(scalar));
		}
		template <int N>
		static inline void unmarshal_byref(int deviceId, const intN<N> &in, qvalue_t &out)
		{
			internal_unmarshal_vecx(deviceId, out, (void *)&in, N, sizeof(int));
		}
		template <int N>
		static inline void unmarshal_byref(int deviceId, const cscalarN<N> &in, qvalue_t &out)
		{
			internal_unmarshal_vecx(deviceId, out, (void *)&in, N, sizeof(cscalar));
		}

		// Structs
		static QUASARHOSTAPI void internal_unmarshal_typed_object(int deviceId, const void *ptr, size_t size, LPCTSTR typeName, qvalue_t &out);

		template <typename T>
		static inline void unmarshal(int deviceId, const T &in, qvalue_t &out)
		{
			internal_unmarshal_typed_object(deviceId, &in, sizeof(T), typeinfo_t<T>::typeName, out);
		}

		// Specialized implementations for all other data types
		template <int N>
		static inline void unmarshal(int deviceId, const scalarN<N> &in, qvalue_t &out) { unmarshal_default(in, out); }
		template <int N>
		static inline void unmarshal(int deviceId, const cscalarN<N> &in, qvalue_t &out) { unmarshal_default(in, out); }
		template <int N>
		static inline void unmarshal(int deviceId, const intN<N> &in, qvalue_t &out) { unmarshal_default(in, out); }

		template <typename T>
		static void unmarshal(int deviceId, const VectorBase<T> &in, qvalue_t &out)
		{
			internal_unmarshal_vector(deviceId, out, in.data, in.dim1);
		}

		template <typename T>
		static void unmarshal(int deviceId, const MatrixBase<T> &in, qvalue_t &out)
		{
			internal_unmarshal_matrix(deviceId, out, in.data, in.dim1, in.dim2);
		}

		template <typename T>
		static void unmarshal(int deviceId, const CubeBase<T> &in, qvalue_t &out)
		{
			internal_unmarshal_cube(deviceId, out, in.data, in.dim1, in.dim2, in.dim3);
		}

		template <typename T, int N>
		static void unmarshal(int deviceId, const NCubeBase<T, N> &in, qvalue_t &out)
		{
			internal_unmarshal_ncube(deviceId, out, in.data, in.dims.el, N);
		}


		template <typename C, typename F>
		static void unmarshal(int deviceId, const function_ptr<C, F> &in, qvalue_t &out)
		{
			internal_unmarshal_lambda(deviceId, out, in.ptr, in.data);
		}

		template <typename T>
		static void unmarshal(int deviceId, object_ptr<T> &in, qvalue_t &out)
		{
			internal_unmarshal_obj(deviceId, out, (T *)in);
		}

		template <typename R>
		static void marshal_arguments(int deviceId, int N, const qvalue_t* inArgs, R& current)
		{
			if (N > 0)
				marshal(deviceId, inArgs[0], current);
		}

		template <typename R, typename ...T>
		static void marshal_arguments(int deviceId, int N, const qvalue_t* inArgs, R& current, T&... args)
		{
			marshal_arguments(deviceId, N, inArgs, current);
			marshal_arguments(deviceId, N - 1, inArgs + 1, args...);
		}

		template <typename R>
		static void unmarshal_arguments(int deviceId, int index, int N, qvalue_t *inArgs, qvalue_t *outArgs, std::tuple<R>*, std::decay_t<R>& current)
		{
			if (index >= N)
				unmarshal(deviceId, current, outArgs[0]);
			else if CONSTEXPR (std::is_lvalue_reference<R>::value)
				unmarshal_byref(deviceId, current, inArgs[0]);
		}

		template <typename R, typename ...T>
		static void unmarshal_arguments(int deviceId, int index, int N, qvalue_t *inArgs, qvalue_t* outArgs, std::tuple<R, T...>*, std::decay_t<R>& current, std::decay_t<T>&... args)
		{
			unmarshal_arguments(deviceId, index, N, inArgs, outArgs, (std::tuple<R>*)nullptr, current);
			unmarshal_arguments(deviceId, index + 1, N, inArgs + 1, outArgs + 1, (std::tuple<T...>*)nullptr, args...);
		}
	};

	// Specialized implementations for all other data types
	template <>
	inline void bindings::marshal<void*>(int deviceId, qvalue_t in, void* &out)
	{
		int dummy; internal_marshal_vector(deviceId, in, out, dummy);
	}

	template <>
	inline void bindings::unmarshal<float>(int deviceId, const float &in, qvalue_t &out) { out = qvalue_t::fromScalar(in); }
	template <>
	inline void bindings::unmarshal<double>(int deviceId, const double &in, qvalue_t &out) { out = qvalue_t::fromScalar(in); }
	template <>
	inline void bindings::unmarshal<int>(int deviceId, const int &in, qvalue_t &out) { out = qvalue_t::fromInt(in); }
	template <>
	inline void bindings::unmarshal<bool>(int deviceId, const bool &in, qvalue_t &out) { out = qvalue_t::fromInt(in); }
	template <>
	inline void bindings::unmarshal<size_t>(int deviceId, const size_t &in, qvalue_t &out) { out = qvalue_t::fromLong(in); }
	template <>
	inline void bindings::unmarshal<long long>(int deviceId, const long long &in, qvalue_t &out) { out = qvalue_t::fromLong(in); }
	template <>
	inline void bindings::unmarshal<cscalar>(int deviceId, const cscalar &in, qvalue_t &out) { out = qvalue_t::fromCScalar(in); }
	template <>
	inline void bindings::unmarshal<LPTSTR>(int deviceId, const LPTSTR &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<scalar1>(int deviceId, const scalar1 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<scalar2>(int deviceId, const scalar2 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<scalar3>(int deviceId, const scalar3 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<scalar4>(int deviceId, const scalar4 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<cscalar1>(int deviceId, const cscalar1 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<cscalar2>(int deviceId, const cscalar2 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<cscalar3>(int deviceId, const cscalar3 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<cscalar4>(int deviceId, const cscalar4 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<int1>(int deviceId, const int1 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<int2>(int deviceId, const int2 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<int3>(int deviceId, const int3 &in, qvalue_t &out) { unmarshal_default(in, out); }
	template <>
	inline void bindings::unmarshal<int4>(int deviceId, const int4 &in, qvalue_t &out) { unmarshal_default(in, out); }

	template <>
	inline void bindings::unmarshal_byref<scalar1>(int deviceId, const scalar1 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*) &in, 1, sizeof(scalar)); }
	template <>
	inline void bindings::unmarshal_byref<scalar2>(int deviceId, const scalar2 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 2, sizeof(scalar)); }
	template <>
	inline void bindings::unmarshal_byref<scalar3>(int deviceId, const scalar3 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 3, sizeof(scalar)); }
	template <>
	inline void bindings::unmarshal_byref<scalar4>(int deviceId, const scalar4 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 4, sizeof(scalar)); }
	template <>
	inline void bindings::unmarshal_byref<cscalar1>(int deviceId, const cscalar1 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 1, sizeof(cscalar)); }
	template <>
	inline void bindings::unmarshal_byref<cscalar2>(int deviceId, const cscalar2 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 2, sizeof(cscalar)); }
	template <>
	inline void bindings::unmarshal_byref<cscalar3>(int deviceId, const cscalar3 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 3, sizeof(cscalar)); }
	template <>
	inline void bindings::unmarshal_byref<cscalar4>(int deviceId, const cscalar4 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 4, sizeof(cscalar)); }
	template <>
	inline void bindings::unmarshal_byref<int1>(int deviceId, const int1 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 1, sizeof(int)); }
	template <>
	inline void bindings::unmarshal_byref<int2>(int deviceId, const int2 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 2, sizeof(int)); }
	template <>
	inline void bindings::unmarshal_byref<int3>(int deviceId, const int3 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 3, sizeof(int)); }
	template <>
	inline void bindings::unmarshal_byref<int4>(int deviceId, const int4 &in, qvalue_t &out) { internal_unmarshal_vecx(deviceId, out, (void*)&in, 4, sizeof(int)); }

}
