#!/bin/bash
## Simple install script for Quasar and Redshift
## Dirk Van Haerenborgh <dirk.vanhaerenborgh@ugent.be>
## Adapted by Simon Donné <sdonn@telin.ugent.be>
## Maintained by Bart Goossens <bart.goossens@ugent.be>

clear

PROGNAME=${0##*/}

usage()
{
cat << EOF
usage: $PROGNAME options

Installs Quasar.

OPTIONS:
   -h, --help  Show this message
   --prefix    Preferred location [optional][unused]
EOF
}

SHORTOPTS="h"
LONGOPTS="help,prefix:,intermediate"
ARGS=$(getopt -s bash -o $SHORTOPTS -l $LONGOPTS --name $PROGNAME -- "$@" )

CURDIR=$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )
PREFIX=$CURDIR

TMPDIR=

eval set -- "$ARGS"

while [ $# -gt 0 ]; do
   case $1 in
      -h|--help)
         usage
         exit 0
         ;;
      --intermediate)
         TMPDIR=$2
         shift
         ;;
      --)
         shift
         break
         ;;
      *)
         shift
         break
         ;;
   esac
   shift
done


## CHECK DEPENDENCIES

#test string type versions
check_version()
{
    local version=$1 check=$2
    local highest=$(echo -e "$version\n$check" | sed '/^$/d' | sort -Vr | head -1)
    [[ "$highest" = "$version" ]] && return 0
    return 1
}

#version comparison
vercomp () {
    if [[ $1 == $2 ]]
    then
        return 0
    fi
    local IFS=.
    local i ver1=($1) ver2=($2)
    # fill empty fields in ver1 with zeros
    for ((i=${#ver1[@]}; i<${#ver2[@]}; i++))
    do
        ver1[i]=0
    done
    for ((i=0; i<${#ver1[@]}; i++))
    do
        if [[ -z ${ver2[i]} ]]
        then
            # fill empty fields in ver2 with zeros
            ver2[i]=0
        fi
        if ((10#${ver1[i]} > 10#${ver2[i]}))
        then
            return 1
        fi
        if ((10#${ver1[i]} < 10#${ver2[i]}))
        then
            return 2
        fi
    done
    return 0
}

#Check mono and its version. If there is a mono version coming with Quasar,
# use that one, unless the installed mono has a newer version
VERSION_STR=`mono --version 2>/dev/null`
if [ $? -ne 0 ]; then
    INSTALLED_MONO_VER='0.0'
else
    INSTALLED_MONO_VER=`echo "$VERSION_STR" | awk '/version/ { print $5 }'`
fi
vercomp $INSTALLED_MONO_VER '5.2.0.175'
if [ $? -eq 2 ] && [ -d "Quasar/mono/bin" ]; then
    MONO_BIN=Quasar/mono/bin/mono
else
    MONO_BIN=$(which mono);
    if [[ -z "$MONO_BIN" ]]; then
	      echo "It appears you do not have mono installed. Exiting...";
        exit 1
    fi
fi

MONO_VERSTR=$($MONO_BIN --version);
MONO_VERARR=( $MONO_VERSTR );
MONO_VERSION=$(echo ${MONO_VERARR[4]} );

MONO_REQUIRED="2.10.8" #We need at least mono 2.10.8
check_version $MONO_VERSION $MONO_REQUIRED
if [[ $? -gt 0 ]] ; then
    echo "You do not have mono $MONO_REQUIRED or higher! Exiting...";
    exit 1
fi

#Check cuda and its version
CUDA_BIN=$(which nvcc);
if [[ -z "$CUDA_BIN" ]]; then
	echo "Warning: It appears you do not have cuda installed.";
else
    CUDA_VERSTR=$($CUDA_BIN --version | grep release);
    CUDA_VERARR=( $CUDA_VERSTR );
    CUDA_VERSION=$(echo ${CUDA_VERARR[4]} | sed 's/,$//' );

    #We need at least cuda 4.0
    CUDA_REQUIRED="4.0"
    check_version $CUDA_VERSION $CUDA_REQUIRED
    if [[ $? -gt 0 ]] ; then
        echo "Warning: Quasar CUDA backend needs at least cuda $CUDA_REQUIRED.";
    fi
fi

## INSTALL

BINDIR=${PREFIX}/bin
INSTALLDIR=${PREFIX}/Quasar
USERCONFDIR="${XDG_CONFIG_HOME:-$HOME/.config}/Quasar/"
LOGDIR="${LOGDIR:-${XDG_CONFIG_HOME:-$HOME/.config}/Quasar/}"
TMPDIR=${TMPDIR:-/tmp/$USER/quasar}

mkdir -p ${BINDIR}
install -d ${LOGDIR}
install -d ${TMPDIR}

cd Quasar

chmod 755 scripts/link_script.sh
chmod 755 scripts/cc_script.sh
chmod 755 scripts/nvcc_script.sh

# configure and install desktop file
install -p redshift.desktop redshift.desktop.bak
sed -i -re "s|^(Exec)=(.*)|\1=${BINDIR}/redshift %F|g" redshift.desktop
sed -i -re "s|^(Icon)=(.*)|\1=${INSTALLDIR}/redshift.png|g" redshift.desktop
sed -i -re "s|^(Path)=(.*)|\1=${INSTALLDIR}|g" redshift.desktop
mkdir -p ~/.local/share/applications
install -m 644 -p redshift.desktop ~/.local/share/applications
install -p redshift.desktop.bak redshift.desktop
rm -f redshift.desktop.bak

#fill in correct values in the shell executables
sed -i -re "s|^(TMPDIR)=(.*)|\1=${TMPDIR}|g" quasar
sed -i -re "s|^(TMPDIR)=(.*)|\1=${TMPDIR}|g" redshift
sed -i -re "s|^(TMPDIR)=(.*)|\1=${TMPDIR}|g" spectroscope
sed -i  -e "s|\${0%%/bin/quasar}/lib/quasar|${INSTALLDIR}|g" quasar
sed -i  -e "s|\${0%%/bin/redshift}/lib/quasar|${INSTALLDIR}|g" redshift
sed -i  -e "s|\${0%%/bin/spectroscope}/lib/quasar|${INSTALLDIR}|g" spectroscope

install -m 755 -p quasar ${BINDIR}
install -m 755 -p redshift ${BINDIR}
install -m 755 -p spectroscope ${BINDIR}

declare -A QUASAR_CONFIG=( [NVCC_PATH]="$INSTALLDIR/nvcc_script.sh"
                           [CC_PATH]="$INSTALLDIR/cc_script.sh"
                           [LINK_PATH]="$INSTALLDIR/link_script.sh" )

declare -A QUASAR_RUNTIME_CONFIG=( [INTERMEDIATE_DIR]=$TMPDIR
                                   [MODULE_DIR]="$INSTALLDIR;$INSTALLDIR/Library;$INSTALLDIR/Samples;$INSTALLDIR/Interop_Samples;$INSTALLDIR/res/data;$INSTALLDIR/res/images"
                                   [DLL_SEARCHPATH64]="$INSTALLDIR/OS_Runtimes/Ubuntu_x64/" )

declare -A QUASAR_REDSHIFT_CONFIG=( [HELP_PATH]="$INSTALLDIR/Doc" )


# now install the config files. We need to configure them first
# config files are XML, we parse them in here

trim() {
    local var="$@"
    shopt -s extglob
    var="${var#"${var%%[![:space:]]*}"}"   # remove leading whitespace
    var="${var%"${var##*[![:space:]]}"}"   # remove trailing whitespace
    shopt -u extglob
    echo -n "$var"
}

toupper() {
    local var=$1
    var=$(echo $var | tr [:lower:] [:upper:])
    echo -n $var
}

NEXT_CONTENT=""

read_dom () {
    local IFS=\<
    read -r -d \> CONTENT ENTITY
    local RET=$?
    TAG_NAME=${ENTITY%% *}
    ATTRIBUTES=${ENTITY#* }
    return $RET
}

parse_dom () {
    if [[ -z $ENTITY ]]; then
        return 1
    fi
    local CONTENT=$(trim $CONTENT)
    local MAP="$( declare -p ${1} )";
    eval "declare -A MAP=${MAP#*=}"

    #change values
    if [[ $TAG_NAME = "setting" ]] ; then
        eval local $ATTRIBUTES
        local NAMEUP=$(toupper $name)
        NEXT_CONTENT=${MAP[$NAMEUP]}
    fi

    #assemble xml again
    if [[ $ENTITY == "/value" ]]; then
        echo "${NEXT_CONTENT:-$CONTENT}<$ENTITY>"
        NEXT_CONTENT=""
    else
        echo -n "<$ENTITY>"
    fi
}

install_xml() {
    NEXT_CONTENT=""
    while read_dom; do
		parse_dom $1
	done < $2 >$3
}

#associative arrays are passed by name
# install_xml QUASAR_CONFIG "Quasar.config.xml" "Quasar.config.xml.tmp"
# mv Quasar.config.xml.tmp Quasar.config.xml

# Only install Quasar.Runtime.config.xml if non-existing
if [ ! -f $USERCONFDIR/Quasar.Runtime.config.xml ]; then
    install_xml QUASAR_RUNTIME_CONFIG "Quasar.Runtime.config.xml" "Quasar.Runtime.config.xml.tmp"
    install -m 644 -p Quasar.Runtime.config.xml.tmp $USERCONFDIR/Quasar.Runtime.config.xml
fi

# Remove the local config file (otherwise this file may be used)
rm -f Quasar.Runtime.config.{xml,tmp}

# Update mcs script if mono is bundled

if [ -f mono/bin/mcs ]; then
cat > mono/bin/mcs << EOF
#!/bin/sh
exec $PREFIX/Quasar/mono/bin/mono $MONO_OPTIONS $PREFIX/Quasar/mono/lib/mono/4.5/mcs.exe "\$@"
EOF
chmod +x mono/bin/mcs
fi

echo "Quasar was installed to this directory"
