//=============================================================================
// Quasar CUDA / OpenMP runtime backend
// (C) 2011-2020 Gepura
//=============================================================================
#ifndef QUASAR_H_INCLUDED
#define QUASAR_H_INCLUDED

#ifndef __CUDACC_RTC__
#include <limits.h>
#else
#define SCHAR_MIN	(-128)
#define SCHAR_MAX	127
#define UCHAR_MAX	255
#define CHAR_MIN	0
#define CHAR_MAX	UCHAR_MAX
#define INT_MAX		2147483647
#define INT_MIN		(-INT_MAX-1)
#define UINT_MAX	0xffffffff
#define SHRT_MAX	32767
#define SHRT_MIN	(-SHRT_MAX-1)
#define USHRT_MAX	0xffff
#define LONG_MAX	2147483647L
#define LONG_MIN	(-LONG_MAX-1)
#define ULONG_MAX	0xffffffffUL
#define LLONG_MAX 9223372036854775807LL
#define LLONG_MIN (-LLONG_MAX - 1)
#define ULLONG_MAX (2ULL * LLONG_MAX + 1)
#define FLT_MAX
#define FLT_RADIX 2
#define FLT_MANT_DIG 24
#define FLT_DIG 6
#define FLT_ROUNDS 1
#define FLT_EPSILON 1.1920929e-07F
#define FLT_MIN_EXP (-125)
#define FLT_MIN 1.17549435e-38F
#define FLT_MIN_10_EXP (-37)
#define FLT_MAX_EXP 128
#define FLT_MAX 3.40282347e+38F
#define FLT_MAX_10_EXP 38
#define DBL_MANT_DIG 53
#define DBL_DIG 15
#define DBL_EPSILON 2.2204460492503131e-16
#define DBL_MIN_EXP (-1021)
#define DBL_MIN 2.2250738585072014e-308
#define DBL_MIN_10_EXP (-307)
#define DBL_MAX_EXP 1024
#define DBL_MAX 1.7976931348623157e+308
#define DBL_MAX_10_EXP 308
#endif

// NVCC compiler in device mode...
#if (defined(__CUDA_ARCH__) || defined(__CUDACC__)) && !defined(TARGET_CUDA)
#define TARGET_CUDA
#endif

#ifdef TARGET_CUDA
#ifndef __CUDACC_RTC__
#include "cuda_runtime.h"
#include <cstdio>
#include <float.h>
#ifdef USE_COOPERATIVE_THREADING
#include <cooperative_groups.h>
#endif
#endif
#else
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <float.h>
#ifdef _OPENMP
#include <omp.h>
#endif
#endif
#ifndef __CUDACC_RTC__
#include <new>
#endif

// Visual C++ intrinsics
#ifdef _MSC_VER
#include <intrin.h>
#endif

//+++++++ INTEGER TYPES +++++++++
#if defined(_MSC_VER) || defined(__CUDACC_RTC__)
// Visual C++: stdint.h file is missing on the default installation...
typedef signed char int8_t;
typedef unsigned char uint8_t;
typedef signed short int16_t;
typedef unsigned short uint16_t;
typedef signed int int32_t;
typedef unsigned int uint32_t;
typedef signed long long int64_t;
typedef unsigned long long uint64_t;
#else
#include <stdint.h>
#endif

// Define the __LP64__ macro (indicating that the code is being compiled in
// 64-bit mode).
#if defined(WIN64) || defined(_WIN64)
#define __LP64__
#endif

#if defined(__GNUC__) && (defined(__x86_64__) || defined(__ppc64__) || defined(__aarch64__)) && \
    !defined(__LP64__)
#define __LP64__
#endif

#define VALUE_TO_STRING(x) #x
#define VAR_NAME_VALUE(var) #var "=" VALUE_TO_STRING(var)

// Determine the version of the C++ standard
#if __cplusplus >= 201703L
#define CPLUSPLUS17
#define CPLUSPLUS14
#elif __cplusplus >= 201402L
#define CPLUSPLUS14
#elif __cplusplus >= 201103L
#define CPLUSPLUS11
#elif defined(_MSC_VER) && _MSC_VER >= 1900
// Visual C++ 2015 supports C++ 11 and C++ 14 (see https://msdn.microsoft.com/en-us/library/hh567368.aspx)
#define CPLUSPLUS14
#elif defined(_MSC_VER) && _MSC_VER >= 1800
// Visual C++ 2013 supports C++ 11
#define CPLUSPLUS11
#elif defined(__GNUC__) && __GNUC__ >= 5
// GCC 5.0
#define CPLUSPLUS11
#endif

#if defined(CPLUSPLUS14) && !defined(CPLUSPLUS11)
#define CPLUSPLUS11
#endif

#if defined(CUDAVER) && CUDAVER == 420
#define __HOST_V__
#ifndef __CUDA_ARCH__
#define __CUDA_ARCH__ 210
#endif
#else
#define __HOST_V__ __host__
#endif

#ifdef CPLUSPLUS17
#define CONSTEXPR constexpr
#define DEFAULT_TEMPLATE_ARG(x) = x
#else
#define CONSTEXPR
#define DEFAULT_TEMPLATE_ARG(x)
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define M_MAXLOG 7.09782712893383996732E2
#define M_MINLOG -7.451332191019412076235E2
#define M_MAXGAM 171.624376956302725
#define M_SQTPI 2.50662827463100050242E0
#define M_SQRTH 7.07106781186547524401E-1
#define M_LOGPI 1.14472988584940017414
#define M_LOG2 0.693147180559945309417

#ifndef M_EPS
#ifdef DBL_SCALAR
#define M_EPS 2.2204460492503131e-016 /*DBL_EPSILON*/
#else
#define M_EPS 1.192092896e-07F /*FLT_EPSILON*/
#endif
#endif

// NAN/INFINITY MACROS
#ifndef NAN
// CUDA does not support std::numeric_limits<scalar>
#ifdef TARGET_CUDA
#ifdef DBL_SCALAR
#define INFINITY __longlong_as_double(0x7ff0000000000000ULL)
#define NAN __longlong_as_double(0xfff8000000000000ULL)
#else
#define INFINITY __int_as_float(0x7f800000)
#define NAN __int_as_float(0x7fffffff)
#endif
#else
#include <limits>
#define INFINITY std::numeric_limits<scalar>::infinity()
#define NAN std::numeric_limits<scalar>::quiet_NaN()
#endif
#endif

#define NEG_INFINITY (-INFINITY)

#undef INLINE
#if defined(__CUDACC_RTC__)
#define INLINE __inline
#define EXPORT
#define IMPORT
#define SELECTANY
#elif defined(__clang__)
#define INLINE inline
#define EXPORT
#define IMPORT
#define SELECTANY
#elif defined(__GNUG__)
#define INLINE inline
#define EXPORT
#define IMPORT
#if defined(_WIN32) || defined(_WIN64)
#define SELECTANY __attribute__((selectany))
#else
#define SELECTANY __attribute__((weak))
#endif
#else
#define INLINE __forceinline
#define EXPORT __declspec(dllexport)
#define IMPORT __declspec(dllimport)
#define SELECTANY __declspec(selectany)
#endif

// Default template parameter value
#if !defined(_MSC_VER) || _MSC_VER >= 1800
#define DEFAULT_TEMPLATEVAL(x,val) x = val
#else
#define DEFAULT_TEMPLATEVAL(x,val) x
#endif

// FOR_ALL helper macro
#define FOR_ALL(X)                                                            \
    for (int i = 0; i < N; i++)                                               \
        (X);

#define ENTRY(x) _##x

#ifdef _MSC_VER
#pragma warning(disable : 4247 4297 4190) // C-linkage related issues throwing
                                          // exceptions from extern "C"
                                          // functions
#pragma warning(disable : 4305 4244)      // Truncation from double to scalar
#endif

//++++++++ MAIN SCALAR DEF ++++++++++
#ifdef DBL_SCALAR
typedef double scalar;
#else
typedef float scalar;
#endif

//+++++++ IFDEF TARGET_CUDA +++++++++
#ifdef TARGET_CUDA

#ifdef DBL_SCALAR
typedef double1 scalar1;
typedef double2 scalar2;
typedef double3 scalar3;
typedef double4 scalar4;
#define make_scalar1(x) make_double1(x)
#define make_scalar2(x, y) make_double2(x, y)
#define make_scalar3(x, y, z) make_double3(x, y, z)
#define make_scalar4(x, y, z, w) make_double4(x, y, z, w)
#else
typedef float1 scalar1;
typedef float2 scalar2;
typedef float3 scalar3;
typedef float4 scalar4;
#define make_scalar1(x) make_float1(x)
#define make_scalar2(x, y) make_float2(x, y)
#define make_scalar3(x, y, z) make_float3(x, y, z)
#define make_scalar4(x, y, z, w) make_float4(x, y, z, w)
#endif

#else

/* NON CUDA TARGETS - do some elementary definitions */

// G++ compiler known _Pragma, not __pragma
#ifdef __GNUG__
#ifndef __pragma
#define __pragma(x) _Pragma(#x)
#endif
#define __align__(x) __attribute__((packed, aligned(x)))
#else
#define __thread __declspec(thread)
#define __align__(x) __declspec(align(x))
#endif

// Vector types cannot be aligned in 32-bit mode, due to the MSVC error
// "Formal parameter with requested alignment won't be aligned"
#ifdef __LP64__
#define __valign__(x)  __align__(x)
#else
#define __valign__(x)
#endif

#define __device__
#define __host__
#define __global__
#define __constant__
#define __shared__
#define __syncthreads() __pragma(omp barrier)
#endif


// Built-in functions
// __threadidx: returns the current thread index
__device__ __HOST_V__ INLINE int __threadidx(void)
{
#if defined(_OPENMP)
    return omp_get_thread_num();
#elif defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (threadIdx.x * blockDim.y + threadIdx.y) * blockDim.z + threadIdx.z;
#else
    return 0;
#endif
}

// __blockidx: returns the current block index
__device__ __HOST_V__ INLINE int __blockidx(void)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (blockIdx.x * gridDim.x + blockIdx.y) * gridDim.z + blockIdx.x;
#else
    return 0;
#endif
}

// __threadcnt: returns the current thread count
__device__ __HOST_V__ INLINE int __threadcnt(void)
{
#if defined(_OPENMP)
    return omp_get_num_threads();
#elif defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return blockDim.x * blockDim.y * blockDim.z;
#else
    return 1;
#endif
}

// Calculating the next power of two of the specified number
INLINE __HOST_V__ __device__ unsigned int NextPow2(unsigned int v)
{
	// See https://en.wikipedia.org/wiki/Find_first_set
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	return 1 << (32 - __clz(v - 1));
#elif defined(__GNUG__)
	return 1 << (32 - __builtin_clz(v - 1));
#elif defined(_MSC_VER)
	unsigned int index;
	return 1 << (_BitScanReverse((unsigned long *)&index, v - 1) ? index + 1 : 0);
#else
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    return ++v;
#endif
}

namespace quasar
{
    //=============================================================================
    // SUPPORT FOR VECTOR / MATRIX / CUBE DATA TYPES
    // NOTE: defining the copy constructors results in a segment violation under
    // GCC
    // when parameters are passed in C-style
    //=============================================================================
    template <typename T> struct VectorBase
    {
        T *data;
        int dim1;

        INLINE __device__ __HOST_V__ VectorBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ VectorBase(T *data, int dim1) : data(data), dim1(dim1)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1; }
    };

    template <typename T> struct MatrixBase
    {
        T *data;
        int dim1;
        int dim2;

        INLINE __device__ __HOST_V__ MatrixBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ MatrixBase(T *data, int dim1, int dim2)
            : data(data), dim1(dim1), dim2(dim2)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2; }
    };

    template <typename T> struct CubeBase
    {
        T *data;
        int dim1;
        int dim2;
        int dim3;

        INLINE __device__ __HOST_V__ CubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ CubeBase(T *data, int dim1, int dim2, int dim3)
            : data(data), dim1(dim1), dim2(dim2), dim3(dim3)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2 * dim3; }
    };

// Specialized classes (only useful when dynamic memory support is enabled)
#if defined(ENABLE_DYNAMIC_KERNEL_MEM)

    template <typename T> struct VectorBase<VectorBase<T> >
    {
        VectorBase<T> *data;
        int dim1;

        INLINE __device__ __HOST_V__ VectorBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ VectorBase(VectorBase<T> *data, int dim1)
            : data(data), dim1(dim1)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1; }
    };

    template <typename T> struct VectorBase<MatrixBase<T> >
    {
        MatrixBase<T> *data;
        int dim1;

        INLINE __device__ __HOST_V__ VectorBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ VectorBase(MatrixBase<T> *data, int dim1)
            : data(data), dim1(dim1)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1; }
    };

    template <typename T> struct VectorBase<CubeBase<T> >
    {
        CubeBase<T> *data;
        int dim1;

        INLINE __device__ __HOST_V__ VectorBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ VectorBase(CubeBase<T> *data, int dim1)
            : data(data), dim1(dim1)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1; }
    };

    template <typename T> struct MatrixBase<VectorBase<T> >
    {
        VectorBase<T> *data;
        int dim1;
        int dim2;

        INLINE __device__ __HOST_V__ MatrixBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ MatrixBase(VectorBase<T> *data, int dim1, int dim2)
            : data(data), dim1(dim1), dim2(dim2)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2; }
    };

    template <typename T> struct MatrixBase<MatrixBase<T> >
    {
        MatrixBase<T> *data;
        int dim1;
        int dim2;

        INLINE __device__ __HOST_V__ MatrixBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ MatrixBase(MatrixBase<T> *data, int dim1, int dim2)
            : data(data), dim1(dim1), dim2(dim2)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2; }
    };

    template <typename T> struct MatrixBase<CubeBase<T> >
    {
        CubeBase<T> *data;
        int dim1;
        int dim2;

        INLINE __device__ __HOST_V__ MatrixBase() : data(NULL) {}
        INLINE __device__ __HOST_V__ MatrixBase(CubeBase<T> *data, int dim1, int dim2)
            : data(data), dim1(dim1), dim2(dim2)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2; }
    };

    template <typename T> struct CubeBase<VectorBase<T> >
    {
        VectorBase<T> *data;
        int dim1;
        int dim2;
        int dim3;

        INLINE __device__ __HOST_V__ CubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__
        CubeBase(VectorBase<T> *data, int dim1, int dim2, int dim3)
            : data(data), dim1(dim1), dim2(dim2), dim3(dim3)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2 * dim3; }
    };

    template <typename T> struct CubeBase<MatrixBase<T> >
    {
        MatrixBase<T> *data;
        int dim1;
        int dim2;
        int dim3;

        INLINE __device__ __HOST_V__ CubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__
        CubeBase(MatrixBase<T> *data, int dim1, int dim2, int dim3)
            : data(data), dim1(dim1), dim2(dim2), dim3(dim3)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2 * dim3; }
    };

    template <typename T> struct CubeBase<CubeBase<T> >
    {
        CubeBase<T> *data;
        int dim1;
        int dim2;
        int dim3;

        INLINE __device__ __HOST_V__ CubeBase() : data(NULL) {}
        INLINE __device__ __HOST_V__
        CubeBase(CubeBase<T> *data, int dim1, int dim2, int dim3)
            : data(data), dim1(dim1), dim2(dim2), dim3(dim3)
        {
        }
        INLINE __device__ __HOST_V__ void addref();
        INLINE __device__ __HOST_V__ void release();
        INLINE __device__ __HOST_V__ int get_numel() const { return dim1 * dim2 * dim3; }
    };

#endif

	// Auxiliary class for fetching default values for certain types
	template <typename T> struct default_val
	{
		static T val;
	};

	template <typename T> T default_val<T>::val;

	// Auxiliary class for left-references, as an alternative for the
	// referencer operator &, but allowing rebinding. This class is
	// used by the C++/CUDA code generator.
	template <typename T>
	class lref
	{
		T* ptr;

	public:
#if !defined(TARGET_CUDA)
		__device__ __HOST_V__ lref() :ptr(&default_val<T>::val) {}
#else
		__device__ __HOST_V__ lref() : ptr(NULL) {}
#endif
        template <typename R>
		__device__ __HOST_V__ lref(R &val) :ptr((T *) &val) {}
		__device__ __HOST_V__ T* operator->() { return ptr; }
		__device__ __HOST_V__ operator T() const { return *ptr; }
		__device__ __HOST_V__ operator T&() { return *ptr; }
		__device__ __HOST_V__ T operator=(const T val) {
#if !defined(TARGET_CUDA)
			if (ptr != &default_val<T>::val) // Do not overwrite the default value!
				*ptr = val;
#else
			if (ptr != NULL)
				*ptr = val;
#endif
			return val;
		}
	};

	// Function pointer template class
	template <typename C, typename F>
	struct function_ptr {
		F* ptr;	 // C-style function pointer
		C* data; // Closure variables
		INLINE __device__ __HOST_V__ function_ptr() {}
		template <typename G>
		INLINE __device__ __HOST_V__ function_ptr(G* ptr, C* data) :ptr(reinterpret_cast<F*>(ptr)), data(data) {}
	};

	template <typename R> class ReturnType {};

	// =========================================================================
	// autocast function
	// provides automatic casting operations. In case ENABLE_DYNAMIC_KERNEL_MEM,
	// is defined, some operations will allocate dynamic kernel memory.
	// =========================================================================
	template <typename R, typename T>
	INLINE R __device__ __HOST_V__ autocast(T value)
	{
		return autocast(value, ReturnType<R>());
	}
	template <typename R, typename T>
	INLINE R __device__ __HOST_V__ autocast(T x, T y) // Two argument version
	{
		return autocast(x, y, ReturnType<R>());
	}
	template <typename R, typename T> // default casting function
	INLINE R __device__ __HOST_V__ autocast(T value, ReturnType<R>)
	{
		return (R)value;
	}

	// Some useful operator functions
	template <typename T> __device__ __HOST_V__ T OP_add(T a, T b) { return a + b; }
	template <typename T> __device__ __HOST_V__ T OP_sub(T a, T b) { return a - b; }
	template <typename T> __device__ __HOST_V__ T OP_mul(T a, T b) { return a * b; }
	template <typename T> __device__ __HOST_V__ T OP_div(T a, T b) { return a / b; }
	template <typename T> __device__ __HOST_V__ T OP_min(T a, T b) { return min(a, b); }
	template <typename T> __device__ __HOST_V__ T OP_max(T a, T b) { return max(a, b); }
	template <typename T> __device__ __HOST_V__ T OP_minus(T a) { return -a; }

} // namespace quasar

//=============================================================================
// SUPPORT FOR NON-CUDA TARGETS
//=============================================================================
//+++++++ IFNDEF TARGET_CUDA +++++++++
#ifndef TARGET_CUDA

/* Built-in types */
struct int1
{
    int x;
};
struct __valign__(8) int2
{
    int x, y;
};
struct int3
{
    int x, y, z;
};
struct __valign__(16) int4
{
    int x, y, z, w;
};
struct scalar1
{
    scalar x;
};
#ifdef DBL_SCALAR
struct __valign__(16) scalar2
#else
struct __valign__(8) scalar2
#endif
{
    scalar x, y;
};
struct scalar3
{
    scalar x, y, z;
};
struct __valign__(16) scalar4
{
    scalar x, y, z, w;
};

/* Constructors for built-in types */
#define MAKE_T4(T) \
	INLINE T##1 make_##T##1(T x)	\
	{	\
		T##1 t = { x }; \
		return t; \
	} \
	INLINE T##2 make_##T##2(T x, T y) \
	{ \
		T##2 t = { x, y }; \
		return t; \
	} \
	INLINE T##3 make_##T##3(T x, T y, T z) \
	{ \
		T##3 t = { x, y, z }; \
		return t; \
	} \
	INLINE T##4 make_##T##4(T x, T y, T z, T w) \
	{ \
		T##4 t = { x, y, z, w }; \
		return t; \
	}
MAKE_T4(scalar)
MAKE_T4(int)

#define M_REC_LN2 1.4426950408889634073599246810019 /* 1/ln(2) */

/* Evaluates polynomial of degree N */
INLINE scalar __device__ __HOST_V__ polevl(scalar x, const scalar *coef, int N)
{
    scalar ans = coef[0];
    for (int i = 1; i <= N; i++)
        ans = ans * x + coef[i];
    return ans;
}

/* Evaluates polynomial of degree N with assumtion that coef[N] = 1.0 */
INLINE scalar __device__ __HOST_V__ p1evl(scalar x, const scalar *coef, int N)
{
    scalar ans = x + coef[0];
    for (int i = 1; i < N; i++)
        ans = ans * x + coef[i];
    return ans;
}

/* Built-in functions - note: we need to overload the standard definitions -
 * otherwise the wrong version could be used (for example abs(int) instead of
 * fabs(float))
 * By including the std namespace, most of these definitions should be included
 */
using namespace std;

// C++ 11: copysign function is already defined!
#ifndef CPLUSPLUS11
INLINE scalar copysign(scalar fTo, scalar fFrom)
{
    scalar t;
    *(unsigned int *)&t = ((*(unsigned int *)&fTo & 0x7FFFFFFF) |
                           (*(unsigned int *)&fFrom & 0x80000000));
    return t;
}

INLINE scalar round(scalar r)
{
    return (r > 0.0) ? floor(r + 0.5) : ceil(r - 0.5);
}

INLINE scalar exp2(scalar x) { return powf(2.0f, x); }
INLINE scalar log2(scalar x) { return log(x) * M_REC_LN2; }
INLINE scalar asinh(scalar xx)
{
    scalar x;
    int sign;
    if (xx == 0)
        return xx;
    if (xx < 0)
    {
        sign = -1;
        x = -xx;
    }
    else
    {
        sign = 1;
        x = xx;
    }
    return sign * log(x + sqrt(x * x + 1));
}
INLINE scalar acosh(scalar x)
{
    if (x < 1.0)
        return NAN;
    return log(x + sqrt(x * x - 1));
}
INLINE scalar atanh(scalar x)
{
    if (x < -1 || x > 1)
        return NAN;
    return 0.5 * log((1.0 + x) / (1.0 - x));
}
INLINE scalar erf(scalar x);
INLINE scalar erfc(scalar a)
{
    scalar x, y, z, p, q;
    scalar P[] = {2.46196981473530512524E-10, 5.64189564831068821977E-1,
                  7.46321056442269912687E0,   4.86371970985681366614E1,
                  1.96520832956077098242E2,   5.26445194995477358631E2,
                  9.34528527171957607540E2,   1.02755188689515710272E3,
                  5.57535335369399327526E2};
    scalar Q[] = {// 1.0
                  1.32281951154744992508E1, 8.67072140885989742329E1,
                  3.54937778887819891062E2, 9.75708501743205489753E2,
                  1.82390916687909736289E3, 2.24633760818710981792E3,
                  1.65666309194161350182E3, 5.57535340817727675546E2};
    scalar R[] = {5.64189583547755073984E-1, 1.27536670759978104416E0,
                  5.01905042251180477414E0,  6.16021097993053585195E0,
                  7.40974269950448939160E0,  2.97886665372100240670E0};
    scalar S[] = {// 1.00000000000000000000E0,
                  2.26052863220117276590E0, 9.39603524938001434673E0,
                  1.20489539808096656605E1, 1.70814450747565897222E1,
                  9.60896809063285878198E0, 3.36907645100081516050E0};
    if (a < 0.0)
        x = -a;
    else
        x = a;
    if (x < 1.0)
        return 1.0 - erf(a);
    z = -a * a;
    if (z < -M_MAXLOG)
    {
        if (a < 0)
            return (2.0);
        else
            return (0.0);
    }
    z = exp(z);
    if (x < 8.0)
    {
        p = polevl(x, P, 8);
        q = p1evl(x, Q, 8);
    }
    else
    {
        p = polevl(x, R, 5);
        q = p1evl(x, S, 6);
    }
    y = (z * p) / q;
    if (a < 0)
        y = 2.0 - y;
    if (y == 0.0)
    {
        if (a < 0)
            return 2.0;
        else
            return (0.0);
    }
    return y;
}
INLINE scalar erf(scalar x)
{
    scalar y, z;
    scalar T[5] = {9.60497373987051638749E0, 9.00260197203842689217E1,
                   2.23200534594684319226E3, 7.00332514112805075473E3,
                   5.55923013010394962768E4};
    scalar U[5] = {3.35617141647503099647E1, 5.21357949780152679795E2,
                   4.59432382970980127987E3, 2.26290000613890934246E4,
                   4.92673942608635921086E4};
    if (abs(x) > 1.0)
        return 1.0 - erfc(x);
    z = x * x;
    y = x * polevl(z, T, 4) / p1evl(z, U, 5);
    return y;
}
#endif
INLINE scalar max(scalar x, scalar y) { return x > y ? x : y; }
INLINE scalar min(scalar x, scalar y) { return x > y ? y : x; }
INLINE scalar saturate(scalar x) { return x < 0 ? 0 : x > 1 ? 1 : x; }
INLINE scalar rsqrt(scalar x) { return 1.0f / sqrtf(x); }
INLINE scalar rcp_rn(scalar x) { return 1.0f / x; }
INLINE scalar rcp_ru(scalar x) { return 1.0f / x; }
INLINE scalar rcp_rz(scalar x) { return 1.0f / x; }
INLINE scalar div_approx(scalar a, scalar b) { return a / b; }

#ifdef _MSC_VER
// Microsoft C library lacks log1p
INLINE double log1p(double x)
{
    if (fabs(x) > 1e-4)
        return log(1.0 + x);
    return (-0.5 * x + 1.0) *
           x; // Taylor approx of log(1.0 + x) - more accurate for small values
}
#endif
INLINE scalar erfinv(scalar _a)
{
    double a = (double)_a;
    double p, q, t, fa;
    volatile union
    {
        double d;
        unsigned long long int l;
    } cvt;

    fa = fabs(a);
    if (fa >= 1.0)
    {
        cvt.l = 0xfff8000000000000ull;
        t = cvt.d; /* INDEFINITE */
        if (fa == 1.0)
        {
            t = a * exp(1000.0); /* Infinity */
        }
    }
    else if (fa >= 0.9375)
    {
        /* Based on: J.M. Blair, C.A. Edwards, J.H. Johnson: Rational Chebyshev
           Approximations for the Inverse of the Error Function. Mathematics of
           Computation, Vol. 30, No. 136 (Oct. 1976), pp. 827-830. Table 59
         */
        t = log1p(-fa);
        t = 1.0 / sqrt(-t);
        p = 2.7834010353747001060e-3;
        p = p * t + 8.6030097526280260580e-1;
        p = p * t + 2.1371214997265515515e+0;
        p = p * t + 3.1598519601132090206e+0;
        p = p * t + 3.5780402569085996758e+0;
        p = p * t + 1.5335297523989890804e+0;
        p = p * t + 3.4839207139657522572e-1;
        p = p * t + 5.3644861147153648366e-2;
        p = p * t + 4.3836709877126095665e-3;
        p = p * t + 1.3858518113496718808e-4;
        p = p * t + 1.1738352509991666680e-6;
        q = t + 2.2859981272422905412e+0;
        q = q * t + 4.3859045256449554654e+0;
        q = q * t + 4.6632960348736635331e+0;
        q = q * t + 3.9846608184671757296e+0;
        q = q * t + 1.6068377709719017609e+0;
        q = q * t + 3.5609087305900265560e-1;
        q = q * t + 5.3963550303200816744e-2;
        q = q * t + 4.3873424022706935023e-3;
        q = q * t + 1.3858762165532246059e-4;
        q = q * t + 1.1738313872397777529e-6;
        t = p / (q * t);
        if (a < 0.0)
            t = -t;
    }
    else if (fa >= 0.75)
    {
        /* Based on: J.M. Blair, C.A. Edwards, J.H. Johnson: Rational Chebyshev
           Approximations for the Inverse of the Error Function. Mathematics of
           Computation, Vol. 30, No. 136 (Oct. 1976), pp. 827-830. Table 39
        */
        t = a * a - .87890625;
        p = .21489185007307062000e+0;
        p = p * t - .64200071507209448655e+1;
        p = p * t + .29631331505876308123e+2;
        p = p * t - .47644367129787181803e+2;
        p = p * t + .34810057749357500873e+2;
        p = p * t - .12954198980646771502e+2;
        p = p * t + .25349389220714893917e+1;
        p = p * t - .24758242362823355486e+0;
        p = p * t + .94897362808681080020e-2;
        q = t - .12831383833953226499e+2;
        q = q * t + .41409991778428888716e+2;
        q = q * t - .53715373448862143349e+2;
        q = q * t + .33880176779595142685e+2;
        q = q * t - .11315360624238054876e+2;
        q = q * t + .20369295047216351160e+1;
        q = q * t - .18611650627372178511e+0;
        q = q * t + .67544512778850945940e-2;
        p = p / q;
        t = a * p;
    }
    else
    {
        /* Based on: J.M. Blair, C.A. Edwards, J.H. Johnson: Rational Chebyshev
           Approximations for the Inverse of the Error Function. Mathematics of
           Computation, Vol. 30, No. 136 (Oct. 1976), pp. 827-830. Table 18
        */
        t = a * a - .5625;
        p = -.23886240104308755900e+2;
        p = p * t + .45560204272689128170e+3;
        p = p * t - .22977467176607144887e+4;
        p = p * t + .46631433533434331287e+4;
        p = p * t - .43799652308386926161e+4;
        p = p * t + .19007153590528134753e+4;
        p = p * t - .30786872642313695280e+3;
        q = t - .83288327901936570000e+2;
        q = q * t + .92741319160935318800e+3;
        q = q * t - .35088976383877264098e+4;
        q = q * t + .59039348134843665626e+4;
        q = q * t - .48481635430048872102e+4;
        q = q * t + .18997769186453057810e+4;
        q = q * t - .28386514725366621129e+3;
        p = p / q;
        t = a * p;
    }
    return (scalar)t;
}
INLINE scalar erfcinv(scalar _a)
{
    double a = (double)_a;
    double t;
    volatile union
    {
        double d;
        unsigned long long int l;
    } cvt;

#if defined(_MSC_VER)
    if (_isnan(a))
#else
    if (isnan(a))
#endif
    {
        return a + a;
    }
    if (a <= 0.0)
    {
        cvt.l = 0xfff8000000000000ull;
        t = cvt.d; /* INDEFINITE */
        if (a == 0.0)
        {
            t = (1.0 - a) * exp(1000.0); /* Infinity */
        }
    }
    else if (a >= 0.0625)
    {
        t = erfinv(1.0 - a);
    }
    else if (a >= 1e-100)
    {
        /* Based on: J.M. Blair, C.A. Edwards, J.H. Johnson: Rational Chebyshev
           Approximations for the Inverse of the Error Function. Mathematics of
           Computation, Vol. 30, No. 136 (Oct. 1976), pp. 827-830. Table 59
        */
        double p, q;
        t = log(a);
        t = 1.0 / sqrt(-t);
        p = 2.7834010353747001060e-3;
        p = p * t + 8.6030097526280260580e-1;
        p = p * t + 2.1371214997265515515e+0;
        p = p * t + 3.1598519601132090206e+0;
        p = p * t + 3.5780402569085996758e+0;
        p = p * t + 1.5335297523989890804e+0;
        p = p * t + 3.4839207139657522572e-1;
        p = p * t + 5.3644861147153648366e-2;
        p = p * t + 4.3836709877126095665e-3;
        p = p * t + 1.3858518113496718808e-4;
        p = p * t + 1.1738352509991666680e-6;
        q = t + 2.2859981272422905412e+0;
        q = q * t + 4.3859045256449554654e+0;
        q = q * t + 4.6632960348736635331e+0;
        q = q * t + 3.9846608184671757296e+0;
        q = q * t + 1.6068377709719017609e+0;
        q = q * t + 3.5609087305900265560e-1;
        q = q * t + 5.3963550303200816744e-2;
        q = q * t + 4.3873424022706935023e-3;
        q = q * t + 1.3858762165532246059e-4;
        q = q * t + 1.1738313872397777529e-6;
        t = p / (q * t);
    }
    else
    {
        /* Based on: J.M. Blair, C.A. Edwards, J.H. Johnson: Rational Chebyshev
           Approximations for the Inverse of the Error Function. Mathematics of
           Computation, Vol. 30, No. 136 (Oct. 1976), pp. 827-830. Table 82
        */
        double p, q;
        t = log(a);
        t = 1.0 / sqrt(-t);
        p = 6.9952990607058154858e-1;
        p = p * t + 1.9507620287580568829e+0;
        p = p * t + 8.2810030904462690216e-1;
        p = p * t + 1.1279046353630280005e-1;
        p = p * t + 6.0537914739162189689e-3;
        p = p * t + 1.3714329569665128933e-4;
        p = p * t + 1.2964481560643197452e-6;
        p = p * t + 4.6156006321345332510e-9;
        p = p * t + 4.5344689563209398450e-12;
        q = t + 1.5771922386662040546e+0;
        q = q * t + 2.1238242087454993542e+0;
        q = q * t + 8.4001814918178042919e-1;
        q = q * t + 1.1311889334355782065e-1;
        q = q * t + 6.0574830550097140404e-3;
        q = q * t + 1.3715891988350205065e-4;
        q = q * t + 1.2964671850944981713e-6;
        q = q * t + 4.6156017600933592558e-9;
        q = q * t + 4.5344687377088206783e-12;
        t = p / (q * t);
    }
    return (scalar)t;
}
#ifndef CPLUSPLUS11
INLINE scalar lgamma(scalar x)
{ // Natural logarithm of the gamma function
    scalar p, q, w, z;
    scalar A[] = {8.11614167470508450300E-4, -5.95061904284301438324E-4,
                  7.93650340457716943945E-4, -2.77777777730099687205E-3,
                  8.33333333333331927722E-2};
    scalar B[] = {-1.37825152569120859100E3, -3.88016315134637840924E4,
                  -3.31612992738871184744E5, -1.16237097492762307383E6,
                  -1.72173700820839662146E6, -8.53555664245765465627E5};
    scalar C[] = {/* 1.00000000000000000000E0, */
                  -3.51815701436523470549E2, -1.70642106651881159223E4,
                  -2.20528590553854454839E5, -1.13933444367982507207E6,
                  -2.53252307177582951285E6, -2.01889141433532773231E6};

    if (x < -34.0)
    {
        q = -x;
        w = lgamma(q);
        p = floor(q);
        if (p == q)
            return NAN;
        z = q - p;
        if (z > 0.5)
        {
            p += 1.0;
            z = p - q;
        }
        z = q * sin(M_PI * z);
        if (z == 0.0)
            return NAN;
        z = M_LOGPI - log(z) - w;
        return z;
    }

    if (x < 13.0)
    {
        z = 1.0;
        while (x >= 3.0)
        {
            x -= 1.0;
            z *= x;
        }
        while (x < 2.0)
        {
            if (x == 0.0)
                return NAN;
            z /= x;
            x += 1.0;
        }
        if (z < 0.0)
            z = -z;
        if (x == 2.0)
            return log(z);
        x -= 2.0;
        p = x * polevl(x, B, 5) / p1evl(x, C, 6);
        return (log(z) + p);
    }

    if (x > 2.556348e305)
        return NAN;
    q = (x - 0.5) * log(x) - x + 0.91893853320467274178;
    if (x > 1.0e8)
        return (q);

    p = 1.0 / (x * x);
    if (x >= 1000.0)
        q += ((7.9365079365079365079365e-4 * p - 2.7777777777777777777778e-3) *
                  p +
              0.0833333333333333333333) /
             x;
    else
        q += polevl(p, A, 4) / x;
    return q;
}

INLINE scalar tgamma_helper(scalar a)
{
    scalar t;
    t = -1.05767296987211380E-003;
    t = t * a + 7.09279059435508670E-003;
    t = t * a - 9.65347121958557050E-003;
    t = t * a - 4.21736613253687960E-002;
    t = t * a + 1.66542401247154280E-001;
    t = t * a - 4.20043267827838460E-002;
    t = t * a - 6.55878234051332940E-001;
    t = t * a + 5.77215696929794240E-001;
    t = t * a + 1;
    return t;
}

INLINE scalar tgamma(scalar a)
{
    float s, xx, x = a;
    if (x >= 0.0f)
    {
        if (x > 36.0f)
            x = 36.0f; /* clamp */
        s = 1.0f;
        xx = x;
        if (x > 34.03f)
        { /* prevent premature overflow */
            xx -= 1.0f;
        }
        while (xx > 1.5f)
        {
            xx = xx - 1.0f;
            s = s * xx;
        }
        if (x >= 0.5f)
        {
            xx = xx - 1.0f;
        }
        xx = tgamma_helper(xx);
        if (x < 0.5f)
        {
            xx = xx * x;
        }
        s = s / xx;
        if (x > 34.03f)
        {
            /* Cannot use s = s * x - s due to intermediate overflow! */
            xx = x - 1.0f;
            s = s * xx;
        }
        return s;
    }
    else
    {
        if (x == floor(x))
        { /* x is negative integer */
            x = NAN;
        }
        if (x < -41.1f)
            x = -41.1f; /* clamp */
        xx = x;
        if (x < -34.03f)
        { /* prevent overflow in intermediate result */
            xx += 6.0f;
        }
        s = xx;
        while (xx < -0.5f)
        {
            xx = xx + 1.0f;
            s = s * xx;
        }
        xx = tgamma_helper(xx);
        s = s * xx;
        s = 1.0f / s;
        if (x < -34.03f)
        {
            xx = x;
            xx *= (x + 1.0f);
            xx *= (x + 2.0f);
            xx *= (x + 3.0f);
            xx *= (x + 4.0f);
            xx *= (x + 5.0f);
            xx = 1.0f / xx;
            s = s * xx;
            if ((a < -42.0f) && !(((int)a) & 1))
            {
                s = -0.0f;
            }
        }
        return s;
    }
}

#endif

#if defined(_MSC_VER)

INLINE int _isfinite_(scalar x) { return _finite(x); }
INLINE int _isinf_(scalar x) { return !(_finite(x) || _isnan(x)); }
INLINE int _isnan_(scalar x) { return _isnan(x); }

// GCC seems to define isfinite, isnan and isinf through macros, which causes
// some problems when applied to cscalar values/vectors, therefore we renamed
// these functions to _isfinite_, _isnan_, _isinf_
#elif defined(__GNUG__)

INLINE int _isfinite_(scalar x) { return isfinite(x); }
INLINE int _isinf_(scalar x) { return !(isfinite(x) || isnan(x)); }
INLINE int _isnan_(scalar x) { return isnan(x); }

#endif

// All different overloads need to be defined, otherwise
// the compiler complains...
INLINE int _isfinite_(int8_t x) { return 0; }
INLINE int _isinf_(int8_t x) { return 0; }
INLINE int _isnan_(int8_t x) { return 0; }
INLINE int _isfinite_(uint8_t x) { return 0; }
INLINE int _isinf_(uint8_t x) { return 0; }
INLINE int _isnan_(uint8_t x) { return 0; }
INLINE int _isfinite_(int16_t x) { return 0; }
INLINE int _isinf_(int16_t x) { return 0; }
INLINE int _isnan_(int16_t x) { return 0; }
INLINE int _isfinite_(uint16_t x) { return 0; }
INLINE int _isinf_(uint16_t x) { return 0; }
INLINE int _isnan_(uint16_t x) { return 0; }
INLINE int _isfinite_(int32_t x) { return 0; }
INLINE int _isinf_(int32_t x) { return 0; }
INLINE int _isnan_(int32_t x) { return 0; }
INLINE int _isfinite_(uint32_t x) { return 0; }
INLINE int _isinf_(uint32_t x) { return 0; }
INLINE int _isnan_(uint32_t x) { return 0; }
INLINE int _isfinite_(int64_t x) { return 0; }
INLINE int _isinf_(int64_t x) { return 0; }
INLINE int _isnan_(int64_t x) { return 0; }
INLINE int _isfinite_(uint64_t x) { return 0; }
INLINE int _isinf_(uint64_t x) { return 0; }
INLINE int _isnan_(uint64_t x) { return 0; }


#define __cosf cosf
#define __sinf sinf

#undef assert
INLINE void assert(bool x)
{
    if (!(x))
    {
        throw "Kernel error";
    }
}

#else // if defined(TARGET_CUDA)

// Define the host function saturate
#ifndef __CUDA_ARCH__
INLINE scalar __device__ __HOST_V__ host_saturate(scalar x) { return x < 0 ? 0 : x > 1 ? 1 : x; }
#define saturate host_saturate
#endif

#if CUDART_VERSION >= 12000
INLINE scalar __device__ saturate(scalar x) { return x < 0 ? 0 : x > 1 ? 1 : x; }
#endif

// All different overloads need to be defined, otherwise
// the compiler complains...
INLINE int __device__ __HOST_V__  _isfinite_(scalar x) { return isfinite(x); }
INLINE int __device__ __HOST_V__  _isnan_(scalar x) { return isnan(x); }
INLINE int __device__ __HOST_V__  _isinf_(scalar x) { return isinf(x); }
INLINE int __device__ __HOST_V__  _isfinite_(int8_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(int8_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(int8_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isfinite_(uint8_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(uint8_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(uint8_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isfinite_(int16_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(int16_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(int16_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isfinite_(uint16_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(uint16_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(uint16_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isfinite_(int32_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(int32_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(int32_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isfinite_(uint32_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(uint32_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(uint32_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isfinite_(int64_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(int64_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(int64_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isfinite_(uint64_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isinf_(uint64_t x) { return 0; }
INLINE int __device__ __HOST_V__  _isnan_(uint64_t x) { return 0; }

INLINE scalar __device__ rcp_rn(scalar x) { return __frcp_rn(x); }
INLINE scalar __device__ rcp_ru(scalar x) { return __frcp_ru(x); }
INLINE scalar __device__ rcp_rz(scalar x) { return __frcp_rz(x); }
INLINE scalar __device__ div_approx(scalar a, scalar b) { return __fdividef(a, b); }


// Implement assert through the trap instruction
// See
// https://devtalk.nvidia.com/default/topic/491119/cuda-programming-and-performance/using-assert-in-cuda-code/
#undef assert

// Uncommented, because CUDA cannot recover from a trap:(
//#define assert(x)	if (!(x)) { __threadfence(); asm("trap;"); }
#define assert(x)

//+++++++ ENDIF TARGET_CUDA +++++++++
#endif

// Custom implementation of a number of functions that seem to be missing
INLINE __device__ __HOST_V__ uint8_t abs(uint8_t a) { return a; }
INLINE __device__ __HOST_V__ uint16_t abs(uint16_t a) { return a; }
INLINE __device__ __HOST_V__ uint32_t abs(uint32_t a) { return a; }
INLINE __device__ __HOST_V__ uint64_t abs(uint64_t a) { return a; }

INLINE __device__ __HOST_V__ uint32_t mod(uint32_t a, uint32_t b) { return a % b; }
INLINE __device__ __HOST_V__ uint64_t mod(uint64_t a, uint64_t b) { return a % b; }

INLINE __device__ __HOST_V__ int pow(int base, int a) // Exponentiation by squaring
{
    int res = 1;
    while (a)
    {
        if (a & 1)
            res *= base;
        a >>= 1;
        base *= base;
    }
    return res;
}
INLINE __device__ __HOST_V__ unsigned int pow(unsigned int base,
                                   unsigned int a) // Exponentiation by squaring
{
    unsigned int res = 1;
    while (a)
    {
        if (a & 1)
            res *= base;
        a >>= 1;
        base *= base;
    }
    return res;
}
template <int N>
INLINE __device__ __HOST_V__ scalar powN(scalar base) // Integer power of a scalar number
{
    scalar res = 1;
    if (N == 0)
        return 1;
    else if (N == 1)
        return base;
    else
    {
        if (N & 1)
            res = base;
        res *= powN<(N >> 1)>(base * base);
    }
    return res;
}

INLINE __device__ __HOST_V__ scalar pow(int a, scalar b) { return powf(a, b); }

#ifdef _MSC_VER
INLINE __device__ __HOST_V__ int floor(int x)
{
    return x;
} // MSVC does not know about this overload
INLINE __device__ __HOST_V__ scalar log(int x)
{
    return log((scalar)x);
} // MSVC does not know about this overload
INLINE __device__ __HOST_V__ scalar exp(int x)
{
    return exp((scalar)x);
} // MSVC does not know about this overload
#endif

// log2 integer implementation (floor(log2(x)))
INLINE __device__ __HOST_V__ int log2(int x)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    int val = __clz(x);
    return CHAR_BIT * sizeof(int) - val - 1;
#elif defined(__GNUG__)
    int val = __builtin_clz(x);
    return CHAR_BIT * sizeof(int) - val - 1;
#elif defined(_MSC_VER)
    unsigned long leading_zero = 0;
    _BitScanReverse(&leading_zero, x);
    return leading_zero;
#else
#error "CountLeadingZeros not implemented for this platform!"
#endif
}

/* Built-in types */
/* At some point in the future, we may use C++03 typedef templates here... */
template <int N> struct intN
{
    int el[N];
	#if !defined(__CUDACC_VER_MAJOR__) || __CUDACC_VER_MAJOR__>=6
    __device__ __HOST_V__ operator quasar::VectorBase<int>()
    {
        return quasar::VectorBase<int>(el, N);
    }
	#endif
};
template <int N> struct scalarN
{
    scalar el[N];
	#if !defined(__CUDACC_VER_MAJOR__) || __CUDACC_VER_MAJOR__>=6
    __device__ __HOST_V__ operator quasar::VectorBase<scalar>()
    {
        return quasar::VectorBase<scalar>(el, N);
    }
	#endif
};

// Need implicit conversion from int4 to intN<4>. This is required for several
// higher dimensional functions. We obtain this by specializing the intN<4>
// class.
template <> struct intN<4>
{
    int el[4];
    __device__ __HOST_V__ intN() {}
	#if !defined(__CUDACC_VER_MAJOR__) || __CUDACC_VER_MAJOR__>=6
    __device__ __HOST_V__ operator quasar::VectorBase<int>()
    {
        return quasar::VectorBase<int>(el, 4);
    }
	#endif
	__device__ __HOST_V__ operator int4()
	{
		return make_int4(el[0], el[1], el[2], el[3]);
	}
    __device__ __HOST_V__ intN(int4 a)
    {
        el[0] = a.x;
        el[1] = a.y;
        el[2] = a.z;
        el[3] = a.w;
    }
};

#if defined(__LP64__) && (!defined(_MSC_VER) || _MSC_VER >= 1910)
#define AUTO_SIMD_ALIGNMENT
#endif

#if defined(DBL_SCALAR) && defined(AUTO_SIMD_ALIGNMENT)
template <> struct __align__(32) scalarN<4>
#elif defined(__LP64__) && defined(AUTO_SIMD_ALIGNMENT)
template <> struct __align__(16) scalarN<4>
#else
template <> struct scalarN<4>
#endif
{
	scalar el[4];
	__device__ __HOST_V__ scalarN() {}
#if !defined(__CUDACC_VER_MAJOR__) || __CUDACC_VER_MAJOR__>=6
	__device__ __HOST_V__ operator quasar::VectorBase<scalar>()
	{
		return quasar::VectorBase<scalar>(el, 4);
	}
#endif
	__device__ __HOST_V__ operator scalar4()
	{
		return make_scalar4(el[0], el[1], el[2], el[3]);
	}
	__device__ __HOST_V__ scalarN(scalar4 a)
	{
		el[0] = a.x;
		el[1] = a.y;
		el[2] = a.z;
		el[3] = a.w;
	}
};

// Aligned data types (for SIMD processing)
#if !defined(__CUDACC_VER_MAJOR__) || __CUDACC_VER_MAJOR__>=6
#ifdef __LP64__
#define SIMD_ALIGNED_VECTORTYPE(N, alignment)	\
	template <> struct __align__(alignment) intN<N> \
	{ \
		int el[N]; \
		__device__ __HOST_V__ operator quasar::VectorBase<int>() { return quasar::VectorBase<int>(el, N); } \
	}; \
	template <> struct __align__(alignment) scalarN<N> \
	{ \
		scalar el[N]; \
		__device__ __HOST_V__ operator quasar::VectorBase<scalar>() { return quasar::VectorBase<scalar>(el, N); } \
	};
#else
#define SIMD_ALIGNED_VECTORTYPE(N, alignment)	\
	template <> struct intN<N> \
	{ \
		int el[N]; \
		__device__ __HOST_V__ operator quasar::VectorBase<int>() { return quasar::VectorBase<int>(el, N); } \
	}; \
	template <> struct scalarN<N> \
	{ \
		scalar el[N]; \
		__device__ __HOST_V__ operator quasar::VectorBase<scalar>() { return quasar::VectorBase<scalar>(el, N); } \
	};
#endif
#elif !defined(DBL_SCALAR)
#define SIMD_ALIGNED_VECTORTYPE(N, alignment)	\
	template <> __align__(alignment) struct intN<N> { int el[N]; }; \
	template <> __align__(alignment) struct scalarN<N> { scalar el[N]; };
#else
#define SIMD_ALIGNED_VECTORTYPE(N, alignment)	\
	template <> __align__(alignment) struct intN<N> { int el[N]; }; 
#endif

#if defined(AUTO_SIMD_ALIGNMENT)
SIMD_ALIGNED_VECTORTYPE(8, 32)
#endif

typedef int64_t			longlong;
typedef uint64_t		ulonglong;
typedef uint16_t		ushort;
typedef uint8_t			uchar;
typedef uint32_t		uint;

#ifndef TARGET_CUDA
struct					longlong1 { longlong x; };
struct __valign__(16)	longlong2 { longlong x, y; };
struct					longlong3 { longlong x, y, z; };
struct __valign__(16)	longlong4 { longlong x, y, z, w; };

struct					char1 { char x; };
struct __valign__(2)	char2 { char x, y; };
struct					char3 { char x, y, z; };
struct __valign__(4)	char4 { char x, y, z, w; };

struct					short1 { short x; };
struct __valign__(4)	short2 { short x, y; };
struct					short3 { short x, y, z; };
struct __valign__(8)	short4 { short x, y, z, w; };

struct					ulonglong1 { ulonglong x; };
struct __valign__(16)	ulonglong2 { ulonglong x, y; };
struct					ulonglong3 { ulonglong x, y, z; };
struct __valign__(16)	ulonglong4 { ulonglong x, y, z, w; };

struct					uchar1 { uchar x; };
struct __valign__(2)	uchar2 { uchar x, y; };
struct					uchar3 { uchar x, y, z; };
struct __valign__(4)	uchar4 { uchar x, y, z, w; };

struct					ushort1 { ushort x; };
struct __valign__(4)	ushort2 { ushort x, y; };
struct					ushort3 { ushort x, y, z; };
struct __valign__(8)	ushort4 { ushort x, y, z, w; };

struct					uint1 { uint x; };
struct __valign__(8)	uint2 { uint x, y; };
struct					uint3 { uint x, y, z; };
struct __valign__(16)	uint4 { uint x, y, z, w; };

struct					float1 { float x; };
struct __valign__(8)	float2 { float x, y; };
struct					float3 { float x, y, z; };
struct __valign__(16)	float4 { float x, y, z, w; };

struct					double1 { double x; };
struct __valign__(16)	double2 { double x, y; };
struct					double3 { double x, y, z; };
struct __valign__(16)	double4 { double x, y, z, w; };

MAKE_T4(short)
MAKE_T4(ushort)
MAKE_T4(uint)
MAKE_T4(longlong)
MAKE_T4(ulonglong)
MAKE_T4(char)
MAKE_T4(uchar)
MAKE_T4(float)
MAKE_T4(double)

#endif	// !defined(TARGET_CUDA)

template <int N> struct charN		{ char el[N]; };
template <int N> struct ucharN		{ unsigned char el[N]; };
template <int N> struct shortN		{ short el[N]; };
template <int N> struct ushortN		{ unsigned short el[N]; };
template <int N> struct uintN		{ unsigned int el[N]; };
template <int N> struct longlongN	{ long long el[N]; };
template <int N> struct ulonglongN  { unsigned long long el[N]; };
template <int N> struct floatN		{ float el[N]; };
template <int N> struct doubleN		{ double el[N]; };


/*
template <> struct __valign__(8)  charN<8>   { char el[8]; };
template <> struct __valign__(8)  ucharN<8>  { char el[8]; };
template <> struct __valign__(16) shortN<8>  { char el[8]; };
template <> struct __valign__(16) ushortN<8> { char el[8]; };*/

namespace quasar
{
	// Helper class
	template <typename T, int N> struct VecBaseType		{ typedef void type; };
	template <int N> struct VecBaseType<scalar, N>		{ typedef scalarN<N> type; };
	template <int N> struct VecBaseType<int, N>			{ typedef intN<N> type; };
	template <int N> struct VecBaseType<char, N>		{ typedef charN<N> type; };
	template <int N> struct VecBaseType<short, N>		{ typedef shortN<N> type; };
	template <int N> struct VecBaseType<longlong, N>	{ typedef longlongN<N> type; };
	template <int N> struct VecBaseType<uint, N>		{ typedef uintN<N> type; };
	template <int N> struct VecBaseType<uchar, N>		{ typedef ucharN<N> type; };
	template <int N> struct VecBaseType<ushort, N>		{ typedef ushortN<N> type; };
	template <int N> struct VecBaseType<ulonglong, N>	{ typedef ulonglongN<N> type; };
	#ifdef DBL_SCALAR
	template <int N> struct VecBaseType<float, N>		{ typedef floatN<N> type; };
	#else
	template <int N> struct VecBaseType<double, N>		{ typedef doubleN<N> type; };
	#endif

	// VecType maps the element type and vector length onto the resulting CUDA vector type.
	template <typename T, int N> struct VecType { typedef typename VecBaseType<T, N>::type type; };

	template <> struct VecType<scalar, 1> { typedef scalar1 type; };
	template <> struct VecType<scalar, 2> { typedef scalar2 type; };
	template <> struct VecType<scalar, 3> { typedef scalar3 type; };
	template <> struct VecType<scalar, 4> { typedef scalar4 type; };

	template <> struct VecType<int64_t, 1> { typedef longlong1 type; };
	template <> struct VecType<int64_t, 2> { typedef longlong2 type; };
	template <> struct VecType<int64_t, 3> { typedef longlong3 type; };
	template <> struct VecType<int64_t, 4> { typedef longlong4 type; };

	template <> struct VecType<int, 1> { typedef int1 type; };
	template <> struct VecType<int, 2> { typedef int2 type; };
	template <> struct VecType<int, 3> { typedef int3 type; };
	template <> struct VecType<int, 4> { typedef int4 type; };

	template <> struct VecType<short, 1> { typedef short1 type; };
	template <> struct VecType<short, 2> { typedef short2 type; };
	template <> struct VecType<short, 3> { typedef short3 type; };
	template <> struct VecType<short, 4> { typedef short4 type; };

	template <> struct VecType<char, 1> { typedef char1 type; };
	template <> struct VecType<char, 2> { typedef char2 type; };
	template <> struct VecType<char, 3> { typedef char3 type; };
	template <> struct VecType<char, 4> { typedef char4 type; };

	template <> struct VecType<uint64_t, 1> { typedef ulonglong1 type; };
	template <> struct VecType<uint64_t, 2> { typedef ulonglong2 type; };
	template <> struct VecType<uint64_t, 3> { typedef ulonglong3 type; };
	template <> struct VecType<uint64_t, 4> { typedef ulonglong4 type; };

	template <> struct VecType<uint32_t, 1> { typedef uint1 type; };
	template <> struct VecType<uint32_t, 2> { typedef uint2 type; };
	template <> struct VecType<uint32_t, 3> { typedef uint3 type; };
	template <> struct VecType<uint32_t, 4> { typedef uint4 type; };

	template <> struct VecType<uint16_t, 1> { typedef ushort1 type; };
	template <> struct VecType<uint16_t, 2> { typedef ushort2 type; };
	template <> struct VecType<uint16_t, 3> { typedef ushort3 type; };
	template <> struct VecType<uint16_t, 4> { typedef ushort4 type; };

	template <> struct VecType<uint8_t, 1> { typedef uchar1 type; };
	template <> struct VecType<uint8_t, 2> { typedef uchar2 type; };
	template <> struct VecType<uint8_t, 3> { typedef uchar3 type; };
	template <> struct VecType<uint8_t, 4> { typedef uchar4 type; };

	#ifndef DBL_SCALAR
	template <> struct VecType<double, 1> { typedef double1 type; };
	template <> struct VecType<double, 2> { typedef double2 type; };
	template <> struct VecType<double, 3> { typedef double3 type; };
	template <> struct VecType<double, 4> { typedef double4 type; };
	#else
	template <> struct VecType<float, 1> { typedef float1 type; };
	template <> struct VecType<float, 2> { typedef float2 type; };
	template <> struct VecType<float, 3> { typedef float3 type; };
	template <> struct VecType<float, 4> { typedef float4 type; };
	#endif

} // namespace quasar

// ARRAY_LENGTH macro. Warning: not to be used for function parameters (since
// they are pointers in C/C++)
#define ARRAY_LENGTH(x, y) (sizeof(x) / sizeof(y))

//=============================================================================
// Some custom functions
//=============================================================================
INLINE scalar __device__ __HOST_V__ frac(scalar x) { return x - floor(x); }
INLINE scalar __device__ __HOST_V__ frac(int x) { return 0; }
INLINE scalar __device__ __HOST_V__ mod(scalar x, scalar y) { return fmod(x, y); }
INLINE scalar __device__ __HOST_V__ sign(scalar x)
{
    return (x == 0) ? 0 : copysign(1.0f, x);
}
INLINE int __device__ __HOST_V__ sign(int x) { return (x == 0) ? 0 : (x < 0) ? -1 : 1; }
INLINE int __device__ __HOST_V__ sign(uint32_t x) { return (x == 0) ? 0 : 1; }
INLINE scalar __device__ __HOST_V__ sdiv(int x, int y) { return (scalar)x / (scalar)y; }
INLINE int __device__ __HOST_V__ idiv(int x, int y) { return x / y; }
INLINE int __device__ __HOST_V__ mod(int x, int y) { return x % y; }
template <typename T, typename R> INLINE T __device__ __HOST_V__ _and_(T a, R b)
{
    return (T)(a & b);
}
template <typename T, typename R> INLINE T __device__ __HOST_V__ _or_(T a, R b)
{
    return (T)(a | b);
}
template <typename T, typename R> INLINE T __device__ __HOST_V__ _xor_(T a, R b)
{
    return (T)(a ^ b);
}
template <typename T, typename R> INLINE T __device__ __HOST_V__ logand(T a, R b)
{
    return (T)((a != T()) && (b != R())); // T() and R() are the zero-elements
}
template <typename T, typename R> INLINE T __device__ __HOST_V__ logor(T a, R b)
{
    return (T)((a != T()) || (b != R()));
}
template <typename T, typename R> INLINE T __device__ __HOST_V__ logxor(T a, R b)
{
    return (T)((a != T()) ^ (b != R()));
}
template <typename T> INLINE T __device__ __HOST_V__ lognot(T a)
{
    return (T)(a == T());
}
template <typename T, typename R> INLINE T __device__ __HOST_V__ _shl_(T a, R b)
{
    return (T)(a << b);
}
template <typename T, typename R> INLINE T __device__ __HOST_V__ _shr_(T a, R b)
{
    return (T)(a >> b);
}
template <typename T> INLINE T __device__ __HOST_V__ _not_(T a) { return ~a; }
INLINE int __device__ __HOST_V__ _shl_(bool a, int b)
{
	return (int)a << b;
}
INLINE int __device__ __HOST_V__ _shr_(bool a, int b)
{
	return (int)a >> b;
}

template <> INLINE scalar __device__ __HOST_V__ _and_(scalar a, scalar b)
{
    return (scalar)((int)a & (int)b);
}
template <> INLINE scalar __device__ __HOST_V__ _or_(scalar a, scalar b)
{
    return (scalar)((int)a | (int)b);
}
template <> INLINE scalar __device__ __HOST_V__ _xor_(scalar a, scalar b)
{
    return (scalar)((int)a ^ (int)b);
}
template <> INLINE scalar __device__ __HOST_V__ _shl_(scalar a, scalar b)
{
    return (scalar)((int)a << (int)b);
}
template <> INLINE scalar __device__ __HOST_V__ _shr_(scalar a, scalar b)
{
    return (scalar)((int)a >> (int)b);
}
template <> INLINE scalar __device__ __HOST_V__ _not_(scalar a)
{
    return (scalar) ~(int)a;
}
INLINE int __device__ __HOST_V__ ind2pos(int sz, int a) { return a; }

INLINE int2 __device__ __HOST_V__ ind2pos(int2 sz, int a)
{
	unsigned x = (unsigned) a / (unsigned) sz.y;
	return make_int2(x, (signed)(a - x * (unsigned) sz.y));
}
INLINE int3 __device__ __HOST_V__ ind2pos(int3 sz, int a)
{
	int3 u;
    unsigned b = (unsigned)a / (unsigned)sz.z;
	u.z = (signed)(a - b * (unsigned)sz.z);
    unsigned c = b / (unsigned)sz.y;
	u.y = (signed)(b - c * (unsigned)sz.y);
	u.x = (signed)c;
	return u;
}
INLINE int4 __device__ __HOST_V__ ind2pos(int4 sz, int a)
{
	int4 u;
    unsigned b = (unsigned)a / (unsigned)sz.w;
	u.w = (signed)(a - b * (unsigned)sz.w);
    unsigned c = b / (unsigned)sz.z;
	u.z = (signed)(b - c * (unsigned)sz.z);
    unsigned d = c / (unsigned)sz.y;
	u.y = (signed)(c - d * (unsigned)sz.y);
	u.x = (int)d;
	return u;
}
template <int N> INLINE intN<N> __device__ __HOST_V__ ind2pos(intN<N> sz, int a)
{
	intN<N> u;
	for (int i = N - 1; i >= 1; i--)
	{
		unsigned b = (unsigned)a / (unsigned)sz.el[i];
		u.el[i] = (signed)(a - b * (unsigned)sz.el[i]);
		a = (signed)b;
	}
	u.el[0] = a;
	return u;
}

INLINE int __device__ __HOST_V__ pos2ind(int sz, int pos) { return pos; }
INLINE int __device__ __HOST_V__ pos2ind(int1 sz, int1 pos) { return pos.x; }
INLINE int __device__ __HOST_V__ pos2ind(int2 sz, int2 pos)
{
    return pos.x * sz.y + pos.y;
}
INLINE int __device__ __HOST_V__ pos2ind(int3 sz, int3 pos)
{
    return (pos.x * sz.y + pos.y) * sz.z + pos.z;
}
INLINE int __device__ __HOST_V__ pos2ind(int4 sz, int4 pos)
{
	return ((pos.x * sz.y + pos.y) * sz.z + pos.z) * sz.w + pos.w;
}
template <int N> INLINE int __device__ __HOST_V__ pos2ind(intN<N> sz, intN<N> pos)
{
    int ind = pos.el[0];
    for (int i = 1; i < N; i++)
    {
        ind *= sz.el[i];
        ind += pos.el[i];
    }
    return ind;
}

INLINE int __device__ __HOST_V__ pos2indchk(int sz, int pos)
{
	return 0 <= pos && pos < sz ? pos : -1;
}
INLINE int __device__ __HOST_V__ pos2indchk(int1 sz, int1 pos)
{
	return pos2indchk(sz.x, pos.x);
}
INLINE int __device__ __HOST_V__ pos2indchk(int2 sz, int2 pos)
{
	return 0 <= pos.x && 0 <= pos.y && pos.x < sz.x && pos.y < sz.y ? pos.x * sz.y + pos.y : -1;
}
INLINE int __device__ __HOST_V__ pos2indchk(int3 sz, int3 pos)
{
	return 0 <= pos.x && 0 <= pos.y && 0 <= pos.z && pos.x < sz.x && pos.y < sz.y && pos.z < sz.z ?
		   (pos.x * sz.y + pos.y) * sz.z + pos.z : -1;
}
INLINE int __device__ __HOST_V__ pos2indchk(int4 sz, int4 pos)
{
	return 0 <= pos.x && 0 <= pos.y && 0 <= pos.z && 0 <= pos.w &&
		   pos.x < sz.x && pos.y < sz.y && pos.z < sz.z && pos.w < sz.w ?
		   ((pos.x * sz.y + pos.y) * sz.z + pos.z) * sz.w + pos.w : -1;
}
template <int N> INLINE int __device__ __HOST_V__ pos2indchk(intN<N> sz, intN<N> pos)
{
	int ind = pos.el[0];
	bool in_bounds = 0 <= pos.el[0] && pos.el[0] < sz.el[0];
	for (int i = 1; i < N; i++)
	{
		in_bounds &= 0 <= pos.el[i] && pos.el[i] < sz.el[i];
		ind *= sz.el[i];
		ind += pos.el[i];
	}
	return in_bounds ? ind : -1;
}

template <int N> INLINE scalar __device__ __HOST_V__ sum(const scalarN<N> &a)
{
    scalar s = 0.0f;
    for (int i = 0; i < N; i++)
        s += a.el[i];
    return s;
}

template <int N> INLINE scalar __device__ __HOST_V__ mean(const scalarN<N> &a)
{
    return sum(a) / N;
}

template <int N> INLINE int __device__ __HOST_V__ sum(intN<N> a)
{
    int s = 0;
    for (int i = 0; i < N; i++)
        s += a.el[i];
    return s;
}

template <int N> INLINE scalar __device__ __HOST_V__ prod(const scalarN<N> &a)
{
    scalar s = 1.0f;
    for (int i = 0; i < N; i++)
        s *= a.el[i];
    return s;
}

template <int N> INLINE int __device__ __HOST_V__ prod(intN<N> a)
{
    int s = 1;
    for (int i = 0; i < N; i++)
        s *= a.el[i];
    return s;
}

template <int N> INLINE int __device__ __HOST_V__ _any_(intN<N> a)
{
    int s = 0;
    for (int i = 0; i < N; i++)
        s |= a.el[i];
    return s;
}

template <int N> INLINE int __device__ __HOST_V__ _all_(intN<N> a)
{
    int s = 1;
    for (int i = 0; i < N; i++)
        s &= a.el[i];
    return s;
}

INLINE scalar2 __device__ __HOST_V__ sincos(scalar a)
{
    scalar c, s;
#if !defined(DBL_SCALAR) && defined(__CUDA_ARCH__)
    __sincosf(a, &c, &s);
#elif defined(__GNUG__) && defined(DBL_SCALAR)
    sincos(a, &c, &s);
#elif defined(__GNUG__) && !defined(DBL_SCALAR)
    sincosf(a, &c, &s);
#else
    c = cosf(a);
    s = sinf(a);
#endif
    return make_scalar2(c, s);
}

template <int N>
INLINE __device__ __HOST_V__ scalar powN(scalar1 base) // Integer power of a scalar number
{
    return make_scalar1(powN<N>(base.x));
}
template <int N>
INLINE __device__ __HOST_V__ scalar2 powN(scalar2 base)
{
    return make_scalar2(powN<N>(base.x), powN<N>(base.y));
}
template <int N>
INLINE __device__ __HOST_V__ scalar3 powN(scalar3 base)
{
    return make_scalar3(powN<N>(base.x), powN<N>(base.y), powN<N>(base.z));
}
template <int N>
INLINE __device__ __HOST_V__ scalar4 powN(scalar4 base)
{
    return make_scalar4(powN<N>(base.x), powN<N>(base.y), powN<N>(base.z), powN<N>(base.w));
}

template <int N, int M>
INLINE __device__ __HOST_V__ scalarN<M> powN(scalarN<M> base) // Integer power of a scalar number
{
    scalarN<M> result;
    for (int i = 0; i < M; i ++)
        result.el[i] = powN<N>(base.el[i]);
    return result;
}

//=============================================================================
// SUPPORT FOR HALF PRECISION FLOATING POINT NUMBERS
//=============================================================================
#if defined(CUDART_VERSION) && CUDART_VERSION >= 9000
#include "cuda_fp16.h"
#define float2half_rn(f)	 __float2half_rn(f)
#define half2float(f)		 __half2float(f)
#define floats2half2_rn(f,g) __floats2half2_rn(f,g)
#define low2half(f)		     __low2half(f)
#define high2half(f)		 __high2half(f)
#define HAS_CUDA_HALF_SUPPORT
#else

#if !defined(HAS_CUDA_HALF_SUPPORT)
struct __half {
	uint16_t val;
	INLINE __HOST_V__ __device__ __half() {}
	INLINE __HOST_V__ __device__ __half(scalar v);
	INLINE __HOST_V__ __device__ operator scalar() const;
};
struct __half2 { uint32_t val; };
#endif

// Following functions are defined in older CUDA versions
#if defined(TARGET_CUDA) && CUDART_VERSION < 9000
INLINE __HOST_V__ __device__ __half float2half_rn(float f)
{
	__half ret;
	ret.val = __float2half_rn(f);
	return ret;
}
INLINE __HOST_V__ __device__ float half2float(__half value)
{
	return __half2float(value.val);
}
#else
// CPU replacement for the __float2half_rn function.
INLINE __HOST_V__ __device__ __half float2half_rn(float f)
{
    uint32_t inu = *(uint32_t *) &f;
    uint32_t t1 = inu & 0x7fffffff;             // Non-sign bits
    uint32_t t2 = inu & 0x80000000;             // Sign bit
    uint32_t t3 = inu & 0x7f800000;             // Exponent
    t1 >>= 13;                                  // Align mantissa on MSB
    t2 >>= 16;                                  // Shift sign bit into position
    t1 -= 0x1c000;                              // Adjust bias
    t1 = (t3 < 0x38800000) ? 0 : t1;            // Flush-to-zero
    t1 = (t3 > 0x8e000000) ? 0x7bff : t1;       // Clamp-to-max
    t1 = (t3 == 0 ? 0 : t1);                    // Denormals-as-zero
    t1 |= t2;                                   // Re-insert sign bit
    return (__half &) t1;
}

// CPU replacement for the __half2float function.
INLINE __HOST_V__ __device__ float half2float(__half value)
{
	uint32_t inu = (uint32_t&)value;
    uint32_t t1 = inu & 0x7fffu;                // Non-sign bits
    uint32_t t2 = inu & 0x8000u;                // Sign but
    uint32_t t3 = inu & 0x7c00u;                // Exponent
    t1 = t1 << 13;                              // Align mantissa on MSB
    t2 = t2 << 16;                              // Shift sign bit into position
    t1 += 0x38000000u;                          // Adjust bias
    t1 = (t3 == 0 ? 0 : t1);                    // Denormals-as-zero
    t1 |= t2;                                   // Reinsert sign bit
    float f = 0.0f;
    *(uint32_t*)&f = t1;
    return f;
}
#endif

INLINE __HOST_V__ __device__ __half2 floats2half2_rn(float f1, float f2)
{
	__half2 y; y.val = (uint32_t)float2half_rn(f1).val | ((uint32_t)float2half_rn(f2).val << 16u);
	return y;
}

INLINE __HOST_V__ __device__ __half2 __halves2half2(__half f1, __half f2)
{
	__half2 y = { f1.val | ((uint32_t)f2.val << 16) };
	return y;
}

INLINE __HOST_V__ __device__ __half low2half(__half2 x) { __half y; y.val = x.val & 0xffffu; return y; }
INLINE __HOST_V__ __device__ __half high2half(__half2 x) { __half y; y.val = (x.val & 0xffff0000u) >> 16u; return y; }

INLINE __HOST_V__ __device__ __half::__half(scalar v) { *this = float2half_rn(v); }
INLINE __HOST_V__ __device__ __half::operator scalar() const { return half2float(*this); }
#endif

// Half vector data type
template <int N> struct __halfN { 
	__half el[N]; 

private:
	
public:
	template <__half fn(__half a, __half b)>
	INLINE __HOST_V__ __device__ __halfN binary_op(__halfN b) const {
		__halfN result; FOR_ALL(result.el[i] = fn(el[i], b.el[i])); return result;
	}
	template <__half fn(__half a)>
	INLINE __HOST_V__ __device__ __halfN unary_op() const {
		__halfN result; FOR_ALL(result.el[i] = fn(el[i])); return result;
	}
	INLINE __HOST_V__ __device__ __halfN operator+ (__halfN b) const { return binary_op<quasar::OP_add>(b); }
	INLINE __HOST_V__ __device__ __halfN operator* (__halfN b) const { return binary_op<quasar::OP_mul>(b); }
	INLINE __HOST_V__ __device__ __halfN operator/ (__halfN b) const { return binary_op<quasar::OP_div>(b); }
	INLINE __HOST_V__ __device__ __halfN operator- (__halfN b) const { return binary_op<quasar::OP_sub>(b); }
	INLINE __HOST_V__ __device__ __halfN operator- () const { return unary_op<quasar::OP_minus>(); }
};

namespace quasar
{
	template <int N> struct VecBaseType<__half, N> { typedef __halfN<N> type; };
	template <> struct VecType<__half, 2> { typedef __half2 type; };

	INLINE __half2 __device__ make___half2(__half x, __half y)
	{
		return __halves2half2(x, y);
	}

	// Half conversions (with scalar) rely on explicit conversion functions
	INLINE __half __device__ autocast(scalar a, ReturnType<__half>)
	{
		return float2half_rn(a);
	}
	INLINE scalar __device__ autocast(__half a, ReturnType<scalar>)
	{
		return half2float(a);
	}
	INLINE __half2 __device__ autocast(scalar2 a, ReturnType<__half2>)
	{
		return floats2half2_rn((float)a.x, (float)a.y);
	}
	INLINE __half2 __device__ autocast(int2 a, ReturnType<__half2>)
	{
		return floats2half2_rn((float)a.x, (float)a.y);
	}
	INLINE scalar2 __device__ autocast(__half2 a, ReturnType<scalar2>)
	{
		return make_scalar2(half2float(low2half(a)), half2float(high2half(a)));
	}
	INLINE __halfN<3> __device__ autocast(scalar3 a, ReturnType<__halfN<3> >)
	{
		__halfN<3> y;
		y.el[0] = float2half_rn(a.x);
		y.el[1] = float2half_rn(a.y);
		y.el[2] = float2half_rn(a.z);
		return y;
	}
	INLINE __halfN<3> __device__ autocast(int3 a, ReturnType<__halfN<3> >)
	{
		__halfN<3> y;
		y.el[0] = float2half_rn((float)a.x);
		y.el[1] = float2half_rn((float)a.y);
		y.el[2] = float2half_rn((float)a.z);
		return y;
	}
	INLINE scalar3 __device__ autocast(__halfN<3> a, ReturnType<scalar3>)
	{
		return make_scalar3(half2float(a.el[0]), half2float(a.el[1]), half2float(a.el[2]));
	}
	INLINE __halfN<4> __device__ autocast(scalar4 a, ReturnType<__halfN<4> >)
	{
		__halfN<4> y;
		y.el[0] = float2half_rn(a.x);
		y.el[1] = float2half_rn(a.y);
		y.el[2] = float2half_rn(a.z);
		y.el[3] = float2half_rn(a.w);
		return y;
	}
	INLINE __halfN<4> __device__ autocast(int4 a, ReturnType<__halfN<4> >)
	{
		__halfN<4> y;
		y.el[0] = float2half_rn((float)a.x);
		y.el[1] = float2half_rn((float)a.y);
		y.el[2] = float2half_rn((float)a.z);
		y.el[3] = float2half_rn((float)a.w);
		return y;
	}
	INLINE scalar4 __device__ autocast(__halfN<4> a, ReturnType<scalar4>)
	{
		return make_scalar4(half2float(a.el[0]), half2float(a.el[1]), half2float(a.el[2]), half2float(a.el[3]));
	}
	template <int N>
	INLINE scalarN<N> __device__ autocast(__halfN<N> a, ReturnType<scalarN<N> >)
	{
		scalarN<N> y;
		FOR_ALL(y.el[i] = half2float(a.el[i]));
		return y;
	}
	template <int N>
	INLINE __halfN<N> __device__ autocast(scalarN<N> a, ReturnType<__halfN<N> >)
	{
		__halfN<N> y;
		FOR_ALL(y.el[i] = float2half_rn(a.el[i]));
		return y;
	}
	template <int N>
	INLINE __halfN<N> __device__ autocast(intN<N> a, ReturnType<__halfN<N> >)
	{
		__halfN<N> y;
		FOR_ALL(y.el[i] = float2half_rn((float)a.el[i]));
		return y;
	}
} // namespace quasar

// Half2 functions
INLINE __half __device__ __HOST_V__ sum(__half2 a)
{
#ifndef HAS_CUDA_HALF_SUPPORT
    return low2half(a) + high2half(a);
#else
    return float2half_rn(half2float(low2half(a)) + half2float(high2half(a)));
#endif
}
INLINE __half __device__ __HOST_V__ mean(__half2 a)
{
#ifndef HAS_CUDA_HALF_SUPPORT
    return (low2half(a) + high2half(a)) * 0.5f;
#else
    return float2half_rn(0.5f * (half2float(low2half(a)) + half2float(high2half(a))));
#endif
}
INLINE __half __device__ __HOST_V__ prod(__half2 a)
{
#ifndef HAS_CUDA_HALF_SUPPORT
    return low2half(a) * high2half(a);
#else
    return float2half_rn(half2float(low2half(a)) * half2float(high2half(a)));
#endif
}
INLINE __half __device__ __HOST_V__ min(__half2 a)
{
    return float2half_rn(min(half2float(low2half(a)), half2float(high2half(a))));
}
INLINE __half __device__ __HOST_V__ max(__half2 a)
{
    return float2half_rn(max(half2float(low2half(a)), half2float(high2half(a))));
}


namespace quasar
{
#if !defined(HALF_SCALAR)
    typedef VectorBase<scalar> Vector;
    typedef MatrixBase<scalar> Matrix;
    typedef CubeBase<scalar> Cube;
#else
    typedef VectorBase<__half> Vector;
    typedef MatrixBase<__half> Matrix;
    typedef CubeBase<__half> Cube;
#endif

} // namespace quasar


//=============================================================================
// SUPPORT FOR COMPLEX NUMBERS
//=============================================================================
template <typename T>
struct cscalar_t
{
    T x;
    T y;

    INLINE __device__ __HOST_V__ cscalar_t() : x(0), y(0) {}
    INLINE __device__ __HOST_V__ cscalar_t(T x) : x(x), y(0) {}
    INLINE __device__ __HOST_V__ cscalar_t(T x, T y) : x(x), y(y) {}
    //INLINE __device__ cscalar(const cscalar &b) : x(b.x), y(b.y) {}
    INLINE __device__ __HOST_V__ cscalar_t operator*(cscalar_t b) const
    {
        return cscalar_t(x * b.x - y * b.y, x * b.y + y * b.x);
    }
    INLINE __device__ __HOST_V__ cscalar_t operator+(cscalar_t b) const
    {
        return cscalar_t(x + b.x, y + b.y);
    }
    INLINE __device__ __HOST_V__ cscalar_t operator-(cscalar_t b) const
    {
        return cscalar_t(x - b.x, y - b.y);
    }
    INLINE __device__ __HOST_V__ cscalar_t operator/(cscalar_t b) const
    {
        if (abs(b.x) < abs(b.y))
        { // implementation which is overflow-safe
            T t = b.x / b.y;
            T d = b.y * (1.0f + t * t);
            return (*this) * cscalar_t(t / d, -1.0f / d);
        }
        else
        {
            T t = b.y / b.x;
            T d = b.x * (1.0f + t * t);
            return (*this) * cscalar_t(1.0f / d, -t / d);
        }
    }
    INLINE __device__ __HOST_V__ cscalar_t operator*=(cscalar_t b)
    {
        return *this = (*this * b);
    }
    INLINE __device__ __HOST_V__ cscalar_t operator+=(cscalar_t b)
    {
        return *this = (*this + b);
    }
    INLINE __device__ __HOST_V__ cscalar_t operator-=(cscalar_t b)
    {
        return *this = (*this - b);
    }
    INLINE __device__ __HOST_V__ cscalar_t operator/=(cscalar_t b)
    {
        return *this = (*this /= b);
    }
    INLINE __device__ __HOST_V__ int operator==(cscalar_t b) { return x == b.x && y == b.y; }
    INLINE __device__ __HOST_V__ int operator!=(cscalar_t b) { return x != b.x || y != b.y; }
    INLINE __device__ __HOST_V__ T square_mag() const { return x * x + y * y; }
    INLINE cscalar_t __device__ __HOST_V__ operator=(cscalar_t b)
    {
        x = b.x;
        y = b.y;
        return *this;
    }
	template <typename R>
	INLINE __device__ __HOST_V__ operator cscalar_t<R>() const {
		return cscalar_t<R>((R)x, (R)y);
	}
};

typedef cscalar_t<scalar> cscalar;

INLINE __device__ __HOST_V__ cscalar make_cscalar(scalar x, scalar y)
{
    return cscalar(x, y);
}
template <typename T>
INLINE __device__ __HOST_V__ T abs(cscalar_t<T> a) { return sqrt(a.square_mag()); }
INLINE __device__ __HOST_V__ scalar angle(scalar a) { return (scalar)0; }
template <typename T>
INLINE __device__ __HOST_V__ T angle(cscalar_t<T> a) { return atan2(a.y, a.x); }
INLINE __device__ __HOST_V__ scalar conj(scalar a) { return a; }
template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> conj(cscalar_t<T> a) { return cscalar_t<T>(a.x, -a.y); }

// Operators
template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> operator/(cscalar_t<T> a, T b)
{
    return cscalar_t<T>(a.x / b, a.y / b);
}
template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> operator/(T a, cscalar_t<T> b)
{
    if (abs(b.x) < abs(b.y))
    { // implementation which is overflow-safe
        T t = b.x / b.y;
        T d = b.y * (1.0f + t * t);
        return cscalar_t<T>(a * t / d, -a / d);
    }
    else
    {
        T t = b.y / b.x;
        T d = b.x * (1.0f + t * t);
        return cscalar_t<T>(a / d, -a * t / d);
    }
}
template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> operator*(cscalar_t<T> a, R b)
{
    return cscalar_t<T>(a.x * b, a.y * b);
}
template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> operator*(R a, cscalar_t<T> b)
{
    return cscalar_t<T>(a * b.x, a * b.y);
}
template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> operator+(cscalar_t<T> a, R b)
{
    return cscalar_t<T>(a.x + b, a.y);
}
template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> operator+(R a, cscalar_t<T> b)
{
    return cscalar_t<T>(a + b.x, b.y);
}
template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> operator-(cscalar_t<T> a, R b)
{
    return cscalar_t<T>(a.x - b, a.y);
}
template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> operator-(R a, cscalar_t<T> b)
{
    return cscalar_t<T>(a - b.x, -b.y);
}
template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> operator-(cscalar_t<T> a) { return cscalar_t<T>(-a.x, -a.y); }

#ifndef __CUDA_ARCH__
#define __cosf cosf
#define __sinf sinf
#endif

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> pow(cscalar_t<T> a, cscalar_t<T> b)
{
    T rm = a.x * a.x + a.y * a.y;
    if (rm == 0)
        return cscalar_t<T>(0, 0);
    T r = log(rm);
    T t = atan2(a.y, a.x);
    T ra = exp(0.5f * b.x * r - b.y * t);
    T rb = 0.5 * b.y * r + b.x * t;
    scalar2 z = sincos(rb);
    return cscalar_t<T>(ra * z.x, ra * z.y);
}

template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> pow(R a, cscalar_t<T> b)
{
    T r = log(a);
    T t = sign(a) * M_PI;
    T ra = exp(b.x * r - b.y * t);
    T rb = b.y * r + b.x * t;
    scalar2 z = sincos(rb);
    return cscalar_t<T>(ra * z.x, ra * z.y);
}

template <typename T, typename R>
INLINE __device__ __HOST_V__ cscalar_t<T> pow(cscalar_t<T> a, R b)
{
    T r = log(a.x * a.x + a.y * a.y);
    T t = atan2(a.y, a.x);
    T ra = exp(0.5f * b * r);
    T rb = b * t;
    scalar2 z = sincos(rb);
    return cscalar_t<T>(ra * z.x, ra * z.y);
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> log(cscalar_t<T> a)
{
    return cscalar_t<T>(0.5 * log(a.x * a.x + a.y * a.y), atan2(a.y, a.x));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> log10(cscalar_t<T> a)
{
    T ln10 = log(10.0f);
    return cscalar_t<T>(0.5 * log(a.x * a.x + a.y * a.y) / ln10,
                   atan2(a.y, a.x) / ln10);
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> log2(cscalar_t<T> a)
{
    T ln2 = log(2.0f);
    return cscalar_t<T>(0.5 * log(a.x * a.x + a.y * a.y) / ln2,
                   atan2(a.y, a.x) / ln2);
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> sin(cscalar_t<T> a)
{
    return cscalar_t<T>(sin(a.x) * cosh(a.y), cos(a.x) * sinh(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> cos(cscalar_t<T> a)
{
    scalar2 z = sincos(a.x);
    return cscalar_t<T>(z.y * cosh(a.y), -z.x * sinh(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> tan(cscalar_t<T> a)
{
    scalar2 z = sincos(a.x);
    T cosr = z.y;
    T sinhi = sinh(a.y);
    T denom = cosr * cosr + sinhi * sinhi;
    return cscalar_t<T>(z.x * cosr / denom, sinhi * cosh(a.y) / denom);
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> exp(cscalar_t<T> a)
{
    T expr = exp(a.x);
    scalar2 z = sincos(a.y);
    return cscalar_t<T>(expr * z.y, expr * z.x);
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> exp2(cscalar_t<T> a)
{
    T expr = exp2(a.x);
    T theta = M_LOG2 * a.y;
    scalar2 z = sincos(theta);
    return cscalar_t<T>(expr * z.y, expr * z.x);
}

template <typename T>
INLINE __device__ __HOST_V__ T sign(cscalar_t<T> a)
{
    return (a.x == 0 && a.y == 0) ? 0 : sign(a.x);
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> sqrt(cscalar_t<T> a)
{
    T r = sqrt(a.x * a.x + a.y * a.y);
    return cscalar_t<T>(sqrt(0.5 * (r + a.x)),
                   sqrt(0.5 * (r - a.x)) * (a.y >= 0 ? 1.0 : -1.0));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> asin(cscalar_t<T> z)
{
    return cscalar_t<T>(0, -1) *
           log(cscalar_t<T>(0, 1) * z +
               sqrt(cscalar_t<T>(1) - z * z)); // -i*ln(iz+(1-z^2)^(1/2))
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> acos(cscalar_t<T> z)
{
    return cscalar_t<T>(0.5 * M_PI) - asin(z);
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> atan(cscalar_t<T> z)
{
    const cscalar_t<T> iz = cscalar_t<T>(0, 1) * z;
    return (cscalar_t<T>(0, 1) * log(cscalar_t<T>(1) - iz) - log(cscalar_t<T>(1) + iz)) * 0.5;
}

template <typename T>
INLINE __device__ __HOST_V__ T real(cscalar_t<T> a) { return a.x; }

template <typename T>
INLINE __device__ __HOST_V__ T imag(cscalar_t<T> a) { return a.y; }

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> ceil(cscalar_t<T> a)
{
    return cscalar_t<T>(ceil(a.x), ceil(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> floor(cscalar_t<T> a)
{
    return cscalar_t<T>(floor(a.x), floor(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> round(cscalar_t<T> a)
{
    return cscalar_t<T>(round(a.x), round(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> frac(cscalar_t<T> a)
{
    return cscalar_t<T>(frac(a.x), frac(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> saturate(cscalar_t<T> a)
{
    return cscalar_t<T>(saturate(a.x), saturate(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> _float_(cscalar_t<T> a) { return a; }
template <typename T>
INLINE __device__ __HOST_V__ cscalar_t<T> _int_(cscalar_t<T> a)
{
    return cscalar_t<T>(int(a.x), int(a.y));
}

template <typename T>
INLINE __device__ __HOST_V__ int _isfinite_(cscalar_t<T> a)
{
    return _isfinite_(a.x) && _isfinite_(a.y);
}
template <typename T>
INLINE __device__ __HOST_V__ int _isinf_(cscalar_t<T> a)
{
    return _isinf_(a.x) || _isinf_(a.y);
}
template <typename T>
INLINE __device__ __HOST_V__ int _isnan_(cscalar_t<T> a)
{
    return _isnan_(a.x) || _isnan_(a.y);
}

// Complex vector types
struct cscalar1
{
	cscalar x;
};
struct cscalar2
{
	cscalar x, y;
};
struct cscalar3
{
	cscalar x, y, z;
};
struct cscalar4
{
	cscalar x, y, z, w;
};
template <int N> struct cscalarN
{
	cscalar el[N];
};

INLINE __device__ __HOST_V__ cscalar1 make_cscalar1(cscalar x)
{
	cscalar1 t = { x };
	return t;
}
INLINE __device__ __HOST_V__ cscalar2 make_cscalar2(cscalar x, cscalar y)
{
	cscalar2 t = { x, y };
	return t;
}
INLINE __device__ __HOST_V__ cscalar3 make_cscalar3(cscalar x, cscalar y, cscalar z)
{
	cscalar3 t = { x, y, z };
	return t;
}
INLINE __device__ __HOST_V__ cscalar4
make_cscalar4(cscalar x, cscalar y, cscalar z, cscalar w)
{
	cscalar4 t = { x, y, z, w };
	return t;
}


// Use variadic template packs when the compiler supports it.

#ifdef CPLUSPLUS14

namespace quasar
{
    // make_ucharN functions and friends
    template <typename T, typename C>
    INLINE __HOST_V__ __device__ void expand_parampack(T *dst, C cur)
    {
        dst[0] = (T)cur;
    }

    template <typename T, typename C, typename ...R>
    INLINE __HOST_V__ __device__ void expand_parampack(T *dst, C cur, R... vals)
    {
        expand_parampack(dst, cur);
        expand_parampack(dst + 1, vals...);
    }
}

#define MAKE_TN(Type, ElemType) \
	template <typename ...T>	\
	INLINE __HOST_V__ __device__ typename quasar::VecType<ElemType, sizeof...(T)>::type make_##Type##N(T... vals) { \
		Type##N<sizeof...(T)> v;	\
		quasar::expand_parampack<ElemType>(v.el, vals...); return v; }

#else

// Note: originally script-generated. Before CUDA 7, NVCC did not support
// functions with variadic arguments.

#define MAKE_TN(R,T) \
    INLINE R##N<5> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e) { \
        R##N<5> y = {a,b,c,d,e}; \
        return y; \
    } \
    INLINE R##N<6> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f) { \
        R##N<6> y = {a,b,c,d,e,f}; \
        return y; \
    } \
    INLINE R##N<7> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g) { \
        R##N<7> y = {a,b,c,d,e,f,g}; \
        return y; \
    } \
    INLINE R##N<8> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h) { \
        R##N<8> y = {a,b,c,d,e,f,g,h}; \
        return y; \
    } \
    INLINE R##N<9> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i) { \
        R##N<9> y = {a,b,c,d,e,f,g,h,i}; \
        return y; \
    } \
    INLINE R##N<10> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j) { \
        R##N<10> y = {a,b,c,d,e,f,g,h,i,j}; \
        return y; \
    } \
    INLINE R##N<11> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k) { \
        R##N<11> y = {a,b,c,d,e,f,g,h,i,j,k}; \
        return y; \
    } \
    INLINE R##N<12> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l) { \
        R##N<12> y = {a,b,c,d,e,f,g,h,i,j,k,l}; \
        return y; \
    } \
    INLINE R##N<13> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m) { \
        R##N<13> y = {a,b,c,d,e,f,g,h,i,j,k,l,m}; \
        return y; \
    } \
    INLINE R##N<14> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n) { \
        R##N<14> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n}; \
        return y; \
    } \
    INLINE R##N<15> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o) { \
        R##N<15> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o}; \
        return y; \
    } \
    INLINE R##N<16> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p) { \
        R##N<16> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p}; \
        return y; \
    } \
    INLINE R##N<17> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q) { \
        R##N<17> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q}; \
        return y; \
    } \
    INLINE R##N<18> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r) { \
        R##N<18> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r}; \
        return y; \
    } \
    INLINE R##N<19> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s) { \
        R##N<19> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s}; \
        return y; \
    } \
    INLINE R##N<20> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t) { \
        R##N<20> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t}; \
        return y; \
    } \
    INLINE R##N<21> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u) { \
        R##N<21> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u}; \
        return y; \
    } \
    INLINE R##N<22> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v) { \
        R##N<22> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v}; \
        return y; \
    } \
    INLINE R##N<23> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w) { \
        R##N<23> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w}; \
        return y; \
    } \
    INLINE R##N<24> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x) { \
        R##N<24> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x}; \
        return y; \
    } \
    INLINE R##N<25> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A) { \
        R##N<25> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A}; \
        return y; \
    } \
    INLINE R##N<26> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B) { \
        R##N<26> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B}; \
        return y; \
    } \
    INLINE R##N<27> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C) { \
        R##N<27> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C}; \
        return y; \
    } \
    INLINE R##N<28> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D) { \
        R##N<28> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D}; \
        return y; \
    } \
    INLINE R##N<29> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E) { \
        R##N<29> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E}; \
        return y; \
    } \
    INLINE R##N<30> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F) { \
        R##N<30> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F}; \
        return y; \
    } \
    INLINE R##N<31> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G) { \
        R##N<31> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G}; \
        return y; \
    } \
    INLINE R##N<32> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H) { \
        R##N<32> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H}; \
        return y; \
    } \
    INLINE R##N<33> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I) { \
        R##N<33> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I}; \
        return y; \
    } \
    INLINE R##N<34> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J) { \
        R##N<34> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J}; \
        return y; \
    } \
    INLINE R##N<35> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K) { \
        R##N<35> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K}; \
        return y; \
    } \
    INLINE R##N<36> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L) { \
        R##N<36> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L}; \
        return y; \
    } \
    INLINE R##N<37> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M) { \
        R##N<37> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M}; \
        return y; \
    } \
    INLINE R##N<38> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N) { \
        R##N<38> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N}; \
        return y; \
    } \
    INLINE R##N<39> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O) { \
        R##N<39> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O}; \
        return y; \
    } \
    INLINE R##N<40> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P) { \
        R##N<40> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P}; \
        return y; \
    } \
    INLINE R##N<41> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q) { \
        R##N<41> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q}; \
        return y; \
    } \
    INLINE R##N<42> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S) { \
        R##N<42> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S}; \
        return y; \
    } \
    INLINE R##N<43> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U) { \
        R##N<43> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U}; \
        return y; \
    } \
    INLINE R##N<44> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V) { \
        R##N<44> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V}; \
        return y; \
    } \
    INLINE R##N<45> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W) { \
        R##N<45> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W}; \
        return y; \
    } \
    INLINE R##N<46> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X) { \
        R##N<46> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X}; \
        return y; \
    } \
    INLINE R##N<47> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y) { \
        R##N<47> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y}; \
        return y; \
    } \
    INLINE R##N<48> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z) { \
        R##N<48> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z}; \
        return y; \
    } \
    INLINE R##N<49> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1) { \
        R##N<49> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1}; \
        return y; \
    } \
    INLINE R##N<50> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1) { \
        R##N<50> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1}; \
        return y; \
    } \
    INLINE R##N<51> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1) { \
        R##N<51> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1}; \
        return y; \
    } \
    INLINE R##N<52> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1) { \
        R##N<52> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1}; \
        return y; \
    } \
    INLINE R##N<53> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1) { \
        R##N<53> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1}; \
        return y; \
    } \
    INLINE R##N<54> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1) { \
        R##N<54> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1}; \
        return y; \
    } \
    INLINE R##N<55> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1) { \
        R##N<55> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1}; \
        return y; \
    } \
    INLINE R##N<56> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1) { \
        R##N<56> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1}; \
        return y; \
    } \
    INLINE R##N<57> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1) { \
        R##N<57> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1}; \
        return y; \
    } \
    INLINE R##N<58> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1,T J1) { \
        R##N<58> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1,J1}; \
        return y; \
    } \
    INLINE R##N<59> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1,T J1,T K1) { \
        R##N<59> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1,J1,K1}; \
        return y; \
    } \
    INLINE R##N<60> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1,T J1,T K1,T L1) { \
        R##N<60> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1,J1,K1,L1}; \
        return y; \
    } \
    INLINE R##N<61> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1,T J1,T K1,T L1,T M1) { \
        R##N<61> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1,J1,K1,L1,M1}; \
        return y; \
    } \
    INLINE R##N<62> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1,T J1,T K1,T L1,T M1,T N1) { \
        R##N<62> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1,J1,K1,L1,M1,N1}; \
        return y; \
    } \
    INLINE R##N<63> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1,T J1,T K1,T L1,T M1,T N1,T O1) { \
        R##N<63> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1,J1,K1,L1,M1,N1,O1}; \
        return y; \
    } \
    INLINE R##N<64> __device__ __HOST_V__ make_##R##N(T a,T b,T c,T d,T e,T f,T g,T h,T i,T j,T k,T l,T m,T n,T o,T p,T q,T r,T s,T t,T u,T v,T w,T x,T A,T B,T C,T D,T E,T F,T G,T H,T I,T J,T K,T L,T M,T N,T O,T P,T Q,T S,T U,T V,T W,T X,T Y,T Z,T A1,T B1,T C1,T D1,T E1,T F1,T G1,T H1,T I1,T J1,T K1,T L1,T M1,T N1,T O1,T P1) { \
        R##N<64> y = {a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,S,U,V,W,X,Y,Z,A1,B1,C1,D1,E1,F1,G1,H1,I1,J1,K1,L1,M1,N1,O1,P1}; \
        return y; \
    } 

// In some circumstances, intN<4> and scalarN<4> are also used
INLINE scalarN<4>   __device__ __HOST_V__ make_scalarN(scalar a, scalar b, scalar c, scalar d) { return make_scalar4(a,b,c,d); }
INLINE intN<4>      __device__ __HOST_V__ make_intN(int a, int b, int c, int d) { return make_int4(a,b,c,d); } 
#endif

MAKE_TN(char, int8_t)
MAKE_TN(uchar, uint8_t)
MAKE_TN(short, int16_t)
MAKE_TN(ushort, uint16_t)
MAKE_TN(uint, uint32_t)
MAKE_TN(int, int)
MAKE_TN(scalar, scalar)
MAKE_TN(__half, __half)
MAKE_TN(cscalar, cscalar)


#undef MAKE_TN

namespace quasar
{
	// =========================================================================
	// Complex-valued vector, matrix and cube types
	// =========================================================================
	#if !defined(HALF_SCALAR)
    typedef VectorBase<cscalar> CVector;
    typedef MatrixBase<cscalar> CMatrix;
    typedef CubeBase<cscalar> CCube;
	#else
    typedef VectorBase<cscalar_t<__half> > CVector;
    typedef MatrixBase<cscalar_t<__half> > CMatrix;
    typedef CubeBase<cscalar_t<__half> > CCube;
	#endif

	// Register the vector type
	template <int N> struct VecBaseType<cscalar, N> { typedef cscalarN<N> type; };

	//==========================================================================
	// Fixed-length sequence functions
	//==========================================================================
	template <typename T>
	INLINE __HOST_V__ __device__ T seq(int a, int step) { return seq(a, step, ReturnType<T>()); }
	template <typename T>
	INLINE __HOST_V__ __device__ T seq(scalar a, scalar step) { return seq(a, step, ReturnType<T>()); }

	// Integer sequence function (implements a..step..b for fixed-length vectors)
	template <int N>
	INLINE __HOST_V__ __device__ intN<N> seq(int a, int step, ReturnType<intN<N> >) {
		intN<N> y; for (int i = 0; i < N; i ++, a += step) y.el[i] = a; return y;
	}	INLINE __HOST_V__ __device__ int1 seq(int a, int s, ReturnType<int1>) { return make_int1(a); }
	INLINE __HOST_V__ __device__ int2 seq(int a, int s, ReturnType<int2>) { return make_int2(a, a + s); }
	INLINE __HOST_V__ __device__ int3 seq(int a, int s, ReturnType<int3>) { return make_int3(a, a + s, a + 2 * s); }
	INLINE __HOST_V__ __device__ int4 seq(int a, int s, ReturnType<int4>) { return make_int4(a, a + s, a + 2 * s, a + 3 * s); }

	// Scalar sequence function (implements a..step..b for fixed-length vectors)
	template <int N>
	INLINE __HOST_V__ __device__ scalarN<N> seq(scalar a, scalar step, ReturnType<scalarN<N> >) {
		scalarN<N> y; for (int i = 0; i < N; i ++, a += step) y.el[i] = a; return y;
	}
	INLINE __HOST_V__ __device__ scalar1 seq(scalar a, scalar s, ReturnType<scalar1>) { return make_scalar1(a); }
	INLINE __HOST_V__ __device__ scalar2 seq(scalar a, scalar s, ReturnType<scalar2>) { return make_scalar2(a, a + s); }
	INLINE __HOST_V__ __device__ scalar3 seq(scalar a, scalar s, ReturnType<scalar3>) { return make_scalar3(a, a + s, a + 2 * s); }
	INLINE __HOST_V__ __device__ scalar4 seq(scalar a, scalar s, ReturnType<scalar4>) { return make_scalar4(a, a + s, a + 2 * s, a + 3 * s); }

	#if !defined(_MSC_VER) || _MSC_VER >= 1800
	#define _explicit explicit
	#else
	#define _explicit
	#endif

	// Extends the basic fixed-length vector types with constructor and casting operators
	// The vector type is aligned according to the settings of the invididual type that it holds
	template <typename T, int N> class vecTN
	{
		typedef typename VecType<T, N>::type V;
		V val;

	public:
		__HOST_V__ __device__ vecTN() :val() {}
		__HOST_V__ __device__ vecTN(const vecTN& v) : val(v.val) {}
		__HOST_V__ __device__ vecTN(V v) : val(v) {}
		__HOST_V__ __device__ vecTN(typename VecType<int, N>::type v) : val(autocast<typename VecType<T, N>::type>(v)) {}
		__HOST_V__ __device__ vecTN(typename VecType<scalar, N>::type v) : val(autocast<typename VecType<T, N>::type>(v)) {}
		__HOST_V__ __device__ T& operator[] (int index) { return ((T *)&val)[index]; }
		__HOST_V__ __device__ T operator[] (int index) const { return ((T *)&val)[index]; }
		__HOST_V__ __device__ operator V() const { return val; }
		_explicit __HOST_V__ __device__ operator typename VecType<int, N>::type() const { return autocast<typename VecType<int, N>::type>(val); }
		_explicit __HOST_V__ __device__ operator typename VecType<scalar, N>::type() const { return autocast<typename VecType<scalar, N>::type>(val); }
	};

	// Standard operators for vecTN
	#define IMPL_VECTN_OP(X)	\
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> operator X(vecTN<T, N> a, vecTN<T, N> b) { vecTN<T, N> c; FOR_ALL(c[i] = a[i] X b[i]); return c; } \
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> operator X(vecTN<T, N> a, int b) { vecTN<T, N> c; FOR_ALL(c[i] = a[i] X b); return c; } \
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> operator X(int a, vecTN<T, N> b) { vecTN<T, N> c; FOR_ALL(c[i] = a X b[i]); return c; } \
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> operator X(vecTN<T, N> a, typename VecType<T, N>::type b) { return a X vecTN<T, N>(b); } \
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> operator X(typename VecType<T, N>::type a, vecTN<T, N> b) { return vecTN<T, N>(a) X b; }

	IMPL_VECTN_OP(+)
	IMPL_VECTN_OP(-)
	IMPL_VECTN_OP(*)
	IMPL_VECTN_OP(/)
	IMPL_VECTN_OP(==)
	IMPL_VECTN_OP(!=)
	IMPL_VECTN_OP(<)
	IMPL_VECTN_OP(>)
	IMPL_VECTN_OP(<=)
	IMPL_VECTN_OP(>=)

	#undef IMPL_VECTN_OP

	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> operator-(vecTN<T, N> a) { vecTN<T, N> c; FOR_ALL(c[i] = -a[i]); return c; }

	INLINE int __device__ __HOST_V__ autocast(scalar value, ReturnType<int>) { return (int)value; }
	INLINE scalar __device__ __HOST_V__ autocast(int value, ReturnType<scalar>) { return (scalar)value; }
	INLINE cscalar __device__ __HOST_V__ autocast(scalar value, ReturnType<cscalar>) { return make_cscalar(value, 0); }
	INLINE cscalar __device__ __HOST_V__ autocast(int value, ReturnType<cscalar>) { return make_cscalar(value, 0); }
	template <typename T, int N, typename R>
	INLINE R __device__ __HOST_V__ autocast(vecTN<T, N> value, ReturnType<R>) { return autocast((typename VecType<T, N>::type)value, ReturnType<R>()); }

	// Helper class, a replacement for std::array since not all NVCC versions support C++11
	template <typename T, int N>
	class array
	{
	private:
		T elem[N];
	public:
		__device__ __HOST_V__ INLINE T & operator[](int index) { return elem[index]; }
		__device__ __HOST_V__ INLINE const T & operator[](int index) const { return elem[index]; }
	};

	// Binary and unary operators on uchar4, short8, ... are cast to vecTN types to allow generic handling
#define IMPL_VECTN2_OP(VecType, ElemType, op) \
	INLINE __HOST_V__ __device__ VecType##1 operator op (VecType##1 a, VecType##1 b) { return (vecTN<ElemType, 1>)a op (vecTN<ElemType, 1>)b; } \
	INLINE __HOST_V__ __device__ VecType##2 operator op (VecType##2 a, VecType##2 b) { return (vecTN<ElemType, 2>)a op (vecTN<ElemType, 2>)b; } \
	INLINE __HOST_V__ __device__ VecType##3 operator op (VecType##3 a, VecType##3 b) { return (vecTN<ElemType, 3>)a op (vecTN<ElemType, 3>)b; } \
	INLINE __HOST_V__ __device__ VecType##4 operator op (VecType##4 a, VecType##4 b) { return (vecTN<ElemType, 4>)a op (vecTN<ElemType, 4>)b; } \
	template <int N>  \
	INLINE __HOST_V__ __device__ VecType##N<N> operator op(VecType##N<N> a, VecType##N<N> b) { return (vecTN<ElemType, N>)a op (vecTN<ElemType, N>)b; }

#define IMPL_VECTN2_UOP(VecType, ElemType, op) \
	INLINE __HOST_V__ __device__ VecType##1 operator op (VecType##1 a) { return op (vecTN<ElemType, 1>)a; } \
	INLINE __HOST_V__ __device__ VecType##2 operator op (VecType##2 a) { return op (vecTN<ElemType, 2>)a; } \
	INLINE __HOST_V__ __device__ VecType##3 operator op (VecType##3 a) { return op (vecTN<ElemType, 3>)a; } \
	INLINE __HOST_V__ __device__ VecType##4 operator op (VecType##4 a) { return op (vecTN<ElemType, 4>)a; } \
	template <int N>  \
	INLINE __HOST_V__ __device__ VecType##N<N> operator op(VecType##N<N> a) { return op (vecTN<ElemType, N>)a; }

#define IMPL_VECTN2_IOP(VecType, ElemType, op) \
	INLINE __HOST_V__ __device__ VecType##1& operator op##= (VecType##1& a, VecType##1 b) { return a = (vecTN<ElemType, 1>)a op (vecTN<ElemType, 1>)b; } \
	INLINE __HOST_V__ __device__ VecType##2& operator op##= (VecType##2& a, VecType##2 b) { return a = (vecTN<ElemType, 2>)a op (vecTN<ElemType, 2>)b; } \
	INLINE __HOST_V__ __device__ VecType##3& operator op##= (VecType##3& a, VecType##3 b) { return a = (vecTN<ElemType, 3>)a op (vecTN<ElemType, 3>)b; } \
	INLINE __HOST_V__ __device__ VecType##4& operator op##= (VecType##4& a, VecType##4 b) { return a = (vecTN<ElemType, 4>)a op (vecTN<ElemType, 4>)b; } \
	template <int N>  \
	INLINE __HOST_V__ __device__ VecType##N<N>& operator op##= (VecType##N<N>& a, VecType##N<N> b) { return a = (vecTN<ElemType, N>)a op (vecTN<ElemType, N>)b; }

#define IMPL_VECTN2_OPS(VecType, ElemType)		\
	IMPL_VECTN2_OP(VecType, ElemType, +)		\
	IMPL_VECTN2_OP(VecType, ElemType, -)		\
	IMPL_VECTN2_OP(VecType, ElemType, *)		\
	IMPL_VECTN2_OP(VecType, ElemType, / )		\
	IMPL_VECTN2_IOP(VecType, ElemType, + )		\
	IMPL_VECTN2_IOP(VecType, ElemType, - )		\
	IMPL_VECTN2_IOP(VecType, ElemType, * )		\
	IMPL_VECTN2_IOP(VecType, ElemType, / )		\
	IMPL_VECTN2_OP(VecType, ElemType, == )		\
	IMPL_VECTN2_OP(VecType, ElemType, != )		\
	IMPL_VECTN2_OP(VecType, ElemType, < )		\
	IMPL_VECTN2_OP(VecType, ElemType, > )		\
	IMPL_VECTN2_OP(VecType, ElemType, <= )		\
	IMPL_VECTN2_OP(VecType, ElemType, >= )		\
	IMPL_VECTN2_UOP(VecType, ElemType, ! )

	IMPL_VECTN2_OPS(char, char);
	IMPL_VECTN2_OPS(uchar, uint8_t);
	IMPL_VECTN2_OPS(short, int16_t);
	IMPL_VECTN2_OPS(ushort, uint16_t);
	IMPL_VECTN2_OPS(uint, uint32_t);

	IMPL_VECTN2_UOP(char, char, -)
	IMPL_VECTN2_UOP(short, int16_t, -)

#undef IMPL_VECTN2_OP
#undef IMPL_VECTN2_IOP
#undef IMPL_VECTN2_OPS
#undef IMPL_VECTN2_UOP

} // namespace quasar


// vecTN<T, N> mathematical functions
namespace
{
	using quasar::vecTN;
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> abs(vecTN<T, N> a) { vecTN<T, N> c; FOR_ALL(c[i] = abs(a[i])); return c; }
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> max(vecTN<T, N> a, vecTN<T, N> b) { vecTN<T, N> c; FOR_ALL(c[i] = max(a[i], b[i])); return c; }
	template <typename T, int N> INLINE __HOST_V__ __device__ vecTN<T, N> min(vecTN<T, N> a, vecTN<T, N> b) { vecTN<T, N> c; FOR_ALL(c[i] = min(a[i], b[i])); return c; }
}

// helper function for boundary handling
INLINE __device__ __HOST_V__ int mirror_ext(int n, int N)
{
    n = abs(n) % (2 * N);
    if (n >= N)
        n = 2 * N - 1 - n;
    return n;
}
INLINE __device__ __HOST_V__ int periodize(int n, int N)
{
    int q = n % N;
    return q < 0 ? q + N : q;
}
INLINE __device__ __HOST_V__ int clamp(int n, int N) { return max(0, min(N, n)); }
INLINE __device__ __HOST_V__ scalar mirror_ext(scalar n, scalar N)
{
    n = fmod(abs(n), 2 * N);
    if (n >= N)
        n = 2 * N - 1 - n;
    return n;
}
INLINE __device__ __HOST_V__ scalar periodize(scalar n, scalar N)
{
    scalar q = fmod(n, N);
    return q < 0 ? q + N : q;
}
INLINE __device__ __HOST_V__ scalar clamp(scalar n, scalar N)
{
    return max((scalar)0, min((scalar)N, n));
}

//=============================================================================
//		ERRORS & ERROR HANDLING
//=============================================================================
#define ERRORCODE_OUTOFBOUNDS 1
#define ERRORCODE_OVERFLOW 2
#define ERRORCODE_USERERROR 3
#define ERRORCODE_NAN_OR_INF 4
#define ERRORCODE_NAN 5
#define ERRORCODE_ASSERTIONFAILED 6
#define ERRORCODE_DYNMEM_BLOCKSIZE_TOOLARGE 7
#define ERRORCODE_DYNMEM_OUTOFMEMORY 8
#define ERRORCODE_DYNMEM_INVALIDSTATE 9

struct _ERRORARGS
{
	const void *target;
	int target_bytes;
	int code;
};

#if defined(__CUDA_ARCH__)
__device__ struct _ERRORARGS _error;
#else
extern "C" {
	EXPORT SELECTANY struct _ERRORARGS _error = {};
}
#endif


//=============================================================================
// Minimum and maximum values of primitive data types
//=============================================================================

#define MINVAL_uint8 0
#define MAXVAL_uint8 UCHAR_MAX
#define MINVAL_uint16 0
#define MAXVAL_uint16 USHRT_MAX
#define MINVAL_uint32 0
#define MAXVAL_uint32 UINT_MAX
#define MINVAL_uint64 0
#define MAXVAL_uint64 ULLONG_MAX
#define MINVAL_char SCHAR_MIN
#define MAXVAL_char SCHAR_MAX
#define MINVAL_uchar 0
#define MAXVAL_uchar UCHAR_MAX
#define MINVAL_int8 SCHAR_MIN
#define MAXVAL_int8 SCHAR_MAX
#define MINVAL_int16 SHRT_MIN
#define MAXVAL_int16 SHRT_MAX
#define MINVAL_int32 INT_MIN
#define MAXVAL_int32 INT_MAX
#define MINVAL_int64 LLONG_MIN
#define MAXVAL_int64 LLONG_MAX
#define MINVAL_int   INT_MIN
#define MAXVAL_int   INT_MAX

#define MINVAL_float -FLT_MAX
#define MAXVAL_float FLT_MAX
#define MINVAL_double -DBL_MAX
#define MAXVAL_double DBL_MAX
#ifdef DBL_SCALAR
#define MINVAL_scalar MINVAL_double
#define MAXVAL_scalar MAXVAL_double
#else
#define MINVAL_scalar MINVAL_float
#define MAXVAL_scalar MAXVAL_float
#endif
#define MINVAL___half -65504.0
#define MAXVAL___half 65504.0

#define MINVAL_cscalar make_cscalar(MINVAL_scalar,MINVAL_scalar)
#define MAXVAL_cscalar make_cscalar(MAXVAL_scalar,MAXVAL_scalar)

//=============================================================================
// INTEGER CONVERSION FUNCTIONS
//=============================================================================
namespace quasar
{
	enum IntegerConversionModes
	{
		ConvUnchecked,	// Unchecked (default)
		ConvChecked,	// Checks the integer range during conversion
		ConvSaturated	// Performs saturation when the minimum/maximum is exceeded
	};

	// The main gateway conversion function
	template <typename R, IntegerConversionModes mode, typename T>
	INLINE R __device__ __HOST_V__ autocast_int(T val)
	{
		return autocast_int(val, mode, ReturnType<R>());
	}

	// The function doing all of the work
	template <typename T, typename R>
	INLINE R __device__  __HOST_V__ autocast_int(T val, R minval, R maxval, IntegerConversionModes mode)
	{
		R ret;
		switch (mode)
		{
		case ConvUnchecked:
			ret = (R)val;
			break;
		case ConvSaturated:
			if (val < (T)minval)
				ret = minval;
			else if (val >(T)maxval)
				ret = maxval;
			else
				ret = (R)val;
			break;
		default:
			if (val < (T)minval || val >(T)maxval)
			{
				_error.target = NULL;
				_error.target_bytes = sizeof(void *);
				_error.code = ERRORCODE_OVERFLOW;
				val = (T)0; // set to zero
				assert(0);
			}
			ret = (R)val;
		}
		return ret;
	}

	// vecTN implementation
	template <int N, typename T, typename R>
	INLINE R __device__ __HOST_V__ autocast_int(vecTN<T, N> value, IntegerConversionModes mode, ReturnType<R>)
	{
		return autocast_int<R>((typename VecType<T, N>::type)value, mode, ReturnType<R>());
	}
}

//=============================================================================
// SIMD vector processing functionality
//=============================================================================
#include "quasar_simd.h"

//=============================================================================
// FUNCTION DEFINITIONS OF VECTOR TYPES
// Note: the parentheses around X are added to prevent macro expansion!
//=============================================================================
#define UNARY_FUNCTION_T(X, T)                                                 \
    INLINE T##1 __device__ __HOST_V__(X)(T##1 a) { return make_##T##1(X(a.x)); }          \
    INLINE T##2 __device__ __HOST_V__(X)(T##2 a) { return make_##T##2(X(a.x), X(a.y)); }  \
    INLINE T##3 __device__ __HOST_V__(X)(T##3 a)                                          \
    {                                                                          \
        return make_##T##3(X(a.x), X(a.y), X(a.z));                            \
    }                                                                          \
    INLINE T##4 __device__ __HOST_V__(X)(T##4 a)                                 \
    {                                                                          \
        return make_##T##4(X(a.x), X(a.y), X(a.z), X(a.w));                   \
    }                                                                          \
    template <int N> INLINE T##N<N> __device__ __HOST_V__(X)(T##N<N> a)        \
    {                                                                          \
        T##N<N> y;                                                             \
        FOR_ALL(y.el[i] = X(a.el[i]));                                         \
        return y;                                                              \
    }
#define UNARY_FUNCTION_TR(X, T, R)                                             \
    INLINE R##1 __device__ __HOST_V__(X)(T##1 a) { return make_##R##1(X(a.x)); }          \
    INLINE R##2 __device__ __HOST_V__(X)(T##2 a) { return make_##R##2(X(a.x), X(a.y)); }  \
    INLINE R##3 __device__ __HOST_V__(X)(T##3 a)                                          \
    {                                                                          \
        return make_##R##3(X(a.x), X(a.y), X(a.z));                            \
    }                                                                          \
    INLINE R##4 __device__ __HOST_V__(X)(T##4 a)                                 \
    {                                                                          \
        return make_##R##4(X(a.x), X(a.y), X(a.z), X(a.w));                    \
    }                                                                          \
    template <int N> INLINE R##N<N> __device__ __HOST_V__(X)(T##N<N> a)         \
    {                                                                          \
        R##N<N> y;                                                             \
        FOR_ALL(y.el[i] = X(a.el[i]));                                         \
        return y;                                                              \
    }
#define BINARY_FUNCTION_T(X, T)                                                \
    INLINE T##1 __device__ __HOST_V__(X)(T##1 a, T##1 b)                         \
    {                                                                          \
        return make_##T##1(X(a.x, b.x));                                       \
    }                                                                          \
    INLINE T##2 __device__ __HOST_V__(X)(T##2 a, T##2 b)                         \
    {                                                                          \
        return make_##T##2(X(a.x, b.x), X(a.y, b.y));                          \
    }                                                                          \
    INLINE T##3 __device__ __HOST_V__(X)(T##3 a, T##3 b)                         \
    {                                                                          \
        return make_##T##3(X(a.x, b.x), X(a.y, b.y), X(a.z, b.z));             \
    }                                                                          \
    INLINE T##4 __device__ __HOST_V__(X)(T##4 a, T##4 b)                         \
    {                                                                          \
        return make_##T##4(X(a.x, b.x), X(a.y, b.y), X(a.z, b.z),              \
                           X(a.w, b.w));                                       \
    }                                                                          \
    template <int N> INLINE T##N<N> __device__ __HOST_V__(X)(T##N<N> a, T##N<N> b) \
    {                                                                          \
        T##N<N> y;                                                             \
        FOR_ALL(y.el[i] = X(a.el[i], b.el[i]));                                \
        return y;                                                              \
    }
#define BINARY_FUNCTION_TR(X, T, A, B)                                         \
    INLINE T##1 __device__ __HOST_V__(X)(A##1 a, B##1 b)                         \
    {                                                                          \
        return make_##T##1(X(a.x, b.x));                                       \
    }                                                                          \
    INLINE T##2 __device__ __HOST_V__(X)(A##2 a, B##2 b)                         \
    {                                                                          \
        return make_##T##2(X(a.x, b.x), X(a.y, b.y));                          \
    }                                                                          \
    INLINE T##3 __device__ __HOST_V__(X)(A##3 a, B##3 b)                         \
    {                                                                          \
        return make_##T##3(X(a.x, b.x), X(a.y, b.y), X(a.z, b.z));             \
    }                                                                          \
    INLINE T##4 __device__ __HOST_V__(X)(A##4 a, B##4 b)                         \
    {                                                                          \
        return make_##T##4(X(a.x, b.x), X(a.y, b.y), X(a.z, b.z),             \
                           X(a.w, b.w));                                       \
    }                                                                          \
    template <int N> INLINE T##N<N> __device__ __HOST_V__(X)(A##N<N> a, B##N<N> b) \
    {                                                                          \
        T##N<N> y;                                                             \
        FOR_ALL(y.el[i] = X(a.el[i], b.el[i]));                                \
        return y;                                                             \
    }
#define BINARY_FUNCTION_TR2(X, R, T)                                           \
    INLINE R##1 __device__ __HOST_V__(X)(T a, T##1 b) { return make_##R##1(X(a, b.x)); }  \
    INLINE R##2 __device__ __HOST_V__(X)(T a, T##2 b)                            \
    {                                                                          \
        return make_##R##2(X(a, b.x), X(a, b.y));                              \
    }                                                                          \
    INLINE R##3 __device__ __HOST_V__(X)(T a, T##3 b)                            \
    {                                                                          \
        return make_##R##3(X(a, b.x), X(a, b.y), X(a, b.z));                  \
    }                                                                          \
    INLINE R##4 __device__ __HOST_V__(X)(T a, T##4 b)                            \
    {                                                                          \
        return make_##R##4(X(a, b.x), X(a, b.y), X(a, b.z), X(a, b.w));       \
    }                                                                          \
    INLINE R##1 __device__ __HOST_V__(X)(T##1 a, T b) { return make_##R##1(X(a.x, b)); }  \
    INLINE R##2 __device__ __HOST_V__(X)(T##2 a, T b)                            \
    {                                                                          \
        return make_##R##2(X(a.x, b), X(a.y, b));                              \
    }                                                                          \
    INLINE R##3 __device__ __HOST_V__(X)(T##3 a, T b)                            \
    {                                                                          \
        return make_##R##3(X(a.x, b), X(a.y, b), X(a.z, b));                   \
    }                                                                          \
    INLINE R##4 __device__ __HOST_V__(X)(T##4 a, T b)                            \
    {                                                                          \
        return make_##R##4(X(a.x, b), X(a.y, b), X(a.z, b), X(a.w, b));        \
    }                                                                          \
    template <int N> INLINE R##N<N> __device__ __HOST_V__(X)(T a, T##N<N> b)    \
    {                                                                          \
        R##N<N> y;                                                             \
        FOR_ALL(y.el[i] = X(a, b.el[i]));                                      \
        return y;                                                              \
    }                                                                          \
    template <int N> INLINE R##N<N> __device__ __HOST_V__(X)(T##N<N> a, T b)    \
    {                                                                          \
        R##N<N> y;                                                             \
        FOR_ALL(y.el[i] = X(a.el[i], b));                                      \
        return y;                                                              \
    }


// Integer/scalar functions
UNARY_FUNCTION_T(acos, scalar);
UNARY_FUNCTION_T(atan, scalar);
BINARY_FUNCTION_T(atan2, scalar);
UNARY_FUNCTION_T(conj, scalar);
UNARY_FUNCTION_T(cos, scalar);
UNARY_FUNCTION_T(sin, scalar);
UNARY_FUNCTION_T(exp, scalar);
UNARY_FUNCTION_T(exp2, scalar);
BINARY_FUNCTION_T(mod, scalar);
UNARY_FUNCTION_T(log, scalar);
UNARY_FUNCTION_T(log2, scalar);
UNARY_FUNCTION_T(log10, scalar);
BINARY_FUNCTION_T(pow, scalar);
BINARY_FUNCTION_TR2(pow, scalar, scalar);
UNARY_FUNCTION_T(saturate, scalar);
UNARY_FUNCTION_T(sinh, scalar);
UNARY_FUNCTION_T(cosh, scalar);
UNARY_FUNCTION_T(tanh, scalar);
#ifdef _MSC_VER
UNARY_FUNCTION_T(asinh, scalar);
UNARY_FUNCTION_T(acosh, scalar);
UNARY_FUNCTION_T(atanh, scalar);
#endif
UNARY_FUNCTION_T(erf, scalar);
UNARY_FUNCTION_T(erfc, scalar);
UNARY_FUNCTION_T(erfinv, scalar);
UNARY_FUNCTION_T(erfcinv, scalar);
UNARY_FUNCTION_T(lgamma, scalar);
UNARY_FUNCTION_T(tgamma, scalar);
UNARY_FUNCTION_TR(_isnan_, scalar, int);
UNARY_FUNCTION_TR(_isinf_, scalar, int);
UNARY_FUNCTION_TR(_isfinite_, scalar, int);
BINARY_FUNCTION_T(periodize, scalar);
BINARY_FUNCTION_T(mirror_ext, scalar);
BINARY_FUNCTION_T(clamp, scalar);
BINARY_FUNCTION_T(logand, scalar);
BINARY_FUNCTION_T(logor, scalar);
BINARY_FUNCTION_T(logxor, scalar);
UNARY_FUNCTION_T(lognot, scalar);

// Integer functions
BINARY_FUNCTION_T(mod, int);
BINARY_FUNCTION_T(_shl_, int);
BINARY_FUNCTION_T(_shr_, int);
BINARY_FUNCTION_T(periodize, int);
BINARY_FUNCTION_T(mirror_ext, int);
BINARY_FUNCTION_T(clamp, int);
UNARY_FUNCTION_T(_isnan_, int);
UNARY_FUNCTION_T(_isinf_, int);
UNARY_FUNCTION_T(_isfinite_, int);
BINARY_FUNCTION_TR(sdiv, scalar, int, int);
BINARY_FUNCTION_TR2(sdiv, scalar, int);
BINARY_FUNCTION_T(idiv, int);
BINARY_FUNCTION_TR2(idiv, int, int);
BINARY_FUNCTION_TR2(pow, int, int);


// Complex functions
UNARY_FUNCTION_TR(abs, cscalar, scalar);
UNARY_FUNCTION_TR(real, cscalar, scalar);
UNARY_FUNCTION_TR(imag, cscalar, scalar);
UNARY_FUNCTION_TR(angle, cscalar, scalar);
UNARY_FUNCTION_T(conj, cscalar);
BINARY_FUNCTION_T(pow, cscalar);
BINARY_FUNCTION_TR(pow, cscalar, cscalar, scalar);
BINARY_FUNCTION_TR(pow, cscalar, scalar, cscalar);
UNARY_FUNCTION_T(log, cscalar);
UNARY_FUNCTION_T(log10, cscalar);
UNARY_FUNCTION_T(log2, cscalar);
UNARY_FUNCTION_T(sin, cscalar);
UNARY_FUNCTION_T(cos, cscalar);
UNARY_FUNCTION_T(tan, cscalar);
UNARY_FUNCTION_T(exp, cscalar);
UNARY_FUNCTION_TR(sign, cscalar, scalar);
UNARY_FUNCTION_T(sqrt, cscalar);
UNARY_FUNCTION_TR(_isnan_, cscalar, int);
UNARY_FUNCTION_TR(_isinf_, cscalar, int);
UNARY_FUNCTION_TR(_isfinite_, cscalar, int);


// Misc functions
#define FUNCTION_LERP_DEF(T)                                                   \
    INLINE T __device__  __HOST_V__ lerp(T a, T b, T d) { return a * (1 - d) + b * d; }    \
    INLINE T##1 __device__  __HOST_V__ lerp(T##1 a, T##1 b, T d)                 \
    {                                                                          \
        return a * (1 - d) + b * d;                                            \
    }                                                                          \
    INLINE T##2 __device__  __HOST_V__ lerp(T##2 a, T##2 b, T d)                 \
    {                                                                          \
        return a * (1 - d) + b * d;                                            \
    }                                                                          \
    INLINE T##3 __device__  __HOST_V__ lerp(T##3 a, T##3 b, T d)                 \
    {                                                                          \
        return a * (1 - d) + b * d;                                            \
    }                                                                          \
    INLINE T##4 __device__  __HOST_V__ lerp(T##4 a, T##4 b, T d)                 \
    {                                                                          \
        return a * (1 - d) + b * d;                                            \
    }                                                                          \
    template <int N> T __device__  __HOST_V__ lerp(T##N<N> a, T##N<N> b, T d)   \
    {                                                                          \
        return a * (1 - d) + b * d;                                            \
    }

FUNCTION_LERP_DEF(scalar);
FUNCTION_LERP_DEF(cscalar);

#undef UNARY_FUNCTION_T
#undef UNARY_FUNCTION_TR
#undef BINARY_FUNCTION_T
#undef BINARY_FUNCTION_TR
#undef BINARY_FUNCTION_TR2
#undef FUNCTION_DOTPROD_DEF
#undef FUNCTION_LERP_DEF

namespace quasar
{
    INLINE __HOST_V__ __device__ void raise_error(const char *msg, int len)
    {
        _error.target = msg;
        _error.target_bytes = len;
        _error.code = ERRORCODE_USERERROR;
#if defined(_MSC_VER) && !defined(TARGET_CUDA)
        throw msg; // Exception will be captured by a surrounding try/catch
                   // block
#endif
    }

#define Q_RAISE_ERROR(msg) raise_error(msg, sizeof(msg))
}

//=============================================================================
// OTHER CONVERSION FUNCTIONS
//=============================================================================

// Additional conversion functions used for hardware texturing
#ifdef TARGET_CUDA
INLINE __device__ __HOST_V__ uint16_t half_to_ushort(__half a)
{
	return (uint16_t&)a;
}
INLINE __device__ __HOST_V__ ushort1 half_to_ushort(quasar::vecTN<__half, 1> a)
{
	return make_ushort1((uint16_t&)a[0]);
}
INLINE __device__ __HOST_V__ ushort2 half_to_ushort(quasar::vecTN<__half, 2> a)
{
	return make_ushort2((uint16_t&)a[0], (uint16_t&)a[1]);
}
INLINE __device__ __HOST_V__ ushort3 half_to_ushort(quasar::vecTN<__half, 3> a)
{
	return make_ushort3((uint16_t&)a[0], (uint16_t&)a[1], (uint16_t&)a[2]);
}
INLINE __device__ __HOST_V__ ushort4 half_to_ushort(quasar::vecTN<__half, 4> a)
{
	return make_ushort4((uint16_t&)a[0], (uint16_t&)a[1], (uint16_t&)a[2], (uint16_t&)a[3]);
}
template <int N>
INLINE __device__ __HOST_V__ ushortN<N> half_to_ushort(quasar::vecTN<__half, N> a)
{
	ushortN<N> y;
	FOR_ALL(y.el[i] = (uint16_t&)a[i]);
	return y;
}
#endif


#undef DECL_CVT
#undef DECL_CVTN

//=============================================================================
//		MATRIX ACCESSORS
//=============================================================================
namespace quasar
{

// Debugging - checking of the unsafe boundaries!

// Boundary checking code ('checked modifiers)
#define CHECK_BOUNDS(x, cond, defval)                                          \
    if (!(cond) && _error.target == NULL)                                      \
    {                                                                          \
        _error.target = (void *)(x).data;                                      \
        _error.target_bytes = sizeof(void *);                                  \
        _error.code = ERRORCODE_OUTOFBOUNDS;                                   \
        assert(0);                                                             \
        return defval;                                                         \
    }
#define V_CHECK_BOUNDS(x, cond, defval)                                        \
    if (!(cond) && _error.target == NULL)                                      \
    {                                                                          \
        _error.target = (void *)&x;                                            \
        _error.target_bytes = sizeof(void *);                                  \
        _error.code = ERRORCODE_OUTOFBOUNDS;                                   \
        assert(0);                                                             \
        return defval;                                                         \
    }

#if defined(CHECK_UNSAFE_BOUNDS)
#define CHECK_BOUNDS_FAST(x, cond, defval)									   \
	if (!(cond)) return defval;
#else
#define CHECK_BOUNDS_FAST(x, cond, defval)
#endif

#if defined(CHECK_NAN_OR_INF)
#undef CHECK_NAN_OR_INF
#define CHECK_NAN_OR_INF(tgt, x)                                               \
    if (_error.target == NULL && !_all_(_isfinite_(x)))                        \
    {                                                                          \
        _error.target = (tgt).data;                                            \
        _error.target_bytes = sizeof(void *);                                  \
        _error.code = ERRORCODE_NAN_OR_INF;                                    \
        assert(0);                                                             \
    }
#elif defined(CHECK_NAN)
#define CHECK_NAN_OR_INF(tgt, x)                                               \
    if (_error.target == NULL && _any_(_isnan_(x)))                            \
    {                                                                          \
        _error.target = (tgt).data;                                            \
        _error.target_bytes = sizeof(void *);                                  \
        _error.code = ERRORCODE_NAN;                                           \
        assert(0);                                                             \
    }
#else
#define CHECK_NAN_OR_INF(tgt, x)
#endif

	// Support for the CUDA __ldg intrinsic
	//
	template <typename T>
	INLINE __device__ T ldg(const T* a) { return *a; }

	// ldg overloads
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 350
	template <> INLINE __device__ scalar ldg(const scalar *a) { return __ldg(a); }
	template <> INLINE __device__ int ldg(const int *a) { return __ldg(a); }
	template <> INLINE __device__ cscalar ldg(const cscalar *a) { return make_cscalar(ldg(&a->x), ldg(&a->y)); }
	template <> INLINE __device__ float1 ldg(const float1 *a) { return make_float1(ldg(&a->x)); }
	template <> INLINE __device__ float2 ldg(const float2 *a) { return __ldg(a); }
	template <> INLINE __device__ float3 ldg(const float3 *a) { return make_float3(ldg(&a->x), ldg(&a->y), ldg(&a->z)); }
	template <> INLINE __device__ float4 ldg(const float4 *a) { return __ldg(a); }
	template <> INLINE __device__ double1 ldg(const double1 *a) { return make_double1(ldg(&a->x)); }
	template <> INLINE __device__ double2 ldg(const double2 *a) { return __ldg(a); }
	template <> INLINE __device__ double3 ldg(const double3 *a) { return make_double3(ldg(&a->x), ldg(&a->y), ldg(&a->z)); }
	template <> INLINE __device__ double4 ldg(const double4 *a) { return make_double4(ldg(&a->x), ldg(&a->y), ldg(&a->z), ldg(&a->w)); }
	template <int N> INLINE __device__ scalarN<N> ldg(const scalarN<N> *a) {
		scalarN<N> y; FOR_ALL(y.el[i] = ldg(&a->el[i])); return y; 
    }
	template <> INLINE __device__ int1 ldg(const int1 *a) { return make_int1(ldg(&a->x)); }
	template <> INLINE __device__ int2 ldg(const int2 *a) { return __ldg(a); }
	template <> INLINE __device__ int3 ldg(const int3 *a) { return make_int3(ldg(&a->x), ldg(&a->y), ldg(&a->z)); }
	template <> INLINE __device__ int4 ldg(const int4 *a) { return __ldg(a); }
	template <int N> INLINE __device__ intN<N> ldg(const intN<N> *a) {
		intN<N> y; FOR_ALL(y.el[i] = ldg(&a->el[i])); return y; }
	template <typename T, int N> INLINE __device__ array<T, N> ldg(const array<T, N> *a) {
		array<T, N> y; FOR_ALL(y[i] = ldg(&(*a)[i])); return y;
	}
#endif
	#define LOAD_CONST(x)		quasar::ldg(x)

    // Modes for handling the array/matrix/cube boundaries
	enum BoundaryAccessMode
	{
		Default = 0,   // Raise an exception (checked)
		Unchecked = 1, // Fastest option - no checking
		Zero = 2,      // Boundary extension with zeros
		Circular = 3,  // Circular extension (note: only supported in host interface)
		Mirror = 4,    // Mirrored extension (note: only supported in host interface)
		Clamp = 5,     // Clamps to the border value (note: only supported in host interface)
		Const = 8,	   // Read from L2 cache (note: only supported by some functions)
		ConstDefault = Const | Default,		// Read from L2 cache with exception checking
		ConstUnchecked = Const | Unchecked, // Read from L2 cache without checks
		ConstZero = Const | Zero			// Read from L2 cache with zero extension
    };

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ lref<T> vector_get_ref_at(const VectorBase<T>& x, int pos)
	{
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, pos >= 0 && pos < x.dim1, lref<T>())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, pos >= 0 && pos < x.dim1, lref<T>())
		}
		else if (pos < 0 || pos >= x.dim1)
		{
			return lref<T>();
		}
		return x.data[pos];
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T vector_get_at(const VectorBase<T>& x, int pos)
	{
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, pos >= 0 && pos < x.dim1, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, pos >= 0 && pos < x.dim1, T())
		}
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, pos >= 0 && pos < x.dim1, T())
			return LOAD_CONST(&((const T *)x.data)[pos]);
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, pos >= 0 && pos < x.dim1, T())
			return LOAD_CONST(&((const T *)x.data)[pos]);
		}
		else if CONSTEXPR (mode == ConstZero)
		{
			if (pos < 0 || pos >= x.dim1)
			{
				return T();
			}
			return LOAD_CONST(&((const T *)x.data)[pos]);
		}
		else if (pos < 0 || pos >= x.dim1)
		{
			return T();
		}
		return x.data[pos];
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T vector_set_at(const VectorBase<T>& x, int pos, T val)
	{
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, pos >= 0 && pos < x.dim1, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, pos >= 0 && pos < x.dim1, T())
		}
		else if (pos < 0 || pos >= x.dim1)
		{
			return T();
		}
		CHECK_NAN_OR_INF(x, val);
		return x.data[pos] = val;
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ lref<T> vector_get_ref_at(const VectorBase<T>& x, int1 pos)
	{
		return vector_get_ref_at<mode>(x, pos.x);
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T vector_get_at(const VectorBase<T>& x, int1 pos)
	{
		return vector_get_at<mode>(x, pos.x);
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T vector_set_at(const VectorBase<T>& x, int1 pos, T val)
	{
		return vector_set_at<mode>(x, pos.x, val);
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ lref<T> matrix_get_ref_at(const MatrixBase<T>& x, int2 pos)
	{
		int index = pos.y + pos.x * x.dim2;
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2, lref<T>())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2, lref<T>())
		}
		else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2))
		{
			return lref<T>();
		}
		return x.data[index];
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T matrix_get_at(const MatrixBase<T>& x, int2 pos)
	{
		int index = pos.y + pos.x * x.dim2;
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2, T())
		}
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2, T())
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2, T())
				return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if CONSTEXPR (mode == ConstZero)
		{
			if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2))
			{
				return T();
			}
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2))
		{
			return T();
		}
		return x.data[index];
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T matrix_set_at(const MatrixBase<T>& x, int2 pos, T val)
	{
		int index = pos.y + pos.x * x.dim2;
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2, T())
		}
		else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2))
		{
			return T();
		}
		CHECK_NAN_OR_INF(x, val);
		return x.data[index] = val;
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ lref<T> cube_get_ref_at(const CubeBase<T>& x, int3 pos)
	{
		int index = (pos.y + pos.x * x.dim2) * x.dim3 + pos.z;
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3, lref<T>())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 * x.dim3, lref<T>())
		}
		else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3))
		{
			return lref<T>();
		}
		return x.data[index];
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T cube_get_at(const CubeBase<T>& x, int3 pos)
	{
		int index = (pos.y + pos.x * x.dim2) * x.dim3 + pos.z;
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 * x.dim3, T())
		}
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 * x.dim3, T())
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3, T())
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if CONSTEXPR (mode == ConstZero)
		{
			if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3))
			{
				return T();
			}
			return LOAD_CONST(&((const T *)x.data)[index]);
		}
		else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3))
		{
			return T();
		}
		return x.data[index];
	}

	template <BoundaryAccessMode mode, typename T>
	INLINE __device__ __HOST_V__ T cube_set_at(const CubeBase<T>& x, int3 pos, T val)
	{
		int index = (pos.y + pos.x * x.dim2) * x.dim3 + pos.z;
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3, T())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 * x.dim3, T())
		}
		else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3))
		{
			return T();
		}
		CHECK_NAN_OR_INF(x, val);
		return x.data[index] = val;
	}

	//========================================================================
	// Vector slice getters/setters
	//========================================================================

	template <BoundaryAccessMode mode, typename T, int N, bool DEFAULT_TEMPLATEVAL(aligned, true)>
    INLINE __device__ __HOST_V__ typename VecType<T, N>::type vector_get_vec_at(const VectorBase<T> &x, int pos)
    {
		typedef typename VecType<T, N>::type R;	// Vector type
		typedef array<T, N> R_unaligned;	// Array type
		R retVal;
		if CONSTEXPR (mode == Default)
		{
			CHECK_BOUNDS(x, pos >= 0 && pos <= x.dim1 - N, R())
		}
		else if CONSTEXPR (mode == Unchecked)
		{
			CHECK_BOUNDS_FAST(x, pos >= 0 && pos <= x.dim1 - N, R())
		}
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, pos >= 0 && pos < x.dim1 - N, R())
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, pos >= 0 && pos < x.dim1 - N, R())
		}
		else if (pos < 0 || pos > x.dim1 - N)
			return R();

		if CONSTEXPR (mode < Const)
		{
			if (aligned)
				retVal = (R &)x.data[pos];
			else
				(R_unaligned &)retVal = *((R_unaligned *)&x.data[pos]);
		}
		else
		{
			if (aligned)
				retVal = LOAD_CONST(((R *)&x.data[pos]));
			else
				(R_unaligned &)retVal = LOAD_CONST(((R_unaligned *)&x.data[pos]));
		}
		return retVal;
    }

    template <BoundaryAccessMode mode, typename T, int N, bool DEFAULT_TEMPLATEVAL(aligned, true)>
    INLINE __device__ __HOST_V__ typename VecType<T, N>::type
    vector_set_vec_at(const VectorBase<T> &x, int pos, typename VecType<T, N>::type val)
    {
		typedef typename VecType<T, N>::type R;	// Vector type
		typedef array<T, N> R_unaligned;	// Array type

        if CONSTEXPR (mode == Default)
        {
            CHECK_BOUNDS(x, pos >= 0 && pos <= x.dim1 - N, R())
        }
        else if CONSTEXPR (mode == Unchecked)
        {
            CHECK_BOUNDS_FAST(x, pos >= 0 && pos <= x.dim1 - N, R());
        }
        else if (pos < 0 || pos > x.dim1 - N)
            return R();

        CHECK_NAN_OR_INF(x, val);

		if CONSTEXPR (aligned)
			(R &)x.data[pos] = val;
		else
			(R_unaligned &)x.data[pos] = (R_unaligned &)val;
        return val;
    }

    template <BoundaryAccessMode mode, typename T, int N, bool DEFAULT_TEMPLATEVAL(aligned, true)>
    INLINE __device__ __HOST_V__ typename VecType<T, N>::type matrix_get_vec_at(const MatrixBase<T> &x, int2 pos)
    {
		typedef typename VecType<T, N>::type R;	// Vector type
		typedef array<T, N> R_unaligned;	// Array type
		R retVal;

		int index = pos.y + pos.x * x.dim2;
		if CONSTEXPR (mode == Default)
		{
            CHECK_BOUNDS(x, pos.x >= 0 && pos.x < x.dim1 && pos.y >= 0 &&
                            pos.y <= x.dim2 - N, R());
		}
		else if CONSTEXPR (mode == Unchecked)
		{
            CHECK_BOUNDS_FAST(x, index >= 0 && index <= x.dim1 * x.dim2 - N, R())
        }
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 - N, R())
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 - N, R())
		}
        else if (pos.x < 0 || pos.y < 0 || pos.x >= x.dim1 || pos.y > x.dim2 - N)
            return R();

		if CONSTEXPR (mode < Const)
		{
			if (aligned)
				retVal = (R &)x.data[index];
			else
				(R_unaligned &)retVal = *((R_unaligned *)&x.data[index]);
		}
		else
		{
			if (aligned)
				retVal = LOAD_CONST(((R *)&x.data[index]));
			else
				(R_unaligned &)retVal = LOAD_CONST(((R_unaligned *)&x.data[index]));
		}
		return retVal;
    }

    template <BoundaryAccessMode mode, typename T, int N, bool DEFAULT_TEMPLATEVAL(aligned, true)>
    INLINE __device__ __HOST_V__ typename VecType<T, N>::type matrix_set_vec_at(const MatrixBase<T> &x, int2 pos,
		typename VecType<T, N>::type val)
    {
		typedef typename VecType<T, N>::type R;	// Vector type
		typedef array<T, N> R_unaligned;	// Array type

        int index = pos.y + pos.x * x.dim2;
        if CONSTEXPR (mode == Default)
		{
            CHECK_BOUNDS(x, pos.x >= 0 && pos.x < x.dim1 && pos.y >= 0 &&
                            pos.y <= x.dim2 - N, R());
		}
		else if CONSTEXPR (mode == Unchecked)
		{
            CHECK_BOUNDS_FAST(x, index >= 0 && index <= x.dim1 * x.dim2 - N, R())
        }
        else if (pos.x < 0 || pos.y < 0 || pos.x >= x.dim1 || pos.y > x.dim2 - N)
            return R();

        CHECK_NAN_OR_INF(x, val);

		if CONSTEXPR (aligned)
			(R &)x.data[index] = val;
		else
			(R_unaligned &)x.data[index] = (R_unaligned &)val;
		return val;
    }

    template <BoundaryAccessMode mode, typename T, int N, bool DEFAULT_TEMPLATEVAL(aligned, true)>
    INLINE __device__ __HOST_V__ typename VecType<T, N>::type cube_get_vec_at(const CubeBase<T> &x, int3 pos)
    {
		typedef typename VecType<T, N>::type R;	// Vector type
		typedef array<T, N> R_unaligned;	// Array type
		R retVal;

		int index = pos.z + (pos.y + pos.x * x.dim2) * x.dim3;
		if CONSTEXPR (mode == Default)
		{
            CHECK_BOUNDS(x, pos.x >= 0 && pos.x < x.dim1 && pos.y >= 0 &&
                            pos.y < x.dim2 && pos.z >= 0 &&
                            pos.z <= x.dim3 - N, R());
		}
		else if CONSTEXPR (mode == Unchecked)
		{
            CHECK_BOUNDS_FAST(x, index >= 0 && index <= x.dim1 * x.dim2 * x.dim3 - N, R())
        }
		else if CONSTEXPR (mode == ConstUnchecked)
		{
			CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 * x.dim3 - N, R())
		}
		else if CONSTEXPR (mode == ConstDefault)
		{
			CHECK_BOUNDS(x, pos.x >= 0 && pos.x < x.dim1 && pos.y >= 0 &&
				pos.y < x.dim2 && pos.z >= 0 && pos.z <= x.dim3 - N, R())
		}
		else if CONSTEXPR (mode == ConstZero)
		{
			if (pos.x < 0 || pos.y < 0 || pos.x >= x.dim1 || pos.y >= x.dim2 ||
				pos.z < 0 || pos.z > x.dim3 - N)
				return R();
		}
        else if (pos.x < 0 || pos.y < 0 || pos.x >= x.dim1 || pos.y >= x.dim2 ||
                  pos.z < 0 || pos.z > x.dim3 - N)
            return R();

		if CONSTEXPR (mode < Const)
		{
			if (aligned)
				retVal = (R &)x.data[index];
			else
				(R_unaligned &)retVal = *((R_unaligned *)&x.data[index]);
		}
		else
		{
			if (aligned)
				retVal = LOAD_CONST(((R *)&x.data[index]));
			else
				(R_unaligned &)retVal = LOAD_CONST(((R_unaligned *)&x.data[index]));
		}
		return retVal;
    }

    template <BoundaryAccessMode mode, typename T, int N, bool DEFAULT_TEMPLATEVAL(aligned, true)>
    INLINE __device__ __HOST_V__ typename VecType<T, N>::type cube_set_vec_at(const CubeBase<T> &x, int3 pos,
		typename VecType<T, N>::type val)
    {
		typedef typename VecType<T, N>::type R;	// Vector type
		typedef array<T, N> R_unaligned;	// Array type

		int index = pos.z + (pos.y + pos.x * x.dim2) * x.dim3;
        if CONSTEXPR (mode == Default)
		{
            CHECK_BOUNDS(x, pos.x >= 0 && pos.x < x.dim1 && pos.y >= 0 &&
                            pos.y < x.dim2 && pos.z >= 0 && pos.z <= x.dim3 - N, R());
		}
		else if CONSTEXPR (mode == Unchecked)
		{
            CHECK_BOUNDS_FAST(x, index >= 0 && index <= x.dim1 * x.dim2 * x.dim3 - N, R())
        }
        else if (pos.x < 0 || pos.y < 0 || pos.x >= x.dim1 || pos.y >= x.dim2 ||
                  pos.z < 0 || pos.z > x.dim3 - N)
            return R();
        CHECK_NAN_OR_INF(x, val);

		if CONSTEXPR (aligned)
			(R &)x.data[index] = val;
		else
			(R_unaligned &)x.data[index] = (R_unaligned &)val;
		return val;
    }

    //========================================================================
	// Fixed length vectors
	//========================================================================

	template <typename T, int N, BoundaryAccessMode mode DEFAULT_TEMPLATE_ARG(Default)>
	INLINE __device__ __HOST_V__ T vecX_get_at(typename VecType<T, N>::type a, int pos)
	{
		return vector_get_at<mode>(VectorBase<T>((T *)&a, N), pos);
	}

	template <typename T, int N, BoundaryAccessMode mode DEFAULT_TEMPLATE_ARG(Default)>
	INLINE __device__ __HOST_V__ T vecX_set_at(typename VecType<T, N>::type &a, int pos, T value)
	{
		return vector_set_at<mode>(VectorBase<T>((T *)&a, N), pos, value);
	}

	template <typename T, int N, BoundaryAccessMode mode DEFAULT_TEMPLATE_ARG(Default)>
	INLINE __device__ __HOST_V__ lref<T> vecX_get_ref_at(typename VecType<T, N>::type &a, int pos)
	{
		return vector_get_ref_at<mode>(VectorBase<T>((T *)&a, N), pos);
	}

	// vecX slice getters/setters
	template <typename T, int N, typename I, int M, BoundaryAccessMode mode DEFAULT_TEMPLATE_ARG(Default)>
	INLINE __device__ __HOST_V__  typename VecType<T, M>::type vecX_set_slice(
		typename VecType<T, N>::type A,
		typename VecType<I, M>::type indices,
		typename VecType<T, M>::type value)
	{
		for (int i = 0; i < M; i++)
			vecX_set_at<T, N, mode>(A, vecX_get_at<I, M, Unchecked>(indices, i), vecX_get_at<T, M, Unchecked>(value, i));
		return value;
	}

	template <typename T, int N, typename I, int M, BoundaryAccessMode mode DEFAULT_TEMPLATE_ARG(Default)>
	INLINE __device__ __HOST_V__ T vecX_set_slice(
		typename VecType<T, N>::type A, typename VecType<I, M>::type indices, T value)
	{
		for (int i = 0; i < M; i++)
			vecX_set_at<T, N, mode>(A, vecX_get_at<I, M, Unchecked>(indices, i), value);
		return value;
	}

	template <typename T, int N, typename I, int M, BoundaryAccessMode mode DEFAULT_TEMPLATE_ARG(Default)>
	INLINE __device__ __HOST_V__ typename VecType<T, M>::type vecX_get_slice(
		typename VecType<T, N>::type A, typename VecType<I, M>::type indices)
	{
		typename VecType<T, M>::type Y;
		for (int i = 0; i < M; i++)
			vecX_set_at<T, M, Unchecked>(Y, i, vecX_get_at<T, N, mode>(A, vecX_get_at<I, M, Unchecked>(indices, i)));
		return Y;
	}
} // namespace quasar

//=============================================================================
// ATOMIC OPERATIONS
//=============================================================================

// Inplace variant
template <typename T>
INLINE __HOST_V__ __device__ T inplaceCAS(T *x, T refval, T newval)
{
	T oldval = *x; *x = oldval == refval ? newval : oldval; return oldval;
}
template <typename T>
INLINE __HOST_V__ __device__ T inplaceExch(T *x, T newval)
{
	T oldval = *x; *x = newval; return oldval;
}



#if defined(_MSC_VER)

#pragma intrinsic(_InterlockedCompareExchange, _InterlockedCompareExchange64, _InterlockedExchangeAdd, _InterlockedExchangeAdd64)

// WIN32 - _InterlockedCompareExchange64 and _InterlockedExchange64 seem not to be defined
#if defined(_WIN32) && !defined(__LP64__) && !defined(TARGET_CUDA)
INLINE __device__ long long _InterlockedCompareExchange64(long long *x,
	long long oldval, long long newval)
{
	__int64 iReturn;
	__asm {
		push esi
		mov esi, x
		mov eax, dword ptr oldval
		mov edx, dword ptr oldval + 4
		mov ebx, dword ptr newval
		mov ecx, dword ptr newval + 4
		lock cmpxchg8b[esi]
		pop esi
		mov dword ptr iReturn, eax
		mov dword ptr iReturn + 4, edx
	};
	return iReturn;
}

INLINE __device__ long long _InterlockedExchange64(long long *x,
	long long newval)
{
	__asm {
		push ebx
		push esi
		mov esi, [x]
		mov ecx, dword ptr[newval]
		mov ebx, dword ptr[newval + 4]
		mov eax, dword ptr[esi]
		mov edx, dword ptr[esi + 4]
l0:     lock cmpxchg8b[esi]
        jne l0
        pop esi
        pop ebx
	}
}

INLINE __device__ long long _InterlockedExchangeAdd64(long long *x,
    long long addval)
{
    __int64 iReturn;
    __asm {
        push esi
        mov esi, x
l1:     mov eax, dword ptr [esi]
        mov edx, dword ptr [esi + 4]
        mov ebx, dword ptr [addval]
        mov ecx, dword ptr [addval + 4]
        add ebx, eax
        adc ecx, edx
        lock cmpxchg8b[esi]
        jne l1
        pop esi
        mov dword ptr iReturn, eax
        mov dword ptr iReturn + 4, edx
    };
    return iReturn;
}

#endif	// defined(_WIN32) && !defined(__LP64__)
#endif  //  defined(_MSC_VER)

// TODO - check MAC versions
// OSAtomicCompareAndSwapPtr should be defined in "#include
// <libkern/OSAtomic.h>"
//#if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__ >= 1050
//#define atomicCAS
// OSAtomicCompareAndSwapPtr
//(old_value, new_value, ptr);

namespace quasar
{

// Define some signed overloads for the existing atomicCAS functions
#if !defined(TARGET_CUDA) && (defined(__GNUG__) || defined(__clang__))
// For GCC/clang, _atomicCAS is defined as a macro. Initially, we used a function,
// but this caused some compilation problems
#define _atomicCAS_macro(x,oldval,newval) __sync_val_compare_and_swap(x,oldval,newval)
#else
#define _atomicCAS_macro(x,oldval,newval) _atomicCAS(x,oldval,newval)
#endif

INLINE __HOST_V__ __device__ long _atomicCAS(long *x, long oldval, long newval)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	return atomicCAS((unsigned int *)(x), (unsigned int)oldval,
		(unsigned int)newval);
#elif defined(_MSC_VER)
	return _InterlockedCompareExchange(x, newval, oldval);
#elif defined(__GNUG__) || defined(__clang__)
	return __sync_val_compare_and_swap(x, oldval, newval);
#else
#   error AtomicCAS not implemented for this platform!
#endif
}

INLINE __HOST_V__ __device__ unsigned int _atomicCAS(unsigned int *x, unsigned int oldval, unsigned int newval)
{
	return _atomicCAS((long *)x, (long)oldval, (long)newval);
}

INLINE __HOST_V__ __device__ long long _atomicCAS(long long *x, long long oldval,
	long long newval)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	return atomicCAS((unsigned long long *)(x), (unsigned long long)oldval,
		(unsigned long long)newval);
#elif defined(_MSC_VER)
	return _InterlockedCompareExchange64(x, newval, oldval);
#elif defined(__GNUG__) || defined(__clang__)
	return __sync_val_compare_and_swap(x, oldval, newval);
#else
#   error AtomicCAS not implemented for this platform!
#endif
}

INLINE __HOST_V__ __device__ unsigned long long _atomicCAS(unsigned long long *x, unsigned long long oldval, unsigned long long newval)
{
	return _atomicCAS((long long *)x, (long long)oldval, (long long)newval);
}

INLINE __HOST_V__ __device__ uint16_t _atomicCAS(uint16_t *x, uint16_t oldval, uint16_t newval)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	int odd = (size_t)x & 1;
	long *base_address = (long *)((size_t)x & ~2);
	long init = *base_address;
	long old_val = odd ? (init & 0xffffu) | (oldval << 16) : (init & 0xffff0000u) | oldval;
	long new_val = odd ? (init & 0xffffu) | (newval << 16) : (init & 0xffff0000u) | newval;
	long result = _atomicCAS(base_address, old_val, new_val);
	return odd ? (uint32_t)result >> 16 : ((uint32_t)result & 0xffffu);
#elif defined(__GNUG__) || defined(__clang__)
	return __sync_val_compare_and_swap(x, oldval, newval);
#elif defined(_MSC_VER)
	return _InterlockedCompareExchange16((short *)(x), (short)(newval), (short)(oldval));
#else
#   error AtomicCAS not implemented for this platform!
#endif
}

#if !defined(TARGET_CUDA) && (defined(__GNUG__) || defined(__clang__))
// For GCC/clang, _atomicExch is defined as a macro. Initially, we used a function,
// but this caused some compilation problems
#define _atomicExch_macro(x, newval) __sync_lock_test_and_set(x, newval)
#else
#define _atomicExch_macro(x, newval) _atomicExch(x, newval)
#endif

INLINE __HOST_V__ __device__ long _atomicExch(long *x, long newval)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	return (long)atomicExch((unsigned int *)x, (unsigned int)newval);
#elif defined(_MSC_VER)
	return _InterlockedExchange(x, newval);
#elif defined(__GNUG__) || defined(__clang__)
	return __sync_lock_test_and_set(x, newval);
#else
#   error AtomicExch not implemented for this platform!
#endif
}

INLINE __HOST_V__ __device__ long long _atomicExch(long long *x, long long newval)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	return (long long)atomicExch((unsigned long long *)x, (unsigned long long)newval);
#elif defined(_MSC_VER)
	return _InterlockedExchange64(x, newval);
#elif defined(__GNUG__) || defined(__clang__)
	return __sync_lock_test_and_set(x, newval);
#else
#   error AtomicExch not implemented for this platform!
#endif
}

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && !defined(DBL_SCALAR)
INLINE __HOST_V__ __device__ scalar _atomicAdd(scalar *a, scalar b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && !defined(DBL_SCALAR)
	return atomicAdd(a, b) + b;
#endif
}
#define FLOAT_ATOMIC_ADD_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(__clang__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ int _atomicAdd(int *a, int b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	return atomicAdd(a, b) + b;
#elif defined(_MSC_VER)
    return (int)_InterlockedExchangeAdd((long *)a, (long)b) + b;
#elif defined(__GNUG__) || defined(__clang__)
	return __sync_add_and_fetch(a, b);
#endif
}
#define INT_ATOMIC_ADD_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(__clang__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ uint32_t _atomicAdd(uint32_t *a, uint32_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicAdd(a, b) + b;
#elif defined(_MSC_VER)
    return (uint32_t)_InterlockedExchangeAdd((long *)a, (long)b) + b;
#elif defined(__GNUG__) || defined(__clang__)
    return __sync_add_and_fetch(a, b);
#endif
}
#define UINT32_ATOMIC_ADD_DEFINED
#endif


#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(__clang__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ int64_t _atomicAdd(int64_t *a, int64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (int64_t) atomicAdd((unsigned long long int *)a, (unsigned long long int)b) + b;
#elif defined(_MSC_VER)
    return (int64_t)_InterlockedExchangeAdd64((long long *)a, (long long)b) + b;
#elif defined(__GNUG__) || defined(__clang__)
    return __sync_add_and_fetch(a, b);
#endif
}
#define INT64_ATOMIC_ADD_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(__clang__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ uint64_t _atomicAdd(uint64_t *a, uint64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (uint64_t) atomicAdd((unsigned long long int*) a, (unsigned long long int)b) + b;
#elif defined(_MSC_VER)
    return (uint64_t)_InterlockedExchangeAdd64((long long int*)a, (long long int)b) + b;
#elif defined(__GNUG__) || defined(__clang__)
    return __sync_add_and_fetch(a, b);
#endif
}
#define UINT64_ATOMIC_ADD_DEFINED
#endif


#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && !defined(DBL_SCALAR)
INLINE __HOST_V__ __device__ scalar _atomicSub(scalar *a, scalar b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && !defined(DBL_SCALAR)
	return atomicAdd(a, -b) - b;
#endif
}
#define FLOAT_ATOMIC_SUB_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ int _atomicSub(int *a, int b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	return atomicAdd(a, -b) - b;
#elif defined(_MSC_VER)
    return (int)_InterlockedExchangeAdd((long *)a, (long) -b) - b;
#else
#pragma omp atomic
    *a -= b;
    return *a;
#endif
}
#define INT_ATOMIC_SUB_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ uint32_t _atomicSub(uint32_t *a, uint32_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicAdd(a, -b) - b;
#elif defined(_MSC_VER)
    return (uint32_t)_InterlockedExchangeAdd((long *) a, -(long)b) - b;
#else
#pragma omp atomic
    *a -= b;
    return *a;
#endif
}
#define UINT32_ATOMIC_SUB_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ int64_t _atomicSub(int64_t *a, int64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (int64_t)atomicAdd((unsigned long long int*)a, (unsigned long long int)-b) - b;
#elif defined(_MSC_VER)
    return (int64_t)_InterlockedExchangeAdd64((long long *)a, (long long)-b) - b;
#else
#pragma omp atomic
    *a -= b;
    return *a;
#endif
}
#define INT64_ATOMIC_SUB_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ uint64_t _atomicSub(uint64_t *a, uint64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (uint64_t)atomicAdd((unsigned long long int*) a, (unsigned long long int)-b) - b;
#elif defined(_MSC_VER)
    return (uint64_t)_InterlockedExchangeAdd64((long long *)a, -(long long)b) - b;
#else
#pragma omp atomic
    *a -= b;
    return *a;
#endif
}
#define UINT64_ATOMIC_SUB_DEFINED
#endif

// Note: 64-bit versions of atomicAnd, atomicOr and atomicXor require compute arch 3.5 or higher
// Atomic AND
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ int _atomicAnd(int *a, int b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicAnd(a, b) & b;
#elif defined(_MSC_VER)
    return (int)_InterlockedAnd((long *)a, (long)b) & b;
#else
#pragma omp atomic
    *a &= b;
    return *a;
#endif
}
#define INT32_ATOMIC_AND_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ uint32_t _atomicAnd(uint32_t *a, uint32_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicAnd(a, b) & b;
#elif defined(_MSC_VER)
    return (uint32_t)_InterlockedAnd((long *)a, (long)b) & b;
#else
#pragma omp atomic
    *a &= b;
    return *a;
#endif
}
#define UINT32_ATOMIC_AND_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 350 || (defined(__GNUG__) || defined(_MSC_VER)) && !defined(__CUDA_ARCH__)
INLINE __HOST_V__ __device__ int64_t _atomicAnd(int64_t *a, int64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (int64_t)atomicAnd((unsigned long long int *)a, (unsigned long long int)b) & b;
#elif defined(_MSC_VER)
    return (int64_t)_InterlockedAnd64((long long *)a, (long long)b) & b;
#else
#pragma omp atomic
    *a &= b;
    return *a;
#endif
}
#define INT64_ATOMIC_AND_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 350 || (defined(__GNUG__) || defined(_MSC_VER)) && !defined(__CUDA_ARCH__)
INLINE __HOST_V__ __device__ uint64_t _atomicAnd(uint64_t *a, uint64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (uint64_t)atomicAnd((unsigned long long int *)a, (unsigned long long int)b) & b;
#elif defined(_MSC_VER)
    return (uint64_t)_InterlockedAnd64((long long *)a, (long long)b) & b;
#else
#pragma omp atomic
    *a &= b;
    return *a;
#endif
}
#define UINT64_ATOMIC_AND_DEFINED
#endif

// Atomic OR
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ int _atomicOr(int *a, int b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicOr(a, b) | b;
#elif defined(_MSC_VER)
    return (int)_InterlockedOr((long *)a, (long)b) | b;
#else
#pragma omp atomic
    *a |= b;
    return *a;
#endif
}
#define INT32_ATOMIC_OR_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ uint32_t _atomicOr(uint32_t *a, uint32_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicOr(a, b) | b;
#elif defined(_MSC_VER)
    return (uint32_t)_InterlockedOr((long *)a, (long)b) | b;
#else
#pragma omp atomic
    *a |= b;
    return *a;
#endif
}
#define UINT32_ATOMIC_OR_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 350 || (defined(__GNUG__) || defined(_MSC_VER)) && !defined(__CUDA_ARCH__)
INLINE __HOST_V__ __device__ int64_t _atomicOr(int64_t *a, int64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (int64_t)atomicOr((unsigned long long int *)a, (unsigned long long int)b) | b;
#elif defined(_MSC_VER)
    return (int64_t)_InterlockedOr64((long long *)a, (long long)b) | b;
#else
#pragma omp atomic
    *a |= b;
    return *a;
#endif
}
#define INT64_ATOMIC_OR_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 350 || (defined(__GNUG__) || defined(_MSC_VER)) && !defined(__CUDA_ARCH__)
INLINE __HOST_V__ __device__ uint64_t _atomicOr(uint64_t *a, uint64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (uint64_t)atomicOr((unsigned long long int *)a, (unsigned long long int)b) | b;
#elif defined(_MSC_VER)
    return (uint64_t)_InterlockedOr64((long long *)a, (long long)b) | b;
#else
#pragma omp atomic
    *a |= b;
    return *a;
#endif
}
#define UINT64_ATOMIC_OR_DEFINED
#endif

// Atomic XOR
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ int _atomicXor(int *a, int b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicXor(a, b) ^ b;
#elif defined(_MSC_VER)
    return (int)_InterlockedXor((long *)a, (long)b) ^ b;
#else
#pragma omp atomic
    *a ^= b;
    return *a;
#endif
}
#define INT32_ATOMIC_XOR_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) || defined(__GNUG__) || defined(_MSC_VER)
INLINE __HOST_V__ __device__ uint32_t _atomicXor(uint32_t *a, uint32_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return atomicXor(a, b) ^ b;
#elif defined(_MSC_VER)
    return (uint32_t)_InterlockedXor((long *)a, (long)b) ^ b;
#else
#pragma omp atomic
    *a ^= b;
    return *a;
#endif
}
#define UINT32_ATOMIC_XOR_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 350 || (defined(__GNUG__) || defined(_MSC_VER)) && !defined(__CUDA_ARCH__)
INLINE __HOST_V__ __device__ int64_t _atomicXor(int64_t *a, int64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (int64_t)atomicXor((unsigned long long int *)a, (unsigned long long int)b) | b;
#elif defined(_MSC_VER)
    return (int64_t)_InterlockedXor64((long long *)a, (long long)b) ^ b;
#else
#pragma omp atomic
    *a ^= b;
    return *a;
#endif
}
#define INT64_ATOMIC_XOR_DEFINED
#endif

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 350 || (defined(__GNUG__) || defined(_MSC_VER)) && !defined(__CUDA_ARCH__)
INLINE __HOST_V__ __device__ uint64_t _atomicXor(uint64_t *a, uint64_t b)
{
#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
    return (uint64_t)atomicXor((unsigned long long int *)a, (unsigned long long int)b) ^ b;
#elif defined(_MSC_VER)
    return (uint64_t)_InterlockedXor64((long long *)a, (long long)b) ^ b;
#else
#pragma omp atomic
    *a ^= b;
    return *a;
#endif
}
#define UINT64_ATOMIC_XOR_DEFINED
#endif

template <typename R, typename T>
INLINE __HOST_V__ __device__ R bit_cast(T t)
{
	return *(R*) &t;
}

// Additional overloaded operators
INLINE __HOST_V__ __device__ float _atomicCAS(float *x, float oldval,
													  float newval)
{
	return bit_cast<float>(_atomicCAS((long *)x,
				bit_cast<long>(oldval),
				bit_cast<long>(newval)));
}

INLINE __HOST_V__ __device__ double _atomicCAS(double *x, double oldval,
													  double newval)
{
	return bit_cast<double>(_atomicCAS((long long*)x,
				bit_cast<long long>(oldval),
				bit_cast<long long>(newval)));
}

INLINE __HOST_V__ __device__ int _atomicCAS(int *x, int oldval, int newval)
{
	return (int)_atomicCAS((long *)x, (long) oldval, (long) newval);
}

INLINE __HOST_V__ __device__ int2 _atomicCAS(int2 *x, int2 oldval, int2 newval)
{
	return bit_cast<int2>(_atomicCAS((long long*)x,
                        bit_cast<long long>(oldval),
			bit_cast<long long>(newval)));
}

INLINE __HOST_V__ __device__ int _atomicExch(int *x, int newval)
{
	return (int)_atomicExch_macro((long *)x, (long) newval);
}

INLINE __HOST_V__ __device__ float _atomicExch(float *x, float newval)
{
    return bit_cast<float>(_atomicExch_macro((long *)x,
				bit_cast<long>(newval)));
}

INLINE __HOST_V__ __device__ double _atomicExch(double *x, double newval)
{
    return bit_cast<double>(_atomicExch_macro((long long*)x,
			bit_cast<long long>(newval)));
}

INLINE __HOST_V__ __device__ int2 _atomicExch(int2 *x, int2 newval)
{
    return bit_cast<int2>(_atomicExch_macro((long long*)x,
			bit_cast<long long>(newval)));
}


// Integers - operator
#define DECL_INT_ATOMIC_OP(name, X)                                            \
    INLINE __device__ __HOST_V__ int _atomic##name(int *pd, int d)             \
    {                                                                          \
        union                                                                  \
        {                                                                      \
            unsigned int iOld;                                                 \
            int dOld;                                                          \
        };                                                                     \
        union                                                                  \
        {                                                                      \
            unsigned int iNew;                                                 \
            int dNew;                                                          \
        };                                                                     \
        for (;;)                                                               \
        {                                                                      \
            iOld = *(unsigned int *)pd;                                        \
            dNew = dOld X d;                                                   \
            if (_atomicCAS_macro((unsigned int *)pd, iOld, iNew) == iOld)      \
                return dNew;                                                   \
        }                                                                      \
    }
// Integers - function
#define DECL_INT_ATOMIC_OP_FN(name, fn)                                        \
    INLINE __device__ __HOST_V__ int _atomic##name(int *pd, int d)             \
    {                                                                          \
        union                                                                  \
        {                                                                      \
            unsigned int iOld;                                                 \
            int dOld;                                                          \
        };                                                                     \
        union                                                                  \
        {                                                                      \
            unsigned int iNew;                                                 \
            int dNew;                                                          \
        };                                                                     \
        for (;;)                                                               \
        {                                                                      \
            iOld = *(unsigned int *)pd;                                        \
            dNew = fn(dOld, d);                                                \
            if (_atomicCAS_macro((unsigned int *)pd, iOld, iNew) == iOld)      \
                return dNew;                                                   \
        }                                                                      \
    }
// Float32 - operator
#define DECL_FLOAT_ATOMIC_OP(name, X)                                          \
    INLINE __device__ __HOST_V__ float _atomic##name(float *pd, float d)       \
    {                                                                          \
        union                                                                  \
        {                                                                      \
            unsigned int iOld;                                                 \
            float dOld;                                                        \
        };                                                                     \
        union                                                                  \
        {                                                                      \
            unsigned int iNew;                                                 \
            float dNew;                                                        \
        };                                                                     \
        for (;;)                                                               \
        {                                                                      \
            iOld = *(unsigned int *)pd;                                        \
            dNew = dOld X d;                                                   \
            if (_atomicCAS_macro((unsigned int *)pd, iOld, iNew) == iOld)      \
                return dNew;                                                   \
        }                                                                      \
    }
// Float32 - function
#define DECL_FLOAT_ATOMIC_OP_FN(name, fn)                                      \
    INLINE __device__ __HOST_V__ float _atomic##name(float *pd, float d)       \
    {                                                                          \
        union                                                                  \
        {                                                                      \
            unsigned int iOld;                                                 \
            float dOld;                                                        \
        };                                                                     \
        union                                                                  \
        {                                                                      \
            unsigned int iNew;                                                 \
            float dNew;                                                        \
        };                                                                     \
        for (;;)                                                               \
        {                                                                      \
            iOld = *(unsigned int *)pd;                                        \
            dNew = fn(dOld, d);                                                \
            if (_atomicCAS_macro((unsigned int *)pd, iOld, iNew) == iOld)      \
                return dNew;                                                   \
        }                                                                      \
    }
// Float64 - operator
#define DECL_DOUBLE_ATOMIC_OP(name, X)                                         \
    INLINE __device__ __HOST_V__ double _atomic##name(double *pd, double d)    \
    {                                                                          \
        union                                                                  \
        {                                                                      \
            long long iOld;                                                    \
            double dOld;                                                       \
        };                                                                     \
        union                                                                  \
        {                                                                      \
            long long iNew;                                                    \
            double dNew;                                                       \
        };                                                                     \
        for (;;)                                                               \
        {                                                                      \
            iOld = *(long long *)pd;                                           \
            dNew = dOld X d;                                                   \
            if (_atomicCAS_macro((long long *)pd, iOld, iNew) == iOld)         \
                return dNew;                                                   \
        }                                                                      \
    }
// Float64 - function
#define DECL_DOUBLE_ATOMIC_OP_FN(name, fn)                                     \
    INLINE __device__ __HOST_V__ double _atomic##name(double *pd, double d)    \
    {                                                                          \
        union                                                                  \
        {                                                                      \
            long long iOld;                                                    \
            double dOld;                                                       \
        };                                                                     \
        union                                                                  \
        {                                                                      \
            long long iNew;                                                    \
            double dNew;                                                       \
        };                                                                     \
        for (;;)                                                               \
        {                                                                      \
            iOld = *(long long *)pd;                                           \
            dNew = fn(dOld, d);                                                \
            if (_atomicCAS_macro((long long *)pd, iOld, iNew) == iOld)         \
                return dNew;                                                   \
        }                                                                      \
    }

// Float16 - operator
#define DECL_HALF_ATOMIC_OP(name, X)                                           \
    INLINE __device__ __HOST_V__ __half _atomic##name(__half *pd, __half d)    \
    {                                                                          \
		__half hOld, hNew;													   \
        for (;;)                                                               \
        {                                                                      \
            hOld = *pd;														   \
            hNew = float2half_rn(half2float(hOld) * half2float(d));            \
            if (_atomicCAS_macro((uint16_t *)pd, (uint16_t&)hOld, (uint16_t&)hNew) == (uint16_t&)hOld)  \
                return hNew;                                                   \
        }                                                                      \
    }

// Float16 - function
#define DECL_HALF_ATOMIC_OP_FN(name, fn)                                       \
    INLINE __device__ __HOST_V__ __half _atomic##name(__half *pd, __half d)    \
    {                                                                          \
		__half hOld, hNew;													   \
        for (;;)                                                               \
        {                                                                      \
            hOld = *pd;														   \
            hNew = float2half_rn(fn(half2float(hOld), half2float(d)));         \
            if (_atomicCAS_macro((uint16_t *)pd, (uint16_t&)hOld, (uint16_t&)hNew) == (uint16_t&)hOld)  \
                return hNew;                                                   \
        }                                                                      \
    }


#ifndef INT_ATOMIC_ADD_DEFINED
    DECL_INT_ATOMIC_OP(Add, +)
#endif
#ifndef INT_ATOMIC_SUB_DEFINED
    DECL_INT_ATOMIC_OP(Sub, -)
#endif
    DECL_INT_ATOMIC_OP(Mul, *)
    DECL_INT_ATOMIC_OP(Div, / )
    DECL_INT_ATOMIC_OP_FN(Pow, pow)
    DECL_INT_ATOMIC_OP_FN(Min, min)
    DECL_INT_ATOMIC_OP_FN(Max, max)

#ifndef FLOAT_ATOMIC_ADD_DEFINED
    DECL_FLOAT_ATOMIC_OP(Add, +)
#endif
#ifndef FLOAT_ATOMIC_SUB_DEFINED
    DECL_FLOAT_ATOMIC_OP(Sub, -)
#endif
    DECL_FLOAT_ATOMIC_OP(Mul, *)
    DECL_FLOAT_ATOMIC_OP(Div, / )
    DECL_FLOAT_ATOMIC_OP_FN(Pow, pow)
    DECL_FLOAT_ATOMIC_OP_FN(Min, min)
    DECL_FLOAT_ATOMIC_OP_FN(Max, max)

#ifndef DOUBLE_ATOMIC_ADD_DEFINED
    DECL_DOUBLE_ATOMIC_OP(Add, +)
#endif
#ifndef DOUBLE_ATOMIC_SUB_DEFINED
    DECL_DOUBLE_ATOMIC_OP(Sub, -)
#endif
    DECL_DOUBLE_ATOMIC_OP(Mul, *)
    DECL_DOUBLE_ATOMIC_OP(Div, / )
    DECL_DOUBLE_ATOMIC_OP_FN(Pow, pow)
    DECL_DOUBLE_ATOMIC_OP_FN(Min, min)
    DECL_DOUBLE_ATOMIC_OP_FN(Max, max)

#if !defined(__GNUG__) && !defined(__clang__)
    DECL_HALF_ATOMIC_OP(Add, +)
    DECL_HALF_ATOMIC_OP(Sub, -)
    DECL_HALF_ATOMIC_OP(Mul, *)
    DECL_HALF_ATOMIC_OP(Div, / )
    DECL_HALF_ATOMIC_OP_FN(Pow, pow)
    DECL_HALF_ATOMIC_OP_FN(Min, min)
    DECL_HALF_ATOMIC_OP_FN(Max, max)
#endif

    //
    // Atomic operations for cscalar numbers
    //
    // TO DO - operation is not atomic yet... A possible solution might be to
    // use
    // the atomicCAS function, although we need 128-bit integers for the
    // DBL_SCALAR
    // case.
    //
	#ifdef HALF_SCALAR
	#define cscalar_atomicT cscalar_t<half_t>
	#else
	#define cscalar_atomicT cscalar
	#endif

    INLINE __device__ __HOST_V__ cscalar_atomicT _atomicAdd(cscalar_atomicT *a, cscalar_atomicT b)
    {
        return cscalar_atomicT(_atomicAdd(&a->x, b.x), _atomicAdd(&a->y, b.y));
    }
    INLINE __device__ __HOST_V__ cscalar_atomicT _atomicSub(cscalar_atomicT *a, cscalar_atomicT b)
    {
        return cscalar_atomicT(_atomicSub(&a->x, b.x), _atomicSub(&a->y, b.y));
    }
    INLINE __device__ __HOST_V__ cscalar_atomicT _atomicMul(cscalar_atomicT *a, cscalar_atomicT b)
    {
        return cscalar_atomicT(_atomicMul(&a->x, b.x), _atomicMul(&a->y, b.y));
    }
    INLINE __device__ __HOST_V__ cscalar_atomicT _atomicDiv(cscalar_atomicT *a, cscalar_atomicT b)
    {
        return cscalar_atomicT(_atomicDiv(&a->x, b.x), _atomicDiv(&a->y, b.y));
    }
    INLINE __device__ __HOST_V__ cscalar_atomicT _atomicPow(cscalar_atomicT *a, cscalar_atomicT b)
    {
        return cscalar_atomicT(_atomicPow(&a->x, b.x), _atomicPow(&a->y, b.y));
    }
	#undef cscalar_atomicT

	// Vector extensions (note: these are non truly atomic)
	#define BINARY_FUNCTION_T(X, T)                                           \
    INLINE T##1 __device__ __HOST_V__ (X)(T##1* a, T##1 b)                     \
    {                                                                          \
        return make_##T##1(X(&a->x, b.x));                                     \
    }                                                                          \
    INLINE T##2 __device__ __HOST_V__ (X)(T##2* a, T##2 b)                     \
    {                                                                          \
        return make_##T##2(X(&a->x, b.x), X(&a->y, b.y));                      \
    }                                                                          \
    INLINE T##3 __device__ __HOST_V__ (X)(T##3* a, T##3 b)                     \
    {                                                                          \
        return make_##T##3(X(&a->x, b.x), X(&a->y, b.y), X(&a->z, b.z));       \
    }                                                                          \
    INLINE T##4 __device__ __HOST_V__ (X)(T##4* a, T##4 b)                     \
    {                                                                          \
        return make_##T##4(X(&a->x, b.x), X(&a->y, b.y), X(&a->z, b.z),        \
                           X(&a->w, b.w));                                     \
    }                                                                          \
    template <int N> INLINE T##N<N> __device__ __HOST_V__ (X)(T##N<N>* a, T##N<N> b) \
    {                                                                          \
        T##N<N> y;                                                             \
        FOR_ALL(y.el[i] = X(&a->el[i], b.el[i]));                              \
        return y;                                                              \
    }

	BINARY_FUNCTION_T(_atomicAdd, int);
	BINARY_FUNCTION_T(_atomicSub, int);
	BINARY_FUNCTION_T(_atomicMul, int);
	BINARY_FUNCTION_T(_atomicDiv, int);
	BINARY_FUNCTION_T(_atomicPow, int);
	BINARY_FUNCTION_T(_atomicMin, int);
	BINARY_FUNCTION_T(_atomicMax, int);

	BINARY_FUNCTION_T(_atomicAdd, scalar);
	BINARY_FUNCTION_T(_atomicSub, scalar);
	BINARY_FUNCTION_T(_atomicMul, scalar);
	BINARY_FUNCTION_T(_atomicDiv, scalar);
	BINARY_FUNCTION_T(_atomicPow, scalar);
	BINARY_FUNCTION_T(_atomicMin, scalar);
	BINARY_FUNCTION_T(_atomicMax, scalar);

#ifndef HALF_SCALAR
	BINARY_FUNCTION_T(_atomicAdd, cscalar);
	BINARY_FUNCTION_T(_atomicSub, cscalar);
	BINARY_FUNCTION_T(_atomicMul, cscalar);
	BINARY_FUNCTION_T(_atomicDiv, cscalar);
	BINARY_FUNCTION_T(_atomicPow, cscalar);
#endif

    // 1. VectorBase<T>
    template <BoundaryAccessMode mode, typename T, T atomic_function(T *, T)>
    INLINE __device__ __HOST_V__ T vector_at(const VectorBase<T> x, int pos, T val)
    {
        if (mode == Default)
        {
            CHECK_BOUNDS(x, pos >= 0 && pos < x.dim1, T());
        }
        else if (mode == Unchecked)
        {
            CHECK_BOUNDS_FAST(x, pos >= 0 && pos < x.dim1, T());
        }
        else if (!(pos >= 0 && pos < x.dim1))
            return T();
        CHECK_NAN_OR_INF(x, val);
        return atomic_function((T *)&x.data[pos], val);
    }

    // 2. MatrixBase<T>
    template <BoundaryAccessMode mode, typename T, T atomic_function(T *, T)>
    INLINE __device__ __HOST_V__ T matrix_at(const MatrixBase<T> x, int2 pos, T val)
    {
        int index = pos.y + pos.x * x.dim2;
        if (mode == Default)
        {
            CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2, T());
        }
        else if (mode == Unchecked)
        {
            CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2, T());
        }
        else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2))
            return T();
        CHECK_NAN_OR_INF(x, val);
        return atomic_function((T *)&x.data[index], val);
    }

    // 3. CubeBase<T>
    template <BoundaryAccessMode mode, typename T, T atomic_function(T *, T)>
    INLINE __device__ __HOST_V__ T cube_at(const CubeBase<T> x, int3 pos, T val)
    {
        int index = pos.z + (pos.y + pos.x * x.dim2) * x.dim3;
        if (mode == Default)
        {
            CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3, T());
        }
        else if (mode == Unchecked)
        {
            CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 * x.dim3, T());
        }
        else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3))
            return T();
        CHECK_NAN_OR_INF(x, val);
        return atomic_function((T *)&x.data[index], val);
    }

	// Atomic functions with 3 arguments
	// 1. VectorBase<T>
    template <BoundaryAccessMode mode, typename T, T atomic_function(T *, T, T)>
    INLINE __device__ __HOST_V__ T vector_at(const VectorBase<T> x, int pos, T ref, T val)
    {
        if (mode == Default)
        {
            CHECK_BOUNDS(x, pos >= 0 && pos < x.dim1, T());
        }
        else if (mode == Unchecked)
        {
            CHECK_BOUNDS_FAST(x, pos >= 0 && pos < x.dim1, T());
        }
        else if (!(pos >= 0 && pos < x.dim1))
            return T();
        CHECK_NAN_OR_INF(x, val);
		CHECK_NAN_OR_INF(x, ref);
        return atomic_function((T *)&x.data[pos], ref, val);
    }

    // 2. MatrixBase<T>
    template <BoundaryAccessMode mode, typename T, T atomic_function(T *, T, T)>
    INLINE __device__ __HOST_V__ T matrix_at(const MatrixBase<T> x, int2 pos, T ref, T val)
    {
        int index = pos.y + pos.x * x.dim2;
        if (mode == Default)
        {
            CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2, T());
        }
        else if (mode == Unchecked)
        {
            CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2, T());
        }
        else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2))
            return T();
        CHECK_NAN_OR_INF(x, val);
		CHECK_NAN_OR_INF(x, ref);
        return atomic_function((T *)&x.data[index], ref, val);
    }

    // 3. CubeBase<T>
    template <BoundaryAccessMode mode, typename T, T atomic_function(T *, T, T)>
    INLINE __device__ __HOST_V__ T cube_at(const CubeBase<T> x, int3 pos, T ref, T val)
    {
        int index = pos.z + (pos.y + pos.x * x.dim2) * x.dim3;
        if (mode == Default)
        {
            CHECK_BOUNDS(x, min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3, T());
        }
        else if (mode == Unchecked)
        {
            CHECK_BOUNDS_FAST(x, index >= 0 && index < x.dim1 * x.dim2 * x.dim3, T());
        }
        else if (!(min(pos) >= 0 && pos.x < x.dim1 && pos.y < x.dim2 && pos.z < x.dim3))
            return T();
        CHECK_NAN_OR_INF(x, val);
		CHECK_NAN_OR_INF(x, ref);
        return atomic_function(
            (T *)&x.data[index], ref, val);
    }

#if defined(TARGET_CUDA) && defined(__CUDA_ARCH__)
	
extern "C" {
    __device__ volatile unsigned int syncgrid_mutex_array_l3[32][32];
    __device__ volatile unsigned int syncgrid_mutex_array_l2[32];
    __device__ volatile unsigned int syncgrid_mutex_array_l1;
}	

    INLINE __device__ void __syncgrid()
    {        
        if ((threadIdx.x | threadIdx.y | threadIdx.z) == 0)
        {            
            int griddim = gridDim.x * gridDim.y * gridDim.z;                  
            int block_id =  ((blockIdx.z * gridDim.y) + blockIdx.y)*gridDim.x + blockIdx.x;
            
            int l3_index = (block_id & 0x1f);
            int l2_index = (block_id >> 5) & 0x1f;
            int l1_index = block_id >> 10;
            
            unsigned int l3_xor_value = 1 << l3_index;
            unsigned int l2_xor_value = 1 << l2_index;
            unsigned int l1_xor_value = 1 << l1_index;  

            unsigned int goalval;

            if (griddim <= 1024)
                goalval = -((syncgrid_mutex_array_l2[0] & 0x80000000) == 0);
            else
                goalval = -((syncgrid_mutex_array_l1 & 0x80000000) == 0);


            if(block_id == griddim -1)
                l3_xor_value = (0xffffffff - l3_xor_value) + 1;

            atomicOr((unsigned int*)&syncgrid_mutex_array_l3[l1_index][l2_index], l3_xor_value);
            
            if (l3_index == 0)
            {    
                while(_atomicCAS((unsigned int *)&syncgrid_mutex_array_l3[l1_index][l2_index], 0xffffffff, 0) != 0xffffffff); 
                    ;             
            
                if (griddim <= 1024)
                {
                    if(l2_index == ((griddim-1) >> 5))    
                    {                                                   
                        l2_xor_value = (0xffffffff - l2_xor_value) + 1; 
                        while(_atomicCAS((unsigned int *)&syncgrid_mutex_array_l2[0], (goalval ^ l2_xor_value), goalval) != (goalval ^ l2_xor_value))
                            ;
                    }
                    else            
                        atomicXor((unsigned int*)&syncgrid_mutex_array_l2[0], l2_xor_value); 
                }     
                else
                {
                    if(l2_index == ((griddim-1) >> 5))                
                        l2_xor_value = (0xffffffff - l2_xor_value) + 1;

                    atomicOr((unsigned int*)&syncgrid_mutex_array_l2[l1_index], l2_xor_value);   
                }    
            }            
            
            if (griddim > 1024)
            {
                if ((block_id & (1024-1)) == 0)
                {         
                    while(_atomicCAS((unsigned int *)&syncgrid_mutex_array_l2[l1_index], 0xffffffff, 0) != 0xffffffff); 
                        ;
                    
                    if(l1_index == ((griddim-1) >> 10)) 
                    {               
                        l1_xor_value = (0xffffffff - l1_xor_value) + 1;
                        while(_atomicCAS((unsigned int *)&syncgrid_mutex_array_l1, (goalval ^ l1_xor_value), goalval) != (goalval ^ l1_xor_value))
                            ;
                    }
                    else
                        atomicXor((unsigned int*)&syncgrid_mutex_array_l1, l1_xor_value);                
                }           
                
                if((block_id & (1024-1)) != 0 || l1_index != ((griddim-1) >> 10))
                    while ((syncgrid_mutex_array_l1 & (0x80000000)) != (goalval & (0x80000000)))
                        ;
            }
            else
            {
                if(l3_index != 0 || l2_index != ((griddim-1) >> 5))
                    while ((syncgrid_mutex_array_l2[0] & (0x80000000)) != (goalval & (0x80000000)))
                        ;
            }
        }
        __syncthreads(); // Synchronize all threads so that they also reach the barrier
        __threadfence(); // Thread fence is important so that the order of memory operation
                            // is preserved across __syncgrid() blocks
    }
#endif

} // namespace quasar

//=============================================================================
// SUPPORT FOR CONSTANT MEMORY (currently, we allow max 8 KB)
//============+================================================================
#define MAX_CONST_MEM 8192
#ifdef TARGET_CUDA
__device__ __constant__ unsigned char const_mem[MAX_CONST_MEM];
#endif

//=============================================================================
// SUPPORT FOR SHARED MEMORY
//=============================================================================
#ifdef CHECK_UNSAFE_BOUNDS
#define HAS_SHARED_MEM_DWORDS
#endif

#if defined(TARGET_CUDA)
extern __shared__ scalar shared_mem[];
__device__ int shared_mem_dwords;
SELECTANY scalar *shared_mem_host = NULL;
SELECTANY int shared_mem_dwords_host = 0;
#elif defined(__QUASAR_NO_EXPORTS__)
extern "C" {
extern scalar *shared_mem;
extern int shared_mem_dwords;
}
#elif defined(QUASARHOST_DLLIMPORT)
extern "C" {
	IMPORT scalar *shared_mem;
	IMPORT int shared_mem_dwords;
}
#else
extern "C" {
EXPORT SELECTANY scalar *shared_mem = NULL;
EXPORT SELECTANY int shared_mem_dwords = 0;
}
#endif
typedef int shmem;

namespace quasar
{

    // Parallel fill algorithm (for memory initizialization)
    template <typename T>
    INLINE void __device__ __HOST_V__ shmem_fill(T *ptr, int nelem, T val)
    {
        int nThreads = __threadcnt();
        int tid = __threadidx();
        for (int i = tid; i < nelem; i += nThreads)
            ptr[i] = val;
    }

    template <typename T>
	INLINE void __device__ __HOST_V__ mem_fill(T *ptr, int nscalars, T val)
	{
		for (int i = 0; i < nscalars; i++)
			ptr[i] = val;
	}

    INLINE void __device__ __HOST_V__ shmem_init(shmem *_shmem) { *_shmem = 0; }

    // allocate shared memory
    template <typename T>
    INLINE __device__ __HOST_V__ T *shmem_allocate(shmem *_shmem, int nelem)
    {
        T *retval;
#define ALIGN(x,a)    (((x)+(a)-1) & ~((a)-1))
        *_shmem = ALIGN(*_shmem, alignof(T)); // Make sure the returned pointer is aligned
#if defined(TARGET_CUDA) && !defined(__CUDA_ARCH__)  // __HOST_V__ function
		retval = (T *)&((uint8_t *)shared_mem_host)[*_shmem];
#if defined(HAS_SHARED_MEM_DWORDS)
		shared_mem_dwords_host = 
#endif
        *_shmem += nelem * sizeof(T);
#else
        retval = (T *)&((uint8_t *) shared_mem)[*_shmem];
        *_shmem += nelem * sizeof(T);
#if defined(HAS_SHARED_MEM_DWORDS)
        if (__threadidx() == 0)
            shared_mem_dwords = *_shmem;
#endif

#endif
        if (*_shmem > 32768) // shared memory overrun
        {
#if defined(TARGET_CUDA) && !defined(__CUDA_ARCH__)
			_error.target = shared_mem_host;
#else
            _error.target = shared_mem;
#endif
            _error.target_bytes = *_shmem;
            _error.code = ERRORCODE_DYNMEM_OUTOFMEMORY;
            assert(0);
        }
        return retval;
    }

    template <typename T>
    INLINE __device__ __HOST_V__ VectorBase<T> shmem_alloc(shmem *_shmem, int dim1)
    {
        return VectorBase<T>(shmem_allocate<T>(_shmem, dim1), dim1);
    }

    template <typename T>
    INLINE __device__ __HOST_V__ MatrixBase<T> shmem_alloc(shmem *_shmem, int dim1,
                                                int dim2)
    {
        return MatrixBase<T>(shmem_allocate<T>(_shmem, dim1 * dim2), dim1, dim2);
    }

    template <typename T>
    INLINE __device__ __HOST_V__ CubeBase<T> shmem_alloc(shmem *_shmem, int dim1, int dim2,
                                              int dim3)
    {
        return CubeBase<T>(shmem_allocate<T>(_shmem, dim1 * dim2 * dim3), dim1,
                         dim2, dim3);
    }


    template <typename T>
    INLINE __device__ __HOST_V__ VectorBase<T> shmem_zeros(shmem *_shmem, int dim1)
    {
        VectorBase<T> v(shmem_allocate<T>(_shmem, dim1), dim1);
        shmem_fill<T>(v.data, v.dim1, (T)0.0f);
        return v;
    }

    template <typename T>
    INLINE __device__ __HOST_V__ MatrixBase<T> shmem_zeros(shmem *_shmem, int dim1,
                                                int dim2)
    {
        MatrixBase<T> m(shmem_allocate<T>(_shmem, dim1 * dim2), dim1, dim2);
        shmem_fill<T>(m.data, m.dim1 * m.dim2, (T)0.0f);
        return m;
    }

    template <typename T>
    INLINE __device__ __HOST_V__ CubeBase<T> shmem_zeros(shmem *_shmem, int dim1, int dim2,
                                              int dim3)
    {
        CubeBase<T> c(shmem_allocate<T>(_shmem, dim1 * dim2 * dim3),
                                  dim1, dim2, dim3);
        shmem_fill<T>(c.data, c.dim1 * c.dim2 * c.dim3, 0.0f);
        return c;
    }

	// Stack allocation functions
	template <typename T>
	INLINE __device__ __HOST_V__ VectorBase<T> stmem_uninit(void *stackp, int dim1)
	{
		return VectorBase<T>((T*)stackp, dim1);
	}

	template <typename T>
	INLINE __device__ __HOST_V__ MatrixBase<T> stmem_uninit(void *stackp, int dim1,
		int dim2)
	{
		return MatrixBase<T>((T*)stackp, dim1, dim2);
	}

	template <typename T>
	INLINE __device__ __HOST_V__ CubeBase<T> stmem_uninit(void *stackp, int dim1, int dim2,
		int dim3)
	{
		return CubeBase<T>((T*)stackp, dim1,
			dim2, dim3);
	}

	template <typename T>
	INLINE __device__ __HOST_V__ VectorBase<T> stmem_zeros(void *stackp, int dim1)
	{
		VectorBase<T> v((T*)stackp, dim1);
		shmem_fill<T>(v.data, v.dim1, (T)0.0f);
		return v;
	}

	template <typename T>
	INLINE __device__ __HOST_V__ MatrixBase<T> stmem_zeros(void *stackp, int dim1,
		int dim2)
	{
		MatrixBase<T> m((T*)stackp, dim1, dim2);
		shmem_fill<T>(m.data, m.dim1 * m.dim2, (T)0.0f);
		return m;
	}

	template <typename T>
	INLINE __device__ __HOST_V__ CubeBase<T> stmem_zeros(void *stackp, int dim1, int dim2,
		int dim3)
	{
		CubeBase<T> c((T*)stackp,
			dim1, dim2, dim3);
		shmem_fill<T>(c.data, c.dim1 * c.dim2 * c.dim3, 0.0f);
		return c;
	}

    //=============================================================================
    // USER ASSERTIONS
    //=============================================================================
    INLINE __device__ __HOST_V__ void user_assert(int x, const char *msg, int len)
    {
        if (!(x))
        {
            _error.target = msg;
            _error.target_bytes = len;
            _error.code = ERRORCODE_ASSERTIONFAILED;
            assert(0);
        }
    }

    INLINE __device__ __HOST_V__ void user_assert(int1 a, const char *msg, int len)
    {
        user_assert(a.x, msg, len);
    }
    INLINE __device__ __HOST_V__ void user_assert(int2 a, const char *msg, int len)
    {
        user_assert(a.x & a.y, msg, len);
    }
    INLINE __device__ __HOST_V__ void user_assert(int3 a, const char *msg, int len)
    {
        user_assert(a.x & a.y & a.z, msg, len);
    }
    INLINE __device__ __HOST_V__ void user_assert(int4 a, const char *msg, int len)
    {
        user_assert(a.x & a.y & a.z & a.w, msg, len);
    }
    template <int N>
    __device__ __HOST_V__ void user_assert(intN<N> a, const char *msg, int len)
    {
        int x = a.el[0];
        for (int i = 1; i < N; i++)
            x &= a.el[i];
        user_assert(x, msg, len);
    }

} // namespace quasar


//=============================================================================
// Parallel memory allocation (to be used from __kernel__ and __device__
// functions)
//=============================================================================

#ifdef ENABLE_DYNAMIC_KERNEL_MEM
#include "quasar_dynmem.h"
#else

namespace quasar
{
    // Smart object pointer - now much simpler
    template <typename T> class object_ptr
    {
      protected:
        T *_ptr;

      public:
        INLINE __device__ object_ptr() : _ptr(NULL) {}
		INLINE __device__ object_ptr(long int nullPtr) : _ptr(NULL) {}
		INLINE __device__ object_ptr(int nullPtr) : _ptr(NULL) {}
        INLINE __device__ object_ptr(T *ptr) : _ptr(ptr) {}
		INLINE __device__ object_ptr(lref<T>* val) : _ptr(&(T &)*val) {}
		INLINE __device__ object_ptr(lref<T> val) : _ptr(&(T &)val) {}
        INLINE __device__ operator T *&() { return (T *&)_ptr; }
        INLINE __device__ T *operator->() { return (T *)_ptr; }
        INLINE __device__ void release() {}
        INLINE __device__ void addref() {}
    };

    template <typename T> INLINE __device__ void VectorBase<T>::addref() {}
    template <typename T> INLINE __device__ void MatrixBase<T>::addref() {}
    template <typename T> INLINE __device__ void CubeBase<T>::addref() {}

    template <typename T> INLINE __device__ void VectorBase<T>::release() {}
    template <typename T> INLINE __device__ void MatrixBase<T>::release() {}
    template <typename T> INLINE __device__ void CubeBase<T>::release() {}

} // namespace quasar

#endif // ENABLE_DYNAMIC_KERNEL_MEM

namespace quasar
{
#ifdef CHECK_UNSAFE_BOUNDS
#define CHECK_LENGTH(cond)                                                     \
    if (!(cond))															   \
    Q_RAISE_ERROR("Vector dimensions don't match")
#else
#define CHECK_LENGTH(cond)
#endif

	// =========================================================================
	// autocast specializations
	// =========================================================================

	// cast from vec to vecX, ivec to ivecX, cvec to cvecX
#define AUTOCAST_IMPL(R, T)                                                     \
	INLINE __device__ __HOST_V__ R##1 autocast(T a, ReturnType<R##1>)           \
	{                                                                           \
		CHECK_LENGTH(1);                                                        \
		return make_##R##1(a.data[0]);                                          \
	}                                                                           \
	INLINE __device__ __HOST_V__ R##2 autocast(T a, ReturnType<R##2>)           \
	{                                                                           \
		CHECK_LENGTH(2);                                                        \
		return make_##R##2(a.data[0], a.data[1]);                               \
	}                                                                           \
	INLINE __device__ __HOST_V__ R##3 autocast(T a, ReturnType<R##3>)           \
	{                                                                           \
		CHECK_LENGTH(3);                                                        \
		return make_##R##3(a.data[0], a.data[1], a.data[2]);                    \
	}                                                                           \
	INLINE __device__ __HOST_V__ R##4 autocast(T a, ReturnType<R##4>)           \
	{                                                                           \
		CHECK_LENGTH(4);                                                        \
		return make_##R##4(a.data[0], a.data[1], a.data[2], a.data[3]);         \
	}                                                                           \
    template <int N>															\
	__device__ __HOST_V__ R##N<N> autocast(T a, ReturnType<R##N<N> >)           \
	{                                                                           \
		CHECK_LENGTH(N);                                                        \
		R##N<N> y;                                                              \
		FOR_ALL(y.el[i] = a.data[i]);                                           \
		return y;                                                               \
	}

	AUTOCAST_IMPL(scalar, Vector)
	AUTOCAST_IMPL(cscalar, CVector)
	AUTOCAST_IMPL(int, VectorBase<int>)

#undef AUTOCAST_IMPL

	// scalar -> complex
#define AUTOCAST_IMPL(T)                                                                    \
    INLINE __device__ __HOST_V__ cscalar1 autocast(T##1 a, T##1 b, ReturnType<cscalar1>)    \
    {                                                                                       \
        return make_cscalar1(cscalar(a.x, b.x));                                            \
    }                                                                                       \
    INLINE __device__ __HOST_V__ cscalar2 autocast(T##2 a, T##2 b, ReturnType<cscalar2>)    \
    {                                                                                       \
        return make_cscalar2(cscalar(a.x, b.x), cscalar(a.y, b.y));                         \
    }                                                                                       \
    INLINE __device__ __HOST_V__ cscalar3 autocast(T##3 a, T##3 b, ReturnType<cscalar3>)    \
    {                                                                                       \
        return make_cscalar3(cscalar(a.x, b.x), cscalar(a.y, b.y), cscalar(a.z, b.z));      \
    }                                                                                       \
    INLINE __device__ __HOST_V__ cscalar4 autocast(T##4 a, T##4 b, ReturnType<cscalar4>)    \
    {                                                                                       \
        return make_cscalar4(cscalar(a.x, b.x), cscalar(a.y, b.y), cscalar(a.z, b.z), cscalar(a.w, b.w));   \
    }                                                                                       \
    template <int N>                                                                        \
    INLINE __device__ __HOST_V__ cscalarN<N> autocast(T##N<N> a, T##N<N> b, ReturnType<cscalarN<N> >) \
    {                                                                                       \
        cscalarN<N> y;                                                                      \
        FOR_ALL(y.el[i] = cscalar(a.el[i], b.el[i]));                                       \
        return y;                                                                           \
    }

	AUTOCAST_IMPL(scalar)
		AUTOCAST_IMPL(int)

#undef AUTOCAST_IMPL

	// =========================================================================
	// static_autocast function: provides the following conversions, without
	// use of dynamic kernel memory
	// =========================================================================
	template <typename R, typename T>
	INLINE R __device__ __HOST_V__ static_autocast(const T &value)
	{
		return static_autocast(value, ReturnType<R>());
	}

	// 1. Fixed-sized vector conversions (these overloads have extra parameters)
	// vecX to vec
	// vecX to matrix (dim1, dim2)
	// vecX to cube (dim1, dim2, dim3)
	// vecX to ncube (ivecN
	// cvecX to vec
	// cvecX to cmatrix
	// cvecX to ccube
	// cvecX to cncube
#define STATIC_AUTOCAST_IMPL(T)															\
	INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##1 &a, ReturnType<VectorBase<T> >) { \
        return VectorBase<T>(const_cast<T*>(&a.x), 1); }	                                            \
    INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##2 &a, ReturnType<VectorBase<T> >) { \
        return VectorBase<T>(const_cast<T*>(&a.x), 2); }                                                \
    INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##3 &a, ReturnType<VectorBase<T> >) { \
        return VectorBase<T>(const_cast<T*>(&a.x), 3); }                                                \
    INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##4 &a, ReturnType<VectorBase<T> >) { \
        return VectorBase<T>(const_cast<T*>(&a.x), 4); }                                                \
    template <int N>																	 \
	INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##N<N> &a, ReturnType<VectorBase<T> >) {   \
        return VectorBase<T>(const_cast<T*>(&a.el[0]), N); }                                                \
    INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##1 &a, int dim1) {      \
        CHECK_LENGTH(dim1==1); return VectorBase<T>(const_cast<T*>(&a.x), dim1); }	                    \
    INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##2 &a, int dim1) {      \
        CHECK_LENGTH(dim1==2); return VectorBase<T>(const_cast<T*>(&a.x), dim1); }                      \
    INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##3 &a, int dim1) {      \
        CHECK_LENGTH(dim1==3); return VectorBase<T>(const_cast<T*>(&a.x), dim1); }                      \
    INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##4 &a, int dim1) {      \
        CHECK_LENGTH(dim1==4); return VectorBase<T>(const_cast<T*>(&a.x), dim1); }                      \
    template <int N>																	 \
	INLINE VectorBase<T> __device__ __HOST_V__ static_autocast(const T##N<N> &a, int dim1) {   \
        CHECK_LENGTH(dim1==N); return VectorBase<T>(const_cast<T*>(&a.el[0]), dim1); }                      \
	INLINE MatrixBase<T> __device__ __HOST_V__ static_autocast(const T##1 &a, int2 dims) { \
        CHECK_LENGTH(prod(dims)==1); return MatrixBase<T>(const_cast<T*>(&a.x), dims.x, dims.y); }	\
    INLINE MatrixBase<T> __device__ __HOST_V__ static_autocast(const T##2 &a, int2 dims) { \
        CHECK_LENGTH(prod(dims)==2); return MatrixBase<T>(const_cast<T*>(&a.x), dims.x, dims.y); }  \
    INLINE MatrixBase<T> __device__ __HOST_V__ static_autocast(const T##3 &a, int2 dims) { \
        CHECK_LENGTH(prod(dims)==3); return MatrixBase<T>(const_cast<T*>(&a.x), dims.x, dims.y); }  \
    INLINE MatrixBase<T> __device__ __HOST_V__ static_autocast(const T##4 &a, int2 dims) { \
        CHECK_LENGTH(prod(dims)==4); return MatrixBase<T>(const_cast<T*>(&a.x), dims.x, dims.y); } \
    template <int N>															   \
	INLINE MatrixBase<T> __device__ __HOST_V__ static_autocast(const T##N<N> &a, int2 dims) { \
        CHECK_LENGTH(prod(dims)==N); return MatrixBase<T>(const_cast<T*>(&a.el[0]), dims.x, dims.y); }      \
	INLINE CubeBase<T> __device__ __HOST_V__ static_autocast(T##1 &a, int3 dims) { \
        CHECK_LENGTH(prod(dims)==1); return CubeBase<T>(const_cast<T*>(&a.x), dims.x, dims.y, dims.z); } \
    INLINE CubeBase<T> __device__ __HOST_V__ static_autocast(T##2 &a, int3 dims) { \
        CHECK_LENGTH(prod(dims)==2); return CubeBase<T>(const_cast<T*>(&a.x), dims.x, dims.y, dims.z); } \
    INLINE CubeBase<T> __device__ __HOST_V__ static_autocast(T##3 &a, int3 dims) { \
        CHECK_LENGTH(prod(dims)==3); return CubeBase<T>(const_cast<T*>(&a.x), dims.x, dims.y, dims.z); } \
    INLINE CubeBase<T> __device__ __HOST_V__ static_autocast(T##4 &a, int3 dims) { \
        CHECK_LENGTH(prod(dims)==4); return CubeBase<T>(const_cast<T*>(&a.x), dims.x, dims.y, dims.z); } \
    template <int N>															\
	INLINE CubeBase<T> __device__ __HOST_V__ static_autocast(const T##N<N> &a, int3 dims) { \
        CHECK_LENGTH(prod(dims)==N); return CubeBase<T>(const_cast<T*>(&a.el[0]), dims.x, dims.y, dims.z); } \
	INLINE MatrixBase<T> __device__ __HOST_V__ static_autocast(const VectorBase<T>& a, int2 dims) { \
        CHECK_LENGTH(prod(dims)==a.dim1); return MatrixBase<T>(const_cast<T*>(a.data), dims.x, dims.y); } \
	INLINE CubeBase<T> __device__ __HOST_V__ static_autocast(const VectorBase<T>& a, int3 dims) { \
        CHECK_LENGTH(prod(dims)==a.dim1); return CubeBase<T>(const_cast<T*>(a.data), dims.x, dims.y, dims.z); }

	STATIC_AUTOCAST_IMPL(scalar)
	STATIC_AUTOCAST_IMPL(cscalar)
	STATIC_AUTOCAST_IMPL(int)

#undef STATIC_AUTOCAST_IMPL

    // 2. Dimensionality changing auto cast
    template <typename T>
    __device__ MatrixBase<T> static_autocast(VectorBase<T> A, ReturnType<MatrixBase<T> >)
	{
		return MatrixBase<T>(A.data, 1, A.dim1);
	}

	template <typename T>
	__device__ VectorBase<T> static_autocast(MatrixBase<T> A, ReturnType<VectorBase<T> >)
	{
		return VectorBase<T>(A.data, A.dim1 * A.dim2);
	}

	template <typename T>
	__device__ CubeBase<T> static_autocast(VectorBase<T> A, ReturnType<CubeBase<T> >)
	{
		return CubeBase<T>(A.data, 1, A.dim1, 1);
	}

	template <typename T>
	__device__ VectorBase<T> static_autocast(CubeBase<T> A, ReturnType<VectorBase<T> >)
	{
		return VectorBase<T>(A.data, A.dim1 * A.dim2 * A.dim3);
	}

	template <typename T>
	__device__ CubeBase<T> static_autocast(MatrixBase<T> A, ReturnType<CubeBase<T> >)
	{
		return CubeBase<T>(A.data, A.dim1, A.dim2, 1);
	}

	template <typename T>
	__device__ MatrixBase<T> static_autocast(CubeBase<T> A, ReturnType<MatrixBase<T> >)
	{
#ifdef CHECK_UNSAFE_BOUNDS
		if (A.dim3 > 1)
			Q_RAISE_ERROR("Cannot convert cube to a matrix: the dimensions don't match!");
#endif
		return MatrixBase<T>(A.data, A.dim1, A.dim2);
	}

    //=============================================================================
    // High-level operations
    //=============================================================================

    template <typename T, T op(T, T)>
    __device__ T sequential_reduction(VectorBase<T> A)
    {
        T total = A.dim1 == 0 ? T() : A.data[0];
        for (int i = 1; i < A.dim1; i++)
            total = op(total, A.data[i]);
        return total;
    }

    template <typename T, T op(T, T)>
    __device__ T sequential_reduction(MatrixBase<T> A)
    {
        T total = A.dim1 == 0 ? T() : A.data[0];
        for (int i = 1; i < A.dim1 * A.dim2; i++)
            total = op(total, A.data[i]);
        return total;
    }

    template <typename T, T op(T, T)>
    __device__ T sequential_reduction(CubeBase<T> A)
    {
        T total = A.dim1 == 0 ? T() : A.data[0];
        for (int i = 1; i < A.dim1 * A.dim2 * A.dim3; i++)
            total = op(total, A.data[i]);
        return total;
    }

    template <typename T> __device__ T sum(VectorBase<T> a)
    {
        return sequential_reduction<T, OP_add>(a);
    }
    template <typename T> __device__ T sum(MatrixBase<T> a)
    {
        return sequential_reduction<T, OP_add>(a);
    }
    template <typename T> __device__ T sum(CubeBase<T> a)
    {
        return sequential_reduction<T, OP_add>(a);
    }

    template <typename T> __device__ T mean(VectorBase<T> a)
    {
        return sequential_reduction<T, OP_add>(a) / a.dim1;
    }
    template <typename T> __device__ T mean(MatrixBase<T> a)
    {
        return sequential_reduction<T, OP_add>(a) / (a.dim1 * a.dim2);
    }
    template <typename T> __device__ T mean(CubeBase<T> a)
    {
        return sequential_reduction<T, OP_add>(a) / (a.dim1 * a.dim2 * a.dim3);
    }

    template <typename T> __device__ T prod(VectorBase<T> a)
    {
        return sequential_reduction<T, OP_mul>(a);
    }
    template <typename T> __device__ T prod(MatrixBase<T> a)
    {
        return sequential_reduction<T, OP_mul>(a);
    }
    template <typename T> __device__ T prod(CubeBase<T> a)
    {
        return sequential_reduction<T, OP_mul>(a);
    }

    using ::min; // Enable overloading for these functions
    using ::max;

    template <typename T> __device__ T min(VectorBase<T> a)
    {
        return sequential_reduction<T, OP_min>(a);
    }
    template <typename T> __device__ T min(MatrixBase<T> a)
    {
        return sequential_reduction<T, OP_min>(a);
    }
    template <typename T> __device__ T min(CubeBase<T> a)
    {
        return sequential_reduction<T, OP_min>(a);
    }

    template <typename T> __device__ T max(VectorBase<T> a)
    {
        return sequential_reduction<T, OP_max>(a);
    }
    template <typename T> __device__ T max(MatrixBase<T> a)
    {
        return sequential_reduction<T, OP_max>(a);
    }
    template <typename T> __device__ T max(CubeBase<T> a)
    {
        return sequential_reduction<T, OP_max>(a);
    }

    template <typename T> __device__ T dotprod(VectorBase<T> A, VectorBase<T> B)
    {
        T total = T();
#ifdef CHECK_UNSAFE_BOUNDS
        if (A.dim1 != B.dim1)
            Q_RAISE_ERROR("Vector dimensions don't match");
#endif
        for (int i = 0; i < A.dim1; i++)
            total += A.data[i] * B.data[i];
        return total;
    }

    template <typename T> __device__ T dotprod(MatrixBase<T> A, MatrixBase<T> B)
    {
        T total = T();
#ifdef CHECK_UNSAFE_BOUNDS
        if (A.dim1 != B.dim1 || A.dim2 != B.dim2)
            Q_RAISE_ERROR("Matrix dimensions don't match");
#endif
        for (int i = 0; i < A.dim1 * A.dim2; i++)
            total += A.data[i] * B.data[i];
        return total;
    }

    template <typename T> __device__ T dotprod(CubeBase<T> A, CubeBase<T> B)
    {
        T total = T();
#ifdef CHECK_UNSAFE_BOUNDS
        if (A.dim1 != B.dim1 || A.dim2 != B.dim2 || A.dim3 != B.dim3)
            Q_RAISE_ERROR("Cube dimensions don't match");
#endif
        for (int i = 0; i < A.dim1 * A.dim2 * A.dim3; i++)
            total += A.data[i] * B.data[i];
        return total;
    }

#ifdef ENABLE_DYNAMIC_KERNEL_MEM

    // Transpose function
    template <typename T>
    __device__ __HOST_V__ MatrixBase<T> transpose(MatrixBase<T> x, const char *elemT)
    {
        MatrixBase<T> y = uninit<T>(elemT, x.dim2, x.dim1);
        for (int m = 0; m < y.dim1; m++)
            for (int n = 0; n < y.dim2; n++)
                y.data[m * y.dim2 + n] = x.data[n * x.dim2 + m];
        return y;
    }

    // Transpose function for vectors (the result is a matrix!)
    template <typename T>
    __device__ __HOST_V__ MatrixBase<T> transpose(VectorBase<T> x, const char *elemT)
    {
		MatrixBase<T> y = uninit<T>(elemT, x.dim1, 1);
        for (int n = 0; n < x.dim1; n++)
            y.data[n] = x.data[n];
        return y;
    };

    // Hermitian transpose function
    template <typename T>
    __device__ __HOST_V__ MatrixBase<T> herm_transpose(MatrixBase<T> x, const char *elemT)
    {
        MatrixBase<T> y = uninit<T>(elemT, x.dim2, x.dim1);
        for (int m = 0; m < y.dim2; m++)
            for (int n = 0; n < y.dim1; n++)
                y.data[m * x.dim1 + n] = conj(x.data[n * x.dim2 + m]);
        return y;
    }

    // Eye function
    INLINE __device__ __HOST_V__ Matrix eye(int N)
    {
        Matrix y = uninit<scalar>("scalar", N, N);
        for (int m = 0; m < N; m++)
            for (int n = 0; n < N; n++)
                y.data[m * N + n] = (m == n);
        return y;
    }

    // Copy function
    template <typename T>
    __device__ __HOST_V__ VectorBase<T> copy(VectorBase<T> x, const char *elemT)
    {
        VectorBase<T> y = uninit<T>(elemT, x.dim1);
        for (int m = 0; m < x.dim1; m++)
            y.data[m] = x.data[m];
        return y;
    }

    template <typename T>
    __device__ __HOST_V__ MatrixBase<T> copy(MatrixBase<T> x, const char *elemT)
    {
        MatrixBase<T> y = uninit<T>(elemT, x.dim1, x.dim2);
        for (int m = 0; m < x.dim1 * x.dim2; m++)
            y.data[m] = x.data[m];
        return y;
    }

    template <typename T>
    __device__ __HOST_V__ CubeBase<T> copy(CubeBase<T> x, const char *elemT)
    {
        CubeBase<T> y = uninit<T>(elemT, x.dim1, x.dim2, x.dim3);
        for (int m = 0; m < x.dim1 * x.dim2 * x.dim3; m++)
            y.data[m] = x.data[m];
        return y;
    }

    // Reshape function (all possible combinations)
    template <typename T>
    __device__ __HOST_V__ VectorBase<T> reshape(VectorBase<T> a, int dim1,
												const char *elemT)
    {
#ifdef CHECK_UNSAFE_BOUNDS
        if (dim1 != a.dim1)
            Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
        return copy(a, elemT);
    }

    template <typename T>
    __device__ __HOST_V__ MatrixBase<T> reshape(VectorBase<T> a, int2 dims,
												const char *elemT)
    {
#ifdef CHECK_UNSAFE_BOUNDS
        if (dims.x * dims.y != a.dim1)
            Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
        return MatrixBase<T>(copy(a, elemT).data, dims.x, dims.y);
    }

    template <typename T>
    __device__ __HOST_V__ CubeBase<T> reshape(VectorBase<T> a, int3 dims,
											const char *elemT)
    {
#ifdef CHECK_UNSAFE_BOUNDS
        if (dims.x * dims.y * dims.z != a.dim1)
            Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
        return CubeBase<T>(copy(a, elemT).data, dims.x, dims.y, dims.z);
    }

    template <typename T>
    __device__ __HOST_V__ VectorBase<T> reshape(MatrixBase<T> a, int dim1,
												const char *elemT)
    {
#ifdef CHECK_UNSAFE_BOUNDS
        if (dim1 != a.dim1 * a.dim2)
            Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
        return VectorBase<T>(copy(a, elemT).data, dim1);
    }

    template <typename T>
    __device__ __HOST_V__ VectorBase<T> reshape(CubeBase<T> a, int dim1, const char *elemT)
    {
#ifdef CHECK_UNSAFE_BOUNDS
        if (dim1 != a.dim1 * a.dim2 * a.dim3)
            Q_RAISE_ERROR("Reshape: the number of elements must not change!");
#endif
        return VectorBase<T>(copy(a, elemT).data, dim1);
    }

    template <typename T>
    __device__ __HOST_V__ MatrixBase<T> reshape(MatrixBase<T> a, int2 dims,
												const char *elemT)
    {
        return MatrixBase<T>(reshape(a, a.dim1 * a.dim2, elemT).data, dims.x, dims.y);
    }
    template <typename T>
    __device__ __HOST_V__ MatrixBase<T> reshape(CubeBase<T> a, int2 dims,
												const char *elemT)
    {
        return MatrixBase<T>(reshape(a, a.dim1 * a.dim2 * a.dim3, elemT).data, dims.x,
                       dims.y);
    }
    template <typename T>
    __device__ __HOST_V__ CubeBase<T> reshape(MatrixBase<T> a, int3 dims,
											  const char *elemT)
    {
        return CubeBase<T>(reshape(a, a.dim1 * a.dim2, elemT).data, dims.x, dims.y, dims.z);
    }
    template <typename T>
    __device__ __HOST_V__ CubeBase<T> reshape(CubeBase<T> a, int3 dims,
											  const char *elemT)
    {
        return CubeBase<T>(reshape(a, a.dim1 * a.dim2 * a.dim3, elemT).data, dims.x, dims.y, dims.z);
    }

    // Seq function
    INLINE __device__ __HOST_V__  Vector seq(scalar a, scalar b, scalar step)
    {
        int count = (int)floor((b - a) / step + 1);
        Vector y = uninit<scalar>("scalar", count);
        for (int n = 0; n < count; n++)
            y.data[n] = a + n * step;
        return y;
    }

    // Seq function
    INLINE __device__ __HOST_V__ Vector seq(int a, int b, int step)
    {
        int count = step == 1 ? b - a + 1 : ((b - a) / step + 1);
        Vector y = uninit<scalar>("scalar", count);
        for (int n = 0; n < count; n++)
            y.data[n] = a + n * step;
        return y;
    }

    // Linspace function
    INLINE __device__ __HOST_V__ Vector linspace(scalar a, scalar b, int count)
    {
        Vector y = uninit<scalar>("scalar", count);
        scalar delta = b - a;
        if (count == 1)
            y.data[0] = a;
        else
        {
            for (int n = 0; n < count;
                 n++) // Note: division because it is the most accurate
                y.data[n] = a + (delta * (scalar)n) / (count - 1);
        }
        return y;
    }

#endif // ENABLE_DYNAMIC_KERNEL_MEM

	// Runtime load balancer for OpenMP: determines the number of CPU threads
	// based on the data dimensions and the kernel complexity.
	INLINE __device__ __HOST_V__ int runtime_loadbalancer(int3 blkDim, int3 gridDim,
		int complexity_level)
	{
		int total = prod(gridDim * blkDim);
		// data dimensions too small or the complexity level too low...
		if ((total < 32768 && complexity_level < 5) || total < 4096)
		{
			return 1;
		}
#ifdef _OPENMP
		return omp_get_max_threads();
#else
		// Compiling without OpenMP...
		return 1;
#endif
	}


#if defined(TARGET_CUDA)

    INLINE __device__ int cuda_calculate_blkdim(int dims)
    {
        return min(1024, NextPow2(dims));
    }

    INLINE __device__ int2 cuda_calculate_blkdim(int2 dims)
    {
        int x = NextPow2(dims.x);
        int y = NextPow2(dims.y);
        while (x * y > 32 && y > 1)
            y /= 2;
        while (x * y > 32 && x > 1)
            x /= 2;
        return make_int2(x, y);
    }

    INLINE __device__ int3 cuda_calculate_blkdim(int3 dims)
    {
        int x = NextPow2(dims.x);
        int y = NextPow2(dims.y);
        int z = NextPow2(dims.z);
        while (x * y * z > 32 && z > 1)
            z /= 2;
        while (x * y * z > 32 && y > 1)
            y /= 2;
        while (x * y * z > 32 && x > 1)
            x /= 2;
        return make_int3(x, y, z);
    }

    INLINE __device__ dim3 make_dim3(int dim) { return dim3(dim, 1, 1); }
    INLINE __device__ dim3 make_dim3(int2 dim) { return dim3(dim.x, dim.y, 1); }
    INLINE __device__ dim3 make_dim3(int3 dim)
    {
        return dim3(dim.x, dim.y, dim.z);
    }

#endif


#if defined(TARGET_CUDA) && defined(USE_COOPERATIVE_THREADING)
	//
	// cg::coalesced_group vector extensions
	//
	struct coalesced_group_ext : public cooperative_groups::coalesced_group
	{
        __device__ __HOST_V__ coalesced_group_ext(cooperative_groups::coalesced_group &&other) :
            cooperative_groups::coalesced_group(std::move(other))
        {
        }

        template <typename T>
        INLINE __device__ T shfl(T a, int n) { return coalesced_group::shfl(a, n); }
        template <typename T>
        INLINE __device__ T shfl_down(T a, int n) { return coalesced_group::shfl_down(a, (unsigned int)n); }
        template <typename T>
        INLINE __device__ T shfl_up(T a, int n) { return coalesced_group::shfl_up(a, (unsigned int)n); }
        template <typename T>
        INLINE __device__ T shfl_xor(T a, int n) { return coalesced_group::shfl_xor(a, (unsigned int)n); }

        INLINE __device__ cscalar shfl(cscalar a, int n)
        {
            return make_cscalar(coalesced_group::shfl(a.x, n), coalesced_group::shfl(a.y, n));
        }

        INLINE __device__ cscalar shfl_down(cscalar a, int n)
        {
            return make_cscalar(coalesced_group::shfl_down(a.x, n), coalesced_group::shfl_down(a.y, (unsigned int)n));
        }

        INLINE __device__ cscalar shfl_up(cscalar a, int n)
        {
            return make_cscalar(coalesced_group::shfl_up(a.x, n), coalesced_group::shfl_up(a.y, (unsigned int)n));
        }

        /*INLINE __device__ cscalar shfl_xor(cscalar a, int n)
        {
            return make_cscalar(coalesced_group::shfl_xor(a.x, n), coalesced_group::shfl_xor(a.y, (unsigned int)n));
        }*/

		
		// Vector-based shuffle
		template <typename T, int N>
		INLINE __device__ typename quasar::VecType<T, N>::type shfl(
			typename quasar::VecType<T, N>::type a, int n)
		{
			typename quasar::VecType<T, N>::type b;
			FOR_ALL((vecX_get_ref_at<T, N, Unchecked>(b, i)) = (coalesced_group::shfl(vecX_get_at<T, N, Unchecked>(a, i), n)));
			return b;
		}
		
		template <typename T, int N>
		INLINE __device__ typename quasar::VecType<T, N>::type shfl_down(
			typename quasar::VecType<T, N>::type a, int n)
		{
			typename quasar::VecType<T, N>::type b;
			FOR_ALL((vecX_get_ref_at<T, N, Unchecked>(b, i)) = (coalesced_group::shfl_down(vecX_get_at<T, N, Unchecked>(a, i), (unsigned int)n)));
			return b;
		}
		
		template <typename T, int N>
		INLINE __device__ typename quasar::VecType<T, N>::type shfl_up(
			typename quasar::VecType<T, N>::type a, int n)
		{
			typename quasar::VecType<T, N>::type b;
			FOR_ALL((vecX_get_ref_at<T, N, Unchecked>(b, i)) = (coalesced_group::shfl_up(vecX_get_at<T, N, Unchecked>(a, i), (unsigned int)n)));
			return b;
		}
		
		template <typename T, int N>
		INLINE __device__ typename quasar::VecType<T, N>::type shfl_xor(
			typename quasar::VecType<T, N>::type a, int n)
		{
			typename quasar::VecType<T, N>::type b;
			FOR_ALL((vecX_get_ref_at<T, N, Unchecked>(b, i)) = (coalesced_group::shfl_xor(vecX_get_at<T, N, Unchecked>(a, i), (unsigned int)n)));
			return b;
		}
	};
#endif

#if defined(TARGET_CUDA) && defined(USE_TENSOR_CORE)

    #include <mma.h>
    using namespace nvcuda;
    
    template <typename Use, int m, int n, int k, typename T, typename Layout=void>
    struct tensor_fragment
    {        
        wmma::fragment<Use, m, n, k, T, Layout> frag;

        INLINE __device__ tensor_fragment()
        {
        }

        INLINE __device__ void set(T value) 
        {
           wmma::fill_fragment(frag, value);
        }

        INLINE __device__ void load_matrix(quasar::VectorBase<T> vector, int start_index, int ldm) 
        {            
            wmma::load_matrix_sync(frag, (T*)&vector.data[start_index], ldm);
        }

        template <int NR, int NC>
        INLINE __device__ void load_matrix(quasar::MatrixBase<T> matrix, intN<NR> rowSlice, intN<NC> colSlice) 
        {           
            int r = rowSlice.el[0];
            int c = colSlice.el[0];  
            wmma::load_matrix_sync(frag, (T*)&matrix.data[r*matrix.dim2 + c], matrix.dim2*(rowSlice.el[1]-rowSlice.el[0]));
        }

        template <int NR, int NC>
        INLINE __device__ void load_matrix(quasar::CubeBase<T> matrix, intN<NR> rowSlice, intN<NC> colSlice, intN<NC> depSlice) 
        {           
            int r = rowSlice.el[0];
            int c = colSlice.el[0];  
            int k = depSlice.el[0];  
        }

        INLINE __device__ void store_matrix(quasar::VectorBase<T> vector, int start_index, int ldm) 
        {
            wmma::store_matrix_sync((T*)&vector.data[start_index], frag, ldm, (wmma::layout_t)mem_layout);
        }

        template <int NR,int NC, int mem_layout >
        INLINE __device__ void store_matrix(quasar::MatrixBase<T> matrix, intN<NR> rowSlice, intN<NC> colSlice) 
        {
            int r = rowSlice.el[0];
            int c = colSlice.el[0]; 
            wmma::store_matrix_sync((T*)&matrix.data[r*matrix.dim2 + c], frag, matrix.dim2*(rowSlice.el[1]-rowSlice.el[0]), (wmma::layout_t)mem_layout);
        }


        template <int NR,int NC, int mem_layout >
        INLINE __device__ void store_matrix(quasar::CubeBase<T> cube, intN<NR> rowSlice, intN<NC> colSlice, intN<NC> depSlice) 
        {
            int r = rowSlice.el[0];
            int c = colSlice.el[0]; 
            int k = depSlice.el[0]; 
        }

        template <typename TA, typename TB, typename LayoutA, typename LayoutB>
        INLINE __device__ void multiply_accumulate_matrix(tensor_fragment<wmma::accumulator, m, n, k, T, Layout> acc_frag, tensor_fragment<wmma::matrix_a, m, n, k, TA, LayoutA> a_frag, tensor_fragment<wmma::matrix_b, m, n, k, TB, LayoutB> b_frag)
        {           
            wmma::mma_sync(frag, a_frag.frag, b_frag.frag, acc_frag.frag);
        }
    };
    
    struct tensor_core
    {
        template <typename Use, int m, int n, int k, typename T, typename Layout>
        INLINE __device__ tensor_fragment<Use, m, n, k, T, Layout> alloc() 
        {            
            return tensor_fragment<Use, m, n, k, T, Layout>();
        }
        
        template <typename Use, int m, int n, int k, typename T, typename Layout, int len>
        INLINE __device__ quasar::VectorBase<tensor_fragment<Use, m, n, k, T, Layout>> alloc_vec() 
        {   
            tensor_fragment<Use, m, n, k, T, Layout> frags[len];
            return quasar::VectorBase<tensor_fragment<Use, m, n, k, T, Layout>>(frags, len);
        }
        
        template <typename Use, int m, int n, int k, typename T, typename Layout, int dim1, int dim2>
        INLINE __device__ quasar::MatrixBase<tensor_fragment<Use, m, n, k, T, Layout>> alloc_mat() 
        {   
            tensor_fragment<Use, m, n, k, T, Layout> frags[dim1*dim2];
            return quasar::MatrixBase<tensor_fragment<Use, m, n, k, T, Layout>>(frags, dim1, dim2);
        }

        template <typename Use, int m, int n, int k, typename T>
        INLINE __device__ tensor_fragment<Use, m, n, k, T> alloc() 
        {            
            return tensor_fragment<Use, m, n, k, T>();
        }
        
        template <typename Use, int m, int n, int k, typename T, int len>
        INLINE __device__ quasar::VectorBase<tensor_fragment<Use, m, n, k, T>> alloc_vec() 
        {            
            tensor_fragment<Use, m, n, k, T> frags[len];
            return quasar::VectorBase<tensor_fragment<Use, m, n, k, T>>(frags, len);
        }
        
        template <typename Use, int m, int n, int k, typename T, int dim1, int dim2>
        INLINE __device__ quasar::MatrixBase<tensor_fragment<Use, m, n, k, T>> alloc_mat() 
        {            
            tensor_fragment<Use, m, n, k, T> frags[dim1*dim2];
            return quasar::MatrixBase<tensor_fragment<Use, m, n, k, T>>(frags, dim1, dim2);
        }
    };
    
#endif

} // namespace quasar


#endif
