%==============================================================================
% colortransform.q
% Per-pixel color transforms
%==============================================================================
import "system.q"
import "imfilter.q"
import "linalg.q"

{!author name="Bart Goossens"}
{!doc category="Image Processing/Colors"}

% Section: colortransforms.q
% Various utility functions for performing color conversions/transformations/etc.

% Function: is_rgb
% 
% Returns true only if the image is an RGB image
is_rgb = (img : cube) -> (size(img,2)==3 || size(img,2)==4)

% Function: is_gray
%
% Returns true only if the image is a grayscale image
is_gray = (img : cube) -> size(img,2)==1

% Function: color_cov
%
% Computes the color band covariance matrix of an image.
function C = color_cov(img : cube, zero_mean = 0)
    p = size(img, 2)
    C = zeros(p * [1,1])
    img_copy = copy(img)
    % subtract the mean...
    if !zero_mean
        img_copy -= repmat(mean(img_copy, 2),[1,1,p])
    endif

    % exploit the symmetry to reduce the number of calculations
    for m=0..p-1
        for n=m..p-1
            C[m,n] = C[n,m] = sum(img_copy[:,:,m] .* img_copy[:,:,n])        
        endfor
    endfor
endfunction

% Function: PCA_color_transform
%
% Computes the color decorrelating transform using Principal Component Analysis
%   (PCA) in the L2 sense
%
% : [U, Ut] = PCA_color_transform(img)
%
% Parameters:
% img - the input image
% U - The decorrelating transform
% Ut - The transpose of the decorrelating transform
function [U, Ut] = PCA_color_transform(img)
    if size(img,2)>1
        img_f = imfilter(img, [[0,-1,0],[-1,4,-1],[0,-1,0]])
        C = color_cov(img_f, true)
        [U,Lambda,V] = svd(C) % Then U is the decorrelating transform
        Ut = transpose(U)
    else
        U = 1
        Ut = 1
    endif
endfunction

% Function: L1PCA_color_transform
%
% Computes the color decorrelating transform using Principal Component Analysis
%   (PCA) in the L1 sense
%
% : [U, Ut] = L1PCA_color_transform(img : cube)
%
% Parameters:
% img - the input image (variable number of channels)
% U - The decorrelating transform
% Ut - The transpose of the decorrelating transform
function [U, Ut] = L1PCA_color_transform(img : cube)
    p = size(img,2)
    if p==1 % Grayscale image input
        U = Ut = 1
        return
    endif

    img_f = imfilter(img, [[0,-1,0],[-1,4,-1],[0,-1,0]])    
    C = color_cov(img_f, true)
    [U,Lambda,V] = svd(C) % Then U is the decorrelating transform

    % Function for computing the cross-covariance between two
    % images
    function C : mat = cross_cov(A : cube, B : cube)
        N = size(A,2)
        
        function [C : mat(N,N)] = __kernel__ kernel[N:int](A : cube'unchecked(:,:,N), B : cube'unchecked(:,:,N), pos : ivec2)
            Av : vec'col(N) = squeeze(A[...pos,:])
            Bv : vec'col(N) = squeeze(B[...pos,:])

            C += Av * transpose(Bv)
        endfunction
        
        match N with
        | 3 ->
            C = parallel_do(size(A,0..1),A,B,kernel[3])
        | _ ->
            C = zeros(N,N)
            for m=0..N-1
                for n=m..N-1
                    C[m,n] = sum(A[:,:,m] .* B[:,:,n])        
                endfor
            endfor
            
            % Symmetry extension        
            {!parallel for}
            for m=1..N-1
                for n=0..N-1
                    if n<m
                        C[m,n]=C[n,m]
                    endif
                endfor
            endfor
        endmatch

    endfunction

    function z : mat = L1PCA_seed(t : cube, V : mat)
        epsilon = 1e-6 % epsilon for single precision floating point
        epsilon2 = 1e-2
        flag = 0
        norm = x : mat -> sqrt(sum(x.^2))
        z = V ./ norm(V)
    
        while flag < 1
            % polarity check
            w = colortransform(z, t)
            w2 = ones(size(w)) - 2.*(w<0)
            % flipping and maximization
            z2 = cross_cov(w2,t)
            z2 /= norm(z2)

            % convergence checks
            if sum(abs(z-z2)) < epsilon
                w2 = sum(abs(w2))
                w3 = sum(w2 == 0)
                if w3 > 0
                    z2 = z2 + epsilon2.*(rand(size(z2))-0.5)
                    z2 /= norm(z2)
                else
                    flag = 1
                endif
            endif
            z = z2
        endwhile
    endfunction

    z = V
    for i=0..p-1
        z2 = L1PCA_seed(img_f, V)
        z[i,:] = z2[i,:]
        img_f = img_f - colortransform(transpose(z2)*z2, img_f)
    endfor
    U = z*3
    Ut = transpose(z)
endfunction

% Function: get_colortransform_from_RGB
% Get the specified color transform [Note - only a couple of linear
% color transforms are supported here]
%
% : [U, Ut] = get_colortransform_from_RGB(colorspace, img = 0)
%
% Parameters:
% colorspace - supported color spaces are "RGB", "PCA", "YCbCr", "YUV"
% img - the input image (in case of "PCA")
% U - The forward color transform
% Uinv - The inverse color transform
function [U : mat(3,3), Uinv : mat(3,3)] = get_colortransform_from_RGB(colorspace : string, img = 0)
    match colorspace with
    | "RGB" ->
        U = Uinv = eye(3)        
    | "PCA" ->
        [U, Uinv] = PCA_color_transform(img)
    | "YCbCr" ->
        U = [[0.299, 0.587, 0.114],
             [-0.168736, -0.331264, 0.5],
             [0.5, -0.418688, -0.081312]]
        Uinv = inv(U)
    | "YUV" ->
        U = [[0.229, 0.587, 0.114],
             [-0.14713, -0.28886, 0.436],
             [0.615, -0.51499, -0.10001]]
        Uinv = inv(U)
    | "xyz" ->
        U = [[0.412453, 0.357580, 0.180423],
             [0.212671, 0.715160, 0.072169],
             [0.019334, 0.119193, 0.950227]]
        Uinv = inv(U)
    % Haematoxylin-Eosin-DAB colorspace
    % From original Ruifrok's paper: A. C. Ruifrok and D. A. Johnston,
    % "Quantification of histochemical staining by color deconvolution.,"
    % Analytical and quantitative cytology and histology / the International
    % Academy of Cytology [and] American Society of Cytology, vol. 23, no. 4,
    % pp. 291-9, Aug. 2001.        
    | "hed" ->
        Uinv = [[0.65, 0.70, 0.29],
                [0.07, 0.99, 0.11],
                [0.27, 0.57, 0.78]]
        U = inv(Uinv)
    % Hematoxylin + DAB
    | "hdx" ->
        Uinv = [[0.650, 0.704, 0.286],
                [0.268, 0.570, 0.776],
                [0.0, 0.0, 0.0]]
        U = inv(Uinv)             
    | _ ->
        error("Undefined color space!")
    endmatch
endfunction

% Function: colortransform
% Apply the specified color transform to an image
%
% : function y : cube = colortransform(C : [mat|scalar], x : cube)
%
% Parameters:
% C - a 3x3 color transform matrix (or a scalar value)
% x - an input image image of dimensions MxNx3
% y - the color transformed output image
function y : cube = colortransform(C : [mat|scalar], x : cube)

    if isscalar(C)
        y = C * x
        return
    endif

    if size(x,2) != size(C,0) || size(x,2) != size(C,1)
        error "Color transform does not have the correct dimensions!"
    endif

    y = uninit(size(x))

    function [] = __kernel__ colortransform_general_kernel(x:cube'unchecked, y:cube'unchecked, _
        C : mat'hwconst, pos : ivec3)
        m = pos[2]
        sum = 0.0
        for n=0..size(x,2)-1
            sum += C[m,n] * x[pos[0],pos[1],n]
        endfor
        y[pos] = sum
    endfunction
    
    function [] = __kernel__ colortransform_detail_kernel[N:int](x : cube'unchecked(:,:,N), _
        y : cube'unchecked(:,:,N), C : mat(N,N), pos : ivec2)
        
        total = zeros(N)
        for m=0..N-1
            for n=0..N-1
                total[m] += C[m,n] * x[pos[0],pos[1],n]
            endfor
        endfor
        y[...pos,:] = total
    endfunction
    
    match size(x,2) with
    | 3 ->
        parallel_do(size(x,0..1),x,y,C,colortransform_detail_kernel[3])
    | 4 ->
        parallel_do(size(x,0..1),x,y,C,colortransform_detail_kernel[4])
    | _ ->
        parallel_do(size(x),x,y,C, colortransform_general_kernel)
    endmatch

endfunction

% Function: rgb2gray
% Conversion from RGB to grayscale
%
% : function y : mat = rgb2gray(x : cube)
function y : mat = rgb2gray(x : cube)
    if size(x,2)==1
        y = x[:,:,0] % already grayscale image
    else
        y = 0.2126 * x[:,:,0] + 0.7152 * x[:,:,1] + 0.0722 * x[:,:,2]
    endif
endfunction

% Function: gray2rgb
% Conversion from grayscale to RGB
%
% : function y : cube(:,:,3) = gray2rgb(x : cube)
function y : cube(:,:,3) = gray2rgb(x : cube)
    if size(x,2)==3
        y = x % already RGB image
    else
        y = repmat(x,[1,1,3])
    endif
endfunction

% Function: compute_hue
% Computes the HUE value of a given RGB color value
%
% : function h : scalar = __device__ compute_hue (c : vec3)
function h : scalar = __device__ compute_hue (c : vec3)
    cmax = max(c)
    cmin = min(c)
    if cmax == cmin
        h = 0.0
    elseif cmax == c[0]
        h = (c[1] - c[2])/(6*(cmax-cmin))
    elseif cmax == c[1]
        h = (c[2] - c[0])/(6*(cmax-cmin)) + 1.0/3
    else
        h = (c[0] - c[1])/(6*(cmax-cmin)) + 2.0/3
    endif
    h = mod(h + 1.0, 1.0) % Output hue value is in range 0-1
endfunction

% Function: rgb2hsv_devicefunc
% Conversion from RGB to Hue-Saturation-Value (HSV)
%
% : function [y : vec3] = __device__ rgb2hsv_devicefunc (c : vec3)
function [y : vec3] = __device__ rgb2hsv_devicefunc (c : vec3)
    cmax = max(c)
    cmin = min(c)
    hue = compute_hue(c)  % hue
    sat = cmax == 0.0 ? 0.0 : 1.0 - cmin / cmax % saturation 
    val = cmax / 255.0 % value
    y = [hue, sat, val]
endfunction

% Function: rgb2hsv_devicefunc
% Conversion from RGB to Hue-Saturation-Luminance (HSL)
%
% : function [y : vec3] = __device__ rgb2hsv_devicefunc (c : vec3)
function [y : vec3] = __device__ rgb2hsl_devicefunc (c : vec3)
    cmax = max(c)/255.0
    cmin = min(c)/255.0
    hue = compute_hue(c)  % hue
    sat = cmax - cmin % saturation 
    lightness = 0.5 * (cmin + cmax)  % lightness
    y = [hue, sat, lightness]
endfunction

% Function: rgb2hsv
% Conversion from RGB to Hue-Saturation-Value (HSV)
%
% : function y = rgb2hsv(x : cube(:,:,3))
function y = rgb2hsv(x : cube(:,:,3))
    y = uninit(size(x))
    parallel_do(size(x,0..1),x,y,__kernel__(x : cube'unchecked, y : cube'unchecked, pos : ivec2) -> _
        y[pos[0],pos[1],0..2] = rgb2hsv_devicefunc(x[pos[0],pos[1],0..2]))
endfunction

% Function: rgb2hsl
% Conversion from RGB to Hue-Saturation-Luminance (HSL)
%
% : function y = rgb2hsl(x : cube(:,:,3))
function y = rgb2hsl(x : cube(:,:,3))
    y = uninit(size(x))
    parallel_do(size(x,0..1),x,y,__kernel__(x : cube'unchecked, y : cube'unchecked, pos : ivec2) -> _
        y[pos[0],pos[1],0..2] = rgb2hsl_devicefunc(x[pos[0],pos[1],0..2]))
endfunction

% Function: rgb2hsv_devicefunc
% Conversion from Hue-Saturation-Value (HSV) to RGB
%
% : function [y : vec3] = __device__ hsv2rgb_devicefunc (c : vec3)
function [y : vec3] = __device__ hsv2rgb_devicefunc (c : vec3)
    h = int(c[0] * 6.0)
    f = frac(c[0] * 6.0)
    v = 255.0 * c[2]    
    p = v * (1 - c[1])
    q = v * (1 - f * c[1])
    t = v * (1 - (1 - f) * c[1])
    match h with
    | 0 -> y = [v, t, p]
    | 1 -> y = [q, v, p]
    | 2 -> y = [p, v, t]
    | 3 -> y = [p, q, v]
    | 4 -> y = [t, p, v]
    | _ -> y = [v, p, q]
    endmatch
endfunction

% Function: hsv2rgb
% Conversion from Hue-Saturation-Value (HSV) to RGB
%
% : function y = __device__ hsv2rgb (x : cube(:,:,3))
function y = hsv2rgb(x : cube(:,:,3))
    y = uninit(size(x))
    parallel_do(size(x,0..1),x,y,__kernel__(x : cube'unchecked, y : cube'unchecked, pos : ivec2) -> _
        y[pos[0],pos[1],0..2] = hsv2rgb_devicefunc(x[pos[0],pos[1],0..2]))
endfunction

% Function: hsl2rgb_devicefunc
% Conversion from Hue-Saturation-Lightness (HSL) to RGB
%
% : function [y : vec3] = __device__ hsl2rgb_devicefunc (c : vec3)
function [y : vec3] = __device__ hsl2rgb_devicefunc (c : vec3)
    h = c[0]
    s = 255.0 * c[1]
    l = 255.0 * c[2]   
    cmax = l + 0.5*s
    cmin = l - 0.5*s
    s = cmax == 0.0 ? 0.0 : 1.0 - cmin / cmax % saturation 
    val = cmax / 255.0
    y = hsv2rgb_devicefunc([h, s, val])
endfunction

% Function: hsl2rgb
% Conversion from Hue-Saturation-Lightness (HSL) to RGB
%
% : function y = hsl2rgb(x : cube(:,:,3))
function y = hsl2rgb(x : cube(:,:,3))
    y = uninit(size(x))
    parallel_do(size(x,0..1),x,y,__kernel__(x : cube'unchecked, y : cube'unchecked, pos : ivec2) -> _
        y[pos[0],pos[1],0..2] = hsl2rgb_devicefunc(x[pos[0],pos[1],0..2]))
endfunction

% Function: rgb2lab
% Conversion from RGB to the CIELAB color space
%
% : function Lab = rgb2lab(img:cube(:,:,3))
% 
% Parameters
% - img: the RGB input image
% - Lab: the CIELAB output image
function Lab = rgb2lab(img:cube'unchecked(:,:,3))
    Lab:'unchecked = uninit(size(img))
    scale = 1.0
    if(max(img)>2)%if we take 1, then due to inaccuracies we may sometimes divide when unnecessary
        scale = 255.0
    endif
    % Set a threshold
    T = 0.008856
    X=__device__ (r:scalar, g:scalar, b:scalar) -> (r*0.412453 + g*0.357580 + b*0.180423)/0.950546;
    Y=__device__ (r:scalar, g:scalar, b:scalar) -> (r*0.212671 + g*0.715160 + b*0.072169);
    Z=__device__ (r:scalar, g:scalar, b:scalar) -> (r*0.019334 + g*0.119193 + b*0.950227)/1.088754;
    f=__device__ (t:scalar, T:scalar) -> T*t^(1/3) + (1-T)*(t*7.787+16/116)
    function [] = __kernel__ color2lab(pos:ivec2)
        [r,g,b] = img[pos[0], pos[1], 0..2]/scale
        x = X(r,g,b)
        y = Y(r,g,b)
        z = Z(r,g,b)
        xt = x > T
        yt = y > T
        zt = z > T
        fX = f(x,xt)
        fY = f(y,yt)
        fZ = f(z,zt)
        Lab[pos[0], pos[1], 0]= yt*(116*y^(1/3)-16)+(1-yt)*(903.3*y)
        Lab[pos[0], pos[1], 1]= 500*(fX-fY)
        Lab[pos[0], pos[1], 2]= 200*(fY-fZ)
    endfunction
    parallel_do(size(img,0..1), color2lab)
endfunction

% Function: lab2rgb
% Conversion from the CIELAB color space to RGB
%
% : function img = lab2rgb(Lab:cube(:,:,3))
% 
% Parameters
% - Lab: the CIELAB input image
% - img: the RGB output image
function img = lab2rgb(Lab:cube'unchecked(:,:,3))
    img : 'unchecked = uninit(size(Lab))
    
    Xn = 0.950546
    Yn = 1.0
    Zn = 1.088754
    fi=__device__ (t:scalar) -> t>6/29 ? t^3 : 3*(6/29)^2*(t-4/29)
    r_c=__device__ (X:scalar, Y:scalar, Z:scalar) -> ( X*3.2404542 - Y*1.5371385 - Z*0.4985314)
    g_c=__device__ (X:scalar, Y:scalar, Z:scalar) -> (-X*0.9692660 + Y*1.8760108 + Z*0.0415560)
    b_c=__device__ (X:scalar, Y:scalar, Z:scalar) -> ( X*0.0556434 - Y*0.2040259 + Z*1.0572252)
    function [] = __kernel__ lab2color(pos:ivec2)
        [L,a,b] = Lab[pos[0], pos[1], 0..2]
        
        Y = Yn * fi((L+16)/116)
        X = Xn * fi((L+16)/116 + a/500)
        Z = Zn * fi((L+16)/116 - b/200)
        
        img[pos[0],pos[1],0..2] = [r_c(X,Y,Z), g_c(X,Y,Z), b_c(X,Y,Z)]
    endfunction
    parallel_do(size(img,0..1), lab2color)
endfunction

% Function: hsl_correct
% Performs Hue-Saturation-Lightness correction of an image
%
% : function [] = hsl_correct(img_in : cube(:,:,3), img_out : cube(:,:,3), add_hue : scalar, _
%        add_saturation : scalar, add_lightness : scalar)
%
% Parameters:
% img_in - input image
% img_out - output image
% add_hue - the amount of hue to add (hue values are in the range [0,1])
% add_saturation - the amount of saturation to add
% add_lightness - the amount of lightness to add
%
% Remarks:
% the values of add_hue, add_saturation, add_lightness can also be negative.
function [] = hsl_correct(img_in : cube(:,:,3), img_out : cube(:,:,3), add_hue : scalar, _
    add_saturation : scalar, add_lightness : scalar)
  
    function [] = __kernel__ HSL_kernel(img_in : cube'unchecked(:,:,3), img_out : cube'unchecked(:,:,3), _
        v : vec3, pos : ivec2)

        hslval = rgb2hsl_devicefunc(img_in[pos[0],pos[1],:])
        hslval[0] = mod(hslval[0] + v[0] + 1.0, 1.0)
        hslval[1] = hslval[1] * v[1]
        hslval[2] = hslval[2] * v[2]
        img_out[pos[0],pos[1],:] = hsl2rgb_devicefunc(hslval)
    endfunction

    parallel_do(size(img_in,0..2),img_in,img_out,[add_hue, 10^add_saturation, 10^add_lightness],HSL_kernel)
endfunction

function [] = main()
    im = imread("lena_big.tif")
    tic()
    [U,Ut] = PCA_color_transform(im)
    toc("PCA_color_transform")
    print U
    tic()
    [U,Ut] = L1PCA_color_transform(im)
    toc("L1PCA_color_transform")
    print U
    imshow(colortransform(-U,im),[])
    return

    y = rgb2hsl(im)
    print max(y[:,:,0]),",",max(y[:,:,1]),",",max(y[:,:,2])
    imshow(y,[])
    y = hsl2rgb(y)
    imshow(y)
    psnr = (x, y) -> 10*log10(255^2/mean((x-y).^2))
    print psnr(im,y)
endfunction