%==============================================================================
% connected_components.q
% Parallel implementation of connected components
%==============================================================================

{!author name="Bart Goossens"}
{!doc category="Image Analysis"}

% Section: connected_components.q
% This file contains a parallel implementation for connected components, using
% shared memory for fast processing.

% Function: connectedcomponents2d
%
% Labels the 4-connected components in an image. Works both for grayscale and 
% binary images, only connecting pixels with the exact same value.
%
% : function [labels] = connectedcomponents2d(x:mat)
%
% Parameters:
%   x - the input data
%   labels - output matrix with the same size of x. Pixels in the same component have the same label value, 
%            but label values are not guaranteed to (and in practice will not) be consecutive from 1.
%
function w = connectedcomponents2d(x : mat)   
    
    % The swizzle algorithm just detects horizontal transitions. Simultaneously
    % a reshaping operation is performed, similar to:
    % reshape(x, [size(x,0)*mp, ceil(size(x,1)/mp)]
    % The mp parameter (number of blocks processed in parallel) is there to
    % ensure that the GPU has enough workload (the risk is that the image has
    % a small height, e.g. 512), while recent GPUs often require 2048 or more
    % threads to be effective. Therefore, mp allows to exploit more parallelism    
    function [] = __kernel__ transition_and_swizzle(x : mat'safe, dx : mat'unchecked,
        mp : int, mpx : int, pos : ivec2)
        
        dx[pos[0]*mp+int(pos[1]/mpx), mod(pos[1],mpx)] = pos[1] == 0 || x[pos] != x[pos-[0,1]]
    end
    
    % The initial scan just assigns a new label whenever the input intensity 
    % changes. We use a sequential algorithm that is applied in parallel to 
    % different regions in the image.
    function [] = __kernel__ horscan_initial(x : mat'unchecked, w : mat'unchecked, 
        starting_label : vec, mp : int, mpx : int, pos : int)
        
        cur_val = -1.
        cur_label = starting_label[pos-1]-1
        p = int(pos / mp)
        idx = mod(pos, mp)
    
        for i=idx*mpx..min((idx+1)*mpx,size(x,1)-1)
            new_val = x[p,i]
            if new_val != cur_val
                % assign a new label
                cur_val = new_val
                cur_label += 1
            endif
            w[p,i] = cur_label
        end              
    end    
    
    % The parallel connected components algorithm just scans west, east,
    % north and south and connects regions until no changes are possible
    % anymore. Furthermore, we divide the image into overlapping blocks.
    % The blocks are just overlapping with one pixel and can be stored
    % in the shared memory of the GPU. The iterative updating process is
    % first repeated until no changes are possible within the block.
    % Cross-border region merging is then obtained by subsequent kernel
    % calls. Since most of the iterative operations are only performed in
    % shared memory, the resulting kernel function is usually quite fast.
    function [updated : int] = __kernel__ scan(x : mat, w : mat, 
        pos : ivec2, blkpos : ivec2, blkdim : ivec2)    
        
        % STEP 1 - cache the input variables in shared memory, as we will
        % likely need to iterate multiple times  
        assert((blkdim[0]+2) * (blkdim[1]+2) <= 2040)      
        xs : mat'unchecked = shared(blkdim[0]+2,blkdim[1]+2)
        ws : mat'unchecked = shared(blkdim[0]+2,blkdim[1]+2)
        changed : vec[int]'unchecked = shared[int](1)
        
        xs[blkpos] = x[pos-[1,1]]
        ws[blkpos] = w[pos-[1,1]]
        
        % Add the borders
        if blkpos[0] < 2
            xs[blkpos+[blkdim[0],0]] = x[pos+[blkdim[0],0]-[1,1]]
            ws[blkpos+[blkdim[0],0]] = w[pos+[blkdim[0],0]-[1,1]]
        endif
        if blkpos[1] < 2
            xs[blkpos+[0,blkdim[1]]] = x[pos+[0,blkdim[1]]-[1,1]]
            ws[blkpos+[0,blkdim[1]]] = w[pos+[0,blkdim[1]]-[1,1]]
        endif
        if max(blkpos) < 2
            xs[blkpos+blkdim] = x[pos+blkdim-[1,1]]
            ws[blkpos+blkdim] = w[pos+blkdim-[1,1]]
        endif
        syncthreads
        
        % STEP 2 - calculate the new labels (segment merging)   
        p = blkpos + [1,1]     
        it = 0
        new_label = ws[p]
        while it < 100
            changed[0] = false
            old_label = copy(new_label)
            val = xs[p]
            if xs[p+[0,1]] == val
                new_label = min(new_label, ws[p+[0,1]])
            endif
            if xs[p+[1,0]] == val
                new_label = min(new_label, ws[p+[1,0]])
            endif
            if xs[p+[0,-1]] == val
                new_label = min(new_label, ws[p+[0,-1]])
            endif
            if xs[p+[-1,0]] == val
                new_label = min(new_label, ws[p+[-1,0]])
            endif
            syncthreads            
            if new_label != old_label
                changed[0] = true % Broadcast to all threads!
                ws[p] = new_label
            endif
            syncthreads
            if changed[0] == 1 % All threads will see the same value
                it += 1
                updated = it
                syncthreads
            else
                break
            endif
        end
    
        % STEP 3 - save the result
        w[pos] = ws[p]
    end                 
    
    
    % First, count the number of labels that we will need per row (for the initial scan)
    % A good initial scan may boost the overall performance of the algorithm. 
    mp = 1                              % Number of blocks to be processed in parallel
    mpx = int(size(x,1)/mp)
    dx = uninit(size(x,0)*mp,mpx)    
    parallel_do(size(x), x, dx, mp, mpx, transition_and_swizzle)
    
    % 'starting_label' is a vector with the label indices to start the initial label scan.
    % Every stride of 'mpx' corresponds to one entry in 'starting_label'. Note that due
    % to the cumulative sum, starting_label is shifted one position. This needs to be 
    % taken into account in the horscan algorithm
    starting_label = cumsum(transpose(sum(dx, 1)))
        
    w = zeros(size(x)) % label image   
    parallel_do(size(x,0)*mp,x,w,starting_label,mp,mpx,horscan_initial)   
    
    for k=0..100
        updated = parallel_do(size(x),x,w,scan)
        if !updated
            break
        endif                   
    end
end

function [] = main()

    img = imread("lena_big.tif")[:,:,1]
    mask = img < 128
    
    imshow(mask,[])
    %warmup for timing
    for k=0..9
        w = connectedcomponents2d(mask)
    end
    %actual timing
    tic()
    for k=0..9
        w = connectedcomponents2d(mask)
    end
    toc()
    imshow(w,[])
end
