%==============================================================================
% Computation of covariance matrices - used by the wavelet denoisaicing code
% (see wavelet_denoisaicing.q)
% Aug 2012 - Bart Goossens
%==============================================================================

{!author name="Bart Goossens"}
{!doc category="Image Processing/Restoration"}

import "linalg.q"

%
% Function: imcov
%
% Computes the covariance matrix of an image with N channels
%
% :function C : mat(N,N) = imcov[N](A : cube(:,:,N))
% :function C : mat = imcov(img : cube, subtract_mean = 1)
%
function C : mat(N,N) = imcov[N](A : cube(:,:,N), subtract_mean = 1)

    function [C : vec(N*(N+1)/2)] = __kernel__ kernel(A : cube'unchecked(:,:,N), mu : mat, pos : ivec2)
        Av : vec'col(N) = squeeze(A[...pos,:]) - mu[pos]

        C += utriang_vprod[N](Av, Av)
    endfunction
   
    if subtract_mean
        mu = mean(A,2)
    else
        mu = zeros(N)
    endif
   
    C : mat(N,N) = utriangm[N](parallel_do(size(A,0..1),A,mu,kernel))/prod(size(A,0..1))

    % Symmetry extension        
    C = utriang_symext[N](C)
endfunction

% Overload that is used when N could not be determined
function C : mat = imcov(img : cube, subtract_mean = 1)
    p = size(img, 2)
    C = zeros(p * [1,1])
    img_copy = copy(img)
    % subtract the mean...
    if !subtract_mean
        img_copy -= repmat(mean(img_copy, 2),[1,1,p])
    endif

    % exploit the symmetry to reduce the number of calculations
    {!interpreted for}
    for m=0..p-1
        for n=m..p-1
            C[m,n] = C[n,m] = sum(img_copy[:,:,m] .* img_copy[:,:,n])        
        endfor
    endfor
endfunction


%
% Function: covariance_4x4
%
% Computes an 4x4 covariance matrix (of a 4-component image, across color
% channels), assuming mean 0.
%
% :function C = covariance_4x4(x : cube(:,:,4))
%
function C = covariance_4x4(x : cube(:,:,4))
    C = imcov(x, false)
endfunction


function C=covariance_8x8(A:cube(:,:,8),subtract_mean=1) 
    C = imcov(A, false)
endfunction 


function [] = main()
    x = randn(128,128,4)
    runs = 100
    tic()
    for k=1..runs
    A = covariance_4x4(x)
    endfor
    toc()
    print A

    x2 = randn(60,60,8)
    
    tic()
    for k=1..runs
    A = covariance_8x8(x2)
    endfor
    toc()
    print A
    
    x2 = randn(60,60,8)
    
    tic()
    for k=1..runs
    A = imcov(x2)
    endfor
    toc()
    print A
endfunction
